﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: assert.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>


// @addtogroup pia_common_debug_macro  Definitions of Debugging Macros Provided by the Pia Library
// @brief  Specifies the debugging macros provided by the Pia library.
//! @{
//! @name  Assert
//! @{

/*!
    @brief  Immediately ends the program if the <tt><var>EXP</var></tt> argument is <tt>false</tt> or <tt>NULL</tt>.

    @details  When you use this macro, all strings that you have output using the <tt>@ref PIA_CACHED_PRINTF</tt> macro are output from the cache.

    @details  This macro is ignored in release builds.

    @param[in] EXP  Specifies to forcibly end the program if this argument is <tt>false</tt> or <tt>NULL</tt>.

    @see  PIA_CACHED_PRINTF
 */
#if NN_PIA_ENABLE_ASSERT
#define PIA_ASSERT(EXP)                                                                \
    if (!(EXP))                                                                        \
    {                                                                                  \
        ::nn::pia::Assert(#EXP, PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_LINE, NULL); \
    }
#else
#define PIA_ASSERT(EXP) ((void)0)
#endif


/*!
    @brief  Immediately ends the program if the <tt><var>EXP</var></tt> argument is <tt>false</tt> or <tt>NULL</tt>.

    @details  When you use this macro, all strings that you have output using the <tt>@ref PIA_CACHED_PRINTF</tt> macro are output from the cache.

    @details  This macro is ignored in release builds.

    @param[in] EXP  Specifies to forcibly end the program if this argument is <tt>false</tt> or <tt>NULL</tt>.
    @param[in] ...  Specifies variable arguments equivalent to those of the <tt>printf</tt> function in the standard library.

    @see  PIA_CACHED_PRINTF
 */
#if NN_PIA_ENABLE_ASSERT
#define PIA_ASSERTMSG(EXP, ...)                                                               \
    if (!(EXP))                                                                               \
    {                                                                                         \
        ::nn::pia::Assert(#EXP, PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_LINE, __VA_ARGS__); \
    }
#else
#define PIA_ASSERTMSG(EXP, ...) ((void)0)
#endif


//! @cond PRIVATE
#define PIA_MACRO_CAT(l, r) l##r
#define PIA_MACRO_EXPANDING_CAT(l, r) PIA_MACRO_CAT(l, r)
//! @endcond

/*!
    @brief  Results in a compiler error if <tt><var>EXP</var></tt> is <tt>false</tt> or <tt>NULL</tt>.

    @details  This is a compile-time (static) assertion macro.
     Specify a constant expression for the <tt><var>EXP</var></tt> argument.

    @param[in] EXP  Specifies a constant expression. A compiler error results if this argument is <tt>false</tt> or <tt>NULL</tt>.
 */
#define PIA_COMPILE_ASSERT(EXP)                  \
    typedef ::nn::pia::PIA_COMPILE_ASSERT_INNER< \
        sizeof(::nn::pia::PIA_COMPILE_ASSERT_FAILURE<static_cast<bool>(EXP)>)> \
PIA_MACRO_EXPANDING_CAT(pia_compile_assert_at_, __LINE__)


//! @}


//! @name  Panic
//! @{

/*!
    @brief  Immediately ends the program.

    @details  When you use this macro, all strings that you have output using the <tt>@ref PIA_CACHED_PRINTF</tt> macro are output from the cache.

    @param[in] ...  Specifies variable arguments equivalent to those of the <tt>printf</tt> function in the standard library.

    @see  PIA_CACHED_PRINTF
 */
#if NN_PIA_ENABLE_ASSERT
#define PIA_PANIC(...) ::nn::pia::Panic(PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_LINE, __VA_ARGS__)
#else
#define PIA_PANIC(...) PIA_BASIC_HALT(__VA_ARGS__)
#endif

//! @}
//! @}


namespace nn
{
namespace pia
{

/*!
@cond PRIVATE
  @brief  Evaluates an assertion.

  @details  This function is meant to be used by macros such as <tt>PIA_ASSERT</tt> or <tt>PIA_ASSERTMSG</tt>.

   We do not recommend calling this function directly.

  @param[in] pExp  Specifies a conditional expression in text format.
  @param[in] pPos  Specifies a location in code in text format. (<tt>__FILE__</tt>, <tt>__FUNCTION__</tt>, <tt>__MODULE__</tt>, and so on.)
  @param[in] line  Specifies the line number.
  @param[in] pMsg  Specifies an additional message to output on assertion.
 */
void Assert(const char* pExp, const char* pPos, u32 line, const char* pMsg, ...);
//! @endcond


/*!
@cond PRIVATE
  @brief  Triggers an immediate panic of the running kernel.

  @details  This function is meant to be used by macros such as <tt>PIA_PANIC</tt>.

   We do not recommend calling this function directly.

  @param[in] pPos  Specifies a location in code in text format. (<tt>__FILE__</tt>, <tt>__FUNCTION__</tt>, <tt>__MODULE__</tt>, and so on.)
  @param[in] line  Specifies the line number.
  @param[in] pMsg  Specifies an additional message to output on assertion.
 */
void Panic(const char* pPos, u32 line, const char* pMsg, ...);
//! @endcond


/*!
@cond PRIVATE
  @brief  Represents a template class for compiler assertions.

  @details  These template classes are meant to be used by macros such as <tt>PIA_COMPILE_ASSERT</tt>.

   We do not recommend using this template class for other purposes.
 */
template <bool b>
struct PIA_COMPILE_ASSERT_FAILURE;
template <>
struct PIA_COMPILE_ASSERT_FAILURE<true>
{
};
template <int x>
struct PIA_COMPILE_ASSERT_INNER
{
};
//! @endcond
}
}  // end of namespace nn::pia
