﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     assertableResult.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>

#if NN_PIA_NIN_WIN
#include <nn/nn_ErrorResult.h>
#endif

namespace nn
{
namespace pia
{

/*!
@cond PRIVATE
@brief  Enables or disables assertions for all <tt>Result</tt> values.
*/
struct AssertableResultAll
{
public:
/*!
@brief  Enables assertions.
*/
    static void EnableAssert()
    {
        m_IsAssertEnable = true;
    }

/*!
@brief  Disables assertions.
*/
    static void DisableAssert()
    {
        m_IsAssertEnable = false;
    }

/*!
@brief  Gets whether assertions are enabled.
@return  Returns <tt>true</tt> if assertions are enabled, and <tt>false</tt> otherwise.
*/
    static bool IsAssertEnable()
    {
        return m_IsAssertEnable;
    }

private:
    static bool m_IsAssertEnable;
};
//! @endcond

#if NN_PIA_NIN_WIN
/*!
@cond PRIVATE
@brief  Represents the <tt>Result</tt> structure used to manage enabling and disabling assertions.
For more information, see <tt>NN_DEFINE_ERROR_RANGE_RESULT</tt> in <tt>result_ErrorResult.h</tt>.
*/
template <int module, int descriptionBegin>
struct AssertableResult
    : public ::nn::result::detail::ErrorResultBase<module, descriptionBegin>
    , public ::nn::result::detail::ErrorRange<module, descriptionBegin, descriptionBegin+1>
{
public:
/*!
@brief  Enables assertions.
*/
    static void EnableAssert()
    {
        m_IsAssertEnable = true;
    }

/*!
@brief  Disables assertions.
*/
    static void DisableAssert()
    {
        m_IsAssertEnable = false;
    }

/*!
@brief  Gets whether assertions are enabled.
@return  Returns <tt>true</tt> if assertions are enabled, and <tt>false</tt> otherwise.
*/
    static bool IsAssertEnable()
    {
        return (m_IsAssertEnable || AssertableResultAll::IsAssertEnable());
    }

private:
    static bool m_IsAssertEnable;
};


template <int TModule, int TDescription>
bool AssertableResult<TModule, TDescription>::m_IsAssertEnable = false;
//! @endcond
#else
/*!
@cond PRIVATE
@brief  Represents the <tt>Result</tt> structure used to manage enabling and disabling assertions.
*/
template <Result::Level TLevel, Result::Summary TSummary, Result::Module TModule, int TDescription>
struct AssertableResult : public nn::Result::Const<TLevel, TSummary, TModule, TDescription>
{
public:
/*!
@brief  Enables assertions.
*/
    static void EnableAssert()
    {
        m_IsAssertEnable = true;
    }

/*!
@brief  Disables assertions.
*/
    static void DisableAssert()
    {
        m_IsAssertEnable = false;
    }

/*!
@brief  Gets whether assertions are enabled.
@return  Returns <tt>true</tt> if assertions are enabled, and <tt>false</tt> otherwise.
*/
    static bool IsAssertEnable()
    {
        return (m_IsAssertEnable || AssertableResultAll::IsAssertEnable());
    }

private:
    static bool m_IsAssertEnable;
};


template <Result::Level TLevel, Result::Summary TSummary, Result::Module TModule, int TDescription>
bool AssertableResult<TLevel, TSummary, TModule, TDescription>::m_IsAssertEnable = false;
//! @endcond
#endif

/*!
@cond PRIVATE
@brief  Wraps the <tt>GetValue()</tt> function of <tt>nn::Result</tt>.
*/
#if NN_PIA_NIN_WIN
inline int32_t PiaResultGetValue(nn::Result result)
{
    return result.GetInnerValueForDebug();
}
#else
inline u32 PiaResultGetValue(nn::Result result)
{
    return result.GetValue();
}
//! @endcond
#endif
}
}  // end of namespace nn::pia

#if NN_PIA_NIN_WIN
/*!
@cond PRIVATE
@brief  Directly enters the NintendoSDK Pia module number.
*/
namespace nn{
namespace pia{
    const int PiaResultModuleNo = 301;
}
}  // end of namespace nn::pia

#define PIA_DEFINE_RESULT(name, level, summary, module, description) \
    typedef ::nn::pia::AssertableResult<(::nn::pia::PiaResultModuleNo), (description)> name
//! @endcond
#else
//! @cond PRIVATE
#define PIA_DEFINE_RESULT(name, level, summary, module, description) \
    typedef ::nn::pia::AssertableResult<(level), (summary), (module), (description)> name
//! @endcond
#endif
