﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     chat_VoiceProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>
#include <pia/chat/chat_definitions.h>

#include <pia/transport/transport_Protocol.h>

#include <pia/common/common_CriticalSection.h>

#pragma diag_suppress 1300
#include <nex.h>
#pragma diag_warning 1300

// Forward declaration.
namespace nn
{
namespace pia
{
namespace transport
{
class UnreliableProtocol;
}
}
}

struct VCTSession;

#if NN_PIA_HF_LIBRARY
namespace nn
{
namespace pia
{
namespace chat
{
class HandsFree;
}
}
}
#endif


namespace nn
{
namespace pia
{
namespace chat
{


/*!
@brief  Contains member functions that provide voice chat functionality.

@details  You cannot create more than one instance of <tt>VoiceProtocol</tt>.

@date 2015-09-01  If the client specified for <tt>RemoveConferenceClient()</tt> cannot be found, the function now returns <tt>ResultNotFound</tt> instead of <tt>ResultInvalidArgument</tt>.
@date 2015-09-01  If an invalid argument is given to <tt>Send()</tt>, the function now returns <tt>ResultInvalidArgument</tt>.
@date 2015-09-01  If there is no audio data to get when <tt>Receive()</tt> is called, the function now returns <tt>ResultNoData</tt>.
@date 2015-08-05  Added notes to the <tt>SetVADClampGain()</tt> and <tt>Send()</tt> functions.
@date 2015-05-18  Made the <tt>GetVADPower</tt> function public.
@date 2015-01-15  Changed the type of the third parameter of the <tt>Receive</tt> function to <tt>nn::pia::StationIndex*</tt>.
@date 2014-11-20  Added information about the thread-safe APIs.
@date 2014-05-12  Made the <tt>IsClientInConference</tt> function private.
@date 2014-05-07  Expanded the <tt>Setting</tt> structure to include a parameter that allows you to configure the maximum number of stations that can participate in a conference chat.
@cond CAFE_DOC
@date 2014-03-07  Added a new API for using the Automatic Gain Control (AGC) feature.
@date 2014-03-07  Added a new API for using the noise suppression feature.
@date 2014-03-07  Added a new API for using the echo cancellation feature.
@date 2014-03-07  Added a new <tt>Send</tt> function (with different arguments from the previous one).
@endcond
@date 2014-02-26  Changed the specifications so that the high-pass filter is enabled by default.
@date 2014-02-25  Corrected an incorrect default value in the description of the <tt>SetVADActiveGain</tt> function.
@date 2014-01-30  Specifying 2-bit ADPCM as the sound codec for main packets is no longer supported.
@date 2014-01-30  Quality alert-related API functions have been made private.
@date 2014-01-23  Changed the specifications so that the <tt>SetVADClampGain</tt> function returns <tt>nn::Result</tt>.
@date 2014-01-23  Changed the specifications so that the <tt> SetQualityAlertThresholdRate</tt> function returns <tt>nn::Result</tt>.
@date 2014-01-23  Changed the specifications so that calling the <tt>SetAudioFrameLength</tt> function in a state where communication is not occurring results in <tt>ResultInvalidState</tt> being returned.
@date 2014-01-22  Added the <tt>IsInCommunication</tt> member function to determine whether communication is taking place.
@date 2014-01-22  Changed the specifications so that the <tt>SetCodec</tt> function returns <tt>ResultInvalidState</tt> when in a state where communication is not occurring.
@date 2013-10-02  Added a member function to set the number of receive buffers for <tt>UnreliableProtocol</tt> to the <tt>Setting</tt> structure.
@date 2013-10-02  Stopped using <tt>ReliableProtocol</tt> internally.
@date 2013-07-12  Defined various constants and enumerations.
@date 2013-07-12  Ported set and get functions for various parameters.
@date 2013-06-18  Initial version.
*/
class VoiceProtocol : public transport::Protocol
{
public:
    static const u32 AUDIO_FRAME_LENGTH_MIN = 64;       //!< Specifies the minimum audio frame length (in milliseconds).
    static const u32 AUDIO_FRAME_LENGTH_MAX = 192;      //!< Specifies the maximum audio frame length (in milliseconds).
    static const u32 AUDIO_FRAME_LENGTH_DEFAULT = 144;  //!< Specifies the default audio frame length (in milliseconds).

    static const size_t AUDIO_DATA_SIZE_PER_MILLI_SEC = 16;                                                      //!< Specifies the data size per millisecond (in bytes).
    static const size_t AUDIO_DATA_SIZE_MIN = (AUDIO_FRAME_LENGTH_MIN * AUDIO_DATA_SIZE_PER_MILLI_SEC);          //!< Specifies the minimum audio data size.
    static const size_t AUDIO_DATA_SIZE_MAX = (AUDIO_FRAME_LENGTH_MAX * AUDIO_DATA_SIZE_PER_MILLI_SEC);          //!< Specifies the maximum audio data size.
    static const size_t AUDIO_DATA_SIZE_DEFAULT = (AUDIO_FRAME_LENGTH_DEFAULT * AUDIO_DATA_SIZE_PER_MILLI_SEC);  //!< Specifies the default audio data size.

    static const size_t MAX_CONFERENCE_CLIENT = 16;
    static const size_t MAX_AUDIO_STREAM = MAX_CONFERENCE_CLIENT - 1;

    static const u32 VAD_CLAMP_GAIN_MAX = 0x3fffffff;

#if NN_PIA_HF_LIBRARY
    static const s32 AGC_LEVEL_MIN = 0;      //!< The smallest value that can be specified for the AGC level.
    static const s32 AGC_LEVEL_MAX = 4;      //!< The largest value that can be specified for the AGC level.
    static const s32 AGC_LEVEL_DEFAULT = 1;  //!< The default value for the AGC level.

    static const s32 NOISE_SUPPRESSOR_LEVEL_MIN = 0;      //!< The smallest value that can be specified for the noise suppression level.
    static const s32 NOISE_SUPPRESSOR_LEVEL_MAX = 2;      //!< The largest value that can be specified for the noise suppression level.
    static const s32 NOISE_SUPPRESSOR_LEVEL_DEFAULT = 1;  //!< The default value for the noise suppression level (recommended).

    static const EcMode ECHO_CANCELLER_MODE_DEFAULT = EC_STANDARD;  //!< The default value for the echo cancellation mode (recommended).
#endif                                                              // end of NN_PIA_HF_LIBRARY

/*!
@brief  Stores parameters used by the <tt>Initialize</tt> function.

@details  This structure contains the various settings for <tt>VoiceProtocol</tt>.
@see  Initialize()
*/
    struct Setting
    {
/*!
@brief  Instantiates the object with default parameters (default constructor). Each member variable is set to the default value.
*/
        Setting(void)
          : codec(CODEC_4BIT_ADPCM),
            frameLength(AUDIO_FRAME_LENGTH_DEFAULT),
            port(10),
            recvBufNum(32),
            maxStationNum(MAX_CONFERENCE_CLIENT)
        {
        }

        Codec codec;           //!< Specifies the audio codec.
        size_t frameLength;    //!< Specifies the audio frame length. The value is in milliseconds.
        u16 port;              //!< Specifies the port number assigned to the <tt>UnreliableProtocol</tt> object used internally by <tt>PiaChat</tt>. Make sure this does not conflict with the port number used by the application.
        size_t recvBufNum;     //!< Specifies the number of receive buffers for the <tt>UnreliableProtocol</tt> object used internally by <tt>PiaChat</tt>.
        size_t maxStationNum;  //!< The maximum number of stations that can participate in a conference chat.
    };


/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).

@details  Do not call this constructor directly from applications.
Use the <tt>@ref Transport::CreateProtocol</tt> function to create an instance.
*/
    VoiceProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destructor.

@details  Do not call this destructor directly from applications.
Use the <tt>@ref Transport::DestroyProtocol</tt> function to destroy an instance.
*/
    virtual ~VoiceProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  This is a dispatch function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the <tt>VoiceProtocol::Initialize</tt> function may not have been called. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Dispatch(void);
    //! @endcond


/*!
@cond PRIVATE
*/
    virtual nn::Result UpdateProtocolEvent(const transport::ProtocolEvent& event);
    //! @endcond


    PIA_PROTOCOL_TYPE_INFO(transport::PROTOCOL_TYPE_VOICE);


/*!
@brief  Initializes the instance.

@details  Call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.
@param[in] setting  Specifies a settings parameter structure for <tt>VoiceProtocol</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>chat</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Indicates that the function failed to allocate memory. Consider providing more memory when calling the <tt>common::Initialize</tt> function. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyInitialized  Indicates that the <tt>VoiceProtocol</tt> instance is already initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was not called between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@see  Finalize
*/
    nn::Result Initialize(const VoiceProtocol::Setting& setting);


/*!
@brief  Finalizes the instance.

@details  This function returns without doing anything if it is called before the <tt>Initialize</tt> function has been called.
@see  Initialize
*/
    void Finalize(void);


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] myStationIndex  Specifies the local <tt>StationIndex</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup(void);
    //! @endcond


/*!
@brief  Determines whether the system is in a communication state.

@details  This API function is thread-safe.

@return  Returns <tt>true</tt> if the system is in a communication state or <tt>false</tt> otherwise.
*/
    bool IsInCommunication(void) const;


/*!
@brief  Packs audio data in the send queue.

@cond CAFE_DOC
@details  If you use this API to send audio data, you cannot use the AGC, noise suppression, or echo cancellation features.
To use these features, use the <tt>@ref Send(const s16*, size_t, const s16*, const s16*)</tt> API function rather than this one.
@endcond

@details  This API function can also be called from threads other than the one that calls the <tt>common::Scheduler::Dispatch()</tt> function.

@details  The assumption is that this API will receive audio data actually picked up by the microphone.
You must be careful when passing synthesized audio data. For more information, see the <tt>SetVADClampGain()</tt> function.

@param[in] pAudioBuf  Specifies a pointer to the start of the audio data.
@param[in] samples  Specifies the sampling rate of the audio data.

@cond CTR_DOC
@param[in] micOffset  Specifies the offset (number of samples) between the latest mic sample and last sample of the input mic data.
@endcond

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@cond CAFE_DOC
@see  Receive, Send(const s16*, size_t, const s16*, const s16*)
@endcond
@cond CTR_DOC
@see  Receive
@endcond

*/
#if NN_PIA_CTR
    nn::Result Send(const void* pAudioBuf, u32 samples, u32 micOffset);
#elif NN_PIA_CAFE
    nn::Result Send(const void* pAudioBuf, u32 samples);
#else
#error  "Invalid platform."
#endif


/*!
@brief  Extracts audio data from the receive buffer.

@details  This function also acts internally to mix the audio if it is playing on more than one channel.
The audio data written to <tt><var>pAudioBuf</var></tt> is in 16-bit/8 kHz format.
This buffer must be aligned to a 32-byte boundary.
@details  This API can also be called from threads other than the one that calls the <tt>common::Scheduler::Dispatch()</tt> function.

@param[out] pAudioBuf  Specifies a pointer to the audio data buffer.
@param[in] frameDataSize  Specifies the size of the audio data buffer.
@param[out] pStationIndexArray  Stores the <tt>StationIndex</tt> array of the client that sent this stream. Specify <tt>NULL</tt> if this is unnecessary.
The number of elements in the array must not exceed the <tt>maxStationNum</tt> member of the <tt>Setting</tt> structure passed when the <tt>Initialize</tt> function was called.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  There was no audio data to get. Handle appropriately in the application.
@cond CAFE_DOC
@see  Send, Send(const s16*, size_t, const s16*, const s16*)
@endcond
@cond CTR_DOC
@see  Send
@endcond

*/
    nn::Result Receive(void* pAudioBuf, u32 frameDataSize, StationIndex* pStationIndexArray);


/*!
@brief  Adds a client to the conference.

@details  The local <tt>StationId</tt> does not need to be specified.
To start a conference, this function must be called by all parties for each other.
For example, if three systems with a <tt>StationId</tt> of 1, 2, and 3 are starting a conference, system 1 must add 2 and 3, system 2 must add 1 and 3, and system 3 must add 1 and 2.
If N is the maximum number of people who can connect in a conference, the maximum number of clients that the local system can add to a conference is N-1.
This function returns an error if called while not in conference mode.

@param[in] id  Specifies the <tt>StationId</tt> of the client to add.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that this function is not in a callable state. Either preparations have not been made for the call, or the maximum number of participating clients has been reached. Programming error. Fix your program so that this error is not returned.


*/
    nn::Result AddConferenceClient(StationId id);


/*!
@cond PRIVATE
@brief  Adds a client to the conference.
*/
    nn::Result AddConferenceClient(StationIndex idx);
    //! @endcond


/*!
@brief  Removes a client from the conference.

@details  Have all participants in the conference call this function, and not just the one that is leaving.

@param[in] id  Specifies the <tt>StationId</tt> of the client to remove.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotFound  The client with the specified ID cannot be found. Handle appropriately in the application.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result RemoveConferenceClient(StationId id);


/*!
@cond PRIVATE
@brief  Removes a client from the conference.
*/
    nn::Result RemoveConferenceClient(StationIndex id);
    //! @endcond


/*!
@brief  Determines whether a client exists in a conference.

@param[in] id  Specifies the <tt>StationId</tt> of the client to search for.

@return  Returns <tt>true</tt> if the client is in the conference or <tt>false</tt> otherwise.
*/
    bool IsClientInConference(StationId id) const;


/*!
@cond PRIVATE
@brief  Determines whether a client exists in a conference.
*/
    bool IsClientInConference(StationIndex idx) const;
    //! @endcond


/*!
@brief  Clears the audio streaming.
*/
    void ClearStreaming(void);


/*!
@brief  Sets the audio codec.

@details  The codec can be changed even while streaming.
You can specify 4-bit ADPCM or 2-bit ADPCM, but 2-bit ADPCM is not supported.
This function returns <tt>ResultInvalidArgument</tt> if an unsupported audio codec is specified.

@param[in] codec  The changed audio codec type for the sender.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. An unsupported codec may have been specified. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Communication might not have been started. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetCodec(enum Codec codec);


/*!
@brief  Gets the currently configured audio codec.

@return  Returns the currently configured audio codec for main packets.
*/
    enum Codec GetCodec(void) const;


/*!
@brief  Sets the audio frame length.

@details  For <tt><var>frameLength</var></tt>, you must specify a value between <tt>@ref AUDIO_FRAME_LENGTH_MIN</tt> and <tt>@ref AUDIO_FRAME_LENGTH_MAX</tt> that is also an integer multiple of 8.

@param[in] frameLength  Specifies the length of the audio frame (in milliseconds).

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result SetAudioFrameLength(u32 frameLength);


/*!
@brief  Gets the currently configured audio frame length.

@details  If you call this function before initialization, the recommended value <tt>@ref AUDIO_FRAME_LENGTH_DEFAULT</tt> is returned.
@details  This API function is thread-safe.

@return  Returns either the value that was set in the <tt><var>frameLength</var></tt> member variable of the <tt>@ref Setting</tt> structure at the time of initialization, or the value set using the <tt>@ref SetAudioFrameLength</tt> function after initialization.
*/
    u32 GetAudioFrameLength(void) const;


/*!
@brief  Gets the packet size of audio data.

@details  This function returns the data size calculated based on the currently configured audio frame length.
You can get the currently configured audio frame length by using the <tt>@ref  GetAudioFrameLength</tt> function.
If called prior to initialization, <tt>@ref AUDIO_DATA_SIZE_DEFAULT</tt>, which corresponds to the recommended audio frame length value in <tt>@ref AUDIO_FRAME_LENGTH_DEFAULT</tt>, is returned.
@details  This API function is thread-safe.

@return  Returns the packet size of audio data.


*/
    size_t GetAudioDataSize(void) const;


/*!
@cond PRIVATE
@brief  Gets the calculated audio scale (square of the average power) of the input audio data.

@details  The <tt><var>length</var></tt> argument specifies the number of bytes of audio data to input.
The audio data must have been sampled at 16-bits/8 kHz.

@param[in] audio  Specifies the audio data.
@param[in] length  Specifies the number of bytes of the audio data.

@return  Returns the audio scale.
*/
    u32 GetAudioScale(const void* audio, u32 length);
    //! @endcond


/*!
@brief  Gets the audio streaming information of the remote station specified by <tt>StationId</tt>.

@param[in] id  Specifies the <tt>StationId</tt> of the remote station from which to get the audio streaming information.
@param[out] pInfo  Specifies a pointer to a <tt>@ref AudioInfo</tt> structure.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetAudioInfo(StationId id, AudioInfo* pInfo) const;


/*!
@cond PRIVATE
@brief  Gets the audio streaming information of the other party specified by <tt>StationIndex</tt>.
*/
    nn::Result GetAudioInfo(StationIndex idx, AudioInfo* pInfo) const;
    //! @endcond


/*!
@brief  Sets the time period not to send audio for <tt>@ref SignalNoiseSkip</tt>.

@param[in] msec  Specifies the time period not to send audio. The value is in milliseconds.
@see  SignalNoiseSkip, GetNoiseSkipTime
*/
    void SetNoiseSkipTime(u32 msec);


/*!
@brief  Gets the time period not to send audio for <tt>@ref SignalNoiseSkip</tt>.

@return  Returns the time period not to send audio. The value is in milliseconds.
@see  SignalNoiseSkip, SetNoiseSkipTime
*/
    u32 GetNoiseSkipTime(void) const;


/*!
@brief  Stops VAD (Voice Activity Detection) from becoming active and sending audio on the sound of button presses when VAD is enabled but not active.

@details  Setting up this function enables you to prevent the VAD feature from activating and sending audio on the sound of button presses when VAD is enabled but not active.
By not sending audio for a certain time period after button presses, you can minimize the button pressing sounds received by other players.
When VAD is active and audio is being sent, the sound of button presses cannot be removed.
Specify the time period not to send audio using the <tt>@ref SetNoiseSkipTime</tt> function.
@see  SetNoiseSkipTime, GetNoiseSkipTime

*/
    void SignalNoiseSkip(void);


/*!
@cond PRIVATE
@brief  This function enables and disables the spare packet feature.

@details  The spare packet feature is enabled by default.
@param[in] enable  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.
*/
    void EnableSparePacket(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the spare packet feature is enabled.

@details  The spare packet feature is enabled by default.
@retval true  Indicates that the spare packet feature is enabled.
@retval false  Indicates that the spare packet feature is disabled.
*/
    bool IsEnableSparePacket(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the audio codec for spare packets.

@details  By default, spare packets use the 2-bit ADPCM codec.
Use the <tt>@ref SetCodec</tt> function to change the codec for main packets.
<b>Note:</b> The only supported codec type for spare packets is 2-bit ADPCM.
Accordingly, this function does not need to be called at this time.
This function returns <tt>ResultInvalidArgument</tt> if an unsupported codec is specified.
@param[in] codec  Specifies the codec type for spare packets.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetSpareCodec(enum Codec codec);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the currently configured audio codec for spare packets.

@details  This value is meaningless if the spare packet feature is not used.
<b>Note:</b> The only supported codec type for spare packets is 2-bit ADPCM.
Accordingly, the value returned by this function will always be <tt>@ref CODEC_2BIT_ADPCM</tt> at this time.
@return  Returns the currently configured codec type for spare packets.
*/
    enum Codec GetSpareCodec(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Enables or disables the quality alert feature.

@details  By default, quality alert is enabled.
When this feature is enabled, if the incoming audio packets sent from a peer are often late in arriving, a quality alert flag is set in the outgoing packets to that peer to notify them of the situation.
@param[in] enable  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.


*/
    void EnableQualityAlert(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the quality feature is enabled.

@retval true  Indicates that the quality alert feature is enabled.
@retval false  Indicates that the quality alert feature is disabled.
*/
    bool IsEnableQualityAlert(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the state of the quality alert flag for the specified <tt>StationIndex</tt>.

@details  If the return value is <tt>true</tt>, packets received from that peer are often late in arriving.
You can find out the state of the local quality alert flag by specifying the local <tt>StationIndex</tt>.
@param[in] id  Specifies the <tt>StationIndex</tt> of the peer.
@retval true  Indicates that the quality alert feature is enabled.
@retval false  Indicates that the quality alert feature is disabled.

*/
    bool GetQualityAlert(StationIndex id) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the package delay threshold at which the quality alert flag is set.

@details  The value is set as a percentage. It is set to 20% by default.
@param[in] rate  Specifies the packet delay threshold (%).

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

@see  GetQualityAlertThresholdRate
*/
    nn::Result SetQualityAlertThresholdRate(u8 rate);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the configured package delay threshold at which the quality alert flag is set.

@return  Returns the packet delay threshold, in percent (%).
@see  SetQualityAlertThresholdRate
*/
    u8 GetQualityAlertThresholdRate(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the duration for keeping the quality alert flag set.

@details  Sets the duration in seconds for the local station to keep its quality alert flag set after receiving a quality alert notification flag from a peer.
The default value is 1800 seconds.
@param[in] sec  Specifies the duration of the flag (in seconds).
@see  GetQualityAlertDuration
*/
    void SetQualityAlertDuration(u32 sec);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the duration for keeping the quality alert flag set.

@details  Gets the duration in seconds for the local station to keep its quality alert flag set after receiving a quality alert notification flag from a peer.
@return  Returns the duration of the flag (in seconds).
@see  SetQualityAlertDuration

*/
    u32 GetQualityAlertDuration(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the number of nearby packets to sample when measuring the packet delay.

@details  This function sets the number of nearby packets to sample when measuring the packet delay.
A larger value increases accuracy, but also reduces responsiveness to changes in connection quality.
The default value is 40 packets.
@param[in] num  Specifies the number of nearby packets to sample.
@see  GetQualityAlertSamplingPacketNum
*/
    void SetQualityAlertSamplingPacketNum(u16 num);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the number of nearby packets to sample when measuring the packet delay.

@return  Returns the number of nearby packets to sample.
@see  SetQualityAlertSamplingPacketNum
*/
    u16 GetQualityAlertSamplingPacketNum(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Enables or disables the drop recover feature.

@details  The drop recover feature is enabled by default.
When the drop recover feature is enabled and packet loss occurs, the audio data in the lost packets is recovered.
When the spare packet feature is enabled and the spare packet of the subsequent packet can be used, the audio data in the current packet can be restored based on the audio data in the spare packet, improving audio quality.
@param[in] flag  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.


*/
    void EnableDropRecover(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the drop recover feature is enabled.

@retval true  Indicates that the drop recover feature is enabled.
@retval false  Indicates that the drop recover feature is disabled.
*/
    bool IsEnableDropRecover(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Enables or disables the VAD (Voice Activity Detection) feature.

@details  The VAD feature is enabled by default.
When VAD is enabled, in addition to evaluating audio inside the <tt>@ref Dispatch</tt> function, decisions are also made automatically about whether to send packets.
@param[in] flag  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.

*/
    void EnableVAD(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the VAD (Voice Activity Detection) feature is enabled or disabled.

@retval true  Indicates that the VAD feature is enabled.
@retval false  Indicates that the VAD feature is disabled.
*/
    bool IsEnableVAD(void) const;
    //! @endcond


/*!
@brief  Gets whether audio is detected using the VAD (Voice Activity Detection) algorithm.

@details  Returns <tt>true</tt> while frames with audio (speech frames) are present, or <tt>false</tt> while the frames without audio (silent frames) are present.
When the VAD feature is disabled, this function always returns <tt>false</tt> regardless of whether there is any voice activity.

@return  Returns <tt>true</tt> if it is determined that audio is present or <tt>false</tt> otherwise.

*/
    bool GetVADActivity(void) const;


/*!
@brief  Gets the average power of audio data in the current audio frame.

@details  The VAD algorithm detects audio based on the size of this value.
This value is equivalent to the average power of the audio data (the square root of the volume scale of the audio data used in the <tt>GetAudioScale</tt> function or <tt>@ref SetVADClampGain</tt> function) divided by a normalization factor (<tt>256</tt>).
@return  Returns the average power of the audio data in the current audio frame.


*/
    u16 GetVADPower(void) const;


/*!
@cond PRIVATE
@brief  Resets the internal state of VAD (Voice Activity Detection).

@details  Call this function after making changes to the microphone gain or similar settings.
*/
    void ResetVAD(void);
    //! @endcond


/*!
@brief  Sets the speech frame detection threshold.

@details  When the previous audio frame is recognized as a silent frame, this threshold value is used to determine whether the current audio frame includes audio.
A frame is recognized as a speech frame when the following conditions are met: <tt>Ts</tt> is the threshold value, <tt>Ec</tt> is the average speech power of the audio data in the current frame, and <tt>En</tt> is the average speech power of the audio data in the four frames immediately prior to the current one. The difference between <tt>Ec</tt> and <tt>En</tt> are compared. <br>
Ec > En * Ts / 100<br />
The <tt><var>value</var></tt> parameter takes the <tt>Ts</tt> value. The default value is 350.
In this case, the threshold of the power difference between <tt>En</tt> and <tt>Ec</tt> is equivalent to <tt>20 * log (350 / 100) = 10.9 dB</tt>.
@param[in] value  Specifies the speech frame detection threshold.
@see  GetVADActiveGain



*/
    void SetVADActiveGain(u16 value);


/*!
@brief  Gets the current speech frame detection threshold.

@return  Returns the speech frame detection threshold.
@see  SetVADActiveGain
*/
    u16 GetVADActiveGain(void) const;


/*!
@brief  Sets the silent frame detection threshold.

@details  When the previous audio frame is recognized as a speech frame, this threshold value is used to determine whether the current audio frame is silent.
A frame is recognized as silent when the following condition is met, where <tt>Tn</tt> is the threshold value, <tt>Ec</tt> is the average speech power of the audio data in the nearest four frames, and <tt>En</tt> is the average speech power of the audio data in the four frames before the last frame to be deemed a silent frame. <br>
Ec < En * Tn / 100<br />
However, the sending of audio frames is not actually stopped until the consecutive number of frames deemed silent reaches the value specified with the <tt>@ref SetVADReleaseTime</tt> function.
The <tt><var>value</var></tt> parameter takes the <tt>Tn</tt> value. The default value is 280.
In this case, the threshold of the power difference between <tt>En</tt> and <tt>Ec</tt> is equivalent to <tt>20 * log (280 / 100) = 8.94 dB</tt>.
@param[in] value  Specifies the silent frame detection threshold.
@see  GetVADInactiveGain




*/
    void SetVADInactiveGain(u16 value);


/*!
@brief  Gets the current silent frame detection threshold.

@return  Returns the silent frame detection threshold.
@see  SetVADInactiveGain
*/
    u16 GetVADInactiveGain(void) const;


/*!
@brief  Sets the time in number of packets before the sending of packets stops after detecting silence.

@details  To prevent VAD from detecting silence in the middle of a conversation because, for example, the user takes a breath, set the release time to at least 500 ms.
Set the parameter to the time to wait in number of packets before actually stopping the sending of packets after silence is detected.
The default is 5 packets (<tt>144 * 5 = 720 ms</tt>).
@param[in] packetNum  Specifies the release time (in number of packets).
@see  GetVADReleaseTime

*/
    void SetVADReleaseTime(u32 packetNum);


/*!
@brief  Gets the time in number of packets before the sending of packets stops after detecting silence.

@return  Returns the release time (in number of packets).
@see  SetVADReleaseTime
*/
    u32 GetVADReleaseTime(void) const;


/*!
@brief  Sets the volume threshold as a volume scale value (the square of the average power) for unconditionally determining silence.

@details  <tt><var>value</var></tt> takes a volume scale value in the range from <tt>0</tt> through <tt>0x3FFFFFFF</tt>, and any audio frame with a volume scale set below this threshold is treated as silence.

The volume scale of the audio data as obtained by the <tt>GetAudioScale</tt> function is used for this value. The default value is 49.

@details  Audio frames with volumes below the value set by this API are treated as silence by the library and the VAD status is not updated.
For this reason, if you want to pass synthesized audio data to a function that sends audio data, you must make sure its volume is higher than the threshold set by this function.
Operations could be impaired if data at very low volume below this threshold (for example, data to fill the memory with zeros) continues to be passed to a function that sends audio data.

@param[in] value  Specifies the volume threshold at which a frame is unconditionally determined to be silent.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

@see  GetVADClampGain




*/
    nn::Result SetVADClampGain(u32 value);


/*!
@brief  Gets the volume threshold as a volume scale value for unconditionally determining silence.

@details  The volume scale value is specified in the range from <tt>0</tt> through <tt>0x3FFFFFFF</tt>.
@return  Returns the volume threshold at which a frame is unconditionally determined to be silent.
@see  SetVADClampGain
*/
    u32 GetVADClampGain(void) const;


/*!
@cond PRIVATE
@brief  Enables or disables the high-pass filter.

@details  By default, the high-pass filter is enabled.
When this feature is enabled, the high-pass filter is applied to the audio being sent inside the <tt>@ref  Dispatch</tt> function to eliminate low-range noise of the audio data.
@param[in] enable  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.


*/
    void EnableHPF(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the high-pass filter is enabled.

@retval true  Indicates that the high-pass filter is enabled.
@retval false  Indicates that the high-pass filter is disabled.
*/
    bool IsEnableHPF(void) const;
//! @endcond


#if NN_PIA_HF_LIBRARY
/*!
@brief  Uses the echo cancellation, AGC, and noise suppression features when sending audio.

@details  This API uses approximately 16 KB of the stack. Make sure that you allocate a stack of adequate size for the thread that calls this API.
@details  This API blocks for about 14 to 20 milliseconds. It blocks because it runs internal audio signal processes related to echo cancellation, AGC, and noise suppression.
@details  This API can also be called from threads other than the one that calls the <tt>common::Scheduler::Dispatch()</tt> function.
@details  The assumption is that this API will receive audio data actually picked up by the microphone.
You must be careful when passing synthesized audio data. For more information, see the <tt>SetVADClampGain()</tt> function.

@param[in] pAudioBuf  The audio data to send (generally microphone input). The sampling frequency is 8 kHz.
@param[in] samples  Number of samples. These samples must correspond to the audio data frame length.
@param[in] pRefLeftIn  Reference data input for the left channel. The sampling frequency is 32 kHz.
@param[in] pRefRightIn  Reference data input for the right channel. The sampling frequency is 32 kHz.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that this API function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
@see  Receive



*/
    nn::Result Send(const s16* pAudioBuf, size_t samples, const s16* pRefLeftIn, const s16* pRefRightIn);


/*!
@brief  Enables or disables echo cancellation.
@details  By default, echo cancellation is enabled.

@param[in] bEnable  true - Enabled, false - Disabled

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), IsEnableEchoCanceller, SetEchoCancellerMode, GetEchoCancellerMode
*/
    nn::Result EnableEchoCanceller(bool bEnable);


/*!
@brief  Gets whether echo cancellation is enabled or disabled.

@return  Returns <tt>true</tt> if echo cancellation is enabled, and returns <tt>false</tt> otherwise.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableEchoCanceller, SetEchoCancellerMode, GetEchoCancellerMode
*/
    bool IsEnableEchoCanceller() const;


/*!
@brief  Sets the echo cancellation operation mode.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableEchoCanceller, IsEnableEchoCanceller, GetEchoCancellerMode
*/
    nn::Result SetEchoCancellerMode(enum EcMode ecMode);


/*!
@brief  Gets the echo cancellation operation mode.

@return  Returns the echo cancellation operation mode.
@retval EC_VOID  Indicates that the echo cancellation operation mode is not in an obtainable state.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableEchoCanceller, IsEnableEchoCanceller, SetEchoCancellerMode
*/
    enum EcMode GetEchoCancellerMode() const;


/*!
@brief  Enables or disables the Automatic Gain Control (AGC) feature.

@details  By default, the AGC feature is enabled.
@param[in] bEnable  true - Enabled, false - Disabled

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), IsEnableAgc, SetAgcLevel, GetAgcLevel
*/
    nn::Result EnableAgc(bool bEnable);


/*!
@brief  Gets whether the Automatic Gain Control (AGC) feature is enabled or disabled.

@return  Returns <tt>true</tt> if the AGC feature is enabled or <tt>false</tt> otherwise.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableAgc, SetAgcLevel, GetAgcLevel
*/
    bool IsEnableAgc() const;


/*!
@brief  Sets the Automatic Gain Control (AGC) level.
Values of at least <tt>AGC_LEVEL_MIN</tt> and no greater than <tt>AGC_LEVEL_MAX</tt> can be set.

@details  If the argument is set to <tt>AGC_LEVEL_MIN</tt>, AGC is not actually being used because the amplification rate is 1×.
@details  The greater the value set, the more the amplification rate increases, but it also becomes easier for feedback to occur. Setting a value of <tt>1</tt> or <tt>2</tt> is recommended.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableAgc, IsEnableAgc, GetAgcLevel
*/
    nn::Result SetAgcLevel(s32 level);


/*!
@brief  Gets the AGC (Automatic Gain Control) level.

@return  Returns a value of at least <tt>AGC_LEVEL_MIN</tt> but not greater than <tt>AGC_LEVEL_MAX</tt> if successful.
@retval -1  The AGC level is not in a state where it can be obtained.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableAgc, IsEnableAgc, SetAgcLevel
*/
    s32 GetAgcLevel() const;


/*!
@brief  Enables or disables noise suppression.
@details  By default, noise suppression is enabled.

@param[in] bEnable  true - Enabled, false - Disabled
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), IsEnableNoiseSuppressor, SetNoiseSuppressorLevel, GetNoiseSuppressorLevel
*/
    nn::Result EnableNoiseSuppressor(bool bEnable);


/*!
@brief  Gets whether noise suppression is enabled or disabled.

@return  Returns <tt>true</tt> if the noise suppression feature is enabled or <tt>false</tt> otherwise.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableNoiseSuppressor, SetNoiseSuppressorLevel, GetNoiseSuppressorLevel
*/
    bool IsEnableNoiseSuppressor() const;

/*!
@brief  Sets the noise suppression level.
Values of at least <tt>NOISE_SUPPRESSOR_LEVEL_MIN</tt> and no greater than <tt>NOISE_SUPPRESSOR_LEVEL_MAX</tt> can be set.
<tt>NOISE_SUPPRESSOR_LEVEL_DEFAULT</tt> is the recommended value.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  Indicates that this API was called when the <tt>VoiceProtocol</tt> instance was not in an initialized state. Programming error. Fix your program so that this error is not returned.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableNoiseSuppressor, IsEnableNoiseSuppressor, GetNoiseSuppressorLevel
*/
    nn::Result SetNoiseSuppressorLevel(s32 level);


/*!
@brief  Gets the noise suppression level.

@return  Returns a value of at least <tt>NOISE_SUPPRESSOR_LEVEL_MIN</tt> but not greater than <tt>NOISE_SUPPRESSOR_LEVEL_MAX</tt> if successful.
@retval –1  The noise suppression level is not in a state where it can be obtained.
@see  Send(const s16*, size_t, const s16*, const s16*), EnableNoiseSuppressor, IsEnableNoiseSuppressor, SetNoiseSuppressorLevel
*/
    s32 GetNoiseSuppressorLevel() const;
#endif  // end of NN_PIA_HF_LIBRARY


#if NN_PIA_CTR
/*!
@cond PRIVATE
@brief  Enables or disables the biquad filter.

@details  By default, the biquad filter is disabled.
When this feature is enabled, the biquad filter is applied inside the <tt>@ref  Dispatch</tt> function to eliminate non-speech noise in the audio being sent.
@param[in] enable  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.


*/
    void EnableBPF(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the biquad filter is enabled.

@retval true  Indicates that the biquad filter is enabled.
@retval false  Indicates that the biquad filter is disabled.
*/
    bool IsEnableBPF(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Enables or disables the echo cancellation feature.

@details  The echo cancellation feature is enabled by default.
When the echo cancellation feature is turned on, echo cancellation runs in the <tt>@ref Send</tt> function.
@param[in] flag  Specify <tt>true</tt> to enable or <tt>false</tt> to disable.

*/
    void EnableEchoCancel(bool enable);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether the echo cancellation feature is enabled.

@retval true  Indicates that the echo cancellation feature is enabled.
@retval false  Indicates that the echo cancellation feature is disabled.
*/
    bool IsEnableEchoCancel(void) const;
    //! @endcond


/*!
@brief  Supplies reference data for echo cancellation.

@details  Call the <tt>nn::snd::GetMixedBusData</tt> function at every sound DSP interrupt (every 5 ms), and pass the 5 ms of 16-bit/32 kHz stereo data obtained by that function to the <tt><var>pBuffer</var></tt> of this function.
This function does nothing if <tt><var>size</var></tt> is not the correct size for 5 milliseconds of 16-bit/32 kHz stereo data.
@param[in] pBuffer  Specifies reference data for echo cancellation.
@param[in] size  Specifies the size of the reference data for echo cancellation.



*/
    void HandleReference(const u8* pBuffer, u32 size);
#endif  // end of NN_PIA_CTR


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


private:
    static const size_t NET_SEND_BUFF_SIZE = 2048;
    static const size_t NET_SEND_DATA_MAX = 80;

    struct SendDataStructure
    {
        u32 sendBuff[NET_SEND_BUFF_SIZE / sizeof(u32)];
        nn::nex::ConnectionID destCid;
        u32 dataSize;
        u32 destCidBmp;
    };

    bool allocProc(size_t maxStationNum);
    void freeProc(void);
    size_t calcAudioBufferSize(size_t maxStationNum) const;

    bool initializeVct(void);
    void finalizeVct(void);

    nn::Result setupProtocols(u16 port, size_t recvBufNum);
    void destroyProtocols(void);

    void updateSessionEndMonitoringData();

    static bool SendDataCallback(u16 cid, const void* buff, u32 size);
    bool SendDataCallbackImpl(u32 cid, const void* pBuff, u32 size);

    static bool SendMultiDataCallback(u32 cidBmp, const void* pBuf, size_t size);
    bool SendMultiDataCallbackImpl(u32 cidBmp, const void* pBuf, size_t size);

    s32 getPassedTime() const;  // Returns the elapsed time since the <tt>Startup</tt> function was called. The value is in milliseconds.

    nn::Result receiveCore(void* pAudioBuf, u32 frameDataSize, u16* pCidList);

    u32 m_UnreliableHandle;
    transport::UnreliableProtocol* m_pUnreliable;

    VCTSession* m_pSession;
    u8* m_pAudioBuffer;
    SendDataStructure* m_pSendDataInfo;
    u32 m_ReqSendDataIndexR;
    u32 m_ReqSendDataIndexW;

    StationIndex m_LocalStationIndex;

    Setting m_Setting;  // The argument that was passed when calling <tt>Initialize</tt> is copied.

    common::CriticalSection m_SendCs;

    size_t m_DispatchCount;  // Specifies the number of times the <tt>Dispatch</tt> function was called.

    u64 m_TotalSendDataSize;
    u64 m_TotalReceiveDataSize;
    common::Time m_StartupTime;

#if NN_PIA_HF_LIBRARY
    HandsFree* m_pHandsFree;
    s16* m_pAudioOutBuf;
#endif

    static VoiceProtocol* s_pVoiceProtocol;  // Singleton pointer.
};
}
}
}  // end of namespace nn::pia::chat
