﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     chat_types.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once


namespace nn
{
namespace pia
{
namespace chat
{


/*!
@brief  Enumerates audio codecs.
*/
enum Codec
{
    CODEC_8BIT_RAW,    //!< Specifies 8-bit 8 kHz raw audio (unsupported).
    CODEC_G711_ULAW,   //!< Specifies G711 u-Law (unsupported).
    CODEC_2BIT_ADPCM,  //!< Specifies 2-bit IMA-ADPCM (non-standard format).
    CODEC_3BIT_ADPCM,  //!< Specifies 3-bit IMA-ADPCM (unsupported).
    CODEC_4BIT_ADPCM,  //!< Specifies 4-bit IMA-ADPCM.
    CODEC_END
};

/*!
@brief  Enumerates the modes of echo cancellation.
*/
#if NN_PIA_CTR
enum EcMode
{
    EC_STANDARD,        //!< Specifies standard echo cancellation.
    EC_HIGHPERFORMANCE  //!< Specifies advanced echo cancellation (only usable in debug mode).
};
#elif NN_PIA_CAFE
enum EcMode
{
    EC_VOID = -1,     //!< Indicates an invalid mode.
    EC_PERFORMANCE,   //!< A mode that prioritizes echo suppression. Sounds may become distorted, however.
    EC_STANDARD,      //!< An intermediate mode between <tt>EC_PERFORMANCE</tt> and <tt>EC_SOUNDQUALITY</tt>.
    EC_SOUNDQUALITY,  //!< A mode that emphasizes not distorting the sound and prioritizes sound quality. Echos may be slightly more frequent, however.
};
#endif


/*!
@brief  Contains audio streaming information.
*/
struct AudioInfo
{
    Codec codec;      /*!< Specifies the audio codec of the peer. */
    u8 padding0_[3];  //   Represents a reserved region.

    u32 sequence; /*!< Specifies the current audio packet sequence number. */

#if NN_PIA_CAFE
    OSTime bufferLatency; /*!< Specifies the jitter buffer latency.                   */
    OSTime jitter;        /*!< Specifies the maximum jitter buffer latency (in ticks). */
    OSTime clockSkew;     /*!< Specifies the clock skew (in ticks).         */
#elif NN_PIA_CTR
    nn::os::Tick bufferLatency; /*!< Specifies the jitter buffer latency.                   */
    nn::os::Tick jitter;        /*!< Specifies the maximum jitter buffer latency (in ticks). */
    nn::os::Tick clockSkew; /*!< Specifies the clock skew (in ticks).         */
#endif

    u32 internalLatency; /*!< Specifies the internal receive latency amount.               */
    u32 bufferCount;     /*!< Specifies the number of packets in the jitter buffer. */
    u32 dropCount;       /*!< Specifies the number of packets that were dropped.       */
    u32 continuousDrop;  /*!< Specifies the most recent number of consecutive dropped packets. */

    u32 jamCount;       /*!< Specifies the number of packets that did not arrive in order.                         */
    u32 duplicateCount; /*!< Specifies the number of duplicate packets.                                       */
    u32 waitCount;      /*!< Specifies the number of times packet playback was deferred and drop recovery attempted.    */
    u32 lateCount;      /*!< Specifies the number of packets that were late due to network congestion or another reason. */

    u32 recentLateCount;    /*!< Specifies the number of packets that were late recently.                                                */
    u32 recentReceiveCount; /*!< Specifies the number of packets received recently.                                                          */
    u32 recentLateRatio;    /*!< Specifies the ratio of packets that were late recently (<tt>recentLateCount / recentReceiveCount</tt>). */
    u32 recoverCount;       /*!< Specifies the number of recovered packets.                                                                        */

    u32 recoverUsingSpareCount; /*!< Specifies the number of packets that were recovered using spare packets. */
    u32 padding1_;              //   Represents a reserved region.
};
}
}
}
