﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_AtomicSharingClone.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_SharingClone.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  This class is a derivative of <tt>CloneBase</tt> and provides functionality for exclusively controlling the lock on sending and for bidirectional sending and receiving.

@date 2013-09-25  Initial version.
*/
class AtomicSharingClone : public SharingClone
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    AtomicSharingClone();


/*!
@brief  Destroys the object (destructor).
*/
    virtual ~AtomicSharingClone();


/*!
@brief  Attempts to acquire a lock on sending.
@details  If this function is called while the state is <tt>@ref LOCK_STATUS_UNLOCKED</tt>, the state transitions to <tt>@ref LOCK_STATUS_TRY_LOCK</tt>.
If the lock successfully acquired, the state changes to <tt>@ref LOCK_STATUS_LOCKED</tt>.
If the lock cannot be acquired because another station has them, the state changes to <tt>@ref LOCK_STATUS_UNLOCKED</tt>.
This function does nothing if called while the state is <tt>@ref LOCK_STATUS_TRY_LOCK</tt> or <tt>@ref LOCK_STATUS_LOCKED</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that the object is not active. (<tt>IsActive</tt> returns <tt>false</tt>. ) Programming error. Fix your program so that this error is not returned.
*/
    nn::Result TryLock();


/*!
@brief  Releases the send lock.
@details  If this function is called while the state is <tt>@ref LOCK_STATUS_LOCKED</tt>, the state transitions to <tt>@ref LOCK_STATUS_UNLOCKED</tt> and other stations can acquire the lock.
If called while the state is <tt>@ref LOCK_STATUS_TRY_LOCK</tt>, the object stops trying to acquire a lock and the state changes to <tt>@ref LOCK_STATUS_UNLOCKED</tt>.
This function does nothing if it is called while the state is <tt>@ref LOCK_STATUS_UNLOCKED</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that the object is not active. (<tt>IsActive</tt> returns <tt>false</tt>. ) Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Unlock();


/*!
@brief  Enumerates the status of the lock.
*/
    enum LockStatus
    {
        LOCK_STATUS_UNLOCKED = 0x10,  //!< Specifies that the station does not have the lock.
        LOCK_STATUS_TRY_LOCK = 0x20,  //!< Specifies that the station is attempting to acquire the lock.
        LOCK_STATUS_LOCKED = 0x30     //!< Specifies that the station has the lock.
    };


/*!
@brief  Gets the status of the lock.
@return  Returns the status of the lock.
*/
    LockStatus GetLockStatus() const
    {
        return static_cast<LockStatus>(m_LockStatus & cLockStatusMask);
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


    //! @cond PRIVATE

protected:
    virtual void OnStartup();
    virtual void OnCleanup();
    virtual void OnEnd();

    virtual void OnExit(StationIndex stationId);
    virtual void OnDisconnect(StationIndex stationId, bool isValid);

    virtual void OnAddParticipant(StationIndex stationIndex);

public:
    virtual void ReceiveReject(StationIndex src, ClockValue clock);
    virtual void ReceiveStop(StationIndex src);

    enum LockResult
    {
        LOCK_RESULT_OK,
        LOCK_RESULT_NG,
        LOCK_RESULT_DEFER
    };

    LockResult ReceiveTryLock(StationIndex src, ClockValue clock, u32 participantBitmap);
    void ReceiveLockOK(StationIndex src, ClockValue clock, u8 count, u32 participantBitmap);
    void ReceiveLockNG(StationIndex src, ClockValue clock);

public:
    static Type GetTypeStatic()
    {
        return TYPE_ATOMIC;
    }
    static Key ToKey(Id id)
    {
        return CloneBase::ToKey(TYPE_ATOMIC, STATION_INDEX_INVALID, id);
    }
    virtual Type GetTypeVirtual() const
    {
        return GetTypeStatic();
    }

public:
    class LockCommandToken : public ClockCommandToken
    {
    public:
        explicit LockCommandToken(AtomicSharingClone* pClone)
          : ClockCommandToken(pClone), m_DestBitmap(0)
        {
        }
        virtual CommandType GetType() const
        {
            return COMMAND_LOCK;
        }
        virtual u32 GetDestStationBitmap() const
        {
            return m_DestBitmap;
        }
        virtual bool IsUnicast() const
        {
            return true;
        }
        void SetDestBitmap(u32 destBitmap)
        {
            PIA_TRACE_EX(common::TRACE_FLAG_CLONE, "called LockCommandToken key:0x%016llx state:%d dest:0x%08x -> 0x%08x", GetClone() != NULL ? GetClone()->GetKey() : 0, GetClone() != NULL ? GetClone()->GetState() : 0, m_DestBitmap, destBitmap);
            m_DestBitmap = destBitmap;
        }
        void Exit(StationIndex stationId)
        {
            PIA_TRACE_EX(common::TRACE_FLAG_CLONE, "called LockCommandToken key:0x%016llx state:%d dest:0x%08x -> 0x%08x", GetClone() != NULL ? GetClone()->GetKey() : 0, GetClone() != NULL ? GetClone()->GetState() : 0, m_DestBitmap, m_DestBitmap & ~(0x1 << stationId));
            m_DestBitmap &= ~(0x1 << stationId);
        }
        virtual ClockValue GetClock() const
        {
            return static_cast<const AtomicSharingClone*>(GetClone())->m_LockClock;
        }

    private:
        ClockValue m_DestBitmap;
    };

    virtual bool IsReadyToSetValue() const;

private:
    enum InnerLockStatus
    {
        eILUnlocked = LOCK_STATUS_UNLOCKED | 0x1,
        eILKeepLock = LOCK_STATUS_UNLOCKED | 0x2,
        eILTryLock = LOCK_STATUS_TRY_LOCK | 0x1,
        eILLocked = LOCK_STATUS_LOCKED | 0x1
    };

    enum
    {
        cLockStatusMask = 0xf0
    };

    InnerLockStatus m_LockStatus;
    ClockValue m_LockClock;
    ClockValue m_UnlockClock;

    LockCommandToken m_LockCommandToken;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const AtomicSharingClone& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
