﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_CloneBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>
#include <pia/common/common_TreeMapNode.h>
#include <pia/common/common_OffsetTreeMap.h>
#include <pia/common/common_OffsetList.h>

namespace nn
{
namespace pia
{
namespace clone
{

class CloneProtocol;

/*!
@brief  This is the base class for managing sending and receiving.
@date 2013-11-25  Added a profiling feature.
@date 2013-07-18  Initial version.

*/
class CloneBase
{
public:
protected:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    CloneBase();


public:
/*!
@brief  Destroys the object.
*/
    virtual ~CloneBase();


/*!
@brief  Gets the <tt>CloneProtocol</tt> this object is registered to.
@return  Returns the <tt>CloneProtocol</tt> this object is registered to.
*/
    const CloneProtocol* GetProtocol() const
    {
        return m_pProtocol;
    }


/*!
@brief  Specifies the type of the ID for identifying clones.
*/
    typedef u32 Id;


/*!
@brief  Gets the ID.
@return  Returns the ID specified when registering the object with <tt>CloneProtocol</tt>. Returns <tt>0</tt> if not registered with <tt>CloneProtocol</tt>.
*/
    Id GetId() const
    {
        return static_cast<Id>(m_TreeMapNode.GetKey() & 0xffffffff);
    }


/*!
@brief  Determines whether the object is registered with <tt>CloneProtocol</tt>.
@return  Returns <tt>true</tt> if it is registered with <tt>CloneProtocol</tt>.
*/
    bool IsRegisteredWithProtocol() const
    {
        return m_pProtocol != NULL;
    }


/*!
@brief  Registers a clone element.
@details  This function must be called while the object is not registered with <tt>CloneProtocol</tt>.
@param[in] pElement  Specifies the clone element to register.
@param[in] elementId  Specifies the ID of the clone element to register.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that this object is already registered with <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that a clone element is already registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result RegisterElement(CloneElementBase* pElement, CloneElementBase::Id elementId);


/*!
@brief  Unregisters a clone element.
@details  This function must be called while the object is not registered with <tt>CloneProtocol</tt>.
@param[in] pElement  Specifies the clone element to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that this object is already registered with <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result UnregisterElement(CloneElementBase* pElement);


/*!
@brief  Unregisters a clone element.
@details  This function must be called while the object is not registered with <tt>CloneProtocol</tt>.
@param[in] elementId  Specifies the ID of the clone element to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that this object is already registered with <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.
@retval ResultNotSet  Indicates that a clone element is not registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result UnregisterElement(CloneElementBase::Id elementId);


/*!
@brief  Searches for the clone element of the specified ID.
@param[in] elementId  Specifies the clone element to search for.
@return  Returns the clone element with the specified ID. Returns <tt>NULL</tt> if it does not exist.
*/
    CloneElementBase* SearchElement(CloneElementBase::Id elementId)
    {
        return m_ElementCollection.Find(elementId);
    }


/*!
@brief  Searches for the clone element of the specified ID.
@param[in] elementId  Specifies the clone element to search for.
@return  Returns the clone element with the specified ID. Returns <tt>NULL</tt> if it does not exist.
*/
    const CloneElementBase* SearchElement(CloneElementBase::Id elementId) const
    {
        return m_ElementCollection.Find(elementId);
    }


/*!
@brief  Determines whether this object is active.
@details  An object is treated as active when it is registered with <tt>CloneProtocol</tt> and can actually send and receive data.
An object is not active when it is in the process of being registered or unregistered with <tt>CloneProtocol</tt>, or when <tt>CloneProtocol</tt> is starting up or being finalized.
@return  Returns <tt>true</tt> if active.
*/
    bool IsActive() const
    {
        return (m_State == STATE_ACTIVE);
    }


    //! @name  Profiling
    //! @{

/*!
@brief  Sets an object to manage send/receive profiling.
@details  Send/receive tasks carried out by registered clone elements and clones themselves can be profiled by configuring <tt>CloneProfiler</tt>.
This calculation includes all exchanges of data between this clone and other clones, so the data exchanges between clones on the local station are also included in the value.
@details  When multiple clone elements are sending data at the same time, it gets calculated on a per-clone basis. In most cases, the total send/receive count of the clones is less than that of the clone elements.
@details  When compression is enabled, the compressed size is used in the calculation.
For this reason, calculations may result in a value that is small relative to the total amount of data sent or received that was measured for registered clone elements.
@details  A <tt>CloneProfiler</tt> instance that is already configured can be released by specifying <tt>NULL</tt> in the arguments.
@param[in] pSendProfiler  Specifies the <tt>CloneProfiler</tt> that manages the send profiling.
Specify <tt>NULL</tt> if not necessary.
@param[in] pReceiveProfiler  Specifies the <tt>CloneProfiler</tt> that manages the receive profiling.
Specify <tt>NULL</tt> if not necessary.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.


*/
    nn::Result SetProfiler(CloneProfilerBase* pSendProfiler, CloneProfilerBase* pReceiveProfiler);


/*!
@brief  Gets the object configured to manage send profiling.
@return  The object configured to manage send profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetSendProfiler() const
    {
        return m_pSendProfiler;
    }


/*!
@brief  Gets the object configured to manage receive profiling.
@return  The object configured to manage receive profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetReceiveProfiler() const
    {
        return m_pReceiveProfiler;
    }


/*!
@brief  Resets the results of send/receive profiling.
@see  SetProfiler
*/
    void ResetProfiler();


/*!
@brief  Gets the send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveSize(StationIndex stationIndex) const;

    //! @}

    virtual void Trace(u64 flag) const;

    //! @cond PRIVATE

public:
    enum Type
    {
        TYPE_SEND = 0x01,
        TYPE_RECEIVE = 0x02,
        TYPE_ATOMIC = 0x03,
        TYPE_SEQUENTIAL = 0x04
    };

    virtual Type GetTypeVirtual() const = 0;


    typedef u64 Key;


    static s32 GetTreeMapNodeOffset()
    {
        return offsetof(CloneBase, m_TreeMapNode);
    }

    void UnregisterElementForce(CloneElementBase* pElement);

    bool CheckElementSize(size_t sizeMax);
    void AttachProtocol(CloneProtocol* pProtocol);
    void DetachProtocol();

    Type GetType() const
    {
        return static_cast<Type>(m_TreeMapNode.GetKey() >> 48);
    }
    StationIndex GetStationIndex() const
    {
        return static_cast<StationIndex>((m_TreeMapNode.GetKey() >> 32) & 0xff);
    }
    Key GetKey() const
    {
        return m_TreeMapNode.GetKey();
    }
    static Key ToKey(Type type, StationIndex stationId, Id id)
    {
        return (static_cast<u64>(type) << 48) | (static_cast<u64>(stationId) << 32) | id;
    }


    enum State
    {
        STATE_INACTIVE = 0,
        STATE_ACTIVE,
        STATE_COMPLETE,
        STATE_ENDING,
        STATE_ENDED,
        STATE_COMPLETE_FOR_UNREGISTER,
        STATE_ENDING_FOR_UNREGISTER,
        STATE_ENDED_FOR_UNREGISTER
    };

    State GetState() const
    {
        return m_State;
    }

    CloneProfilerBase* GetSendProfilerPtr()
    {
        return m_pSendProfiler;
    }
    CloneProfilerBase* GetReceiveProfilerPtr()
    {
        return m_pReceiveProfiler;
    }

public:
    void Startup();
    void Cleanup();
    void End();
    void Unregister();
    void ClearData();

    void Enter(StationIndex stationId)
    {
        OnEnter(stationId);
    }
    void Exit(StationIndex stationId)
    {
        OnExit(stationId);
    }
    void Disconnect(StationIndex stationId, bool isValid)
    {
        OnDisconnect(stationId, isValid);
    }

    void UpdateState();
    bool IsReceiveData() const
    {
        return m_State != STATE_INACTIVE;
    }

    bool IsSendingData(StationIndex stationId, bool isDebug = false) const;

protected:
    virtual void OnStartup()
    {
    }
    virtual void OnCleanup()
    {
    }
    virtual void OnEnd()
    {
    }
    virtual void OnComplete()
    {
    }
    virtual bool OnCheckEnd()
    {
        return true;
    }

    virtual void OnEnter(StationIndex stationId)
    {
        (void)stationId;
    }
    virtual void OnExit(StationIndex stationId)
    {
        (void)stationId;
    }
    virtual void OnDisconnect(StationIndex stationId, bool isValid)
    {
        (void)stationId;
        (void)isValid;
    }
    virtual void OnRegister()
    {
    }

public:
    virtual bool IsReadyToSetValue() const
    {
        return false;
    }
    virtual u32 GetDestBitmap() const = 0;
    virtual bool IsInRequest() const
    {
        return false;
    }

    virtual u8 GetLocalRegisterCount() const
    {
        return 0;
    }
    virtual u64 GetRegisterCountAll() const
    {
        return 0;
    }
    virtual bool IsMatchLocalRegisterCount(u64 registerCountAll) const
    {
        NN_PIA_DUMMY_PARAM(registerCountAll);
        return true;
    }

    void AddSendData(CloneElementBase::ISendToken* pToken);
    void AddSendDelayData(CloneElementBase::ResendableSendToken* pToken);
    void CancelSendData(CloneElementBase::ISendToken* pToken);
    void ErrorDropEvent();


private:
    typedef common::OffsetList<CloneElementBase::ISendToken> SendDataList;
    typedef common::OffsetList<CloneElementBase::ResendableSendToken> ResendableDataList;

public:
    class DataTokenAccessor
    {
    public:
        DataTokenAccessor(CloneBase* pClone, SystemTime time, SystemTime resendSpan, SystemTime maxSendDelay, bool isUnicast, StationIndex stationIndex);
        CloneElementBase::ISendToken* Get() const
        {
            return m_pToken;
        }
        void Reset(bool isUnicast, StationIndex stationIndex, SystemTime resendSpan);
        void Next();
        void Commit(u32 dispatchCount, u32 destBitmap);

    private:
        void NextToken(bool isFirst);
        void NextUnicastToken();

    private:
        CloneBase* m_pClone;
        SystemTime m_Time;
        SystemTime m_ResendSpan;
        SystemTime m_MaxSendDelay;
        CloneElementBase::ISendToken* m_pToken;
        bool m_IsUnicast;
        StationIndex m_TargetStationIndex;
        u32 m_TargetBitmap;

        enum ListState
        {
            eLSFirst,
            eLSResend,
            eLSDeferment,
            eLSEnd,
            eLSNoData,
            eLSComitted
        };
        ListState m_ListState;
    };

protected:
    void AddDataDest(StationIndex stationId);
    void RemoveDataDest(StationIndex stationId);
    void AddDataParticipant(StationIndex stationId);
    void ComplementData(StationIndex stationId);
    void LockData();

public:
    enum CommandType
    {
        COMMAND_ANNOUNCE,
        COMMAND_REQUEST,
        COMMAND_END,
        COMMAND_LOCK
    };

    enum
    {
        INVALID_COMMAND_INDEX = 0
    };

    class CommandToken
    {
    protected:
        explicit CommandToken(CloneBase* pClone)
          : m_ListNode(), m_pClone(pClone), m_LastSendTime(INVALID_SYSTEM_TIME)
        {
        }

    public:
        bool IsInList() const
        {
            return m_ListNode.IsFreeListNode() == false;
        }
        virtual CommandType GetType() const = 0;
        virtual u32 GetDestStationBitmap() const = 0;
        Id GetCloneId() const
        {
            return m_pClone->GetId();
        }
        SystemTime GetLastSendTime(StationIndex stationIndex) const
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                return m_LastSendTime;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                return m_LastSendTimeArray[stationIndex];
            }
        }
        void SetLastSendTime(SystemTime time, StationIndex stationIndex)
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                m_LastSendTime = time;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                m_LastSendTimeArray[stationIndex] = time;
            }
        }
        const CloneBase* GetClone() const
        {
            return m_pClone;
        }
        static s32 GetListNodeOffset()
        {
            return offsetof(CommandToken, m_ListNode);
        }
        CloneBase* GetClonePtr()
        {
            return m_pClone;
        }
        // Overridden to return <tt>true</tt> in token classes that resend using unicast.
        virtual bool IsUnicast() const
        {
            return false;
        }

    private:
        common::ListNode m_ListNode;
        CloneBase* m_pClone;
        SystemTime m_LastSendTime;
        SystemTime m_LastSendTimeArray[MAX_STATION_NUM];
    };


    class ClockCommandToken : public CommandToken
    {
    protected:
        ClockCommandToken(CloneBase* pClone)
          : CommandToken(pClone)
        {
        }

    public:
        virtual ClockValue GetClock() const = 0;
    };


    class DataToken
    {
    public:
        explicit DataToken(CloneBase* pClone)
          : m_ListNode(), m_pClone(pClone)
        {
        }
        bool IsInList() const
        {
            return m_ListNode.IsFreeListNode() == false;
        }
        CloneBase* GetClone()
        {
            return m_pClone;
        }
        bool CheckEraseFromList() const
        {
            return m_pClone->IsSendDataEmpty();
        }
        static s32 GetListNodeOffset()
        {
            return offsetof(DataToken, m_ListNode);
        }

    private:
        common::ListNode m_ListNode;
        CloneBase* m_pClone;
    };


private:
    // Copying is prohibited.
    CloneBase(const CloneBase& rhs);
    CloneBase& operator=(const CloneBase& rhs);


protected:
    CloneProtocol* GetProtocolPtr()
    {
        return m_pProtocol;
    }

    void RequestInitialData();


protected:
    bool IsSendDataEmpty() const
    {
        return m_SendDataList.IsEmpty() && m_SendDelayDataList.IsEmpty() && m_ResendDataList.IsEmpty();
    }

private:
    common::TreeMapNode<Key> m_TreeMapNode;
    CloneProtocol* m_pProtocol;

    State m_State;

    typedef common::OffsetTreeMap<CloneElementBase::Id, CloneElementBase> ElementCollection;
    ElementCollection m_ElementCollection;


    SendDataList m_SendDataList;
    ResendableDataList m_ResendDataList;
    ResendableDataList m_SendDelayDataList;

    DataToken m_DataToken;

private:
    CloneProfilerBase* m_pSendProfiler;
    CloneProfilerBase* m_pReceiveProfiler;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const CloneBase& obj, Arg* pArg = NULL);

private:
    // n2920: For the client-monitoring feature.
    void SetElementNumToMonitoringData();
    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
