﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_CloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/common/common_TreeMapNode.h>
#include <pia/common/common_ListNode.h>

namespace nn
{
namespace pia
{
namespace clone
{

class CloneBase;
class CloneProtocol;
class IDataPacker;
class IDataUnpacker;
class CloneProfilerBase;

/*!
@brief  This is the base class for managing data that is sent and received.
@date 2013-11-25  Added a profiling feature.
@date 2013-07-18  Initial version.

*/
class CloneElementBase
{
protected:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    CloneElementBase();


public:
/*!
@brief  Destroys the object.
*/
    virtual ~CloneElementBase();


/*!
@brief  Specifies the type of the ID for identifying clone elements.
*/
    typedef u16 Id;


/*!
@brief  Gets the ID.
@return  Returns the ID specified when registering the object with <tt>CloneBase</tt>. Returns <tt>0</tt> if the object is not registered with <tt>CloneBase</tt>.
(If <tt>0</tt> was specified as the ID when registering the object, <tt>0</tt> may be returned as a valid value.)
*/
    Id GetId() const
    {
        return m_TreeMapNode.GetKey();
    }


/*!
@brief  Gets the <tt>CloneBase</tt> this object is registered to.
@return  Returns the <tt>CloneBase</tt> this object is registered to.
*/
    const CloneBase* GetCloneBase() const
    {
        return m_pCloneBase;
    }


/*!
@brief  Gets the <tt>CloneProtocol</tt> of the <tt>CloneBase</tt> this object is registered to.
@return  Returns the <tt>CloneProtocol</tt> of the <tt>CloneBase</tt> this object is registered to.
*/
    const CloneProtocol* GetProtocol() const;


/*!
@brief  Determines whether the object is registered with <tt>CloneBase</tt>.
@return  Returns <tt>true</tt> if it is registered with <tt>CloneBase</tt>.
*/
    bool IsRegisteredWithCloneBase() const
    {
        return m_pCloneBase != NULL;
    }


/*!
@brief  Determines whether the object is registered with <tt>CloneProtocol</tt>.
@return  Returns <tt>true</tt> if it is registered with <tt>CloneProtocol</tt>.
*/
    bool IsRegisteredWithProtocol() const;


/*!
@brief  Determines whether a value can be set.
@return  Returns <tt>true</tt> if a value can be set.
*/
    bool IsReadyToSetValue() const;


    //! @name  Profiling
    //! @{

/*!
@brief  Sets an object to manage send/receive profiling.
@details  Clone element send/receive tasks can be profiled by configuring <tt>CloneProfiler</tt>.
This calculation includes all exchanges of data between this clone element and other clone elements, so the data exchanges between clone elements on the local station are also included in the value.
@details  Even when compression is enabled, the uncompressed size is used in the calculation.
@details  A <tt>CloneProfiler</tt> instance that is already configured can be released by specifying <tt>NULL</tt> in the arguments.
@param[in] pSendProfiler  Specifies the <tt>CloneProfiler</tt> that manages the send profiling.
Specify <tt>NULL</tt> if not necessary.
@param[in] pReceiveProfiler  Specifies the <tt>CloneProfiler</tt> that manages the receive profiling.
Specify <tt>NULL</tt> if not necessary.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result SetProfiler(CloneProfilerBase* pSendProfiler, CloneProfilerBase* pReceiveProfiler);


/*!
@brief  Gets the object configured to manage send profiling.
@return  The object configured to manage send profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetSendProfiler() const
    {
        return m_pSendProfiler;
    }


/*!
@brief  Gets the object configured to manage receive profiling.
@return  The object configured to manage receive profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetReceiveProfiler() const
    {
        return m_pReceiveProfiler;
    }


/*!
@brief  Resets the results of send/receive profiling.
@see  SetProfiler
*/
    void ResetProfiler();


/*!
@brief  Gets the send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveSize(StationIndex stationIndex) const;

    //! @}


    //! @cond PRIVATE

public:
    static s32 GetTreeMapNodeOffset()
    {
        return offsetof(CloneElementBase, m_TreeMapNode);
    }

    virtual bool IsSupportedClone(int cloneType) const
    {
        NN_PIA_DUMMY_PARAM(cloneType);
        return true;
    }
    void AttachCloneBase(CloneBase* pCloneBase);
    void DetachCloneBase();

    CloneProfilerBase* GetSendProfilerPtr()
    {
        return m_pSendProfiler;
    }
    CloneProfilerBase* GetReceiveProfilerPtr()
    {
        return m_pReceiveProfiler;
    }

    enum Type
    {
        TYPE_UNRELIABLE = 0x1000,
        TYPE_RELIABLE = 0x2000,
        TYPE_EVENT = 0x3000,
        TYPE_RECKONING = 0x4000,
        TYPE_RELIABLE_LARGE = 0x5000
    };

    virtual u16 GetType() const = 0;
    virtual size_t GetSize() const = 0;

    virtual bool IsInitialized() const
    {
        return true;
    }

    virtual void ClearData() = 0;
    virtual void AddDest(StationIndex stationId)
    {
        NN_PIA_DUMMY_PARAM(stationId);
    }
    virtual void RemoveDest(StationIndex stationId)
    {
        NN_PIA_DUMMY_PARAM(stationId);
    }
    virtual void AddParticipant(StationIndex stationId)
    {
        NN_PIA_DUMMY_PARAM(stationId);
    }
    virtual void Complement(StationIndex stationId)
    {
        NN_PIA_DUMMY_PARAM(stationId);
    }
    virtual void Lock()
    {
    }

    virtual void RequestInitialData()
    {
    }

    u32 GetStationNum() const;

public:
    class ISendToken
    {
    public:
        ISendToken()
          : m_ListNode()
        {
        }
        bool IsInList() const
        {
            return m_ListNode.IsFreeListNode() == false;
        }
        virtual Type GetElementType() const = 0;
        virtual u32 GetDestBitmap() const = 0;
        virtual bool IsResend() const
        {
            return false;
        }
        virtual size_t GetChunkSize() const = 0;
        virtual void WriteChunk(void* pChunkBuffer) const = 0;
        virtual CloneElementBase* GetElement() = 0;
        static s32 GetListNodeOffset()
        {
            return offsetof(ISendToken, m_ListNode);
        }

    private:
        common::ListNode m_ListNode;
    };


    class ResendableSendToken : public ISendToken
    {
    public:
        ResendableSendToken()
          : ISendToken(), m_LastSendTime(INVALID_SYSTEM_TIME), m_IsSendDelay(false), m_SendCount(0)
        {
            for (int i = 0; i < MAX_STATION_NUM; ++i)
            {
                m_LastSendTimeArray[i] = INVALID_SYSTEM_TIME;
                m_SendCountArray[i] = 0;
            }
        }
        virtual bool IsResend() const
        {
            return true;
        }
        SystemTime GetLastSendTime(StationIndex stationIndex) const
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                return m_LastSendTime;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                return m_LastSendTimeArray[stationIndex];
            }
        }
        bool IsSendDelay() const
        {
            return m_IsSendDelay;
        }
        void SetLastSendTime(SystemTime time, StationIndex stationIndex)
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                m_LastSendTime = time;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                m_LastSendTimeArray[stationIndex] = time;
            }
        }
        void EnableSendDelay(bool isSendDelay)
        {
            m_IsSendDelay = isSendDelay;
        }
        // Overridden to return <tt>true</tt> in token classes that resend using unicast.
        virtual bool IsUnicast() const
        {
            return false;
        }
        void SetSendCount(u32 sendCount, StationIndex stationIndex)
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                m_SendCount = sendCount;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                m_SendCountArray[stationIndex] = sendCount;
            }
        }
        void IncrementSendCount(StationIndex stationIndex)
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                ++m_SendCount;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                ++m_SendCountArray[stationIndex];
            }
        }
        u32 GetSendCount(StationIndex stationIndex) const
        {
            if (stationIndex == STATION_INDEX_INVALID)
            {
                return m_SendCount;
            }
            else
            {
                PIA_ASSERT(IsUnicast());
                PIA_ASSERT(stationIndex < MAX_STATION_NUM);
                return m_SendCountArray[stationIndex];
            }
        }

    private:
        SystemTime m_LastSendTime;
        SystemTime m_LastSendTimeArray[MAX_STATION_NUM];
        bool m_IsSendDelay;
        u32 m_SendCount;
        u32 m_SendCountArray[MAX_STATION_NUM];
    };


    class ConversibleResendableSendToken : public ResendableSendToken
    {
    public:
        ConversibleResendableSendToken()
          : ResendableSendToken(), m_IsResend(false)
        {
        }
        virtual bool IsResend() const
        {
            return m_IsResend;
        }
        void SetResend(bool isResend)
        {
            m_IsResend = isResend;
        }

    private:
        bool m_IsResend;
    };

protected:
    void AddSendData(ISendToken* pToken);
    void AddSendDelayData(ResendableSendToken* pToken);
    void CancelSendData(ISendToken* pToken);

protected:
    void ErrorDropEvent();

private:
    // Copying and assignment are prohibited.
    CloneElementBase(const CloneElementBase& rhs);
    CloneElementBase& operator=(const CloneElementBase& rhs);

private:
    common::TreeMapNode<Id> m_TreeMapNode;
    CloneBase* m_pCloneBase;

public:
    template <typename Element, typename CP>
    class DefinitionBase
    {
    protected:
        DefinitionBase()
        {
            CP::GetElementDefinition()->Register(Element::GetTypeStatic(), Element::Receive, Element::GetDataChunkHeaderSize());
        }
    };

    template <typename Element>
    class Definition : public DefinitionBase<Element, CloneProtocol>
    {
    public:
        Definition()
          : DefinitionBase<Element, CloneProtocol>()
        {
        }
    };

private:
    CloneProfilerBase* m_pSendProfiler;
    CloneProfilerBase* m_pReceiveProfiler;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const CloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
