﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: clone_CloneProfiler.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneProfilerBase.h>
#include <pia/assert.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
    @brief  This class manages <tt>PiaClone</tt> profiling.

    @details  It adds a feature that profiles the data each instance sends/receives by registering a <tt>CloneProfiler</tt> instance for the <tt>CloneProtocol</tt>, <tt>CloneBase</tt>, and <tt>CloneElementBase</tt> instances.
                <tt>CloneProtocol</tt> send/receive tasks can be profiled by configuring instance.
    @details  You must use a separate <tt>CloneProfiler</tt> instance for each instance of <tt>CloneProtocol</tt>,
                <tt>CloneBase</tt>, and <tt>CloneElementBase</tt> to be profiled at the same time.

    @tparam STATION_NUM  The number of stations to measure. Normally you specify the largest number of stations participating in the session.
    @tparam LATEST_BUFFER_SIZE  Measures the data sent and received the last <tt>LATEST_BUFFER_SIZE</tt> times the <tt>common::Scheduler::Dispatch()</tt> function was called.
                                Specify <tt>0</tt> if not necessary.

    @see  CloneProtocol::SetProfiler, CloneBase::SetProfiler, CloneElementBase::SetProfiler

    @date 2013-11-25  Initial version.
*/
template <u32 STATION_NUM, u32 LATEST_BUFFER_SIZE>
class CloneProfiler : public CloneProfilerBase
{
public:
/*!
    @brief  Instantiates an object with default parameters (default constructor).
*/
    CloneProfiler()
      : CloneProfilerBase(
            (LATEST_BUFFER_SIZE > 0) ? reinterpret_cast<u32*>(&m_aBuffer[cDispatchCountBufferOffset]) : NULL,
            (LATEST_BUFFER_SIZE > 0) ? &m_aBuffer[cCountBufferOffset] : NULL,
            (LATEST_BUFFER_SIZE > 0) ? reinterpret_cast<u16*>(&m_aBuffer[cSizeBufferOffset]) : NULL,
            reinterpret_cast<u16*>(&m_aBuffer[cTotalCountBufferOffset]),
            reinterpret_cast<u32*>(&m_aBuffer[cTotalSizeBufferOffset]),
            cStationNum,
            LATEST_BUFFER_SIZE)
    {
        PIA_COMPILE_ASSERT(STATION_NUM > 0);
    };


private:
    enum
    {
        cStationNum = STATION_NUM < 32 ? STATION_NUM : 32,
        cDispatchCountBufferSize = sizeof(u32) * LATEST_BUFFER_SIZE,
        cCountBufferSize = sizeof(u8) * LATEST_BUFFER_SIZE * cStationNum,
        cSizeBufferSize = sizeof(u16) * LATEST_BUFFER_SIZE * cStationNum,
        cTotalCountBufferSize = sizeof(u16) * cStationNum,
        cTotalSizeBufferSize = sizeof(u32) * cStationNum,
        cDispatchCountBufferOffset = 0,
        cTotalSizeBufferOffset = cDispatchCountBufferOffset + cDispatchCountBufferSize,
        cTotalCountBufferOffset = cTotalSizeBufferOffset + cTotalSizeBufferSize,
        cSizeBufferOffset = cTotalCountBufferOffset + cTotalCountBufferSize,
        cCountBufferOffset = cSizeBufferOffset + cSizeBufferSize,
        cBufferSize = cCountBufferOffset + cCountBufferSize
    };

    PIA_ATTRIBUTE_ALIGN(4) u8 m_aBuffer[cBufferSize];
};
}
}
}  // end of namespace nn::pia::clone
