﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_CloneProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_PacketDefine.h>
#include <pia/common/common_OffsetTreeMap.h>
#include <pia/common/common_OffsetList.h>
#include <pia/transport/transport_Protocol.h>
#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_SendClone.h>
#include <pia/clone/clone_ReceiveClone.h>
#include <pia/clone/clone_AtomicSharingClone.h>
#include <pia/clone/clone_SequentialSharingClone.h>

namespace nn
{
namespace pia
{
namespace clone
{

class Clock;
class IDataPacker;
class IDataUnpacker;
class CompressDataUnpacker;


/*!
@brief  Represents a protocol for sharing values between stations. <br />

@date 2014-08-01  Changed specifications specific to the arguments and success conditions for the <tt>RegisterReceiveClone</tt>, <tt>UnregisterReceiveClone</tt>, and <tt>SearchReceiveClone</tt> functions.
@date 2014-07-17  Changed specifications specific to success conditions for the <tt>RegisterReceiveClone</tt>, <tt>UnregisterReceiveClone</tt>, and <tt>SearchReceiveClone</tt> functions.
@date 2014-07-17  Changed the names of the <tt>SetDefermentPeriod</tt> and <tt>GetDefermentPeriod</tt> functions to the <tt>SetMaxSendDelay</tt> and <tt>GetMaxSendDelay</tt> functions, respectively.
@date 2014-01-30  Added constants to specify the compression level of synchronization data, and made any other compression levels unsupported.
@date 2013-11-25  Added a profiling feature.
@date 2013-11-19  Corrected the spelling of <tt>Setting::m_RegulationRate</tt>.
@date 2013-07-18  Initial version.
*/
class CloneProtocol : public transport::Protocol
{
public:
    PIA_PROTOCOL_TYPE_INFO(nn::pia::transport::PROTOCOL_TYPE_CLONE);

/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
*/
    CloneProtocol();
    //! @endcond


/*!
@cond PRIVATE
@brief  Destructor.
*/
    virtual ~CloneProtocol();
    //! @endcond


/*!
@brief  Compares the order.
@param[in] clock  Specifies the time of the comparison.
@param[in] stationA  Specifies the <tt>StationIndex</tt> of a station to compare.
@param[in] stationB  Specifies the <tt>StationIndex</tt> of a station to compare.
@return  When the time is <tt>clock</tt>, returns a negative value when <tt><var>stationB</var></tt> is treated as if it occurred before <tt><var>stationA</var></tt>. Returns a positive value when <tt><var>stationB</var></tt> is treated as if it occurred after <tt><var>stationA</var></tt>.
@details  This function is never called when <tt><var>stationA</var></tt> and <tt><var>stationB</var></tt> are the same value.
@details  This function must never give contradictory order relationships at any particular time.
For example, if <tt>ClockPriorityJudgeFunction(clock, stationA, stationB) < 0</tt> is true, <tt>ClockPriorityJudgeFunction(clock, stationB, stationA) > 0</tt> must also be true.
When <tt>ClockPriorityJudgeFunction(clock, stationA, stationB) < 0</tt> and <tt>ClockPriorityJudgeFunction(clock, stationB, stationC) < 0</tt> are both true, <tt>ClockPriorityJudgeFunction(clock, stationA, stationC) < 0</tt> must also be true.
@details  At different times, however, the order relationship does not need to match.
Just because <tt>ClockPriorityJudgeFunction(clock1, stationA, stationB) < 0</tt> is true does not mean it is a problem for <tt>ClockPriorityJudgeFunction(clock2, stationA, stationB) > 0</tt> to also be true.




*/
    typedef int (*ClockPriorityJudgeFunction)(ClockValue clock, StationIndex stationA, StationIndex stationB);


    static const u32 COMPRESSION_LEVEL_NONE = 0;    //!< Specifies the compression level (no compression) for the data to send specified in <tt>@ref Setting::dataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_LOW = 1;     //!< Specifies the compression level (speed emphasized) for the data to send specified in <tt>@ref Setting::dataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_MIDDLE = 5;  //!< Specifies the compression level for the data to send specified in <tt>@ref Setting::dataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_HIGH = 9;    //!< Specifies the compression level (compression rate emphasized) for the data to send specified in <tt>@ref Setting::dataCompressionLevel</tt>.

/*!
@brief  Stores settings specified during initialization.
@details  Specify the same values on stations being communicated with.
*/
    struct Setting
    {
    public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
@details  This function initializes the object using default values.
*/
        Setting()
          : clockType(CLOCK_TYPE_RTC),
            clockPerSec(1000.0f),
            clockPerFrame(1),
            clockRegulationRate(0.1f),
            clockPriorityJudgeMethod(CLOCK_PRIORITY_JUDGE_METHOD_SIMPLE),
            clockPriorityJudgeFunction(NULL),
            dataCompressionLevel(COMPRESSION_LEVEL_NONE)
        {
        }


/*!
@brief  Enumerates the standard used for advancing the clock.
*/
        enum ClockType
        {
            CLOCK_TYPE_RTC = 0,  //!< Specifies that the clock advances based on RTC.
            CLOCK_TYPE_FRAME     //!< Specifies that the clock advances based on the number of times <tt>CloneProtocol::UpdateClock</tt> is called.
        };


/*!
@brief  Specifies the standard to use for advancing the clock.
@details  If <tt>CLOCK_TYPE_RTC</tt> is specified, the clock advances based on RTC.
After a session begins and the clock has been regulated, time differences between systems do not need to be considered. However, the clock may not advance the same amount for each frame depending on when <tt>CloneProtocol::UpdateClock</tt> is called.
Specify the speed at which to advance the clock in <tt>clockPerSec</tt>.
@details  If <tt>CLOCK_TYPE_FRAME</tt> is specified, usually the clock advances a set value each time <tt>CloneProtocol::UpdateClock</tt> is called.
However, if the number of calls to <tt>CloneProtocol::UpdateClock</tt> differs between stations due to lag or other factors, the clock may no longer match between stations.
Specify the speed at which to advance the clock in <tt>clockPerFrame</tt>.
@details  For more information about managing the clock, see the programming manual.
@see  clockPerSec
@see  clockPerFrame


*/
        ClockType clockType;


/*!
@brief  Specifies the value to advance the clock per second.
@details  This setting is ignored if <tt>clockType</tt> is set to something other than <tt>CLOCK_TYPE_RTC</tt>.
@see  clockType
*/
        f32 clockPerSec;


/*!
@brief  Specifies the value to advance the clock per frame.
@details  This setting is ignored if <tt>clockType</tt> is set to something other than <tt>CLOCK_TYPE_FRAME</tt>.
@see  clockType
*/
        u32 clockPerFrame;


/*!
@brief  Specifies the rate at which to speed up or slow down the clock when re-regulating the clock.
@details  When the clock is re-regulated using <tt>CloneProtocol::RegulateClock</tt>, it advances faster or slower than usual at the rate specified here.
For example, if <tt>0.1</tt> is specified and the station clock is behind the clocks of other stations, the clock advances at a speed of 110%.
On the other hand, if the clock is ahead of the other stations, it advances at a speed of 90%.
*/
        f32 clockRegulationRate;


/*!
@brief  Enumerates the method for determining which station was first when an event occurs at the time clock value.
*/
        enum ClockPriorityJudgeMethod
        {
            CLOCK_PRIORITY_JUDGE_METHOD_SIMPLE = 0,  //!< Specifies that the station with a smaller station ID value was first.
            CLOCK_PRIORITY_JUDGE_METHOD_IMPARTIAL,   //!< Specifies that the priority is changed based on the clock value so that no single station has an advantage over the long term.
            CLOCK_PRIORITY_JUDGE_METHOD_USER_DEFINE  //!< Specifies that the application defines the method to use.
        };


/*!
@brief  Specifies the method for determining which station was first when an event occurs at the time clock value.
@details  If <tt>CLOCK_PRIORITY_JUDGE_METHOD_USER_DEFINE</tt> is specified, specify the function to use for determining the order in <tt>clockPriorityJudgeFunction</tt>.
@see  clockPriorityJudgeFunction
*/
        ClockPriorityJudgeMethod clockPriorityJudgeMethod;


/*!
@brief  Specifies the function for determining which station was first when an event occurs at the time clock value.
@details  This setting is ignored if <tt>clockPriorityJudgeMethod</tt> is set to something other than <tt>CLOCK_PRIORITY_JUDGE_METHOD_USER_DEFINE</tt>.
@see  clockPriorityJudgeMethod;
*/
        ClockPriorityJudgeFunction clockPriorityJudgeFunction;


/*!
@brief  Specifies the compression level for data to send.
@details  Not compressed when <tt>@ref CloneProtocol::COMPRESSION_LEVEL_NONE</tt> is specified.
Specifying <tt>@ref CloneProtocol::COMPRESSION_LEVEL_LOW</tt> emphasizes speed; <tt>@ref CloneProtocol::COMPRESSION_LEVEL_HIGH</tt> emphasizes the compression rate.
Specifying <tt>@ref CloneProtocol::COMPRESSION_LEVEL_MIDDLE</tt> results in a level of compression in between the other two.
The default value is <tt>@ref CloneProtocol::COMPRESSION_LEVEL_NONE</tt>.
Other values are not supported.
Execution time varies significantly based on the CPU performance of the platform.
*/
        u32 dataCompressionLevel;
    };


/*!
@brief  Initialization. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.
@param[in] setting  Specifies various settings.
Specify the same values on all stations. Operations are undefined when different values are specified.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@see  Finalize
*/
    nn::Result Initialize(const Setting& setting);


/*!
@brief  Finalizes the class instance.
@see  Initialize
*/
    void Finalize();


/*!
@brief  Gets the maximum value that can be retrieved from the <tt>SerializePolicy::GetSize</tt> function of the clone element to be used.
@details  If a clone with a registered clone element larger than this value is passed to the <tt>RegisterXXXClone</tt> function, <tt>@ref ResultInvalidArgument</tt> is returned.
@details  This value is the maximum size that can fit in a single packet.
The value differs depending on the communication module being used and the MTU setting.
This value can be retrieved after the <tt>Initialize</tt> function has been called.
@return  Returns the maximum value that can be retrieved from the <tt>SerializePolicy::GetSize</tt> function of the clone element to be used.
*/
    size_t GetElementSizeMax() const
    {
        return m_ElementSizeMax;
    }


/*!
@brief  Registers a <tt>SendClone</tt> to the specified ID.
@param[in] pClone  Specifies the <tt>SendClone</tt> to register.
@param[in] cloneId  Specifies the ID to register to.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the object is already registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result RegisterSendClone(SendClone* pClone, CloneBase::Id cloneId)
    {
        return RegisterClone(SendClone::ToKey(cloneId), pClone);
    }


/*!
@brief  Unregisters a <tt>SendClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Therefore, even if this function succeeds, <tt><var>pClone</var></tt> cannot be reused or re-registered to the same ID until <tt>pClone->@ref SendClone::IsRegisteredWithProtocol "IsRegisteredWithProtocol"</tt> returns <tt>false</tt>.
@param[in] pClone  Specifies the <tt>SendClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. The specified <tt>SendClone</tt> is not registered to this <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterSendClone(SendClone* pClone)
    {
        return UnregisterClone(pClone);
    }


/*!
@brief  Unregisters a <tt>SendClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>SendClone</var></tt> cannot be reused or reregistered to the same ID until <tt>SendClone::IsRegisteredWithProtocol</tt> returns <tt>false</tt>.
@param[in] cloneId  Specifies the ID of the <tt>SendClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultNotSet  Indicates that the object is not registered to the specified ID. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterSendClone(CloneBase::Id cloneId)
    {
        return UnregisterClone(SendClone::ToKey(cloneId));
    }


/*!
@brief  Searches for and gets the <tt>SendClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>SendClone</tt> to search for.
@return  Returns the <tt>SendClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    SendClone* SearchSendClone(CloneBase::Id cloneId)
    {
        return static_cast<SendClone*>(m_CloneCollection.Find(SendClone::ToKey(cloneId)));
    }


/*!
@brief  Searches for and gets the <tt>SendClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>SendClone</tt> to search for.
@return  Returns the <tt>SendClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    const SendClone* SearchSendClone(CloneBase::Id cloneId) const
    {
        return static_cast<const SendClone*>(m_CloneCollection.Find(SendClone::ToKey(cloneId)));
    }


/*!
@brief  Registers a <tt>ReceiveClone</tt> to the specified ID.
@param[in] pClone  Specifies the <tt>ReceiveClone</tt> to register.
@param[in] sendStationIndex  Specifies the <tt>@ref StationIndex</tt> of the sender with the <tt>SendClone</tt>.
@param[in] cloneId  Specifies the ID to register to.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the object is already registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result RegisterReceiveClone(ReceiveClone* pClone, StationIndex sendStationIndex, CloneBase::Id cloneId);


/*!
@brief  Unregisters a <tt>ReceiveClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>pClone</var></tt> cannot be reused or re-registered to the same ID until <tt>pClone->@ref ReceiveClone::IsRegisteredWithProtocol "IsRegisteredWithProtocol"</tt> returns <tt>false</tt>.
(<tt>ReceiveClone</tt> may also send for other reasons, for example when serving as a proxy when a station is improperly disconnected.)
@param[in] pClone  Specifies the <tt>ReceiveClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. The specified <tt>ReceiveClone</tt> is not registered to this <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterReceiveClone(ReceiveClone* pClone)
    {
        return UnregisterClone(pClone);
    }


/*!
@brief  Unregisters a <tt>ReceiveClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>ReceiveClone</var></tt> cannot be reused or reregistered to the same ID until <tt>ReceiveClone::IsRegisteredWithProtocol</tt> returns <tt>false</tt>.
(<tt>ReceiveClone</tt> may also send for other reasons, for example when serving as a proxy when a station is improperly disconnected.)
@param[in] sendStationIndex  Specifies the <tt>@ref StationIndex</tt> of the sender with the <tt>SendClone</tt>.
@param[in] cloneId  Specifies the ID of the <tt>ReceiveClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotSet  Indicates that the object is not registered to the specified ID. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterReceiveClone(StationIndex sendStationIndex, CloneBase::Id cloneId);


/*!
@brief  Searches for and gets the <tt>ReceiveClone</tt> object with the specified ID.
@param[in] sendStationIndex  Specifies the <tt>@ref StationIndex</tt> of the sender with the <tt>SendClone</tt>.
@param[in] cloneId  Specifies the ID of the <tt>ReceiveClone</tt> to search for.
@return  Returns the <tt>ReceiveClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    ReceiveClone* SearchReceiveClone(StationIndex sendStationIndex, CloneBase::Id cloneId);


/*!
@brief  Searches for and gets the <tt>ReceiveClone</tt> object with the specified ID.
@param[in] sendStationIndex  Specifies the <tt>@ref StationIndex</tt> of the sender with the <tt>SendClone</tt>.
@param[in] cloneId  Specifies the ID of the <tt>ReceiveClone</tt> to search for.
@return  Returns the <tt>ReceiveClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    const ReceiveClone* SearchReceiveClone(StationIndex sendStationIndex, CloneBase::Id cloneId) const;


/*!
@brief  Registers a <tt>AtomicSharingClone</tt> to the specified ID.
@param[in] pClone  Specifies the <tt>AtomicSharingClone</tt> to register.
@param[in] cloneId  Specifies the ID to register to.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the object is already registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result RegisterAtomicSharingClone(AtomicSharingClone* pClone, CloneBase::Id cloneId)
    {
        return RegisterClone(AtomicSharingClone::ToKey(cloneId), pClone);
    }


/*!
@brief  Unregisters a <tt>AtomicSharingClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>pClone</var></tt> cannot be reused or re-registered to the same ID until <tt>pClone->@ref AtomicSharingClone::IsRegisteredWithProtocol "IsRegisteredWithProtocol"</tt> returns <tt>false</tt>.
(The object may send without acquiring a lock for other reasons, for example when serving as a proxy when a station is improperly disconnected.)
@param[in] pClone  Specifies the <tt>AtomicSharingClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. The specified <tt>AtomicSharingClone</tt> is not registered to this <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterAtomicSharingClone(AtomicSharingClone* pClone)
    {
        return UnregisterClone(pClone);
    }


/*!
@brief  Unregisters a <tt>AtomicSharingClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>AtomicSharingClone</var></tt> cannot be reused or reregistered to the same ID until <tt>AtomicSharingClone::IsRegisteredWithProtocol</tt> returns <tt>false</tt>.
(The object may send without acquiring a lock for other reasons, for example when serving as a proxy when a station is improperly disconnected.)
@param[in] cloneId  Specifies the ID of the <tt>AtomicSharingClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultNotSet  Indicates that the object is not registered to the specified ID. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterAtomicSharingClone(CloneBase::Id cloneId)
    {
        return UnregisterClone(AtomicSharingClone::ToKey(cloneId));
    }


/*!
@brief  Searches for and gets the <tt>AtomicSharingClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>AtomicSharingClone</tt> to search for.
@return  Returns the <tt>AtomicSharingClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    AtomicSharingClone* SearchAtomicSharingClone(CloneBase::Id cloneId)
    {
        return static_cast<AtomicSharingClone*>(m_CloneCollection.Find(AtomicSharingClone::ToKey(cloneId)));
    }


/*!
@brief  Searches for and gets the <tt>AtomicSharingClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>AtomicSharingClone</tt> to search for.
@return  Returns the <tt>AtomicSharingClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    const AtomicSharingClone* SearchAtomicSharingClone(CloneBase::Id cloneId) const
    {
        return static_cast<const AtomicSharingClone*>(m_CloneCollection.Find(AtomicSharingClone::ToKey(cloneId)));
    }


/*!
@brief  Registers a <tt>SequentialSharingClone</tt> to the specified ID.
@param[in] pClone  Specifies the <tt>SequentialSharingClone</tt> to register.
@param[in] cloneId  Specifies the ID to register to.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the object is already registered to the specified ID. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result RegisterSequentialSharingClone(SequentialSharingClone* pClone, CloneBase::Id cloneId)
    {
        return RegisterClone(SequentialSharingClone::ToKey(cloneId), pClone);
    }


/*!
@brief  Unregisters a <tt>SequentialSharingClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>pClone</var></tt> cannot be reused or re-registered to the same ID until <tt>pClone->@ref SequentialSharingClone::IsRegisteredWithProtocol "IsRegisteredWithProtocol"</tt> returns <tt>false</tt>.
@param[in] pClone  Specifies the <tt>SequentialSharingClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. The specified <tt>AtomicSharingClone</tt> is not registered to this <tt>CloneProtocol</tt>. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterSequentialSharingClone(SequentialSharingClone* pClone)
    {
        return UnregisterClone(pClone);
    }


/*!
@brief  Unregisters a <tt>SequentialSharingClone</tt>.
@details  If a send is in progress, this function waits until the send is complete before unregistering.
Even if this function succeeds, <tt><var>SequentialSharingClone</var></tt> cannot be reused or reregistered to the same ID until <tt>SequentialSharingClone::IsRegisteredWithProtocol</tt> returns <tt>false</tt>.
@param[in] cloneId  Specifies the ID of the <tt>SequentialSharingClone</tt> to unregister.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultNotSet  Indicates that the object is not registered to the specified ID. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result UnregisterSequentialSharingClone(CloneBase::Id cloneId)
    {
        return UnregisterClone(SequentialSharingClone::ToKey(cloneId));
    }


/*!
@brief  Searches for and gets the <tt>SequentialSharingClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>SequentialSharingClone</tt> to search for.
@return  Returns the <tt>SequentialSharingClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    SequentialSharingClone* SearchSequentialSharingClone(CloneBase::Id cloneId)
    {
        return static_cast<SequentialSharingClone*>(m_CloneCollection.Find(SequentialSharingClone::ToKey(cloneId)));
    }


/*!
@brief  Searches for and gets the <tt>SequentialSharingClone</tt> object with the specified ID.
@param[in] cloneId  Specifies the ID of the <tt>SequentialSharingClone</tt> to search for.
@return  Returns the <tt>SequentialSharingClone</tt> with the specified ID. Returns <tt>NULL</tt> when an object with the specified ID cannot be found.
*/
    const SequentialSharingClone* SearchSequentialSharingClone(CloneBase::Id cloneId) const
    {
        return static_cast<const SequentialSharingClone*>(m_CloneCollection.Find(SequentialSharingClone::ToKey(cloneId)));
    }


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] localStationIndex  Specifies the <tt>StationIndex</tt> of the local station. Specify a value in the range from <tt>STATION_INDEX_1</tt> to <tt><var>maxConnections</var></tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Startup(StationIndex localStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  The main process.

@return  Returns <tt>nn::ResultSuccess</tt>.
*/
    virtual nn::Result Dispatch();
    //! @endcond


/*!
@cond PRIVATE
@brief  Switches the connection status.

@param[in] event  Specifies the event representing the operation.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the function was called in an invalid state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result UpdateProtocolEvent(const transport::ProtocolEvent& event);
    //! @endcond


/*!
@brief  Enumerates the operational states of <tt>CloneProtocol</tt>.
*/
    enum State
    {
        STATE_INACTIVE = 0x10,  //!< Specifies inactive.
        STATE_START = 0x20,     //!< Specifies starting up.
        STATE_ACTIVE = 0x30,    //!< Specifies active.
        STATE_STOP = 0x40       //!< Specifies stopping.
    };


/*!
@brief  Gets the current operational state.
@return  Returns the current operational state.
*/
    State GetState() const
    {
        return static_cast<State>(m_InnerState & cStateMask);
    }


/*!
@brief  Determines whether the <tt>CloneProtocol</tt> of the specified station is active.
@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station to be checked if active.
@return  Returns <tt>true</tt> if the specified station is active.
*/
    bool IsActiveStation(StationId stationId) const;


/*!
@brief  Starts operation of <tt>CloneProtocol</tt>.
@details  This function can be called when the operational state is <tt>@ref STATE_INACTIVE</tt> after communication starts.
When this function succeeds, the object state changes to <tt>@ref STATE_START</tt> and begins the startup process, after which it transitions to <tt>@ref STATE_ACTIVE</tt>.
After calling this function, you must periodically call the <tt>UpdateClock</tt> function.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that the station is not communicating or the state is not <tt>@ref nn::pia::clone::CloneProtocol::STATE_INACTIVE "STATE_INACTIVE"</tt>. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Start();


/*!
@brief  Stops operation of <tt>CloneProtocol</tt>.
@details  When this function is called while the operational state is <tt>@ref STATE_START</tt> or <tt>@ref STATE_ACTIVE</tt>, the state changes to <tt>@ref STATE_STOP</tt> and then to <tt>@ref STATE_INACTIVE</tt> after finalization is complete.
You must continue to call the <tt>UpdateClock</tt> function while the state is <tt>@ref STATE_STOP</tt>.
This function does nothing if called while the state is <tt>@ref STATE_INACTIVE</tt> or <tt>@ref STATE_STOP</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that communication is not active. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Stop();


/*!
@brief  Advances the clock by one frame.
@details  The clock managed by <tt>CloneProtocol</tt> only advances when this function is called.
Be sure to call this function periodically while the operational state is either <tt>@ref STATE_START</tt>, <tt>@ref STATE_ACTIVE</tt>, or <tt>@ref STATE_STOP</tt>.
Set the amount each call advances the clock in the <tt>Setting</tt> object that is specified when calling the <tt>Initialize</tt> function.
If the interval between calls is shorter than the precision of the clock value, the clock may stay the same but it never goes backward.
@see  GetClock
*/
    void UpdateClock();


/*!
@brief  Gets the current clock.
@return  Returns the current clock. Returns <tt>@ref INVALID_CLOCK</tt> if a valid clock does not exist because, for example, the object is not active.
@see  UpdateClock
*/
    ClockValue GetClock() const;


/*!
@brief  Starts re-regulating the clock.
@details  The clock managed by the <tt>CloneProtocol</tt> for this station is re-adjusted to be in line with the average clock of the other stations.
Discrepancies are fixed by advancing the clock at a faster or slower rate than usual after examining the clocks of the other stations.
The rate itself is set in the <tt>Setting::clockRegulationRate</tt> member of the <tt>Setting</tt> object that is specified when the <tt>Initialize</tt> function is called.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidState  Indicates that the state is not <tt>@ref nn::pia::clone::CloneProtocol::STATE_ACTIVE "STATE_ACTIVE"</tt> or the clock is already being regulated. Programming error. Fix your program so that this error is not returned.
@see  IsRegulatingClock
*/
    nn::Result RegulateClock();


/*!
@brief  Gets whether the clock is being regulated.
@return  Returns <tt>true</tt> if the clock is being regulated..
@see  RegulateClock
*/
    bool IsRegulatingClock() const;


/*!
@brief  (For debugging.) Sets the time until the clock value overflows.

@details  Sets the time from when synchronous communication starts until the clock value overflows.
@details  This function is for debugging. Do not include it in retail products.

@param[in] span  If the <tt>@ref Setting::ClockType</tt> is <tt>@ref Setting::CLOCK_TYPE_RTC</tt>, this is the time (in seconds). If it is <tt>@ref Setting::CLOCK_TYPE_FRAME</tt>, this is the number of frames.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetClockOverFlowSpanForDebug(u32 span);


/*!
@brief  Enumerates the error state.
*/
    enum ErrorType
    {
        ERROR_TYPE_NOTHING = 0,    //!< Specifies no error.
        ERROR_TYPE_DROP_EVENT,     //!< Specifies an inconsistency in event management. Some event notifications may not have been sent.
                                   //!< This error can occur when <tt>EventCloneElement</tt> is used with <tt>AtomicSharingClone</tt>.
        ERROR_TYPE_CLOCK_OVERFLOW  //!< Specifies the clock value overflowed. Widespread inconsistencies may be occurring.
    };


/*!
@brief  Gets the error state.
@details  When an error occurs, the values being sent or received may be invalid. We recommend stopping the match and ending the session with the <tt>Stop</tt> function.
This value is reset when the <tt>Start</tt> function is called again.
When multiple error types occur, the one with the highest <tt>@ref ErrorType</tt> takes precedence.
@return  Returns the error state.

*/
    ErrorType GetError() const
    {
        return m_Error;
    }


/*!
@brief  Sets the maximum time to defer the sending of data.
@details  In some clone elements, you can specify whether to send data immediately after it is set, or to wait until there is other data to send.
Even if you defer the sending, the data is sent by itself when the time specified here elapses.
Deferring the sending of data can be expected to reduce the number of packets sent, but it takes longer for the data to arrive after it is set.
@param[in] maxSendDelay  Specifies the maximum time to defer sending. The value is in milliseconds. The maximum value that can be specified is <tt>10000</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@see  GetMaxSendDelay
*/
    nn::Result SetMaxSendDelay(u16 maxSendDelay);


/*!
@brief  Gets the maximum time to defer the sending of data.
@return  Returns the maximum time to defer the sending of data. The value is in milliseconds.
@see  SetMaxSendDelay
*/
    u16 GetMaxSendDelay() const
    {
        return m_MaxSendDelay;
    }


/*!
@brief  Specifies the ratio of the resend interval for the base RTT obtained from profiling.
@details  The default value of the ratio is 1.25. The resend interval is the value of (RTT * <tt><var>ratio</var></tt>). The maximum value of the resend interval is 2500 milliseconds.
@param[in] ratio  The ratio of the resend interval for the base RTT obtained from profiling Specify a value larger than <tt>0</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetResendSpanRatio(f32 ratio);


    //! @name  Profiling
    //! @{

/*!
@brief  Sets an object to manage send/receive profiling.
@details  <tt>CloneProtocol</tt> send/receive tasks can be profiled by configuring <tt>CloneProfiler</tt>.
This calculation includes the data <tt>CloneProtocol</tt> actually sends through and receives from the communication pathway.
In clone profiling, the data the local station sends and receives is included in the calculation, but this data is resolved in <tt>CloneProtocol</tt> and is not included in the <tt>CloneProtocol</tt> profiling calculations.
@details  When compression is enabled, the compressed size is used in the calculation.
@details  A <tt>CloneProfiler</tt> instance that is already configured can be released by specifying <tt>NULL</tt> in the arguments.
@param[in] pSendProfiler  Specifies the <tt>CloneProfiler</tt> that manages the send profiling.
Specify <tt>NULL</tt> if not necessary.
@param[in] pReceiveProfiler  Specifies the <tt>CloneProfiler</tt> that manages the receive profiling.
Specify <tt>NULL</tt> if not necessary.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result SetProfiler(CloneProfilerBase* pSendProfiler, CloneProfilerBase* pReceiveProfiler);


/*!
@brief  Gets the object configured to manage send profiling.
@return  The object configured to manage send profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetSendProfiler() const
    {
        return m_pSendProfiler;
    }


/*!
@brief  Gets the object configured to manage receive profiling.
@return  The object configured to manage receive profiling. Returns <tt>NULL</tt> if nothing has been set.
@see  SetProfiler
*/
    const CloneProfilerBase* GetReceiveProfiler() const
    {
        return m_pReceiveProfiler;
    }


/*!
@brief  Resets the results of send/receive profiling.
@see  SetProfiler
*/
    void ResetProfiler();


/*!
@brief  Gets the send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The send count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, sent during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, sent since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage send profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalSendSize(StationIndex stationIndex) const;


/*!
@brief  Gets the receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The receive count for the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times <tt>common::Scheduler::Dispatch</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The total amount of data, in bytes, received during the last <tt>@ref CloneProfiler "CloneProfiler::LATEST_BUFFER_SIZE"</tt> times of <tt>common::Scheduler::Dispatch</tt> being called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler
*/
    u32 GetProfiledLatestReceiveSize(StationIndex stationIndex) const;


/*!
@brief  Gets the number of times received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The number of times received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveCount(StationIndex stationIndex) const;


/*!
@brief  Gets the amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
@param[in] stationIndex  The <tt>StationIndex</tt> that is subject to profiling.
@return  The amount of data, in bytes, received since <tt>ResetProfiler</tt> was called.
Returns <tt>0</tt> if no object is configured to manage receive profiling.
@see  SetProfiler, ResetProfiler
*/
    u32 GetProfiledTotalReceiveSize(StationIndex stationIndex) const;

    //! @}


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


public:
/*!
@cond PRIVATE
@brief  Gets the resend interval. (For debugging.)
*/
    SystemTime GetResendSpan() const
    {
        return m_ResendSpan;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the send interval used in unicasts. (For debugging.)
*/
    SystemTime GetResendSpanForUnicast(StationIndex stationIndex) const
    {
        return m_ResendSpanForUnicast[stationIndex];
    }
    //! @endcond

    u32 GetStationNum() const
    {
        return m_StationNum;
    }

    //! @cond PRIVATE

private:
    static int SimpleClockPriorityJudgeFunction(ClockValue clock, StationIndex stationA, StationIndex stationB);
    static int ImpartialClockPriorityJudgeFunction(ClockValue clock, StationIndex stationA, StationIndex stationB);

public:
    StationIndex GetLocalStationIndex() const
    {
        return m_LocalStationIndex;
    }
    u32 GetParticipantBitmap() const
    {
        return m_ParticipantBitmap;
    }
    SystemTime GetSystemTime() const
    {
        return m_SystemTime;
    }
    u32 GetDispatchCount() const
    {
        return m_DispatchCount;
    }

    void ErrorDropEvent()
    {
        SetError(ERROR_TYPE_DROP_EVENT);
    }

    // n2920: Added for <tt>CloneBase</tt> to reference.
    u32 GetValidStationBitmap() const
    {
        return m_ValidStationBitmap;
    }

    u32 GetNoticeParticipationBitmap() const
    {
        return m_NoticeParticipationBitmap;
    }

private:
    bool IsInitialized() const
    {
        return (m_StationNum > 0);
    }
    bool IsInCommunication() const
    {
        return m_LocalStationIndex != STATION_INDEX_INVALID;
    }

    enum InnerState
    {
        eISInactive = STATE_INACTIVE | 0x1,
        eISSyncClock = STATE_START | 0x1,
        eISNoticeStart = STATE_START | 0x2,
        eISActive = STATE_ACTIVE | 0x1,
        eISComplete = STATE_STOP | 0x1,
        eISStop = STATE_STOP | 0x2
    };
    InnerState GetInnerState() const
    {
        return m_InnerState;
    }

    void UpdateCloneState();
    void UpdateState();
    void UpdateResendSpan();

    void ReceiveMessages();
    void ReceiveMessage(StationIndex src, const void* cpMessage, size_t size);
    bool UpdateReceivedMessageSequentialId(StationIndex src, const void* cpMessage);

    bool SendUnicastMessage(StationIndex dest, void* pMessage, size_t messageSize);
    bool SendMulticastMessage(u32 destBitmap, void* pMessage, size_t messageSize);
    void SetSequentialIdToMessage(void* pMessage);

    bool SendUnicastCloneCommand(StationIndex dest, CloneBase::Type cloneType, StationIndex cloneStationIndex, CloneBase::Id cloneId, u8 command);
    bool SendUnicastClockCloneCommand(StationIndex dest, CloneBase::Type cloneType, StationIndex cloneStationIndex, CloneBase::Id cloneId, u8 command, ClockValue clock);
    bool SendUnicastClockAndCountCloneCommand(StationIndex dest, CloneBase::Type cloneType, StationIndex cloneStationIndex, CloneBase::Id cloneId, u8 command, ClockValue clock, u8 count);
    bool SendUnicastClockAndCountAndParticipantCloneCommand(StationIndex dest, CloneBase::Type cloneType, StationIndex cloneStationIndex, CloneBase::Id cloneId, u8 command, ClockValue clock, u8 count, u32 participantBitmap);

    void UpdateSystemTime();

    void SendClockRequest();
    void ReceiveClockRequest(StationIndex src, const void* cpMessage, size_t size);
    void ReceiveClockReplyMessage(StationIndex src, const void* cpMessage, size_t size);

    void NoticeParticipation();
    void ReceiveEnterMessage(StationIndex src, const void* cpMessage, size_t size);
    void ReceiveExitMessage(StationIndex src, const void* cpMessage, size_t size);
    void ReceiveEnterAckMessage(StationIndex src, const void* cpMessage, size_t size);
    void ReceiveExitAckMessage(StationIndex src, const void* cpMessage, size_t size);

    struct DispatchCountData
    {
    public:
        void Clear()
        {
            m_MyCount = cInvalidDispatchCount;
            m_ReceivedCount = 0;
        }
        u32 m_MyCount;
        u16 m_ReceivedCount;
    };

private:
    size_t m_ElementSizeMax;

    u32 m_StationNum;
    StationIndex m_LocalStationIndex;
    u32 m_ValidStationBitmap; // Bitmap to directly link to CONNECT and DISCONNECT.
    u32 m_ParticipantBitmap; // Consider the bitmap this way: <tt>CloneProtocol</tt> joins when it receives <tt>eMTEnter</tt> and does not join when it receives <tt>eMTExit</tt>.
    InnerState m_InnerState;

    Clock* m_pClock;

    SystemTime m_SystemTime;
    SystemTime m_ResendSpan;
    SystemTime m_ResendSpanForUnicast[MAX_STATION_NUM];
    f32 m_ResendSpanRatio;
    SystemTime m_LastUpdateResendSpanTime;
    SystemTime m_LastParticipationCommandTime;
    SystemTime m_LastClockCommandTime;

    SystemTime m_MaxSendDelay;

    static const SystemTime cDefaultResendSpan = 500;

    u32 m_DispatchCount;
    DispatchCountData* m_paReceivedDispatchCount;
    static const u32 cCheckDispatchCount = 0x00003fff;

public:
    static const u32 cInvalidDispatchCount = 0xffffffff;

private:
    enum
    {
        cStateMask = 0xf0
    };

    u32 m_NoticeParticipationBitmap; // The bitmap that Ack by peers checks when the local station calls <tt>Start()</tt> and <tt>Stop()</tt> on <tt>CloneProtocol</tt>.

private:
    nn::Result RegisterClone(CloneBase::Key key, CloneBase* pClone);
    nn::Result UnregisterClone(CloneBase::Key key);
    nn::Result UnregisterClone(CloneBase* pClone);

    typedef common::OffsetTreeMap<CloneBase::Key, CloneBase> CloneCollection;
    CloneCollection m_CloneCollection;


public:
    void AddSendCommand(CloneBase::CommandToken* pToken);
    void CancelSendCommand(CloneBase::CommandToken* pToken);

private:
    void SendCommand();
    u32 SendCommandCore(CloneBase::CommandToken* pToken, bool isUnicast, u32 destBitmap, void* pMessage, size_t messageSize);

    typedef common::OffsetList<CloneBase::CommandToken> CommandTokenList;
    CommandTokenList m_SendCommandTokenList;
    CommandTokenList m_ResendCommandTokenList;

public:
    void AddSendData(CloneBase::DataToken* pToken);
    void CancelSendData(CloneBase::DataToken* pToken);

private:
    void SendData();
    bool SendCloneData(CloneBase::DataToken* pToken);
    void ReceiveData(StationIndex src, CloneBase* pClone, const void* cpMessage, size_t messageSize);

    typedef common::OffsetList<CloneBase::DataToken> DataTokenList;
    DataTokenList m_DataTokenList;

    struct DataMessageBuffer
    {
    public:
        void Clear()
        {
            m_DestBitmap = 0;
            m_Size = 0;
        }

        u32 m_DestBitmap;
        u16 m_Size;
        u16 m_Dummy;
        u8 m_Message[common::PROTOCOL_MESSGAE_PAYLOAD_SIZE_MAX];
    };

    DataMessageBuffer m_DataMessageBuffer;
    IDataPacker* m_pDataPacker;
    IDataUnpacker* m_pDataUnpacker;
    CompressDataUnpacker* m_pCompressedDataUnpacker;

public:
    ClockPriorityJudgeFunction GetClockPriorityJudgeFunction() const
    {
        return m_ClockPriorityJudgeFunction;
    }

private:
    ClockPriorityJudgeFunction m_ClockPriorityJudgeFunction;

private:
    void SetError(ErrorType errorType);

    ErrorType m_Error;

private:
    CloneProfilerBase* m_pSendProfiler;
    CloneProfilerBase* m_pReceiveProfiler;

    u32 m_SendMessageSequentialId;
    u32 m_ReceivedMessageSequentialId[MAX_STATION_NUM];
#if NN_PIA_ENABLE_TRACE
    // If the same <tt>SequentialId</tt> is received twice in a row, the first is for message information tracing.
    struct MessageHeaderForDebug
    {
        u8 m_FormatVersion;
        u8 m_MessageType;
        u16 m_DispatchCount;
    };
    MessageHeaderForDebug m_ReceivedMessageHeaderForDebug[MAX_STATION_NUM];
#endif

public:
    typedef bool (*ReceiveElementFunction)(CloneElementBase* pElement, IDataPacker* pAckPacker, const void* cpChunk, u16 chunkSize, CloneElementBase::Id elementId, StationIndex src, StationIndex localStationIndex, u32 dispatchCount);

    class ElementDefinition
    {
    public:
        ElementDefinition();
        void Register(u16 type, ReceiveElementFunction receiveFunction, size_t dataChunkHeaderSize);
        ReceiveElementFunction GetReceiveFunction(u16 type) const;
        size_t GetDataChunkHeaderSizeMax() const
        {
            return m_DataChunkHeaderSizeMax;
        }
        static const u32 cTypeNum = 16;

    private:
        ReceiveElementFunction m_aReceiveFunction[cTypeNum];
        size_t m_DataChunkHeaderSizeMax;
    };

    static ElementDefinition* GetElementDefinition();


protected:
    virtual Clock* CreateRtcClock(f32 clockPerSec, f32 regulationRate);
    virtual size_t TransportGetMaxStationNum();
    virtual size_t PayloadSizeManagerGetMtuSize();
    virtual size_t SignatureManagerGetSignatureSizeMax();
    virtual const common::Time& TransportGetDispatchedTime();
    virtual s32 StationGetRtt(StationIndex stationId);
    virtual s32 StationGetRttPulseInterval();
    virtual size_t PacketHandlerGetPayloadSizeLimit();
    virtual bool PacketHandlerSendMessageById(StationIndex dest, const void* cpData, size_t size);
    virtual bool PacketHandlerSendMessageByBitmap(u32 destBitmap, const void* cpData, size_t size);
    virtual void PacketHandlerReadIterationBegin(void** ppIterator);
    virtual bool PacketHandlerReadIterationIsEnd(const void* cpIterator);
    virtual void PacketHandlerReadIterationNext(void* pIterator);
    virtual void PacketHandlerReadIteratorGet(StationIndex* pSrc, const void** cppPayloadPtr, size_t* pPayloadSize, const void* cpIterator);

public:
    template <int TEST_ID, typename Arg>
    static void Test(const CloneProtocol& obj, Arg* pArg = NULL);

private:
    // n2920: For the client-monitoring feature.
    u32 m_UpdateCount;
    u32 m_ResendCount;
    nn::pia::common::Time m_ActivationStartTime;
    u32 m_CompCount;
    f32 m_CompRateTotal;
    f32 m_CompAverage;
    u32 m_SendCount;
    u32 m_SendDataTotal;

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
