﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_EventCloneElement.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_EventCloneElementBase.h>
#include <pia/clone/clone_SerializePolicyDefinition.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  Manages the sending and receiving of events.

@tparam Value_  Specifies the value type to specify when raising an event. You must specify a value even if none is required.
@tparam BUFFER_SIZE  Specifies the size of the event buffer. Specify a positive value.
@tparam SerializePolicy_  Specifies the algorithm to use for serializing \a Value.

@date 2013-07-18  Initial version.
*/
template <typename Value_, size_t BUFFER_SIZE, typename SerializePolicy_ = HostByteOrderSerializePolicy<Value_> >
class EventCloneElement : public EventCloneElementBase
{
public:
/*!
@brief  Specifies the argument type to be sent with this event.
*/
    typedef Value_ Value;

/*!
@brief  Specifies the algorithm to use for serializing \a Value.
*/
    typedef SerializePolicy_ SerializePolicy;


/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    EventCloneElement();


/*!
@brief  Destroys the object.
*/
    virtual ~EventCloneElement()
    {
    }


/*!
@brief  Raises an event.
@param[in] value  The event's value.
@param[in] isEnableSendDelay  Specifying <tt>true</tt> defers the sending of data. The data is then sent along with other immediate-send items among the elements registered to the same clone.
The maximum deferment time can be specified using the <tt>CloneProtocol::SetMaxSendDelay</tt> function. After that time elapses, the data is sent even if there is no other data to send.
This feature can reduce the number of packets sent, but also delays the arrival of packets by the deferment time.
Specify <tt>false</tt> to send the data immediately.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Specifies that the registered clone cannot send in its current state. Programming error. Fix your program so that this error is not returned.
@retval ResultBufferIsFull  Specifies that the event cannot be raised because the buffer is full. Handle appropriately in the application.
*/
    nn::Result SetValue(const Value& value, bool isEnableSendDelay = false);


/*!
@brief  References the next event.
@details  The event currently being referenced can be accessed until the <tt>HandleNext</tt> function is called again.
@return  Returns the value for the newly referenced event. Returns <tt>NULL</tt> when there is no next event.
*/
    const Value* HandleNext()
    {
        HandleNextCore();
        return GetValue();
    }


/*!
@brief  Gets the pointer to the value for when the current event was raised.
@return  Returns the pointer to the value for when the current event was raised. The return value is the same as the value returned when <tt>HandleNext</tt> was last called. Returns <tt>NULL</tt> when no event is being referenced.
*/
    const Value* GetValue() const
    {
        return (GetHandlingEventToken() != NULL) ? &(static_cast<const EventToken*>(GetHandlingEventToken())->GetValue()) : NULL;
    }


/*!
@brief  Gets the pointer to the issued time for when the current event was raised.
@return  Returns the pointer to the issued time for when the current event was raised. Returns <tt>NULL</tt> when no event is being referenced.
*/
    const ClockValue* GetClock() const
    {
        return (GetHandlingEventToken() != NULL) ? (&GetHandlingEventToken()->GetClock()) : NULL;
    }


/*!
@brief  Gets the index of the station that raised the event being referenced.
@return  Returns the index of the station that raised the event being referenced. If no event is being referenced, <tt>STATION_INDEX_INVALID</tt> is returned.
*/
    StationIndex GetSetterStationIndex() const
    {
        return (GetHandlingEventToken() != NULL) ? GetHandlingEventToken()->GetSetterStationIndex() : STATION_INDEX_INVALID;
    }


/*!
@brief  Gets the index of the event currently being referenced.
@return  Returns the index of the event currently being referenced. If no event is being referenced, <tt>0</tt> is returned.
*/
    Index GetEventIndex() const
    {
        return (GetHandlingEventToken() != NULL) ? GetHandlingEventToken()->GetIndex() : 0;
    }

protected:
    virtual size_t GetSize() const
    {
        return SerializePolicy::GetSize();
    }


private:
    class EventToken : public EventTokenBase
    {
    public:
        EventToken()
          : EventTokenBase(), m_Value()
        {
        }
        void SetValue(const Value& value)
        {
            m_Value = value;
        }
        const Value& GetValue() const
        {
            return m_Value;
        }

        virtual void Serialize(void* pBuffer) const;
        virtual void Deserialize(const void* cpBuffer);

    private:
        Value m_Value;
    };


    class EventTokenBuffer : public EventTokenBufferBase
    {
    public:
        explicit EventTokenBuffer(EventCloneElementBase* pElement);

    protected:
        virtual EventTokenBase* At(int idx)
        {
            return &m_EventTokenArray[idx];
        }
        virtual const EventTokenBase* At(int idx) const
        {
            return &m_EventTokenArray[idx];
        }

    private:
        EventToken m_EventTokenArray[BUFFER_SIZE];
    };


    EventTokenBuffer m_EventTokenBuffer;
};


//! @cond

template <typename Value, size_t BUFFER_SIZE, typename SerializePolicy>
EventCloneElement<Value, BUFFER_SIZE, SerializePolicy>::EventCloneElement()
  : EventCloneElementBase(&m_EventTokenBuffer),
    m_EventTokenBuffer(this)
{
    PIA_COMPILE_ASSERT(BUFFER_SIZE >= cBufferSizeMin && BUFFER_SIZE <= cBufferSizeMax);
}


template <typename Value, size_t BUFFER_SIZE, typename SerializePolicy>
nn::Result EventCloneElement<Value, BUFFER_SIZE, SerializePolicy>::SetValue(const Value& value, bool isEnableSendDelay)
{
    EventTokenBase* pTokenBase;
    Result r = SetValueCore(&pTokenBase, isEnableSendDelay);
    if (r.IsSuccess() && pTokenBase != NULL)
    {
        EventToken* pToken = static_cast<EventToken*>(pTokenBase);
        pToken->SetValue(value);
    }
    return r;
}


template <typename Value, size_t BUFFER_SIZE, typename SerializePolicy>
void EventCloneElement<Value, BUFFER_SIZE, SerializePolicy>::EventToken::Serialize(void* pBuffer) const
{
    SerializePolicy::Serialize(pBuffer, m_Value);
}


template <typename Value, size_t BUFFER_SIZE, typename SerializePolicy>
void EventCloneElement<Value, BUFFER_SIZE, SerializePolicy>::EventToken::Deserialize(const void* cpBuffer)
{
    SerializePolicy::Deserialize(&m_Value, cpBuffer);
}


template <typename Value, size_t BUFFER_SIZE, typename SerializePolicy>
EventCloneElement<Value, BUFFER_SIZE, SerializePolicy>::EventTokenBuffer::EventTokenBuffer(EventCloneElementBase* pElement)
  : EventTokenBufferBase(BUFFER_SIZE)
{
    for (u32 i = 0; i < BUFFER_SIZE; ++i)
    {
        m_EventTokenArray[i].Init(pElement);
        m_EventTokenArray[i].SetState(EventTokenBase::EVENT_TOKEN_STATE_EMPTY);
        m_EventTokenArray[i].SetIndex(static_cast<Index>(i));
    }
}

//! @endcond
}
}
}  // end of namespace nn::pia::clone
