﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_EventCloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>
#include <pia/common/common_OffsetList.h>

namespace nn
{
namespace pia
{
namespace clone
{

class IDataPacker;

/*!
@brief  This is the base class for managing the sending and receiving of events.

@date 2014-06-03  Fixed a bug where if the first sent event was dropped, the second event or later events were treated as the first event.
@date 2014-03-12  Added a function that gets the current number of event buffers in use.
@date 2013-07-18  Initial version.
*/
class EventCloneElementBase : public CloneElementBase
{
public:
/*!
@brief  Destroys the object.
*/
    virtual ~EventCloneElementBase();


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


/*!
@brief  Gets whether a drop event error occurred.
@details  If a station disconnects abnormally when using <tt>AtomicSharingClone</tt>, an event that was supposed to be received may be lost.
If this error occurs in any of the <tt>EventCloneElement</tt> objects, the <tt>CloneProtocol::GetError</tt> function returns <tt>CloneProtocol::ERROR_TYPE_DROP_EVENT</tt>.
@return  Returns <tt>true</tt> if a drop event error occurred.
@see  CloneProtocol::GetError


*/
    bool IsDropEvent() const
    {
        return m_IsDropEvent;
    }

/*!
@brief  Gets the current number of event buffers in use.

@details  Sometimes the return value is not zero (0) when it is registered to <tt>ReceiveClone</tt>, <tt>AtomicSharingClone</tt>, or <tt>SequentialSharingClone</tt>.
When subsequent data is received, data that can be deleted will be deleted from the buffer.

@return  Returns the number of event buffers in use.

*/
    size_t GetUsingBufferNum() const;


/*!
@cond PRIVATE
@brief  Gets the station from which the event, specified within the event buffer, is waiting for an Ack.

@param[in] index  The index of the target event, when the earliest event in the event buffer has an index of 0.
Set a value equal or larger than 0 and less than the number of events used by the buffer currently.

@return  Returns a bitmap that indicates the station index for the station from which the specified event is waiting for an Ack.
@see  GetUsingBufferNum
*/
    u32 GetBufferEventWaitingStationIndexBitmap(int index) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the clock value for the event specified within the event buffer.

@param[in] index  The index of the target event, when the earliest event in the event buffer has an index of 0.
Set a value equal or larger than 0 and less than the number of events used by the buffer currently.

@return  Returns a clock value for the specified event. Returns <tt>0</tt> when no event is targeted.
@see  GetUsingBufferNum
*/
    ClockValue GetBufferEventClock(int index) const;
    //! @endcond


    //! @cond PRIVATE

public:
    virtual u16 GetType() const
    {
        return GetTypeStatic();
    }
    virtual bool IsSupportedClone(int cloneType) const;

    static u16 GetTypeStatic()
    {
        return TYPE_EVENT;
    }

    static bool Receive(
        CloneElementBase* pElement,
        IDataPacker* pAckPacker,
        const void* cpChunk,
        u16 chunkSize,
        Id elementId,
        StationIndex src,
        StationIndex localStationIndex,
        u32 dispatchCount);

    static size_t GetDataChunkHeaderSize();

public:
    typedef u16 Index;

private:
    enum ReceiveDataResponse
    {
        RESPONSE_NOTHING,
        RESPONSE_ACK,
        RESPONSE_PRESERVER_MIGRATION
    };
    ReceiveDataResponse ReceiveData(const void* cpData, size_t size, StationIndex src, StationIndex setter, ClockValue clock, u64 registerCountAll, u32 destBitmap, Index index, Index eraseIndex);
    void ReceiveAck(StationIndex src, Index index);
    void ReceivePreserverMigration(StationIndex src, Index index);

    void ReceiveInitial(StationIndex src, Index index);
    void ReceiveInitialAck(StationIndex src);

public:
    virtual void ClearData();

    virtual void RemoveDest(StationIndex stationId);
    virtual void AddParticipant(StationIndex stationId);
    virtual void Complement(StationIndex stationId);
    virtual void Lock();

protected:
    static const size_t cBufferSizeMin = 2;
    static const size_t cBufferSizeMax = 0x3fff;

public:
    class EventSendToken : public CloneElementBase::ResendableSendToken
    {
    public:
        enum TokenType
        {
            TOKEN_INITIAL,
            TOKEN_EVENT
        };
        virtual TokenType GetTokenType() const = 0;
    };


    class EventInitialToken : public EventSendToken
    {
    public:
        explicit EventInitialToken(EventCloneElementBase* pElement)
          : EventSendToken(), m_pElement(pElement)
        {
        }

        virtual Type GetElementType() const
        {
            return TYPE_EVENT;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_INITIAL;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

        Id GetId() const
        {
            return m_pElement->GetId();
        }

        Index GetInitialIndex() const
        {
            return m_pElement->m_pEventTokenBuffer->GetNextIndex();
        }
        void SetDestBitmap(u32 destBitmap)
        {
            m_DestBitmap = destBitmap;
        }

    private:
        EventCloneElementBase* m_pElement;
        u32 m_DestBitmap;
    };


    class EventTokenBase : public EventSendToken
    {
    public:
        EventTokenBase()
          : EventSendToken()
        {
        }
        void Init(EventCloneElementBase* pElement)
        {
            m_pElement = pElement;
        }

        virtual Type GetElementType() const
        {
            return TYPE_EVENT;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_EVENT;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }
        virtual bool IsUnicast() const
        {
            return true;
        }

        Index GetIndex() const
        {
            return m_Index;
        }
        StationIndex GetSetterStationIndex() const
        {
            return m_SetterStationIndex;
        }
        StationIndex GetPreserverStationIndex() const
        {
            return m_PreserverStationIndex;
        }
        const ClockValue& GetClock() const
        {
            return m_Clock;
        }
        u64 GetRegisterCountAll() const
        {
            return m_RegisterCountAll;
        }

        void SetDestBitmap(u32 destBitmap)
        {
            m_DestBitmap = destBitmap;
        }
        void RemoveDestBmp(u32 bmp)
        {
            m_DestBitmap &= ~bmp;
        }
        void RemoveDestAll()
        {
            m_DestBitmap = 0;
        }
        void SetIndex(Index index)
        {
            m_Index = index;
        }
        void SetSetterStationIndex(StationIndex stationId)
        {
            m_SetterStationIndex = stationId;
        }
        void SetPreserverStationIndex(StationIndex stationId)
        {
            m_PreserverStationIndex = stationId;
        }
        void SetClock(ClockValue clock)
        {
            m_Clock = clock;
        }
        void SetRegisterCountAll(u64 count)
        {
            m_RegisterCountAll = count;
        }

        enum State
        {
            EVENT_TOKEN_STATE_EMPTY,
            EVENT_TOKEN_STATE_SET,
            EVENT_TOKEN_STATE_RECEIVED,
            EVENT_TOKEN_STATE_HANDLED
        };

        State GetState() const
        {
            return m_State;
        }
        void SetState(State state)
        {
            m_State = state;
        }

        Id GetId() const
        {
            return m_pElement->GetId();
        }
        virtual size_t GetSize() const
        {
            return m_pElement->GetSize();
        }
        virtual void Serialize(void* pBuffer) const = 0;
        virtual void Deserialize(const void* cpBuffer) = 0;
        Index GetEraseIndex() const
        {
            return m_pElement->m_pEventTokenBuffer->GetEraseIndex();
        }

        void CancelSendData()
        {
            m_pElement->CancelSendData(this);
        }
        void AddSendData()
        {
            m_pElement->AddSendData(this);
        }

    private:
        EventCloneElementBase* m_pElement;
        u32 m_DestBitmap;
        Index m_Index;
        StationIndex m_SetterStationIndex;
        StationIndex m_PreserverStationIndex;
        ClockValue m_Clock;
        u64 m_RegisterCountAll;
        State m_State;
    };


    class EventTokenBufferBase
    {
    protected:
        explicit EventTokenBufferBase(u32 size);

    public:
        void Clear();
        void InitIndex(Index index);
        EventTokenBase* Get(Index index);
        EventTokenBase* AssignNext();
        EventTokenBase* Assign(Index index);
        void Recycle();
        bool ShiftHead();
        Index GetHeadIndex() const
        {
            return At(m_HeadIdx)->GetIndex();
        }
        bool IsEmpty() const
        {
            return (m_Num == 0);
        }
        Index GetEraseIndex() const
        {
            return m_EraseIndex;
        }
        Index GetNextIndex() const;

        void RemoveDest(StationIndex stationId);
        void Complement(StationIndex stationId, u32 destBitmap);
        void Lock();

        size_t GetUsingBufferNum()
        {
            return m_Num;
        }

        u32 GetBufferEventWaitingStationIndexBitmap(int index) const;
        ClockValue GetBufferEventClock(int index) const;

    protected:
        virtual EventTokenBase* At(int idx) = 0;
        virtual const EventTokenBase* At(int idx) const = 0;

    private:
        int m_BufferSize;
        int m_HeadIdx;
        int m_Num;

        Index m_EraseIndex;
    };

protected:
    explicit EventCloneElementBase(EventTokenBufferBase* pEventTokenBuffer);

protected:
    nn::Result SetValueCore(EventTokenBase** ppTokenBase, bool isEnableSendDelay);
    void HandleNextCore();
    const EventTokenBase* GetHandlingEventToken() const
    {
        return m_pHandlingEventToken;
    }


private:
    enum IndexState
    {
        EVENT_INDEX_STATE_INVALID,
        EVENT_INDEX_STATE_INIT,
        EVENT_INDEX_STATE_ADJUST,
        EVENT_INDEX_STATE_VALID
    };

    IndexState m_IndexState;

    Index m_HandledIndex;
    EventTokenBase* m_pHandlingEventToken;
    Index m_PreIndex;

    EventInitialToken m_InitialToken;
    EventTokenBufferBase* m_pEventTokenBuffer;

    bool m_IsDropEvent;

    static Definition<EventCloneElementBase> s_Definition;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const EventCloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
