﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_ReceiveClone.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneBase.h>


namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  This class is a derivative of <tt>CloneBase</tt> and provides functionality for managing unidirectional receiving.
@date 2013-07-18  Initial version.

*/
class ReceiveClone : public clone::CloneBase
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    ReceiveClone();


/*!
@brief  Destructor.
*/
    virtual ~ReceiveClone();


/*!
@brief  Gets the <tt>@ref StationIndex</tt> that was specified as the sender when registering this object with <tt>CloneProtocol</tt>.
@return  Returns the <tt>@ref StationIndex</tt> of the sender. Returns <tt>@ref STATION_INDEX_INVALID</tt> if not registered with <tt>CloneProtocol</tt>.
*/
    StationIndex GetSendStationIndex() const
    {
        return GetStationIndex();
    }


/*!
@brief  Determines whether there is a <tt>SendClone</tt> that sends to this <tt>ReceiveClone</tt>.
@return  Returns <tt>true</tt> if there is a <tt>SendClone</tt> that sends to this <tt>ReceiveClone</tt>.
@see  SendClone::IsConnected, SendClone::IsConnectedWithAll
*/
    bool IsConnected() const
    {
        return IsActive() && m_IsReceiving;
    }


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


public:
    // For debugging.
/*!
@cond PRIVATE
@brief  Gets whether the object is receiving. (For debugging.)
*/
    bool IsReceiving() const
    {
        return m_IsReceiving;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether a connection is being requested. (For debugging.)
*/
    bool IsRequest() const
    {
        return m_RequestCommandToken.IsInList();
    }
    //! @endcond


    //! @cond PRIVATE

public:
    static Type GetTypeStatic()
    {
        return TYPE_RECEIVE;
    }
    static Key ToKey(StationIndex stationId, Id id)
    {
        return CloneBase::ToKey(TYPE_RECEIVE, stationId, id);
    }
    virtual Type GetTypeVirtual() const
    {
        return GetTypeStatic();
    }

protected:
    virtual void OnStartup();
    virtual void OnCleanup();
    virtual void OnEnd();
    virtual bool OnCheckEnd();

    virtual void OnEnter(StationIndex stationId);
    virtual void OnExit(StationIndex stationId);
    virtual void OnDisconnect(StationIndex stationId, bool isValid);
    virtual void OnRegister();

public:
    virtual u32 GetDestBitmap() const;
    virtual bool IsInRequest() const
    {
        return m_RequestCommandToken.IsInList();
    }

    virtual u8 GetLocalRegisterCount() const
    {
        return m_LocalRegisterCount;
    }
    virtual bool IsMatchLocalRegisterCount(u64 registerCountAll) const;
    void IncrementLocalRegisterCount();

public:
    bool ReceiveAnnounce();
    void ReceiveAccept(ClockValue clock);
    void ReceiveReject(ClockValue clock);
    void ReceiveEnd();
    void ReceiveStop(StationIndex src);

public:
    class RequestCommandToken : public ClockCommandToken
    {
    public:
        explicit RequestCommandToken(ReceiveClone* pClone)
          : ClockCommandToken(pClone), m_Clock(INVALID_CLOCK)
        {
        }
        virtual CommandType GetType() const
        {
            return COMMAND_REQUEST;
        }
        virtual u32 GetDestStationBitmap() const
        {
            return (0x1 << (static_cast<const ReceiveClone*>(GetClone())->GetStationIndex()));
        }
        virtual ClockValue GetClock() const
        {
            return m_Clock;
        }
        void SetClock(ClockValue clock)
        {
            m_Clock = clock;
        }

    private:
        ClockValue m_Clock;
    };

private:
    RequestCommandToken m_RequestCommandToken;
    bool m_IsReceiving;

    u8 m_LocalRegisterCount; // Register counter for the local station (2-bit).

public:
    template <int TEST_ID, typename Arg>
    static void Test(const ReceiveClone& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
