﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_ReliableCloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>


namespace nn
{
namespace pia
{
namespace clone
{


class IDataPacker;

/*!
@brief  This is the base class for managing the sending and receiving of reliable data.

@date 2013-07-18  Initial version.
*/
class ReliableCloneElementBase : public CloneElementBase
{
protected:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    ReliableCloneElementBase();

public:
/*!
@brief  Destroys the object (destructor).
*/
    virtual ~ReliableCloneElementBase();


/*!
@brief  Gets whether a valid value can be obtained using <tt>@ref ReliableCloneElement::GetValue "GetValue"</tt>.
@return  Returns <tt>true</tt> if a valid value can be obtained using <tt>@ref ReliableCloneElement::GetValue "GetValue"</tt>.
*/
    bool IsValidValue() const
    {
        return m_SetterStationIndex != STATION_INDEX_INVALID;
    }


/*!
@brief  Gets the clock for when the value was set.
@return  Returns the clock for when the value was set. This value is undefined when <tt>IsValidValue</tt> returns <tt>false</tt>.
*/
    ClockValue GetClock() const
    {
        return m_Clock;
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


public:
/*!
@cond PRIVATE
@brief  Gets the <tt>StationIndex</tt> that set the current value. (For debugging.)
*/
    StationIndex GetSetter() const
    {
        return m_SetterStationIndex;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the <tt>StationIndex</tt> responsible for sending the current value. (For debugging.)
*/
    StationIndex GetPreserver() const
    {
        return m_PreserverStationIndex;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the destination bitmap for the current value. (For debugging.)
*/
    u32 GetDestBitmap() const
    {
        return m_DestBitmap;
    }
    //! @endcond

    //! @cond PRIVATE

public:
    virtual u16 GetType() const
    {
        return GetTypeStatic();
    }

    static u16 GetTypeStatic()
    {
        return TYPE_RELIABLE;
    }
    static bool Receive(CloneElementBase* pElement, IDataPacker* pAckPacker, const void* cpChunk, u16 chunkSize, Id elementId, StationIndex src, StationIndex localStationIndex, u32 dispatchCount);
    static size_t GetDataChunkHeaderSize();

private:
    bool ReceiveRequest(StationIndex src);
    bool ReceiveData(const void* cpData, size_t size, StationIndex src, StationIndex setter, ClockValue clock, u32 destBitmap);
    void ReceiveNoData(StationIndex src);
    void ReceivePreserverMigration(StationIndex src, StationIndex setter, ClockValue clock);
    void ReceiveAck(StationIndex src, StationIndex setter, ClockValue clock);

public:
    virtual void ClearData();

    virtual void AddDest(StationIndex stationId);
    virtual void RemoveDest(StationIndex stationId);
    virtual void Complement(StationIndex stationId);

    virtual void RequestInitialData();

private:
    class ReliableSendToken : public CloneElementBase::ResendableSendToken
    {
    public:
        enum TokenType
        {
            TOKEN_REQUEST,
            TOKEN_DATA
        };
        virtual TokenType GetTokenType() const = 0;
    };


    class ReliableRequestToken : public ReliableSendToken
    {
    public:
        explicit ReliableRequestToken(ReliableCloneElementBase* pElement)
          : ReliableSendToken(), m_pElement(pElement)
        {
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_pElement->m_DestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_REQUEST;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

    private:
        Id GetId() const
        {
            return m_pElement->GetId();
        }

    private:
        ReliableCloneElementBase* m_pElement;
    };


    class ReliableDataToken : public ReliableSendToken
    {
    public:
        explicit ReliableDataToken(ReliableCloneElementBase* pElement)
          : ReliableSendToken(), m_pElement(pElement)
        {
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_pElement->m_DestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_DATA;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }
        virtual bool IsUnicast() const
        {
            return true;
        }

    private:
        size_t GetSize() const
        {
            return m_pElement->GetSize();
        }
        Id GetId() const
        {
            return m_pElement->GetId();
        }
        ClockValue GetClock() const
        {
            return m_pElement->m_Clock;
        }
        StationIndex GetSetterStationIndex() const
        {
            return m_pElement->m_SetterStationIndex;
        }
        void Serialize(void* pBuffer) const
        {
            m_pElement->Serialize(pBuffer);
        }

    private:
        ReliableCloneElementBase* m_pElement;
    };


    class ReliableInitialToken : public CloneElementBase::ISendToken
    {
    public:
        explicit ReliableInitialToken(ReliableCloneElementBase* pElement)
          : ISendToken(), m_pElement(pElement)
        {
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBitmap;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

    private:
        size_t GetSize() const
        {
            return m_pElement->GetSize();
        }
        bool IsValidValue() const
        {
            PIA_ASSERT(m_pElement->IsValidValue());
            return true;
        }
        Id GetId() const
        {
            return m_pElement->GetId();
        }
        ClockValue GetClock() const
        {
            return m_pElement->m_Clock;
        }
        StationIndex GetSetterStationIndex() const
        {
            return m_pElement->m_SetterStationIndex;
        }
        void Serialize(void* pBuffer) const
        {
            m_pElement->Serialize(pBuffer);
        }

    public:
        void AddDest(StationIndex stationId)
        {
            m_DestBitmap |= (0x1 << stationId);
        }
        void RemoveDest(StationIndex stationId)
        {
            m_DestBitmap &= (0x1 << stationId);
        }
        void ClearDest()
        {
            m_DestBitmap = 0;
        }

    private:
        ReliableCloneElementBase* m_pElement;
        u32 m_DestBitmap;
    };


protected:
    nn::Result SetValueCore();

    virtual void Serialize(void* pBuffer) const = 0;
    virtual void Deserialize(const void* cpData) = 0;
    virtual void ClearValue() = 0;

private:
    StationIndex m_SetterStationIndex;
    StationIndex m_PreserverStationIndex;
    ClockValue m_Clock;
    u32 m_DestBitmap;
    ReliableRequestToken m_RequestToken;
    ReliableDataToken m_DataToken;
    ReliableInitialToken m_InitialToken;

    static Definition<ReliableCloneElementBase> s_Definition;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const ReliableCloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
