﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_ReliableLargeCloneElement.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_ReliableLargeCloneElementBase.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
    @brief  Manages sending and receiving reliable large-sized data.

    @tparam Value_  Specifies the type of value to manage. The function that initializes the value, <tt>void Value_::Clear</tt>, must be defined.
    @tparam DIVIDE_SIZE  The size of a single data unit, post division. Specify a positive value. A value larger than <tt>@ref CloneProtocol::GetElementSizeMax</tt> cannot be used.
    @tparam SerializePolicy_  Specifies the algorithm to use for serializing \a Value.

    @date 2014-10-08  Initial version.
*/
template <typename Value_, size_t DIVIDE_SIZE = 1094, typename SerializePolicy_ = HostByteOrderSerializePolicy<Value_> >
class ReliableLargeCloneElement : public ReliableLargeCloneElementBase
{
public:
/*!
    @brief  Specifies the type of value managed by this object.
*/
    typedef Value_ Value;


/*!
    @brief  Specifies the algorithm to use for serializing \a Value.
*/
    typedef SerializePolicy_ SerializePolicy;



/*!
    @brief  Instantiates the object with default parameters (default constructor).
*/
    ReliableLargeCloneElement() : ReliableLargeCloneElementBase(m_SerializedBuffer, m_aDivideInfo, SIZE, DIVIDE_SIZE, DIVIDE_NUM),
        m_Value()
    {
    }


/*!
    @brief  Destructor.
*/
    virtual ~ReliableLargeCloneElement()
    {
    }


/*!
    @brief  Sets values.
    @param[in] value  Specifies the flags to set.
    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
    @retval ResultInvalidState  Indicates that the registered <tt>CloneBase</tt> cannot send in its current state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidTiming  Indicates that a value cannot be set at this time. The clock must be advanced to set a value. Handle appropriately in the application.
*/
    nn::Result SetValue(const Value& value);


/*!
    @brief  Gets a value.
    @return  Returns the value managed by this object. This value is undefined when <tt>IsValidValue</tt> returns <tt>false</tt>.
*/
    const Value& GetValue() const
    {
        return m_Value;
    }


    //! @cond PRIVATE

protected:
    virtual void Serialize()
    {
        SerializePolicy::Serialize(m_SerializedBuffer, m_Value);
    }
    virtual void Deserialize()
    {
        SerializePolicy::Deserialize(&m_Value, m_SerializedBuffer);
    }
    virtual void ClearValue()
    {
        m_Value.Clear();
    }


private:
    static const size_t SIZE = SerializePolicy::SIZE;
    static const size_t DIVIDE_LAST_SIZE = (SIZE - 1) % DIVIDE_SIZE + 1;
    static const u32 DIVIDE_NUM = (SIZE - DIVIDE_LAST_SIZE) / DIVIDE_SIZE + 1;

private:
    Value m_Value;
    u8 m_SerializedBuffer[SIZE];
    DivideInfo m_aDivideInfo[DIVIDE_NUM];

    //! @endcond
};


//! @cond
template <typename Value, size_t DIVIDE_SIZE, typename SerializePolicy>
nn::Result ReliableLargeCloneElement<Value, DIVIDE_SIZE, SerializePolicy>::SetValue(const Value& value)
{
    PIA_COMPILE_ASSERT(DIVIDE_NUM <= 0xffff);

    nn::Result r = SetValueCore();
    if (r.IsSuccess())
    {
        m_Value = value;
        PostSetValue();
    }
    return r;
}
//! @endcond
}
}
}  // end of namespace nn::pia::clone
