﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_ReliableLargeCloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>


namespace nn
{
namespace pia
{
namespace clone
{


class IDataPacker;
enum ReliableLargeChunkType;


/*!
    @brief  The base class for managing the sending and receiving of reliable large-sized data.

    @date 2014-10-08  Initial version.
*/
class ReliableLargeCloneElementBase : public CloneElementBase
{
public:
/*!
    @brief  Destructor.
*/
    virtual ~ReliableLargeCloneElementBase();


/*!
    @brief  Gets whether a valid value can be obtained using <tt>@ref ReliableCloneElement::GetValue "GetValue"</tt>.
    @return  Returns <tt>true</tt> if a valid value can be obtained using <tt>@ref ReliableCloneElement::GetValue "GetValue"</tt>.
*/
    bool IsValidValue() const
    {
        return m_ValueSetterStationIndex != STATION_INDEX_INVALID;
    }


/*!
    @brief  Gets the clock for when the value was set.
    @return  Returns the clock for when the value was set. This value is undefined when <tt>IsValidValue</tt> returns <tt>false</tt>.
*/
    ClockValue GetClock() const
    {
        return m_ValueClock;
    }


/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


public:
/*!
    @cond PRIVATE
    @brief  Gets the <tt>StationIndex</tt> that set the current value. (For debugging.)
*/
    StationIndex GetSetter() const
    {
        return m_ValueSetterStationIndex;
    }
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Gets the <tt>StationIndex</tt> responsible for sending the current value. (For debugging.)
*/
    StationIndex GetPreserver() const
    {
        return m_ValuePreserverStationIndex;
    }
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Gets the destination bitmap for the current value. (For debugging.)
*/
    u32 GetDestBitmap(u32 divideIndex) const
    {
        return (divideIndex < m_DivideNum) ? m_aDivideInfo[divideIndex].m_ValueDestBitmap : 0;
    }
    //! @endcond

    //! @cond PRIVATE

public:
    virtual u16 GetType() const
    {
        return GetTypeStatic();
    }

    static u16 GetTypeStatic()
    {
        return TYPE_RELIABLE_LARGE;
    }
    static bool Receive(CloneElementBase* pElement, IDataPacker* pAckPacker, const void* cpChunk, u16 chunkSize, Id elementId, StationIndex src, StationIndex localStationIndex, u32 dispatchCount);
    static size_t GetDataChunkHeaderSize();

private:
    bool ReceiveRequest(StationIndex src, u16 divideIndexOffset, u32 divideIndexFlag);
    ReliableLargeChunkType ReceiveData(const void* cpData, size_t size, StationIndex src, StationIndex setter, ClockValue clock, u16 divideIndex, u32 destBitmap);
    void ReceiveNoData(StationIndex src);
    void ReceivePreserverMigration(StationIndex src, StationIndex setter, ClockValue clock);
    void ReceiveAck(StationIndex src, StationIndex setter, ClockValue clock, u16 divideIndex);

public:
    virtual void ClearData();

    virtual void AddDest(StationIndex stationIndex);
    virtual void RemoveDest(StationIndex stationIndex);
    virtual void Complement(StationIndex stationIndex);

    virtual void RequestInitialData();

private:
    class ReliableLargeSendToken : public CloneElementBase::ResendableSendToken
    {
    public:
        enum TokenType
        {
            TOKEN_REQUEST,
            TOKEN_DATA
        };
        virtual TokenType GetTokenType() const = 0;
    };

    class ReliableLargeRequestToken : public ReliableLargeSendToken
    {
    public:
        explicit ReliableLargeRequestToken(ReliableLargeCloneElementBase* pElement)
          : ReliableLargeSendToken(), m_pElement(pElement), m_DestBitmap(0)
        {
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE_LARGE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_REQUEST;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

    public:
        void SetDest(u32 destBitmap) {m_DestBitmap = destBitmap;}

    private:
        Id GetId() const
        {
            return m_pElement->GetId();
        }

    private:
        ReliableLargeCloneElementBase* m_pElement;
        u32 m_DestBitmap;
    };

protected:
    class ReliableLargeDataToken : public ReliableLargeSendToken
    {
    public:
        explicit ReliableLargeDataToken() : ReliableLargeSendToken() {}
        void Init(ReliableLargeCloneElementBase* pElement, u32 divideIndex)
        {
            m_pElement = pElement;
            m_DivideIndex = divideIndex;
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE_LARGE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_pElement->m_aDivideInfo[m_DivideIndex].m_SerializedDestBitmap;
        }
        virtual TokenType GetTokenType() const
        {
            return TOKEN_DATA;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }
        virtual bool IsUnicast() const
        {
            return true;
        }

    private:
        size_t GetSize() const
        {
            return (m_DivideIndex < m_pElement->m_DivideNum - 1) ? m_pElement->m_DivideSize : m_pElement->m_DivideLastSize;
        }
        Id GetId() const
        {
            return m_pElement->GetId();
        }
        ClockValue GetClock() const
        {
            return m_pElement->m_SerializedClock;
        }
        StationIndex GetSetterStationIndex() const
        {
            return m_pElement->m_SerializedSetterStationIndex;
        }

    private:
        ReliableLargeCloneElementBase* m_pElement;
        u32 m_DivideIndex;
    };


    class ReliableLargeInitialToken : public CloneElementBase::ISendToken
    {
    public:
        explicit ReliableLargeInitialToken() : ISendToken() {}
        void Init(ReliableLargeCloneElementBase* pElement, u32 divideIndex)
        {
            m_pElement = pElement;
            m_DivideIndex = divideIndex;
        }
        virtual Type GetElementType() const
        {
            return TYPE_RELIABLE_LARGE;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBitmap;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

    private:
        size_t GetSize() const
        {
            return (m_DivideIndex < m_pElement->m_DivideNum - 1) ? m_pElement->m_DivideSize : m_pElement->m_DivideLastSize;
        }
        bool IsValidValue() const
        {
            PIA_ASSERT(m_pElement->IsValidValue());
            return true;
        }
        Id GetId() const
        {
            return m_pElement->GetId();
        }
        ClockValue GetClock() const
        {
            return m_pElement->m_SerializedClock;
        }
        StationIndex GetSetterStationIndex() const
        {
            return m_pElement->m_SerializedSetterStationIndex;
        }

    public:
        void AddDest(StationIndex stationId)
        {
            m_DestBitmap |= (0x1 << stationId);
        }
        void RemoveDest(StationIndex stationId)
        {
            m_DestBitmap &= (0x1 << stationId);
        }
        void ClearDest()
        {
            m_DestBitmap = 0;
        }

    private:
        ReliableLargeCloneElementBase* m_pElement;
        u32 m_DivideIndex;
        u32 m_DestBitmap;
    };

protected:
    struct DivideInfo
    {
    public:
        DivideInfo() : m_ValueDestBitmap(0), m_SerializedDestBitmap(0), m_SerializedIsValid(false), m_DataToken(), m_InitialToken() {}

        u32 m_ValueDestBitmap;
        u32 m_SerializedDestBitmap;
        bool m_SerializedIsValid;
        ReliableLargeDataToken m_DataToken;
        ReliableLargeInitialToken m_InitialToken;
    };


protected:

    ReliableLargeCloneElementBase(u8* pBuffer, DivideInfo* aDivideInfo, size_t serializedSize, size_t divideSize, u32 divideNum);

    virtual size_t GetSize() const  {return m_DivideSize;}
    bool IsEqualValueAndSerialized() const
    {
        return m_ValueClock == m_SerializedClock && m_ValueSetterStationIndex == m_SerializedSetterStationIndex;
    }
    void GetRequestDivideIndex(u16* pOffset, u32* pFlag);

protected:
    nn::Result SetValueCore();
    void PostSetValue();
    virtual void Serialize() = 0;
    virtual void Deserialize() = 0;
    virtual void ClearValue() = 0;


private:

    StationIndex m_ValueSetterStationIndex;
    StationIndex m_ValuePreserverStationIndex;
    ClockValue m_ValueClock;

    StationIndex m_SerializedSetterStationIndex;
    StationIndex m_SerializedPreserverStationIndex;
    ClockValue m_SerializedClock;

    u8* m_SerializedBuffer;
    DivideInfo* m_aDivideInfo;

    size_t m_SerializedSize;
    size_t m_DivideSize;
    size_t m_DivideLastSize;
    u32 m_DivideNum;

    ReliableLargeRequestToken m_RequestToken;

#if 0
    StationIndex m_SetterStationIndex;
    StationIndex m_PreserverStationIndex;
    ClockValue m_Clock;
    u32 m_DestBitmap;
    ReliableRequestToken m_RequestToken;
    ReliableDataToken m_DataToken;
    ReliableInitialToken m_InitialToken;
#endif

    static Definition<ReliableLargeCloneElementBase> s_Definition;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const ReliableLargeCloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
