﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_ReverseSerializePolicy.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/common/common_ByteOrder.h>


namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  Reverses the byte order of the source data and then serializes it.

@tparam Type_  Specifies the type of the data to be serialized.

@details  This class is only implemented for the following types: <tt>u8</tt>, <tt>s8</tt>, <tt>u16</tt>, <tt>s16</tt>, <tt>u32</tt>, <tt>s32</tt>, <tt>u64</tt>, <tt>s64</tt>, <tt>f32</tt>, and <tt>f64</tt>.
To use this class with a type, you must implement a <tt>Reverse</tt> function corresponding to that type.

For example:
@code
struct Val
                {
u32 a;
f32 b;
                };
@endcode
You would implement this type as follows.
@code
template<> void ReverseSerializePolicy<Val>::Reverse(Val* pTo, const Val* cpFrom)
                {
ReverseValue(&pTo->a, &cpFrom->a);
ReverseValue(&pTo->b, &cpFrom->b);
                }
@endcode


@date 2013-07-18  Initial version.
*/
template <typename Type_>
class ReverseSerializePolicy
{
public:
/*!
@brief  Defines the type of the data to be serialized.
*/
    typedef Type_ Type;


/*!
@cond PRIVATE
@brief  Serializes data.
@param[out] pBuffer  Specifies the output buffer.
@param[in] value  Specifies the data to serialize.
*/
    static void Serialize(void* pBuffer, const Type& value)
    {
        Reverse(reinterpret_cast<Type*>(pBuffer), &value);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Deserializes data.
@param[out] pValue  Specifies the place to save the deserialized data.
@param[in] cpData  Specifies the byte array to deserialize.
*/
    static void Deserialize(Type* pValue, const void* cpData)
    {
        Reverse(pValue, reinterpret_cast<const Type*>(cpData));
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the size of the serialized data.
@details  Cannot use a value larger than <tt>@ref nn::pia::clone::CloneProtocol::GetElementSizeMax() "CloneProtocol::GetElementSizeMax"</tt>.
@return  Returns the size of the serialized data.
*/
    static size_t GetSize()
    {
        return sizeof(Type);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Returns the size of the serialized data.
@details  This value, and not the value of <tt>@ref GetSize</tt>, is used when using <tt>@ref nn::pia::clone::ReliableLargeCloneElement</tt>.
*/
    static const size_t SIZE = sizeof(Type);
    //! @endcond

private:
/*!
@brief  Reverses the byte order.

@details  This function is only implemented for the following types: <tt>u8</tt>, <tt>s8</tt>, <tt>u16</tt>, <tt>s16</tt>, <tt>u32</tt>, <tt>s32</tt>, <tt>u64</tt>, <tt>s64</tt>, <tt>f32</tt>, and <tt>f64</tt>.
To use this class with a type, you must implement a <tt>Reverse</tt> function corresponding to that type.
@attention  In some cases, the alignment for <span class="argument">pTo</span> or <span class="argument">cpFrom</span> is invalid.

@param[out] pTo  Specifies the place to output the reversed data.
@param[in] cpFrom  Specifies the source data.
*/
    static void Reverse(Type* pTo, const Type* cpFrom);


/*!
@brief  Reverses the byte order of a basic type value.

@tparam T  Specifies the type of value to reverse.

@param[out] pTo  Specifies the place to output the reversed value.
@param[in] cpFrom  Specifies the source data.
*/
    template <typename T>
    static void ReverseValue(T* pTo, const T* cpFrom)
    {
        common::ByteOrder::Swap<sizeof(T)>(pTo, cpFrom);
    }
};

//! @cond

template <>
inline void ReverseSerializePolicy<u8>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<s8>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<u16>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<s16>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<u32>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<s32>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<u64>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<s64>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<f32>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}
template <>
inline void ReverseSerializePolicy<f64>::Reverse(Type* pTo, const Type* pFrom)
{
    ReverseValue(pTo, pFrom);
}

//! @endcond
}
}
}  // end of namespace nn::pia::clone
