﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_SendClone.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneBase.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  This class is a derivative of <tt>CloneBase</tt> and provides functionality for managing unidirectional sending.
@date 2013-07-18  Initial version.

*/
class SendClone : public clone::CloneBase
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    SendClone();


/*!
@brief  Destructor.
*/
    virtual ~SendClone();


/*!
@brief  Determines whether there is a <tt>ReceiveClone</tt> to receive data from this <tt>SendClone</tt>.
@details  Note that if the paired <tt>ReceiveClone</tt> is unregistered in the middle of an operation, the change is not applied until after something takes place, like data being sent or received by the application or internally, the ending of <tt>CloneProtocol</tt> communications, or a station leaving the session.
@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the recipient.
@return  Returns <tt>true</tt> if there is a <tt>ReceiveClone</tt> to receive data from this <tt>SendClone</tt>.
@see  IsConnectedWithAll, ReceiveClone::IsConnected

*/
    bool IsConnected(StationId stationId) const;


/*!
@brief  Determines whether all stations have a <tt>ReceiveClone</tt> to receive data from this <tt>SendClone</tt>.
@details  The stations that are assessed are those for which the <tt>CloneProtocol::IsActiveStation</tt> function returns <tt>true</tt>.
@details  Note that if the paired <tt>ReceiveClone</tt> is unregistered in the middle of an operation, the change is not applied until something takes place, like data being sent or received by the application or internally.
@return  Returns <tt>true</tt> if all stations have a <tt>ReceiveClone</tt> to receive data from this <tt>SendClone</tt>.
@see  IsConnected, ReceiveClone::IsConnected

*/
    bool IsConnectedWithAll() const;


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


public:
/*!
@cond PRIVATE
@brief  Gets the receiver. (For debugging.)
*/
    u32 GetReceiverBmp() const
    {
        return m_ReceiverBmp;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether startup is being announced. (For debugging.)
*/
    bool IsAnnounce() const
    {
        return m_AnnounceCommandToken.IsInList();
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the destination of the startup announcement. (For debugging.)
*/
    u32 GetAnnounceBmp() const
    {
        return m_AnnounceCommandToken.GetDestStationBitmap();
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether the end is being announced. (For debugging.)
*/
    bool IsEnding() const
    {
        return m_EndCommandToken.IsInList();
    }
    //! @endcond


    //! @cond PRIVATE
public:
    static Type GetTypeStatic()
    {
        return TYPE_SEND;
    }
    static Key ToKey(Id id)
    {
        return CloneBase::ToKey(TYPE_SEND, STATION_INDEX_INVALID, id);
    }
    virtual Type GetTypeVirtual() const
    {
        return GetTypeStatic();
    }


protected:
    virtual void OnStartup();
    virtual void OnCleanup();
    virtual void OnEnd();
    virtual void OnComplete();
    virtual bool OnCheckEnd();

    //virtual void OnEnter(StationIndex stationId) {(void)stationId;}
    virtual void OnExit(StationIndex stationId);
    virtual void OnDisconnect(StationIndex stationId, bool isValid);

public:
    bool ReceiveRequest(StationIndex src, ClockValue clock, u8 count);
    void ReceiveAnnounceAck(StationIndex src);
    void ReceiveStop(StationIndex src);

public:
    virtual bool IsReadyToSetValue() const;
    virtual u32 GetDestBitmap() const;

    virtual u64 GetRegisterCountAll() const
    {
        return m_RegisterCountAll;
    }
    void SetRegisterCountAll(StationIndex stationIndex, u8 count);

public:
    class AnnounceCommandToken : public CommandToken
    {
    public:
        explicit AnnounceCommandToken(SendClone* pClone)
          : CommandToken(pClone), m_DestBitmap(0)
        {
        }
        virtual CommandType GetType() const
        {
            return COMMAND_ANNOUNCE;
        }
        virtual u32 GetDestStationBitmap() const
        {
            return m_DestBitmap;
        }
        void SetDestBitmap(u32 destBitmap)
        {
            m_DestBitmap = destBitmap;
        }
        void Exit(StationIndex stationId)
        {
            m_DestBitmap &= ~(0x1 << stationId);
        }

    private:
        u32 m_DestBitmap;
    };


    class EndCommandToken : public CommandToken
    {
    public:
        explicit EndCommandToken(SendClone* pClone)
          : CommandToken(pClone)
        {
        }
        virtual CommandType GetType() const
        {
            return COMMAND_END;
        }
        virtual u32 GetDestStationBitmap() const
        {
            return static_cast<const SendClone*>(GetClone())->m_ReceiverBmp;
        }
    };


private:
    u32 m_ReceiverBmp;

    AnnounceCommandToken m_AnnounceCommandToken;
    EndCommandToken m_EndCommandToken;

    u64 m_RegisterCountAll; // Register counters for all stations (each 2-bit).

public:
    template <int TEST_ID, typename Arg>
    static void Test(const SendClone& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
