﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_SharingClone.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneBase.h>


namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  This class is a derivative of <tt>CloneBase</tt> and provides functionality for bidirectional sending and receiving.
@date 2013-09-25  Initial version.

*/
class SharingClone : public CloneBase
{
protected:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    SharingClone();

public:
/*!
@brief  Destructor.
*/
    virtual ~SharingClone();


/*!
@brief  Determines whether another station's <tt>SharingClone</tt> exchanges data with this <tt>SharingClone</tt>.
@details  Note that if the paired <tt>SharingClone</tt> is unregistered in the middle of an operation, the change is not applied until after something takes place, like data being sent or received by the application or internally, the ending of <tt>CloneProtocol</tt> communications, or a station leaving the session.
@param[in] stationId  Specifies the paired <tt>@ref nn::pia::StationId "StationId"</tt>.
@return  Returns <tt>true</tt> if there is a <tt>SharingClone</tt> that exchanges data with this <tt>SharingClone</tt>.
@see  IsConnectedWithAll

*/
    bool IsConnected(StationId stationId) const;


/*!
@brief  Determines whether all stations have a <tt>SharingClone</tt> that exchanges data with this <tt>SharingClone</tt>.
@details  The stations that are assessed are those for which the <tt>CloneProtocol::IsActiveStation</tt> function returns <tt>true</tt>.
@details  Note that if the paired <tt>SharingClone</tt> is unregistered in the middle of an operation, the change is not applied until something takes place, like data being sent or received by the application or internally.
@return  Returns <tt>true</tt> if all stations have a <tt>SharingClone</tt> that exchanges data with this <tt>SharingClone</tt>.
@see  IsConnected

*/
    bool IsConnectedWithAll() const;


public:
/*!
@cond PRIVATE
@brief  Gets the receiver. (For debugging.)
*/
    u32 GetParticipantBmp() const
    {
        return m_ParticipantBmp;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the destination of the connection request. (For debugging.)
*/
    u32 GetRequestBmp() const
    {
        return m_RequestCommandToken.GetDestStationBitmap();
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether a connection is being requested. (For debugging.)
*/
    bool IsRequest() const
    {
        return m_RequestCommandToken.IsInList();
    }
    //! @endcond

    virtual void Trace(u64 flag) const;

    //! @cond PRIVATE

protected:
    virtual void OnStartup();
    virtual void OnCleanup();
    virtual void OnEnd();
    virtual bool OnCheckEnd();

    virtual void OnExit(StationIndex stationId);
    virtual void OnDisconnect(StationIndex stationId, bool isValid);
    virtual void OnRegister();

    virtual void OnAddParticipant(StationIndex stationIndex)
    {
        NN_PIA_DUMMY_PARAM(stationIndex);
    }

public:
    enum RequestResponse
    {
        REQUEST_RESPONSE_ACCEPT,
        REQUEST_RESPONSE_REJECT,
        REQUEST_RESPONSE_IGNORE
    };
    RequestResponse ReceiveRequest(StationIndex src, ClockValue clock, u8 count);
    void ReceiveAccept(StationIndex src, ClockValue clock, u8 count);
    virtual void ReceiveReject(StationIndex src, ClockValue clock);
    virtual void ReceiveStop(StationIndex src);

public:
    virtual u32 GetDestBitmap() const;
    virtual bool IsInRequest() const
    {
        return m_RequestCommandToken.IsInList();
    }

    virtual u8 GetLocalRegisterCount() const
    {
        return m_LocalRegisterCount;
    }
    virtual u64 GetRegisterCountAll() const
    {
        return m_RegisterCountAll;
    }
    virtual bool IsMatchLocalRegisterCount(u64 registerCountAll) const;
    void IncrementLocalRegisterCount();
    void SetRegisterCountAll(StationIndex stationIndex, u8 count);

protected:
    bool IsSendingRequest(StationIndex stationIndex) const;

public:
    class RequestCommandToken : public ClockCommandToken
    {
    public:
        explicit RequestCommandToken(SharingClone* pClone)
          : ClockCommandToken(pClone), m_DestBitmap(0), m_Clock(INVALID_CLOCK)
        {
        }
        virtual CommandType GetType() const
        {
            return COMMAND_REQUEST;
        }
        virtual u32 GetDestStationBitmap() const
        {
            return m_DestBitmap;
        }
        void SetDestBitmap(u32 destBitmap)
        {
            m_DestBitmap = destBitmap;
        }
        void Exit(StationIndex stationId)
        {
            m_DestBitmap &= ~(0x1 << stationId);
        }
        virtual ClockValue GetClock() const
        {
            return m_Clock;
        }
        void SetClock(ClockValue clock)
        {
            m_Clock = clock;
        }

    private:
        u32 m_DestBitmap;
        ClockValue m_Clock;
    };

private:
    u32 m_ParticipantBmp;

    RequestCommandToken m_RequestCommandToken;

    u8 m_LocalRegisterCount; // Register counter for the local station (2-bit).
    u64 m_RegisterCountAll; // Register counters for all stations (each 2-bit).

public:
    template <int TEST_ID, typename Arg>
    static void Test(const SharingClone& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
