﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_StraightSerializePolicy.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <memory>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  Serializes the source data structure without changes.

@tparam Type_  Specifies the type of the data to be serialized.

@date 2013-07-18  Initial version.
*/
template <typename Type_>
class StraightSerializePolicy
{
public:
/*!
@brief  Defines the type of the data to be serialized.
*/
    typedef Type_ Type;


/*!
@cond PRIVATE
@brief  Serializes data.
@param[out] pBuffer  Specifies the output buffer.
@param[in] value  Specifies the data to serialize.
*/
    static void Serialize(void* pBuffer, const Type& value)
    {
#if NN_PIA_CAFE
        if (reinterpret_cast<u32>(pBuffer) % __alignof(Type) != 0)
        {
            OSBlockMove(pBuffer, &value, sizeof(Type), FALSE);
        }
        else
        {
            *reinterpret_cast<Type*>(pBuffer) = value;
        }
#elif NN_PIA_CTR
        nn::nstd::MemCpy(pBuffer, &value, sizeof(Type));
#else
        *reinterpret_cast<Type*>(pBuffer) = value;
#endif
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Deserializes data.
@param[out] pValue  Specifies the place to save the deserialized data.
@param[in] cpData  Specifies the byte array to deserialize.
*/
    static void Deserialize(Type* pValue, const void* cpData)
    {
#if NN_PIA_CAFE
        if (reinterpret_cast<u32>(cpData) % __alignof(Type) != 0)
        {
            OSBlockMove(pValue, cpData, sizeof(Type), FALSE);
        }
        else
        {
            *pValue = *reinterpret_cast<const Type*>(cpData);
        }
#elif NN_PIA_CTR
        nn::nstd::MemCpy(pValue, cpData, sizeof(Type));
#else
        *pValue = *reinterpret_cast<const Type*>(cpData);
#endif
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the size of the serialized data.
@details  Cannot use a value larger than <tt>@ref nn::pia::clone::CloneProtocol::GetElementSizeMax() "CloneProtocol::GetElementSizeMax"</tt>.
@return  Returns the size of the serialized data.
*/
    static size_t GetSize()
    {
        return sizeof(Type);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Returns the size of the serialized data.
@details  This value, and not the value of <tt>@ref GetSize</tt>, is used when using <tt>@ref nn::pia::clone::ReliableLargeCloneElement</tt>.
*/
    static const size_t SIZE = sizeof(Type);
    //! @endcond

};
}
}
}  // end of namespace nn::pia::clone
