﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_UnreliableCloneElement.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_UnreliableCloneElementBase.h>
#include <pia/clone/clone_SerializePolicyDefinition.h>

namespace nn
{
namespace pia
{
namespace clone
{


/*!
@brief  Manages the sending and receiving of unreliable data.

@tparam Value_  Specifies the type of value to manage.
@tparam SerializePolicy_  Specifies the algorithm to use for serializing <span class="argument">Value</span>.

@date 2013-07-18  Initial version.
*/
template <typename Value_, typename SerializePolicy_ = HostByteOrderSerializePolicy<Value_> >
class UnreliableCloneElement : public UnreliableCloneElementBase
{
public:
/*!
@brief  Specifies the type of value managed by this object.
*/
    typedef Value_ Value;


/*!
@brief  Specifies the algorithm to use for serializing <span class="argument">Value</span>.
*/
    typedef SerializePolicy_ SerializePolicy;


/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    UnreliableCloneElement()
      : UnreliableCloneElementBase(), m_Value()
    {
    }


/*!
@brief  Destroys the object (destructor).
*/
    virtual ~UnreliableCloneElement()
    {
    }


/*!
@brief  Sets values.
@param[in] value  Specifies the flags to set.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the registered <tt>CloneBase</tt> cannot send in its current state. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetValue(const Value& value);


/*!
@brief  Gets a value.
@return  Returns the value managed by this object. This value is undefined when <tt>IsValidValue</tt> returns <tt>false</tt>.
*/
    const Value& GetValue() const
    {
        return m_Value;
    }


    //! @cond PRIVATE

protected:
    virtual size_t GetSize() const
    {
        return SerializePolicy::GetSize();
    }
    virtual void Serialize(void* pBuffer) const
    {
        SerializePolicy::Serialize(pBuffer, m_Value);
    }
    virtual void Deserialize(const void* cpData)
    {
        SerializePolicy::Deserialize(&m_Value, cpData);
    }
    virtual void ClearValue()
    {
        m_Value = Value();
    }


private:
    Value m_Value;

    //! @endcond
};


//! @cond
template <typename Value, typename SerializePolicy>
nn::Result UnreliableCloneElement<Value, SerializePolicy>::SetValue(const Value& value)
{
    bool isCopyValue;
    nn::Result r = SetValueCore(&isCopyValue);
    if (isCopyValue)
    {
        m_Value = value;
    }
    return r;
}
//! @endcond
}
}
}  // end of namespace nn::pia::clone
