﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     clone_UnreliableCloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/clone/clone_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>


namespace nn
{
namespace pia
{
namespace clone
{


class IDataPacker;

/*!
@brief  This is the base class for managing the sending and receiving of unreliable data.

@date 2013-07-18  Initial version.
*/
class UnreliableCloneElementBase : public CloneElementBase
{
protected:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    UnreliableCloneElementBase();

public:
/*!
@brief  Destroys the object (destructor).
*/
    virtual ~UnreliableCloneElementBase();


/*!
@brief  Gets whether a valid value can be obtained using <tt>@ref UnreliableCloneElement::GetValue "GetValue"</tt>.
@return  Returns <tt>true</tt> if a valid value can be obtained using <tt>@ref UnreliableCloneElement::GetValue "GetValue"</tt>.
*/
    bool IsValidValue() const
    {
        return m_IsValidValue;
    }


/*!
@brief  Gets the clock for when the value was set.
@return  Returns the clock for when the value was set. This value is undefined when <tt>@ref IsValidValue</tt> returns <tt>false</tt>.
*/
    ClockValue GetClock() const
    {
        return m_Clock;
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


    //! @cond PRIVATE

public:
    virtual u16 GetType() const
    {
        return GetTypeStatic();
    }

    static u16 GetTypeStatic()
    {
        return TYPE_UNRELIABLE;
    }
    static bool Receive(CloneElementBase* pElement, IDataPacker* pAckPacker, const void* cpChunk, u16 chunkSize, Id elementId, StationIndex src, StationIndex localStationIndex, u32 dispatchCount);
    static size_t GetDataChunkHeaderSize();

private:
    void ReceiveData(const void* cpData, size_t size, ClockValue clock);

public:
    virtual void ClearData();

private:
    class UnreliableSendToken : public CloneElementBase::ISendToken
    {
    public:
        explicit UnreliableSendToken(UnreliableCloneElementBase* pElement)
          : CloneElementBase::ISendToken(), m_pElement(pElement)
        {
        }
        virtual Type GetElementType() const
        {
            return TYPE_UNRELIABLE;
        }
        virtual u32 GetDestBitmap() const;
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

    private:
        Id GetId() const
        {
            return m_pElement->GetId();
        }
        size_t GetSize() const
        {
            return m_pElement->GetSize();
        }
        ClockValue GetClock() const
        {
            return m_pElement->m_Clock;
        }
        void Serialize(void* pBuffer) const
        {
            m_pElement->Serialize(pBuffer);
        }

    private:
        UnreliableCloneElementBase* m_pElement;
    };

protected:
    nn::Result SetValueCore(bool* pIsCopyValue);

    virtual void Serialize(void* pBuffer) const = 0;
    virtual void Deserialize(const void* cpData) = 0;
    virtual void ClearValue() = 0;

private:
    bool m_IsValidValue;
    ClockValue m_Clock;
    UnreliableSendToken m_SendDataToken;

    static Definition<UnreliableCloneElementBase> s_Definition;

public:
    template <int TEST_ID, typename Arg>
    static void Test(const UnreliableCloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::clone
