﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Api.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


// This macro is used to access the version number of the SDK used by <tt>NN_PIA_COMMON_SET_PROGRAM_INFORMATION</tt>.
#if NN_PIA_CTR
#define NN_PIA_SDK_VER NN_CURRENT_VERSION_NUMBER
#elif NN_PIA_CAFE
#define NN_PIA_SDK_VER CAFE_OS_SDK_VERSION
#elif NN_PIA_WIN || NN_PIA_NIN_WIN
#define NN_PIA_SDK_VER 0xffffffff
#else
#error  "Invalid platform."
#endif

// @addtogroup  pia_common_debug_macro  Definitions of Debugging Macros Provided by the Pia Library
// @brief  Specifies the debugging macros provided by the Pia library.
//! @{
//!     @name  Others
//!     @{

/*!
@def  NN_PIA_COMMON_SET_PROGRAM_INFORMATION
@brief  This macro registers information about the SDK and NEX library used by the application with the PIA library.

@details  This macro is meant to be called after the <tt>common::Initialize</tt> function.
To access the constant that represents the NEX version number, you must first use the <CODE>#include</CODE> directive to process the NEX header files.

*/
#define NN_PIA_COMMON_SET_PROGRAM_INFORMATION() \
    nn::pia::common::SetProgramInformation(     \
        NN_PIA_SDK_VER,                         \
        NEX_VERSION_MAJOR,                      \
        NEX_VERSION_MINOR,                      \
        NEX_VERSION_MICRO)

//!     @}
//! @}


//  Contains non-member functions in the common module.

namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  Indicates whether the <tt>common</tt> module is initialized.

@return  Returns <tt>true</tt> if initialized, and <tt>false</tt> otherwise.
*/
bool IsInitialized(void);


/*!
@brief  Initializes the <tt>common</tt> module.

@details  The heap used by the Pia library is configured during initialization.
Next, the function sets up an instance of the <tt>nn::pia::common::Log</tt> class so that macros like <tt>PIA_LOG</tt> can be used.
@if CTR_DOC
Pia uses thread-local storage implicitly to efficiently manage threads internally.
For more information, see the Programming Manual.
@endif

@param[in] pMem  Specifies a pointer to the start address of the memory block passed to the Pia library.
The buffer must be 4-byte aligned.
@param[in] size  Specifies the size of the memory block passed in <span class="argument">pMem</span>. The value is in bytes.
The library asserts and stops execution if the size is not large enough to construct the heap.

@return  Returns a <tt>Result</tt> value indicating success if initialization succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidArgument  Specifies that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyInitialized  The module is already initialized. Programming error. Fix your program so that this error is not returned.


*/
nn::Result Initialize(void* pMem, u32 size);


/*!
@brief  Finalizes the <tt>common</tt> module.

@details  This function destroys the instance of <tt>nn::pia::common::Log</tt> created by the <tt>@ref Initialize</tt> function, and frees the memory supplied to the Pia library by the application in the call to <tt>@ref Initialize</tt>.
This function does nothing if the <tt>common</tt> module is not initialized when it is called.

*/
void Finalize(void);


/*!
@brief  Declares the start of setup.

@details  Call this function before creating a singleton in the <tt>common</tt> module.
Always call this function and <tt>@ref EndSetup</tt> even if your application does not create any singletons in the <tt>common</tt> module.

@return  Returns a <tt>Result</tt> value indicating success if it is called at the right time. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultNotInitialized  The function for initializing the <tt>common</tt> module has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.

*/
nn::Result BeginSetup(void);


/*!
@brief  Declares the end of setup.

@details  Call this function after you are finished creating singletons and using other features of the <tt>common</tt> module.
Calling this function configures memory to be used by the <tt>common</tt> module singletons and optimizes memory management within the library.
Always call <tt>@ref BeginSetup</tt> and this function even if your application does not create any singletons in the <tt>common</tt> module.

@return  Returns a <tt>Result</tt> value indicating success if it is called at the right time. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultNotInitialized  The function for initializing the <tt>common</tt> module has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.


*/
nn::Result EndSetup(void);

/*!
@brief  Gets the amount of memory being used by the Pia library at the point this function is called.

@details  Do not call this function between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions of each module. If you call this function, an incorrect value is returned.

@return  Returns the amount of memory being used by the Pia library at the point this function was called.
If the <tt>common</tt> module is not set up, or if internal errors occur, <tt>0</tt> is returned.

@retval 0  The Pia library has not been set up.


*/
size_t GetMemoryUsage(void);


/*!
@cond PRIVATE
@brief  Registers program information.

@details  Registers the NEX and SDK version numbers used by the application.
@endcond
*/
void SetProgramInformation(u32 sdkVer, u8 nexMajorVer, u8 nexMinorVer, u8 nexMicroVer);


/*!
@cond PRIVATE
@brief  Indicates whether an application is executing code between calls to the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.
@endcond
*/
bool IsInSetupMode(void);


/*!
@cond PRIVATE
@brief  Prints debugging information.
@endcond
*/
void PrintDebugInfo(void);


/*!
@cond PRIVATE
@brief  Prints information about the environment.
@endcond
*/
void PrintEnvironmentInfo(void);


/*!
@brief  The callback type passed to <tt>RegisterInternalPrintCallback</tt>.

@param[in] pStr  Pointer to the head of the string. The string is not necessarily terminated by a null character.
@param[in] len  The string length.
*/
typedef void (*InternalPrintCallback)(const char* pStr, s32 len);


/*!
@brief  Sets the callback that hooks the Pia console output.

@param[in] pCallback  Specifies a pointer to the callback function to register.
@return  Returns a pointer to the callback function that was registered before this function was called.
<tt>NULL</tt> is returned if no callback function was registered.
*/
InternalPrintCallback RegisterInternalPrintCallback(InternalPrintCallback pCallback);


#if NN_PIA_EXPERIMENT_USE_PRINT_HOOK
/*!
@cond PRIVATE
@brief  Prints the string.

@param[in] cpFormat  Specifies the format string.
@param[in] varg  Specifies the argument list.
*/
void VPrintf(const char* cpFormat, va_list varg);
//! @endcond


/*!
@cond PRIVATE
@brief  Prints the string.

@param[in] cpFormat  Specifies the format string.
@param[in] Specifies  variable length arguments.
*/
    void Printf(const char* cpFormat, ...);
//! @endcond
#endif // end of NN_PIA_EXPERIMENT_USE_PRINT_HOOK
}
}
}  // end of namespace nn::pia::common
