﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_ByteOrder.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  Converts the byte order.

@date 2012-04-04  Initial version.
*/
class ByteOrder
{
public:
    static inline u16 Swap16(u16 v)
    {
        return (v << 8) | (v >> 8);
    }

    static inline u32 Swap32(u32 v)
    {
        u32 w = ((v & 0xff00ff00) >> 8) | ((v & 0x00ff00ff) << 8);
        return (w >> 16) | (w << 16);
    }

    static inline u64 Swap64(u64 v)
    {
        u64 w = ((v & 0xff00ff00ff00ff00) >> 8) | ((v & 0x00ff00ff00ff00ff) << 8);
        u64 x = ((w & 0xffff0000ffff0000) >> 16) | ((w & 0x0000ffff0000ffff) << 16);
        return (x >> 32) | (x << 32);
    }

/*!
@brief  Converts a <tt>u16</tt> value from network byte order to host byte order.

@param[in] net  Specifies the <tt>u16</tt> value in network byte order.

@return  Returns the <tt>u16</tt> value in host byte order.
*/
    static inline u16 NetworkToHost16(u16 net)
    {
#if NN_PIA_ENDIAN_BIG
        return net;
#else
        return Swap16(net);
#endif
    }


/*!
@brief  Converts a <tt>u32</tt> value from network byte order to host byte order.

@param[in] net  Specifies the <tt>u32</tt> value in network byte order.

@return  Returns the <tt>u32</tt> value in host byte order.
*/
    static inline u32 NetworkToHost32(u32 net)
    {
#if NN_PIA_ENDIAN_BIG
        return net;
#else
        return Swap32(net);
#endif
    }


/*!
@brief  Converts a <tt>u64</tt> value from network byte order to host byte order.

@param[in] net  Specifies the <tt>u64</tt> value in network byte order.

@return  Returns the <tt>u64</tt> value in host byte order.
*/
    static inline u64 NetworkToHost64(u64 net)
    {
#if NN_PIA_ENDIAN_BIG
        return net;
#else
        return Swap64(net);
#endif
    }


/*!
@brief  Converts a <tt>u16</tt> value from host byte order to network byte order.

@param[in] host  Specifies the <tt>u16</tt> value in host byte order.

@return  Returns the <tt>u16</tt> value in network byte order.
*/
    static inline u16 HostToNetwork16(u16 host)
    {
#if NN_PIA_ENDIAN_BIG
        return host;
#else
        return Swap16(host);
#endif
    }


/*!
@brief  Converts a <tt>u32</tt> value from host byte order to network byte order.

@param[in] host  Specifies the <tt>u32</tt> value in host byte order.

@return  Returns the <tt>u32</tt> value in network byte order.
*/
    static inline u32 HostToNetwork32(u32 host)
    {
#if NN_PIA_ENDIAN_BIG
        return host;
#else
        return Swap32(host);
#endif
    }


/*!
@brief  Converts a <tt>u64</tt> value from host byte order to network byte order.

@param[in] host  Specifies the <tt>u64</tt> value in host byte order.

@return  Returns the <tt>u64</tt> value in network byte order.
*/
    static inline u64 HostToNetwork64(u64 host)
    {
#if NN_PIA_ENDIAN_BIG
        return host;
#else
        return Swap64(host);
#endif
    }


/*!
@cond PRIVATE
@brief  Reverses the order of bytes in the specified byte array.
@tparam Size  Specifies the number of bytes in the byte array being reversed. You can only specify one of <tt>1</tt>, <tt>2</tt>, <tt>4</tt>, or <tt>8</tt>.
@param[out] pTo  Specifies the address to write the result of conversion to.
@param[in] cpFrom  Specifies the byte array to be converted.
*/
    template <size_t Size>
    static inline void Swap(void* pTo, const void* cpFrom);
    //! @endcond
};


//! @cond
template <>
inline void ByteOrder::Swap<1>(void* pTo, const void* cpFrom)
{
    *reinterpret_cast<u8*>(pTo) = *reinterpret_cast<const u8*>(cpFrom);
}

template <>
inline void ByteOrder::Swap<2>(void* pTo, const void* cpFrom)
{
    *reinterpret_cast<u16*>(pTo) = Swap16(*reinterpret_cast<const u16*>(cpFrom));
}

template <>
inline void ByteOrder::Swap<4>(void* pTo, const void* cpFrom)
{
    *reinterpret_cast<u32*>(pTo) = Swap32(*reinterpret_cast<const u32*>(cpFrom));
}

template <>
inline void ByteOrder::Swap<8>(void* pTo, const void* cpFrom)
{
    reinterpret_cast<u32*>(pTo)[0] = Swap32(reinterpret_cast<const u32*>(cpFrom)[1]);
    reinterpret_cast<u32*>(pTo)[1] = Swap32(reinterpret_cast<const u32*>(cpFrom)[0]);
}
//! @endcond
}
}
}  // end of namespace nn::pia::common
