﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_CachedPrint.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_CriticalSection.h>


// @addtogroup  pia_common_debug_macro  Definitions of Debugging Macros Provided by the Pia Library
// @brief  Specifies the debugging macros provided by the Pia library.
//! @{
//!     @name  Log Output
//!     @{

/*!
@def  PIA_CACHED_PRINTF
@brief  Caches console output. When no instance of <tt>CachedPrint</tt> has been generated, console output operates normally.
@param[in] ...  Indicates that arguments are the same as <tt>printf</tt> in the standard library.

*/
#if NN_PIA_UNDER_DEVELOP
#define PIA_CACHED_PRINTF(...)                                              \
    if (::nn::pia::common::CachedPrint::GetInstance() != NULL)              \
    {                                                                       \
        ::nn::pia::common::CachedPrint::GetInstance()->Printf(__VA_ARGS__); \
    }                                                                       \
    else                                                                    \
    {                                                                       \
        PIA_BASIC_REPORT(__VA_ARGS__);                                      \
    }
#else
#define PIA_CACHED_PRINTF(...) \
    {                          \
    }
#endif


/*!
@def  PIA_CACHED_VPRINTF
@brief  Caches console output. When no instance of <tt>CachedPrint</tt> has been generated, console output operates normally.
@param[in] format  Specifies the output format string.
@param[in] varg  A variable argument list.

*/
#if NN_PIA_UNDER_DEVELOP
#define PIA_CACHED_VPRINTF(format, varg)                                          \
    if (::nn::pia::common::CachedPrint::GetInstance() != NULL)                    \
    {                                                                             \
        ::nn::pia::common::CachedPrint::GetInstance()->VPrintf((format), (varg)); \
    }                                                                             \
    else                                                                          \
    {                                                                             \
        PIA_BASIC_VREPORT((format), (varg));                                      \
    }
#else
#define PIA_CACHED_VPRINTF(format, varg) \
    {                                    \
        (void)(format);                  \
        (void)(varg);                    \
    }
#endif

//!     @}
//! @}


namespace nn
{
namespace pia
{
namespace common
{

/*!
@brief  Caches console output.
This class is used for debugging. Do not use it in production versions of code.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2012-05-01  Initial version.

*/
class CachedPrint : public common::RootObject
{
public:
/*!
@brief  Creates an instance (singleton pattern).
@details  Call this member function after calling <tt>nn::pia::common::BeginSetup</tt>, but before calling <tt>nn::pia::common::EndSetup</tt>.

@param[in] bufferSize  Specifies the size of the buffer to be cached (in bytes). Specify a value equal to or greater than <tt>512</tt>.
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  The <tt>common</tt> module has not been initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was not called between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.




*/
    static nn::Result CreateInstance(u32 bufferSize);


/*!
@brief  Destroys the instance (singleton pattern).
@details  This function does nothing and returns if called when an instance has not been created.


*/
    static void DestroyInstance();


/*!
@brief  Gets a pointer to the <tt>CachedPrint</tt> instance (singleton pattern).
@return  Returns a <tt>NULL</tt> pointer if the instance cannot be created.

*/
    static CachedPrint* GetInstance()
    {
        return s_pInstance;
    }


/*!
@brief  Prints information useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    void Trace(u64 flag) const;


/*!
@cond PRIVATE
@brief  Writes text strings to the cache.
@param[in] cpFormat  Specifies the format string.
@param[in] varg  Specifies the argument list.

*/
    void VPrintf(const char* cpFormat, va_list varg);
    //! @endcond


/*!
@cond PRIVATE
@brief  Writes text strings to the cache.
@param[in] cpFormat  Specifies the format string.
@param[in] Specifies  variable length arguments.

*/
    void Printf(const char* cpFormat, ...);
    //! @endcond


/*!
@brief  Clears the string currently being cached.
*/
    void Clear();


/*!
@brief  Outputs the string currently being cached and then clears it.
*/
    void Flush();


/*!
@cond PRIVATE
@brief  Gets the string currently being cached. Strings are null-terminated. (This function has not been verified. Do not use it yet.)
@param[out] pBuffer  Specifies the buffer used to get the string.
@param[out] pLength  The length of the string to get.
@param[in] bufferSize  Specifies the buffer size (in bytes).
@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function will return <tt>true</tt>.
@retval ResultInvalidArgument  Indicates an invalid argument. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates no string is currently being cached. A <tt>NULL</tt> terminator is still written to the specified buffer. Handle appropriately in the application.



*/
    nn::Result Read(char* pBuffer, size_t* pLength, size_t bufferSize);
    //! @endcond


private:
    // Not implemented.
    CachedPrint();
    CachedPrint(const CachedPrint& rhs);
    bool operator==(const CachedPrint& rhs) const;
    bool operator!=(const CachedPrint& rhs) const;
    CachedPrint& operator=(const CachedPrint& rhs);

    explicit CachedPrint(u32 bufferSize);
    ~CachedPrint();

    static CachedPrint* s_pInstance;

    char* m_pBuffer;
    u32 m_Size;

    u32 m_Top;
    u32 m_Next;
    u32 m_Repeat;

    CriticalSection m_CriticalSection;
};
}
}
}  // end of namespace nn::pia::common
