﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_CallContext.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_RootObject.h>

#define USE_CANCEL_REQUEST 1  // n1769


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  Represents the calling context.

@details  The calling context is primarily used to log the execution status of asynchronous functions.

@date 2013-12-16  Removed a previous prohibition on calling the <tt>Cancel</tt> function in the <tt>CallContext</tt> passed to the Pia API that does not support cancellations.
@date 2013-12-13  Added the <tt>IsInProgress</tt> function.
@date 2013-12-13  Made the <tt>GetState</tt> function private.
@date 2012-09-21  No longer inherits <tt>Event</tt>.
@date 2012-09-20  Added a description of the conditions under which <tt>Cancel</tt> can be called.
@date 2012-05-14  Added the <tt>CompletionCallback</tt> parameter so values can be specified by applications.
@date 2012-05-14  Initial version.
*/
class CallContext : public RootObject
{
public:
    //n1589:todo2: It would be better to allow the user to select an arbitrary argument.
/*!
@brief  This <tt>typedef</tt> defines the callback function that is executed when the asynchronous operation finishes.

@details  Register the callback using the <tt>CompletionCallback</tt> or <tt>RegisterCompletionCallback</tt> function so that a callback function is called when asynchronous processing completes.

@see  RegisterCompletionCallback

*/
    typedef void (*CompletionCallback)(nn::Result result, void* pUserArg);

/*!
@cond PRIVATE
@brief  Enumerates <tt>@ref CallContext</tt> states.
*/
    enum State
    {
        Available,       //!< Indicates that execution is possible.
        CallInProgress,  //!< Indicates that the operation is executing. Indicates that a context in this state cannot be used as the <tt>Context</tt> of another asynchronous process.
        CallSuccess,     //!< Indicates that the operation succeeded and ended.
        CallFailure,     //!< Indicates that the operation failed and ended.
        CallCancelled    //!< Indicates that the process was canceled.
    };
    //! @endcond

/*!
@brief  Instantiates the object (constructor).

@details  The state is set to <tt>Available</tt>.
*/
    CallContext(void);


/*!
@brief  Destroys the object.

@details  The destructor asserts if the <tt>CallContext</tt> state is <tt>CallInProgress</tt> when it is called.
*/
    ~CallContext(void);


/*!
@brief  Sets the callback function executed when asynchronous processing completes.

@param[in] pCallback  Specifies a pointer to the callback function called when asynchronous processing completes.
@param[in] pUserArg  Specifies the value given in the second argument when calling <span class="argument">pCallback</span>.
@see  CompletionCallback
*/
    void RegisterCompletionCallback(CompletionCallback pCallback, void* pUserArg = NULL);

/*!
@cond PRIVATE
@brief
*/
    bool InitiateCall();
    //! @endcond


/*!
@cond PRIVATE
@brief
*/
    void SignalSuccess(nn::Result result);
    //! @endcond


/*!
@cond PRIVATE
@brief
*/
    void SignalFailure(nn::Result result);
//! @endcond


#if USE_CANCEL_REQUEST
/*!
@cond PRIVATE
@brief
*/
    void SignalCancel();
//! @endcond
#else
/*!
@cond PRIVATE
@brief
*/
    void ForceCancel();
//! @endcond
#endif


#if USE_CANCEL_REQUEST
/*!
@brief  Requests the cancellation of asynchronous processing.

@details  Note that calling this API function does not mean that the asynchronous process is canceled immediately.
After this API function is called, you must periodically call the <tt>IsCompleted</tt> function to check the state of completion. Wait until <tt>true</tt> is returned.

@see  IsCompleted, GetResult


*/
    void Cancel(void);
#else
/*!
@brief  Cancels asynchronous processing.

@details  If this function is called in a <tt>CallContext</tt> other than <tt>CallInProgress</tt>, it does nothing and returns.

After asynchronous processing is terminated using this function, the valued obtained by <tt>GetResult</tt> is <tt>ResultCancelled</tt>.

Asynchronous processes can only be canceled at those times clearly indicated in the description of asynchronous processes that use this <tt>CallContext</tt>.
Do not call this function on other asynchronous processes.

@see  GetResult


*/
    void Cancel(void);
#endif


/*!
@cond PRIVATE
@brief
*/
    void Complete(nn::Result result);
    //! @endcond


/*!
@brief  Determines whether an asynchronous process is in progress.

@return  Returns <tt>true</tt> if an asynchronous process is in progress, and <tt>false</tt> otherwise.
*/
    bool IsInProgress() const
    {
        return m_State == CallInProgress;
    }


/*!
@brief  Determines whether an asynchronous process has completed.

@return  Returns <tt>true</tt> if the asynchronous process has completed (including cases where it has failed); returns <tt>false</tt> otherwise.
*/
    bool IsCompleted() const
    {
        return m_State == CallSuccess ||
               m_State == CallFailure ||
               m_State == CallCancelled;
    }


/*!
@cond PRIVATE
@brief  Sets the state.

@param[in] state  Specifies the state to set.
@see  GetState
*/
    void SetState(State state);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the current state.

@return  Returns the current <tt>CallContext</tt>.
*/
    State GetState() const
    {
        return m_State;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the <tt>nn::Result</tt> value representing the execution result of the asynchronous process.

@param[in] result  Specifies the <tt>nn::Result</tt> value of the asynchronous process.
*/
    void SetResult(nn::Result result);
    //! @endcond


/*!
@brief  Gets the <tt>nn::Result</tt> value representing the execution result of the asynchronous process.

@return  Returns the <tt>nn::Result</tt> value of the asynchronous process.
*/
    nn::Result GetResult() const
    {
        return m_Result;
    }


#if USE_CANCEL_REQUEST
/*!
@cond PRIVATE
@brief  Determines whether the <tt>Cancel</tt> function has been called.
*/
    bool IsCancelRequested() const
    {
        return m_bCancelRequested;
    }
//! @endcond
#endif

    void Reset();

protected:
    State m_State;
    nn::Result m_Result;
    CompletionCallback m_CompletionCallback;
    void* m_pUserArg;
#if USE_CANCEL_REQUEST
    bool m_bCancelRequested;
#endif

private:
/*!
@brief  This copy constructor is private.
*/
    CallContext(const CallContext&);

/*!
@brief  This assignment operator is private.
*/
    CallContext& operator=(const CallContext&);
};


typedef CallContext AsyncContext;
}
}
}  // end of namespace nn::pia::common
