﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: common_DateTime.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

namespace nn
{
namespace pia
{
namespace common
{
/*!
  @brief  Class that represents time.
*/
class DateTime : public nn::pia::common::RootObject
{
public:
/*!
    @brief  Instantiates an object with default parameters (default constructor).
*/
    DateTime() :
        m_Year(0), m_Month(0), m_Day(0),
        m_Hour(0), m_Minute(0), m_Second(0),
        m_IsRegistered(false)
    {
    }

/*!
    @brief  The copy constructor.

    @param[in] rhs  Specifies the <tt>DateTime</tt> to copy.
*/
    DateTime(const DateTime& rhs)
    {
        m_Year = rhs.m_Year;
        m_Month = rhs.m_Month;
        m_Day = rhs.m_Day;
        m_Hour = rhs.m_Hour;
        m_Minute = rhs.m_Minute;
        m_Second = rhs.m_Second;
        m_IsRegistered = rhs.m_IsRegistered;
    }

/*!
    @brief  Instantiates an object with the specified date and time.

    @param[in] year  Year.
    @param[in] month  Month.
    @param[in] day  Day.
    @param[in] hour  Hour.
    @param[in] minute  Minutes.
    @param[in] second  Seconds.
*/
    DateTime(u16 year, u8 month, u8 day, u8 hour, u8 minute, u8 second)
    {
        m_Year = year;
        m_Month = month;
        m_Day = day;
        m_Hour = hour;
        m_Minute = minute;
        m_Second = second;
        m_IsRegistered = true;
    }

/*!
    @brief  Assignment operator.
    @param[in] rhs  Assignment date and time.
*/
    DateTime& operator=(const DateTime& rhs)
    {
        m_Year = rhs.m_Year;
        m_Month = rhs.m_Month;
        m_Day = rhs.m_Day;
        m_Hour = rhs.m_Hour;
        m_Minute = rhs.m_Minute;
        m_Second = rhs.m_Second;
        m_IsRegistered = rhs.m_IsRegistered;
        return *this;
    }

/*!
    @brief  Specifies the year.
    @details  Specify a value in the range from 1 to 9999.
    @param[in] year  Year.
*/
    void SetYear(u16 year)
    {
        m_Year = year;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified year.
    @return  Returns the specified year.
*/
    u16 GetYear() const
    {
        return m_Year;
    }

/*!
    @brief  Specifies the month.
    @details  Specify a value in the range from 1 to 12.
    @param[in] month  Month.
*/
    void SetMonth(u8 month)
    {
        m_Month = month;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified month.
    @return  Returns the specified month.
*/
    u8 GetMonth() const
    {
        return m_Month;
    }

/*!
    @brief  Specifies the day.
    @details  Specify a value in the range from 1 to 31.
    @param[in] day  Day.
*/
    void SetDay(u8 day)
    {
        m_Day = day;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified day.
    @return  Returns the specified day.
*/
    u8 GetDay() const
    {
        return m_Day;
    }

/*!
    @brief  Specifies the hour.
    @details  Specify a value in the range from 0 to 23.
    @param[in] hour  Hour.
*/
    void SetHour(u8 hour)
    {
        m_Hour = hour;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified hour.
    @return  Returns the specified hour.
*/
    u8 GetHour() const
    {
        return m_Hour;
    }

/*!
    @brief  Specifies the minutes.
    @details  Specify a value in the range from 0 to 59.
    @param[in] minute  Minutes.
*/
    void SetMinute(u8 minute)
    {
        m_Minute = minute;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified minutes.
    @return  Returns the specified minutes.
*/
    u8 GetMinute() const
    {
        return m_Minute;
    }

/*!
    @brief  Specifies the seconds.
    @details  Specify a value in the range from 0 to 59.
    @param[in] second  Seconds.
*/
    void SetSecond(u8 second)
    {
        m_Second = second;
        m_IsRegistered = true;
    }

/*!
    @brief  Gets the specified seconds.
    @return  Returns the specified seconds.
*/
    u8 GetSecond() const
    {
        return m_Second;
    }

/*!
    @brief  Gets whether a value has been specified.
    @return  Returns <tt>true</tt> if a value is specified.
*/
    bool IsRegistered() const
    {
        return m_IsRegistered;
    }

/*!
    @brief  Gets whether the specified date and time is correct.
    @return  Returns <tt>true</tt> if the specified date and time is correct.
*/
    bool IsValid() const
    {
        if (m_Year == 0 || m_Year > 9999)
        {
            return false;
        }
        if (m_Month == 0 || m_Month > 12)
        {
            return false;
        }
        if (m_Day == 0)
        {
            return false;
        }
        else if (m_Month == 2)
        {
            u8 days = (m_Year % 400 == 0 || (m_Year % 100 != 0 && m_Year % 4 == 0)) ? 29 : 28;
            if (m_Day > days)
            {
                return false;
            }
        }
        else if (m_Month == 4 || m_Month == 6 || m_Month == 9 || m_Month == 11)
        {
            if (m_Day > 30)
            {
                return false;
            }
        }
        else
        {
            if (m_Day > 31)
            {
                return false;
            }
        }
        if (m_Hour > 23)
        {
            return false;
        }
        if (m_Minute > 59)
        {
            return false;
        }
        if (m_Second > 59)
        {
            return false;
        }
        return true;
    }

/*!
    @brief  Clears the specified date and time.
*/
    void Clear()
    {
        m_Year = 0;
        m_Month = 0;
        m_Day = 0;
        m_Hour = 0;
        m_Minute = 0;
        m_Second = 0;
        m_IsRegistered = false;
    }

/*!
    @brief  The destructor.
*/
    virtual ~DateTime()
    {
    }

private:
    u16 m_Year;      //!<  1-9999
    u8  m_Month;     //!<  1-12
    u8  m_Day;       //!<  1-31
    u8  m_Hour;      //!<  0-23
    u8  m_Minute;    //!<  0-59
    u8  m_Second;    //!<  0-59
    bool m_IsRegistered;
};

}
}
}  // end of namespace nn::pia::common
