﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_ErrorHandler.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_Trace.h>

#if NN_PIA_ENABLE_TRACE

/*!
@brief  Outputs a trace of <tt>Result</tt> in the Pia module.

@param result  Specifies the value of <tt>Result</tt>.
*/
#define PIA_TRACE_RESULT(result)                                                                                       \
    \
{                                                                                                               \
        PIA_TRACE(nn::pia::common::TRACE_FLAG_ALWAYS, "%s, line: %d", PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_LINE); \
        nn::pia::common::ErrorHandler::TraceResult(result);                                                            \
    \
}

#else  // else of NN_PIA_ENABLE_TRACE

#define PIA_TRACE_RESULT(result) (void)(result)

#endif  // end of NN_PIA_ENABLE_TRACE

namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  This class is used for error handling.

@date 2012-07-11  Deleted the <tt>Severity</tt>-related code that was scheduled for removal.
@date 2011-12-07  Added a table to the <tt>GetSeverity</tt> page that shows the relationship between each <tt>Result</tt> used in the Pia module and the associated severity.
*/
class ErrorHandler
{
public:
/*!
@brief  Displays detailed information for <tt>nn::Result</tt> values.
@param[in] result  The <tt>nn::Result</tt> value to trace.
*/
    static void TraceResult(const nn::Result& result);

/*!
@brief  Displays detailed information for <tt>nn::Result</tt> values.

@param[in] traceFlags  Specifies trace flags. For more information, see <tt>@ref TraceFlag</tt>.
@param[in] result  The <tt>nn::Result</tt> value to trace.
*/
    static void TraceResult(u64 traceFlags, const nn::Result& result);

/*!
@brief  Determines whether there is an error in the Pia module.
@param[in] result  Specifies the <tt>nn::Result</tt> value to check.
@return  Returns <tt>true</tt> if the value is a Pia module <tt>Result</tt>, and returns <tt>false</tt> otherwise.
*/
    static bool IsPiaError(const nn::Result& result);

/*!
@brief  Gets the Pia module in which the error occurred.
@param[in] result  Specifies the <tt>nn::Result</tt> value to be checked by the module.
@return  Returns the module type using <tt>@ref nn::pia::ModuleType</tt>.
*/
    static ModuleType GetModuleTypeFromResult(const nn::Result& result);


private:
/*!
@brief  This constructor is private because it does not create an instance.
*/
    ErrorHandler(void);

/*!
@brief  This destructor is private.
*/
    ~ErrorHandler(void);

/*!
@brief  This copy constructor is private.
*/
    ErrorHandler(const ErrorHandler&);

/*!
@brief  This assignment operator is private.
*/
    ErrorHandler& operator=(const ErrorHandler&);
};
}
}
}  // end of namespace nn::pia::common
