﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Event.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#if NN_PIA_CTR
#include <nn/os.h>
#endif

namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  This class represents events.

@date 2012-10-30  Added <tt>TryWait</tt>.
@date 2012-09-21  Made functions public.
*/
class Event
{
public:
/*!
@brief  Instantiates the object.

@details  You must explicitly call <tt>Initialize</tt> to use an event.
*/
    Event();

/*!
@brief  Destroys the object.
*/
    ~Event();

/*!
@brief  Initializes an event.

@param[in] bManualReset  Specify <tt>true</tt> for manual reset, or <tt>false</tt> for automatic reset.
@see  Finalize
*/
    void Initialize(bool bManualReset = true);

/*!
@brief  Destroys an event.

@see  Initialize
*/
    void Finalize();


/*!
@brief  Waits for the event to enter the signal state.

@see  Signal
*/
    void Wait();


/*!
@brief  Determines whether the event is in the signal state.

@details  Clears the signal state if the event is an automatic reset event.

@see  Wait
*/
    bool TryWait();


/*!
@brief  Puts the event in the signal state.

@see  Wait, Reset
*/
    void Signal();


/*!
@brief  Puts the event in the non-signal state.

@see  Signal
*/
    void Reset();


#if NN_PIA_CTR
/*!
@brief  Gets this <tt>nn::os::LightEvent</tt> instance.
@return  Returns this <tt>nn::os::LightEvent</tt> instance.
*/
    nn::os::LightEvent* GetRawLightEvent()
    {
        return &m_OsEvent;
    }
#endif


protected:
#if NN_PIA_CTR
    nn::os::LightEvent m_OsEvent;
#elif NN_PIA_WIN || NN_PIA_NIN_WIN
    HANDLE m_Handle;
#elif NN_PIA_CAFE
    OSEvent m_Event;
#else
#error  "Invalid platform";
#endif

private:
/*!
@brief  This copy constructor is private.
*/
    Event(const Event&);

/*!
@brief  This assignment operator is private.
*/
    Event& operator=(const Event&);
};
}
}
}  // end of namespace nn::pia::common
