﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_FixedRingBuffer.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  This class represents a fixed length ring buffer.

@date 2012-05-14  Initial version.
*/
template <typename T, int N>
class FixedRingBuffer : public common::RootObject
{
public:
/*!
@brief  Instantiates the object.
*/
    FixedRingBuffer(void)
    {
        Clear();
    }


/*!
@brief  Destroys the object.
*/
    ~FixedRingBuffer(void)
    {
    }


/*!
@brief  Empties the ring buffer.
*/
    void Clear(void)
    {
        m_Size = 0;
        m_BeginIndex = 0;
    }


/*!
@brief  Adds an element to the end. This function asserts if executed when the buffer is full.
*/
    void PushBack(T value)
    {
        PIA_ASSERT(!IsFull());

        pushBack(value);
    }


/*!
@brief  Adds an element to the end. Destroys the element at the start of the buffer and overwrites if executed when the buffer is full. This version does not assert.
*/
    void PushBackForce(T value)
    {
        pushBack(value);
    }


/*!
@brief  Removes one element from the start of the buffer. This function asserts if called when the buffer is empty.
*/
    void PopFront(void)
    {
        PIA_ASSERT(!IsEmpty());

        incBeginIndex();
        --m_Size;
    }


/*!
@brief  Determines whether the ring buffer is full.
*/
    bool IsFull(void) const
    {
        return m_Size == N;
    }


/*!
@brief  Determines whether there are any elements in the ring buffer.
*/
    bool IsEmpty(void) const
    {
        return m_Size == 0;
    }


/*!
@brief  Gets the number of elements in the ring buffer.
*/
    size_t GetSize(void) const
    {
        return m_Size;
    }


/*!
@brief  Provides access using the [] operators (non-<tt>const</tt> version).
*/
    T& operator[](size_t n)
    {
        return m_Values[getIndex(n)];
    }


/*!
@brief  Provides access using the [] operators (<tt>const</tt> version).
*/
    const T& operator[](size_t n) const
    {
        return m_Values[getIndex(n)];
    }


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const
    {
        (void)flag;  // Not implemented.
    }


private:
    // Add an element to the end. When the elements are full, the element at the start is overwritten.
    void pushBack(T value)
    {
        m_Values[getEndIndex()] = value;
        if (IsFull())
        {
            incBeginIndex();
        }
        else
        {
            ++m_Size;
        }
    }

    // Increments the start index.
    void incBeginIndex(void)
    {
        m_BeginIndex = (m_BeginIndex + 1) % N;
    }

    // Currently, the position that the lead element is in <tt>m_Values[]</tt> is returned as a subscript.
    size_t getBeginIndex(void) const
    {
        return m_BeginIndex;
    }

    // Currently, the position that the element after the last element is in <tt>m_Values[]</tt> is returned as a subscript.
    size_t getEndIndex(void) const
    {
        return (getBeginIndex() + GetSize()) % N;
    }

    // Returns a specific index value (a subscript value in <tt>m_Values[]</tt>).
    size_t getIndex(size_t n) const
    {
        PIA_ASSERT(n < N);
        return (m_BeginIndex + n) % N;
    }

    size_t m_Size;        // Number of elements maintained.
    size_t m_BeginIndex;  // Supports the subscript of the value displayed with [0].
    T m_Values[N];        // The instance of the elements.
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
