﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_InetAddress.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once


#include <pia/common/common_definitions.h>
#include <pia/common/common_RootObject.h>
#include <pia/common/common_ByteOrder.h>

namespace nn
{
namespace pia
{
namespace common
{

// Forward declaration.
class String;

/**
* @brief  An IPv4 address structure.
*/
struct InAddr
{
    u32 addr;
};


//n1589
#define NN_PIA_COMMON_INET_ADDRESS_STRING_LEN_MAX sizeof(_T("255.255.255.255"))


/*!
@brief  Class that represents an Internet address.

@details  This class contains information such as the IP address and port number.

@date 2012-04-19  Initial version.
*/
class InetAddress : public RootObject
{
public:
    static const size_t SERIALIZE_SIZE_MAX =
        sizeof(u32)     // m_Address
        + sizeof(u16);  // m_Port

/*!
@brief  Instantiates the object with default parameters (default constructor).
Sets the IP address and port number being cleared to zero.
*/
    InetAddress();

/*!
@brief  Instantiates the object and initializes it with the IP address and port number specified.

@param[in] uiAddress  The IP address.
@param[in] uiPort  Specifies the port number.
*/
    InetAddress(u32 uiAddress, u16 uiPort);

/*!
@brief  Copy constructor.

@param[in] oInetAddress  Specifies the object being copied.
*/
    InetAddress(const InetAddress& oInetAddress);

/*!
@brief  Destroys the object.
*/
    ~InetAddress();

/*!
@brief  Instantiates the object.
*/
    void Init()
    {
        m_Address = 0;
        m_Port = 0;
    }

/*!
@brief  Sets the IP address in host byte order.

@param[in] uiAddress  Specifies the IP address (in host byte order).
@see  GetAddress
*/
    void SetAddress(const u32 uiAddress)
    {
        m_Address = uiAddress;
    }

/*!
@brief  Gets the IP address in host byte order.

@return  Returns the IP address set for this object (in host byte order).
@see  SetAddress
*/
    u32 GetAddress() const
    {
        return m_Address;
    }

/*!
@brief  Sets the port number in host byte order.

@param[in] ui16Port  Specifies the port number (in host byte order).
@see  GetPort
*/
    void SetPort(const u16 ui16Port)
    {
        m_Port = ui16Port;
    }

/*!
@brief  Sets the port number in host byte order.

@return  Returns the port number set for this object (in host byte order).
@see  SetPort
*/
    u16 GetPort() const
    {
        return m_Port;
    }

/*!
@brief  Gets the key for comparison.

@return  Returns the key for comparison.
*/
    s64 GetKey() const;

/*!
@brief  This is the assignment operator.

@param[in] oInetAddress  The object to assign.
@return  See this object.
*/
    InetAddress& operator=(const InetAddress& oInetAddress);

/*!
@brief  Gets the address string notation.
Address strings are represented in a format such as <tt>255.255.255.255:80</tt>.

@param[out] pString  Specifies a pointer to the <tt>String</tt> object used to set the address string notation.
*/
    void GetAddressString(String* pString) const;

/*!
@brief  Prints information that is useful for debugging.

@param[in] uiTraceFlag  Bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 uiTraceFlag) const;

/*!
@brief  This is an equality operator.

@param[in] oInetAddress  Specifies the object to compare.
@return  Returns <tt>true</tt> if this object and the one being compared to are the same; returns <tt>false</tt> otherwise.
*/
    bool operator==(const InetAddress& oInetAddress) const
    {
        return (GetKey() == oInetAddress.GetKey());
    }

/*!
@brief  This is an equality operator.

@param[in] oInetAddress  Specifies the object to compare.
@return  Returns <tt>true</tt> if this object and the one being compared to are not the same; returns <tt>false</tt> otherwise.
*/
    bool operator!=(const InetAddress& oInetAddress) const
    {
        return !operator==(oInetAddress);
    }

/*!
@brief  Comparison operator.

@param[in] oInetAddress  Specifies the object to compare.
@return
*/
    bool operator<(const InetAddress& oInetAddress) const
    {
        return (GetKey() < oInetAddress.GetKey());
    }

/*!
@cond PRIVATE
@brief  Gets the length of the object as a serialized byte array. This value is in bytes.

@return  Returns the length of the serialized data. This value is in bytes.
@see  Serialize, Deserialize
*/
    size_t GetSerializedSize(void) const
    {
        return sizeof(u32) + sizeof(u16);
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Serializes the object.

@param[out] pBuffer  Holds the serialized data.
@param[out] pDataLen  Receives a pointer to the length of the serialized data that was written to the buffer. This value is in bytes.
@param[in] bufferSize  Specifies the size of the buffer that was specified in <span class="argument">pBuffer</span>.
@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function will return <tt>true</tt>.
@retval ResultInvalidArgument  Indicates that an argument is invalid. (For example, a NULL pointer was specified.) This function also returns this error if the size specified by <span class="argument">pBuffer</span> or <span class="argument">bufferSize</span> is too small.
@see  Deserialize, GetSerializedSize
*/
    nn::Result Serialize(bit8* pBuffer, size_t* pDataLen, size_t bufferSize) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Restores an object from serialized data.

@param[in] pData  Pointer to the serialized data.

@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function will return <tt>true</tt>.
@retval ResultInvalidArgument  Indicates that an argument is invalid. (For example, a NULL pointer was specified.)
@see  Serialize, GetSerializedSize
*/
    nn::Result Deserialize(const bit8* pData);
    //! @endcond

/*!
@cond PRIVATE
@brief  Determines whether a valid address and port have been set.

@return  Returns <tt>true</tt> if valid and <tt>false</tt> otherwise.
*/
    bool IsValid(void) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Determines whether the specified address is valid.

@return  Returns <tt>true</tt> if valid and <tt>false</tt> otherwise.
*/
    bool IsValidAddress(void) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Determines whether the specified port is valid.

@return  Returns <tt>true</tt> if valid and <tt>false</tt> otherwise.
*/
    bool IsValidPort(void) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Determines whether the address is a private address.

@return  Returns <tt>true</tt> if private, and returns <tt>false</tt> otherwise.
*/
    bool IsPrivate(void) const;
    //! @endcond


protected:
    u32 m_Address;
    u16 m_Port;
};
}
}
}  // end of namespace nn::pia::common
