﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Job.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_ListNode.h>
#include <pia/common/common_Time.h>

namespace nn
{
namespace pia
{
namespace common
{

/*!
@cond PRIVATE
@brief  Represents the base class for jobs.

@details  To use job-related features, override the pure virtual function <tt>ExecuteCore</tt> in the derived class.

*/
class Job : public RootObject
{
public:
/*!
@brief  Enumerates job states.
*/
    enum State
    {
        STATE_INITIAL,   //! Indicates the state immediately after a job instance was created.
        STATE_WAITING,   //! Indicates the job is currently waiting. The job does not run until the specified time has elapsed.
        STATE_SUSPEND,   //! Indicates the job has been suspended.
        STATE_READY,     //! Indicates the job can be executed.
        STATE_RUNNING,   //! Indicates the job is currently executing.
        STATE_COMPLETE,  //! Indicates that job execution has completed.
        STATE_MAX        //! Provided as a sentinel to indicate the last <tt>State</tt> enumerated type.
    };


/*!
@brief  Instantiates the object.

@details  Because this class includes pure virtual functions, you must create a class that inherits this base class to use job-related features.

*/
    Job();


/*!
@brief  Destroys the object.
*/
    virtual ~Job();


/*!
@brief  Resets a job so it can be reused.

@details  Call this function when you want to reuse a job that has been executed by the scheduler.
Calling this member function is equivalent to placing the instance in the same state as immediately after the constructor is called.
The function asserts if it is called on a job whose status is <tt>STATE_RUNNING</tt> (if it is called while <tt>ExecuteCore</tt> is running).
The function succeeds if it is called in other states, but a warning message is output if it is called in any state other than <tt>STATE_INITIAL</tt> or <tt>STATE_COMPLETE</tt>.


*/
    virtual void Reset(bool isCheck = true);


/*!
@brief  Puts a job currently in initial status (<tt>STATE_INITIAL</tt>) in executable status (<tt>STATE_READY</tt>).

@details  This function asserts if it is called when the job is in any other state.

@param[in] isBackground  Specify <tt>true</tt> if using <tt>BackgroundScheduler</tt> to run this function in another thread.
*/
    void Ready(bool isBackground = false);


/*!
@brief  Puts a job currently in suspended status (<tt>STATE_SUSPEND</tt>) in executable status (<tt>STATE_READY</tt>).

@details  This function asserts if it is called when the job is in any other state.

@param[in] isBackground  Specify <tt>true</tt> if using <tt>BackgroundScheduler</tt> to run this function in another thread.
*/
    void Resume(bool isBackground = false);


/*!
@brief  Stops the job.

@details  The functionality is the same as the <tt>Reset</tt> function, except that it does not output warnings.
*/
    void Cancel()
    {
        Reset(false);
    }


/*!
@brief  Gets the job's state.

@return  Returns the job's current state.
*/
    State GetState() const;


/*!
@brief  Gets whether the job is currently running.
@return  Returns <tt>true</tt> if the job's status is <tt>WAITING</tt>, <tt>READY</tt>, <tt>RUNNING</tt>, or <tt>SUSPEND</tt>.
*/
    bool IsRunning() const;


/*!
@brief  Gets whether the job is running in the foreground.
@return  Returns <tt>true</tt> if the job's state is <tt>WAITING</tt>, <tt>READY</tt>, or <tt>RUNNING</tt> in the foreground.
*/
    bool IsForeground() const;

/*!
@brief  Gets whether the job is running in the background.
@return  Returns <tt>true</tt> if the job's state is <tt>WAITING</tt>, <tt>READY</tt>, or <tt>RUNNING</tt> in the background.
*/
    bool IsBackground() const;


/*!
@brief  Gets how many times the job has been run.

@return  Returns <tt>0</tt> if the job has never been run by calling the <tt>Execute</tt> function.
The counter increments after the <tt>ExecuteCore</tt> function is called.
*/
    u32 GetExecutedCount() const
    {
        return m_ExecutedCount;
    }


/*!
@brief  Gets the next time when a job in a waiting state will be executed.

@return  Returns the time in ticks.
*/
    const Time& GetExecutionTime() const
    {
        return m_ExecutionTime;
    }


/*!
@cond PRIVATE
@brief  Executes a job.

@details  This member function is meant to be called by the scheduler.
Applications should not call this function directly.
*/
    void Execute(bool isBackground);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the offset to the list node.
*/
    static s32 GetListNodeOffset()
    {
        return offsetof(Job, m_ListNode);
    }
    //! @endcond


/*!
@brief  Indicates the type of execution result.
*/
    enum ExecuteState
    {
        EXECUTE_STATE_CONTINUE,             //! Indicates execution is not complete. Continue execution.
        EXECUTE_STATE_COMPLETE,             //! Indicates execution has completed.
        EXECUTE_STATE_CANCEL,               //! Indicates execution has been canceled. (Currently no different from <tt>COMPLETE</tt>.)
        EXECUTE_STATE_SUSPEND,              //! Indicates execution was suspended.
        EXECUTE_STATE_WAIT,                 //! Indicates execution is not complete, but will resume after a set period of time.
        EXECUTE_STATE_WAIT_NEXT,            //! Indicates execution is not complete. Resume execution with the next call to <tt>Scheduler::Dispatch</tt>. Jobs in this state cannot be used in the background.
        EXECUTE_STATE_CONTINUE_MAIN,        //! Indicates execution is not complete. Next, execute in the main thread using the <tt>Dispatch</tt> function.
        EXECUTE_STATE_CONTINUE_BACKGROUND,  //! Indicates execution is not complete. The job will run in the background next.
        EXECUTE_STATE_MAX                   //! Provided as a sentinel to indicate the last <tt>ExecuteState</tt> enumerated type.
    };


/*!
@brief  Represents the result of calling the <tt>Execute</tt> function.
*/
    struct ExecuteResult
    {
    public:
/*!
@brief  Instantiates the object.

@details  If you specify a state other than <tt>EXECUTE_STATE_WAIT</tt>, call this overload, and only specify the <tt>ExecuteState</tt> value.
You can pass a value of type <tt>ExecuteState</tt> in the return value of the <tt>ExecuteCore</tt> function directly, because type conversion is performed implicitly.
*/
        ExecuteResult(ExecuteState executeState)
          : m_ExecuteState(executeState), m_WaitTime(0)
        {
            PIA_ASSERT(executeState >= 0 && executeState < EXECUTE_STATE_MAX);
        }

/*!
@brief  Instantiates the object.

@details  If you specify <tt>EXECUTE_STATE_WAIT</tt>, use this overload to specify the state in addition to the wait time.
*/
        ExecuteResult(ExecuteState executeState, u16 waitTime)
          : m_ExecuteState(executeState), m_WaitTime(waitTime)
        {
            PIA_ASSERT(executeState == EXECUTE_STATE_WAIT || waitTime == 0);
        }

        u8 m_ExecuteState;
        u8 m_Reserved;
        u16 m_WaitTime;
    };


protected:
/*!
@brief  This pure virtual function is used to code the processing that the job executes.

@details  Use an override of this function to code the actual work done by the job.
*/
    virtual ExecuteResult ExecuteCore() = 0;


private:
    enum InnerState
    {
        INNER_STATE_INITIAL,
        INNER_STATE_FG_WAITING,
        INNER_STATE_FG_READY,
        INNER_STATE_FG_RUNNING,
        INNER_STATE_BG_WAITING,
        INNER_STATE_BG_READY,
        INNER_STATE_BG_RUNNING,
        INNER_STATE_SUSPEND,
        INNER_STATE_COMPLETE,
        INNER_STATE_NUM
    };


/*!
@brief  Gets a string representation of the job state.

@return  Returns a pointer to a string representation of the job state.
*/
    static const char* GetInnerStateString(InnerState innerState);


    // Specifies the job's state.
    volatile InnerState m_InnerState;

    // Indicates how many times the <tt>Execute</tt> function has been executed.
    u32 m_ExecutedCount;

    // Indicates how much time is left until a job in waiting status can be switched to ready status.
    Time m_ExecutionTime;

    // Represents a list node.
    ListNode m_ListNode;
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
