﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_LatestMedian.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#include <pia/common/common_FixedRingBuffer.h>

#include <algorithm>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  This class is used to find the median of the last <tt>n</tt> values.

@date 2013-10-16  Added the <tt>GetSize</tt> function to get the number of elements.
@date 2013-01-22  Changed specifications so that the median of the <tt>n</tt> nearest values is also found.
@date 2012-06-14  Added the <tt>IsEmpty</tt> function.
@date 2012-05-14  Initial version.
*/
template <typename T, int N>
class LatestMedian : public common::RootObject
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    LatestMedian(void)
    {
        Clear();
    }


/*!
@brief  Destructor.
*/
    ~LatestMedian(void)
    {
    }


/*!
@brief  Clears elements. Restores the object to its state immediately after the constructor was executed.
*/
    void Clear(void)
    {
        m_Values.Clear();
    }


/*!
@brief  Adds the latest value. Only <tt>n</tt> values can be maintained. The earliest value is discarded.
*/
    void Add(T value)
    {
        m_Values.PushBackForce(value);
    }


/*!
@brief  Gets the median. Do not call this function frequently because it performs sorting internally.
This function returns <tt>0</tt> if called when there are no elements.
*/
    T GetMedian(void) const
    {
        return GetMedian(N);
    }


/*!
@brief  Gets the median of the last <tt>n</tt> values.
If there are less than <tt>n</tt> values, the median of those present is calculated.

Do not call this function frequently because it performs sorting internally.
This function returns <tt>0</tt> if it is called when there are no elements.
*/
    T GetMedian(size_t n) const
    {
        PIA_ASSERT(n <= N);

        // Checks the number of elements.
        size_t sz = m_Values.GetSize();
        PIA_ASSERT(sz <= N);
        if (sz == 0)
        {
            return 0;  // The function is guaranteed to return <tt>0</tt> if there are no elements.
        }

        // If there are less than <tt>n</tt> values, the median of those present is calculated.
        size_t num = (sz < n) ? sz : n;  // <tt>num</tt> specifies the number of samples being used this time.

        // An array is prepared separately and sorted.
        // To get the last <tt>num</tt> values, read them from the end of the array.
        T buf[N];
        for (u32 i = 0; i < num; ++i)
        {
            size_t idx = sz - 1 - i;  // <tt>sz-1</tt> is the subscript of the last element in the array.
            buf[i] = m_Values[idx];
        }
        std::sort(buf, buf + num);

        // Gets the subscript of the middle value. The median calculation differs depending on whether the number of elements is odd or even.
        int midIdx = num / 2;
        if (num % 2 == 0)
        {
            return static_cast<T>(buf[midIdx - 1] + buf[midIdx]) / static_cast<T>(2);
        }
        else
        {
            return buf[midIdx];
        }
    }


/*!
@brief  Determines whether the array is empty.
*/
    bool IsEmpty(void) const
    {
        return m_Values.IsEmpty();
    }


/*!
@brief  Determines whether the container is full of elements.
*/
    bool IsFull(void) const
    {
        return m_Values.IsFull();
    }


/*!
@brief  Gets the number of elements.
*/
    size_t GetSize(void) const
    {
        return m_Values.GetSize();
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const
    {
        (void)flag;  // Not implemented.
    }


private:
    FixedRingBuffer<T, N> m_Values;
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
