﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: common_Log.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_RootObject.h>
#include <pia/common/common_Time.h>

namespace nn
{
namespace pia
{
namespace common
{


/*!
  @brief  This class is used for logging.

  @details  You can set a level for the log and switch between
              showing and hiding the log for each level.

  @date 2015-06-19  Added the <tt>PIA_LOG_EX()</tt> macro.
  @date 2012-05-30  Times can now also be printed.
  @date 2012-05-30  Made the <tt>Write()</tt> function private.
  @date 2012-04-19  Initial version.
*/
class Log : public RootObject
{
public:
    // Specifies a bit mask representing the log level.
    static const u32 LEVEL_DEBUG = 0x00000001 << 0;    //!<  Specifies the debug information level used inside Pia.
    static const u32 LEVEL_INFO = 0x00000001 << 1;     //!<  Specifies the detailed information level used inside Pia.
    static const u32 LEVEL_WARNING = 0x00000001 << 2;  //!<  Specifies the warning level. (This should be changed, but can be ignored.)
    static const u32 LEVEL_ERROR = 0x00000001 << 3;    //!<  Specifies the error level (recoverable).
    static const u32 LEVEL_FATAL = 0x00000001 << 4;    //!<  Specifies the error level (irrecoverable).
    static const u32 LEVEL_ALWAYS = 0x00000001 << 5;   //!<  Specifies the log level that is constantly displayed.

    //n1589: Optimized code for checking <tt>IsFlagSet</tt>.
/*!
    @cond  PRIVATE
*/
    static bool IsFlagSetStatic(u64 flag)
    {
#if NN_PIA_RELEASE
        NN_PIA_DUMMY_PARAM_IN_RELEASE(flag);
        return false;
#else
        if (GetGlobalObject() == NULL)
        {
            return false;
        }

        return GetGlobalObject()->IsFlagSet(flag);
#endif
    }
    //! @endcond

/*!
    @brief  The constructor. Objects are constructed with all bits of the log level bit mask set to 1.
*/
    Log(void);

/*!
    @brief  The constructor. A mask representing the log level can be specified as a parameter.

    @param[in] mask  Specifies a bit mask representing the log level.
*/
    explicit Log(u32 mask);

/*!
    @brief  The destructor.
*/
    ~Log();

/*!
    @brief  Sets the mask. This allows you to suppress log messages of a particular level.

    @details  This member function allows you to filter log information being output.
           Note that a log for which <tt>LEVEL_ALWAYS</tt> has been specified is
           always output regardless of the mask set by this member function.

    @param[in] mask  Specifies a bit mask representing the log level.
*/
    void SetMask(u32 mask)
    {
        m_mask = mask;
    }

/*!
    @brief  Indicates whether the specified flags are set.

    @details  The result of <tt>IsFlagSet(TRACE_FLAG_NEVER)</tt> is <tt>true</tt>.

    @return  Returns <tt>true</tt> if the particular flag is set, or <tt>false</tt> otherwise.

    @param[in] flag  Specifies the trace flags to query.
*/
    bool IsFlagSet(u64 flag) const
    {
#if NN_PIA_RELEASE
        NN_PIA_DUMMY_PARAM_IN_RELEASE(flag);
        return false;
#else
        return (m_mask & flag) == flag;
#endif
    }

/*!
    @cond  PRIVATE
    @brief  Writes text strings to the console. There is support for variable arguments.

    @param[in] mask  Specifies a bit mask representing the log level.
    @param[in] pStr  Specifies a string having the same format as that used by the standard C library function <tt>printf</tt>.
*/
    void Write(u32 mask, const char* pStr, ...);
    //! @endcond


/*!
    @cond  PRIVATE
    @brief  Writes caption text strings to the console.

    @param[in] mask  Specifies a bit mask representing the log level.
*/
    void WriteCaption(u32 mask);
    //! @endcond


/*!
    @brief  Enables and disables the time display.
    @details  When enabled, it displays the time passed from the base time in units of milliseconds, such as "[Pia:12345]..."
    @param[in] isPrint  Specify <tt>true</tt> to display the time.
*/
    void SetPrintTime(bool isPrint)
    {
        m_IsPrintTime = isPrint;
    }


/*!
    @brief  Gets whether time display is enabled or disabled.
    @return  Returns <tt>true</tt> if times are printed.
*/
    bool IsPrintTime() const
    {
        return m_IsPrintTime;
    }


/*!
    @brief  Resets the base time for the time display to the current time.
*/
    void ResetBaseTime()
    {
        m_BaseTime.SetNow();
    }


/*!
    @brief  Registers a global log object.

    @details  This function registers the specified <tt>Log</tt> object as a global log object.
           The <tt>PIA_LOG()</tt> macro outputs a log using the <tt>Log</tt> object registered by this function.

    @param[in] pLog  Specifies a pointer to the <tt>Log</tt> object to be registered.
*/
    static void SetGlobalObject(Log* pLog)
    {
        s_pLog = pLog;
    }

/*!
    @brief  Gets a pointer to the <tt>Log</tt> object set using <tt>SetGlobalObject()</tt>.

    @details  Calling <tt>nn::pia::common::Initialize()</tt> implicitly creates a <tt>Log</tt> instance
           internally and sets this instance as the global log object.

    @return  Returns a pointer to the instance registered as the global log object.
*/
    static Log* GetGlobalObject(void)
    {
        return s_pLog;
    }


private:
/*!
    @brief  This copy constructor is private.
*/
    Log(const Log&);

/*!
    @brief  This assignment operator is private.
*/
    Log& operator=(const Log&);


    // The pointer indicating the global object.
    static Log* s_pLog;

    // Mask settings are maintained for each <tt>Log</tt> instance.
    u32 m_mask;

    Time m_BaseTime;
    bool m_IsPrintTime;
};


/*!
  @brief  This macro is provided to output a simple log.

  @details  This macro uses the <tt>Log</tt> object set as the global log object to output log information.
         It automatically adds line headings and newline characters.

    @param[in] mask  Specifies a bit mask representing the log.
*/
#if NN_PIA_UNDER_DEVELOP
#define PIA_LOG(mask, ...)                                                       \
    \
{                                                                         \
        nn::pia::common::Log* pPiaLog = nn::pia::common::Log::GetGlobalObject(); \
        if (pPiaLog)                                                             \
        {                                                                        \
            pPiaLog->WriteCaption(mask);                                         \
            pPiaLog->Write(mask, __VA_ARGS__);                                   \
            pPiaLog->Write(mask, "\n");                                          \
        }                                                                        \
    \
}
#else
#define PIA_LOG(mask, ...) {}
#endif

/*!
  @brief  This macro is provided to output a log including filenames and function names.

  @details  This macro uses the <tt>Log</tt> object set as the global log object to output log information.
           It automatically adds line headings and newline characters.

  @param[in] mask  Specifies a bit mask representing the log.
*/
#if NN_PIA_UNDER_DEVELOP
#define PIA_LOG_EX(mask, ...)                                                               \
{                                                                                           \
        nn::pia::common::Log* pPiaLog = nn::pia::common::Log::GetGlobalObject();            \
        if (pPiaLog)                                                                        \
        {                                                                                   \
            pPiaLog->WriteCaption(mask);                                                    \
            pPiaLog->Write(mask, "%s %s:", PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_FUNC); \
            pPiaLog->Write(mask, __VA_ARGS__);                                              \
            pPiaLog->Write(mask, "\n");                                                     \
        }                                                                                   \
}
#else
#define PIA_LOG_EX(mask, ...) {}
#endif


// Macro to collect <tt>return</tt> from <tt>Result</tt> and log outputs.
// The log level is constantly <tt>LEVEL_ERROR</tt>. Only pass <tt>Result</tt> values
// that are equivalent to programming errors to this macro.
// This macro only returns the result in the release build.
#if NN_PIA_UNDER_DEVELOP
#define PIA_RETURN_RESULT_WITH_LOG(result, ...)                                                                      \
{                                                                                                                    \
        nn::pia::common::Log* pPiaLog = nn::pia::common::Log::GetGlobalObject();                                     \
        if (pPiaLog)                                                                                                 \
        {                                                                                                            \
            u32 logLevelError_ = nn::pia::common::Log::LEVEL_ERROR;                                                  \
            pPiaLog->WriteCaption(logLevelError_);                                                                   \
            pPiaLog->Write(logLevelError_, "[%s] %s %s:", #result, PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_FUNC);  \
            pPiaLog->Write(logLevelError_, __VA_ARGS__);                                                             \
            pPiaLog->Write(logLevelError_, "\n");                                                                    \
        }                                                                                                            \
        PIA_RETURN_RESULT(result);                                                                                   \
}
#else
#define PIA_RETURN_RESULT_WITH_LOG(result, ...) { return result(); }
#endif

}
}
}  // end of namespace nn::pia::common
