﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_MemoryLeakChecker.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  This class is used to automatically detect memory leaks inside the Pia library.

@date 2014-02-03  Improved the description of <tt>Set/Check</tt>.
*/
class MemoryLeakChecker
{
public:
/*!
@brief  Instantiates the object.
*/
    MemoryLeakChecker(void);


/*!
@brief  Destroys the object.
*/
    ~MemoryLeakChecker(void);


/*!
@brief  Sets the starting point to check for memory leaks.

@details  This API function is designed to be called prior to starting an initialization process on a Pia module that you suspect is leaking memory. Pia is designed to first initialize the <tt>common</tt> module, and then initialize the higher-level modules in sequence. If you suspect that higher-level modules, including the <tt><i>xxx</i></tt> module, are leaking memory, execute this API function prior to calling the <tt>nn::pia::<i>xxx</i>::Initialize</tt> function.
@details  Do not call this function between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions of each module.

@return  Returns a successful <tt>Result</tt> if the call completes without an error.
@retval ResultNotInitialized  Returned if the <tt>nn::pia::common::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Returned if this function is called at the wrong time. Call the <tt>nn::pia::common::EndSetup</tt> function at regular intervals. Programming error. Fix your program so that this error is not returned.

@see  Check





*/
    nn::Result Set(void);


/*!
@brief  Determines whether a memory leak has occurred.

@details  This API function is designed to be called after finalizing a Pia module that you suspect is leaking memory. Pia is designed to finalize modules in sequence, starting with the highest-level modules. If you suspect that higher-level modules, including the <tt><i>xxx</i></tt> module, are leaking memory, execute this API function after calling the <tt>nn::pia::<i>xxx</i>::Finalize</tt> function.
@details  Do not call this function between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions of each module.

@return  Returns a successful <tt>Result</tt> if the call completes without an error and no memory leak is suspected.
@retval ResultMemoryLeak  Returned if a memory leak is suspected. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  Returned if the <tt>nn::pia::common::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Returned if this function is called at the wrong time. Call the <tt>nn::pia::common::EndSetup</tt> function at regular intervals. Programming error. Fix your program so that this error is not returned.

@see  Set





*/
    nn::Result Check(void) const;


private:
/*!
@brief  This copy constructor is private.
*/
    MemoryLeakChecker(const MemoryLeakChecker&);

/*!
@brief  This assignment operator is private.
*/
    MemoryLeakChecker& operator=(const MemoryLeakChecker&);

    // Maintain the amount of available memory.
    u32 m_FreeSize;
};
}
}
}  // end of namespace nn::pia::common
