﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_ObjList.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_OffsetList.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  Represents a doubly linked list used to manage memory for storing objects.

@details  Before using the list, call the <tt>SetBuffer</tt> function and set the memory to use.

@param T  Specifies the type of object being managed.
*/
template <typename T>
class ObjList : public nn::pia::common::ListBase
{
protected:
    struct Node
    {
    public:
        Node()
          : m_Item()
        {
        }
        Node(const Node& n)
          : m_Item(n.m_Item)
        {
        }

    public:
        ListNode m_Link;
        T m_Item;
    };

    enum
    {
        c_LinkOffset = offsetof(Node, m_Link),
        c_ItemOffset = offsetof(Node, m_Item)
    };

public:
/*!
@brief  Represents the type of object being stored in the list.
*/
    typedef T ValueType;

    enum
    {
/*!
@brief  Specifies the size of the element management class.

@details  The class uses a buffer of the size given by <tt>c_NodeSize * (max_number_of_elements)</tt>.
*/
        c_NodeSize = sizeof(Node)
    };


/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    ObjList()
      : ListBase(), m_FreeList(), m_pBuffer(NULL), m_LimitNum(0)
    {
        m_FreeList.InitOffset(offsetof(Node, m_Link));
    }


/*!
@brief  Gets the required buffer size.

@return  Returns the required buffer size.
*/
    static u32 GetNeedBufferSize(u32 limitNum)
    {
        return limitNum * c_NodeSize;
    }


/*!
@brief  Sets the memory used to store the objects.

@details  Memory of the size specified by <tt>GetNeedBufferSize(limitNum)</tt> (that is, <tt>limitNum * c_NodeSize</tt>) is required.

@param limitNum  Specifies the maximum number of elements to store in the list.
@param pBuffer  Specifies the address of the memory used by the list.
*/
    void SetBuffer(u32 limitNum, void* pBuffer)
    {
        if (limitNum > 0)
        {
            if (pBuffer != NULL)
            {
                m_pBuffer = reinterpret_cast<Node*>(pBuffer);
                m_LimitNum = limitNum;
                for (u32 i = 0; i < limitNum; ++i)
                {
                    new (&m_pBuffer[i].m_Link) ListNode();
                    m_FreeList.PushBack(&m_pBuffer[i]);
                }
            }
            else
            {
                PIA_ASSERTMSG(false, "pBuf is null");
            }
        }
        else
        {
            PIA_ASSERTMSG(false, "limitNum[%d] must be larger than zero", limitNum);
        }
    }


/*!
@brief  Gets whether a buffer has been set.

@return  Returns <tt>true</tt> if a buffer has been set.
*/
    bool IsBufferReady() const
    {
        return m_pBuffer != NULL;
    }


/*!
@brief  Gets the maximum number of elements that can be stored in the list.

@return  Returns the maximum number of elements that can be stored in the list.
*/
    u32 MaxSize() const
    {
        return m_LimitNum;
    }


/*!
@brief  Gets whether the number of elements in the list has reached the maximum.

@return  Returns <tt>true</tt> if the number of elements in the list has reached the maximum.
*/
    bool IsFull() const
    {
        return m_FreeList.IsEmpty();
    }


/*!
@brief  Deletes all elements.

@details  The function calls the destructor on all stored elements.
*/
    void Clear()
    {
        ListNode* pNode = Terminator()->Next();
        while (pNode != Terminator())
        {
            ListNode* pN = pNode;
            pNode = pNode->Next();
            InitListNode(pN);
            m_FreeList.PushFront(ListNodeToNode(pN));
        }
        Init();
    }


/*!
@brief  Creates an element and appends it to the end of the list.

@details  If the number of elements being stored has reached the maximum, the release version returns <tt>NULL</tt>, and other versions assert.

@return  Returns the element that was added.
*/
    T* BirthBack()
    {
        if (IsFull())
        {
            PIA_ASSERTMSG(false, "buffer full");
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        PushBackNode(&(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and appends it to the end of the list.

@details  The function returns <tt>NULL</tt> if the number of elements being stored has reached the maximum.

@return  Returns the element that was added.
*/
    T* TryBirthBack()
    {
        if (IsFull())
        {
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        PushBackNode(&(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and adds it to the start of the list.

@details  If the number of elements being stored has reached the maximum, the release version returns <tt>NULL</tt>, and other versions assert.

@return  Returns the element that was added.
*/
    T* BirthFront()
    {
        if (IsFull())
        {
            PIA_ASSERTMSG(false, "buffer full");
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        PushFrontNode(&(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and adds it to the start of the list.

@details  The function returns <tt>NULL</tt> if the number of elements being stored has reached the maximum.

@return  Returns the element that was added.
*/
    T* TryBirthFront()
    {
        if (IsFull())
        {
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        PushFrontNode(&(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and inserts it in front of the specified element.

@details  If the number of elements being stored has reached the maximum, the release version returns <tt>NULL</tt>, and other versions assert.

@param pBasis  Specifies the base element.

@return  Returns the element that was added.
*/
    T* BirthBefore(const T* pBasis)
    {
        if (IsFull())
        {
            PIA_ASSERTMSG(false, "buffer full");
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        InsertBeforeNode(&(ObjToNode(pBasis)->m_Link), &(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and inserts it in front of the specified element.

@details  The function returns <tt>NULL</tt> if the number of elements being stored has reached the maximum.

@param pBasis  Specifies the base element.

@return  Returns the element that was added.
*/
    T* TryBirthBefore(const T* pBasis)
    {
        if (IsFull())
        {
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        InsertBeforeNode(&(ObjToNode(pBasis)->m_Link), &(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and inserts it after the specified element.

@details  If the number of elements being stored has reached the maximum, the release version returns <tt>NULL</tt>, and other versions assert.

@param pBasis  Specifies the base element.

@return  Returns the element that was added.
*/
    T* BirthAfter(const T* pBasis)
    {
        if (IsFull())
        {
            PIA_ASSERTMSG(false, "buffer full");
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        InsertAfterNode(&(ObjToNode(pBasis)->m_Link), &(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Creates an element and inserts it after the specified element.

@details  The function returns <tt>NULL</tt> if the number of elements being stored has reached the maximum.

@param pBasis  Specifies the base element.

@return  Returns the element that was added.
*/
    T* TryBirthAfter(const T* pBasis)
    {
        if (IsFull())
        {
            return NULL;
        }
        Node* pNode = m_FreeList.PopBack();
        new (&pNode->m_Item) T();
        InsertAfterNode(&(ObjToNode(pBasis)->m_Link), &(pNode->m_Link));
        return &(pNode->m_Item);
    }


/*!
@brief  Removes the specified element.

@details  Operations are undefined if the specified element is not in the list.

@param pObj  Specifies the element to remove.
*/
    void Erase(T* pObj)
    {
        Node* pNode = ObjToNode(pObj);
        pObj->~T();
        EraseNode(&(pNode->m_Link));
        m_FreeList.PushFront(pNode);
    }


/*!
@brief  Gets the element at the start of the list.

@return  Returns the element at the start of the list.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* Front()
    {
        ListNode* pLink = FrontNode();
        if (pLink == NULL)
        {
            return NULL;
        }
        return &(ListNodeToNode(pLink)->m_Item);
    }


/*!
@brief  Gets the last element in the list.

@return  Returns the last element in the list.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* Back()
    {
        ListNode* pLink = BackNode();
        if (pLink == NULL)
        {
            return NULL;
        }
        return &(ListNodeToNode(pLink)->m_Item);
    }


/*!
@brief  Gets the element before the specified element.

@param pObj  Specifies the base element.
Operations are undefined if the specified element is not in the list.

@return  Returns the element before the specified element.
Returns <tt>NULL</tt> if the first element in the list is specified as the base.
*/
    T* Prev(const T* pObj)
    {
        ListNode* pPrevLink = ObjToNode(pObj)->m_Link.Prev();
        if (pPrevLink == Terminator())
        {
            return NULL;
        }
        return &(ListNodeToNode(pPrevLink)->m_Item);
    }


/*!
@brief  Gets the element after the specified element.

@param pObj  Specifies the base element.
Operations are undefined if the specified element is not in the list.

@return  Returns the element after the specified element.
Returns <tt>NULL</tt> if the last element in the list is specified as the base.
*/
    T* Next(const T* pObj)
    {
        ListNode* pNextLink = ObjToNode(pObj)->m_Link.Next();
        if (pNextLink == Terminator())
        {
            return NULL;
        }
        return &(ListNodeToNode(pNextLink)->m_Item);
    }


/*!
@brief  Checks whether the specified element is in the list.

@param pObj  Specifies the element to check.

@return  Returns <tt>true</tt> if the element is in the list.
*/
    bool IsInclude(const T* pObj) const
    {
        return IsIncludeNode(&(ObjToNode(pObj)->m_Link));
    }


/*!
@brief  Represents an iterator for the <tt>ObjList</tt> class.
*/
    class Iterator
    {
        friend class ObjList<T>;

    private:
        Iterator(const ObjList<T>* pList, typename ObjList<T>::Node* pNode)
          : m_pList(pList), m_pNode(pNode)
        {
        }

    public:
        Iterator& operator++()
        {
            m_pNode = m_pList->ListNodeToNode(m_pNode->m_Link.Next());
            return *this;
        }

        bool operator==(const Iterator& x) const
        {
            return m_pNode == x.m_pNode;
        }
        bool operator!=(const Iterator& x) const
        {
            return m_pNode != x.m_pNode;
        }

        T& operator*()
        {
            return m_pNode->m_Item;
        }
        T* operator->()
        {
            return &(m_pNode->m_Item);
        }

    private:
        const ObjList<T>* m_pList;
        typename ObjList<T>::Node* m_pNode;
    };


/*!
@brief  Represents a <tt>const</tt> iterator for the <tt>ObjList</tt> class.
*/
    class ConstIterator
    {
        friend class ObjList<T>;

    private:
        ConstIterator(const ObjList<T>* pList, typename ObjList<T>::Node* pNode)
          : m_pList(pList), m_pNode(pNode)
        {
        }

    public:
        ConstIterator& operator++()
        {
            m_pNode = m_pList->ListNodeToNode(m_pNode->m_Link.Next());
            return *this;
        }

        bool operator==(const ConstIterator& x) const
        {
            return m_pNode == x.m_pNode;
        }
        bool operator!=(const ConstIterator& x) const
        {
            return m_pNode != x.m_pNode;
        }

        const T& operator*() const
        {
            return m_pNode->m_Item;
        }
        const T* operator->() const
        {
            return &(m_pNode->m_Item);
        }

    private:
        const ObjList<T>* m_pList;
        typename ObjList<T>::Node* m_pNode;
    };


/*!
@brief  Represents a <tt>robust</tt> iterator for the <tt>ObjList</tt> class.

@details  The iterator remains valid even if the element it points to is removed.
The iterator is no longer valid if an element is added or an element it does not point to is removed.
*/
    class RobustIterator
    {
        friend class ObjList<T>;

    private:
        RobustIterator(const ObjList<T>* pList, typename ObjList<T>::Node* pNode)
          : m_pList(pList), m_pNode(pNode)
        {
            m_pNextLink = m_pNode->m_Link.Next();
        }

    public:
        RobustIterator& operator++()
        {
            m_pNode = m_pList->ListNodeToNode(m_pNextLink);
            m_pNextLink = m_pNode->m_Link.Next();
            return *this;
        }

        bool operator==(const RobustIterator& x) const
        {
            return m_pNode == x.m_pNode;
        }
        bool operator!=(const RobustIterator& x) const
        {
            return m_pNode != x.m_pNode;
        }

        T& operator*()
        {
            return m_pNode->m_Item;
        }
        T* operator->()
        {
            return &(m_pNode->m_Item);
        }

    private:
        const ObjList<T>* m_pList;
        typename ObjList<T>::Node* m_pNode;
        ListNode* m_pNextLink;
    };


/*!
@brief  Gets an Iterator pointing to the element at the start of the list.

@return  Returns an iterator pointing to the element at the start of the list.
*/
    Iterator Begin() const
    {
        return Iterator(this, ListNodeToNode(Terminator()->Next()));
    }


/*!
@brief  Gets an <tt>Iterator</tt> object pointing to the element after the last element in the list.

@return  Returns an <tt>Iterator</tt> object pointing to the element after the last element in the list.
*/
    Iterator End() const
    {
        return Iterator(this, ListNodeToNode(Terminator()));
    }


/*!
@brief  Gets a <tt>const</tt> iterator pointing to the element at the start of the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element at the start of the list.
*/
    ConstIterator ConstBegin() const
    {
        return ConstIterator(this, ListNodeToNode(Terminator()->Next()));
    }


/*!
@brief  Gets a <tt>const</tt> iterator pointing to the element after the last element in the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element after the last element in the list.
*/
    ConstIterator ConstEnd() const
    {
        return ConstIterator(this, ListNodeToNode(Terminator()));
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element at the start of the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element at the start of the list.
*/
    RobustIterator RobustBegin() const
    {
        return RobustIterator(this, ListNodeToNode(Terminator()->Next()));
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.
*/
    RobustIterator RobustEnd() const
    {
        return RobustIterator(this, ListNodeToNode(Terminator()));
    }

private:
    static Node* ObjToNode(const T* pObj)
    {
        return reinterpret_cast<Node*>((reinterpret_cast<u32>(pObj)) - c_ItemOffset);
    }

    static Node* ListNodeToNode(const ListNode* pListNode)
    {
        return reinterpret_cast<Node*>((reinterpret_cast<u32>(pListNode)) - c_LinkOffset);
    }


private:
    OffsetList<Node> m_FreeList;
    Node* m_pBuffer;
    u32 m_LimitNum;
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
