﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_OffsetList.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_ListBase.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  Represents a doubly linked list that maintains a <tt>ListNode</tt> instance for each element member.

@details  Before using the list, you must specify the location of the <tt>ListNode</tt> instance in the element, as shown below.

class Node
          {
public:
static s32 GetListNodeOffset() {return offsetof(Node, m_ListNode);}
Private:
ListNode m_ListNode;
          };

OffsetList<Node> list;
list.InitOffset(Node::GetListNodeOffset());

@param T  Specifies the type of object being managed.


*/
template <typename T>
class OffsetList : public nn::pia::common::ListBase
{
public:
    typedef T Node;


/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    OffsetList(void)
      : ListBase(), m_Offset(-1)
    {
    }


/*!
@brief  Sets the offset from the start address of <tt>T</tt> to the <tt>ListNode</tt> member.

@details  You must call this function before use.

@param offset  Specifies the offset.
*/
    void InitOffset(s32 offset)
    {
        m_Offset = offset;
    }


/*!
@brief  Removes all elements from the list.
*/
    void Clear()
    {
        ClearNode();
    }


/*!
@brief  Adds an element to the end of the list.

@param pObj  Specifies the element to add.
*/
    void PushBack(T* pObj)
    {
        PushBackNode(ObjToListNode(pObj));
    }


/*!
@brief  Adds an element to the start of the list.

@param pObj  Specifies the element to add.
*/
    void PushFront(T* pObj)
    {
        PushFrontNode(ObjToListNode(pObj));
    }


/*!
@brief  Removes and returns the last element in the list.

@return  Returns the popped element.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* PopBack()
    {
        return ListNodeToObjWithNullCheck(PopBackNode());
    }


/*!
@brief  Removes and returns the first element on the list.

@return  Returns the popped element.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* PopFront()
    {
        return ListNodeToObjWithNullCheck(PopFrontNode());
    }


/*!
@brief  Adds an element in front of the specified element.

@param pBasis  Specifies the base element.
Operations are undefined if the specified element is not in the list.
@param pObj  Specifies the element to add.
*/
    void InsertBefore(const T* pBasis, T* pObj)
    {
        InsertBeforeNode(ObjToListNode(pBasis), ObjToListNode(pObj));
    }


/*!
@brief  Adds an element after the specified element.

@param pBasis  Specifies the base element.
Operations are undefined if the specified element is not in the list.
@param pObj  Specifies the element to add.
*/
    void InsertAfter(const T* pBasis, T* pObj)
    {
        InsertAfterNode(ObjToListNode(pBasis), ObjToListNode(pObj));
    }


/*!
@brief  Removes the specified element.

@details  Operations are undefined if the specified element is not in the list.

@param pObj  Specifies the element to remove.
*/
    void Erase(T* pObj)
    {
        EraseNode(ObjToListNode(pObj));
    }


/*!
@brief  Gets the element at the start of the list.

@return  Returns the element at the start of the list.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* Front() const
    {
        return ListNodeToObjWithNullCheck(FrontNode());
    }


/*!
@brief  Gets the last element in the list.

@return  Returns the last element in the list.
Returns <tt>NULL</tt> if the list is empty.
*/
    T* Back() const
    {
        return ListNodeToObjWithNullCheck(BackNode());
    }


/*!
@brief  Gets the element before the specified element.

@param pObj  Specifies the base element.
Operations are undefined if the specified element is not in the list.

@return  Returns the element before the specified element.
Returns <tt>NULL</tt> if the first element in the list is specified as the base.
*/
    T* Prev(const T* pObj) const
    {
        ListNode* pPrev = ObjToListNode(pObj)->Prev();
        if (pPrev == Terminator())
        {
            return NULL;
        }
        else
        {
            return ListNodeToObj(pPrev);
        }
    }


/*!
@brief  Gets the element after the specified element.

@param pObj  Specifies the base element.
Operations are undefined if the specified element is not in the list.

@return  Returns the element after the specified element.
Returns <tt>NULL</tt> if the last element in the list is specified as the base.
*/
    T* Next(const T* pObj) const
    {
        ListNode* pNext = ObjToListNode(pObj)->Next();
        if (pNext == Terminator())
        {
            return NULL;
        }
        else
        {
            return ListNodeToObj(pNext);
        }
    }


/*!
@brief  Checks whether the specified element is in the list.

@param pObj  Specifies the element to check.

@return  Returns <tt>true</tt> if the element is in the list.
*/
    bool IsInclude(const T* pObj) const
    {
        return IsIncludeNode(ObjToListNode(pObj));
    }


/*!
@brief  Rotates the list so the specified element is at the start of the list.

@param pObj  Specifies the element to put at the start.
Operations are undefined if the specified element is not in the list.
*/
    void Rotate(const T* pObj)
    {
        RotateNode(ObjToListNode(pObj));
    }


/*!
@brief  Represents an iterator for <tt>OffsetList</tt>.
*/
    class Iterator
    {
        friend class OffsetList<T>;

    private:
        Iterator(const OffsetList<T>* pList, T* pObj)
          : m_pList(pList), m_pObj(pObj)
        {
        }

    public:
        Iterator& operator++()
        {
            m_pObj = m_pList->ListNodeToObj(m_pList->ObjToListNode(m_pObj)->Next());
            return *this;
        }

        bool operator==(const Iterator& x) const
        {
            return m_pObj == x.m_pObj;
        }
        bool operator!=(const Iterator& x) const
        {
            return m_pObj != x.m_pObj;
        }

        T& operator*()
        {
            return *m_pObj;
        }
        T* operator->()
        {
            return m_pObj;
        }

        const OffsetList<T>* GetList() const
        {
            return m_pList;
        }

    private:
        const OffsetList<T>* m_pList;
        T* m_pObj;
    };


/*!
@brief  Represents a <tt>const</tt> iterator for <tt>OffsetList</tt>.
*/
    class ConstIterator
    {
        friend class OffsetList<T>;

    private:
        ConstIterator(const OffsetList<T>* pList, const T* pObj)
          : m_pList(pList), m_pObj(pObj)
        {
        }

    public:
        ConstIterator& operator++()
        {
            m_pObj = m_pList->ListNodeToObj(m_pList->ObjToListNode(m_pObj)->Next());
            return *this;
        }

        bool operator==(const ConstIterator& x) const
        {
            return m_pObj == x.m_pObj;
        }
        bool operator!=(const ConstIterator& x) const
        {
            return m_pObj != x.m_pObj;
        }

        const T& operator*()
        {
            return *m_pObj;
        }
        const T* operator->()
        {
            return m_pObj;
        }

        const OffsetList<T>* GetList() const
        {
            return m_pList;
        }

    private:
        const OffsetList<T>* m_pList;
        const T* m_pObj;
    };


/*!
@brief  Represents a <tt>robust</tt> iterator for <tt>OffsetList</tt>.

@details  The iterator remains valid even if the element it points to is removed.
The iterator is no longer valid if an element is added or an element it does not point to is removed.
*/
    class RobustIterator
    {
        friend class OffsetList<T>;

    private:
        RobustIterator(const OffsetList<T>* pList, T* pObj)
          : m_pList(pList), m_pObj(pObj)
        {
            m_pNextNode = m_pList->ObjToListNode(pObj)->Next();
        }

    public:
        RobustIterator& operator++()
        {
            m_pObj = m_pList->ListNodeToObj(m_pNextNode);
            m_pNextNode = m_pList->ObjToListNode(m_pObj)->Next();
            return *this;
        }

        bool operator==(const RobustIterator& x) const
        {
            return m_pObj == x.m_pObj;
        }
        bool operator!=(const RobustIterator& x) const
        {
            return m_pObj != x.m_pObj;
        }

        T& operator*()
        {
            return *m_pObj;
        }
        T* operator->()
        {
            return m_pObj;
        }

        const OffsetList<T>* GetList() const
        {
            return m_pList;
        }

    private:
        const OffsetList<T>* m_pList;
        T* m_pObj;
        ListNode* m_pNextNode;
    };


/*!
@brief  Gets an Iterator pointing to the element at the start of the list.

@return  Returns an iterator pointing to the element at the start of the list.
*/
    Iterator Begin() const
    {
        return Iterator(this, ListNodeToObj(Terminator()->Next()));
    }


/*!
@brief  Gets an <tt>Iterator</tt> object pointing to the element after the last element in the list.

@return  Returns an <tt>Iterator</tt> object pointing to the element after the last element in the list.
*/
    Iterator End() const
    {
        return Iterator(this, ListNodeToObj(Terminator()));
    }


/*!
@brief  Gets a <tt>const</tt> iterator pointing to the element at the start of the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element at the start of the list.
*/
    ConstIterator ConstBegin() const
    {
        return ConstIterator(this, ListNodeToObj(Terminator()->Next()));
    }


/*!
@brief  Gets a <tt>const</tt> iterator pointing to the element after the last element in the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element after the last element in the list.
*/
    ConstIterator ConstEnd() const
    {
        return ConstIterator(this, ListNodeToObj(Terminator()));
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element at the start of the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element at the start of the list.
*/
    RobustIterator RobustBegin() const
    {
        return RobustIterator(this, ListNodeToObj(Terminator()->Next()));
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.
*/
    RobustIterator RobustEnd() const
    {
        return RobustIterator(this, ListNodeToObj(Terminator()));
    }


private:
    s32 m_Offset;

    ListNode* ObjToListNode(const T* pObj) const
    {
        return reinterpret_cast<ListNode*>((reinterpret_cast<u32>(pObj)) + m_Offset);
    }

    T* ListNodeToObj(const ListNode* pNode) const
    {
        return reinterpret_cast<T*>((reinterpret_cast<u32>(pNode)) - m_Offset);
    }

    T* ListNodeToObjWithNullCheck(const ListNode* pNode) const
    {
        if (pNode)
            return reinterpret_cast<T*>((reinterpret_cast<u32>(pNode)) - m_Offset);
        else
            return NULL;
    }
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
