﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_OffsetTreeMap.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_TreeMap.h>

namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  Represents a tree map that maintains a <tt>TreeMapNode</tt> instance for each element member.
*/
template <typename Key_, typename Value_>
class OffsetTreeMap : public TreeMap<Key_>
{
public:
    typedef Key_ Key;
    typedef Value_ Value;
    typedef TreeMapNode<Key> Node;


/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    OffsetTreeMap()
      : TreeMap<Key>(), m_Offset(-1)
    {
    }


/*!
@brief  Sets the offset from the start address of <tt>Value</tt> to the <tt>TreeMapNode</tt> member.
@details  You must call this function before use.
@param[in] offset  Specifies the offset.
*/
    void InitOffset(s32 offset)
    {
        m_Offset = offset;
    }


/*!
@brief  Deletes all elements.
*/
    void Clear()
    {
        TreeMap<Key>::ClearNode();
    }


/*!
@brief  Finds the element that corresponds to the specified key.
@param[in] key  Specifies the key to find.
@return  Returns the element that corresponds to the key. Returns <tt>NULL</tt> if it does not exist.
*/
    Value* Find(const Key& key) const
    {
        return NodeToValueWithNullCheck(TreeMap<Key>::FindNode(key));
    }


/*!
@brief  Adds an element that corresponds to the specified key.
@details  If an element with the same key already exists, it is replaced.
@param[in] key  Specifies the key to add.
@param[in] pValue  Specifies the element to add.
@return  Returns the old element if an element was replaced. Otherwise, returns <tt>NULL</tt>.
*/
    Value* Insert(const Key& key, Value* pValue)
    {
        return NodeToValueWithNullCheck(TreeMap<Key>::InsertNode(key, ValueToNode(pValue)));
    }


/*!
@brief  Removes the element with the specified key.
@details  If an element with the specified key does not exist, nothing is erased.
@param[in] key  Specifies the key to erase.
@return  Returns the erased element if one was erased. Returns <tt>NULL</tt> if nothing was erased.
*/
    Value* Erase(const Key& key)
    {
        return NodeToValueWithNullCheck(TreeMap<Key>::EraseNode(key));
    }


/*!
@brief  Removes the specified element.
@details  Operations are undefined if anything other than an element of this <tt>TreeMap</tt> is specified.
@param[in] pValue  Specifies the element to erase.
*/
    void Erase(Value* pValue)
    {
        EraseNode(ValueToNode(pValue));
    }


/*!
@brief  Gets the element at the start of the list.
@return  Returns the element at the start of the list. Returns <tt>NULL</tt> if the list is empty.
*/
    Value* Front() const
    {
        return NodeToValueWithNullCheck(TreeMap<Key>::FrontNode());
    }


/*!
@brief  Gets the last element in the list.
@return  Returns the element at the end of the list.
*/
    Value* Back() const
    {
        return NodeToValueWithNullCheck(TreeMap<Key>::BackNode());
    }


/*!
@brief  Gets the element before the specified element.
@details  Operations are undefined if anything other than an element of this <tt>TreeMap</tt> is specified.
@param[in] pValue  Specifies the base element.
@return  Returns the previous element. Returns <tt>NULL</tt> if the first element in the list is specified as the base.
*/
    Value* Prev(const Value* pValue) const
    {
        return NodeToValueWithNullCheck(ValueToNode(pValue)->Prev());
    }


/*!
@brief  Gets the element after the specified element.
@details  Operations are undefined if anything other than an element of this <tt>TreeMap</tt> is specified.
@param[in] pValue  Specifies the base element.
@return  Returns the next element. Returns <tt>NULL</tt> if the last element in the list is specified as the base.
*/
    Value* Next(const Value* pValue) const
    {
        return NodeToValueWithNullCheck(ValueToNode(pValue)->Next());
    }


/*!
@brief  Determines whether the specified element is included in this <tt>TreeMap</tt>.
@param[in] pValue  Specifies the element to check for.
@return  Returns <tt>true</tt> if the element is in this <tt>TreeMap</tt>.
*/
    bool IsInclude(const Value* pValue) const
    {
        return IsIncludeNode(ValueToNode(pValue));
    }


/*!
@brief  Represents an iterator.
*/
    class Iterator
    {
        friend class OffsetTreeMap;

    private:
        Iterator(const OffsetTreeMap* pTreeMap, Node* pNode)
          : m_pTreeMap(pTreeMap), m_pNode(pNode)
        {
            m_pValue = m_pTreeMap->NodeToValueWithNullCheck(pNode);
        }

    public:
        Iterator& operator++()
        {
            m_pNode = m_pNode->Next();
            m_pValue = m_pTreeMap->NodeToValueWithNullCheck(m_pNode);
            return *this;
        }

        bool operator==(const Iterator& x) const
        {
            return m_pValue == x.m_pValue;
        }
        bool operator!=(const Iterator& x) const
        {
            return m_pValue != x.m_pValue;
        }

        Value& operator*()
        {
            return *m_pValue;
        }
        Value* operator->()
        {
            return m_pValue;
        }

    private:
        const OffsetTreeMap* m_pTreeMap;
        Value* m_pValue;
        Node* m_pNode;
    };


/*!
@brief  Represents a <tt>const</tt> iterator.
*/
    class ConstIterator
    {
        friend class OffsetTreeMap;

    private:
        ConstIterator(const OffsetTreeMap* pTreeMap, const Node* pNode)
          : m_pTreeMap(pTreeMap), m_pNode(pNode)
        {
            m_pValue = m_pTreeMap->NodeToValueWithNullCheck(pNode);
        }

    public:
        ConstIterator& operator++()
        {
            m_pNode = m_pNode->Next();
            m_pValue = m_pTreeMap->NodeToValueWithNullCheck(m_pNode);
            return *this;
        }

        bool operator==(const ConstIterator& x) const
        {
            return m_pValue == x.m_pValue;
        }
        bool operator!=(const ConstIterator& x) const
        {
            return m_pValue != x.m_pValue;
        }

        const Value& operator*()
        {
            return *m_pValue;
        }
        const Value* operator->()
        {
            return m_pValue;
        }

    private:
        const OffsetTreeMap* m_pTreeMap;
        const Value* m_pValue;
        const Node* m_pNode;
    };


/*!
@brief  Represents a <tt>robust</tt> iterator
@details  The iterator remains valid even if the element it points to is removed.
The iterator is no longer valid if an element is added or an element it does not point to is removed.
*/
    class RobustIterator
    {
        friend class OffsetTreeMap;

    private:
        RobustIterator(const OffsetTreeMap* pTreeMap, Node* pNode)
          : m_pTreeMap(pTreeMap)
        {
            m_pValue = m_pTreeMap->NodeToValueWithNullCheck(pNode);
            m_pNextNode = (pNode != NULL) ? pNode->Next() : NULL;
        }

    public:
        RobustIterator& operator++()
        {
            if (m_pNextNode != NULL)
            {
                m_pValue = m_pTreeMap->NodeToValue(m_pNextNode);
                m_pNextNode = m_pNextNode->Next();
            }
            else
            {
                m_pValue = NULL;
            }
            return *this;
        }

        bool operator==(const RobustIterator& x) const
        {
            return m_pValue == x.m_pValue;
        }
        bool operator!=(const RobustIterator& x) const
        {
            return m_pValue != x.m_pValue;
        }

        Value& operator*()
        {
            return *m_pValue;
        }
        Value* operator->()
        {
            return m_pValue;
        }

    private:
        const OffsetTreeMap* m_pTreeMap;
        Value* m_pValue;
        Node* m_pNextNode;
    };


/*!
@brief  Gets an <tt>Iterator</tt> object pointing to the element at the start of the list.

@return  Returns an <tt>Iterator</tt> object pointing to the element at the start of the list.
*/
    Iterator Begin() const
    {
        return Iterator(this, TreeMap<Key>::FrontNode());
    }


/*!
@brief  Gets an <tt>Iterator</tt> object pointing to the element after the last element in the list.

@return  Returns an <tt>Iterator</tt> object pointing to the element after the last element in the list.
*/
    Iterator End() const
    {
        return Iterator(this, NULL);
    }


/*!
@brief  Gets an <tt>Iterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.

@return  Returns an <tt>Iterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.
*/
    Iterator LowerBound(const Key& key) const
    {
        return Iterator(this, TreeMap<Key>::LowerBoundNode(key));
    }


/*!
@brief  Gets a <tt>ConstIterator</tt> object pointing to the element at the start of the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element at the start of the list.
*/
    ConstIterator ConstBegin() const
    {
        return ConstIterator(this, TreeMap<Key>::FrontNode());
    }


/*!
@brief  Gets a <tt>ConstIterator</tt> object pointing to the element after the last element in the list.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element after the last element in the list.
*/
    ConstIterator ConstEnd() const
    {
        return ConstIterator(this, NULL);
    }


/*!
@brief  Gets a <tt>ConstIterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.

@return  Returns a <tt>ConstIterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.
*/
    ConstIterator ConstLowerBound(const Key& key) const
    {
        return ConstIterator(this, TreeMap<Key>::LowerBoundNode(key));
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element at the start of the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element at the start of the list.
*/
    RobustIterator RobustBegin() const
    {
        return RobustIterator(this, TreeMap<Key>::FrontNode());
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element after the last element in the list.
*/
    RobustIterator RobustEnd() const
    {
        return RobustIterator(this, NULL);
    }


/*!
@brief  Gets a <tt>RobustIterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.

@return  Returns a <tt>RobustIterator</tt> object pointing to the element with the smallest key that is equal to or greater than the specified key.
*/
    RobustIterator RobustLowerBound(const Key& key) const
    {
        return RobustIterator(this, TreeMap<Key>::LowerBoundNode(key));
    }

private:
    s32 m_Offset;

    Node* ValueToNode(const Value* pObj) const
    {
        return reinterpret_cast<Node*>((reinterpret_cast<u32>(pObj)) + m_Offset);
    }

    Value* NodeToValue(const Node* pNode) const
    {
        return reinterpret_cast<Value*>((reinterpret_cast<u32>(pNode)) - m_Offset);
    }

    Value* NodeToValueWithNullCheck(const Node* pNode) const
    {
        if (pNode)
            return reinterpret_cast<Value*>((reinterpret_cast<u32>(pNode)) - m_Offset);
        else
            return NULL;
    }
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
