﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Scheduler.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_OffsetList.h>
#include <pia/common/common_Time.h>
#include <pia/common/common_CriticalSection.h>


namespace nn
{
namespace pia
{
namespace common
{

class Time;
class Job;
class BackgroundScheduler;

/*!
@brief  Represents a job scheduler.
@details  The scheduler object is used to asynchronously execute processes defined by the <tt>Job</tt> class.
This class is not thread-safe.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2012-07-03  Changed the background scheduler priority check.
@date 2012-06-15  The <tt>Dispatch</tt> function used to return <tt>nn::Result</tt>, but no longer returns a value.
@date 2012-05-21  Added a feature for running jobs in a different thread. The parameters and return values of <tt>CreateInstance</tt> have been changed accordingly.
@date 2012-04-04  Initial version.



*/
class Scheduler : public nn::pia::common::RootObject
{
public:
/*!
@brief  Specifies the default priority value threads used by the background scheduler.
*/
#if NN_PIA_CTR
    static const s32 DEFAULT_BACKGROUND_THREAD_PRIORITY = 24;
#elif NN_PIA_WIN || NN_PIA_NIN_WIN
    static const s32 DEFAULT_BACKGROUND_THREAD_PRIORITY = THREAD_PRIORITY_BELOW_NORMAL;
#elif NN_PIA_CAFE
    static const s32 DEFAULT_BACKGROUND_THREAD_PRIORITY = 24;
#else
#error  "Invalid platform";
#endif


/*!
@brief  Creates an instance (singleton pattern).
@details  Call this member function after calling <tt>nn::pia::common::BeginSetup</tt>, but before calling <tt>nn::pia::common::EndSetup</tt>.
@param[in] backgroundThreadPriority  Specifies the priority of the thread used by the background scheduler. Be sure to specify a priority lower than the thread used to call Pia member functions.
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  The <tt>common</tt> module has not been initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was not called between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates an argument is invalid. Programming error. Fix your program so that this error is not returned.





*/
    static nn::Result CreateInstance(s32 backgroundThreadPriority = DEFAULT_BACKGROUND_THREAD_PRIORITY);


/*!
@brief  Destroys the instance (singleton pattern).
@details  If this function is called when an instance has not been created, it does nothing and returns.


*/
    static void DestroyInstance();


/*!
@brief  Gets a pointer to the <tt>Scheduler</tt> instance (singleton pattern).
@return  Returns a <tt>NULL</tt> pointer if the instance cannot be created.

*/
    static Scheduler* GetInstance()
    {
        return s_pInstance;
    }


/*!
@brief  Dispatches jobs.
@details  Jobs are run by calling this function.

@param[in] timeout  Specifies the timeout value serving as the upper limit for job execution. The value is in milliseconds.
There is, however, no guarantee that jobs will occur within the timeout period.
A value of zero is interpreted as if no timeout was specified.

*/
    void Dispatch(u32 timeout = 0);


/*!
@cond PRIVATE
@brief  Registers a job.

@details  Do not call this function from another thread as it is called by a job.
*/
    void EntryJob(Job* pJob, bool isBackground);
    //! @endcond


/*!
@cond PRIVATE
@brief  Registers a job that waits for the next call to <tt>Dispatch</tt>.

@details  Do not call this function from another thread as it is called by a job.
*/
    void EntryJobNext(Job* pJob);
    //! @endcond


/*!
@cond PRIVATE
@brief  Removes a job.

@details  Do not call this function from another thread as it is called by a job.
*/
    void ResetJob(Job* pJob);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the time at which <tt>Dispatch</tt> was called.
*/
    const Time& GetDispatchedTime() const
    {
        return m_DispatchedTime;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the job-related <tt>CriticalSection</tt>.
*/
    CriticalSection* GetCriticalSection()
    {
        return &m_CriticalSection;
    }
    //! @endcond
/*!
@cond PRIVATE
@brief  Gets the number of jobs currently waiting.
*/
    u32 GetJobNum() const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets monitoring data.
*/
    void SetMonitoringData();
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;

private:
/*!
@brief  Instantiates the <tt>Scheduler</tt> object.
*/
    explicit Scheduler(s32 backgroundThreadPriority);

/*!
@brief  Destroys the <tt>Scheduler</tt>.
*/
    ~Scheduler();


    Scheduler(const Scheduler&);
    Scheduler& operator=(const Scheduler&);

    // A pointer to a singleton instance.
    static Scheduler* s_pInstance;


    //! The job list type.
    typedef OffsetList<Job> JobList;

    //! Job lists.
    JobList m_JobList;
    JobList m_CountBasedJobList;

    //! Specifies the dispatch time.
    Time m_DispatchedTime;

    //! The number of times <tt>Dispatch</tt> is called.
    u32 m_DispatchCalls;

    mutable CriticalSection m_CriticalSection;

    BackgroundScheduler* m_pBackgroundScheduler;
};
}
}
}  // end of namespace nn::pia::common
