﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_SignatureSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  This class maintains signature setting information.

@date 2012-11-30  Initial version.
*/
class SignatureSetting
{
public:
    static const u32 SIGNATURE_KEY_DATA_SIZE_MAX = 32;  //!< Maximum number of bytes for the key data used for the signature.

/*!
@brief  Indicates the type of signature algorithm.
*/
    enum Mode
    {
        MODE_NOTHING = 0,  //!< No signature.
        MODE_HMAC_MD5      //!< HMAC-MD5 signature.
    };


/*!
@brief  Instantiates an object. Initializes the object instance using the signature algorithm and key data specified.

@param[in] mode  Specifies the type of signature algorithm.
@param[in] cpKeyData  Specifies the address of the signature data. Data being referenced must be maintained until this instance is destroyed or reinitialized.
@param[in] keySize  Specifies the size of the key data, in bytes.
*/
    SignatureSetting(Mode mode, const void* cpKeyData, size_t keySize);


/*!
@brief  Instantiates an object. Initializes the object using the predefined algorithm (HMAC-MD5) and the specified key data.

@param[in] cpKeyData  Specifies the address of the signature data. Data being referenced must be maintained until this instance is destroyed or reinitialized.
@param[in] keySize  Specifies the size of the key data, in bytes.
*/
    SignatureSetting(const void* cpKeyData, size_t keySize);


/*!
@brief  Instantiates an object. Initializes the instance without a signature.
*/
    SignatureSetting()
      : m_Mode(MODE_NOTHING), m_cpKeyData(NULL), m_KeySize(0)
    {
    }


/*!
@brief  Sets signature settings.

@param[in] mode  Specifies the type of signature algorithm.
@param[in] cpKeyData  Specifies the address of the signature data. Data being referenced must be maintained until this instance is destroyed or reinitialized.
@param[in] keySize  Specifies the size of the key data, in bytes.

@return  Returns a <tt>Result</tt> that indicates success if configured correctly.
@retval ResultInvalidArgument  Indicates an invalid argument. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Set(Mode mode, const void* cpKeyData, size_t keySize);


/*!
@brief  Gets the signature algorithm that is currently specified.
@return  Returns the signature algorithm that is currently specified.
*/
    Mode GetMode() const
    {
        return m_Mode;
    }


/*!
@brief  Gets the address of the key data that is currently specified.
@return  Returns the address of the key data that is currently specified.
*/
    const void* GetKeyData() const
    {
        return m_cpKeyData;
    }


/*!
@brief  Gets the size of the key data that is currently specified.
@return  Returns the address of the key data that is currently specified.
*/
    size_t GetKeySize() const
    {
        return m_KeySize;
    }

    virtual void Trace(u64 flag) const;

/*!
@brief  Specifies an instance of the signature setting class that has no signature.
*/
    static const SignatureSetting NO_SIGNATURE;

private:
    Mode m_Mode;
    const void* m_cpKeyData;
    size_t m_KeySize;
};
}
}
}  // end of namespace nn::pia::common
