﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_SignatureSettingWithKeyBuffer.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>
#include <pia/common/common_SignatureSetting.h>

namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  This class represents signature setting information and includes an internal signature key buffer.

@param KeySize  Specifies the key size (in bytes). A buffer of this size is configured.

@date 2012-11-30  Initial version.
*/
template <size_t KeySize>
class SignatureSettingWithKeyBuffer : public SignatureSetting
{
public:
/*!
@brief  Instantiates an object. Initializes the object instance using the specified signature algorithm.

@details  Write the key to the address retrieved by using the <tt>GetKeyBufferPrt</tt> function.

@param[in] mode  Specifies the type of signature algorithm.
*/
    explicit SignatureSettingWithKeyBuffer(Mode mode)
      : SignatureSetting(mode, m_KeyBuffer, KeySize)
    {
        ClearKeyBuffer();
    }


/*!
@brief  Instantiates an object. The instance is initialized using the predefined signature algorithm (HMAC-MD5).

@details  Write the key to the address retrieved by using the <tt>GetKeyBufferPrt</tt> function.
*/
    SignatureSettingWithKeyBuffer()
      : SignatureSetting(MODE_HMAC_MD5, m_KeyBuffer, KeySize)
    {
        ClearKeyBuffer();
    }


/*!
@brief  Gets a pointer to the key data buffer.

@return  Returns a pointer to the key data buffer.
*/
    u8* GetKeyBufferPtr()
    {
        return m_KeyBuffer;
    }


/*!
@brief  Sets the signature algorithm.

@param[in] mode  Specifies the type of signature algorithm.

@return  Returns a <tt>Result</tt> that indicates success if configured correctly.
@retval ResultInvalidArgument  Indicates an invalid argument. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetMode(Mode mode)
    {
        return Set(mode, m_KeyBuffer, KeySize);
    }


    // Clears the content of the key data buffer.
    void ClearKeyBuffer()
    {
        for(size_t i=0; i<KeySize; ++i)
        {
            m_KeyBuffer[i] = 0;
        }
    }


private:
    // Because <tt>m_cpKeyData</tt> cannot specify <tt>m_KeyBuffer</tt> after this function is used, we have made sure it cannot be called externally.
    // However, it can be called incorrectly by accessing it as <tt>SignatureSetting*</tt>.
    nn::Result Set(Mode mode, const void* cpKeyData, size_t keySize)
    {
        return SignatureSetting::Set(mode, cpKeyData, keySize);
    }

private:
    u8 m_KeyBuffer[KeySize];
};
}
}
}  // end of namespace nn::pia::common
