﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_SimpleContainer.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@cond PRIVATE
@brief  Represents a simple container having a fixed size.
*/
template <typename T, int N>
class SimpleContainer : public nn::pia::common::RootObject
{
public:
    typedef T* Iterator;             //!< Defines a non-<tt>const</tt> iterator type.
    typedef const T* ConstIterator;  //!<   Defines a <tt>const</tt> iterator type.

/*!
@brief  Gets the iterator for the instance at the start of the container. (<tt>const</tt> version)
If there are no valid elements in the container, this function returns the same iterator as the one returned by the <tt>End</tt> function.
@return  Returns an iterator for the instance at the start of the container.
@see  End
*/
    ConstIterator Begin(void) const
    {
        return &(m_Buffer[0]);  // This is fine, because when there are no elements in the container, the <tt>End</tt> function returns <tt>&(m_Buffer[0])</tt>.
    }


/*!
@brief  Gets the iterator for the instance at the start of the container (non-<tt>const</tt> version).
If there are no valid elements, this function returns the same iterator as the one returned by the <tt>End</tt> function.
@return  Returns an iterator for the instance at the start of the container.
@see  End
*/
    Iterator Begin(void)
    {
        return &(m_Buffer[0]);  // This is fine, because when there are no elements in the container, the <tt>End</tt> function returns <tt>&(m_Buffer[0])</tt>.
    }


/*!
@brief  Gets the iterator for the instance at the end of the container (<tt>const</tt> version).
@return  Returns the iterator for the instance at the end of the container.
@see  Begin
*/
    ConstIterator End(void) const
    {
        return &(m_Buffer[m_Elements]);
    }


/*!
@brief  Gets the iterator for the instance at the end of the container (non-<tt>const</tt> version).
@return  Returns the iterator for the instance at the end of the container.
@see  Begin
*/
    Iterator End(void)
    {
        return &(m_Buffer[m_Elements]);
    }


/*!
@brief  Instantiates the object.
*/
    SimpleContainer(void)
      : m_Elements(0),
        m_Buffer()
    {
    }


/*!
@brief  This is the copy constructor.
*/
    SimpleContainer(const SimpleContainer& rhs)
    {
        *this = rhs;
    }


/*!
@brief  Destroys the object.
*/
    virtual ~SimpleContainer(void)
    {
    }


/*!
@brief  Adds an element to the container.

@return  Returns a <tt>Result</tt> value for which <tt>IsSuccess</tt> returns <tt>true</tt> if added successfully.
@retval ResultBufferIsFull  Specifies that no more elements can be added because the container is full.
@see  IsFull, Clear, Erase
*/
    nn::Result Add(const T& t)
    {
        if (IsFull())
        {
            PIA_RETURN_RESULT(ResultBufferIsFull);
        }
        else
        {
            m_Buffer[m_Elements++] = t;
            return ResultSuccess();
        }
    }


/*!
@brief  Removes the element pointed to by the iterator from the container.
Elements after the element removed are shifted toward the front.
A bit more CPU cost is therefore incurred when dealing with large elements or removing an element near the front.

@param[in] it  Specifies a valid iterator retrieved from this container.
@return  Returns an iterator pointing to the next element after the one removed. Returns the same iterator as the <tt>End</tt> function if the last element was removed.
@see  Add

*/
    Iterator Erase(Iterator it)
    {
        if ((Begin() <= it) && (it < End()))
        {
            // Shift subsequent elements.
            for (Iterator i = it; (i + 1) != End(); ++i)
            {
                *i = *(i + 1);
            }
            PIA_ASSERT(m_Elements > 0);
            --m_Elements;
            // Because the elements were shifted to the front, the removed location becomes the location of the next element in the container.
            // Fortunately, deleting the last element makes this location the same as what the <tt>End</tt> function returns.
            return it;
        }
        else
        {
            // Specifies an invalid iterator. (It might be out of range.) Does nothing.
            return End();
        }
    }

/*!
@brief  Removes all elements in the container.

@see  IsFull, Clear
*/
    void Clear(void)
    {
        m_Elements = 0;
    }


/*!
@brief  Determines whether the container is empty.

@return  Returns <tt>true</tt> if the container is empty, and <tt>false</tt> if there are any elements in it.
@see  IsFull, GetSize
*/
    bool IsEmpty(void) const
    {
        return m_Elements == 0;
    }


/*!
@brief  Determines whether the container is full.

@return  Returns <tt>true</tt> if no more elements can be added; returns <tt>false</tt> otherwise.
@see  IsEmpty, GetSize
*/
    bool IsFull(void) const
    {
        return m_Elements == N;
    }


/*!
@brief  Gets the number of elements in the container.

@return  Returns the number of valid elements in the container.
@see  IsEmpty, IsFull
*/
    size_t GetSize(void) const
    {
        return m_Elements;
    }


/*!
@brief  This is the assignment operator.

@return  Returns a reference to this instance.
*/
    SimpleContainer& operator=(const SimpleContainer& rhs)
    {
        m_Elements = rhs.m_Elements;

        for (int i = 0; i < N; ++i)
        {
            m_Buffer[i] = rhs.m_Buffer[i];
        }

        return *this;
    }


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const
    {
        PIA_TRACE(flag, "----- SimpleContainer::Trace() called. -----");
        PIA_TRACE(flag, "Number of elements: %d", m_Elements);
        PIA_TRACE(flag, "----- SimpleContainer::Trace() end.    -----");
    }


private:
    // Specifies the number of valid elements included in the container.
    size_t m_Elements;

    // Represents the element instances.
    T m_Buffer[N];
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
