﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_StationAddress.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once


#include <pia/common/common_definitions.h>

#include <pia/common/common_InetAddress.h>


namespace nn
{
namespace pia
{
namespace common
{

/*!
@brief  Represents an address used to uniquely identify stations.
@if CTR_DOC
@date 2014-05-29  Made the <tt>StationAddress::GetUdsNodeId</tt> and <tt>StationAddress::SetUdsNodeId</tt> functions private.
@endif
@if CAFE_DOC
@date 2013-07-11  Made the <tt>StationAddress::GetUdsNodeId</tt> and <tt>StationAddress::SetUdsNodeId</tt> functions private.
@endif
@date 2013-05-17  Changed the documentation because of the deprecation of <tt>StationAddress::GetUdsNodeId</tt> and <tt>StationAddress::SetUdsNodeId</tt>.
@date 2013-02-22  Added greater than and less than comparison functions and operators.
@date 2012-04-23  Removed an unneeded destructor.
@date 2012-04-19  Initial version.
*/
class StationAddress : public nn::pia::common::RootObject
{
public:
    static const size_t SERIALIZE_SIZE_MAX =
        common::InetAddress::SERIALIZE_SIZE_MAX + sizeof(u16);  // m_Extension

/*!
@brief  Instantiates the object.
*/
    StationAddress(void);


/*!
@brief  This is the copy constructor.
*/
    StationAddress(const StationAddress& rhs);


/*!
@cond PRIVATE
@if CTR_DOC
@brief  Gets a UDS node ID. [Deprecated]

@details  This function is deprecated. It will be removed in a future release.
Although this function is retained for backward compatibility, use <tt>@ref local::UdsNode::ConvertStationAddressToUdsNodeId</tt> and other conversion functions.

This function can be used to retrieve Node IDs from a valid <tt>StationAddress</tt>. Use <tt>UdsNodeId</tt> for local communication with host migration disabled and <tt>TransportNodeId</tt> when host migration is enabled.

@return  Returns the node ID.

@see  local::UdsNode::ConvertStationAddressToUdsNodeId, local::UdsNode::ConvertStationAddressToTransportNodeId
@endif

*/
    u16 GetUdsNodeId(void) const
    {
        return m_Extension;
    }
    //! @endcond

/*!
@cond PRIVATE
@if CTR_DOC
@brief  Sets a UDS node ID. [Deprecated]

@details  This function is deprecated. It will be removed in a future release.
Although this function is retained for backward compatibility, use <tt>@ref local::UdsNode::ConvertUdsNodeIdToStationAddress</tt> and other conversion functions.

@param[in] nodeId  Specifies the UDS node ID.
@return  This function always succeeds.

@see  local::UdsNode::ConvertUdsNodeIdToStationAddress, local::UdsNode::ConvertTransportNodeIdToStationAddress
@endif
*/
    nn::Result SetUdsNodeId(u16 nodeId);
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the extension ID.

@details  This function can be used to retrieve the extension IDs from a valid <tt>StationAddress</tt>. Use <tt>UdsNodeId</tt> for local communication with host migration disabled and <tt>TransportNodeId</tt> when host migration is enabled.

@return  Returns the extension ID.

*/
    u16 GetExtensionId(void) const
    {
        return m_Extension;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the extension ID.

@param[in] extensionId  Specifies the extension ID to set.
@return  This function always succeeds.
*/
    nn::Result SetExtensionId(u16 extensionId);
    //! @endcond


/*!
@brief  Gets a <tt>const</tt> reference to an <tt>nn::pia::common::InetAddress</tt> instance.

@return  Returns a <tt>const</tt> reference to an <tt>nn::pia::common::InetAddress</tt> instance.
*/
    const nn::pia::common::InetAddress& GetInetAddress(void) const
    {
        return m_Address;
    }


/*!
@brief  Gets a reference to an <tt>nn::pia::common::InetAddress</tt> instance.

@return  Returns a reference to an <tt>nn::pia::common::InetAddress</tt> instance.
*/
    nn::pia::common::InetAddress& GetInetAddress(void)
    {
        return m_Address;
    }


/*!
@brief  Sets the <tt>nn::pia::common::InetAddress</tt> instance.

@param[in] addr  Specifies the Internet address to set.
@return  This function always succeeds.
*/
    nn::Result SetInetAddress(const nn::pia::common::InetAddress& addr);


/*!
@brief  Clears the <tt>StationAddress</tt>.
*/
    void Clear(void);


/*!
@brief  Gets the length of the object as a serialized byte array. The value is in bytes.

@return  Returns the length of the serialized data. The value is in bytes.
@see  Serialize, Deserialize
*/
    size_t GetSerializedSize(void) const;


/*!
@brief  Serializes the object.

@param[out] pBuffer  Pointer to a buffer for holding the serialized data.
@param[out] pDataLen  Holds the length of the serialized data that was written to the buffer. The value is in bytes.
@param[in] bufferSize  Specifies the size of the buffer that was specified in <span class="argument">pBuffer</span>.
@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function will return <tt>true</tt>. Your application's implementation must ensure that this function does not return any errors.
@retval ResultInvalidArgument  Indicates that an argument is invalid. (For example, a NULL pointer was specified.) The function also returns this error if <span class="argument">pBuffer</span> or the size specified in <span class="argument">bufferSize</span> is too small. Programming error. Fix your program so that this error is not returned.
@see  Deserialize, GetSerializedSize
*/
    nn::Result Serialize(bit8* pBuffer, size_t* pDataLen, size_t bufferSize) const;


/*!
@brief  Restores an object from serialized data.

@param[in] pData  Pointer to the serialized data.

@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function will return <tt>true</tt>. Your application's implementation must ensure that this function does not return any errors.
@retval ResultInvalidArgument  Indicates that an argument is invalid. (For example, a NULL pointer was specified.) Programming error. Fix your program so that this error is not returned.
@see  Serialize, GetSerializedSize
*/
    nn::Result Deserialize(const bit8* pData);


/*!
@cond PRIVATE
@brief  Determines whether the specified address is valid.

@return  <tt>true</tt> if the address is valid, and <tt>false</tt> otherwise.
*/
    bool IsValid(void) const;
    //! @endcond


/*!
@brief  This is an equality operator. Determines whether two <tt>StationAddress</tt> objects are equal.

@return  Returns <tt>true</tt> if the objects are equal, and <tt>false</tt> otherwise.
*/
    bool operator==(const StationAddress& rhs) const;


/*!
@brief  This is an equality operator. Determines whether two <tt>StationAddress</tt> objects are not equal.

@return  Returns <tt>true</tt> if the objects are not equal, and <tt>false</tt> otherwise.
*/
    bool operator!=(const StationAddress& rhs) const
    {
        return !(*this == rhs);
    }


/*!
@brief  This is the assignment operator.

@return  Returns a reference to the <tt>this</tt> object.
*/
    StationAddress& operator=(const StationAddress& rhs);


/*!
@brief  Comparison operator.

@return  <tt>true</tt> if this object is smaller.
*/
    bool operator<(const StationAddress& rhs) const;


/*!
@brief  Comparison operator.

@return  <tt>true</tt> if this object is larger.
*/
    bool operator>(const StationAddress& rhs) const;


/*!
@brief  Comparison function (greater than/less than/equal).
@param[in] a  The left-hand side object to compare.
@param[in] b  The right-hand side object to compare.
@return  Returns <tt>-1</tt> if <tt>b</tt> is less than <tt>a</tt>, <tt>1</tt> if <tt>b</tt> is greater than <tt>a</tt>, and <tt>0</tt> if they are equal.
*/
    static int Compare(const StationAddress& a, const StationAddress& b);

/*!
@cond PRIVATE
@brief  Gets the address string.

@param[in] pString  Pointer to the instance of the string class.
*/
    void GetAddressString(String* pString) const;
    //! @endcond

/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

private:
    common::InetAddress m_Address;
    u16 m_Extension;  // For expansion. At the current time, stores the UDS node ID.
};
}
}
}  // end of namespace nn::pia::common
