﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_StepSequenceJob.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#include <pia/common/common_RootObject.h>
#include <pia/common/common_Job.h>
#include <pia/common/common_Trace.h>


namespace nn
{
namespace pia
{
namespace common
{

#define NN_PIA_JOBSTEP(STEP) Step(static_cast<StepMethod>(&STEP), #STEP)


/*!
@cond PRIVATE
@brief  Represents the step execution jobs of the base class.

@details  Using this class enables you to execute jobs function-by-function.

*/
class StepSequenceJob : public nn::pia::common::Job
{
public:
    typedef Job::ExecuteResult (StepSequenceJob::*StepMethod)();

    class Step : public nn::pia::common::RootObject
    {
    public:
        Step(void)
          : m_pStepMethod(NULL), m_pName(NULL)
        {
        }
        Step(StepMethod pStepMethod, const char* pName)
          : m_pStepMethod(pStepMethod), m_pName(pName)
        {
        }

        virtual ~Step(void)
        {
        }

        const char* GetName(void) const
        {
            return m_pName ? m_pName : "(no name)";
        }

        Job::ExecuteResult Execute(StepSequenceJob* pJob)
        {
            return (pJob->*m_pStepMethod)();
        }

        bool operator==(const Step& rhs) const;
        bool operator!=(const Step& rhs) const { return !(*this == rhs); }

        virtual void Trace(u64 flag) const
        {
            PIA_TRACE(flag, "StepSequenceJob::Step::Trace() called. (%s)", GetName());
        }

    private:
        StepMethod m_pStepMethod;
        const char* m_pName;
    };


/*!
@brief  Instantiates an object.
*/
    StepSequenceJob(void);


/*!
@brief  Destroys the object.
*/
    virtual ~StepSequenceJob(void);


    virtual void Reset(bool isCheck = true);


    void CancelStepSequence()
    {
        m_IsCancelStepSequence = true;
    }

    virtual void CancelCleanup()
    {
    }

/*!
@brief  Registers a function to be sequentially executed.

@details  This function makes the settings necessary to execute functions as a batch process.

@param[in] step  Specifies the <tt>Step</tt> instance containing the function to execute next.
This instance is usually created by the <tt>NN_PIA_JOBSTEP</tt> macro.
*/
    void SetStep(const Step& step)
    {
        m_NextStep = step;
    }

    void WaitForCompletion(u32 sleepTime);

/*!
@brief  Sets tracing flags.

@param[in] flags  Specifies tracing flags.
*/
    void SetTraceFlags(u64 flags)
    {
        m_TraceFlags = flags;
    }


/*!
@brief  Gets tracing flags.
*/
    u64 GetTraceFlags(void) const
    {
        return m_TraceFlags;
    }

/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

protected:
    // Override.
    virtual Job::ExecuteResult ExecuteCore();

    // Returns the instance registered with the <tt>SetStep</tt> function.
    const Step& GetStep(void) const
    {
        return m_NextStep;
    }

private:
    // The instance bundled in the next job to be executed.
    Step m_NextStep;

    volatile bool m_IsCancelStepSequence;

    // The trace flag for this instance.
    u64 m_TraceFlags;
};
//! @endcond
}
}
}  // end of namespace nn::pia::common
