﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_String.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  Class for representing strings.

@details  This class includes a 128-byte buffer used to store strings.

@date 2012-04-19  Initial version.
*/
class String : public nn::pia::common::RootObject
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    String()
    {
        m_Buffer[0] = '\0';
    }


/*!
@brief  This is the copy constructor.

@param[in] rhs  Specifies the string to copy.
*/
    String(const String& rhs);


/*!
@brief  Instantiates the object built from a C string.

@param[in] pStr  Specifies a pointer to a C string.
*/
    String(const char* pStr);


/*!
@brief  Gets a pointer to the string buffer.

@return  Specifies a pointer to the string buffer.
*/
    const char* GetCharArray() const
    {
        return m_Buffer;
    }


/*!
@brief  This is an equality operator. It determines whether two <tt>String</tt> objects are equal.

@param[in] rhs  Specifies the string to compare.

@return  Returns <tt>true</tt> if the objects are equal, and <tt>false</tt> otherwise.
*/
    bool operator==(const String& rhs) const;


/*!
@brief  This is an equality operator. It determines whether two <tt>String</tt> objects are not equal.

@param[in] rhs  Specifies the string to compare.

@return  Returns <tt>true</tt> if the objects are not equal, and <tt>false</tt> otherwise.
*/
    bool operator!=(const String& rhs) const
    {
        return !(*this == rhs);
    }


/*!
@brief  This is the assignment operator.

@param[in] rhs  Specifies the string to assign.

@return  Returns a reference to this instance.
*/
    String& operator=(const String& rhs);


/*!
@brief  Builds a string according to the format.

@param[in] pFormatString  Specifies the format string.
@param[in] varg  Specifies a variable argument list.

@return  Returns the length of the string after writing it.
*/
    s32 FormatV(const char* pFormatString, va_list varg);


/*!
@brief  Builds a string according to the format.

@param[in] pFormatString  Specifies the format string.
@param[in] ...  Specifies variable arguments.

@return  Returns the length of the string after writing it.
*/
    s32 Format(const char* pFormatString, ...);


/*!
@brief  Calculates the length of a string.

@return  Specifies the length of the string.
*/
    u32 StrLen() const;


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

private:
    enum
    {
        c_Length = 128
    };

    char m_Buffer[c_Length];
};
}
}
}  // end of namespace nn::pia::common
