﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_StringStream.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>
#include <pia/common/common_RootObject.h>

#define NN_PIA_T(x) x

//n2788: Added 0-padded hexadecimal display.
#define NN_PIA_ENABLE_ZERO_FILL_HEX 1

namespace nn
{
namespace pia
{
namespace common
{

#define NN_PIA_STRINGSTREAM_INITIAL_SIZE 128

//n2619: Determine whether to enable processing when a global <tt>new</tt> is invoked outside <tt>BeginSetup</tt> and <tt>EndSetup</tt>. (Basically, always set to 0.)
//
#define PIA_ENABLE_RESIZE_BUFFER 0

/*!
@cond PRIVATE
@brief  Represents the string stream class.
Concatenates strings using the manipulator (<<).
*/
class StringStream : public RootObject
{
public:
/*!
@brief  Instantiates the object.
*/
    StringStream(void);

/*!
@brief  Destroys the object.
*/
    virtual ~StringStream(void);

    // Returns the stream length.
    u32 GetLength() const;

    // Returns the <tt>m_szBuffer</tt> size (capacity).
    u32 GetSize() const
    {
        return m_uiSize;
    }

    const char* CStr() const
    {
        return m_szBuffer;
    }

    void Clear();

    // Streaming of basic types.
    StringStream& operator<<(const char* szText);
    StringStream& operator<<(const void* pPointer);
    StringStream& operator<<(bool bValue);
    StringStream& operator<<(f64 dValue);
    StringStream& operator<<(f32 fValue);

    StringStream& operator<<(u8 uiValue)
    {
        return StreamNumber(uiValue);
    }
    StringStream& operator<<(s8 iValue)
    {
        return StreamNumber(iValue);
    }
    StringStream& operator<<(u16 uiValue)
    {
        return StreamNumber(uiValue);
    }
    StringStream& operator<<(s16 iValue)
    {
        return StreamNumber(iValue);
    }
    StringStream& operator<<(u32 uiValue)
    {
        return StreamNumber(uiValue);
    }
    StringStream& operator<<(s32 iValue)
    {
        return StreamNumber(iValue);
    }

    StringStream& operator<<(u64 uiValue);
    StringStream& operator<<(s64 uiValue);

    StringStream& operator<<(const StringStream& oSS);

    // Manipulators.
    StringStream& operator<<(StringStream& (*pf)(StringStream&))
    {
        return (*pf)(*this);
    }

#if !NN_PIA_ENABLE_ZERO_FILL_HEX
    StringStream& Manip_hex()
    {
        m_Hex = true;
        return *this;
    }
    StringStream& Manip_dec()
    {
        m_Hex = false;
        return *this;
    }
#else   // !NN_PIA_ENABLE_ZERO_FILL_HEX
    StringStream& Manip_hex()
    {
        m_Hex = true;
        m_ZeroHex = false;
        return *this;
    }
    StringStream& Manip_dec()
    {
        m_Hex = false;
        m_ZeroHex = false;
        return *this;
    }
    StringStream& Manip_zerohex()
    {
        m_Hex = true;
        m_ZeroHex = true;
        return *this;
    }
#endif  // !NN_PIA_ENABLE_ZERO_FILL_HEX

    StringStream& Manip_showbase()
    {
        m_ShowBase = true;
        return *this;
    }
    StringStream& Manip_noshowbase()
    {
        m_ShowBase = false;
        return *this;
    }

    StringStream& Manip_boolalpha()
    {
        m_BoolAlpha = true;
        return *this;
    }
    StringStream& Manip_noboolalpha()
    {
        m_BoolAlpha = false;
        return *this;
    }

    const char& operator[](u32 uiPos) const
    {
        return m_szBuffer[uiPos];
    }

protected:
private:
/*!
@brief  This copy constructor is private.
*/
    StringStream(const StringStream&);

/*!
@brief  This assignment operator is private.
*/
    StringStream& operator=(const StringStream&);

#if PIA_ENABLE_RESIZE_BUFFER
    void FreeBuffer();
    void FreeBuffer(char* szBuffer);
    void ResizeBuffer(u32 uiSize);
#endif
    void AddBaseIfRequired();

    StringStream& StreamNumber(u32 uiValue);
    StringStream& StreamNumber(s32 iValue);
    void TestFreeRoom(u32 uiNewSize);

    //n2619
    s32 AddStringToRoom(const char* szBuffer);

private:
    char* m_szBuffer;
    u32 m_uiSize;
    char* m_szCurrentPosition;

    char m_szInitialBuffer[NN_PIA_STRINGSTREAM_INITIAL_SIZE];
    //n2619
    char m_szBufferToAdd[NN_PIA_STRINGSTREAM_INITIAL_SIZE];

    bool m_Hex;
    bool m_ShowBase;
    bool m_BoolAlpha;

#if NN_PIA_ENABLE_ZERO_FILL_HEX
    //n2788:
    bool m_ZeroHex;
#else
    u8 m_padding1;  // Avoids the compiler warning.
#endif  // NN_PIA_ENABLE_ZERO_FILL_HEX
};

// Manipulators.

extern StringStream& hex(StringStream& oStringStream);
extern StringStream& dec(StringStream& oStringStream);
extern StringStream& endl(StringStream& oStringStream);
extern StringStream& showbase(StringStream& oStringStream);
extern StringStream& noshowbase(StringStream& oStringStream);
extern StringStream& boolalpha(StringStream& oStringStream);
extern StringStream& noboolalpha(StringStream& oStringStream);

#if NN_PIA_ENABLE_ZERO_FILL_HEX
extern StringStream& zerohex(StringStream& oStringStream);
#endif  // NN_PIA_ENABLE_ZERO_FILL_HEX
}
}
}  // end of namespace nn::pia::common
//! @endcond
