﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Time.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_TimeSpan.h>

namespace nn
{
namespace pia
{
namespace common
{

/*!
@brief  Class that represents time.

@date 2012-06-06  Made the class inherit from \c RootObject.
@date 2012-04-20  Changed the internal representation to be a direct representation of a system tick. Consequently, removed <tt>Time(u64)</tt> and <tt>operator u64</tt>, which handled the raw values directly.
@date 2012-04-20  Added a \c TimeSpan class to represent time spans (deltas). The arguments and return values of operators that handle time spans have been changed to instances of the \c TimeSpan class.
@date 2012-04-20  Added the \c SetNow function.
@date 2012-04-20  Removed \c ConvertFromSeconds and \c ConvertToSeconds functions, and gave equivalent functionality to <tt>TimeSpan::SetFromSec(f32)</tt> and <tt>TimeSpan::GetSecF32</tt> functions.
@date 2012-04-13  Changed the specifications of <tt>GetTime</tt> function to return the time elapsed since the <tt>nn::pia::common::Initialize</tt> function was called.
@date 2012-04-13  Removed the <tt>Reset</tt> function.
@date 2012-04-13  The <tt>nn::pia::common::Initialize</tt> and <tt>nn::pia::common::Finalize</tt> functions now initialize and finalize the \c Time class.
@date 2012-04-04  Initial version.
*/
class Time : public RootObject
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    Time()
      : m_Tick(0)
    {
    }


    explicit Time(u64 tick)
      : m_Tick(tick)
    {
    }

/*!
@brief  Copy constructor.

@param[in] rhs  A <tt>Time</tt> instance that holds the value used to initialize this instance.
*/
    Time(const Time& rhs)
      : m_Tick(rhs.m_Tick)
    {
    }


/*!
@brief  Makes this object represent the current time.
*/
    void SetNow();


/*!
@brief  Gets a \c Time instance representing the current time.

@return  A <tt>Time</tt> instance representing the current time.
*/
    static Time GetTime()
    {
        Time time;
        time.SetNow();
        return time;
    }


/*!
@brief  Assignment operator.

@param[in] rhs  The <tt>Time</tt> instance to assign to this instance.

@return  Returns a reference to this instance.
*/
    Time& operator=(const Time& rhs)
    {
        m_Tick = rhs.m_Tick;
        return *this;
    }


/*!
@brief  Adds time to this instance's time.

@param[in] span  The time to add.

@return  Returns a reference to this instance.
*/
    Time& operator+=(const TimeSpan& span)
    {
        m_Tick += span.m_TickSpan;
        return *this;
    }


/*!
@brief  Subtracts time from this instance.

@param[in] span  The time to subtract.

@return  Returns a reference to this instance.
*/
    Time& operator-=(const TimeSpan& span)
    {
        m_Tick -= span.m_TickSpan;
        return *this;
    }


/*!
@brief  Calculates a <tt>Time</tt> object holding the amount of time added to this instance.

@param[in] span  The time to add.

@return  A <tt>Time</tt> representing the amount of time that was added.
*/
    Time operator+(const TimeSpan& span) const
    {
        Time time;
        time.m_Tick = m_Tick + span.m_TickSpan;
        return time;
    }


/*!
@brief  Calculates a <tt>Time</tt> object holding the amount of time subtracted from this instance.

@param[in] span  The time to subtract.

@return  A <tt>Time</tt> representing the amount of time that was subtracted.
*/
    Time operator-(const TimeSpan& span) const
    {
        Time time;
        time.m_Tick = m_Tick - span.m_TickSpan;
        return time;
    }


/*!
@brief  Calculates a time delta.

@param[in] rhs  The base time.

@return  A <tt>TimeSpan</tt> object representing the time delta.
*/
    TimeSpan operator-(const Time& rhs) const
    {
        TimeSpan span;
        span.m_TickSpan = m_Tick - rhs.m_Tick;
        return span;
    }


/*!
@brief  Equality operator. Determines whether two <tt>Time</tt> objects are equal.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  Returns <tt>true</tt> if they represent the same time, and <tt>false</tt> otherwise.
*/
    bool operator==(const Time& rhs) const
    {
        return (m_Tick == rhs.m_Tick);
    }


/*!
@brief  Equality operator. Determines whether two <tt>Time</tt> objects differ.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  <tt>true</tt> if they represent different times, and <tt>false</tt> otherwise.
*/
    bool operator!=(const Time& rhs) const
    {
        return (m_Tick != rhs.m_Tick);
    }


/*!
@brief  Comparison operator. Returns \c true if this instance is earlier.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  \c true if this instance is earlier, or \c false otherwise.
*/
    bool operator<(const Time& rhs) const
    {
        return (m_Tick < rhs.m_Tick);
    }


/*!
@brief  Comparison operator. Returns \c true if this instance is earlier or equal.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  \c true if this instance is earlier or equal, or \c false otherwise.
*/
    bool operator<=(const Time& rhs) const
    {
        return (m_Tick <= rhs.m_Tick);
    }


/*!
@brief  Comparison operator. Returns \c true if this instance is later.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  \c true if this instance is later, or \c false otherwise.
*/
    bool operator>(const Time& rhs) const
    {
        return (m_Tick > rhs.m_Tick);
    }


/*!
@brief  Comparison operator. Returns \c true if this instance is later or equal.

@param[in] rhs  The <tt>Time</tt> instance to compare with.

@return  \c true if this instance is later or equal, or \c false otherwise.
*/
    bool operator>=(const Time& rhs) const
    {
        return (m_Tick >= rhs.m_Tick);
    }


/*!
@cond PRIVATE
@brief  Gets the time from the current time to the specified time.

@param[in] timeout  A time span (delta).

@return  The time from now until <SPAN class="argument">timeout</SPAN>.
*/
    static Time ConvertTimeoutToDeadline(const TimeSpan& timeout)
    {
        return (GetTime() += timeout);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the raw tick value stored in the instance.

@return  The tick value stored in the instance.
@see  SetTick
*/
    u64 GetTick(void) const
    {
        return m_Tick;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the raw tick value of the instance.

@param[in] tick  The raw tick count.
@see  GetTick
*/
    void SetTick(u64 tick)
    {
        m_Tick = tick;
    }
    //! @endcond

/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;

    static const Time c_Max;  //< maximum time value.
    static const Time c_Min;  //< minimum time value.

private:
    static Time Create(u64 tick);

    u64 m_Tick;
};
}
}
}  // end of namespace nn::pia::common
