﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_TimeSpan.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_RootObject.h>

namespace nn
{
namespace pia
{
namespace common
{

/*!
@brief  Type that represents a time interval. The value is in milliseconds.
*/
typedef s32 TimeInterval;


class Time;

/*!
@brief  Class for representing time spans. This class can perform type conversions with numeric values in milliseconds.

@date 2012-07-12  Added support so that this instance can be initialized as a static variable without problems.
@date 2012-05-24  Added the addition and subtraction operators to the <tt>TimeSpan</tt> class.
@date 2012-04-20  Initial version.
*/
class TimeSpan : public common::RootObject
{

    friend class Time;

public:
/*!
@brief  Default constructor.
*/
    TimeSpan()
      : m_TickSpan(0)
    {
    }


/*!
@brief  Constructor that initializes the instance from a numeric value representing milliseconds.

@param[in] msec  Specifies the initial value, in milliseconds.
*/
    TimeSpan(s32 msec)
    {
        SetFromMSec(msec);
    }


/*!
@brief  Constructor that initializes the instance from a numeric value representing milliseconds.

@param[in] msec  Specifies the initial value, in milliseconds.
*/
    TimeSpan(s64 msec)
    {
        SetFromMSec(msec);
    }


/*!
@brief  Copy constructor.

@param[in] rhs  A <tt>TimeSpan</tt> instance that holds the values used to initialize this instance.
*/
    TimeSpan(const TimeSpan& rhs)
      : m_TickSpan(rhs.m_TickSpan)
    {
    }


/*!
@brief  Assignment operator.

@param[in] rhs  The <tt>TimeSpan</tt> instance to assign to this instance.

@return  Returns a reference to this instance.
*/
    TimeSpan& operator=(const TimeSpan& rhs)
    {
        m_TickSpan = rhs.m_TickSpan;
        return *this;
    }


/*!
@brief  An operator that assigns a numeric value in milliseconds.

@param[in] msec  The time in milliseconds.

@return  Returns a reference to this instance.
*/
    TimeSpan& operator=(s32 msec)
    {
        SetFromMSec(msec);
        return *this;
    }


/*!
@brief  An operator that assigns a numeric value in milliseconds.

@param[in] msec  The time in milliseconds.

@return  Returns a reference to this instance.
*/
    TimeSpan& operator=(s64 msec)
    {
        SetFromMSec(msec);
        return *this;
    }


/*!
@brief  An operator that converts a time object to an <tt>s32</tt> value in milliseconds.

@return  The time in milliseconds.
*/
    operator s32() const
    {
        return GetMSec();
    }


/*!
@brief  An operator that converts a time object to an <tt>s64</tt> value in milliseconds.

@return  The time in milliseconds.
*/
    operator s64() const
    {
        return GetMSecS64();
    }


/*!
@brief  Sets the time in milliseconds from an <tt>s32</tt> value.

@param[in] msec  The time in milliseconds.
*/
    void SetFromMSec(s32 msec)
    {
        m_TickSpan = GetTicksPerMSec() * msec;
    }


/*!
@brief  Sets the time in milliseconds from an <tt>s64</tt> value.

@param[in] msec  The time in milliseconds.
*/
    void SetFromMSec(s64 msec)
    {
        m_TickSpan = GetTicksPerMSec() * msec;
    }


/*!
@brief  Sets the time in seconds from an <tt>s32</tt> value.

@param[in] sec  The time in seconds.
*/
    void SetFromSec(s32 sec)
    {
        m_TickSpan = GetTicksPerSec() * sec;
    }


/*!
@brief  Sets the time in seconds from an <tt>s64</tt> value.

@param[in] sec  The time in seconds.
*/
    void SetFromSec(s64 sec)
    {
        m_TickSpan = GetTicksPerSec() * sec;
    }


/*!
@brief  Sets the time in seconds from an <tt>f32</tt> value.

@param[in] sec  The time in seconds.
*/
    void SetFromSec(f32 sec)
    {
        m_TickSpan = static_cast<s64>(GetTicksPerSec() * sec);
    }


/*!
@brief  Gets the time in milliseconds as an <tt>s32</tt> value.

@return  The time in milliseconds.
*/
    s32 GetMSec() const
    {
        return static_cast<s32>(m_TickSpan / GetTicksPerMSec());
    }


/*!
@brief  Gets the time in milliseconds as an <tt>s64</tt> value.

@return  The time in milliseconds.
*/
    s64 GetMSecS64() const
    {
        return m_TickSpan / GetTicksPerMSec();
    }


/*!
@brief  Gets the time in seconds as an <tt>s32</tt> value.

@return  The time in seconds.
*/
    s32 GetSec() const
    {
        return static_cast<s32>(m_TickSpan / GetTicksPerSec());
    }


/*!
@brief  Gets the time in seconds as an <tt>s64</tt> value.

@return  The time in seconds.
*/
    s64 GetSecS64() const
    {
        return m_TickSpan / GetTicksPerSec();
    }


/*!
@brief  Gets the time in seconds as an <tt>f32</tt> value.

@return  The time in seconds.
*/
    f32 GetSecF32() const
    {
        return static_cast<f32>(m_TickSpan) / static_cast<f32>(GetTicksPerSec());
    }


/*!
@brief  Equality operator. It determines whether two <tt>TimeSpan</tt> objects are equal.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  Returns <tt>true</tt> if they represent the same time, and <tt>false</tt> otherwise.
*/
    bool operator==(const TimeSpan& rhs) const
    {
        return (m_TickSpan == rhs.m_TickSpan);
    }


/*!
@brief  Equality operator. It determines whether two <tt>TimeSpan</tt> objects are equal.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if they represent the same time, and <tt>false</tt> otherwise.
*/
    bool operator==(s32 rhs) const
    {
        return (*this == TimeSpan(rhs));
    }


/*!
@brief  Equality operator. Determines whether two <tt>TimeSpan</tt> objects are equal.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if they represent the same time, and <tt>false</tt> otherwise.
*/
    bool operator==(s64 rhs) const
    {
        return (*this == TimeSpan(rhs));
    }


/*!
@brief  Equality operator. It determines whether two <tt>TimeSpan</tt> objects differ.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  <tt>true</tt> if they represent different times, and <tt>false</tt> otherwise.
*/
    bool operator!=(const TimeSpan& rhs) const
    {
        return (m_TickSpan != rhs.m_TickSpan);
    }


/*!
@brief  Equality operator. Determines whether two <tt>TimeSpan</tt> objects differ.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  <tt>true</tt> if they represent different times, and <tt>false</tt> otherwise.
*/
    bool operator!=(s32 rhs) const
    {
        return (*this != TimeSpan(rhs));
    }


/*!
@brief  Equality operator. Determines whether two <tt>TimeSpan</tt> objects differ.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  <tt>true</tt> if they represent different times, and <tt>false</tt> otherwise.
*/
    bool operator!=(s64 rhs) const
    {
        return (*this != TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is shorter; otherwise, returns <tt>false</tt>.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is shorter, and <tt>false</tt> otherwise.
*/
    bool operator<(const TimeSpan& rhs) const
    {
        return (m_TickSpan < rhs.m_TickSpan);
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is shorter; otherwise, returns <tt>false</tt>.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is shorter, and <tt>false</tt> otherwise.
*/
    bool operator<(s32 rhs) const
    {
        return (*this < TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is shorter; otherwise, returns <tt>false</tt>.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is shorter, and <tt>false</tt> otherwise.
*/
    bool operator<(s64 rhs) const
    {
        return (*this < TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.
*/
    bool operator<=(const TimeSpan& rhs) const
    {
        return (m_TickSpan <= rhs.m_TickSpan);
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.
*/
    bool operator<=(s32 rhs) const
    {
        return (*this <= TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is less than or equal, and <tt>false</tt> otherwise.
*/
    bool operator<=(s64 rhs) const
    {
        return (*this <= TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is longer; otherwise returns <tt>false</tt>.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is longer, and <tt>false</tt> otherwise.
*/
    bool operator>(const TimeSpan& rhs) const
    {
        return (m_TickSpan > rhs.m_TickSpan);
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is longer; otherwise returns <tt>false</tt>.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is longer, and <tt>false</tt> otherwise.
*/
    bool operator>(s32 rhs) const
    {
        return (*this > TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is longer; otherwise returns <tt>false</tt>.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is longer, and <tt>false</tt> otherwise.
*/
    bool operator>(s64 rhs) const
    {
        return (*this > TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is greater than or equal; otherwise returns <tt>false</tt>.

@param[in] rhs  The <tt>TimeSpan</tt> instance to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is greater than or equal, and <tt>false</tt> otherwise.
*/
    bool operator>=(const TimeSpan& rhs) const
    {
        return (m_TickSpan >= rhs.m_TickSpan);
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is greater than or equal; otherwise returns <tt>false</tt>.

@param[in] rhs  An <tt>s32</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is greater than or equal, and <tt>false</tt> otherwise.
*/
    bool operator>=(s32 rhs) const
    {
        return (*this >= TimeSpan(rhs));
    }


/*!
@brief  Comparison operator. Returns <tt>true</tt> if the time represented by this instance is greater than or equal; otherwise returns <tt>false</tt>.

@param[in] rhs  An <tt>s64</tt> value in milliseconds to compare with.

@return  Returns <tt>true</tt> if the time represented by this instance is greater than or equal, and <tt>false</tt> otherwise.
*/
    bool operator>=(s64 rhs) const
    {
        return (*this >= TimeSpan(rhs));
    }


/*!
@brief  Adds time to this instance.

@param[in] rhs  The time span to add.

@return  Returns a reference to this instance.
*/
    TimeSpan& operator+=(const TimeSpan& rhs)
    {
        m_TickSpan += rhs.m_TickSpan;
        return *this;
    }


/*!
@brief  Subtracts time from this instance.

@param[in] rhs  The time span to subtract.

@return  Returns a reference to this instance.
*/
    TimeSpan& operator-=(const TimeSpan& rhs)
    {
        m_TickSpan -= rhs.m_TickSpan;
        return *this;
    }


/*!
@brief  Calculates a <tt>TimeSpan</tt> object holding the amount of time added to this instance.

@param[in] rhs  The time span to add.

@return  A <tt>Time</tt> object representing the amount of time that was added.
*/
    TimeSpan operator+(const TimeSpan& rhs) const
    {
        TimeSpan t;
        t.m_TickSpan = m_TickSpan + rhs.m_TickSpan;
        return t;
    }


/*!
@brief  Calculates a <tt>TimeSpan</tt> object holding the amount of time subtracted from this instance.

@param[in] rhs  The time span to subtract.

@return  A <tt>TimeSpan</tt> object representing the amount of time that was subtracted.
*/
    TimeSpan operator-(const TimeSpan& rhs) const
    {
        TimeSpan t;
        t.m_TickSpan = m_TickSpan - rhs.m_TickSpan;
        return t;
    }


/*!
@cond PRIVATE
@brief  Gets the raw tick count.

@return  The raw tick count.
*/
    s64 GetTick() const
    {
        return m_TickSpan;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the raw tick count.

@param[in] tick  The raw tick count.
*/
    void SetTick(s64 tick)
    {
        m_TickSpan = tick;
    }
    //! @endcond


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;

private:
    s64 m_TickSpan;

    static const s64& GetTicksPerMSec();
    static const s64& GetTicksPerSec();
};
}
}
}  // end of namespace nn::pia::common
