﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Trace.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_StringStream.h>
#include <pia/common/common_EmptyStream.h>
#include <pia/common/common_Time.h>


namespace nn
{
namespace pia
{
namespace common
{

#if NN_PIA_ENABLE_TRACE
/*!
@cond PRIVATE
@brief  Outputs trace information.
*/
#define PIA_TRACE(FLAG, ...)                                                               \
    \
{                                                                                   \
        nn::pia::common::Trace* pNnPiaCommonTrace = nn::pia::common::Trace::GetInstance(); \
        if (pNnPiaCommonTrace)                                                             \
        {                                                                                  \
            if (pNnPiaCommonTrace->IsFlagSet(FLAG))                                        \
            {                                                                              \
                pNnPiaCommonTrace->WriteCaption();                                         \
                pNnPiaCommonTrace->Write(__VA_ARGS__);                                     \
                pNnPiaCommonTrace->Write("\n");                                            \
            }                                                                              \
        }                                                                                  \
    \
}
//! @endcond

/*!
@cond PRIVATE
@brief  Converts a <tt>bool</tt> value to a string.
*/
#define PIA_BOOL_TO_TEXT(VALUE) ((VALUE) ? "true" : "false")
//! @endcond


/*!
@cond PRIVATE
@brief  Takes stream-format arguments and outputs them to trace.
*/
#define PIA_TRACE_STREAM(FLAG, ARGS)                                                       \
    \
{                                                                                   \
        nn::pia::common::Trace* pNnPiaCommonTrace = nn::pia::common::Trace::GetInstance(); \
        if (pNnPiaCommonTrace)                                                             \
        {                                                                                  \
            if (pNnPiaCommonTrace->IsFlagSet(FLAG))                                        \
            {                                                                              \
                pNnPiaCommonTrace->WriteCaption();                                         \
                nn::pia::common::StringStream nnPiaCommonStringStream;                     \
                (nnPiaCommonStringStream) << ARGS << NN_PIA_T("\n");                       \
                pNnPiaCommonTrace->Write(nnPiaCommonStringStream.CStr());                  \
            }                                                                              \
        }                                                                                  \
    \
}
//! @endcond

/*!
@cond PRIVATE
@brief  Adds the function name to the trace output.
*/
#define PIA_TRACE_EX(FLAG, ...)                                                                     \
    \
{                                                                                            \
        nn::pia::common::Trace* pNnPiaCommonTrace = nn::pia::common::Trace::GetInstance();          \
        if (pNnPiaCommonTrace)                                                                      \
        {                                                                                           \
            if (pNnPiaCommonTrace->IsFlagSet(FLAG))                                                 \
            {                                                                                       \
                pNnPiaCommonTrace->WriteCaption();                                                  \
                pNnPiaCommonTrace->Write("%s %s:", PIA_CODE_POSITION_FILE, PIA_CODE_POSITION_FUNC); \
                pNnPiaCommonTrace->Write(__VA_ARGS__);                                              \
                pNnPiaCommonTrace->Write("\n");                                                     \
            }                                                                                       \
        }                                                                                           \
    \
}
//! @endcond


#else  // else of NN_PIA_ENABLE_TRACE

#define PIA_TRACE(FLAG, ...) (void)(FLAG)
#define PIA_BOOL_TO_TEXT(VALUE)
#define PIA_TRACE_STREAM(FLAG, ARGS)                \
    {                                               \
        nn::pia::common::EmptyStream(FLAG) << ARGS; \
    }
#define PIA_TRACE_EX(FLAG, ...) (void)(FLAG)

#endif  // end of NN_PIA_ENABLE_TRACE

//n1589: 8-bit assignment is made for each module. We may adjust this if it proves insufficient.
/*!
@brief  Defines the tracing flag type.

@details  Use by passing it as the first argument to the <tt>PIA_TRACE</tt> macro.
*/

typedef u64 TraceFlag;

/*!
@name  TraceFlag
@anchor TraceFlag
     @{
*/
static const TraceFlag TRACE_FLAG_NEVER = 0LL;        //!< The flag indicating that this is never displayed.
static const TraceFlag TRACE_FLAG_ALWAYS = 1LL << 0;  //!< The flag indicating that this is always displayed.

static const TraceFlag TRACE_FLAG_COMMON = 1LL << 1;        //!< Flag for tracing the <tt>PiaCommon</tt> module.
static const TraceFlag TRACE_FLAG_COMMON_EVENT = 1LL << 2;  //!< Flag for tracing events in the <tt>PiaCommon</tt> module.
static const TraceFlag TRACE_FLAG_COMMON_JOB = 1LL << 3;    //!< Flag for tracing jobs in the <tt>PiaCommon</tt> module.

static const TraceFlag TRACE_FLAG_LOCAL = 1LL << 5;            //!< Flag for tracing the <tt>PiaLocal</tt> module.
static const TraceFlag TRACE_FLAG_LOCAL_TRANSPORT = 1LL << 6;  //!< Flag for tracing sending and receiving in the <tt>PiaLocal</tt> module.
static const TraceFlag TRACE_FLAG_LOCAL_MIGRATION = 1LL << 7;  //!< Flag for tracing host migration in the <tt>PiaLocal</tt> module.

static const TraceFlag TRACE_FLAG_INET = 1LL << 9;                  //!< Flag for tracing the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_INIT = 1LL << 10;            //!< Flag for tracing initialization and finalization in the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_SOCKET = 1LL << 11;          //!< Flag for tracing sockets in the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_SOCKET_DETAIL = 1LL << 12;   //!< Flag for verbose tracing of sockets in the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_SOCKET_SEND = 1LL << 13;     //!< Flag for tracing send operations in the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_SOCKET_RECEIVE = 1LL << 14;  //!< Flag for tracing receive operations in the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_NAT = 1LL << 15;             //!< Flag for tracing NAT traversal by the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_NAT_DETAIL = 1LL << 16;      //!< Flag for verbose tracing of NAT traversal by the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_NAT_PROBE = 1LL << 17;       //!< Flag for verbose tracing of NAT traversal by the <tt>PiaInet</tt> module.
static const TraceFlag TRACE_FLAG_INET_GATEWAY = 1LL << 18;         //!< Flag for tracing gateway services in the <tt>PiaInet</tt> module. (There are no plans to put gateway services into production.)

static const TraceFlag TRACE_FLAG_TRANSPORT = 1LL << 19;                          //!< Flag for tracing the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_INIT = 1LL << 20;                     //!< Flag for tracing initialization and finalization in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_KEEP_ALIVE = 1LL << 21;               //!< Flag for tracing keep-alives in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_RTT = 1LL << 22;                      //!< Flag for tracing the round-trip time (RTT) protocol in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_BUFFER = 1LL << 23;                   //!< Flag for tracing buffers in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_PACKET_SEND = 1LL << 24;              //!< Flag for tracing packet sending in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_PACKET_RECEIVE = 1LL << 25;           //!< Flag for tracing packet reception in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_PROTOCOLMESSAGE_SEND = 1LL << 26;     //!< Flag for tracing <tt>ProtocolMessage</tt> sending in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_PROTOCOLMESSAGE_RECEIVE = 1LL << 27;  //!< Flag for tracing <tt>ProtocolMessage</tt> reception in the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_RELAY = 1LL << 28;                    //!< Flag for tracing the relay feature of the <tt>PiaTransport</tt> module.
static const TraceFlag TRACE_FLAG_TRANSPORT_RELAY_DETAIL = 1LL << 29;             //!< Flag for tracing the details of the relay feature in the <tt>PiaTransport</tt> module.

static const TraceFlag TRACE_FLAG_SESSION = 1LL << 30;            //!< Flag for tracing the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_CONNECT = 1LL << 31;    //!< Flag for tracing connections in the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_LEAVE = 1LL << 32;      //!< Flag for tracing disconnections in the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_UPDATE = 1LL << 33;     //!< Flag for tracing updates in the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_MIGRATION = 1LL << 34;  //!< Flag for tracing host migration in the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_INIT = 1LL << 36;       //!< Flag for tracing initialization and finalization in the <tt>PiaSession</tt> module.
static const TraceFlag TRACE_FLAG_SESSION_JOINT = 1LL << 37;      //!< Flag for tracing joint sessions in the <tt>PiaSession</tt> module.

static const TraceFlag TRACE_FLAG_SYNC = 1LL << 39;  //!< Flag for tracing the <tt>PiaSync</tt> module.

static const TraceFlag TRACE_FLAG_CLONE = 1LL << 41;  //!< Flag for tracing the <tt>PiaClone</tt> module.

static const TraceFlag TRACE_FLAG_RECKONING = 1LL << 43;  //!< Flag for tracing the <tt>PiaReckoning</tt> module.

static const TraceFlag TRACE_FLAG_CHAT = 1LL << 45;  //!< Flag for tracing the <tt>PiaChat</tt> module.

static const TraceFlag TRACE_FLAG_UTIL = 1LL << 47;  //!< Flag for tracing the <tt>PiaUtil</tt> module.

static const TraceFlag TRACE_FLAG_SERVICE = 1LL << 49;                   //!< Flag for tracing the <tt>PiaService</tt> module.
static const TraceFlag TRACE_FLAG_SERVICE_TRANSPORT = 1LL << 50;         //!< Flag for tracing the <tt>PiaService</tt> module.
static const TraceFlag TRACE_FLAG_SERVICE_TRANSPORT_DETAIL = 1LL << 51;  //!< Flag for tracing the <tt>PiaService</tt> module.
static const TraceFlag TRACE_FLAG_SERVICE_RELAY_SERVER = 1LL << 52;      //!< Flag for tracing the <tt>PiaService</tt> module.

static const TraceFlag TRACE_FLAG_BRAIN = 1LL << 54;        //!< Flag for tracing the <tt>PiaBrain</tt> module.
static const TraceFlag TRACE_FLAG_BRAIN_DETAIL = 1LL << 55; //!< Flag for tracing the <tt>PiaBrain</tt> module.

static const TraceFlag TRACE_FLAG_ALL = TRACE_FLAG_ALWAYS;  //!< Flag for tracing everything.

static const TraceFlag TRACE_FLAG_DEFAULT = TRACE_FLAG_ALWAYS;  //!< The default value for <tt>TraceFlag</tt>.

/*!
end of name TraceFlag
     @}
*/


/*!
@brief  This class is the trace class.

@details  Trace code is disabled in release builds.

@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance()</tt> when an instance already exists.
@date 2013-11-21  Reorganized the trace flags. The bit positions of the flags are now different from what they were before.
@date 2013-07-18  Changed the operations of <tt>IsFlagSet</tt>. The function now returns <tt>true</tt> if multiple flags are specified and any of the flags are raised.
@date 2012-06-07  <tt>TraceFlag</tt> and its value were changed from enumerated to <tt>u64</tt> constants.
@date 2012-05-30  Times can now also be printed.
@date 2012-04-19  Initial version.
*/
class Trace : public nn::pia::common::RootObject
{
public:
/*!
@brief  Creates the <tt>Trace</tt> class instance (singleton pattern).

@details  Call this member function after calling <tt>nn::pia::common::BeginSetup</tt>, but before calling <tt>nn::pia::common::EndSetup</tt>.

@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  The <tt>common</tt> module has not been initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was not called between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.

*/
    static nn::Result CreateInstance(void);


/*!
@brief  Destroys the <tt>Trace</tt> class instance (singleton pattern).

@details  Destroys the <tt>Trace</tt> class instance.
This function returns without performing any action if it is called before an instance has been created.
*/
    static void DestroyInstance(void);


/*!
@brief  Gets the <tt>Trace</tt> class instance (singleton pattern).

@details  Returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>CreateInstance()</tt> function.

@return  Returns a pointer to the <tt>Trace</tt> instance.

*/
    static Trace* GetInstance(void)
    {
        return s_pInstance;
    }

    //n1589: Optimized code for checking <tt>IsFlagSet</tt>.
/*!
@cond PRIVATE
*/
    static bool IsFlagSetStatic(u64 flag)
    {
#if NN_PIA_ENABLE_TRACE
        if (Trace::GetInstance() == NULL)
        {
            return false;
        }

        return Trace::GetInstance()->IsFlagSet(flag);
#else
        NN_PIA_DUMMY_PARAM(flag);
        return false;
#endif
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Outputs a string.
Do not call this directly. It is called from a macro.
*/
    void Write(const char* pFmt, ...);
    //! @endcond


/*!
@cond PRIVATE
@brief  Outputs a caption string.
Do not call this directly. It is called from a macro.
*/
    void WriteCaption();
    //! @endcond

/*!
@brief  Enables trace flags.

@details  Use this method to enable additional trace flags.

@param[in] flags  Specifies the bitwise OR of the <tt>@ref TraceFlag</tt> values to enable.
*/
    void SetFlag(u64 flags)
    {
        m_flags |= flags;
    }

/*!
@brief  Disables the specified trace flags.

@details  Disables the specified trace flags.
However, <tt>TRACE_FLAG_ALWAYS</tt> is excluded.

@param[in] flags  Specifies the bitwise OR of the <tt>@ref TraceFlag</tt> values to disable.
*/
    void ClearFlag(u64 flags)
    {
        m_flags = m_flags & (~flags);
    }

/*!
@brief  Indicates whether the specified flags are set.

@details  The result of <tt>IsFlagSet(TRACE_NEVER)</tt> is <tt>true</tt>.
Returns <tt>true</tt> if any of the specified flags are raised when multiple flags have been specified by argument.

@return  Returns <tt>true</tt> if the particular flag is set, and <tt>false</tt> if not.

@param[in] flag  Specifies the trace flags to query.

*/
    bool IsFlagSet(u64 flag) const
    {
#if NN_PIA_ENABLE_TRACE
        return (m_flags & flag) != 0;
#else
        NN_PIA_DUMMY_PARAM(flag);
        return false;
#endif
    }

/*!
@brief  Clears all trace flags, except for <tt>TRACE_FLAG_ALWAYS</tt>.
*/
    void ClearFlag(void)
    {
        m_flags = TRACE_FLAG_DEFAULT;
    }


/*!
@brief  Enables and disables the time display.
When enabled, it displays the time passed from the base time in units of milliseconds, such as "[Pia:12345]..."
@param[in] isPrint  Specify <tt>true</tt> to display the time.
*/
    void SetPrintTime(bool isPrint)
    {
        m_IsPrintTime = isPrint;
    }


/*!
@brief  Gets whether time display is enabled or disabled.
@return  Returns <tt>true</tt> if times are printed.
*/
    bool IsPrintTime() const
    {
        return m_IsPrintTime;
    }


/*!
@brief  Resets the base time for the time display to the current time.
*/
    void ResetBaseTime()
    {
        m_BaseTime.SetNow();
    }

private:
/*!
@brief  The constructor. Only the <tt>TRACE_FLAG_ALWAYS</tt> flag is set.
*/
    Trace();

/*!
@brief  Destructor.
*/
    ~Trace(void)
    {
    }


/*!
@brief  This copy constructor is private.
*/
    Trace(const Trace&);

/*!
@brief  This assignment operator is private.
*/
    Trace& operator=(const Trace&);

    u64 m_flags;
    Time m_BaseTime;
    bool m_IsPrintTime;

    static Trace* s_pInstance;
};
}
}
}  // end of namespace nn::pia::common
