﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_Watermark.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#include <pia/common/common_RootObject.h>


namespace nn
{
namespace pia
{
namespace common
{


/*!
@brief  Contains member functions for profiling things like peak values in the Pia library's resource usage (unsupported).

@date 2012-04-04  Initial version.
*/
class Watermark : public RootObject
{
public:
/*!
@brief  Specifies the maximum length of a name that can be set using the <tt>SetName</tt> function.

@see  SetName
*/
    static const size_t MAX_NAME_LENGTH = 63;


/*!
@brief  Instantiates the object with default parameters (default constructor).
The instance is constructed with <tt>Update</tt> functionality disabled.
*/
    Watermark(void);


/*!
@brief  Destroys the object.
*/
    ~Watermark(void);


/*!
@brief  Enables updating values with the <tt>Update</tt> function.

@details  If you call <tt>Update</tt> while updating is enabled, minimum and maximum values are updated as necessary, values that can be retrieved with <tt>GetLatestValue</tt> are updated, and the counter that tracks the number of calls to <tt>Update</tt> is incremented.
@see  DisableUpdate, IsUpdateEnabled


*/
    void EnableUpdate(void)
    {
        m_Enabled = true;
    }


/*!
@brief  Disables updating values with <tt>Update</tt>.

@details  If updating is disabled, these values (the minimum, maximum, updated value, and number of calls to the <tt>Update</tt> function) are not updated even if you call <tt>Update</tt>.

@see  EnableUpdate, IsUpdateEnabled
*/
    void DisableUpdate(void)
    {
        m_Enabled = false;
    }


/*!
@brief  Returns whether <tt>Update</tt> functionality is enabled.

@return  Returns <tt>true</tt> if updating values with <tt>Update</tt> is enabled; returns <tt>false</tt> otherwise.
@see  EnableUpdate, DisableUpdate
*/
    bool IsUpdateEnabled(void) const
    {
        return m_Enabled;
    }


/*!
@brief  Sets a name for the instance.

@param[in] pName  Specifies the name to set. The specified string must be <tt>NULL</tt>-terminated and no longer than <tt>MAX_NAME_LENGTH</tt>.
The function asserts if an invalid pointer is specified or the name is too long.
@see  GetName
*/
    void SetName(const char* pName);


/*!
@brief  Gets the name of an instance.

@return  Returns a pointer to the character array storing the name.
@see  SetName
*/
    const char* GetName(void) const
    {
        return m_Name;
    }


/*!
@brief  Re-initializes the values maintained by the instance (maximum, minimum, updated values, and the number of calls to <tt>Update</tt>).

@details  Calling this function does not change the name set using <tt>SetName</tt> or change whether <tt>Update</tt> functionality is enabled or disabled.
@see  SetName, EnableUpdate, DisableUpdate, IsUpdateEnabled
*/
    void Reset(void);


/*!
@brief  Updates the value.
@details  If you call this function while <tt>Update</tt> functionality is enabled, the values being used so far for the maximum, minimum, update value, and number of calls to <tt>Update</tt> are all updated appropriately.
If you call this function while <tt>Update</tt> functionality is disabled, it does nothing and returns.

@param[in] value  Specifies the value being monitored.
@see  EnableUpdate, DisableUpdate, IsUpdateEnabled, GetMaxValue, GetMinValue, GetLatestValue, GetUpdateCount




*/
    void Update(s64 value);


/*!
@brief  Gets the maximum value currently being used by <tt>Update</tt>.

@see  Update, GetMinValue, GetLatestValue, GetUpdateCount
*/
    s64 GetMaxValue(void) const
    {
        return m_MaxValue;
    }


/*!
@brief  Gets the minimum value currently being used by <tt>Update</tt>.

@see  Update, GetMaxValue, GetLatestValue, GetUpdateCount
*/
    s64 GetMinValue(void) const
    {
        return m_MinValue;
    }


/*!
@brief  Gets the update value passed in the last call to <tt>Update</tt>.

@see  Update, GetMinValue, GetMaxValue, GetUpdateCount
*/
    s64 GetLatestValue(void) const
    {
        return m_LatestValue;
    }

/*!
@brief  Gets the number of times <tt>Update</tt> has been called so far.

@see  Update, GetMaxValue, GetMinValue, GetLatestValue
*/
    s64 GetUpdateCount(void) const
    {
        return m_UpdateCount;
    }


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


private:
    s64 m_MinValue;     // Specifies the minimum value.
    s64 m_MaxValue;     // Specifies the maximum value.
    s64 m_LatestValue;  // Specifies the latest value.
    u64 m_UpdateCount;  // Specifies the number of times the <tt>Update</tt> function has been called.

    bool m_Enabled;  // Specifies whether the <tt>Update</tt> feature is available.

    char m_Name[MAX_NAME_LENGTH + 1];  // Specifies the name applied to the watermark instance.
};
}
}
}  // end of namespace nn::pia::common
