﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     common_WatermarkManager.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#include <pia/common/common_Singleton.h>
#include <pia/common/common_Watermark.h>


#define NN_PIA_WATERMARK_UPDATE(KEY, VALUE)                                                 \
    \
if(nn::pia::common::WatermarkManager::GetInstance())                                        \
    \
{                                                                                    \
        nn::pia::common::WatermarkManager::GetInstance()->GetWatermark(KEY)->Update(VALUE); \
    \
}


#define NN_PIA_WATERMARK_SETNAME(KEY, NAME)                                                 \
    \
if(nn::pia::common::WatermarkManager::GetInstance())                                        \
    \
{                                                                                    \
        nn::pia::common::WatermarkManager::GetInstance()->GetWatermark(KEY)->SetName(NAME); \
    \
}


#define NN_PIA_WATERMARK_TRACE(KEY, FLAG)                                                 \
    \
if(nn::pia::common::WatermarkManager::GetInstance())                                      \
    \
{                                                                                  \
        nn::pia::common::WatermarkManager::GetInstance()->GetWatermark(KEY)->Trace(FLAG); \
    \
}


namespace nn
{
namespace pia
{
namespace common
{

/*!
@brief  This class manages a collection of Watermark instances (not supported).
@details  The <tt>WatermarkManager</tt> class is provided for managing collections of <tt>Watermark</tt> instances.
You can use it to get information like the peak buffer usage of the Pia library.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2013-05-14  Added <tt>KEY_UNRELIABLE_PROTOCOL_RECEIVE_BUFFER_NUM</tt>.
@date 2013-04-01  Removed <tt>KEY_PACKETPARSER_BUFFUR_NUM</tt> and <tt>KEY_PACKETCREATOR_BUFFUR_NUM</tt>.
@date 2013-03-21  Added keys for <tt>Reliable</tt> send and receive buffers in <tt>SessionProtocol</tt>.
@date 2012-05-09  Added keys for <tt>ReliableProtocol</tt> and <tt>StationProtocolReliable</tt> send and receive buffers.
@date 2012-04-25  Added documentation.
@date 2012-04-04  Initial version.


*/
class WatermarkManager : public nn::pia::common::RootObject
{
public:
    static const int WATERMARK_INSTANCES = 10;  //!< The maximum number of <tt>Watermark</tt> instances managed by the <tt>WatermarkManager</tt> class.

    // This key is for differentiating watermark instances. It must not be duplicated.
    static const int KEY_SENDTHREADSTREAM_BUFFER_NUM = 0;     //!< This key is for the buffer usage of the send thread.
    static const int KEY_RECEIVETHREADSTREAM_BUFFER_NUM = 1;  //!< This key is for the buffer usage of the receive thread.
    static const int KEY_INET_RECEIVE_BUFFER_NUM = 2;
    static const int KEY_RELIABLE_PROTOCOL_SEND_BUFFER_NUM = 3;             //!< This key is for the send buffer usage of the <tt>ReliableProtocol</tt> instance.
    static const int KEY_RELIABLE_PROTOCOL_RECEIVE_BUFFER_NUM = 4;          //!< This key is for the receive buffer usage of the <tt>ReliableProtocol</tt> instance.
    static const int KEY_STATION_PROTOCOL_RELIABLE_SEND_BUFFER_NUM = 5;     //!< This key is for the send buffer usage of the <tt>StationProtocolReliable</tt> instance.
    static const int KEY_STATION_PROTOCOL_RELIABLE_RECEIVE_BUFFER_NUM = 6;  //!< This key is for the receive buffer usage of the <tt>StationProtocolReliable</tt> instance.
    static const int KEY_SESSION_PROTOCOL_RELIABLE_SEND_BUFFER_NUM = 7;     //!< This key is for the <tt>Reliable</tt> send buffer usage of the <tt>SessionProtocol</tt> instance.
    static const int KEY_SESSION_PROTOCOL_RELIABLE_RECEIVE_BUFFER_NUM = 8;  //!< This key is for the <tt>Reliable</tt> receive buffer usage of the <tt>SessionProtocol</tt> instance.
    static const int KEY_UNRELIABLE_PROTOCOL_RECEIVE_BUFFER_NUM = 9;        //!< This key is for the receive buffer usage of the <tt>UnreliableProtocol</tt> instance.
    static const int KEY_MAX = WATERMARK_INSTANCES - 1;


/*!
@brief  Creates the class instance in the <tt>common</tt> module (singleton pattern).
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. Your application's implementation must ensure that this function does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>common</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the timing at which the <tt>CreateInstance</tt> function was called is incorrect. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.

*/
    static nn::Result CreateInstance(void);


/*!
@brief  Destroys the class instance in the <tt>common</tt> module (singleton pattern).
*/
    static void DestroyInstance(void);


/*!
@brief  Gets the class instance in the <tt>common</tt> module (singleton pattern).
@details  Returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>@ref CreateInstance</tt> function.
@return  Returns a pointer to the instance.



*/
    static WatermarkManager* GetInstance(void)
    {
        return s_pInstance;
    }


/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
*/
    WatermarkManager(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    ~WatermarkManager(void);
    //! @endcond


/*!
@brief  Gets a <tt>Watermark</tt> instance (non-<tt>const</tt> version).
@param[in] key  Specifies the key. The following relation must hold: <tt>0 <= key <= KEY_MAX</tt>.
The function asserts if a key outside this range is specified.
@return  On success, returns a pointer to a <tt>Watermark</tt> instance. On failure, returns a <tt>NULL</tt> pointer.

*/
    Watermark* GetWatermark(int key);


/*!
@brief  Gets a <tt>Watermark</tt> instance (<tt>const</tt> version).
@param[in] key  Specifies the key. The following relation must hold: <tt>0 <= key <= KEY_MAX</tt>.
The function asserts if a key outside this range is specified.
@return  On success, returns a pointer to a <tt>Watermark</tt> instance. On failure, returns a <tt>NULL</tt> pointer.

*/
    const Watermark* GetWatermark(int key) const;


/*!
@brief  Enables the <tt>Update</tt> feature of all <tt>Watermark</tt> instances managed by the <tt>WatermarkManager</tt> instance.
@see  DisableAllWatermark

*/
    void EnableAllWatermark(void);


/*!
@brief  Disables the <tt>Update</tt> feature of all <tt>Watermark</tt> instances managed by the <tt>WatermarkManager</tt> instance.
@see  EnableAllWatermark

*/
    void DisableAllWatermark(void);


/*!
@brief  Prints information useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;

private:
    Watermark m_Watermark[WATERMARK_INSTANCES];

    static WatermarkManager* s_pInstance;  // The only instantiation of this class
};
}
}
}  // end of namespace nn::pia::common
