/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_Api.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>


// This is the group of non-member functions for the <tt>inet</tt> module.

namespace nn
{
namespace pia
{
namespace inet
{


/*!
      @brief  Indicates whether the <tt>inet</tt> module is initialized.

      @return  Returns <tt>true</tt> if initialized, or <tt>false</tt> otherwise.
*/
bool IsInitialized(void);


/*!
    @brief  Stores parameters used by the <tt>Initialize</tt> function.
*/
struct Setting
{
    size_t mtuSize;  //!< Specify the MTU of a packet. The value is in bytes. Specify a value in the range from <tt>@ref MIN_MTU_SIZE</tt> through <tt>@ref MAX_MTU_SIZE</tt>.
};

const size_t MAX_MTU_SIZE = 1364;      //!< Defines the maximum MTU that can be specified in <tt>@ref Setting</tt>.
const size_t MIN_MTU_SIZE = 576;       //!< Defines the minimum MTU that can be specified in <tt>@ref Setting</tt>
const size_t DEFAULT_MTU_SIZE = 1240;  //!< Defines the default MTU that is specified in <tt>@ref Setting</tt>.

/*!
    @brief  (For debugging.) This structure stores the debug feature settings that are passed to <tt>@ref nn::pia::inet::SetDebugSetting</tt>.
*/
struct DebugSetting
{
    DebugSetting(bool argIsEnableNatTraversalFailureEmulation = false)
        : isEnableNatTraversalFailureEmulation(argIsEnableNatTraversalFailureEmulation)
    {
    }
    bool isEnableNatTraversalFailureEmulation;  //!< This setting makes NAT transversal fail for debugging purposes.
};

/*!
      @brief  Initializes the <tt>inet</tt> module.

      @return  Returns a <tt>Result</tt> value indicating success if initialization succeeds. You must make sure that the implementation of this function in your application does not return any errors.

      @retval nn::pia::ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
      @retval nn::pia::ResultAlreadyInitialized  Indicates that the module is already initialized. Programming error. Fix your program so that this error is not returned.

      @see  Setting
*/
nn::Result Initialize(const Setting& setting);


/*!
      @brief  Finalizes the <tt>inet</tt> module.
*/
void Finalize(void);


/*!
      @brief  Declares the start of setup.

      @details  Call this function before creating a singleton in the <tt>inet</tt> module.
              Call this function and the <tt>@ref EndSetup</tt> function even when the application
              does not create any singletons in the <tt>inet</tt> module.

      @return  Returns a <tt>Result</tt> value indicating success if it is called at the right time. You must make sure that the implementation of this function in your application does not return any errors.

      @retval nn::pia::ResultNotInitialized  Indicates that the function for initializing the <tt>inet</tt> module has not been called. Programming error. Fix your program so that this error is not returned.
      @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
nn::Result BeginSetup(void);


/*!
      @brief  Declares the end of setup.

      @details  Call this function after creation of a singleton in the <tt>inet</tt> module is complete.
              Calling this function configures memory to be used by the <tt>inet</tt> module singletons,
              and optimizes memory management within the library.
              Always call the <tt>@ref BeginSetup</tt> function and this function even if the application
              does not create any singletons in the <tt>inet</tt> module.

      @return  Returns a <tt>Result</tt> value indicating success if it is called at the right time. You must make sure that the implementation of this function in your application does not return any errors.

      @retval nn::pia::ResultNotInitialized  Indicates that the function for initializing the <tt>inet</tt> module has not been called. Programming error. Fix your program so that this error is not returned.
      @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
nn::Result EndSetup(void);


/*!
    @cond  PRIVATE
    @brief  Gets whether an application is executing code between calls to the <tt>BeginSetup()</tt> and <tt>EndSetup()</tt> functions.
    @endcond
*/
bool IsInSetupMode(void);

/*!
    @brief  (For debugging.) Configures the <tt>inet</tt> debugging feature.

    @details  This function can be called when there is no <tt>inet</tt> instance.

    If <tt><var>isEnableNatErrorEmulation</var></tt> of the <tt>DebugSetting</tt> structure is set to <tt>true</tt>, this setting makes NAT transversal fail on an error.

    If this function is not called, all related debugging features are disabled.

    @attention  Do not call this function in retail ROMs.

    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. This function always succeeds.
*/
nn::Result SetDebugSetting(const DebugSetting& debugSetting);

/*!
    @cond  PRIVATE
    @brief  Returns the setting for the debugging feature.
    @endcond
*/
DebugSetting* GetDebugSetting();

/*!
    @if CAFE_DOC
      @brief  Stops the network interface.
      @details  Use this function to debug the handling of abnormal disconnection
               of the network interface during communication.

               If the network interface is stopped, it is automatically restarted.
               Note, however, that the other communication-related libraries continue to run while the interface is stopped and
               fall into an inoperable state. This necessitates a restart of those libraries.
    @endif
*/
void ControlNetworkIfDown();

}
}
}  // end of namespace nn::pia::inet
