﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     inet_LanCreateSessionSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_LanSessionTypes.h>
#include <pia/session/session_CreateSessionSetting.h>

namespace nn
{
namespace pia
{
namespace inet
{
//! @cond CAFE_DOC

/*!
@brief  Class that manages settings specific to LAN matchmaking session creation.

@date 2014-02-03  Initial version.

@see  nn::pia::session::CreateSessionSetting
*/
class LanCreateSessionSetting : public nn::pia::session::CreateSessionSetting
{
public:
/*!
@brief  The default constructor.
*/
    LanCreateSessionSetting();

/*!
@brief  Destroys the object.
*/
    virtual ~LanCreateSessionSetting();

/*!
@brief  Sets the game mode for the session to create.
@param[in] gameMode  The game mode specified as the session to create.
*/
    void SetGameMode(u32 gameMode);
/*!
@brief  Gets the specified game mode.
@return  Returns the specified game mode.
*/
    u32 GetGameMode() const;

/*!
@brief  Sets the session type to create.
@param[in] sessionType  The session type specified as the session to create.
*/
    void SetSessionType(LanSessionType sessionType);
/*!
@brief  Gets the specified session type.
@return  Returns the specified session type.
*/
    LanSessionType GetSessionType() const;

/*!
@brief  Sets the description string for the session to create.
@param[in] description  The description string specified for the session to create.
*/
    void SetDescription(const wchar_t* description);
/*!
@brief  Gets the specified description string.
@return  Returns the specified description string.
*/
    const wchar_t* GetDescription() const;

/*!
@brief  Specifies the attributes for the session to create.
@param[in] index  The index of the attribute specified for the session to create.
@param[in] value  The value of the attribute specified for the session to create.
*/
    void SetAttribute(u32 index, u32 value);
/*!
@brief  Gets the specified attribute.
@param[in] index  Index for the attribute to get.
@return  Returns a specified attribute value.
*/
    u32 GetAttribute(u32 index) const;

/*!
@brief  Sets application-defined data.
@param[in] pData  The buffer for the application-defined data specified for the session to create.
@param[in] size  The size of the application-defined data specified for the session to create.
@return  Returns a <tt>Result</tt> value indicating success if the application-defined data is set successfully.

@retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer was <tt>NULL</tt> or the specified data size was too big. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetApplicationData(const u8* pData, size_t size);
/*!
@brief  Gets specified application-defined data.
@param[in] pDataBuf  Buffer for copying the specified application-defined data.
@param[in] pDataSize  Size of the data written to the copy buffer.
@param[in] dataBufSize  Indicates the copy buffer size.
@return  Returns a <tt>Result</tt> value indicating success if the application-defined data is acquired successfully.

@retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer is <tt>NULL</tt> or the pointer returned by the copied size is <tt>NULL</tt>. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultInvalidState  Indicates that no application-defined data is set. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultBufferShortage  Indicates that the set application-defined data cannot be copied because it is greater in size than the buffer. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetApplicationData(u8* pDataBuf, size_t* pDataSize, size_t dataBufSize) const;

/*!
@brief  Gets the size of the specified application-defined data.
@return  Returns the size of the application-defined data.
*/
    size_t GetApplicationDataSize() const
    {
        return m_ApplicationDataSize;
    }

/*!
@brief  Sets whether participation in this session is possible.

@details  Sets whether the created session is to be created as joinable.

@param[in] isOpen  When set to <tt>true</tt>, the session created by this <tt>LanCreateSessionSetting</tt> function is joinable for the first time.
*/
    void SetOpenParticipation(bool isOpen)
    {
        m_OpenParticipation = isOpen;
    }

/*!
@brief  Gets whether participation in this session is possible.

@return  Returns <tt>true</tt> if the session is joinable.
*/
    bool GetOpenParticipation() const
    {
        return m_OpenParticipation;
    }

/*!
@brief  Sets the matchmaking priority when a session with the same requirements is found.

@details  Sets the matchmaking priority for the created session.
If a value that exceeds <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is provided as an argument, <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is set instead.

@param[in] priority  The matchmaking priority being set.
*/
    void SetSelectionPriority(u8 priority);

/*!
@cond PRIVATE
@brief  Gets the matchmaking priority when a session with the same requirements is found.

@details  Gets the matchmaking priority for the session being created.

@return  Returns the matchmaking priority being set.
*/
    u8 GetSelectionPriority() const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the trace flag. For more information, see the <tt>@ref TraceFlag</tt> type.
@endcond
*/
    void Trace(u64 flag) const;


protected:
    // Game mode.
    u32 m_GameMode;

    LanSessionType m_SessionType;

    // The description string.
    wchar_t m_Description[LAN_DESCRIPTION_DATA_BUFFER_MAX_SIZE];

    // Attribute.
    u32 m_Attribute[LAN_ATTRIBUTE_MAX_SIZE];

    u8 m_ApplicationDataBuffer[LAN_APPLICATION_DATA_BUFFER_MAX_SIZE];
    size_t m_ApplicationDataSize;

    bool m_OpenParticipation;

private:
    // The copy constructor is sealed.
    LanCreateSessionSetting(const LanCreateSessionSetting&);

    // The assignment operator is sealed.
    LanCreateSessionSetting& operator=(const LanCreateSessionSetting&);
};

//! @endcond
}
}
}  // end of namespace nn::pia::inet
