﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     inet_LanFacade.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>
#include <pia/inet/inet_LanNetworkFactory.h>

namespace nn
{
namespace pia
{
namespace transport
{
class StationConnectionInfo;
}
}
}

namespace nn
{
namespace pia
{
namespace inet
{
class SocketOutputStream;
class SocketInputStream;
class Socket;
class LanMatchmakeUpdateJob;

//! @cond CAFE_DOC

/*!
@brief  Represents the LAN matchmaking facade (<tt>facade</tt>, interface wrapper) class.
@details  This feature is for library developers. It is normally not available to application developers.
When in a LAN environment within a single subnet, the network-related processes (such as sending/receiving or session management) do not use the NEX library.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-04-30  Initial version.

*/
class LanFacade : public nn::pia::common::RootObject
{
public:
/*!
@brief  Creates an instance (singleton pattern).
@return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
@retval nn::pia::ResultNotInitialized  Indicates that the <tt>inet</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.

*/
    static nn::Result CreateInstance(void);

/*!
@brief  Destroys the instance (singleton pattern).
@details  This function does nothing and returns if called when an instance has not been created.


*/
    static void DestroyInstance(void);

/*!
@brief  Gets the pointer to the <tt>LanFacade</tt> instance (singleton pattern).
@return  Returns a <tt>NULL</tt> pointer if the instance cannot be created.

*/
    static LanFacade* GetInstance(void)
    {
        return s_pInstance;
    }

/*!
@cond PRIVATE
@brief  Starts the <tt>LanFacade</tt> feature.
@return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
@retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.

*/
    virtual nn::Result Startup();
    //! @endcond

/*!
@cond PRIVATE
@brief  Cleans up the <tt>LanFacade</tt> feature.
*/
    virtual void Cleanup();
    //! @endcond

/*!
@cond PRIVATE
@brief  Starts a LAN session.
*/
    virtual nn::Result StartLanSession();
    //! @endcond

/*!
@cond PRIVATE
@brief  Terminates a LAN session.
*/
    virtual void StopLanSession();
    //! @endcond

/*!
@brief  Registers a local address.
@details  Registers the IP address assigned locally. Call this function before calling the <tt>@ref session::Session::Startup</tt> function.
@return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
@retval nn::pia::ResultInvalidArgument  Indicates that there is an error in an argument. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Bind(u32 localAddress);

/*!
@brief  Deletes the information on the registered local address.
@details  Call this function after calling the <tt>@ref session::Session::Cleanup</tt> function.
*/
    void Unbind();

/*!
@cond PRIVATE
@brief  Gets the local address.
*/
    u32 GetLocalAddress() const
    {
        return m_LocalAddress;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the local <tt>PrincipalId</tt>.
*/
    u32 GetLocalPrincipalId() const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the local <tt>StationConnectionInfo</tt>.
*/
    transport::StationConnectionInfo& GetLocalStationConnectionInfo();
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the LAN session update job.
*/
    LanMatchmakeUpdateJob* GetUpdateJob()
    {
        return m_pUpdateJob;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the trace flag. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;
    //! @endcond

private:
    // A pointer to a singleton instance.
    static LanFacade* s_pInstance;

/*!
@brief  This constructor is private because the singleton pattern is used.
*/
    LanFacade(void);

/*!
@brief  This destructor is private because the singleton pattern is used.
*/
    virtual ~LanFacade(void);

/*!
@brief  This copy constructor is private because the singleton pattern is used.
*/
    LanFacade(const LanFacade&);

/*!
@brief  This assignment operator is private because the singleton pattern is used.
*/
    LanFacade& operator=(const LanFacade&);

    virtual nn::Result Initialize();
    virtual void Finalize();

    nn::Result BindRandomPort();
    u16 GetRandomPort();

    bool m_IsStartedLanMesh;
    u32 m_LocalAddress;
    Socket* m_pSocket;
    SocketOutputStream* m_pOutputStream;
    SocketInputStream* m_pInputStream;

    LanMatchmakeUpdateJob* m_pUpdateJob;
};

//! @endcond
}
}
}  // end of namespace nn::pia::inet
