﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     inet_LanNetworkFactory.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>
#include <pia/transport/transport_NetworkFactory.h>

namespace nn
{
namespace pia
{
namespace inet
{
//! @cond CAFE_DOC

/*!
@brief  This class creates the necessary classes for LAN networks on the same subnet. It implements the factory pattern.

@details  This feature is for library developers. It is normally not available to application developers.
When in a LAN environment within a single subnet, configuring the <tt>LanNetworkFactory</tt> class for <tt>@ref nn::pia::transport::Transport</tt> and <tt>@ref nn::pia::session::Session</tt> enables the use of network-related process (sending and receiving of data, session management, and so on) without using the NEX library.

@see  nn::pia::transport::Transport::Setting, nn::pia::session::Session::Setting

@date 2014-04-30  Initial version.


*/
class LanNetworkFactory : public nn::pia::transport::NetworkFactory
{
public:

/*!
@name  Finish Initialization
     @{
*/

/*!
@brief  Instantiates an object.
*/
    LanNetworkFactory() {}

/*!
@brief  Destroys the object.
*/
    virtual ~LanNetworkFactory() {}

/*!
end of name Finish Initialization
     @}
*/

/*!
@cond PRIVATE
@brief  Instantiates an <tt>InputStream</tt> object.
*/
    virtual nn::pia::common::IPacketInput* CreateInputStream();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates an <tt>OutputStream</tt> object.
*/
    virtual nn::pia::common::IPacketOutput* CreateOutputStream();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>ConnectStationJob</tt> object.
*/
    virtual nn::pia::transport::ConnectStationJob* CreateConnectStationJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>DisconnectStationJob</tt> object.
*/
    virtual nn::pia::transport::DisconnectStationJob* CreateDisconnectStationJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateCreateMeshJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>JoinMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateJoinMeshJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateLeaveMeshJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Builds the required protocol.
*/
    virtual nn::Result CreateProtocols();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>ProcessHostMigrationJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateProcessHostMigrationJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveWithHostMigrationJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateLeaveWithHostMigrationJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Creates a <tt>KickoutManageJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateKickoutManageJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Constructs a concrete instance of the <tt>PacketHandler</tt> class.
*/
    virtual transport::PacketHandler* CreatePacketHandler();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Gets whether a signature is required.
*/
    virtual bool IsSignatureRequired() const;
/*! @endcond */

/*!
@cond PRIVATE
@brief  Creates a <tt>Storage</tt> object for setting the signature.
*/
    virtual nn::pia::session::SignatureSettingStorage* CreateSignatureSettingStorage();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Gets whether multiple-candidate host migration is allowed.
*/
    virtual bool IsMultiCandidatePermitted() const;
/*! @endcond */

/*!
@cond PRIVATE
@brief  Gets whether the use of the relay communication feature is allowed.
*/
    virtual bool IsRelayConnectionPermitted() const;
/*! @endcond */

/*!
@cond PRIVATE
@brief  Gets whether the use of joint session features is allowed.
*/
    virtual bool IsJointSessionPermitted() const;
/*! @endcond */

/*!
@cond PRIVATE
@brief  Gets the total size of the header portion of packets at the network layer. The value is in bytes.
*/
    virtual size_t GetHeaderSize() const
    {
        return 20 + 8;
    }  // IP header (20 bytes) + UDP header (8 bytes)
/*! @endcond */

/* For the auto-matchmaking feature. */
/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchCreateSessionJob();
/*! @endcond */


/*!
@cond PRIVATE
@brief  Instantiates an <tt>AutoMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchAutoMatchmakeSessionJob();
/*! @endcond */


/*!
@cond PRIVATE
@brief  Instantiates a <tt>BrowseMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchBrowseMatchmakeSessionJob();
/*! @endcond */


/*!
@cond PRIVATE
@brief  Instantiates a <tt>JoinSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchJoinSessionJob();
/*! @endcond */


/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchLeaveSessionJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>DestroySessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchDestroySessionJob();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateGenerateMatchmakeSystemPasswordJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateGenerateMatchmakeSystemPasswordJob();
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateClearMatchmakeSystemPasswordJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateClearMatchmakeSystemPasswordJob();
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the length of the session's <tt>MatchmakeSessionSystemPassword</tt>.
Generates a maintenance region from this value.
*/
    virtual u32 GetMatchmakeSessionSystemPasswordSize() const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Generates a maintenance region for the session's <tt>MatchmakeSessionSystemPassword</tt>.
*/
    virtual wchar_t* CreateMatchmakeSessionSystemPassword(u32 size);
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>JointSessionJob</tt> object for joint sessions.
*/
    virtual common::StepSequenceJob* CreateJointSessionJob();
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>ModifyAttributeJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateModifyAttributeJob();
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateSessionSetting</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateSessionSettingJob();
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateApplicationBufferJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateApplicationDataJob();
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates an <tt>IMatchmakeSession</tt> object for the session.
*/
    virtual session::IMatchmakeSession* CreateMatchmakeSession();
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates an <tt>ISessionInfoList</tt> object for the session.
*/
    virtual session::ISessionInfoList* CreateSessionInfoList(u8 browseSessionInfoListNum);
/*! @endcond */

/*!
@cond PRIVATE
@brief  Instantiates a <tt>MeshLayerController</tt> object for the session.
*/
    virtual session::MeshLayerController* CreateMatchMeshLayerController();
/*! @endcond */
#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    virtual session::IMatchmakeReferee* CreateMatchmakeReferee();

    virtual common::StepSequenceJob* CreateRefereeStartGameJob();

    virtual common::StepSequenceJob* CreateRefereeEndGameJob();

    virtual common::StepSequenceJob* CreateRefereeEndGameWithoutReportJob();

    virtual common::StepSequenceJob* CreateRefereeGetOrCreateStatsJob();
#endif

/*!
@brief  Gets whether a call to <tt>Bind</tt> is required.
*/
    virtual bool GetNecessaryBind() const;

/*!
@brief  Gets whether a server is required.
*/
    virtual bool GetNecessaryServer() const;

/*!
@brief  Gets the host migration mode.
*/
    virtual session::HostMigrationMode GetHostMigrationMode() const;

/*!
@brief  Gets the maximum number of session search results.
*/
    virtual u8 GetBrowseSessionInfoListMax() const;

/* For the auto-matchmaking feature. */

protected:
private:
};

//! @endcond
}
}
}  // end of namespace nn::pia::inet
