﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_LanSessionSearchCriteria.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_SessionSearchCriteria.h>
#include <pia/inet/inet_LanSessionTypes.h>

namespace nn
{
namespace pia
{
namespace session
{
class ISessionInfo;
}
}
}


namespace nn
{
namespace pia
{
namespace inet
{
//! @cond CAFE_DOC

class LanMatchmakeSessionInfo;
/*!
  @brief  The search condition class used when searching for a LAN matchmaking session.

  @date 2014-02-03  Initial version.

  @see  nn::pia::session::SessionSearchCriteria
*/
class LanSessionSearchCriteria : public nn::pia::session::SessionSearchCriteria
{
public:
/*!
    @brief  The default constructor.
*/
    LanSessionSearchCriteria();

/*!
    @brief  The destructor.
*/
    virtual ~LanSessionSearchCriteria();

/*!
    @brief  Sets the minimum participants in the search criteria.
    @param[in] min  The minimum participants value to set in the search criteria.
*/
    void SetMinParticipants(u16 min)
    {
        m_MinParticipantsRangeMax = min;
        m_MinParticipantsRangeMin = min;
    }

/*!
    @brief  Sets a range of minimum participants in the search criteria.
    @param[in] min  The minimum value of the range to set for minimum participants in the search criteria.
    @param[in] max  The maximum value of the range to set for minimum participants in the search criteria.
*/
    void SetMinParticipantsRange(u16 min, u16 max)
    {
        m_MinParticipantsRangeMax = max;
        m_MinParticipantsRangeMin = min;
    };

/*!
    @brief  Sets the maximum participants in the search criteria.
    @param[in] max  The maximum participants value to set in the search criteria.
*/
    void SetMaxParticipants(u16 max)
    {
        m_MaxParticipantsRangeMax = max;
        m_MaxParticipantsRangeMin = max;
    }

/*!
    @brief  Sets a range of maximum participants in the search criteria.
    @param[in] min  The minimum value of the range to set for maximum participants in the search criteria.
    @param[in] max  The maximum value of the range to set for maximum participants in the search criteria.
*/
    void SetMaxParticipantsRange(u16 min, u16 max)
    {
        m_MaxParticipantsRangeMax = max;
        m_MaxParticipantsRangeMin = min;
    };

/*!
    @brief  Sets whether to search only for sessions that are open to participants in the search criteria.
    @param[in] bOpenedOnly  Specify <tt>true</tt> to search only for sessions that are open to participants.
*/
    void SetOpenedOnly(bool bOpenedOnly)
    {
        m_bOpenedOnly = bOpenedOnly;
    }

/*!
    @brief  Sets whether to search only for sessions that have openings in the search criteria.
    @param[in] bVacantOnly  Specify <tt>true</tt> to search only for sessions that have openings.
*/
    void SetVacantOnly(bool bVacantOnly)
    {
        m_bVacantOnly = bVacantOnly;
    }

/*!
    @brief  Sets the search conditions to game mode.
    @param[in] gameMode  Specifies the game mode search conditions.
*/
    void SetGameMode(u32 gameMode)
    {
        m_GameMode = gameMode;
    }
/*!
    @brief  Sets the session type search conditions.
    @param[in] sessionType  Specifies the session type search conditions.
*/
    void SetSessionType(LanSessionType sessionType)
    {
        m_SessionType = sessionType;
    }
/*!
    @brief  Sets a value to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] value  Specifies the value of an attribute.
*/
    void SetAttribute(u32 index, u32 value);
/*!
    @brief  Sets multiple values to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] value  Specifies a list of values for an attribute.
    @param[in] attributeNum  Specifies the number of values for an attribute.
*/
    void SetAttribute(u32 index, u32* value, size_t attributeNum);
/*!
    @brief  Sets a range of values to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] min  Specifies the minimum value of an attribute.
    @param[in] max  Specifies the maximum value of an attribute.
*/
    void SetAttributeRange(u32 index, u32 min, u32 max);

/*!
    @brief  Sets the selection methods for joining sessions when random matchmaking is in effect.
    @param[in] selectionMethod  Specifies the selection method to set for the session.
*/
    void SetRandomSessionSelectionMethod(u32 selectionMethod);

/*!
    @cond  PRIVATE
    @brief  Determines whether a specified matchmaking session matches the search conditions.
    @param[in] pInfo  Specifies the matchmaking session to check to determine whether it matches the search conditions.
*/
    bool IsHit(session::ISessionInfo* pInfo);
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Serializes <tt>LanSessionSearchCriteria</tt>.
*/
    nn::Result Serialize(bit8* pBuffer, size_t* pDataSize, size_t bufferSize) const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Deserializes <tt>LanSessionSearchCriteria</tt>.
*/
    nn::Result Deserialize(const bit8* pData, size_t dataSize);
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Gets the size of <tt>LanSessionSearchCriteria</tt> post-serialization.
*/
    size_t GetSerializedSize() const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Prints information that is useful for debugging.
    @param[in] flag  Specifies the trace flag. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    virtual void Trace(u64 flag) const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Resets the member variables.
*/
    void Clear();
    //! @endcond

/*!
    @brief  Assignment operator.
    @param[in] r  The object to assign.
    @return  See this object.
*/
    LanSessionSearchCriteria& operator=(const LanSessionSearchCriteria& r);

private:
    // The copy constructor is sealed.
    LanSessionSearchCriteria(const LanSessionSearchCriteria&);

    // Invalid number of participants setting.
    static const u16 INVALID_PARICIPANTS_NUM = 0xffff;

    // Minimum participants range.
    u16 m_MinParticipantsRangeMax;
    u16 m_MinParticipantsRangeMin;

    // Maximum participants range.
    u16 m_MaxParticipantsRangeMax;
    u16 m_MaxParticipantsRangeMin;

    // Whether to list only open sessions.
    bool m_bOpenedOnly;
    // Whether to remove full sessions from the results.
    bool m_bVacantOnly;

    // Game mode.
    u32 m_GameMode;

    // Session type.
    LanSessionType m_SessionType;

    // Attributes.
    u32 m_AttributeList[LAN_ATTRIBUTE_MAX_SIZE][LAN_ATTRIBUTE_ARRAY_MAX_SIZE];
    size_t m_AttributeSizeList[LAN_ATTRIBUTE_MAX_SIZE];
    u32 m_AttributeRangeMin[LAN_ATTRIBUTE_MAX_SIZE];
    u32 m_AttributeRangeMax[LAN_ATTRIBUTE_MAX_SIZE];
    bool m_bAttributeRangeUsed[LAN_ATTRIBUTE_MAX_SIZE];
};

//! @endcond
}
}
}  // End of namespace nn::pia::inet.
