﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexCreateSessionSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_CreateSessionSetting.h>
#include <pia/inet/inet_NexSessionTypes.h>

namespace nn
{
namespace pia
{
namespace inet
{
/*!
  @brief  Class that manages settings specific to session creation.

  @date 2015-03-20  Added APIs to set parameters for scored matchmaking.
  @date 2014-05-16  Added the <tt>SetSessionUserPassword()</tt> function.
  @date 2014-02-03  Initial version.

  @see  nn::pia::session::CreateSessionSetting
 */
class NexCreateSessionSetting : public nn::pia::session::CreateSessionSetting
{
public:
/*!
    @brief  The default constructor.
*/
    NexCreateSessionSetting();

/*!
    @brief  The destructor.
*/
    virtual ~NexCreateSessionSetting();

/*!
    @brief  Sets the game mode for the new session.
    @param[in] gameMode  Specifies the game mode for the new session.
*/
    void SetGameMode(u32 gameMode);
/*!
    @brief  Gets the specified game mode.
    @return  Returns the specified game mode.
*/
    u32 GetGameMode() const;

/*!
    @brief  Sets the session type for the new session.
    @param[in] sessionType  Specifies the session type for the new session.
*/
    void SetSessionType(SessionType sessionType);
/*!
    @brief  Gets the specified session type.
    @return  Returns the specified session type.
*/
    SessionType GetSessionType() const;

/*!
    @brief  Sets the description string for the new session.
    @param[in] description  Specifies the description string for the new session.
*/
    void SetDescription(const wchar_t* description);
/*!
    @brief  Gets the specified description string.
    @return  Returns the specified description string.
*/
    const wchar_t* GetDescription() const;

/*!
    @brief  Sets the attributes for the new session.
    @param[in] index  Specifies the index of the attribute for the new session.
    @param[in] value  Specifies the value of the attribute for the new session.
*/
    void SetAttribute(u32 index, u32 value);
/*!
    @brief  Gets the specified attribute.
    @param[in] index  Specifies the index for the attribute to get.
    @return  Returns the specified attribute value.
*/
    u32 GetAttribute(u32 index) const;

/*!
    @brief  Sets application-defined data.
    @details  Specify a size that does not exceed <tt>inet::APPLICATION_DATA_BUFFER_MAX_SIZE</tt>.
    @param[in] pData  Specifies the buffer for the application-defined data for the new session.
    @param[in] size  Specifies the size of the application-defined data the new session.
    @return  Returns a <tt>Result</tt> value indicating success if the application-defined data is set successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer was <tt>NULL</tt> or the specified data size was too big. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetApplicationData(const u8* pData, size_t size);
/*!
    @brief  Gets specified application-defined data.
    @param[in] pDataBuf  Specifies a pointer to a buffer from which to copy the specified application-defined data.
    @param[in] pDataSize  Specifies the size of the data written to the copy buffer.
    @param[in] dataBufSize  Specifies the copy buffer size.
    @return  Returns a <tt>Result</tt> value indicating success if the application-defined data is acquired successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer is <tt>NULL</tt> or the pointer returned by the copied size is <tt>NULL</tt>. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that no application-defined data is set. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultBufferShortage  Indicates that the set application-defined data cannot be copied because it is greater in size than the buffer. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetApplicationData(u8* pDataBuf, size_t* pDataSize, size_t dataBufSize) const;

/*!
    @brief  Gets the size of the specified application-defined data.
    @return  Returns the size of the application-defined data.
*/
    size_t GetApplicationDataSize() const
    {
        return m_ApplicationDataSize;
    }
/*!
    @brief  Sets whether participation in this session is possible.

    @details  Sets whether the new session is to be created as joinable. When not set, it is created as not joinable.

    @param[in] isOpen  When set to <tt>true</tt>, the new session created by this <tt>NexCreateSessionSetting</tt> function is joinable.
*/
    void SetOpenParticipation(bool isOpen);

/*!
    @cond PRIVATE
    @brief  Gets whether participation in this session is possible.

    @return  Returns <tt>true</tt> if the session is joinable.
*/
    bool GetOpenParticipation() const;
    //! @endcond

/*!
    @brief  Sets the user password.

    @details  Sets a user password for the new session.

    @param[in] pUserPassword  Specifies the user password.
    @param[in] passwordLength  Specifies the string length of the user password.
*/
    void SetSessionUserPassword(const wchar_t* pUserPassword, size_t passwordLength);

/*!
    @cond PRIVATE
    @brief
*/
    const wchar_t* GetSessionUserPassword() const;
    //! @endcond

/*!
    @brief  Sets the matchmaking priority when a session with the same requirements is found.

    @details  Sets the matchmaking priority for the new session.
     If unspecified, it is set to <tt>0</tt>.
     Enabled only when searches are done with
     <tt>nn::pia::inet::SELECTION_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY</tt> or <tt>SELECTION_METHOD_SCORE_BASED</tt>.
     If a value that exceeds <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is provided as an argument, <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is set instead.

    @param[in] priority  Specifies the matchmaking priority.
*/
    void SetSelectionPriority(u8 priority);

/*!
    @cond PRIVATE
    @brief  Gets the matchmaking priority.

    @details  Gets the matchmaking priority for the new session.

    @return  Returns the matchmaking priority.
*/
    u8 GetSelectionPriority() const;
    //! @endcond

/*!
    @brief  Sets the rating value to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @param[in] ratingValue  Specifies the rating value.
*/
    void SetRatingValue(u32 ratingValue);
/*!
    @cond PRIVATE
*/
    u32 GetRatingValue() const;
    bool IsSetRatingValue() const;
    //! @endcond

/*!
    @brief  Sets the disconnection rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] disconnectionRate  Specifies the disconnection rate.
*/
    void SetDisconnectionRate(u32 disconnectionRate);
/*!
    @cond PRIVATE
*/
    u32 GetDisconnectionRate() const;
    bool IsSetDisconnectionRate() const;
    //! @endcond

/*!
    @brief  Sets the violation rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] violationRate  Specifies the violation rate.
*/
    void SetViolationRate(u32 violationRate);
/*!
    @cond PRIVATE
*/
    u32 GetViolationRate() const;
    bool IsSetViolationRate() const;
    //! @endcond

/*!
    @brief  Sets whether to use position information (longitude and latitude, country code) for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  When not setting whether to use position information, the score for comparing distance uses the default score for rating.
    @param[in] b  Specifies whether to use position information.
*/
    void UseGeoIp(bool b);
/*!
    @cond PRIVATE
*/
    bool GetUseGeoIp() const;
    bool IsSetUseGeoIp() const;
    //! @endcond

/*!
    @brief  Sets the country code to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  Sets the country code to use when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available.
     The specified country code is not used when location information is available.
     When a country code is not set, and when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available, the default value is used as the score for comparing country codes.
    @param[in] countryCode  Specifies the country code.
*/
    void SetNintendoCountryCode(u32 countryCode);
/*!
    @cond PRIVATE
*/
    u32 GetNintendoCountryCode() const;
    bool IsSetNintendoCountryCode() const;
    //! @endcond

/*!
    @brief  Sets the IP address for debugging to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  Use this IP address instead of your own to get location information using <tt>GeoIP</tt>. Only available for debugging.
    @param[in] pIpAddress  Specifies the replacement IP address.
    @param[in] ipAddressLength  Specifies the IP address string length.
*/

    void SetOverrideIpAddress(const wchar_t* pIpAddress, size_t ipAddressLength);

/*!
    @cond PRIVATE
*/
    const wchar_t* GetOverrideIpAddress() const;
    bool IsSetOverrideIpAddress() const;
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Determines whether to set parameters for score-based matching.
*/
    bool IsSetScoreBasedParameter() const;
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Updates the header information that indicates the usage state of the application-defined data region.
*/
    void UpdateApplicationDataHeader(const u8* pData, size_t dataSize);
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Gets data from the application-defined data.
*/
    void GetApplicationDataInternal(void* pVectorTmp) const;
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Copies the application-defined data.
*/
    void CopyApplicationDataToVector(void* pVectorTmp) const;
    //! @endcond

    // Prints information that is useful for debugging.
    void Trace(u64 flag) const;


/*!
    @cond PRIVATE
*/
    void SetExtraParticipants(u16 num);
    u16 GetExtraParticipants() const;
    //! @endcond

/*!
        @brief  Sets the associated community ID.
        @details  This session can only be found in the search results for players who used the same community ID for searching in <tt>inet::NexSessionSearchCriteria::SetCommunityId</tt>.
         In addition, it is required when linking with the NEX ranking feature. For information about linking with the ranking feature, see the documentation about persistent gathering for the NEX.
*/
    void SetCommunityId(u32 gid);
/*!
        @cond PRIVATE
        @brief  Gets the associated community ID.
*/
    u32 GetCommunityId();
    //! @endcond

    // SetApplicationBuffer
    // OpenParticipate

protected:
    // Game mode.
    u32 m_GameMode;

    SessionType m_SessionType;

    // The description string.
    wchar_t m_Description[DESCRIPTION_DATA_BUFFER_MAX_SIZE + 1];

    // Attributes.
    u32 m_Attribute[ATTRIBUTE_MAX_SIZE];

    u8 m_ApplicationDataBuffer[APPLICATION_DATA_BUFFER_MAX_SIZE_INTERNAL];
    size_t m_ApplicationDataSize;
    size_t m_ApplicationDataInternalOffset;

    bool m_OpenParticipation;

    u8 m_ProgressScore;

    // Parameters for score-based matching.
    u32 m_RatingValue;
    bool m_IsSetRatingValue;

    u32 m_ViolationRate;
    bool m_IsSetViolationRate;

    u32 m_DisconnectionRate;
    bool m_IsSetDisconnectionRate;

    bool m_UseGeoIp;
    bool m_IsSetUseGeoIp;

    u32 m_CountryCode;
    bool m_IsSetCountryCode;

    wchar_t m_OverrideIpAddress[MAX_IP_ADDRESS_LENGTH + 1];
    bool m_IsSetOverrideIpAddress;

    u16 m_ExtraParticipants;

    u32 m_CommunityId;

#if NN_PIA_NEX_NEW_FEATURE_FOR_MATCHMAKE
    // Pointer to the user password configured for the new session.
    wchar_t m_UserPassword[nn::pia::inet::SESSION_USER_PASSWORD_LENGTH + 1];
#endif

private:
    // The copy constructor is sealed.
    NexCreateSessionSetting(const NexCreateSessionSetting&);

    // The assignment operator is sealed.
    NexCreateSessionSetting& operator=(const NexCreateSessionSetting&);
};
}
}
}  // End of namespace nn::pia::inet.
