﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexDebugSession.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>

#include <pia/inet/inet_NexSessionTypes.h>
#include <pia/inet/inet_NexCreateSessionSetting.h>
#include <pia/inet/inet_NexUpdateSessionSetting.h>
#include <pia/common/common_CallContext.h>

#if NN_PIA_CTR
#pragma diag_suppress 1300  //n1769: Pragma for ARMCC.
#endif
#include <nex.h>
#if NN_PIA_CTR
#pragma diag_warning 1300  //n1769: Pragma for ARMCC.
#endif

namespace nn
{
namespace pia
{
namespace session
{
// Forward declarations.
class IMatchmakeSession;
}
}
}
namespace nn
{
namespace pia
{
namespace inet
{
class NexDebugSessionJob;

/*!
@brief  (For debugging.) This class handles dummy sessions for debugging.
*/
class NexDebugSession : public common::RootObject
{
public:
/*!
    @brief  Creates an instance (singleton pattern).

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultNotInitialized  Indicates that the <tt>inet</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
*/
    static Result CreateInstance();

/*!
    @brief  Destroys the instance (singleton pattern).
    @details  This function does nothing and returns if called when an instance has not been created.
*/
    static void DestroyInstance();

/*!
    @brief  Gets a pointer to the <tt>NexDebugSession</tt> instance (singleton pattern).

    @return  Returns a <tt>NULL</tt> pointer if the instance cannot be created.
*/
    static NexDebugSession* GetInstance()
    {
        return s_pInstance;
    }
/*!
    @brief  Starts the <tt>NexDebugSession</tt> function.
    @details  Call this after <tt>NexFacade::Bind</tt>.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Create a <tt>NexFacade</tt> instance, and verify whether <tt>NexFacade::Bind</tt> is called. Programming error. Fix your program so that this error is not returned.
*/
    Result Startup();

/*!
    @brief  Cleans up the <tt>NexDebugSession</tt> feature.
    @details  Call before <tt>NexFacade::Unbind</tt>.
*/
    void Cleanup();

/*!
    @brief  Builds a dummy session.
    @details  This function starts an asynchronous process that creates a dummy session.
              The local station becomes the host of the created dummy session. The clients cannot participate in a dummy session.
              The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
              Use <tt>@ref IsCompletedCreateDummySession</tt> to check whether the process has completed and <tt>@ref GetCreateDummySessionResult</tt> to check the result.
              You can use <tt>@ref GetCreatedDummySessionId</tt> to get the session ID of the constructed dummy session.

    @param[in] createSessionSetting  Specifies the settings to use when creating the session.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that <tt>Session::Startup</tt> is not called. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedCreateDummySession, Checks the completion of <tt>GetCreateDummySessionResult</tt>, <tt>GetCreatedDummySessionId</tt>,
*/
    Result CreateDummySessionAsync(const NexCreateSessionSetting& createSessionSetting);

/*!
    @brief  and <tt>CreateDummySessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref CreateDummySessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CreateDummySessionAsync, <tt>GetCreateDummySessionResult</tt>
*/
    bool IsCompletedCreateDummySession();

/*!
        @brief  Gets the result of <tt>CreateDummySessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref CreateDummySessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
        @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultInvalidState  Attempted to construct 21 or more dummy sessions. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.

        @see  CreateDummySessionAsync, <tt>IsCompletedCreateDummySession</tt>
*/
    Result GetCreateDummySessionResult() const;

/*!
        @brief  Gets the session ID for the dummy session that was created.
        @return  Returns the session ID for the dummy session that was created.
*/
    u32 GetCreatedDummySessionId() const;

/*!
    @brief  Deletes a dummy session.
    @details  This function starts an asynchronous process that deletes a dummy session.
              The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
              Use <tt>@ref IsCompletedDestroyDummySession</tt> to check whether the process has completed and <tt>@ref GetDestroyDummySessionResult</tt> to check the result.

    @param[in] sessionId  The session ID for the dummy session that was deleted.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that <tt>Session::Startup</tt> is not called. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedDestroyDummySession, <tt>GetDestroyDummySessionResult</tt>
*/
    Result DestroyDummySessionAsync(u32 sessionId);

/*!
    @brief  Checks the completion of <tt>DestroyDummySessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref DestroyDummySessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  DestroyDummySessionAsync, <tt>GetDestroyDummySessionResult</tt>
*/
    bool IsCompletedDestroyDummySession();

/*!
        @brief  Gets the result of <tt>DestroyDummySessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref DestroyDummySessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
        @retval nn::pia::ResultInvalidState  Attempted to delete a session that was not the host. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNotFound  The session you tried to delete was not found. Handle appropriately in the application.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.

        @see  DestroyDummySessionAsync, <tt>IsCompletedDestroyDummySession</tt>
*/
    Result GetDestroyDummySessionResult() const;

/*!
    @brief  Updates dummy session settings.
    @details  Starts asynchronous processing for updating dummy session settings.
              The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
              Use <tt>@ref IsCompletedUpdateDummySession</tt> to check whether the process has completed and <tt>@ref GetUpdateDummySessionResult</tt> to check the result.

    @param[in] sessionId  The session ID for the dummy session for which settings are updated.
    @param[in] updateSessionSetting  Update settings for the session.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that <tt>Session::Startup</tt> is not called. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedUpdateDummySession, <tt>GetUpdateDummySessionResult</tt>
*/
    Result UpdateDummySessionAsync(u32 sessionId, const NexUpdateSessionSetting& updateSessionSetting);

/*!
    @brief  Checks the completion of <tt>UpdateDummySessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref UpdateDummySessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  UpdateDummySessionAsync, <tt>GetUpdateDummySessionResult</tt>
*/
    bool IsCompletedUpdateDummySession();

/*!
        @brief  Gets the result of <tt>UpdateDummySessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref UpdateDummySessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
        @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultInvalidState  Attempted to update a session that was not the host. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNotFound  The session you tried to update was not found. Handle appropriately in the application.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.

        @see  UpdateDummySessionAsync, <tt>IsCompletedUpdateDummySession</tt>
*/
    Result GetUpdateDummySessionResult() const;

/*!
    @cond  PRIVATE
*/
    /// @endcond

private:
    // A pointer to a singleton instance.
    static NexDebugSession* s_pInstance;

/*!
    @brief  This constructor is private because the singleton pattern is used.
*/
    NexDebugSession();

/*!
    @brief  This destructor is private because the singleton pattern is used.
*/
    virtual ~NexDebugSession();

/*!
    @brief  This copy constructor is private because the singleton pattern is used.
*/
    NexDebugSession(const NexDebugSession&);

/*!
    @brief  This assignment operator is private because the singleton pattern is used.
*/
    NexDebugSession& operator=(const NexDebugSession&);

    virtual Result initialize();
    virtual void finalize();

    nn::nex::MatchmakeExtensionClient m_MatchMakingClient;

    NexDebugSessionJob* m_pDebugSessionJob;
    bool m_IsStartedup;
    common::AsyncContext m_InternalAsyncContext;
    u32 m_AsyncProcessType;
};
}
}
}  // End of namespace nn::pia::inet.
