﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexFacade.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>
#include <pia/common/common_SignatureSetting.h>
#include <pia/transport/transport_NetworkFactory.h>

#if NN_PIA_NEX_3

#if NN_PIA_CTR
#pragma diag_suppress 1300
#endif

#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Transport/Interface/StationURL.h>
#include <RendezVous/Core/src/Client/Credentials.h>
#include <RendezVous/Core/src/Common/Constants/NotificationEvents.h>
#if NN_PIA_ENABLE_LOOSE_COUPLING_NEX
#include <RendezVous/Services/NATTraversal/src/Client/NATTraversalClient.h>
#endif

#include <RendezVous/Services/MatchMaking.h>

#include <RendezVous/Services/Jugem/Login/src/Client/NgsFacade.h>

#if NN_PIA_ENABLE_NGS_BRIDGE
// n1589: Not included in NgsFacade.h.
#include <RendezVous\/Services/Jugem/Login/src/Client/NgsBridgeInterface.h>
#endif

#include <RendezVous/Services/Jugem/Matchmake/src/Client/MatchmakeExtensionClient.h>

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
#include <RendezVous/Services/Jugem/MatchmakeReferee/src/Client/MatchmakeRefereeClient.h>
#endif

#if NN_PIA_CTR
#pragma diag_default 1300
#endif

#else  // Not NN_PIA_NEX_3.

#include <Platform/Core/String.h>
#include <Platform/Core/qList.h>
#include <Plugins/Transport/Interface/StationURL.h>
#include <Services/Foundation/Client/Credentials.h>
#include <Services/MatchMaking.h>
#if NN_PIA_ENABLE_LOOSE_COUPLING_NEX
#include <Services/NATTraversalClient.h>
#endif

#include <RVPackages/Jugem/Game/Login/Client/NgsFacade.h>
#include <RVPackages/Jugem/Game/Matchmake/Client/MatchmakeExtensionClient.h>

#endif  // end NN_PIA_NEX_3

namespace nn
{
namespace pia
{
namespace common
{
class String;
}
}
}

namespace nn
{
namespace pia
{
namespace transport
{
class StationConnectionInfo;
}
}
}

namespace nn
{
namespace pia
{
namespace inet
{
#if NN_PIA_ENABLE_LOOSE_COUPLING_NEX
class NexNatRelay;
#endif
class NexNatTraversalClient;
class NexNotificationDataJob;
class NatTraverser;
class NatProperty;
class SocketOutputStream;
class SocketInputStream;
class Socket;

/*!
  @brief  Represents the NEX library facade (interface wrapper) class.

  @details  This is not the facade for the <tt>nn::pia::inet::Nex*</tt> classes, but for the NEX library classes.

  @date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance()</tt> when an instance already exists.
  @date 2014-04-01  Added the <tt>Bind()</tt> and <tt>Unbind()</tt> functions and <tt>LoginInfo</tt>. In conjunction with that, modified the arguments for the <tt>Startup()</tt> function.
  @date 2014-03-25  Deleted the <tt>GetNatProperty()</tt> function. Added the <tt>GetNatPropertyMapping()</tt>, <tt>GetNatPropertyFiltering()</tt>, <tt>GetNatPropertyPortIncrement()</tt>, <tt>GetNatPropertyPublicPort()</tt>, and <tt>GetNatPropertyInterfacePort()</tt> functions.
  @date 2014-03-18  Deleted the <tt>ConvertNexStationURLToStationConnectionInfo()</tt> function. Made the <tt>ConvertNexInetAddressToInetAddress()</tt>, <tt>ConvertInetAddressToNexInetAddress()</tt>, and <tt>ConvertStationLocationToNexStationUrl()</tt> functions private.
  @date 2014-01-15  Made changes so that the name resolution process that was being done in the <tt>StartNatSessionAsync()</tt> function now uses the socket library directly.
  @date 2013-12-18  Added the <tt>StartNatSessionAsync()</tt>, <tt>IsCompletedStartNatSession()</tt>, <tt>GetStartNatSessionResult()</tt>, and <tt>CancelStartNatSessionAsync()</tt> functions.
  @date 2013-11-27  Deleted <tt>ResultDnsFailed</tt> from the <tt>StartNatSession()</tt> return values and added it to the <tt>Result</tt> values that could be returned through <tt>CallContext</tt> when asynchronous processes fail.
  @date 2013-09-18  Added the game server ID as the third parameter of the <tt>Startup()</tt> function.
  @date 2013-09-05  Added the fact that the NEX allocator is called inside the <tt>Startup()</tt> function.
  @date 2013-07-02  Changed the first argument of the <tt>Startup()</tt> function from <tt>nn::nex::Credentials*</tt> to <tt>nn::nex::NgsFacade*</tt>.
  @date 2012-12-03  Added <tt>ConvertNexSignatureKeyToSignatureSetting</tt>.
  @date 2012-12-03  Added <tt>ConvertNexStationUrlToStationConnectionInfo</tt>.
  @date 2012-12-03  Deprecated <tt>ConvertNexStationURLToStationConnectionInfo</tt>.
  @date 2012-12-03  Renamed the <tt>ConvertStationLocationToNexStationURL()</tt> function to <tt>ConvertStationLocationToNexStationUrl()</tt>.
  @date 2012-12-03  Renamed the <tt>ConvertNexStationURLToStationLocation()</tt> function to <tt>ConvertNexStationUrlToStationLocation()</tt>.
  @date 2012-11-29  Moved the parts of the <tt>nn::pia::transport::StationLocation</tt> function that bring in a dependency on NEX here.
  @date 2012-11-21  Added the <tt>ConvertStationLocationToNexStationURL()</tt> and <tt>ConvertNexStationURLToStationLocation()</tt> functions.
  @date 2012-11-21  Added the <tt>ConvertNexInetAddressToInetAddress()</tt> and <tt>ConvertInetAddressToNexInetAddress()</tt> functions.
  @date 2012-05-17  Added the <tt>NexFacade::UpdateGatheringID</tt> function to get the <tt>GatheringID</tt> of the matchmaking session from the application.
  @date 2012-05-16  Added parameters to the <tt>NexFacade::Startup function</tt>.
  @date 2012-04-06  Initial version.
 */
class NexFacade : public nn::pia::common::RootObject
{
public:
    //n1589: It must match NEX.
/*!
    @brief  The NAT port mapping type.
*/
    enum Mapping
    {
        MAPPING_UNKNOWN = 0,  //!<  An unknown type.
        MAPPING_EIM,          //!<  Specifies endpoint independent mapping (EIM).
        MAPPING_EDM           //!<  Specifies endpoint dependent mapping, symmetric (EDM).
    };

    //n1589: It must match NEX.
/*!
    @brief  The NAT filtering type.
*/
    enum Filtering
    {
        FILTERING_UNKNOWN = 0,  //!<  An unknown type.
        FILTERING_PIF,          //!<  Specifies port independent filtering (PIF).
        FILTERING_PDF           //!<  Specifies port dependent filtering (PDF).
    };

/*!
    @brief  Information about the NEX game server that was logged in to.
*/
    struct LoginInfo
    {
        LoginInfo()
            : pNgsFacade(NULL), gameServerId(0)
        {
        }
        nn::nex::NgsFacade* pNgsFacade;                                    //!<  The <tt>nn::nex::NgsFacade</tt> instance that was logged in to on the game server.
        u32 gameServerId;                                                  //!<  The game server ID specified when logging in to the game server.
    };

/*!
    @brief  Creates an instance (singleton pattern).

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultNotInitialized  Indicates that the <tt>inet</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
 */
    static nn::Result CreateInstance(void);

/*!
    @brief  Destroys the instance (singleton pattern).

    @details  This function does nothing and returns if called when an instance has not been created.
     Returns from the function without doing anything.
 */
    static void DestroyInstance(void);

/*!
    @brief  Gets the pointer to the <tt>NexFacade</tt> instance (singleton pattern).

    @return  Returns a null pointer if the instance cannot be created.
 */
    static NexFacade* GetInstance(void)
    {
        return s_pInstance;
    }

/*!
    @brief  Converts <tt>nn::nex::StationURL</tt> to <tt>@ref nn::pia::transport::StationConnectionInfo</tt>.

    @details  Converts the connection point information obtained from <tt>nn::nex::MatchmakeExtensionClient::GetMeshURLs</tt> to Pia station connection information.
     The converted station connection information can be used with <tt>@ref nn::pia::session::Mesh::JoinMeshAsync</tt>.

    @param[in] nexStationUrlList  Specifies the connection to convert. There must be one or two elements.
    @param[in] pStationConnectionInfo  Specifies the station connection to convert to.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
 */
    static nn::Result ConvertNexStationUrlToStationConnectionInfo(
        const nn::nex::qList<nn::nex::StationURL>& nexStationUrlList,
        nn::pia::transport::StationConnectionInfo* pStationConnectionInfo);

/*!
    @brief  Converts a <tt>NEX</tt> <tt>StationURL</tt> instance to a <tt>PiaTransport</tt> <tt>StationLocation</tt> instance.
    @param[in] url  Specifies the NEX <tt>StationURL</tt> instance.
    @param[out] pLocation  Specifies the <tt>StationLocation</tt> instance of <tt>PiaTransport</tt>.
    @see  ConvertStationLocationToNexStationUrl
 */
    static void ConvertNexStationUrlToStationLocation(const nn::nex::StationURL& url, transport::StationLocation* pLocation);

/*!
    @cond PRIVATE
    @brief  Converts a NEX <tt>InetAddress</tt> to a Pia <tt>InetAddress</tt>.

    @details  This function is thread-safe.

    @param[in] nexAddress  Specifies the NEX <tt>InetAddress</tt>.
    @param[out] pAddress  Specifies the Pia <tt>InetAddress</tt>.
    @endcond
 */
    static void ConvertNexInetAddressToInetAddress(const nn::nex::InetAddress& nexAddress, nn::pia::common::InetAddress* pAddress);

/*!
    @cond PRIVATE
    @brief  Converts a Pia <tt>InetAddress</tt> to a NEX <tt>InetAddress</tt>.

    @details  This function is thread-safe.

    @param[in] address  Specifies the Pia <tt>InetAddress</tt>.
    @param[out] pNexAddress  Specifies the NEX <tt>InetAddress</tt>.
    @endcond
 */
    static void ConvertInetAddressToNexInetAddress(const nn::pia::common::InetAddress& address, nn::nex::InetAddress* pNexAddress);

/*!
    @cond PRIVATE
    @brief  Converts a <tt>PiaTransport</tt> <tt>StationLocation</tt> instance to a NEX <tt>StationURL</tt> instance.
    @param[in] location  Specifies the <tt>StationLocation</tt> instance of <tt>PiaTransport</tt>.
    @param[out] pUrl  Specifies the NEX <tt>StationURL</tt> instance.
    @see  ConvertNexStationUrlToStationLocation
    @endcond
 */
    static void ConvertStationLocationToNexStationUrl(const transport::StationLocation& location, nn::nex::StationURL* pUrl);

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsBehindNat(const transport::StationLocation& location);

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsPublic(const transport::StationLocation& location);

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsGlobal(const transport::StationLocation& location);

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsEdmMapping(const transport::StationLocation& location);

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsEimMapping(const transport::StationLocation& location);

/*!
    @cond PRIVATE
    @endcond
 */
    static void ConvertStationLocationToStationUrlNexList(
        const nn::nex::qList<transport::StationLocation>& stationLocationList,
        nn::nex::qList<nn::nex::StationURL>* pStationUrlList);

/*!
    @cond PRIVATE
    @endcond
 */
    static void ConvertStationUrlToStationLocationNexList(
        const nn::nex::qList<nn::nex::StationURL>& stationUrlList,
        nn::nex::qList<transport::StationLocation>* pStationLocationList);

/*!
    @cond PRIVATE
    @brief  Converts an <tt>nex::String</tt> to a <tt>nn::pia::common::String</tt>.
    @param[in] nexString  Specifies the <tt>nex::String</tt> to convert.
    @param[out] pString  Specifies the <tt>nn::pia::common::String</tt> to convert to.
    @endcond
 */
    static void ConvertNexStringToString(const nn::nex::String& nexString, nn::pia::common::String* pString);

/*!
    @brief  Converts an <tt>nn::nex::qVector<nn::nex::qByte></tt> shared session key to an <tt>nn::pia::common::SignatureSetting</tt> instance.

    @details  Converts the shared session key obtained from the game server during matchmaking into a key for the class used to configure the packet-signing feature.
     Either configure the converted <tt>nn::pia::common::SignatureSetting</tt> instance with a buffer for the key beforehand, or use the <tt>@ref nn::pia::common::SignatureSettingWithKeyBuffer</tt> class.
     The converted packet-signing settings are required by the <tt>@ref nn::pia::session::Mesh::Startup</tt> function.

    @param[in] nexSignatureKey  Specifies the <tt>nn::nex::qVector<nn::nex::qByte></tt> instance to convert from.
    @param[out] pSignatureSetting  Specifies the <tt>nn::pia::common::SignatureSetting</tt> instance to convert to.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Make sure that the buffer in the <tt>nn::pia::common::SignatureSetting</tt> instance is at least as large as the original data. Programming error. Fix your program so that this error is not returned.
 */

    static nn::Result ConvertNexSignatureKeyToSignatureSetting(
        const nn::nex::qVector<nn::nex::qByte>& nexSignatureKey,
        nn::pia::common::SignatureSetting* pSignatureSetting);

/*!
    @cond PRIVATE
    @endcond
 */
    static void EnableLanNetwork(bool o)
    {
        PIA_TRACE_EX(common::TRACE_FLAG_INET, "EnableLanNetwork:%d", o);
        s_LanNetwork = o;
    }

/*!
    @cond PRIVATE
    @endcond
 */
    static bool IsLanNetwork()
    {
        return s_LanNetwork;
    }

/*!
    @cond PRIVATE
    @endcond
 */
    nn::Result CreateProtocols();

/*!
    @brief  Registers information for the game server that was logged in to.

    @details  This function registers the game server ID, and an instance of the <tt>nn::nex::NgsFacade</tt> class when a login to the NEX game server succeeds with the Pia library.
     Call it in the interval between logging in to the NEX game server and calling the <tt>NexFacade::Startup</tt> function.

    @param[in] pLoginInfo  Specifies a pointer to the NEX game server information <tt>LoginInfo</tt>.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @see  Unbind, Startup
*/
    virtual nn::Result Bind(LoginInfo* pLoginInfo);

/*!
    @brief  Deletes the information on the registered game server.

    @details  This function deletes the game server ID's registration information, and an instance of the <tt>nn::nex::NgsFacade</tt> class from the Pia library when a login to the NEX game server succeeds.
     Call this function in the interval between the call to the <tt>NexFacade::Cleanup</tt> function and logging out from the NEX game server.
    @see  Bind, Startup
*/
    virtual void Unbind();

/*!
    @brief  Notifies the game server of game-specific events.
    @details  Notifies friends logged in to the same game server of events that have been uniquely defined in the game.
     Log in to the gamer server and call this function in the interval between the calls to <tt>Bind</tt> and <tt>Unbind</tt>.
     Limit the frequency of event notifications so there are no more than 10 notifications in any minute.

    @param[in] gameNotificationType  The game-defined event type (<tt>nn::nex::NotificationEvents::GameNotificationEvent1</tt> through <tt>nn::nex::NotificationEvents::GameNotificationEvent8</tt>).
    @param[in] param1  Specifies an event parameter.
    @param[in] param2  Specifies an event parameter.
    @param[in] strParam  Specifies an event parameter string. Limit the string to no longer than 256 characters.
    @param[in] strParamLength  The length of the event parameter string. Limit the string to no longer than 256 characters.
    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
    Result UpdateNotificationDataAsync(nn::nex::NotificationEvents::NotificationEvents gameNotificationType, u32 param1, u32 param2, const wchar_t* strParam, size_t strParamLength);

/*!
    @brief  Checks for completion of asynchronous processing by <tt>UpdateNotificationDataAsync</tt>.
    @details  After <tt>UpdateNotificationDataAsync</tt> has been called, this function checks the completion of the asynchronous process.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise.
*/
    bool IsCompletedUpdateNotificationData() const;

/*!
    @brief  Gets the result of the <tt>UpdateNotificationDataAsync</tt> asynchronous process.
    @details  This function gets the result after <tt>UpdateNotificationDataAsync</tt> has been called, and <tt>IsCompletedUpdateNotificationData</tt> has confirmed that the process is complete.

    @return  Returns the result of the <tt>UpdateNotificationDataAsync</tt> asynchronous process. The following results are returned when the process ends.
    @retval ResultSuccess  The asynchronous process succeeded.
    @retval nn::pia::ResultInvalidState  Indicates that the <tt>UpdateNotificationDataAsync</tt> function was called at the wrong time, or the asynchronous process has not completed. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
*/
    Result GetUpdateNotificationDataResult() const;

/*!
    @brief  Gets the game-specific events that were sent as notifications to the game server.
    @details  Gets events, uniquely defined by the game, that friends logged in to the same game server have sent as notifications.
     Log in to the gamer server and call this function in the interval between the calls to <tt>Bind</tt> and <tt>Unbind</tt>.
     You normally get events using an event handler. Use this function to get events only during a login.

    @param[in] gameNotificationTypeBitMask  Specify the bitmask that represents the game-defined event type to acquire. The <i>n</i>th bit corresponds to <tt>nn::nex::NotificationEvents::GameNotificationEventN</tt>.
    @param[out] pRetrievedEventList  Specifies the location for storing the acquired event list. Maintain an instance for the specified list until asynchronous processing completes.
    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
    Result RetrieveNotificationDataAsync(u8 gameNotificationTypeBitMask, nn::nex::qList<nn::nex::NotificationEvent>* pRetrievedEventList);

/*!
    @brief  Checks for completion of the <tt>RetrieveNotificationDataAsync</tt> asynchronous process.
    @details  After calling <tt>RetrieveNotificationDataAsync</tt>, use this function to check the completion of the asynchronous process.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise.
*/
    bool IsCompletedRetrieveNotificationData() const;

/*!
    @brief  Gets the result of the <tt>RetrieveNotificationDataAsync</tt> asynchronous process.
    @details  This function gets the result after <tt>RetrieveNotificationDataAsync</tt> has been called, and <tt>IsCompletedRetrieveNotificationData</tt> has confirmed that the process is complete.

    @return  Returns the result of <tt>RetrieveNotificationDataAsync</tt> asynchronous processing. The following results are returned when the process ends.
    @retval ResultSuccess  The asynchronous process succeeded.
    @retval nn::pia::ResultInvalidState  Indicates that the <tt>RetrieveNotificationDataAsync</tt> function was called at the wrong time, or that the asynchronous process has not completed. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
*/
    Result GetRetrieveNotificationDataResult() const;

/*!
    @brief  Starts the <tt>NexFacade</tt> function. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  Call this function after logging in to the NEX game server and running the <tt>Bind</tt> function.
     Use the NEX allocator to create instances of classes used for internal processing inside this function. These instances are managed with the NEX library and are destroyed
     when the <tt>nn::nex::NgsFacade</tt> instance is destroyed. A new instance is not created if an instance of this class is already being managed by NEX.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @if CTR_DOC
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time or that the NEX thread mode is not supported by Pia. (The Pia library only supports unsafe transport buffer thread mode or unsafe user thread mode. We recommend using unsafe transport buffer thread mode with 3.<I>x</I> versions of NEX. Programming error. Fix your program so that this error is not returned.
    @endif
    @if CAFE_DOC
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time or that the NEX thread mode is not supported by Pia. (The Pia library only supports unsafe transport buffer thread mode.) Programming error. Fix your program so that this error is not returned.
    @endif
    @see  Bind
*/
    virtual nn::Result Startup();

/*!
    @brief  Cleans up the <tt>NexFacade</tt> function. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  Call before logging out from the NEX game server.
 */
    virtual void Cleanup();

/*!
    @cond PRIVATE
    @brief  Starts a NAT session (P2P communication). This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class. (To be removed in subsequent versions.)

    @details  Starts a NAT session and performs the necessary preprocessing for P2P communication.
     If calling this function, do so before calling the <tt>nn::pia::session::Mesh::Startup</tt> function.
     The port number used for P2P communication changes every time this function is run.

    The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
    You can monitor the completion and success or failure of the process by checking the <tt>CallContext</tt> function specified in the arguments.

    This function performs DNS name resolution on the server for NAT-type detection, but only the first time it is called after the application starts.
    The name resolution process runs in a background thread during asynchronous processes.
    @if CTR_DOC
    The name resolution process uses the <tt>nn::socket::GetAddrInfo()</tt> function of the socket library. Memory is allocated from the working memory specified using the <tt>nn::socket::Initialize()</tt> function when
    the name resolution process is run. That memory is later freed when the name resolution process ends.
    @endif
    @if CAFE_DOC
    The name resolution process uses the <tt>SOGetAddrInfo</tt> function of the socket library. Memory is allocated when the name resolution process is run. That memory
    is later freed when the name resolution process ends. <b>Important:</b> If the memory allocator for the name resolution process is registered from within the application using the
    <tt>SOResolverRegisterAllocator()</tt> function, the allocator is called from a background thread. The allocator must
    be thread-safe.
    @endif

    If the WAN cable of the router is unplugged before name resolution finishes, it might take 30 or more seconds for this asynchronous process to complete because it needs to wait for the name resolution process timeout.
    This process may take over 30 seconds to complete. The name resolution process cannot be canceled.

    @if CAFE_DOC
    When the call to this function succeeds, the background-download feature of the system is shut down.
    The background-download feature is then enabled when the <tt>StopNatSession()</tt> function is called.

    <b>Note:</b> Background downloads may still be performed after calling this function because background stress mode only depends on the DevMenu settings.
    @endif

    This function is deprecated. It will be removed in a future release. Use the <tt>@ref StartNatSessionAsync</tt> function instead.

    <b>The following results may be returned by the <tt>CallContext</tt> instance when the asynchronous process ends.</b>
    - <tt>nn::ResultSuccess</tt>: The asynchronous process succeeded.
    - <tt>nn::pia::ResultCancelled</tt>: The asynchronous process was canceled. Perform cleanup.
    - <tt>nn::pia::ResultInvalidState</tt>: Called at the wrong time. Programming error. Fix your program so that this error is not returned.
    - <tt>nn::pia::ResultNetworkConnectionIsLost</tt>: The network is not available. The wireless switch may be turned off or there may be a problem with the access point. Reconnect to AC.
    - <tt>nn::pia::ResultNatCheckFailed</tt>: The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    - <tt>nn::pia::ResultDnsFailed</tt>: DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.

    @param[in] pCallContext  Specifies the <tt>CallContext</tt> pointer to check the execution progress status.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result StartNatSession(
        common::CallContext* pCallContext);
    //! @endcond

/*!
    @brief  Starts a NAT session (P2P communication). This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  Starts a NAT session and performs the necessary preprocessing for P2P communication.
     If calling this function, do so before calling the <tt>nn::pia::session::Mesh::Startup</tt> function.
     The port number used for P2P communication changes every time this function is run.

    The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
    Use <tt>@ref IsCompletedStartNatSession</tt> to check whether the process has completed and <tt>@ref GetStartNatSessionResult</tt> to check the result.

    This function performs DNS name resolution on the server for NAT-type detection, but only the first time it is called after the application starts.
    The name resolution process runs in a background thread during asynchronous processes.
    @if CTR_DOC
    The name resolution process uses the <tt>nn::socket::GetAddrInfo()</tt> function of the socket library. Memory is allocated from the working memory specified using the <tt>nn::socket::Initialize()</tt> function when
    the name resolution process is run. That memory is later freed when the name resolution process ends.
    @endif
    @if CAFE_DOC
    The name resolution process uses the <tt>SOGetAddrInfo</tt> function of the socket library. Memory is allocated when the name resolution process is run. That memory
    is later freed when the name resolution process ends. <b>Important:</b> If the memory allocator for the name resolution process is registered from within the application using the
    <tt>SOResolverRegisterAllocator()</tt> function, the allocator is called from a background thread. The allocator must
    be thread-safe.
    @endif

    If the WAN cable of the router is unplugged before name resolution finishes, it might take 30 or more seconds for this asynchronous process to complete because it needs to wait for the name resolution process timeout.
    This process may take over 30 seconds to complete. The name resolution process cannot be canceled.

    @if CAFE_DOC
    When the call to this function succeeds, the background-download feature of the system is shut down.
    The background-download feature is then enabled when the <tt>StopNatSession()</tt> function is called.

    <b>Note:</b> Background downloads may still be performed after calling this function because background stress mode only depends on the DevMenu settings.
    @endif

    <b>The following results may be returned by the <tt>@ref GetStartNatSessionResult</tt> instance when the asynchronous process completes.</b>
    - <tt>nn::ResultSuccess</tt>: The asynchronous process succeeded.
    - <tt>nn::pia::ResultCancelled</tt>: The asynchronous process was canceled. Perform cleanup.
    - <tt>nn::pia::ResultInvalidState</tt>: Called at the wrong time. Programming error. Fix your program so that this error is not returned.
    - <tt>nn::pia::ResultNetworkConnectionIsLost</tt>: The network is not available. The wireless switch may be turned off or there may be a problem with the access point. Reconnect to AC.
    - <tt>nn::pia::ResultNatCheckFailed</tt>: The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    - <tt>nn::pia::ResultDnsFailed</tt>: DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result StartNatSessionAsync();

/*!
    @brief  Checks the completion of <tt>StartNatSessionAsync</tt> asynchronous processing. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  This function checks the completion of asynchronous processing after <tt>StartNatSessionAsync</tt> is called.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise.
*/
    virtual bool IsCompletedStartNatSession() const;

/*!
    @brief  Gets the result of <tt>StartNatSessionAsync</tt> asynchronous processing. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  This function gets the results after <tt>StartNatSessionAsync</tt> is called and <tt>IsCompletedStartNatSession</tt> confirms that the process has completed.

    @return  Returns the result of <tt>StartNatSessionAsync</tt> asynchronous processing. The following results are returned when the process ends.
    @retval nn::ResultSuccess  The asynchronous process succeeded.
    @retval nn::pia::ResultCancelled  The asynchronous process was canceled. Perform cleanup.
    @retval nn::pia::ResultInvalidState  The <tt>StartNatSessionAsync</tt> function was called at the wrong time, or the asynchronous process has not completed. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off or there may be a problem with the access point. Reconnect to AC.
    @retval nn::pia::ResultNatCheckFailed  The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultDnsFailed  DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
*/
    virtual nn::Result GetStartNatSessionResult() const;

/*!
    @brief  Cancels <tt>StartNatSessionAsync</tt> asynchronous processes. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  This function cancels asynchronous processing after <tt>StartNatSessionAsync</tt> is called.
     The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
     The library quickly acts to cancel the asynchronous process after this function is called, but check
     using the <tt>@ref IsCompletedStartNatSession()</tt> function that the cancellation has completed.

    @return  Returns the result of canceling <tt>StartNatSessionAsync</tt> asynchronous processing. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Returned when no <tt>StartNatSessionAsync</tt> asynchronous processing is being performed at the time this function is called. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result CancelStartNatSessionAsync();

/*!
    @brief  Stops a NAT session (P2P communication). This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  Call this function before calling the <tt>nn::pia::session::Mesh::Cleanup</tt> function.
 */
    virtual void StopNatSession();

/*!
    @cond PRIVATE
    @endcond
 */
    nn::Result CompleteStartNatSession(u16 port);

/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the trace flag. For more information, see the <tt>@ref TraceFlag</tt> API reference.
 */
    virtual void Trace(u64 flag) const;

/*!
    @brief  Checks whether the <tt>NexFacade</tt> feature is in a started state.

    @return  Returns <tt>true</tt> when started up.
 */
    bool IsStartedUp()
    {
        return m_IsStartedUp;
    }

/*!
    @brief  Checks whether the NAT session (P2P communication) has started.

    @return  Returns <tt>true</tt> if started.
 */
    bool IsStartedNatSession()
    {
        return m_IsStartedNatSession;
    }

/*!
    @cond PRIVATE
    @endcond
 */
    NatTraverser* GetNatTraverser()
    {
        return m_pNatTraverser;
    }

/*!
    @cond PRIVATE
    @brief  Gets a pointer to the <tt>MatchmakeExtensionClient</tt> instance.
    @return  Returns the pointer to <tt>MatchmakeExtensionClient</tt>.
    @endcond
 */
    nn::nex::MatchmakeExtensionClient* GetMatchMakingClient()
    {
        return &m_MatchMakingClient;
    }

#if NN_PIA_ENABLE_NGS_BRIDGE
/*!
    @cond PRIVATE
    @brief  Gets a pointer to the <tt>NgsFacade</tt> instance specified in the <tt>Startup</tt> function.

    @return  Returns a pointer to the <tt>NgsFacade</tt> instance.
    @endcond
 */
    nn::nex::NgsBridgeInterface* GetNgsInterface()
    {
        return m_pNgsInterface;
    }
#else
/*!
    @cond PRIVATE
    @brief  Gets a pointer to the <tt>NgsFacade</tt> instance specified in the <tt>Startup</tt> function.

    @return  Returns a pointer to the <tt>NgsFacade</tt> instance.
    @endcond
 */
    nn::nex::NgsFacade* GetNgsInterface()
    {
        return m_pNgsInterface;
    }
#endif

/*!
    @cond PRIVATE
    @brief  Gets the <tt>GameServerId</tt> specified in the <tt>Startup</tt> function.

    @return  Returns the <tt>GameServerId</tt>.
    @endcond
 */
    u32 GetGameServerId() const
    {
        return m_GameServerId;
    }

/*!
    @brief  Updates the <tt>GatheringID</tt> of the matchmaking session managed by <tt>NexFacade</tt>. This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

    @details  Any <tt>GatheringID</tt> updated by this function can be used by features such as host migration.

    @param[in] gatheringID  Specify the <tt>GatheringID</tt> of the matchmaking session you are in.
 */
    void UpdateGatheringID(nn::nex::GatheringID gatheringID);

/*!
    @cond PRIVATE
    @brief  Gets the <tt>GatheringID</tt> specified in the <tt>UpdateGatheringID</tt> function.

    @return  Returns the <tt>GatheringID</tt>.
    @endcond
 */
    nn::nex::GatheringID GetGatheringID()
    {
        return m_GatheringID;
    }

/*!
    @brief  Gets the local NAT port mapping type.
    @details  Gets the local NAT port mapping type.
     A correct value is available after the asynchronous process for <tt>NexFacade::StartNatSessionAsync</tt> completes.
    @return  Returns the NAT port mapping type.
 */
    Mapping GetNatPropertyMapping();
/*!
    @brief  Gets the local NAT filtering type.
    @details  Gets the local NAT filtering type.
     A correct value is available after the asynchronous process for <tt>NexFacade::StartNatSessionAsync</tt> completes.
    @return  Returns the NAT filtering type.
 */
    Filtering GetNatPropertyFiltering();
/*!
    @brief  Gets the local NAT port increment.
    @details  Gets the local NAT port increment.
     A correct value is available after the asynchronous process for <tt>NexFacade::StartNatSessionAsync</tt> completes.
    @return  Returns the NAT port increment.
 */
    u32 GetNatPropertyPortIncrement();
/*!
    @brief  Gets the Internet port for the local NAT.
    @details  Gets the Internet port for the local NAT.
     A correct value is available after the asynchronous process for <tt>NexFacade::StartNatSessionAsync</tt> completes.
    @return  Returns the Internet port for the NAT.
 */
    u16 GetNatPropertyPublicPort();
/*!
    @brief  Gets the LAN port for the local NAT.
    @details  Gets the LAN port for the local NAT.
     A correct value is available after the asynchronous process for <tt>NexFacade::StartNatSessionAsync</tt> completes.
    @return  Returns the LAN port for the NAT.
 */
    u16 GetNatPropertyInterfacePort();

/*!
    @cond PRIVATE
    @brief  Sets the base NAT-traversal timeout.

    @details  Standard applications must not change the timeout due to its significant effect on connectivity.

    @param[in] timeout  Specifies the timeout. The value is in milliseconds.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. The function returns this error when called between the <tt>Startup</tt> and <tt>Cleanup</tt> functions.
    @endcond
 */
    nn::Result SetNatTraversalTimeout(s32 timeout);

/*!
    @cond PRIVATE
    @endcond
 */
    s32 GetNatTraversalTimeout()
    {
        return m_NatTraversalTimeout;
    }

/*!
    @cond PRIVATE
    @brief  Sets the EDM-extension timeout for NAT traversal.

    @details  Standard applications must not change the timeout due to its significant effect on connectivity.

    @param[in] timeout  Specifies the timeout. The value is in milliseconds.

    @return  Returns the result of the function. If this function call fails, one or more of the following <tt>Result</tt> values is returned. You must make sure that the implementation of this function in your application does not return any errors.
    @retval nn::pia::ResultInvalidState  Indicates that the function was called at the wrong time. The function returns this error when called between the <tt>Startup</tt> and <tt>Cleanup</tt> functions.
    @endcond
 */
    nn::Result SetNatTraversalTimeoutEdmExtension(s32 timeout);

/*!
    @cond PRIVATE
    @endcond
 */
    s32 GetNatTraversalTimeoutEdmExtension()
    {
        return m_NatTraversalTimeoutEdmExtension;
    }


#if NN_PIA_CAFE
/*!
@cond PRIVATE
@brief  Sets whether to disable the background download feature of the system when there is a valid NAT session (during P2P communication).

@details  If <tt>true</tt> is specified for <tt><var>isSuspendBossDaemon</var></tt>, the system background download feature is suspended when the <tt>@ref StartNATSession()</tt>
function is called and re-enabled when the <tt>@ref StopNATSession()</tt> function is called. If you specify <tt>false</tt>, these functions have no effect on the operations of the background download feature.
Regardless of the setting, there is no effect on the foreground download feature of the BOSS library.
If you specify <tt>true</tt>, the library automatically calls the <tt>Initialize()</tt> and <tt>Finalize()</tt> functions of the <tt>ndm</tt> module.

The default is <tt>true</tt> (disable). If your application uses ndm features directly, use this function to change this
setting to <tt>false</tt> before calling the <tt>@ref StartNATSession()</tt> function.

@param[in] isSuspendBossDaemon  Specifies whether to disable the background download feature of the system when there is a valid NAT session. Set to <tt>true</tt> to cancel.
 */
    void SetSuspendBossDaemonOnNATSession(bool isSuspendBossDaemon)
    {
        m_IsSuspendBossDaemon = isSuspendBossDaemon;
    }
    //! @endcond

/*!
      @cond PRIVATE
      @brief  Gets whether the background download feature of the system is disabled when there is a valid NAT session (during P2P communication).

      @return  Returns whether calling the <tt>StartNatSession()</tt> function disables the background download feature of the system.
 */
    bool GetSuspendBossDaemonOnNATSession() const
    {
        return m_IsSuspendBossDaemon;
    }
//! @endcond
#endif

/*!
      @cond PRIVATE
      @brief  Adds an <tt>nn::nex::NotificationEventHandler</tt> instance to <tt>NgsFacade</tt> for internal Pia processing.

      @return  Returns whether the event handler was successfully added.
      @endcond
 */
    bool RegisterNexNotificationEventHandler4Pia(nn::nex::NotificationEventHandler* pEventHandler);

/*!
      @cond PRIVATE
      @brief  Removes the <tt>nn::nex::NotificationEventHandler</tt> instance from <tt>NgsFacade</tt> for internal Pia processing.

      @return  Returns whether the event handler was successfully removed.
      @endcond
 */
    bool UnregisterNexNotificationEventHandler4Pia(nn::nex::NotificationEventHandler* pEventHandler);

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    nn::nex::MatchmakeRefereeClient* GetMatchmakeRefereeClient()
    {
        return m_pMatchmakeRefereeClient;
    }
#endif

private:
    // A pointer to a singleton instance.
    static NexFacade* s_pInstance;
    static bool s_LanNetwork;

/*!
    @brief  This constructor is private because the singleton pattern is used.
 */
    NexFacade(void);

/*!
    @brief  This destructor is private because the singleton pattern is used.
 */
    virtual ~NexFacade(void);

/*!
    @brief  This copy constructor is private because the singleton pattern is used.
 */
    NexFacade(const NexFacade&);

/*!
    @brief  This assignment operator is private because the singleton pattern is used.
 */
    NexFacade& operator=(const NexFacade&);

    virtual nn::Result Initialize();
    virtual void Finalize();

#if NN_PIA_CAFE
    void suspendBossDaemon();
    void resumeBossDaemon();
#endif

    virtual nn::Result startNatSessionCore(common::CallContext* pCallContext);

    u32 m_GameServerId;
#if NN_PIA_ENABLE_NGS_BRIDGE
    nn::nex::NgsBridgeInterface* m_pNgsInterface;
#else
    nn::nex::NgsFacade* m_pNgsInterface;
#endif
    nn::nex::MatchmakeExtensionClient m_MatchMakingClient;
    nn::nex::GatheringID m_GatheringID;

    u8* m_pNatTraversalClientBuffer;
#if NN_PIA_ENABLE_LOOSE_COUPLING_NEX
    nn::nex::NATTraversalRelayClient* m_pNatTraversalRelayClient;
    NexNatRelay* m_pNexNatRelay;
#else
    NexNatTraversalClient* m_pNatTraversalClient;
#endif

    NatTraverser* m_pNatTraverser;

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    u8* m_pMatchmakeRefereeClientBuffer;
    nn::nex::MatchmakeRefereeClient* m_pMatchmakeRefereeClient;
#endif
    Socket* m_pSocket;
    SocketOutputStream* m_pOutputStream;
    SocketInputStream* m_pInputStream;

    bool m_IsStartedUp;
    bool m_IsStartedNatSession;

    s32 m_NatTraversalTimeout;
    s32 m_NatTraversalTimeoutEdmExtension;

#if NN_PIA_CAFE
    bool m_IsSuspendBossDaemon;
    s32 m_NdmInitializeCount;
#endif

    nn::nex::NotificationEventHandler* m_pNexNotificationEventHandler;

    NexNotificationDataJob* m_pNexNotificationDataJob;

    u8 m_InternalAsyncProcessType;
    common::AsyncContext m_InternalAsyncConext;
};
}
}
}  // End of namespace nn::pia::inet.
