﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexJoinSessionSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_JoinSessionSetting.h>
#include <pia/inet/inet_NexSessionTypes.h>

namespace nn
{
namespace pia
{
namespace inet
{
/*!
  @brief  Class that manages settings specific to NEX session participation.

  @date 2014-05-16  Added the <tt>SetSessionUserPassword()</tt> and <tt>SetSessionSystemPassword()</tt> functions.
  @date 2014-04-30  Added the <tt>SetSessionId()</tt> and <tt>GetSessionId()</tt> functions.
  @date 2014-02-03  Initial version.

  @see  nn::pia::session::JoinSessionSetting
*/

class NexJoinSessionSetting : public nn::pia::session::JoinSessionSetting
{
public:
/*!
    @brief  The default constructor.
*/
    NexJoinSessionSetting();

/*!
    @brief  The destructor.
*/
    virtual ~NexJoinSessionSetting();

/*!
    @brief  Configures a target session ID.
    @param[in] id  The target session ID.
*/
    virtual void SetSessionId(u32 id);

/*!
    @cond  PRIVATE
    @brief  Gets a target session ID.
    @return  Returns the target session ID.
*/
    virtual u32 GetSessionId() const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Sets the maximum number of participants in the target session.
    @param[in] num  The maximum number of participants in the target session.
*/
    virtual void SetMaxParticipants(u16 num);
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Gets the maximum number of participants in the target session.
    @return  Returns the maximum number of participants in the target session.
*/
    virtual u16 GetMaxParticipants() const;
    //! @endcond

/*!
    @brief  Sets the session's user password.
    @details  If a user password is already set for the session, this function confirms that they match. The default is an empty string.
    @param[in] pUserPassword  The user password.
    @param[in] passwordLength  The user password's string length.
*/
    void SetSessionUserPassword(const wchar_t* pUserPassword, size_t passwordLength);

/*!
    @cond  PRIVATE
    @brief  Gets the user password for the session.
    @return  Returns the user password for the session.
*/
    const wchar_t* GetSessionUserPassword() const;
    //! @endcond

/*!
    @brief  Sets the session's system password.
    @details  If a system password is already set for the session, this function confirms that they match. The default is an empty string.
    @param[in] pSystemPassword  The system password.
    @param[in] passwordLength  The system password's string length.
*/
    void SetSessionSystemPassword(const wchar_t* pSystemPassword, size_t passwordLength);

/*!
    @cond  PRIVATE
    @brief  Gets the system password for the session.
    @return  Returns the system password for the session.
*/
    const wchar_t* GetSessionSystemPassword() const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Prints information that is useful for debugging.
*/
    virtual void Trace(u64 flag) const;
    //! @endcond

protected:
#if NN_PIA_NEX_NEW_FEATURE_FOR_MATCHMAKE
    // Pointer to the user password configured for the new session.
    wchar_t m_UserPassword[nn::pia::inet::SESSION_USER_PASSWORD_LENGTH + 1];
    wchar_t m_SystemPassword[nn::pia::inet::SESSION_SYSTEM_PASSWORD_LENGTH + 1];
#endif

private:
    u32 m_SessionId;
    u16 m_MaxParticipants;

    // The copy constructor is sealed.
    NexJoinSessionSetting(const NexJoinSessionSetting&);

    // The assignment operator is sealed.
    NexJoinSessionSetting& operator=(const NexJoinSessionSetting&);
};
}
}
}  // End of namespace nn::pia::inet.
