﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     inet_NexLanNetworkFactory.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/inet/inet_definitions.h>
#include <pia/transport/transport_NetworkFactory.h>

namespace nn
{
namespace pia
{
namespace inet
{

/*!
@cond PRIVATE
@brief  This class creates the necessary classes for networks using the NEX library (LAN communication) (factory pattern).

@details  This feature is for library developers. It is normally not available to application developers.
If the game server is on the same LAN, applications can execute network-related processes that are compatible with the NEX library (such as sending and receiving data, and managing sessions) without any NAT traversal. To do so, an application must pass an instance of the <tt>NexLanNetwork</tt> factory class to the <tt>@ref nn::pia::transport::Transport</tt> or <tt>@ref nn::pia::session::Session</tt> class objects.

@see  nn::pia::transport::Transport::Setting, nn::pia::session::Mesh::Setting

@date 2014-08-01  Added the <tt>CreateMatchOpenRecruitmentJob</tt> and <tt>CreateMatchCloseRecruitmentJob</tt> functions.
@date 2013-07-17  Initial version.


*/
class NexLanNetworkFactory : public nn::pia::transport::NetworkFactory
{
public:

/*!
@brief  Instantiates an object.
*/
    NexLanNetworkFactory() {}

/*!
@brief  Destroys the object.
*/
    virtual ~NexLanNetworkFactory() {}

/*!
@brief  Instantiates an <tt>InputStream</tt> object.
*/
    virtual nn::pia::common::IPacketInput* CreateInputStream();

/*!
@brief  Instantiates an <tt>OutputStream</tt> object.
*/
    virtual nn::pia::common::IPacketOutput* CreateOutputStream();

/*!
@brief  Instantiates a <tt>ConnectStationJob</tt> object.
*/
    virtual nn::pia::transport::ConnectStationJob* CreateConnectStationJob();

/*!
@brief  Instantiates a <tt>DisconnectStationJob</tt> object.
*/
    virtual nn::pia::transport::DisconnectStationJob* CreateDisconnectStationJob();

/*!
@brief  Instantiates a <tt>CreateMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateCreateMeshJob();

/*!
@brief  Instantiates a <tt>JoinMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateJoinMeshJob();

/*!
@brief  Instantiates a <tt>LeaveMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateLeaveMeshJob();

/*!
@brief  Instantiates a <tt>GenerateMatchmakeSystemPasswordJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateGenerateMatchmakeSystemPasswordJob();

/*!
@brief  Instantiates a <tt>ClearMatchmakeSystemPasswordJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateClearMatchmakeSystemPasswordJob();

/*!
@brief  Gets the length of the session's <tt>MatchmakeSessionSystemPassword</tt>.
Generates a maintenance region from this value.
*/
    virtual u32 GetMatchmakeSessionSystemPasswordSize() const;

/*!
@brief  Generates a maintenance region for the session's <tt>MatchmakeSessionSystemPassword</tt>.
*/
    virtual wchar_t* CreateMatchmakeSessionSystemPassword(u32 size);

/*!
@brief  Builds the required protocol.
*/
    virtual nn::Result CreateProtocols();

/*!
@brief  Instantiates a <tt>ProcessHostMigrationJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateProcessHostMigrationJob();

/*!
@brief  Instantiates a <tt>LeaveWithHostMigrationJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateLeaveWithHostMigrationJob();

/*!
@brief  Creates a <tt>KickoutManageJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateKickoutManageJob();

/*!
@brief  Constructs a concrete instance of the <tt>PacketHandler</tt> class.
*/
    virtual transport::PacketHandler* CreatePacketHandler();

/*!
@brief  Gets whether a signature is required.
*/
    virtual bool IsSignatureRequired() const;

/*!
@brief  Creates a <tt>Storage</tt> object for setting the signature.
*/
    virtual nn::pia::session::SignatureSettingStorage* CreateSignatureSettingStorage();

/*!
@brief  Gets whether multiple-candidate host migration is allowed.
*/
    virtual bool IsMultiCandidatePermitted() const;

/*!
@brief  Gets whether the use of the relay communication feature is allowed.
*/
    virtual bool IsRelayConnectionPermitted() const;

/*!
@brief  Gets whether the use of joint session features is allowed.
*/
    virtual bool IsJointSessionPermitted() const;

/*!
@brief  Gets the total size of the header portion of packets at the network layer. The value is in bytes.
*/
    virtual size_t GetHeaderSize() const
    {
        return 20 + 8;
    }  // IP header (20 bytes) + UDP header (8 bytes)

/* For the auto-matchmaking feature. */
/*!
@brief  Instantiates a <tt>CreateSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchCreateSessionJob();


/*!
@brief  Instantiates an <tt>AutoMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchAutoMatchmakeSessionJob();


/*!
@brief  Instantiates a <tt>BrowseMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchBrowseMatchmakeSessionJob();


/*!
@brief  Instantiates a <tt>JoinSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchJoinSessionJob();


/*!
@brief  Instantiates a <tt>LeaveSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchLeaveSessionJob();


/*!
@brief  Instantiates a <tt>DestroySessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchDestroySessionJob();

/*!
@brief  Instantiates a <tt>JointSessionJob</tt> object for joint sessions.
*/
    virtual common::StepSequenceJob* CreateJointSessionJob();

/*!
@brief  Instantiates a <tt>ModifyAttributeJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateModifyAttributeJob();

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateSessionSetting</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateSessionSettingJob();
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateApplicationBufferJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateApplicationDataJob();
    //! @endcond

/*!
@brief  Instantiates an <tt>IMatchmakeSession</tt> object for the session.
*/
    virtual session::IMatchmakeSession* CreateMatchmakeSession();

/*!
@brief  Instantiates an <tt>ISessionInfoList</tt> object for the session.
*/
    virtual session::ISessionInfoList* CreateSessionInfoList(u8 browseSessionInfoListNum);

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    virtual session::IMatchmakeReferee* CreateMatchmakeReferee();

    virtual common::StepSequenceJob* CreateRefereeStartGameJob();

    virtual common::StepSequenceJob* CreateRefereeEndGameJob();

    virtual common::StepSequenceJob* CreateRefereeEndGameWithoutReportJob();

    virtual common::StepSequenceJob* CreateRefereeGetOrCreateStatsJob();
#endif
/*!
@brief  Instantiates a <tt>MeshLayerController</tt> object for the session.
*/
    virtual session::MeshLayerController* CreateMatchMeshLayerController();

    virtual bool GetNecessaryBind() const;

    virtual bool GetNecessaryServer() const;

    virtual session::HostMigrationMode GetHostMigrationMode() const;

    virtual u8 GetBrowseSessionInfoListMax() const;

/* For the auto-matchmaking feature. */

protected:
private:
};
/*! @endcond */
}
}
}  // end of namespace nn::pia::inet
