﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexSessionInfo.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_ISessionInfo.h>
#include <pia/inet/inet_NexSessionTypes.h>
#include <pia/common/common_DateTime.h>

namespace nn
{
namespace pia
{
namespace inet
{

/*!
    @brief  Data class for Internet communication sessions.
*/
class NexSessionInfo : public session::ISessionInfo
{
public:
    NexSessionInfo();
    virtual ~NexSessionInfo();

/*!
    @brief  Gets the session's game mode.
    @return  Returns the session's game mode.
*/
    virtual u32 GetGameMode() const;

/*!
    @brief  Gets the session's ID.
    @return  Returns the session ID.
*/
    virtual u32 GetSessionId() const;

/*!
    @brief  Gets the number of people participating in a session.
    @return  Returns the number of people participating in a session.
*/
    virtual u32 GetCurrentParticipants() const;

/*!
    @brief  Gets the minimum number of people who can participate in a session.
    @return  Returns the minimum number of people who can participate in a session.
*/
    virtual u32 GetMinParticipants() const;

/*!
    @brief  Gets the maximum number of people who can participate in a session.
    @return  Returns the maximum number of people who can participate in a session.
*/
    virtual u32 GetMaxParticipants() const;

/*!
    @brief  Gets whether a session is open for recruitment.
    @return  Returns <tt>true</tt> if the session is open for recruitment, and returns <tt>false</tt> if it is closed.
*/
    virtual bool IsOpened() const;

/*!
    @brief  Gets the session's application-defined data.
    @details  Gets the application-defined data configured for the session.
    @param[out] pBuffer  Pointer to the buffer for copying the session's application-defined data.
    @param[in] size  Specifies the size of the buffer required for copying the session's application-defined data.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result GetApplicationDataBuf(void* pBuffer, size_t size);

/*!
    @brief  Gets the length of the session's application-defined data.
    @return  Returns the length of the session's application-defined data.
*/
    virtual size_t GetApplicationDataSize() const;

/*!
    @brief  Gets the session's attributes.
    @param[out] pAttribute  Pointer to the attribute container for the session for the specified index.
    @param[in] idx  Specifies the index for the attribute to get.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result GetAttribute(u32* pAttribute, u8 idx) const;

/*!
    @brief  Gets the session's description string.
    @param[out] pBuffer  Pointer to the buffer for copying the session's description string.
    @param[in] size  Size of the buffer for copying the session's description string.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result GetDescription(wchar_t* pBuffer, size_t size) const;

/*!
    @brief  Gets the length of the session's description string.
    @return  Returns the length of the session's description string.
*/
    virtual size_t GetDescriptionSize() const;

/*!
    @brief  Checks whether a user password is set for the session.
    @return  Returns <tt>true</tt> if a user password is set.
*/
    virtual bool IsRestrictedByUserPassword() const;

/*!
    @cond PRIVATE
    @brief  Checks whether a system password is set for the session.
    @return  Returns <tt>true</tt> if a system password is set.
*/
    virtual bool IsRestrictedBySystemPassword() const;
    /// @endcond

/*!
    @brief  Gets the session type.
    @return  Returns the configured session type.
*/
    virtual SessionType GetSessionType() const;

/*!
    @brief  Gets the principal ID of the session host.
    @return  Returns the principal ID of the session host.
*/
    virtual u32 GetHostPrincipalId() const;

/*!
    @brief  Gets the principal ID specified in the session search conditions.
    @return  Returns the principal ID specified in the session search conditions.
*/
    virtual u32 GetTargetPrincipalId() const;

/*!
    @brief  Gets the matchmaking priority for the session.
    @return  Returns the matchmaking priority for the session.
*/
    virtual u8 GetSelectionPriority() const;

/*!
    @brief  Gets the clock to use when the session begins.
    @return  Returns the clock to use when the session begins.
*/
    virtual const common::DateTime& GetStartedTime() const;

/*!
    @brief  Clears session information.
*/
    virtual void Clear();

/*!
    @brief  Copies session information.
    @details  Use this when you want the application to retain session information.
    @param[in] rhs  Session information to copy.
*/
    virtual void Copy(const NexSessionInfo& rhs);

/*!
    @brief  Prints information that is useful for debugging.
    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
 */
    virtual void Trace(u64 flag);

/*!
    @cond PRIVATE
*/
    virtual void SetGameMode(u32 gameMode);
    virtual void SetSessionId(u32 sessionId);
    virtual void SetCurrentParticipants(u32 participants);
    virtual void SetMinParticipants(u32 participants);
    virtual void SetMaxParticipants(u32 participants);
    virtual void SetOpen(bool isOpened);
    virtual void SetAttribute(u32 attribute, u8 index);
    virtual void SetDescription(const void* pDesc, size_t size);
    virtual void SetApplicationData(const void* pData, size_t size);
    virtual void SetUserPasswordEnabled(bool isEnabled);
    virtual void SetSystemPasswordEnabled(bool isEnabled);
    virtual void SetSessionType(SessionType sessionType);
    virtual void SetHostPrincipalId(u32 principalId);
    virtual void SetTargetPrincipalId(u32 principalId);
    virtual void SetSelectionPriority(u8 priority);
    virtual void SetStartedTime(const common::DateTime& time);
    /// @endcond

private:
    u32 m_GameMode;
    u32 m_SessionId;
    u32 m_CurrentParticipants;
    u32 m_MinParticipants;
    u32 m_MaxParticipants;
    bool m_bIsOpened;

    u32 m_Attribute[ATTRIBUTE_MAX_SIZE];
    wchar_t m_Description[DESCRIPTION_DATA_BUFFER_MAX_SIZE + 1];
    size_t m_DescriptionSize;
    u8 m_ApplicationData[APPLICATION_DATA_BUFFER_MAX_SIZE_INTERNAL];
    size_t m_ApplicationDataSize;
    bool m_bUserPasswordEnabled;
    bool m_bSystemPasswordEnabled;

    SessionType m_SessionType;
    u32 m_HostPrincipalId;
    u32 m_TargetPrincipalId;
    u8 m_SelectionPriority;
    common::DateTime m_StartedTime;
};
}
}
}  // End of namespace nn::pia::inet.
