/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexSessionSearchCriteria.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_SessionSearchCriteria.h>
#include <pia/inet/inet_NexSessionTypes.h>

#if NN_PIA_CTR
#pragma diag_suppress 1300  //n1769: Pragma for ARMCC.
#endif
#include <nex.h>
#if NN_PIA_CTR
#pragma diag_warning 1300  //n1769: Pragma for ARMCC.
#endif

namespace nn
{
namespace pia
{
namespace inet
{

/*!
  @brief  The search condition class used when searching for a session.

  @date 2015-03-20  Added APIs to set parameters for scored matchmaking.
  @date 2014-02-03  Initial version.

  @see  nn::pia::session::SessionSearchCriteria
*/
class NexSessionSearchCriteria : public nn::pia::session::SessionSearchCriteria
{
public:
    enum NexCondMask
    {
        NexCondMask_MinParticipation = 0,
        NexCondMask_MaxParticipation,
        NexCondMask_OpenedOnly,
        NexCondMask_VacantOnly,
        NexCondMask_GameMode,
        NexCondMask_SessionType,
        NexCondMask_Attribute,
        NexCondMask_ExcludeNonHost,
        NexCondMask_ExcludeUserPassword,
        NexCondMask_ScoreBasedSettingIndex,
        NexCondMask_ScoreBasedRatingValue,
        NexCondMask_ScoreBasedDisconnectionRate,
        NexCondMask_ScoreBasedViolationRate,
        NexCondMask_ScoreBasedCountryCode,
        NexCondMask_ScoreBasedGeoIp,
        NexCondMask_ScoreBasedOverrideIp,
        NexCondMask_Community,
        NexCondMask_Max
    };

/*!
    @brief  The default constructor.
*/
    NexSessionSearchCriteria();

/*!
    @brief  The destructor.
*/
    virtual ~NexSessionSearchCriteria();

/*!
    @brief  Sets the minimum participants in the search criteria.
    @param[in] min  The minimum participants value to set in the search criteria.
*/
    void SetMinParticipants(u16 min);

/*!
    @brief  Sets a range of minimum participants in the search criteria.
    @param[in] min  The minimum value of the range to set for minimum participants in the search criteria.
    @param[in] max  The maximum value of the range to set for minimum participants in the search criteria.
*/
    void SetMinParticipantsRange(u16 min, u16 max);

/*!
    @brief  Sets the maximum participants in the search criteria.
    @param[in] max  The maximum participants value to set in the search criteria.
*/
    void SetMaxParticipants(u16 max);

/*!
    @brief  Sets a range of maximum participants in the search criteria.
    @param[in] min  The minimum value of the range to set for maximum participants in the search criteria.
    @param[in] max  The maximum value of the range to set for maximum participants in the search criteria.
*/
    void SetMaxParticipantsRange(u16 min, u16 max);

/*!
    @brief  Sets whether to search only for sessions that are open to participants in the search criteria.
    @param[in] bOpenedOnly  Specify <tt>true</tt> to search only for sessions that are open to participants.
*/
    void SetOpenedOnly(bool bOpenedOnly);

/*!
    @brief  Sets whether to search only for sessions that have openings in the search criteria.
    @param[in] bVacantOnly  Specify <tt>true</tt> to search only for sessions that have openings.
*/
    void SetVacantOnly(bool bVacantOnly);

/*!
    @brief  Sets the search conditions to game mode.
    @details  When no search conditions are set, it matches the game mode of any value.
    @param[in] gameMode  Specifies the game mode search conditions.
*/
    void SetGameMode(u32 gameMode);

/*!
    @brief  Sets the session type search conditions.
    @details  If no conditions are specified, the default value (<tt>SESSION_TYPE_ANYBODY</tt>) will be used.
    @param[in] sessionType  Specifies the session type search conditions.
*/
    void SetSessionType(SessionType sessionType);

/*!
    @brief  Sets a value to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] value  Specifies the value of an attribute.
*/
    void SetAttribute(u32 index, u32 value);
/*!
    @brief  Sets multiple values to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] value  Specifies a list of values for an attribute.
    @param[in] attributeNum  Specifies the number of values for an attribute.
*/
    void SetAttribute(u32 index, u32* value, size_t attributeNum);
/*!
    @brief  Sets a range of values to use when searching for a single attribute.
    @param[in] index  Specifies the index of an attribute.
    @param[in] min  Specifies the minimum value of an attribute.
    @param[in] max  Specifies the maximum value of an attribute.
*/
    void SetAttributeRange(u32 index, u32 min, u32 max);
/*!
    @brief  Sets the selection methods for joining sessions when random matchmaking is in effect.
    @param[in] selectionMethod  Specifies the selection method to set for the session.
*/
    void SetRandomSessionSelectionMethod(SelectionMethod selectionMethod);

/*!
    @brief  Sets whether to exclude sessions with no hosts from the search. (By default, they are excluded.)
    @param[in] bExclude  Specifies whether to exclude sessions with no hosts from the search. Set to <tt>true</tt> to enable.
*/
    void SetExcludeNonHostPid(bool bExclude);

/*!
    @brief  Sets whether to exclude sessions for which a user password was set from the search. (By default, they are excluded.)
    @param[in] bExclude  Specifies whether to exclude sessions for which a user password was set from the search. Set to <tt>true</tt> to enable.
*/
    void SetExcludeUserPasswordSet(bool bExclude);

/*!
    @brief  Sets the setting index for the score calculation method to use during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  <tt>0</tt> is used when a setting index is not set.
    @param[in] settingIndex  Specifies the score calculation setting index.
*/
    void SetScoreSettingIndex(u32 settingIndex);

/*!
    @brief  Sets the rating value to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  When not set, the score for the rating value difference uses a default score to rate.
    @param[in] ratingValue  Specifies the rating value.
*/
    void SetRatingValue(u32 ratingValue);

/*!
    @brief  Sets the disconnection rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  When not set, the score for the disconnection rate difference uses a default score to rate. We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] disconnectionRate  Specifies the disconnection rate.
*/
    void SetDisconnectionRate(u32 disconnectionRate);

/*!
    @brief  Sets the violation rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  When not set, the score for the violation rate difference uses a default score to rate. We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] violationRate  Specifies the violation rate.
*/
    void SetViolationRate(u32 violationRate);

/*!
    @brief  Sets whether to use position information (longitude and latitude, country code) for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  When not setting whether to use position information, the score for comparing distance uses the default score for rating.
    @param[in] b  Specifies whether to use position information.
*/
    void SetUseGeoIp(bool b);

/*!
    @brief  Sets the country code to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  Sets the country code to use when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available.
                The specified country code is not used when location information is available.
                When a country code is not set, and when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available, the default value is used as the score for comparing country codes.
    @param[in] countryCode  Specifies the country code.
*/
    void SetNintendoCountryCode(u32 countryCode);

/*!
    @brief  Sets the IP address for debugging to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  Use this IP address instead of your own to get location information using <tt>GeoIP</tt>. Only available for debugging.
    @param[in] pIpAddress  Specifies the replacement IP address.
    @param[in] ipAddressLength  Specifies the IP address string length.
*/
    void SetOverrideIpAddress(const wchar_t* pIpAddress, size_t ipAddressLength);

/*!
        @brief  Sets the ID for the associated community (persistent gathering).
*/
    void SetCommunityId(u32 gid);
/*!
        @cond  PRIVATE
        @brief  Gets the ID for the associated community (persistent gathering).
*/
    u32 GetCommunityId();
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Converts the configured information to the condition and search range criteria used by the NEX classes.
*/
    void ConvertToNexMatchmakeSessionSearchCriteria(
        nn::nex::MatchmakeSessionSearchCriteria* pCriteria,
        nn::nex::MatchmakeParam* pParam,
        nn::nex::ResultRange* pResultRange,
        int reportNum = 2);
    //! @endcond

/*!
    @brief  Assignment operator.
    @param[in] r  The object to assign.
    @return  See this object.
*/
    NexSessionSearchCriteria& operator=(const NexSessionSearchCriteria& r);

/*!
    @cond  PRIVATE
    @brief  Prints information that is useful for debugging.
*/
    virtual void Trace(u64 flag) const;
    //! @endcond

private:
    // The copy constructor is sealed.
    NexSessionSearchCriteria(const NexSessionSearchCriteria&);

    // Invalid number of participants setting.
    static const u16 INVALID_PARICIPANTS_NUM = 0xffff;

    // Minimum participants range.
    u16 m_MinParticipantsRangeMax;
    u16 m_MinParticipantsRangeMin;

    // Maximum participants range.
    u16 m_MaxParticipantsRangeMax;
    u16 m_MaxParticipantsRangeMin;

    // Whether to list only open sessions.
    bool m_bOpenedOnly;
    // Whether to remove full sessions from the results.
    bool m_bVacantOnly;

    // Game mode.
    u32 m_GameMode;
    // Session type.
    SessionType m_SessionType;

    // Attributes.
    u32 m_AttributeList[ATTRIBUTE_MAX_SIZE][ATTRIBUTE_ARRAY_MAX_SIZE];
    size_t m_AttributeSizeList[ATTRIBUTE_MAX_SIZE];
    u32 m_AttributeRangeMin[ATTRIBUTE_MAX_SIZE];
    u32 m_AttributeRangeMax[ATTRIBUTE_MAX_SIZE];
    bool m_bAttributeRangeUsed[ATTRIBUTE_MAX_SIZE];

    // Exclude sessions for which a host is not yet set.
    bool m_bExcludeNonHostPid;
    // Exclude sessions for which a user password is set.
    bool m_bExcludeUserPasswordSet;
    // Set the way to choose the session to join from the search results.
    SelectionMethod m_SelectionMethod;

    u32 m_ScoreBasedSettingIndex;
    u32 m_ScoreBasedRatingValue;
    u32 m_ScoreBasedDisconnectionRate;
    u32 m_ScoreBasedViolationRate;
    u32 m_ScoreBasedCountryCode;
    bool m_IsScoreBasedGeoIpUsed;

    wchar_t m_IpAddress[MAX_IP_ADDRESS_LENGTH + 1];

    u32 m_CommunityId;

    u32 m_SetCondMask;
};
}
}
}  // End of namespace nn::pia::inet.
