﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: inet_NexUpdateSessionSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_UpdateSessionSetting.h>
#include <pia/inet/inet_NexSessionTypes.h>

#if NN_PIA_CTR
#pragma diag_suppress 1300  //n1769: Pragma for ARMCC.
#endif
#include <nex.h>
#if NN_PIA_CTR
#pragma diag_warning 1300  //n1769: Pragma for ARMCC.
#endif

namespace nn
{
namespace pia
{
namespace inet
{
/*!
  @brief  Base class that manages information specific to updating session conditions.
  @details  Only the settings of the value specified for the update are updated. No specification is needed for settings that are not being updated.

  @date 2015-03-20  Added APIs to set parameters for scored matchmaking.
  @date 2014-01-21  Initial version.
*/
class NexUpdateSessionSetting : public nn::pia::session::UpdateSessionSetting
{
public:
/*!
    @brief  The default constructor.
*/
    NexUpdateSessionSetting();

/*!
    @brief  The destructor.
*/
    virtual ~NexUpdateSessionSetting();

/*!
    @brief  Specifies the session type for the session.
    @details  Updates are only possible from <tt>inet::SESSION_TYPE_FRIEND</tt> to <tt>inet::SESSION_TYPE_ANYBODY</tt>.
    @param[in] sessionType  Specifies the session type.
*/
    void SetSessionType(SessionType sessionType);
/*!
    @brief  Gets the specified session type.
    @return  Returns the specified session type.
*/
    SessionType GetSessionType() const;

/*!
    @brief  Specifies the description string for the session.
    @param[in] description  The description string.
*/
    void SetDescription(const wchar_t* description);
/*!
    @brief  Gets the specified description string.
    @return  Returns the specified description string.
*/
    const wchar_t* GetDescription() const;

/*!
    @cond  PRIVATE
    @brief  Sets attributes for the session.
    @param[in] index  Specifies the index of the attribute for the session.
    @param[in] value  Specifies the value of the attribute for the session.
*/
    //void SetAttribute(u32 index, u32 value);
    //! @endcond

/*!
    @brief  Sets attributes for the session.
    @param[in] attributes  Specifies an array of attributes.
*/
    //void SetAttribute(u32 index, u32 value);
    void SetAttributes(u32 attributes[nn::pia::inet::ATTRIBUTE_MAX_SIZE]);
/*!
    @brief  Gets the specified attribute.
    @param[in] index  Specifies the index for the attribute to get.
    @return  Returns the specified attribute value.
*/
    u32 GetAttribute(u32 index) const;

/*!
    @brief  Sets application-defined data.
    @details  Specify a size that does not exceed <tt>inet::APPLICATION_DATA_BUFFER_MAX_SIZE</tt>.
    @param[in] pData  Specifies a pointer to the buffer for the application-defined data for the session.
    @param[in] size  Specifies the size of the application-defined data for the session.
    @return  Returns a <tt>Result</tt> value indicating success if the application-defined data is set successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer was <tt>NULL</tt> or the specified data size was too big. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetApplicationData(const u8* pData, size_t size);
/*!
    @brief  Gets specified application-defined data.
    @param[in] pDataBuf  Specifies a pointer to a buffer from which to copy the specified application-defined data.
    @param[in] pDataSize  Specifies the size of the data written to the copy buffer.
    @param[in] dataBufSize  Specifies the copy buffer size.
    @return  Returns a <tt>Result</tt> value indicating success if the application-defined data is acquired successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that either the buffer pointer is <tt>NULL</tt> or the pointer returned by the copied size is <tt>NULL</tt>. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that no application-defined data is set. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultBufferShortage  Indicates that the set application-defined data cannot be copied because it is greater in size than the buffer. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetApplicationData(u8* pDataBuf, size_t* pDataSize, size_t dataBufSize) const;

/*!
    @brief  Sets whether participation in this session is possible.

    @details  Specifies whether the session can be joined.

    @param[in] isOpen  Specifies whether the session updated by this <tt>NexUpdateSessionSetting</tt> function is joinable. Set to <tt>true</tt> to enable.
*/
    void SetOpenParticipation(bool isOpen);

/*!
    @cond  PRIVATE
    @brief  Gets whether participation in this session is possible.

    @return  Returns <tt>true</tt> if the session is joinable.
*/
    bool GetOpenParticipation() const;
    //! @endcond

/*!
    @brief  Sets the user password.

    @details  Specifies the session's user password for the session.

    @param[in] pUserPassword  Specifies the user password.
    @param[in] passwordLength  Specifies the string length of the user password.
*/
    void SetSessionUserPassword(const wchar_t* pUserPassword, size_t passwordLength);

/*!
    @cond  PRIVATE
    @brief
*/
    const wchar_t* GetSessionUserPassword() const;
    //! @endcond

/*!
    @brief  Specifies the matchmaking priority when a session with the same requirements is found.

    @details  Specifies the matchmaking priority for the new session.
                Enabled only when searches are performed with
                <tt>nn::pia::inet::SELECTION_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY</tt> and <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
                If a value that exceeds <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is provided as an argument, <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt> is set instead.

    @param[in] priority  Specifies the matchmaking priority.
*/
    void SetSelectionPriority(u8 priority);

/*!
    @brief  Sets the minimum number of participants for the session.
    @param[in] min  Specifies the minimum number of participants for the new session.
*/
    void SetMinParticipants(u16 min);

/*!
    @brief  Gets the specified minimum number of participants.
    @return  Returns the minimum number of participants.
*/
    u16 GetMinParticipants() const;

/*!
    @brief  Sets the maximum number of participants for the session.
    @param[in] max  Specifies the maximum number of participants for the new session.
*/
    void SetMaxParticipants(u16 max);

/*!
    @brief  Gets the maximum number of participants.
    @return  Returns the maximum number of participants.
*/
    u16 GetMaxParticipants() const;

/*!
    @brief  Updates the rating value to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @param[in] ratingValue  Specifies the rating value.
*/
    void SetRatingValue(u32 ratingValue);
/*!
    @cond  PRIVATE
*/
    u32 GetRatingValue() const;
    bool IsSetRatingValue() const;
    //! @endcond

/*!
    @brief  Updates the disconnection rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] disconnectionRate  Specifies the disconnection rate.
*/
    void SetDisconnectionRate(u32 disconnectionRate);
/*!
    @cond  PRIVATE
*/
    u32 GetDisconnectionRate() const;
    bool IsSetDisconnectionRate() const;
    //! @endcond

/*!
    @brief  Updates the violation rate to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  We recommend using NEX's <tt>MatchmakeReferee</tt> feature.
    @param[in] violationRate  Specifies the violation rate.
*/
    void SetViolationRate(u32 violationRate);
/*!
    @cond  PRIVATE
*/
    u32 GetViolationRate() const;
    bool IsSetViolationRate() const;
    //! @endcond

/*!
    @brief  Updates the country code to use for comparison during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @details  Sets the country code to use when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available.
                The specified country code is not used when location information is available.
                When a country code is not set, and when you have configured to not use location information with <tt>UseGeoIP()</tt> or when you have configured to use that location information but it was not available, the default value is used as the score for comparing country codes.
    @param[in] countryCode  Specifies the country code.
*/
   void SetNintendoCountryCode(u32 countryCode);
/*!
    @cond  PRIVATE
*/
    u32 GetNintendoCountryCode() const;
    bool IsSetNintendoCountryCode() const;
    //! @endcond

/*!
    @brief  Sets the IP address to use for getting a distance over the Internet during <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>.
    @param[in] b  Specifies whether to update <tt>GeoIp</tt>.
*/
    void SetUpdateGeoIp(bool b);

/*!
    @cond  PRIVATE
    @brief  Gets whether to update <tt>GeoIp</tt>.

    @return  Specifies whether to update <tt>GeoIp</tt>.
*/
    bool GetUpdateGeoIp() const;
    //! @endcond

/*!
    @cond  PRIVATE
*/
    bool IsSetScoreBaseParam() const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Gets the matchmaking priority.

    @details  Gets the matchmaking priority for the session.

    @return  Returns the matchmaking priority.
*/
    u8 GetSelectionPriority() const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Updates the header information that indicates the usage state of the application-defined data region.
*/
    void UpdateApplicationDataHeader(const u8* pData, size_t dataSize);
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Gets data from the application-defined data.
*/
    void GetApplicationDataInternal(void* pVectorTmp) const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Copies the application-defined data.
*/
    void CopyApplicationDataToVector(void* pVectorTmp) const;
    //! @endcond

/*!
    @cond  PRIVATE
    @brief  Copies the application-defined data.
*/
    u32 GetModificationFlag() const;
    //! @endcond

    void Reset();

    // Prints information that is useful for debugging.
    void Trace(u64 flag) const;

    // SetApplicationBuffer
    // OpenParticipate

protected:
    u32 m_ModificationFlag;

    SessionType m_SessionType;

    u16 m_MinParticipants;
    u16 m_MaxParticipants;

    // The description string.
    wchar_t m_Description[DESCRIPTION_DATA_BUFFER_MAX_SIZE + 1];

    // Attributes.
    u32 m_Attribute[ATTRIBUTE_MAX_SIZE];

    u8 m_ApplicationDataBuffer[APPLICATION_DATA_BUFFER_MAX_SIZE_INTERNAL];
    size_t m_ApplicationDataSize;
    size_t m_ApplicationDataInternalOffset;

    bool m_OpenParticipation;

    u8 m_ProgressScore;

#if NN_PIA_NEX_NEW_FEATURE_FOR_MATCHMAKE
    // Pointer to the user password that is configured for the session.
    wchar_t m_UserPassword[nn::pia::inet::SESSION_USER_PASSWORD_LENGTH + 1];
#endif

    // Scored parameters.
    u32 m_RatingValue;
    bool m_IsSetRatingValue;

    u32 m_DisconnectionRate;
    bool m_IsSetDisconnectionRate;

    u32 m_ViolationRate;
    bool m_IsSetViolationRate;

    u32 m_CountryCode;
    bool m_IsSetCountryCode;

    bool m_IsUpdateGeoIp;

private:
    // The copy constructor is sealed.
    NexUpdateSessionSetting(const NexUpdateSessionSetting&);

    // The assignment operator is sealed.
    NexUpdateSessionSetting& operator=(const NexUpdateSessionSetting&);

/*!
    @cond  PRIVATE
    @brief  Adds a change flag to the locations that changed.
*/
    void AddModificationFlag(u32 bit);
    //! @endcond
};
}
}
}  // End of namespace nn::pia::inet.
