﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/definitions.h>

/*!
  @namespace  nn::pia::local
  @brief  Represents the <tt>PiaLocal</tt> namespace.

  @date 2014-04-04  Removed the <tt>UdsNode</tt> function. Use the <tt>LocalNetwork</tt> function instead.
  @date 2012-10-29  Categorized member functions to improve readability of the API reference.
  @date 2012-10-19  Modified comments in the header field in accordance with changes to the format of the API reference.
  @date 2012-07-19  Changed the name of the "NEX Local Communication Linking Feature" to "NetZ Local Communication Linking Feature."
  @date 2012-07-10  Deleted old <tt>Result</tt>-related codes that had been left in the <tt>local</tt> namespace for compatibility reasons.
  @date 2012-06-01  Revised the description of parameters (second parameter) passed to the <tt>local::UdsUpdateEventCallback</tt> function.
  @date 2012-05-01  Added the header files included in <tt>local.h</tt>. All headers required to use the <tt>local</tt> module are included if you include <tt>local.h</tt>. (This does not include the NEX local network link feature.)
  @date 2012-04-10  Initial version. Revision histories are listed individually for Pia 1.7.0 or later.
*/
namespace nn
{
namespace pia
{
namespace local
{
/*! @name  Finish Initialization
    @{
*/

/*!
              @brief  Initializes the <tt>local</tt> module.

              @details  Execute the <tt>nn::pia::common::Initialize</tt> function in advance.

              @return  Returns a <tt>Result</tt> value indicating success if initialization succeeds.

              @retval ResultInvalidState  Indicates that <tt>nn::pia::common</tt> has not been initialized.
              @retval ResultAlreadyInitialized  Indicates that the module is already initialized.
*/
nn::Result Initialize();

/*!
              @brief  Gets whether the <tt>local</tt> module is initialized.

              @return  Returns <tt>true</tt> if initialized, or <tt>false</tt> otherwise.
*/
bool IsInitialized(void);

/*!
              @brief  Declares the start of setup.

              @details  Call this function before initializing objects in the <tt>local</tt> module (before calling <tt>@ref LocalNetwork::CreateInstance</tt> and <tt>@ref LocalNetwork::Initialize</tt>).

              @attention  This function does not operate correctly if it is called during another module's setup processing.
                      After calling the <tt>@ref BeginSetup</tt> function, perform initialization in the <tt>local</tt> module, and promptly call the <tt>@ref EndSetup</tt> function to complete setup.

              @return  Returns a <tt>Result</tt> value indicating success if it is called at the right time.

              @retval ResultNotInitialized  Indicates that the <tt>local</tt> module has not been initialized.
              @retval ResultInvalidState  Indicates that the function was called at the wrong time.
*/
nn::Result BeginSetup(void);

/*!
              @brief  Declares the end of setup.

              @details  Call this function after initializing objects below the <tt>local</tt> module (after calling <tt>@ref LocalNetwork::CreateInstance</tt> and <tt>@ref LocalNetwork::Initialize</tt>).
                       Calling this function configures memory to be used by the <tt>local</tt> module singletons and optimizes memory management within the library.

              @attention  This function does not operate correctly if it is called during another module's setup processing.
                      After calling the <tt>@ref BeginSetup</tt> function, perform initialization in the <tt>local</tt> module, and promptly call the <tt>@ref EndSetup</tt> function to complete setup.

              @return  Returns a <tt>Result</tt> value indicating success if it is called at the right time.

              @retval ResultNotInitialized  Indicates that the <tt>local</tt> module has not been initialized.
              @retval ResultInvalidState  Indicates that the function was called at the wrong time.
*/
nn::Result EndSetup(void);

/// @cond PRIVATE
bool IsDuringSetup(void);
/// @endcond

/*!
              @brief  Finalizes the <tt>local</tt> module.

              @details  Processing of all <tt>PiaLocal</tt> modules must be finalized before executing this function.
                      Calling this function has no effect if the <tt>@ref nn::pia::local::Initialize</tt> function has not been called.
*/
void Finalize(void);
/*!
    end of Finish Initialization section.
    @}
*/
}
}
}

#include <pia/local/local_LocalNetwork.h>
#include <pia/local/local_UdsNetworkSetting.h>
#include <pia/local/local_UdsNetworkFactory.h>
#include <pia/local/local_LocalFacade.h>
#include <pia/local/local_LocalSessionInfo.h>
#include <pia/local/local_LocalSessionTypes.h>
#include <pia/local/local_LocalSessionSearchCriteria.h>
#include <pia/local/local_LocalSessionStationInfo.h>
#include <pia/local/local_UdsCreateSessionSetting.h>
#include <pia/local/local_UdsJoinSessionSetting.h>
#include <pia/local/local_UdsSessionInfo.h>
