﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     local_LocalFacade.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/local/local_definitions.h>
#include <pia/local/local_LocalNetworkSetting.h>
#include <pia/common/common_RootObject.h>

namespace nn
{
namespace pia
{
namespace transport
{

class StationConnectionInfo;
}
}
}  // end of namespace nn::pia::transport

namespace nn
{
namespace pia
{
namespace local
{


/*!
@class  nn::pia::local::LocalFacade
@brief  Represents the facade (interface wrapper) class for the local communication library.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-03-03  Initial version.

*/
class LocalFacade : public nn::pia::common::RootObject
{
public:
/*!
@name  Singleton Instance Control
     @{
*/

/*!
@brief  Creates the <tt>LocalFacade</tt> class instance (singleton pattern).
@details  You must call this function after the <tt>@ref LocalNetwork::CreateInstance</tt> function.

@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully.
@retval ResultNotInitialized  Indicates that the <tt>local</tt> module is not initialized.
@retval ResultInvalidState  Indicates that the instance was created at the wrong time.
@retval ResultAlreadyExists  Indicates that an instance has already been created.


*/
    static nn::Result CreateInstance(void);


/*!
@brief  Destroys the <tt>LocalFacade</tt> class instance (singleton pattern).
*/
    static void DestroyInstance(void);

/*!
@brief  Gets the <tt>LocalFacade</tt> class instance (singleton pattern).
@details  Returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>@ref CreateInstance</tt> function.
@return  Returns a pointer to the <tt>LocalFacade</tt> instance.



*/
    static LocalFacade* GetInstance(void)
    {
        return s_pInstance;
    }

/*!
end of name Singleton Instance Control
     @}
*/
/*!
@name  Finish Initialization
     @{
*/

/*!
@brief  Starts the <tt>LocalFacade</tt>.
@details  This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.
This function has the following requirements.
<ul>
<li> A <tt>@ref  nn::pia::transport::Transport</tt> instance must be available. </li>
<li> You must be connected to a local network. </li>
</ul>
The function fails if these conditions are not met.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if startup succeeds. The following results are returned on failure.
@retval ResultInvalidState  Indicates that availability conditions have not been met.


*/
    virtual nn::Result Startup();


/*!
@brief  Cleans up the <tt>LocalFacade</tt>.
@details  This function can only be called when directly manipulated by the <tt>session::Mesh</tt> class.

*/
    virtual void Cleanup();

/*!
end of name Finish Initialization
     @}
*/

/*!
@brief  Gets the <tt>StationConnectionInfo</tt> of the session host.
@param[out] pStationConnectionInfo  Stores the session host's <tt>StationConnectionInfo</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function return <tt>true</tt> if conversion succeeds. The following results are returned on failure.
@retval ResultInvalidState  The state of the session host is invalid.
@retval ResultInvalidArgument  Indicates that an invalid pointer was specified as the storage destination.


*/
    nn::Result GetHostStationConnectionInfo(nn::pia::transport::StationConnectionInfo* pStationConnectionInfo) const;

    /// @cond PRIVATE
/*!
@brief  Destructor.
*/
    virtual ~LocalFacade(void);
    /// @endcond


    /// @cond PRIVATE
/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;
    /// @endcond


/*!
@cond PRIVATE
*/
    static void LocalFacadeUpdateEventCallback(LocalUpdateEvent updateEvent, u8 transportId, void* pParam);
//! @endcond

#if 0
/*!
@name  Registering and Unregistering Event Callbacks
     @{
*/

/*!
@brief  Registers the callback function called when the UDS network state changes.
@details  This function wraps the <tt>@ref  UdsNode::RegisterUpdateEventCallback</tt> function.
If you are using <tt>PiaSession</tt> and register or unregister <tt>UdsUpdateEventCallback</tt> objects, use <tt>@ref  RegisterUpdateEventCallback</tt> and <tt>@ref  UnregisterUpdateEventCallback</tt>, regardless of whether the <tt>PiaSession</tt> module has been initialized.
This function is a static function, because it can register and unregister callbacks even if there is no instance of the <tt>LocalFacade</tt> class.
@param[in] callback  Specifies a pointer to the callback function being registered.
@param[in] pParam  Specifies the pointer passed to the callback function when it is called.
@see  UnregisterUpdateEventCallback, UdsNode::RegisterUpdateEventCallback, UdsNode::UnregisterUpdateEventCallback





*/
    static void RegisterUpdateEventCallback(UdsUpdateEventCallback callback, void* pParam);

/*!
@brief  Unregisters the registered callback function for network state changes.
@details  This function wraps the <tt>@ref  UdsNode::UnregisterUpdateEventCallback</tt> function.
If you are using <tt>PiaSession</tt> and register or unregister <tt>UdsUpdateEventCallback</tt> objects, use <tt>@ref  RegisterUpdateEventCallback</tt> and <tt>@ref  UnregisterUpdateEventCallback</tt>, regardless of whether the <tt>PiaSession</tt> module has been initialized.
This function is a static function, because it can register and unregister callbacks even if there is no instance of the <tt>UdsFacade</tt> class.
@see  RegisterUpdateEventCallback, UdsNode::RegisterUpdateEventCallback, UdsNode::UnregisterUpdateEventCallback




*/
    static void UnregisterUpdateEventCallback();

/*!
end of name Register and Unregister Event Callbacks
     @}
*/
    static void PrepareNodeIdCallback(u16 localNodeId, u16 hostNodeId, void* pParam);

    NODE_ID_TYPE GetHostTransportNodeId();
    NODE_ID_TYPE GetLocalTransportNodeId();

#endif

protected:
private:
/*!
@name  Get StationConnectionInfo
     @{
*/

/*!
@brief  Converts a <tt>TransportId</tt> to a <tt>StationConnectionInfo</tt>.
@param[in] transportId  Specifies the <tt>TransportId</tt> you want to convert.
@param[out] pStationConnectionInfo  Specifies a pointer to the converted <tt>StationConnectionInfo</tt> instance.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function return <tt>true</tt> if conversion succeeds. The following results are returned on failure.
@retval ResultInvalidArgument  Indicates that an invalid <tt>TransportId</tt> was specified, or an invalid out pointer was specified.


*/
    nn::Result ConvertTransportIdToStationConnectionInfo(
        u8 transportNodeId,
        nn::pia::transport::StationConnectionInfo* pStationConnectionInfo) const;

/*!
end of name Get <tt>StationConnectionInfo</tt>
     @}
*/

private:
    /// @cond PRIVATE
/*!
@brief  This default constructor is private because this class implements the singleton pattern.
*/
    LocalFacade(void);
    /// @endcond

    /// @cond PRIVATE
/*!
@brief  This copy constructor is private because the singleton pattern is used.
*/
    LocalFacade(const LocalFacade&);
    /// @endcond

    /// @cond PRIVATE
/*!
@brief  This assignment operator is private because the singleton pattern is used.
*/
    LocalFacade& operator=(const LocalFacade&);
    /// @endcond


    // A pointer to a singleton instance.
    static LocalFacade* s_pInstance;

#if 0
    common::CriticalSection m_Lock;
    NODE_ID_TYPE m_HostTransportNodeId;
    NODE_ID_TYPE m_LocalTransportNodeId;
#endif
    bool m_IsStarted;

#if 0
    static UdsUpdateEventCallback s_pfAppUpdateEventCallback;
    static void* s_pAppUpdateEventCallbackParam;
#endif
};
}
}
}  // end of namespace nn::pia::local
