﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     local_LocalNetwork.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/local/local_LocalNetworkSetting.h>

#include <pia/common/common_RootObject.h>
#include <pia/common/common_CallContext.h>
#include <pia/common/common_CriticalSection.h>

namespace nn
{
namespace pia
{
namespace common
{

class StationAddress;
}
}
}  // end of namespace nn::pia::local

namespace nn
{
namespace pia
{
namespace local
{

// Forward declarations.
class LocalNetworkManager;
class LocalMigrationManager;
class LocalAroundNetworkSearchManager;
class LocalCreateNetworkJob;
class LocalDestroyNetworkJob;
class LocalScanNetworkJob;
class LocalConnectNetworkJob;
class LocalDisconnectNetworkJob;
class LocalForceDisconnectNetworkJob;
class LocalBackgroundProcessJob;
class LocalNetworkSetting;
class UdsNetworkSetting;
class LocalNetworkDescription;
class UdsNetworkDescription;
class LocalCreateNetworkSetting;
class LocalConnectNetworkSetting;
class LocalScanNetworkSetting;

/*!
@brief  Class for handling the local network.

@date 2014-10-23  The <tt>Initialize</tt> function now stops on an assert if device memory is passed to Pia.
@date 2014-03-03  Initial version.

*/
class LocalNetwork : public nn::pia::common::RootObject
{
public:
/*!
@brief  Creates the <tt>LocalNetwork</tt> class instance (singleton pattern).

@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully.

@retval ResultNotInitialized  Indicates that the <tt>local</tt> module is not initialized.
@retval ResultInvalidState  Indicates that the instance was created at the wrong time.
@retval ResultAlreadyExists  Indicates that an instance has already been created.
*/
    static nn::Result CreateInstance();


/*!
@brief  Destroys the <tt>LocalNetwork</tt> class instance (singleton pattern).
*/
    static void DestroyInstance();

/*!
@brief  Gets the <tt>LocalNetwork</tt> class instance (singleton pattern).

@details  Returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>@ref Initialize</tt> function.

@return  Returns a pointer to the <tt>LocalNetwork</tt> instance.

*/
    static LocalNetwork* GetInstance()
    {
        return s_pInstance;
    }

private:
    LocalNetwork();
    virtual ~LocalNetwork();
/*!
@brief  This copy constructor is private.
*/
    LocalNetwork(const LocalNetwork&);

/*!
@brief  This assignment operator is private.
*/
    LocalNetwork& operator=(const LocalNetwork&);

public:
/*!
@brief  Initializes the instance. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.

@attention  Device memory cannot be used as the <tt>LocalNetwork</tt> receive buffer, so <tt>common::Initialize</tt> stops on an assert if device memory is passed.

@param[in] setting  Specifies various settings.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

@see  Finalize

*/
    nn::Result Initialize(const LocalNetworkSetting& setting);


/*!
@brief  Finalizes the instance.

@see  Initialize
*/
    void Finalize();


/*!
@brief  Startup process.

@attention  Do not transition to sleep mode during a call to this function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.

@see  Cleanup
*/
    nn::Result Startup();


/*!
@brief  Cleanup process.

@see  Startup
*/
    void Cleanup();


/*!
@brief  Enumerates the reason for the disconnection.
*/
    enum DisconnectReason
    {
        DISCONNECT_REASON_BEFORE_COMMUNICATION = 0,  //!< Communication has not yet started.
        DISCONNECT_REASON_NETWORK_IS_AVAILABLE,      //!< Connection being maintained.
        DISCONNECT_REASON_REQUEST_FROM_MYSELF,       //!< Disconnected from the network due to local action.
        DISCONNECT_REASON_REQUEST_FROM_SYSTEM,       //!< Disconnected from the network due to a system request (transition to wireless off mode or sleep mode).
        DISCONNECT_REASON_DISCARDED_FROM_NETWORK,    //!< Kicked off the network due to an instruction from the host. Alternatively, the client was disconnected due to degraded connectivity and the host's internal state.
        DISCONNECT_REASON_CONNECTION_LOST,           //!< The connection could not be maintained because a beacon could not be received from the host for a certain period due to degraded connectivity.
        DISCONNECT_REASON_UNKNOWN                    //!< Disconnected for an unknown reason.
    };


/*!
@brief  Represents the state that allows or denies new participants.
*/
    enum AllowParticipatingState
    {
        ALLOW_PARTICIPATING_STATE_NONE = 0,                //!< Not connected to the network.
        ALLOW_PARTICIPATING_STATE_ALLOW,                   //!< New participants are allowed.
        ALLOW_PARTICIPATING_STATE_DISALLOW,                //!< New participants are denied.
        ALLOW_PARTICIPATING_STATE_ALLOW_ONLY_RECONNECTION  //!< Only reconnections are allowed.
    };


/*!
@cond PRIVATE
*/
    nn::Result CreateNetwork(nn::pia::common::CallContext* pCallContext, const LocalCreateNetworkSetting* pCreateNetworkSetting);
    //! @endcond

/*!
@brief  Creates the local network.

@details  Receives the network settings and starts the asynchronous process that creates the local network.
The device that creates the network becomes the local network host.
The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
Use <tt>@ref IsCompletedCreateNetwork</tt> to check whether the process has completed and <tt>@ref GetCreateNetworkResult</tt> to check the result.

@param[in] pCreateNetworkSetting  Settings for the network to create.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the process is successfully started.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that processes are already running. Alternatively, another asynchronous process is running. Or, the station is already connected. Programming error. Fix your program so that this error is not returned.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  IsCompletedCreateNetwork, GetCreateNetworkResult
*/
    nn::Result CreateNetworkAsync(const LocalCreateNetworkSetting* pCreateNetworkSetting);

/*!
@brief  Checks the completion of <tt>CreateNetworkAsync</tt> asynchronous processing.

@details  This function checks the completion of asynchronous processes started with <tt>@ref CreateNetworkAsync</tt>.

@return  Returns <tt>true</tt> if the asynchronous process has completed; <tt>false</tt> otherwise. Returns <tt>false</tt> if the asynchronous process has not started or if another asynchronous process has been started.

@see  CreateNetworkAsync, GetCreateNetworkResult
*/
    bool IsCompletedCreateNetwork();

/*!
@brief  Gets the result of <tt>CreateNetworkAsync</tt> asynchronous processing.

@details  This function gets the results after an asynchronous process started by <tt>@ref CreateNetworkAsync</tt> ends.

@return  Returns the result of <tt>CreateNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultNotInitialized  Indicates that the communication library is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Alternatively, results cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultTimeOut  Timed out. Perform cleanup.
@retval ResultCancelled  Canceled. Perform cleanup.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  CreateNetworkAsync, IsCompletedCreateNetwork
*/
    nn::Result GetCreateNetworkResult();


/*!
@cond PRIVATE
*/
    nn::Result DestroyNetwork(nn::pia::common::CallContext* pCallContext);
    //! @endcond

/*!
@brief  Destroys the local network.

@details  Starts an asynchronous process that destroys the local network.
The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
Use <tt>@ref IsCompletedDestroyNetwork</tt> to check whether the process has completed and <tt>@ref GetDestroyNetworkResult</tt> to check the result.
After the process is completed, clean up <tt>LocalNetwork</tt>.

@attention  This function can only be called by the local network host.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the process is successfully started.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that processes are already running. Alternatively, another asynchronous process is running. Alternatively, the device is not the local network host. Programming error. Fix your program so that this error is not returned.

@see  IsCompletedDestroyNetwork, GetDestroyNetworkResult
*/
    nn::Result DestroyNetworkAsync();

/*!
@brief  Checks the completion of <tt>DestroyNetworkAsync</tt> asynchronous processing.

@details  This function checks the completion of asynchronous processes started with <tt>@ref DestroyNetworkAsync</tt>.

@return  Returns <tt>true</tt> if the asynchronous process has completed; <tt>false</tt> otherwise. Returns <tt>false</tt> if the asynchronous process has not started or if another asynchronous process has been started.

@see  DestroyNetworkAsync, GetDestroyNetworkResult
*/
    bool IsCompletedDestroyNetwork();

/*!
@brief  Gets the result of <tt>DestroyNetworkAsync</tt> asynchronous processing.

@details  This function gets the results after an asynchronous process started by <tt>@ref DestroyNetworkAsync</tt> ends.

@return  Returns the result of <tt>DestroyNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultNotInitialized  Indicates that the communication library is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Alternatively, results cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultCancelled  Canceled. Perform cleanup.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  DestroyNetworkAsync, IsCompletedDestroyNetwork
*/
    nn::Result GetDestroyNetworkResult();


/*!
@cond PRIVATE
*/
    nn::Result ConnectNetwork(nn::pia::common::CallContext* pCallContext, const LocalConnectNetworkSetting* pConnectNetworkSetting);
    //! @endcond

/*!
@brief  Connects to the local network.

@details  Starts an asynchronous process that connects to the local network. The device that connects becomes a local host client.
The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
Use <tt>@ref IsCompletedConnectNetwork</tt> to check whether the process has completed and <tt>@ref GetConnectNetworkResult</tt> to check the result.
You can also cancel the process by calling the <tt>@ref CancelConnectNetworkAsync</tt> function during the process.
Even if you cancel the process, wait until you confirm the process completion using <tt>@ref IsCompletedConnectNetwork</tt>.
If you get results that indicate that the process failed (or was canceled), clean up <tt>LocalNetwork</tt> and start it again.

@param[in] pConnectNetworkSetting  Settings for the network to connect to.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the process is successfully started.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that processes are already running. Alternatively, another asynchronous process is running. Or, the station is already connected. Programming error. Fix your program so that this error is not returned.
@retval ResultDifferentVersion  Indicates a different application version or the local network version of the target network. Perform cleanup.

@see  IsCompletedConnectNetwork, GetConnectNetworkResult, CancelConnectNetworkAsync
*/
    nn::Result ConnectNetworkAsync(const LocalConnectNetworkSetting* pConnectNetworkSetting);

/*!
@brief  Checks the completion of <tt>ConnectNetworkAsync</tt> asynchronous processing.

@details  This function checks the completion of asynchronous processes started with <tt>@ref ConnectNetworkAsync</tt>.

@return  Returns <tt>true</tt> if the asynchronous process has completed; <tt>false</tt> otherwise. Returns <tt>false</tt> if the asynchronous process has not started or if another asynchronous process has been started.

@see  ConnectNetworkAsync, GetConnectNetworkResult, CancelConnectNetworkAsync
*/
    bool IsCompletedConnectNetwork();

/*!
@brief  Gets the result of <tt>ConnectNetworkAsync</tt> asynchronous processing.

@details  This function gets the results after an asynchronous process started by <tt>@ref ConnectNetworkAsync</tt> ends.

@return  Returns the result of <tt>ConnectNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultNotInitialized  Indicates that the communication library is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Alternatively, results cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidNetwork  Indicates that the connection target network was not found. Perform cleanup.
@retval ResultNetworkIsFull  The maximum number of connections has already been exceeded. Perform cleanup.
@retval ResultJoinRequestDenied  The host denied the attempt to join. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultTimeOut  Timed out.
It could also be returned when <tt>@ref nn::pia::local::UdsConnectNetworkSetting::m_Passphrase</tt> is incorrect. Perform cleanup.
@retval ResultCancelled  Canceled. Perform cleanup.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  ConnectNetworkAsync, IsCompletedConnectNetwork, CancelConnectNetworkAsync
*/
    nn::Result GetConnectNetworkResult();

/*!
@brief  Cancels <tt>ConnectNetworkAsync</tt> asynchronous processes.

@details  This function cancels asynchronous processes started by <tt>ConnectNetworkAsync</tt>.
The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
The library promptly cancels asynchronous processes after this function is called, but use the <tt>@ref IsCompletedConnectNetwork</tt> function to check whether the cancellation has completed.

@return  Returns the result of canceling <tt>ConnectNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultInvalidState  Returned when no <tt>ConnectNetworkAsync</tt> asynchronous processing is being performed. Programming error. Fix your program so that this error is not returned.

@see  ConnectNetworkAsync, IsCompletedConnectNetwork, GetConnectNetworkResult

*/
    nn::Result CancelConnectNetworkAsync();


/*!
@cond PRIVATE
*/
    nn::Result DisconnectNetwork(nn::pia::common::CallContext* pCallContext);
    //! @endcond

/*!
@brief  Disconnects from the local network.

@details  Starts an asynchronous process that disconnects from the local network.
When the local device is the local network host, this function succeeds if host migration is enabled, and the manual host migration process starts.
The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
Use <tt>@ref IsCompletedDisconnectNetwork</tt> to check whether the process has completed and <tt>@ref GetDisconnectNetworkResult</tt> to check the result.
After the process is completed, clean up <tt>LocalNetwork</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the process is successfully started.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that processes are already running. Alternatively, another asynchronous process is running.
Alternatively, the local device is the local network host, but host migration is disabled. Programming error. Fix your program so that this error is not returned.

@see  IsCompletedDisconnectNetwork, GetDisconnectNetworkResult
*/
    nn::Result DisconnectNetworkAsync();

/*!
@brief  Checks the completion of <tt>DisconnectNetworkAsync</tt> asynchronous processing.

@details  This function checks the completion of asynchronous processes started with <tt>@ref DisconnectNetworkAsync</tt>.

@return  Returns <tt>true</tt> if the asynchronous process has completed; <tt>false</tt> otherwise. Returns <tt>false</tt> if the asynchronous process has not started or if another asynchronous process has been started.

@see  DisconnectNetworkAsync, GetDisconnectNetworkResult
*/
    bool IsCompletedDisconnectNetwork();

/*!
@brief  Gets the result of <tt>DisconnectNetworkAsync</tt> asynchronous processing.

@details  This function gets the results after an asynchronous process started by <tt>@ref DisconnectNetworkAsync</tt> ends.

@return  Returns the result of <tt>DisconnectNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultNotInitialized  Indicates that the communication library is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Alternatively, results cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultCancelled  Canceled. Perform cleanup.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  DisconnectNetworkAsync, IsCompletedDisconnectNetwork
*/
    nn::Result GetDisconnectNetworkResult();


/*!
@cond PRIVATE
*/
    nn::Result ScanNetwork(nn::pia::common::CallContext* pCallContext, u32 localCommunicationId, u8 subId);
    //! @endcond

/*!
@brief  Scans for nearby local networks.

@details  Starts an asynchronous process that scans for nearby local networks. Fails if called while connecting to a network.
The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
Use <tt>@ref IsCompletedScanNetwork</tt> to check whether the process has completed and <tt>@ref GetScanNetworkResult</tt> to check the result.
You can also cancel the process by calling the <tt>@ref CancelScanNetworkAsync</tt> function during the process.
Even if you cancel the process, wait until you can confirm the process completion using <tt>@ref IsCompletedScanNetwork</tt>.
Use <tt>@ref GetNetworkDescriptionList</tt> and <tt>@ref GetNetworkDescription</tt> to get information on the local networks that were found through scanning.
Use <tt>@ref GetStationInfoList</tt> to get information on the connected stations that were found through scanning.
Get the strength of the signal sent from the network by calling <tt>@ref GetRadioStrength</tt>.
These items of information become available after the network scanning ends successfully. If scanning starts again, they can no longer be obtained.
The information is also no longer available when you begin the process of creating or connecting to a network.

@param[in] localCommunicationId  Specifies the local communication ID. Specify the value generated by <tt>@ref CreateLocalCommunicationId</tt>.
@param[in] subId  Specifies the ID for identifying the communication mode. The application can set any arbitrary value for this parameter.
To search all IDs, specify <tt>0xff</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the process is successfully started.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that processes are already running. Alternatively, another asynchronous process is running. Alternatively, the system is already connected to the network. Programming error. Fix your program so that this error is not returned.

@see  IsCompletedScanNetwork, GetScanNetworkResult, CancelScanNetworkAsync


*/
    nn::Result ScanNetworkAsync(u32 localCommunicationId, u8 subId);

/*!
@brief  Checks the completion of <tt>ScanNetworkAsync</tt> asynchronous processing.

@details  This function checks the completion of asynchronous processes started with <tt>@ref ScanNetworkAsync</tt>.

@return  Returns <tt>true</tt> if the asynchronous process has completed; <tt>false</tt> otherwise. Returns <tt>false</tt> if the asynchronous process has not started or if another asynchronous process has been started.

@see  ScanNetworkAsync, GetScanNetworkResult, CancelScanNetworkAsync
*/
    bool IsCompletedScanNetwork();

/*!
@brief  Gets the result of <tt>ScanNetworkAsync</tt> asynchronous processing.

@details  This function gets the results after an asynchronous process started by <tt>@ref ScanNetworkAsync</tt> ends.

@return  Returns the result of <tt>ScanNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultNotInitialized  Indicates that the communication library is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Alternatively, results cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultNegligibleFault  Canceled.
@retval ResultErrorOccurred  Indicates that an internal error occurred. Perform cleanup.

@see  ScanNetworkAsync, IsCompletedScanNetwork, CancelScanNetworkAsync
*/
    nn::Result GetScanNetworkResult();

/*!
@brief  Cancels <tt>ScanNetworkAsync</tt> asynchronous processes.

@details  This function cancels asynchronous processes started by <tt>ScanNetworkAsync</tt>.
The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
The library promptly cancels asynchronous functions after this function is called, but use the <tt>@ref IsCompletedScanNetwork</tt> function to check whether the cancellation has completed.

@return  Returns the result of canceling <tt>ScanNetworkAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultInvalidState  Returned when no <tt>ConnectNetworkAsync</tt> asynchronous processing is being performed. Programming error. Fix your program so that this error is not returned.

@see  ScanNetworkAsync, IsCompletedScanNetwork, GetScanNetworkResult

*/
    nn::Result CancelScanNetworkAsync();


/*!
@brief  Creates a local communication ID from a unique ID.

@details  This function generates a 32-bit value (the <I>local communication ID</I>) to use for local communication from the 20-bit unique ID (the title-specific ID) assigned by the Nintendo Licensing department.
If you have not been assigned a unique ID by the Nintendo Licensing department, use one of the unique IDs for game software prototypes (<tt>0xFF000</tt> through <tt>0xFF3FF</tt>).

@attention  However, for your retail product, be sure to get a unique ID from the Nintendo Licensing department.

@param[in] uniqueId  Specifies a unique ID. To communicate more than one title, specify the unique ID of any of the titles.
@param[in] isTrial  Specifies the flag for handling cases where the same unique ID is shared by both the final product and trial version. If you do not want the final product and trial version to communicate with each other, specify a value of <tt>true</tt> for trial versions. <b>Note:</b> Always set this flag to <tt>false</tt> in retail products.

@return  Returns the created local communication ID.
*/
    u32 CreateLocalCommunicationId(u32 uniqueId, bool isTrial = false) const;


/*!
@brief  Gets the beacon data set by the application.

@param[out] pBuffer  Buffer that stores the data obtained.
@param[out] pDataSize  Specifies the size of the data obtained.
@param[in] bufferSize  The size of the buffer that stores the data obtained.
@param[in] pNetworkDescription  The information for the network whose data you want to get.

@return  Returns a <tt>Result</tt> value indicating success if data was successfully obtained from the network.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  There was no data to get.
*/
    nn::Result GetApplicationData(void* pBuffer, size_t* pDataSize, size_t bufferSize, const LocalNetworkDescription* pNetworkDescription) const;

/*!
@brief  Gets the size of the beacon data set by the application.

@param[out] pDataSize  Specifies the size of the beacon data set by the application.
@param[in] pNetworkDescription  The information for the network whose beacon data size you want to get.

@return  Returns a <tt>Result</tt> value indicating success if data was successfully obtained from the network.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  There was no data to get.
*/
    nn::Result GetApplicationDataSize(size_t* pDataSize, const LocalNetworkDescription* pNetworkDescription) const;

/*!
@brief  Gets the data set in the beacon.

@attention  This function is intended to be called by a device connected to a local network.

@attention  This function may not return correct results if called during local host migration.
Do not call it during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@param[out] pBuffer  Specifies the buffer to write data to.
@param[out] pDataSize  Specifies the length of the retrieved data.
@param[in] bufferSize  Specifies the size of the buffer to write data to.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultNotInCommunication  In a state where communication is not possible. Either a disconnect will shortly be detected, or host migration is about to begin. The function may succeed if called after host migration completes.
*/
    nn::Result GetApplicationDataFromBeacon(void* pBuffer, size_t* pDataSize, size_t bufferSize) const;

/*!
@brief  Gets the size of the data set in the beacon.

@attention  This function is intended to be called by a device connected to a local network.

@attention  This function may not return correct results if called during local host migration.
Do not call it during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@param[out] pDataSize  The size of the data set in the beacon.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
@retval ResultNotInCommunication  In a state where communication is not possible. Either a disconnect will shortly be detected, or host migration is about to begin. The function may succeed if called after host migration completes.
*/
    nn::Result GetApplicationDataSizeFromBeacon(size_t* pDataSize) const;

/*!
@brief  Sets data in the beacon.

@details  Only the local network host can execute this function.

@attention  This function may not return correct results if called during local host migration.
Do not call it during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@param[in] pData  Specifies a pointer to the data to set.
@param[in] dataSize  Specifies the size of the data to set. The size must be no greater than the value that can be obtained with <tt>@ref GetBeaconApplicationDataSizeMax</tt>.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Or, it was disconnected from the local network. Perform cleanup.
*/
    nn::Result SetApplicationDataToBeacon(const void* pData, size_t dataSize);

/*!
@brief  Gets the local network information.

@details  Gets a list of the local network information discovered by the <tt>@ref ScanNetworkAsync</tt> function.

@param[out] paLocalNetworkDescriptionList  Specifies the buffer that stores the obtained network information.
@param[out] pValidNetworkNum  The number of networks for which network information was obtained.
@param[in] networkDescriptionListSize  The size of the buffer that stores the obtained network information.

@return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetNetworkDescriptionList(LocalNetworkDescription* paLocalNetworkDescriptionList, u32* pValidNetworkNum, size_t networkDescriptionListSize);

/*!
@brief  Gets the local network information.

@details  Gets the local network information discovered by the <tt>@ref ScanNetworkAsync</tt> function.

@param[out] pLocalNetworkDescription  Specifies the buffer that stores the network information obtained.
@param[in] networkDescriptionIndex  The index of the network information you want to get.

@return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.

@retval ResultNotInitialized  Indicates that the <tt>LocalNetwork</tt> is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetNetworkDescription(LocalNetworkDescription* pLocalNetworkDescription, u32 networkDescriptionIndex);


/*!
@brief  Gets information about the specified station.

@param[out] pLocalStationInfo  Specifies the buffer that stores the station information obtained.
@param[in] address  The address of the station whose information you want to get.

@return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.

@retval ResultNotInitialized  Indicates <tt>LocalNetwork</tt> has not been started. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotFound  The target station was not found. You might have been disconnected.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
*/
    nn::Result GetStationInfo(LocalStationInfo* pLocalStationInfo, const common::StationAddress& address) const;

/*!
@brief  Gets a list of information about the stations participating in the specified network.

@param[out] paLocalStationInfoList  Specifies the buffer that stores the list of station information obtained. The array length must be equal to or greater than the maximum number of stations that can connect to the specified network.
@param[in] listLength  Specifies the length of the <var>paLocalStationInfoList</var> array. This must be equal to or greater than the maximum number of stations that can connect to the specified network.
@param[in] networkListIndex  The index of the network whose information you want to get.

@return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.

@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that no networks are found. Alternatively, data cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates that information could not be obtained.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Perform cleanup.
*/
    nn::Result GetStationInfoList(LocalStationInfo* paLocalStationInfoList, u8 listLength, u32 networkListIndex) const;


/*!
@brief  Gets the largest size of data the application can set in the beacon.

@return  Returns the largest size of data the application can set in the beacon.
*/
    size_t GetBeaconApplicationDataSizeMax() const;

/*!
@brief  Gets the number of networks discovered.

@return  Returns the number of networks discovered.
*/
    u32 GetFoundNetworkNum() const
    {
        return m_FoundNetworkNum;
    }


/*!
@brief  Gets whether the local device is the local network host.

@return  Returns whether the local device is the local network host.
*/
    bool IsHost() const;


/*!
@brief  Gets whether the local device is a local network client.

@return  Returns whether the local device is a local network client.
*/
    bool IsClient() const;


/*!
@brief  Gets whether there is a connection to the network.

@attention  Note that <tt>false</tt> could be returned during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@return  Returns whether there is a connection to the network.

*/
    bool IsConnected() const
    {
        return IsHost() || IsClient();
    }


/*!
@brief  Gets whether local host migration is in progress.

@return  Returns whether local host migration is in progress.
*/
    bool IsDuringHostMigration() const;


/*!
@brief  Gets whether local host migration is enabled.

@return  Returns whether local host migration is enabled.
*/
    bool IsEnableHostMigration() const;


/*!
@brief  Gets the link level.

@details  There are four link levels, from <tt>0</tt> through <tt>3</tt>. <tt>0</tt>: Communication quality is extremely poor, or communication has not been established. <tt>1</tt>: Communication quality is poor. <tt>2</tt>: Communication quality is not very good. <tt>3</tt>: Communication quality is good.

@attention  This function may not return correct results if called during local host migration.
Do not call it during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@return  Returns the communication quality of the network currently connected to. If not connected, <tt>0</tt> is returned.

*/
    u8 GetLinkLevel() const;


/*!
@brief  Parses a network and gets the signal strength sent from that network.

@details  There are four levels of signal strength, from <tt>0</tt> through <tt>3</tt>. <tt>0</tt>: Communication quality is extremely poor, or communication has not been established. <tt>1</tt>: Communication quality is poor. <tt>2</tt>: Communication quality is not very good. <tt>3</tt>: Communication quality is good.

@attention  The link level retrieved corresponds to the local network information stored when <tt>@ref ScanNetworkAsync</tt> is executed.
To get the saved local network information list, use the <tt>@ref GetNetworkDescriptionList</tt> function.

@param[out] pRadioStrength  Specifies a pointer to the variable that records the signal strength. The data is stored only if the network can be analyzed correctly.
@param[in] descriptionListIndex  Specifies the variable for specifying the network to parse. Specify the network number you want to analyze by index, starting from <tt>0</tt> for the first entry in the local network information list.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultInvalidState  No information is stored in the local network information list. Alternatively, data cannot be obtained in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result GetRadioStrength(u8* pRadioStrength, u32 descriptionListIndex) const;


/*!
@brief  Gets the IEEE 802.11b/g wireless channel in use.

@attention  This function may not return correct results if called during local host migration.
Do not call it during host migration (while the <tt>@ref LocalNetwork::IsDuringHostMigration</tt> function returns <tt>true</tt>).

@return  Returns the current channel. If not connected or if the attempt to get the channel fails, <tt>0</tt> is returned.
*/
    u8 GetNetworkChannel() const;

/*!
@brief  Disallows new clients from joining the network. This function does not affect clients that are currently connected.

@details  This feature is intended to deny connections by new clients attempting to connect during interactive gaming.
When you want to allow connections again, call the <tt>@ref AllowParticipating</tt> function.

@note  In some cases, the information set by this function is not carried over to the new network after host migration occurs due to abnormal disconnection of the host.

@attention  This function can only be called by the local network host.

@param[in] isDisallowToReconnect  If set to <tt>false</tt>, only the clients that were disconnected after calling this function can reconnect.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Or, it was disconnected from the local network. Perform cleanup.
*/
    nn::Result DisallowParticipating(bool isDisallowToReconnect = true);

/*!
@brief  Allows new clients to join the network.

@details  This function re-enables connections from new clients that were denied by the <tt>@ref DisallowParticipating</tt> function.

@note  In some cases, the information set by this function is not carried over to the new network after host migration occurs due to abnormal disconnection of the host.

@attention  This function can only be called by the local network host.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Or, it was disconnected from the local network. Perform cleanup.
*/
    nn::Result AllowParticipating();

/*!
@cond PRIVATE
@brief  Kicks the specified client off the network.

@attention  This function can only be called by the local network host.

@param[in] address  The <tt>@ref common::StationAddress</tt> for the client to kick off.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Or, it was disconnected from the local network. Perform cleanup.
*/
    nn::Result EjectClient(const common::StationAddress& address);
    //! @endcond

/*!
@cond PRIVATE
@brief  Kicks all clients off the network.

@attention  This function can only be called by the local network host.

@return  Returns a <tt>Result</tt> value indicating success on successful completion.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidConnection  Indicates that the device entered wireless off mode. Or, it was disconnected from the local network. Perform cleanup.
*/
    nn::Result EjectClientAll();
    //! @endcond

/*!
@brief  Gets the reason for being disconnected from the network.

@details  Resets to <tt>@ref DISCONNECT_REASON_BEFORE_COMMUNICATION</tt> when <tt>@ref Startup</tt> or <tt>@ref Cleanup</tt> is called.

@return  Returns the reason for being disconnected from the network.
*/
    DisconnectReason GetDisconnectReason()
    {
        return m_DisconnectReason;
    }

/*!
@cond PRIVATE
*/
    void SetDisconnectReason(DisconnectReason reason)
    {
        m_DisconnectReason = reason;
    }
    //! @endcond

/*!
@brief  Gets the state that indicates whether new participants are allowed or denied.

@return  Returns whether new participants are allowed or denied.
*/
    AllowParticipatingState GetAllowParticipatingState() const;

/*!
@cond PRIVATE
*/
    void SetAllowParticipatingState(AllowParticipatingState reason)
    {
        m_AllowParticipatingState = reason;
    }
    //! @endcond

/*!
@brief  Gets whether nearby network search is enabled.

@return  Returns whether nearby network search is enabled.
*/
    bool IsEnableAroundNetworkSearch() const;

/*!
@brief  Begins nearby network search.

@details  Only the local network host can execute this function.
Only succeeds when nearby network search is enabled.
After this API is called, start notification is also sent to the clients.

@return  Returns a <tt>Result</tt> value indicating a successful start to nearby network search.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result StartAroundNetworkSearch();

/*!
@brief  Terminates nearby network search.

@details  Only the local network host can execute this function.
Only succeeds when nearby network search is enabled.
After termination, any discovered networks are no longer available.
After this API is called, termination notification is also sent to the clients.

@return  Returns a <tt>Result</tt> value indicating a successful termination to nearby network search.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result StopAroundNetworkSearch();

/*!
@brief  Gets whether nearby network search has begun.

@return  Returns whether nearby network search has begun.
*/
    bool IsAroundNetworkSearchActivated() const;

/*!
@brief  Sets the parameters for nearby network search.

@details  Settings are not shared with other stations. To do so, configure each and every station.
Only succeeds when nearby network search is enabled.

@param[in] setting  The parameters for nearby network search.

@return  Returns a <tt>Result</tt> value indicating a successful setting was made.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetAroundNetworkSearchSetting(const LocalAroundNetworkSearchSetting& setting);

/*!
@brief  Gets the information for networks discovered during nearby network search.

@details  Also succeeds if zero networks were discovered.
Only succeeds when nearby network search is enabled.

@param[out] paLocalAroundNetworkInfoList  The buffer storing the obtained network information.
@param[out] pValidNetworkNum  The number of networks for which network information was obtained.
@param[in] networkInfoListSize  The size of the buffer storing the obtained network information.

@return  Returns a <tt>Result</tt> value indicating that information was obtained successfully.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetAroundNetworkInfoList(LocalAroundNetworkInfo* paLocalAroundNetworkInfoList, u32* pValidNetworkNum, size_t networkInfoListSize) const;

/*!
@cond PRIVATE
*/
    void SleepBeforeSend();
    //! @endcond

/*!
@cond PRIVATE
@brief  Sets the sleep interval to use before sending data.

@param[in] sleepTime  The sleep time to use before sending data. The value is in milliseconds.
*/
    void SetSleepBeforeSendTime(s64 sleepTime);
    //! @endcond

/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


/*!
@cond PRIVATE
*/
    u8 GetTransportId() const;
    LocalNetworkDescription* GetNetworkDescription(u32 index);
    LocalNetworkManager* GetLocalNetworkManager()
    {
        return m_pLocalNetworkManager;
    }
    LocalMigrationManager* GetLocalMigrationManager();
    LocalAroundNetworkSearchManager* GetLocalAroundNetworkSearchManager();
    LocalCreateNetworkJob* GetLocalCreateNetworkJob()
    {
        return m_pLocalCreateNetworkJob;
    }
    LocalBackgroundProcessJob* GetLocalBackgroundProcessJob()
    {
        return m_pLocalBackgroundProcessJob;
    }

    void* GetReceiveBuffer()
    {
        return m_pReceiveBuffer;
    }
    size_t GetReceiveBufferSize() const
    {
        return m_ReceiveBufferSize;
    }
    void* GetScanBuffer()
    {
        return m_pScanBuffer;
    }
    const void* GetScanBuffer() const
    {
        return m_pScanBuffer;
    }
    size_t GetScanBufferSize() const
    {
        return m_ScanBufferSize;
    }
    common::CriticalSection* GetScanBufferLock()
    {
        return &m_ScanBufferLock;
    }

    void RegisterUpdateEventCallback(LocalUpdateEventCallback callback, void* pParam);
    void UnregisterUpdateEventCallback();
    void ProcessUpdateEventDisconnected(u8 transportId);
    void ProcessUpdateEventMigrationStarted();
    void SetIsEnableGetScanData(bool isEnable)
    {
        m_IsEnableGetScanData = isEnable;
    }
    bool IsEnableGetScanData() const
    {
        return m_IsEnableGetScanData && !IsDuringHostMigration() && !IsConnected();
    }

    u32 GetSessionId() const;
    u32 GetSessionIdFromNetworkDescription(const LocalNetworkDescription* pNetworkDescription) const;
    bool IsForceDisconnect() const
    {
        return m_IsForceDisconnect;
    }
    void SetIsForceDisconnect(bool flag)
    {
        m_IsForceDisconnect = flag;
    }
    bool IsExpectDisconnectByKickout() const
    {
        return m_IsExpectDisconnectByKickout;
    }
    void SetIsExpectDisconnectByKickout(bool flag)
    {
        m_IsExpectDisconnectByKickout = flag;
    }

    u8 GetNetworkParticipantNum() const;

    bool IsAsyncProcessInProgress() const;
    //! @endcond

private:
    nn::Result InitializeCore(const UdsNetworkSetting& setting);
    void CreateJobs();
    void DestroyJobs();
    void CleanupJobs();

    // For asynchronous processing.
    enum AsyncProcessingName
    {
        ASYNC_NOTHING = 0,
        ASYNC_CREATE_NETWORK,
        ASYNC_DESTROY_NETWORK,
        ASYNC_SCAN_NETWORK,
        ASYNC_CONNECT_NETWORK,
        ASYNC_DISCONNECT_NETWORK
    };
    void ResetInternalAsyncState();
    bool IsCompletedAsyncProcess(AsyncProcessingName processingName);
    nn::Result GetAsyncProcessResult(AsyncProcessingName processingName);
    nn::Result CancelAsyncProcess(AsyncProcessingName processingName);

private:
    static LocalNetwork* s_pInstance;
    static bool s_IsInitialized;

    LocalNetworkType m_LocalNetworkType;
    LocalNetworkSetting* m_pLocalNetworkSetting;
    LocalNetworkDescription* m_paLocalNetworkDescriptionArray;
    u32 m_FoundNetworkNum;

    LocalNetworkManager* m_pLocalNetworkManager;
    LocalMigrationManager* m_pLocalMigrationManager;
    LocalAroundNetworkSearchManager* m_pLocalAroundNetworkSearchManager;

    LocalCreateNetworkJob* m_pLocalCreateNetworkJob;
    LocalDestroyNetworkJob* m_pLocalDestroyNetworkJob;
    LocalScanNetworkJob* m_pLocalScanNetworkJob;
    LocalConnectNetworkJob* m_pLocalConnectNetworkJob;
    LocalDisconnectNetworkJob* m_pLocalDisconnectNetworkJob;
    LocalForceDisconnectNetworkJob* m_pLocalForceDisconnectNetworkJob;
    LocalBackgroundProcessJob* m_pLocalBackgroundProcessJob;

    void* m_pReceiveBuffer;                    //!< Specifies the receive buffer. The buffer must be 4096-byte aligned, and its size must be a multiple of 4096.
    size_t m_ReceiveBufferSize;                //!< Specifies the receive buffer size.
    bool m_IsReceiveBufferAllocatedByApplication; //!< Flag that indicates that the receive buffer was specified using a parameter from within the application.
    void* m_pScanBuffer;                       //!< Specifies the scan buffer.
    size_t m_ScanBufferSize;                   //!< Specifies the scan buffer size.
    common::CriticalSection m_ScanBufferLock;  //!< Scan buffer lock.

    DisconnectReason m_DisconnectReason;

    // For asynchronous processing.
    common::CallContext m_InternalContext;
    AsyncProcessingName m_CurrentAsyncProcessingName;

    LocalUpdateEventCallback m_pfUpdateEventCallback;
    void* m_pUpdateEventCallbackParam;

    // Whether new participants are allowed or denied.
    AllowParticipatingState m_AllowParticipatingState;

    // Flag indicating whether scan data can be retrieved.
    bool m_IsEnableGetScanData;

    // For sleep prior to sending.
    bool m_IsNeedSleepBeforeSend;
    s64 m_LastSendTick;
    s64 m_SleepBeforeSendTick;

    // Flag for disconnection from the application, and then calling the destructor.
    bool m_IsForceDisconnect;
    // Flag for anticipating a disconnect due to being kicked out.
    bool m_IsExpectDisconnectByKickout;
};
}
}
}  // end of namespace nn::pia::local
