﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_LocalNetworkFactory.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/local/local_definitions.h>
#include <pia/transport/transport_NetworkFactory.h>


namespace nn
{
namespace pia
{
namespace local
{


/*!
  @class  nn::pia::local::LocalNetworkFactory
  @brief  Base class for creating the classes required for a network that uses the <tt>LocalNetwork</tt> class. It implements the factory pattern.

  @details  This class is never used by an application. Use a class that inherits this class.

  @see  nn::pia::local::UdsNetworkFactory

  @date 2014-08-01  Added the <tt>CreateMatchOpenRecruitmentJob()</tt> and <tt>CreateMatchCloseRecruitmentJob()</tt> functions.
  @date 2014-03-03  Initial version.
*/
class LocalNetworkFactory : public nn::pia::transport::NetworkFactory
{
public:
/*!
    @cond  PRIVATE
*/


/*!
    @brief  The destructor.
*/
    virtual ~LocalNetworkFactory(void);

/*!
    @brief  Instantiates an <tt>InputStream</tt> object.
*/
    virtual nn::pia::common::IPacketInput* CreateInputStream();

/*!
    @brief  Instantiates an <tt>OutputStream</tt> object.
*/
    virtual nn::pia::common::IPacketOutput* CreateOutputStream();


/*!
    @brief  Instantiates a <tt>ConnectStationJob</tt> object.
*/
    virtual nn::pia::transport::ConnectStationJob* CreateConnectStationJob();


/*!
    @brief  Instantiates a <tt>DisconnectStationJob</tt> object.
*/
    virtual nn::pia::transport::DisconnectStationJob* CreateDisconnectStationJob();


/*!
    @brief  Instantiates a <tt>CreateMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateCreateMeshJob();


/*!
    @brief  Instantiates a <tt>JoinMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateJoinMeshJob();


/*!
    @brief  Instantiates a <tt>LeaveMeshJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateLeaveMeshJob();


/*!
    @brief  Instantiates a <tt>ProcessHostMigrationJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateProcessHostMigrationJob();


/*!
    @brief  Instantiates a <tt>LeaveWithHostMigrationJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateLeaveWithHostMigrationJob();


/*!
    @brief  Creates a <tt>KickoutManageJob</tt> object.
*/
    virtual nn::pia::common::StepSequenceJob* CreateKickoutManageJob();


/*!
    @brief  Constructs a concrete instance of the <tt>PacketHandler</tt> class.
*/
    virtual transport::PacketHandler* CreatePacketHandler();


/*!
    @brief  Gets whether a signature is required.
*/
    virtual bool IsSignatureRequired() const;


/*!
    @brief  Instantiates a <tt>SignatureSettingStorage</tt> object.
*/
    virtual session::SignatureSettingStorage* CreateSignatureSettingStorage();


/*!
    @brief  Gets the total size of the header portion of packets at the network layer. The value is in bytes.
*/
    virtual size_t GetHeaderSize() const
    {
        return 10;
    }


/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    virtual void Trace(u64 flag) const;

/* For the auto-matchmaking feature. */
/*!
    @brief  Instantiates a <tt>CreateSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchCreateSessionJob();


/*!
    @brief  Instantiates an <tt>AutoMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchAutoMatchmakeSessionJob();


/*!
    @brief  Instantiates a <tt>BrowseMatchmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchBrowseMatchmakeSessionJob();


/*!
    @brief  Instantiates a <tt>JoinSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchJoinSessionJob();


/*!
    @brief  Instantiates a <tt>LeaveSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchLeaveSessionJob();


/*!
    @brief  Instantiates a <tt>DestroySessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchDestroySessionJob();

/*!
    @brief  Instantiates a <tt>JointSessionJob</tt> object for joint sessions.
*/
    virtual common::StepSequenceJob* CreateJointSessionJob();

/*!
    @brief  Instantiates a <tt>ModifyAttributeJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateModifyAttributeJob();

/*!
    @brief  Instantiates an <tt>UpdateSessionSetting</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateSessionSettingJob();

/*!
    @cond  PRIVATE
    @brief  Instantiates an <tt>UpdateApplicationBufferJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateApplicationDataJob();
    //! @endcond

/*!
    @brief  Instantiates a <tt>GenerateMatchmakeSystemPasswordJob</tt> object for the session.
           This object is not needed locally, so <tt>NULL</tt> is returned.
*/
    virtual common::StepSequenceJob* CreateGenerateMatchmakeSystemPasswordJob();

/*!
    @brief  Instantiates a <tt>ClearMatchmakeSystemPasswordJob</tt> object for the session.
           This object is not needed locally, so <tt>NULL</tt> is returned.
*/
    virtual common::StepSequenceJob* CreateClearMatchmakeSystemPasswordJob();

    virtual u32 GetMatchmakeSessionSystemPasswordSize() const;

    virtual wchar_t* CreateMatchmakeSessionSystemPassword(u32 size);

/*!
    @brief  Instantiates an <tt>IMatchmakeSession</tt> object for the session.
*/
    virtual session::IMatchmakeSession* CreateMatchmakeSession() = 0;

/*!
    @brief  Instantiates an <tt>ISessionInfoList</tt> object for the session.
*/
    virtual session::ISessionInfoList* CreateSessionInfoList(u8 browseSessionInfoListNum) = 0;

/*!
    @brief  Instantiates a <tt>MeshLayerController</tt> object for the session.
*/
    virtual session::MeshLayerController* CreateMatchMeshLayerController();

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    virtual session::IMatchmakeReferee* CreateMatchmakeReferee();
/*!
    @brief  Instantiates a <tt>RefereeStartGameJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateRefereeStartGameJob();

/*!
    @brief  Instantiates a <tt>RefereeEndGame</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateRefereeEndGameJob();

/*!
    @brief  Instantiates a <tt>RefereeEndGameWithoutReportJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateRefereeEndGameWithoutReportJob();

/*!
    @brief  Instantiates a <tt>RefereeGetOrCreateStatsJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateRefereeGetOrCreateStatsJob();
#endif

    virtual bool GetNecessaryBind() const;

    virtual bool GetNecessaryServer() const;

    virtual session::HostMigrationMode GetHostMigrationMode() const;

    virtual u8 GetBrowseSessionInfoListMax() const = 0;


/* For the auto-matchmaking feature. */


/*!
    @endcond
*/

protected:
/*!
    @cond  PRIVATE
    @brief  Instantiates an object with default parameters (default constructor).
*/
    LocalNetworkFactory(void);
    //! @endcond

private:
};
}
}
}  // End of namespace nn::pia::local.
