﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_LocalNetworkSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/definitions.h>
#include <pia/common/common_RootObject.h>

namespace nn
{
namespace pia
{
namespace local
{

/*!
@cond  PRIVATE
*/
static const u8 LOCAL_NETWORK_VERSION = 1;
//! @endcond

static const u8 STATION_MAX_ENTRY = 12;      //!<  The maximum number of nodes that can connect. The local network host is included in the number of connected nodes.
static const u32 SCAN_NETWORK_NUM_MAX = 16;  //!<  Defines the maximum number of search results that can be stored at one time when performing a network search.

static const u32 DEFAULT_AROUND_NETWORK_SEARCH_SCAN_INTERVAL = 200;     //!<  Default value, in milliseconds, for the scan interval during nearby network search.
static const u16 DEFAULT_AROUND_NETWORK_SEARCH_SCAN_TIME = 20;          //!<  Default value, in milliseconds, for the scan time during nearby network search.
static const u32 DEFAULT_AROUND_NETWORK_SEARCH_NETWORK_LIFE_TIME = 5000;//!<  Default value, in milliseconds, for the holding period for a discovered network during nearby network search.

/*!
@cond  PRIVATE
@brief  Local network state update event.
*/
enum LocalUpdateEvent
{
    LocalUpdateEventUnknown = 0,      //!<  Specifies an unknown state other than one of the following defined states. This state does not occur as long as the process is running normally.
    LocalUpdateEventDisconnected,     //!<  The event that occurs when disconnected from the local network.
    LocalUpdateEventMigrationStarted  //!<  Host migration has started.
};

typedef void (*LocalUpdateEventCallback)(LocalUpdateEvent e, u8 transportId, void* pParam);
//! @endcond

/*!
@cond  PRIVATE
*/
enum LocalNetworkType
{
    LOCAL_NETWORK_TYPE_UDS = 0
};
//! @endcond

/*!
  @brief  Class for configuring a local network.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsNetworkSetting

  @date 2014-03-03  Initial version.
*/
class LocalNetworkSetting : public nn::pia::common::RootObject
{
protected:
    LocalNetworkSetting()
      : m_IsEnableHostMigration(false),
        m_IsEnableAroundNetworkSearch(false),
        m_ApplicationVersion(0)
    {
    }

public:
    virtual LocalNetworkType GetLocalNetworkType() const = 0;

    bool m_IsEnableHostMigration;  //!<  Enables or disables local network host migration. Enabled if <tt>true</tt>; disabled if <tt>false</tt>.
    bool m_IsEnableAroundNetworkSearch;  //!<  Enables or disables nearby network search. Enabled if <tt>true</tt>; disabled if <tt>false</tt>.
    u8 m_ApplicationVersion;       //!<  Specifies the application version. Applications running a different version than this value can find the created network but not connect to it.
};


/*!
  @brief  Class for configuring the creation of a local network.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsCreateNetworkSetting

  @date 2014-03-03  Initial version.
*/
class LocalCreateNetworkSetting : public nn::pia::common::RootObject
{
protected:
    LocalCreateNetworkSetting()
    {
    }
};


/*!
  @brief  Class that holds information about a local network.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsNetworkDescription

  @date 2014-03-03  Initial version.
*/
class LocalNetworkDescription : public nn::pia::common::RootObject
{
public:
    virtual u32 GetCurrentParticipants() const = 0;
    virtual u32 GetMaxParticipants() const = 0;
    virtual bool IsOpened() const = 0;
    virtual u32 GetLocalCommunicationId() const = 0;
    virtual u8 GetSubId() const = 0;
    virtual u16 GetChannel() const = 0;
    virtual void GetBssid(bit8* pBssidBuffer) const = 0;

protected:
    LocalNetworkDescription()
    {
    }

public:
    virtual void Copy(const LocalNetworkDescription* description) = 0;
};


/*!
  @brief  Class for configuring a connection to a local network.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsConnectNetworkSetting

  @date 2014-03-03  Initial version.
*/
class LocalConnectNetworkSetting : public nn::pia::common::RootObject
{
protected:
    LocalConnectNetworkSetting()
      : m_pLocalNetworkDescription(NULL)
    {
    }

public:
    LocalNetworkDescription* m_pLocalNetworkDescription;  //!<  Information about the target network.
};

/*!
@cond  PRIVATE
The following declarations are for internal use by the library.
*/
class LocalScanNetworkSetting : public nn::pia::common::RootObject
{
protected:
    LocalScanNetworkSetting()
      : m_paLocalNetworkDescriptionArray(NULL),
        m_pFoundNetworkNum(NULL)
    {
    }

public:
    LocalNetworkDescription* m_paLocalNetworkDescriptionArray;  // SCAN_NETWORK_NUM_MAX
    u32* m_pFoundNetworkNum;
};

class LocalConnectionStatus : public nn::pia::common::RootObject
{
protected:
    LocalConnectionStatus()
    {
    }

public:
    virtual u8 GetParticipantNum() const = 0;
};

class LocalBeaconSystemData
{
public:
    u32 m_NetworkId;          //!<  Specifies the network ID determined by the host. This ID is used to determine whether to connect to a network during host migration.
    u32 m_SessionId;          //!<  Session ID that does not change during host migration.
    u8 m_NetworkVersion;      //!<  Specifies the local network version.
    u8 m_ApplicationVersion;  //!<  Specifies the application version set by the application.
    u8 m_Reserved[6];

protected:
    LocalBeaconSystemData()
      : m_NetworkId(0),
        m_SessionId(0),
        m_NetworkVersion(LOCAL_NETWORK_VERSION),
        m_ApplicationVersion(0)
    {
        memset(m_Reserved, 0, sizeof(m_Reserved));
    }
};


/*!
Internal library declarations end here.
@endcond
*/


/*!
  @brief  Class that holds information about stations connected to a local network.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsStationInfo

  @date 2014-03-03  Initial version.
*/
class LocalStationInfo : public nn::pia::common::RootObject
{
protected:
    LocalStationInfo()
      : m_Role(STATION_INFO_ROLE_NONE)
    {
    }

public:
/*!
    @brief  Indicates the station's role (not connected, host, or client).
*/
    enum STATION_INFO_ROLE
    {
        STATION_INFO_ROLE_NONE = 0,  //!<  Not connected
        STATION_INFO_ROLE_HOST,      //!<  Host.
        STATION_INFO_ROLE_CLIENT     //!<  Client
    };
    STATION_INFO_ROLE m_Role;  //!<  The station's role (not connected, host, or client).
};


/*!
  @brief  Class that holds information about the network discovered using nearby network search.

  @details  This class is never instantiated by an application.
           Use a class that inherits this class.

  @see  nn::pia::local::UdsAroundNetworkInfo

  @date 2015-02-10  Initial version.
*/
class LocalAroundNetworkInfo : public nn::pia::common::RootObject
{
protected:
    LocalAroundNetworkInfo(){}

public:
};

/*!
  @brief  Class for configuring the nearby network search feature.

  @date 2015-02-10  Initial version.
*/
class LocalAroundNetworkSearchSetting : public nn::pia::common::RootObject
{
public:
    LocalAroundNetworkSearchSetting()
        : m_LocalCommunicationId(0)
        , m_SubId(0xff)
        , m_ScanInterval(DEFAULT_AROUND_NETWORK_SEARCH_SCAN_INTERVAL)
        , m_ScanTime(DEFAULT_AROUND_NETWORK_SEARCH_SCAN_TIME)
        , m_NetworkInfoLifeTime(DEFAULT_AROUND_NETWORK_SEARCH_NETWORK_LIFE_TIME)
    {
    }

    u32 m_LocalCommunicationId; //!<  Specifies the local communication ID. Specify the value generated by <tt>@ref LocalNetwork::CreateLocalCommunicationId</tt>.
    u8 m_SubId;                 //!<  Specifies an ID for identifying the communication mode. To search all IDs, specify <tt>0xff</tt>.
    u32 m_ScanInterval;         //!<  The scan interval, in milliseconds.
    u16 m_ScanTime;             //!<  The scan time, in milliseconds.
    u32 m_NetworkInfoLifeTime;  //!<  The holding period, in milliseconds, for the discovered network.
};

}
}
}  // End of namespace nn::pia::local.
