﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_LocalSessionInfo.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_ISessionInfo.h>

namespace nn
{
namespace pia
{
namespace session
{
// Forward declarations.
class SessionStationInfo;
}
}
}

namespace nn
{
namespace pia
{
namespace local
{

class LocalNetworkDescription;

/*!
    @brief  Data class for local communication sessions.
*/
class LocalSessionInfo : public session::ISessionInfo
{
public:
    LocalSessionInfo();
    virtual ~LocalSessionInfo();

/*!
    @brief  Gets the minimum number of people who can participate in a session.
    @return  Returns the minimum number of people who can participate in a session.
*/
    virtual u32 GetMinParticipants() const;

/*!
    @brief  Gets the session's application-defined data.
    @details  Gets the beacon data set by the application.
    @param[out] pBuffer  Pointer to the buffer for copying the session's application-defined data.
    @param[in] size  Specifies the size of the buffer required for copying the session's application-defined data.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @if CTR_DOC
    @retval ResultNoData  There was no data to get. Handle appropriately in the application.
    @endif
*/
    virtual nn::Result GetApplicationDataBuf(void* pBuffer, size_t size) = 0;

/*!
    @brief  Gets the length of the session's application-defined data.
    @param[out] pSize  Pointer to the container for the length of the session's application-defined data.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result GetApplicationDataSize(size_t* pSize) = 0;

/*!
    @if CTR_DOC
    @brief  Gets the session's signal strength.
    @param[out] pRadioStrength  Pointer to the signal strength container.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @endif
*/
    virtual nn::Result GetRadioStrength(u8* pRadioStrength) const;

/*!
    @if CTR_DOC
    @brief  Gets the information for stations currently participating in sessions.
    @details  Succeeds only when local communication is ongoing.
    @param[out] pSessionStationInfo  Specifies a pointer to the station information container.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @retval ResultNoData  Indicates that information could not be obtained. Returned only during local communications. Handle appropriately in the application.
    @retval ResultInvalidConnection  Indicates that the device entered wireless off mode. You must restart. Returned only during local communications. Perform cleanup.
    @endif
*/
    virtual nn::Result GetSessionStationInfo(session::SessionStationInfo* pSessionStationInfo) const = 0;

/*!
    @if CTR_DOC
    @brief  Gets the BSSID (the host's MAC address).
    @details  The size of the BSSID is <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes. Set a buffer of this size in the parameter.
    @param[out] pBssidBuffer  Buffer that stores the BSSID. The size must be <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @endif
*/
    virtual nn::Result GetBssid(bit8* pBssidBuffer) const = 0;

/*!
    @brief  Clears session information.
*/
    virtual void Clear();

    //! @cond PRIVATE
    virtual const LocalNetworkDescription* GetLocalNetworkDescription() const = 0;
    virtual void SetLocalNetworkDescription(LocalNetworkDescription* pDescription) = 0;
    void SetRadioStrength(u32 index); // <tt>index</tt> is the index when performing a network search.
    virtual void SetSessionStationInfo(u32 index) = 0; // <tt>index</tt> is the index when performing a network search.
    //! @endcond

protected:
    bool m_IsNetworkDescriptionSet;
    u8 m_RadioStrength;
};
}
}
}  // End of namespace nn::pia::local.
