﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_LocalSessionSearchCriteria.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_SessionSearchCriteria.h>

namespace nn
{
namespace pia
{
namespace local
{
// Forward declarations.
class LocalNetworkDescription;

/*!
  @brief  The search condition class used when searching for a matchmaking session using <tt>LocalNetwork</tt>.

  @date 2014-03-03  Initial version.

  @see  nn::pia::session::SessionSearchCriteria
 */
class LocalSessionSearchCriteria : public nn::pia::session::SessionSearchCriteria
{
public:
    // The constructor.
    LocalSessionSearchCriteria()
      : session::SessionSearchCriteria(),
        m_MaxParticipantsRangeMax(INVALID_PARICIPANTS_NUM),
        m_MaxParticipantsRangeMin(INVALID_PARICIPANTS_NUM),
        m_bOpenedOnly(true),
        m_bVacantOnly(false),
        m_LocalCommunicationId(0),
        m_SubId(0)
    {
    }

    // The destructor.
    virtual ~LocalSessionSearchCriteria()
    {
    }

/*!
    @brief  Sets the maximum participants in the search criteria.
    @param[in] max  The maximum participants value to set in the search criteria.
*/
    void SetMaxParticipants(u16 max)
    {
        m_MaxParticipantsRangeMax = max;
        m_MaxParticipantsRangeMin = max;
    }

/*!
    @brief  Sets a range of maximum participants in the search criteria.
    @param[in] min  The minimum value of the range to set for maximum participants in the search criteria.
    @param[in] max  The maximum value of the range to set for maximum participants in the search criteria.
*/
    void SetMaxParticipantsRange(u16 min, u16 max)
    {
        m_MaxParticipantsRangeMax = max;
        m_MaxParticipantsRangeMin = min;
    };

/*!
    @brief  Gets the minimum value of the range of maximum participants set by the search criteria.
    @return  Returns the minimum value of the range of maximum participants set by the search criteria.
*/
    u16 GetMaxParticipantsRangeMin() const
    {
        return m_MaxParticipantsRangeMin;
    }

/*!
    @brief  Gets the maximum value of the range of maximum participants set by the search criteria.
    @return  Returns the maximum value of the range of maximum participants set by the search criteria.
*/
    u16 GetMaxParticipantsRangeMax() const
    {
        return m_MaxParticipantsRangeMax;
    }

/*!
    @brief  Sets whether to search only for sessions that are open to participants in the search criteria.
    @param[in] bOpenedOnly  Specify <tt>true</tt> to search only for sessions that are open to participants.
*/
    void SetOpenedOnly(bool bOpenedOnly)
    {
        m_bOpenedOnly = bOpenedOnly;
    }

/*!
    @brief  Sets whether to search only for sessions that have openings in the search criteria.
    @param[in] bVacantOnly  Specify <tt>true</tt> to search only for sessions that have openings.
*/
    void SetVacantOnly(bool bVacantOnly)
    {
        m_bVacantOnly = bVacantOnly;
    }

/*!
    @brief  Sets the local communication ID.

    @details  Specifies the value generated by <tt>@ref LocalNetwork::CreateLocalCommunicationId</tt>.

    @param[in] localCommunicationId  Specifies the local communication ID.
 */
    void SetLocalCommunicationId(u32 localCommunicationId)
    {
        m_LocalCommunicationId = localCommunicationId;
    }

/*!
    @cond PRIVATE
 */
    u32 GetLocalCommunicationId() const
    {
        return m_LocalCommunicationId;
    }
    //! @endcond

/*!
    @brief  Sets the ID for identifying the communication mode. To search all IDs, specify <tt>0xff</tt>.

    @param[in] subId  Specifies an ID for identifying the communication mode.
 */
    void SetSubId(u8 subId)
    {
        m_SubId = subId;
    }

/*!
    @cond PRIVATE
 */
    u8 GetSubId() const
    {
        return m_SubId;
    }
    //! @endcond

/*!
    @cond PRIVATE
 */
    bool IsHit(LocalNetworkDescription* pDescription);
    //! @endcond

    LocalSessionSearchCriteria& operator=(const LocalSessionSearchCriteria& r)
    {
        session::SessionSearchCriteria::operator=(r);
        m_MaxParticipantsRangeMax = r.m_MaxParticipantsRangeMax;
        m_MaxParticipantsRangeMin = r.m_MaxParticipantsRangeMin;
        m_bOpenedOnly = r.m_bOpenedOnly;
        m_bVacantOnly = r.m_bVacantOnly;
        m_LocalCommunicationId = r.m_LocalCommunicationId;
        m_SubId = r.m_SubId;
        return *this;
    }

    // Prints information that is useful for debugging.
    virtual void Trace(u64 flag) const
    {
        NN_PIA_DUMMY_PARAM(flag);
    }


private:
    // The copy constructor is sealed.
    LocalSessionSearchCriteria(const LocalSessionSearchCriteria&);

    // Invalid number of participants setting.
    static const u16 INVALID_PARICIPANTS_NUM = 0xffff;

    // Maximum participants range.
    u16 m_MaxParticipantsRangeMax;
    u16 m_MaxParticipantsRangeMin;

    // Whether to list only open sessions.
    bool m_bOpenedOnly;
    // Whether to remove full sessions from the results.
    bool m_bVacantOnly;

    u32 m_LocalCommunicationId;
    u8 m_SubId;
};
}
}
}  // End of namespace nn::pia::local.
