﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_UdsNetworkSetting.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/definitions.h>

#include <pia/local/local_LocalNetworkSetting.h>

namespace nn
{
namespace pia
{
namespace local
{

static const u32 UDS_RECEIVE_BUFFER_SIZE_MIN = 4 * 4096;  //!<  Defines the minimum receive buffer size.
static const u32 UDS_SCAN_BUFFER_SIZE_MIN = 1 * 1024;     //!<  Minimum size of the scan buffer.

static const size_t UDS_PASSPHRASE_LENGTH_MIN = nn::uds::UDS_PASSPHRASE_LENGTH_MIN;      //!<  Minimum size of the passphrase for generating the encrypted key used in communication.
static const size_t UDS_PASSPHRASE_LENGTH_MAX = nn::uds::UDS_PASSPHRASE_LENGTH_MAX;      //!<  Maximum size of the passphrase for generating the encryption key used in communication.
static const size_t UDS_APPLICATION_DATA_SIZE_MAX = nn::uds::NET_DESC_APPDATA_SIZE_MAX;  //!<  Maximum size of the application data that can be set for the beacon. The size available for the application to use will be less than this amount because the size includes a region for system use.

/*!
  @brief  Class for configuring networks that use the UDS library.

  @date 2014-03-03  Initial version.
*/
class UdsNetworkSetting : public LocalNetworkSetting
{
public:
    UdsNetworkSetting()
      : LocalNetworkSetting(),
        m_ReceiveBufferSize(0),
        m_pReceiveBuffer(NULL),
        m_ScanBufferSize(0),
        m_SendOption(nn::uds::NO_WAIT),
        m_ReceiveOption(nn::uds::NO_WAIT),
        m_pUserName(NULL)
    {
    }

/*!
    @cond  PRIVATE
*/
    virtual LocalNetworkType GetLocalNetworkType() const
    {
        return LOCAL_NETWORK_TYPE_UDS;
    }
    //! @endcond

    size_t m_ReceiveBufferSize;           //!<  Specifies the receive buffer size. The size must be at least <tt>@ref local::UDS_RECEIVE_BUFFER_SIZE_MIN</tt> and a multiple of 4096.
    void* m_pReceiveBuffer;               //!<  Specifies the receive buffer size. This value should be specified if you want to allocate the buffer from within the application. It must be 4096-byte aligned. If left unspecified, a buffer is allocated within the Pia library as usual.
    size_t m_ScanBufferSize;              //!<  Specifies the scan buffer size. The size must be at least <tt>@ref local::UDS_SCAN_BUFFER_SIZE_MIN</tt>.
    u8 m_SendOption;                      //!<  Specifies UDS send options. Available options include <tt>nn::uds::NO_WAIT</tt> and <tt>nn::uds::FORCE_DIRECT_BC</tt>. Use the OR operator to set multiple options.
    u8 m_ReceiveOption;                   //!<  Specifies UDS receive options. Always set <tt>nn::uds::NO_WAIT</tt> to on. Use the OR operator to set multiple options.
    nn::cfg::CTR::UserName* m_pUserName;  //!<  Returns the user name.
};

/*!
  @brief  Class for configuring the creation of networks that use the UDS library.

  @date 2014-03-03  Initial version.
*/
class UdsCreateNetworkSetting : public LocalCreateNetworkSetting
{
public:
/*!
    @brief  The default constructor.
*/
    UdsCreateNetworkSetting()
      : LocalCreateNetworkSetting(),
        m_SubId(0),
        m_MaxEntry(0),
        m_LocalCommunicationId(0),
        m_PassphraseLength(0),
        m_Channel(0),
        m_ApplicationDataSize(0)
    {
        memset(m_Passphrase, 0, sizeof(m_Passphrase));
        memset(m_ApplicationData, 0, sizeof(m_ApplicationData));
    }

    u8 m_SubId;                                           //!<  Specifies an ID for identifying the communication mode. Set this parameter to a value between <tt>0x00</tt> and <tt>0xFE</tt>.
    u8 m_MaxEntry;                                        //!<  Maximum number of stations that can connect to the network.
    u32 m_LocalCommunicationId;                           //!<  Specifies the local communication ID. Specify the value generated by <tt>@ref LocalNetwork::CreateLocalCommunicationId</tt>.
    char m_Passphrase[UDS_PASSPHRASE_LENGTH_MAX];         //!<  Specifies the passphrase used to encrypt the wireless layer. The size must be at least <tt>@ref UDS_PASSPHRASE_LENGTH_MIN</tt> but no greater than <tt>@ref UDS_PASSPHRASE_LENGTH_MAX</tt>.
    size_t m_PassphraseLength;                            //!<  Specifies the size of the passphrase used to encrypt the wireless layer. Must be at least <tt>@ref UDS_PASSPHRASE_LENGTH_MIN</tt> but no greater than <tt>@ref UDS_PASSPHRASE_LENGTH_MAX</tt>.
    u8 m_Channel;                                         //!<  Specifies the channel to use for communication. Specify channel <tt>0</tt> (automatic), <tt>1</tt>, <tt>6</tt>, or <tt>11</tt>. When this function runs on production devices, the channel is always selected automatically.
    u8 m_ApplicationData[UDS_APPLICATION_DATA_SIZE_MAX];  //!<  The optional data to set for the beacon. The size must be no greater than the value that can be obtained with <tt>@ref LocalNetwork::GetBeaconApplicationDataSizeMax</tt>.
    size_t m_ApplicationDataSize;                         //!<  Size of the optional data to set for the beacon. The size must be no greater than the value that can be obtained using <tt>@ref LocalNetwork::GetBeaconApplicationDataSizeMax</tt>.
};

/*!
  @brief  Class that holds information about networks that use the UDS library.

  @date 2014-03-03  Initial version.
*/
class UdsNetworkDescription : public LocalNetworkDescription
{
public:
/*!
    @brief  The default constructor.
*/
    UdsNetworkDescription()
      : LocalNetworkDescription()
    {
    }

/*!
    @brief  Gets the number of stations currently participating in the network.

    @details  The number of stations that can be obtained is the number at the point when the network is scanned.

    @return  Returns the number of stations participating in the network.
*/
    virtual u32 GetCurrentParticipants() const
    {
        return m_NetworkDescription.GetNowEntry();
    }

/*!
    @brief  Gets the maximum number of stations that can participate in the network.

    @return  Returns the maximum number of stations that can participate in the network.
*/
    virtual u32 GetMaxParticipants() const
    {
        return m_NetworkDescription.GetMaxEntry();
    }

/*!
    @brief  Gets whether the network allows participation.

    @details  The state that can be obtained is the state at the point when the network is scanned.

    @return  Returns whether the network allows participation.

    @see  LocalNetwork::DisallowParticipating, LocalNetwork::AllowParticipating
*/
    virtual bool IsOpened() const
    {
        return m_NetworkDescription.CanConnectAsClient();
    }

    virtual u32 GetLocalCommunicationId() const
    {
        return m_NetworkDescription.GetLocalCommunicationId();
    }

/*!
    @brief  Gets the ID for identifying the communication mode.

    @return  Returns the ID for identifying the communication mode.
*/
    virtual u8 GetSubId() const
    {
        return m_NetworkDescription.GetSubId();
    }

/*!
    @brief  Gets the channel being used for the communication.

    @details  There is normally no need to reference this.

    @return  Returns the channel being used for communication.
*/
    virtual u16 GetChannel() const
    {
        return m_NetworkDescription.GetChannel();
    }

/*!
    @brief  Gets the BSSID (the host's MAC address).

    @details  The size of the BSSID is <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes. Set a buffer of this size in the parameter.

    @param[out] pBssidBuffer  Buffer that stores the BSSID. The size must be <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes.
*/
    virtual void GetBssid(bit8* pBssidBuffer) const
    {
        if (!PIA_IS_VALID_POINTER(pBssidBuffer))
        {
            return;
        }
        memcpy(pBssidBuffer, m_NetworkDescription.GetBssid(), nn::uds::MAC_ADDRESS_SIZE);
    }

/*!
    @cond  PRIVATE
*/
    const nn::uds::NetworkDescription& GetNetworkDescription() const
    {
        return m_NetworkDescription;
    }
    //! @endcond

/*!
    @cond  PRIVATE
*/
    nn::uds::NetworkDescription& GetNetworkDescription()
    {
        return m_NetworkDescription;
    }
    //! @endcond

/*!
    @cond  PRIVATE
*/
    void SetNetworkDescription(const nn::uds::NetworkDescription& networkDescription)
    {
        m_NetworkDescription = networkDescription;
    }
    //! @endcond

    // n2908: Copy <tt>NetworkDescription</tt>.
    virtual void Copy(const LocalNetworkDescription* description)
    {
        const UdsNetworkDescription* pUdsDesc = static_cast<const UdsNetworkDescription*>(description);
        m_NetworkDescription = pUdsDesc->m_NetworkDescription;
    }

private:
    nn::uds::NetworkDescription m_NetworkDescription;  //!<  Network information.
};

/*!
  @brief  Class for configuring connections to networks that use the UDS library.

  @date 2014-03-03  Initial version.
*/
class UdsConnectNetworkSetting : public LocalConnectNetworkSetting
{
public:
    UdsConnectNetworkSetting()
      : LocalConnectNetworkSetting(),
        m_PassphraseLength(0)
    {
        memset(m_Passphrase, 0, sizeof(m_Passphrase));
    }

    char m_Passphrase[UDS_PASSPHRASE_LENGTH_MAX];  //!<  Specifies the passphrase used to encrypt the wireless layer. The size must be at least <tt>@ref UDS_PASSPHRASE_LENGTH_MIN</tt> but no greater than <tt>@ref UDS_PASSPHRASE_LENGTH_MAX</tt>.
    u8 m_PassphraseLength;                         //!<  Specifies the size of the passphrase used to encrypt the wireless layer. Must be at least <tt>@ref UDS_PASSPHRASE_LENGTH_MIN</tt> but no greater than <tt>@ref UDS_PASSPHRASE_LENGTH_MAX</tt>.
};


/*!
@cond  PRIVATE
The following declarations are for internal use by the library.
*/
class UdsScanNetworkSetting : public LocalScanNetworkSetting
{
public:
    UdsScanNetworkSetting()
      : LocalScanNetworkSetting(),
        m_SubId(0),
        m_LocalCommunicationId(0),
        m_pScanBuffer(NULL),
        m_ScanBufferSize(0)
    {
    }

    u8 m_SubId;
    u32 m_LocalCommunicationId;

    void* m_pScanBuffer;   // Same as for <tt>UdsNetworkSetting</tt>.
    u32 m_ScanBufferSize;  // Same as for <tt>UdsNetworkSetting</tt>.
};

class UdsNetworkConnectionStatus : public LocalConnectionStatus
{
public:
    UdsNetworkConnectionStatus()
      : LocalConnectionStatus()
    {
    }

    virtual u8 GetParticipantNum() const
    {
        return m_ConnectionStatus.nowEntry;
    }

    nn::uds::ConnectionStatus m_ConnectionStatus;
};

class UdsBeaconSystemData : public LocalBeaconSystemData
{
public:
    UdsBeaconSystemData()
      : LocalBeaconSystemData()
    {
    }
};

class UdsBeaconSystemDataHostMigration : public LocalBeaconSystemData
{
public:
    UdsBeaconSystemDataHostMigration()
    {
        memset(m_MacAddress, 0, sizeof(m_MacAddress));
        memset(m_ReservedUdsHostMigration, 0, sizeof(m_ReservedUdsHostMigration));
    }

    bit8 m_MacAddress[nn::uds::MAC_ADDRESS_SIZE];  //!<  For determining the network to connect to after host migration.
    u8 m_ReservedUdsHostMigration[6];
};

/*!
Internal library declarations end here.
@endcond
*/


/*!
  @brief  Class that holds information about stations that are connected to a network that uses the UDS library.

  @date 2014-03-03  Initial version.
*/
class UdsStationInfo : public LocalStationInfo
{
public:
    UdsStationInfo()
    {
    }

    nn::uds::ScrambledLocalFriendCode m_ScrambledLocalFriendCode;  //!<  Information that can be converted into a local friend code.
    nn::cfg::UserName m_UserName;                                  //!<  The user name.
};


/*!
  @brief  Class that holds information about a network discovered via the nearby network search feature, using the UDS library.

  @date 2015-02-10  Initial version.
*/
class UdsAroundNetworkInfo : public LocalAroundNetworkInfo
{
public:
    UdsAroundNetworkInfo() :
        LocalAroundNetworkInfo()
    {
    }

    UdsNetworkDescription m_NetworkDescription;             //!<  Information about the network.
    UdsStationInfo m_NodeInformation[STATION_MAX_ENTRY];    //!<  Information about the stations connected to the network.
};

}
}
}  // End of namespace nn::pia::local.
