﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: local_UdsSessionInfo.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/local/local_LocalSessionInfo.h>
#include <pia/local/local_UdsNetworkSetting.h>

namespace nn
{
namespace pia
{
namespace local
{

/*!
    @brief  Data class for local communication sessions.
*/
class UdsSessionInfo : public local::LocalSessionInfo
{
public:
    UdsSessionInfo();
    virtual ~UdsSessionInfo();

/*!
    @brief  Copies session information.
    @details  Use this when you want the application to retain session information.
    @param[in] rhs  Session information to copy.
*/
    void Copy(UdsSessionInfo& rhs);

/*!
    @brief  Clears session information.
*/
    virtual void Clear();

/*!
    @brief  Gets the session's game mode.
    @details  A communication mode ID is cast as a <tt>u32</tt> argument and returned.
    @return  Returns the game mode.
*/
    virtual u32 GetGameMode() const;

/*!
    @brief  Gets the session's ID.
    @return  Returns the session ID.
*/
    virtual u32 GetSessionId() const;

/*!
    @brief  Gets the number of people participating in a session.
    @return  Returns the number of people participating in a session.
*/
    virtual u32 GetCurrentParticipants() const;

/*!
    @brief  Gets the maximum number of people who can participate in a session.
    @return  Returns the maximum number of people who can participate in a session.
*/
    virtual u32 GetMaxParticipants() const;

/*!
    @brief  Gets the session's application-defined data.
    @details  Gets the beacon data set by the application.
    @param[out] pBuffer  Pointer to the buffer for copying the session's application-defined data.
    @param[in] size  Specifies the size of the buffer required for copying the session's application-defined data.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @if CTR_DOC
    @retval ResultNoData  There was no data to get. Handle appropriately in the application.
    @endif
*/
    virtual nn::Result GetApplicationDataBuf(void* pBuffer, size_t size);

/*!
    @brief  Gets the length of the session's application-defined data.
    @param[out] pSize  Pointer to the container for the length of the session's application-defined data.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result GetApplicationDataSize(size_t* pSize);


/*!
    @brief  Gets whether a session is open for recruitment.
    @return  Returns <tt>true</tt> if the session is open for recruitment, and returns <tt>false</tt> if it is closed.
*/
    virtual bool IsOpened() const;

/*!
    @if CTR_DOC
    @brief  Gets the information for stations currently participating in sessions.
    @param[out] pSessionStationInfo  Specifies a pointer to the station information container.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @retval ResultNoData  Indicates that information could not be obtained. Returned only during local communications. Handle appropriately in the application.
    @retval ResultInvalidConnection  Indicates that the device entered wireless off mode. You must restart. Returned only during local communications. Perform cleanup.
    @endif
*/
    virtual nn::Result GetSessionStationInfo(session::SessionStationInfo* pSessionStationInfo) const;

/*!
    @if CTR_DOC
    @brief  Gets the BSSID (the host's MAC address).
    @details  The size of the BSSID is <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes. Set a buffer of this size in the parameter.
    @param[out] pBssidBuffer  Buffer that stores the BSSID. The size must be <tt>nn::uds::MAC_ADDRESS_SIZE</tt> bytes.
    @return  Returns a <tt>Result</tt> value indicating success if information was successfully obtained.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not in a state where data can be retrieved. Programming error. Fix your program so that this error is not returned.
    @endif
*/
    virtual nn::Result GetBssid(bit8* pBssidBuffer) const;

/*!
    @brief  Prints information that is useful for debugging.
    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    virtual void Trace(u64 flag);

    //! @cond PRIVATE
    virtual const LocalNetworkDescription* GetLocalNetworkDescription() const;
    virtual void SetLocalNetworkDescription(LocalNetworkDescription* pDescription);

    virtual void SetSessionStationInfo(u32 index);
    //! @endcond

private:
    UdsNetworkDescription m_UdsNetworkDescription;
    UdsStationInfo m_UdsStationInfoArray[STATION_MAX_ENTRY];
};
}
}
}  // End of namespace nn::pia::local.
