﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     reckoning_ReckoningCloneElement.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/reckoning/reckoning_definitions.h>
#include <pia/reckoning/reckoning_ReckoningCloneElementBase.h>
#include <pia/clone/clone_SerializePolicyDefinition.h>
#include <pia/clone/clone_VoidMemberConcealer.h>
#include <pia/clone/clone_CloneProtocol.h>


namespace nn
{
namespace pia
{
namespace reckoning
{

/*!
@brief  Manages the sending and receiving of values while using dead reckoning.

@tparam ReckoningStrategy_  Specifies the type of class that defines the reckoning algorithm.
@tparam SerializePolicy_  Specifies the algorithm to use for serializing <tt>ReckoningStrategy::Sample</tt>.

@details  Some of the configured values are saved and sent as a sample on which to predict the current value, eliminating the need to send all of the configured values.

The following members must be defined in the class specified for <tt>ReckoningStrategy</tt>.

Types

@li  @code Value @endcode
Specifies the type of value that is set and retrieved using <tt>SetValue</tt> and <tt>GetValue</tt>.
@li  @code Sample @endcode
Specifies the type of the sample. The value is actually sent and received using this type.
@li  @code SetValueArg @endcode
Specifies the type of value that is set when calling <tt>SetValue</tt>, and is passed to the <tt>ReckoningStrategy::CheckSample</tt> and <tt>ReckoningStrategy::MakeSample</tt> functions.
Specify <tt>void</tt> if this is not required.
@li  @code Work @endcode
Specifies the type that stores the pre-calculated result for prediction. Specify <tt>void</tt> if this is not required.

Constants

@li  @code static const size_t BUFFER_SIZE @endcode
Specifies the number of samples to retain.

Functions

@li  @code bool Estimate(
Value* pValue,
const nn::pia::reckoning::ReckoningCloneElementBase::SampleAccessor<Sample>& accessor,
nn::pia::clone::ClockValue clock,
Work* pWork); @endcode
Calculate the prediction.
If the prediction could be calculated, set the value in <tt>*pValue</tt> and return <tt>true</tt>.
If it could not be calculated, return <tt>false</tt>.
@li  @code bool CheckSample(
bool* pIsReliable,
const Value& value,
const SetValueArg* cpSetValueArg,
const nn::pia::reckoning::ReckoningCloneElementBase::SampleAccessor<Sample>& accessor,
nn::pia::clone::ClockValue clock,
Work* pWork); @endcode
Determine whether to use the <tt>value</tt> set with <tt>SetValue</tt> as a sample.
Return <tt>true</tt> to use it as a sample.
If you want to use reliable communication when sending the sample, set <tt>*pIsReliable</tt> to <tt>true</tt>.
@li  @code void MakeSample(
Sample* pSample,
const Value& value,
const SetValueArg* cpSetValueArg,
const nn::pia::reckoning::ReckoningCloneElementBase::SampleAccessor<Sample>& accessor,
nn::pia::clone::ClockValue clock,
Work* pWork); @endcode
Calculate the value to save as a sample.
This is called when <tt>CheckSample</tt> determines the value is to be used as a sample and the buffer was successfully allocated.
Set the sample in <tt>*pSample</tt>.
@li  @code void OnUpdateSample(
const nn::pia::reckoning::ReckoningCloneElementBase::SampleAccessor<Sample>& accessor,
int index,
Work* pWork); @endcode
This is called when a sample is added.
Update the <tt>pWork</tt> value as necessary.

For more information, see the Programming Manual.

@date 2013-10-28  Initial version.

*/
template <
    typename ReckoningStrategy_,
    typename SerializePolicy_ = clone::HostByteOrderSerializePolicy<typename ReckoningStrategy_::Sample> >
class ReckoningCloneElement : public ReckoningCloneElementBase
{
public:
/*!
@brief  Defines the type of the dead reckoning algorithm.
*/
    typedef ReckoningStrategy_ ReckoningStrategy;


/*!
@brief  Specifies the type of value managed by this object.
*/
    typedef typename ReckoningStrategy::Value Value;


/*!
@brief  Defines the type of the sample.
*/
    typedef typename ReckoningStrategy::Sample Sample;


/*!
@brief  Specifies the type of value that is passed to <tt>ReckoningStrategy::CheckSample</tt> and <tt>ReckoningStrategy::MakeSample</tt> from <tt>SetValue</tt>.
*/
    typedef typename ReckoningStrategy::SetValueArg SetValueArg;


/*!
@brief  Defines the type of the working buffer.
*/
    typedef typename ReckoningStrategy::Work Work;


/*!
@brief  Specifies the algorithm to use for serializing <tt>Sample</tt>.
*/
    typedef SerializePolicy_ SerializePolicy;


/*!
@brief  Specifies the size of the buffer that holds samples.
*/
    static const size_t BUFFER_SIZE = ReckoningStrategy::BUFFER_SIZE;


/*!
@brief  Instantiates the object (constructor).
@param[in] pReckoningStrategy  Specifies an instance of the dead reckoning algorithm.
You can skip specifying the algorithm here, but you must set a valid instance using <tt>SetReckoningStrategy</tt> before passing this object to <tt>CloneBase::RegisterElement</tt>.

*/
    explicit ReckoningCloneElement(ReckoningStrategy* pReckoningStrategy = NULL);


/*!
@brief  Destroys the object (destructor).
*/
    virtual ~ReckoningCloneElement()
    {
    }


/*!
@brief  Sets an instance of the dead reckoning algorithm.
@details  This cannot be set when the object is registered to a <tt>@ref clone::CloneBase "CloneBase"</tt>.
@param[in] pReckoningStrategy  Specifies the dead reckoning algorithm to set.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the object is registered to a <tt>CloneBase</tt>. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetReckoningStrategy(ReckoningStrategy* pReckoningStrategy);


/*!
@brief  Resets the currently set instance of the dead reckoning algorithm.
@details  This cannot be reset when the object is registered to a <tt>@ref clone::CloneBase "CloneBase"</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the object is registered to a <tt>CloneBase</tt>. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result ResetReckoningStrategy()
    {
        return SetReckoningStrategy(NULL);
    }


/*!
@brief  Sets values.
@param[in] value  Specifies the flags to set.
@param[in] cpSetValueArg  Specifies the value to pass to <tt>ReckoningStrategy::CheckSample</tt> and <tt>ReckoningStrategy::MakeSample</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the registered <tt>CloneBase</tt> cannot send in its current state. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetValue(const Value& value, const SetValueArg* cpSetValueArg = NULL);


/*!
@brief  Gets a value.
@return  Returns the value managed by this object. This value is undefined when <tt>@ref IsValidValue</tt> returns <tt>false</tt>.
*/
    const Value& GetValue()
    {
        UpdateValue();
        return *m_ValueWork.GetT1Ptr();
    }


    //! @cond PRIVATE

public:
    virtual bool IsInitialized() const
    {
        return PIA_IS_VALID_POINTER(m_pReckoningStrategy);
    }

protected:
    virtual size_t GetSize() const
    {
        return SerializePolicy::GetSize();
    }
    virtual void Serialize(void* pBuffer, const ReckoningSendToken* cpToken);
    virtual void ClearValue()
    {
        m_ValueWork = clone::VoidMemberConcealer<Value, Work>();
    }
    virtual bool EstimateValue(clone::ClockValue clock);

    virtual void Deserialize(ReckoningSendToken* pToken, const void* cpData);
    virtual void OnUpdateSample(int idx);

private:
    ReckoningSendToken* m_apToken[BUFFER_SIZE];
    SampleReckoningSendToken<Sample> m_aSampleBuffer[BUFFER_SIZE];

    ReckoningStrategy* m_pReckoningStrategy;

    clone::VoidMemberConcealer<Value, Work> m_ValueWork;

    //! @endcond
};


//! @cond

template <typename ReckoningStrategy, typename SerializePolicy>
ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::ReckoningCloneElement(ReckoningStrategy* pReckoningStrategy)
  : ReckoningCloneElementBase(m_apToken, BUFFER_SIZE),
    m_pReckoningStrategy(pReckoningStrategy),
    m_ValueWork()
{
    PIA_COMPILE_ASSERT(BUFFER_SIZE > 0 && BUFFER_SIZE <= LONG_MAX);

    for (u32 i = 0; i < static_cast<u32>(BUFFER_SIZE); ++i)
    {
        m_aSampleBuffer[i].Init(this);
        m_apToken[i] = &m_aSampleBuffer[i];
    }
}


template <typename ReckoningStrategy, typename SerializePolicy>
nn::Result ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::SetReckoningStrategy(ReckoningStrategy* pReckoningStrategy)
{
    if (IsRegisteredWithCloneBase())
    {
        PIA_RETURN_RESULT(ResultInvalidState);
    }
    m_pReckoningStrategy = pReckoningStrategy;
    return ResultSuccess();
}


template <typename ReckoningStrategy, typename SerializePolicy>
nn::Result ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::SetValue(const Value& value, const SetValueArg* cpSetValueArg)
{
    bool isSend;
    nn::Result r = SetValueCore(&isSend);

    if (isSend)
    {
        const clone::ClockValue clock = GetProtocol()->GetClock();
        const SampleAccessor<Sample> accessor(GetSampleBuffer());

        bool isReliable = false;
        bool isSample = m_pReckoningStrategy->CheckSample(&isReliable, value, cpSetValueArg, accessor, clock, m_ValueWork.GetT2Ptr());
        if (isSample)
        {
            int idx;
            ReckoningSendToken* pToken = AssignTokenForSend(&idx, isReliable);
            PIA_ASSERT(pToken != NULL && idx == 0);

            SampleReckoningSendToken<Sample>* pSampleToken = static_cast<SampleReckoningSendToken<Sample>*>(pToken);
            m_pReckoningStrategy->MakeSample(pSampleToken->GetSamplePtr(), value, cpSetValueArg, accessor, clock, m_ValueWork.GetT2Ptr());
            m_pReckoningStrategy->OnUpdateSample(accessor, idx, m_ValueWork.GetT2Ptr());
        }
    }

    return r;
}


template <typename ReckoningStrategy, typename SerializePolicy>
void ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::Serialize(void* pBuffer, const ReckoningSendToken* cpToken)
{
    const SampleReckoningSendToken<Sample>* cpSampleToken = static_cast<const SampleReckoningSendToken<Sample>*>(cpToken);
    SerializePolicy::Serialize(pBuffer, cpSampleToken->GetSample());
}


template <typename ReckoningStrategy, typename SerializePolicy>
bool ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::EstimateValue(clone::ClockValue clock)
{
    const SampleAccessor<Sample> accessor(GetSampleBuffer());
    return m_pReckoningStrategy->Estimate(m_ValueWork.GetT1Ptr(), accessor, clock, m_ValueWork.GetT2Ptr());
}


template <typename ReckoningStrategy, typename SerializePolicy>
void ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::Deserialize(ReckoningSendToken* pToken, const void* cpData)
{
    SampleReckoningSendToken<Sample>* pSampleToken = static_cast<SampleReckoningSendToken<Sample>*>(pToken);
    SerializePolicy::Deserialize(pSampleToken->GetSamplePtr(), cpData);
}


template <typename ReckoningStrategy, typename SerializePolicy>
void ReckoningCloneElement<ReckoningStrategy, SerializePolicy>::OnUpdateSample(int idx)
{
    const SampleAccessor<Sample> accessor(GetSampleBuffer());
    m_pReckoningStrategy->OnUpdateSample(accessor, idx, m_ValueWork.GetT2Ptr());
}


//! @endcond
}
}
}  // end of namespace nn::pia::reckoning
