﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     reckoning_ReckoningCloneElementBase.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/reckoning/reckoning_definitions.h>
#include <pia/clone/clone_CloneElementBase.h>
#include <pia/common/common_OffsetList.h>


namespace nn
{
namespace pia
{
namespace clone
{
class CloneProtocol;
class IDataPacker;
}
}
}


namespace nn
{
namespace pia
{
namespace reckoning
{

/*!
@brief  This is the base class for managing the sending and receiving of data while using dead reckoning.

@date 2013-10-28  Initial version.
*/
class ReckoningCloneElementBase : public clone::CloneElementBase
{
public:
/*!
@brief  Destroys the object (destructor).
*/
    virtual ~ReckoningCloneElementBase();


/*!
@brief  Gets whether a valid value can be obtained using <tt>@ref ReckoningCloneElement::GetValue "GetValue"</tt>.
@return  Returns <tt>true</tt> if a valid value can be obtained using <tt>@ref ReckoningCloneElement::GetValue "GetValue"</tt>.
*/
    bool IsValidValue();


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


    //! @cond PRIVATE

public:
    virtual u16 GetType() const
    {
        return GetTypeStatic();
    }
    virtual bool IsSupportedClone(int cloneType) const;

    static u16 GetTypeStatic()
    {
        return TYPE_RECKONING;
    }
    static bool Receive(CloneElementBase* pElement, clone::IDataPacker* pAckPacker, const void* cpChunk, u16 chunkSize, Id elementId, StationIndex src, StationIndex myStationIndex, u32 dispatchCount);
    static size_t GetDataChunkHeaderSize();

private:
    void ReceiveUnreliable(const void* cpData, size_t size, StationIndex src, StationIndex setter, clone::ClockValue clock);
    void ReceiveReliable(const void* cpData, size_t size, StationIndex src, StationIndex setter, clone::ClockValue clock, u32 destBmp);
    void ReceiveAck(StationIndex src, StationIndex setter, clone::ClockValue clock);

public:
    virtual void ClearData();

    virtual void AddDest(StationIndex stationId);
    virtual void RemoveDest(StationIndex stationId);
    virtual void Adjust(StationIndex stationId);

    virtual void RequestInitialData();

public:
    class ReckoningSendToken : public ConversibleResendableSendToken
    {
    public:
        ReckoningSendToken()
          : ConversibleResendableSendToken()
        {
        }
        void Init(ReckoningCloneElementBase* pElement)
        {
            m_pElement = pElement;
        }

        virtual Type GetElementType() const
        {
            return TYPE_RECKONING;
        }
        virtual u32 GetDestBitmap() const
        {
            return m_DestBmp;
        }
        virtual size_t GetChunkSize() const;
        virtual void WriteChunk(void* pChunkBuffer) const;
        virtual CloneElementBase* GetElement()
        {
            return m_pElement;
        }

        Id GetId() const
        {
            return m_pElement->GetId();
        }
        size_t GetSize() const
        {
            return m_pElement->GetSize();
        }
        clone::ClockValue GetClock() const
        {
            return m_Clock;
        }
        void Serialize(void* pBuffer) const
        {
            m_pElement->Serialize(pBuffer, this);
        }

        StationIndex GetSetterStationIndex() const
        {
            return m_SetterStationIndex;
        }

        void SetDestBitmap(u32 destBmp)
        {
            m_DestBmp = destBmp;
        }
        void SetClock(clone::ClockValue clock)
        {
            m_Clock = clock;
        }
        void SetSetterStationIndex(StationIndex stationId)
        {
            m_SetterStationIndex = stationId;
        }

        void AddSendData()
        {
            m_pElement->AddSendData(this);
        }
        void CancelSendData()
        {
            m_pElement->CancelSendData(this);
        }

    private:
        ReckoningCloneElementBase* m_pElement;
        u32 m_DestBmp;
        clone::ClockValue m_Clock;
        StationIndex m_SetterStationIndex;
    };

protected:
    template <typename Sample>
    class SampleReckoningSendToken : public ReckoningSendToken
    {
    public:
        SampleReckoningSendToken()
          : ReckoningSendToken()
        {
        }
        Sample* GetSamplePtr()
        {
            return &m_Sample;
        }
        const Sample& GetSample() const
        {
            return m_Sample;
        }

    private:
        Sample m_Sample;
    };

protected:
    ReckoningCloneElementBase(ReckoningSendToken** papBuffer, u32 bufferSize);

protected:
    void UpdateValue();
    nn::Result SetValueCore(bool* pIsSend);
    ReckoningSendToken* AssignTokenForSend(int* pIdx, bool isResendable);

    virtual void Serialize(void* pBuffer, const ReckoningSendToken* cpToken) = 0;
    virtual void ClearValue() = 0;
    virtual bool EstimateValue(clone::ClockValue clock) = 0;

    virtual void Deserialize(ReckoningSendToken* pToken, const void* cpData) = 0;
    virtual void OnUpdateSample(int idx) = 0;

protected:
    class SampleBuffer
    {
    public:
        SampleBuffer(ReckoningSendToken** apBuffer, int size);

        void Clear();
        ReckoningSendToken* Get(int idx);
        const ReckoningSendToken* Get(int idx) const;
        bool IsHead(ReckoningSendToken* pToken) const
        {
            return (pToken == m_apBuffer[m_HeadIdx]);
        }
        ReckoningSendToken* Search(clone::ClockValue clock);
        ReckoningSendToken* Assign(int* pIdx, clone::ClockValue clock);
        int GetNum() const
        {
            return m_Num;
        }

        void AddDest(StationIndex stationId, bool IsResponsible);
        void RemoveDest(StationIndex stationId);
        void Adjust();

    private:
        ReckoningSendToken** m_apBuffer;
        int m_Size;
        int m_HeadIdx;
        int m_Num;
    };

    const SampleBuffer& GetSampleBuffer() const
    {
        return m_SampleBuffer;
    }

    //! @endcond

public:
/*!
@brief  Contains member functions used to access values in the sample buffer.
@tparam Sample  Specifies the sample type.
*/
    template <typename Sample>
    class SampleAccessor
    {
    public:
        //! @cond PRIVATE
        explicit SampleAccessor(const SampleBuffer& cBuffer)
          : m_cBuffer(cBuffer)
        {
        }
        //! @endcond


/*!
@brief  Gets the number of saved samples.
*/
        int GetNum() const
        {
            return m_cBuffer.GetNum();
        }


/*!
@brief  Gets the clock for when the sample was set.
@param[in] index  Specifies the index of the sample. A smaller value indicates a later sample.
@return  Returns the clock for when the sample was set.
*/
        clone::ClockValue GetClock(int index) const
        {
            PIA_ASSERT(index >= 0 && index < GetNum());
            return m_cBuffer.Get(index)->GetClock();
        }


/*!
@brief  Gets a sample.
@param[in] index  Specifies the index of the sample. A smaller value indicates a later sample.
@return  Returns the sample.
*/
        const Sample& GetSample(int index) const
        {
            PIA_ASSERT(index >= 0 && index < GetNum());
            return static_cast<const SampleReckoningSendToken<Sample>*>(m_cBuffer.Get(index))->GetSample();
        }

    private:
        const SampleBuffer& m_cBuffer;
    };


    //! @cond PRIVATE

private:
    SampleBuffer m_SampleBuffer;
    clone::ClockValue m_Clock;
    bool m_IsValidValue;
    StationIndex m_ResponsibleStationIndex;

    static Definition<ReckoningCloneElementBase> s_Definition;

public:
#if 0
    void Test(const CloneProtocol* cpProtocol);
#endif

public:
    template <int TEST_ID, typename Arg>
    static void Test(const ReckoningCloneElementBase& obj, Arg* pArg = NULL);

    //! @endcond
};
}
}
}  // end of namespace nn::pia::reckoning
