﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     reckoning_Simple3dReckoningStrategy.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/reckoning/reckoning_definitions.h>
#include <pia/reckoning/reckoning_Vector3f.h>
#include <pia/reckoning/reckoning_ReckoningCloneElementBase.h>
#include <pia/clone/clone_CloneBase.h>

namespace nn
{
namespace pia
{
namespace reckoning
{

/*!
@brief  The <tt>ReckoningStrategy</tt> class interpolates 3D vectors using linear evaluation functions.

@date 2013-10-29  Initial version.
*/
class Simple3dReckoningStrategy
{
public:
    typedef Vector3f Value;  //!< Specifies the type of value that is set and retrieved using <tt>SetValue</tt> and <tt>GetValue</tt>.

/*!
@brief  Contains a value and a stop flag as a set.
*/
    struct SampleValue
    {
        Value m_Value;
        bool m_IsStop;
    };
    typedef SampleValue Sample;  //!< Specifies the actual type of the sample that is sent and received.

    typedef bool SetValueArg;  //!< Specifies the type of data to pass to <tt>CheckSample</tt> and <tt>MakeSample</tt> when calling <tt>SetValue</tt>.

/*!
@brief  Stores the pre-calculated result for prediction.
*/
    struct Work {
        Value m_Param;
        Value m_ParamBase;
        nn::pia::clone::ClockValue m_BaseClock;
    };

    static const size_t BUFFER_SIZE = 10;  //!< Specifies the size of the buffer that holds samples.

    typedef ReckoningCloneElementBase::SampleAccessor<Sample> Accessor;  //!< Defines the accessor for the sample buffer.

/*!
@brief  Instantiates the object with default parameters (default constructor).
@param[in] threshold  Specifies the threshold value to use for determination.
*/
    Simple3dReckoningStrategy(f32 threshold);

/*!
@brief  Sets the threshold to use when determining whether to use the <tt><var>value</var></tt> set using <tt>SetValue</tt> as a sample.
@param[in] threshold  Specifies the threshold value to use for determination.
*/
    void SetThreshold(f32 threshold);

/*!
@brief  Gets the threshold to use when determining whether to use the <tt><var>value</var></tt> set using <tt>SetValue</tt> as a sample.
@return  Returns the threshold value to use for determination.
*/
    f32 GetThreshold()
    {
        return m_Threshold;
    }

/*!
@brief  Calculates the prediction.
@param[out] pValue  Outputs the prediction.
@param[in] accessor  Specifies the accessor for the sample buffer.
@param[in] clock  Specifies the current clock.
@param[in] cpWork  Specifies the pre-calculated result for calculating the prediction.
@return  Returns <tt>true</tt> and stores the prediction in <tt><var>pValue</var></tt> if successful.
*/
    bool Estimate(Value* pValue, const Accessor& accessor, nn::pia::clone::ClockValue clock, const Work* cpWork) const;

/*!
@brief  Determines whether to use the <tt><var>value</var></tt> set with <tt>SetValue</tt> as a sample.
@param[out] pIsReliable  Outputs whether reliable communication needs to be used when sending the sample.
@param[in] value  Specify the value that was set with <tt>SetValue</tt>.
@param[in] cpSetValueArg  Specify the value that was set in the second parameter of <tt>SetValue</tt>. This flag specifies whether <tt><var>value</var></tt> is stopped.
@param[in] accessor  Specifies the accessor for the sample buffer.
@param[in] clock  Specifies the current clock.
@param[in] cpWork  Specifies the pre-calculated result for calculating the prediction.
@return  Returns <tt>true</tt> if the value will be used as a sample.
*/
    bool CheckSample(bool* pIsReliable, const Value& value, const SetValueArg* cpSetValueArg, const Accessor& accessor, nn::pia::clone::ClockValue clock, const Work* cpWork) const;

/*!
@brief  Calculates the value to save as a sample.
@details  This is called when <tt>CheckSample</tt> determines the value is to be used as a sample and the buffer was successfully allocated.
@param[out] pSample  Outputs the sample.
@param[in] value  Specify the value that was set with <tt>SetValue</tt>.
@param[in] cpSetValueArg  Specify the value that was set in the second parameter of <tt>SetValue</tt>. This flag specifies whether <tt><var>value</var></tt> is stopped.
@param[in] accessor  Specifies the accessor for the sample buffer.
@param[in] clock  Specifies the current clock.
@param[in] cpWork  Specifies the pre-calculated result for calculating the prediction.
*/
    void MakeSample(Sample* pSample, const Value& value, const SetValueArg* cpSetValueArg, const Accessor& accessor, nn::pia::clone::ClockValue clock, const Work* cpWork) const;

/*!
@brief  This is called when a sample is added.
@details  This function updates the parameters used to calculate the prediction.
@param[in] accessor  Specifies the accessor for the sample buffer.
@param[in] index  Specifies the index of the sample.
@param[in] pWork  Specifies the pre-calculated result for calculating the prediction.
*/
    void OnUpdateSample(const Accessor& accessor, int index, Work* pWork) const;

/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;

protected:
private:
    f32 m_Threshold;
    s32 m_DelayClock;
};
}
}
}  // end of namespace nn::pia::reckoning
