﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     reckoning_Vector3f.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/reckoning/reckoning_definitions.h>


namespace nn
{
namespace pia
{
namespace reckoning
{

/*!
@brief  The <tt>Vector3f</tt> class provides functionality for the strategy used to handle 3D vectors.

@date 2013-10-29  Initial version.
*/
class Vector3f
{
public:
    f32 x, y, z;

/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    Vector3f()
      : x(0.f), y(0.f), z(0.f) {};

/*!
@brief  Instantiates the object and initializes the element with the specified arguments.
@param[in] x, y, z  Specifies the values to use for initialization.
*/
    Vector3f(const f32 x, const f32 y, const f32 z)
    {
        this->x = x;
        this->y = y;
        this->z = z;
    }

/*!
@brief  Sets the zero vector.
*/
    void MakeZero()
    {
        x = y = z = 0.f;
    }

/*!
@brief  Gets the square of the distance between two points.
@param[in] v  Specifies the vector to find the distance of.
@return  Returns the square of the distance.
*/
    f32 SquaredDistance(const Vector3f& v) const
    {
        f32 dx = this->x - v.x;
        f32 dy = this->y - v.y;
        f32 dz = this->z - v.z;
        return dx * dx + dy * dy + dz * dz;
    }

/*!
@brief  Adds the specified vector to the target vector.
@param[out] out  Stores the target vector.
@param[in] v  Specifies the vector to add.
*/
    static void Add(Vector3f* out, const Vector3f& v)
    {
        out->x += v.x;
        out->y += v.y;
        out->z += v.z;
        return;
    }

/*!
@brief  Multiplies the target vector by the specified scalar value.
@param[out] out  Stores the target vector.
@param[in] t  Specifies the scalar value.
*/
    static void Mul(Vector3f* out, const f32& t)
    {
        out->x *= t;
        out->y *= t;
        out->z *= t;
    }

/*!
@brief  Adds the specified vector and assigns the result.
@param[in] v  Specifies the vector to add.
@return  Returns the vector after addition.
*/
    Vector3f& operator+=(const Vector3f& v)
    {
        x += v.x;
        y += v.y;
        z += v.z;
        return *this;
    }

/*!
@brief  Divides by the specified scalar value and assigns the result.
@param[in] t  Specifies the value to divide by.
@return  Returns the vector after division.
*/
    Vector3f& operator/=(const f32& t)
    {
        x /= t;
        y /= t;
        z /= t;
        return *this;
    }

protected:
private:
};
}
}
}  // end of namespace nn::pia::reckoning
