﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     result.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


//This is the header for using the <tt>nn::Result</tt> type in Pia.

#pragma once

#include <pia/platform.h>
#include <pia/assertableResult.h>

/*!
@namespace  nn::pia
@brief  Represents the namespace for the Pia API.

@details  Most of the functions in the Pia API are not thread-safe.
Do not take it as a guarantee that a function will be thread-safe if nothing is written about it being either thread-safe or not thread-safe in its specific entry within the API Reference.
(We will note thread-safe functions in future revisions to the documentation.)

@date 2014-10-29  Added <tt>ResultNexInternalError</tt>.
@date 2014-09-12  Added <tt>ResultSessionConnectionIsLost</tt>, <tt>ResultInProgress</tt>, and <tt>ResultNotSet</tt>.
@date 2014-09-11  Consolidated all of the Pia Results into this header file.
@date 2014-07-16  Added <tt>ResultTemporaryUnavailable</tt>.
@date 2013-03-26  Added <tt>MODULE_TYPE_DB</tt>.
@date 2012-07-25  The <tt>NN_PIA_ENABLE_TRACE</tt> macro is now also set to a valid value in Development builds.
@date 2012-07-18  Added information about <tt>DESCRIPTION_BUFFER_IS_FULL</tt> and <tt>ResultBufferIsFull</tt>.
@date 2012-07-12  Added <tt>MODULE_TYPE_UTIL</tt>.
@date 2012-07-10  Deleted old <tt>Result</tt> definitions that had been kept for compatibility.
@date 2012-04-05  Merged generic <tt>Result</tt> values used in each module. They are now defined in the <tt>nn::pia</tt> namespace.
@date 2011-11-17  Excluding some exceptions, noted that the Pia API is not guaranteed to be thread-safe.
@date 2011-11-14  Expanded on the documentation of <tt>ModuleType</tt> and <tt>ErrorSeverity</tt>.
@date 2011-11-08  Created a page for the <tt>nn::pia</tt> namespace.

*/
namespace nn
{
namespace pia
{

/*!
@brief  Enumerates modules in the Pia library.
You get one of these values when you pass a <tt>nn::Result</tt> value to the <tt>@ref nn::pia::common::ErrorHandler::GetModuleTypeFromResult</tt> function. Use this value to determine the module in which an error occurred.
*/
enum ModuleType
{
    MODULE_TYPE_RESERVED = 0,    //!< Reserved value. The <tt>GetModuleTypeFromResult</tt> function never returns this value.
    MODULE_TYPE_PIA,             //!< Represents a generic <tt>Result</tt> value.
    MODULE_TYPE_COMMON,          //!< Represents a <tt>PiaCommon</tt> <tt>Result</tt> value.
    MODULE_TYPE_LOCAL,           //!< Represents a <tt>PiaLocal</tt> <tt>Result</tt> value.
    MODULE_TYPE_TRANSPORT,       //!< Represents a <tt>PiaTransport</tt> <tt>Result</tt> value.
    MODULE_TYPE_SYNC,            //!< Represents a <tt>PiaSync</tt> <tt>Result</tt> value.
    MODULE_TYPE_INET,            //!< Represents a <tt>PiaInet</tt> <tt>Result</tt> value.
    MODULE_TYPE_SESSION,         //!< Represents a <tt>PiaSession</tt> <tt>Result</tt> value.
    MODULE_TYPE_UTIL,            //!< Represents a <tt>PiaUtil</tt> <tt>Result</tt> value.
    MODULE_TYPE_SERVICE,         //!< Represents a <tt>PiaService</tt> <tt>Result</tt> value.
    MODULE_TYPE_DB,              //!< Represents a <tt>PiaDB</tt> <tt>Result</tt> value.
    MODULE_TYPE_CHAT,            //!< Represents a <tt>PiaChat</tt> <tt>Result</tt> value.
    MODULE_TYPE_CLONE,           //!< Represents a <tt>PiaClone</tt> <tt>Result</tt> value.
    MODULE_TYPE_RECKONING,       //!< Represents a <tt>PiaReckonng</tt> <tt>Result</tt> value.
    MODULE_TYPE_BRAIN,           //!< Represents a <tt>PiaBrain</tt> <tt>Result</tt> value.
    MODULE_TYPE_UNKNOWN = 0x0f,  //!< Indicates that a Pia module is not the source.
    MODULE_TYPE_MAX = 0x10       //!< The maximum value. The application is not required to use this value.
};

/*!
@cond PRIVATE
*/
const u32 RESULT_MODULE_TYPE_FIELD = 4;  //<tt>ModuleType</tt> is represented by 4 bits.
const u32 RESULT_DESCRIPTION_FIELD = 5;  //<tt>Description</tt> can be represented by 5 bits.
const u32 RESULT_DESCRIPTION_MIN = 0;    //The minimum value for <tt>Description</tt>.
const u32 RESULT_DESCRIPTION_MAX = 31;   //The maximum value for <tt>Description</tt>.

const u32 NN_PIA_ERROR_BIT           = MODULE_TYPE_PIA       << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_COMMON_ERROR_BIT    = MODULE_TYPE_COMMON    << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_LOCAL_ERROR_BIT     = MODULE_TYPE_LOCAL     << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_TRANSPORT_ERROR_BIT = MODULE_TYPE_TRANSPORT << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_SYNC_ERROR_BIT      = MODULE_TYPE_SYNC      << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_INET_ERROR_BIT      = MODULE_TYPE_INET      << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_SESSION_ERROR_BIT   = MODULE_TYPE_SESSION   << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_SESSION_EX_ERROR_BIT= MODULE_TYPE_UTIL      << RESULT_DESCRIPTION_FIELD;  // (Note) Out of the <tt>description</tt> assigned to <tt>session</tt>, so <tt>util</tt> is being applied intentionally.
const u32 NN_PIA_CLONE_ERROR_BIT     = MODULE_TYPE_CLONE     << RESULT_DESCRIPTION_FIELD;
const u32 NN_PIA_BRAIN_ERROR_BIT     = MODULE_TYPE_BRAIN     << RESULT_DESCRIPTION_FIELD;

/*!
@brief  Enumerates general-purpose Pia <tt>Result</tt> values.
*/
enum Description
{
    DESCRIPTION_ALLOCATION_FAILED                         = NN_PIA_ERROR_BIT | 0,            //!< Indicates a failure to allocate memory or some other resource.
    DESCRIPTION_ALREADY_INITIALIZED                       = NN_PIA_ERROR_BIT | 1,            //!< Initialization is already complete.
    DESCRIPTION_BROKEN_DATA                               = NN_PIA_ERROR_BIT | 2,            //!< An error indicating that data received over this connection is corrupt, or may have been tampered with.
    DESCRIPTION_BUFFER_SHORTAGE                           = NN_PIA_ERROR_BIT | 3,            //!< The buffer passed in the argument is too small.
    DESCRIPTION_CANCELLED                                 = NN_PIA_ERROR_BIT | 4,            //!< The asynchronous process received a cancellation request.
    DESCRIPTION_NETWORK_CONNECTION_IS_LOST                = NN_PIA_ERROR_BIT | 5,            //!< The connection was terminated.
    DESCRIPTION_INVALID_ARGUMENT                          = NN_PIA_ERROR_BIT | 6,            //!< An invalid argument was passed to an API function.
    DESCRIPTION_INVALID_STATE                             = NN_PIA_ERROR_BIT | 7,            //!< An API function was called at the wrong time. Indicates that the library was not in the correct state to call the function that returned this <tt>Result</tt> value.
    DESCRIPTION_NO_DATA                                   = NN_PIA_ERROR_BIT | 8,            //!< No data.
    DESCRIPTION_NOT_FOUND                                 = NN_PIA_ERROR_BIT | 9,            //!< Requested item not found.
    DESCRIPTION_NOT_IMPLEMENTED                           = NN_PIA_ERROR_BIT | 10,           //!< Not implemented.
    DESCRIPTION_NOT_INITIALIZED                           = NN_PIA_ERROR_BIT | 11,           //!< Not initialized.
    DESCRIPTION_BUFFER_IS_FULL                            = NN_PIA_ERROR_BIT | 12,           //!< The action failed because the buffer required to complete the action temporarily ran out of space. It also may be caused by network congestion. It may succeed if you try again after some time has passed.
    DESCRIPTION_TIME_OUT                                  = NN_PIA_ERROR_BIT | 13,           //!< The asynchronous process timed out.
    DESCRIPTION_ALREADY_EXISTS                            = NN_PIA_ERROR_BIT | 14,           //!< Already exists.
    DESCRIPTION_CONTAINER_IS_FULL                         = NN_PIA_ERROR_BIT | 15,           //!< The container is already full.
    DESCRIPTION_TEMPORARY_UNAVAILABLE                     = NN_PIA_ERROR_BIT | 16,           //!< In a temporarily unusable state.
    DESCRIPTION_IN_PROGRESS                               = NN_PIA_ERROR_BIT | 17,           //!< The indicated process is already in progress.
    DESCRIPTION_NOT_SET                                   = NN_PIA_ERROR_BIT | 18,           //!< Values that are supposed to be set in advance are not set.

    DESCRIPTION_MEMORY_LEAK                               = NN_PIA_COMMON_ERROR_BIT | 0,     //!< There may be a memory leak in the Pia library.

    DESCRIPTION_NOT_FOUND_NETWORK                         = NN_PIA_LOCAL_ERROR_BIT | 4,      //!< Indicates that a network could not be found.
    DESCRIPTION_INVALID_NODE                              = NN_PIA_LOCAL_ERROR_BIT | 5,      //!< Indicates that the target node does not exist on the network.
    DESCRIPTION_NOTHING_PACKET                            = NN_PIA_LOCAL_ERROR_BIT | 6,      //!< Indicates that there are no packets to receive.
    DESCRIPTION_NEGLIGIBLE_FAULT                          = NN_PIA_LOCAL_ERROR_BIT | 7,      //!< Indicates a failure that can be ignored.
    DESCRIPTION_INVALID_CONNECTION                        = NN_PIA_LOCAL_ERROR_BIT | 8,      //!< Indicates that the station is not connected to the network.
    DESCRIPTION_HOST_MIGRATION_FAILURE                    = NN_PIA_LOCAL_ERROR_BIT | 9,      //!< Local communication host migration failed.
    DESCRIPTION_DIFFERENT_VERSION                         = NN_PIA_LOCAL_ERROR_BIT | 10,     //!< The local network version or the application version is different.
    DESCRIPTION_ERROR_OCCURRED                            = NN_PIA_LOCAL_ERROR_BIT | 11,     //!< An error occurred.
    DESCRIPTION_INVALID_NETWORK                           = NN_PIA_LOCAL_ERROR_BIT | 12,     //!< Cannot connect to this network.
    DESCRIPTION_NETWORK_IS_FULL                           = NN_PIA_LOCAL_ERROR_BIT | 13,     //!< The network is full.

    DESCRIPTION_CONNECTION_FAILED                         = NN_PIA_TRANSPORT_ERROR_BIT | 1,  //!< Connection succeeded.
    DESCRIPTION_CREATE_STATION_FAILED                     = NN_PIA_TRANSPORT_ERROR_BIT | 2,  //!< Indicates that creating the <tt>Station</tt> instance failed.
    DESCRIPTION_VERSION_MISMATCHED                        = NN_PIA_TRANSPORT_ERROR_BIT | 3,  //!< Indicates that communication protocol version numbers differ.
    DESCRIPTION_IS_NOT_IN_COMMUNICATION                   = NN_PIA_TRANSPORT_ERROR_BIT | 4,  //!< Indicates that stations are not communicating.
    DESCRIPTION_TABLE_IS_FULL                             = NN_PIA_TRANSPORT_ERROR_BIT | 5,  //!< Indicates that no more entries can be added because the table is full.

    DESCRIPTION_SYNC_BASE                                 = NN_PIA_SYNC_ERROR_BIT | 0,       //!< Avoid making the width wide. Do not use.
    DESCRIPTION_SYNC_NOT_INITIALIZED                      = NN_PIA_SYNC_ERROR_BIT | 1,       //!< Not initialized.
    DESCRIPTION_SYNC_INVALID_STATE                        = NN_PIA_SYNC_ERROR_BIT | 2,       //!< Not allowed to execute the function in this state.
    DESCRIPTION_SYNC_INVALID_ARGUMENT                     = NN_PIA_SYNC_ERROR_BIT | 3,       //!< There was a problem with an argument.
    DESCRIPTION_SYNC_DATA_IS_NOT_ARRIVED_YET              = NN_PIA_SYNC_ERROR_BIT | 4,       //!< Data has not yet arrived from all stations.
    DESCRIPTION_SYNC_OUT_OF_RESOURCE                      = NN_PIA_SYNC_ERROR_BIT | 5,       //!< Insufficient resources.
    DESCRIPTION_SYNC_SYNCHRONIZATION_IS_LOST              = NN_PIA_SYNC_ERROR_BIT | 6,       //!< Synchronization could not be maintained, and the process lapsed into an error state.
    DESCRIPTION_SYNC_FAILED_TO_SYNCHRONIZE                = NN_PIA_SYNC_ERROR_BIT | 7,       //!< Could not synchronize with the station host's start of synchronization.
    DESCRIPTION_SYNC_INVALID_DEVICE                       = NN_PIA_SYNC_ERROR_BIT | 8,       //!< Failed because you did not specify a target device during initialization.
    DESCRIPTION_SYNC_TO_BE_IMPLEMENTED                    = NN_PIA_SYNC_ERROR_BIT | 9,       //!< Not implemented.
    DESCRIPTION_SYNC_DATA_IS_NOT_SET                      = NN_PIA_SYNC_ERROR_BIT | 10,      //!< The data to send has not been specified.
    DESCRIPTION_SYNC_MAX                                  = NN_PIA_SYNC_ERROR_BIT | RESULT_DESCRIPTION_MAX,

    DESCRIPTION_NAT_CHECK_FAILED                          = NN_PIA_INET_ERROR_BIT | 1,       //!< The NAT check failed.
    DESCRIPTION_IN_USED                                   = NN_PIA_INET_ERROR_BIT | 2,
    DESCRIPTION_DNS_FAILED                                = NN_PIA_INET_ERROR_BIT | 3,       //!< The DNS name resolution prior to the NAT check failed.
    DESCRIPTION_NEX_INTERNAL_ERROR                        = NN_PIA_INET_ERROR_BIT | 4,       //!< Indicates that the game server returned an unexpected result code.

    DESCRIPTION_REQUEST_DENIED                            = NN_PIA_SESSION_ERROR_BIT | 0,    //!< Request denied.
    DESCRIPTION_STATION_CONNECTION_FAILED                 = NN_PIA_SESSION_ERROR_BIT | 1,    //!< Failed to connect stations.
    DESCRIPTION_MESSAGE_NOT_HANDLED                       = NN_PIA_SESSION_ERROR_BIT | 2,    //!< Could not process the received <tt>ProtocolMessageOld</tt> object.
    DESCRIPTION_FAILED_CREATE_PROTOCOL                    = NN_PIA_SESSION_ERROR_BIT | 3,    //!< System protocol generation failed.
    DESCRIPTION_FAILED_STARTUP_PROTOCOL                   = NN_PIA_SESSION_ERROR_BIT | 4,    //!< The system protocol start process failed.
    DESCRIPTION_MESH_IS_FULL                              = NN_PIA_SESSION_ERROR_BIT | 5,    //!< The mesh is full.
    DESCRIPTION_INVALID_MESSAGE                           = NN_PIA_SESSION_ERROR_BIT | 6,    //!< Handled an invalid system message.
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_UNKNOWN   = NN_PIA_SESSION_ERROR_BIT | 7,    //!< Failed to connect stations. The NAT type is unknown.
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_EIM       = NN_PIA_SESSION_ERROR_BIT | 8,    //!< Failed to connect stations. The NAT type is <tt>EIM</tt>. (Removed.)
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_EDM       = NN_PIA_SESSION_ERROR_BIT | 9,    //!< Failed to connect stations. The NAT type is <tt>EDM</tt>. (Removed.)
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EIM_REMOTE_EIM = NN_PIA_SESSION_ERROR_BIT | 10,   //!< Failed to connect stations. The NAT type for both local and remote stations is EIM.
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EIM_REMOTE_EDM = NN_PIA_SESSION_ERROR_BIT | 11,   //!< Failed to connect stations. The NAT type was EIM for the local station and EDM for the remote station.
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EDM_REMOTE_EIM = NN_PIA_SESSION_ERROR_BIT | 12,   //!< Failed to connect stations. The NAT type was EDM for the local station and EIM for the remote station.
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EDM_REMOTE_EDM = NN_PIA_SESSION_ERROR_BIT | 13,   //!< Failed to connect stations. The NAT type for both local and remote stations is EDM.
    DESCRIPTION_RELAY_FAILED_NO_CANDIDATE                 = NN_PIA_SESSION_ERROR_BIT | 14,   //!< The relay connection failed. (There was no relay candidate.)
    DESCRIPTION_RELAY_FAILED_RTT_LIMIT                    = NN_PIA_SESSION_ERROR_BIT | 15,   //!< The relay connection failed. (The RTT limit was exceeded.)
    DESCRIPTION_RELAY_FAILED_RELAY_NUM_LIMIT              = NN_PIA_SESSION_ERROR_BIT | 16,   //!< The relay connection failed. (The limit on the number of relays was exceeded.)
    DESCRIPTION_RELAY_FAILED_UNKNOWN                      = NN_PIA_SESSION_ERROR_BIT | 17,   //!< The relay connection failed. (Details unknown.)
    DESCRIPTION_NAT_TRAVERSAL_REQUEST_TIMEOUT             = NN_PIA_SESSION_ERROR_BIT | 18,   //!< Failed to connect stations. The NAT traversal request timed out.
    DESCRIPTION_SESSION_IS_NOT_FOUND                      = NN_PIA_SESSION_ERROR_BIT | 19,   //!< The session was not found.
    DESCRIPTION_MATCHMAKE_SESSION_IS_FULL                 = NN_PIA_SESSION_ERROR_BIT | 20,   //!< The matchmaking session is full.
    DESCRIPTION_DENIED_BY_PARTICIPANTS                    = NN_PIA_SESSION_ERROR_BIT | 21,   //!< The local station is blacklisted by a user in the session.
    DESCRIPTION_PARTICIPANT_IN_BLACKLIST                  = NN_PIA_SESSION_ERROR_BIT | 22,   //!< A user on the local station's blacklist is in the session.
    DESCRIPTION_GAME_SERVER_MAINTENANCE                   = NN_PIA_SESSION_ERROR_BIT | 23,   //!< The server is under maintenance.
    DESCRIPTION_MESH_CONNECTION_IS_LOST                   = NN_PIA_SESSION_ERROR_BIT | 24,   //!< The mesh is disconnected.
    DESCRIPTION_SESSION_USER_PASSWORD_UNMATCH             = NN_PIA_SESSION_ERROR_BIT | 25,   //!< Attempted to join a session set with a user password, but the user password did not match.
    DESCRIPTION_SESSION_SYSTEM_PASSWORD_UNMATCH           = NN_PIA_SESSION_ERROR_BIT | 26,   //!< Attempted to join a session set with a system password, but the system password did not match.
    DESCRIPTION_SESSION_IS_CLOSED                         = NN_PIA_SESSION_ERROR_BIT | 27,   //!< Attempted to join a session that is closed.
    DESCRIPTION_COMPANION_STATION_IS_OFFLINE              = NN_PIA_SESSION_ERROR_BIT | 28,   //!< Attempted to get a companion station to join in the joint session that was not logged in to the server.
    DESCRIPTION_HOST_IS_NOT_FRIEND                        = NN_PIA_SESSION_ERROR_BIT | 29,   //!< Attempted to join a session with a host that is not a friend.
    DESCRIPTION_SESSION_CONNECTION_IS_LOST                = NN_PIA_SESSION_ERROR_BIT | 30,   //!< The session is disconnected.

    DESCRIPTION_COMPANION_STATION_IS_LEFT                 = NN_PIA_SESSION_EX_ERROR_BIT | 0, //!< Attempted to specify companion stations for the joint session that were disconnected.
    DESCRIPTION_COMPANION_STATION_DENIED                  = NN_PIA_SESSION_EX_ERROR_BIT | 1, //!< Attempted to specify companion stations for the joint session that rejected join-in.
    DESCRIPTION_SESSION_MIGRATION_FAILED                  = NN_PIA_SESSION_EX_ERROR_BIT | 2, //!< Processes failed that were necessary for session transitions in joint sessions.
    DESCRIPTION_GAME_SERVER_PROCESS_ABORTED               = NN_PIA_SESSION_EX_ERROR_BIT | 3, //!< The game server process aborted.
    DESCRIPTION_SESSION_WRONG_STATE                       = NN_PIA_SESSION_EX_ERROR_BIT | 4, //!< The state of the joined session was irregular.
    DESCRIPTION_SESSION_INCONSISTENT_STATE                = NN_PIA_SESSION_EX_ERROR_BIT | 5, //!< The session is in an inconsistent state.
    DESCRIPTION_SESSION_INVALID_STATE                     = NN_PIA_SESSION_EX_ERROR_BIT | 6, //!< The <tt>Session</tt> API is being called in the wrong way.

    DESCRIPTION_INVALID_TIMING                            = NN_PIA_CLONE_ERROR_BIT | 1
};
/** @endcond */

/*!
@class  nn::pia::ResultAllocationFailed
@brief  Indicates a failure to allocate memory or other resources. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultAllocationFailed,
    Result::LEVEL_USAGE,
    Result::SUMMARY_OUT_OF_RESOURCE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_ALLOCATION_FAILED);

/*!
@class  nn::pia::ResultAlreadyInitialized
@brief  Indicates that initialization is already complete. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultAlreadyInitialized,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_ALREADY_INITIALIZED);

/*!
@class  nn::pia::ResultBufferShortage
@brief  Indicates that the buffer passed as an argument is too small. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultBufferShortage,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INVALID_ARGUMENT,
    Result::MODULE_NN_PIA,
    DESCRIPTION_BUFFER_SHORTAGE);

/*!
@class  nn::pia::ResultBrokenData
@brief  Indicates that data coming in through the communication line is corrupted or may have been tampered with.
*/
PIA_DEFINE_RESULT(
    ResultBrokenData,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_BROKEN_DATA);

/*!
@class  nn::pia::ResultCancelled
@brief  Indicates that the asynchronous process received a cancellation request.
*/
PIA_DEFINE_RESULT(
    ResultCancelled,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_CANCELLED,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CANCELLED);

/*!
@class  nn::pia::ResultNetworkConnectionIsLost
@brief  Indicates that the connection was already terminated.
*/
PIA_DEFINE_RESULT(
    ResultNetworkConnectionIsLost,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NETWORK_CONNECTION_IS_LOST);

/*!
@class  nn::pia::ResultInvalidArgument
@brief  Indicates that an invalid argument was passed to an API function. One case where this result would be returned is when a NULL pointer was passed to a function. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultInvalidArgument,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INVALID_ARGUMENT,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_ARGUMENT);


/*!
@class  nn::pia::ResultInvalidState
@brief  Indicates that the API function was called at the wrong time. Indicates that the Pia library was not in the correct state to call the function that returned this <tt>Result</tt> value.
*/
PIA_DEFINE_RESULT(
    ResultInvalidState,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_STATE);

/*!
@class  nn::pia::ResultNoData
@brief  Indicates that the data was not found.
*/
PIA_DEFINE_RESULT(
    ResultNoData,
    Result::LEVEL_TEMPORARY,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NO_DATA);

/*!
@class  nn::pia::ResultNotFound
@brief  Indicates that the requested item was not found.
*/
PIA_DEFINE_RESULT(
    ResultNotFound,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_NOT_FOUND,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOT_FOUND);

/*!
@class  nn::pia::ResultNotImplemented
@brief  Indicates that a feature is not implemented. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultNotImplemented,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOT_IMPLEMENTED);


/*!
@class  nn::pia::ResultNotInitialized
@brief  Indicates that the object or module that called the function is not initialized. Applications must be implemented to ensure this <tt>Result</tt> value is never returned.
*/
PIA_DEFINE_RESULT(
    ResultNotInitialized,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOT_INITIALIZED);

/*!
@class  nn::pia::ResultBufferIsFull
@brief  Indicates that the action failed because the buffer required for the action was temporarily full. This may be caused by network congestion. It may succeed if you try again after some time has passed.
*/
PIA_DEFINE_RESULT(
    ResultBufferIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_OUT_OF_RESOURCE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_BUFFER_IS_FULL);

/*!
@class  nn::pia::ResultTimeOut
@brief  Indicates that an asynchronous process timed out.
*/
PIA_DEFINE_RESULT(
    ResultTimeOut,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_STATUS_CHANGED,
    Result::MODULE_NN_PIA,
    DESCRIPTION_TIME_OUT);

/*!
@class  nn::pia::ResultAlreadyExists
@brief  Indicates that the object already exists.
*/
PIA_DEFINE_RESULT(
    ResultAlreadyExists,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_ALREADY_EXISTS);

/*!
@class  nn::pia::ResultContainerIsFull
@brief  Indicates that the container is full.
*/
PIA_DEFINE_RESULT(
    ResultContainerIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CONTAINER_IS_FULL);

/*!
@class  nn::pia::ResultTemporaryUnavailable
@brief  Result code that indicates that the API is temporarily unusable.
*/
PIA_DEFINE_RESULT(
    ResultTemporaryUnavailable,
    Result::LEVEL_TEMPORARY,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_TEMPORARY_UNAVAILABLE);

/*!
@class  nn::pia::ResultInProgress
@brief  Result code indicating that the specified process is already in progress.
*/
PIA_DEFINE_RESULT(
    ResultInProgress,
    Result::LEVEL_STATUS,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_IN_PROGRESS);

/*!
@class  nn::pia::ResultNotSet
@brief  Result code indicating that values that are supposed to be set in advance are not set.
*/
PIA_DEFINE_RESULT(
    ResultNotSet,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOT_SET);


/*!
@class  nn::pia::ResultMemoryLeak
@brief  Result code indicating that there may be a memory leak in the Pia library.
*/
PIA_DEFINE_RESULT(
    ResultMemoryLeak,
    Result::LEVEL_USAGE,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_MEMORY_LEAK);


/*!
@class  nn::pia::ResultNotFoundNetwork
@brief  Indicates that the network was not found.
*/
PIA_DEFINE_RESULT(
    ResultNotFoundNetwork,
    Result::LEVEL_TEMPORARY,
    Result::SUMMARY_NOT_FOUND,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOT_FOUND_NETWORK);

/*!
@class  nn::pia::ResultInvalidNode
@brief  Indicates that the node being sent to could not be found on the network.
*/
PIA_DEFINE_RESULT(
    ResultInvalidNode,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_NOT_FOUND,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_NODE);

/*!
@class  nn::pia::ResultNothingPacket
@brief  Indicates that packets expected to be received could not be found.
*/
PIA_DEFINE_RESULT(
    ResultNothingPacket,
    Result::LEVEL_TEMPORARY,
    Result::SUMMARY_NOT_FOUND,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NOTHING_PACKET);

/*!
@class  nn::pia::ResultNegligibleFault
@brief  Indicates an error that can be ignored.
*/
PIA_DEFINE_RESULT(
    ResultNegligibleFault,
    Result::LEVEL_TEMPORARY,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NEGLIGIBLE_FAULT);

/*!
@class  nn::pia::ResultInvalidConnection
@brief  Indicates that the connection state is invalid or that wireless mode is off.
*/
PIA_DEFINE_RESULT(
    ResultInvalidConnection,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_CONNECTION);

/*!
@class  nn::pia::ResultHostMigrationFailure
@brief  Indicates that local network host migration failed.
Hosts leaving the network must be handled appropriately or communication cannot continue.
*/
PIA_DEFINE_RESULT(
    ResultHostMigrationFailure,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_STATUS_CHANGED,
    Result::MODULE_NN_PIA,
    DESCRIPTION_HOST_MIGRATION_FAILURE);

/*!
@class  nn::pia::ResultDifferentVersion
@brief  Indicates that the local network version or the application version of the target network is different from that of the local station.
*/
PIA_DEFINE_RESULT(
    ResultDifferentVersion,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_DIFFERENT_VERSION);

/*!
@class  nn::pia::ResultErrorOccurred
@brief  Indicates that an internal error occurred.
*/
PIA_DEFINE_RESULT(
    ResultErrorOccurred,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_ERROR_OCCURRED);

/*!
@class  nn::pia::ResultInvalidNetwork
@brief  Indicates that a the system cannot connect to this network.
*/
PIA_DEFINE_RESULT(
    ResultInvalidNetwork,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_NETWORK);

/*!
@class  nn::pia::ResultNetworkIsFull
@brief  The network is full.
*/
PIA_DEFINE_RESULT(
    ResultNetworkIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NETWORK_IS_FULL);


/*!
@class  nn::pia::ResultConnectionFailed
@brief  Indicates that connecting to the network failed for some reason. (System error.)
This result never gets back to the application.
*/
PIA_DEFINE_RESULT(
    ResultConnectionFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CONNECTION_FAILED);


/*!
@class  nn::pia::ResultCreateStationFailed
@brief  Indicates that the new <tt>Station</tt> could not be created. (System error.)
This result never gets back to the application.
*/
PIA_DEFINE_RESULT(
    ResultCreateStationFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INTERNAL,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CREATE_STATION_FAILED);


/*!
@class  nn::pia::ResultIncompatibleFormat
@brief  Represents incompatible formats.
*/
PIA_DEFINE_RESULT(
    ResultIncompatibleFormat,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_VERSION_MISMATCHED);

/*!
@class  nn::pia::ResultNotInCommunication
@brief  Not currently in communication.
*/
PIA_DEFINE_RESULT(
    ResultNotInCommunication,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_IS_NOT_IN_COMMUNICATION);

/*!
@class  nn::pia::ResultTableIsFull
@brief  Indicates that no more entries can be added because the table is full. (System error.)
This result never gets back to the application.
*/
PIA_DEFINE_RESULT(
    ResultTableIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_TABLE_IS_FULL);


/*!
@class  nn::pia::ResultDataIsNotArrivedYet
@brief  Data has not yet arrived from all stations.

@details  This error is generated when a problem occurs in data synchronization, because of a degenerated communication state, for example.
If the function returns this error, have the application stop the progress of the game and wait for the next picture frame.
Note that this error is not necessarily encountered by all stations at the same time, or for the same number of times.
*/
PIA_DEFINE_RESULT(
    ResultDataIsNotArrivedYet,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_NOTHING_HAPPENED,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SYNC_DATA_IS_NOT_ARRIVED_YET);


/*!
@class  nn::pia::ResultDataIsNotSet
@brief  The data to send has not been specified.
*/
PIA_DEFINE_RESULT(
    ResultDataIsNotSet,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_NOTHING_HAPPENED,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SYNC_DATA_IS_NOT_SET);


/*!
@class  nn::pia::ResultNatCheckFailed
@brief  Indicates that the NAT check failed.
*/
PIA_DEFINE_RESULT(
    ResultNatCheckFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_CHECK_FAILED);

/*!
@cond PRIVATE
@class  nn::pia::ResultInUse
@brief  Indicates that the object is already being used.
*/
PIA_DEFINE_RESULT(
    ResultInUse,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_IN_USED);
//!@endcond


/*!
@class  nn::pia::ResultDnsFailed
@brief  Indicates that DNS resolution failed.
*/
PIA_DEFINE_RESULT(
    ResultDnsFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_DNS_FAILED);


/*!
@class  nn::pia::ResultNexInternalError
@brief  Indicates that an error occurred in NEX.
*/
PIA_DEFINE_RESULT(
    ResultNexInternalError,
    Result::LEVEL_FATAL,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NEX_INTERNAL_ERROR);


/*!
@class  nn::pia::ResultJoinRequestDenied
@brief  The join request was rejected by the session host.
*/
PIA_DEFINE_RESULT(
    ResultJoinRequestDenied,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_REQUEST_DENIED);


/*!
@class  nn::pia::ResultStationConnectionFailed
@brief  Failed to connect stations.
*/
PIA_DEFINE_RESULT(
    ResultStationConnectionFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_STATION_CONNECTION_FAILED);


/*!
@class  nn::pia::ResultMessageNotHandled
@brief  Could not process the received <tt>ProtocolMessageOld</tt> object.
*/
PIA_DEFINE_RESULT(
    ResultMessageNotHandled,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_MESSAGE_NOT_HANDLED);


/*!
@class  nn::pia::ResultFailedCreateProtocol
@brief  Failed to generate the system protocol.
*/
PIA_DEFINE_RESULT(
    ResultFailedCreateProtocol,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_FAILED_CREATE_PROTOCOL);


/*!
@class  nn::pia::ResultFailedStartupProtocol
@brief  Failed to start the system protocol.
*/
PIA_DEFINE_RESULT(
    ResultFailedStartupProtocol,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_FAILED_STARTUP_PROTOCOL);


/*!
@class  nn::pia::ResultMeshIsFull
@brief  Could not join the requested mesh because it was full.
*/
PIA_DEFINE_RESULT(
    ResultMeshIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_MESH_IS_FULL);


/*!
@class  nn::pia::ResultInvalidSystemMessage
@brief  An invalid message was returned as a response to a join request.
*/
PIA_DEFINE_RESULT(
    ResultInvalidSystemMessage,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_MESSAGE);

/*!
@class  nn::pia::ResultStationConnectionNatTraversalFailedUnknown
@brief  NAT traversal between stations failed. The NAT type is unknown.
*/
PIA_DEFINE_RESULT(
    ResultStationConnectionNatTraversalFailedUnknown,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_UNKNOWN);

/*!
@class  nn::pia::ResultStationConnectionNatTraversalFailedEim
@brief  NAT traversal between stations failed. The NAT type is <tt>EIM</tt>. (Removed.)
*/
PIA_DEFINE_RESULT(
    ResultStationConnectionNatTraversalFailedEim,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_EIM);

/*!
@class  nn::pia::ResultStationConnectionNatTraversalFailedEdm
@brief  NAT traversal between stations failed. The NAT type is <tt>EDM</tt>. (Removed.)
*/
PIA_DEFINE_RESULT(
    ResultStationConnectionNatTraversalFailedEdm,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_CONNECTION_NAT_TRAVERSAL_FAILED_EDM);

/*!
@class  nn::pia::ResultNatTraversalFailedBothEim
@brief  NAT traversal between stations failed. The NAT type for both local and remote stations is EIM.
*/
PIA_DEFINE_RESULT(
    ResultNatTraversalFailedBothEim,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EIM_REMOTE_EIM);

/*!
@class  nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm
@brief  NAT traversal between stations failed. The NAT type was EIM for the local station and EDM for the remote station.
*/
PIA_DEFINE_RESULT(
    ResultNatTraversalFailedLocalEimRemoteEdm,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EIM_REMOTE_EDM);

/*!
@class  nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim
@brief  NAT traversal between stations failed. The NAT type was EDM for the local station and EIM for the remote station.
*/
PIA_DEFINE_RESULT(
    ResultNatTraversalFailedLocalEdmRemoteEim,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EDM_REMOTE_EIM);

/*!
@class  nn::pia::ResultNatTraversalFailedBothEdm
@brief  NAT traversal between stations failed. The NAT type for both local and remote stations is EDM.
*/
PIA_DEFINE_RESULT(
    ResultNatTraversalFailedBothEdm,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_TRAVERSAL_FAILED_LOCAL_EDM_REMOTE_EDM);


/*!
@class  nn::pia::ResultRelayFailedNoCandidate
@brief  The relay connection failed. (There was no relay candidate.)
*/
PIA_DEFINE_RESULT(
    ResultRelayFailedNoCandidate,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_RELAY_FAILED_NO_CANDIDATE);

/*!
@class  nn::pia::ResultRelayFailedRttLimit
@brief  The relay connection failed. (The RTT limit was exceeded.)
*/
PIA_DEFINE_RESULT(
    ResultRelayFailedRttLimit,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_RELAY_FAILED_RTT_LIMIT);
/*!
@class  nn::pia::ResultRelayFailedRelayNumLimit
@brief  The relay connection failed. (The limit on the number of relays was exceeded.)
*/
PIA_DEFINE_RESULT(
    ResultRelayFailedRelayNumLimit,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_RELAY_FAILED_RELAY_NUM_LIMIT);
/*!
@class  nn::pia::ResultRelayFailedUnknown
@brief  The relay connection failed. (Details unknown.)
*/
PIA_DEFINE_RESULT(
    ResultRelayFailedUnknown,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_RELAY_FAILED_UNKNOWN);

/*!
@class  nn::pia::ResultNatTraversalRequestTimeout
@brief  NAT traversal between stations failed. The NAT traversal request timed out.
*/
PIA_DEFINE_RESULT(
    ResultNatTraversalRequestTimeout,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_NAT_TRAVERSAL_REQUEST_TIMEOUT);

/*!
@class  nn::pia::ResultSessionIsNotFound
@brief  The session no longer exists.
*/
PIA_DEFINE_RESULT(
    ResultSessionIsNotFound,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_IS_NOT_FOUND);

/*!
@class  nn::pia::ResultMatchmakeSessionIsFull
@brief  Attempted to join a matchmaking session that is full.
*/
PIA_DEFINE_RESULT(
    ResultMatchmakeSessionIsFull,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_MATCHMAKE_SESSION_IS_FULL);

/*!
@class  nn::pia::ResultDeniedByParticipants
@brief  The local station is blacklisted by a user in the session.
*/
PIA_DEFINE_RESULT(
    ResultDeniedByParticipants,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_DENIED_BY_PARTICIPANTS);

/*!
@class  nn::pia::ResultParticipantInBlackList
@brief  A user on the local station's blacklist is in the session.
*/
PIA_DEFINE_RESULT(
    ResultParticipantInBlackList,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_PARTICIPANT_IN_BLACKLIST);

/*!
@class  nn::pia::ResultGameServerMaintenance
@brief  The game server is down for maintenance.
*/
PIA_DEFINE_RESULT(
    ResultGameServerMaintenance,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_GAME_SERVER_MAINTENANCE);

/*!
@class  nn::pia::ResultSessionUserPasswordUnmatch
@brief  Attempted to join a session set with a user password, but the user password did not match.
*/
PIA_DEFINE_RESULT(
    ResultSessionUserPasswordUnmatch,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_USER_PASSWORD_UNMATCH);

/*!
@class  nn::pia::ResultSessionSystemPasswordUnmatch
@brief  Attempted to join a session set with a system password, but the system password did not match.
*/
PIA_DEFINE_RESULT(
    ResultSessionSystemPasswordUnmatch,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_SYSTEM_PASSWORD_UNMATCH);

/*!
@class  nn::pia::ResultMeshConnectionIsLost
@brief  The mesh is disconnected.
*/
PIA_DEFINE_RESULT(
    ResultMeshConnectionIsLost,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_MESH_CONNECTION_IS_LOST);

/*!
@class  nn::pia::ResultSessionIsClosed
@brief  Attempted to join a session that is closed.
*/
PIA_DEFINE_RESULT(
    ResultSessionIsClosed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_IS_CLOSED);

/*!
@class  nn::pia::ResultCompanionStationIsOffline
@brief  (For joint sessions) Attempted to add a station that was not logged in to the server.
*/
PIA_DEFINE_RESULT(
    ResultCompanionStationIsOffline,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_COMPANION_STATION_IS_OFFLINE);

/*!
@class  nn::pia::ResultHostIsNotFriend
@brief  Attempted to join a session with a host that is not a friend.
*/
PIA_DEFINE_RESULT(
    ResultHostIsNotFriend,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_HOST_IS_NOT_FRIEND);


/*!
@class  nn::pia::ResultSessionConnectionIsLost
@brief  The session was disconnected.
*/
PIA_DEFINE_RESULT(
    ResultSessionConnectionIsLost,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_CONNECTION_IS_LOST);

/*!
@class  nn::pia::ResultCompanionStationIsLeft
@brief  Attempted to specify companion stations for the joint session that were disconnected.
*/
PIA_DEFINE_RESULT(
    ResultCompanionStationIsLeft,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_COMPANION_STATION_IS_LEFT);

/*!
@class  nn::pia::ResultCompanionStationDenied
@brief  Attempted to specify companion stations for the joint session that rejected join-in.
*/
PIA_DEFINE_RESULT(
    ResultCompanionStationDenied,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_COMPANION_STATION_DENIED);

/*!
@class  nn::pia::ResultSessionMigrationFailed
@brief  Processes failed that were necessary for session transitions in joint sessions.
*/
PIA_DEFINE_RESULT(
    ResultSessionMigrationFailed,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_MIGRATION_FAILED);

/*!
@class  nn::pia::ResultGameServerProcessAborted
@brief  The game server process aborted.
*/
PIA_DEFINE_RESULT(
    ResultGameServerProcessAborted,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_GAME_SERVER_PROCESS_ABORTED);

/*!
@class  nn::pia::ResultSessionWrongState
@brief  The state of the joined session was irregular.
*/
PIA_DEFINE_RESULT(
    ResultSessionWrongState,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_WRONG_STATE);

/*!
@class  nn::pia::ResultSessionInvalidState
@brief  The <tt>Session</tt> API is being called in the wrong way.
*/
PIA_DEFINE_RESULT(
    ResultSessionInvalidState,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_INVALID_STATE);


/*!
@class  nn::pia::ResultSessionInconsistentState
@brief  There is an inconsistent state in the session.
*/
PIA_DEFINE_RESULT(
    ResultSessionInconsistentState,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_SESSION_INCONSISTENT_STATE);

/*!
@class  nn::pia::ResultInvalidTiming
@brief  Indicates that the action could not execute because of a timing problem.
*/
PIA_DEFINE_RESULT(
    ResultInvalidTiming,
    Result::LEVEL_PERMANENT,
    Result::SUMMARY_INVALID_STATE,
    Result::MODULE_NN_PIA,
    DESCRIPTION_INVALID_TIMING);


#define PIA_RETURN_RESULT(result)                                            \
    {                                                                        \
        PIA_ASSERTMSG(result::IsAssertEnable() == false, "Result:" #result); \
        return result();                                                     \
    }
}
}  // end of namespace nn::pia
