﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     session_Api.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_definitions.h>


namespace nn
{
namespace pia
{
namespace session
{


/*!
@name  Module Initialization and Finalization
@{
*/
/*!
@brief  Initializes the <tt>PiaSession</tt> module.
Execute the <tt>nn::pia::common::Initialize</tt> function in advance.

@details  Memory that the <tt>session</tt> module dynamically allocates and uses is allocated from memory passed by the <tt>@ref nn::pia::common::Initialize</tt> function.

@return  Returns a <tt>Result</tt> value indicating success if called at the right time. Your application's implementation must ensure that this function does not return any errors.
@retval ResultInvalidState  Indicates that the <tt>common</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyInitialized  Indicates that the function has already been executed and initialization has completed. Programming error. Fix your program so that this error is not returned.

*/
nn::Result Initialize();

/*!
@brief  Declares the start of setup.

@details  Call before executing initialization processing beneath the <tt>session</tt> module. In other words, call before executing the <tt>@ref Session::CreateInstance</tt> function.

Because this function does not run correctly if competing with another module's setup processing, perform initialization in the <tt>local</tt> module, and promptly call the <tt>@ref EndSetup</tt> function to complete setup.

@return  Returns a <tt>Result</tt> value indicating success if called at the right time. Your application's implementation must ensure that this function does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>session::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called again during setup. Programming error. Fix your program so that this error is not returned.



*/
nn::Result BeginSetup(void);

/// @cond PRIVATE
/*!
@brief  Checks whether setup is currently underway.

@details  The application should not call this function. It is provided for internal use. <br>

@return  Returns <tt>true</tt> between calls to the <tt>@ref session::BeginSetup</tt> and <tt>@ref session::EndSetup</tt> functions.
*/
bool IsDuringSetup(void);
/// @endcond

/*!
@brief  Declares the end of setup.

@details  Call after executing initialization processing beneath the <tt>session</tt> module. In other words, call after executing the <tt>@ref Session::CreateInstance</tt> function.

Calling this function determines the amount of memory used by singletons and other objects in the <tt>session</tt> module, and optimizes memory management in the module.

Because this function does not run correctly if competing with another module's setup processing, perform initialization in the <tt>local</tt> module, and promptly call the <tt>@ref EndSetup</tt> function to complete setup.

@return  Returns a <tt>Result</tt> value indicating success if called at the right time. Your application's implementation must ensure that this function does not return any errors.

@retval ResultNotInitialized  Indicates that the <tt>session::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.




*/
nn::Result EndSetup(void);

/*!
@brief  Returns whether the <tt>PiaSession</tt> module is initialized.

@return  Returns <tt>true</tt> if initialized, and <tt>false</tt> otherwise.
*/
bool IsInitialized(void);

/*!
@brief  Finalizes the <tt>PiaSession</tt> module.
Call this function before calling the <tt>@ref nn::pia::common::Finalize</tt> function.

@details  Processing of all <tt>PiaSession</tt> modules must be finalized before executing this function.
Calling this function has no effect if the <tt>@ref nn::pia::session::Initialize</tt> function has not been called.
*/
void Finalize(void);
/*!
@}
*/
}
}
}  // end of namespace nn::pia::session
