﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: session_Mesh.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_RootObject.h>
#include <pia/common/common_CallContext.h>
#include <pia/common/common_SignatureSetting.h>

#include <pia/session/session_definitions.h>
#include <pia/session/session_types.h>

#include <pia/transport/transport_Station.h>
#include <pia/transport/transport_StationLocationContainer.h>
#include <pia/transport/transport_StationConnectionInfo.h>

namespace nn
{
namespace nex
{
class StationURL;
}
}

namespace nn
{
namespace pia
{
namespace common
{
struct CryptoSetting;
class MonitoringDataSender;
}
}
}

namespace nn
{
namespace pia
{
namespace transport
{
class StationProtocol;
class ConnectStationJob;
class DisconnectStationJob;
class ProcessConnectionRequestJob;
class MissingStationHandler;
class NetworkFactory;
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
class BandwidthCheckerProtocol;
#endif
}
}
}

namespace nn
{
namespace pia
{
namespace session
{

class CreateMeshJob;
class JoinMeshJob;
class LeaveMeshJob;
class ProcessJoinRequestJob;
class ProcessUpdateMeshJob;
class DestroyMeshJob;
class ProcessDestroyMeshJob;
class ProcessHostMigrationJob;
class LeaveWithHostMigrationJob;
class SignatureSettingStorage;
#if NN_PIA_EXPERIMENT_RELAY_CONNECTION_TOPOLOGY
class RelayRouteManageJob;
#endif
class KickoutManageJob;
class MeshProtocol;
class SyncClockProtocol;
class MeshEventListener;

static const u16 PIA_SESSION_STATION_MAX_DEFAULT = 4;

/*!
  @class  nn::pia::session::Mesh
  @brief  Contains member functions that manage connections with other devices.

  @details  This class implements the singleton pattern. Only a single instance can be created.

  @attention  This class is for directly manipulating meshes. Do not use the <tt>Mesh</tt> class when using sessions. Do not use the <tt>Session</tt> class when directly manipulating meshes.

  @details  Some of the internal processes of the <tt>Mesh</tt> class are registered as jobs in <tt>@ref nn::pia::common::Scheduler</tt>. To use the <tt>Mesh</tt> features, you must make periodic calls to
            <tt>@ref nn::pia::common::Scheduler::Dispatch</tt>.

  @date 2015-05-20  Added the <tt>Mesh::GetUplinkBitRate</tt> function.
  @date 2015-05-20  Added the <tt>bitRateCheckMode</tt> member to the <tt>Setting</tt> structure to enable configuration of the bit rate detection feature.
  @date 2015-05-20  Added the <tt>uplinkBitRateLowerLimit</tt> and <tt>bitRateCheckPacketSize</tt> members to the <tt>StartupSetting</tt> structure to enable configuration of the bit rate detection feature.
  @date 2015-05-11  Provided the <tt>StartupSetting</tt> structure to consolidate the arguments when calling the <tt>Startup()</tt> function.
  @date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance()</tt> when an instance already exists.
  @date 2014-09-17  Changed the <tt>Result</tt> returned when a <tt>StationId</tt> that has already been kicked out is specified for <tt>Mesh::KickoutStation</tt>.
  @if CTR_DOC
  @date 2014-05-30  Changed the notations for the cautions when local communication for <tt>Mesh::Startup</tt> and host migration are enabled.
  @endif
  @date 2014-05-15  Added the <tt>Mesh::CheckConnection</tt> function that returns a <tt>Result</tt> indicating whether <tt>Mesh</tt> is in a communicable state.
  @if CTR_DOC
  @date 2014-05-02  Added a note to the description of <tt>Mesh::Startup(const transport::Station::PlayerName*, const common::CryptoSetting*, const common::SignatureSetting&, const transport::Station::IdentificationToken*)</tt> about the affect of the UdsNode connection monitoring feature when the host migration feature is enabled for local communication.
  @date 2014-04-17  Added a note to the description of <tt>Mesh::Startup(u32, u32, const transport::Station::PlayerName*, const common::CryptoSetting*, const common::SignatureSetting&, const transport::Station::IdentificationToken*)</tt> about the affect of the UdsNode connection monitoring feature when the host migration feature is enabled for local communication.
  @endif
  @date 2014-02-10  Added a note to the descriptions of both <tt>Mesh::RelayEmulation</tt> and <tt>Mesh::ResetRelayEmulation</tt> about when the functions can be called.
  @date 2014-02-03  Organized the results that are returned when <tt>Mesh::JoinMeshAsync</tt> or asynchronous processes fail, and increased the number of reasons for distinguishable asynchronous process failures.
  @date 2013-12-18  Removed the synchronous processing version of <tt>Mesh::CreateMesh</tt>.
  @date 2013-12-11  Fixed a bug where the asynchronous processing result was different from the actual reason for failure when <tt>Mesh::JoinMesh</tt> failed because its request was rejected due to the internal state of the host's library.
  @date 2013-12-04  Added an overload with arguments to <tt>Mesh::Startup</tt> to pass a structure for the encryption settings.
  @if CAFE_DOC
  @date 2013-10-07  <tt>ResultInvalidArgument</tt> was being returned when a player name with an undefined language code was passed to <tt>Mesh::Startup</tt>, but this was changed to not return an error when the language code is undefined.
  @endif
  @date 2013-08-09  Fixed a bug where the internal state would become inconsistent if <tt>Mesh::DestroyMesh</tt> was called while joining a mesh, or if a station called <tt>Mesh::KickoutStation</tt> on itself.
  @date 2013-08-09  Corrected the descriptions of the return values of <tt>Mesh::GetConnectionRoute</tt> and <tt>Mesh::RelayEmulation</tt>. (No changes were made to conditions that result in an error or <tt>Result</tt> types.)
  @date 2013-08-06  Fixed a bug where meshes would disconnect during host migration if the single-candidate host migration feature was enabled while a relay type network topology was also enabled.
  @date 2013-07-26  Removed <tt>Mesh::GetIdentificationToken</tt>.
  @date 2013-07-10  Changed the specifications of <tt>Mesh::Startup</tt> about the timeout used with the <tt>Mesh::JoinMesh</tt> process so that it cannot be changed by applications.
  @date 2013-07-10  Removed some <tt>Result</tt> types that can be returned when asynchronous processing of <tt>Mesh::JoinMesh</tt> fails.
  @date 2013-06-25  Added the <tt>Mesh::RelayEmulation</tt> and <tt>Mesh::ResetRelayEmulation</tt> functions for debugging the relay feature.
  @date 2013-06-25  Added an overload to <tt>Mesh::CreateInstance</tt> allowing you to enable or disable the use of debugging of the relay feature.
  @date 2013-06-18  Changed the specification to check whether lower modules can use host migration when <tt>Mesh::Startup</tt> is called while host migration is enabled.
  @date 2013-06-13  Fixed a bug where calls to <tt>Mesh::LeaveMesh</tt> would fail if they were called at the same time a station was being kicked from the mesh using <tt>Mesh::KickoutStation</tt>.
  @date 2013-05-28  Changed the default timeout for <tt>Mesh::JoinMesh</tt> processing (defined by the <tt>NN_PIA_SESSION_JOIN_TIMEOUT_DEFAULT</tt> constant) from 20,000 milliseconds to 30,000 milliseconds.
  @date 2013-05-15  Added the <tt>Mesh::KickoutStation</tt> function for kicking out an arbitrary station from the mesh.
  @date 2013-05-15  Added the <tt>Mesh::DisconnectReason</tt> enumerated type to represent reasons why a station is unable to connect to a mesh, and the <tt>Mesh::GetDisconnectReason</tt> function to get this reason.
  @date 2013-04-24  Added the <tt>Mesh::GetConnectionRoute</tt> function for debugging relaying.
  @date 2013-04-19  Fixed a bug where the state could become inconsistent if a host migration started while another station was joining the mesh, under certain timing and other conditions.
  @date 2013-04-08  Added available settings to use as the network topology for the meshes built using the <tt>Mesh</tt> class.
  @date 2013-04-08  Added the following functions for changing settings when relaying is enabled: <tt>Mesh::SetMaxNumOfRelayRoutesAssignedToStation</tt> and <tt>Mesh::SetMaxRtt</tt>.
  @date 2013-03-29  Changed the <tt>Mesh::Startup</tt> function to set the player name.
  @date 2013-03-14  Added an asynchronous version of the <tt>Mesh::CreateMesh</tt> function.
  @date 2013-01-08  Fixed a low-frequency bug where other clients could get into an inconsistent state if a client left a mesh and then tried to join the same one again.
  @date 2012-12-20  Fixed a bug where an inconsistent internal state could occur if the other client disconnected at certain times while a new client was joining the mesh.
  @date 2012-12-19  Reviewed the value of <tt>nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</tt> (the default setting for the interval at which to send keep-alives), and changed it from 2000 to 1000.
  @date 2012-12-18  Fixed a bug where clients already in a mesh could be dropped if a new client failed to join it.
  @date 2012-12-18  Added a <tt>Result</tt> type that can be returned when asynchronous processing of a <tt>Mesh::JoinMesh</tt> fails.
  @date 2012-12-10  Fixed a timing bug where host migration could fail when it was expected to succeed.
  @date 2012-12-10  Fixed a bug where local host migration would fail for everyone unless it succeeded for all the remaining clients.
  @date 2012-12-10  Fixed a bug where some clients were highly likely to fail to join a mesh when multiple clients tried to join at the same time.
  @date 2012-11-30  Changed the specifications to configure the packet-signing feature when calling the <tt>Mesh::Startup</tt> function. The parameter specification was changed.
  @date 2012-11-26  Fixed a bug where the internal status of the mesh host could become inconsistent when a client failed to join the session, depending on the progress status of other processes.
  @date 2012-11-13  Fixed a bug where the mesh client could never detect an abnormal connection status if communication with another client was lost.
  @date 2012-10-29  Categorized member functions to improve readability of the API reference.
  @date 2012-08-22  Fixed a bug in the link between the server matchmaking process for Internet communication and the mesh participation process.
  @date 2012-08-22  As a result of this fix, communication with <tt>PiaMesh</tt> in earlier versions is no longer compatible.
  @date 2012-08-20  Fixed a bug where disconnect events with the same <tt>StationIndex</tt> were duplicated in events sent in a callback for a mesh state change notification.
  @date 2012-08-07  Fixed a bug where continuing host migration by using local communication frequently failed.
  @date 2012-07-18  Made minor amendments to the descriptions of the <tt>Mesh::Startup</tt>, <tt>Mesh::Cleanup</tt>, and related functions.
  @date 2012-07-10  Improved the speed and accuracy of host migration failure assessments during host migration.
  @date 2012-07-05  Added a <tt>Mesh::Startup</tt> overload that enables you to specify any settings, such as the timeout period for the mesh participation process.
  @date 2012-07-03  Added a <tt>Result</tt> type that can be returned when asynchronous processing of a <tt>Mesh::JoinMesh</tt> fails.
  @date 2012-07-02  Fixed a bug where when running a <tt>Mesh::DestroyMesh</tt>, the mesh destroy message sometimes did not reach the client.
  @date 2012-06-11  Fixed a bug where a station could not join a session using the <tt>Mesh::JoinMesh</tt> function if another station left the mesh while the function was executing.
  @date 2012-05-30  Fixed a bug where clients did not receive notifications when the mesh host called the <tt>Mesh::DestroyMesh</tt> function.
  @date 2012-05-30  Fixed a bug where disconnecting from a peer at nearly the same time the <tt>Mesh::LeaveMesh</tt> function starts would result in a timeout wait even though the peer had already disconnected.
  @date 2012-05-22  Implemented host migration. To enable or disable the host migration feature, use the <tt>Mesh::Setting</tt> structure, which is passed when you create a <tt>Mesh</tt> instance.
  @date 2012-05-22  Added the <tt>Mesh::LeaveMeshWithHostMigration</tt> function.
  @date 2012-05-22  Added host migration-related events to the events that are notified with the state change notification callback for the mesh.
  @date 2012-05-22  Changed the name of the event enumerated type from <tt>Event</tt> to <tt>EventType</tt> for the state change notification callback for the mesh.
  @date 2012-05-15  Added a function to get and set the synchronization pulse interval of the synchronization clock.
  @date 2012-05-11  Added the <tt>GetTime</tt> function, which can get the elapsed time (called the synchronization clock) from the time that the mesh was created.
  @date 2012-05-01  Added the <tt>Mesh::Startup</tt> function.
  @date 2012-05-01  Implemented the shared identification token feature.
  @date 2012-04-24  Added a feature to notify, in a callback, the changes to mesh information when joining or leaving stations.
  @date 2012-04-16  Changed the maximum specification value of connectible stations from <tt>10</tt> to <tt>12</tt>.
  @date 2012-04-12  Listed <tt>const</tt> information for functions in which the internal state does not change.
  @date 2012-04-12  Changed <tt>Mesh::DestroyMesh</tt> so that it is an asynchronous process.
  @date 2012-04-12  Added an enumerated type to indicate the <tt>Mesh</tt> state and a function for getting the state.
  @date 2012-04-12  Implemented a keep-alive feature.
  @date 2012-04-06  Initial version.
*/
class Mesh : public nn::pia::common::RootObject
{
    friend class CreateMeshJob;
    friend class JoinMeshJob;
    friend class LeaveMeshJob;
    friend class ProcessJoinRequestJob;
    friend class ProcessUpdateMeshJob;
    friend class DestroyMeshJob;
    friend class ProcessDestroyMeshJob;
    friend class ProcessHostMigrationJob;
    friend class LeaveWithHostMigrationJob;
    friend class KickoutManageJob;
    friend class MeshProtocol;
    friend class MeshEventListener;

public:
/*!
    @brief  This structure stores settings passed to <tt>@ref nn::pia::session::Mesh::CreateInstance</tt>.
*/
    struct Setting
    {
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        Setting()
            : bitRateCheckMode(session::BitRateCheck_Disable)
        {
        }
#endif
        nn::pia::transport::NetworkFactory* pNetworkFactory;  //!<  Pointer to an instance of the concrete class that inherits <tt>@ref nn::pia::transport::NetworkFactory</tt> (factory pattern). It is used only when creating a <tt>Mesh</tt> instance.
        nn::pia::session::NetworkTopology networkTopology;    //!<  Specifies the network topology setting for the mesh.
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        nn::pia::session::BitRateCheckMode bitRateCheckMode;  //!<  Settings for the bitrate detection feature.
#endif
    };

/*!
    @brief  This structure stores settings passed to <tt>@ref nn::pia::session::Mesh::Startup</tt>.
*/
    struct StartupSetting
    {
        StartupSetting()
            : maxSilenceTime(NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT),
            keepAliveSendingInterval(NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT),
            bUsingHostMigration(true),
            pPlayerName(NULL),
            pCryptoSetting(NULL),
            signatureSetting(nn::pia::common::SignatureSetting::NO_SIGNATURE),
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
            uplinkBitRateLowerLimit(NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT),
            bitRateCheckPacketSize(0),
            bitRateMeasuringSpan(1000),
            isSkipBitRateCheck(false),
#endif
            pToken(NULL)
        {
        }

        u32 maxSilenceTime;                                                 //!<  Specifies the maximum allowable time without communication (in milliseconds). The default value is <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT</tt>. The maximum value that can be set is <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_MAX</tt>, the minimum value is  <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_MIN</tt>.
        u32 keepAliveSendingInterval;                                       //!<  Specifies the send interval for keep-alive (in milliseconds). The default value is <tt>session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</tt>.
        bool bUsingHostMigration;                                           //!<  Sets the use of the host migration feature. The default is <tt>true</tt> (enabled).
        const nn::pia::transport::Station::PlayerName* pPlayerName;         //!<  Specifies the name and language code associated with the local station. Other members of the mesh access this information.
        const nn::pia::common::CryptoSetting* pCryptoSetting;               //!<  Specifies a pointer that points to a structure for configuring encryption. No encryption is configured if <tt>NULL</tt> is specified. The default is <tt>NULL</tt>.
        nn::pia::common::SignatureSetting signatureSetting;                 //!<  Specifies a properly configured instance of the class used to specify packet-signature settings. The default value for this parameter is <tt>unsigned</tt>.
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        s32 uplinkBitRateLowerLimit;   //!<  Specifies the minimum rate required by the application when using the bitrate detection feature. The value is in bits per second. The default value is <tt>session::NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT</tt>.
        size_t bitRateCheckPacketSize; //!<  Specifies the IP packet size used for testing when using the bitrate detection feature. The value is in bytes. The expected value is the average size of IP packet sent and received by the application. It must be a multiple of 4. The default is <tt>0</tt>.
        s32 bitRateMeasuringSpan;      //!<  Sets the time to spend measuring when the bit rate detection feature is used. The value is in milliseconds. The default is <tt>1000 msec</tt>. Changing this value from the default is not recommended.
        bool isSkipBitRateCheck;       //!<  Set to skip detection processes when using bitrate detection. Specify <tt>true</tt> to skip. The default is <tt>false</tt>.
#endif
        const nn::pia::transport::Station::IdentificationToken* pToken;     //!<  Specifies a pointer to the identification token indicating the local device. If <tt>NULL</tt> is specified, an identification token with all data set to <tt>NULL</tt> is used. The default is <tt>NULL</tt>.

    };

/*!
    @brief  (For debugging.) This structure stores the debug feature settings that are passed to <tt>@ref nn::pia::session::Mesh::SetDebugSetting</tt>.
*/
    struct DebugSetting
    {
        DebugSetting(bool argIsEnableRelayEmulation = false, bool argIsDisableCheckMaxSilenceTime = false)
          : isEnableRelayEmulation(argIsEnableRelayEmulation),
            isDisableCheckMaxSilenceTime(argIsDisableCheckMaxSilenceTime)
        {
        }
        bool isEnableRelayEmulation;  //!<  Settings for the connection route emulation feature for debugging.
        bool isDisableCheckMaxSilenceTime; //!<  Does not check the range set for the maximum time without communication.
    };

/*!
    @cond  PRIVATE
    @brief  Enumerates mesh states.
*/
    enum State
    {
        STATE_DISCONNECTED,  //!<  Specifies that the local device is not connected to other devices and communication is not possible.
        STATE_CONNECTED      //!<  Specifies that the local device is connected to other devices and communication is possible.
    };
    //! @endcond

/*!
    @brief  This enumerated type represents reasons why a <tt>Mesh</tt> object cannot communicate.

    @details  This information can be retrieved using the <tt>Mesh::GetDisconnectReason</tt> function.
*/
    enum DisconnectReason
    {
        UNKNOWN_REASON = 0,    //!<  Specifies that the reason for the disconnection is unknown. One way to get this value is by calling the <tt>Mesh::Cleanup</tt> function during mesh communication (without destroying or leaving the mesh).
        NOT_YET_COMMUNICATED,  //!<  Specifies that no communication has been performed. You get this value if you never build or join a mesh after calling the <tt>Mesh::Startup</tt> function.
        OPERATION_OF_OWN,      //!<  Specifies that the disconnection was caused by the local station. You get this value when the local station left the mesh by calling the <tt>Mesh::DestroyMesh</tt> or <tt>Mesh::LeaveMesh</tt> function, or the equivalent.
        OPERATION_OF_OTHER,    //!<  Specifies that the disconnection was caused by another station in the mesh. One way to get this value is when the local station is a client, and the host destroys the mesh.
        KICKOUT_BY_USER,       //!<  Specifies that the local station was kicked out from the mesh because of a call to the <tt>Mesh::KickoutStation</tt> function.
        KICKOUT_BY_SYSTEM,     //!<  Specifies that the disconnection from the mesh was caused by the library. One way to get this value is when relaying is enabled, and the library disconnects you from the mesh because it restricts the number of relay requests.
        INCONSISTENT_INFO,     //!<  Specifies that the disconnection was caused by a desync. (The mesh information was no longer consistent between participants.) One way to get this value is when the station counts do not match.
        MIGRATION_FAIL,        //!<  Specifies that the disconnection was caused by a failed host migration.
        EXTERNAL_FACTOR,       //!<  Specifies that the disconnection was caused by something other than the <tt>Mesh</tt> object. One way to get this value is when the network connection is lost.
        MIGRATION_FATAL_ERROR  //!<  Specifies that a fatal error occurred during host migration. This error can occur when the library detects that a mesh has broken up.
    };


/*!
    @brief  Enumerates connection path emulation settings. (For debugging.)

    @details  This is used as the third argument to <tt>Mesh::RelayEmulation</tt>.
*/
    enum EmulationType
    {
        EMULATION_TYPE_NONE = 0,  //!<  Specifies no emulation.
        EMULATION_TYPE_RELAY      //!<  Specifies the relay connection. (No direct connection is possible.)
    };


/*!
    @brief  This enumerated type is passed as a parameter to the callback that sends notifications about changes in the mesh state.

    @details  It represents the possible changes to the mesh state.
*/
    enum EventType
    {
        EVENT_JOIN,           //!<  Indicates that a station has joined.
        EVENT_LEAVE,          //!<  A station has left.
        EVENT_HOST_CHANGED,   //!<  Specifies that the mesh host has migrated. (Only occurs when host migration is enabled.)
        EVENT_MIGRATION_FAIL  //!<  Indicates that host migration failed. (Only occurs when host migration is enabled.)
    };

/*!
    @brief  This callback function type is called when the mesh state changes.

    @details  <b>Parameters</b>
            <ul>
            <li> eventType: Enumerated type indicating what the state change was. For more information, see the <tt>@ref EventType</tt> enumerated type.</li>
            <li> id: The <tt>StationId</tt> of the station that changed its state.</li>
            </ul>
*/
    typedef void (*MeshEventCallback)(EventType eventType, StationId id);


/*!
    @brief  Declares the type of the callback function for determining whether to approve join requests from clients.

    @details  If the callback returns <tt>true</tt>, the join request is approved. If it returns <tt>false</tt>, the join request is denied.

            <b>Parameters</b>
            <ul>
            <li> pRequesterToken: Pointer to the identification token for the station that sent the join request.</li>
            </ul>
*/
    typedef bool (*JoiningMeshApprovalCallback)(const nn::pia::transport::Station::IdentificationToken* pRequesterToken);

    /// @cond PRIVATE
/*!
    @brief  The callback function for adjusting to any secondary host candidate position when using multiple candidate host migration.

    @details  The secondary host candidate position can be adjusted if any host candidate position (in ascending order) is returned for the <tt><var>targetId</var></tt> parameter.
*/
    typedef u8 (*CheckHostCandidateRankingCallback)(StationIndex targetId, bool isHost);
    /// @endcond

    /// @cond PRIVATE
    typedef u32 (*GetSessionIdCallback)();
    /// @endcond

/*!
     @name  Singleton Instance Control
     @{
*/

/*!
    @brief  Creates an instance of <tt>Mesh</tt>.

    @details  This function must be executed between calls to the <tt>@ref nn::pia::session::BeginSetup</tt> and <tt>@ref nn::pia::session::EndSetup</tt> functions.

    @details  If the relay emulation feature for debugging has been set by the <tt>@ref SetDebugSetting</tt> function,
                and the network topology is not set to <tt>NetworkTopology_RelayMesh</tt>, the call to this function fails.

    @attention  Do not call <tt>@ref SetDebugSetting</tt> in retail ROMs.

    @param[in] setting  Specifies various settings.

    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. You must make sure that the implementation of this function in your application does not return any errors.
            If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultNotInitialized  Indicates that the <tt>session</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. This error is returned when the function is executed outside calls to the <tt>@ref nn::pia::session::BeginSetup</tt> and <tt>@ref nn::pia::session::EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the settings specified in the argument are invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
*/
    static nn::Result CreateInstance(const nn::pia::session::Mesh::Setting& setting);


/*!
    @brief  Destroys the instance of <tt>Mesh</tt>.
*/
    static void DestroyInstance();


/*!
    @brief  Gets an instance of the <tt>Mesh</tt> object.

    @details  The <tt>Mesh</tt> class is a singleton. Only one instance of this class is automatically created.
            The pointer to this single instance can be obtained by calling this function.

    @return  Returns a pointer to an instance of the <tt>Mesh</tt> object.
*/
    static Mesh* GetInstance()
    {
        return s_pMesh;
    }


/*!
    @brief  (For debugging.) Sets the debugging features for <tt>Mesh</tt>.

    @details  This function can be called even when a <tt>Mesh</tt> instance does not yet exist. Call it before calling the <tt>@ref CreateInstance</tt> function to create a <tt>Mesh</tt> instance.

                If the <tt>isEnableRelayEmulation</tt> member of the <tt>DebugSetting</tt> structure is set to <tt>true</tt>, the <tt>Session</tt> will be configured to use the relay emulation feature for debugging.
                When the session is configured to use the relay emulation feature, the <tt>@ref CreateInstance</tt> function will not succeed unless the network topology is set to <tt>NetworkTopology_RelayMesh</tt>.

                If this function is not called, all related debugging features are disabled.

    @attention  Do not call this function in retail ROMs.

    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. You must make sure that the implementation of this function in your application does not return any errors.
            If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Returns when the <tt>Mesh</tt> instance has already been created. Programming error. Fix your program so that this error is not returned.
*/
    static nn::Result SetDebugSetting(const DebugSetting& debugSetting);

/*!
     end of name Singleton Instance Control
     @}
*/

/*!
     @name  Finish Initialization
     @{
*/

/*!
    @brief  Starts the <tt>Mesh</tt>.

    @details  The use of host migration is set by the <tt>bUsingHostMigration</tt> member of the <tt>StartupSetting</tt> structure. Set to <tt>true</tt> to enable. The default is <tt>true</tt>.
             Specify the name and language code associated with the local station in the <tt>pPlayerName</tt> member of <tt>StartupSetting</tt>. The name must be no longer than 16 characters (not including the terminating null character).
    @if CAFE_DOC
             Set the language code to the language used by the application. Use one of the values defined in <tt>nn::fp::Language</tt>.
             The name and language code you specify are shared with all members of the mesh. When the local station joins a mesh, all mesh participants are added to Player History. When another
             client joins the mesh, the new participant's name and language code are looked up and automatically
             added to Player History using the <tt>nn::fp::AddRecentPlayRecord()</tt> function from within the library.
             For more information about Player History, see the <i>Pia Programming Manual</i> for the Friend Presence library
             and the <i>Pia API Reference</i> for the <tt>nn::fp::AddRecentPlayRecord()</tt> function.
    @endif
    @if CTR_DOC
             Specify a language code if your application determines the languages of other regions. You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage()</tt> function if you want to use the language setting of the system.
    @endif

            You can configure the communication encryption feature in the <tt>pCryptoSetting</tt> member of the <tt>StartupSetting</tt> structure.
            Communication is not encrypted if <tt>NULL</tt> is specified. The default is <tt>NULL</tt>.

            You can configure the packet-signing feature in the <tt>signatureSetting</tt> member of the <tt>StartupSetting</tt> structure.
            The unsigned option is not allowed for Internet communication. The <tt>@ref Startup</tt> function fails if unsigned is specified.

            In the <tt>pToken</tt> member of the <tt>StartupSetting</tt> structure, you can set an identification token that indicates the local host.
            This identification token is shared with peers during the connection process, and it is associated with and maintained in <tt>@ref nn::pia::transport::Station</tt>.
            You are not required to specify an identification token if you do not need to use one. The default is <tt>NULL</tt>.

            The <tt>@ref Startup</tt> function is the only way to specify and change an identification token that indicates the local host.
            Note that there is no supported method for changing an identification token while a mesh is in use.
    @if CTR_DOC

            When local communications and host migration are enabled, host migration can begin after detecting a communication problem with the host on the local network.
            Note that the time period used by the local network to determine communication problems is not related to the maximum allowed silence time in <tt>Session</tt>.
    @endif

            If the bitrate detection feature is enabled, you must set the <tt>uplinkBitRateLowerLimit</tt> and <tt>bitRateCheckPacketSize</tt> members of the <tt>@ref StartupSetting</tt> structure.
            Both members default to an invalid value. Set the minimum bitrate (in bits per second) required by the application in <tt>uplinkBitRateLowerLimit</tt> and the IP packet size (in bytes) to use for measurements in <tt>bitRateCheckPacketSize</tt>.

    @param[in] startupSetting  Specifies the settings to use for startup.

    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Returned when already in startup or a similar state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates a problem with the name associated with the station, a problem with the configuration of the packet-signing feature, or that an invalid pointer was specified in one of the arguments. Programming error. Fix your program so that this error is not returned.

    @see  Cleanup, transport::Station::GetIdentificationToken, RegisterJoiningMeshApprovalCallback
*/
    nn::Result Startup(Mesh::StartupSetting& startupSetting);

    /// @cond PRIVATE
/*!
    @brief  Starts the <tt>Mesh</tt>.

    @details  Host migration is configured with the first argument. Set to <tt>true</tt> to enable.
             Specify the name and language code associated with the local station in the second parameter. The name must be no longer than 16 characters (not including the terminating null character).
    @if CAFE_DOC
             Set the language code to the language used by the application. Use one of the values defined in <tt>nn::fp::Language</tt>.
             The name and language code you specify are shared with all members of the mesh. When the local station joins a mesh, all mesh participants are added to Player History. When another
             client joins the mesh, the new participant's name and language code are looked up and automatically
             added to Player History using the <tt>nn::fp::AddRecentPlayRecord()</tt> function from within the library.
             For more information about Player History, see the <i>Pia Programming Manual</i> for the Friend Presence library
             and the <i>Pia API Reference</i> for the <tt>nn::fp::AddRecentPlayRecord()</tt> function.
    @endif
    @if CTR_DOC
             Specify a language code if your application determines the languages of other regions. You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage()</tt> function if you want to use the language setting of the system.
    @endif

            You can configure the communication encryption feature in the third argument.
            Communication is not encrypted if <tt>NULL</tt> is specified.

            You can configure the packet-signing feature in the fourth argument.
            The unsigned option is not allowed for Internet communication. The <tt>@ref Startup</tt> function fails if unsigned is specified.

            In the fifth parameter, you can set an identification token that indicates the local host.
            This identification token is shared with peers during the connection process, and it is associated with and maintained in <tt>@ref nn::pia::transport::Station</tt>.
            You are not required to specify an identification token if you do not need to use one.

            The <tt>@ref Startup</tt> function is the only way to specify and change an identification token that indicates the local host.
            Note that there is no supported method for changing an identification token while a mesh is in use.
    @if CTR_DOC

            When local communications and host migration are enabled, host migration can begin after detecting a communication problem with the host on the local network.
            Note that the time period used by the local network to determine communication problems is not related to the maximum allowed silence time in <tt>Session</tt>.
    @endif

    @param[in] bUsingHostMigration  Sets the use of the host migration feature.
    @param[in] pPlayerName  Specifies the name and language code associated with the local station. Other members of the mesh access this information.
    @param[in] pCryptoSetting  Specifies a pointer that points to a structure for configuring encryption. No encryption is configured if <tt>NULL</tt> is specified.
    @param[in] signatureSetting  Specifies a properly configured instance of the class used to specify packet-signature settings. The default value for this parameter is <tt>unsigned</tt>.
    @param[in] pToken  Specifies a pointer to the identification token indicating the local device. If <tt>NULL</tt> is specified, an identification token with all data set to <tt>NULL</tt> is used.
    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. You must make sure that the implementation of this function in your application does not return any errors.
            If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Returned when already in startup or a similar state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates a problem with the name associated with the station, a problem with the configuration of the packet-signing feature, or that an invalid pointer was specified in one of the arguments. Programming error. Fix your program so that this error is not returned.

    @see  Cleanup, transport::Station::GetIdentificationToken, RegisterJoiningMeshApprovalCallback
*/
    //nn::Result Startup(
    //    bool bUsingHostMigration,
    //    const nn::pia::transport::Station::PlayerName* pPlayerName,
    //    const nn::pia::common::CryptoSetting* pCryptoSetting,
    //    const nn::pia::common::SignatureSetting& signatureSetting = nn::pia::common::SignatureSetting::NO_SIGNATURE,
    //    const nn::pia::transport::Station::IdentificationToken* pToken = NULL);

/*!
    @brief  Starts the <tt>Mesh</tt>.

    @details  Specify the maximum allowable time without communication in the first argument. If the period without communication with the station exceeds the specified period, that station is treated as disconnected.
                Specify the send interval for keep alive in the second argument. Keep alive is a feature that keeps connections alive during periods of no communication by having the system send data to stations the application is not sending data to.

            The default values for the first and second arguments are listed below. The following defaults are used if you call an overload without configuration arguments.
            <ul>
            <li>First argument: <tt><var>nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT</var></tt></li>
            <li>Second argument: <tt><var>nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</var></tt></li>
            </ul>
            If you customize these values, make sure to test the actual results as you adjust them.

            Host migration is configured with the third argument. Set to <tt>true</tt> to enable.

            Specify the name and language code associated with the local station in the fourth argument. The name must be no longer than 16 characters (not including the terminating null character).
    @if CAFE_DOC
             Set the language code to the language used by the application. Use one of the values defined in <tt>nn::fp::Language</tt>.
             The name and language code you specify are shared with all members of the mesh. When the local station joins a mesh, all mesh participants are added to Player History. When another
             client joins the mesh, the new participant's name and language code are looked up and automatically
             added to Player History using the <tt>nn::fp::AddRecentPlayRecord()</tt> function from within the library.
             For more information about Player History, see the <i>Pia Programming Manual</i> for the Friend Presence library
             and the <i>Pia API Reference</i> for the <tt>nn::fp::AddRecentPlayRecord()</tt> function.
    @endif
    @if CTR_DOC
             Specify a language code if your application determines the languages of other regions. You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage()</tt> function if you want to use the language setting of the system.
    @endif

            You can configure the communication encryption feature in the fifth argument.
            Communication is not encrypted if <tt>NULL</tt> is specified.

            You can configure the packet-signing feature in the sixth argument.
            The unsigned option is not allowed for Internet communication. The <tt>@ref Startup</tt> function fails if unsigned is specified.

            In the seventh argument, you can set an identification token that indicates the local host.
            This identification token is shared with peers during the connection process, and it is associated with and maintained in <tt>@ref nn::pia::transport::Station</tt>.
            You are not required to specify an identification token if you do not need to use one.

            The <tt>@ref Startup</tt> function is the only way to specify and change an identification token that indicates the local host.
            Note that there is no supported method for changing an identification token while a mesh is in use.
    @if CTR_DOC

            When local communications and host migration are enabled, host migration can begin after detecting a communication problem with the host on the local network.
            Note that the time period used by the local network to determine communication problems is not related to the maximum allowed silence time in <tt>Session</tt>.
    @endif

    @param[in] maxSilenceTime  Specifies the maximum allowable time without communication (in milliseconds).
    @param[in] keepAliveSendingInterval  Specifies the send interval for keep-alive (in milliseconds).
    @param[in] bUsingHostMigration  Sets the use of the host migration feature.
    @param[in] pPlayerName  Specifies the name and language code associated with the local station. Other members of the mesh access this information.
    @param[in] pCryptoSetting  Specifies a pointer that points to a structure for configuring encryption. No encryption is configured if <tt>NULL</tt> is specified.
    @param[in] signatureSetting  Specifies a properly configured instance of the class used to specify packet-signature settings. The default value for this parameter is <tt>unsigned</tt>.
    @param[in] pToken  Specifies a pointer to the identification token indicating the local device. If <tt>NULL</tt> is specified, an identification token with all data set to <tt>NULL</tt> is used.
    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. You must make sure that the implementation of this function in your application does not return any errors.
            If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Returned when already in startup or a similar state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates a problem with the name associated with the station, a problem with the configuration of the packet-signing feature, or that an invalid pointer was specified in one of the arguments. Programming error. Fix your program so that this error is not returned.

    @see  Cleanup, transport::Station::GetIdentificationToken, RegisterJoiningMeshApprovalCallback
*/
//    nn::Result Startup(
//        u32 maxSilenceTime,
//        u32 keepAliveSendingInterval,
//        bool bUsingHostMigration,
//        const nn::pia::transport::Station::PlayerName* pPlayerName,
//        const nn::pia::common::CryptoSetting* pCryptoSetting,
//        const nn::pia::common::SignatureSetting& signatureSetting = nn::pia::common::SignatureSetting::NO_SIGNATURE,
//#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
//        s32 bandwidthLimit = NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT,
//#endif
//        const nn::pia::transport::Station::IdentificationToken* pToken = NULL);
    /// @endcond


/*!
    @brief  Cleans up the <tt>Mesh</tt>.

    @details  Call <tt>@ref Startup</tt> again if you want to reuse the <tt>Mesh</tt> instance after you have called the <tt>@ref Cleanup</tt> function.
            For Internet communication, call it again from the start of the NAT session (<tt>@ref nn::pia::inet::NexFacade::StartNatSessionAsync</tt>).

    @see  Startup
*/
    void Cleanup();

/*!
     end of name Finish Initialization
     @}
*/

/*!
     @name  Building, Destroying, Joining, and Leaving Meshes
     @{
*/

/*!
    @cond  PRIVATE
    @brief  Builds a mesh. (Old specification.)

    @details  The local device acts as the mesh host.

           Starts building a mesh asynchronously. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           You can monitor the completion of the process by checking the <tt>CallContext</tt> specified in the arguments.

           You must set up the <tt>@ref nn::pia::transport::Transport</tt> instance beforehand.

            <b>The following results may be returned by the <tt>CallContext</tt> instance if the asynchronous process fails.</b>
            <ul>
            <li> ResultInvalidState: This function is not in a usable state. Programming error. Fix your program so that this error is not returned.</li>
            </ul>

    @param[in] pCallContext  Specifies the <tt>CallContext</tt> pointer to check the execution progress status.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the local device successfully starts building a mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the function cannot be called in the specified <tt>CallContext</tt>. Programming error. Fix your program so that this error is not returned.

    @see  DestroyMesh, LeaveMeshWithHostMigration
*/
    nn::Result CreateMesh(nn::pia::common::CallContext* pCallContext);
    //!@endcond

/*!
    @brief  Builds a mesh.

    @details  The local device acts as the mesh host.

           Starts building a mesh asynchronously. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedCreateMesh</tt> to check whether the process has completed and <tt>@ref GetCreateMeshResult</tt> to check the result.

           You must set up the <tt>@ref nn::pia::transport::Transport</tt> instance beforehand.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the local device successfully starts building a mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedCreateMesh, GetCreateMeshResult
    @see  DestroyMeshAsync, LeaveMeshWithHostMigrationAsync
*/
    nn::Result CreateMeshAsync();

/*!
    @brief  Checks the completion of <tt>CreateMeshAsync</tt> asynchronous processing.

    @details  This function checks the completion of asynchronous processes started with <tt>@ref CreateMeshAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CreateMeshAsync, GetCreateMeshResult
*/
    bool IsCompletedCreateMesh();

/*!
    @brief  Gets the result of <tt>CreateMeshAsync</tt> asynchronous processing.

    @details  This function gets the results after an asynchronous process started by <tt>@ref CreateMeshAsync</tt> ends.

    @return  Returns the result of <tt>CreateMeshAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when <tt>CreateMeshAsync</tt> is called at the wrong time. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.

    @see  CreateMeshAsync, IsCompletedCreateMesh
*/
    nn::Result GetCreateMeshResult();


/*!
    @cond  PRIVATE
    @brief  Destroys a mesh. (Old specification.)

    @details  Only the mesh host can call this function.

           Starts an asynchronous mesh destruction process. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           You can monitor the completion of the process by checking the <tt>CallContext</tt> specified in the arguments.
           <b>Note:</b> If the mesh destruction process starts normally, the asynchronous process does not fail.

           Within the <tt>Mesh</tt> destruction process, all participants are notified that the <tt>Mesh</tt> will be ended.
           For stations that receive the notification, the internal state of their mesh is initialized and becomes <tt>STATE_DISCONNECTED</tt>.

    @param[in] pCallContext  Specifies the <tt>CallContext</tt> pointer to check the execution progress status.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the mesh is successfully destroyed. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that a node other than the mesh host called the function. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the function cannot be called in the specified <tt>CallContext</tt>. Programming error. Fix your program so that this error is not returned.

    @see  CreateMesh, LeaveMeshWithHostMigration
*/
    nn::Result DestroyMesh(nn::pia::common::CallContext* pCallContext);
    //!@endcond

/*!
    @brief  Destroys a mesh.

    @details  Only the mesh host can call this function.

           Starts an asynchronous mesh destruction process. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedDestroyMesh</tt> to check whether the process has completed and <tt>@ref GetDestroyMeshResult</tt> to check the result.
           <b>Note:</b> If the mesh destruction process starts normally, the asynchronous process does not fail.

           Within the <tt>Mesh</tt> destruction process, all participants are notified that the <tt>Mesh</tt> will be ended.
           For stations that receive the notification, the internal state of their mesh is initialized and becomes <tt>STATE_DISCONNECTED</tt>.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the mesh is successfully destroyed. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when called by a station other than the mesh host or when another asynchronous process has already started. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedDestroyMesh, GetDestroyMeshResult
    @see  CreateMeshAsync, LeaveMeshWithHostMigrationAsync
*/
    nn::Result DestroyMeshAsync();

/*!
    @brief  Checks the completion of <tt>DestroyMeshAsync</tt> asynchronous processing.

    @details  This function checks the completion of asynchronous processes started with <tt>@ref DestroyMeshAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  DestroyMeshAsync, GetDestroyMeshResult
*/
    bool IsCompletedDestroyMesh();

/*!
    @brief  Gets the result of <tt>DestroyMeshAsync</tt> asynchronous processing.

    @details  This function gets the results after an asynchronous process started by <tt>@ref DestroyMeshAsync</tt> ends.
                <b>Note:</b> If the mesh destruction process starts normally, the asynchronous process does not fail.

    @return  Returns the result of <tt>DestroyMeshAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.

    @see  DestroyMeshAsync, IsCompletedDestroyMesh
*/
    nn::Result GetDestroyMeshResult();


/*!
    @cond  PRIVATE
    @brief  Leaves a mesh. Also specifies that host migration will start for mesh participants. (Old specification.)

    @details  Only the mesh host can call this function. This function can be used only when the host migration feature is enabled.

           Starts an asynchronous mesh leaving process. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           You can monitor the completion of the process by checking the <tt>CallContext</tt> specified in the arguments.
           <b>Note:</b> The asynchronous process never fails if the leave mesh process starts normally.

           All session participants are notified that host migration has started within the asynchronous process.
           The host migration process begins on the stations that received the notification, and the mesh continues
           under the new mesh host even after communication is terminated with the old mesh host that called this function.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the mesh is successfully left. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that a node other than the mesh host called this function, or that host migration is disabled. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the function cannot be called in the specified <tt>CallContext</tt>. Programming error. Fix your program so that this error is not returned.

    @see  CreateMesh, DestroyMesh
*/
    nn::Result LeaveMeshWithHostMigration(nn::pia::common::CallContext* pCallContext);
    //!@endcond

/*!
    @brief  Leaves a mesh. Also specifies that host migration will start for mesh participants.

    @details  Only the mesh host can call this function. This function can be used only when the host migration feature is enabled.

           Starts an asynchronous mesh leaving process. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedLeaveMeshWithHostMigration</tt> to check whether the process has completed and <tt>@ref GetLeaveMeshWithHostMigrationResult</tt> to check the result.
           <b>Note:</b> The asynchronous process never fails if the leave mesh process starts normally.

           All session participants are notified that host migration has started within the asynchronous process.
           The host migration process begins on the stations that received the notification, and the mesh continues
           under the new mesh host even after communication is terminated with the old mesh host that called this function.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the mesh is successfully left. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when called by a station other than the mesh host, when the host migration feature is disabled, or when another asynchronous process has already started. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedLeaveMeshWithHostMigration, GetLeaveMeshWithHostMigrationResult
    @see  CreateMeshAsync, DestroyMeshAsync
*/
    nn::Result LeaveMeshWithHostMigrationAsync();

/*!
    @brief  Checks the completion of <tt>LeaveMeshWithHostMigrationAsync</tt> asynchronous processing.

    @details  This function checks the completion of asynchronous processes started with <tt>@ref LeaveMeshWithHostMigrationAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  LeaveMeshWithHostMigrationAsync, GetLeaveMeshWithHostMigrationResult
*/
    bool IsCompletedLeaveMeshWithHostMigration();

/*!
    @brief  Gets the result of <tt>LeaveMeshWithHostMigrationAsync</tt> asynchronous processing.

    @details  This function gets the results after an asynchronous process started by <tt>@ref LeaveMeshWithHostMigrationAsync</tt> ends.
                <b>Note:</b> If the mesh destruction process starts normally, the asynchronous process does not fail.

    @return  Returns the result of <tt>LeaveMeshWithHostMigrationAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.

    @see  LeaveMeshWithHostMigrationAsync, IsCompletedLeaveMeshWithHostMigration
*/
    nn::Result GetLeaveMeshWithHostMigrationResult();


    /// @cond PRIVATE
/*!
    @brief  Specifies whether join requests are allowed.

    @return  Returns <tt>0xFF</tt> if join requests are allowed or another value if they are denied.
*/
    u8 CheckApprovalJoin(transport::Station* pTargetStation);
    /// @endcond


/*!
    @cond  PRIVATE
    @brief  Joins a mesh. (Old specification.)

    @details  This function starts an asynchronous processing required to join the mesh. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            You can monitor the completion and success or failure of the process by checking the <tt>CallContext</tt> specified in the second argument.
            You can also cancel the process by calling the <tt>@ref nn::pia::common::CallContext::Cancel</tt> function during the process.

            <b>The following results may be returned by the <tt>CallContext</tt> instance when the asynchronous process ends.</b>
            <ul>
            <li> <tt>ResultSuccess</tt>: The asynchronous process succeeded.</li>
            <li> <tt>ResultCancelled</tt>: The asynchronous process was canceled. Perform cleanup.</li>
            <li> <tt>ResultInvalidState</tt>: Connection is not available in this state. Programming error. Fix your program so that this error is not returned.</li>
            <li> <tt>ResultStationConnectionFailed</tt>: The connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultStationConnectionNatTraversalFailedUnknown</tt>: NAT traversal connection failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultNatTraversalFailedBothEim</tt>: NAT traversal connection failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultNatTraversalFailedLocalEimRemoteEdm</tt>: NAT traversal connection failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultNatTraversalFailedLocalEdmRemoteEim</tt>: NAT traversal connection failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultNatTraversalFailedBothEdm</tt>: NAT traversal connection failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultNatTraversalRequestTimeout</tt>: NAT traversal connection failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.</li>
            <li> <tt>ResultJoinRequestDenied</tt>: The mesh host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultMeshIsFull</tt>: Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultRelayFailedNoCandidate</tt>: The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultRelayFailedRttLimit</tt>: The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultRelayFailedRelayNumLimit</tt>: The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultRelayFailedUnknown</tt>: The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.</li>
            <li> <tt>ResultInvalidSystemMessage</tt>: The process was canceled because an invalid response was received from the mesh host. Programming error. Fix your program so that this error is not returned.</li>
            <li> <tt>ResultIncompatibleFormat</tt>: Specifies that the peer is using an incompatible communication method. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.</li>
            </ul>

            If you execute other processes without destroying the <tt>Mesh</tt> instance after execution fails, or if you cancel execution using the
            <tt>@ref nn::pia::common::CallContext::Cancel</tt> function, do so after you first clean up <tt>Mesh</tt> by calling <tt>@ref Cleanup</tt>.
            For Internet communication, call it again from the start of the NAT session (<tt>@ref nn::pia::inet::NexFacade::StartNatSession</tt>).

    @param[in] stationConnectionInfo  Specifies information for connecting to the host of the mesh being joined.
    @param[in] pCallContext  Specifies the <tt>CallContext</tt> pointer to check the execution progress status.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the station successfully joins the mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidArgument  Indicates that the <tt>CallContext</tt> specified in the arguments was not in a usable state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  The mesh has already been joined, or the <tt>JoinMesh</tt> asynchronous process already started. Programming error. Fix your program so that this error is not returned.

    @see  LeaveMesh
*/
    nn::Result JoinMesh(
        const nn::pia::transport::StationConnectionInfo& stationConnectionInfo,
        nn::pia::common::CallContext* pCallContext);
    //!@endcond

/*!
    @brief  Joins a mesh.

    @details  This function starts an asynchronous processing required to join the mesh. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedJoinMesh</tt> to check whether the process has completed and <tt>@ref GetJoinMeshResult</tt> to check the result.
            You can also cancel the process by calling the <tt>@ref CancelJoinMeshAsync</tt> function during the process.

            If you execute other processes without destroying the <tt>Mesh</tt> instance after execution fails, or if you cancel execution using the
            <tt>@ref CancelJoinMeshAsync</tt> function, do so after you first clean up <tt>Mesh</tt> by calling <tt>@ref Cleanup</tt>.
            For Internet communication, call it again from the start of the NAT session (<tt>@ref nn::pia::inet::NexFacade::StartNatSessionAsync</tt>).

    @param[in] stationConnectionInfo  Specifies information for connecting to the host of the mesh being joined.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the station successfully joins the mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that the mesh has already been joined or another asynchronous process has already been started. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedJoinMesh, GetJoinMeshResult, CancelJoinMeshAsync
    @see  LeaveMeshAsync
*/
    nn::Result JoinMeshAsync(const nn::pia::transport::StationConnectionInfo& stationConnectionInfo);

/*!
    @brief  Checks the completion of <tt>JoinMeshAsync</tt> asynchronous processing.

    @details  This function checks the completion of asynchronous processes started with <tt>@ref JoinMeshAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  JoinMeshAsync, GetJoinMeshResult, CancelJoinMeshAsync
*/
    bool IsCompletedJoinMesh();

/*!
    @brief  Gets the result of <tt>JoinMeshAsync</tt> asynchronous processing.

    @details  This function gets the results after an asynchronous process started by <tt>@ref JoinMeshAsync</tt> ends.

    @return  Returns the result of <tt>JoinMeshAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultCancelled  The asynchronous process was canceled. Perform cleanup.
    @retval ResultInvalidState  Returned when the station is not in a connection-ready state. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval ResultStationConnectionFailed  The connection process failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultStationConnectionNatTraversalFailedUnknown  The NAT traversal connection process failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultNatTraversalFailedBothEim  The NAT traversal connection process failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultNatTraversalFailedLocalEimRemoteEdm  The NAT traversal connection process failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultNatTraversalFailedLocalEdmRemoteEim  The NAT traversal connection process failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultNatTraversalFailedBothEdm  The NAT traversal connection process failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultNatTraversalRequestTimeout  The NAT traversal connection process failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval ResultJoinRequestDenied  The mesh host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultMeshIsFull  Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultRelayFailedNoCandidate  The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultRelayFailedRttLimit  The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultRelayFailedRelayNumLimit  The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultRelayFailedUnknown  The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval ResultInvalidSystemMessage  The process was canceled because an invalid response was received from the mesh host. Programming error. Fix your program so that this error is not returned.
    @retval ResultIncompatibleFormat  Indicates that the peer is using an incompatible communication format. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.

    @see  JoinMeshAsync, IsCompletedJoinMesh, CancelJoinMeshAsync
*/
    nn::Result GetJoinMeshResult();

/*!
    @brief  Cancels <tt>JoinMeshAsync</tt> asynchronous processes.

    @details  This function cancels asynchronous processes started by <tt>JoinMeshAsync</tt>.
                The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
                The library quickly acts to cancel the asynchronous process after this function is called, but check
                using the <tt>@ref IsCompletedJoinMesh()</tt> function that the cancellation has completed.

    @return  Returns the result of canceling <tt>JoinMeshAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when no <tt>JoinMeshAsync</tt> asynchronous processing is being performed. Programming error. Fix your program so that this error is not returned.

    @see  JoinMeshAsync, IsCompletedJoinMesh, GetJoinMeshResult
*/
    nn::Result CancelJoinMeshAsync();


/*!
    @cond  PRIVATE
    @brief  Leaves a joined mesh. (Old specification.)

    @details  Only the mesh clients can call this function.

            Starts an asynchronous process to leave the mesh. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            You can monitor the completion and success or failure of the process by checking the <tt>CallContext</tt> function specified in the arguments.

    @param[in] pCallContext  Specifies the <tt>CallContext</tt> pointer to check the execution progress status.

    @return  Returns a <tt>Result</tt> value indicating success if the station successfully starts the process for leaving the mesh.
    @retval ResultInvalidState  Indicates that a mesh host called this function. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the function cannot be called in the specified <tt>CallContext</tt>. Programming error. Fix your program so that this error is not returned.

    @see  JoinMesh
*/
    nn::Result LeaveMesh(nn::pia::common::CallContext* pCallContext);
    //!@endcond

/*!
    @brief  Leaves a joined mesh.

    @details  Only the mesh clients can call this function.

            Starts an asynchronous process to leave the mesh. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedJoinMesh</tt> to check whether the process has completed and <tt>@ref GetJoinMeshResult</tt> to check the result.
            <b>Note:</b> The asynchronous process never fails if the leave mesh process starts normally.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the mesh is successfully left. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when called by the mesh host or when another asynchronous process has already started. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedLeaveMesh, GetLeaveMeshResult
    @see  JoinMeshAsync
*/
    nn::Result LeaveMeshAsync();

/*!
    @brief  Checks the completion of <tt>LeaveMeshAsync</tt> asynchronous processing.

    @details  This function checks the completion of asynchronous processes started with <tt>@ref LeaveMeshAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  LeaveMeshAsync, GetLeaveMeshResult
*/
    bool IsCompletedLeaveMesh();

/*!
    @brief  Gets the result of <tt>LeaveMeshAsync</tt> asynchronous processing.

    @details  This function gets the results after an asynchronous process started by <tt>@ref LeaveMeshAsync</tt> ends.
                <b>Note:</b> If the mesh destruction process starts normally, the asynchronous process does not fail.

    @return  Returns the result of <tt>LeaveMeshAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.

    @see  LeaveMeshAsync, IsCompletedLeaveMesh
*/
    nn::Result GetLeaveMeshResult();


/*!
    @brief  Kicks a station from the mesh.

    @details  Only the mesh host can call this function.

            A kick notice is conveyed to the station assigned the <tt>StationId</tt> specified in
            the argument, and that station leaves the mesh upon receiving the notice.

            Because of latency in transmitting the disconnect notification and other factors, there may be a delay between successfully calling this function, and the station leaving the mesh.

    @param[in] id  Specifies the station ID assigned to the station being kicked.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the station is successfully kicked from the mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Indicates that a node other than the mesh host called the function. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the <tt>StationId</tt> specified in the arguments belongs to the mesh host, or that no station in the mesh has that ID. Programming error. Fix your program so that this error is not returned.
    @retval ResultInProgress  Indicates that the process to kick the station having the <tt>StationId</tt> specified in the arguments has already started. Handle appropriately in the application.
    @retval ResultBufferIsFull  Indicates that the buffer for kicking stations or the buffer for sending kick notifications is full. It may succeed if you try again after some time has passed. Handle appropriately in the application.
*/
    nn::Result KickoutStation(nn::pia::StationId id);

    /// @cond PRIVATE
    nn::Result KickoutStationByIndex(nn::pia::StationIndex index);
    /// @endcond

/*!
     end of name Building, Destroying, Joining, and Leaving Meshes
     @}
*/

/*!
     @name  Mesh Analysis
     @{
*/

/*!
    @brief  Gets the number of valid stations connected to the current mesh.

    @return  Returns the number of connected stations.
*/
    u16 GetStationNum() const
    {
        return m_StationCount;
    }


/*!
    @brief  Gets the <tt>StationId</tt> of the mesh host.

    @details  This function returns <tt>STATION_ID_INVALID</tt> if a mesh host has not been established.

    @return  Returns the <tt>StationId</tt> representing the mesh host.
*/
    nn::pia::StationId GetHostStationId() const;


/*!
    @brief  Gets the <tt>StationId</tt> that specifies the local device.

    @details  This function returns <tt>STATION_ID_INVALID</tt> if the local device's <tt>StationId</tt> has not been established.

    @return  Returns the <tt>StationId</tt> representing the local station.
*/
    nn::pia::StationId GetLocalStationId() const;


    /// @cond PRIVATE
    nn::pia::StationIndex GetHostStationIndex() const
    {
        return m_HostStationIndex;
    }

    nn::pia::StationIndex GetLocalStationIndex() const
    {
        return m_LocalStationIndex;
    }
    /// @endcond


/*!
    @brief  Gets whether the local device is the mesh host.

    @details  This function always returns <tt>false</tt> if the local device is not joined in a mesh.

    @return  Returns <tt>true</tt> if the local device is the mesh host, or <tt>false</tt> otherwise.
*/
    bool IsHost() const
    {
        if (m_LocalStationIndex > STATION_INDEX_MAX)
        {
            return false;
        }
        return m_LocalStationIndex == m_HostStationIndex;
    }


/*!
    @brief  Checks whether an arbitrary <tt>StationId</tt> within the mesh is currently valid.

    @param[in] id  Specifies the <tt>StationId</tt> being checked.

    @return  Returns <tt>true</tt> if the specified <tt>StationId</tt> is valid, or <tt>false</tt> otherwise.
*/
    bool CheckStationIdIsValid(nn::pia::StationId id) const;


    /// @cond PRIVATE
    bool CheckStationIndexIsValid(nn::pia::StationIndex index) const;

/*!
    @brief  Gets a <tt>StationIndex</tt> that is currently not used in the mesh.

    @return  Returns <tt>STATION_INDEX_INVALID</tt> if there are no valid unused <tt>StationIndex</tt> values.
*/
    nn::pia::StationIndex GetUsableStationIndex();


/*!
    @brief  Gets a bitmap indicating the currently valid station indexes in the mesh.

    @details  <tt>StationIndex</tt> values are obtained by shifting bits. <tt>STATION_INDEX_1</tt> is <tt>0x00000001</tt>, <tt>STATION_INDEX_2</tt> is <tt>0x00000002</tt>, and so on.

    @return  Returns a bitmap corresponding to the <tt>StationIndex</tt> instances that are currently active in the session.
*/
    bit32 GetValidStationIndexBitmap() const
    {
        return m_StationIndexUsedFlags;
    };
    /// @endcond

/*!
    @brief  Confirms whether the state of the current <tt>Mesh</tt> is communicable.

    @details  Indicates in the <tt>Result</tt> whether the current <tt>Mesh</tt> is in a communicable state. When not in a communicable state, use <tt>@ref GetDisconnectReason</tt> to check, in detail, the reasons for entering that state.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if communication is possible. Otherwise, one of the following <tt>Result</tt> values is returned.
    @retval ResultMeshConnectionIsLost  The <tt>Mesh</tt> is in an incommunicable state. Perform cleanup.

    @see  GetDisconnectReason
*/
    nn::Result CheckConnectionError() const;


/*!
    @brief  Gets the reason why the <tt>Mesh</tt> object was disconnected.

    @details  This function always returns <tt>UNKNOWN_REASON</tt> if the <tt>Mesh</tt> object is in a communicable state.

    @return  Returns a <tt>DisconnectReason</tt> value indicating the reason why <tt>Mesh</tt> disconnected.

    @see  CheckConnectionError
*/
    DisconnectReason GetDisconnectReason() const;

/*!
    @brief  Determines whether the bitrate value may be found using the bitrate detection feature.

    @return  Returns <tt>true</tt> if the bitrate value can be obtained. <tt>true</tt> is also returned when detection fails on a timeout, or cancel.
*/
    bool IsCompleteBitRateCheck();

/*!
    @brief  Gets the uplink bitrate of the connection measured by the bitrate detection feature.

    @details  Note that the value returned by this function is simply a measurement of the amount of data received over a set
             amount of time and does not necessarily represent the actual bandwidth of the connection.
             Even if the connection is capable of a high bitrate, if an application has
             low requirements, the amount of data sent and received during the measurement
             process is also small, which results in a lower bits-per-second value.

    @param[out] pBps  Stores the measured bitrate. The value is in bits per second. A negative value is set if the bitrate could not be measured.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the call succeeds.

    @retval ResultNotInitialized  Indicates that the instance has not been initialized. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates that the function was called at the wrong time. It is possible that the function was called while the bitrate detection feature was disabled. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetUplinkBitRate(s32* pBps);

/*!
     end of name Mesh Analysis
     @}
*/

/*!
     @name  Registering and Unregistering Callback Functions
     @{
*/

/*!
    @brief  Registers the callback function that is called when the mesh state changes.

    @details  This function registers the callback that is called when the session mesh changes (when a station joins or exits).

            The mesh state change callback function is generally called by the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            As an exception, it can also be called when the <tt>@ref CreateMeshAsync</tt> and <tt>@ref DestroyMeshAsync</tt> functions are called. It can also be called when the <tt>@ref Cleanup</tt> function is called.

            Do not include any processes that block for long periods of time in the registering function.
            In addition, do not call <tt>session</tt> module functions within the function that you are registering.

    @param[in] callback  Specifies a pointer to the callback function being registered.

    @see  UnregisterMeshEventCallback
*/
    void RegisterMeshEventCallback(MeshEventCallback callback);


/*!
    @brief  Unregisters the registered callback function for mesh state changes.

    @see  RegisterMeshEventCallback
*/
    void UnregisterMeshEventCallback();


/*!
    @brief  Registers the callback used to approve or deny requests received by the host from clients seeking to join the mesh.

    @details  The callback function for deciding whether to approve <tt>Join</tt> requests is called by the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.

            Do not include any processes that block for long periods of time in the registering function.

            The ID token of the station that sent the <tt>Join</tt> request is passed as an argument to the callback function.

            If you do not register a callback function, <tt>Join</tt> requests are always approved unless the mesh is full.

    @param[in] callback  Specifies a pointer to the callback function being registered.

    @see  Startup, UnregisterJoiningMeshApprovalCallback
*/
    void RegisterJoiningMeshApprovalCallback(JoiningMeshApprovalCallback callback);


/*!
    @brief  Unregisters the registered <tt>Join</tt> request callback function for approving or denying the request.

    @see  RegisterJoiningMeshApprovalCallback
*/
    void UnregisterJoiningMeshApprovalCallback();


    /// @cond PRIVATE
/*!
    @brief  Registers the callback function for adjusting to any secondary host candidate position when using multiple candidate host migration.

    @details  The callback function for adjusting to any secondary host candidate position is called by the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.

            Do not include any processes that block for long periods of time in the registering function.

    @param[in] callback  Specifies a pointer to the callback function being registered.
    @return  If it was possible to call this function while multiple candidate host migration was enabled, the <tt>IsSuccess</tt> function returns a <tt>Result</tt> value of <tt>true</tt>.

    @see  UnregisterCheckHostCandidateRankingCallback
*/
    nn::Result RegisterCheckHostCandidateRankingCallback(CheckHostCandidateRankingCallback callback);


/*!
    @brief  Unregisters the callback function registered for adjusting to any secondary host candidate position.

    @see  RegisterCheckHostCandidateRankingCallback
*/
    void UnregisterCheckHostCandidateRankingCallback();
    /// @endcond


    /// @cond PRIVATE
    void RegisterEventListener(MeshEventListener* pEventListener);
    void UnregisterEventListener();

    void RegisterGetSessionIdCallback(GetSessionIdCallback callback);
    void UnregisterGetSessionIdCallback();
    /// @endcond


/*!
     end of name Registering and Unregistering Callback Functions
     @}
*/


    /// @cond PRIVATE
/*!
    @brief  Gets the <tt>MeshProtocol</tt> object generated automatically by the system.

    @return  Returns a pointer to a <tt>MeshProtocol</tt> instance.
*/
    MeshProtocol* GetMeshProtocol()
    {
        return m_pMeshProtocol;
    }
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  Gets the <tt>StationProtocol</tt> object generated automatically by the system.

    @return  Returns a pointer to a <tt>StationProtocol</tt> instance.
*/
    nn::pia::transport::StationProtocol* GetStationProtocol()
    {
        return m_pStationProtocol;
    }
    /// @endcond


/*!
     @name  Synchronous Clock
     @{
*/

/*!
    @brief  Specifies the synchronous clock (the time, in milliseconds, since the mesh was created).

    @details  The synchronous clock is synchronized among all the stations in the mesh.
                Clients joined in the mesh periodically send a clock-synchronization pulse to the mesh host.
                The host then responds to these pulses, ensuring synchronization between the host and clients.
                By default, synchronization pulses are sent every two seconds,
                but the application can also change this interval. For more information, see <tt>SetSyncClockPulseInterval()</tt> and
                <tt>GetSyncClockPulseInterval()</tt>.

    @return  Returns the elapsed time, in milliseconds, since the mesh was created. A negative value is returned when a mesh client out of synchronization
                with the mesh host calls this function.
    @see  SetSyncClockPulseInterval, GetSyncClockPulseInterval
*/
    s64 GetTime(void) const;


/*!
    @brief  Specifies the send interval of the synchronization pulse for the synchronous clock.

    @param[in] interval  Specifies the interval at which the synchronization pulse is sent. The value is in milliseconds.
    @see  GetTime, GetSyncClockPulseInterval
*/
    void SetSyncClockPulseInterval(s32 interval);


/*!
    @brief  Gets the send interval of the synchronization pulse for the synchronous clock.

    @return  Returns the interval at which synchronization pulses are sent. The value is in milliseconds.
    @see  GetTime, SetSyncClockPulseInterval
*/
    s32 GetSyncClockPulseInterval(void) const;

/*!
     end of name Synchronous Clock
     @}
*/


/*!
     @name  Relay Feature
     @{
*/

/*!
    @brief  Sets the maximum number of forwarding requests that can be assigned to a single station when relaying is enabled.

    @details  The library uses this value as a metric when searching for and selecting relay paths.
                If you do not set a value, the library uses [<em>maximum mesh participants</em>] × [<em>maximum mesh participants</em>] as the default.

                We recommend that you set this to the same value for all participants in a mesh. (Operation is not guaranteed if different values are set on different clients.)

                This function is only available when the network topology allows relaying, and <tt>@ref Startup</tt> has not yet been called.

    @param[in] num  Specifies the maximum value.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval ResultInvalidState  Indicates that relay connections are disabled or the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetMaxNumOfRelayRoutesAssignedToStation(u16 num);


/*!
    @brief  Sets the maximum total RTT for a relay route when relaying is enabled.

    @details  The library uses this value as a metric when searching for and selecting relay paths.
                If you do not set a value, the library uses the default value of <tt>nn::pia::session::NN_PIA_SESSION_MAX_RTT_DEFAULT</tt>.

                We recommend that you set this to the same value for all participants in a mesh. (Operation is not guaranteed if different values are set on different clients.)

                This function is only available when the network topology allows relaying, and <tt>@ref Startup</tt> has not yet been called.

    @param[in] rtt  Specifies the maximum value. The value is in milliseconds.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval ResultInvalidState  Indicates that relay connections are disabled or the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetMaxRtt(u16 rtt);

/*!
    @brief  Gets the connection path between the local station and an arbitrary station index when relaying is enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.
                The connection path between the local station and <tt><var>targetIndex</var></tt> is stored in <tt><var>pRelayIndex</var></tt>, according to the following rules.
                (The value of <tt><var>pRelayIndex</var></tt> does not change if the function could not retrieve the connection path.)
                <ul>
                <li>Direct connection: <tt><var>targetIndex</var></tt></li>
                <li>Relay connection: <tt>StationIndex</tt> of the relay requester.</li>
                <li>Local station: <tt>STATION_INDEX_INVALID</tt></li>
                <li>Not connected: <tt>STATION_INDEX_INVALID</tt></li>
                </ul>

    @param[in] targetIndex  Specifies the <tt>StationIndex</tt> with the connection path to retrieve.
    @param[out] pRelayIndex  Specifies the <tt>StationIndex</tt> indicating the connection path.

    @return  Returns a <tt>Result</tt> value indicating success if the value can be retrieved successfully.
    @retval ResultInvalidState  Indicates that relay connections are disabled. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  The specified <tt>StationIndex</tt> is invalid, or the pointer specified in the arguments is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetConnectionRoute(StationIndex targetIndex, StationIndex* pRelayIndex);

/*!
    @brief  Specifies connection path emulation settings between any two <tt>StationIndex</tt> objects when relaying is enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.

    @details  When you call this function, make sure that you have enabled the connection path emulation feature by using the <tt>@ref Mesh::SetDebugSetting</tt> function.

    @details  If <tt>EMULATION_TYPE_RELAY</tt> is set, the connection method for the target <tt>StationIndex</tt> is always a relay connection
                (even when, for example, a direct connection is possible).

    @details  Call this function when the terminal is not connected to a mesh.
                If you call this function when the terminal has already joined a mesh, there is no guarantee that changing the connection path emulation settings will work.

    @param[in] index1  Specifies a <tt>StationIndex</tt>.
    @param[in] index2  Specifies a <tt>StationIndex</tt>.
    @param[in] type  Specifies the connection method emulation setting.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval ResultInvalidState  Indicates that relay connections are disabled. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that the <tt>StationIndex</tt> is invalid, or the specified <tt>EmulationType</tt> is invalid. Programming error. Fix your program so that this error is not returned.

    @see  ResetRelayEmulation
*/
    nn::Result RelayEmulation(StationIndex index1, StationIndex index2, EmulationType type);

/*!
    @brief  Initializes connection path emulation settings when relay connections are enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.

    @details  All connections are initialized to use the connection path emulation configured by the <tt>Mesh::RelayEmulation</tt> function.
                (This is equivalent to setting <tt>EMULATION_TYPE_NONE</tt> for all combinations of <tt>StationIndex</tt>.)

    @details  Call this function when the terminal is not connected to a mesh.
                If you call this function when the terminal has already joined a mesh, there is no guarantee that changing the connection path emulation settings will work.

    @see  RelayEmulation
*/
    void ResetRelayEmulation();

/*!
     end of name Relay Feature
     @}
*/


    /// @cond PRIVATE
/*!
    @brief  Sets a flag for determining whether host migration has started.
    @param[in] val  Specifies that the flag is in this state after the function is called.
*/
    void SetHostMigrationStartFlag(bool val);
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  Gets a flag for determining whether host migration has started.
    @return  Returns the flag state.
*/
    bool GetHostMigrationStartFlag();
    /// @endcond

    //private
    ProcessHostMigrationJob* GetProcessHostMigrationJobPtr()
    {
        return m_pProcessHostMigrationJob;
    }

    KickoutManageJob* GetKickoutManageJobPtr()
    {
        return m_pKickoutManageJob;
    }

    //private
    ProcessDestroyMeshJob* GetProcessDestroyMeshJobPtr()
    {
        return m_pProcessDestroyMeshJob;
    }

    void SetFatalError(DisconnectReason reason);

    /// @cond PRIVATE
/*!
    @brief  Sets whether to add players to Player History automatically.
    @param[in] val  Specify <tt>true</tt> to add players automatically.
*/
    void SetAutoAddRecentPlayRecord(bool val)
    {
        m_AutoAddRecentPlayRecord = val;
    }
/// @endcond

#if NN_PIA_EXPERIMENT_SHARE_CONNECTION_INFO_EARLY
    //private
    ProcessUpdateMeshJob* GetProcessUpdateMeshJobPtr()
    {
        return m_pProcessUpdateMeshJob;
    }
#endif

    //private
    NetworkTopology GetNetworkTopologyType()
    {
        return m_NetworkTopologyType;
    }

    /// @cond PRIVATE
/*!
    @brief  Gets the location where the signature setting is stored.
*/
    const SignatureSettingStorage* GetSignatureSettingStorage() const
    {
        return m_pSignatureSettingStorage;
    }
    /// @endcond

    /// @cond PRIVATE
    static nn::Result SetDebugSettingForInternal(const DebugSetting& debugSetting);
    /// @endcond

    /// @cond PRIVATE
    void NotifyLeaveStationAddress(const common::StationAddress& leftStationAddress);
    /// @endcond

/*!
     @name  Other
     @{
*/

/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    void TraceStationConnectionInfoList(u64 flag);

    /// @cond PRIVATE
    bool IsUsingHostMigration();  //Gets whether host migration is in use.
    int GetJoinMeshSequence();      // Gets the progress of joining the mesh.
    void SetSkipCleanupReport(bool isSkip);
    void SetSkipReport(bool isSkip);
    void SendReport(u8 type);
    bool IsProcessedAtMeshBegin() const;
    bool IsProcessingLeaveMesh() const;
    State GetMeshState() const;
    void IsDenyJoinRequestBySession(bool isDeny);   // While <tt>true</tt> is set, participation in the <tt>Mesh</tt> instance fails on denial from <tt>Session</tt> and <tt>ResultJoinRequestDenied</tt> is returned.
    void CleanupMonitoringBeginData();
    /// @endcond

#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
    /// @cond PRIVATE
    BitRateCheckMode GetBitRateCheckMode();
    nn::pia::transport::BandwidthCheckerProtocol* GetBandwidthCheckerProtocol();
    /// @endcond
#endif

    /// @cond PRIVATE
    s32 GetProcessUpdateMeshJobStepSequence() const;
    /// @endcond

/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    virtual void Trace(u64 flag) const;

/*!
     end of name Others
     @}
*/

protected:
    static Mesh* s_pMesh;
    static DebugSetting s_DebugSetting;

    static nn::Result setDebugSettingImpl(const DebugSetting& debugSetting);

private:
    /// @cond PRIVATE
/*!
    @brief  Instantiates an object. (DO NOT USE.)
*/
    Mesh(const nn::pia::session::Mesh::Setting& setting);
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  Destroys the object. (DO NOT USE.)
*/
    virtual ~Mesh(void);
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  This copy constructor is private.
*/
    Mesh(const Mesh&);
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  This assignment operator is private.
*/
    Mesh& operator=(const Mesh&);
    /// @endcond

    nn::Result Initialize(const nn::pia::session::Mesh::Setting& setting, bool isEnableRelayEmulation);

    nn::Result SetupProtocols();  // System protocol preparations.

    void CleanupStatus();  // Called at the end of the mesh that initializes the internal state.
    void CleanupJobs();    // Called at the end of the mesh that initializes the job the mesh manages.

    void CleanupStationsJobs();

    void FixConnectedId(nn::pia::StationIndex id);
    void UnfixDisconnectedId(nn::pia::StationIndex id);

    void StartUse(nn::pia::StationIndex id);  // Changes the status to "in use."
    void StopUse(nn::pia::StationIndex id);   // Changes the status to "not in use."

    void ResetStationIndexUsage();  // Resets the use status of the <tt>StationIndex</tt>.

    // Records the times when the mesh was successfully created and joined.
    void SetMeshStartTime(common::Time startTime)
    {
        m_MeshStartTime = startTime;
        m_MeshStartTimeValidFlag = true;
    }

    // Checks whether the values recorded for when the mesh was successfully created or joined are valid.
    bool CheckMeshStartTimeIsValid() const
    {
        return m_MeshStartTimeValidFlag;
    }

    // Calculates the difference between an arbitrary time and when the mesh was successfully created or joined.
    common::TimeSpan CalcDiffWithMeshStartTime(common::Time target) const
    {
        return (target - m_MeshStartTime);
    }

    void ProcessingAtMeshBegin();
    void ProcessingAtMeshEnd(DisconnectReason reason);
    void ProcessingAtEndHM();
    void MonitoringProcess(DisconnectReason reason, u8 type);

    enum AsyncProcessingName
    {
        ASYNC_NOTHING = 0,
        ASYNC_CREATE_SESSION,
        ASYNC_DESTROY_SESSION,
        ASYNC_LEAVE_SESSION_WITH_HM,
        ASYNC_JOIN_SESSION,
        ASYNC_LEAVE_SESSION
    };

    void ResetInternalAsyncState();
    bool IsCompletedAsyncProcess(AsyncProcessingName processingName);
    nn::Result GetAsyncProcessResult(AsyncProcessingName processingName);
    nn::Result CancelAsyncProcess(AsyncProcessingName processingName);

    nn::Result joinMeshCore(
        const nn::pia::transport::StationConnectionInfo& stationConnectionInfo,
        nn::pia::common::CallContext* pCallContext);

    void SetMeshState(State meshState);
    void NoticeMeshEvent(EventType eventType, StationIndex idx);

    CreateMeshJob* m_pCreateMeshJob;
    JoinMeshJob* m_pJoinMeshJob;
    LeaveMeshJob* m_pLeaveMeshJob;
    ProcessJoinRequestJob* m_pProcessJoinRequestJob;
    ProcessUpdateMeshJob* m_pProcessUpdateMeshJob;
    DestroyMeshJob* m_pDestroyMeshJob;
    ProcessDestroyMeshJob* m_pProcessDestroyMeshJob;
    ProcessHostMigrationJob* m_pProcessHostMigrationJob;
    LeaveWithHostMigrationJob* m_pLeaveWithHostMigrationJob;
    SignatureSettingStorage* m_pSignatureSettingStorage;
#if NN_PIA_EXPERIMENT_RELAY_CONNECTION_TOPOLOGY
    RelayRouteManageJob* m_pRelayRouteManageJob;
#endif
    KickoutManageJob* m_pKickoutManageJob;
    nn::pia::transport::MissingStationHandler* m_pMissingStationHandler;

    MeshEventCallback m_MeshEventCallback;
    JoiningMeshApprovalCallback m_JoiningMeshApprovalCallback;
    CheckHostCandidateRankingCallback m_CheckHostCandidateRankingCallback;
    GetSessionIdCallback m_GetSessionIdCallback;

    u32 m_MeshProtocolHandle;
    MeshProtocol* m_pMeshProtocol;
    nn::pia::transport::StationProtocol* m_pStationProtocol;

#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
    u32 m_BandwidthCheckerHandle;
    nn::pia::transport::BandwidthCheckerProtocol* m_pBandwidthCheckerProtocol;
#endif

    u16 m_StationCount;
    u16 m_StationMax;
    bool m_IsApprovalJoin;

    // n2908: Approve/deny join-in flag that is changed only when <tt>OpenSession</tt> or <tt>CloseSession</tt> is handled by <tt>Session</tt>.
    bool m_IsDenyJoinRequestBySession;

    nn::pia::StationIndex m_HostStationIndex;
    nn::pia::StationIndex m_LocalStationIndex;

    State m_State;
    DisconnectReason m_DisconnectReason;
    HostMigrationMode m_HostMigrationMode;
    NetworkTopology m_NetworkTopologyType;
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
    BitRateCheckMode m_BitRateCheckMode;
#endif

    bool m_IsEnableHostMigration;

    bit32 m_StationIndexUsedFlags;
    u32 m_UpdateCount;

    //n1769
    u32 m_SyncClockProtocolHandle;

    nn::pia::common::Time m_MeshStartTime;
    bool m_MeshStartTimeValidFlag;

    bool m_IsStarted;

    bool m_HostMigrationStartFlag;

    bool m_AutoAddRecentPlayRecord;

    common::MonitoringDataSender* m_pMonitoringDataSender;

    common::CallContext m_InternalContext;
    AsyncProcessingName m_CurrentAsyncProcessingName;

    MeshEventListener* m_pMeshEventListener;

    bool m_IsSkipSendingReport;
    bool m_IsSkipCleanupReport;
};
}
}
}  // end of namespace nn::pia::session
