﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: session_SessionSearchCriteria.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_definitions.h>

namespace nn
{
namespace pia
{
namespace session
{
/*!
    @brief  The base class for the search criteria used when searching for sessions.
    @details  This class is never instantiated by an application. Use a class that inherits this class.
    @date 2014-02-27  Initial version.
@if CTR_DOC
    @see  nn::pia::local::LocalSessionSearchCriteria
@endif
    @see  nn::pia::inet::NexSessionSearchCriteria
 */
class SessionSearchCriteria : public nn::pia::common::RootObject
{
public:
/*!
    @cond PRIVATE
    @brief  Type of search criteria.
*/
    enum SearchCriteriaKey
    {
        SearchCriteriaKey_Standard = 0,             //!<  Set standard search criteria such as the number of participants.
        SearchCriteriaKey_OwnerPrincipalId,         //!<  Set the principal ID of the owner as the search criterion.
        SearchCriteriaKey_ParticipantPrincipalId,   //!<  Set the principal ID of the participant as the search criterion.
        SearchCriteriaKey_Unused                    //!<  Do not set search criteria.
    };
    //! @endcond

    // The destructor.
    virtual ~SessionSearchCriteria() {};

/*!
    @brief  Specifies the range of search results to get.
    @details  If you specify <tt>0xFFFFFFFF</tt> for the offset only during Internet communication, the search results will be sorted randomly and
     it will get the specified number of entries.
    @param[in] offset  Starting position of the entries to get as search results from the list of entries that match the search criteria.
    @param[in] size  Number of entries to get as search results from the list of entries that match the search criteria.
*/
    virtual void SetResultRange(u32 offset, u32 size)
    {
        m_ResultOffset = offset;
        m_ResultSize = size;
    }

/*!
    @cond PRIVATE
*/
    u32 GetResultRangeOffset() const
    {
        return m_ResultOffset;
    }
    //! @endcond

/*!
    @cond PRIVATE
*/
    u32 GetResultRangeSize() const
    {
        return m_ResultSize;
    }
    //! @endcond

/*!
    @cond PRIVATE
    @brief  Gets the search condition type.
*/
    SearchCriteriaKey GetKeyType() const
    {
        return m_KeyType;
    }
    //! @endcond

/*!
    @brief  Assignment operator.
    @param[in] r  The object to assign.
    @return  See this object.
*/
    SessionSearchCriteria& operator=(const SessionSearchCriteria& r)
    {
        m_KeyType = r.m_KeyType;
        m_ResultOffset = r.m_ResultOffset;
        m_ResultSize = r.m_ResultSize;
        return *this;
    }

    // Prints information that is useful for debugging.
    virtual void Trace(u64 flag) const
    {
        NN_PIA_DUMMY_PARAM(flag);
    }

protected:
    // The Search key.
    SearchCriteriaKey m_KeyType;

    // Search range.
    u32 m_ResultOffset;
    u32 m_ResultSize;

    // The constructor.
    SessionSearchCriteria()
      : m_KeyType(SearchCriteriaKey_Standard),
        m_ResultOffset(0),
        m_ResultSize(20) {};

private:
    // The copy constructor is sealed.
    SessionSearchCriteria(const SessionSearchCriteria&);
};
}
}
}  // end of namespace nn::pia::session
