﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     sync.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/definitions.h>


/*!
@namespace  nn::pia::sync
@brief  Represents the <tt>PiaSync</tt> namespace.

@if CTR_DOC
@date 2013-04-09  Removed the following: <tt>HidSync</tt>, <tt>SyncProtocolOld</tt>, <tt>StationAddressTable</tt>, <tt>ResultInternalStateMismatch</tt>, <tt>ResultSynchronizationIsLost</tt>, <tt>ResultFailedToSynchronize</tt>, and <tt>ResultInvalidDevice</tt>.
@date 2013-01-23  Increased the maximum number of stations available when using the <tt>HidSync</tt> and <tt>SyncProtocolOld</tt> classes.
@endif
@date 2013-01-18  Added <tt>ResultDataIsNotSet</tt> to the <tt>Result</tt> values.

@if CAFE_DOC
@date 2012-12-25  Initial version.
@endif

@if WIN_DOC
@date 2012-12-25  Initial version.
@endif

@if CTR_DOC
@date 2012-10-29  Categorized member functions to improve readability of the API reference.
@date 2012-10-19  Modified comments in the header field in accordance with changes to the format of the API reference.
@date 2012-07-10  Deleted old <tt>Result</tt>-related codes that had been left in the <tt>sync</tt> namespace for compatibility reasons.
@date 2012-05-07  The data sharing feature for the accelerometer and gyro sensor is now officially supported.
@date 2012-05-07  Changed the Japanese spelling of 'sensor' in the Nintendo 3DS Terminology Guide.
@date 2012-04-05  In accordance with the revision of the policy for managing results throughout Pia, the results for the <tt>sync</tt> module have changed, and the references have been made to adhere to those changes.
@date 2012-03-12  The <tt>APP_SPECIFIC_DEVICE</tt> defined by the <tt>DeviceId</tt> enumerated type can now be used by applications as an application-specific device. (Equivalent to the Beta version.)
@date 2012-03-03  Added a caution when sharing gyro sensor data. For more information, see <tt>sync::SyncDeviceProfile</tt>.
@date 2012-02-22  Sharing of accelerometer and gyro sensor data has been implemented on a trial basis. Data sharing is equivalent to an Alpha version. Correct operation is not guaranteed at present.
@date 2011-12-07  The severity of <tt>ResultSyncToBeImplemented</tt> changed from <tt>SEVERITY_SHOULD_BE_HANDLED</tt> to <tt>SEVERITY_PROGRAMMING_ERROR</tt>. The severity of <tt>ResultSyncDataIsNotArrivedYet</tt> changed from <tt>SEVERITY_IGNORABLE</tt> to <tt>SEVERITY_SHOULD_BE_HANDLED</tt>.
@date 2011-11-29  Added <tt>ResultSyncInvalidDevice</tt> to the <tt>Result</tt> values.
@date 2011-11-16  Provided the <tt>sync::ErrorReason</tt> enumerated type.
@date 2011-11-16  Discontinued the original plans to provide specifications for callback during synchronization start. Deleted a callback type definition.
@date 2011-11-10  Added <tt>STATE_MASTER_REINITIALIZING</tt> and <tt>STATE_CLIENT_REINITIALIZING</tt> to the <tt>sync::State</tt> enumeration.
@date 2011-11-10  Added information about the <tt>Initialize</tt> function's return values, which had been previously left out.
@date 2011-10-31  Changed specifications for initialization processing beneath the <tt>sync</tt> module, so that it is now performed during the setup period from the execution of the <tt>sync::BeginSetup</tt> function until the execution of the <tt>sync::EndSetup</tt> function.
@date 2011-10-31  Removed arguments from the <tt>sync::Initialize</tt> function.
@date 2011-10-31  Changed specifications for the <tt>Initialize</tt> function's return value from <tt>void</tt> to <tt>nn::Result</tt>.
@date 2011-10-21  Added documentation.
@date 2011-10-19  Released the alpha version.
@endif
*/
namespace nn
{
namespace pia
{
namespace sync
{
/*! @name  Finish Initialization
    @{
*/

/*!
@brief  Initializes the <tt>PiaSync</tt> module.
Execute the <tt>nn::pia::common::Initialize</tt> function in advance.

@details  Memory that the <tt>sync</tt> module dynamically allocates and uses is allocated from memory passed by the <tt>@ref nn::pia::common::Initialize</tt> function.

@return  Returns a <tt>Result</tt> value indicating success if it is called at the right time.
@retval nn::pia::ResultInvalidState  Indicates that the function has already been executed and initialization has completed.

*/
nn::Result Initialize();

/*!
@brief  Declares the start of setup.

@details  Call this function before initializing the <tt>sync</tt> module.

@attention  This function does not operate correctly if called during another module's setup processing.
After calling the <tt>@ref sync::BeginSetup</tt> function, perform initialization in the <tt>sync</tt> module, and promptly call the <tt>@ref sync::EndSetup</tt> function to complete setup.

@return  Returns a <tt>Result</tt> value indicating success if it is called at the right time.
@retval nn::pia::ResultNotInitialized  Indicates that the <tt>sync::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultInvalidState  Indicates that this function is being called during the setup period. Programming error. Fix your program so that this error is not returned.
*/
nn::Result BeginSetup(void);

/*!
@cond PRIVATE
@brief  Checks whether setup is currently underway.

@details  The application should not call this function. It is provided for internal use.

@return  Returns <tt>true</tt> if this function is called between calls to the <tt>@ref sync::BeginSetup</tt> and <tt>@ref sync::EndSetup</tt> functions.
*/
bool IsSetupNow(void);
/** @endcond */

/*!
@brief  Declares the end of setup.

@details  Call this function after initializing the <tt>sync</tt> module.
Calling this function establishes memory to be consumed by the singleton <tt>sync</tt> module and other modules, and optimizes memory management within the library.

@attention  This function does not operate correctly if it is called during another module's setup processing.
After calling the <tt>@ref BeginSetup</tt> function, perform initialization in the <tt>sync</tt> module, and promptly call the <tt>@ref EndSetup</tt> function to complete setup.

@return  Returns a <tt>Result</tt> value indicating success if it is called at the right time.

@retval nn::pia::ResultNotInitialized  Indicates that the <tt>sync::Initialize</tt> function has not been called. Programming error. Fix your program so that this error is not returned.
@retval nn::pia::ResultInvalidState  Indicates that this function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
*/
nn::Result EndSetup(void);

/*!
@brief  Indicates whether the <tt>PiaSync</tt> module is initialized.

@return  Returns <tt>true</tt> if initialized, and <tt>false</tt> otherwise.
*/
bool IsInitialized(void);

/*!
@brief  Finalizes the <tt>PiaSync</tt> module.
Call this function before calling the <tt>@ref nn::pia::common::Finalize</tt> function.

@details  All processing for the <tt>PiaSync</tt> module must be ended before executing this function.
Calling this function has no effect if the <tt>@ref nn::pia::sync::Initialize</tt> function has not been called.
*/
void Finalize(void);
/*!
end of name Finish Initialization
    @}
*/
}
}
}

#include <pia/sync/sync_SyncProtocol.h>
