﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     sync_SyncProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once


#include <pia/transport/transport_Protocol.h>

#include <pia/transport/transport_ProtocolEvent.h>
#if NN_PIA_TRANSPORT_V3
#include <pia/common/common_PacketDefine.h>
#else
#include <pia/common/common_ProtocolMessageOld.h>
#endif
#include <pia/transport/transport_definitions.h>


// Forward declaration.
namespace nn
{
namespace pia
{
namespace common
{
class ZlibCompressor;
class ZlibUncompressor;
}
}
}


namespace nn
{
namespace pia
{
namespace sync
{

/*!
@brief  This protocol is used for synchronized communication. <br/>
Full mesh is the only network topology type that is supported. Note that star and relay mesh types are not supported.
@date 2014-01-30  Added constants to specify the compression level of synchronization data, and made any other compression levels unsupported.
@date 2013-11-22  Changed internal processes and reduced the amount of synchronization data sent.
@date 2013-11-22  Enabled a skip feature when sending synchronized data on the same frame by default and removed the configuration parameter (<tt>Setting::m_SendSkipMax</tt>).
@date 2013-10-25  Added the <tt>EndAlone</tt> function for unilaterally ending synchronization.
@date 2013-10-25  Added a feature for compressing synchronized data.
@date 2013-10-25  Added a skip feature when sending synchronized data on the same frame.
@date 2013-08-30  Changed the message headers. This breaks backward compatibility.
@date 2013-08-26  Removed the <tt>Session::m_SendPeriod</tt> function. Set the same parameters by calling <tt>Start</tt>.
@date 2013-08-26  Removed <tt>Start(bit32 usingDataIdBitmap)</tt> and <tt>Start(bit32 usingDataIdBitmap, u32 delay)</tt>. Use <tt>Start(bit32 usingDataIdBitmap, u32 delay, u32 sendPeriod)</tt> when synchronized communication starts.
@date 2013-05-20  Promoted the feature to change delay during synchronization to the officially supported version.
@date 2013-04-12  Added a function to calculate the maximum data size that would fit into a single packet of synchronization data to be sent.
@date 2013-04-12  Added a feature to change the delay after initialization. (The feature to change the delay during synchronization is still in beta.)
@date 2013-04-12  Changed the type of the <tt>Setting</tt> member.
@date 2013-04-12  Changed the number of Data IDs that can be configured.
@date 2013-02-19  Renamed <tt>STATE_READY</tt> to <tt>STATE_WAITING</tt>.
@date 2013-02-15  Added a function to get the settings specified in the <tt>Initialize</tt> function.
@date 2013-02-15  Changed the type of the <tt>Setting</tt> member.
@date 2013-01-18  Greatly revised the state transitions, and changed the operations of each function.
@date 2013-01-18  It is now possible to send large files.
@date 2012-12-25  Initial version.
@if CTR_DOC
@date 2012-12-25  Renamed the existing <tt>SyncProtocol</tt> class to <tt>SyncProtocolOld</tt>.
@endif

*/
class SyncProtocol : public transport::Protocol
{
public:
/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
@details  Applications should not call this constructor directly.
Use the <tt>@ref transport::Transport::CreateProtocol</tt> function to create an instance.

*/
    SyncProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destructor.
@details  Do not call this destructor directly from applications.
Use the <tt>@ref transport::Transport::DestroyProtocol</tt> function to free an instance.

*/
    virtual ~SyncProtocol(void);
    //! @endcond


    PIA_PROTOCOL_TYPE_INFO(nn::pia::transport::PROTOCOL_TYPE_SYNC);


    //! Defines the maximum input latency specified that can be specified for <tt>@ref Setting::m_MaxDelay</tt>, as a number of frames.
    static const u8 FRAME_DELAY_MAX = 32;

    //! Defines the number of data IDs that can be set.
    static const u32 DATA_ID_NUM = 16;

    //! Defines the reserved data ID.
    static const u32 RESERVED_DATA_ID_NUM = 1;
    static const u32 RESERVED_DATA_ID_CHANGE_DELAY = DATA_ID_NUM + 0;

    static const u32 SEND_SKIP_MAX = 2;
    static const u32 DATA_FIRST_SEND_COUNT = 3;

    static const u32 COMPRESSION_LEVEL_NONE = 0;    //!< Specifies the compression level (no compression) for the synchronization data specified in <tt>@ref Setting::m_DataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_LOW = 1;     //!< Specifies the compression level (speed emphasized) for the synchronization data specified in <tt>@ref Setting::m_DataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_MIDDLE = 5;  //!< Specifies the compression level for the synchronization data specified in <tt>@ref Setting::m_DataCompressionLevel</tt>.
    static const u32 COMPRESSION_LEVEL_HIGH = 9;    //!< Specifies the compression level (compression rate emphasized) for the synchronization data specified in <tt>@ref Setting::m_DataCompressionLevel</tt>.

/*!
@brief  Stores settings specified during initialization.
@details  Specify the same values on stations using data synchronization.
*/
    struct Setting
    {
    public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
@details  This function initializes the object using default values.
*/
        Setting();

/*!
@brief  Sets the size of the data to send for each data ID.
@details  Specify <tt>0</tt> for unused data IDs.
You can use the <tt>@ref SyncProtocol::GetDataUnitSizeMax</tt> function to find out the largest size this value can be set to.
@see  SyncProtocol::GetDataUnitSizeMax
*/
        size_t m_DataUnitSize[DATA_ID_NUM + RESERVED_DATA_ID_NUM];

/*!
@brief  Defines the maximum number of frames of input delay.
@details  Specify a value between 1 and <tt>FRAME_DELAY_MAX</tt>.
@if CTR_DOC
@details  <b>Note:</b> To set a value equivalent to the delays set in the <tt>HidSync</tt> and <tt>SyncProtocolOld</tt> classes, you must specify a value 1 greater than the values set in those classes.
@endif
*/
        u32 m_MaxDelay;

/*!
@brief  Ends synchronization due to a timeout if packets cannot be received consecutively for the specified number of frames (number of calls to <tt>@ref SyncProtocol::Step</tt>).
@details  If you specify 0, the object continues to wait for packets forever, and data synchronization never times out.
*/
        u32 m_TimeoutFrame;

/*!
@brief  Enables or disables use of the change delay feature while communication is synchronized.
*/
        bool m_IsUsingChangeDelay;

/*!
@brief  Sets the compression level for synchronized data.
@details  Not compressed when <tt>@ref SyncProtocol::COMPRESSION_LEVEL_NONE</tt> is specified.
Specifying <tt>@ref SyncProtocol::COMPRESSION_LEVEL_LOW</tt> emphasizes speed; <tt>@ref SyncProtocol::COMPRESSION_LEVEL_HIGH</tt> emphasizes the compression rate.
Specifying <tt>@ref SyncProtocol::COMPRESSION_LEVEL_MIDDLE</tt> results in an intermediate level of compression between the other two.
The default value is <tt>@ref SyncProtocol::COMPRESSION_LEVEL_NONE</tt>.
Other values are not supported.
Execution time varies significantly based on the CPU performance of the platform.
*/
        u32 m_DataCompressionLevel;
    };


/*!
@brief  Gets the maximum number of bytes that can be sent in one data unit.
@details  The value obtained depends on the parameter specified when initializing the <tt>transport::Transport</tt> instance.
@see  Returns the maximum number of bytes that can be sent in one data unit.
@see  Setting::m_DataUnitSize



*/
    size_t GetDataUnitSizeMax() const;


/*!
@brief  Initializes the instance. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.

@param[in] setting  Specifies various settings.
Specify the same values on all stations using data synchronization. Operations are undefined when different values are specified.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@see  Finalize


*/
    nn::Result Initialize(const Setting& setting);


/*!
@brief  Finalizes the instance.
@see  Initialize

*/
    void Finalize();


/*!
@brief  Gets the size of data having the data ID specified to the <tt>Initialize</tt> function.
@param[in] dataId  Specifies the data ID of the data being checked.
@return  Returns the size of the specified data.


*/
    size_t GetDataSize(u32 dataId) const;


/*!
@brief  Gets the maximum input delay specified to the <tt>Initialize</tt> function.
@return  Returns the maximum number of frames to delay input.

*/
    u32 GetMaxDelay() const
    {
        return static_cast<u32>(m_MaxDelay);
    }

/*!
@brief  Gets the input delay actually specified.
@return  Returns the number of input delay frames.

*/
    u32 GetDelay() const
    {
        return static_cast<u32>(m_Delay);
    }


/*!
@brief  Gets the timeout setting specified in the <tt>Initialize</tt> function.
@return  Returns the timeout setting.

*/
    u32 GetTimeoutFrame() const
    {
        return static_cast<u32>(m_TimeoutFrame);
    }


/*!
@brief  Gets the send interval specified to the <tt>Start</tt> function.
@return  Returns the send interval.

*/
    u32 GetSendPeriod() const
    {
        return m_SendPeriod;
    }

/*!
@brief  Sets the send interval.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.


*/
    nn::Result SetSendPeriod(u32 sendPeriod);


/*!
@brief  Advances synchronization by one frame.
@details  Call this function during each frame when the state is other than <tt>STATE_NOT_SYNCHRONIZED</tt>. Call this function at the beginning of each frame.
When <tt>@ref NeedSetData</tt> returns <tt>true</tt>, specify the data to send to the <tt>@ref SetData</tt> function before calling this function.
This function can also be called when the state is <tt>STATE_NOT_SYNCHRONIZED</tt>.
@ref EndAlone After being called successfully, this function fails when called during individual synchronization finalization processes. It must be called every frame to keep the internal processing going.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that communication is not active. Programming error. Fix your program so that this error is not returned.
@retval ResultDataIsNotArrivedYet  Indicates that the frame was not advanced because the data being synchronized has not arrived from other stations.
Try calling this function again after calling the <tt>common::Scheduler::Dispatch</tt> function. Handle appropriately in the application.
@retval ResultDataIsNotSet  Indicates that the local station has not yet set the data to send. Programming error. Fix your program so that this error is not returned.
@retval ResultTemporaryUnavailable  An individual synchronization finalization process is in progress.
@see  SetData






*/
    nn::Result Step();


/*!
@brief  Starts synchronization.
@details  If this function is successfully called when the state is <tt>STATE_NOT_SYNCHRONIZED</tt>, the state transitions to <tt>STATE_WAITING</tt>.
If a network node starts synchronization after this function is called, the state transitions to <tt>STATE_STARTING</tt> and <tt>STATE_SYNCHRONIZED</tt>.
@param[in] usingDataIdBitmap  Specifies a bit flag representing the data IDs (among those specified by <tt>Initialize</tt>) to send from this station during this round of synchronized communication.
Bits are ordered from low to high, with the lowest-order bit corresponding to data ID 0.
You can specify separate values for each station.

@param[in] delay  Specifies the delay requested by this station.
The largest delay value among stations that specify a delay is actually used.
This delay value must be less than or equal to the delay specified during initialization.
@param[in] sendPeriod  Specifies the send/resend interval for the configured data.
Each time that <tt>SyncProtocol::Step</tt> is called the number of times specified by this parameter, send/resend is executed when <tt>common::Scheduler::Dispatch</tt> is called.
Specify a value of <tt>1</tt> or greater.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@see  End




*/
    nn::Result Start(bit32 usingDataIdBitmap, u32 delay, u32 sendPeriod);


/*!
@brief  Ends synchronized communication.
@details  If this function can be called when the state is <tt>STATE_WAITING</tt>, <tt>STATE_STARTING</tt>, or <tt>STATE_SYNCHRONIZED</tt>, and the function succeeds, the state transitions to <tt>STATE_ENDING</tt>.
When network nodes are subsequently notified that communication has ended, the state transitions to <tt>STATE_NOT_SYNCHRONIZED</tt>.
If you use this function to end synchronization, normally the reason for the end of communication is <tt>END_REASON_END</tt> for the local station and <tt>END_REASON_OTHER_STATION</tt> for other stations.
However, these are not necessarily the values if communication is simultaneously terminated by different nodes for different reasons.
Immediately after this function is called, you must call the <tt>Step</tt> function for several frames until <tt>STATE_NOT_SYNCHRONIZED</tt> results before calling the <tt>Start</tt> function again.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@see  Start





*/
    nn::Result End();

/*!
@brief  Unilaterally ends synchronization for only the local station.
@details  This function can be called while the state is <tt>STATE_SYNCHRONIZED</tt>.
<br/>
This function can get the synchronization data for each frame. If there is a notification that an individual synchronization finalization has been done on another station after this function has been called, however, the synchronization data that station sent cannot be obtained. This can lead to the data getting out of sync, so make sure to call this function after getting the received synchronization data.
The end frame is set to the last frame for which send data has already been set.
The peers are notified of the end frame and the local station transitions to the <tt>STATE_ENDED_ALONE</tt> state after confirming that all peers have reached the end frame.
After this function is called successfully, <tt>@ref Step</tt> fails while the state is <tt>STATE_SYNCHRONIZED</tt>. Note that the frame can no longer advance when this happens, so the local station never reaches the end frame.
Nonetheless, the <tt>@ref Step</tt> function must still be called in every frame to advance internal processes.
If you use this function to end synchronization, normally the reason for the end of communication is <tt>END_REASON_END_ALONE</tt> for the local station.
However, these are not necessarily the values if communication is simultaneously terminated by different nodes for different reasons.
Once synchronization ends with all peers while in the <tt>STATE_ENDED_ALONE</tt> state, the state transitions to <tt>STATE_NOT_SYNCHRONIZED</tt>.
Also, the other stations can maintain synchronization even if the local station leaves the session in the <tt>STATE_ENDED_ALONE</tt> state, or after transitioning to the subsequent <tt>STATE_NOT_SYNCHRONIZED</tt> state.
However, if the local station disconnects in the <tt>STATE_WAITING</tt> state because a new station connected after the transition to the <tt>STATE_ENDED_ALONE</tt> state and the other stations attempted synchronization again, the other stations end synchronization and transition to the <tt>STATE_NOT_SYNCHRONIZED</tt> state.
Also, stations that ended in the <tt>STATE_ENDED_ALONE</tt> state can restart synchronization after transitioning to the <tt>STATE_NOT_SYNCHRONIZED</tt> state, but if those stations restart synchronization or leave the session, the stations other than the local station cannot be synchronized.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. or that the end frame has already been set. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
















*/
    nn::Result EndAlone();

/*!
@brief  Gets the end frame for the specified station.
@details  This function can be called when the state is something other than <tt>STATE_NOT_SYNCHRONIZED</tt> or <tt>STATE_WAITING</tt>.
If synchronization ends before the <tt>STATE_SYNCHRONIZED</tt> state is reached, <tt>endFrame</tt> becomes <tt>0</tt>, but note that synchronization actually ends before that.
Also note that the station ending synchronization may end before reaching this frame.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates that the end frame is not set. Handle appropriately in the application.



*/
    nn::Result GetEndFrameNo(StationId stationId, u32* endFrameNo) const;

/*!
@brief  Represents the current state of synchronization.
*/
    enum State
    {
        STATE_NOT_SYNCHRONIZED = 0,  //!< Synchronization is not active.
        STATE_WAITING,               //!< Start synchronization and wait for other stations to start synchronization.
        STATE_STARTING,              //!< Synchronization is starting. Although the data to send must be set, it cannot yet be retrieved.
        STATE_SYNCHRONIZED,          //!< Synchronization is active. The data to send must be set, and it can be retrieved.
        STATE_ENDING,                //!< Synchronization is ending.
        STATE_ENDED_ALONE            //!< The local station unilaterally ended synchronization.
    };


/*!
@brief  Gets the current state of synchronization.
@return  Returns the current state of synchronization.

*/
    State GetState() const
    {
        return m_State;
    }


/*!
@brief  Gets whether it is necessary to set the data to send.
@details  You must set the data to send during <tt>STATE_STARTING</tt> and <tt>STATE_SYNCHRONIZED</tt>.
If the return value of a call to <tt>Step</tt> indicates that execution failed, there is no need to reconfigure send data before the next call to <tt>Step</tt> because the frame has not advanced.
@return  Returns <tt>true</tt> if any of the required data IDs are not set.
Returns <tt>false</tt> if it is not necessary to set send data, or the data to send has already been set.
@see  SetData




*/
    bool NeedSetData() const;


/*!
@brief  Gets whether it is necessary to set the data to send for the specified data ID.
@details  You must set the data to send during <tt>STATE_STARTING</tt> and <tt>STATE_SYNCHRONIZED</tt>.
When the return value of a call to <tt>Step</tt> indicates that execution failed, you do not need to reconfigure send data before the next call to <tt>Step</tt> because the frame has not advanced.
@param[in] dataId  Specifies the data ID to check whether it is necessary to send data for.
@return  Returns <tt>true</tt> when send data needs to be set for the specified data ID.
Returns <tt>false</tt> if it is not necessary to set send data, or the data to send has already been set.
@see  SetData





*/
    bool NeedSetData(u32 dataId) const;


/*!
@brief  Gets whether received data can be retrieved.
@details  Received data can be retrieved while the state is <tt>STATE_SYNCHRONIZED</tt>.
However, when you are using the feature to change the delay while communication is synchronized, use <tt>GetIsNoDataFrame</tt> to check whether data can be retrieved during the current frame. During some frames you cannot retrieve data even if this function returns <tt>true</tt>.
@return  Returns <tt>true</tt> if received data can be retrieved.
@see  GetData




*/
    bool CanGetData() const;


/*!
@brief  Gets the number of frames since synchronization started.
@details  While the state is <tt>STATE_SYNCHRONIZED</tt>, every successful call to the <tt>Step</tt> function advances one frame.
@return  Returns the number of frames since synchronization started.
<b>Note:</b> This function returns <tt>0</tt> if synchronization is not active, and also returns <tt>0</tt> on the first frame when synchronization starts.


*/
    u32 GetFrameNo() const;


/*!
@brief  Gets whether the specified station has joined synchronization.
@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station being checked.
@return  Returns <tt>true</tt> if the specified station has joined synchronization.


*/
    bool CheckEntry(StationId stationId) const;


/*!
@brief  Gets a bit flag representing the list of data IDs that the specified station is sending.
@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station being checked.
@return  Returns a list of data IDs that the specified station is sending.
<tt>0</tt> is returned for stations not engaging in synchronization (stations for which <tt>@ref CheckEntry</tt> returns <tt>false</tt>).


*/
    bit32 GetUsingDataIdBitmap(StationId stationId) const;


/*!
@brief  Represents the reason why synchronization was ended.
*/
    enum EndReason
    {
        END_REASON_NOTHING = 0,        //!< Synchronization has not ended.
        END_REASON_END,                //!< Communication ended because of a call to the <tt>@ref End</tt> function by the station.
        END_REASON_OTHER_STATION,      //!< Communication ended because another station ended synchronization.
        END_REASON_CHANGE_CONNECTION,  //!< Communication ended because there was a change in the stations in the session.
        END_REASON_TIMEOUT,            //!< Communication ended because no data was received for <tt>@ref Setting::m_TimeoutFrame</tt> frames.
        END_REASON_END_ALONE           //!< Communication ended because of a call to the <tt>@ref EndAlone</tt> function by the station.
    };


/*!
@brief  Gets the reason why synchronization was ended.
@details  This value is reset if the <tt>@ref Start</tt> function is called to start synchronization again.
@return  Returns the reason why synchronization ended.


*/
    EndReason GetLastEndReason() const;


/*!
@brief  Sets data to send during synchronization.
@details  If the <tt>NeedSetData</tt> function returns <tt>true</tt>, you must set data before calling the <tt>Step</tt> function.

@param[in] dataId  Specifies the data ID of data being set.
@param[in] cpData  Specifies the data being set. The size is specified in the call to the <tt>Initialize</tt> function.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the data has already been set. Programming error. Fix your program so that this error is not returned.
@see  NeedSetData, Step



*/
    nn::Result SetData(u32 dataId, const void* cpData);


/*!
@brief  Gets the data received during synchronization.
@details  If the <tt>CanGetData</tt> and <tt>CheckEntry</tt> functions return <tt>true</tt> for a station, it can retrieve data for data IDs that have bits set in the bitmap returned from the <tt>GetUsingDataIdBitmap</tt> function.
If you use the feature to change the delay while communication is synchronized, data cannot be retrieved in some frames. The function returns <tt>ResultInvalidState</tt> in these frames.
Use the <tt>@ref GetIsNoDataFrame</tt> function to determine whether data can be retrieved in this frame.

@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station that is sending the data to retrieve.
@param[in] dataId  Specifies the data ID of the data to retrieve.
@param[out] pBuffer  Specifies a pointer to a buffer to store the data. It must be the size specified in the <tt>Initialize</tt> function.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Alternatively, it could indicate that data cannot be retrieved in this frame. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.


*/
    nn::Result GetData(StationId stationId, u32 dataId, void* pBuffer) const
    {
        return this->GetData(stationId, dataId, pBuffer, 0);
    }


/*!
@cond PRIVATE
@brief  Gets data for a future frame received via synchronized communication.
@details  If the <tt>CanGetData</tt> and <tt>CheckEntry</tt> functions return <tt>true</tt> for a station, it can retrieve data for data IDs that have bits set in the bitmap returned from the <tt>GetUsingDataIdBitmap</tt> function.
You can specify a frame offset. This is the offset from the current frame to the frame for which you want to retrieve synchronization data.
If you use the feature to change the delay while communication is synchronized, data cannot be retrieved in some frames. The function returns <tt>ResultInvalidState</tt> in these frames.
Use the <tt>@ref GetIsNoDataFrame</tt> function to determine whether data can be retrieved in this frame.
The function returns <tt>ResultNoData</tt> if no data has been received for the specified future frame.

@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station that is sending the data to retrieve.
@param[in] dataId  Specifies the data ID of the data to retrieve.
@param[out] pBuffer  Specifies a pointer to a buffer to store the data. It must be the size specified in the <tt>Initialize</tt> function.
@param[in] frameOffset  Specifies the offset from the current frame to the frame you want to retrieve synchronization data for.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that this function cannot be called in the current <tt>@ref nn::pia::sync::SyncProtocol::State "State"</tt>. Alternatively, it could indicate that data cannot be retrieved in this frame. Or, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates that there is no data. Handle appropriately in the application.


*/
    nn::Result GetData(StationId stationId, u32 dataId, void* pBuffer, u32 frameOffset) const;
    //! @endcond


/*!
@brief  Gets a pointer to the data received during synchronization.
@details  If the <tt>CanGetData</tt> and <tt>CheckEntry</tt> functions return <tt>true</tt> for a station, it can retrieve data for data IDs that have bits set in the bitmap returned from the <tt>GetUsingDataIdBitmap</tt> function.
The retrieved pointer is valid until the next call to the <tt> Step</tt> function.
It has the size specified in the <tt>Initialize</tt> function.

@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station that is sending the data to retrieve.
@param[in] dataId  Specifies the data ID of the data to retrieve.
@return  Returns a pointer to data received during synchronization.
This function returns <tt>NULL</tt> if there is no available data.


*/
    const void* GetDataPtr(StationId stationId, u32 dataId) const;


/*!
@cond PRIVATE
@brief  Gets the number of frames from the current frame for which the local station has received synchronization data from the specified station.
@details  The number for the available frames that have received data does not include the current frame.
If you specify the ID of the local station in <tt>stationId</tt>, this function returns the value <tt>configured delay - 1</tt>.
The function returns zero if the <tt>stationId</tt> value is invalid, or if the <tt>GetUsingDataIdBitmap</tt> function returns zero for that station.

@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station that sent the received data.
@return  Returns the amount of synchronization data for future frames received from the specified station.


*/
    u32 GetReceivedDataNum(StationId stationId) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the number of frames that the local station is ahead from the perspective of the specified station.
@details  Estimates the number of frames that the local station is ahead from the perspective of the specified station. The function bases its estimate on the number of frames of data received, and the RTT.
If the local station is ahead, the function returns a positive value. It returns a negative value if the local station is behind.
The actual number of frames that the local station is ahead or behind may be different from this value, for example if some data was not received due to packet loss.
Additionally, the numbers of frames ahead estimated by each station do not necessarily match up.
@param[in] stationId  Specifies the <tt>@ref StationId</tt> of the station with which to compare frame counts.
@param[in] frameRate  Specifies the frame rate of the application.
@return  Returns the number of frames that the local station is ahead from the perspective of the specified station.



*/
    s32 EstimateLeadingFrameNum(StationId stationId, u32 frameRate) const;
    //! @endcond

/*!
@brief  Sets a change delay request as synchronization data.
@details  The requested delay must be no larger than the delay specified at initialization.
If you request a delay of zero, the library assumes that you have not requested a change.
Even if this function succeeds, the delay is not guaranteed to change to the requested value.
If more than one request is received in the same frame, the largest requested delay value gets precedence.
If the delay is already being changed in response to a request when another delay-change request arrives as synchronization data, the newly arrived request is ignored.
@param[in] newDelay  Specifies the requested delay value.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called in an invalid state. Alternatively, it indicates that stations are not communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that the data has already been set. Programming error. Fix your program so that this error is not returned.




*/
    nn::Result RequestToChangeDelay(u32 newDelay);

/*!
@brief  Gets whether synchronization data cannot be retrieved for this frame.
@details  If you are using the feature to change the delay while communication is synchronized, use this function to check whether data can be retrieved in the current frame. Data cannot be retrieved for some frames when you use this feature.
You are not required to use this function if you are not using the function for changing the delay while communication is synchronized.
@param[out] isNoDataFrame  Specifies whether synchronization data is available in this frame.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called in an invalid state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.




*/
    nn::Result GetIsNoDataFrame(bool* isNoDataFrame) const
    {
        return this->GetIsNoDataFrame(isNoDataFrame, 0);
    }

/*!
@cond PRIVATE
@brief  Gets whether synchronization data cannot be retrieved for this frame.
@details  If you are using the feature to change the delay while communication is synchronized, use this function to check whether data can be retrieved in the current frame. Data cannot be retrieved for some frames when you use this feature.
You are not required to use this function if you are not using the function for changing the delay while communication is synchronized.
@param[out] isNoDataFrame  Specifies whether synchronization data is available in this frame.
@param[in] frameOffset  Specifies the offset from the current frame.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called in an invalid state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.




*/
    nn::Result GetIsNoDataFrame(bool* isNoDataFrame, u32 frameOffset) const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the preconfigured maximum frame offset beyond which synchronization data cannot be retrieved.
@details  There is a predetermined limit to the number of frames after the current one for which synchronization data can be retrieved. This function returns that value as an offset from the current frame.
@return  Returns the preconfigured maximum frame offset beyond which synchronization data cannot be retrieved.


*/
    u32 GetMaxFrameOffsetDecidedIfNoDataFrameOrNot() const;
    //! @endcond

/*!
@brief  Calculates the maximum total size of data that can be sent as a single packet of synchronization data, given the specified number of data IDs.
@details  This function calculates the maximum amount of synchronization data that can be sent. This maximum amount of data is the value to use when you need to send (delay * 2) frames of synchronization data.
@param[in] delay  Specifies the delay value.
@param[in] dataIdNum  Specifies the number of data IDs that you are using.
@return  Returns the maximum total size of data that can be sent as a single packet of synchronization data, given the specified number of data IDs.



*/
    size_t CalcMaxSyncDataSizeSentInOnePacket(u32 delay, u32 dataIdNum);


/*!
@cond PRIVATE
@brief  Advances to the next frame without setting data to send when a frame is dropped.
@details  When a frame is dropped, call this function instead of the <tt>SyncProtocol::Step</tt> function. On success, you can advance one frame without setting data to send. The function also decrements by 1 the offset of the frame for which data being sent is set. Execution fails if this function is called in any state other than <tt>STATE_SYNCHRONIZED</tt>. This cannot be used at the same time as the feature for changing the delay during synchronization.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
On failure, returns the same <tt>Result</tt> value as the <tt>SyncProtocol::Step</tt> function, or <tt>ResultInvalidState</tt>.
@retval ResultInvalidState  Indicates that the frame offset for the input delay frame count plus the data being sent is less than <tt>1</tt>, or that this function was not called in a <tt>@ref nn::pia::sync::SyncProtocol::State</tt> state. Programming error. Fix your program so that this error is not returned.



*/
    nn::Result StepWithoutSetData();
    //! @endcond


/*!
@cond PRIVATE
@brief  Adjusts the frame offset for setting send data.
@details  Call this after setting the data to send when skipping a frame or when processing slows. On success, increments by 1 the offset of the frame for which to set send data. Execution fails if this function is called in any state other than <tt>STATE_SYNCHRONIZED</tt>. This cannot be used at the same time as the feature for changing the delay during synchronization.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the frame offset for the input delay frame count plus the data being sent is greater than the maximum input frame delay, or that this function was not called in a <tt>@ref nn::pia::sync::SyncProtocol::State</tt> state. Programming error. Fix your program so that this error is not returned.
@retval ResultDataIsNotSet  Indicates that the local station has not yet set the data to send. Programming error. Fix your program so that this error is not returned.



*/
    nn::Result EndSetDataInNoProcessFrame();
    //! @endcond


/*!
@cond PRIVATE
@brief  Indicates the offset of the frame for which to set send data.
*/
    s32 GetSetDataOffset()
    {
        return m_SetDataOffset;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Specifies whether to use the feature for checking if data is already received and suppressing resends.
*/
    void SetIsUsingReceivedDataCheck(bool isUsingReceivedDataCheck);
    //! @endcond


/*!
@cond PRIVATE
@brief  Starts communication.
@param[in] myStationIndex  The <tt>StationIndex</tt> value of the local station. Specify a value in the range from <tt>STATION_INDEX_1</tt> to <span class="argument">maxConnections</span>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultNotInitialized  Indicates that the instance is not initialized.
@retval ResultInvalidState  Indicates that stations are already communicating.
@retval ResultInvalidArgument  Indicates that an argument is invalid.



*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  The main process.
@return  Returns <tt>nn::ResultSuccess</tt>.

*/
    virtual nn::Result Dispatch();
    //! @endcond


/*!
@cond PRIVATE
@brief  Switches the connection status.
@param[in] event  Specifies the event representing the operation.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called in an invalid state.
@retval ResultInvalidArgument  Indicates that an argument is invalid.



*/
    virtual nn::Result UpdateProtocolEvent(const transport::ProtocolEvent& event);
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


private:
    struct StationInfo
    {
    public:
        void Clear();

    public:
        bool m_IsConnect;
        bool m_IsSync;

        u16 m_SyncId;
        u32 m_UsingDataIdBitmap;

        u32 m_PreparedFrameNo;
        bit32 m_ReceivedDataBitmap;

        u32 m_LastReceiveStepCount;

        bool m_IsEndAlone;
        u32 m_EndFrameNo;
        bool m_IsEndAloneInCurrentSession;
    };

    bool IsInitialized() const
    {
        return (m_StationNum > 0);
    }
    bool IsInCommunication() const
    {
        return (m_LocalStationIndex != STATION_INDEX_INVALID);
    }

    void CheckTimeout();

    void DispatchReceive();
    void ReceiveStartMessage(StationIndex stationIndex, u16 syncId, u32 usingDataIdBitmap, s32 delay, bit16 connectedStationBitmap);
    void ReceiveEndMessage(StationIndex stationIndex, u16 syncId);
    void ReceiveStartEndMessage(StationIndex stationIndex, u16 syncId);
    void ReceiveEndAloneMessage(StationIndex stationIndex, u16 syncId, u32 endFrameNo);
    void ReceiveStartAckMessage(StationIndex stationIndex, u16 syncId);
    void ReceiveEndAckMessage(StationIndex stationIndex, u16 syncId);
    void ReceiveEndAloneAckMessage(StationIndex stationIndex, u16 syncId);
    void ReceiveDataMessage(StationIndex stationIndex, u16 syncId, const void* cpMessage, size_t size, bool isCompressedData);

    void DispatchSend();
    bool SendCommand(StationIndex stationIndex, int type, u16 syncId = 0);
    bool SendCommandMulti(u32 destBmp, int type);
    void SendData();
    void ResetSendDataFrameNo();
    bool CreateDataChunk(size_t* pSize, bool forceDisableDataCompression = false);

    void UpdatePreparedFrameNo();
    void UpdateReceivedDataBitmap();
    bool IsAllReceivedData(u32 sendDataFrameNo);

    u64 MakeSendFrameBitmap(u32 dataFirstSendCount) const;
    bool IsSendFrameNo(u32 sendDataFrameNo);
    u32* GetSendProcessCountPtr(u32 frameNo)
    {
        return &m_paSendProcessCnt[GetDataBufferIdx(m_LocalStationIndex, frameNo)];
    }

    StationInfo* GetStationInfo(StationIndex stationIndex);
    const StationInfo* GetStationInfo(StationIndex stationIndex) const
    {
        return const_cast<SyncProtocol*>(this)->GetStationInfo(stationIndex);
    }

    int GetDataBufferIdx(StationIndex stationIndex, u32 frameNo) const;

    u32* GetValidDataFlag(StationIndex stationIndex, u32 frameNo)
    {
        return &m_paValidDataFlag[stationIndex * m_BufferFrameNum + GetDataBufferIdx(stationIndex, frameNo)];
    }
    const u32* GetValidDataFlag(StationIndex stationIndex, u32 frameNo) const
    {
        return const_cast<SyncProtocol*>(this)->GetValidDataFlag(stationIndex, frameNo);
    }

    u32* GetNoDataFrameFlag(u32 frameNo)
    {
        return &m_paNoDataFrameFlag[GetDataBufferIdx(m_LocalStationIndex, frameNo)];
    }
    const u32* GetNoDataFrameFlag(u32 frameNo) const
    {
        return const_cast<SyncProtocol*>(this)->GetNoDataFrameFlag(frameNo);
    }

    void* GetDataBuffer(StationIndex stationIndex, u32 frameNo, u32 dataId)
    {
        PIA_ASSERT(dataId < DATA_ID_NUM + RESERVED_DATA_ID_NUM);
        return &m_pDataBuffer[(stationIndex * m_BufferFrameNum + GetDataBufferIdx(stationIndex, frameNo)) * m_TotalDataSize + m_DataInfo[dataId].m_Offset];
    }
    const void* GetDataBuffer(StationIndex stationIndex, u32 frameNo, u32 dataId) const
    {
        return const_cast<SyncProtocol*>(this)->GetDataBuffer(stationIndex, frameNo, dataId);
    }

    // Dynamically change the delay.
    nn::Result ChangeDelay(u32 newDelay);
    bool IsUsingChangeDelay() const
    {
        return m_DataInfo[RESERVED_DATA_ID_CHANGE_DELAY].m_Size != 0;
    }

    bool EndImpl(EndReason reason);

    void CheckEndedAlone();

    bit16 CalcConnectedStationBitmap() const;

    // For the <tt>StationIndex</tt>.
    bool CheckEntry(StationIndex stationIndex) const;
    nn::Result GetData(StationIndex stationIndex, u32 dataId, void* pBuffer) const
    {
        return this->GetData(stationIndex, dataId, pBuffer, 0);
    }
    nn::Result GetData(StationIndex stationIndex, u32 dataId, void* pBuffer, u32 frameOffset) const;
    bit32 GetUsingDataIdBitmap(StationIndex stationIndex) const;

private:
    u32 m_StationNum;
    StationIndex m_LocalStationIndex;

    State m_State;

#if NN_PIA_TRANSPORT_V3
    static const u32 cMessageBufferSize = common::PROTOCOL_MESSGAE_PAYLOAD_SIZE_MAX / sizeof(u32) * sizeof(u32);
#else
    static const u32 cMessageBufferSize = common::ProtocolMessageOld::MESSAGE_SIZE_MAX / sizeof(u32) * sizeof(u32);
#endif
    u32 m_MessageBuffer[cMessageBufferSize / sizeof(u32)];
    u32* m_pCompressedMessageBuffer;

    StationInfo* m_paStationInfo;
    u32* m_paValidDataFlag;
    u32* m_paNoDataFrameFlag;
    u8* m_pDataBuffer;
    s32* m_pDelayBuffer;

    struct
    {
        u32 m_Offset;
        u32 m_Size;
    } m_DataInfo[DATA_ID_NUM + RESERVED_DATA_ID_NUM];

    u32 m_TotalDataSize;

    s32 m_Delay;
    s32 m_MaxDelay;
    s32 m_BufferFrameNum;
    u32 m_SendPeriod;
    s32 m_TimeoutFrame;

    u32 m_CurrentFrameNo;
    u32 m_PreparedFrameNo;
    bit32 m_ReceivedDataBitmap;  // In the frames after <span class="argument">m_PreparedFrameNo</span> + 1, data has already been received from all stations for this frame.
    bool m_IsUsingReceivedDataCheck;

    s32 m_CurrentDataBufferIdx;

    EndReason m_LastEndReason;

    u32 m_StepCount;
    u32 m_StepCountOffset;
    bool m_IsSendStart;
    bool m_IsSendEnd;
    bool m_IsSendEndAlone;
    u32 m_ParticipantBmp;
    u32 m_SendStartDestBmp;
    u32 m_SendEndDestBmp;
    u32 m_SendEndAloneDestBmp;
    u32 m_SendDataFrameNo;
    u32 m_SendDataId;
    bool m_IsResetSendData;

    // For dynamically changing the delay.
    s32 m_NewDelay;
    u32 m_ChangeDelayRequestFrameNo;

    // Offset of the frame for which to set send data.
    s32 m_SetDataOffset;

    // Flag for stepping without checking whether data to send has been set.
    bool m_IsStepWithoutSetData;

    // To compress and uncompress data:
    bool m_IsEnableDataCompression;
    u32 m_DataCompressionLevel;
    static const u32 cZlibCompressorWindowBits = 12;
    static const u32 cZlibCompressorMemLevel = 5;
    static const u32 cZlibCompressorWorkMemorySize = 38596;
    common::ZlibCompressor* m_pZlibCompressor;
    u32* m_paZlibCompressorMemory;
    common::ZlibUncompressor* m_pZlibUncompressor;
    u32* m_paZlibUncompressorMemory;

    // To skip sending the same data:
    u32 m_SendSkipMax;
    u32 m_SendSkip;
    u32 m_SameDataCount;
    u32 m_PreSendFrameNoMin;
    u32 m_PreSendFrameNoMax;

    // Determines whether to send synchronization data.
    u32* m_paSendProcessCnt;      // Buffer for counting the number of times sending all frames was attempted.
    u32 m_SendProcessFrameNoMin;  // The smallest value of the frame numbers for which a send attempt was made.
    u32 m_SendProcessFrameNoMax;  // The largest value of the frame numbers for which a send attempt was made.
    u32 m_DataFirstSendCount;     // The first send count setting for synchronization data.
    u64 m_SendFrameBitmap;        // Bitmap for send determinations.

    // For <tt>MonitoringData</tt>.
    u8 m_MonitoringMinDelay;                  // The minimum input delay value during game communication.
    u8 m_MonitoringMaxDelay;                  // The maximum input delay value during game communication.
    u8 m_MonitoringMinSendPeriod;             // The minimum value for the sending frequency.
    u8 m_MonitoringMaxSendPeriod;             // The maximum value for the sending frequency.
    u32 m_MonitoringStepCount;                // Number of times <tt>Step</tt> has been called.
    u32 m_MonitoringPreSendFrameNoMaxBuf;     // For saving the largest frame number sent last time.
    u32 m_MonitoringResendCountedFrameNoMax;  // The largest frame number counted as a resent frame.
    u8 m_MonitoringResendFrameCnt;            // For counting the number of frames resent.
    u8 m_MonitoringMaxResendFrameNum;         // The maximum value of the number of frames resent.
};
}
}
}  // end of namespace nn::pia::sync
