﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_NetworkFactory.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

namespace nn
{
namespace pia
{
namespace common
{
class StepSequenceJob;
class IPacketInput;
class IPacketOutput;
class SignatureSetting;
class MonitoringDataSender;
}
}
}  // end of namespace nn::pia::common

namespace nn
{
namespace pia
{
namespace session
{
class SignatureSettingStorage;
class IMatchmakeSession;
class ISessionInfoList;
class MeshLayerController;
enum HostMigrationMode;
#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
class IMatchmakeReferee;
#endif
}
}
}  // end of namespace nn::pia::session

namespace nn
{
namespace pia
{
namespace transport
{

class ConnectStationJob;
class DisconnectStationJob;
class PacketHandler;
class MissingStationHandler;

#if NN_PIA_EXPERIMENT_RELAY
class IRelayAddressTable;
#endif

/*!
@brief  An interface (factory pattern) for generating the classes needed by the network.
@details  This class is never instantiated by an application. Use a class that inherits this class.

@date 2013-05-16  Added the <tt>IsMultiCandidatePermitted</tt> function.
@date 2013-03-28  Added the <tt>CreatePacketHandler</tt> function.
@date 2013-03-21  Added the <tt>CreateSignatureSettingStorage</tt> function.
@date 2013-03-14  Added the <tt>CreateCreateMeshJob</tt> function.
@date 2012-11-27  Added the <tt>IsSignatureRequired</tt> function.
@if CTR_DOC
@date 2012-04-23  To improve readability, changed the See Also section to refer to <tt>@ref nn::pia::local::UdsNetworkFactory</tt> and <tt>@ref nn::pia::inet::NexNetworkFactory</tt>.
@endif
@if CAFE_DOC
@date 2012-04-23  To improve readability, changed the See Also section to refer to <tt>@ref nn::pia::inet::NexNetworkFactory</tt>.
@endif
@date 2012-04-06  Initial version.
@if CTR_DOC
@see  nn::pia::local::UdsNetworkFactory
@endif
@see  nn::pia::inet::NexNetworkFactory
*/
class NetworkFactory : public common::RootObject
{
public:
/*!
@cond PRIVATE
@brief  Virtual destructor.
*/
    virtual ~NetworkFactory(void)
    {
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>ConnectStationJob</tt> object.
*/
    virtual ConnectStationJob* CreateConnectStationJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>DisconnectStationJob</tt> object.
*/
    virtual DisconnectStationJob* CreateDisconnectStationJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates an <tt>InputStream</tt> object.
*/
    virtual common::IPacketInput* CreateInputStream() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateMeshJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateCreateMeshJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>JoinMeshJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateJoinMeshJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveMeshJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateLeaveMeshJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates an <tt>OutputStream</tt> object.
*/
    virtual common::IPacketOutput* CreateOutputStream() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>ProcessHostMigrationJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateProcessHostMigrationJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveWithHostMigrationJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateLeaveWithHostMigrationJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Creates a <tt>KickoutManageJob</tt> object.
*/
    virtual common::StepSequenceJob* CreateKickoutManageJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Builds the required protocol.
*/
    virtual nn::Result CreateProtocols()
    {
        return ResultSuccess();
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Constructs a concrete instance of the <tt>PacketHandler</tt> class.
*/
    virtual PacketHandler* CreatePacketHandler() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets whether a signature is required.
*/
    virtual bool IsSignatureRequired() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Creates storage for holding the signature key.
*/
    virtual session::SignatureSettingStorage* CreateSignatureSettingStorage() = 0;
    //! @endcond

#if NN_PIA_EXPERIMENT_RELAY
/*!
@cond PRIVATE
@brief  Generates an instance of a class that inherits the <tt>IRelayAddressTable</tt> class.
*/
    virtual IRelayAddressTable* CreateRelayAddressTable() = 0;
//! @endcond
#endif


#if NN_PIA_ENABLE_FEEDBACK
/*!
@cond PRIVATE
@brief  Creates a <tt>FeedbackProtocol</tt> object.
@return  Returns a handle to a <tt>FeedbackProtocol</tt> object.
*/
    virtual u32 CreateFeedbackProtocol()
    {
        return 0;
    }
//! @endcond
#endif

/*!
@cond PRIVATE
@brief  Gets whether multiple-candidate host migration is allowed.
*/
    virtual bool IsMultiCandidatePermitted() const
    {
        return false;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether the use of the relay communication feature is allowed.
*/
    virtual bool IsRelayConnectionPermitted() const
    {
        return false;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether the use of joint session features is allowed.
*/
    virtual bool IsJointSessionPermitted() const
    {
        return false;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Creates an instance of a class that inherits the <tt>MonitoringDataSender</tt> class.
*/
    virtual common::MonitoringDataSender* CreateMonitoringDataSender()
    {
        return NULL;
    };
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the total size of the header portion of packets at the network layer. This value is in bytes.
In IPv4 environments, for example, that would mean the IP header (20 bytes) plus the UDP header (8 bytes), which would be 28 bytes.
*/
    virtual size_t GetHeaderSize() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Constructs a concrete instance of the <tt>MissingStationHandler</tt> class.
*/
    virtual MissingStationHandler* CreateMissingStationHandler()
    {
        return NULL;
    }
    //! @endcond

/* For the auto-matchmaking feature. */
/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchCreateSessionJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates an <tt>AutoSessionmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchAutoMatchmakeSessionJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>BrowseSessionmakeSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchBrowseMatchmakeSessionJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>JoinSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchJoinSessionJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>LeaveSessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchLeaveSessionJob() = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Instantiates a <tt>DestroySessionJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateMatchDestroySessionJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateGenerateMatchmakeSystemPasswordJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateGenerateMatchmakeSystemPasswordJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>CreateClearMatchmakeSystemPasswordJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateClearMatchmakeSystemPasswordJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the length of the session's <tt>MatchmakeSessionSystemPassword</tt>.
Generates a maintenance region from this value.
*/
    virtual u32 GetMatchmakeSessionSystemPasswordSize() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Generates a maintenance region for the session's <tt>MatchmakeSessionSystemPassword</tt>.
*/
    virtual wchar_t* CreateMatchmakeSessionSystemPassword(u32 size) = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>JointSessionJob</tt> object for joint sessions.
*/
    virtual common::StepSequenceJob* CreateJointSessionJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>ModifyAttributeJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateModifyAttributeJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateSessionSetting</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateSessionSettingJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>UpdateApplicationBufferJob</tt> object for the session.
*/
    virtual common::StepSequenceJob* CreateUpdateApplicationDataJob() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates an <tt>ISessionInfoList</tt> object for the session.
*/
    virtual session::ISessionInfoList* CreateSessionInfoList(u8 browseSessionInfoListNum) = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates an <tt>IMatchmakeSession</tt> object for the session.
*/
    virtual session::IMatchmakeSession* CreateMatchmakeSession() = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Instantiates a <tt>MeshLayerController</tt> object for the session.
*/
    virtual session::MeshLayerController* CreateMatchMeshLayerController() = 0;
    //! @endcond

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    virtual session::IMatchmakeReferee* CreateMatchmakeReferee() = 0;

    virtual common::StepSequenceJob* CreateRefereeStartGameJob() = 0;

    virtual common::StepSequenceJob* CreateRefereeEndGameJob() = 0;

    virtual common::StepSequenceJob* CreateRefereeEndGameWithoutReportJob() = 0;

    virtual common::StepSequenceJob* CreateRefereeGetOrCreateStatsJob() = 0;
#endif

/*!
@cond PRIVATE
@brief  Gets whether a call to <tt>Bind</tt> is required.
*/
    virtual bool GetNecessaryBind() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether a server is required.
*/
    virtual bool GetNecessaryServer() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the host migration mode.
*/
    virtual session::HostMigrationMode GetHostMigrationMode() const = 0;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets the maximum number of session search results.
*/
    virtual u8 GetBrowseSessionInfoListMax() const = 0;
    //! @endcond

/* For the auto-matchmaking feature. */

protected:
/*!
@cond PRIVATE
*/
    NetworkFactory()
    {
    }
    //! @endcond
};
}
}
}  // end of namespace nn::pia::transport
