﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_PacketAnalysisData.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_ProtocolId.h>

#include <pia/common/common_Time.h>


namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  Stores information required for tabulating the results of an analysis of the data in sent and received packets.

@date 2013-12-10  Corrected an inaccurate description of <tt>passedMilliSec</tt>.
@date 2013-12-02  The increases in the number of sent packets, sent protocol messages, and sent data due to multicasting are now also included in the calculation.
@date 2013-11-21  The total packet size is now calculated using the size of packets in the network layer (for example, packets that include IPv4 or UDP headers).
@date 2013-11-14  Added a member variable that maintains the count from the point at which measurement starts.
@date 2013-10-17  Revised specifications.
@date 2013-10-15  Initial version.
*/
struct PacketAnalysisData
{
    static const size_t MAX_ENTRIES = 32;   //!< Specifies the maximum entries.
    static const size_t MAX_NAME_LEN = 32;  //!< Specifies the longest name that can be assigned to this object.


/*!
@brief  Contains information required for tabulating analysis results for each <tt>ProtocolId</tt>.
*/
    struct Entry
    {
        ProtocolId protocolId;  //!< ProtocolId
        size_t totalNum;        //!< Specifies the number of protocol messages that correspond to <tt>ProtocolId</tt>.
        size_t totalDataSize;   //!< Specifies the total amount of data, in bytes, corresponding to <tt>ProtocolId</tt>.
        size_t sumTotalNum;     //!< Specifies the number of protocol messages that correspond to <tt>ProtocolId</tt>. This value is not cleared by the <tt>ClearCounters</tt> function.
        u64 sumTotalDataSize;   //!< Specifies the total amount of data, in bytes, corresponding to <tt>ProtocolId</tt>. This value is not cleared by the <tt>ClearCounters</tt> function.
    };

    struct Entry entry[MAX_ENTRIES];  //!< Specifies an array of entries.
    size_t nowEntryNum;               //!< Specifies the current number of entries.
    s32 passedMilliSec;               //!< Specifies the elapsed time, in milliseconds, since the execution of the <tt>PiaTransport</tt> startup process.
    char name[MAX_NAME_LEN];          //!< Specifies the name to attach to the instance. The string must be terminated with <tt>\0</tt>.
    size_t totalPacketNum;            //!< Specifies the number of packets analyzed.
    size_t totalPacketSize;           //!< Specifies the total size of the packets analyzed.

    size_t sumTotalPacketNum;  //!< Specifies the number of packets counted since analysis started. This value is not cleared by the <tt>ClearCounters</tt> function.
    u64 sumTotalPacketSize;    //!< Specifies the total size of the packets counted since analysis started. This value is not cleared by the <tt>ClearCounters</tt> function.

/*!
@brief  Clears all members except <tt>name</tt>.
*/
    void ClearExceptName();


/*!
@brief  Clears the <tt>totalNum</tt>, <tt>totalDataSize</tt>, <tt>passedMilliSec</tt>, <tt>totalPacketNum</tt>, and <tt>totalPacketSize</tt> members.
*/
    void ClearCounters();


/*!
@brief  Prints the data.

@param[in] bDetail  Specify <tt>true</tt> to get verbose output.
*/
    void Print(bool bDetail) const;


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;


    static const size_t INVALID_IDX = 0xffffffff;  // Represents an invalid subscript.
    size_t GetIndex(ProtocolId id) const;          // Returns the entry that matches <tt>id</tt> if one exists; otherwise returns <tt>INVALID_IDX</tt>.
    size_t CreateIndex(ProtocolId id);             // Returns <tt>INVALID_IDX</tt> if the entry is full and a new subscript cannot be allocated.
};
}
}
}  // end of namespace nn::pia::transport
