﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_Protocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/common/common_ListNode.h>
#include <pia/transport/transport_ProtocolID.h>
#include <pia/transport/transport_ProtocolEvent.h>

#define PIA_STATISTIC_PROTOCOL_BLOCK_TIME_ENABLE 0
#if PIA_STATISTIC_PROTOCOL_BLOCK_TIME_ENABLE
#include <pia/common/private/common_Statistic.h>
#include <pia/common/common_TimeSpan.h>
#endif

namespace nn
{
namespace pia
{
namespace transport
{

class PacketHandler;

/*!
@brief  Protocol base class.

@date 2013-03-28  Moved namespace from \c nn::pia::common to \c nn::pia::transport.
@date 2012-04-04  Initial version.
*/
class Protocol : public common::RootObject
{
public:
/*!
@brief  Instantiates the object.
*/
    Protocol(void);


/*!
@brief  Destroys the object.
*/
    virtual ~Protocol(void);


/*!
@cond PRIVATE
@brief  Configures the packet handler to use.

@param[in] pPacketHandler  The packet handler.
*/
    void SetPacketHandler(PacketHandler* pPacketHandler)
    {
        m_pPacketHandler = pPacketHandler;
    }
    //! @endcond


/*!
@brief  Prints information useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


/*!
@cond PRIVATE
@brief  Gets the offset to the list node.

@return  The offset to the list node.
*/
    static u32 GetListNodeOffset()
    {
        return offsetof(Protocol, m_ListNode);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the port.

@param[in] port  The port number.
*/
    void SetPort(u16 port);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the protocol ID.

@return  The protocol ID.
*/
    ProtocolId GetProtocolId() const
    {
        return m_ProtocolId;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the protocol type.

@return  The protocol type.
*/
    virtual u16 GetProtocolType() const = 0;
    //! @endcond


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] myStationIndex  The <tt>StationID</tt> value of the local station. Specify a value in the range from <tt>STATION_INDEX_1</tt> to <span class="argument">maxConnections</span>.

@return  If execution succeeds, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  The main process.

@return  If execution succeeds, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    virtual nn::Result Dispatch();
    //! @endcond


/*!
@cond PRIVATE
@brief  Switches connection status.

@param[in] event  Specifies the event representing the operation.

@return  If execution succeeds, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    virtual nn::Result UpdateProtocolEvent(const ProtocolEvent& event);
//! @endcond


#if PIA_STATISTIC_PROTOCOL_BLOCK_TIME_ENABLE
    Statistic<TimeSpan>* GetBlockTimeStatistic()
    {
        return &m_BlockTimeStatistic;
    }
#endif

#if NN_PIA_EXPERIMENT_PROTOCOL_FILTERING
/*!
@cond PRIVATE
@brief  Function for controlling whether to filter out messages received from old stations.


@return  Returns <tt>true</tt> (allow filtering) by default.
Override this function so that it returns <tt>false</tt> in classes that inherit from the <tt>Protocol</tt> class that do not normally allow filtering.

*/
    virtual bool IsEnableProtocolFiltering() const;
//! @endcond
#endif

protected:
/*!
@cond PRIVATE
@brief  Gets the packet handler.

@return  A pointer to the packet handler.
*/
    PacketHandler* GetPacketHandler()
    {
        return m_pPacketHandler;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the packet handler (<tt>const</tt> version).

@return  A \c const pointer to the packet handler.
*/
    const PacketHandler* GetPacketHandler() const
    {
        return m_pPacketHandler;
    }
    //! @endcond


private:
    common::ListNode m_ListNode;
    ProtocolId m_ProtocolId;
    PacketHandler* m_pPacketHandler;

#if PIA_STATISTIC_PROTOCOL_BLOCK_TIME_ENABLE
    Statistic<TimeSpan> m_BlockTimeStatistic;
#endif
};
}
}
}  // end of namespace nn::pia::transport


#define PIA_PROTOCOL_TYPE_INFO(protocolType) \
    static u16 GetProtocolTypeStatic()       \
    {                                        \
        return protocolType;                 \
    }                                        \
    virtual u16 GetProtocolType() const      \
    {                                        \
        return GetProtocolTypeStatic();      \
    }
