﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_ProtocolManager.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/common/common_OffsetList.h>
#include <pia/transport/transport_ProtocolID.h>
#include <pia/transport/transport_ProtocolEvent.h>
#include <pia/transport/transport_Protocol.h>

namespace nn
{
namespace pia
{
namespace transport
{

class PacketHandler;

/*!
@cond PRIVATE
@brief  Manages protocols.

@date 2012-04-06  Initial version.
*/
class ProtocolManager : public common::RootObject
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    ProtocolManager(void);


/*!
@brief  Destructor.
*/
    ~ProtocolManager(void);


/*!
@brief  Initialization.
@return  <tt>nn::ResultSuccess</tt>.
*/
    nn::Result Initialize();


/*!
@brief  Finalizes.
*/
    void Finalize();


/*!
@brief  Generates an instance of a protocol of type <tt>T</tt>.

@param[in] port  The value of the protocol port.

@return  A handle to the generated protocol. If the same protocol ID already exists, 0 is returned.
Also returns zero after the \c Startup function has been called, because the object cannot be created in this case.
*/
    template <typename T>
    u32 CreateProtocol(u16 port = 0)
    {
        if (m_IsRunning)
        {
            return 0;
        }

        transport::Protocol* pProtocol = new (AllocProtocol(sizeof(T))) T();
        transport::ProtocolId id(T::GetProtocolTypeStatic(), port);
        return CreateProtocolImpl(pProtocol, id);
    }


/*!
@brief  Destroys the protocol instance.

@param[in] protocolHandle  The handle for the protocol to delete.
*/
    void DestroyProtocol(u32 protocolHandle);


/*!
@brief  Gets an instance of a protocol of type <tt>T</tt>.

@param[in] protocolHandle  The handle for the protocol to get.

@return  The address of the protocol instance. Returns \c NULL if the protocol for the specified handle does not exist or is not of type <tt>T</tt>.
*/
    template <typename T>
    T* GetProtocol(u32 protocolHandle)
    {
        return static_cast<T*>(SearchProtocol(ProtocolId(protocolHandle), T::GetProtocolTypeStatic()));
    }


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] pPacketHandler  The \c PacketHandler instance to use.

@return  On success, returns a \c Result value for which the \c IsSuccess function will return \c true.
*/
    nn::Result Startup(PacketHandler* pPacketHandler);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  Calls the \c Protocol::Startup function of each protocol.

@param[in] myStationIndex  The <tt>StationIndex</tt> value of the local station.

@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    nn::Result StartupProtocols(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Calls the \c Protocol::Cleanup function of each protocol.
*/
    void CleanupProtocols();
    //! @endcond


/*!
@cond PRIVATE
@brief  The main process.

@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    nn::Result Dispatch();
    //! @endcond


/*!
@cond PRIVATE
@brief  Switches connection status.

@param[in] event  Event representing the operation.

@return  On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt>.
*/
    nn::Result UpdateProtocolEvent(const ProtocolEvent& event);
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

#if PIA_STATISTIC_PROTOCOL_BLOCK_TIME_ENABLE
    void ResetProtocolStatistic();
    void PrintProtocolStatistic(u32 printMaxDataNum);
#endif

private:
    void* AllocProtocol(u32 size);

    u32 CreateProtocolImpl(Protocol* pProtocol, ProtocolId id);

    Protocol* SearchProtocol(ProtocolId protocolId, u16 protocolType);

private:
    typedef common::OffsetList<Protocol> ProtocolList;

    ProtocolList m_ProtocolList;

    bool m_IsRunning;
};
//! @endcond
}
}
}  // end of namespace nn::pia::transport
