﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_ReliableBroadcastProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_Protocol.h>
#include <pia/common/common_Time.h>


namespace nn
{
namespace pia
{
namespace transport
{
class ProtocolMessageWriter;
class ProtocolMessageReader;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  Represents a protocol for sending a large volume of data to all stations.

@date 2014-07-16  Changed the specifications so that <tt>ResultTemporaryUnavailable</tt> is returned when the <tt>StartSend</tt> or <tt>StartReceive</tt> function is called during joint sessions or during state transitions between sessions.
@date 2014-07-10  Fixed broken links in the API reference.
@date 2014-07-01  Changed the specifications of <tt>TransferSetting</tt>. Use <tt>TransferSetting::GetSourceStationId</tt> to get the sender's <tt>StationId</tt>.
@date 2014-04-02  Revised the documentation for the <tt>IsXXXX</tt> functions in the <tt>Configuration</tt> class.
@date 2013-05-27  Fixed an incorrect return value sometimes returned from the <tt>Cancel</tt> function.
@date 2013-04-25  Refactored code. This breaks backward compatibility.
@date 2012-03-29  The value specified in the <tt>SetThroughputLimit</tt> function is now set as the default value when communication starts.
@date 2012-10-29  Improved throughput by tuning the way that data is sent.
@date 2012-10-17  Fixed a bug where the <tt>GetProgress</tt> function might halt on an assertion.
@date 2012-09-18  Added the <tt>SetThroughputLimit</tt> and <tt>GetThroughputLimit</tt> functions.
@date 2012-09-10  Tuned throughput.
@date 2012-08-21  Initial version.
*/
class ReliableBroadcastProtocol : public transport::Protocol
{
public:
    PIA_PROTOCOL_TYPE_INFO(PROTOCOL_TYPE_RELIABLE_BROADCAST);

/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).

@details  Do not call this constructor directly from applications.
Use the <tt>@ref Transport::CreateProtocol</tt> function to create an instance.
*/
    ReliableBroadcastProtocol();
    //! @endcond


/*!
@cond PRIVATE
@brief  Destructor.

@details  Do not call this destructor directly from applications.
Use the <tt>@ref Transport::DestroyProtocol</tt> function to destroy an instance.
*/
    virtual ~ReliableBroadcastProtocol();
    //! @endcond


/*!
@brief  Initializes the instance. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.

@see  Finalize
*/
    nn::Result Initialize();


/*!
@brief  Finalizes the class instance.
@see  Initialize
*/
    void Finalize();


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] myStationIndex  The <tt>StationIndex</tt> value of the local station. Specify a value in the range from <tt>STATION_INDEX_1</tt> to <var>maxConnections</var>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  The main process.

@return  Returns <tt>nn::ResultSuccess</tt>.
*/
    virtual nn::Result Dispatch();
    //! @endcond


/*!
@cond PRIVATE
@brief  Switches the connection status.

@param[in] event  Specifies the event representing the operation.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that communication is not active. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result UpdateProtocolEvent(const transport::ProtocolEvent& event);
    //! @endcond

private:
/*!
@cond PRIVATE
@brief  Enumerates the operational modes of the protocol.
*/
    enum Mode
    {
        MODE_INVALID = 0x00000000,
        MODE_WAIT = 0x01000000,
        MODE_SEND = 0x02000000,
        MODE_RECEIVE = 0x03000000
    };
    //! @endcond


/*!
@cond PRIVATE
@brief  Enumerates the progress statuses of the protocol.
*/
    enum Condition
    {
        CONDITION_WAIT = 0x00010000,
        CONDITION_IN_PROGRESS = 0x00020000,
        CONDITION_SUCCESS = 0x00030000,
        CONDITION_FAILURE = 0x00040000
    };
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the current operation mode.
@return  Returns the current operation mode.
*/
    Mode GetMode() const
    {
        return static_cast<Mode>(m_State & 0xFF000000);
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the current progress.
@return  Returns the current progress.
*/
    Condition GetCondition() const
    {
        return static_cast<Condition>(m_State & 0x00FF0000);
    }
    //! @endcond


public:
/*!
@brief  Enumerates the operational states.
*/
    enum State
    {
        STATE_INVALID = MODE_INVALID | CONDITION_WAIT,             //!< Indicates not participating in a session.
        STATE_WAIT = MODE_WAIT | CONDITION_WAIT,                   //!< Indicates currently doing nothing.
        STATE_SENDING = MODE_SEND | CONDITION_IN_PROGRESS,         //!< Indicates sending.
        STATE_SEND_SUCCESS = MODE_SEND | CONDITION_SUCCESS,        //!< Indicates that sending was successful.
        STATE_SEND_FAILURE = MODE_SEND | CONDITION_FAILURE,        //!< Indicates failure to send.
        STATE_RECEIVING = MODE_RECEIVE | CONDITION_IN_PROGRESS,    //!< Indicates receiving.
        STATE_RECEIVE_SUCCESS = MODE_RECEIVE | CONDITION_SUCCESS,  //!< Indicates that receiving was successful.
        STATE_RECEIVE_FAILURE = MODE_RECEIVE | CONDITION_FAILURE,  //!< Indicates failure to receive.

        // The following are for internal management.
        //! @cond PRIVATE
        STATE_SEND_CANCELING = STATE_SENDING | 0x1,                         // Indicates currently canceling the send operation locally.
        STATE_SEND_REJECTING = STATE_SENDING | 0x2,                         // Indicates that the receiver is currently rejecting and canceling the send operation.
        STATE_SEND_ABANDONING = STATE_SENDING | 0x3,                        // Indicates currently abandoning the send operation in favor of a send operation from another system.
        STATE_SEND_COMPLETING = STATE_SENDING | 0x4,                        // Indicates that data has reached all session members and completion is being notified.
        STATE_SEND_CANCELED = STATE_SEND_FAILURE | 0x1,                     // Indicates that the send operation was canceled locally.
        STATE_SEND_REJECTED = STATE_SEND_FAILURE | 0x2,                     // Indicates that the send operation was rejected by the receiving station.
        STATE_SEND_ABANDONED = STATE_SEND_FAILURE | 0x3,                    // Indicates that the send operation was abandoned because of send priority arbitration.
        STATE_RECEIVE_READY = STATE_RECEIVING | 0x1,                        // Indicates currently ready to receive.
        STATE_RECEIVE_CANCELING = STATE_RECEIVING | 0x2,                    // Indicates currently canceling the receive operation.
        STATE_RECEIVE_CANCELED_BEFORE_START = STATE_RECEIVE_FAILURE | 0x1,  // Indicates that the receive operation was canceled locally (before the receive operation began).
        STATE_RECEIVE_CANCELED = STATE_RECEIVE_FAILURE | 0x2,               // Indicates that the receive operation was canceled locally.
        STATE_RECEIVE_INTERRUPTED = STATE_RECEIVE_FAILURE | 0x3,            // Indicates failed to receive because the sender canceled the send operation.
        STATE_RECEIVE_MISMATCH = STATE_RECEIVE_FAILURE | 0x4                // Indicates that receive settings do not match the request.
                                                                            //! @endcond
    };


/*!
@brief  Gets the current operational state.
@return  Returns the current operational state.
*/
    State GetState() const
    {
        return static_cast<State>(m_State & 0xFFFF0000);
    }


/*!
@brief  Gets whether asynchronous send and receive operations are currently executing.

@details  This function returns <tt>true</tt> if the <tt>GetState</tt> function returns <tt>STATE_SENDING</tt> or <tt>STATE_RECEIVING</tt>.

@return  Returns <tt>true</tt> if asynchronous processing of send and receive operations is in progress.
@see  GetState
*/
    bool IsRunning() const
    {
        return (GetCondition() == CONDITION_IN_PROGRESS);
    }


/*!
@brief  Represents methods used to transfer data.

@details  We recommend you use the constants provided for typical cases, because some combinations of settings are meaningless.

*/
    class Configuration
    {
    public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
Initialized with the same values as <tt>CURRENT_CONNECTION_SYNC</tt>.
*/
        Configuration()
          : m_Value(CURRENT_CONNECTION_SYNC.m_Value)
        {
        }

/*!
@cond PRIVATE
@brief  Converts from <tt>u16</tt>.
*/
        Configuration(u16 config)
          : m_Value(config)
        {
        }
        //! @endcond


/*!
@cond PRIVATE
@brief  Converts to <tt>u16</tt>.
*/
        operator u16() const
        {
            return m_Value;
        }
        //! @endcond


/*!
@brief  Gets whether to also send data to stations that join the session after the send operation begins.
@return  Returns <tt>true</tt> if data is also sent to stations that join the session while the send operation is ongoing.
Returns <tt>false</tt> if data is sent only to stations that are in the session at the time that the send operation starts.
@see  SetSendToNewcomer
*/
        bool IsSendToNewcomer() const
        {
            return IsFlagOn(FLAG_SEND_TO_NEWCOMER);
        }

/*!
@brief  Gets the setting that controls whether to synchronize the completion of receive operations.
@return  Returns <tt>true</tt> if the state changes to <tt>STATE_RECEIVE_SUCCESS</tt> when all stations have finished receiving the data.
Returns <tt>false</tt> if each station transitions to <tt>STATE_RECEIVE_SUCCESS</tt> when it has received all of the data.
@see  SetSynchronizeCompletion
*/
        bool IsSynchronizeCompletion() const
        {
            return IsFlagOn(FLAG_SYNCRONIZE_COMPLETION);
        }

/*!
@brief  Gets the setting for determining whether to maintain a state of <tt>STATE_SENDING</tt> even after data has been sent to all receiving stations.

@details  If this is set to <tt>true</tt>, the state will not change to <tt>STATE_SEND_SUCCESS</tt> automatically.
If you call the <tt>ReliableBroadcastProtocol::FinishSendEndless</tt> function, subsequent operations are the same as if <tt>false</tt> was specified for this setting. The state transitions to <tt>STATE_SEND_SUCCESS</tt> after data has been sent to all receiving stations.

@return  Returns <tt>true</tt> if the <tt>STATE_SENDING</tt> state is to be maintained after data is sent to all receiving stations and also if data will be sent to any stations that join later.
Returns <tt>false</tt> if the state changes to <tt>STATE_SEND_SUCCESS</tt> after data has been sent to all recipients and no more data will be sent.

@see  SetSendEndless

*/
        bool IsSendEndless() const
        {
            return IsFlagOn(FLAG_SEND_ENDLESS);
        }

/*!
@brief  Gets the setting for determining whether a send operation had failed or been abandoned when the receiving station rejects a receive operation.
@return  Returns <tt>true</tt> if the state transitions to <tt>STATE_SEND_FAILURE</tt>, and the entire send operation fails when a transmission is rejected.
Returns <tt>false</tt> if only the rejected station does not receive the data, and the station continues to send to the other recipients.
@see  SetFailIfRejected
*/
        bool IsFailIfRejected() const
        {
            return IsFlagOn(FLAG_FAIL_IF_REJECTED);
        }

/*!
@brief  Gets the setting for determining whether a send operation automatically fails on one station if two stations try to send each other data at the same time.

@details  If this setting is <tt>true</tt> and there is a conflict between sending stations, the data being sent from the sending station with the largest <tt>StationIndex</tt> value fails.
If two stations are sending data, and this setting is <tt>true</tt> for one and <tt>false</tt> for the other, the send operation sent from the station for which it is set to <tt>true</tt> fails.
If the setting is <tt>false</tt> for both stations, each station maintains a state of <tt>STATE_SENDING</tt> while waiting for the other station to receive. Neither station ever completes its send operation unless the application handles this case appropriately.

@return  Returns <tt>true</tt> if the station stops transmission and goes to the <tt>STATE_SEND_FAILURE</tt> state when it gets a receive request for which the <tt>IsMediateRequest</tt> returns <tt>false</tt>, or a receive request from a receiving station with a <tt>StationIndex</tt> lower than that of the local station.
Returns <tt>false</tt> if the station does not automatically go to the <tt>STATE_SEND_FAILURE</tt> state, depending on the receive requests it gets from recipients.

@see  SetMediateRequest


*/
        bool IsMediateRequest() const
        {
            return IsFlagOn(FLAG_MEDIATE_REQUEST);
        }


/*!
@brief  Sets whether to also send data to stations that join the session after the send operation begins.

@details  For more information about the meaning of this setting, see the <tt>IsSendToNewcomer</tt> function.

@param[in] isOn  Set to <tt>true</tt> to enable the setting.
@see  IsSendToNewcomer
*/
        void SetSendToNewcomer(bool isOn)
        {
            SetFlag(FLAG_SEND_TO_NEWCOMER, isOn);
        }

/*!
@brief  Sets whether to synchronize completion of receive operations.

@details  For more information about the meaning of this setting, see the <tt>IsSynchronizeCompletion</tt> function.

@param[in] isOn  Set to <tt>true</tt> to enable the setting.
@see  IsSynchronizeCompletion
*/
        void SetSynchronizeCompletion(bool isOn)
        {
            SetFlag(FLAG_SYNCRONIZE_COMPLETION, isOn);
        }

/*!
@brief  Sets whether the instance maintains a state of <tt>STATE_SENDING</tt> even after data has been successfully sent to all receiving stations.

@details  For more information about the meaning of this setting, see the <tt>IsSendEndless</tt> function.

@param[in] isOn  Set to <tt>true</tt> to enable the setting.
@see  IsSendEndless
*/
        void SetSendEndless(bool isOn)
        {
            SetFlag(FLAG_SEND_ENDLESS, isOn);
        }

/*!
@brief  Sets whether to treat a send operation as a send failure and suspend sending data when a receiving station rejects receiving data.

@details  For more information about the meaning of this setting, see the <tt>IsFailIfRejected</tt> function.

@param[in] isOn  Set to <tt>true</tt> to enable the setting.
@see  IsFailIfRejected
*/
        void SetFailIfRejected(bool isOn)
        {
            SetFlag(FLAG_FAIL_IF_REJECTED, isOn);
        }

/*!
@brief  Sets whether a send operation automatically fails on one station when two stations try to send data to each other at the same time.

@details  For more information about the meaning of this setting, see the <tt>IsMediateRequest</tt> function.

@param[in] isOn  Set to <tt>true</tt> to enable the setting.
@see  IsMediateRequest
*/
        void SetMediateRequest(bool isOn)
        {
            SetFlag(FLAG_MEDIATE_REQUEST, isOn);
        }


/*!
@brief  Represents the configuration for synchronized transfer of data to currently connected stations.

@details  This configuration assumes that stations will proceed to the next state in the sequence if one of the stations can successfully send data to all of the other stations after the deadline for joining the session has passed.

@li  <tt>IsSendToNewcomer</tt>: <tt>false</tt>
@li  <tt>IsSynchronizeCompletion</tt>: <tt>true</tt>
@li  <tt>IsSendEndless</tt>: <tt>false</tt>
@li  <tt>IsFailIfRejected</tt>: <tt>true</tt>
@li  <tt>IsMediateRequest</tt>: <tt>true</tt>
*/
        static const Configuration CURRENT_CONNECTION_SYNC;

/*!
@brief  Represents the configuration for asynchronously sending data to currently connected stations.

@details  Unlike <tt>CURRENT_CONNECTION_SYNC</tt>, this configuration assumes that receiving stations do not need to wait for data to be delivered to other receiving stations.

@li  <tt>IsSendToNewcomer</tt>: <tt>false</tt>
@li  <tt>IsSynchronizeCompletion</tt>: <tt>false</tt>
@li  <tt>IsSendEndless</tt>: <tt>false</tt>
@li  <tt>IsFailIfRejected</tt>: <tt>false</tt>
@li  <tt>IsMediateRequest</tt>: <tt>true</tt>
*/
        static const Configuration CURRENT_CONNECTION_ASYNC;

/*!
@brief  Represents the configuration for sending data to any stations that may join the session, in addition to currently connected stations, and waits for all stations to finish receiving data.

@details  This configuration assumes that data being shared is sent in advance while awaiting matchmaking, that sending is ended by <tt>FinishSendEndless</tt> after matchmaking ends, and that all stations then proceed to the next state in the sequence.

@li  <tt>IsSendToNewcomer</tt>: <tt>true</tt>
@li  <tt>IsSynchronizeCompletion</tt>: <tt>true</tt>
@li  <tt>IsSendEndless</tt>: <tt>true</tt>
@li  <tt>IsFailIfRejected</tt>: <tt>true</tt>
@li  <tt>IsMediateRequest</tt>: <tt>false</tt>


*/
        static const Configuration SHARE_DATA_SYNC;

/*!
@brief  Represents the configuration for continuing to send data to stations that may join the session in addition to currently connected stations.

@details  This setting is for a use case where specific data is sent to stations as they join the session. The stations then proceed to the next sequence in the order they receive the data.

@li  <tt>IsSendToNewcomer</tt>: <tt>true</tt>
@li  <tt>IsSynchronizeCompletion</tt>: <tt>false</tt>
@li  <tt>IsSendEndless</tt>: <tt>true</tt>
@li  <tt>IsFailIfRejected</tt>: <tt>false</tt>
@li  <tt>IsMediateRequest</tt>: <tt>false</tt>

*/
        static const Configuration SHARE_DATA_ASYNC;


    private:
        enum Flag
        {
            FLAG_SEND_TO_NEWCOMER = 1 << 0,
            FLAG_SYNCRONIZE_COMPLETION = 1 << 1,
            FLAG_SEND_ENDLESS = 1 << 2,
            FLAG_FAIL_IF_REJECTED = 1 << 3,
            FLAG_MEDIATE_REQUEST = 1 << 4
        };

        bool IsFlagOn(Flag flag) const
        {
            return ((m_Value & flag) == flag);
        }
        void SetFlag(Flag flag, bool isOn)
        {
            if (isOn)
            {
                m_Value |= flag;
            }
            else
            {
                m_Value &= ~flag;
            }
        }

        u16 m_Value;
    };


/*!
@brief  Stores data transfer settings.
*/
    struct TransferSetting
    {
        StationIndex m_SourceStationIndex;     //!< The sender's <tt>StationIndex</tt> value.
        u32 m_DataSize;                        //!< Specifies the size (in bytes) of the data being transferred.
        u32 m_UserData;                        //!< Specifies the user information set by the sender.
        Configuration m_Configuration;         //!< Specifies the method used to transfer data.
        StationId GetSourceStationId() const;  //!< Gets the sender's <tt>StationId</tt>.
    };


/*!
@brief  Gets the current data transfer settings.

@return  Retrieves send settings if the function is executed during or after a send operation, or the receive settings if it is during or after a receive operation.
Returns <tt>NULL</tt> if no data has been received, such as when the receive starts before a receive request comes in, or when the <tt>Cancel</tt> function is executed in that state.
*/
    const TransferSetting* GetCurrentTransferSetting() const;


/*!
@brief  Gets information about whether the specified station is to receive data currently being sent.

@details  Normally, all other stations participating in the session receive data.
If <tt>@ref Configuration::IsSendToNewcomer</tt> is <tt>false</tt>, stations that joined the session after the start of the transmissions are excluded as recipients.
If <tt>@ref Configuration::IsFailIfRejected</tt> is <tt>false</tt>, stations that rejected the transmission or for which reception failed are excluded as recipients.

@param[in] stationId  Specifies the <tt>StationId</tt> of the station that you are checking to determine whether it is a receiving station.

@return  Returns <tt>true</tt> if the specified station is a receiving station, and <tt>false</tt> otherwise.
Returns <tt>false</tt> if the local station is currently sending data and the operation has not completed.
*/
    bool IsSendTarget(nn::pia::StationId stationId) const;

/*!
@cond PRIVATE
@brief  Gets the current data transfer settings.
*/
    bool IsSendTarget(nn::pia::StationIndex stationIndex) const;
    //! @endcond


/*!
@brief  Gets the progress status of the send/receive operation.

@details  The value representing progress of a send operation can decrease while sending data because if the number of receiving stations changes.

@return  Returns the amount of data (in bytes) that has been sent and received.
You can retrieve the average amount of data sent to receiving stations during or after a send operation.
You can get the amount of the data received by the local station while or after receiving it.
*/
    size_t GetProgress() const;


/*!
@brief  Gets the progress status of data being sent to a particular receiving station.

@param[in] stationId  Specifies the <tt>StationId</tt> of the station receiving the data.

@return  Returns the amount of data (in bytes) that has been sent to the specified station.
<tt>0</tt> is returned if there is no send operation in progress or completed, or if an invalid <tt>StationId</tt> was specified.
*/
    size_t GetSendProgress(nn::pia::StationId stationId) const;


/*!
@cond PRIVATE
@brief  Gets the progress status of data being sent to a particular receiving station.
*/
    size_t GetSendProgress(nn::pia::StationIndex stationIndex) const;
    //! @endcond


/*!
@brief  Starts sending data.

@details  If this function call succeeds, the state of the station changes to <tt>STATE_SENDING</tt>.
If the call fails with a return value of <tt>ResultInvalidArgument</tt>, the <tt>Reset</tt> function is called.

@details  Data cannot be sent across joint session processes.

@param[in] cpData  Specifies a pointer to the data being sent.
You must maintain the specified buffer while the state is <tt>STATE_SENDING</tt> (while <tt>IsRunning</tt> returns <tt>true</tt>).
There is no guarantee that data will be sent in order. You cannot just rewrite the first <tt>GetProgress()</tt> bytes of data in the buffer.
@param[in] dataSize  Specifies the size of the data to send.
@param[in] configuration  Specifies operations during data transfer.
@param[in] userData  Specifies additional information included in the data to send.
Applications running on the receiving station can view this information before receiving the data, and use it for purposes such as identifying the data.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that the station has not joined the session or that another send/receive operation is in progress. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
*/
    nn::Result StartSend(const void* cpData, size_t dataSize, Configuration configuration, u32 userData = 0);


/*!
@brief  Stops sending data sent using any configuration for which the <tt>Configuration::IsSendEndless</tt> function returns <tt>true</tt>.
For more information about operations, see the <tt>Configuration::IsSendEndless</tt> function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Not currently sending data, or not configured for the <tt>@ref nn::pia::transport::ReliableBroadcastProtocol::Configuration::IsSendEndless "Configuration::IsSendEndless"</tt> function to return <tt>true</tt>. Programming error. Fix your program so that this error is not returned.

@see  Configuration::IsSendEndless
*/
    nn::Result FinishSendEndless();


/*!
@brief  Starts receiving data.

@details  To start receiving data on the local station in response to a <tt>GetRequest</tt> call, call this function with arguments specified in accordance with the settings for receiving data.
If you know in advance what to expect (because of the application's sequence, for example), you can call this function before the receive request comes so that you waste less time receiving the data than if you were to wait for the receive request to arrive.
If this function call succeeds, the local station enters <tt>STATE_RECEIVING</tt> state.
If the call fails with a return value of <tt>ResultInvalidArgument</tt>, the <tt>Reset</tt> function is called.

@details  Data cannot be received across joint session processes.

@param[out] pBuffer  Specifies the buffer that receives the data.
Data cannot be accessed while the local station's state is <tt>STATE_RECEIVING</tt> (while <tt>IsRunning</tt> returns <tt>true</tt>).
Data is not guaranteed to be in order. You may not be able to access the first number of bytes given by <tt>GetProgress</tt>.
@param[in] bufferSize  Specifies the size of the receive buffer.
If the size of the requested data is larger than this value, this function succeeds, but the station enters <tt>STATE_RECEIVE_FAILURE</tt> state.
@param[in] sourceStationId  Specifies the station the data is being received from.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that the station has not joined the session or that another send/receive operation is in progress. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.

@see  GetRequest

*/
    nn::Result StartReceive(void* pBuffer, size_t bufferSize, StationId sourceStationId);


/*!
@cond PRIVATE
@brief  Starts receiving data.
*/
    nn::Result StartReceive(void* pBuffer, size_t bufferSize, StationIndex sourceStationIndex);
    //! @endcond


/*!
@brief  Stops sending or receiving data.

@details  If this function is called when the state is <tt>STATE_SENDING</tt> and the action is then canceled in the <tt>common::Scheduler::Dispatch</tt> function, the station ordinarily goes to <tt>STATE_SEND_FAILURE</tt>.
However, the station may also enter <tt>STATE_SEND_SUCCESS</tt>, depending on the timing.
The receiving station also transitions to <tt>STATE_RECEIVE_FAILURE</tt> (or <tt>STATE_RECEIVE_SUCCESS</tt>).

If this function is called when the state is <tt>STATE_RECEIVING</tt> and the action is then canceled in the <tt>common::Scheduler::Dispatch</tt> function, the station ordinarily goes to <tt>STATE_RECEIVE_FAILURE</tt>.
However, the station may also enter <tt>STATE_RECEIVE_SUCCESS</tt>, depending on the timing. If this happens, you can access received data.
Sending itself is suspended if the station is not configured to run in a way that the <tt>Configuration::IsFailIfRejected(</tt> function returns <tt>true</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that data is not currently being sent/received. Programming error. Fix your program so that this error is not returned.

@see  StartSend, StartReceive
*/
    nn::Result Cancel();


/*!
@brief  Discards information about the result of sending and receiving data and returns to the <tt>STATE_WAIT</tt> state.

@details  This function fails if data is currently being sent or received. To call the function at such times, first call the <tt>Cancel</tt> function to suspend the send or receive operation, and then call this function.
Rather than always using this function, you can also start sending and receiving data directly from the following states: <tt>STATE_SEND_SUCCESS</tt>, <tt>STATE_SEND_FAILURE</tt>, <tt>STATE_RECEIVE_SUCCESS</tt>, and <tt>STATE_RECEIVE_FAILURE</tt>

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that the station has not joined in the session or is currently sending or receiving data. Programming error. Fix your program so that this error is not returned.

*/
    nn::Result Reset();


/*!
@brief  Gets receive requests that have been delivered.

@details  To get this data, prepare a buffer of the required size and call the <tt>StartReceive</tt> function.
Call the <tt>RejectRequest</tt> function if you do not want to receive the data.
When multiple receive requests have arrived, you can get the receive request with the highest priority (the request from the sender with the lowest <tt>StationIndex</tt>, regardless of the value of the <tt>Configuration::IsMediateRequest</tt> function).

@return  Returns a pointer to the receive request.
This function returns <tt>NULL</tt> if a receive request has not been delivered.

@see  StartReceive, RejectRequest
*/
    const TransferSetting* GetRequest() const;


/*!
@brief  Gets receive requests from a particular station.

@details  To get this data, prepare a buffer of the required size and call the <tt>StartReceive</tt> function.

@param[in] stationId  Specifies the <tt>StationId</tt> of the station you want to get a receive request from.

@return  Returns a pointer to the receive request.
This function returns <tt>NULL</tt> if a receive request has not been delivered.

@see  StartReceive, RejectRequest
*/
    const TransferSetting* GetRequest(nn::pia::StationId stationId) const;


/*!
@cond PRIVATE
@brief  Gets receive requests from a particular station.
*/
    const TransferSetting* GetRequest(nn::pia::StationIndex stationIndex) const;
    //! @endcond


/*!
@brief  Rejects receive requests retrieved with the <tt>GetRequest</tt> function.

@details  The send operation itself is suspended when the transfer protocol is configured such that the <tt>@ref Configuration::IsFailIfRejected</tt> function returns <tt>true</tt>.
Calling this function on a receive request currently being received is equivalent to calling the <tt>Cancel</tt> function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that the <tt>GetRequest</tt> function has returned <tt>NULL</tt>. Programming error. Fix your program so that this error is not returned.

@see  GetRequest
*/
    nn::Result RejectRequest();


/*!
@brief  Rejects receive requests retrieved with the <tt>@ref GetRequest ( nn::pia::StationId )</tt> function.

@details  The send operation itself is suspended when the transfer protocol is configured such that the <tt>@ref Configuration::IsFailIfRejected</tt> function returns <tt>true</tt>.
Calling this function on a receive request currently being received is equivalent to calling the <tt>Cancel</tt> function.

@param[in] stationId  Specifies the <tt>StationId</tt> of the station you want to reject a receive request from.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

@retval ResultInvalidState  Indicates that the <tt>GetRequest ( nn::pia::StationId )</tt> function has returned <tt>NULL</tt>. Programming error. Fix your program so that this error is not returned.

@see  GetRequest
*/
    nn::Result RejectRequest(nn::pia::StationId stationId);


/*!
@cond PRIVATE
@brief  Rejects receive requests retrieved with the <tt>@ref GetRequest ( StationIndex )</tt> function.
*/
    nn::Result RejectRequest(nn::pia::StationIndex stationIndex);
    //! @endcond


/*!
@brief  Sets the maximum amount of data that this protocol sends per call to the <tt>common::Scheduler::Dispatch</tt> function.

@details  The average amount of data sent by this protocol per call to the <tt>common::Scheduler::Dispatch</tt> function is kept under the value specified to this function.
Data may still momentarily be sent at rates greater than specified.

To temporarily stop sending data, a sender must specify <tt>0</tt> in this function. Setting a value of <tt>0</tt> completely stops this protocol from sending data.
Change the setting to a positive value if you want to start sending data again.
If you try to cancel data transfer by calling the <tt>Cancel</tt> function while sending is paused, the cancellation does not complete and notification of the cancellation is not sent while <tt>0</tt> is specified for this value.
Conversely, the data transfer does not proceed if a receiving station sets this value to <tt>0</tt>, because the receiving station also sends small amounts of data (such as acknowledgement packets).

Use the default value for the best performance.
The default is six packets worth of data. This value is not constant because the size of a single packet can vary depending on the MTU and other settings.

Set this value after communication starts, because the default value is reset when communication starts.

@param[in] throughputLimit  Specifies the amount of send data to configure (in bytes).

@see  GetThroughputLimit





*/
    void SetThroughputLimit(u32 throughputLimit)
    {
        m_ThroughputLimit = throughputLimit;
    }


/*!
@brief  Gets the maximum amount of data that this protocol can send per call to the <tt>common::Scheduler::Dispatch</tt> function.

@return  Returns the amount of data that the protocol sends per call to the <tt>common::Scheduler::Dispatch</tt> function.

@see  SetThroughputLimit
*/
    u32 GetThroughputLimit() const
    {
        return m_ThroughputLimit;
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


private:
    static const u16 cInvalidMessageIdx = 0xffff;

    // Manage progress.
    class Progress
    {
    public:
        u32 GetCount() const;
        bool IsOn(u32 idx) const;
        bool IsAllOff() const;
        bool IsComplete() const
        {
            return m_FirstOffIdx >= m_UnitNum;
        }

    protected:
        static const u32 cFlagBlockNum = 4;
        u32 m_UnitNum;
        u32 m_FirstOffIdx;
        u32 m_Flag[cFlagBlockNum];
    };

    // Manage progress while receiving.
    class ReceiverProgress : public Progress
    {
    public:
        void Clear(u32 unitNum);
        bool Raise(u32 idx);
        void Get(u32* pOffset, u64* pFlag) const;
    };

    // Manage progress while sending.
    class SenderProgress : public Progress
    {
    public:
        void Clear(u32 unitNum);
        bool Update(u32 offset, u64 flag);
        bool IsReceivable(u32 offset) const;
        void UpdateMessageIdx(u32 unitIdx, u16 messageIdx);
        u32 GetFirstOffIdx() const
        {
            return m_FirstOffIdx;
        }
        u32 GetUnitNum() const
        {
            return m_UnitNum;
        }
        u16 GetSendMessageIdx(u32 unitIdx) const;

    private:
        static const u32 cDataSize = cFlagBlockNum * 32;

        u16 m_SendMessageIdx[cDataSize];
    };


    // Message types.
    enum MessageType
    {
        eMTInvalid = 0x00,
        eMTRequest = 0x11,
        eMTData = 0x12,
        eMTCancel = 0x18,
        eMTComplete = 0x19,
        eMTDataAck = 0x21,
        eMTCommandAck = 0x28,
        eMTReject = 0x29
    };


    // Access messages.
    class MessageAccessor
    {
    public:
        MessageAccessor()
          : m_Type(eMTInvalid)
        {
        }
        bool IsSenderMessage() const
        {
            return (m_Type & 0xf0) == 0x10;
        }
        bool IsReceiverMessage() const
        {
            return (m_Type & 0xf0) == 0x20;
        }
        bool IsDataMessage() const
        {
            return m_Type == eMTData;
        }

        u32 GetMessageSize() const;
        static bool Unpack(MessageAccessor* pAccessor, const ProtocolMessageReader& reader, u32* pMessageIdx);
        static bool Pack(ProtocolMessageWriter* pWriter, const MessageAccessor& accessor, u32 messageIdx);

    public:
        MessageType m_Type;
        u16 m_Id;

        u32 m_TotalSize;
        u16 m_UnitSize;
        u32 m_UserData;
        Configuration m_Configuration;

        u32 m_Index;
        const void* m_cpPayload;
        u16 m_PayloadSize;

        u32 m_AckOffset;
        u64 m_AckFlag;
    };


    // Internal control of transfer settings.
    struct InnerTransferSetting : public TransferSetting
    {
    public:
        bool IsValid() const
        {
            return (m_DataSize > 0);
        }
        bool IsPriorTo(const TransferSetting& setting) const;

    public:
        u16 m_Id;
        u16 m_UnitSize;
        u32 m_UnitNum;
    };


    // Sending station states.
    enum SenderState
    {
        eSSNothing,
        eSSSending,
        eSSRejected,
        eSSFinished
    };

    // Receiving station states.
    enum ReceiverState
    {
        eRSNothing,
        eRSReceiving,
        eRSRejecting,
        eRSFinished,
        eRSRejected
    };

    // Manage the status of all communicating stations.
    class Station
    {
    public:
        Station();
        void Initialize(ReliableBroadcastProtocol* pProtocol);

        void Startup(StationIndex stationId);
        void Cleanup();

        void HandleMessage(const MessageAccessor& accessor);
        bool SendAck();

        void ForceReserveReject()
        {
            m_ReservedAckFlag |= eRAFReject;
        }

        void RejectSend();

        void ClearReceiverState()
        {
            m_ReceiverState = eRSNothing;
        }
        void StartReceive(u16 receiveId, u32 unitNum);

        bool IsConnected() const
        {
            return MyStationIndex() != STATION_INDEX_INVALID;
        }
        StationIndex GetStationIndex() const
        {
            return MyStationIndex();
        }
        SenderState GetSenderState() const
        {
            return m_SenderState;
        }
        ReceiverState GetReceiverState() const
        {
            return m_ReceiverState;
        }
        const InnerTransferSetting& GetRequest() const
        {
            return m_Request;
        }
        const SenderProgress* GetProgress() const
        {
            return &m_Progress;
        }

        u16 GetReturnedMessageIdx() const
        {
            return m_ReturnedMessageIdx;
        }
        u16 GetReceivedMessageIdx() const
        {
            return m_ReceivedMessageIdx;
        }

        void HandleMessageIdx(u32 messageIdx);

        u32 GetNextSendUnitIndex(bool isOnlyPrior, u32 dispatchCount) const;

        bool IsNeedSendRequest() const
        {
            return (GetReceiverState() == eRSReceiving) && m_IsNeedSendRequest;
        }

        void RecordSendData(u32 dispatchCount, u16 unitIdx, u16 messageIdx);

    private:
        void UpdateRequest(const MessageAccessor& accessor);
        bool CheckRequest(const MessageAccessor& accessor) const;

        StationIndex& MyStationIndex()
        {
            return m_Request.m_SourceStationIndex;
        }
        StationIndex MyStationIndex() const
        {
            return m_Request.m_SourceStationIndex;
        }
        u16& SenderSequenceId()
        {
            return m_Request.m_Id;
        }
        u16 SenderSequenceId() const
        {
            return m_Request.m_Id;
        }

    private:
        ReliableBroadcastProtocol* m_pProtocol;

        InnerTransferSetting m_Request;
        SenderState m_SenderState;

        enum ReservedAckFlag
        {
            eRAFNone = 0x0,
            eRAFRequest = 0x1,
            eRAFData = 0x2,
            eRAFReject = 0x4,
            eRAFCommand = 0x8
        };
        u32 m_ReservedAckFlag;

        SenderProgress m_Progress;
        u16 m_ReceivingId;
        ReceiverState m_ReceiverState;

        u16 m_ReturnedMessageIdx;
        u16 m_ReceivedMessageIdx;

        bool m_IsNeedSendRequest;

        u32 m_LastSentDataDispatchCount;
    };

    // Is the other station valid?
    bool IsValidStation(StationIndex stationId) const
    {
        return (stationId >= STATION_INDEX_1) && (stationId < static_cast<StationIndex>(m_MaxConnections)) && (stationId != m_LocalStationIndex);
    }

    // Get information about another station.
    Station* GetStation(StationIndex stationId)
    {
        PIA_ASSERT(IsValidStation(stationId));
        return &m_paStation[stationId < m_LocalStationIndex ? stationId : stationId - 1];
    }

    // Get information about another station.
    const Station* GetStation(StationIndex stationId) const
    {
        PIA_ASSERT(IsValidStation(stationId));
        return &m_paStation[stationId < m_LocalStationIndex ? stationId : stationId - 1];
    }

    static u32 CalcProgress(const InnerTransferSetting& transferSetting, const Progress& progress);

    void UpdateState();
    bool DispatchSendAck();
    bool DispatchSendCommand();
    bool DispatchSendData();
    void DispatchOnReceiveCanceling();

    void PushRequest(const MessageAccessor& accessor, const Station& sender);
    bool PushData(const MessageAccessor& accessor, const Station& sender);
    void PushComplete(const MessageAccessor& accessor, const Station& sender);
    void PushCancel(const MessageAccessor& accessor, const Station& sender);

    bool CreateMessage(const MessageAccessor& accessor, Station* pDestinationStation);
    u32 IssueMessageIdx(const Station* cpDestStation);

    static bool IsSameSetting(const InnerTransferSetting& setting, const MessageAccessor& accessor);


    u32 m_MaxConnections;

    State m_State;
    StationIndex m_LocalStationIndex;

    InnerTransferSetting m_TransferSetting;

    void* m_pReceiveBuffer;
    ReceiverProgress m_ReceiveProgress;

    const void* m_cpSendBuffer;
    u16 m_NextSendId;
    u32 m_NextSendCommandStationIndex;
    u32 m_NextSendDataStationIndex;
    bool m_IsFinishSendEndless;

    u32 m_NextSendAckStationIndex;

    u16 m_SentMessageIdx;

    u32 m_ThroughputLimit;
    u32 m_RestThroughputLimit;

    u32 m_DispatchCount;

    Station* m_paStation;

    common::Time m_StartSendTime;

    static const u32 cDefaultPacketNumPerDispatch;
};
}
}
}  // end of namespace nn::pia::transport
