/*---------------------------------------------------------------------------*
  Project: Pia
  File: transport_ReliableProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/common/common_types.h>
#include <pia/transport/transport_Protocol.h>

#include <pia/transport/transport_ProtocolEvent.h>

namespace nn
{
namespace pia
{
namespace transport
{

class ReliableSlidingWindow;

/*!
  @brief  This protocol sends transmissions that are guaranteed to arrive.

  @date 2014-07-17  Changed the specifications so that <tt>ResultNotFound</tt> is returned when the destination <tt>StationId</tt> specified in the <tt>Send()</tt> function is not found.
  @date 2014-07-16  Changed the specifications so that <tt>ResultTemporaryUnavailable</tt> is returned when <tt>Send()</tt> is called while state transitions between sessions or joint sessions are in progress.
  @date 2014-03-24  Added a formula to the <tt>Initialize()</tt> API Reference for determining the approximate amount of memory it consumes.
  @date 2012-12-14  Improved the method of tuning the resend interval.
  @date 2012-10-29  It now automatically tunes the resend interval.
  @date 2012-07-17  Added <tt>ResultNotInCommunication</tt> to the return values of the <tt>Send()</tt> and <tt>Receive()</tt> functions.
  @date 2012-07-09  You are no longer required to set the maximum number of connections in the <tt>Initialize()</tt> function.
  @date 2012-07-05  Added descriptions of the send and receive operations and buffer usage.
  @date 2012-04-24  You can now send to all members by specifying the recipient as <tt>STATION_INDEX_ALL</tt> in the <tt>Send()</tt> function.
  @date 2012-04-06  Initial version.
*/
class ReliableProtocol : public transport::Protocol
{
public:
/*!
    @cond  PRIVATE
    @brief  Instantiates an object with default parameters (default constructor).

    @details  Do not call this constructor directly from applications.
           Use the <tt>@ref Transport::CreateProtocol</tt> function to create an instance.
*/
    ReliableProtocol(void);
    //! @endcond


/*!
    @cond  PRIVATE
    @brief  The destructor.

    @details  Do not call this destructor directly from applications.
           Use the <tt>@ref Transport::DestroyProtocol</tt> function to destroy an instance.
*/
    virtual ~ReliableProtocol(void);
    //! @endcond


    PIA_PROTOCOL_TYPE_INFO(PROTOCOL_TYPE_RELIABLE);


/*!
    @brief  Initializes the instance. Call between the <tt>BeginSetup()</tt> and <tt>EndSetup()</tt> functions.

    @details  You can measure the actual number of send/receive buffers used by specifying
            <tt>@ref common::WatermarkManager::KEY_RELIABLE_PROTOCOL_SEND_BUFFER_NUM</tt> or
            <tt>@ref common::WatermarkManager::KEY_RELIABLE_PROTOCOL_RECEIVE_BUFFER_NUM</tt> in
            <tt>@ref common::WatermarkManager</tt>.
    @details  Use the following formula to calculate the approximate amount of memory consumed by <tt>ReliableProtocol</tt>.
             (<tt><var>sendBufferSize</var></tt> * <tt>1500</tt> + <tt><var>receiveBufferSize</var></tt> * <tt>1500</tt>) * (<tt>Transport::Setting::maxStationNum</tt> - <tt>1</tt>)

    @param[in] sendBufferSize  Specifies the size of the send buffer. Specify a value of <tt>1</tt> or greater.
                                        Send buffers are used to hold data sent using the <tt>Send()</tt> function until the sender can confirm that it has arrived to the recipient.
    @param[in] receiveBufferSize  Specifies the size of the receive buffer. Specify a value of <tt>1</tt> or greater.
                                        Receive buffers are used to hold received data until the <tt>Receive()</tt> function is called.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.

    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.

    @see  Finalize
*/
    nn::Result Initialize(u32 sendBufferSize, u32 receiveBufferSize);


/*!
    @brief  Finalizes.

    @see  Initialize
*/
    void Finalize();


/*!
    @cond  PRIVATE
    @brief  Starts communication.

    @param[in] myStationIndex  Specifies the <tt>StationIndex</tt> value of the local station. Specify a value in the range from <tt>STATION_INDEX_1</tt> to <tt><var>maxConnections</var></tt>.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.

    @retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
    @cond  PRIVATE
    @brief  Ends communication.
*/
    virtual void Cleanup();
    //! @endcond


/*!
    @cond  PRIVATE
    @brief  The main process.

    @return  Returns <tt>nn::ResultSuccess</tt>.
*/
    virtual nn::Result Dispatch();
    //! @endcond


/*!
    @cond  PRIVATE
    @brief  Switches the connection status.

    @param[in] event  Specifies the event representing the operation.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.

    @retval ResultInvalidState  Indicates that the function was called in an invalid state. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result UpdateProtocolEvent(const transport::ProtocolEvent& event);
    //! @endcond


/*!
    @brief  Sends data.

    @details  The function divides the send data that is set with this function into packet-sized chunks and stores the chunks in buffers.
            Next, the <tt>@ref common::Scheduler::Dispatch()</tt> function creates packets from the data consecutively, and sends them.
            If the library cannot confirm that a packet has arrived after waiting for a reasonable time, it resends it.
            The time until a resent is adjusted automatically, based on the round-trip time (RTT).
            After the library confirms that a packet has arrived, it deletes the data from the buffer. At this point, the buffer can be reused for sending additional data.

    @param[in] destinationStationId  Specifies the <tt>@ref StationId</tt> of the recipient. Specify <tt>STATION_ID_ALL</tt> to send to all currently connected peers.
    @param[in] pData  Specifies the address of the data to send.
    @param[in] dataSize  Specifies the length, in bytes, of the data to send.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.

    @retval ResultInvalidState  <tt>ReliableProtocol</tt> might not have been initialized, or the session setup might not have been completed. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultNotInCommunication  Indicates that the local station is not connected to the specified recipient. The specified peer is not in this session. Handle appropriately in the application.
    @retval ResultBufferIsFull  Indicates that there is not enough space in the send buffer. If this happens frequently, consider increasing the send buffer size specified in the <tt>Initialize()</tt> function. Handle appropriately in the application.
    @retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
    @retval ResultNotFound  Indicates that the specified destination was not found. Handle appropriately in the application.

    @see  Receive
*/
    nn::Result Send(StationId destinationStationId, const void* pData, size_t dataSize);


/*!
    @cond  PRIVATE
    @brief  Sends data.
*/
    nn::Result Send(StationIndex destinationStationIndex, const void* pData, size_t dataSize);
    //! @endcond


/*!
    @brief  Receives data.

    @details  The library parses received packets when the <tt>@ref common::Scheduler::Dispatch()</tt> function is called, and accumulates them in buffers.
            When the library can confirm that all the data chunks created by the <tt>Send()</tt> function have been received, it joins them together with this function. You can then retrieve the data using this function.
            The library then frees the storage buffer, and it can be used for the next receive operation.

    @param[out] pSourceStationId  Specifies a pointer to a buffer containing the <tt>@ref StationId</tt> of the sender of the data that was received.
    @param[out] pBuffer  Specifies a pointer to a buffer for holding the received data.
    @param[out] pDataSize  Specifies a pointer to a buffer for holding the size of the received data.
    @param[in] bufferSize  Specifies the size of the buffer (in bytes) for holding the received data.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds.

    @retval ResultInvalidState  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultNotInCommunication  Indicates that the station is not actively communicating. The peer is not in the current session. Handle appropriately in the application.
    @retval ResultNoData  Indicates that no data was received. Handle appropriately in the application.
    @retval ResultBrokenData  Indicates that the received data is too large to fit in the receive buffer or cannot all be written to the buffer passed into the function. Start over from the session startup step.
                                    This result could be returned because of invalid data from the sender, such as cheats, requiring handling at run time.
                                    If the sender's behavior is valid, the program must be fixed so that this result is not returned. Handle appropriately in the application.

    @see  Send
*/
    nn::Result Receive(StationId* pSourceStationId, void* pBuffer, size_t* pDataSize, size_t bufferSize);


/*!
    @cond  PRIVATE
    @brief  Receives data.
*/
    nn::Result Receive(StationIndex* pSourceStationIndex, void* pBuffer, size_t* pDataSize, size_t bufferSize);
    //! @endcond


/*!
    @brief  Gets whether the specified node is communicating.

    @details  If the specified peer is in the same session as the local station, <tt>true</tt> will be returned.

    @param[in] stationId  Returns the <tt>@ref StationId</tt> of the peer.

    @return  Returns <tt>true</tt> if the node is communicating.
*/
    bool IsInCommunication(StationId stationId) const;


/*!
    @cond  PRIVATE
    @brief  Gets whether the specified node is communicating.
*/
    bool IsInCommunication(StationIndex stationId) const;
    //! @endcond


    // Resend time calculation callback type.
    // <tt>rtt</tt> takes the round-trip time (RTT) in milliseconds. <tt>resendCount</tt> takes the number of times the data has been resent.
    // The callback's return value is the resend time. The value is in milliseconds.
    typedef s32 (*ResendTimeCallback)(s32 rtt, size_t resendCount);


/*!
    @cond  PRIVATE
    @brief  Registers a callback function that calculates the resend time.
            No callback function is registered by default.

    @return  Returns a pointer to the callback function held until just before the call.
*/
    static ResendTimeCallback RegisterResendTimeCallback(ResendTimeCallback cb);
    //! @endcond


/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>TraceFlag</tt> reference.
*/
    virtual void Trace(u64 flag) const;

#if NN_PIA_EXPERIMENT_RELIABLE_DETAIL_INFO
    void PrintDebugInfo(void) const;
    void ClearAllResendCount(void);
    u32 GetAllResendCount(void) const;
#endif

private:
    ReliableSlidingWindow* GetSlidingWindow(StationIndex stationId);
    const ReliableSlidingWindow* GetSlidingWindow(StationIndex stationId) const;

    nn::Result ReceiveImpl(StationIndex* pSourceStationIndex, void* pBuffer, size_t* pDataSize, size_t bufferSize, bool bCheckStationIdTable);

private:
    StationIndex m_LocalStationIndex;
    u32 m_MaxConnections;
    ReliableSlidingWindow* m_paSlidingWindow;
    u32 m_DispatchSlidingWindowIdx;

    static ResendTimeCallback s_pfResendTimeCallback;  // Resend time calculation callback.
};
}
}
}  // End of namespace nn::pia::transport.
