﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_ReliableSlidingWindow.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>
#include <pia/common/common_types.h>
#include <pia/transport/transport_ProtocolID.h>
#include <pia/common/common_PacketDefine.h>


// Forward declaration.
namespace nn
{
namespace pia
{
namespace common
{
class String;
}
}
}

namespace nn
{
namespace pia
{
namespace transport
{

class ProtocolMessageReader;
class PacketHandler;

/*!
@cond PRIVATE
@brief  Represents a sliding window for reliable sending and receiving.

@date 2014-01-29  Provided an interface to the resend time calculation callback.
@date 2013-01-22  Added functions to get and set the sample count when getting RTT.
@date 2013-01-22  Added functions to get and set the timeout rate for resending.
@date 2012-07-17  Added the <tt>GetString</tt> function.
@date 2012-04-06  Initial version.
*/
class ReliableSlidingWindow : public common::RootObject
{
public:
/*!
@brief  Instantiates the object with default parameters (default constructor).
*/
    ReliableSlidingWindow();


/*!
@brief  Destroys the object.
*/
    virtual ~ReliableSlidingWindow();


/*!
@brief  Initializes the object.

@param[in] sendBufferSize  Specifies the size of the send buffer. Specify a value of <tt>1</tt> or greater.
@param[in] receiveBufferSize  Specifies the size of the receive buffer. Specify a value of <tt>1</tt> or greater.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyInitialized  Indicates that the instance is already initialized. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Initialize(u32 sendBufferSize, u32 receiveBufferSize);


/*!
@brief  Finalizes the object.
*/
    void Finalize();


/*!
@brief  Starts communication.

@param[in] pPacketHandler  Specifies the packet handler to use.
@param[in] protocolId  Specifies the protocol ID to use for communication.
@param[in] sourceStationIndex  Specifies the ID of the local station. Specify a value between <tt>STATION_INDEX_1</tt> and <tt>STATION_INDEX_MAX</tt>.
@param[in] destinationStationIndex  Specifies the ID of the station you are communicating with. Specify a value between <tt>STATION_INDEX_1</tt> and <tt>STATION_INDEX_MAX</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Startup(PacketHandler* pPacketHandler, u32 protocolId, StationIndex sourceStationIndex, StationIndex destinationStationIndex);


/*!
@brief  Starts communication as the host in the <tt>Station</tt> protocol.

@param[in] pPacketHandler  Specifies the packet handler to use.
@param[in] protocolId  Specifies the protocol ID to use for communication.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result StartupForStationHost(PacketHandler* pPacketHandler, u32 protocolId);


/*!
@brief  Starts communication as a client in the <tt>Station</tt> protocol.

@param[in] pPacketHandler  Specifies the packet handler to use.
@param[in] protocolId  Specifies the protocol ID to use for communication.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultNotInitialized  Indicates that the instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that stations are already communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result StartupForStationChild(PacketHandler* pPacketHandler, u32 protocolId);


/*!
@brief  Sets the <tt>StationIndex</tt>.

@param[in] sourceStationIndex  Specifies the ID of the local station. Specify a value between <tt>STATION_INDEX_1</tt> and <tt>STATION_INDEX_MAX</tt>.
@param[in] destinationStationIndex  Specifies the ID of the station you are communicating with. Specify a value between <tt>STATION_INDEX_1</tt> and <tt>STATION_INDEX_MAX</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the station is not communicating or that the <tt>StationID</tt> has already been set. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetStationIndex(StationIndex sourceStationIndex, StationIndex destinationStationIndex);


/*!
@brief  Ends communication.
*/
    void Cleanup();


/*!
@brief  Gets whether communication is active.

@return  Returns <tt>true</tt> if currently communicating.
*/
    bool IsInCommunication() const;


/*!
@brief  Parses a received message.

@param[in] reader  Specifies the received <tt>ProtocolMessageReader</tt>.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the station is not actively communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultBrokenData  Indicates that the data in the specified <tt>ProtocolMessageOld</tt> object was invalid. Handle appropriately in the application.
@retval ResultBufferIsFull  Indicates that the receive buffer is full. Handle appropriately in the application.
*/
    nn::Result AnalyzeProtocolMessage(const ProtocolMessageReader& reader);


/*!
@brief  The main process.

@param[in] pPacketHandler  Specifies the <tt>PacketHandler</tt> instance to use.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the station is not actively communicating. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result Dispatch(PacketHandler* pPacketHandler);


/*!
@brief  Gets received data.

@param[out] pBuffer  Returns a pointer to a buffer for holding the received data.
@param[out] pDataSize  Returns a pointer to a buffer for holding the size of the received data.
@param[in] bufferSize  Specifies the size of the buffer (in bytes) for holding the received data.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the station is not actively communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates that no data was received. Handle appropriately in the application.
@retval ResultBrokenData  Indicates that the received data is too large to fit in the receive buffer or cannot all be written to the buffer passed into the function. Try the communication again. Handle appropriately in the application.
*/
    nn::Result PopData(void* pBuffer, size_t* pDataSize, size_t bufferSize);


/*!
@brief  Adds data to send.

@param[in] pData  Specifies a pointer to the data to send.
@param[in] dataSize  Specifies the size of the data to send.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.

@retval ResultInvalidState  Indicates that the station is not actively communicating. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultBufferIsFull  Indicates that there is not enough space in the send buffer. Handle appropriately in the application.
*/
    nn::Result PushData(const void* pData, size_t dataSize);


/*!
@brief  Checks whether data to send can be added.

@param[in] dataSize  Specifies the size of the data to send.

@return  Returns <tt>true</tt> if data can be added, and <tt>false</tt> otherwise.
*/
    bool CanPushData(size_t dataSize);


/*!
@brief  Gets the protocol ID.

@return  Returns the protocol ID.
*/
    u32 GetProtocolId() const
    {
        return m_ProtocolId;
    }


/*!
@brief  Gets the local <tt>StationIndex</tt>.

@return  Returns the local <tt>StationIndex</tt>.
*/
    StationIndex GetSource() const
    {
        return m_SourceStationIndex;
    }


/*!
@brief  Gets the <tt>StationIndex</tt> of the peer.

@return  Returns the <tt>StationIndex</tt> of the peer.
*/
    StationIndex GetDestination() const
    {
        return m_DestinationStationIndex;
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

/*!
@cond PRIVATE
@brief  Gets information that is useful for debugging.

@param[out] pString  Returns a pointer to a <tt>String</tt> instance.
*/
    void GetString(common::String* pString) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the number of elements used in the sending buffer.
@return  Returns the number of elements used in the sending buffer.
*/
    u32 GetSendBufferDataNum() const
    {
        return m_SendBuffer.GetDataNum();
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the number of elements used in the receive buffer.
@return  Returns the number of elements used in the receive buffer.
*/
    u32 GetReceiveBufferDataNum() const
    {
        return m_ReceiveBuffer.GetDataNum();
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the timeout rate for resending. The resend timeout is calculated using the product of this rate and the RTT.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments is invalid (such as a negative number).
*/
    static nn::Result SetRetransmissionTimeOutRate(f32 rate);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the timeout rate for resending.
@return  Returns a rate that can be multiplied with the RTT.
*/
    static f32 GetRetransmissionTimeOutRate(void)
    {
        return s_RtoRate;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the number of samples to use when getting the RTT.
The lower the value of <span class="argument">num</span>, the more the obtained RTT value is affected by the current conditions. Although this makes the RTT value more variable, it makes the value less stable.

@param[in] num  Specifies the number of samples to use when calculating the RTT.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that an argument is invalid.
*/
    static nn::Result SetRttSamplingNum(size_t num);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the number of samples that this class uses to calculate RTT.

@return  Returns the number of samples used to calculate the RTT.
*/
    static size_t GetRttSamplingNum(void)
    {
        return s_RttSamplingNum;
    }
    //! @endcond


    // Resend time calculation callback type.
    typedef s32 (*ResendTimeCallback)(s32 rtt, size_t resendCount);


/*!
@cond PRIVATE
@brief  Specifies a callback function that calculates the resend time.
No callback is registered by default.

@return  A pointer to the callback function registered before this function was called.
*/
    static ResendTimeCallback RegisterResendTimeCallback(ResendTimeCallback cb);
//! @endcond


#if NN_PIA_EXPERIMENT_RELIABLE_DETAIL_INFO
    void PrintDebugInfo(void) const;

    void ClearResendCount(void);
    u32 GetResendCount(void) const;
#endif

private:
    // Copying and comparisons are not implemented
    ReliableSlidingWindow(const ReliableSlidingWindow& rhs);
    bool operator==(const ReliableSlidingWindow& rhs) const;
    bool operator!=(const ReliableSlidingWindow& rhs) const;
    ReliableSlidingWindow& operator=(const ReliableSlidingWindow& rhs);


    void StartupCommon(PacketHandler* pPacketHandler);

private:
    // The header attached to <tt>ProtocolMessageOld</tt>.
    struct MessageHeader
    {
        u16 m_Flags;
        u16 m_PayloadSize;
        u32 m_Reserved;
        u32 m_SequenceId;
        u32 m_AckOrigin;
        u64 m_AckFlag;
    };

    // The maximum payload size that fits in a <tt>ProtocolMessage</tt> instance.
    static const u32 c_PayloadSizeMax = (common::PROTOCOL_MESSGAE_PAYLOAD_SIZE_MAX / sizeof(u64) * sizeof(u64)) - sizeof(MessageHeader);

    // The message payload.
    struct Payload
    {
        bit8 m_Buffer[c_PayloadSizeMax];
    };

    // Data structure of <tt>ProtocolMessageOld</tt>.
    struct MessageData
    {
        MessageHeader m_Header;
        Payload m_Payload;
    };

    // The data for the send buffer.
    struct SendData
    {
        u64 m_ResendTime;
        u32 m_DataSize;
        u16 m_SendCount;
        MessageData m_MessageData;
    };

    // The data for the receive buffer.
    struct ReceiveData
    {
        bool m_IsValid;
        bool m_IsComplete;
        u16 m_PayloadSize;
        Payload m_Payload;
    };

    // Common processes for the send and receive buffers.
    template <typename T>
    class Buffer : public common::RootObject
    {
    public:
        Buffer();
        ~Buffer();

        nn::Result Initialize(u32 size);
        void Finalize();

        void Clear();

        T* GetData(u32 sequenceId);

        u32 GetSize() const
        {
            return m_Size;
        }
        u32 GetStartSequenceId() const
        {
            return m_StartSequenceId;
        }
        u32 GetDataNum() const
        {
            return m_DataNum;
        }

        void Withdraw(u32 num);

        class Iterator
        {
        public:
            Iterator(Buffer<T>* pBuffer)
              : m_pBuffer(pBuffer)
            {
                Begin();
            }

            T* Get()
            {
                PIA_ASSERT(IsEnd() == false);
                u32 idx = m_pBuffer->m_StartIdx + m_Offset;
                if (idx >= m_pBuffer->m_Size)
                {
                    idx -= m_pBuffer->m_Size;
                }
                return &(m_pBuffer->m_pData[idx]);
            }

            u32 Offset() const
            {
                return m_Offset;
            }

            void Begin()
            {
                m_Offset = 0;
            }
            bool IsEnd() const
            {
                return m_Offset >= m_pBuffer->m_DataNum;
            }
            void Next()
            {
                if (IsEnd() == false)
                {
                    ++m_Offset;
                }
            }

        private:
            Buffer<T>* m_pBuffer;
            u32 m_Offset;
        };

    protected:
        T* m_pData;
        u32 m_Size;
        u32 m_StartIdx;
        u32 m_StartSequenceId;
        u32 m_DataNum;
    };


    // Send buffer.
    class SendBuffer : public Buffer<SendData>
    {
    public:
        u32 GetAvailableDataNum() const
        {
            return m_Size - m_DataNum;
        }
        SendData* Assign();
    };


    // Receive buffer.
    class ReceiveBuffer : public Buffer<ReceiveData>
    {
    public:
        ReceiveData* Assign(u32 sequenceId);
    };

private:
    transport::ProtocolId m_ProtocolId;
    StationIndex m_SourceStationIndex;
    StationIndex m_DestinationStationIndex;

    SendBuffer m_SendBuffer;
    ReceiveBuffer m_ReceiveBuffer;

    u32 m_PayloadSize;

    u64 m_DispatchTime;
    u64 m_DefaultResendTimeSpan;
    u32 m_AckOrigin;
    u64 m_AckFlag;
    bool m_IsSendAck;

#if NN_PIA_EXPERIMENT_RELIABLE_DETAIL_INFO
    u32 m_ResendCount;  // The number of resends.
#endif

    size_t m_LargeDataSendSize;
    common::Time m_LargeDataSendStartTime;

    static f32 s_RtoRate;                              // Resend timeout rate (coefficient).
    static size_t s_RttSamplingNum;                    // Number of samples used to calculate RTT.
    static ResendTimeCallback s_pfResendTimeCallback;  // Resend time calculation callback.
};
//! @endcond
}
}
}  // end of namespace nn::pia::transport
