﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_RoundRobinUnreliableProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>
#include <pia/transport/transport_UnreliableProtocol.h>

namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  This class inherits the <tt>UnreliableProtocol</tt> class, and adds features to send data while distributing the load.
@date 2013-03-19  Initial version.


*/
class RoundRobinUnreliableProtocol : public UnreliableProtocol
{
public:
/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
@details  Applications should not call this constructor directly.
Use the <tt>@ref Transport::CreateProtocol</tt> function to create an instance.

*/
    RoundRobinUnreliableProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
@details  Applications should not call this destructor directly.
Use the <tt>@ref Transport::DestroyProtocol</tt> function to destroy an instance.

*/
    virtual ~RoundRobinUnreliableProtocol(void);
    //! @endcond


    PIA_PROTOCOL_TYPE_INFO(PROTOCOL_TYPE_ROUNDROBIN_UNRELIABLE);


/*!
@cond PRIVATE
@brief  Starts communication.
@param[in] myStationIndex  Indicates the local <tt>StationIndex</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.


*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup(void);
    //! @endcond


/*!
@brief  Sends data to multiple stations.
@details  Sends data to the <tt>stationNum</tt> stations specified. Sends to stations in order of <tt>StationIndex</tt>, from lowest to highest, starting with the next largest <tt>StationIndex</tt> after that of the local station.
The next time the call is made, the starting point is the last <tt>StationIndex</tt> that the station sent data to.
For example, consider a session in which stations A, B, C, D, E, and F are participating. The local station is F. If <tt>SendToRoundRobin</tt> is called with a recipient count of 2, data is sent as follows.
First call: <tt>SendToRoundRobin(Data1)</tt> — Data1 is sent to A and B.
Second call: <tt>SendToRoundRobin(Data2)</tt> — Data2 is sent to C and D.
Third call: <tt>SendToRoundRobin(Data3)</tt> — Data3 is sent to E and A.
Fourth call: <tt>SendToRoundRobin(Data4)</tt> — Data4 is sent to B and C.
Fifth call: <tt>SendToRoundRobin(Data5)</tt> — Data5 is sent to D and E.

Calling this function alone does not cause the data to actually be sent.
This function copies the data passed in by the application into the Pia library's packet buffer.
The Pia library's send thread periodically sends the packets.
@param[in] pData  Specifies a pointer to the start of the data to send.
@param[in] dataSize  Specifies the size of the data to send. The value is in bytes. This size must not be greater than the value returned by the <tt>GetDataSizeLimit</tt> function.
@param[in] stationNum  Specifies the number of stations you want to send data to.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. This error is also returned when the size of the data to send is too large. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>Initialize</tt> function may not have been called. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInCommunication  Indicates that communication is not possible. Handle appropriately in the application.
@see  Send, Receive, MAX_DATA_SIZE, GetDataSizeLimit, IsInCommunication






*/
    nn::Result SendToRoundRobin(const bit8* pData, size_t dataSize, u32 stationNum);


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


private:
    u32 m_LatestPariticipatingBitmap;
    StationIndex m_LastSentStationIndex;
};
}
}
}  // end of namespace nn::pia::transport
