﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_SequenceIdController.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>


namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  Manages the sequence ID of the packet.
In conjunction with this, includes a profiling feature for packet loss.

@date 2013-03-25  Removed the <tt>GetProfiledToAllPacketNum</tt> function and added the <tt>GetProfiledMulticastPacketNum</tt> function.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-06-21  Made the constructor and destructor private.
@date 2012-04-19  Made it possible to count the packets addressed to <tt>STATION_INDEX_ALL</tt>.
@date 2012-04-06  Initial version.
*/
class SequenceIdController : public nn::pia::common::RootObject
{
public:
/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
*/
    SequenceIdController(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    virtual ~SequenceIdController(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Initiates communication.
@return  Returns <tt>nn::ResultSuccess</tt>.
*/
    nn::Result Startup();
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    void Cleanup();
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the next sequence ID to send.
@return  The next sequence ID to send.
*/
    u16 GetNextSendSequenceId();
    //! @endcond


/*!
@cond PRIVATE
@brief  Checks a received sequence ID.
@return  Returns <tt>true</tt> if it is a valid sequence ID, or <tt>false</tt> otherwise.
*/
    bool CheckReceivedSequenceId(u16 sequenceId);
    //! @endcond


/*!
@cond PRIVATE
@brief  Profiles the packets to <tt>STATION_INDEX_ALL</tt>.
*/
    void CountReceivedToAllPacket();
    //! @endcond


/*!
@brief  Starts profiling.
*/
    void StartProfile();


/*!
@brief  Stops profiling.
*/
    void StopProfile();


/*!
@brief  Gets the total number of profiled packets.
This function is thread-safe.
@return  The total number of profiled packets.
*/
    u16 GetProfiledTotalPacketNum() const
    {
        return m_ProfileTotalPacketCount;
    }


/*!
@brief  Gets the number of profiled packet losses.
This function is thread-safe.
@return  The number of profiled packet losses.
*/
    u16 GetProfiledPacketLossNum() const
    {
        return m_ProfilePacketLossCount;
    }


/*!
@brief  Gets the number of multicast packets profiled.
This function is thread-safe.
@return  The number of multicast packets profiled.
*/
    u16 GetProfiledMulticastPacketNum() const
    {
        return m_ProfileMulticastPacketCount;
    }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


/*!
@cond PRIVATE
@brief  Invalid sequence ID value.
@return  Invalid sequence ID value.
*/
    static u16 GetInvalidSequenceId()
    {
        return c_InvalidSequenceId;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the total number of profiled packets (for <tt>TransportAnalyzer</tt>).
@return  The total number of profiled packets.
*/
    u32 GetProfiledTotalPacketNumForTransportAnalyzer() const
    {
        return m_ProfileTotalPacketCountForTransportAnalyzer;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the number of profiled packet losses (for <tt>TransportAnalyzer</tt>).
@return  The number of profiled packet losses.
*/
    u32 GetProfiledPacketLossNumForTransportAnalyzer() const
    {
        return m_ProfilePacketLossCountForTransportAnalyzer;
    }
    //! @endcond

private:
    // Will not implement.
    SequenceIdController(const SequenceIdController& rhs);
    bool operator==(const SequenceIdController& rhs) const;
    bool operator!=(const SequenceIdController& rhs) const;
    SequenceIdController& operator=(const SequenceIdController& rhs);


    u16 m_SendSequenceId;
    u16 m_LastReceiveSequenceId;

    u16 m_ProfileTotalPacketCount;
    u16 m_ProfilePacketLossCount;
    u16 m_ProfileMulticastPacketCount;
    bool m_IsProfiling;

    u32 m_ProfileTotalPacketCountForTransportAnalyzer;
    u32 m_ProfilePacketLossCountForTransportAnalyzer;

    static const u16 c_InvalidSequenceId = 0;
};
}
}
}  // end of namespace nn::pia::transport
