﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_Station.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/common/common_StationAddress.h>
#include <pia/common/common_PacketDefine.h>
#include <pia/common/common_Time.h>

#include <pia/transport/transport_SequenceIdController.h>


// Forward declarations.
namespace nn
{
namespace pia
{
namespace common
{
class StepSequenceJob;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{


// Forward declarations.
class StationProtocol;
class StationProtocolReliable;
class ConnectStationJob;
class DisconnectStationJob;
class ProcessConnectionRequestJob;
class ReliableSlidingWindow;
class NetworkFactory;


/*!
@brief  Represents a station. A station refers to a machine participating in a Pia session.

@date 2015-02-19  Made the <tt>GetStationIndex</tt> function public.
@date 2013-10-16  Added <tt>GetRttSamplingNum</tt> to get the number of samples to use for calculating RTT.
@date 2013-08-08  Added <tt>SetInitialRttPulseInterval</tt> and <tt>GetInitialRttPulseInterval</tt>.
@date 2013-05-01  Added the <tt>GetIdentificationToken</tt>, <tt>GetPlayerName</tt>, and <tt>GetPrincipalId</tt> functions.
@date 2013-03-28  Added definitions for <tt>Station::IdentificationInfo</tt> and <tt>Station::PlayerName</tt>.
@date 2013-01-22  Added the <tt>GetRtt</tt> function that allows users to specify the number of samples to use when calculating RTT.
@date 2013-01-17  Added the <tt>SetRttPulseInterval</tt> and <tt>GetRttPulseInterval</tt> member functions for setting and getting the pulse interval used when calculating RTT.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-06-21  Made functions and constants, the use of which were not necessary for applications, private.
@date 2012-06-14  Added <tt>Station::GetRtt</tt> function and <tt>Station::INVALID_RTT</tt> definition.
@date 2012-05-30  Made the send and receive functions private.
@date 2012-05-01  Added a definition for <tt>Station::IdentificationToken</tt>.
@date 2012-04-28  Fixed an error in the notation of the return values for the <tt>Station::Startup</tt> function.
@date 2012-04-06  Initial version.
*/
class Station : public common::RootObject
{
    friend class StationProtocolReliable;

public:
    static const size_t MAX_DATA_SIZE = common::PROTOCOL_MESSGAE_PAYLOAD_SIZE_MAX;  // Maximum data size that can be sent or received with <tt>Send/Receive</tt>. The value is in bytes.
    static const s32 INVALID_RTT = -1;                                              //!< An invalid value that the <tt>GetRtt</tt> function can return.

/*!
@brief  Enumerates <tt>Station</tt> states.
*/
    enum StationState
    {
        STATE_INVALID,             //!< Indicates an invalid <tt>StationState</tt>.
        STATE_INITIAL,             //!< Indicates that this station has already been initialized.
        STATE_CREATING_SESSION,    //!< Indicates a state in which a request has been made to connect to this station, and that the connection process has started.
        STATE_JOINING_SESSION,     //!< Indicates a state in which a request to connect from this station has been received, and that the connection process has started.
        STATE_WAITING_CONNECTION,  //!< Indicates a state of waiting for communication with this station to be established.
        STATE_PARTICIPATING,       //!< Indicates a state in which communication with this station is possible.
        STATE_LEAVING,             //!< Indicates a state in which communication with this station is being disconnected.
        STATE_MAX = STATE_LEAVING  //!< This is a sentinel value indicating the maximum value of <tt>StationState</tt>.
    };


/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).

@details  The <tt>StationIndex</tt> of the instance is set to <tt>STATION_INDEX_INVALID</tt>.
*/
    Station(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    ~Station(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Initializes the <tt>Station</tt> instance.

@param[in] pFactory  Specifies the concrete class that inherits <tt>NetworkFactory</tt>. Job instances are generated using this factory.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  Finalize
*/
    nn::Result Initialize(NetworkFactory* pFactory);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the <tt>Station</tt> instance.

@see  Initialize
*/
    void Finalize(void);
    //! @endcond


/*!
@brief  Gets the <tt>StationIndex</tt>.

@details  This function is thread-safe.
@details  Don't call this function when the application is using the joint session feature.
This results in undefined operations.

@return  Returns the <tt>StationIndex</tt> set in the instance.
*/
    StationIndex GetStationIndex(void) const
    {
        return m_StationIndex;
    }


/*!
@cond PRIVATE
@brief  Sets the <tt>StationIndex</tt>.
@details  Use this function in conjunction with <tt>SetStationId</tt>.
If possible, call this function at nearly the same time as <tt>SetStationId</tt>. The goal is to preserve the consistency of <tt>StationIndex</tt> and <tt>StationId</tt> as they are in the <tt>Station</tt> instance.

It is likely that situations will arise where <tt>StationIndex</tt> will be set before <tt>StationId</tt>. When that happens, make sure that you subsequently call <tt>SetStationId</tt>.


@param[in] idx  Specifies the <tt>StationIndex</tt> to set in this instance.
@see  GetStationIndex, SetStationId
*/
    void SetStationIndex(StationIndex idx);
    //! @endcond


/*!
@brief  Gets the <tt>StationId</tt>.

@return  Returns the <tt>StationId</tt> corresponding to the <tt>Station</tt> instance.
*/
    StationId GetStationId(void) const
    {
        return m_StationId;
    }


/*!
@cond PRIVATE
@brief  Sets the <tt>StationId</tt>.
@details  Use this function in conjunction with <tt>SetStationIndex</tt>.
If possible, call this function at nearly the same time as <tt>SetStationIndex</tt>. The goal is to preserve the consistency of <tt>StationIndex</tt> and <tt>StationId</tt> as they are in the <tt>Station</tt> instance.

It is likely that situations will arise where <tt>StationIndex</tt> will be set before <tt>StationId</tt>. When that happens, make sure that you subsequently call <tt>SetStationId</tt>.


@param[in] id  Specifies the <tt>StationId</tt> to set in this instance.
@see  GetStationId, SetStationIndex
*/
    void SetStationId(StationId id);
    //! @endcond


/*!
@brief  Gets the station address.

@return  Returns a const reference for the <tt>StationAddress</tt> that is set.
*/
    const common::StationAddress& GetStationAddress(void) const
    {
        return m_StationAddress;
    }


/*!
@cond PRIVATE
@brief  Sets the station address.

@param[in] addr  Specifies the <tt>StationAddress</tt> to set in this instance.
@see  GetStationAddress
*/
    void SetStationAddress(const common::StationAddress& addr)
    {
        m_StationAddress = addr;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>ConnectStationJob</tt> instance (non-<tt>const</tt> version).
@return  Returns a pointer to the <tt>ConnectStationJob</tt> instance.
*/
    ConnectStationJob* GetConnectStationJob(void)
    {
        return m_pConnectStationJob;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>ConnectStationJob</tt> instance (<tt>const</tt> version).
@return  Returns a <tt>const</tt> pointer to the <tt>ConnectStationJob</tt> instance.
*/
    const ConnectStationJob* GetConnectStationJob(void) const
    {
        return m_pConnectStationJob;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>DisconnectStationJob</tt> instance (non-<tt>const</tt> version).
@return  Returns a pointer to the <tt>DisconnectStationJob</tt> instance.
*/
    DisconnectStationJob* GetDisconnectStationJob(void)
    {
        return m_pDisconnectStationJob;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>DisconnectStationJob</tt> instance (<tt>const</tt> version).
@return  Returns a <tt>const</tt> pointer to the <tt>DisconnectStationJob</tt> instance.
*/
    const DisconnectStationJob* GetDisconnectStationJob(void) const
    {
        return m_pDisconnectStationJob;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>ProcessConnectionRequestJob</tt> instance (non-<tt>const</tt> version).
@return  Returns a pointer to the <tt>ProcessConnectionRequestJob</tt> instance.
*/
    ProcessConnectionRequestJob* GetProcessConnectionRequestJob(void)
    {
        return m_pProcessConnectionRequestJob;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>ProcessConnectionRequestJob</tt> instance (<tt>const</tt> version).
@return  Returns a <tt>const</tt> pointer to the <tt>ProcessConnectionRequestJob</tt> instance.
*/
    const ProcessConnectionRequestJob* GetProcessConnectionRequestJob(void) const
    {
        return m_pProcessConnectionRequestJob;
    }
    //! @endcond


/*!
@brief  Gets a pointer to the <tt>SequenceIdController</tt> instance (non-<tt>const</tt> version).
This function is thread-safe.
@return  Returns a pointer to the <tt>SequenceIdController</tt> instance.
*/
    SequenceIdController* GetSequenceIdController(void)
    {
        return &m_SequenceIdController;
    }


/*!
@brief  Gets a pointer to the <tt>SequenceIdController</tt> instance (<tt>const</tt> version).
This function is thread-safe.
@return  Returns a <tt>const</tt> pointer to the <tt>SequenceIdController</tt> instance.
*/
    const SequenceIdController* GetSequenceIdController(void) const
    {
        return &m_SequenceIdController;
    }


/*!
@cond PRIVATE
@brief  Gets the <tt>ConnectionId</tt> for the sending station.
@return  Returns the <tt>ConnectionId</tt> for the sending station.
*/
    u8 GetSendingConnectionId(void) const
    {
        return m_SendConnectionId;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the <tt>ConnectionId</tt> for the sending station.
@param[in] id  Specifies the <tt>ConnectionId</tt> for sending to set in this instance.
*/
    void SetSendingConnectionId(u8 id)
    {
        m_SendConnectionId = id;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the <tt>ConnectionId</tt> for the station to receive data.
@return  Returns the<tt>ConnectionId</tt> of the receiving station.
*/
    u8 GetReceivingConnectionId(void) const
    {
        return m_RecvConnectionId;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the <tt>ConnectionId</tt> for the station to receive data.
@param[in] id  Specifies the <tt>ConnectionId</tt> for receiving to set in this instance.
*/
    void SetReceivingConnectionId(u8 id)
    {
        m_RecvConnectionId = id;
    }
    //! @endcond


/*!
@brief  Gets the state of the <tt>Station</tt> instance.

@details  This function is thread-safe.

@return  Returns the state of the <tt>Station</tt>.
*/
    StationState GetState(void) const
    {
        return m_StationState;
    }


/*!
@cond PRIVATE
@brief  Sets the state of the <tt>Station</tt>.

@param[in] state  Specifies the state to set.
@see  GetState
*/
    void SetState(StationState state)
    {
        m_StationState = state;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Performs <tt>Station</tt> startup processing.
@details  After calling this function, set the <tt>StationId</tt> using the <tt>SetStationId</tt> function. Preserve the consistency between <tt>StationIndex</tt> and <tt>StationId</tt>.

@details  After this function is called, the <tt>StationId</tt> for this instance is set to <tt>STATION_ID_INVALID</tt>.

@param[in] pStationProtocol  Specifies a pointer to the <tt>StationProtocol</tt> to set in this instance.
@param[in] idx  Specifies the <tt>StationIndex</tt> to set in this instance.
@param[in] addr  Specifies the <tt>StationAddress</tt> object to set in this instance.
@return  Returns <tt>true</tt> if successful; otherwise returns <tt>false</tt>.
@see  Cleanup, SetStationId
*/
    bool Startup(StationProtocol* pStationProtocol, StationIndex idx, const common::StationAddress& addr);
    //! @endcond


/*!
@cond PRIVATE
@brief  Performs <tt>Station</tt> startup processing.
The <tt>StationIndex</tt> of the instance is set to <tt>STATION_INDEX_INVALID</tt>.
@details  After this function is called, the <tt>StationId</tt> for this instance is set to <tt>STATION_ID_INVALID</tt>.

@param[in] pStationProtocol  Specifies a pointer to the <tt>StationProtocol</tt> to set in this instance.
@param[in] addr  Specifies the <tt>StationAddress</tt> object to set in this instance.
@return  Returns <tt>true</tt> if successful; otherwise returns <tt>false</tt>.
@see  Cleanup
*/
    bool Startup(StationProtocol* pStationProtocol, const common::StationAddress& addr);
    //! @endcond


/*!
@cond PRIVATE
@brief  Performs <tt>Station</tt> startup processing.
The <tt>StationIndex</tt> of this instance is set to <tt>STATION_INDEX_INVALID</tt> and <tt>StationAddress</tt> is set to the same state as if the <tt>StationAddress::Clear</tt> function had been run.

@details  After this function is called, the <tt>StationId</tt> for this instance is set to <tt>STATION_ID_INVALID</tt>.

@param[in] pStationProtocol  Specifies a pointer to the <tt>StationProtocol</tt> to set in this instance.
@return  Returns <tt>true</tt> if successful; otherwise returns <tt>false</tt>.
@see  Cleanup
*/
    bool Startup(StationProtocol* pStationProtocol);
    //! @endcond


/*!
@cond PRIVATE
@brief  Cleans up the <tt>Station</tt> instance.
This function is paired with the <tt>Startup</tt> function. The content set by the <tt>Startup</tt> function is cleared.

@see  Startup
*/
    void Cleanup(void);
    //! @endcond


/*!
@brief  Gets the round-trip time (RTT) between the local <tt>Station</tt> and this <tt>Station</tt>.

@details  The system sends pulses to other stations periodically to get the round-trip time (RTT).
At particular times, however, the system may have never sent pulses to or received responses from other stations. One example is when it has just connected to the session.
At these times, the <tt>GetRtt</tt> function returns <tt>INVALID_RTT</tt>, but if you call the <tt>GetRtt</tt> function after a short time has passed, the correct value is returned.
This function returns the median value of multiple samples.
The value obtained from this function is relatively stable and not easily affected by minor changes in network environment.

@return  Returns the RTT value. The value is in milliseconds.
@retval INVALID_RTT  Specifies a negative value that indicates that the RTT value cannot be obtained in this state.
@see  SetRttPulseInterval, GetRttPulseInterval, SetInitialRttPulseInterval, GetInitialRttPulseInterval, GetRttSamplingNum



*/
    s32 GetRtt(void) const;


/*!
@brief  Gets the round-trip time (RTT) between the local <tt>Station</tt> and this <tt>Station</tt>. You can specify the number of samples to use for calculating RTT.

@details  The system stores several samples obtained from RTT pulse transmissions and responses.
The version of the <tt>GetRtt</tt> function that does not take arguments returns the median value for all stored sample values. The version that takes an argument returns the median value for the specified number of samples.
When specified, the latest samples are used. If you specify a small value, the RTT value strongly reflects the most recent network connection status.
However, if the value of the argument is too small, the RTT value is susceptible to fluctuation.
@param[in] samplingNum  Specifies the number of most recent samples to use in calculating the RTT.
@return  Returns the RTT value. The value is in milliseconds.
@retval INVALID_RTT  Specifies a negative value that indicates that the RTT value cannot be obtained in this state.
@see  SetRttPulseInterval, GetRttPulseInterval, SetInitialRttPulseInterval, GetInitialRttPulseInterval, GetRttSamplingNum


*/
    s32 GetRtt(size_t samplingNum) const;


/*!
@cond PRIVATE
@brief  Gets the most recent round-trip time (RTT) measured between the local station and this one.

@return  Returns the most recent RTT value measured by the system. The value is in milliseconds.
@retval INVALID_RTT  Specifies a negative value that indicates that the RTT value cannot be obtained in this state.

@see  GetRtt, SetRttPulseInterval, GetRttPulseInterval, SetInitialRttPulseInterval, GetInitialRttPulseInterval, GetRttSamplingNum
*/
    s32 GetLatestRtt(void) const;
    //! @endcond


/*!
@brief  Gets the number of RTT (Round Trip Time) samples held by the system.

@details  The system sends pulses to other stations periodically to get the round-trip time (RTT), and the time required for the response is saved as an RTT sample.
At particular times, however, the system cannot obtain a sufficient number of samples because it has just begun sending pulses and receiving responses. One example is when it has just connected to the session.
This function returns the number of samples that are used by functions such as <tt>GetRtt</tt> to calculate the RTT value. It gives an idea of the reliability and precision of the RTT value.

@return  Returns the number of samples held by the system.
@retval 0  A value of <tt>0</tt> specifies a state in which the RTT value could not be obtained.
@see  GetRtt, SetRttPulseInterval, GetRttPulseInterval, SetInitialRttPulseInterval, GetInitialRttPulseInterval



*/
    size_t GetRttSamplingNum(void) const;


/*!
@brief  Sets the interval for sending pulses for RTT calculation.

@param[in] msec  Specifies the interval at which to send pulses. The value is in milliseconds.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Returned if a non-positive integer is specified. Programming error. Fix your program so that this error is not returned.
@see  GetRttPulseInterval, GetRtt, SetInitialRttPulseInterval, GetInitialRttPulseInterval, GetRttSamplingNum
*/
    static nn::Result SetRttPulseInterval(s32 msec);


/*!
@brief  Gets the interval for sending pulses for RTT calculation.

@return  Returns the specified interval for sending pulses. The value is in milliseconds.
@see  SetRttPulseInterval, GetRtt, SetInitialRttPulseInterval, GetInitialRttPulseInterval, GetRttSamplingNum
*/
    static s32 GetRttPulseInterval(void);


/*!
@brief  Sets the interval for sending the initial pulse for RTT calculation.

@details  The pulse for calculating the RTT is sent starting when the station joins a session.
There are not enough samples available to calculate the RTT immediately after a station joins a session.
The system sends pulses at a shorter interval than usual until enough samples are available to quickly get the number of samples required.
This member function sets the interval at which pulses are set during this initial period.

@param[in] msec  Specifies the interval at which to send pulses initially. The value is in milliseconds.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Returned if a non-positive integer is specified. Programming error. Fix your program so that this error is not returned.
@see  GetInitialRttPulseInterval, GetRtt, SetRttPulseInterval, GetRttPulseInterval, GetRttSamplingNum


*/
    static nn::Result SetInitialRttPulseInterval(s32 msec);


/*!
@brief  Gets the interval for sending initial pulses for RTT calculation.

@details  The pulse for calculating the RTT is sent starting when the station joins a session.
There are not enough samples available to calculate the RTT immediately after a station joins a session.
The system sends pulses at a shorter interval than usual until enough samples are available to quickly get the number of samples required.
This member function gets the interval at which pulses are set during this initial period.

@return  Returns the specified interval for sending pulses. The value is in milliseconds.
@see  SetInitialRttPulseInterval, GetRtt, SetRttPulseInterval, GetRttPulseInterval, GetRttSamplingNum


*/
    static s32 GetInitialRttPulseInterval(void);


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>StationProtocol</tt> instance stored in this <tt>Station</tt> instance (non-<tt>const</tt> version).

@return  Returns a pointer to a <tt>StationProtocol</tt> instance. Returns a <tt>NULL</tt> pointer if no <tt>StationProtocol</tt> instance has been configured.
@see  Startup
*/
    StationProtocol* GetStationProtocol(void)
    {
        return m_pStationProtocol;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to the <tt>StationProtocol</tt> instance stored in this <tt>Station</tt> instance (<tt>const</tt> version).

@return  Returns a pointer to a <tt>StationProtocol</tt> instance. Returns a <tt>NULL</tt> pointer if no <tt>StationProtocol</tt> instance has been configured.
@see  Startup
*/
    const StationProtocol* GetStationProtocol(void) const
    {
        return m_pStationProtocol;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Records the last time that a packet was sent.

@details  This function is mainly called from <tt>KeepAliveSender</tt>.
@param[in] t  Specifies the time the packet was sent.
@see  GetLatestPacketSentTime
*/
    void SetLatestPacketSentTime(const common::Time& t)
    {
        m_LatestPacketSentTime = t;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the last time that a packet was sent.

@details  This function is mainly called from <tt>KeepAliveSender</tt>.
@return  Returns the time that a packet was last sent.
@see  SetLatestPacketSentTime
*/
    const common::Time& GetLatestPacketSentTime(void) const
    {
        return m_LatestPacketSentTime;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Records the last time that a packet was received.

@details  This function is mainly called from <tt>KeepAliveReceiver</tt>.
@param[in] t  Specifies the time the packet was received.
@see  GetLatestPacketReceivedTime
*/
    void SetLatestPacketReceivedTime(const common::Time& t)
    {
        m_LatestPacketReceivedTime = t;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the last time that a packet was received.

@details  This function is mainly called from the session layer. It is used for keep-alive determinations.
@return  Returns the last time that a packet was received.
@see  SetLatestPacketReceivedTime

*/
    const common::Time& GetLatestPacketReceivedTime(void) const
    {
        return m_LatestPacketReceivedTime;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Forcibly finalizes and initializes station connection and disconnection jobs.
*/
    void CleanupJobs();
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


    static const size_t IDENTIFICATION_TOKEN_MAX_DATA_SIZE = 32;  //!< Specifies the maximum data size for an identification token.
    static const size_t PLAYER_NAME_MAX_DATA_SIZE = 17;           //!< Specifies the maximum data size for a player name. The final character is for the terminating null character.

/*!
@brief  Stores the ID token associated with a <tt>Station</tt>.

@details  The ID token is shared with connected stations.
*/
    struct IdentificationToken
    {
        bit8 data[IDENTIFICATION_TOKEN_MAX_DATA_SIZE];  //!< The ID token data array.
    };

/*!
@brief  Stores the player name struct associated with a <tt>Station</tt>.

@details  The name is shared with connected stations.
*/
    struct PlayerName
    {
        s16 data[PLAYER_NAME_MAX_DATA_SIZE];  //!< Specifies the player name as a data array.
        u8 length;                            //!< Specifies the length of the player name. The name must be no longer than 16 characters (not including the terminating <tt>null</tt> character).
        u8 language;                          //!< Specifies the language of the player name. @if CAFE_DOC Use one of the values defined in <tt>nn::fp::Language</tt>. @endif @if CTR_DOC You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage</tt> function if you want to use the language setting of the system. @endif
    };

/*!
@cond PRIVATE
@brief  Stores ID information (the token and name) associated with a <tt>Station</tt>.
*/
    struct IdentificationInfo
    {
        IdentificationToken token;
        PlayerName name;
        u8 platform;
        u32 gid;
    };
    //! @endcond

/*!
@brief  Gets the ID token of the station.

@details  If called by the local station, this function gets the ID token set in the <tt>@ref session::Session::Startup</tt> function.
If called by a station other than the station, this function gets the ID token shared at time of connection.
If no token has been configured, this function succeeds but the obtained ID token data is filled with <tt>NULL</tt> values.

The values of the ID token specified as an out parameter never change if the function fails to acquire an ID token.

@param[out] pToken  Specifies the ID token of the <tt>Station</tt> if it was successfully retrieved.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the ID token is successfully retrieved. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultInvalidState  Specifies that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Specifies that an invalid pointer was specified as an argument. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Specifies that the function was not able to obtain the ID token of the specified station. The sharing of the ID token might not have been completed. Handle appropriately in the application.
*/
    nn::Result GetIdentificationToken(IdentificationToken* pToken);

/*!
@brief  Gets the name of the station.

@details  For a local station, gets the name set in the <tt>@ref session::Session::Startup</tt> function.
For a non-local station, gets the name shared on connection.

The value of the name specified as an out parameter never changes if the function fails to acquire a name.

@param[out] pName  Specifies the name of the <tt>Station</tt> if successfully retrieved.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the name is successfully retrieved. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultInvalidState  Specifies that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Specifies that an invalid pointer was specified as an argument. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Specifies that the function was not able to get the name of the specified station. The sharing of the name might not have been completed. Handle appropriately in the application.
*/
    nn::Result GetPlayerName(PlayerName* pName);

/*!
@brief  Gets the principal ID of the station.

@details  For a local station, gets the principal ID specified in the <tt>nn::nex::Credentials</tt> function.
You can get this value after the <tt>@ref inet::NexFacade::StartNatSessionAsync</tt> asynchronous process completes.

For a non-local station, gets the principal ID shared on connection.

The value of the principal ID specified as an out parameter never changes if the function fails to acquire the principal ID.

@param[out] pPrincipalId  Specifies the principal ID of the <tt>Station</tt> if successfully retrieved.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the name is successfully retrieved. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
@retval ResultInvalidState  Specifies that the function cannot be executed in this state. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Specifies that an invalid pointer was specified as an argument. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Specifies that the function was not able to obtain the principal ID of the specified station. The sharing of the ID token might not have been completed. Handle appropriately in the application.
*/
    nn::Result GetPrincipalId(u32* pPrincipalId);

/*!
@cond PRIVATE
@brief  Sets whether the connection route to the target <tt>Station</tt> is a direct connection.
*/
    void SetConnectionRoute(bool isDirect)
    {
        m_ConnectionRouteIsDirect = isDirect;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether the connection route to the target <tt>Station</tt> is a relay connection.
@return  Returns <tt>true</tt> if it is a relay connection, and <tt>false</tt> otherwise.
*/
    bool IsConnectionRouteRelay() const;
    //! @endcond

/*!
@cond PRIVATE
@brief  Gets whether the connection route to the target <tt>Station</tt> is a direct connection.
@return  Returns <tt>true</tt> if it is a direct connection, and <tt>false</tt> otherwise.
*/
    bool IsConnectionRouteDirect() const;
    //! @endcond


#if NN_PIA_EXPERIMENT_HANDSHAKE_CHECK
/*!
@cond PRIVATE
@brief  Returns whether the target station has participated in the session.
@return  Returns <tt>true</tt> when participating and <tt>false</tt> otherwise.
@see  SetJoinFlag
*/
    bool GetJoinFlag() const { return m_JoinFlag; }
    //! @endcond

/*!
@cond PRIVATE
@brief  Sets whether the target station has participated in the session.
Expect it to be called when the <tt>Join</tt> and <tt>Leave</tt> events occur.
@param[in] flag  Set to <tt>true</tt> when marking as "participated," and to <tt>false</tt> when marking as "not participated."
@see  GetJoinFlag
*/
    void SetJoinFlag(bool flag) { m_JoinFlag = flag; }
    //! @endcond
#endif  // end of NN_PIA_EXPERIMENT_HANDSHAKE_CHECK

/*!
    @cond PRIVATE
*/
    bool GetDisconnectImmidiatelyFlag() const { return m_DisconnectImmidiately; }
    //! @endcond

/*!
    @cond PRIVATE
*/
    void SetDisconnectImmidiatelyFlag(bool flag) { m_DisconnectImmidiately = flag; }
    //! @endcond
private:
    // A helper function prepared to bring together the common processes that occur in the <tt>Startup</tt> and <tt>Cleanup</tt> functions.
    // The first argument switches it between the startup and cleanup processes.
    // The second and subsequent arguments are set unconditionally to member variables (that is, there is no argument check).
    void helperStartupCleanup(bool bStartup, StationProtocol* pStationProtocol, StationIndex idx, const common::StationAddress& addr, StationState state);

    common::StationAddress m_StationAddress;
    StationIndex m_StationIndex;
    StationId m_StationId;
    StationState m_StationState;
    ReliableSlidingWindow* m_pReliableSlidingWindow;
    StationProtocol* m_pStationProtocol;
    ConnectStationJob* m_pConnectStationJob;
    DisconnectStationJob* m_pDisconnectStationJob;
    ProcessConnectionRequestJob* m_pProcessConnectionRequestJob;
    SequenceIdController m_SequenceIdController;
    u8 m_SendConnectionId;
    u8 m_RecvConnectionId;

    bool m_ConnectionRouteIsDirect;

    common::Time m_LatestPacketSentTime;
    common::Time m_LatestPacketReceivedTime;

    bool m_DisconnectImmidiately;

#if NN_PIA_EXPERIMENT_HANDSHAKE_CHECK
    bool m_JoinFlag;
#endif
};
}
}
}  // end of namespace nn::pia::transport
