﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_StationConnectionInfo.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>
#include <pia/transport/transport_StationLocation.h>

namespace nn
{
namespace pia
{
namespace transport
{

/*!
@brief  Represents an address used to uniquely identify stations.

@details  This class includes two instances of <tt>StationLocation</tt>. One is for the main address and one is for a sub-address.

@date 2012-04-16  Added a serialized group of API functions for streamlining size during serialization.
@date 2012-04-06  Initial version.
*/
class StationConnectionInfo : public nn::pia::common::RootObject
{
public:
/*!
@brief  Instantiates the object.
*/
    StationConnectionInfo(void);

/*!
@brief  This is the copy constructor.
@param[in] o  Specifies the argument to the copy constructor.
*/
    StationConnectionInfo(const StationConnectionInfo& o);

/*!
@brief  Destroys the object.
*/
    virtual ~StationConnectionInfo(void);

    void SetStationConnectionInfo(const StationConnectionInfo& o);

/*!
@brief  This is the assignment operator.

@param[in] o  The <tt>StationConnectionInfo</tt> object to assign.
@return  Returns a reference to the <tt>this</tt> object.
*/
    StationConnectionInfo& operator=(const StationConnectionInfo& o);

    bool operator==(const StationConnectionInfo& o) const;

    bool operator!=(const StationConnectionInfo& o) const
    {
        return !(*this == o);
    }

    bool HasPrivateAddress() const
    {
        return m_SubLocation.GetStationAddress().GetInetAddress().IsValid();
    }

    bool IsSamePublicAddress(const StationConnectionInfo& o) const
    {
        return m_MainLocation.IsSameAddress(o.m_MainLocation);
    }

/*!
@brief  Gets a <tt>const</tt> reference to the <tt>StationLocation</tt> (main address).

@returns Returns a <tt>const</tt> reference to the <tt>StationLocation</tt> (main address).
*/
    const StationLocation& GetMainLocation(void) const
    {
        return m_MainLocation;
    }

/*!
@brief  Gets a reference to the <tt>StationLocation</tt> (main address).

@returns Returns a reference to the <tt>StationLocation</tt> (main address).
*/
    StationLocation& GetMainLocation(void)
    {
        return m_MainLocation;
    }

/*!
@brief  Sets the <tt>StationLocation</tt> (main address).

@param[in] o  Specifies the <tt>StationLocation</tt> to set.
*/
    void SetMainLocation(const StationLocation& o)
    {
        m_MainLocation.SetStationLocation(o);
    }

/*!
@brief  Gets a <tt>const</tt> reference to the <tt>StationLocation</tt> (sub-address).

@returns Returns a <tt>const</tt> reference to the <tt>StationLocation</tt> (sub-address).
*/
    const StationLocation& GetSubLocation(void) const
    {
        return m_SubLocation;
    }

/*!
@brief  Gets a reference to the <tt>StationLocation</tt> (sub-address).

@returns Returns a reference to the <tt>StationLocation</tt> (sub-address).
*/
    StationLocation& GetSubLocation(void)
    {
        return m_SubLocation;
    }

/*!
@brief  Sets the <tt>StationLocation</tt> (sub-address).

@param[in] o  Specifies the <tt>StationLocation</tt> to set.
*/
    void SetSubLocation(const StationLocation& o)
    {
        m_SubLocation.SetStationLocation(o);
    }

/*!
@brief  Gets a <tt>const</tt> reference to the main address of the <tt>StationLocation</tt>.
The main address is intended to be used for public addresses.

@return  Returns a <tt>const</tt> reference to the main address of the <tt>StationLocation</tt>.
*/
    const StationLocation& GetPublicLocation(void) const
    {
        return m_MainLocation;
    }

/*!
@brief  Gets a reference to the main address of the <tt>StationLocation</tt>.
The main address is intended to be used for public addresses.

@return  Returns a reference to the main address of the <tt>StationLocation</tt>.
*/
    StationLocation& GetPublicLocation(void)
    {
        return m_MainLocation;
    }

/*!
@brief  Gets a <tt>const</tt> reference to the sub-address of the <tt>StationLocation</tt>.
The sub-address is intended to be used for private addresses.

@return  Returns a <tt>const</tt> reference to the sub-address of the <tt>StationLocation</tt>.
*/
    const StationLocation& GetPrivateLocation(void) const
    {
        return m_SubLocation;
    }

/*!
@brief  Gets a reference to the sub-address of the <tt>StationLocation</tt>.
The sub-address is intended to be used for private addresses.

@return  Returns a reference to the sub-address of the <tt>StationLocation</tt>.
*/
    StationLocation& GetPrivateLocation(void)
    {
        return m_SubLocation;
    }

    virtual size_t GetSerializedSize(void) const;
    virtual nn::Result Serialize(bit8* pBuffer, size_t* pSerializedSize, size_t bufferSize) const;
    virtual nn::Result Deserialize(const bit8* pData);

    virtual size_t GetStreamlinedSerializedSize(void) const;
    virtual nn::Result StreamlinedSerialize(bit8* pBuffer, size_t* pSerializedSize, size_t bufferSize) const;
    virtual nn::Result StreamlinedDeserialize(const bit8* pData);

/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

private:
    size_t LittleMembersSerialize(bit8* pBuffer, const StationLocation& src) const;
    size_t LittleMembersDeserialize(StationLocation* pDest, const bit8* pData);

    StationLocation m_MainLocation;
    StationLocation m_SubLocation;
};
}
}
}
