﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_StationLocation.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>
#include <pia/common/common_StationAddress.h>

#if !NN_PIA_NEX_3
namespace nn
{
namespace nex
{
class InetAddress;
class StationURL;
}
}
#endif  // end of !NN_PIA_NEX_3

namespace nn
{
namespace pia
{
namespace inet
{
class NexFacade;
}
}
}

namespace nn
{
namespace pia
{
namespace transport
{

// Forward declaration.
class StationConnectionInfo;

/*!
@brief  Represents an address used to uniquely identify stations.

@date 2012-11-29  Revised code so that the <tt>StationLocation</tt> class does not depend on NEX in NEX 3.X environments.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-04-06  Initial version.
*/
class StationLocation : public nn::pia::common::RootObject
{
    friend class StationConnectionInfo;
    friend class inet::NexFacade;

public:
#if !NN_PIA_NEX_3
/*!
@brief  Converts an NEX <tt>InetAddress</tt> to a Pia <tt>InetAddress</tt>.

@details  This function is thread-safe. This function is deprecated.

@param[in] nexAddress  Specifies the NEX <tt>InetAddress</tt>.
@param[out] pAddress  Specifies the Pia <tt>InetAddress</tt>.
*/
    static void ConvertNexInetAddressToInetAddress(const nn::nex::InetAddress& nexAddress, nn::pia::common::InetAddress* pAddress);

/*!
@brief  Converts a Pia <tt>InetAddress</tt> to an NEX <tt>InetAddress</tt>.

@details  This function is thread-safe. This function is deprecated.

@param[in] address  Specifies the Pia <tt>InetAddress</tt>.
@param[out] pNexAddress  Specifies the NEX <tt>InetAddress</tt>.
*/
    static void ConvertInetAddressToNexInetAddress(const nn::pia::common::InetAddress& address, nn::nex::InetAddress* pNexAddress);

    static void TraceNexAddress(u64 uiTraceFlag, const nn::nex::InetAddress& nexAddress);
#endif  // end of !NN_PIA_NEX_3

/*!
@brief  Instantiates the object.
*/
    StationLocation(void);


/*!
@brief  Instantiates a copy of the object.
*/
    StationLocation(const StationLocation& o);


/*!
@brief  Destroys the object.
*/
    virtual ~StationLocation(void);

    void SetStationLocation(const StationLocation& o);

/*!
@brief  This is the assignment operator.

@return  Returns a reference to the <tt>this</tt> object.
*/
    StationLocation& operator=(const StationLocation& o);

    bool operator==(const StationLocation& o) const;

    bool operator!=(const StationLocation& o) const
    {
        return !(*this == o);
    }

    bool IsSameAddress(const StationLocation& o) const
    {
        return m_Address.GetInetAddress().GetAddress() == o.m_Address.GetInetAddress().GetAddress();
    }

    bool IsSameServerCIdAndUdsNodeId(const StationLocation& o) const
    {
        return (m_ServiceConnectionId == o.m_ServiceConnectionId) &&
               (m_Address.GetExtensionId() == o.m_Address.GetExtensionId());
    }

/*!

@brief  Gets a <tt>const</tt> reference to the <tt>nn::pia::common::StationAddress</tt> instance.

@return  Returns a <tt>const</tt> reference to the <tt>nn::pia::common::StationAddress</tt> instance.
*/
    const nn::pia::common::StationAddress& GetStationAddress(void) const
    {
        return m_Address;
    }

/*!
@brief  Gets a reference to the <tt>nn::pia::common::StationAddress</tt> instance.

@return  Returns a reference to the <tt>nn::pia::common::StationAddress</tt> instance.
*/
    nn::pia::common::StationAddress& GetStationAddress(void)
    {
        return m_Address;
    }

/*!
@brief  Sets the <tt>nn::pia::common::StationAddress</tt> instance.

@param[in] addr  Specifies the Internet address to set.
*/
    void SetStationAddress(const nn::pia::common::StationAddress& addr);

#if !NN_PIA_NEX_3
/*!
@brief  Gets an <tt>nn::nex::StationURL</tt> object.
This function is deprecated.
@param[out] pUrl  Specifies a pointer to the retrieved <tt>StationURL</tt>.
*/
    void GetStationURL(nn::nex::StationURL* pUrl) const;

/*!
@brief  Sets an <tt>nn::nex::StationURL</tt> object.
This function is deprecated.
@param[in] url  Specifies the <tt>StationURL</tt> to set.
*/
    void SetStationURL(const nn::nex::StationURL& url);
#endif  // end of !NN_PIA_NEX_3

    virtual size_t GetSerializedSize(void) const;
    virtual nn::Result Serialize(bit8* pBuffer, size_t* pSerializedSize, size_t bufferSize) const;
    virtual nn::Result Deserialize(const bit8* pData);

    void SetNexServerConnectionId(u32 o)
    {
        m_ServiceConnectionId = o;
    }

    u32 GetNexServerConnectionId() const
    {
        return m_ServiceConnectionId;
    }

    void SetNexUrlType(u8 o)
    {
        m_NexUrlType = o;
    }

    u8 GetNexUrlType() const
    {
        return m_NexUrlType;
    }

    void SetProbeRequestInitiation(u8 o)
    {
        m_IsProbeRequestInitiation = o;
    }

    u8 IsProbeRequestInitiation() const
    {
        return m_IsProbeRequestInitiation;
    }

    void SetNatTypeFlag(u8 flags)
    {
        m_NatTypeFlags = flags;
    }

    u8 GetNatTypeFlag() const
    {
        return m_NatTypeFlags;
    }

    bool IsNatTypeFlagSet(u8 flag) const
    {
        return (m_NatTypeFlags & flag) == flag;
    }

    void GetLocationString(common::String* pString) const;

    void SetPrincipalID(u32 principalId)
    {
        m_NexPrincipalId = principalId;
    }

    u32 GetPrincipalID() const
    {
        return m_NexPrincipalId;
    }

    void SetNatMapping(u8 o)
    {
        m_NatMapping = o;
    }

    u8 GetNatMapping() const
    {
        return m_NatMapping;
    }

    void SetNatFiltering(u8 o)
    {
        m_NatFiltering = o;
    }

    u8 GetNatFiltering() const
    {
        return m_NatFiltering;
    }

    void SetNexConnectionId(u32 o)
    {
        m_NexConnectionId = o;
    }

    u32 GetNexConnectionId() const
    {
        return m_NexConnectionId;
    }

    void SetNexStreamId(u8 o)
    {
        m_NexStreamId = o;
    }

    u8 GetNexStreamId() const
    {
        return m_NexStreamId;
    }

    void SetNexStreamType(u8 o)
    {
        m_NexStreamType = o;
    }

    u8 GetNexStreamType() const
    {
        return m_NexStreamType;
    }

/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

private:
    common::StationAddress m_Address;

    //n1589: Can avoid NEX dependency by making these a different type
    // nn::nex::PrincipalID
    u32 m_NexPrincipalId;
    // nn::nex::ConnectionID
    u32 m_NexConnectionId;
    // nn::nex::ConnectionID
    u32 m_ServiceConnectionId;

    // nn::nex::StationURL::URLType
    u8 m_NexUrlType;
    // nn::nex::Stream::ID
    u8 m_NexStreamId;
    // nn::nex::Stream::Type
    u8 m_NexStreamType;

    // nn::nex::NATProperties::MappingProperties
    u8 m_NatMapping;
    // nn::nex::NATProperties::FilteringProperties
    u8 m_NatFiltering;

    u8 m_NatTypeFlags;

    u8 m_IsProbeRequestInitiation;
};
}
}
}
