﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_StationManager.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_Singleton.h>
#include <pia/transport/transport_Station.h>

#include <pia/common/common_ObjList.h>
#include <pia/common/common_SimpleContainer.h>


// Forward declarations (<tt>PiaCommon</tt>).
namespace nn
{
namespace pia
{
namespace common
{
class InetAddress;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  Represents a class for unifying multiple <tt>Station</tt> instances.
@details  The <tt>Transport</tt> class constructs and destroys the <tt>StationManager</tt> instance.
There is no need for the application to call the <tt>CreateInstance</tt> or <tt>DestroyInstance</tt> functions of the <tt>StationManager</tt> class.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-07-15  Changed how the <tt>GetStationAddress</tt> and <tt>GetStation</tt> functions operate when they are passed invalid arguments.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-04-06  Initial version.



*/
class StationManager : public common::RootObject
{
public:
    typedef common::ObjList<Station> TStationList;
    typedef common::SimpleContainer<Station*, static_cast<int>(MAX_STATION_NUM)> TStationPtrArray;

/*!
@brief  Represents an iterator for referring to the <tt>Station</tt>s participating in a session (non-<tt>const</tt> version).
*/
    class Iterator
    {
    public:
        Iterator(TStationPtrArray::Iterator it)
          : m_Iterator(it)
        {
        }
        Station& operator*(void)
        {
            return **m_Iterator;
        }
        Station* operator->(void)
        {
            return *m_Iterator;
        }
        bool operator==(const Iterator& rhs) const
        {
            return m_Iterator == rhs.m_Iterator;
        }
        bool operator!=(const Iterator& rhs) const
        {
            return !(*this == rhs);
        }
        Iterator& operator++(void)
        {
            ++m_Iterator;
            return *this;
        }

    private:
        TStationPtrArray::Iterator m_Iterator;
    };

/*!
@brief  Represents an iterator for referring to the <tt>Station</tt>s participating in a session (<tt>const</tt> version).
*/
    class ConstIterator
    {
    public:
        ConstIterator(TStationPtrArray::ConstIterator it)
          : m_ConstIterator(it)
        {
        }
        const Station& operator*(void)
        {
            return **m_ConstIterator;
        }
        const Station* operator->(void)
        {
            return *m_ConstIterator;
        }
        bool operator==(const ConstIterator& rhs) const
        {
            return m_ConstIterator == rhs.m_ConstIterator;
        }
        bool operator!=(const ConstIterator& rhs) const
        {
            return !(*this == rhs);
        }
        ConstIterator& operator++(void)
        {
            ++m_ConstIterator;
            return *this;
        }

    private:
        TStationPtrArray::ConstIterator m_ConstIterator;
    };


/*!
@brief  Creates the <tt>StationManager</tt> class instance (singleton pattern).
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully.
@retval ResultNotInitialized  Indicates that the <tt>transport</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the timing at which the <tt>CreateInstance</tt> function was called is incorrect. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@see  DestroyInstance, GetInstance

*/
    static nn::Result CreateInstance(void);


/*!
@brief  Destroys the <tt>StationManager</tt> class instance (singleton pattern).
@see  CreateInstance, GetInstance
*/
    static void DestroyInstance(void);


/*!
@brief  Gets the <tt>StationManager</tt> class instance (singleton pattern).
This function returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>@ref CreateInstance</tt> function.
This function is thread-safe.
@return  Returns a pointer to the instance.
@see  CreateInstance, DestroyInstance


*/
    static StationManager* GetInstance(void)
    {
        return s_pInstance;
    }


/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
*/
    StationManager(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    ~StationManager(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Initializes the <tt>StationManager</tt> instance.
@param[in] pFactory  Specifies a pointer to an instance of a concrete class that inherits <tt>NetworkFactory</tt> (factory pattern).
@param[in] maxStations  Specifies the maximum number of stations to manage.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Indicates that the function failed to allocate memory. Either provide more memory when calling the <tt>common::Initialize</tt> function, or set the <tt>maxStations</tt> argument to a smaller value. Programming error. Fix your program so that this error is not returned.
@see  Finalize

*/
    nn::Result Initialize(NetworkFactory* pFactory, size_t maxStations);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the <tt>StationManager</tt> instance.
@see  Initialize

*/
    void Finalize(void);
    //! @endcond


/*!
@brief  Returns the iterator of the <tt>Station</tt> instance (non-<tt>const</tt> version).
@return  Returns the iterator for the first element of <tt>Station</tt> instances managed by <tt>StationManager</tt>.
@see  End


*/
    Iterator Begin(void)
    {
        return Iterator(m_ValidStationPtr.Begin());
    }


/*!
@brief  Returns the iterator of the <tt>Station</tt> instance (non-<tt>const</tt> version).
@return  Returns the iterator for the last element of <tt>Station</tt> instances managed by <tt>StationManager</tt>.
@see  Begin


*/
    Iterator End(void)
    {
        return Iterator(m_ValidStationPtr.End());
    }


/*!
@brief  Returns the iterator of the <tt>Station</tt> instance ( <tt>const</tt> version).
@return  Returns the iterator for the first element of <tt>Station</tt> instances managed by <tt>StationManager</tt>.
@see  End


*/
    ConstIterator Begin(void) const
    {
        return ConstIterator(m_ValidStationPtr.Begin());
    }


/*!
@brief  Returns the iterator of the <tt>Station</tt> instance ( <tt>const</tt> version).
@return  Returns the iterator for the last element of <tt>Station</tt> instances managed by <tt>StationManager</tt>.
@see  Begin


*/
    ConstIterator End(void) const
    {
        return ConstIterator(m_ValidStationPtr.End());
    }


/*!
@cond PRIVATE
@brief  Creates a <tt>Session</tt> instance for the local station.
@return  On success, returns a pointer to the created <tt>Station</tt> instance.
Returns <tt>NULL</tt> on failure (such as when a local station has already been created).
@see  CreateStation

*/
    Station* CreateLocalStation(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Creates a <tt>Station</tt> instance.
@return  On success, returns a pointer to the created <tt>Station</tt> instance.
This function returns <tt>NULL</tt> if the number of stations has already reached the maximum number of stations specified when <tt>Initialize</tt> was called.
@see  CreateLocalStation, Initialize


*/
    Station* CreateStation(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the specified <tt>Station</tt> instance.
@param[in] pStation  Specifies a pointer to the <tt>Station</tt> instance to destroy. This pointer must have been obtained from the <tt>StationManager</tt> instance.
@see  CreateStation, CreateLocalStation

*/
    void DestroyStation(Station* pStation);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the <tt>Station</tt> instance with the specified <tt>StationIndex</tt>.
@param[in] idx  Specifies the <tt>StationIndex</tt> of the station to destroy.
@see  CreateStation, CreateLocalStation

*/
    void DestroyStation(StationIndex idx);
    //! @endcond


/*!
@brief  Gets the local station (non-<tt>const</tt> version).
@details  This function is thread-safe.

@return  Returns a valid pointer if a local station exists. If none exists, a <tt>NULL</tt> pointer is returned.

*/
    Station* GetLocalStation(void)
    {
        return m_pLocalStation;
    }


/*!
@brief  Gets the local station ( <tt>const</tt> version).
@details  This function is thread-safe.

@return  Returns a valid pointer if a local station exists. If none exists, a <tt>NULL</tt> pointer is returned.

*/
    const Station* GetLocalStation(void) const
    {
        return m_pLocalStation;
    }


/*!
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationId</tt> (non-<tt>const</tt> version).
@param[in] id  Specifies the <tt>StationId</tt>.
@return  If the <tt>Station</tt> corresponding to the specified <tt>StationId</tt> exists, a pointer to that <tt>Station</tt> is returned. A <tt>NULL</tt> pointer is returned if the <tt>StationId</tt> does not exist or is invalid.
@see  GetStationByAddress

*/
    Station* GetStation(StationId id);


/*!
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationId</tt> ( <tt>const</tt> version).
@param[in] id  Specifies the <tt>StationId</tt>.
@return  If the <tt>Station</tt> corresponding to the specified <tt>StationId</tt> exists, a pointer to that <tt>Station</tt> is returned. A <tt>NULL</tt> pointer is returned if the <tt>StationId</tt> does not exist or is invalid.
@see  GetStationByAddress

*/
    const Station* GetStation(StationId id) const;


/*!
@cond PRIVATE
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationIndex</tt> (non-<tt>const</tt> version).
@param[in] idx  The <tt>StationIndex</tt>.
@return  If the <tt>Station</tt> corresponding to the specified <tt>StationIndex</tt> exists, a pointer to that <tt>Station</tt> is returned. A <tt>NULL</tt> pointer is returned if the <tt>StationIndex</tt> does not exist or is invalid.
@see  GetStationByAddress

*/
    Station* GetStation(StationIndex idx);
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationIndex</tt> ( <tt>const</tt> version).
@param[in] idx  The <tt>StationIndex</tt>.
@return  If the <tt>Station</tt> corresponding to the specified <tt>StationIndex</tt> exists, a pointer to that <tt>Station</tt> is returned. A <tt>NULL</tt> pointer is returned if the <tt>StationIndex</tt> does not exist or is invalid.
@see  GetStationByAddress

*/
    const Station* GetStation(StationIndex idx) const;
    //! @endcond


/*!
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationAddress</tt> (non-<tt>const</tt> version).
@param[in] addr  Specifies the station address.
@return  If the <tt>Station</tt> with the specified station address exists, a pointer to that <tt>Station</tt> is returned. If none exists, a <tt>NULL</tt> pointer is returned.
@see  GetStationByAddress

*/
    Station* GetStation(const common::StationAddress& addr);


/*!
@brief  Gets the <tt>Station</tt> corresponding to the <tt>StationAddress</tt> ( <tt>const</tt> version).
@param[in] addr  Specifies the station address.
@return  If the <tt>Station</tt> with the specified station address exists, a pointer to that <tt>Station</tt> is returned. If none exists, a <tt>NULL</tt> pointer is returned.
@see  GetStationByAddress

*/
    const Station* GetStation(const common::StationAddress& addr) const;


/*!
@brief  Gets the station with the specified Internet address (non-<tt>const</tt> version).
@param[in] addr  Specifies the Internet address.
@return  If the <tt>Station</tt> corresponding to the specified Internet address exists, a pointer to that <tt>Station</tt> is returned. If none exists, a <tt>NULL</tt> pointer is returned.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  GetStation

*/
    Station* GetStationByAddress(const common::InetAddress& addr);


/*!
@brief  Gets the <tt>Station</tt> corresponding to the Internet address ( <tt>const</tt> version).
@param[in] addr  Specifies the Internet address.
@return  If the <tt>Station</tt> corresponding to the specified Internet address exists, a pointer to that <tt>Station</tt> is returned. If none exists, a <tt>NULL</tt> pointer is returned.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  GetStation

*/
    const Station* GetStationByAddress(const common::InetAddress& addr) const;


/*!
@brief  Gets the <tt>StationAddress</tt> for the <tt>Station</tt> corresponding to the <tt>StationId</tt>.
@param[out] pAddr  Specifies a pointer to a <tt>StationAddress</tt> instance that will hold the result.
@param[in] id  Specifies the <tt>StationId</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. This <tt>Result</tt> is returned if the <tt>StationId</tt> is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotFound  Indicates that the <tt>Station</tt> specified in the arguments was not found. Handle appropriately in the application.

*/
    nn::Result GetStationAddress(common::StationAddress* pAddr, StationId id) const;


/*!
@cond PRIVATE
@brief  Gets the <tt>StationAddress</tt> for the <tt>Station</tt> corresponding to the <tt>StationIndex</tt>.
@param[out] pAddr  Specifies a pointer to a <tt>StationAddress</tt> instance that will hold the result.
@param[in] idx  The <tt>StationIndex</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. This <tt>Result</tt> is returned if the <tt>StationIndex</tt> is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultNotFound  Indicates that the <tt>Station</tt> specified in the arguments was not found. Handle appropriately in the application.

*/
    nn::Result GetStationAddress(common::StationAddress* pAddr, StationIndex idx) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Returns the number of currently valid <tt>Station</tt> instances.
@return  Returns the number of currently valid <tt>Station</tt> instances.

*/
    size_t GetValidStationNum(void) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Returns a bitmap with bits set corresponding to the <tt>StationIndex</tt> of the stations in the session.
@param[in] bIncludeMyself  Specify <tt>true</tt> if you want to include the bit corresponding to the local station <tt>StationIndex</tt> in the bitmap.

@return  Returns a bitmap value. The least significant bit (LSB) corresponds to <tt>STATION_INDEX_1</tt>.

*/
    u32 GetParticipatingStationBitmap(bool bIncludeMyself) const;
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets called by <tt>session</tt>. Sets the <tt>StationIndex</tt> for the session host.
*/
    void SetHostStationIndex(StationIndex idx)
    {
        m_HostStationIndex = idx;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets the <tt>StationIndex</tt> of the host.
*/
    StationIndex GetHostStationIndex(void) const
    {
        return m_HostStationIndex;
    }
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


private:
    TStationList m_StationList;  // Manage the station instance.
    u8* m_pWorkBuf;              // The buffer for <span class="argument">m_StationList</span>

    TStationPtrArray m_ValidStationPtr;    // Manage the pointer to the currently enabled station.
    TStationPtrArray m_InvalidStationPtr;  // Manage the pointer to the currently disabled station.

    bool isInStationList(const Station* pStation) const;
    bool isInPtrArray(const TStationPtrArray& ptrArray, const Station* pStation) const;

    Station* m_pLocalStation;

    StationIndex m_HostStationIndex;

    static StationManager* s_pInstance;  // The only instantiation of this class
};
}
}
}  // end of namespace nn::pia::transport
