﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_ThreadStreamManager.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

// Forward declarations (common).
namespace nn
{
namespace pia
{
namespace common
{
class IPacketInput;
class IPacketOutput;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{

// Forward declaration.
class TransportThreadStreamOld;
class NetworkFactory;
class SendThreadStream;
class ReceiveThreadStream;

/*!
@brief  Provides a unified interface for managing sending and receiving threads.
@details  This class manages sending and receiving threads, including their creation and destruction.
It also provides functions for setting priority levels for sending or receiving threads and sleep intervals, and functions that have latency emulation features.
An instance of this class is created when the <tt>Transport</tt> class is initialized.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-04-02  Changed the specifications of the <tt>SetSendThreadPacketLossRatio</tt> and <tt>SetReceiveThreadPacketLossRatio</tt> functions.
<tt>ResultInvalidState</tt> is returned when the packet loss emulation feature is not enabled.
@date 2014-04-02  Changed the specifications of the <tt>CreateInstance</tt> function. You can now specify whether to enable or disable packet loss emulation.
@date 2014-04-02  Revised the reference. The reference now refers to <tt>Transport::DebugSetting</tt> for details about the buffer size for latency emulation.
@date 2013-10-18  Fixed errors in the descriptions of <tt>GetSendThreadSendDataSize</tt> and <tt>ClearSendThreadSendDataSize</tt>.
@date 2013-10-18  Fixed errors in the descriptions of <tt>GetSendThreadSendPacketNum</tt> and <tt>ClearSendThreadSendPacketNum</tt>.
@date 2013-10-15  Added a feature to analyze and tabulate the data in sent and received packets.
@date 2013-09-02  Added functions to get and clear the send and receive data size.
@date 2012-08-24  Added a function for packet-loss emulation.
@date 2012-08-08  Added functions to get and clear the number of send and receive packets.
@date 2012-07-11  Added a description of how to set the priority for sending and receiving threads.
@date 2012-07-05  Added the <tt>SetSendThreadLatencyEmulation</tt> and <tt>SetReceiveThreadLatencyEmulation</tt> functions.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-05-24  Initial version.



*/
class ThreadStreamManager : public common::RootObject
{
public:
#if NN_PIA_CTR
    static const s32 SEND_THREAD_DEFAULT_PRIORITY = 10;     //!< Specifies the default priority for sending threads.
    static const s32 RECEIVE_THREAD_DEFAULT_PRIORITY = 10;  //!< Specifies the default priority for receiving threads.
#elif NN_PIA_WIN || NN_PIA_NIN_WIN
    static const s32 SEND_THREAD_DEFAULT_PRIORITY = THREAD_PRIORITY_ABOVE_NORMAL;  //!< Specifies the default priority for sending threads.
    static const s32 RECEIVE_THREAD_DEFAULT_PRIORITY = THREAD_PRIORITY_ABOVE_NORMAL;  //!< Specifies the default priority for receiving threads.
#elif NN_PIA_CAFE
    static const s32 SEND_THREAD_DEFAULT_PRIORITY = 10;  //!< Specifies the default priority for sending threads.
    static const s32 RECEIVE_THREAD_DEFAULT_PRIORITY = 10;  //!< Specifies the default priority for receiving threads.
#else
#error  "Invalid platform"
#endif

    static const s32 SEND_THREAD_DEFAULT_SLEEP_SPAN = 5;     //!< Specifies the default sleep duration (span) of sending threads. The value is in milliseconds.
    static const s32 RECEIVE_THREAD_DEFAULT_SLEEP_SPAN = 5;  //!< Specifies the default sleep duration (span) of receiving threads. The value is in milliseconds.

/*!
@cond PRIVATE
@brief  Creates an instance of the <tt>ThreadStreamManager</tt> class (singleton pattern).
@details  This function is called from within the Pia library. Applications are not required to call this function directly.
@param[in] pFactory  Specifies a pointer to a concrete class that inherits the <tt>NetworkFactory</tt> class.
@param[in] sendThreadMaxPacketNum  Specifies the number of buffers in the send thread.
@param[in] receiveThreadMaxPacketNum  Specifies the number of buffers in the receive thread.
@param[in] sendThreadLatencyEmulationPacketNum  Specifies the number of buffers for emulating sending latency. (Set to zero to disable this feature.)
@param[in] receiveThreadLatencyEmulationPacketNum  Specifies the number of buffers for emulating reception latency. (Set to zero to disable this feature.)
@param[in] bPacketLossEmulation  Specifies whether packet loss emulation is enabled.
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully.
@retval ResultInvalidArgument  Indicates that an argument is invalid. (For example, a <tt>NULL</tt> pointer was specified.) Programming error. Fix your program so that this error is not returned.
@retval ResultNotInitialized  Indicates that the <tt>transport</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the instance was created at the wrong time. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.




*/
    static nn::Result CreateInstance(NetworkFactory* pFactory,
                                     size_t sendThreadMaxPacketNum,
                                     size_t receiveThreadMaxPacketNum,
                                     size_t sendThreadLatencyEmulationPacketNum,
                                     size_t receiveThreadLatencyEmulationPacketNum,
                                     bool bPacketLossEmulation);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the <tt>ThreadStreamManager</tt> class instance (singleton pattern).
@details  This function is called from within the Pia library. Applications are not required to call this function directly.


*/
    static void DestroyInstance(void);
    //! @endcond


/*!
@brief  Gets the <tt>ThreadStreamManager</tt> class instance (singleton pattern).
@details  This function returns a NULL pointer if an instance has not already been created.
@return  Returns a pointer to the <tt>ThreadStreamManager</tt> instance.


*/
    static ThreadStreamManager* GetInstance(void)
    {
        return s_pInstance;
    }


/*!
@cond PRIVATE
@brief  Starts the send and receive threads.
@details  This function is called from within the Pia library. Applications are not required to call this function directly.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@see  Cleanup


*/
    nn::Result Startup(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Cleans up the send and receive threads.
@details  This function is called from within the Pia library. Applications are not required to call this function directly.
@see  Startup


*/
    void Cleanup(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    ~ThreadStreamManager(void);
    //! @endcond


/*!
@brief  Sets the priority of the send thread.
@details  In normal use, set the priority of the sending thread higher than that of the thread in which the application calls the <tt>common::Scheduler::Dispatch</tt> function.
@param[in] prio  Specifies the priority of the send thread. We recommend adjusting these values to suit your application's specifications.
@see  GetSendThreadPriority, common::Scheduler::Dispatch



*/
    void SetSendThreadPriority(s32 prio);


/*!
@brief  Gets the priority of the send thread.
@details  This function is thread-safe.

@return  Returns the priority of the send thread.
@see  SetSendThreadPriority

*/
    s32 GetSendThreadPriority(void) const;


/*!
@brief  Sets the priority of the receive thread.
@details  In normal use, set the priority of the receiving thread higher than that of the thread in which the application calls the <tt>common::Scheduler::Dispatch</tt> function.
@param[in] prio  Specifies the priority of the receive thread. We recommend adjusting these values to suit your application's specifications.
@see  GetReceiveThreadPriority, common::Scheduler::Dispatch



*/
    void SetReceiveThreadPriority(s32 prio);


/*!
@brief  Gets the priority of the receive thread.
@details  This function is thread-safe.

@return  Returns the priority of the receive thread.
@see  SetReceiveThreadPriority

*/
    s32 GetReceiveThreadPriority(void) const;


/*!
@brief  Sets the sleep interval of the send thread.
@details  The send thread sleeps periodically. This function sets the sleep interval.

@param[in] span  Specifies the periodic interval at which to make the send thread sleep. The value is in milliseconds. We recommend adjusting these values to suit your application's specifications.
@see  GetSendThreadSleepTimeSpan

*/
    void SetSendThreadSleepTimeSpan(u32 span);


/*!
@brief  Gets the sleep interval of the send thread.
@details  This function is thread-safe.

@return  Returns the sleep interval of the send thread. The value is in milliseconds.
@see  SetSendThreadSleepTimeSpan

*/
    u32 GetSendThreadSleepTimeSpan(void) const;


/*!
@brief  Sets the sleep interval of the receive thread.
@details  The receive thread sleeps periodically. This function sets the sleep interval.

@param[in] span  Specifies the sleep interval for the receive thread. The value is in milliseconds. We recommend adjusting these values to suit your application's specifications.
@see  GetReceiveThreadSleepTimeSpan

*/
    void SetReceiveThreadSleepTimeSpan(u32 span);


/*!
@brief  Gets the sleep interval of the receive thread.
@details  This function is thread-safe.

@return  Returns the sleep interval of the receive thread. The value is in milliseconds.
@see  SetReceiveThreadSleepTimeSpan

*/
    u32 GetReceiveThreadSleepTimeSpan(void) const;


/*!
@brief  Sets the send latency time to emulate.
@details  The sending latency emulation feature does not immediately send the packets that are created, but instead temporarily saves them to a buffer. After the prescribed amount of time passes, it then sends the saved packets.
For this reason, if you increase the maximum delay time without preparing a buffer of an appropriate size, the buffer overflows and packets appear to be lost frequently.
For more information about setting the buffer size, see the <tt>Transport::DebugSetting</tt> structure.
@param[in] minLatency  Specifies the minimum latency. The value is in milliseconds.
@param[in] maxLatency  Specifies the maximum latency. The value is in milliseconds.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the latency emulation feature may be disabled. Programming error. Fix your program so that this error is not returned.
@see  SetReceiveThreadLatencyEmulation, Transport::DebugSetting






*/
    nn::Result SetSendThreadLatencyEmulation(s32 minLatency, s32 maxLatency);


/*!
@brief  Sets the receive latency time to emulate.
@details  The receiving latency emulation feature does not immediately pass the received packets to higher layers, but instead temporarily saves them to a buffer. After the prescribed amount of time passes, it then passes the saved packets to the higher layers.
For this reason, if you increase the maximum delay time without preparing a buffer of an appropriate size, the buffer overflows and packets appear to be lost frequently.
For more information about setting the buffer size, see the <tt>Transport::DebugSetting</tt> structure.
@param[in] minLatency  Specifies the minimum latency. The value is in milliseconds.
@param[in] maxLatency  Specifies the maximum latency. The value is in milliseconds.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the latency emulation feature may be disabled. Programming error. Fix your program so that this error is not returned.
@see  SetSendThreadLatencyEmulation, Transport::DebugSetting






*/
    nn::Result SetReceiveThreadLatencyEmulation(s32 minLatency, s32 maxLatency);


/*!
@brief  Gets the number of times the send thread attempted to send packets.
@details  The send thread maintains a count of the packets that it attempted to send.
This function returns that count.
@return  Returns the total number of packets that the thread attempted to send.
@see  ClearSendThreadSendPacketNum


*/
    size_t GetSendThreadSendPacketNum(void) const;


/*!
@brief  Clears the sent packet count.
@details  This function clears the counter representing the number of packets that the thread attempted to send. You can use the <tt>GetSendThreadSendPacketNum</tt> function to get the value of this counter.
@see  GetSendThreadSendPacketNum



*/
    void ClearSendThreadSendPacketNum(void);


/*!
@brief  Gets the number of times the receive thread has successfully received packets.
@details  The receive thread maintains a count of the packets that it successfully receives.
This function returns that count.
@return  Returns the total number of packets that have been received successfully.
@see  ClearReceiveThreadReceivePacketNum


*/
    size_t GetReceiveThreadReceivePacketNum(void) const;


/*!
@brief  Clears the received packet count.
@details  This functions clears the counter representing the number of packets successfully received. You can use the <tt>GetReceiveThreadReceivePacketNum</tt> function to get the value of this counter.
@see  GetReceiveThreadReceivePacketNum



*/
    void ClearReceiveThreadReceivePacketNum(void);


/*!
@brief  Gets the total amount of data that the send thread attempted to send.
@details  The send thread maintains a count of the amount of data it attempted to send.
This function returns that count.
@return  Returns the total amount of data that the thread attempted to send. The value is in bytes.
@see  ClearSendThreadSendDataSize


*/
    u64 GetSendThreadSendDataSize(void) const;


/*!
@brief  Clears the counter tracking the amount of data sent.
@details  This function clears the counter representing the amount of data that the thread attempted to send. You can use the <tt>GetSendThreadSendDataSize</tt> function to get the value of this counter.
@see  GetSendThreadSendDataSize



*/
    void ClearSendThreadSendDataSize(void);


/*!
@brief  Gets the total amount of data successfully received in the receive thread.
@details  The receive thread maintains a count of the amount of data it successfully receives.
This function returns that count.
@return  Returns the total amount of data that has been received successfully. The value is in bytes.
@see  ClearReceiveThreadReceiveDataSize


*/
    u64 GetReceiveThreadReceiveDataSize(void) const;


/*!
@brief  Clears the counter tracking the amount of data received.
@details  This functions clears the counter representing the amount of data successfully received. You can use the <tt>GetReceiveThreadSendDataSize</tt> function to get the value of this counter.
@see  GetReceiveThreadReceiveDataSize



*/
    void ClearReceiveThreadReceiveDataSize(void);


/*!
@brief  Sets the packet loss rate to emulate on the send thread.
@details  You can use this packet loss emulation feature to emulate cases in which packets were sent successfully, but were later lost for some reason.
Setting a value other than <tt>0%</tt> in the argument emulates that rate of packet loss by having the sending thread drop the specified percentage of packets.
By default, the packet loss rate on the send thread is set to 0%.
@param[in] lossPercentage  Specifies a value in the range from 0 through 100.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  The packet loss emulation feature is not set to enabled. Programming error. Fix your program so that this error is not returned.
@see  GetSendThreadPacketLossRatio, Transport::SetDebugSetting





*/
    nn::Result SetSendThreadPacketLossRatio(s32 lossPercentage);


/*!
@brief  Gets the packet loss rate emulated on the send thread.
@return  Returns the packet loss rate set for the send thread. The value is in the range from 0 through 100.
@see  SetSendThreadPacketLossRatio

*/
    s32 GetSendThreadPacketLossRatio(void) const;


/*!
@brief  Sets the packet loss rate to emulate on the receive thread.
@details  You can use this packet loss emulation feature to emulate cases in which incoming packets were lost for some reason.
Setting a value other than <tt>0%</tt> in the argument will emulate that rate of packet loss by having the receiving thread drop the specified percentage of packets.
By default, the packet loss rate on the receive thread is set to 0%.
@param[in] lossPercentage  Specifies a value in the range from 0 through 100.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  The packet loss emulation feature is not set to enabled. Programming error. Fix your program so that this error is not returned.
@see  GetReceiveThreadPacketLossRatio, Transport::SetDebugSetting




*/
    nn::Result SetReceiveThreadPacketLossRatio(s32 lossPercentage);


/*!
@brief  Gets the packet loss rate to emulate on the receive thread.
@return  Returns the packet loss rate set for the receive thread. The value is in the range from 0 through 100.
@see  SetReceiveThreadPacketLossRatio

*/
    s32 GetReceiveThreadPacketLossRatio(void) const;


/*!
@cond PRIVATE
@brief  Gets a pointer to a <tt>ReceiveThreadStreamOld</tt> instance (non-<tt>const</tt> version).
*/
    ReceiveThreadStream* GetReceiveStream(void)
    {
        return m_pReceiveStream;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to a <tt>ReceiveThreadStreamOld</tt> instance (<tt>const</tt> version).
*/
    const ReceiveThreadStream* GetReceiveStream(void) const
    {
        return m_pReceiveStream;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to a <tt>SendThreadStreamOld</tt> instance (non-<tt>const</tt> version).
*/
    SendThreadStream* GetSendStream(void)
    {
        return m_pSendStream;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Gets a pointer to a <tt>SendThreadStreamOld</tt> instance (<tt>const</tt> version).
*/
    const SendThreadStream* GetSendStream(void) const
    {
        return m_pSendStream;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Provided for <tt>NexFacade</tt>.
*/
    common::IPacketInput* GetInputStream(void)
    {
        return m_pInputStream;
    }
    //! @endcond

/*!
@cond PRIVATE
@brief  Provided for <tt>NexFacade</tt>.
*/
    common::IPacketOutput* GetOutputStream(void)
    {
        return m_pOutputStream;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets monitoring data.
*/
    void SetMonitoringData(void);
    //! @endcond


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    virtual void Trace(u64 flag) const;


#if NN_PIA_ENABLE_STATISTIC_THREAD_STREAM_BLOCK_TIME
    void ResetThreadStreamStatistic();
    void PrintThreadStreamStatistic(u32 printMaxDataNum);
#endif

private:
/*!
@brief  No default constructor is used. A factory pointer is assumed.
*/
    ThreadStreamManager(void);

/*!
@brief  Instantiates the object with the specified arguments.
*/
    ThreadStreamManager(NetworkFactory* pFactory,
                        size_t sendThreadMaxPacketNum,
                        size_t receiveThreadMaxPacketNum,
                        size_t sendThreadLatencyEmulationPacketNum,
                        size_t receiveThreadLatencyEmulationPacketNum,
                        bool bPacketLossEmulation);

/*!
@brief  This copy constructor is private because the singleton pattern is used.
*/
    ThreadStreamManager(const ThreadStreamManager&);

/*!
@brief  This assignment operator is private because the singleton pattern is used.
*/
    ThreadStreamManager& operator=(const ThreadStreamManager&);


    // Specify variables that receive the pointers created by the factory.
    common::IPacketInput* m_pInputStream;
    common::IPacketOutput* m_pOutputStream;

    ReceiveThreadStream* m_pReceiveStream;
    SendThreadStream* m_pSendStream;

    static ThreadStreamManager* s_pInstance;
};
}
}
}  // end of namespace nn::pia::transport
