﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_Transport.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_Singleton.h>
#include <pia/transport/transport_ProtocolManager.h>

#include <pia/common/common_Time.h>
#include <pia/common/common_Job.h>

#if NN_PIA_ENABLE_FEEDBACK
#include <pia/common/common_Feedback.h>
#endif


// Forward declarations (common).
namespace nn
{
namespace pia
{
namespace common
{
class IPacketInput;
class IPacketOutput;
class StationAddress;
struct CryptoSetting;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{


// Forward declarations (transport).
class NetworkFactory;
class StationConnectionInfo;
class KeepAliveSender;
class KeepAliveReceiver;
class ThreadStreamManager;
class IRelayAddressTable;
class RelayRouteManager;
class StationPacketHandler;
class Station;
class StationIdTable;

#if NN_PIA_ENABLE_FEEDBACK
#define NN_PIA_SEND_FEEDBACKDATA                                      \
    if (nn::pia::transport::Transport::GetInstance())                 \
    {                                                                 \
        NN_PIA_FEEDBACK();                                            \
        nn::pia::transport::Transport::GetInstance()->SendFeedback(); \
    }                                                                 \
    nn::pia::common::Feedback::Dummy
#endif

/*!
@brief  Provides core functionality of the <tt>transport</tt> module.

@date 2015-04-24  Fixed an error in the description of the arguments passed to <tt>ConvertToStationId</tt>.
@date 2015-04-24  Noted that the <tt>ConvertToStationIndex</tt> and <tt>ConvertToStationId</tt> functions have different behavior when used with the Session API and the Mesh API.
@date 2014-11-05  Renamed the <tt>GetConnectionStatus</tt> function to <tt>CheckConnectionError</tt>.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-07-17  Added the new <tt>CheckTransportStatus</tt> function.
@date 2014-06-26  Added the new <tt>ConvertToStationId</tt> and <tt>ConvertToStationIndex</tt> functions.
@date 2014-04-08  Provided the <tt>DebugSetting</tt> structure as the default constructor, and made it initialize all member variables with the default values.
@date 2014-04-03  Renamed the <tt>GetMasterStationConnectionInfo</tt> function to <tt>GetHostStationConnectionInfo</tt>.
@date 2014-04-02  Added the new <tt>SetDebugSetting</tt> function. Use this function to make debugging settings for things like latency emulation.
@date 2014-04-02  Removed the members related to latency emulation from the <tt>Transport::Setting</tt> structure.
@date 2013-11-22  Added the <tt>analysisInterval</tt> member variable to the <tt>Transport::Setting</tt> structure and made it possible to automatically print measurement results from the transport analysis feature.
@date 2013-05-24  Made the <tt>SetKeepAliveInterval</tt> function private. Specify with the <tt>Session::Startup</tt> function.
@date 2013-04-01  Removed the <tt>maxPacketCreatorBufferNum</tt> and <tt>maxPacketParserBufferNum</tt> members from the <tt>Transport::Setting</tt> struct.
@date 2012-12-03  The <tt>Transport::GetConnectionStatus</tt> function can now return <tt>ResultInvalidState</tt>.
@date 2012-11-30  Removed the <tt>Transport::MAX_SIGNATURE_KEY_SIZE</tt> constant.
@date 2012-11-30  Removed signature-related parameters from the version of the <tt>Transport::Setting</tt> struct constructor with parameters.
@date 2012-11-30  Removed signature-related member variables from the <tt>Transport::Setting</tt> struct.
@date 2012-11-27  Changed the specifications of the <tt>Transport::CreateInstance</tt> function. The function now returns an error <tt>Result</tt> value if the signature feature is not enabled on networks where signed packets are recommended.
@date 2012-07-25  Added <tt>Transport::GetLocalStationConnectionInfo</tt> and <tt>Transport::GetMasterStationConnectionInfo</tt>.
@date 2012-07-18  Added error handling for the <tt>Transport::GetConnectionStatus</tt> function.
@date 2012-07-05  Prepared a constructor for the <tt>Transport::Setting</tt> structure.
@date 2012-07-05  Added member variables for latency emulation to the <tt>Setting</tt> structure.
@date 2012-07-04  Updated descriptions of <tt>Setting</tt> structure members.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-06-15  Added the <tt>Transport::GetConnectionStatus</tt> function.
@date 2012-06-15  Made the following functions private: <tt>Transport::Startup</tt>, <tt>Transport::Cleanup</tt>, and <tt>Transport::Dispatch</tt>.
@date 2012-06-01  Added the <tt>Transport::Startup</tt> and <tt>Transport::Cleanup</tt> functions.
@date 2012-05-25  Changed the <tt>Result</tt> values that the <tt>Transport::Dispatch</tt> function can return.
@date 2012-05-09  Extended the <tt>Transport::Setting</tt> struct to add the ability to sign packets.
@date 2012-04-28  Noted that the <tt>Transport::Dispatch</tt> function can return <tt>ResultInvalidState</tt>.
@date 2012-04-28  Revised the description of the <tt>Transport::Dispatch</tt> function.
@date 2012-04-25  Described memory settings using the <tt>Transport::Setting</tt> struct.
@date 2012-04-21  Revised the description of the <tt>Transport::Dispatch</tt> function.
@date 2012-04-18  The size of internal buffers can now be specified in the <tt>Transport::CreateInstance</tt> function.
@date 2012-04-12  Added functions related to keep-alive processing.
@date 2012-04-06  Initial version.
*/
class Transport : public common::RootObject
{
public:
/*!
@brief  Stores parameters passed to the <tt>CreateInstance</tt> function.

@details  The buffer parameter you configure in this struct has a direct effect on the memory usage of the <tt>PiaTransport</tt> class.
You can set this parameter to a large value to reduce the frequency of <tt>ResultBufferIsFull</tt> errors caused by buffer exhaustion when sending and receiving data. This helps make communication function more smoothly, although of course it increases memory usage.
The amount of memory required by the <tt>PiaTransport</tt> class depends on the communication specifications of your application.
The library provides the <tt>common::WatermarkManager()</tt> class to help tune these parameters.
You can use this watermark feature to discover the peak buffer usage when your program actually runs. It helps application developers determine the optimum buffer size to use in the Pia library.
@see  CreateInstance, common::WatermarkManager, ThreadStreamManager





*/
    struct Setting
    {

/*!
@brief  Instantiates the object with default parameters (default constructor). Each member variable is set to <tt>0</tt> or <tt>NULL</tt>.
*/
        Setting(void);

/*!
@brief  Instantiates the object with arguments. The values specified in the arguments are set in the variables of each member.

@param[in] pFactory  Specifies a pointer to an instance of a concrete class that inherits <tt>NetworkFactory</tt> (factory pattern).
@param[in] maxStationNum  Specifies the maximum number of <tt>Station</tt>s that can join a single network.
@param[in] maxSendThreadBufferNum  Specifies the number of buffers that the Pia library's sending thread uses to send packets (about 1.5 KB per buffer). You must set this value to 2 or greater.
@param[in] maxReceiveThreadBufferNum  Specifies the number of buffers that the Pia library's receiving thread uses to receive packets (about 1.5 KB per buffer). You must set this value to 2 or greater.
@param[in] analysisInterval  The interval at which measurement results from the <tt>TransportAnalyzer</tt> feature are automatically output to the console. The unit is seconds. When <tt>0</tt> is specified, output is not automatic, and it is left to the application to get the data.
*/
        Setting(
            NetworkFactory* pFactory,
            size_t maxStationNum = 0,
            size_t maxSendThreadBufferNum = 0,
            size_t maxReceiveThreadBufferNum = 0,
            s32 analysisInterval = 0);

        NetworkFactory* pFactory;          //!< Pointer to an instance of the concrete class that inherits <tt>NetworkFactory</tt> (factory pattern).
        size_t maxStationNum;              //!< Specifies the maximum number of <tt>Station</tt>s that can join a single network.
        size_t maxSendThreadBufferNum;     //!< The number of buffers that the sending thread of the Pia library uses to send packets (about 1.5 KB per buffer). You must set this value to 2 or greater.
        size_t maxReceiveThreadBufferNum;  //!< The number of buffers that the receiving thread of the Pia library uses to receive packets (about 1.5 KB per buffer). You must set this value to 2 or greater.
        s32 analysisInterval;              //!< The interval at which measurement results from the <tt>TransportAnalyzer</tt> feature are automatically output to the console. The unit is seconds. When <tt>0</tt> is specified, output is not automatic, and it is left to the application to get the data.
    };


/*!
@cond PRIVATE
@brief  Instantiates an object. Do not have the application call this constructor directly. Use the <tt>CreateInstance</tt> function instead.
*/
    Transport(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.
*/
    ~Transport(void);
    //! @endcond


/*!
@brief  Creates a <tt>Transport</tt> class instance (singleton pattern).

@param[in] setting  Specifies the <tt>Setting</tt> struct that parameters are stored in.

@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Specifies that the <tt>transport</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Specifies that the timing at which the <tt>CreateInstance</tt> function was called is incorrect. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Specifies that an argument is invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Specifies that the function failed to allocate memory. You must either increase the memory provided to Pia with the <tt>common::Initialize</tt> function or try setting parameters that do not consume as much memory. Programming error. Fix your program so that this error is not returned.
@see  DestroyInstance, GetInstance, Setting
*/
    static nn::Result CreateInstance(const Transport::Setting& setting);


/*!
@brief  Destroys the <tt>Transport</tt> class instance (singleton pattern).
If no instances exist, this function does nothing.

@see  CreateInstance, GetInstance
*/
    static void DestroyInstance(void);


/*!
@brief  Gets the <tt>Transport</tt> class instance (singleton pattern).
This function returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>CreateInstance()</tt> function.
This function is thread-safe.

@return  Returns a pointer to a <tt>Transport</tt> instance.
@see  CreateInstance, DestroyInstance

*/
    static Transport* GetInstance(void)
    {
        return s_pInstance;
    }


/*!
@cond PRIVATE
@brief  Executes the startup procedure.

@details  Call this function after the <tt>transport::EndSetup</tt> function has been called and the initialization process of the <tt>transport</tt> module has ended.
@param[in] cpRelayNodeAddress  Specifies a pointer to the address of the relay node.
It is okay to use a pointer to a temporary instance.
Specify <tt>NULL</tt> if you are not using relay nodes.
@param[in] cpCryptoSetting  The pointer to the encryption settings.
Communication is not encrypted if <tt>NULL</tt> is specified.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  It could be that either the <tt>Startup</tt> function has already been called or that it is being called between calls to the <tt>transport::BeginSetup</tt> and <tt>transport::EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@see  Cleanup, transport::EndSetup



*/
    nn::Result Startup(const common::StationAddress* cpRelayNodeAddress = NULL, const common::CryptoSetting* cpCryptoSetting = NULL);
    //! @endcond


/*!
@cond PRIVATE
@brief  Executes the cleanup procedure.

@details  Call this function before finalizing the <tt>transport</tt> module.

@see  Startup
*/
    void Cleanup(void);
    //! @endcond


/*!
@brief  Generates an instance of a protocol of type <tt>T</tt>.

@details  @tparam T  Specifies the protocol type.
@param[in] port  Specifies the value of the protocol port.

@return  Returns a handle to the generated protocol. If the same protocol ID already exists, <tt>0</tt> is returned.
@see  DestroyProtocol, GetProtocol
*/
    template <typename T>
    u32 CreateProtocol(u16 port = 0)
    {
        return m_ProtocolManager.CreateProtocol<T>(port);
    }


/*!
@brief  Destroys the protocol instance.

@param[in] protocolHandle  Specifies the handle for the protocol to delete.
@see  CreateProtocol, GetProtocol
*/
    void DestroyProtocol(u32 protocolHandle)
    {
        m_ProtocolManager.DestroyProtocol(protocolHandle);
    }


/*!
@brief  Gets an instance of a protocol of type <tt>T</tt>.

@details  @tparam T  Specifies the protocol type.
@param[in] protocolHandle  Specifies the handle for the protocol to get.

@return  Returns the address of the protocol instance. Returns <tt>NULL</tt> if the protocol for the specified handle does not exist or is not of type <tt>T</tt>.
@see  CreateProtocol, DestroyProtocol
*/
    template <typename T>
    T* GetProtocol(u32 protocolHandle)
    {
        return m_ProtocolManager.GetProtocol<T>(protocolHandle);
    }


/*!
@brief  Gets the connection status of the <tt>transport</tt> module.

@details  This function is thread-safe.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> value returns <tt>true</tt> if there are no problems with the network connection.
@retval ResultInvalidState  Specifies that the application may not have called an appropriate initialization or setup function.
This <tt>Result</tt> is returned, for example, if you attempt to connect to the Internet when the socket library has not been initialized.
Programming error. Fix your program so that this error is not returned.
@cond CTR_DOC
@retval ResultNetworkConnectionIsLost  Indicates that the network connection was lost. The wireless switch may be turned off or there may be a problem with the access point.
If this error is returned in a local connection environment, reinitialize <tt>PiaLocal</tt> and the higher-level modules.
If this error is returned in an Internet connection environment, connect to the Internet again (<tt>nn::ac::Connect</tt>). You could also reinitialize, as with a local connection environment.
Handle appropriately in the application.
@endcond
@cond CAFE_DOC
@retval ResultNetworkConnectionIsLost  Indicates that the network connection was lost.
One possible cause is a problem with the access point.
If this error is returned in an Internet connection environment, connect to the Internet again (<tt>nn::ac::Connect</tt>).
Reconnect to AC.
@endcond

@see  util::NetworkErrorCodeConverter::Convert



*/
    nn::Result CheckConnectionError(void) const
    {
        return m_ConnectionError;
    }


/*!
@cond PRIVATE
@brief  Gets the last time that the <tt>dispatch</tt> function was called.

@return  Returns the last time that the <tt>dispatch</tt> function was called.
@see  dispatch
*/
    const common::Time& GetDispatchedTime(void) const
    {
        return m_DispatchedTime;
    }
    //! @endcond


/*!
@brief  Gets the maximum number of connections (<tt>Station</tt> instances) set when <tt>CreateInstance</tt> was called.

@details  This function is thread-safe.

@return  Returns the maximum number of stations that can be connected to.
@see  CreateInstance
*/
    size_t GetMaxStationNum(void) const
    {
        return m_MaxStationNum;
    }


/*!
@cond PRIVATE
@brief  Gets the protocol manager.

@return  Returns the address of the protocol manager.
*/
    ProtocolManager* GetProtocolManager()
    {
        return &m_ProtocolManager;
    }
    //! @endcond


/*!
@cond PRIVATE
@brief  Sets the interval at which to send keep-alive packets.

@param[in] msec  Specifies the interval at which to send keep-alive packets. The value is in milliseconds.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidArgument  Specifies that there is an error in one or more arguments (for example, an argument is negative). Programming error. Fix your program so that this error is not returned.
@see  GetKeepAliveInterval
*/
    nn::Result SetKeepAliveInterval(s32 msec);
    //! @endcond


/*!
@brief  Gets the interval at which to send keep-alive packets.

@details  This function is thread-safe.

@return  Returns the sending interval for keep-alive packets. The value is in milliseconds.
@see  SetKeepAliveInterval
*/
    s32 GetKeepAliveInterval(void) const;


/*!
@brief  Enables the keep-alive process.

@see  DisableKeepAlive, IsKeepAliveEnabled
*/
    void EnableKeepAlive(void);


/*!
@brief  Disables the keep-alive process.

@see  EnableKeepAlive, IsKeepAliveEnabled
*/
    void DisableKeepAlive(void);


/*!
@brief  Gets whether the keep-alive process is enabled.

@details  This function is thread-safe.

@return  Returns <tt>true</tt> if it is enabled, and <tt>false</tt> if it is disabled.
@see  EnableKeepAlive, DisableKeepAlive
*/
    bool IsKeepAliveEnabled(void) const;


/*!
@brief  Gets the address of the local device (<tt>@ref StationConnectionInfo</tt>).

@details  The address that is obtained when you are not connected to the network is undefined.

@param[out] pInfoOut  Specifies a pointer to the object that holds the local address.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  GetHostStationConnectionInfo
*/
    nn::Result GetLocalStationConnectionInfo(StationConnectionInfo* pInfoOut);


/*!
@brief  Gets the address of the session host (<tt>@ref StationConnectionInfo</tt>).

@details  The address that is obtained when you are not connected to the network is undefined.

@param[out] pInfoOut  Specifies a pointer to the object that holds the address obtained for the session host.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  GetLocalStationConnectionInfo
*/
    nn::Result GetHostStationConnectionInfo(StationConnectionInfo* pInfoOut);


/*!
@cond PRIVATE
@brief  Gets a pointer to the instance of the relay address table.
*/
    IRelayAddressTable* GetRelayAddressTable(void)
    {
        return m_pRelayAddressTable;
    }
    //! @endcond


    /// @cond PRIVATE
/*!
@brief  Gets a pointer to the instance of the relay request route manager class.
*/
    RelayRouteManager* GetRelayRouteManagerPtr(void)
    {
        return m_pRelayRouteManager;
    }

/*!
@brief  Sets a pointer to the instance of the relay request route manager class.
*/
    void SetRelayRouteManagerPtr(RelayRouteManager* ptr)
    {
        m_pRelayRouteManager = ptr;
    }
//! @endcond


#if NN_PIA_ENABLE_FEEDBACK
    /// @cond PRIVATE
/*!
@brief  Sends feedback data.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. The <tt>Transport</tt> instance may not have been completely initialized or set up. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SendFeedback(void);

/*!
@brief  Sets the RTT value in the feedback data.
*/
    void SetFeedbackNetworkRtt();
/// @endcond
#endif  // end of NN_PIA_ENABLE_FEEDBACK

    /// @cond PRIVATE
    void SetMonitoringNetworkRtt(bool isBegin);
    //! @endcond

    /// @cond PRIVATE
    StationPacketHandler* GetStationPacketHandler(void)
    {
        return m_pPacketHandler;
    }
    //! @endcond


/*!
@brief  Converts a <tt>StationId</tt> to a <tt>@ref StationIndex</tt>.

@details  The function returns a failure <tt>Result</tt> if the local station is not participating in the session or the function is called during a joint session.
@details  This function can be called inside the session state-change event callback to which the EVENT_JOIN and EVENT_LEAVE events are reported.
@details  This function works differently depending on whether it is used with the Session API or the Mesh API.
When used with the Mesh API, the conversion succeeds even if a <tt>StationId</tt> that is not a mesh participant is specified.

@param[out] pIdx  A pointer to where the converted <tt>@ref StationIndex</tt> is stored.
@param[in] id  The <tt>StationId</tt> to convert.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Not in a state where the conversion process can be run. The setup for <tt>session</tt> might not have been completed. Programming error. Fix your program so that this error is not returned.
@retval ResultNotFound  Indicates that the station specified in the <tt>@ref nn::pia::StationId "StationId"</tt> is not participating in the session. Handle appropriately in the application.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@see  ConvertToStationId, session::Session::SessionEventCallback


*/
    nn::Result ConvertToStationIndex(StationIndex* pIdx, StationId id) const;


/*!
@cond PRIVATE
@brief  Converts <tt>StationId</tt> to <tt>@ref StationIndex</tt>. No checks of arguments or states are performed.

@details  Because this function omits various checks, if it is called haphazardly, you run the risk of crashing the program with a memory access violation or some other issue.

@param[in] id  The <tt>StationId</tt> to convert.
@return  <tt>STATION_INDEX_INVALID</tt> is returned when conversion fails.
*/
    StationIndex ConvertToStationIndexUnsafe(StationId id) const;
    //! @endcond


/*!
@brief  Converts a <tt>@ref nn::pia::StationId</tt> to a <tt>StationId</tt>.

@details  The function returns a failure <tt>Result</tt> if the local station is not participating in the session or the function is called during a joint session.
@details  This function can be called inside the session state-change event callback to which the EVENT_JOIN and EVENT_LEAVE events are reported.
@details  This function works differently depending on whether it is used with the Session API or the Mesh API.
When used with the Mesh API, the conversion succeeds even if a <tt>StationIndex</tt> that is not a mesh participant is specified.

@param[out] pId  Pointer to where the converted <tt>StationId</tt> is stored.
@param[in] idx  The <tt>@ref StationIndex</tt> to convert.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Not in a state where the conversion process can be run. The setup for <tt>session</tt> might not have been completed. Programming error. Fix your program so that this error is not returned.
@retval ResultNotFound  Indicates that the station specified in the <tt>@ref nn::pia::StationIndex "StationIndex"</tt> is not participating in the session. Handle appropriately in the application.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@see  ConvertToStationIndex, session::Session::SessionEventCallback


*/
    nn::Result ConvertToStationId(StationId* pId, StationIndex idx) const;


/*!
@cond PRIVATE
@brief  Converts <tt>@ref StationIndex</tt> to <tt>StationId</tt>. No checks of arguments or states are performed.

@details  Because this function omits various checks, if it is called haphazardly, you run the risk of crashing the program with a memory access violation or some other issue.

@param[in] idx  The <tt>StationIndex</tt> to convert.
@return  <tt>STATION_ID_INVALID</tt> is returned when conversion fails.
*/
    StationId ConvertToStationIdUnsafe(StationIndex idx) const;
    //! @endcond


/*!
@brief  Checks the state of <tt>Transport</tt>, and determines whether sending and receiving processes can be executed in that state.

@return  Returns a <tt>Result</tt> value for which <tt>IsSuccess()</tt> returns <tt>true</tt> if the <tt>session</tt> module is finished being set up and the various API functions can be called without issue.
@retval ResultInvalidState  The setup of the <tt>session</tt> module is not complete. Programming error. Fix your program so that this error is not returned.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@see  ConvertToStationIndex, ConvertToStationId

*/
    nn::Result CheckTransportStatus() const
    {
        PIA_ASSERT((m_TransportStatus.IsSuccess())
                || (ResultInvalidState::Includes(m_TransportStatus))
                || (ResultTemporaryUnavailable::Includes(m_TransportStatus)));
        return m_TransportStatus;
    }


/*!
@cond PRIVATE
@brief  Sets the <tt>Transport</tt> state. Call this function from <tt>session</tt>.

@param[in] state  Sets the <tt>Transport</tt> state with a <tt>Result</tt> value. Three kinds of <tt>Result</tt> values can be set: <tt>ResultInvalidState</tt>, <tt>ResultTemporaryUnavailable</tt> and <tt>ResultSuccess()</tt>.
Pass <tt>ResultSuccess</tt> to the argument when the <tt>session</tt> preparations are complete, <tt>ResultTemporaryUnavailable</tt> when a joint session process starts, and <tt>ResultInvalidState</tt> when destroying the <tt>session</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the call succeeds.
@retval ResultInvalidArgument  Specifies that an argument is invalid. These three <tt>Result</tt> values are the only values that can be specified for <tt>state</tt>. Programming error. Fix your program so that this error is not returned.



*/
    nn::Result SetTransportStatus(nn::Result state);
    //! @endcond


    // Controls whether to convert the <tt>StationId</tt> by using the <tt>StationIdTable</tt> class, or by converting directly.
    void EnableStationIdTable()
    {
        m_bEnableStationIdTable = true;
    }
    void DisableStationIdTable()
    {
        m_bEnableStationIdTable = false;
    }
    bool IsEnableStationIdTable() const
    {
        return m_bEnableStationIdTable;
    }

    // <tt>session</tt> is supposed to get an instance of the conversion table class with this API function, and then execute updates and similar processes.
    //
    const StationIdTable* GetStationIdTable() const
    {
        return m_pStationIdTable;
    }
    StationIdTable* GetStationIdTable()
    {
        return m_pStationIdTable;
    }

/*!
@brief  Structure that collects the parameters to pass to the <tt>SetDebugSetting</tt> function.
*/
    struct DebugSetting
    {
        DebugSetting()
          : sendThreadLatencyEmulationPacketNum(0),
            receiveThreadLatencyEmulationPacketNum(0),
            bPacketLossEmulation(false)
        {
        }  //!< Default constructor. Initializes all of the member variables.

        size_t sendThreadLatencyEmulationPacketNum;     //!< The number of buffers used by the receive-latency emulation feature (about 1.5 KB per buffer).
        size_t receiveThreadLatencyEmulationPacketNum;  //!< The number of buffers used by the receive-latency emulation feature (about 1.5 KB per buffer).
        bool bPacketLossEmulation;                      //!< Set to <tt>true</tt> when using the packet loss emulation feature.
    };


/*!
@brief  Enables debugging features.

@details  Enables debugging features based on the content set in the arguments.
Never call this API function in a product ROM.
@details  When you use this API function, you must call it before calling the <tt>Transport::CreateInstance</tt> function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Returned if this function is called at the wrong time. Programming error. Fix your program so that this error is not returned.

*/
    static nn::Result SetDebugSetting(const struct DebugSetting& setting);


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;

    //n1589: Needed for <tt>NexFacade</tt>.
    common::IPacketInput* GetInputStream(void);
    common::IPacketOutput* GetOutputStream(void);

    void OutputStreamUpdateEvent(void);

    // Disconnection event.
    enum DisconnectionEventType
    {
        EVENT_LEAVE,  // Normal departure from a session.
        EVENT_REJOIN  // Leave temporarily to join another session.
    };

    // Event listener.
    typedef DisconnectionEventType (*DisconnectionEventListener)(void* pStationIdTableEntry);

    void RegisterEventListener(DisconnectionEventListener func)
    {
        m_EventListener = func;
    }
    DisconnectionEventListener GetEventListener()
    {
        return m_EventListener;
    }

private:
/*!
@cond PRIVATE
@brief  This is a dispatch function.

@details  Use this function to pass data between the <tt>PiaTransport</tt> sending and receiving threads, or execute the send or receive processes for a particular protocol.
This function never blocks waiting for data to be sent or received, because it does not call send or receive functions in the wireless module. (The calls to sending and receiving functions in the wireless module are made in the <tt>PiaTransport</tt> sending and receiving threads.)
The assumption is that this function will be called at least once for each game frame.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Specifies that protocol initialization or some other process may not be finished. Programming error. Fix your program so that this error is not returned.




*/
    nn::Result dispatch(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Initializes the <tt>Transport</tt> instance.

@details  Initializes the factory to use and configures the maximum station count.

@param[in] setting  Specifies the <tt>Setting</tt> struct that parameters are stored in.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Specifies that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@see  finalize
*/
    nn::Result initialize(const Transport::Setting& setting);
    //! @endcond


/*!
@cond PRIVATE
@brief  Finalizes the instance.
@see  initialize
*/
    void finalize(void);
    //! @endcond


/*!
@brief  This copy constructor is private.
*/
    Transport(const Transport&);

/*!
@brief  This assignment operator is private.
*/
    Transport& operator=(const Transport&);


/*!
@brief  Updates the time immediately following a call to the <tt>Dispatch</tt> function.

@details  To reduce CPU load, only call this once per call to the <tt>Dispatch</tt> function.
This time is used for keep-alive and other processing.
Updating the time only once per dispatch is typically enough because a high level of accuracy is not required.
@see  GetDispatchedTime
*/
    void updateDispatchedTime(void)
    {
        m_DispatchedTime = common::Time::GetTime();
    }


    void updateTransportAnalyzer(void);


    // Job to call <tt>dispatch</tt>.
    class DispatchJob : public common::Job
    {
    protected:
        virtual ExecuteResult ExecuteCore()
        {
            Transport::GetInstance()->dispatch();
            return Job::EXECUTE_STATE_WAIT_NEXT;
        }
    };

    common::Time m_DispatchedTime;

    ProtocolManager m_ProtocolManager;
    StationPacketHandler* m_pPacketHandler;

    size_t m_MaxStationNum;

    bool m_StartupFlag;

    nn::Result m_ConnectionError;
    DispatchJob m_DispatchJob;

    IRelayAddressTable* m_pRelayAddressTable;
    RelayRouteManager* m_pRelayRouteManager;

#if NN_PIA_ENABLE_FEEDBACK
    u32 m_FeedbackSenderProtocolHandle;
#endif

    s32 m_AnalysisInterval;              // The interval, in seconds, for automatic calls to the <tt>TransportAnalyzer::Update</tt> function. In seconds.
    common::Time m_AnalysisUpdatedTime;  // The time when the <tt>TransportAnalyzer::Update</tt> function was last called.

    DisconnectionEventListener m_EventListener;

    StationIdTable* m_pStationIdTable;
    bool m_bEnableStationIdTable;

    nn::Result m_TransportStatus;  // The state of <tt>Transport</tt>, controlled by <tt>session</tt>.

    static struct DebugSetting s_DebugSetting;

    static Transport* s_pInstance;
};
}
}
}  // end of namespace nn::pia::transport
