﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_TransportAnalyzer.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_TransportAnalysisData.h>
#include <pia/common/common_Time.h>


namespace nn
{
namespace pia
{
namespace transport
{


// Forward declaration.
class ConnectionAnalyzer;

/*!
@brief  The <tt>TransportAnalyzer</tt> class reports on packet analysis results, RTT, packet loss rates, and so on.
@date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance</tt> when an instance already exists.
@date 2014-07-01  Added a feature that lets you pass analysis data as a text string to a user-defined function.
@date 2013-11-25  Added the <tt>SetMarker</tt> function.
@date 2013-11-14  Made the <tt>CreateInstance</tt>, <tt>DestroyInstance</tt>, <tt>Initialize</tt>, <tt>Finalize</tt>, <tt>Startup</tt>, and <tt>Cleanup</tt> functions private.
@date 2013-11-14  Made changes so that instances are implicitly created within <tt>PiaTransport</tt>.
@date 2013-10-29  Initial version.


*/
class TransportAnalyzer : public common::RootObject
{
public:
/*!
@brief  Defines the type of the print callback used for transport analysis data.
@details  Use this type when you want to pass an output string containing transport analysis data to a user-defined function.
@param[in] pStr  Takes the transport analysis data as a text string. Does not add a newline character.

*/
    typedef void (*AnalysisPrintCallback)(const char* pStr);


/*!
@cond PRIVATE
@brief  Creates an instance of the <tt>TransportAnalyzer</tt> class (singleton pattern).
@return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>transport</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the timing at which the <tt>CreateInstance</tt> function was called is incorrect. Make this call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Indicates that the function failed to allocate memory. Allocate more memory to the Pia library in <tt>common::Initialize</tt>. Programming error. Fix your program so that this error is not returned.
@see  DestroyInstance, GetInstance, Setting
@endcond

*/
    static nn::Result CreateInstance();


/*!
@cond PRIVATE
@brief  Destroys an instance of the <tt>TransportAnalyzer</tt> class (singleton pattern).
If no instances exist, this function does nothing.
@see  CreateInstance, GetInstance
@endcond

*/
    static void DestroyInstance();


/*!
@brief  Gets the <tt>TransportAnalyzer</tt> class instance (singleton pattern).
This function returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>CreateInstance</tt> function.
This function is thread-safe.
@return  Returns a pointer to the <tt>TransportAnalyzer</tt> instance.
@see  CreateInstance, DestroyInstance


*/
    static TransportAnalyzer* GetInstance()
    {
        return s_pInstance;
    }

/*!
@cond PRIVATE
@brief  Executes the initialization procedure.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if initialization was successful. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultAlreadyInitialized  Indicates that the <tt>Initialize</tt> function was already run. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Indicates that the function failed to allocate memory. Allocate more memory to the Pia library in <tt>common::Initialize</tt>. Programming error. Fix your program so that this error is not returned.
@endcond

*/
    nn::Result Initialize();


/*!
@cond PRIVATE
@brief  Executes the finalization procedure.
@endcond
*/
    void Finalize();


/*!
@cond PRIVATE
@brief  Starts analyzing.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the procedure was successful. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>TransportAnalyzer</tt> instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.
@endcond

*/
    nn::Result Startup();


/*!
@cond PRIVATE
@brief  Ends the analysis.
@endcond
*/
    void Cleanup();


/*!
@brief  Updates the data.
@details  The application does not need to call this API function if (A) a positive value has been set for the <tt>analysisInterval</tt> member variable in the <tt>Transport::Setting</tt> structure and (B) automatic printing of analysis data is enabled.
When the application gets analysis data when the automatic printing feature is disabled and a value of <tt>0</tt> is set for the <tt>analysisInterval</tt> member variable, update the data by calling this API function explicitly.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if the procedure was successful. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>TransportAnalyzer</tt> instance is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.






*/
    nn::Result Update();


/*!
@brief  Gets the data.
*/
    const struct TransportAnalysisData& GetTransportAnalysisData() const
    {
        return m_Data;
    }


/*!
@brief  Outputs a string to the console to delimit flags.
@param[in] pComment  A pointer to the string that is the name of the delimiter line. The string cannot include whitespace characters. Also, <tt>NULL</tt> cannot be specified.
@retval ResultInvalidArgument  Specifies that an argument is invalid. <tt>NULL</tt> pointers and strings that include whitespace characters cannot be provided as arguments. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was called at the wrong time. Programming error. Fix your program so that this error is not returned.


*/
    nn::Result SetMarker(const char* pComment);

/*!
@brief  Registers the print callback.
@param[in] pCallback  Specifies a pointer to the callback function to register.
@return  Returns the print callback that was registered before this function was called.


*/
    AnalysisPrintCallback RegisterPrintCallback(AnalysisPrintCallback pCallback);


    // Constructor.
    TransportAnalyzer();


    // Destructor.
    ~TransportAnalyzer();


/*!
@cond PRIVATE
@brief  Writes the monitoring data.
@endcond
*/
    void SetMonitoringData();


/*!
@brief  Prints information that is useful for debugging.
@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.

*/
    void Trace(u64 flag) const;

private:
    ConnectionAnalyzer* m_pConnectionAnalyzer;
    bool m_bStartup;
    struct TransportAnalysisData m_Data;

    common::Time m_StartupTime;  // The time when the <tt>Startup</tt> function was called.

    static TransportAnalyzer* s_pInstance;
};
}
}
}  // end of namespace nn::pia::transport
