﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     transport_UnreliableProtocol.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/transport/transport_definitions.h>

#include <pia/transport/transport_Protocol.h>
#include <pia/common/common_PacketDefine.h>


// Forward declaration.
namespace nn
{
namespace pia
{
namespace transport
{
class ReceivedDataContainer;
}
}
}


namespace nn
{
namespace pia
{
namespace transport
{


/*!
@brief  Represents the unreliable protocol class, with which data may be lost.
<tt>UnreliableProtocol</tt> does not guarantee that data arrives in the order it was sent.
It may not be possible to receive data with the <tt>Receive</tt> function in the same order it was sent with the <tt>Send</tt> function.
@date 2014-09-10  Corrected incorrect information about when the <tt>GetDataSizeLimit</tt> function can be called.
@date 2014-07-17  Changed the specifications so that <tt>ResultTemporaryUnavailable</tt> is returned when <tt>SendToAll</tt> is called while state transitions between sessions or joint sessions are in progress.
@date 2014-07-17  Changed the specifications so that <tt>ResultNotFound</tt> is returned when the destination <tt>StationId</tt> specified in the <tt>Send</tt> function is not found.
@date 2014-07-16  Changed the specifications so that <tt>ResultTemporaryUnavailable</tt> is returned when the <tt>Send</tt>/<tt>Receive</tt> function is called while state transitions between sessions or joint sessions are in progress.
@date 2014-03-10  Added information about when the <tt>GetDataSizeLimit</tt> function can be called.
@date 2014-03-10  Fixed errors in the description of the <tt>MAX_DATA_SIZE</tt> value. The correct value is <tt>1430</tt>.
@date 2013-07-05  Modified the <tt>Initialize</tt> function. <tt>Initialize</tt> now succeeds when called between calls to <tt>BeginSetup</tt> and <tt>EndSetup</tt> of another module.
@date 2013-05-14  Fixed the documentation of the <tt>Finalize</tt> function.
@date 2013-04-12  Changed the specifications of <tt>Send</tt> and <tt>SendToAll</tt> so that a <tt>ResultBufferIsFull</tt> object is never returned.
@date 2013-04-05  Made <tt>GetDataSizeLimit</tt> into a non-static member function.
@date 2013-04-01  Changed the specifications of the <tt>Receive</tt> function. Reordered the dummy parameters.
@date 2013-04-01  Changed the specifications of the <tt>Initialize</tt> function. If the <tt>Initialize</tt> function is called when the object has already been initialized with a call to <tt>Initialize</tt>, the function now returns <tt>ResultAlreadyInitialized</tt>.
@date 2013-04-01  Changed the specifications of the <tt>Initialize</tt> function. You can now specify the number of receive buffers in the function arguments.
@date 2013-03-19  Made major changes to the internal implementation of the <tt>UnreliableProtocol</tt> class, due to changes to the <tt>transport</tt> implementation.
@date 2013-02-04  Fixed a bug where data received in the old session would remain in the buffer after the session ended, and could be retrieved in a new session.
@date 2012-07-17  Changed the <tt>Send</tt>, <tt>SendToAll</tt>, and <tt>Receive</tt> functions. Made it so that <tt>ResultNotInCommunication</tt> is returned when communication is not possible.
@date 2012-07-17  Added the <tt>IsInCommunication</tt> function.
@date 2012-07-04  Added a description of detailed operations of the <tt>Receive</tt>, <tt>Send</tt> and <tt>SendToAll</tt> functions.
@date 2012-06-28  Clearly indicated which functions are thread-safe.
@date 2012-05-31  Changed the functionality of the <tt>Receive</tt> function. When the size of the read buffer is insufficient, <tt>ResultBufferShortage</tt> is now returned.
@date 2012-05-30  Changed the functionality of the <tt>Send</tt> and <tt>SendToAll</tt> functions. You can no longer send data in excess of the value of the <tt>GetDataSizeLimit</tt> function.
@date 2012-05-30  Added the <tt>GetDataSizeLimit</tt> function. This function gets the maximum transmission unit.
@date 2012-04-28  It was mistakenly indicated that the <tt>Initialize</tt> function would return <tt>ResultInvalidArgument</tt>. This statement was deleted.
@date 2012-04-28  Specifications changed so that the <tt>Send</tt> and <tt>SendToAll</tt> functions return <tt>ResultInvalidState</tt> when <tt>UnreliableProtocol</tt> is not initialized.
@date 2012-04-28  Fixed a bug where <tt>ResultInvalidState</tt> was left out of the list of return values for the <tt>Receive</tt> function.
@date 2012-04-06  Initial version.
*/
class UnreliableProtocol : public transport::Protocol
{
public:
/*!
@brief  Defines the maximum data size that can be sent and received at one time using <tt>UnreliableProtocol</tt>. This value is in bytes.

@details  The value of this constant is <tt>1430</tt>, but if the application sets a smaller MTU value, the amount of data that can actually be sent or received is smaller than this value.

*/
    static const size_t MAX_DATA_SIZE = common::PROTOCOL_MESSGAE_PAYLOAD_SIZE_MAX;


/*!
@brief  Gets the maximum size of data that can be sent in one transmission. This value is in bytes.

@details  The value returned by this function is affected by the MTU value set by the application and by whether packets are signed.
This function is thread-safe.
@details  If your application uses mesh networks, call this function after the mesh startup process completes.
The correct value is not returned if the function is called sooner.
@details  If your application uses sessions, call this function after joining a session.
The correct value is not returned if this function is called sooner.

@return  Returns the maximum amount of data that can be sent at one time per call to the <tt>Send</tt> or <tt>SendToAll</tt> functions. This value is in bytes.

@see  Send, SendToAll


*/
    size_t GetDataSizeLimit(void);


/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).

@details  Do not call this constructor directly from applications.
Use the <tt>@ref Transport::CreateProtocol</tt> function to create an instance.
*/
    UnreliableProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  Destroys the object.

@details  Do not call this destructor directly from applications.
Use the <tt>@ref Transport::DestroyProtocol</tt> function to destroy an instance.
*/
    virtual ~UnreliableProtocol(void);
    //! @endcond


/*!
@cond PRIVATE
@brief  This is a dispatch function.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidState  Indicates that the <tt>UnreliableProtocol::Initialize</tt> function may not have been called. Programming error. Fix your program so that this error is not returned.
*/
    virtual nn::Result Dispatch(void);
    //! @endcond


    virtual nn::Result UpdateProtocolEvent(const ProtocolEvent& event);


    PIA_PROTOCOL_TYPE_INFO(PROTOCOL_TYPE_UNRELIABLE);


/*!
@brief  Initializes the instance.

@details  Call between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions.


@param[in] recvBufNum  Specifies the number of receive buffers. Each receive buffer uses about 1,500 bytes of memory.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
@retval ResultNotInitialized  Indicates that the <tt>transport</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultAllocationFailed  Indicates that the function failed to allocate memory. Consider providing more memory when calling the <tt>common::Initialize</tt> function. Programming error. Fix your program so that this error is not returned.
@retval ResultAlreadyInitialized  Indicates that the <tt>UnreliableProtocol</tt> instance is already initialized. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidArgument  Indicates that an argument is invalid. One cause of this error is when <tt>recvBufNum</tt> is set to zero. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the function was not called between the <tt>BeginSetup</tt> and <tt>EndSetup</tt> functions. Programming error. Fix your program so that this error is not returned.
@see  Finalize
*/
    nn::Result Initialize(size_t recvBufNum = 32);


/*!
@brief  Finalizes the instance.

@details  This function returns without doing anything if it is called before the <tt>Initialize</tt> function has been called.

@see  Initialize
*/
    void Finalize(void);


/*!
@cond PRIVATE
@brief  Starts communication.

@param[in] myStationIndex  Specifies the local <tt>StationIndex</tt>.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.

*/
    virtual nn::Result Startup(StationIndex myStationIndex);
    //! @endcond


/*!
@cond PRIVATE
@brief  Ends communication.
*/
    virtual void Cleanup(void);
    //! @endcond


/*!
@brief  Loads received data into a buffer.

@details  Receives data. However, calling this function does not trigger a receive operation in the wireless module.
A receiving thread in the Pia library periodically receives packets from the wireless module.
The packets received by the receiving thread are then copied to a packet analysis process buffer that restores the data when the <tt>PiaTransport</tt> dispatch runs.
This function copies that copy of the data restored in this way into a buffer provided by the application.

@param[out] pSrcId  Specifies a pointer to the <tt>StationId</tt> of the station the data was received from.
@param[out] pRecvBuf  Specifies a pointer to the buffer that data is read into.
@param[out] pRecvDataSize  Specifies a pointer to the size of data actually read. This value is in bytes.
@param[in] recvBuffSize  Specifies the size of the buffer that receives the data. This value is in bytes.

@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>UnreliableProtocol::Initialize</tt> function may not have been called, or that the <tt>session</tt> setup might not have been completed. Programming error. Fix your program so that this error is not returned.
@retval ResultBufferShortage  Indicates that the receiving buffer is too small. Returned when the amount of incoming data is larger than the size specified by the <var>recvBuffSize</var> parameter. Programming error. Fix your program so that this error is not returned.
@retval ResultNoData  Indicates that the receiving buffer is empty and no data could be read. Handle appropriately in the application.
@retval ResultNotInCommunication  Indicates that communication is not possible. Handle appropriately in the application.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@see  Send, SendToAll, MAX_DATA_SIZE


*/
    nn::Result Receive(StationId* pSrcId, bit8* pRecvBuf, size_t* pRecvDataSize, size_t recvBuffSize);


/*!
@cond PRIVATE
@brief  Loads received data into a buffer.
*/
    nn::Result Receive(StationIndex* pSrcId, bit8* pRecvBuf, size_t* pRecvDataSize, size_t recvBuffSize);
    //! @endcond


/*!
@brief  Sends data to a specific station.

@details  Sends data. However, calling this function alone does not actually cause the data to be sent.
This function copies the data passed in by the application into a buffer that the Pia library uses to generate packets.
The data copied into the packet generation buffer is packaged into packets, and passed to the Pia internal transmit thread by running the Pia dispatch process afterwards. The actual send process periodically passes packets from the send thread to the wireless module.

@param[in] destId  Specifies the <tt>StationId</tt> of the recipient.
@param[in] pData  Specifies a pointer to the start of the data to send.
@param[in] dataSize  Specifies the size of the data to send. This value is in bytes. This size must not be greater than the value returned by the <tt>GetDataSizeLimit</tt> function.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. This error is also returned when the size of the data to send is too large. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>UnreliableProtocol::Initialize</tt> function may not have been called, or that the <tt>session</tt> setup might not have been completed. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInCommunication  Indicates that communication is not possible. Handle appropriately in the application.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@retval ResultNotFound  Indicates that the specified destination was not found.
@see  SendToAll, Receive, MAX_DATA_SIZE, GetDataSizeLimit, IsInCommunication


*/
    nn::Result Send(StationId destId, const bit8* pData, size_t dataSize);


/*!
@cond PRIVATE
@brief  Sends data to a specific station.
*/
    nn::Result Send(StationIndex destId, const bit8* pData, size_t dataSize);
    //! @endcond


/*!
@brief  Sends data to all stations.

@details  Sends data. However, calling this function alone does not actually cause the data to be sent.
This function copies the data passed in by the application into a buffer that the Pia library uses to generate packets.
The data copied into the packet generation buffer is packaged into packets, and passed to the Pia internal transmit thread by running the Pia dispatch process afterwards. The actual send process periodically passes packets from the send thread to the wireless module.

@param[in] pData  Specifies a pointer to the start of the data to send.
@param[in] dataSize  Specifies the size of the data to send. This value is in bytes. This size must not be greater than the value returned by the <tt>GetDataSizeLimit</tt> function.
@return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess</tt> function returns <tt>true</tt> if execution succeeds.
@retval ResultInvalidArgument  Indicates that one or more arguments are invalid. This error is also returned when the size of the data to send is too large. Programming error. Fix your program so that this error is not returned.
@retval ResultInvalidState  Indicates that the <tt>UnreliableProtocol::Initialize</tt> function may not have been called. Programming error. Fix your program so that this error is not returned.
@retval ResultNotInCommunication  Indicates that communication is not possible. Handle appropriately in the application.
@retval ResultTemporaryUnavailable  The API function is temporarily unavailable because the joint session process is in progress. Handle appropriately in the application.
@see  Send, Receive, MAX_DATA_SIZE, GetDataSizeLimit, IsInCommunication


*/
    nn::Result SendToAll(const bit8* pData, size_t dataSize);


/*!
@cond PRIVATE
@brief  Sends data to multiple stations.
*/
    nn::Result SendToMulti(u32 stationBitmap, const bit8* pData, size_t dataSize);
    //! @endcond


/*!
@brief  Determines whether communication is possible.

@return  Returns <tt>true</tt> if communication is possible; returns <tt>false</tt> otherwise.
*/
    bool IsInCommunication(void) const;


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    virtual void Trace(u64 flag) const;


private:
    nn::Result sendImpl(StationIndex destId, const bit8* pData, size_t dataSize);

protected:
    void* m_pRawBuffer;  // Raw memory passed to <tt>m_pReceiveBuffer</tt>.
    ReceivedDataContainer* m_pReceiveBuffer;

    StationIndex m_LocalStationIndex;
};
}
}
}  // end of namespace nn::pia::transport
