/*---------------------------------------------------------------------------*
  Project: Pia
  File: types.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

// Moved definitions according to the platform.
#include <pia/platform.h>


namespace nn
{
namespace pia
{
/*!
@brief  Definition of the <tt>StationId</tt> identifying the station within the session.

@details  A unique value is assigned to each station participating in the session.
            You will never see the same value for another participating station.
            That said, after participating in a session, if a station leaves,
            its ID might be assigned to another session.
*/
struct StationId
{
    u32 uniqueId;   //!< A unique station value.
    u32 reserved;   //!< This is a reserved region.

    StationId()
      : uniqueId(0),
        reserved(0)
    {
    }

    StationId(u64 arg)
    {
*this = arg;
    }

    StationId& operator=(const StationId& rhs)
    {
        uniqueId = rhs.uniqueId;
        reserved = rhs.reserved;
        return *this;
    }

    StationId& operator=(u64 arg)
    {
        uniqueId = arg % 0x100000000;
        reserved = static_cast<u32>(arg / 0x100000000);
        return *this;
    }

    bool operator==(const StationId& rhs) const
    {
        return (uniqueId == rhs.uniqueId) && (reserved == rhs.reserved);
    }

    bool operator!=(const StationId& rhs) const
    {
        return !(*this == rhs);
    }

    u64 GetValue() const
    {
        return (static_cast<u64>(uniqueId) << 0) | (static_cast<u64>(reserved) << 32);
    }
};


const u64 STATION_ID_INVALID_VALUE = 0xFFFFFFFFFFFFFFFD; //!< Value indicating invalid <tt>StationId</tt>.
const u64 STATION_ID_HOST_VALUE    = 0xFFFFFFFFFFFFFFFE; //!< Value indicating host <tt>StationId</tt>.
const u64 STATION_ID_ALL_VALUE     = 0xFFFFFFFFFFFFFFFF; //!< Value indicating <tt>StationId</tt> for all participants.



/*!
  @brief  Gets the <tt>const</tt> reference of the <tt>StationId</tt> object representing the invalid <tt>StationId</tt>.
            You can also use the <tt>STATION_ID_INVALID</tt> macro.
*/
const StationId& GetStationIdInvalid();


/*!
  @brief  Gets the <tt>const</tt> reference of the <tt>StationId</tt> object representing the host <tt>StationId</tt>.
            You can also use the <tt>STATION_ID_HOST</tt> macro.
*/
const StationId& GetStationIdHost();


/*!
  @brief  Gets the <tt>const</tt> reference indicating the <tt>StationId</tt> object to specify all session participants when sending and receiving data.
            You can also use the <tt>STATION_ID_ALL</tt> macro.
*/
const StationId& GetStationIdAll();


#define STATION_ID_INVALID GetStationIdInvalid()
#define STATION_ID_HOST    GetStationIdHost()
#define STATION_ID_ALL     GetStationIdAll()


}
}  // end of namespace nn::pia
