﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     util_Api.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/util/util_definitions.h>


namespace nn
{
namespace pia
{
namespace util
{

/*!
@brief  Enumerates the types of handling expected.
*/
    enum HandlingType
    {
        HANDLING_VOID,             //!< Invalid value.
        HANDLING_IGNORABLE,        //!< Can be ignored.
        HANDLING_RETRY,            //!< Try to execute again after some time has passed.
        HANDLING_CLEANUP,          //!< A cleanup process is needed.
        HANDLING_LOGOUT,           //!< A logout process is needed.
        HANDLING_LEAVE,            //!< A process for leaving a session is needed.
        HANDLING_LEAVE_AND_CLEANUP,//!< After leaving a session, execute a cleanup process.
        HANDLING_RECONNECT,        //!< A process for reconnecting is needed.
        HANDLING_REINITIALIZE,     //!< Reinitialization is required.
        HANDLING_PROGRAMMING_ERROR //!< Programming error. The source code must be revised.
    };


/*!
@if CTR_DOC
@brief  Enumerates the policies for using the Error/EULA applet.
@endif
@if CAFE_DOC
@brief  Enumerates the policies for using the error viewer.
@endif
*/
    enum ViewerType
    {
        VIEWER_VOID,            //!< Invalid value.
        VIEWER_SHOULD_USE,      //!< Supposed to be used.
        VIEWER_MAY_USE,         //!< Okay to use.
        VIEWER_SHOULD_NOT_USE   //!< Do not use.
    };


/*!
@brief  Determines whether the value is a Pia result.

@param[in] r  The result value for the determination.

@return  Returns <tt>true</tt> if the value is a Pia <tt>Result</tt>, and returns <tt>false</tt> otherwise.
*/
    bool IsPiaResult(const nn::Result& r);


/*!
@brief  Returns the type of handling expected for the <tt>Result</tt> from Pia.

@param[in] r  The result value that determines the type of handling.

@return  Returns an enumerated value for the handling that corresponds to the <tt>Result</tt>.
Returns <tt>HANDLING_VOID</tt> if passed a non-Pia <tt>Result</tt> or a <tt>Result</tt> that Pia only uses internally.

*/
    enum HandlingType GetHandlingType(const nn::Result& r);


/*!
@if CTR_DOC
@brief  Determines whether the Pia <tt>Result</tt> is one for which the Error/EULA applet may be used.
@details  This API determines whether it is appropriate to pass the network error code that results from feeding the result of a failed Pia API call to <tt>NetworkErrorCodeConverter::Convert</tt> on to the Error/EULA applet.



@details  For error results returned during local communication, however, do not use the Error/EULA applet, regardless of the return value of this API.

@endif
@if CAFE_DOC
@brief  Determines whether the Pia <tt>Result</tt> is one for which the error viewer may be used.
@details  This API determines whether it is appropriate to pass the network error code that results from feeding the result of a failed Pia API call to <tt>NetworkErrorCodeConverter::Convert</tt> on to the error viewer.



@endif

@param[in] r  The result value for the determination.

@return  Returns the <tt>ViewerType</tt> that corresponds with the <tt>Result</tt> passed in.
Returns <tt>VIEWER_VOID</tt> when a non-Pia <tt>Result</tt> is passed.
@see  NetworkErrorCodeConverter::Convert
*/
    enum ViewerType GetViewerType(const nn::Result& r);


/*!
@brief  Returns a string representation of the Pia <tt>Result</tt>.

@param[in] r  The Pia <tt>Result</tt> value or <tt>ResultSuccess</tt>.

@return  Returns that string representation when the Pia <tt>Result</tt> or <tt>ResultSuccess</tt> is passed.
Returns an empty string ("") when any other <tt>Result</tt> is passed.
*/
    const char* GetResultString(const nn::Result& r);

}
}
}  // end of namespace nn::pia::util
