﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     util_NetworkErrorCodeConverter.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/util/util_definitions.h>

#include <pia/common/common_RootObject.h>


namespace nn
{
namespace pia
{
namespace util
{


/*!
@cond CTR_DOC
@brief  This class converts <tt>nn::Result</tt> type errors into network error codes passed to the error/EULA applet.
@endcond
@cond CAFE_DOC
@brief  This class converts <tt>nn::Result</tt> type errors into network error codes passed to Error Viewer.
@endcond

@date 2015-04-16  Fixed a bug where <tt>ResultNexInternalError</tt> objects could not be converted using the <tt>Convert</tt> function after the <tt>Session::Cleanup</tt> function was called, causing the library to stop on an assertion.
@date 2014-11-04  Changed the operations of the <tt>Convert</tt> function.
@date 2014-08-07  Changed Pia so that it halts on an assertion if something other than a <tt>Result</tt> is passed to the <tt>Convert</tt> function.
@date 2012-11-06  Replaced locations that coded non-Pia <tt>Result</tt> values passed to <tt>Convert</tt> as immediate values with symbol names. The actual values passed remain the same.
@date 2012-06-21  Initial version.
*/
class NetworkErrorCodeConverter : public common::RootObject
{
public:
#if NN_PIA_CTR
    static const u32 NN_PIA_ERROR_CODE_BASE = 180000;  // 018-xxxx has been assigned to CTR Pia.
#elif NN_PIA_WIN || NN_PIA_NIN_WIN
    static const u32 NN_PIA_ERROR_CODE_BASE = 180000;  // Although the numbers have been matched to CTR, is the error display applet used in the Windows environment?
#elif NN_PIA_CAFE
    static const u32 NN_PIA_ERROR_CODE_BASE = 1180000;  // 118-xxxx has been assigned to Cafe Pia.
#else
#error  "Invalid platform"
#endif

/*!
@cond PRIVATE
@brief  Instantiates the object with default parameters (default constructor).
@endcond
*/
    NetworkErrorCodeConverter(void);


/*!
@cond PRIVATE
@brief  Destroys the object.
@endcond
*/
    ~NetworkErrorCodeConverter(void);


/*!
@brief  Converts Pia library <tt>Result</tt> values to network error codes.

@details  Converts the <tt>Result</tt> type value returned by the Pia library to a network error code.
@details  As an exception, returns a network error code corresponding to the NEX internal error for this <tt>Result</tt> if the <tt>ResultNexInternalError</tt> is passed.
There are several types of internal error, so the return value when passed <tt>ResultNexInternalError</tt> can be undefined.
@details  If <tt> ResultNexInternalError</tt> is passed to this function when the Pia API has not yet returned <tt>ResultNexInternalError</tt>, this function asserts.
@if CTR_DOC
The converted values can be passed to the Error/EULA applet.
@endif
@if CAFE_DOC
The converted values can be passed to the Error Viewer.
@endif
If <tt>IsSuccess</tt> is <tt>true</tt> for the argument, <tt>0</tt> is returned. If an argument not defined in the Pia library is passed, Pia raises an assertion. (In release builds, it returns <tt>NN_PIA_ERROR_CODE_BASE</tt>.)
This function converts the results defined in the Pia library into unique identification codes, with the exception of <tt>ResultNexInternalError</tt>, but it is not necessary to show all codes as errors to the user.
For information about the error codes to show,
@if CTR_DOC
For more information, see the <i>CTR Guidelines: Internet</i> or the network error code list.
@endif
@if CAFE_DOC
See the chapter about displaying errors in <i>Wii U Guidelines</i>.
@endif

@param[in] r  Specifies the <tt>nn::Result</tt> value to convert to a network error code.
@return  Returns the corresponding network error code.






*/
    static u32 Convert(const nn::Result& r);


/*!
@cond PRIVATE
@brief  Sets the network error codes for errors that occur in communication with the game server.

@param[in] networkErrorCode  Sets the value converted from the NEX's <tt>qResult</tt> into a network error code.
@endcond
*/
    static void SetNexNetworkErrorCode(u32 code) { s_NexNetworkErrorCode = code; }


/*!
@cond PRIVATE
@brief  Clears the network error code set by the <tt>SetNexNetworkErrorCode</tt> function.
@endcond
*/
    static void ClearNexNetworkErrorCode(void) { s_NexNetworkErrorCode = 0; }


/*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> type.
*/
    void Trace(u64 flag) const;

protected:
private:
    static u32 s_NexNetworkErrorCode;
};
}
}
}  // end of namespace nn::pia::util
