﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SampleFramework.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_definitions.h>

#include "SampleFramework_Color.h"
#include "SampleFramework_ButtonDefines.h"

#if defined(NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST)
#include "PiaTestLib/PiaTest.h"
#endif  // End of NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST


namespace nn
{
namespace pia
{
namespace sampleframework
{

// Forward declaration.
class SampleFrameworkImplBase;


/*!
  @brief  This is a simple framework class. It is meant to be used in the Pia sample programs.
 */
class SampleFramework
{
public:
    // This structure holds the configuration parameters for creating a <tt>SampleFramework</tt> instance.
    struct Setting
    {
        Setting(void)
          : bUpperScreenOnly(false), bRawTouchpad(false)
        {
        }
        bool bUpperScreenOnly;  // This parameter is only meaningful for the Cafe environment. Set to <tt>true</tt> to only display the top screen in expanded view in the Wii U GamePad.
        bool bRawTouchpad;      // This parameter is only meaningful for the Cafe environment. If set to <tt>true</tt>, it gets the raw touch panel coordinates.
    };


    /*!
    @brief  Creates an instance (singleton pattern).
    @return  Returns <tt>true</tt> if the instance was created successfully, and <tt>false</tt> otherwise.
     

*/
    static bool CreateInstance(void);


    /*!
    @brief  Creates an instance with the properties defined in the parameters (singleton pattern).
    @param[in] setting  Specifies the <tt>Setting</tt> structure instance used to set the framework.
    @return  Returns <tt>true</tt> if the instance was created successfully, and <tt>false</tt> otherwise.
     
*/
    static bool CreateInstance(const SampleFramework::Setting& setting);


    /*!
    @brief  Destroys the instance (singleton pattern).
     */
    static void DestroyInstance(void);


    /*!
    @brief  Gets a pointer to the instance (singleton pattern).
    @return  The pointer to the <tt>SampleFramework</tt> instance.
     */
    static SampleFramework* GetInstance(void);


    /*!
    @brief  Determines whether an instance has been created (singleton pattern).
    @return  Returns the result of the determination.
     */
    static bool IsCreated(void);


    ~SampleFramework(void);


    /*!
    @brief  Initializes the sound features of the <tt>SampleFramework </tt> instance.
           An audio file to stream using <tt>@ref PlayBGM</tt> or play using <tt>@ref PlayShortSound</tt> can be specified as an argument.
            
           A BGM audio file will be streamed. Specify a BCWAV file that can be looped. 
           Specify special-effects audio files as WAV files. 
           This function starts an audio thread and a thread for loading files for streaming playback.
    @param[in] StreamFileName  Specifies the name of a BGM (background music) audio file.
    @param[in] VoiceFileName  Specifies an array of sound-effect audio filenames.
    @param[in] VoiceFileName  Specifies the number of sound-effect audio files.
    @return  Returns <tt>true</tt> if the initialization succeeded, and <tt>false</tt> otherwise.
     

*/
    bool SoundInitialize(
        const char streamFileName[],
        const char* voiceFileName[] = NULL,
        const u32 voiceFileNum = 0);


    /*!
    @brief  Finalizes the sound features of the <tt>SampleFramework</tt> instance.
     */
    void SoundFinalize(void);


    /*!
    @brief  Starts BGM audio playback.
           Starts streaming the BGM audio file you specified as an argument to the <tt>@ref SoundInitialize function</tt>. 
           This function loops playback until you call the <tt>@ref PauseBGM</tt> or <tt>@ref StopBGM</tt> function, or until you finalize the sound feature by calling the <tt>@ref SoundFinalize</tt> function.
           
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool PlayBGM(void);


    /*!
    @brief  Pauses BGM audio.
           This function pauses playback if you are playing the BGM audio file specified in the argument to the <tt>@ref SoundInitialize</tt> function by using the <tt>@ref PlayBGM</tt> function.
           Call <tt>@ref PlayBGM</tt> again to continue playback.
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool PauseBGM(void);


    /*!
    @brief  Stops BGM audio playback.
    This function stops the playback of the BGM audio file you specified as an argument to the <tt>@ref SoundInitialize</tt> function.
    The next time that the <tt>@ref PlayBGM</tt> function is called, the audio file plays back from the beginning.
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool StopBGM(void);


    /*!
    @brief  Starts playback of a sound effect.
           This function plays back one of the sound effect audio files you specified as an argument to the <tt>@ref SoundInitialize</tt> function.
           The audio file index depends on its position in the array you passed to the <tt>@ref SoundInitialize</tt> function.
           There is a limit to the number of sound effects that can be played at one time (<tt>SHORT_SOUND_MAX</tt>).
           If you call this function and specify a new sound effect while the maximum number of simultaneous sound effects are already playing, it stops playback of the first of the current sound effects to begin playback, and plays back the specified sound effect.
           
    @param[in] voiceFileIndex  Specifies the index of the audio file that you want to play.
    @return  Returns <tt>true</tt> if starting playback succeeded, and <tt>false</tt> otherwise.
     


*/
    bool PlayShortSound(const u32 voiceFileIndex);


    /*!
    @brief  Pauses all sound effects that are currently playing.
           This function pauses all sound effects being played back using the <tt>@ref PlayShortSound</tt> function. 
           You can resume playback of all the files by calling the <tt>@ref PlayShortSoundAll</tt> function.
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool PauseShortSoundAll(void);


    /*!
    @brief  Resumes playback of all paused sound effects.
    This function resumes playback of all sound effects paused using the <tt>@ref PauseShortSoundAll</tt> function.
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool PlayShortSoundAll(void);


    /*!
    @brief  Stops all sound effects currently playing.
           This function stops all sound effects being played back using the <tt>@ref PlayShortSound</tt> function. 
           Calling the <tt>@ref PlayShortSoundAll</tt> function does not resume playback.
    @return  Returns <tt>true</tt> if execution succeeds, and <tt>false</tt> otherwise.
     

*/
    bool StopShortSoundAll(void);


    /*!
    @brief  Clears the screen.
     */
    void ClearScreen(void);


    /*!
    @brief  Specifies the color used to clear the screen.
    @param[in] clr  Specifies the fill color to use when the <tt>@ref EndFrame</tt> function is called.
     */
    void SetClearColor(const struct Color& clr);


    /*!
    @brief  Specifies the font color.
    The default is white (1.0f, 1.0f, 1.0f, 1.0f).
    @param[in] clr  Specifies the font color to use when text is drawn using the <tt>@ref DrawText</tt> function.
     

*/
    void SetFontColor(const struct Color& clr);

    /*!
    @brief  Specifies the font size. This function sets the width and height specified by <tt>pixel</tt>.
           The default font size is 8.
     */
    void SetFontSize(int pixel);

    /*!
    @brief  Gets the font width.
     */
    int GetFontWidth(void) const;

    /*!
    @brief  Gets the font height.
     */
    int GetFontHeight(void) const;

    /*!
    @brief  Draws text at the specified position. The upper-left screen coordinate is (0, 0).
    @param[in] x  Specifies the x-coordinate.
    @param[in] y  Specifies the y-coordinate.
    @param[in] fmt  Specifies the text string in <tt>printf</tt> format.
     
*/
    void DrawText(s32 x, s32 y, const char* fmt, ...);


    /*!
    @brief  Draws a rectangle.
    @param[in] x  Specifies the x-coordinate of the rectangle's upper-left corner.
    @param[in] y  Specifies the y-coordinate of rectangle's upper-left corner.
    @param[in] w  Specifies the rectangle width.
    @param[in] h  Specifies the rectangle height.
    @param[in] clr  Specifies the rectangle color.
     
*/
    void DrawRect(s32 x, s32 y, s32 w, s32 h, const struct Color& clr);


    /*!
    @brief  Sets the size to use when drawing a point.
    @param[in] pointSize  Specifies the size of a point in pixels.
     
*/
    void SetPointSize(f32 pointSize);


    /*!
    @brief  Draws a point.
    @param[in] x  Specifies the x-coordinate of the point's center.
    @param[in] y  Specifies the y-coordinate of the point's center.
    @param[in] clr  Specifies the point's color.
     
*/
    void DrawPoint(s32 x, s32 y, const struct Color& clr);


    /*!
    @brief  Sets the thickness to use when drawing lines.
    @param[in] lineWidth  Specifies the line thickness in pixels.
     
*/
    void SetLineWidth(f32 lineWidth);


    /*!
    @brief  Draws a line segment.
    @param[in] x0  Specifies the x-coordinate of the start point of the line segment.
    @param[in] y0  Specifies the y-coordinate of the start point of the line segment.
    @param[in] x1  Specifies the x-coordinate of the end point of the line segment.
    @param[in] y1  Specifies the y-coordinate of the end point of the line segment.
    @param[in] clr  Specifies the line segment's color.
     
*/
    void DrawLine(s32 x0, s32 y0, s32 x1, s32 y1, const struct Color& clr);


    /*!
    @brief  Changes the screen to render, and clears the buffer.
           After calling this function, calls to the <tt>@ref DrawText</tt> and <tt>@ref DrawRect</tt> functions draw to the new target screen.
           During the call to the <tt>@ref BeginFrame</tt> function, the render target returns to the top screen.
    @param[in] target  Specifies the top screen if <tt>true</tt>, or the bottom screen if <tt>false</tt>.
     

*/
    void ChangeRenderTarget(bool target);


    void SwapBuffers(void);

    /*!
    @brief  This function is meant to be called at the beginning of the main loop.
    Use this function if you want to draw anything before calling the <tt>@ref BeginFrame</tt> function.
    Calling this function does not update the screen. In this sense, it is the same as a dropped frame.
     
*/
    void SkipFrame(void);

    /*!
    @brief  This function is meant to be called at the beginning of the main loop.
    Call this function at the beginning of an application's main loop.
    Actions such as getting key input are performed inside this function.
     
*/
    void BeginFrame(void);


    /*!
    @brief  This function is meant to be called at the beginning of the main loop.
    This function is nearly the same as the <tt>BeginFrame</tt> function, except that it does not clear the screen.
     
*/
    void BeginFrameNoRender(void);

    /*!
    @brief  This function is meant to be called at the end of the main loop.
    Call this function at the end of an application's main loop.
    It displays the results of rendering.
     
*/
    void EndFrame(void);


    /*!
    @brief  This function is meant to be called at the end of the main loop.
    Call this function at the end of an application's main loop.
    This function is nearly the same as the <tt>EndFrame</tt> function, except that it does not swap framebuffers.
     
*/
    void EndFrameNoRender(void);


    /*!
    @brief  Gets the frame count.
    The frame count is incremented automatically.
     
*/
    u32 GetFrameCount(void) const;


    /*!
    @brief  Sets the frames per second (FPS).
    @param[in] fps  Specifies the frames per second. The value specified must be a factor of 60.
    @return  Returns <tt>true</tt> on success. Returns <tt>false</tt> if the value is invalid as a setting.
     

*/
    bool SetFramePerSecond(s32 fps);


    /*!
    @brief  Gets the FPS setting.
    @return  Returns the current FPS setting.
     
*/
    s32 GetFramePerSecond(void) const;


    /*!
    @brief  Determines whether a button is pressed.
    @param[in] button  Specifies a bitwise OR of <tt>nn::hid::BUTTON_XXX</tt> values.
    @return  Returns the result of the determination.
     */
    bool IsHold(bit32 button) const;


    /*!
    @brief  Determines whether a button is being pressed at this moment.
    @param[in] button  Specifies a bitwise OR of <tt>nn::hid::BUTTON_XXX</tt> values.
    @return  Returns the result of the determination.
     */
    bool IsTrigger(bit32 button) const;


    bit32 GetTrigger(void) const;

    /*!
    @brief  Determines whether a button was released momentarily.
    @param[in] button  Specifies a bitwise OR of <tt>nn::hid::BUTTON_XXX</tt> values.
    @return  Returns the result of the determination.
     */
    bool IsRelease(bit32 button) const;


    /*!
    @brief  Determines whether the touch panel was momentarily touched by the stylus.
    @return  Returns the result of the determination.
     */
    bool IsPenTriggered(void) const;


    /*!
    @brief  Determines whether the stylus is in contact with the touch panel.
    @return  Returns the result of the determination.
     */
    bool IsPenTouched(void) const;


    /*!
    @brief  Determines whether the stylus was momentarily released from the touch panel.
    @return  Returns the result of the determination.
     */
    bool IsPenReleased(void) const;


    /*!
    @brief  Gets the coordinates where the stylus last touched the touch panel.
    @param[out] x  Specifies the x-coordinate.
    @param[out] y  Specifies the y-coordinate.
    @return  Returns <tt>true</tt> if the coordinates were retrieved successfully. Returns <tt>false</tt> if the touch panel has not been touched at all.
     */
    bool GetLatestPenPosition(u16* x, u16* y) const;


    /*!
    @brief  Gets the input of the Circle Pad in the x direction.
    @return  Returns the input value in the x direction, normalized to a range from -1.0 through 1.0.
     */
    f32 GetAnalogX(void);


    /*!
    @brief  Gets the input of the Circle Pad in the y direction.
    @return  Returns the input value in the y direction, normalized to a range from -1.0 through 1.0.
     */
    f32 GetAnalogY(void);


    /*!
    @brief  Gets the input value of the Circle Pad.
    @param[out] x  Specifies the input value in the x direction, normalized to a range from -1.0 through 1.0.
    @param[out] y  Specifies the input value in the y direction, normalized to a range from -1.0 through 1.0.
     */
    void GetAnalog(f32* x, f32* y);


    /*!
    @brief  Gets the accelerometer value.
    @param[out] x  Specifies the left direction of the +Control Pad. This parameter must not specify a NULL pointer.
    @param[out] y  Specifies the surface direction of the touch panel. This parameter must not specify a NULL pointer.
    @param[out] x  Specifies the up direction of the +Control Pad. This parameter must not specify a NULL pointer.
    @return  Returns <tt>true</tt> if the value was retrieved successfully, and <tt>false</tt> otherwise.
     */
    bool GetAccelerationData(s16* x, s16* y, s16* z) const;


    u32 WaitForButton(u32 button);

    /*!
    @brief  Prints information about the current thread.
     */
    void PrintCurrentThreadInfo(void) const;


    /*!
    @brief  Puts the current thread to sleep.
    @param[in] msec  Specifies the time to make the thread sleep. The value is in milliseconds.
     */
    void SleepCurrentThreadMilliseconds(s32 msec);


    // {Secret}
    // Gets the device memory.
    void* AllocateDeviceMemory(size_t size);


    // {Secret}
    // Restores the device memory.
    void FreeDeviceMemory(void* ptr);


    /*!
    @brief  Determines whether this function was called from the main thread.
    @return  Returns the result of the determination.
     */
    bool IsInMainThread(void) const;


    // {Secret}
    // Code to identify the padding warning.
    void Checker(void);


    /*!
    @brief  Gets a string representing the elapsed time since the program began execution.
    @param[out] str  Specifies the result in the following format: 00:00:00 (HH::MM::SS). Allocate a buffer of sufficient size.
     */
    void GetPassedTimeString(char str[]);


    /*!
    @brief  Sets the <tt>Flush</tt> interval for <tt>CachedPrint</tt>.
    @param[in] span  Specifies the number of frames at which to call the <tt>Flush</tt> function.    
     */
    void SetCachedPrintFlushSpan(u32 span);


    /*!
    @brief  Determines whether a system termination request has been received.
     */
    bool IsSystemTerminationRequested(void) const;


    /*!
    @brief  Sets the current state of the program (for library verification).
    @param[in] sceneNo  Specifies the scene number for the sample demo.
    @param[in] isConnected  Specify <tt>true</tt> if currently communicating; otherwise specify <tt>false</tt>.
    @param[in] isHost  Specify <tt>true</tt> if the local station is the host, otherwise specify <tt>false</tt>.
    @param[in] stationNum  Specifies the number of stations communicating.
    @param[in] linkLevel  Specifies the signal strength (in the range from 0 through 3) of wireless communication.
     */
    void SetProgramStatus(
        u32 sceneNo,
        bool isConnected,
        bool isHost = false,
        u32 stationNum = 0,
        u32 linkLevel = 0);

    /*!
    @brief  Sets all current program parameters in an array. (for library verification).
    @param[in] paramArray  Specifies the array in which to set parameters.
    @param[in] paramNum  Specifies the number of the elements in <span class="argument">paramArray</span>.
     */
    void SetProgramParamArray(
        bit32 paramArray[],
        size_t paramNum);

    /*!
    @brief  Sets one of the current program parameters. (for library verification).
    @param[in] paramIndex  Specifies the index of the parameter to set.
    @param[in] paramVar  Specifies the parameter.
     */
    void SetProgramParam(
        size_t paramIndex,
        bit32 paramVar);

#if defined(NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST)
    PiaTest::PiaTestCommand GetLastCommandFromHost();
#endif  // End of NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST

private:
    SampleFramework(void);
    SampleFramework(const SampleFramework::Setting& setting);
    void constructorImpl(const SampleFramework::Setting& setting);

    SampleFrameworkImplBase* m_pImpl;

    static SampleFramework* s_pInstance;
};
}
}
}  // end of namespace nn::pia::sampleframework
