﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SampleFrameworkImplCtr.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>
#if NN_PIA_CTR

#include "SampleFrameworkImplBase.h"

#include <nn/snd.h>


namespace demo
{
class RenderSystemDrawing;
}

namespace nn
{
namespace pia
{
namespace sampleframework
{

// Somehow it seems the PadStatus or PadReader constructor requires that nn::hid::Initialize be called.
// 
// Plan for this by preparing an object exclusively for initializing and releasing <tt>hid</tt>.
class Hid
{
public:
    Hid(void)
    {
        nn::Result result = nn::hid::Initialize();
        NN_UTIL_PANIC_IF_FAILED(result);
    }
    ~Hid(void)
    {
        nn::hid::Finalize();
    }
};


class SampleFrameworkImplCtr : public SampleFrameworkImplBase
{
public:
    SampleFrameworkImplCtr(void);
    virtual ~SampleFrameworkImplCtr(void);

    virtual bool SoundInitialize(const char streamFileName[], const char* voiceFileName[] = NULL, const u32 voiceFileNum = 0);
    virtual void SoundFinalize(void);
    virtual bool PlayBGM(void);
    virtual bool PauseBGM(void);
    virtual bool StopBGM(void);
    virtual bool PlayShortSound(const u32 voiceFileIndex);
    virtual bool PauseShortSoundAll(void);
    virtual bool PlayShortSoundAll(void);
    virtual bool StopShortSoundAll(void);
    virtual void ClearScreen(void);
    virtual void SetClearColor(const struct Color& clr);
    virtual void SetFontColor(const struct Color& clr);
    virtual void SetFontSize(int pixel);
    virtual int GetFontWidth(void) const;
    virtual int GetFontHeight(void) const;
    virtual void DrawText(s32 x, s32 y, const char* pText);
    virtual void DrawRect(s32 x, s32 y, s32 w, s32 h, const struct Color& clr);
    virtual void SetPointSize(f32 pointSize);
    virtual void DrawPoint(s32 x, s32 y, const struct Color& clr);
    virtual void SetLineWidth(f32 lineWidth);
    virtual void DrawLine(s32 x0, s32 y0, s32 x1, s32 y1, const struct Color& clr);
    virtual void ChangeRenderTarget(bool target);
    virtual void SwapBuffers(void);
    virtual void SkipFrame(void);
    virtual void BeginFrame(void);
    virtual void BeginFrameNoRender(void);
    virtual void EndFrame(void);
    virtual void EndFrameNoRender(void);
    virtual u32 GetFrameCount(void) const;
    virtual bool SetFramePerSecond(s32 fps);
    virtual s32 GetFramePerSecond(void) const;
    virtual bool IsHold(bit32 button) const;
    virtual bool IsTrigger(bit32 button) const;
    virtual bit32 GetTrigger() const;
    virtual bool IsRelease(bit32 button) const;
    virtual bool IsPenTriggered(void) const;
    virtual bool IsPenTouched(void) const;
    virtual bool IsPenReleased(void) const;
    virtual bool GetLatestPenPosition(u16* x, u16* y) const;
    virtual f32 GetAnalogX(void);
    virtual f32 GetAnalogY(void);
    virtual void GetAnalog(f32* x, f32* y);
    virtual bool GetAccelerationData(s16* x, s16* y, s16* z) const;
    virtual u32 WaitForButton(u32 button);
    virtual void PrintCurrentThreadInfo(void) const;
    virtual void SleepCurrentThreadMilliseconds(s32 msec);
    virtual void* AllocateDeviceMemory(size_t size);
    virtual void FreeDeviceMemory(void* ptr);
    virtual bool IsInMainThread(void) const;
    virtual void Checker(void);
    virtual void GetPassedTimeString(char str[]);
    virtual void SetCachedPrintFlushSpan(u32 span);
    virtual bool IsSystemTerminationRequested(void) const;
    virtual void SetProgramStatus(u32 sceneNo, bool isConnected, bool isHost, u32 stationNum, u32 linkLevel);
    virtual void SetProgramParamArray(bit32 paramArray[], size_t paramNum);
    virtual void SetProgramParam(size_t paramIndex, bit32 paramVar);

#if defined(NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST)
    virtual PiaTest::PiaTestCommand GetLastCommandFromHost();
#endif  // End of NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST

private:
    static const size_t GX_HEAP_SIZE = 0x800000;

    static const u16 INVALID_TOUCH_POS = 0xffff;

    static const s32 INVALID_DISPLAY = -1;

    static const s32 VSYNCS_PER_SECOND = 60;

    static const s16 INVALID_ACCELEROMETER_VALUE = 1024;

    static const int FONT_WIDTH = 8;   // Default width of characters drawn with <tt>DrawText</tt>. The unit is pixels.
    static const int FONT_HEIGHT = 8;  // Default height of characters drawn with <tt>DrawText</tt>. The unit is pixels.

    void beginFrameImpl(bool bRender);
    void endFrameImpl(bool bRender);

    static SampleFrameworkImplCtr* s_pInstance;  // Instance for a singleton.

    Hid m_hid;  // You must call this before <tt>m_padStatus</tt> or <tt>m_padReader</tt>!
    NN_PADDING3;

    uptr m_heapForGx;

    nn::hid::PadStatus m_padStatus;
    nn::hid::PadReader m_padReader;

    nn::hid::AccelerometerStatus m_accStatus;  // s16, s16, s16
    NN_PADDING2;
    nn::hid::AccelerometerReader m_accReader;

    bool m_accValid;
    NN_PADDING3;

    nn::hid::TouchPanelStatus m_tpStatus;
    nn::hid::TouchPanelStatus m_tpStatusPrev;

    NN_PADDING4;
    nn::hid::TouchPanelReader m_tpReader;  // Include 64-bit member variables.

    u16 m_latestX;  // X-coordinate where the stylus last touched.
    u16 m_latestY;  // Y-coordinate where the stylus last touched.

    u32 m_FrameCount;

    demo::RenderSystemDrawing* m_pSampleFramework;

    nn::fnd::ExpHeap m_appHeap;

    s32 m_renderingTarget;

    s32 m_vsyncInterval;  // Variable to support FPS.60 -> 1, 30 -> 2, 20 -> 3, ...

    GLint m_beginFrameVsync;  // <tt>VSYNC</tt> value when the <tt>BeginFrame</tt> function was called.

    Color m_FontColor;  // The color used by <tt>DrawText</tt>.

    bool m_SoundInitialized;  // Initialized flag for sound features.
    NN_PADDING3;

    const char* m_StreamFileName;  // Name of the sound file for background music.
    const char** m_VoiceFileName;  // Sound file for sound effects.
    u32 m_VoiceFileNum;            // Number of sound effect sound files.

    nn::os::CriticalSection m_csSoundParamLock;

    nn::snd::Voice* m_pVoiceBGM;                        // Voices for background music.
    static const u32 SHORT_SOUND_MAX = 8;               // Maximum number of sound effects that can be output simultaneously.
    nn::snd::Voice* m_pVoice[SHORT_SOUND_MAX];          // Voices for sound effects.
    nn::snd::WaveBuffer m_WaveBuffer[SHORT_SOUND_MAX];  // <tt>WaveBuffer</tt> used for sound effects.
    u32 m_CurrentVoiceId;                               // Current position of <tt>Voice&WaveBuffer</tt> used for sound effects.

    nn::os::Thread m_SoundThread;  // Sound thread.
    static const s32 SOUND_THREAD_PRIORITY = 2;
    static const size_t SOUND_THREAD_STACK_SIZE = 4096;
    bool m_SoundThreadFlag;
    NN_PADDING3;

    static void SoundThreadFunc(void* arg);  // Sound thread process-related code.

    nn::os::Thread m_FileReadThread;  // Thread for loading files for stream playback.
    static const s32 FILE_READ_THREAD_PRIORITY = 4;
    static const size_t FILE_READ_THREAD_STACK_SIZE = 4096;
    bool m_FileReadThreadFlag;
    NN_PADDING3;

    static void FileReadThreadFunc(void* arg);  // File load thread processing function.

    static const u32 WAVE_BUFFER_NUM = 4;         // Number of buffers used for stream playback.
    static const size_t WAVE_BUFFER_SIZE = 2048;  // Buffer size used for stream playback.

    // Queue for communication between the sound thread and the file loading thread.
    nn::os::BlockingQueue m_Queue;
    static const size_t QUEUE_BUFFER_SIZE = WAVE_BUFFER_NUM;
    uptr m_QueueBuffer[QUEUE_BUFFER_SIZE];

    const nn::snd::Bcwav::WaveInfo* m_pWaveBGMInfo;
    nn::fs::FileReader m_FileReader;
    size_t m_streamHeaderSize;

    // Get the loop information and sample length that can be obtained with the next read operation.
    s32 GetAvailableSampleLength(s32 size, bool& isLooped);
    s32 m_FilePosition;  // Read position.
    s32 m_CurrentStreamBufferId;
    bool m_bResetOffset;
    NN_PADDING3;

    // Code for storing parsing data obtained from sound files for sound effects.
    bit8** m_pMemory;
    nn::snd::SampleFormat* m_pFormat;
    s32* m_pChannelCount;
    s32* m_pSampleRate;
    s32* m_pSampleLength;

    static const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    bit8 romfsBuffer[ROMFS_BUFFER_SIZE];

    u32 m_CachedPrintFlushSpan;
    static const u32 c_DefaultCahcedPrintFlushSpan = 30;

    int m_FontWidth;
    int m_FontHeight;

#if defined(NN_PIA_INTERNAL_FRAMEWORK_FOR_PIATEST) || defined(NN_PIA_INTERNAL_FRAMEWORK)
    bool m_bToggleWifiSwitch;
#endif
};
}
}
}  // end of namespace nn::pia::sampleframework
#endif //PLATFORM
