﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     AppTrace.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_StringStream.h>
#include <pia/common/common_EmptyStream.h>

#if NN_PIA_UNDER_DEVELOP
#define PIA_SAMPLE_TRACE(FLAG, ...)                                                                                             \
    \
{                                                                                                                        \
        AppTrace* pNnPiaCommonAppTrace = AppTrace::GetInstance();                                                               \
        if (pNnPiaCommonAppTrace)                                                                                               \
        {                                                                                                                       \
            if (pNnPiaCommonAppTrace->GetFlag() & (FLAG))                                                                       \
            {                                                                                                                   \
                pNnPiaCommonAppTrace->Write(__VA_ARGS__);                                                                       \
                pNnPiaCommonAppTrace->Write("\n");                                                                              \
            }                                                                                                                   \
        }                                                                                                                       \
        else                                                                                                                    \
        {                                                                                                                       \
            PIA_SAMPLE_ASSERTMSG(false, "AppTrace instance is not ready.  Did you call nn::pia::common::AppTrace::CreateInstance()?"); \
        }                                                                                                                       \
    \
}

#define PIA_SAMPLE_TRACE_STREAM(FLAG, ARGS)                                  \
    \
{                                                                     \
        AppTrace* pNnPiaCommonAppTrace = AppTrace::GetInstance();            \
        if (pNnPiaCommonAppTrace)                                            \
        {                                                                    \
            if (pNnPiaCommonAppTrace->GetFlag() & (FLAG))                    \
            {                                                                \
                nn::pia::common::StringStream nnPiaCommonStringStream;       \
                (nnPiaCommonStringStream) << ARGS << NN_PIA_T("\n");         \
                pNnPiaCommonAppTrace->Write(nnPiaCommonStringStream.CStr()); \
            }                                                                \
        }                                                                    \
    \
}

#else
#define PIA_SAMPLE_TRACE(FLAG, ...) (void)(FLAG)
#define PIA_SAMPLE_TRACE_STREAM(FLAG, ARGS)         \
    {                                               \
        nn::pia::common::EmptyStream(FLAG) << ARGS; \
    }
#endif


/*!
    @brief  Contains member functions that provide tracing functionality for use by the application.

    Trace code is disabled in release builds.
 */
class AppTrace
{
public:
    /*!
    @brief  Gets the <tt>AppTrace</tt> class instance (singleton pattern).

              This function returns a <tt>NULL</tt> pointer if the instance has not yet been created using the <tt>@ref CreateInstance</tt> function.
              

    @return  Returns a pointer to the <tt>AppTrace</tt> instance.
     */
    static AppTrace* GetInstance(void)
    {
        return &s_Instance;
    }

    /*!
    @brief  Instantiates the object.

           Because the <tt>AppTrace</tt> class uses the singleton pattern, we would normally make the constructor private. In this case, we have other reasons to make it public.
           
     */
    AppTrace(void)
      : m_padding(0), m_flags(0)
    {
    }

    /*!
      :private
    @brief  Destructor.
     */
    virtual ~AppTrace(void)
    {
    }

    // {secret}
    // We do not recommend calling this member function from application code. It is meant to be called from macros.
    // 
    void Write(const char* pFmt, ...);

    /*!
    @brief  Enables the specified trace flags.

    Use this method to add the trace flags you want to enable.

    @param[in] flags  Specifies the bitwise OR of the trace flags to enable.
    */
    void SetFlag(u64 flags)
    {
        m_flags |= flags;
    }

    /*!
    @brief  Disables the specified trace flags.

    @param[in] flags  Specifies the bitwise OR of the trace flags to disable.
    */
    void ClearFlag(u64 flags)
    {
        m_flags = m_flags & (~flags);
    }

    /*!
    @brief  Gets the enabled trace flags.

    @return  Returns the enabled trace flags.
    */
    u64 GetFlag(void) const
    {
        return m_flags;
    }

    /*!
    @brief  Indicates whether the specified flags are set.

    @param[in] flag  Specifies the trace flags to query.

@return  Returns <tt>true</tt> if the specified flags are set, and <tt>false</tt> otherwise.
    */
    bool IsFlagSet(u64 flag) const
    {
        return (m_flags & flag) == flag;
    }

    /*!
    @brief  Disables all trace flags.
     */
    void ClearFlag(void)
    {
        m_flags = 0;
    }


    /*!
@brief  Prints information that is useful for debugging.

@param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref common::TraceFlag</tt> type.
     */
    virtual void Trace(u64 flag) const;

protected:
private:
    /*!
@brief  This copy constructor is private.
     */
    AppTrace(const AppTrace&);

    /*!
@brief  This assignment operator is private.
     */
    AppTrace& operator=(const AppTrace&);

    // Padding.
    u32 m_padding;

    // The object must be constructed with 8-byte alignment, because this member variable is 64 bits.
    // 
    u64 m_flags;

    static AppTrace s_Instance;
};
