﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     ErrorHandling.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "ErrorHandling.h"
#include "SampleLib.h"

#include <pia/common/common_ErrorHandler.h>
#include <pia/common/common_Log.h>

#include <cstdlib>
#include <string.h>

namespace
{
const char* STRING_PIA_MODULE_TYPE[] =
    {
        "Reserved",
        "Pia",
        "PiaCommon",
        "PiaLocal",
        "PiaTransport",
        "PiaSync",
        "PiaInet",
        "PiaSession",
        "NotDefined"
    };

const char* STRING_PIA_ERROR_SEVERITY[] =
    {
        "NO_PROBLEM",
        "PROGRAMMING_ERROR",
        "IGNORABLE",
        "SHOULD_BE_HANDLED",
        "MUST_BE_REINITIALIZED",
        "UNKNOWN"
    };

const char* STRING_PIA_ERROR[] =
    {
        "AllocationFailed",
        "AlreadyInitialized",
        "BrokenData",
        "BufferShortage",
        "Cancelled",
        "ConnectionIsLost",
        "InvalidArgument",
        "InvalidState",
        "NoData",
        "NotFound",
        "NotInpremented",
        "NotInitialized",
        "BufferIsFull",
        "TimeOut"
    };

const char* STRING_PIA_COMMON_ERROR[] =
    {
        "NotInitialized",
        "MemoryLeak",
        "InvalidArgument",
        "AlreadyInitialized",
        "BufferIsFull",
        "InternalError",
        "InvalidState"
    };

const char* STRING_PIA_LOCAL_ERROR[] =
    {
        "NotInitialized",
        "InvalidArgument",
        "AlreadyInitialized",
        "InvalidState",
        "NotFoundNetwork",
        "InvalidNode",
        "NothingPacket",
        "NegligibleFault",
        "InvalidConnection",
        "HostMigrationFailure"
    };

const char* STRING_PIA_TRANSPORT_ERROR[] =
    {
        "DESCRIPTION_SUCCESS",
        "ConnectionFailed",
        "CreateStationFailed",
        "IncompatibleFormat"
    };

const char* STRING_PIA_SYNC_ERROR[] =
    {
        "Base",
        "NotInitialized",
        "InvalidState",
        "InvalidArgument",
        "DataIsNotArrivedYet",
        "OutOfResource",
        "SynchronizationIsLost",
        "FailedToSynchronize",
        "InvalidDevice",
        "ToBeImplemented"
    };

const char* STRING_PIA_UTIL_ERROR[] =
    {
        "none"
    };


#if 0
    inline void CopyString(char8 *szDestination, const char16 *szSource)
    {
        const s32 iLength=std::wcslen(szSource)+1;
        std::wcstombs(szDestination, szSource, iLength);
    }

    inline void CopyString(char16 *szDestination, const char8 *szSource)
    {
        const s32 iLength = std::strlen(szSource)+1;
        std::mbstowcs(szDestination, szSource, iLength);
    }
#else
inline void CopyString(char* szDestination, const char* szSource)
{
    strcpy(szDestination, szSource);
}
#endif

}  // end of anonymous namespace


nn::Result ErrorHandling::s_result = nn::ResultSuccess();
char ErrorHandling::s_filename[FILENAME_MAX_LEN] = {};
u32 ErrorHandling::s_lineNo = 0;

bool ErrorHandling::s_bInitialized = false;

// Initializes the module.
void ErrorHandling::Initialize(void)
{
    if (!s_bInitialized)
    {
        s_bInitialized = true;
    }
}


// End processing.
void ErrorHandling::Finalize(void)
{
    if (s_bInitialized)
    {
        s_bInitialized = false;
    }
}


// Gets whether an error was detected.
bool ErrorHandling::IsErrorDetected(void)
{
    PIA_SAMPLE_ASSERT(s_bInitialized);

    return s_result.IsFailure();
}


nn::Result ErrorHandling::GetResultCode(void)
{
    return s_result;
}


const char* ErrorHandling::GetFilename(void)
{
    return s_filename;
}


u32 ErrorHandling::GetLineNo(void)
{
    return s_lineNo;
}


void ErrorHandling::PrintResult(nn::Result result)
{
    if (nn::pia::common::ErrorHandler::IsPiaError(result))
    {
        NN_PIA_DUMMY_PARAM(STRING_PIA_MODULE_TYPE);
        NN_PIA_DUMMY_PARAM(STRING_PIA_ERROR_SEVERITY);
        NN_PIA_DUMMY_PARAM(STRING_PIA_ERROR);
        NN_PIA_DUMMY_PARAM(STRING_PIA_COMMON_ERROR);
        NN_PIA_DUMMY_PARAM(STRING_PIA_LOCAL_ERROR);
        NN_PIA_DUMMY_PARAM(STRING_PIA_TRANSPORT_ERROR);
        NN_PIA_DUMMY_PARAM(STRING_PIA_SYNC_ERROR);
        NN_PIA_DUMMY_PARAM(STRING_PIA_UTIL_ERROR);
#if !(NN_PIA_UNDER_DEVELOP)
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Value:   0x%08x(Pia)", nn::pia::PiaResultGetValue(result));
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Module:  %02d", nn::pia::common::ErrorHandler::GetModuleTypeFromResult(result));
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %04d", result.GetDescription() & 0x001F);
#else
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Value:   0x%08x(Pia)", nn::pia::PiaResultGetValue(result));
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Module:  %s", STRING_PIA_MODULE_TYPE[nn::pia::common::ErrorHandler::GetModuleTypeFromResult(result)]);
        switch (nn::pia::common::ErrorHandler::GetModuleTypeFromResult(result))
        {
        case nn::pia::MODULE_TYPE_PIA:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        case nn::pia::MODULE_TYPE_COMMON:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_COMMON_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        case nn::pia::MODULE_TYPE_LOCAL:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_LOCAL_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        case nn::pia::MODULE_TYPE_TRANSPORT:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_TRANSPORT_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        case nn::pia::MODULE_TYPE_SYNC:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_SYNC_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        case nn::pia::MODULE_TYPE_UTIL:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Desc: %s", STRING_PIA_UTIL_ERROR[(result.GetDescription() & 0x001F)]);
        }
        break;
        default:
        {
            PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Unknown Error");
        }
        break;
        }
#endif
    }
    else
    {
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Value:   0x%08x(Not Pia)", nn::pia::PiaResultGetValue(result));
#if !NN_PIA_NIN_WIN
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Level:       %4d", result.GetLevel());
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Summary:     %4d", result.GetSummary());
#endif
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Module:      %4d", result.GetModule());
        PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Description: %4d", result.GetDescription());
    }
}
// Output the result code string expression maintained by this class to the debug output.
void ErrorHandling::Print(void)
{
    char buf[FILENAME_MAX_LEN];
    CopyString(buf, GetFilename());
    PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, "\n----- ErrorHandling Start -----");
    PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Filename:    %s", buf);
    PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, " Line:        %d", GetLineNo());

    PrintResult(GetResultCode());
    PIA_LOG(nn::pia::common::Log::LEVEL_ERROR, "----- ErrorHandling End -----\n");
}


// Clears the <tt>nn::Result</tt> type value maintained by this class.
void ErrorHandling::Clear(void)
{
    s_result = nn::ResultSuccess();
    s_filename[0] = '\0';
    s_lineNo = 0;
}


// n1769
// Forcibly sets the result code. This function is provided so that you can intentionally cause an error to verify that your error handling sequence is correct.
// 
void ErrorHandling::SetResultCode(nn::Result result)
{
    s_result = result;
}
