/*---------------------------------------------------------------------------*
  Project: Pia
  File: Indexer.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/types.h>


// The class for converting StationIDs into sequential index values.
template <size_t N>
class Indexer
{
public:
    static const size_t INVALID_INDEX = 0xffffffff;
    static const u64 INVALID_STATION_ID_VALUE = nn::pia::STATION_ID_INVALID_VALUE;
    Indexer()
    {
        for (size_t i = 0; i < N; ++i)
        {
            m_aStationId[i] = nn::pia::StationId(INVALID_STATION_ID_VALUE);
        }
    }

    // Registers a new StationID.
    // When a StationID is registered, the index value corresponding to that StationID can be gotten using the <tt>GetIndex</tt> function.
    bool Add(nn::pia::StationId id)
    {
        PIA_SAMPLE_ASSERT(id != nn::pia::StationId(INVALID_STATION_ID_VALUE));

        for (size_t i = 0; i < N; ++i)
        {
            if (m_aStationId[i] == nn::pia::StationId(INVALID_STATION_ID_VALUE))
            {
                m_aStationId[i] = id;
                return true;
            }
        }

        PIA_SAMPLE_ASSERTMSG(false, "No more space!");
        return false;
    }

    // Deletes a registered StationID.
    bool Remove(nn::pia::StationId id)
    {
        PIA_SAMPLE_ASSERT(id != nn::pia::StationId(INVALID_STATION_ID_VALUE));

        for (size_t i = 0; i < N; ++i)
        {
            if (m_aStationId[i] == id)
            {
                m_aStationId[i] = nn::pia::StationId(INVALID_STATION_ID_VALUE);
                return true;
            }
        }

        PIA_SAMPLE_ASSERTMSG(false, "StationId is not found!");
        return false;
    }

    // Gets the index value from the StationID.
    // If registered correctly, the function returns a value greater than 0 and less than N.
    // If the specified StationID has not been registered by <tt>Add</tt>, the function returns INVALID_INDEX.
    size_t GetIndex(nn::pia::StationId id) const
    {
        for (size_t i = 0; i < N; ++i)
        {
            if (m_aStationId[i] == id)
            {
                return i;
            }
        }

        // Not found.
        return INVALID_INDEX;
    }


    // Gets the StationID from the index value.
    // If not registered, the function returns STATION_ID_INVALID.
    nn::pia::StationId GetStationId(size_t idx)
    {
        PIA_SAMPLE_ASSERTMSG(idx < N, "Indexer::GetStationId() called, but idx is out of range.");

        return m_aStationId[idx];
    }


    // Returns the number of valid indices.
    size_t GetValidIndexNum() const
    {
        size_t sum = 0;
        for (size_t i = 0; i < N; ++i)
        {
            if (m_aStationId[i] != nn::pia::StationId(INVALID_STATION_ID_VALUE))
            {
                ++sum;
            }
        }
        return sum;
    }

private:
    nn::pia::StationId m_aStationId[N];
};
