﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Profiler.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "Profiler.h"
#include "Print.h"
#include "Defines.h"

#include <pia/assert.h>

#include <pia/transport.h>
#include <pia/transport/transport_StationManager.h>
#include <pia/transport/transport_Station.h>
#include <pia/session/session_Mesh.h>
#include <pia/transport/transport_ThreadStreamManager.h>

namespace
{
#if NN_PIA_CAFE
PIA_ATTRIBUTE_ALIGN(4) u8 sStack[256];
#endif
}  // end of anonymous namespace

Profiler* Profiler::s_pInstance = NULL;

void Profiler::CreateInstance()
{
    if (s_pInstance == NULL)
    {
        s_pInstance = new Profiler();
    }
}

void Profiler::DestroyInstance()
{
    if (s_pInstance != NULL)
    {
        delete s_pInstance;
        s_pInstance = NULL;
    }
}


Profiler::Profiler()
  : m_CountThread(),
#if NN_PIA_CTR || NN_PIA_CAFE
    m_Alarm(),
#endif
    m_CurrentProfileCount(0),
    m_IsCounting(false),
    m_IsExitIdleThread(false),
    m_CurrentProfileStartTime(),
#if NN_PIA_CTR || NN_PIA_CAFE
    m_ProfileSpan(0),
#endif
    m_pCurrentProfileResult(NULL)
{
#if NN_PIA_CTR
    m_Alarm.Initialize();
    m_CountThread.Start(RunIdleThread, this, 256, nn::os::LOWEST_THREAD_PRIORITY);
#elif NN_PIA_CAFE
    OSCreateAlarm(&m_Alarm);
    OSSetAlarmUserData(&m_Alarm, this);
    m_CountThread.Start(RunIdleThread, this, sStack, 256, 0, OS_PRIORITY_MAX);
#endif
}

Profiler::~Profiler()
{
    while (m_IsCounting)
    {
#if NN_PIA_CTR
        nn::os::Thread::Sleep(m_ProfileSpan);
#elif NN_PIA_CAFE
        OSSleepTicks(m_ProfileSpan);
#endif
    }

    m_IsExitIdleThread = true;
    while (m_CountThread.IsAlive())
    {
        Thread::Sleep(1);
    }

    m_CountThread.Detach();
    m_CountThread.Finalize();
}


void Profiler::StartProfilingIdle(u32 second)
{
    if (m_IsCounting)
    {
        return;
    }

    StartCount(second, &m_IdleProfileResult);

#if NN_PIA_CTR
    nn::os::Thread::Sleep(m_ProfileSpan);
#elif NN_PIA_CAFE
    OSSleepTicks(m_ProfileSpan);
#endif
    while (m_IsCounting)
    {
        // Just in case, wait.
    }
}

void Profiler::StartProfilingBase(u32 second)
{
    if (m_IsCounting)
    {
        return;
    }

    StartCount(second, &m_BaseProfileResult);
}

void Profiler::StartProfilingTarget(u32 second)
{
    if (m_IsCounting)
    {
        return;
    }

    StartProfilePacketLoss();
    StartCount(second, &m_TargetProfileResult);
    StartPacketProfile();  //n1769
}

void Profiler::FinishProfiling()
{
    StopCount();
    StopProfilePacketLoss();
    StopPacketProfile();  //n1769
    Report();
    ReportPacketLoss();
    ReportPacketProfile();  //n1769
}

void Profiler::Report()
{
    if (m_IsCounting)
    {
        return;
    }

    PIA_SAMPLE_PRINT("-- Profiler --\n");

    u16 stationNum = 0;
    if (nn::pia::session::Mesh::GetInstance() != NULL)
    {
        stationNum = nn::pia::session::Mesh::GetInstance()->GetStationNum();
    }
    PIA_SAMPLE_PRINT("PiaSession StationNum: %d\n", stationNum);

    //    PIA_SAMPLE_PRINT("ProfileSpan : %d ms\n", static_cast<s32>(m_ProfileSpan.GetMilliSeconds()));

    if (m_IdleProfileResult.m_PredictedCount > 0)
    {
        PIA_SAMPLE_PRINT("Idle Profile Result: %d (%d [ms], %d)\n", m_IdleProfileResult.m_PredictedCount, m_IdleProfileResult.m_Period.GetMSec(), m_IdleProfileResult.m_Count);
    }
    else
    {
        PIA_SAMPLE_PRINT("Idle Profile Result: N/A\n");
    }

    if (m_BaseProfileResult.m_PredictedCount > 0)
    {
        PIA_SAMPLE_PRINT("Base Profile Result: %d (%d [ms], %d)\n", m_BaseProfileResult.m_PredictedCount, m_BaseProfileResult.m_Period.GetMSec(), m_BaseProfileResult.m_Count);
    }
    else
    {
        PIA_SAMPLE_PRINT("Base Profile Result: N/A\n");
    }

    if (m_TargetProfileResult.m_PredictedCount > 0)
    {
        PIA_SAMPLE_PRINT("Target Profile Result: %d (%d [ms], %d)\n", m_TargetProfileResult.m_PredictedCount, m_TargetProfileResult.m_Period.GetMSec(), m_TargetProfileResult.m_Count);
    }
    else
    {
        PIA_SAMPLE_PRINT("Target Profile Result: N/A\n");
    }

    if (
            m_IdleProfileResult.m_PredictedCount > 0 && m_BaseProfileResult.m_PredictedCount > 0 && m_TargetProfileResult.m_PredictedCount > 0 && m_BaseProfileResult.m_PredictedCount >= m_TargetProfileResult.m_PredictedCount)
    {
        f32 rate = static_cast<f32>(m_BaseProfileResult.m_PredictedCount - m_TargetProfileResult.m_PredictedCount) / m_IdleProfileResult.m_PredictedCount * 100;
        PIA_SAMPLE_PRINT("Profile Result: %.2f %%\n", rate);
    }
    else
    {
        PIA_SAMPLE_PRINT("Profile Result : N/A\n");
    }

    if (nn::pia::common::CachedPrint::GetInstance() != NULL)
    {
        nn::pia::common::CachedPrint::GetInstance()->Flush();
    }
}

void Profiler::StartCount(u32 second, ProfileResult* pProfileResult)
{
    PIA_SAMPLE_PRINT("Profiler::StartCount:%d\n", second);
    m_pCurrentProfileResult = pProfileResult;
    m_CurrentProfileCount = 0;
    m_CurrentProfileStartTime = nn::pia::common::Time::GetTime();
    m_IsCounting = true;

    if (second == 0)
    {
#if NN_PIA_CTR || NN_PIA_CAFE
        m_ProfileSpan = 0;
#endif
        return;
    }

#if NN_PIA_CTR
    m_ProfileSpan = nn::fnd::TimeSpan::FromSeconds(second);
    m_Alarm.SetOneShot(m_ProfileSpan, AlarmHandler, this);
#elif NN_PIA_CAFE
    m_ProfileSpan = OSSecondsToTicks(second);
    OSSetAlarm(&m_Alarm, m_ProfileSpan, AlarmHandler);
#endif
}

void Profiler::StopCount()
{
    m_IsCounting = false;
    m_pCurrentProfileResult->m_Count = m_CurrentProfileCount;
    nn::pia::common::Time currentTime = nn::pia::common::Time::GetTime();
    m_pCurrentProfileResult->m_Period = currentTime - m_CurrentProfileStartTime;
    m_pCurrentProfileResult->CalcPredictedCount(m_IdleProfileResult.m_Period);
    PIA_SAMPLE_PRINT("Profiler::StopCount:%d\n", m_pCurrentProfileResult->m_Period.GetMSec());

    //    PIA_SAMPLE_PRINT("Profiler::StopCount:%llu, %llu\n", (u64)currentTime, (u64)m_CurrentProfileStartTime);

    m_pCurrentProfileResult = NULL;
}

void Profiler::Count()
{
#if NN_PIA_CTR
    const s32 cTicks = static_cast<s32>(nn::os::Tick::TICKS_PER_SECOND / 1000);
#endif

    while (m_IsExitIdleThread == false)
    {
#if NN_PIA_CTR
        nn::os::ARM::SpinWaitCpuCycles(cTicks);
#elif NN_PIA_CAFE
        OSWaitMicroseconds(1);
//OSSleepMilliseconds( 1 );
#endif
        if (m_IsCounting)
        {
            ++m_CurrentProfileCount;
        }
    }
}


#if NN_PIA_CTR
void Profiler::RunIdleThread(void* pProfiler)
{
    reinterpret_cast<Profiler*>(pProfiler)->Count();
}

void Profiler::AlarmHandler(void* param, bool canceled)
{
    (void)canceled;

    Profiler* pProfiler = reinterpret_cast<Profiler*>(param);
    pProfiler->FinishProfiling();
}
#elif NN_PIA_CAFE
int Profiler::RunIdleThread(int, void* pProfiler)
{
    reinterpret_cast<Profiler*>(pProfiler)->Count();
    return 0;
}

void Profiler::AlarmHandler(OSAlarm* alarm, OSContext* context)
{
    (void)context;

    Profiler* pProfiler = reinterpret_cast<Profiler*>(OSGetAlarmUserData(alarm));
    pProfiler->FinishProfiling();
}
#endif

void Profiler::StartProfilePacketLoss()
{
    nn::pia::transport::StationManager* pStationManager = nn::pia::transport::StationManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pStationManager))
    {
        for (nn::pia::transport::StationManager::Iterator it = pStationManager->Begin(); it != pStationManager->End(); ++it)
        {
            nn::pia::transport::Station* pStation = &(*it);
            if (PIA_IS_VALID_POINTER(pStation))
            {
                pStation->GetSequenceIdController()->StartProfile();
            }
        }
    }
}

void Profiler::StopProfilePacketLoss()
{
    nn::pia::transport::StationManager* pStationManager = nn::pia::transport::StationManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pStationManager))
    {
        for (nn::pia::transport::StationManager::Iterator it = pStationManager->Begin(); it != pStationManager->End(); ++it)
        {
            nn::pia::transport::Station* pStation = &(*it);
            if (PIA_IS_VALID_POINTER(pStation))
            {
                pStation->GetSequenceIdController()->StopProfile();
            }
        }
    }
}

void Profiler::ReportPacketLoss()
{
    nn::pia::transport::StationManager* pStationManager = nn::pia::transport::StationManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pStationManager))
    {
        PIA_SAMPLE_PRINT("-- Packet Loss --\n");
        for (nn::pia::transport::StationManager::Iterator it = pStationManager->Begin(); it != pStationManager->End(); ++it)
        {
            nn::pia::transport::Station* pStation = &(*it);
            if (PIA_IS_VALID_POINTER(pStation))
            {
                u16 totalPacketNum = pStation->GetSequenceIdController()->GetProfiledTotalPacketNum();
                u16 packetLossNum = pStation->GetSequenceIdController()->GetProfiledPacketLossNum();
                f32 rate = (totalPacketNum > 0) ? static_cast<f32>(packetLossNum) * 100.0f / totalPacketNum : 0.0f;
                u16 multicastPacketNum = pStation->GetSequenceIdController()->GetProfiledMulticastPacketNum();
                PIA_SAMPLE_PRINT("StationId:%lld  SendTo(Lost/Total (ratio)): (%3d / %3d (%3.2f%%))  Multicast: %d\n", pStation->GetStationId().GetValue(), packetLossNum, totalPacketNum, rate, multicastPacketNum);
            }
        }
    }
}


static s32 s_ReceivePacket = 0;
static s32 s_SendPacket = 0;

void Profiler::StartPacketProfile()
{
    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pMgr))
    {
        pMgr->ClearReceiveThreadReceivePacketNum();
        pMgr->ClearSendThreadSendPacketNum();
    }
}

void Profiler::StopPacketProfile()
{
    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pMgr))
    {
        s_ReceivePacket = pMgr->GetReceiveThreadReceivePacketNum();
        s_SendPacket = pMgr->GetSendThreadSendPacketNum();
    }
}

/*!
n2908: Get the profiled data.
*/
void Profiler::GetPacketLoss(nn::pia::StationId stationId, u16& totalPacketNum, u16& packetLossNum, u16& multicastPacketNum)
{
    nn::pia::transport::StationManager* pStationManager = nn::pia::transport::StationManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pStationManager))
    {
        nn::pia::transport::Station* pStation = pStationManager->GetStation(stationId);
        if (PIA_IS_VALID_POINTER(pStation))
        {
            totalPacketNum = pStation->GetSequenceIdController()->GetProfiledTotalPacketNum();
            packetLossNum = pStation->GetSequenceIdController()->GetProfiledPacketLossNum();
            multicastPacketNum = pStation->GetSequenceIdController()->GetProfiledMulticastPacketNum();
        }
    }
}

void Profiler::ReportPacketProfile()
{
    PIA_SAMPLE_PRINT("-- ThreadStreamManager --\n");
    PIA_SAMPLE_PRINT("   ReceivePacket: %d\n", s_ReceivePacket);
    PIA_SAMPLE_PRINT("   SendPacket:    %d\n", s_SendPacket);
}


#if 0

void Profiler::SetProfileSpan(const nn::fnd::TimeSpan& span)
{
    PIA_SAMPLE_ASSERTMSG(span > nn::fnd::TimeSpan(0), "span must be plus.");

    if (m_IsCounting)
    {
        return;
    }

    m_ProfileSpan = span;
    m_ProfileCountIdle = -1;
    m_ProfileCountBase = -1;
    m_ProfileCount = -1;
}

#endif
