﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Profiler.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>
#include <pia/common/common_Time.h>
#include "Thread.h"

class ProfileResult
{
public:
    s32 m_Count;
    s32 m_PredictedCount;
    nn::pia::common::TimeSpan m_Period;

    ProfileResult()
    {
        m_Count = 0;
        m_PredictedCount = 0;
        m_Period = 0;
    }

    void CalcPredictedCount(nn::pia::common::TimeSpan timeSpan)
    {
        if (m_Period == 0)
        {
            m_PredictedCount = 0;
        }
        else
        {
            m_PredictedCount = (s32)(m_Count * timeSpan.GetMSecS64() / m_Period.GetMSecS64());
        }
    }
};

class Profiler
{
public:
    /*!
    Creates the singleton instance.

    You must call the nn::os::InitializeAlarmSystem function before calling this function.
     */
    static void CreateInstance();


    /*!
Destroys the singleton instance.
     */
    static void DestroyInstance();


    /*!
Gets the singleton instance.
     */
    static Profiler* GetInstance()
    {
        return s_pInstance;
    }


    /*!
Measures the load when the system is idle. This function returns after it finishes measuring for the specified length of time.
     */
    void StartProfilingIdle(u32 second);


    /*!
Starts profiling when there is no communication ongoing. This function measures from the point it is called until the specified length of time has elapsed.
     */
    void StartProfilingBase(u32 second);


    /*!
Starts profiling when there is communication ongoing. This function measures from the point it is called until the specified length of time has elapsed.
     */
    void StartProfilingTarget(u32 second);

    void FinishProfiling();

    /*!
@brief  Outputs the profiling results to the console.
     */
    void Report();


    /*!
Gets whether measurement is ongoing.

Returns true if measurement is ongoing.
     */
    bool IsCounting() const
    {
        return m_IsCounting;
    }

    /*!
Enables you to get the total number of packets and the number of lost packets so that they can be displayed in the application.
@param[in] stationId  The sending station.
@param[out] totalPacketNum  The number of packets that could have been received.
@param[out] packetLossNum  The number of packets lost.
     */
    void GetPacketLoss(nn::pia::StationId stationId, u16& totalPacketNum, u16& packetLossNum, u16& multicastPacketNum);


#if 0
    /*!
@brief  Sets the length of time to profile. If you do not set a value, the default value of 10 seconds is used.

@param[in] span  Specifies the length of time to profile.
     */
    void SetProfileSpan(const nn::fnd::TimeSpan& span);


    /*!
@brief  Gets the profiling results when the system was idle.

@return  Returns the profiling results from when the system was idle.
     */
    s32 GetIdleCount() const {return m_ProfileCountIdle;}


    /*!
@brief  Gets the results of profiling when no communication is ongoing.

@return  Returns the results of profiling when no communication is ongoing.
     */
    s32 GetStandAloneCount() const {return m_ProfileCountBase;}


    /*!
@brief  Gets the results of profiling when communication is ongoing.

@return  Returns the results of profiling when communication is ongoing.
     */
    s32 GetCommunicationCount() const {return m_ProfileCount;}


    /*!
@brief  Sets the profiling results when the system is idle.

@param[in] count  Specifies the profiling results from when the system was idle.
     */
    void PresetIdleCount(s32 count) {m_ProfileCountIdle = count;}


    /*!
@brief  Sets the results of profiling when no communication is ongoing.

@param[in] count  Specifies the results of profiling when no communication is ongoing.
     */
    void PresetStandAloneCount(s32 count) {m_ProfileCountBase = count;}


    /*!
@brief  Sets the results of profiling when communication is ongoing.

@param[in] count  Specifies the results of profiling when communication is ongoing.
     */
    void PresetCommunicationCount(s32 count) {m_ProfileCount = count;}
#endif

private:
    Profiler();

    ~Profiler();

    void StartCount(u32 second, ProfileResult* pProfileResult);

    void StopCount();

    void Count();


#if NN_PIA_CTR
    static void RunIdleThread(void* pProfiler);
    static void AlarmHandler(void* param, bool canceled);
#elif NN_PIA_CAFE
    static int RunIdleThread(int, void* pProfiler);
    static void AlarmHandler(OSAlarm* alarm, OSContext* context);
#endif

    void StartProfilePacketLoss();

    void StopProfilePacketLoss();

    void ReportPacketLoss();

    void StartPacketProfile();  //n1769

    void StopPacketProfile();  //n1769

    void ReportPacketProfile();  //n1769

    PIA_ATTRIBUTE_ALIGN(8) static Profiler* s_pInstance;

    PIA_ATTRIBUTE_ALIGN(8) Thread m_CountThread;
#if NN_PIA_CTR
    nn::os::Alarm m_Alarm;
#elif NN_PIA_CAFE
    OSAlarm m_Alarm;
#endif

    volatile s32 m_CurrentProfileCount;
    volatile bool m_IsCounting;
    volatile bool m_IsExitIdleThread;

    nn::pia::common::Time m_CurrentProfileStartTime;
#if NN_PIA_CTR
    nn::fnd::TimeSpan m_ProfileSpan;
#elif NN_PIA_CAFE
    OSTime m_ProfileSpan;
#endif

    ProfileResult m_IdleProfileResult;
    ProfileResult m_BaseProfileResult;
    ProfileResult m_TargetProfileResult;
    ProfileResult* m_pCurrentProfileResult;
};
