﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Statistic.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/common/common_TimeSpan.h>
#include "Print.h"
#include <float.h>
#include <limits.h>  // for LLONG_MIN, LLONG_MAX, ULLONG_MAX

#define STATISTIC_DETAIL 1

template <typename T>
class Statistic
{
public:
    Statistic()
    {
        Reset();
    }

    void Reset();

    void Record(T value);

    inline u32 GetCount() const
    {
        return m_Count;
    }

    inline T GetMax() const
    {
        return m_Max;
    }

    inline T GetMin() const
    {
        return m_Min;
    }

    inline T GetAverage() const;

#if STATISTIC_DETAIL
    static const u32 c_MaxDataNum = 8;
    void Print(const char* pCaption, u32 printMadDataNum = c_MaxDataNum);
#else
    void Print(const char* pCaption, u32 printMadDataNum);
#endif

private:
    static inline T MinValue();
    static inline T MaxValue();
    static inline T ZeroValue();
    static inline void PrintValue(const T& v);

private:
    u32 m_Count;
    T m_Sum;
    T m_Max;
    T m_Min;

#if STATISTIC_DETAIL
    struct Data
    {
        u32 m_Idx;
        T m_Value;
    };
    Data m_aMaxData[c_MaxDataNum];
#endif
};

template <typename T>
void Statistic<T>::Reset()
{
    m_Count = 0;
    m_Sum = ZeroValue();
    m_Max = MinValue();
    m_Min = MaxValue();

#if STATISTIC_DETAIL
    for (int i = 0; i < c_MaxDataNum; i++)
    {
        m_aMaxData[i].m_Idx = ULONG_MAX;
        m_aMaxData[i].m_Value = MinValue();
    }
#endif
}

template <typename T>
void Statistic<T>::Record(T value)
{
    m_Sum += value;
    if (value < m_Min)
    {
        m_Min = value;
    }
    if (value > m_Max)
    {
        m_Max = value;
    }
    ++m_Count;

#if STATISTIC_DETAIL
    for (int i = 0; i < c_MaxDataNum; i++)
    {
        if (value > m_aMaxData[i].m_Value)
        {
            for (int j = c_MaxDataNum - 1; j > i; j--)
            {
                m_aMaxData[j].m_Idx = m_aMaxData[j - 1].m_Idx;
                m_aMaxData[j].m_Value = m_aMaxData[j - 1].m_Value;
            }
            m_aMaxData[i].m_Idx = m_Count - 1;
            m_aMaxData[i].m_Value = value;
            break;
        }
    }
#endif
}

template <typename T>
#if STATISTIC_DETAIL
void Statistic<T>::Print(const char* pCaption, u32 printMaxDataNum)
#else
void Statistic<T>::Print(const char* pCaption)
#endif
{
    PIA_SAMPLE_PRINT("%s min:", pCaption);
    PrintValue(GetMin());
    PIA_SAMPLE_PRINT(" max:");
    PrintValue(GetMax());
    PIA_SAMPLE_PRINT(" ave:");
    PrintValue(GetAverage());
    PIA_SAMPLE_PRINT(" cnt:%u\n", GetCount());

#if STATISTIC_DETAIL
    u32 num = (printMaxDataNum < c_MaxDataNum) ? printMaxDataNum : c_MaxDataNum;
    for (u32 i = 0; i < num; ++i)
    {
        if (m_aMaxData[i].m_Idx >= m_Count)
        {
            break;
        }
        PIA_SAMPLE_PRINT("  %2d: %6u : ", i, m_aMaxData[i].m_Idx);
        PrintValue(m_aMaxData[i].m_Value);
        PIA_SAMPLE_PRINT("\n");
    }
#endif
}


template <>
inline s32 Statistic<s32>::MinValue()
{
    return LONG_MIN;
}
template <>
inline s32 Statistic<s32>::MaxValue()
{
    return LONG_MAX;
}
template <>
inline s32 Statistic<s32>::ZeroValue()
{
    return 0;
}
template <>
inline void Statistic<s32>::PrintValue(const s32& v)
{
    PIA_SAMPLE_PRINT("%d", v);
}

template <>
inline u32 Statistic<u32>::MinValue()
{
    return 0;
}
template <>
inline u32 Statistic<u32>::MaxValue()
{
    return ULONG_MAX;
}
template <>
inline u32 Statistic<u32>::ZeroValue()
{
    return 0;
}
template <>
inline void Statistic<u32>::PrintValue(const u32& v)
{
    PIA_SAMPLE_PRINT("%u", v);
}

template <>
inline s64 Statistic<s64>::MinValue()
{
    return LLONG_MIN;
}
template <>
inline s64 Statistic<s64>::MaxValue()
{
    return LLONG_MAX;
}
template <>
inline s64 Statistic<s64>::ZeroValue()
{
    return 0;
}
template <>
inline void Statistic<s64>::PrintValue(const s64& v)
{
    PIA_SAMPLE_PRINT("%lld", v);
}

template <>
inline u64 Statistic<u64>::MinValue()
{
    return 0;
}
template <>
inline u64 Statistic<u64>::MaxValue()
{
    return ULLONG_MAX;
}
template <>
inline u64 Statistic<u64>::ZeroValue()
{
    return 0;
}
template <>
inline void Statistic<u64>::PrintValue(const u64& v)
{
    PIA_SAMPLE_PRINT("%llu", v);
}

template <>
inline f32 Statistic<f32>::MinValue()
{
    return FLT_MIN;
}
template <>
inline f32 Statistic<f32>::MaxValue()
{
    return FLT_MAX;
}
template <>
inline f32 Statistic<f32>::ZeroValue()
{
    return 0.0f;
}
template <>
inline void Statistic<f32>::PrintValue(const f32& v)
{
    PIA_SAMPLE_PRINT("%f", v);
}

template <>
inline f64 Statistic<f64>::MinValue()
{
    return DBL_MIN;
}
template <>
inline f64 Statistic<f64>::MaxValue()
{
    return DBL_MAX;
}
template <>
inline f64 Statistic<f64>::ZeroValue()
{
    return 0.0;
}
template <>
inline void Statistic<f64>::PrintValue(const f64& v)
{
    PIA_SAMPLE_PRINT("%f", v);
}

template <>
inline nn::pia::common::TimeSpan Statistic<nn::pia::common::TimeSpan>::MinValue()
{
    nn::pia::common::TimeSpan t;
    t.SetTick(LLONG_MIN);
    return t;
}
template <>
inline nn::pia::common::TimeSpan Statistic<nn::pia::common::TimeSpan>::MaxValue()
{
    nn::pia::common::TimeSpan t;
    t.SetTick(LLONG_MAX);
    return t;
}
template <>
inline nn::pia::common::TimeSpan Statistic<nn::pia::common::TimeSpan>::ZeroValue()
{
    return nn::pia::common::TimeSpan(0);
}
template <>
inline void Statistic<nn::pia::common::TimeSpan>::PrintValue(const nn::pia::common::TimeSpan& v)
{
    PIA_SAMPLE_PRINT("%fms", v.GetSecF32() * 1000.0f);
}

template <typename T>
inline T Statistic<T>::GetAverage() const
{
    return (m_Count > 0) ? (m_Sum / static_cast<T>(m_Count)) : ZeroValue();
}

template <>
inline nn::pia::common::TimeSpan Statistic<nn::pia::common::TimeSpan>::GetAverage() const
{
    nn::pia::common::TimeSpan t;
    t.SetTick((m_Count > 0) ? (m_Sum.GetTick() / m_Count) : 0);
    return t;
}
