﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     MenuConfig.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "MenuConfig.h"

#include <pia/session.h>

#include "Types.h"
#include "Config.h"
#include "Misc.h"
#include "Global.h"
#include "Sndmic.h"

#include "SampleLib/SampleLib.h"

#include "Shared_Chat_Setup.h"

// Macro prepared for items which can be chosen to be either enabled or disabled.
#define DEFINE_NEXT_PREV_FUNC(enableApi)           \
    \
virtual void Next()                                \
    {                                              \
        Chat_GetVoiceProtocol()->enableApi(true);  \
    }                                              \
    \
virtual void Prev()                                \
    {                                              \
        Chat_GetVoiceProtocol()->enableApi(false); \
    }

// Macro prepared for items which can be chosen to be either enabled or disabled.
#define DEFINE_DRAW_ENABLE_DISABLE(title, isEnableApi)                                                           \
    \
virtual void Draw(s32 x, s32 y) const                                                                            \
    \
{                                                                                                         \
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance(); \
        if (PIA_IS_VALID_POINTER(f))                                                                             \
        {                                                                                                        \
            bool bEnabled = Chat_GetVoiceProtocol()->isEnableApi();                                              \
            f->DrawText(x, y, "%-16s : %s", title, ENABLE_DISABLE_TEXT[bEnabled]);                               \
        }                                                                                                        \
    \
}


// Macro for creating <tt>Next</tt> and <tt>Prev</tt> functions for setting and getting numerical values that change at set intervals.
#define DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(varType, setterApi, getterApi, inc, maxValue, minValue)                \
    \
virtual void Next()                                                                                                \
    \
{                                                                                                           \
        varType var = clamp(static_cast<varType>(Chat_GetVoiceProtocol()->getterApi() + inc), minValue, maxValue); \
        Chat_GetVoiceProtocol()->setterApi(var);                                                                   \
    \
}                                                                                                           \
    \
virtual void Prev()                                                                                                \
    \
{                                                                                                           \
        varType var = clamp(static_cast<varType>(Chat_GetVoiceProtocol()->getterApi() - inc), minValue, maxValue); \
        Chat_GetVoiceProtocol()->setterApi(var);                                                                   \
    \
}

#define DEFINE_GETTER_DRAW_FUNC(fmt, getterApi)                                                                  \
    \
virtual void Draw(s32 x, s32 y) const                                                                            \
    \
{                                                                                                         \
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance(); \
        if (PIA_IS_VALID_POINTER(f))                                                                             \
        {                                                                                                        \
            f->DrawText(x, y, fmt, Chat_GetVoiceProtocol()->getterApi());                                        \
        }                                                                                                        \
    \
}


namespace
{
#if (NN_PIA_CTR && USE_PRIVATE_CHAT_API) || NN_PIA_CAFE
const char* ENABLE_DISABLE_TEXT[] = { "Disable", "Enable" };
#endif
const char* CODEC_TEXT[] = { "8bit raw", "G711 u-Law", "2bit ADPCM", "3bit ADPCM", "4bit ADPCM" };


class ItemCodec : public ItemBase
{
public:
    ItemCodec()
    {
    }
    virtual ~ItemCodec()
    {
    }

    virtual void Next()
    {
        Chat_GetVoiceProtocol()->SetCodec(nn::pia::chat::CODEC_4BIT_ADPCM);
    }

    virtual void Prev()
    {
        Chat_GetVoiceProtocol()->SetCodec(nn::pia::chat::CODEC_2BIT_ADPCM);
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            enum nn::pia::chat::Codec codec = Chat_GetVoiceProtocol()->GetCodec();
            f->DrawText(x, y, "%-16s : %s", "Codec", CODEC_TEXT[codec]);
        }
    }

    virtual void Commit()
    {
    }
};


#if USE_PRIVATE_CHAT_API
class ItemSparePacket : public ItemBase
{
public:
    ItemSparePacket()
    {
    }
    virtual ~ItemSparePacket()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableSparePacket);
    DEFINE_DRAW_ENABLE_DISABLE("Spare Packet", IsEnableSparePacket);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API


#if USE_PRIVATE_CHAT_API
class ItemSpareCodec : public ItemBase
{
public:
    ItemSpareCodec()
    {
    }
    virtual ~ItemSpareCodec()
    {
    }

    virtual void Next()
    {
        Chat_GetVoiceProtocol()->SetSpareCodec(nn::pia::chat::CODEC_4BIT_ADPCM);
    }

    virtual void Prev()
    {
        Chat_GetVoiceProtocol()->SetSpareCodec(nn::pia::chat::CODEC_2BIT_ADPCM);
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            enum nn::pia::chat::Codec codec = Chat_GetVoiceProtocol()->GetSpareCodec();
            f->DrawText(x, y, "%-16s : %s", "Spare Codec", CODEC_TEXT[codec]);
        }
    }

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API


#if USE_PRIVATE_CHAT_API
class ItemDropRecover : public ItemBase
{
public:
    ItemDropRecover()
    {
    }
    virtual ~ItemDropRecover()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableDropRecover);
    DEFINE_DRAW_ENABLE_DISABLE("Drop Recover", IsEnableDropRecover);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API


#if USE_PRIVATE_CHAT_API
class ItemHighPassFilter : public ItemBase
{
public:
    ItemHighPassFilter()
    {
    }
    virtual ~ItemHighPassFilter()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableHPF);
    DEFINE_DRAW_ENABLE_DISABLE("High-Pass filter", IsEnableHPF);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API


#if NN_PIA_CTR
#if USE_PRIVATE_CHAT_API
class ItemBiquadFilter : public ItemBase
{
public:
    ItemBiquadFilter()
    {
    }
    virtual ~ItemBiquadFilter()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableBPF);
    DEFINE_DRAW_ENABLE_DISABLE("Biquad filter", IsEnableBPF);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API


#if USE_PRIVATE_CHAT_API
class ItemEchoCanceller : public ItemBase
{
public:
    ItemEchoCanceller()
    {
    }
    virtual ~ItemEchoCanceller()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableEchoCancel);
    DEFINE_DRAW_ENABLE_DISABLE("Echo Canceller", IsEnableEchoCancel);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API

#endif  // end of NN_PIA_CTR


#if USE_PRIVATE_CHAT_API
class ItemVad : public ItemBase
{
public:
    ItemVad()
    {
    }
    virtual ~ItemVad()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableVAD);
    DEFINE_DRAW_ENABLE_DISABLE("VAD", IsEnableVAD);

    virtual void Commit()
    {
    }
};
#endif  // end of USE_PRIVATE_CHAT_API

class ItemActiveGain : public ItemBase
{
public:
    static const u16 GAIN_MAX = 990;
    static const u16 GAIN_MIN = 10;

    ItemActiveGain()
    {
    }
    virtual ~ItemActiveGain()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(u16, SetVADActiveGain, GetVADActiveGain, 10, GAIN_MAX, GAIN_MIN);
    DEFINE_GETTER_DRAW_FUNC("Active Gain      : %d", GetVADActiveGain);

    virtual void Commit()
    {
    }
};


class ItemInactiveGain : public ItemBase
{
public:
    static const u16 GAIN_MAX = 990;
    static const u16 GAIN_MIN = 10;

    ItemInactiveGain()
    {
    }
    virtual ~ItemInactiveGain()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(u16, SetVADInactiveGain, GetVADInactiveGain, 10, GAIN_MAX, GAIN_MIN);
    DEFINE_GETTER_DRAW_FUNC("Inactive Gain    : %d", GetVADInactiveGain);

    virtual void Commit()
    {
    }
};


class ItemClampGain : public ItemBase
{
public:
    static const u32 GAIN_MAX = nn::pia::chat::VoiceProtocol::VAD_CLAMP_GAIN_MAX;
    static const u32 GAIN_MIN = 0;

    ItemClampGain()
    {
    }
    virtual ~ItemClampGain()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(u32, SetVADClampGain, GetVADClampGain, 1, GAIN_MAX, GAIN_MIN);
    DEFINE_GETTER_DRAW_FUNC("Clamp Gain       : %d", GetVADClampGain);

    virtual void Commit()
    {
    }
};


class ItemReleaseTime : public ItemBase
{
public:
    static const u32 PACKET_MAX = 10000;
    static const u32 PACKET_MIN = 0;
    ItemReleaseTime()
    {
    }
    virtual ~ItemReleaseTime()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(u32, SetVADReleaseTime, GetVADReleaseTime, 1, PACKET_MAX, PACKET_MIN);
    DEFINE_GETTER_DRAW_FUNC("Release Time     : %d", GetVADReleaseTime);

    virtual void Commit()
    {
    }
};


class ItemMicGain : public ItemBase
{
public:
#if NN_PIA_CTR
    ItemMicGain(SndMicDevice* pSndMicDevice)
      : m_Gain(GAIN_DEFAULT), m_pSndMicDevice(pSndMicDevice)
    {
    }
#elif NN_PIA_CAFE
    ItemMicGain(SndMicDevice* pSndMicDevice)
      : m_pSndMicDevice(pSndMicDevice)
    {
    }
#endif
    virtual ~ItemMicGain()
    {
    }

#if NN_PIA_CTR
    virtual void Next()
    {
        m_Gain = clamp(static_cast<u8>(m_Gain + GAIN_PITCH), GAIN_MIN, GAIN_MAX);
        m_pSndMicDevice->MicGainSet(m_Gain);
    }

    virtual void Prev()
    {
        m_Gain = clamp(static_cast<u8>(m_Gain - GAIN_PITCH), GAIN_MIN, GAIN_MAX);
        m_pSndMicDevice->MicGainSet(m_Gain);
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "%-16s : %d", "Mic Gain", m_Gain);
        }
    }
#elif NN_PIA_CAFE
    virtual void Next()
    {
        u32 gain = 0;
        m_pSndMicDevice->GetGainLevelMic(&gain);
        if (gain <= GAIN_MAX - GAIN_PITCH)
        {
            m_pSndMicDevice->SetGainLevelMic(gain + GAIN_PITCH, 0);
        }
    }

    virtual void Prev()
    {
        u32 gain = 0;
        m_pSndMicDevice->GetGainLevelMic(&gain);
        if (gain >= GAIN_MIN + GAIN_PITCH)
        {
            m_pSndMicDevice->SetGainLevelMic(gain - GAIN_PITCH, 0);
        }
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            u32 gain = 0;
            m_pSndMicDevice->GetGainLevelMic(&gain);
            f->DrawText(x, y, "%-16s : %d", "Mic Gain", gain);
        }
    }
#endif

    virtual void Commit()
    {
    }

private:
#if NN_PIA_CTR
    static const u8 GAIN_DEFAULT = 60;
    static const u8 GAIN_MIN = 31;
    static const u8 GAIN_MAX = 100;
    static const u8 GAIN_PITCH = 1;
    u8 m_Gain;
#elif NN_PIA_CAFE
    static const u32 GAIN_MIN = 0;
    static const u32 GAIN_MAX = 6400;
    static const u32 GAIN_PITCH = 128;
#endif
    SndMicDevice* m_pSndMicDevice;
};


#if NN_PIA_CTR
#define VOL_DEFAULT 1.0f
#define VOL_MIN 0.0f
#define VOL_MAX 20.0f
#define VOL_PITCH 0.1f
#elif NN_PIA_CAFE
#define VOL_DEFAULT static_cast<s32>((SPEAKER_GAIN_IDX_MAX - 1))
#define VOL_MIN static_cast<s32>(0)
#define VOL_MAX static_cast<s32>((SPEAKER_GAIN_IDX_MAX - 1))
#define VOL_PITCH static_cast<s32>(1)
#endif
class ItemSpeakerVolume : public ItemBase
{
public:
    ItemSpeakerVolume(SndMicDevice* pSndMicDevice)
      : m_Volume(VOL_DEFAULT), m_pSndMicDevice(pSndMicDevice)
    {
    }
    virtual ~ItemSpeakerVolume()
    {
    }

    virtual void Next()
    {
        m_Volume = clamp(m_Volume + VOL_PITCH, VOL_MIN, VOL_MAX);
#if NN_PIA_CTR
        m_pSndMicDevice->SetVolume(m_Volume);
#elif NN_PIA_CAFE
        m_pSndMicDevice->SetDRCVolume(static_cast<u32>(m_Volume));
#endif
    }

    virtual void Prev()
    {
        m_Volume = clamp(m_Volume - VOL_PITCH, VOL_MIN, VOL_MAX);
#if NN_PIA_CTR
        m_pSndMicDevice->SetVolume(m_Volume);
#elif NN_PIA_CAFE
        m_pSndMicDevice->SetDRCVolume(static_cast<u32>(m_Volume));
#endif
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
#if NN_PIA_CTR
            f->DrawText(x, y, "%-16s : %1.1f", "Speaker Volume", m_Volume);
#elif NN_PIA_CAFE
            f->DrawText(x, y, "%-16s : %d", "Speaker Volume", m_Volume);
#endif
        }
    }

    virtual void Commit()
    {
    }

private:
#if NN_PIA_CTR
    f32 m_Volume;
#elif NN_PIA_CAFE
    s32 m_Volume;
#endif
    SndMicDevice* m_pSndMicDevice;
};
#undef VOL_DEFAULT
#undef VOL_MIN
#undef VOL_MAX
#undef VOL_PITCH


#if NN_PIA_CTR
#define VOL_DEFAULT 0.2f
#define VOL_MIN 0.0f
#define VOL_MAX 10.0f
#define VOL_PITCH 0.1f
class ItemBgmVolume : public ItemBase
{
public:
    ItemBgmVolume(SndMicDevice* pSndMicDevice)
      : m_Volume(VOL_DEFAULT), m_pSndMicDevice(pSndMicDevice)
    {
    }
    virtual ~ItemBgmVolume()
    {
    }

    virtual void Next()
    {
        m_Volume = clamp(m_Volume + VOL_PITCH, VOL_MIN, VOL_MAX);
        m_pSndMicDevice->SetBGMVolume(m_Volume);
    }

    virtual void Prev()
    {
        m_Volume = clamp(m_Volume - VOL_PITCH, VOL_MIN, VOL_MAX);
        m_pSndMicDevice->SetBGMVolume(m_Volume);
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "%-16s : %1.1f", "BGM Volume", m_Volume);
        }
    }

    virtual void Commit()
    {
    }

private:
    f32 m_Volume;
    SndMicDevice* m_pSndMicDevice;
};
#undef VOL_DEFAULT
#undef VOL_MIN
#undef VOL_MAX
#undef VOL_PITCH
#elif NN_PIA_CAFE
class ItemBgmVolume : public ItemBase
{
public:
    ItemBgmVolume(SndMicDevice* pSndMicDevice)
      : m_pSndMicDevice(pSndMicDevice)
    {
    }
    virtual ~ItemBgmVolume()
    {
    }

    virtual void Next()
    {
    }
    virtual void Prev()
    {
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "%-16s : (not available)", "BGM Volume");
        }
    }

    virtual void Commit()
    {
    }
    SndMicDevice* m_pSndMicDevice;
};
#endif


class ItemNoiseSkipTime : public ItemBase
{
public:
    static const u32 MSEC_MAX = 1000;
    static const u32 MSEC_MIN = 0;

    ItemNoiseSkipTime()
    {
    }
    virtual ~ItemNoiseSkipTime()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(u32, SetNoiseSkipTime, GetNoiseSkipTime, 100, MSEC_MAX, MSEC_MIN);
    DEFINE_GETTER_DRAW_FUNC("Noise Skip Time  : %d", GetNoiseSkipTime);

    virtual void Commit()
    {
    }
};


#if NN_PIA_CAFE
class ItemAgc : public ItemBase
{
public:
    ItemAgc()
    {
    }
    virtual ~ItemAgc()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableAgc);
    DEFINE_DRAW_ENABLE_DISABLE("AGC", IsEnableAgc);

    virtual void Commit()
    {
    }
};


class ItemNoiseSuppressor : public ItemBase
{
public:
    ItemNoiseSuppressor()
    {
    }
    virtual ~ItemNoiseSuppressor()
    {
    }

    DEFINE_NEXT_PREV_FUNC(EnableNoiseSuppressor);
    DEFINE_DRAW_ENABLE_DISABLE("NoiseSuppressor", IsEnableNoiseSuppressor);

    virtual void Commit()
    {
    }
};


class ItemAgcLevel : public ItemBase
{
public:
    static const s32 AGC_MAX = nn::pia::chat::VoiceProtocol::AGC_LEVEL_MAX;
    static const s32 AGC_MIN = nn::pia::chat::VoiceProtocol::AGC_LEVEL_MIN;

    ItemAgcLevel()
    {
    }
    virtual ~ItemAgcLevel()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(s32, SetAgcLevel, GetAgcLevel, 1, AGC_MAX, AGC_MIN);
    DEFINE_GETTER_DRAW_FUNC("AGC Level        : %d", GetAgcLevel);

    virtual void Commit()
    {
    }
};


class ItemNoiseSuppressorLevel : public ItemBase
{
public:
    static const s32 NS_MAX = nn::pia::chat::VoiceProtocol::NOISE_SUPPRESSOR_LEVEL_MAX;
    static const s32 NS_MIN = nn::pia::chat::VoiceProtocol::NOISE_SUPPRESSOR_LEVEL_MIN;

    ItemNoiseSuppressorLevel()
    {
    }
    virtual ~ItemNoiseSuppressorLevel()
    {
    }

    DEFINE_SETTER_GETTER_NEXT_PREV_FUNC(s32, SetNoiseSuppressorLevel, GetNoiseSuppressorLevel, 1, NS_MAX, NS_MIN);
    DEFINE_GETTER_DRAW_FUNC("NS  Level        : %d", GetNoiseSuppressorLevel);

    virtual void Commit()
    {
    }
};


class ItemEchoCanceller : public ItemBase
{
public:
    ItemEchoCanceller()
      : m_Mode(0)
    {
    }
    virtual ~ItemEchoCanceller()
    {
    }

    virtual void Next()
    {
        changeMode(1);
    }

    virtual void Prev()
    {
        changeMode(-1);
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            if (Chat_GetVoiceProtocol()->IsEnableEchoCanceller())
            {
                static const char* TEXT[] = { "PERFORMANCE", "STANDARD", "SOUNDQUALITY" };
                f->DrawText(x, y, "%-16s : %s", "EC mode", TEXT[Chat_GetVoiceProtocol()->GetEchoCancellerMode()]);
            }
            else
            {
                f->DrawText(x, y, "%-16s : Disabled", "EC mode");
            }
        }
    }

    virtual void Commit()
    {
    }

private:
    s32 m_Mode;  // Associate <tt>-1</tt> with echo cancellation "Disabled."

    void changeMode(int inc)
    {
        const s32 MIN = static_cast<s32>(nn::pia::chat::EC_PERFORMANCE);
        const s32 MAX = static_cast<s32>(nn::pia::chat::EC_SOUNDQUALITY);
        bool bChangeFlag = false;
        bool bEnabled = Chat_GetVoiceProtocol()->IsEnableEchoCanceller();
        s32 mode = bEnabled ? static_cast<s32>(Chat_GetVoiceProtocol()->GetEchoCancellerMode()) : -1;  // -1 is for use with Disable
        if (inc == -1)
        {
            if (mode >= MIN)
            {
                --mode;
                bChangeFlag = true;
            }
        }
        else if (inc == 1)
        {
            if (mode < MAX)
            {
                ++mode;
                bChangeFlag = true;
            }
        }
        else
        {
            PIA_SAMPLE_ASSERT(false);
            return;
        }

        if (bChangeFlag)
        {
            if (mode == -1)
            {
                Chat_GetVoiceProtocol()->EnableEchoCanceller(false);
            }
            else
            {
                Chat_GetVoiceProtocol()->EnableEchoCanceller(true);
                Chat_GetVoiceProtocol()->SetEchoCancellerMode(static_cast<nn::pia::chat::EcMode>(mode));
            }
        }
    }
};
#endif  // end of NN_PIA_CAFE


class ItemConferenceModeMenu : public ItemBase
{
public:
    ItemConferenceModeMenu()
    {
    }
    virtual ~ItemConferenceModeMenu()
    {
    }

    virtual void Next()
    {
    }
    virtual void Prev()
    {
    }
    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "Return...");
        }
    }
    virtual void Commit()
    {
    }
    virtual void Exec()
    {
        *static_cast<bool*>(GetUserData()) = true;
    }
};

}  // end of anonymous namespace


MenuConfig::MenuConfig(SndMicDevice* pSndMicDevice)
  : m_bReadyToReturnConference(false)
{
    PIA_SAMPLE_ASSERT(pSndMicDevice != NULL);

    // Fills the pointer array with NULL.
    for (size_t i = 0; i < MAX_ITEMS; ++i)
    {
        m_paItem[i] = NULL;
    }

    // Registers the item object in the menu.
    size_t idx = 0;
    m_paItem[idx++] = new ItemCodec();
#if USE_PRIVATE_CHAT_API
    m_paItem[idx++] = new ItemSparePacket();
    m_paItem[idx++] = new ItemSpareCodec();
    m_paItem[idx++] = new ItemDropRecover();
    m_paItem[idx++] = new ItemHighPassFilter();
#endif
#if NN_PIA_CTR
#if USE_PRIVATE_CHAT_API
    m_paItem[idx++] = new ItemBiquadFilter();
    m_paItem[idx++] = new ItemEchoCanceller();
#endif
#endif
#if USE_PRIVATE_CHAT_API
    m_paItem[idx++] = new ItemVad();
#endif
    m_paItem[idx++] = new ItemActiveGain();
    m_paItem[idx++] = new ItemInactiveGain();
    m_paItem[idx++] = new ItemClampGain();
    m_paItem[idx++] = new ItemReleaseTime();
    m_paItem[idx++] = new ItemMicGain(pSndMicDevice);
    m_paItem[idx++] = new ItemSpeakerVolume(pSndMicDevice);
    m_paItem[idx++] = new ItemBgmVolume(pSndMicDevice);
    m_paItem[idx++] = new ItemNoiseSkipTime();
#if NN_PIA_CAFE
    m_paItem[idx++] = new ItemAgc();
    m_paItem[idx++] = new ItemNoiseSuppressor();
    m_paItem[idx++] = new ItemAgcLevel();
    m_paItem[idx++] = new ItemNoiseSuppressorLevel();
    m_paItem[idx++] = new ItemEchoCanceller();
#endif

    // Last.
    m_paItem[idx] = new ItemConferenceModeMenu;
    ;
    m_paItem[idx]->SetUserData(&m_bReadyToReturnConference);

    for (size_t i = 0; i < MAX_ITEMS; ++i)
    {
        if (m_paItem[i] != NULL)
        {
            m_MenuImpl.Register(m_paItem[i]);
        }
        else
        {
            break;
        }
    }
}


MenuConfig::~MenuConfig()
{
    // Clean up the constructed item objects.
    for (int i = MAX_ITEMS - 1; i >= 0; --i)
    {
        if (m_paItem[i] != NULL)
        {
            delete m_paItem[i];
            m_paItem[i] = NULL;
        }
    }
}


void MenuConfig::Dispatch()
{
    m_MenuImpl.Dispatch();
}


void MenuConfig::Draw() const
{
    m_MenuImpl.Draw(16, 50);
}
