﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     MenuParticipants.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "MenuParticipants.h"

#include <pia/session.h>

#include "Types.h"
#include "Config.h"
#include "Misc.h"
#include "Global.h"

#include "SampleLib/SampleLib.h"

#include "Shared_Chat_Setup.h"


namespace
{


nn::pia::StationIndex ConvertToStationIndex(nn::pia::StationId id)
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    if (!PIA_IS_VALID_POINTER(pTransport))
    {
        return nn::pia::STATION_INDEX_INVALID;
    }

    nn::pia::StationIndex idx;
    nn::Result ret = pTransport->ConvertToStationIndex(&idx, id);
    if (ret.IsFailure())
    {
        return nn::pia::STATION_INDEX_INVALID;
    }

    return idx;
}


class ItemParticipant : public ItemBase
{
public:
    ItemParticipant(size_t idx)
      : m_Idx(idx)
    {
    }
    virtual ~ItemParticipant()
    {
    }

    virtual void Next()
    {
    }
    virtual void Prev()
    {
    }

    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            static const char* STR[] = { "LEAVE", "PARTICIPATE" };

            nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

            nn::pia::StationId id = g_Indexer.GetStationId(m_Idx);
            const char* pMsg;
            if (pSession->CheckStationIdIsValid(id))
            {
                pMsg = STR[Chat_GetVoiceProtocol()->IsClientInConference(id)];
            }
            else
            {
                pMsg = STR[0];
            }

            f->DrawText(x, y, "Member[ID:0x%04x] %s", ConvertToStationIndex(id), pMsg);
        }
    }

    virtual void Commit()
    {
    }

    virtual void Exec()
    {
        nn::pia::chat::VoiceProtocol* pVoice = Chat_GetVoiceProtocol();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pVoice));

        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

        nn::pia::StationId id = g_Indexer.GetStationId(m_Idx);
        if (pSession->CheckStationIdIsValid(id))
        {
            if (pVoice->IsClientInConference(id))
            {
                pVoice->RemoveConferenceClient(id);
            }
            else
            {
                pVoice->AddConferenceClient(id);
            }
        }
    }

private:
    size_t m_Idx;
};


class ItemConnectedModeMenu : public ItemBase
{
public:
    ItemConnectedModeMenu()
    {
    }
    virtual ~ItemConnectedModeMenu()
    {
    }

    virtual void Next()
    {
    }
    virtual void Prev()
    {
    }
    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "Stop Conference...");
        }
    }
    virtual void Commit()
    {
    }
    virtual void Exec()
    {
        *static_cast<bool*>(GetUserData()) = true;
    }
};


class ItemChangeSetting : public ItemBase
{
public:
    ItemChangeSetting()
    {
    }
    virtual ~ItemChangeSetting()
    {
    }

    virtual void Next()
    {
    }
    virtual void Prev()
    {
    }
    virtual void Draw(s32 x, s32 y) const
    {
        nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
        if (PIA_IS_VALID_POINTER(f))
        {
            f->DrawText(x, y, "Change Setting...");
        }
    }
    virtual void Commit()
    {
    }
    virtual void Exec()
    {
        *static_cast<bool*>(GetUserData()) = true;
    }
};


}  // end of anonymous namespace


MenuParticipants::MenuParticipants()
  : m_bReadyToReturnConnected(false),
    m_bReadyToGoChangeSetting(false)
{
    // Fills the pointer array with NULL.
    for (size_t i = 0; i < MAX_ITEMS; ++i)
    {
        m_paItem[i] = NULL;
    }

    // Registers the item object in the menu.
    for (size_t i = 0; i < MAX_CLIENTS; ++i)
    {
        m_paItem[i] = new ItemParticipant(i);
    }
    m_paItem[MAX_CLIENTS] = new ItemChangeSetting;
    m_paItem[MAX_CLIENTS]->SetUserData(&m_bReadyToGoChangeSetting);
    m_paItem[MAX_CLIENTS + 1] = new ItemConnectedModeMenu;
    ;
    m_paItem[MAX_CLIENTS + 1]->SetUserData(&m_bReadyToReturnConnected);

    for (size_t i = 0; i < MAX_ITEMS; ++i)
    {
        if (m_paItem[i] != NULL)
        {
            m_MenuImpl.Register(m_paItem[i]);
        }
        else
        {
            break;
        }
    }
}


MenuParticipants::~MenuParticipants()
{
    // Clean up the constructed item objects.
    for (int i = MAX_ITEMS - 1; i >= 0; --i)
    {
        if (m_paItem[i] != NULL)
        {
            delete m_paItem[i];
            m_paItem[i] = NULL;
        }
    }
}


void MenuParticipants::Dispatch()
{
    m_MenuImpl.Dispatch();
}


void MenuParticipants::Draw() const
{
    m_MenuImpl.Draw(16, 54);
}
