﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Misc.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "Misc.h"

#pragma diag_suppress 1300
#include <nex.h>
#pragma diag_warning 1300

#include <pia/session.h>


nn::pia::sampleframework::SampleFramework* getSampleFrameworkInstance()
{
    return nn::pia::sampleframework::SampleFramework::GetInstance();
}


void beginFrame(bool bNexDispatch)
{
    if (getSampleFrameworkInstance())
    {
        getSampleFrameworkInstance()->BeginFrame();
        getSampleFrameworkInstance()->SwapBuffers();
    }

    // Dispatching is also required when nn::nex::Scheduler is created.
    if (PIA_IS_VALID_POINTER(nn::nex::Scheduler::GetInstance()) && bNexDispatch)
    {
        nn::nex::Scheduler::GetInstance()->DispatchAll();
    }

    // The job's dispatch process (execution of job inserted by JoinMesh).
    if (PIA_IS_VALID_POINTER(nn::pia::common::Scheduler::GetInstance()))
    {
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
    }
}


void endFrame()
{
    if (getSampleFrameworkInstance())
    {
        getSampleFrameworkInstance()->EndFrame();
    }
}


void changeRenderTarget(bool bUpper)
{
    if (getSampleFrameworkInstance())
    {
        getSampleFrameworkInstance()->ChangeRenderTarget(bUpper);
    }
}


void drawText(s32 x, s32 y, const char* pFmt, ...)
{
    if (getSampleFrameworkInstance())
    {
        const size_t BUFSIZE = 1024;
        static char textBuf[BUFSIZE];

        va_list va;
        va_start(va, pFmt);
        vsnprintf(textBuf, BUFSIZE, pFmt, va);
        va_end(va);

        getSampleFrameworkInstance()->DrawText(x, y, textBuf);
    }
}


void setClearColor(f32 r, f32 g, f32 b)
{
    if (getSampleFrameworkInstance())
    {
        getSampleFrameworkInstance()->SetClearColor(nn::pia::sampleframework::Color(r, g, b, 1.0f));
    }
}


void swapBuffers()
{
    if (getSampleFrameworkInstance())
    {
        getSampleFrameworkInstance()->SwapBuffers();
    }
}


bool isCommander()
{
    nn::pia::session::Mesh* pMesh = nn::pia::session::Mesh::GetInstance();
    if (PIA_IS_VALID_POINTER(pMesh))
    {
        return pMesh->IsHost();
    }
    else
    {
        return false;
    }
}


bool isConnectedSession()
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    if (PIA_IS_VALID_POINTER(pSession))
    {
        return (pSession->CheckConnectionError().IsSuccess() &&
                pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION);
    }
    else
    {
        PIA_BASIC_REPORT("Session instance is not ready.\n");
        return false;
    }
}


#if USE_VALIDATION_MODE
void Validation::createTriangleWave(s16* pOut, size_t sampleNum)
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pOut));
    PIA_SAMPLE_ASSERT(sampleNum > 0);

    const s16 MAX_AMP = 30000;                     // Maximum amplitude.
    const u32 WAVELENGTH = 8;                      // Wavelength.
    const int SLOPE = MAX_AMP / (WAVELENGTH / 4);  // Slope.

    for (size_t i = 0; i < sampleNum; ++i)
    {
        int w = i % WAVELENGTH;
        if (w < WAVELENGTH / 2)
        {
            pOut[i] = static_cast<s16>(SLOPE * w - MAX_AMP);
        }
        else
        {
            pOut[i] = static_cast<s16>(-SLOPE * w + 3 * MAX_AMP);
        }
    }
}


int Validation::s_Mode = 0;
s16 Validation::s_TriangleWaveData[TRIANGLE_WAVE_DATA_SAMPLING_NUM];
bool Validation::s_bPlayTriangleWave = false;


void Validation::initialize()
{
    createTriangleWave(s_TriangleWaveData, TRIANGLE_WAVE_DATA_SAMPLING_NUM);
}


void Validation::renderValidationMode()
{
    nn::pia::sampleframework::SampleFramework* framework = nn::pia::sampleframework::SampleFramework::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(framework));
    framework->DrawText(100, 200, "Validation: %s", s_bPlayTriangleWave ? "Triangle wave" : "mic input");
    nn::pia::transport::ThreadStreamManager* pManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    if (PIA_IS_VALID_POINTER(pManager))
    {
        framework->DrawText(100, 210, "Packet loss emu: %d %%", pManager->GetSendThreadPacketLossRatio());
    }
    else
    {
        framework->DrawText(100, 210, "Packet loss emu: -");
    }
}


void Validation::packetLossEmulation(int percentage)
{
    nn::pia::transport::ThreadStreamManager* pManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    nn::Result r = pManager->SetSendThreadPacketLossRatio(percentage);
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
}


void Validation::gotoNextValidation()
{
    ++s_Mode;
    s_Mode = s_Mode % ITEMS;

    switch (s_Mode)
    {
    case 0:
        s_bPlayTriangleWave = false;
        packetLossEmulation(0);
        break;
    case 1:
        s_bPlayTriangleWave = false;
        packetLossEmulation(10);
        break;
    case 2:
        s_bPlayTriangleWave = false;
        packetLossEmulation(30);
        break;
    case 3:
        s_bPlayTriangleWave = true;
        packetLossEmulation(0);
        break;
    default:
        PIA_SAMPLE_ASSERT(false);
        break;
    }
}
#endif  // end of USE_VALIDATION_MODE
