﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneConference.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SceneConference.h"

#include <pia/session.h>
#include <pia/transport.h>

#include "Config.h"
#include "Misc.h"
#include "Types.h"
#include "MenuParticipants.h"
#include "MenuConfig.h"
#include "Sndmic.h"
#include "HomeButtonProcess.h"
#include "Global.h"

#include "Shared_Chat_Setup.h"


namespace
{
SndMicDevice s_SndMicDevice;
SndMicDevice* p_SndMicDevice;

#if NN_PIA_CAFE
u32 s_micGainMax;
u32 s_micGainMin;
u32 s_micGainStep;
u32 s_micDigitalGain;
u32 s_micDigitalFlag;
#endif

#if NN_PIA_CTR
void startConference()
{
    SndMicDevice* sndmic = &s_SndMicDevice;
    SndMicConfigStructure sndmicConfig;

    // Set Snd/Mic Config.
    sndmicConfig.frame_msec = Chat_GetVoiceProtocol()->GetAudioFrameLength();
    sndmicConfig.mic_gain = 60;

    // Initialize Snd/Mic.
    if (sndmic->Initialize(&sndmicConfig, Chat_GetVoiceProtocol()) != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Initialize snd or mic.\n");
    }

    // Start Snd/Mic.
    if (sndmic->Start() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Start snd or mic.\n");
    }

    p_SndMicDevice = sndmic;
}

#elif NN_PIA_CAFE
// Starts a conference. Thread for voice transmissions.
void startConference()
{
    SND_PARAM sndConfig;
    sndConfig.freq = SND_FREQ_08KHZ;
    sndConfig.frameSample = Chat_GetVoiceProtocol()->GetAudioFrameLength();

    SndMicDevice* sndmic = &s_SndMicDevice;
    sndmic->InitializeMicSndCommon(&sndConfig, Chat_GetVoiceProtocol());

    MIC_STATE_INFO micStateInfo;
    sndmic->GetStateMic(&micStateInfo);
    s_micGainMax = micStateInfo.micGainMax;
    s_micGainMin = micStateInfo.micGainMin;
    s_micGainStep = micStateInfo.micGainStep;
    s_micDigitalGain = 0x1000;
    s_micDigitalFlag = 0;

    p_SndMicDevice = sndmic;
}

#endif  // end of NN_PIA_CAFE

void stopConference()
{
    SndMicDevice* sndmic = &s_SndMicDevice;

    nn::pia::transport::StationManager* pMgr = nn::pia::transport::StationManager::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

    for (nn::pia::transport::StationManager::Iterator it = pMgr->Begin(); it != pMgr->End(); ++it)
    {
        Chat_GetVoiceProtocol()->RemoveConferenceClient(it->GetStationId());
    }

#if NN_PIA_CTR
    if (sndmic->IsPlayingBGM())
    {
        sndmic->StopBGM();
    }

    // Stop Snd/Mic
    if (sndmic->Stop() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Stop snd or mic.\n");
    }

    // Finalize Snd/Mic
    if (sndmic->Finalize() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Finalize snd or mic.\n");
    }
#elif NN_PIA_CAFE
    // Stop Snd/Mic
    if (sndmic->StopMic() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Stop mic.\n");
    }
    if (sndmic->StopSnd() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Stop snd.\n");
    }

    // Finalize Snd/Mic
    if (sndmic->FinalizeMic() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Finalize mic.\n");
    }
    if (sndmic->FinalizeSnd() != SNDMIC_ERROR_NONE)
    {
        PIA_BASIC_REPORT("Error:Finalize snd.\n");
    }

    sndmic->FinalizeMicSndCommon();
#endif

    p_SndMicDevice = NULL;
}


#if NN_PIA_CTR
void processCtrHomeButton()
{
    if (nn::applet::IsExpectedToProcessHomeButton())
    {
        s_SndMicDevice.Pause();  // Pause
        nn::applet::ProcessHomeButtonAndWait();
        nn::applet::ClearHomeButtonState();
        nngxUpdateState(NN_GX_STATE_ALL);
        nngxValidateState(NN_GX_STATE_ALL, GL_TRUE);
        Chat_GetVoiceProtocol()->ClearStreaming();
        s_SndMicDevice.Pause();  // End pause.
    }
}
#endif  // NN_PIA_CTR


#if 0
// All stations that can participate in the conference are joined in to the conference at the time this function is called.
void participateAll()
{
    nn::pia::transport::StationManager* pMgr = nn::pia::transport::StationManager::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

    for(nn::pia::transport::StationManager::Iterator it = pMgr->Begin(); it != pMgr->End(); ++it)
    {
        nn::pia::StationId id = it->GetStationId();
        bool bStationIdValid = nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(id);
        bool bParticipate = Chat_GetVoiceProtocol()->IsClientInConference(id);
        bool bMyself = nn::pia::session::Session::GetInstance()->GetLocalStationId() == id;
        if (bStationIdValid && !bParticipate && !bMyself)  // A station joins if not joined in and not the local station.
        {
            Chat_GetVoiceProtocol()->AddConferenceClient(id);
        }
    }
}
#endif


//static void dispAudioInformation( SampleChatGlobal *pGlobal )
void drawAudioInformation(s32 idx)
{
    nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::StationId id = g_Indexer.GetStationId(static_cast<size_t>(idx));
    if (id == nn::pia::STATION_ID_INVALID)
    {
        f->DrawText(0, 8 * 0, "----------------------------------------------------");
        f->DrawText(0, 8 * 2, "<<< Audio Information for index %d is not ready. >>>", idx);
        f->DrawText(0, 8 * 4, "----------------------------------------------------");
        return;
    }

    nn::pia::chat::AudioInfo audioInfo;
    Chat_GetVoiceProtocol()->GetAudioInfo(id, &audioInfo);

    f->DrawText(0, 8 * 0, "----------------------------------------------------");
    f->DrawText(0, 8 * 2, "<<<<          Audio Information...              >>>>");
    f->DrawText(0, 8 * 4, "----------------------------------------------------");
    f->DrawText(0, 8 * 6, " StationId                 : 0x%016llx", id.GetValue());
    f->DrawText(0, 8 * 7, " Sequence                  : %ld", audioInfo.sequence);
    f->DrawText(0, 8 * 8, " Internal Latency          : %ld", audioInfo.internalLatency);
#if NN_PIA_CTR
    f->DrawText(0, 8 * 9, " Buffer Latency            : %ld", static_cast<u32>(audioInfo.bufferLatency.ToTimeSpan().GetMilliSeconds()));
    f->DrawText(0, 8 * 10, " Jitter                    : %ld", static_cast<u32>(audioInfo.jitter.ToTimeSpan().GetMilliSeconds()));
    f->DrawText(0, 8 * 11, " Clock Skew                : %ld", static_cast<u32>(audioInfo.clockSkew.ToTimeSpan().GetMilliSeconds()));
#else
    f->DrawText(0, 8 * 9, " Buffer Latency            : %ld", static_cast<u32>(OSTicksToMilliseconds(audioInfo.bufferLatency)));
    f->DrawText(0, 8 * 10, " Jitter                    : %ld", static_cast<u32>(OSTicksToMilliseconds(audioInfo.jitter)));
    f->DrawText(0, 8 * 11, " Clock Skew                : %ld", static_cast<u32>(OSTicksToMilliseconds(audioInfo.clockSkew)));
#endif
    f->DrawText(0, 8 * 12, " Buffer Count              : %ld", audioInfo.bufferCount);
    f->DrawText(0, 8 * 13, " Drop Count                : %ld", audioInfo.dropCount);
    f->DrawText(0, 8 * 14, " Continuous Drop           : %ld", audioInfo.continuousDrop);
    f->DrawText(0, 8 * 15, " Jam Count                 : %ld", audioInfo.jamCount);
    f->DrawText(0, 8 * 16, " Duplicate Count           : %ld", audioInfo.duplicateCount);
    f->DrawText(0, 8 * 17, " Wait Count                : %ld", audioInfo.waitCount);
    f->DrawText(0, 8 * 18, " Late Count                : %ld", audioInfo.lateCount);
    f->DrawText(0, 8 * 19, " Recent Late Count         : %ld", audioInfo.recentLateCount);
    f->DrawText(0, 8 * 20, " Recent Receive Count      : %ld", audioInfo.recentReceiveCount);
    f->DrawText(0, 8 * 21, " Recent Late Ratio         : %ld", audioInfo.recentLateRatio);
    f->DrawText(0, 8 * 22, " Recover Count             : %ld", audioInfo.recoverCount);
    f->DrawText(0, 8 * 23, " Recover Using Spare Count : %ld", audioInfo.recoverUsingSpareCount);
    f->DrawText(0, 8 * 25, " VAD Value                 : %d", Chat_GetVoiceProtocol()->GetVADActivity());
    f->DrawText(0, 8 * 26, " VAD Power                 : %u", Chat_GetVoiceProtocol()->GetVADPower());
    f->DrawText(0, 8 * 28, "----------------------------------------------------");
}

void drawConferenceCaption()
{
    nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();

    f->DrawText(0, 0, "----------------------------------------");
    f->DrawText(0, 10, "<<<<      Conference working.       >>>>");
    f->DrawText(0, 20, "----------------------------------------");
    f->DrawText(0, 32, "  MY StationId(0x%016llx)", nn::pia::session::Session::GetInstance()->GetLocalStationId().GetValue());
#if NN_PIA_CTR
    f->DrawText(0, 40, "  Press Y => BGM On/Off.");
#endif
}

void drawConfigCaption()
{
    nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();

    f->DrawText(0, 8 * 0, "----------------------------------------");
    f->DrawText(0, 8 * 2, "<<<<   Change Conference setting.   >>>>");
    f->DrawText(0, 8 * 4, "----------------------------------------");
    f->DrawText(0, 8 * 5, "Press Y => BGM On/Off.");
}


void sceneConferenceCore()
{
    MenuParticipants menuParticipants;
    MenuConfig menuConfig(&s_SndMicDevice);

    bool bConfig = false;  // Used as a switch for <tt>menuParticipants</tt> and <tt>menuConfig </tt>.

    s32 audioInfoIdx = 0;
    while (1)
    {
#if NN_PIA_CTR
        processCtrHomeButton();
#elif NN_PIA_CAFE
        if (HomeButtonProcess::Process(&s_SndMicDevice))
        {
            continue;
        }
#endif

        beginFrame();

        if (bConfig)
        {
            menuConfig.Dispatch();
        }
        else
        {
            menuParticipants.Dispatch();
        }

        changeRenderTarget(true);  // Upper screen.
        drawAudioInformation(audioInfoIdx);
        swapBuffers();

        changeRenderTarget(false);  // Lower screen.
        if (bConfig)
        {
            drawConfigCaption();
            menuConfig.Draw();
        }
        else
        {
            drawConferenceCaption();
            menuParticipants.Draw();
        }
        swapBuffers();

        endFrame();

        // Noise skip determination.
        if (getSampleFrameworkInstance()->IsTrigger(0xffffffff) || getSampleFrameworkInstance()->IsRelease(0xffffffff))
        {
            Chat_GetVoiceProtocol()->SignalNoiseSkip();
        }

        // BGM On/Off.
        if (getSampleFrameworkInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y) && (p_SndMicDevice != NULL))
        {
            if (p_SndMicDevice->IsPlayingBGM())
            {
                p_SndMicDevice->StopBGM();
            }
            else
            {
                p_SndMicDevice->StartBGM();
            }
        }

        // Switch <tt>AudioInfo</tt> display using L/R Buttons.
        if (getSampleFrameworkInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
        {
            audioInfoIdx = (audioInfoIdx + MAX_CLIENTS - 1) % MAX_CLIENTS;
        }
        if (getSampleFrameworkInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            audioInfoIdx = (audioInfoIdx + 1) % MAX_CLIENTS;
        }

#if USE_VALIDATION_MODE
        if (getSampleFrameworkInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            Validation::gotoNextValidation();
        }
#endif  // enf of USE_VALIDATION_MODE

        // Determination of scene transitions and switch to the config screen, etc.
        if (bConfig)
        {
            if (menuConfig.IsReadyToReturnConference())
            {
                menuConfig.ResetFlag();
                bConfig = false;
            }
        }
        else
        {
            if (menuParticipants.IsReadyToGoChangeSetting())
            {
                menuParticipants.ResetFlag();
                bConfig = true;
            }
            if (menuParticipants.IsReadyToReturnConnected())
            {
                break;
            }
        }

        // Escape if the session connection has been dropped.
        if (!isConnectedSession())
        {
            break;
        }
    }
}

}  // end of anonymous namespace


s32 sceneConference()
{
    setClearColor(0.0f, 0.0f, 0.9f);

    startConference();

    sceneConferenceCore();

    stopConference();

    return SCENE_CONNECTED;
}
