﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneNetwork.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SceneNetwork.h"

#include <pia/inet.h>
#include <pia/session.h>
#include <pia/util.h>

#include "Shared_Common_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Chat_Setup.h"

#include "SampleLib/SampleLib.h"

#include "Types.h"
#include "Misc.h"
#include "Config.h"
#include "Global.h"


namespace
{
const size_t PIA_HEAP_MEMORY_SIZE = 8 * 1024 * 1024;
u8 s_PiaHeapMemory[PIA_HEAP_MEMORY_SIZE] PIA_ATTRIBUTE_ALIGN(4);


// Process during connection and disconnection.
void SessionEventCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    nn::Result r;
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
        g_Indexer.Add(stationId);
        break;
    case nn::pia::session::Session::EVENT_LEAVE:
        g_Indexer.Remove(stationId);
        break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
        break;
    case nn::pia::session::Session::EVENT_HOST_MIGRATION_FAILED:
        // Host migration fails.
        break;
    default:
        PIA_SAMPLE_ASSERTMSG(false, "Unsupported event(%d).", eventType);
        break;
    }
}


// Initializes the Pia modules.
void initPiaModules()
{
    nn::Result r;

// Start with <tt>common</tt>.
#if 1
    u64 traceFlags =
        nn::pia::common::TRACE_FLAG_COMMON |
        nn::pia::common::TRACE_FLAG_INET |
        nn::pia::common::TRACE_FLAG_TRANSPORT |
        nn::pia::common::TRACE_FLAG_TRANSPORT_RELAY |
        nn::pia::common::TRACE_FLAG_SESSION |
        nn::pia::common::TRACE_FLAG_CHAT;
#else
    u64 traceFlags = 0LL;
#endif
    r = Common_InitializePiaCommon(
        s_PiaHeapMemory,
        PIA_HEAP_MEMORY_SIZE,
        traceFlags,
        true);  // <tt>CachedPrint</tt> enabled.
    PIA_SAMPLE_ASSERT(r.IsSuccess());

#if 0
        // Setting for a less verbose log.
        nn::pia::common::Log::GetGlobalObject()->SetMask(
            nn::pia::common::Log::LEVEL_ERROR |
            nn::pia::common::Log::LEVEL_FATAL
            );
#endif

    // Initializes NEX before initializing PiaInet.
    r = Nex_InitializeNex(
        1 * 1024 * 1024,                                // Memory for NEX.
        512 * 1024,                                     // Emergency memory for NEX.
        nn::nex::Core::ThreadModeUnsafeTransportBuffer  // Thread mode.
        );

    // Next do <tt>inet</tt>.
    r = Inet_InitializePiaInet();
    PIA_SAMPLE_ASSERT(r.IsSuccess());

#if USE_VALIDATION_MODE
    {
        struct nn::pia::transport::Transport::DebugSetting debugSetting;
        debugSetting.sendThreadLatencyEmulationPacketNum = 0;
        debugSetting.receiveThreadLatencyEmulationPacketNum = 0;
        debugSetting.bPacketLossEmulation = true;

        nn::Result ret = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(ret);
    }
#endif

    // Next do <tt>transport</tt>.
    nn::pia::transport::Transport::Setting setting;
    setting.pFactory = Inet_GetNetworkFactory();
    setting.maxStationNum = MAX_CLIENTS;
    setting.maxSendThreadBufferNum = 16;
    setting.maxReceiveThreadBufferNum = 32;
    r = Transport_InitializePiaTransport(
        setting,
        0,     // <tt>UnreliableProtocol</tt> not in use.
        NULL,  // <tt>UnreliableProtocol</tt> not in use.
        0,     // <tt>ReliableProtocol</tt> not in use.
        NULL   // <tt>ReliableProtocol</tt> not in use.
        );
    PIA_SAMPLE_ASSERT(r.IsSuccess());

    // Next do <tt>chat</tt>.
    r = Chat_InitializePiaChat(g_MaxStationNum);
    PIA_SAMPLE_ASSERT(r.IsSuccess());

// Next do <tt>session</tt>.
#if ENABLE_RELAY_EMULATION
    {
        // If you are performing relay emulation, you need to call the settings function before creating the session instance.
        nn::pia::session::Session::DebugSetting debugSetting;
        debugSetting.isEnableRelayEmulation = true;
        nn::Result ret = nn::pia::session::Session::SetDebugSetting(debugSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(ret);
    }
#endif

    // Initializes the session.
    nn::pia::session::Session::Setting sessionSetting;
    sessionSetting.pNetworkFactory = setting.pFactory;
#if ENABLE_RELAY_EMULATION
    sessionSetting.networkTopology = nn::pia::session::NetworkTopology_RelayMesh;
#else
    sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
#endif

    r = Session_InitializePiaSession(sessionSetting, static_cast<bool>(ENABLE_RELAY_EMULATION));
    PIA_SAMPLE_ASSERT(r.IsSuccess());

    nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SessionEventCallback);
}


// Finalizes the Pia modules. In reverse order of the <tt>initPia</tt> function.
void finiPiaModules()
{
    nn::pia::session::Session::GetInstance()->UnregisterSessionEventCallback();
    PIA_SAMPLE_ASSERT(g_Indexer.GetValidIndexNum() == 0);

    Session_FinalizePiaSession();

    Chat_FinalizePiaChat();

    Transport_FinalizePiaTransport(0, 0);  // <tt>UnreliableProtocol and</tt>ReliableProtocol</tt> not in use.

    Inet_FinalizePiaInet();

    Nex_FinalizeNex();

    Common_FinalizePiaCommon();
}


#if ENABLE_RELAY_EMULATION
void setRelayEmulation()
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

    // Initialize relay emulation before configuring in case the previous settings remain.
    pSession->ResetRelayEmulation();

    // 2-3 is a relay connection.
    PIA_BASIC_REPORT("Set RelayEmulation: 2-3\n");
    nn::Result r = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3, nn::pia::session::Mesh::EMULATION_TYPE_RELAY);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
}
#endif  // end of ENABLE_RELAY_EMULATION


void drawNetworkConnectingCaption()
{
    nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
    if (PIA_IS_VALID_POINTER(f))
    {
        f->DrawText(0, 8 * 4, "----------------------------------------");
        f->DrawText(0, 8 * 6, "<<<<       Network Connecting       >>>>");
        f->DrawText(0, 8 * 8, "----------------------------------------");
    }
}


void drawNetworkDisconnectingCaption()
{
    nn::pia::sampleframework::SampleFramework* f = nn::pia::sampleframework::SampleFramework::GetInstance();
    if (PIA_IS_VALID_POINTER(f))
    {
        f->DrawText(0, 8 * 4, "----------------------------------------");
        f->DrawText(0, 8 * 6, "<<<<     Network Disconnecting      >>>>");
        f->DrawText(0, 8 * 8, "----------------------------------------");
    }
}

}  // end of anonymous namespace


s32 sceneOpenNetwork()
{
    initPiaModules();

    setClearColor(0.0f, 0.3f, 0.0f);

    {
        // At this point in time, NEX dispatches are not performed. This is done in order to avoid the appearance of the following message.
        // (Error) A long delay has been observed between two calls to Scheduler::Dispatch
        beginFrame(false);

        changeRenderTarget(false);  // Render on the lower screen
        drawNetworkConnectingCaption();
        swapBuffers();

        endFrame();

        nn::Result r;

        // Connects to the Internet.
        r = Inet_InitializeNetwork();
        if (r.IsFailure())
        {
            PIA_BASIC_REPORT("Inet_InitializeNetwork() failed...\n");
            return SCENE_CLOSE_NETWORK;
        }

        // Log in to the game server.
        r = Inet_Login();
        if (r.IsFailure())
        {
            PIA_BASIC_REPORT("Inet_Login() failed...\n");
            return SCENE_CLOSE_NETWORK;
        }

        // Process for starting up the <tt>Session</tt> class.
        nn::pia::session::Session::StartupSetting startupSetting;
        startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

        r = Session_StartupSession(startupSetting);
        if (r.IsFailure())
        {
            PIA_BASIC_REPORT("Session_StartupSession() failed...\n");
            return SCENE_CLOSE_NETWORK;
        }

        // Match automatically.
        nn::pia::inet::NexCreateSessionSetting sessionSetting;
        sessionSetting.SetGameMode(g_GameMode);
        sessionSetting.SetMaxParticipants(g_MaxStationNum);
        sessionSetting.SetOpenParticipation(true);

        const size_t searchSettingNum = 1;
        nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
        {
            criteria[0].SetGameMode(g_GameMode);
            criteria[0].SetMaxParticipants(g_MaxStationNum);
        }
        SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinRandomSession(&sessionSetting, criteria, searchSettingNum, &r);
        if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            PIA_BASIC_REPORT("Session_JoinRandomSession:failed");
            PIA_SAMPLE_LOG_RESULT(r);
            return SCENE_CLOSE_NETWORK;
        }

#if ENABLE_RELAY_EMULATION
        // Set relay emulation.
        setRelayEmulation();
#endif

        bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
        if (isHost)
        {
            PIA_BASIC_REPORT("You're host station.");
        }
        else
        {
            PIA_BASIC_REPORT("You're client station.");
        }
    }

    PIA_BASIC_REPORT("sceneOpenNetwork() success.\n");
    return SCENE_CONNECTED;
}


s32 sceneCloseNetwork()
{
    setClearColor(0.0f, 0.3f, 0.3f);

    {
        beginFrame();

        changeRenderTarget(false);  // Render on the lower screen
        drawNetworkDisconnectingCaption();
        swapBuffers();

        endFrame();

        // Destroys or leaves the session.
        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

        bool bCallLeaveSession = false;
        nn::Result r = pSession->CheckConnectionError();
        if (r.IsSuccess() && (pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION))
        {
            bCallLeaveSession = true;
        }
        else if (r.IsFailure())
        {
            enum nn::pia::util::HandlingType handling = nn::pia::util::GetHandlingType(r);
            switch(handling)
            {
            case nn::pia::util::HANDLING_LEAVE_AND_CLEANUP:
                bCallLeaveSession = true;
                break;
            case nn::pia::util::HANDLING_RECONNECT:
                break;
            default:
                // The GetHandlingType function is capable of returning values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT, but when the Result objects that can be returned by the Session::CheckConnectionError function are considered, values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT are never returned.
                // 
                // 
                // An assert statement is provided for that reason.
                PIA_SAMPLE_ASSERTMSG(false, "Unexpected result.\n");
                break;
            }
        }

        if (bCallLeaveSession)
        {
            Session_LeaveSession();
        }

        // Clean up
        Session_CleanupSession();

        Inet_Logout();
        Inet_FinalizeNetwork();
    }

    finiPiaModules();

    return SCENE_TOP;
}
