﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Nex_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"

#if NN_PIA_CAFE
#include <cafe/mem.h>
#endif

#if ENABLE_SERVERLESS
nn::Result Nex_InitializeNex(
    size_t memorySize,
    size_t emergencyMemorySize,
    nn::nex::Core::ThreadMode threadMode)
{
    NN_PIA_DUMMY_PARAM(memorySize);
    NN_PIA_DUMMY_PARAM(emergencyMemorySize);
    NN_PIA_DUMMY_PARAM(threadMode);
    return nn::ResultSuccess();
}

void Nex_FinalizeNex()
{
}

bool Nex_IsNgsConnected()
{
    return true;
}

bool Nex_CheckInternalError()
{
    return false;
}

nn::nex::NgsFacade* Nex_GetNgsFacadeInstance()
{
    return NULL;
}

u32 Nex_GetInternalErrorCode()
{
    return 0;
}

#else

namespace
{
// Multiple calls for finalization are OK.
bool s_Initialized = false;

// Main memory size
size_t s_MemorySize = 0;

// Emergency memory size
size_t s_EmergencyMemorySize = 0;

// Main memory.
void* s_pMemory = NULL;

// Emergency memory used when the heap is depleted.
void* s_pEmergencyMemory = NULL;

// Memory usage.
size_t s_MemUsage = 0;
size_t s_MemMaxUsage = 0;

#if NN_PIA_CTR
// The heap for NEX.
nn::fnd::ThreadSafeExpHeap s_NexHeap;
#elif NN_PIA_CAFE
MEMHeapHandle s_NexHeap;
#endif

#if NN_PIA_NEX_3
//P2P signature key. This must be changed for each application.
const nn::nex::qByte SIGNATURE_KEY[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09 };
#endif

// Specify the variable for storing the error code after conversion by the error converter.
u32 s_InternalErrorCode;
u32 s_LastAPIErrorCode;

// Instance of <tt>NgsFacade </tt>.
nn::nex::NgsFacade* s_pNgsFacade = NULL;

#if !NN_PIA_WIN
// For the community ID
nn::nex::MatchmakeExtensionClient* s_pMatchmakeClient = NULL;
#endif

//-----------------------------------------------------------------------------------------
// Declaration of internal functions
//-----------------------------------------------------------------------------------------
void InitializeRuntimeMemory(size_t memorySize, size_t emergencyMemorySize);
void FinalizeRuntimeMemory();

void* InnerAlloc(size_t size);
void InnerFree(void* ptr);

void* AllocFunction(size_t size);
void FreeFunction(void* pointer);

void AcquireReservedMemory();
void ReleaseReservedMemory();
}

//-----------------------------------------------------------------------------------------
// Implementation of external functions
//-----------------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------------
// Initialize NEX.
//-----------------------------------------------------------------------------------------
nn::Result Nex_InitializeNex(
    size_t memorySize,
    size_t emergencyMemorySize,
    nn::nex::Core::ThreadMode threadMode)
{
    PIA_SAMPLE_LOG("Nex_InitializeNex\n");

    // Initialize the NEX heap.
    PIA_SAMPLE_LOG("Nex_InitializeNex:InitializeRuntimeMemory\n");
    InitializeRuntimeMemory(memorySize, emergencyMemorySize);

    // Initialize global instances.
    nn::nex::GlobalVariables::AcquireInstance();

    // Initialize error handling.
    Nex_InitializeErrorHandling();

    // Thread settings.
    nn::nex::Core::SetThreadMode(threadMode);

    // Network topology settings. (Only needs to be called when using NetZ.)
    nn::nex::NetZ::SetNetworkTopology(nn::nex::NetworkTopology::Peer2Peer);

    // Sets the P2P portion's signature key. (Only needs to be called when using NetZ.)
    nn::nex::Network::GetInstance()->SetP2PDataPacketDefaultSignatureKey(SIGNATURE_KEY, sizeof(SIGNATURE_KEY));

    // Creates an <tt>NgsFacade</tt> instance.
    s_pNgsFacade = qNew nn::nex::NgsFacade();
    PIA_SAMPLE_ASSERT(s_pNgsFacade != NULL);

    s_pMatchmakeClient = qNew nn::nex::MatchmakeExtensionClient();
    PIA_SAMPLE_ASSERT(s_pMatchmakeClient != NULL);
    PIA_SAMPLE_LOG("Created MatchmakeExtensionClient for PersistentGathering.\n");

    // Set the NEX trace.
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    s_Initialized = true;
    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------------------
// Finalize NEX.
//-----------------------------------------------------------------------------------------
void Nex_FinalizeNex()
{
    if (!s_Initialized)
    {
        PIA_SAMPLE_LOG("Nex_FinalizeNex was already finalized.\n");
        return;
    }

    PIA_SAMPLE_LOG("Nex_FinalizeNex:\n");

    // Destroys the <tt>NgsFacade</tt> instance.
    if (s_pNgsFacade)
    {
        qDelete s_pNgsFacade;
        s_pNgsFacade = NULL;
    }

    if (s_pMatchmakeClient)
    {
        qDelete s_pMatchmakeClient;
        s_pMatchmakeClient = NULL;
        PIA_SAMPLE_LOG("Deleted MatchmakeExtensionClient for PersistentGathering.\n");
    }

    // Finalize global instances.
    nn::nex::GlobalVariables::ReleaseInstance();

    // Finalize error handling.
    Nex_FinalizeErrorHandling();

    // Finalize the NEX heap.
    FinalizeRuntimeMemory();

    s_Initialized = false;
}

//-----------------------------------------------------------------------------------------
// Output the protocol call context error to the log.
//-----------------------------------------------------------------------------------------
void Nex_ReportProtocolCallContext(nn::nex::ProtocolCallContext& context, const char* pHeaderText)
{
    // Suppress warnings about unused parameters.
    (void)context;

    if (pHeaderText)
    {
        PIA_SAMPLE_LOG(pHeaderText);
    }

    PIA_SAMPLE_LOG("Nex_ReportProtocolCallContext:(Facility:%d Code:0x%X NErr:%d)\n",
                   context.GetOutcome().GetFacility(),
                   context.GetOutcome().GetReturnCode(),
                   nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(context.GetOutcome()));
}

//-----------------------------------------------------------------------------------------
// Function for getting the <tt>nn::nex::NgsFacade</tt> instance.
//-----------------------------------------------------------------------------------------
nn::nex::NgsFacade* Nex_GetNgsFacadeInstance()
{
    return s_pNgsFacade;
}

bool Nex_IsNgsConnected()
{
    if (s_pNgsFacade)
    {
        if (s_pNgsFacade->IsConnected())
        {
            return true;
        }
    }
    return false;
}

#if !NN_PIA_WIN
nn::nex::MatchmakeExtensionClient* Nex_GetMatchmakeExtensionClientInstance()
{
    return s_pMatchmakeClient;
}
#endif

//-----------------------------------------------------------------------------------------
// Related to error handling.
//-----------------------------------------------------------------------------------------
void Nex_InitializeErrorHandling()
{
    s_InternalErrorCode = 0;
    s_LastAPIErrorCode = 0;
}

void Nex_FinalizeErrorHandling()
{
    s_InternalErrorCode = 0;
    s_LastAPIErrorCode = 0;
}


//-----------------------------------------------------------------------------------------
// This function is for checking errors occurring in or reported by NEX. (The main thread calls this function at regular intervals.)
//-----------------------------------------------------------------------------------------
bool Nex_CheckInternalError()
{
    bool errorOccurred = false;

    nn::nex::NetZ* pNetZ = nn::nex::NetZ::GetInstance();
    if (pNetZ)
    {
        // After the NetZ instance has been created, check whether a fatal error has occurred in NetZ.
        nn::nex::qResult result = pNetZ->GetFatalError();
        if (result == false)
        {
            // A fatal error has occurred.
            PIA_SAMPLE_LOG("Nex_CheckInternalError: A fatal error occurred. Code:0x%08x\n", result.GetReturnCode());
            s_InternalErrorCode = nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(result);
            errorOccurred = true;
        }
    }

    // Return the results of the check. (Return true if the error cannot be ignored, and false otherwise.)
    return errorOccurred;
}

//-----------------------------------------------------------------------------------------
// This function checks the error details if a call to the NEX API fails.
//-----------------------------------------------------------------------------------------
bool Nex_CheckLastAPIError()
{
    bool errorOccurred = false;

    // Output to the log, and do not perform any special handling.
    const size_t SAMPLE_ERROR_STRING_LENGTH = 64;
    nn::nex::qChar16 errorStringChar16[SAMPLE_ERROR_STRING_LENGTH] = {};
    char errorString[SAMPLE_ERROR_STRING_LENGTH] = {};

    // Get the latest error stored in the nn::nex::SystemError class.
    u32 errorResult = nn::nex::SystemError::GetLast();

    // Get the error string, and convert it to log output.
    nn::nex::SystemError::GetErrorString(1, reinterpret_cast<nn::nex::qChar*>(errorStringChar16), SAMPLE_ERROR_STRING_LENGTH);
    int ret = ::std::wcstombs(errorString, errorStringChar16, SAMPLE_ERROR_STRING_LENGTH);
    if (ret == -1)
    {
        // Conversion failed.
        ::std::strncpy(errorString, "wcstombs failed", SAMPLE_ERROR_STRING_LENGTH);
    }

    // For safety, explicitly insert a terminating character.
    errorString[SAMPLE_ERROR_STRING_LENGTH - 1] = '\0';

    PIA_SAMPLE_LOG("CheckNexLastError: Code:0x%08x [%s]\n", errorResult, errorString);

    // Convert to a network error code.
    s_LastAPIErrorCode = nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(errorResult);

    // Return the results of the check. (Return true if the error cannot be ignored, and false otherwise.)
    return errorOccurred;
}

u32 Nex_GetInternalErrorCode()
{
    return s_InternalErrorCode;
}

u32 Nex_GetLastAPIErrorCode()
{
    return s_LastAPIErrorCode;
}

size_t Nex_GetPeakMemoryUsage()
{
    return s_MemMaxUsage;
}


//-----------------------------------------------------------------------------------------
// Implement internal functions.
//-----------------------------------------------------------------------------------------
namespace
{
//-----------------------------------------------------------------------------------------
// Initialize the NEX heap (ExpHeap).
//-----------------------------------------------------------------------------------------
void InitializeRuntimeMemory(size_t memorySize, size_t emergencyMemorySize)
{
    s_MemorySize = memorySize;
    s_EmergencyMemorySize = emergencyMemorySize;

    size_t total = memorySize + emergencyMemorySize;
    PIA_SAMPLE_LOG("InitializeRuntimeMemory:MemorySize:%d, ReservedMemorySize:%d, Total:%d\n", s_MemorySize, s_EmergencyMemorySize, total);

#if NN_PIA_CTR
    // Allocate work memory.
    s_pMemory = new u8[total];
    // Initialize the heap.
    s_NexHeap.Initialize(NN_ANY_TO_UPTR(s_pMemory), total);
#elif NN_PIA_WIN
    // Allocate work memory.
    s_pMemory = new u8[total];
#elif NN_PIA_CAFE
    // Allocate work memory from the default heap (aligned to 8 bytes).
    s_pMemory = MEMAllocFromDefaultHeapEx(total, 8);
    // Initialize the heap.
    s_NexHeap = MEMCreateExpHeapEx(s_pMemory, total, MEM_HEAP_OPT_0_CLEAR | MEM_HEAP_OPT_THREAD_SAFE);
#endif

    // Register the memory function.
    nn::nex::MemoryManager::SetBasicMemoryFunctions(AllocFunction, FreeFunction);

    if (s_EmergencyMemorySize != 0)
    {
        // Allocate reserved memory for an emergency.
        AcquireReservedMemory();
    }
}

//-----------------------------------------------------------------------------------------
// Finalize the NEX heap (ExpHeap).
//-----------------------------------------------------------------------------------------
void FinalizeRuntimeMemory()
{
    if (s_EmergencyMemorySize != 0)
    {
        // If there is any remaining memory reserved for an emergency, release it.
        ReleaseReservedMemory();
    }

    PIA_SAMPLE_PRINT(
        "FinalizeRuntimeMemory:Nex memory usage: Now memory usage:%d, Max memory usage:%d / %d\n",
        s_MemUsage,
        s_MemMaxUsage,
        s_MemorySize + s_EmergencyMemorySize);

#if NN_PIA_CTR
    // Memory is leaking.
    if (s_MemUsage > 0)
    {
        // Display the memory that is leaking.
        s_NexHeap.Dump();
        // Disable the memory that is leaking.
        s_NexHeap.Invalidate();
    }
    // Destroy the heap.
    s_NexHeap.Finalize();
    // Destroy work memory.
    delete[] s_pMemory;
#elif NN_PIA_WIN
    // Destroy work memory.
    delete[] s_pMemory;
#elif NN_PIA_CAFE
    // Memory is leaking.
    if (s_MemUsage > 0)
    {
        // Dump the heap information.
        MEMDumpHeap(s_NexHeap);
    }
    // Destroy the heap.
    MEMDestroyExpHeap(s_NexHeap);
    // Deallocate work memory.
    MEMFreeToDefaultHeap(s_pMemory);
#endif

    s_MemUsage = 0;
    s_MemMaxUsage = 0;
}

#if (NN_PIA_NEX_3 == 0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP == 0)
//-----------------------------------------------------------------------------------------
// Initialize the NEX heap (MultiUnitHeap).
//-----------------------------------------------------------------------------------------
void InitializeMultiUnitHeap(size_t memorySize, size_t emergencyMemorySize, bool useNetZ)
{
    s_MemorySize = memorySize;
    s_EmergencyMemorySize = emergencyMemorySize;

    size_t total = memorySize + emergencyMemorySize;
    PIA_SAMPLE_LOG("InitializeMultiUnitHeap:MemorySize:%d, ReservedMemorySize:%d, Total:%d\n", s_MemorySize, s_EmergencyMemorySize, total);

    s_pMemory = new u8[total];

    const int LEVEL = 5;
    size_t unitSize[LEVEL];
    size_t unitObjs[LEVEL];
    if (useNetZ)
    {
        unitSize[0] = 100;
        unitObjs[0] = 4000;
        unitSize[1] = 300;
        unitObjs[1] = 2000;
        unitSize[2] = 1200;
        unitObjs[2] = 800;
        unitSize[3] = 1800;
        unitObjs[3] = 100;
        unitSize[4] = 8 * 1024;
        unitObjs[4] = 40;
    }
    else
    {
        unitSize[0] = 256;
        unitObjs[0] = 1024;
        unitSize[1] = 2048;
        unitObjs[1] = 256;
        unitSize[2] = 4 * 1024;
        unitObjs[2] = 32;
        unitSize[3] = 8 * 1024;
        unitObjs[3] = 4;
        unitSize[4] = 24 * 1024;
        unitObjs[4] = 1;
    }

    bool ret = TMultiUnitHeap::Initialize(s_pMemory, total, unitSize, unitObjs, LEVEL, emergencyMemorySize);
    PIA_SAMPLE_ASSERTMSG(ret, "InitializeMultiUnitHeap() failed.");
    NN_PIA_DUMMY_PARAM(ret);
}

//-----------------------------------------------------------------------------------------
// Finalize the NEX heap (MultiUnitHeap).
//-----------------------------------------------------------------------------------------
void FinalizeMultiUnitHeap(void)
{
    size_t memSize = Nex_GetMultiUnitHeapPeakMemorySize();
    if (memSize > 0)
    {
        PIA_SAMPLE_LOG("MultiUnitHeap peak memory usage: %d bytes (%d KB)\n", memSize, memSize / 1024);
    }
    else
    {
        PIA_SAMPLE_LOG("Perhaps MultiUnitHeap memory peak recording feature is disabled.\n");
    }

    TMultiUnitHeap::Finalize();

    delete[] s_pMemory;
}
#endif

//-----------------------------------------------------------------------------------------
// Application-defined memory allocation function
//-----------------------------------------------------------------------------------------
void* AllocFunction(size_t size)
{
    void* p = InnerAlloc(size);

    if (p == NULL)
    {
        PIA_SAMPLE_LOG("AllocFunction: No enough memory. Reserved memory will be used. Please shutdown NEX library.\n");
        ReleaseReservedMemory();

        p = InnerAlloc(size);
        PIA_SAMPLE_ASSERT(p != NULL);
        if (p == NULL)
        {
            PIA_SAMPLE_PANIC("Reserved memory is not enough");
        }
    }

    return p;
}

//-----------------------------------------------------------------------------------------
// Application-defined memory deallocation function
//-----------------------------------------------------------------------------------------
void FreeFunction(void* pointer)
{
    InnerFree(pointer);
}

//-----------------------------------------------------------------------------------------
// Allocate memory for an emergency.
//-----------------------------------------------------------------------------------------
void AcquireReservedMemory()
{
    if (s_pEmergencyMemory == NULL)
    {
        PIA_SAMPLE_LOG("AcquireReservedMemory\n");

        size_t nowUsage = s_MemUsage;
        size_t nowMaxUsage = s_MemMaxUsage;

        s_pEmergencyMemory = AllocFunction(s_EmergencyMemorySize);

        s_MemUsage = nowUsage;
        s_MemMaxUsage = nowMaxUsage;
    }
}

//-----------------------------------------------------------------------------------------
//Deallocate memory for an emergency.
//-----------------------------------------------------------------------------------------
void ReleaseReservedMemory()
{
    if (s_pEmergencyMemory)
    {
        PIA_SAMPLE_LOG("ReleaseReservedMemory\n");

        size_t nowUsage = s_MemUsage;
        size_t nowMaxUsage = s_MemMaxUsage;

        FreeFunction(s_pEmergencyMemory);

        s_pEmergencyMemory = NULL;

        s_MemUsage = nowUsage;
        s_MemMaxUsage = nowMaxUsage;
    }
}

//-----------------------------------------------------------------------------------------
// Internal memory function called from AllocFunction
//-----------------------------------------------------------------------------------------
void* InnerAlloc(size_t size)
{
#if NN_PIA_CTR
    void* pBuffer = s_NexHeap.Allocate(size);
#elif NN_PIA_WIN
    void* pBuffer = malloc(size);
#elif NN_PIA_CAFE
    void* pBuffer = MEMAllocFromExpHeapEx(s_NexHeap, size, 8);
#endif
    if (pBuffer == NULL)
    {
        return NULL;
    }

#if NN_PIA_CTR
    s_MemUsage += s_NexHeap.GetSizeOf(pBuffer);
    if (s_MemUsage > s_MemMaxUsage)
    {
        s_MemMaxUsage = s_MemUsage;
    }
#elif NN_PIA_CAFE
    s_MemUsage += MEMGetSizeForMBlockExpHeap(pBuffer);
    if (s_MemUsage > s_MemMaxUsage)
    {
        s_MemMaxUsage = s_MemUsage;
    }
#endif

    return pBuffer;
}

//-----------------------------------------------------------------------------------------
// Internal memory function called from FreeFunction
//-----------------------------------------------------------------------------------------
void InnerFree(void* ptr)
{
#if NN_PIA_CTR
    s_MemUsage -= s_NexHeap.GetSizeOf(ptr);
    s_NexHeap.Free(ptr);
#elif NN_PIA_CAFE
    s_MemUsage -= MEMGetSizeForMBlockExpHeap(ptr);
    MEMFreeToExpHeap(s_NexHeap, ptr);
#elif NN_PIA_WIN
    free(ptr);
#endif
}

}  // Namespace.

#endif  // ENABLE_SERVERLESS
