﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SndmicCtr.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*
 * Includes
 *---------------------------------------------------------------------------*/
#include "SndmicCtr.h"

#include "SampleFramework/include/SampleFramework.h"
#include "Misc.h"


/*---------------------------------------------------------------------------*
    Const
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
 *  Define
 *---------------------------------------------------------------------------*/
#define ADD_BUFFER_NUM (2)                               // Number of the added buffer.
#define SOUND_FREQ_8KHZ (NN_SND_HW_I2S_CLOCK_32KHZ / 4)  //  8 kHz sampling (8182).
//#define FREQ8KHZ_TIME_NS        (1000000000/SOUND_FREQ_8KHZ)    // 122217.646354288786 nsec
#define FREQ8KHZ_TIME_NS (1000000000 / SOUND_FREQ_8KHZ - 0.89)  // Fine adjustment. Not synchronized with the microphone sampling.
#define MIC_CAPTURE_MARGIN_MS (12)
#define MIC_CAPTURE_MARGIN_SAMPLE (MIC_CAPTURE_MARGIN_MS* nn::pia::chat::VoiceProtocol::AUDIO_DATA_SIZE_PER_MILLI_SEC / 2)
#define ROMFILE_WAVE_0 (L"rom:/noise.wav")
#define ROMFILE_WAVE_1 (L"rom:/busy.wav")
#define ROMFILE_WAVE_2 (L"rom:/kart_title.wav")

/*---------------------------------------------------------------------------*
 * Global Variables
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
    Prototype
 *---------------------------------------------------------------------------*/
void SoundThreadFunc(SndMicDevice* sndmicDev);
void FrameProcessThreadFunc(SndMicDevice* sndmicDev);


nn::pia::chat::VoiceProtocol* s_pVoiceProtocol = NULL;


/*---------------------------------------------------------------------------*
  Name: Initialize
  Description: Initialize Snd and Mic Device
  Arguments: *config      - sndmic config parameter.
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SNDMIC_ERROR_PARAM - Parameter error.
              :   SND_ERROR_DSP      - DSP error.
              :   MIC_ERROR_LIB      - mic library error.
              :   SND_ERROR_LIB      - snd library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Initialize(SndMicConfigStructure* config, nn::pia::chat::VoiceProtocol* pVoiceProtocol)
{
    nn::pia::sampleframework::SampleFramework* pFw = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::Result Result;
    nn::snd::MixParam Mix;
    nn::fs::FileReader fileReader;
    WaveFmt fmt;

    /* Check the arguments. */
    if ((pFw == NULL) || (config == NULL))
    {
        return SNDMIC_ERROR_PARAM;
    }

    // After checking the arguments, set a pointer to an instance of VoiceProtocol.
    if (pVoiceProtocol == NULL)
    {
        return SNDMIC_ERROR_PARAM;
    }
    SAMPLE_ASSERT(s_pVoiceProtocol == NULL);
    s_pVoiceProtocol = pVoiceProtocol;

    /* Config parameter copy. */
    m_Config.frame_msec = config->frame_msec;
    m_Config.mic_gain = config->mic_gain;

    /* Frame Size... */
    m_AudioFrameSize = m_Config.frame_msec * 8 * sizeof(s16);

    if (m_AudioFrameSize < 1056)
    {
        m_SoundBufferSize = 1056;
    }
    else
    {
        m_SoundBufferSize = m_AudioFrameSize;
    }

    /* Initialize Variables... */
    m_SoundStatus = SND_PROCESS_NONE;
    m_CurrBufferIdx = 0;
    m_SoundBufferReqIdx = 0;
    m_MicTimingError = false;
    m_Pause = 0;

    m_MicStatus = MIC_PROCESS_NONE;

    m_MicSyncEvent.Initialize(false);
    m_MicSyncEvent.ClearSignal();
    m_MicStopEvent.Initialize(false);
    m_MicStopEvent.ClearSignal();

    /**********   MIC Initialization   **********/
    if (nn::mic::Initialize().IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* MIC-Buffer == 4096*2byte */
    if (nn::mic::SetBuffer(m_MicBuffer, sizeof(m_MicBuffer)).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Set MIC-Amp. */
    if (nn::mic::GetAmpGain(&m_MicGain).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    if (nn::mic::SetAmpGain(m_Config.mic_gain).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Power On MIC. */
    if (nn::mic::SetAmp(true).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /***** Update process status *****/
    m_MicStatus = MIC_PROCESS_INIT;


    /**********   SND Initialization   **********/
    /* Initialize DSP and load the DSP component. */
    Result = nn::dsp::Initialize();
    if (Result.IsFailure())
    {
        return SND_ERROR_DSP;
    }

    Result = nn::dsp::LoadDefaultComponent();
    if (Result.IsFailure())
    {
        return SND_ERROR_DSP;
    }

    Result = nn::snd::Initialize();
    if (Result.IsFailure())
    {
        return SND_ERROR_LIB;
    }


    /* Prepare <tt>SoundBuffer</tt>. */
    for (int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++)
    {
        /* Allocate SoundBuffer (as described in Programming Manual, SoundBuffer must be at least 1025 bytes) */
        m_SoundBuffer[Idx] = reinterpret_cast<u8*>(pFw->AllocateDeviceMemory(m_SoundBufferSize));
        std::memset(m_SoundBuffer[Idx], 0x0, m_SoundBufferSize);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(m_SoundBuffer[Idx]), m_SoundBufferSize);
    }

    /* Prepare <tt>WaveBuffer</tt>. */
    for (int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++)
    {
        /* Register to WaveBuffer. */
        nn::snd::InitializeWaveBuffer(&m_WaveBuffer[Idx]);
        m_WaveBuffer[Idx].bufferAddress = reinterpret_cast<void*>(m_SoundBuffer[Idx]);
        m_WaveBuffer[Idx].sampleLength = nn::snd::GetSampleLength(m_AudioFrameSize, nn::snd::SAMPLE_FORMAT_PCM16, 1);
        m_WaveBuffer[Idx].loopFlag = false;
    }

    /* Prepare the volume mix. */
    Mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT] = 1.0f;
    Mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 1.0f;

    /* Prepare voice. */
    m_pVoice = nn::snd::AllocVoice(128, NULL, NULL);
    m_pVoice->SetSampleRate(SOUND_FREQ_8KHZ);                /* Set the sampling rate (8 kHz).  */
    m_pVoice->SetSampleFormat(nn::snd::SAMPLE_FORMAT_PCM16); /* Specify the format (PCM16).       */
    m_pVoice->SetChannelCount(1);                            /* Set the number of channels (mono).          */
    m_pVoice->SetPitch(1.0f);                                /* Set the pitch.                    */
                                                             //    m_pVoice->SetVolume( 10.0f );                              /* Set the volume.                */
    m_pVoice->SetVolume(1.0f);                               /* Volume settings.                */
    m_pVoice->SetMixParam(Mix);                              /* Set the volume mix.           */
    for (int Idx = 0; Idx < ADD_BUFFER_NUM; Idx++)
    {
        m_pVoice->AppendWaveBuffer(&m_WaveBuffer[Idx]); /* Add the buffer.                  */
    }

    /* Prepare background music. */
    /* Load background music data. */
    Result = fileReader.TryInitialize(ROMFILE_WAVE_2);
    NN_UTIL_PANIC_IF_FAILED(Result);
    m_pWaveMemory[2] = reinterpret_cast<u8*>(pFw->AllocateDeviceMemory(GetWaveLength(&fileReader) + m_AudioFrameSize));
    if (m_pWaveMemory[2] == NULL)
    {
        SAMPLE_PRINT("Error:Allocate memory error.\n");
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[2].buf = m_pWaveMemory[2];
    LoadWave(&fileReader, &fmt, &m_WaveData[2]);
    fileReader.Finalize();

    nn::snd::FlushDataCache(reinterpret_cast<uptr>(m_WaveData[2].buf), m_WaveData[2].size);

    /* Register to WaveBuffer. */
    nn::snd::InitializeWaveBuffer(&m_MusicWaveBuffer);
    m_MusicWaveBuffer.bufferAddress = reinterpret_cast<void*>(m_WaveData[2].buf);
    m_MusicWaveBuffer.sampleLength = nn::snd::GetSampleLength(m_WaveData[2].size, nn::snd::SAMPLE_FORMAT_PCM16, 1);
    m_MusicWaveBuffer.loopFlag = true;

    /* Prepare the voice for background music. */
    m_pMusic = nn::snd::AllocVoice(128, NULL, NULL);
    m_pMusic->SetSampleRate(SOUND_FREQ_8KHZ);                /* Set the sampling rate (8 kHz).  */
    m_pMusic->SetSampleFormat(nn::snd::SAMPLE_FORMAT_PCM16); /* Specify the format (PCM16).       */
    m_pMusic->SetChannelCount(1);                            /* Set the number of channels (mono).        */
    m_pMusic->SetPitch(1.0f);                                /* Set the pitch.                    */
    m_pMusic->SetVolume(0.2f);                               /* Volume settings.                */
    m_pMusic->SetMixParam(Mix);                              /* Set the volume mix.           */
    m_pMusic->AppendWaveBuffer(&m_MusicWaveBuffer);          /* Add the buffer.                  */

    /* Set the master volume. */
    nn::snd::SetMasterVolume(1.0f);

    /* Load noise data. */
    Result = fileReader.TryInitialize(ROMFILE_WAVE_0);
    NN_UTIL_PANIC_IF_FAILED(Result);
    m_pWaveMemory[0] = reinterpret_cast<u8*>(pFw->AllocateDeviceMemory(GetWaveLength(&fileReader) + m_AudioFrameSize));
    if (m_pWaveMemory[0] == NULL)
    {
        SAMPLE_PRINT("Error:Allocate memory error.\n");
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[0].buf = m_pWaveMemory[0];
    LoadWave(&fileReader, &fmt, &m_WaveData[0]);
    fileReader.Finalize();
    m_FrameCount[0] = static_cast<u32>((m_WaveData[0].size + m_AudioFrameSize - 1) / m_AudioFrameSize);

    /* Load busy data. */
    Result = fileReader.TryInitialize(ROMFILE_WAVE_1);
    NN_UTIL_PANIC_IF_FAILED(Result);
    m_pWaveMemory[1] = reinterpret_cast<u8*>(pFw->AllocateDeviceMemory(GetWaveLength(&fileReader) + m_AudioFrameSize));
    if (m_pWaveMemory[1] == NULL)
    {
        SAMPLE_PRINT("Error:Allocate memory error.\n");
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[1].buf = m_pWaveMemory[1];
    LoadWave(&fileReader, &fmt, &m_WaveData[1]);
    fileReader.Finalize();
    m_FrameCount[1] = static_cast<u32>((m_WaveData[1].size + m_AudioFrameSize - 1) / m_AudioFrameSize);

    m_FramePlayReq = false;

    /***** Update process status *****/
    m_SoundStatus = SND_PROCESS_INIT;


    /****************************** Thread Registration *************************/
    /* Register the DSP callback function. */
    m_SoundThread.StartUsingAutoStack(SoundThreadFunc, this, (int)4096, (int)2);

    /* Register the thread for processing frame data. */
    m_FrameProcessThread.StartUsingAutoStack(FrameProcessThreadFunc, this, (int)4096, (int)(NN_OS_DEFAULT_THREAD_PRIORITY - 2));

    m_bChatSendReceiveReady = false;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: Start
  Description: Start playing the stream.
  Arguments: -
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Start(void)
{
    nn::Result Result;

    /* Start the mic using the FrameProcess thread. */
    /* Update the status when changing from STOP to START to match the start timing. */
    if (m_MicStatus == MIC_PROCESS_STOP)
    {
        m_MicStatus = MIC_PROCESS_INIT;
    }


    /**********   SND Start **********/
    /* Check the process status. */
    if ((m_SoundStatus != SND_PROCESS_INIT) && (m_SoundStatus != SND_PROCESS_STOP))
    {
        return SND_ERROR_SEQ;
    }

    /* Start playback. */
    m_pVoice->SetState(nn::snd::Voice::STATE_PLAY);

    /* Update the process status. */
    m_SoundStatus = SND_PROCESS_PLAY;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: Stop
  Description: Stop playing the stream.
  Arguments: -
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Stop(void)
{

    /**********   MIC Stop   **********/
    /* Check the process status. */
    if (m_MicStatus != MIC_PROCESS_START)
    {
        return MIC_ERROR_SEQ;
    }

    /* Update the process status. */
    m_MicStatus = MIC_PROCESS_STOP;
    /* Wait until processing ends. */
    m_MicStopEvent.Wait();


    /**********   SND Stop   **********/
    /* Check the process status. */
    if (m_SoundStatus != SND_PROCESS_PLAY)
    {
        return SND_ERROR_SEQ;
    }

    /* End playback. */
    if (m_pVoice->GetState() == nn::snd::Voice::STATE_PLAY)
    {
        m_pVoice->SetState(nn::snd::Voice::STATE_STOP);
    }

    /* Update the process status. */
    m_SoundStatus = SND_PROCESS_STOP;


    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: SetVolume
  Description: Set the voice volume.
  Arguments: volume - Value of the volume.
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetVolume(f32 volume)
{
    nn::Result Result;

    /* Check the process status. */
    if (m_SoundStatus != SND_PROCESS_PLAY)
    {
        return SND_ERROR_SEQ;
    }

    /* Set Volume... */
    if (m_pVoice != NULL)
    {
        m_pVoice->SetVolume(volume);
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: SetBGMVolume
  Description: Set BGM volume.
  Arguments: volume - Value of the BGM volume.
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetBGMVolume(f32 volume)
{
    nn::Result Result;

    /* Check the process status. */
    if (m_SoundStatus != SND_PROCESS_PLAY)
    {
        return SND_ERROR_SEQ;
    }

    /* Set Volume... */
    if (m_pMusic != NULL)
    {
        m_pMusic->SetVolume(volume);
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: StartBGM
  Description: Start playing BGM.
  Arguments: -
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StartBGM(void)
{
    nn::Result Result;

    /* Check the process status. */
    if (m_SoundStatus != SND_PROCESS_PLAY)
    {
        return SND_ERROR_SEQ;
    }

    /* Start background music playback. */
    m_pMusic->SetState(nn::snd::Voice::STATE_PLAY);

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: StopBGM
  Description: Stop playing BGM.
  Arguments: -
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StopBGM(void)
{
    nn::Result Result;

    /* Check the process status. */
    if (m_SoundStatus != SND_PROCESS_PLAY)
    {
        return SND_ERROR_SEQ;
    }

    /* Stop background music playback */
    if (m_pMusic->GetState() == nn::snd::Voice::STATE_PLAY)
    {
        m_pMusic->SetState(nn::snd::Voice::STATE_PAUSE);
    }

    return SNDMIC_ERROR_NONE;
}


bool SndMicDevice::IsPlayingBGM(void) const
{
    return m_pMusic->GetState() == nn::snd::Voice::STATE_PLAY;
}


/*---------------------------------------------------------------------------*
  Name: Finalize
  Description: Finalize the snd and mic device.
  Arguments: -
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - No error.
              :   SND_ERROR_SEQ   - snd sequence error.
              :   SNDMIC_ERROR_PARAM - Parameter error.
              :   SND_ERROR_DSP      - DSP error.
              :   SND_ERROR_LIB      - snd library error.
              :   MIC_ERROR_SEQ   - Mic sequence error.
              :   MIC_ERROR_LIB      - mic library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Finalize(void)
{
    nn::pia::sampleframework::SampleFramework* pFw = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::Result Result;

    /**********   MIC Finalizing   **********/
    /* Check the process status. */
    if ((m_MicStatus != MIC_PROCESS_STOP) && (m_MicStatus != MIC_PROCESS_INIT))
    {
        return MIC_ERROR_SEQ;
    }

    /* Power off the mic. */
    if (nn::mic::SetAmp(false).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Return the mic amp setting. */
    if (nn::mic::SetAmpGain(m_MicGain).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Free the mic buffer. */
    if (nn::mic::ResetBuffer().IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Finalize MIC. */
    if (nn::mic::Finalize().IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    /* Update the process status. */
    m_MicStatus = MIC_PROCESS_NONE;

    /**********   Finalize the frame data thread   **********/
    m_FrameProcessThread.Join();
    m_FrameProcessThread.Finalize();


    /**********   SND Finalizing   **********/
    /* Check the process status. */
    if ((m_SoundStatus != SND_PROCESS_STOP) && (m_SoundStatus != SND_PROCESS_INIT))
    {
        return SND_ERROR_SEQ;
    }

    /* Update the process status. */
    m_SoundStatus = SND_PROCESS_NONE;

    /* Destroy the sound thread. */
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    /* Destroy voice. */
    nn::snd::FreeVoice(m_pVoice);
    nn::snd::FreeVoice(m_pMusic);

    /* Destroy the sound buffer. */
    for (int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++)
    {
        pFw->FreeDeviceMemory(m_SoundBuffer[Idx]);
    }

    /* Finalize SND. */
    Result = nn::snd::Finalize();
    if (Result.IsFailure())
    {
        return SND_ERROR_LIB;
    }

    /* Finalize DSP. */
    Result = nn::dsp::UnloadComponent();
    if (Result.IsFailure())
    {
        return SND_ERROR_DSP;
    }
    nn::dsp::Finalize();


    if (m_pWaveMemory[0] != NULL)
    {
        pFw->FreeDeviceMemory(m_pWaveMemory[0]);
        m_pWaveMemory[0] = NULL;
    }

    if (m_pWaveMemory[1] != NULL)
    {
        pFw->FreeDeviceMemory(m_pWaveMemory[1]);
        m_pWaveMemory[1] = NULL;
    }

    if (m_pWaveMemory[2] != NULL)
    {
        pFw->FreeDeviceMemory(m_pWaveMemory[2]);
        m_pWaveMemory[2] = NULL;
    }


    m_MicStopEvent.Finalize();
    m_MicSyncEvent.Finalize();

    s_pVoiceProtocol = NULL;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: SoundThreadFuncImpl
  Description: Sound thread function for sending the parameter to DSP (Core).
  Arguments: -
  Returns: -
 *---------------------------------------------------------------------------*/
void SndMicDevice::SoundThreadFuncImpl(void)
{
    bool store_req = false;

    while (true)
    {
        /* Wait for data from DSP. */
        nn::snd::WaitForDspSync();

        if (m_SoundStatus == SND_PROCESS_NONE)
        {
            break;
        }

        if (m_SoundStatus == SND_PROCESS_PLAY)
        {
            /* Triggering MIC here to make sure that SND and MIC are started at a specific timing as much as possible */
            if (m_MicStatus == MIC_PROCESS_INIT)
            {
                m_MicSyncEvent.Signal();
            }

            /* Check the status of WaveBuffer. If playback is done, add the buffer. */
            if ((m_pVoice != NULL) &&
                (m_WaveBuffer[m_CurrBufferIdx].status == nn::snd::WaveBuffer::STATUS_DONE) &&
                (m_WaveBuffer[(m_CurrBufferIdx + ADD_BUFFER_NUM) % SOUND_BUFFER_NUM].status == nn::snd::WaveBuffer::STATUS_FREE))
            {
                /* Re-add the buffer. */
                m_pVoice->AppendWaveBuffer(&m_WaveBuffer[(m_CurrBufferIdx + ADD_BUFFER_NUM) % SOUND_BUFFER_NUM]);
                store_req = true;
                m_SoundBufferReqIdx = m_CurrBufferIdx;
                /* Update CurrBufferIndx (the index of the current buffer). */
                m_CurrBufferIdx = (m_CurrBufferIdx + 1) % SOUND_BUFFER_NUM;
            }

            if (store_req)
            {
                /* Get playback data. */
                if (GetReceiveAudioData(m_SoundBuffer[m_SoundBufferReqIdx], m_AudioFrameSize))
                {
                    if (m_FramePlayReq)
                    {
                        if (m_FramePlayNum < m_FrameCount[m_WaveNumbber])
                        {
                            std::memcpy(m_SoundBuffer[m_SoundBufferReqIdx], m_pPlayWave, m_AudioFrameSize);
                            m_FramePlayNum++;
                            m_pPlayWave += m_AudioFrameSize;
                        }
                        else
                        {
                            m_FramePlayReq = false;
                        }
                    }

                    /* Flush the cache. */
                    nn::snd::FlushDataCache(reinterpret_cast<uptr>(m_SoundBuffer[m_SoundBufferReqIdx]), m_AudioFrameSize);

                    /* Re-add the buffer. */
                    nn::snd::InitializeWaveBuffer(&m_WaveBuffer[m_SoundBufferReqIdx]);
                    m_WaveBuffer[m_SoundBufferReqIdx].bufferAddress =
                        reinterpret_cast<void*>(m_SoundBuffer[m_SoundBufferReqIdx]);
                    m_WaveBuffer[m_SoundBufferReqIdx].sampleLength =
                        nn::snd::GetSampleLength(m_AudioFrameSize, nn::snd::SAMPLE_FORMAT_PCM16, 1);
                    m_WaveBuffer[m_SoundBufferReqIdx].loopFlag = false;

                    store_req = false;
                }
            }

            /* Send reference mixed data... */
            if (nn::snd::GetMixedBusData(m_MixedData, NN_SND_SAMPLES_PER_FRAME))
            {
                s_pVoiceProtocol->HandleReference((u8*)m_MixedData, NN_SND_SAMPLES_PER_FRAME * 2 * sizeof(s16));
            }
        }
        /* Send arguments to DSP. */
        nn::snd::SendParameterToDsp();
    }
}


/*---------------------------------------------------------------------------*
  Name: RomfileWavePlay
  Description: Play Romfile wave data.
  Arguments: number - Wave data number.
  Returns: -
 *---------------------------------------------------------------------------*/
void SndMicDevice::RomfileWavePlay(u32 number)
{
    m_FramePlayReq = true;
    m_FramePlayNum = 0;
    m_WaveNumbber = number;
    m_pPlayWave = m_WaveData[number].buf;
}


/*---------------------------------------------------------------------------*
  Name: SoundThreadFunc
  Description: Sound thread function for sending the parameter to DSP (Wrapper).
  Arguments: sndmicDev - Pointer to the SndMicDevice class.
  Returns: -
 *---------------------------------------------------------------------------*/
void SoundThreadFunc(SndMicDevice* sndmicDev)
{
    sndmicDev->SoundThreadFuncImpl();
}


/*---------------------------------------------------------------------------*
  Name: MicGainSet
  Description: Change the mic gain.
  Arguments: gain - Set the mic gain value.
  Returns: SNDMIC_ERROR_STATUS
              :  SNDMIC_ERROR_NONE  - No error.
              :  MIC_ERROR_SEQ   - Sequence error.
              :  MIC_ERROR_LIB   - Library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::MicGainSet(u8 gain)
{
    /* Check the process status. */
    if (m_MicStatus == MIC_PROCESS_NONE)
    {
        return MIC_ERROR_SEQ;
    }

    /* Set the mic gain. */
    if (nn::mic::SetAmpGain(gain).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: MicGainGet
  Description: Get the mic gain.
  Arguments: *gain - Set the mic gain value.
  Returns: SNDMIC_ERROR_STATUS
              :  SNDMIC_ERROR_NONE  - No error.
              :  MIC_ERROR_SEQ   - Sequence error.
              :  MIC_ERROR_LIB   - Library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::MicGainGet(u8* gain) const
{
    /* Check the process status. */
    if (m_MicStatus == MIC_PROCESS_NONE)
    {
        return MIC_ERROR_SEQ;
    }

    if (nn::mic::GetAmpGain(gain).IsFailure())
    {
        return MIC_ERROR_LIB;
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name: GetReceiveAudioData
  Description: Get ReceiveAudio data.
  Arguments: *gain - Set the mic gain value.
  Returns: true    Success.
              : false   No data.
 *---------------------------------------------------------------------------*/
bool SndMicDevice::GetReceiveAudioData(u8* audioBuf, u16 length)
{
    /* Check the index. */
    if (m_SndBufReadIdx == m_SndBufWriteIdx)
    {
        return false;
    }
    else
    {
        for (int i = 0; i < length; i++)
        {
            audioBuf[i] = m_SndBuf[m_SndBufReadIdx][i];
        }
        m_SndBufReadIdx = (m_SndBufReadIdx + 1) % 4;

        return true;
    }
}


/*---------------------------------------------------------------------------*
  Name: GetSndMicErrorFlag
  Description: Get the SndMic error flag.
  Arguments: 
  Returns: true    An error occurred.
              : false   No error.
 *---------------------------------------------------------------------------*/
bool SndMicDevice::GetSndMicErrorFlag(void)
{
    bool ErrorFlag;

    if (m_MicTimingError)
    {
        ErrorFlag = true;
        m_MicTimingError = false;
    }
    else
    {
        ErrorFlag = false;
    }

    return ErrorFlag;
}


/*---------------------------------------------------------------------------*
  Name: Pause
  Description: Pause or restart.
  Arguments: -
  Returns: 0 Restart.
              : 1 Paused.
 *---------------------------------------------------------------------------*/
u8 SndMicDevice::Pause(void)
{
    m_Pause ^= 1;
    if (m_Pause)
    {
        int i;
        for (i = 0; i < 4; i++)
        {
            std::memset(&m_SndBuf[i], 0, nn::pia::chat::VoiceProtocol::AUDIO_DATA_SIZE_MAX);
        }
    }
    return m_Pause;
}


void SndMicDevice::Dispatch(void)
{
}


void SndMicDevice::sendWrapper(void* pData, size_t samples, u32 offset)
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(s_pVoiceProtocol));

    if (s_pVoiceProtocol->IsInCommunication())
    {
        s_pVoiceProtocol->Send(pData, samples, offset);
    }
}


void SndMicDevice::receiveWrapper(void* pData, u32 frameDataSize, nn::pia::StationIndex stationIndexArray[])
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(s_pVoiceProtocol));

    if (s_pVoiceProtocol->IsInCommunication())
    {
        s_pVoiceProtocol->Receive(pData, frameDataSize, stationIndexArray);
    }
}


/*---------------------------------------------------------------------------*
  Name: FrameProcessThreadImpl
  Description: Frame process thread function (Core).
  Arguments: -
  Returns: -
 *---------------------------------------------------------------------------*/
void SndMicDevice::FrameProcessThreadImpl(void)
{
    uptr Position;
    u32 micOffset;
    nn::Result Result;
    nn::pia::StationIndex outStationIndexArray[nn::pia::chat::VoiceProtocol::MAX_AUDIO_STREAM];
    //    char zero_buf[192*8*2];

    nn::os::Timer timer(false);

    m_SndBufWriteIdx = 0;
    m_SndBufReadIdx = 0;


    while (true)
    {
        if (m_MicStatus == MIC_PROCESS_NONE)
        {
            break;
        }

        /**********   MIC Start   **********/
        /* Check the process status. */
        if (m_MicStatus != MIC_PROCESS_INIT)
        {
            SAMPLE_PRINT("MIC_ERROR_SEQ !!\n");
            break;
        }

        /* Wait for the signal from the sound thread. */
        m_MicSyncEvent.Wait();

        /* Set the half position. */
        m_MicBuffHalfPosition = (uptr)m_MicBuffer + m_AudioFrameSize;
        m_MicPrevPosition = 1;

        /* Start sampling. */
        Result = nn::mic::StartSampling(
            nn::mic::SAMPLING_TYPE_SIGNED_16BIT,
            nn::mic::SAMPLING_RATE_8180,
            0,
            m_AudioFrameSize * 2,
            true);

        /* StartSampling error check. */
        if (Result.IsFailure())
        {
            /* Continue processing because control can reach this point with the system closed.     */
            SAMPLE_PRINT("StartSampling error !!\n");
            m_MicTimingError = true;
        }
        else
        {
            m_MicTimingError = false;
        }

        /* Update the process status. */
        m_MicStatus = MIC_PROCESS_START;


        timer.StartPeriodic(nn::fnd::TimeSpan::FromNanoSeconds(FREQ8KHZ_TIME_NS * 8 * m_Config.frame_msec),
                            nn::fnd::TimeSpan::FromNanoSeconds(FREQ8KHZ_TIME_NS * 8 * m_Config.frame_msec));

        while ((m_MicStatus == MIC_PROCESS_START) && (m_SoundStatus == SND_PROCESS_PLAY))
        {
            timer.Wait(); /* Start for each sound frame. */

            if (m_MicTimingError)
            {
                /* Restart mic sampling. */
                /* Set the half position. */
                m_MicBuffHalfPosition = (uptr)m_MicBuffer + m_AudioFrameSize;
                m_MicPrevPosition = 0;

                /* Start sampling. */
                Result = nn::mic::StartSampling(
                    nn::mic::SAMPLING_TYPE_SIGNED_16BIT,
                    nn::mic::SAMPLING_RATE_8180,
                    0,
                    m_AudioFrameSize * 2,
                    true);

                /* StartSampling error check. */
                if (Result.IsFailure())
                {
                    /* Continue processing because control can reach this point with the system closed.     */
                    SAMPLE_PRINT("StartSampling error !!\n");
                    m_MicTimingError = true;
                }
                else
                {
                    m_MicTimingError = false;
                }
            }

            /* The timer is not synchronized with mic sampling. Keep a margin when getting the mic. */
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(MIC_CAPTURE_MARGIN_MS));

            /* Get the mic sampling position. */
            if (nn::mic::GetLastSamplingAddress(&Position).IsFailure())
            {
                SAMPLE_PRINT("Cannot get Mic address!!\n");
                //break;
            }

            /* Get mic data. */
            nn::pia::chat::VoiceProtocol* pChat = s_pVoiceProtocol;
            if (m_MicPrevPosition == 1)
            {
                if ((((Position + 2) >= m_MicBuffHalfPosition) && ((Position + 2) < (m_MicBuffHalfPosition + m_AudioFrameSize))) && (m_MicTimingError == false))
                {
                    /* Handle mic data. */
                    micOffset = static_cast<u32>(Position + 2 - m_MicBuffHalfPosition) / 2;
                    if (m_Pause == 0)
                    {
                        if (pChat)
                        {
                            void* pData = NULL;
                            size_t sz = 0;
                            u32 offset = 0;
#if USE_VALIDATION_MODE
                            if (Validation::isPlayTriangleWave())
                            {
                                pData = Validation::getTriangleWaveData();
                                sz = Validation::TRIANGLE_WAVE_DATA_SAMPLING_NUM;
                                offset = 0;
                            }
                            else
#endif
                            {

                                pData = (void*)m_MicBuffer;
                                sz = m_AudioFrameSize / sizeof(s16);
                                offset = micOffset;
                            }
                            sendWrapper(pData, sz, offset);
                        }
                    }
                    m_MicPrevPosition = 0;
                }
                else
                {
/* The timer is not synchronized with mic sampling, so control can reach this point. */
/* Control also reaches this point when the system is closed.     */
#if 1
                    SAMPLE_PRINT("Wrong Mic Position 1 !!\n");
#endif
                    /* Pass muted data. */
                    micOffset = MIC_CAPTURE_MARGIN_SAMPLE;
                    std::memset(m_zero_buf, 0x0, m_AudioFrameSize);
                    if (m_Pause == 0)
                    {
                        if (pChat)
                        {
                            sendWrapper(m_zero_buf, m_AudioFrameSize / sizeof(s16), micOffset);
                        }
                    }

                    /* This position is weird. Restart mic sampling. */
                    m_MicTimingError = true;
                }
            }
            else
            {
                /* m_MicPrevPosition == 0  */
                if ((((Position + 2) < m_MicBuffHalfPosition) || ((Position + 2) == (m_MicBuffHalfPosition + m_AudioFrameSize))) && (m_MicTimingError == false))
                {
                    /* Handle mic data. */
                    micOffset = static_cast<u32>(Position + 2 - reinterpret_cast<u32>(m_MicBuffer)) / 2;
                    if ((Position + 2) == (m_MicBuffHalfPosition + m_AudioFrameSize))
                    {
                        micOffset = 0;
                    }
                    if (m_Pause == 0)
                    {
                        if (pChat)
                        {
                            void* pData = NULL;
                            size_t sz = 0;
                            u32 offset = 0;
#if USE_VALIDATION_MODE
                            if (Validation::isPlayTriangleWave())
                            {
                                pData = Validation::getTriangleWaveData();
                                sz = Validation::TRIANGLE_WAVE_DATA_SAMPLING_NUM;
                                offset = 0;
                            }
                            else
#endif
                            {
                                pData = (void*)m_MicBuffHalfPosition;
                                sz = m_AudioFrameSize / sizeof(s16);
                                offset = micOffset;
                            }

                            sendWrapper(pData, sz, offset);
                        }
                    }
                    m_MicPrevPosition = 1;
                }
                else
                {
/* The timer is not synchronized with mic sampling, so control can reach this point. */
/* Control also reaches this point when the system is closed.     */
#if 1
                    SAMPLE_PRINT("Wrong Mic Position 0 !!\n");
#endif
                    /* Pass muted data. */
                    micOffset = MIC_CAPTURE_MARGIN_SAMPLE;
                    if (m_Pause == 0)
                    {
                        if (pChat)
                        {
                            sendWrapper(m_zero_buf, m_AudioFrameSize / sizeof(s16), micOffset);
                        }
                    }

                    /* This position is weird. Restart mic sampling. */
                    m_MicTimingError = true;
                }
            }

            /* Set snd data. */
            if (m_Pause == 0)
            {
                //N2050 20130329
                if (pChat)
                {
                    receiveWrapper((void*)m_SndBuf[m_SndBufWriteIdx], m_AudioFrameSize, outStationIndexArray);
                }
            }

            m_SndBufWriteIdx = (m_SndBufWriteIdx + 1) % 4;

            m_bChatSendReceiveReady = true;
        }

        /* Stop sampling. */
        if (nn::mic::StopSampling().IsFailure())
        {
            SAMPLE_PRINT("StopSampling Error!!\n");
        }

        /* Notification when the microphone stops. */
        m_MicStopEvent.Signal();

        /* Wait until the status changes. */
        while ((m_MicStatus != MIC_PROCESS_NONE) && (m_SoundStatus != SND_PROCESS_INIT))
        {
            timer.Wait();
        }
        timer.Stop();
        if (m_MicStatus == MIC_PROCESS_NONE)
        {
            break;
        }
    }

    timer.Finalize();
}


/*---------------------------------------------------------------------------*
  Name: FrameProcessThreadFunc
  Description: Frame process thread function (Wrapper).
  Arguments: sndmicDev - Pointer to the SndMicDevice class.
  Returns: -
 *---------------------------------------------------------------------------*/
void FrameProcessThreadFunc(SndMicDevice* sndmicDev)
{
    sndmicDev->FrameProcessThreadImpl();
}


/* EOF. */
