﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Wave.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
// Includes
#include "Wave.h"

#include "Misc.h"


/*---------------------------------------------------------------------------*/
// Constants

#define MAKE_WAVE_TAG_VALUE(c1, c2, c3, c4) \
    (c1 | (c2 << 8) | (c3 << 16) | (c4 << 24))


/*---------------------------------------------------------------------------*/
// Typedef
typedef struct
{
    u32 tag;
    u32 size;
    u32 type;
} WaveRiff;

typedef struct
{
    u32 tag;
    u32 size;
} WaveChunk;


/*---------------------------------------------------------------------------*/
// Functions

#if NN_PIA_CTR
/*---------------------------------------------------------------------------*
  Name:         LoadWave

  Description:  Load wave data.

  Arguments:    filepath: File path.
                fmt: Store the fmt chunk of the Wave data.
                data: Store the data chunk of the Wave data.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void LoadWaveCore(nn::fs::FileReader* reader, WaveFmt* fmt, WaveData* data, bool isSizeOnly)
{
    SAMPLE_ASSERT(reader);

    reader->SetPosition(0);

    WaveRiff riff;     // RIFF header.
    WaveChunk chunk;   // CHUNK.
    u32 readbyte = 0;  // Used to check file loading.
    u32 fsize = reader->GetSize();

    // Load the RIFF header.
    reader->Read((u8*)&riff, sizeof(WaveRiff));

    // Estimate the size.
    readbyte += sizeof(WaveRiff);

    // Confirm that it matches RIFF.
    if (riff.tag != MAKE_WAVE_TAG_VALUE('R', 'I', 'F', 'F'))
    {
        SAMPLE_PANIC("Not RIFF format.\n");
    }
    // Confirm that it matches WAVE.
    if (riff.type != MAKE_WAVE_TAG_VALUE('W', 'A', 'V', 'E'))
    {
        SAMPLE_PANIC("Not Wave format.\n");
    }

    // Load until it is larger than the file.
    while (fsize > readbyte)
    {
        // Load CHUNK.
        reader->Read((u8*)&chunk, sizeof(WaveChunk));
        readbyte += sizeof(WaveChunk);

        // "fmt"
        if (chunk.tag == MAKE_WAVE_TAG_VALUE('f', 'm', 't', ' '))
        {
            reader->Read((u8*)fmt, sizeof(WaveFmt));
            readbyte += sizeof(WaveFmt);

            // If it is an extended region, read and ignore.
            if (chunk.size > sizeof(WaveFmt))
            {
                u16 ext_size;

                reader->Read((u8*)&ext_size, sizeof(ext_size));
                readbyte += sizeof(ext_size);

                readbyte += chunk.size;
                if (readbyte == fsize)
                    break;

                reader->Seek(ext_size, nn::fs::POSITION_BASE_CURRENT);
            }
        }
        // "data"
        else if (chunk.tag == MAKE_WAVE_TAG_VALUE('d', 'a', 't', 'a'))
        {
            data->size = chunk.size;

            if (isSizeOnly)
            {
                return;
            }
            else
            {
                reader->Read((u8*)data->buf, data->size);
                // If it is an 8-bit WAV file, it must be converted from unsigned to signed.
                if (fmt->quantumBits == 8)
                {
                    s8* p = (s8*)data->buf;
                    for (int i = 0; i < data->size; i++)
                    {
                        p[i] -= 128;
                    }
                }
            }
            readbyte += chunk.size;
        }
        // Other cases.
        else
        {
            readbyte += chunk.size;
            if (readbyte == fsize)
                break;

            // Ignore.
            reader->Seek(chunk.size, nn::fs::POSITION_BASE_CURRENT);
        }
    }
}

#endif

/*---------------------------------------------------------------------------*/
/* End of file. */
/*---------------------------------------------------------------------------*/
