﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Object.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include "ReckoningStrategy.h"
#include "SampleConfig.h"

namespace
{
// Type of the data being shared.
struct StructValue
{
    u32 m_V1;
    f32 m_V2;
#if PIA_TEST_CODE
    enum
    {
        cStructValueSizeMax = 1000 / 4
    };
    u32 m_V3[cStructValueSizeMax];
#endif
};
struct SampleLargeData
{
    static const size_t DATA_SIZE = 100 * 1024;
    u32 dataArray[DATA_SIZE / sizeof(u32)];
    u32 hash;
    void Clear()
    {
        memset(dataArray, 0, sizeof(dataArray));
        hash = 0;
    }
};
}

// Reverse the StructValue bytes so it can be serialized.
template <>
void nn::pia::clone::ReverseSerializePolicy<StructValue>::Reverse(StructValue* pTo, const StructValue* cpFrom)
{
    ReverseValue(&pTo->m_V1, &cpFrom->m_V1);
    ReverseValue(&pTo->m_V2, &cpFrom->m_V2);
#if PIA_TEST_CODE
    for (int i = 0; i < StructValue::cStructValueSizeMax; ++i)
    {
        ReverseValue(&pTo->m_V3[i], &cpFrom->m_V3[i]);
    }
#endif
}

// The base class for the game object that corresponds to Clone.
class Object
{
public:
    Object():m_SampleLargeDataPreHash(0)
    {
    }
    virtual void Operate();
    virtual void Update()
    {
    }
    virtual bool Cleanup()
    {
        return true;
    }
    virtual void RegisterClone() = 0;
    virtual void UnregisterClone() = 0;
    virtual bool IsCloneRegistered() const = 0;
    bool CleanupCore(nn::pia::clone::CloneBase* pClone);
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor)
    {
        NN_PIA_DUMMY_PARAM(cColor);
        return y;
    }
    void PrintValue(int y);
    void ResetProfiler()
    {
        GetClonePtr()->ResetProfiler();
        m_ElementR1.ResetProfiler();
    }
    void TraceProfiling() const;
    virtual bool IsEnableReceive() const
    {
        return true;
    }

    nn::pia::clone::CloneBase::Id m_CloneId;
    nn::pia::clone::UnreliableCloneElement<u64> m_ElementU1;
    nn::pia::clone::UnreliableCloneElement<f32, nn::pia::clone::NetworkByteOrderSerializePolicy<f32> > m_ElementU2;
    nn::pia::clone::UnreliableCloneElement<StructValue, nn::pia::clone::NetworkByteOrderSerializePolicy<StructValue> > m_ElementU3;
    nn::pia::clone::ReliableCloneElement<u32> m_ElementR1;
    nn::pia::clone::EventCloneElement<u32, 8> m_ElementE1;
    nn::pia::reckoning::ReckoningCloneElement<ReckoningStrategy> m_ElementC1;

    nn::pia::clone::ReliableLargeCloneElement<SampleLargeData> m_ElementL1;

    nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> m_SendProfiler;
    nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> m_ReceiveProfiler;

    nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> m_SendProfilerForElement;
    nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> m_ReceiveProfilerForElement;

    static ReckoningStrategy s_ReckoningStrategy;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr() = 0;
    virtual const nn::pia::clone::CloneBase& GetClone() const = 0;
    virtual const char* GetCloneTypeText() const = 0;
    void OperateElement();
    void UpdateElement();
    u32 m_SampleLargeDataPreHash;
#if PIA_TEST_CODE
public:
    virtual void OperateForPiaTest();
    virtual void UpdateForPiaTest();
    virtual void UpdatePiaTestData() = 0;

protected:
    virtual nn::pia::StationIndex GetSendStationIndex()
    {
        return nn::pia::STATION_INDEX_INVALID;
    }
    virtual void SetParamSendUnreliable(u32 v)
    {
        NN_PIA_DUMMY_PARAM(v);
    }
    virtual void SetParamSendReliable(u32 v)
    {
        NN_PIA_DUMMY_PARAM(v);
    }
    virtual void SetParamSendEvent(u32 v)
    {
        NN_PIA_DUMMY_PARAM(v);
    }
    virtual void SetParamSendReckoning(u32 v)
    {
        NN_PIA_DUMMY_PARAM(v);
    }
    const char* GetConnectionStatusText();
    virtual bool IsConnected(nn::pia::StationId sid)
    {
        NN_PIA_DUMMY_PARAM(sid);
        return false;
    }
#endif
};


// A game object that uses SendClone.
class SendObject : public Object
{
public:
    SendObject();
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual void RegisterClone();
    virtual void UnregisterClone();
    virtual bool IsCloneRegistered() const
    {
        return m_Clone.IsRegisteredWithProtocol();
    }
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    virtual bool IsEnableReceive() const
    {
        return false;
    }
    nn::pia::clone::SendClone m_Clone;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr()
    {
        return &m_Clone;
    }
    virtual const nn::pia::clone::CloneBase& GetClone() const
    {
        return m_Clone;
    }
    virtual const char* GetCloneTypeText() const
    {
        return "S";
    }

#if PIA_TEST_CODE
public:
    virtual void UpdatePiaTestData();

protected:
    virtual void SetParamSendUnreliable(u32 v);
    virtual void SetParamSendEvent(u32 v);
    virtual void SetParamSendReliable(u32 v);
    virtual void SetParamSendReckoning(u32 v);
    virtual bool IsConnected(nn::pia::StationId sid)
    {
        return m_Clone.IsConnected(sid);
    }
#endif
};

// A game object that uses ReceiveClone.
class ReceiveObject : public Object
{
public:
    ReceiveObject();
    void InitCloneTypeText()
    {
        snprintf(m_CloneTypeText, sizeof(m_CloneTypeText), "R%1d", m_StationIndex);
    }
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual void RegisterClone();
    virtual void UnregisterClone();
    virtual bool IsCloneRegistered() const
    {
        return m_Clone.IsRegisteredWithProtocol();
    }
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::ReceiveClone m_Clone;
    nn::pia::StationIndex m_StationIndex;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr()
    {
        return &m_Clone;
    }
    virtual const nn::pia::clone::CloneBase& GetClone() const
    {
        return m_Clone;
    }
    virtual const char* GetCloneTypeText() const
    {
        return m_CloneTypeText;
    }
    char m_CloneTypeText[4];

#if PIA_TEST_CODE
public:
    virtual void UpdatePiaTestData();

protected:
    virtual nn::pia::StationIndex GetSendStationIndex()
    {
        return m_Clone.GetSendStationIndex();
    }
#endif
};

// A game object that uses AtomicSharingClone.
class AtomicObject : public Object
{
public:
    AtomicObject();
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual void RegisterClone();
    virtual void UnregisterClone();
    virtual bool IsCloneRegistered() const
    {
        return m_Clone.IsRegisteredWithProtocol();
    }
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::AtomicSharingClone m_Clone;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr()
    {
        return &m_Clone;
    }
    virtual const nn::pia::clone::CloneBase& GetClone() const
    {
        return m_Clone;
    }
    virtual const char* GetCloneTypeText() const
    {
        return "A";
    }

#if PIA_TEST_CODE
public:
    virtual void UpdatePiaTestData();

protected:
    virtual void SetParamSendUnreliable(u32 v);
    virtual void SetParamSendEvent(u32 v);
    virtual void SetParamSendReliable(u32 v);
    virtual void SetParamSendReckoning(u32 v);
    virtual bool IsConnected(nn::pia::StationId sid)
    {
        return m_Clone.IsConnected(sid);
    }
#endif
};

// A game object that uses SequentialSharingClone.
class SequentialObject : public Object
{
public:
    SequentialObject();
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual void RegisterClone();
    virtual void UnregisterClone();
    virtual bool IsCloneRegistered() const
    {
        return m_Clone.IsRegisteredWithProtocol();
    }
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::SequentialSharingClone m_Clone;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr()
    {
        return &m_Clone;
    }
    virtual const nn::pia::clone::CloneBase& GetClone() const
    {
        return m_Clone;
    }
    virtual const char* GetCloneTypeText() const
    {
        return "Q";
    }

#if PIA_TEST_CODE
public:
    virtual void UpdatePiaTestData();

protected:
    virtual void SetParamSendUnreliable(u32 v);
    virtual void SetParamSendReliable(u32 v);
    virtual bool IsConnected(nn::pia::StationId sid)
    {
        return m_Clone.IsConnected(sid);
    }
#endif
};
