﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/util.h>
#include "Object.h"
#include "SampleConfig.h"


nn::pia::StationId g_StationIdList[MAX_STATIONS];

nn::pia::StationIndex ConvertStationIdToStationIndex(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
    if (result.IsFailure())
    {
        stationIndex = nn::pia::STATION_INDEX_INVALID;
    }
    return stationIndex;
}

#if PIA_TEST_CODE
bool g_IsEnableEventSendDelay = false;
#endif

namespace
{
#if PIA_TEST_CODE
nn::pia::clone::CloneProtocol::Setting::ClockType s_ClockType = nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_RTC;
#endif

// Indicates the mode that is executing.
// With inet, starts with automatch, and then determines whether to be the host or a client.
enum GameMode
{
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_MASTER,
    GAME_MODE_LOCAL_CLIENT,
#endif
    GAME_MODE_INET_AUTOMATCH,
    GAME_MODE_INET_MASTER,
    GAME_MODE_INET_CLIENT,
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_ERROR,
#endif
    GAME_MODE_INET_ERROR,
    GAME_MODE_EXIT
};


// Determines the mode to execute.
GameMode DecideGameMode()
{
#if USE_PIA_LOCAL
    PIA_SAMPLE_PRINT("Please select game mode.\n [A]LocalHost\n [B]LocalClient\n [X]InetAutomatch\n [START]Quit\n");
#else
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
#endif
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
#if USE_PIA_LOCAL
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalHost\n");
            gameMode = GAME_MODE_LOCAL_MASTER;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalClient\n");
            gameMode = GAME_MODE_LOCAL_CLIENT;
            isDecided = true;
        }
#endif
#if PIA_TEST_CODE
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R) && pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            if (s_ClockType == nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_RTC)
            {
                s_ClockType = nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_FRAME;
            }
            else
            {
                s_ClockType = nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_RTC;
            }
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R) && pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            g_IsEnableEventSendDelay = !g_IsEnableEventSendDelay;
        }
        pSampleFramework->SetProgramParam(CLOCK_TYPE, s_ClockType);
        pSampleFramework->SetProgramParam(ENABLE_EVENT_SEND_DELAY, g_IsEnableEventSendDelay ? 1 : 0);
        pSampleFramework->SetProgramParam(SESSION_PROCESSING, isDecided ? 1 : 0);
#endif

        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        pSampleFramework->DrawText(10, 20, "Select game mode.");
#if USE_PIA_LOCAL
        pSampleFramework->DrawText(10, 30, "A: LocalHost");
        pSampleFramework->DrawText(10, 40, "B: LocalClient");
        pSampleFramework->DrawText(10, 50, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 60, "START: Quit this application");
#else
        pSampleFramework->DrawText(10, 30, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 40, "START: Quit this application");
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return gameMode;
        }
    }
}
}

namespace
{
static nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> s_ProtocolSendProfiler;
static nn::pia::clone::CloneProfiler<MAX_STATIONS, 120> s_ProtocolReceivedProfiler;

static const u32 cSendReceiveObjectNum = 1;
static const u32 cAtomicObjectNum = 2;
static const u32 cSequentialObjectNum = 1;
static const u32 cObjectNum = (1 + MAX_STATIONS) * cSendReceiveObjectNum + cAtomicObjectNum + cSequentialObjectNum;
static SendObject s_SendObject[cSendReceiveObjectNum];
static ReceiveObject s_ReceiveObject[cSendReceiveObjectNum][MAX_STATIONS];
static AtomicObject s_AtomicObject[cAtomicObjectNum];
static SequentialObject s_SequentialObject[cSequentialObjectNum];
static Object* s_apObject[cObjectNum];

static int s_Cursor = cObjectNum;

static nn::pia::common::Time s_StartTime;
static bool s_IsClockInvalid;


void ClearStationIdList()
{
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        g_StationIdList[i] = nn::pia::STATION_ID_INVALID;
    }
}

void RegisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(g_StationIdList[stationIndex] == nn::pia::STATION_ID_INVALID);
    g_StationIdList[stationIndex] = stationId;
}

void UnregisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(g_StationIdList[stationIndex] != nn::pia::STATION_ID_INVALID);
    g_StationIdList[stationIndex] = nn::pia::STATION_ID_INVALID;
}

// Declarations of callback functions.
void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG("Sample Event JOIN detect. StationId(0x%016llx)\n", stationId.GetValue());
        RegisterStationId(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect. StationId(0x%016llx)\n", stationId.GetValue());
        UnregisterStationId(stationId);
    }
    break;
    }
}


// Initialize the objects.
void InitObjects()
{
    u32 i = 0;
    for (u32 j = 0; j < cSendReceiveObjectNum; ++j)
    {
        s_SendObject[j].m_CloneId = j;
        s_apObject[i] = &s_SendObject[j];
        ++i;
    }
    for (u32 k = 0; k < MAX_STATIONS; ++k)
    {
        for (u32 j = 0; j < cSendReceiveObjectNum; ++j)
        {
            s_ReceiveObject[j][k].m_CloneId = j;
            s_ReceiveObject[j][k].m_StationIndex = static_cast<nn::pia::StationIndex>(k);
            s_ReceiveObject[j][k].InitCloneTypeText();
            s_apObject[i] = &s_ReceiveObject[j][k];
            ++i;
        }
    }
    for (u32 j = 0; j < cAtomicObjectNum; ++j)
    {
        s_AtomicObject[j].m_CloneId = j;
        s_apObject[i] = &s_AtomicObject[j];
        ++i;
    }
    for (u32 j = 0; j < cSequentialObjectNum; ++j)
    {
        s_SequentialObject[j].m_CloneId = j;
        s_apObject[i] = &s_SequentialObject[j];
        ++i;
    }
    PIA_SAMPLE_ASSERT(i == cObjectNum);
}


// Operation by key input.
bool Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    // Exit the GameLoop function if START is pressed.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        return true;
    }

    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
    {
        --s_Cursor;
        if (s_Cursor < 0)
        {
            s_Cursor = cObjectNum;
        }
    }
    else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
    {
        ++s_Cursor;
        if (s_Cursor >= cObjectNum + 1)
        {
            s_Cursor = 0;
        }
    }

#if PIA_TEST_CODE
#else
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L))
    {
        Clone_GetCloneProtocol()->ResetProfiler();
        for (u32 i = 0; i < cObjectNum; ++i)
        {
            s_apObject[i]->ResetProfiler();
        }
    }
#endif

#if PIA_TEST_CODE
#else
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
    {
        PIA_SAMPLE_PRINT("==== Profiling ====\n");
        PIA_SAMPLE_PRINT("           Recent        Total\n");
        PIA_SAMPLE_PRINT("           Size   Count  Size   Count\n");

        {
            nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();
            PIA_SAMPLE_PRINT("Protocol\n");
            for (u32 i = 0; i < MAX_STATIONS; ++i)
            {
                nn::pia::StationIndex stationIndex = static_cast<nn::pia::StationIndex>(i);

                if (pCloneProtocol->GetSendProfiler())
                {
                    PIA_SAMPLE_PRINT("  Send (%d) %7d %4d  %7d %4d\n",
                                     i,
                                     pCloneProtocol->GetProfiledLatestSendSize(stationIndex),
                                     pCloneProtocol->GetProfiledLatestSendCount(stationIndex),
                                     pCloneProtocol->GetProfiledTotalSendSize(stationIndex),
                                     pCloneProtocol->GetProfiledTotalSendCount(stationIndex));
                }
                if (pCloneProtocol->GetReceiveProfiler())
                {
                    PIA_SAMPLE_PRINT("  Recv (%d) %7d %4d  %7d %4d\n",
                                     i,
                                     pCloneProtocol->GetProfiledLatestReceiveSize(stationIndex),
                                     pCloneProtocol->GetProfiledLatestReceiveCount(stationIndex),
                                     pCloneProtocol->GetProfiledTotalReceiveSize(stationIndex),
                                     pCloneProtocol->GetProfiledTotalReceiveCount(stationIndex));
                }
            }
        }

        for (u32 i = 0; i < cObjectNum; ++i)
        {
            s_apObject[i]->TraceProfiling();
        }
    }
#endif

#if PIA_TEST_CODE
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L) && pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
    {
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            for (int i = 0; i < cObjectNum; ++i)
            {
                if (!s_apObject[i]->IsCloneRegistered())
                {
                    s_apObject[i]->RegisterClone();
                }
            }
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            for (int i = 0; i < cObjectNum; ++i)
            {
                if (s_apObject[i]->IsCloneRegistered())
                {
                    s_apObject[i]->UnregisterClone();
                }
            }
        }
    }
    else if(pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L))
    {
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            nn::pia::transport::ThreadStreamManager* pThreadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pThreadStreamMgr));
            nn::Result result;
            result = pThreadStreamMgr->SetSendThreadPacketLossRatio(100);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = pThreadStreamMgr->SetReceiveThreadPacketLossRatio(100);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            nn::pia::transport::ThreadStreamManager* pThreadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pThreadStreamMgr));
            nn::Result result;
            result = pThreadStreamMgr->SetSendThreadPacketLossRatio(0);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = pThreadStreamMgr->SetReceiveThreadPacketLossRatio(0);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }
    }
#endif

    if (s_Cursor < cObjectNum)
    {
        s_apObject[s_Cursor]->Operate();
    }
    else
    {
        nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            switch (pCloneProtocol->GetState())
            {
            case nn::pia::clone::CloneProtocol::STATE_INACTIVE:
            {
                nn::Result r = pCloneProtocol->Start();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
                s_StartTime = nn::pia::common::Time::GetTime();
                s_IsClockInvalid = false;
            }
            break;

            case nn::pia::clone::CloneProtocol::STATE_START:
            case nn::pia::clone::CloneProtocol::STATE_ACTIVE:
            {
                nn::Result r = pCloneProtocol->Stop();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            }
            break;
            }
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            if (pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE &&
                pCloneProtocol->IsRegulatingClock() == false)
            {
                nn::Result r = pCloneProtocol->RegulateClock();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                PIA_SAMPLE_PRINT("Called RegulateClock()\n");
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            }
        }
#if PIA_TEST_CODE
#else
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            // Check Search function.
            PIA_SAMPLE_PRINT("Search Clone\n");
            nn::pia::StationIndex sidx;
            nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&sidx, nn::pia::session::Session::GetInstance()->GetLocalStationId());
            if (result.IsFailure())
            {
                sidx = nn::pia::STATION_INDEX_INVALID;
            }
            for (u32 i = 0; i < cObjectNum; ++i)
            {
                nn::pia::clone::CloneBase::Id cid = static_cast<nn::pia::clone::CloneBase::Id>(i);
                nn::pia::clone::SendClone* pSendClone = pCloneProtocol->SearchSendClone(cid);
                nn::pia::clone::ReceiveClone* pReceiveClone = pCloneProtocol->SearchReceiveClone(sidx, cid);
                nn::pia::clone::AtomicSharingClone* pAtomicClone = pCloneProtocol->SearchAtomicSharingClone(cid);
                nn::pia::clone::SequentialSharingClone* pSequentialClone = pCloneProtocol->SearchSequentialSharingClone(cid);
                if (pSendClone)
                {
                    PIA_SAMPLE_PRINT("index:%d Clone Type: Send\n", cid);
                }
                if (pReceiveClone)
                {
                    PIA_SAMPLE_PRINT("index:%d local station index:%d Clone Type: Receive\n", cid, sidx);
                }
                if (pAtomicClone)
                {
                    PIA_SAMPLE_PRINT("index:%d Clone Type: Atomic\n", cid);
                }
                if (pSequentialClone)
                {
                    PIA_SAMPLE_PRINT("index:%d Clone Type: Sequential\n", cid);
                }
            }
            // Check IsActive function.
            for (u32 i = 0; i < MAX_STATIONS; ++i)
            {
                nn::pia::StationId sid = g_StationIdList[i];
                if (pCloneProtocol->IsActiveStation(sid))
                {
                    PIA_SAMPLE_PRINT("Station%d (%016llx) is active.\n", i, sid.GetValue());
                }
            }
        }
#endif
    }

    return false;
}


#if PIA_TEST_CODE
void UpdatePiaTestData(GameMode gameMode, bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    bool isInet = false;
    switch (gameMode)
    {
    case GAME_MODE_INET_MASTER:
    case GAME_MODE_INET_CLIENT:
    case GAME_MODE_INET_AUTOMATCH:
    {
        isInet = true;
    }
    break;
    }
    PIA_SAMPLE_ASSERT(pSession);
    pSampleFramework->SetProgramStatus(
        isExit ? 0 : (isInet ? 4 : 3),
        (pSession->CheckConnectionError().IsSuccess() &&
         pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION),
        pSession->GetLocalStationId() == pSession->GetHostStationId(),
        pSession->GetStationNum());
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    pSampleFramework->SetProgramParam(MAX_STATION_NUM, MAX_STATIONS);
    pSampleFramework->SetProgramParam(STATION_INDEX, pSession->GetLocalStationIndex());
    pSampleFramework->SetProgramParam(CLONE_PROTOCOL_STATE, pCloneProtocol->GetState());
    pSampleFramework->SetProgramParam(CLOCK_TYPE, s_ClockType);
    pSampleFramework->SetProgramParam(ENABLE_EVENT_SEND_DELAY, g_IsEnableEventSendDelay ? 1 : 0);
    pSampleFramework->SetProgramParam(CLOCK_VALUE, pCloneProtocol->GetClock());
    pSampleFramework->SetProgramParam(CURSOR_POS, s_Cursor);

    for (u32 i = 0; i < cObjectNum; ++i)
    {
        s_apObject[i]->UpdatePiaTestData();
    }
}
#endif

// Draw on the screen.
void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    static const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    static const nn::pia::sampleframework::Color cCursorColor(1.0f, 0.3f, 0.3f, 1.0f);
    static const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);
    static const nn::pia::sampleframework::Color cColorR(1.0f, 0.4f, 0.4f, 1.0f);
    static const nn::pia::sampleframework::Color cColorG(0.4f, 1.0f, 0.4f, 1.0f);
    static const nn::pia::sampleframework::Color cColorB(0.4f, 0.4f, 1.0f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);
    {
        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "LocalHost";
            }
            else
            {
                cGameModeText = "LocalClient";
            }
        }
        break;
#endif
        case GAME_MODE_INET_AUTOMATCH:
        {
            cGameModeText = "InetAutomatch";
        }
        break;
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "InetHost";
            }
            else
            {
                cGameModeText = "InetClient";
            }
        }
        break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%d", pSession->GetStationNum());
        pSampleFramework->DrawText(300, 20, "Index:%d", ConvertStationIdToStationIndex(pSession->GetLocalStationId()));
    }

    {
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    {
        pSampleFramework->SetFontColor(s_Cursor == cObjectNum ? cCursorColor : cDefaultColor);

        const char* cpStateText = "";
        switch (pCloneProtocol->GetState())
        {
        case nn::pia::clone::CloneProtocol::STATE_INACTIVE:
        {
            cpStateText = "Inactive";
        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_START:
        {
            cpStateText = "Start";
        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_ACTIVE:
        {
            cpStateText = "Active";
        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_STOP:
        {
            cpStateText = "Stop";
        }
        break;
        }
        pSampleFramework->DrawText(10, 225, "State:%s", cpStateText);

        pSampleFramework->SetFontColor(cDefaultColor);
    }

    {
        nn::pia::clone::ClockValue clock = pCloneProtocol->GetClock();
        if (clock == nn::pia::clone::INVALID_CLOCK)
        {
            pSampleFramework->DrawText(170, 225, "Clock:--------");
        }
        else
        {
#if PIA_TEST_CODE
            if (s_ClockType == nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_RTC)
            {
                pSampleFramework->DrawText(170, 225, "Clock:%8.2f%c", static_cast<f32>(clock) / 1000.0f, pCloneProtocol->IsRegulatingClock() ? 'R' : ' ');
            }
            else
            {
                pSampleFramework->DrawText(170, 225, "Clock:%08d%c", clock, pCloneProtocol->IsRegulatingClock() ? 'R' : ' ');
            }
#else
            pSampleFramework->DrawText(170, 225, "Clock:%8.2f%c", static_cast<f32>(clock) / 1000.0f, pCloneProtocol->IsRegulatingClock() ? 'R' : ' ');
#endif
            if (s_IsClockInvalid == false)
            {
                nn::pia::common::Time time = nn::pia::common::Time::GetTime();
                s32 diff = time - s_StartTime;
                PIA_SAMPLE_PRINT("span of start to invalid clock : %d\n", diff);
                s_IsClockInvalid = true;
            }
        }
    }

    {
        static u32 printPacketNum = 0;
        if (pSampleFramework->GetFrameCount() % 30 == 0)
        {
            static u32 lastPacketNum = 0;
            u32 packetNum = nn::pia::transport::ThreadStreamManager::GetInstance()->GetSendThreadSendPacketNum();
            printPacketNum = packetNum - lastPacketNum;
            lastPacketNum = packetNum;
        }
        pSampleFramework->DrawText(350, 225, "%2d", printPacketNum);
    }

    {
        int y = 30;
        for (u32 i = 0; i < cObjectNum; ++i)
        {
            y = s_apObject[i]->Print(y, (i == s_Cursor) ? cCursorColor : cDefaultColor);
        }
    }

    pSampleFramework->SetFontColor(cDefaultColor);
}


void Update()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    if (pCloneProtocol->GetState() != nn::pia::clone::CloneProtocol::STATE_INACTIVE)
    {
        pCloneProtocol->UpdateClock();
    }

    if (pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_START ||
        pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE)
    {
        if (pCloneProtocol->GetError() != nn::pia::clone::CloneProtocol::ERROR_TYPE_NOTHING)
        {
            // Detected an error, so stop.
            const char* cpErrorText = "";
            switch (pCloneProtocol->GetError())
            {
            case nn::pia::clone::CloneProtocol::ERROR_TYPE_CLOCK_OVERFLOW:
            {
                cpErrorText = "ClockOverflow";
            }
            break;
            case nn::pia::clone::CloneProtocol::ERROR_TYPE_DROP_EVENT:
            {
                cpErrorText = "DropEvent";
            }
            break;
            }
            PIA_SAMPLE_PRINT("Error: %s\n", cpErrorText);
            pCloneProtocol->Stop();
        }
    }

#if PIA_TEST_CODE
    static int frameCnt = 0;
    const int SEND_PERIOD = 10;
    bool isSend = (frameCnt % SEND_PERIOD == 0);
    ++frameCnt;
    if (isSend)
    {
        for (u32 i = 0; i < cObjectNum; ++i)
        {
            s_apObject[i]->UpdateForPiaTest();
        }
    }
#else
    for (u32 i = 0; i < cObjectNum; ++i)
    {
        s_apObject[i]->Update();
    }
#endif
}

bool Cleanup()
{
    // Unregister the Clone from CloneProtocol, and then Unregister the Element from Clone.
    bool isCleanupAll = true;
    for (u32 i = 0; i < cObjectNum; ++i)
    {
        isCleanupAll &= s_apObject[i]->Cleanup();
    }
    return isCleanupAll;
}

// Main loop during communication.
void MainLoop(GameMode gameMode, bool* pIsLeaveNeeded)
{
    nn::Result result;
    while (true)
    {
        bool isExit = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        isExit |= Operate();
        if (isExit)
        {
            *pIsLeaveNeeded = true;
        }

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (gameMode == GAME_MODE_INET_MASTER || gameMode == GAME_MODE_INET_CLIENT)
        {
            if (Inet_IsConnected() == false)
            {
                // Ends because the AC has disconnected.
                isExit = true;
                *pIsLeaveNeeded = false;
            }
        }

        result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
        if (result.IsSuccess())
        {
            Update();
        }
        else
        {
            // Ends because the session was lost.
            isExit = true;
            nn::pia::util::HandlingType handlingType = nn::pia::util::GetHandlingType(result);
            if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP)
            {
                *pIsLeaveNeeded = true;
            }
            else if(handlingType == nn::pia::util::HANDLING_RECONNECT)
            {
                *pIsLeaveNeeded = false;
            }
        }

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        Draw(gameMode);


#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, isExit);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SESSION_PROCESSING, isExit ? 1 : 0);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            return;
        }
    }
}


}  // end of anonymous namespace

namespace
{
const size_t MEMSIZE = 10 * 1024 * 1024;             // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.
}


namespace
{
// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const u32 SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const u32 SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

#define SAMPLE_NEX_THREAD_MODE nn::nex::Core::ThreadModeUnsafeTransportBuffer

const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 45000;
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;

#if USE_PIA_LOCAL
const u32 SAMPLE_TITLE_ID = 0xFF330;
const u8 SAMPLE_SUB_ID = 0x00;
const char SAMPLE_PASSPHRASE[] = "aaabbccc0123456789";
const char SAMPLE_SIGNATURE_KEY_DATA[] = "cloneprotocol_key";
#endif
}

// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;
    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(60);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.1f, 0.1f, 0.1f, 1.0f));

#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontSize(5);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(MAX_STATION_NUM, MAX_STATIONS);
#endif
    }

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    int randomSeed;
#if NN_PIA_CTR
    randomSeed = static_cast<int>(nn::os::Tick::GetSystemCurrent());
#elif NN_PIA_CAFE
    randomSeed = static_cast<int>(OSGetTick());
#endif
    ::std::srand(randomSeed);

    InitObjects();

    while (true)
    {
        ClearStationIdList();

        GameMode gameMode = DecideGameMode();
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
#if PIA_TEST_CODE
            u64 traceFlags =
                nn::pia::common::TRACE_FLAG_CLONE |
                nn::pia::common::TRACE_FLAG_LOCAL |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_SESSION |
                nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
                nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
                nn::pia::common::TRACE_FLAG_SESSION_UPDATE;
#else
            u64 traceFlags = nn::pia::common::TRACE_FLAG_CLONE;
#endif
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlags, true);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            nn::pia::common::Log::GetGlobalObject()->SetMask(nn::pia::common::Log::LEVEL_ERROR);

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            {
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID, ENABLE_HOST_MIGRATION);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
            {
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
            }
            break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            nn::pia::transport::Transport::Setting setting;
            setting.pFactory = pFactory;
            setting.maxStationNum = MAX_STATIONS;
            setting.maxSendThreadBufferNum = 16;
            setting.maxReceiveThreadBufferNum = 24;
#if PIA_TEST_CODE
            nn::pia::transport::Transport::DebugSetting debugSetting;
            debugSetting.sendThreadLatencyEmulationPacketNum = 128;
            debugSetting.receiveThreadLatencyEmulationPacketNum = 256;
            debugSetting.bPacketLossEmulation = true;
            result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#endif

            result = Transport_InitializePiaTransport(setting);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            // Set up clone.
            nn::pia::clone::CloneProtocol::Setting cloneSetting;
            {
#if PIA_TEST_CODE
                cloneSetting.clockType = s_ClockType;
                cloneSetting.clockPerSec = 1000.0f;
                cloneSetting.clockPerFrame = 1;
                cloneSetting.clockPriorityJudgeMethod = nn::pia::clone::CloneProtocol::Setting::CLOCK_PRIORITY_JUDGE_METHOD_SIMPLE;
                cloneSetting.dataCompressionLevel = nn::pia::clone::CloneProtocol::COMPRESSION_LEVEL_NONE;
#else
#if 1
                cloneSetting.clockType = nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_RTC;
                cloneSetting.clockPerSec = 1000.0f;
#else
                cloneSetting.clockType = nn::pia::clone::CloneProtocol::Setting::CLOCK_TYPE_FRAME;
                cloneSetting.clockPerFrame = 1;
#endif
                cloneSetting.clockPriorityJudgeMethod = nn::pia::clone::CloneProtocol::Setting::CLOCK_PRIORITY_JUDGE_METHOD_SIMPLE;
                cloneSetting.dataCompressionLevel = nn::pia::clone::CloneProtocol::COMPRESSION_LEVEL_MIDDLE;
#endif
            }
            Clone_InitializePiaClone(cloneSetting);

            result = Clone_GetCloneProtocol()->SetProfiler(&s_ProtocolSendProfiler, &s_ProtocolReceivedProfiler);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#if PIA_TEST_CODE
            result = Clone_GetCloneProtocol()->SetMaxSendDelay(170);
#else
            result = Clone_GetCloneProtocol()->SetMaxSendDelay(30);
#endif
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = Clone_GetCloneProtocol()->SetResendSpanRatio(1.3f);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#if 0
            // This feature is for debugging.
            result = Clone_GetCloneProtocol()->SetClockOverFlowSpanForDebug(100);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#endif

            // Set up reckoning.
            Reckoning_InitializePiaReckoning();

            // Sets up the session.
            nn::pia::session::Session::Setting sessionSetting;
            sessionSetting.pNetworkFactory = pFactory;
            sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
            sessionSetting.browsedSessionInfoListNum = 8;
#if PIA_TEST_CODE
            result = Session_InitializePiaSession(sessionSetting, false, false, true);
#else
            result = Session_InitializePiaSession(sessionSetting);
#endif
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionCallback);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }

        // Connect to the network connections and create a session.
        {

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            {
                // Start Session.
#if PIA_TEST_CODE
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;
                startupSetting.maxSilenceTime = MAX_SILENT_TIME;

                result = Session_StartupSession(startupSetting);
#else
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
#endif
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the matchmaking session to create.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                        createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
#if PIA_TEST_CODE
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;
                startupSetting.maxSilenceTime = MAX_SILENT_TIME;

                result = Session_StartupSession(startupSetting);
#else
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
#endif
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
                    searchCriteriaList[0].SetResultRange(0, 8);
                }

                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connects to the Internet.
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }
#if PIA_TEST_CODE
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;
                startupSetting.maxSilenceTime = MAX_SILENT_TIME;

                result = Session_StartupSession(startupSetting);
#else
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);

#endif
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }
                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
                if (isHost)
                {
                    // Became the host as a result of the matching.
                    gameMode = GAME_MODE_INET_MASTER;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;

            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        bool isLeaveNeeded = false;

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
        {
            MainLoop(gameMode, &isLeaveNeeded);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }
        }
        break;
        }


        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            {
                if (isLeaveNeeded)
                {
                    // Leaves the session.
                    // In this sample, the return value is ignored because finalization will take place later.
                    Session_LeaveSession();
                }
            }
            break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
            }
            break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_Logout();
                Inet_FinalizeNetwork();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, true);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SESSION_PROCESSING, 0);
#endif

        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Reckoning_FinalizePiaReckoning();
            Clone_FinalizePiaClone();
            Transport_FinalizePiaTransport();
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
                Local_FinalizePiaLocal();
            }
            break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
            Common_FinalizePiaCommon();
        }
    }
    while (!Cleanup())
    {
    }
    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
