﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     EventChecker.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include "SampleConfig.h"

class EventChecker
{
public:
    enum EVENT_BIT
    {
        EVENT_BIT_SESSION_HOST_CHANGED          = 0x1 << 0,
        EVENT_BIT_HOST_MIGRATION_FAILED         = 0x1 << 1,
        EVENT_BIT_JOINT_SESSION_HOST_CHANGED    = 0x1 << 2,
        EVENT_BIT_START_SESSION_CREATE_JOINT    = 0x1 << 3,
        EVENT_BIT_END_SESSION_CREATE_JOINT      = 0x1 << 4,
        EVENT_BIT_START_SESSION_JOIN_JOINT      = 0x1 << 5,
        EVENT_BIT_END_SESSION_JOIN_JOINT        = 0x1 << 6,
        EVENT_BIT_START_SESSION_JOINT_RANDOM    = 0x1 << 7,
        EVENT_BIT_END_SESSION_JOINT_RANDOM      = 0x1 << 8,
        EVENT_BIT_START_SESSION_DESTROY_JOINT   = 0x1 << 9,
        EVENT_BIT_END_SESSION_DESTROY_JOINT     = 0x1 << 10,
        EVENT_BIT_START_SESSION_LEAVE_JOINT     = 0x1 << 11,
        EVENT_BIT_END_SESSION_LEAVE_JOINT       = 0x1 << 12,
        EVENT_BIT_SESSION_MIGRATION_FAILED      = 0x1 << 13
    };

    enum CHECK_TARGET
    {
        CHECK_TARGET_NONE = 0,
        CHECK_TARGET_SESSION_MIGRATION,
        CHECK_TARGET_HOST_MIGRATION
    };

    EventChecker()
    {
        Reset();
    }

    void Reset()
    {
        m_CheckTarget = CHECK_TARGET_NONE;
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            m_JoinStationIdList[i] = nn::pia::STATION_ID_INVALID;
            m_LeaveStationIdList[i] = nn::pia::STATION_ID_INVALID;
            m_ExpectedLeaveStationIdList[i] = nn::pia::STATION_ID_INVALID;
        }
        UpdateValidStationIdList();
        m_EventFlag = 0;
        m_ExpectedEventFlag = 0;
    }

    bool IsCheckEnabled() const
    {
        return m_CheckTarget != CHECK_TARGET_NONE;
    }

    CHECK_TARGET GetCheckTarget() const
    {
        return m_CheckTarget;
    }

    void CheckBegin(CHECK_TARGET target)
    {
        PIA_SAMPLE_LOG("[EventChecker] CheckBegin %d\n", target);
        PIA_SAMPLE_ASSERT(m_CheckTarget != target);
        Reset();
        m_CheckTarget = target;
    }

    void CheckEnd(CHECK_TARGET target)
    {
        NN_PIA_DUMMY_PARAM_IN_RELEASE(target);
        PIA_SAMPLE_LOG("[EventChecker] CheckEnd %d\n", target);
        PIA_SAMPLE_ASSERT(m_CheckTarget == target);
        PIA_SAMPLE_ASSERT(IsCheckEnabled());
        Reset();
    }

    u32 GetEventFlag() const
    {
        return m_EventFlag;
    }

    u32 GetExpectedEventFlag() const
    {
        return m_ExpectedEventFlag;
    }

    void SetEventFlag(EVENT_BIT eventBit)
    {
        if (!IsCheckEnabled())
        {
            return;
        }
        if (eventBit != EVENT_BIT_SESSION_HOST_CHANGED && eventBit != EVENT_BIT_JOINT_SESSION_HOST_CHANGED)
        {
            PIA_SAMPLE_ASSERT((m_EventFlag & eventBit) == 0);
        }
        m_EventFlag |= eventBit;
    }

    void SetExpectedEventFlag(EVENT_BIT eventBit)
    {
        m_ExpectedEventFlag |= eventBit;
    }

    void UnsetExpectedEventFlag(EVENT_BIT eventBit)
    {
        m_ExpectedEventFlag &= ~eventBit;
    }

    void SetJoinEventStationId(nn::pia::StationId stationId)
    {
        if (!IsCheckEnabled())
        {
            return;
        }
        PIA_SAMPLE_ASSERT(stationId != nn::pia::STATION_ID_INVALID);
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_ASSERT(m_JoinStationIdList[i] != stationId);
        }
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (m_JoinStationIdList[i] == nn::pia::STATION_ID_INVALID)
            {
                m_JoinStationIdList[i] = stationId;
                return;
            }
        }
        PIA_SAMPLE_ASSERT(false);
    }

    void SetLeaveEventStationId(nn::pia::StationId stationId)
    {
        if (!IsCheckEnabled())
        {
            return;
        }
        PIA_SAMPLE_ASSERT(stationId != nn::pia::STATION_ID_INVALID);
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_ASSERT(m_LeaveStationIdList[i] != stationId);
        }
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (m_LeaveStationIdList[i] == nn::pia::STATION_ID_INVALID)
            {
                m_LeaveStationIdList[i] = stationId;
                return;
            }
        }
        PIA_SAMPLE_ASSERT(false);
    }

    void SetExpectedLeaveEventStationId(nn::pia::StationId stationId)
    {
        PIA_SAMPLE_ASSERT(stationId != nn::pia::STATION_ID_INVALID);
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_ASSERT(m_ExpectedLeaveStationIdList[i] != stationId);
        }
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (m_ExpectedLeaveStationIdList[i] == nn::pia::STATION_ID_INVALID)
            {
                m_ExpectedLeaveStationIdList[i] = stationId;
                return;
            }
        }
        PIA_SAMPLE_ASSERT(false);
    }

    void UpdateValidStationIdList()
    {
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            m_ValidStationIdList[i] = nn::pia::STATION_ID_INVALID;
        }
        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
        if (!PIA_IS_VALID_POINTER(pSession))
        {
            return;
        }
        int index = 0;
        nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
        for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
        {
            m_ValidStationIdList[index] = *it;
            ++index;
        }
    }

    void CheckStationIdList() const
    {
        // Valid + Join - Leave = current

        nn::pia::StationId tmpValidStationIdList[SAMPLE_GAME_PLAYER_MAX];
        nn::pia::StationId tmpJoinStationIdList[SAMPLE_GAME_PLAYER_MAX];
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            tmpValidStationIdList[i] = m_ValidStationIdList[i];
            tmpJoinStationIdList[i] = m_JoinStationIdList[i];
        }

        // Should not be included in both tmpValid and tmpJoin.
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (tmpValidStationIdList[i] != nn::pia::STATION_ID_INVALID)
            {
                for (int j = 0; j < SAMPLE_GAME_PLAYER_MAX; ++j)
                {
                    // When both tmpValid and tmpJoin are included, there will be an exit mid-way, and although it may be the case that a peer will have joined, in this instance, a Leave should have been triggered.
                    // 
                    if (tmpValidStationIdList[i] == tmpJoinStationIdList[j])
                    {
                        bool isFound = false;
                        NN_PIA_DUMMY_PARAM_IN_RELEASE(isFound);
                        for (int k = 0; k < SAMPLE_GAME_PLAYER_MAX; ++k)
                        {
                            if (tmpValidStationIdList[i] == m_LeaveStationIdList[k])
                            {
                                isFound = true;
                                break;
                            }
                            if (tmpJoinStationIdList[j] == m_LeaveStationIdList[k])
                            {
                                isFound = true;
                                break;
                            }
                        }
                        PIA_SAMPLE_ASSERT(isFound);
                    }
                }
            }
        }

        // Leave should be in tmpValid or tmpJoin. 
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (m_LeaveStationIdList[i] != nn::pia::STATION_ID_INVALID)
            {
                bool isFound = false;
                NN_PIA_DUMMY_PARAM_IN_RELEASE(isFound);
                for (int j = 0; j < SAMPLE_GAME_PLAYER_MAX; ++j)
                {
                    if (tmpValidStationIdList[j] == m_LeaveStationIdList[i])
                    {
                        isFound = true;
                        tmpValidStationIdList[j] = nn::pia::STATION_ID_INVALID;
                        break;
                    }
                    if (tmpJoinStationIdList[j] == m_LeaveStationIdList[i])
                    {
                        isFound = true;
                        tmpJoinStationIdList[j] = nn::pia::STATION_ID_INVALID;
                        break;
                    }
                }
                PIA_SAMPLE_ASSERT(isFound);
            }
        }

        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            // tmpValid and tmpJoin should currently be valid.
            if (tmpValidStationIdList[i] != nn::pia::STATION_ID_INVALID)
            {
                PIA_SAMPLE_ASSERT(pSession->CheckStationIdIsValid(tmpValidStationIdList[i]));
            }
            if (tmpJoinStationIdList[i] != nn::pia::STATION_ID_INVALID)
            {
                PIA_SAMPLE_ASSERT(pSession->CheckStationIdIsValid(tmpJoinStationIdList[i]));
            }
        }

        nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
        for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
        {
            PIA_SAMPLE_ASSERT(*it != nn::pia::STATION_ID_INVALID);
            // Should exist in tmpValid or tmpJoin.
            bool isFound = false;
            NN_PIA_DUMMY_PARAM_IN_RELEASE(isFound);
            for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
            {
                if (tmpValidStationIdList[i] == *it)
                {
                    isFound = true;
                    tmpValidStationIdList[i] = nn::pia::STATION_ID_INVALID;
                    break;
                }
                if (tmpJoinStationIdList[i] == *it)
                {
                    isFound = true;
                    tmpJoinStationIdList[i] = nn::pia::STATION_ID_INVALID;
                    break;
                }
            }
            PIA_SAMPLE_ASSERT(isFound);
        }

        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            // tmpValid and tmpJoin should be empty.
            PIA_SAMPLE_ASSERT(tmpValidStationIdList[i] == nn::pia::STATION_ID_INVALID);
            PIA_SAMPLE_ASSERT(tmpJoinStationIdList[i] == nn::pia::STATION_ID_INVALID);
        }
    }

    void CheckExpectedLeaveStationIdList() const
    {
        // Determine whether ExpectedLeave is included in Leave.
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            if (m_ExpectedLeaveStationIdList[i] != nn::pia::STATION_ID_INVALID)
            {
                bool isFound = false;
                NN_PIA_DUMMY_PARAM_IN_RELEASE(isFound);
                for (int j = 0; j < SAMPLE_GAME_PLAYER_MAX; ++j)
                {
                    if (m_ExpectedLeaveStationIdList[i] == m_LeaveStationIdList[j])
                    {
                        isFound = true;
                        break;
                    }
                }
                PIA_SAMPLE_ASSERT(isFound);
            }
        }
    }

    void Trace() const
    {
        PIA_SAMPLE_LOG("--- EventChecker Trace Begin ---\n");
        PIA_SAMPLE_LOG("Target:%d Event:0x%08x ExpectedEvent:0x%08x\n", m_CheckTarget, m_EventFlag, m_ExpectedEventFlag);
        PIA_SAMPLE_LOG("Valid: ");
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_LOG("[%d]%lld ", i, m_ValidStationIdList[i].GetValue());
        }
        PIA_SAMPLE_LOG("\n");
        PIA_SAMPLE_LOG("Join: ");
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_LOG("[%d]%lld ", i, m_JoinStationIdList[i].GetValue());
        }
        PIA_SAMPLE_LOG("\n");
        PIA_SAMPLE_LOG("Leave: ");
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_LOG("[%d]%lld ", i, m_LeaveStationIdList[i].GetValue());
        }
        PIA_SAMPLE_LOG("\n");
        PIA_SAMPLE_LOG("ExpectedLeave: ");
        for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
        {
            PIA_SAMPLE_LOG("[%d]%lld ", i, m_ExpectedLeaveStationIdList[i].GetValue());
        }
        PIA_SAMPLE_LOG("\n");
        PIA_SAMPLE_LOG("--- EventChecker Trace End ---\n");
    }

private:
    CHECK_TARGET m_CheckTarget;
    bit32 m_EventFlag;
    bit32 m_ExpectedEventFlag;
    nn::pia::StationId m_ValidStationIdList[SAMPLE_GAME_PLAYER_MAX];
    nn::pia::StationId m_JoinStationIdList[SAMPLE_GAME_PLAYER_MAX];
    nn::pia::StationId m_LeaveStationIdList[SAMPLE_GAME_PLAYER_MAX];
    nn::pia::StationId m_ExpectedLeaveStationIdList[SAMPLE_GAME_PLAYER_MAX];
};

// For checking events.
extern EventChecker g_SessionMigrationEventChecker;
extern EventChecker g_HostMigrationEventChecker;
