﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     PrepareInternet.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>
#include <pia/transport/transport_StationConnectionInfo.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"
#include "SampleLib/Thread.h"

#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"

#include "SampleConfig.h"

#include <nn/ac.h>

#if NN_PIA_CTR
#include <nn/os.h>
#endif


/*!

    @brief  CPP file with processing specific to Internet communication.

*/

namespace
{
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

// Displays the progress of <tt>InitializePiaModules</tt> on the screen.
void InitializingProcessRender(u32 processingStep)
{
    // Gets the signal strength information.
    u32 linkLevel = GetLinkLevel();
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_PREPARE_INET,
        false,
        false,
        0,
        linkLevel);

    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    int line = 0;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo]", PROGRAM_NAME);
    line += FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Inet Initializing process.");
    line += FONT_SIZE;

    line += FONT_SIZE;

    // Progress display.
    // Steps that have been processed are displayed in white; steps that have not been processed are displayed in red.
    {
        const char* stepStrings[] = {
            "Initialize NEX",
            "Initialize Pia modules",
            "Initialize Network",
            "Login server"
        };  // 4 step.
        const u32 STEP_MAX = 4;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   START");
        line += FONT_SIZE;
        for (u32 step = 0; step < STEP_MAX; step++)
        {
            if (step == processingStep)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "     |");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
            }
            else
            {
                line += FONT_SIZE;
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", stepStrings[step]);
            line += FONT_SIZE;
        }
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   END");
        line += FONT_SIZE;
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

    line = SCREEN_HEIGHT - FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
    line += FONT_SIZE;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
}

}  // end of anonymous namespace

#if NN_PIA_CTR
// Function for getting the signal strength of the communication with the access point, and returning a converted value in the range from 0 through 3.
u32 GetInternetLinkLevel()
{
    nn::ac::CTR::LinkLevel acLinkLevel = nn::ac::CTR::GetLinkLevel();

    switch (acLinkLevel)
    {
    case nn::ac::CTR::LINK_LEVEL_3:
        return 3;
    case nn::ac::CTR::LINK_LEVEL_2:
        return 2;
    case nn::ac::CTR::LINK_LEVEL_1:
        return 1;
    case nn::ac::CTR::LINK_LEVEL_0:
    default:
        return 0;
    }
}
#endif

/* ---< Setup when conducting Internet communication >--- */
// Initializes NEX, initializes the Pia modules, connects to the Internet, and logs in to the game server.
bool InitializePiaModules()
{
    nn::Result result;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_PREPARE_INET,
        false,
        false,
        0);

    // Because the blocking process continues for a long time, the progress is displayed on the screen.
    u32 step = 0;
    InitializingProcessRender(step);
    step++;

    /* Start processing NEX-related initialization. */

    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    /* Finished processing NEX-related initialization. */

    InitializingProcessRender(step);
    step++;

    PIA_SAMPLE_LOG("NEX initialize completed.\n");

    /* Start of the Pia modules initialization process. */

    result = Inet_InitializePiaInet();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    InitializePiaTransport();

    InitializePiaSession();

    InitializePiaClone();

    /* End of the Pia modules initialization process. */

    InitializingProcessRender(step);
    step++;

    PIA_SAMPLE_LOG("Pia modules initialize completed.\n");

    /* Preparing for Internet communication. */
    // Internal call to nn::ac::Connect .
    // This blocks for a long time, so running it on the main thread is not recommended, but we do it here in this sample because the emphasis is on readability.
    // 
    result = Inet_InitializeNetwork();
    if (result.IsFailure())
    {
#if NN_PIA_CTR
        if (result == nn::ac::ResultWifiOff())
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_WIRELESS_OFF);
        }
        else
#elif NN_PIA_WIN
#error  "Not implemented."
#elif NN_PIA_CAFE
//n1769 Cafe AC does not have <tt>ResultWifiOff</tt>, so the if statement of ↑ is skipped.
#else
#error  "Invalid platform."
#endif
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        }
        PIA_SAMPLE_LOG("Inet_InitializeNetwork failed.\n");

        SetNetworkErrorResult(result);

        return false;
    }

    /* Finished preparing Internet connection processing. */

    InitializingProcessRender(step);
    step++;

    PIA_SAMPLE_LOG("Inet_InitializeNetwork complete.\n");

    /* Start of the process of logging in to the server. */

    // Calls are made to the (asynchronous) APIs for logging in to the friend server and game server.
    // For the sake of readability, this sample program waits for the asynchronous processes to end.
    result = Inet_Login();
    if (result.IsFailure())
    {
        // Log in fails.
        SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        SetNetworkErrorResult(result);
        PIA_SAMPLE_LOG("Inet_Login failed.\n");
        return false;
    }

    /* End of the process of logging in to the server. */

    InitializingProcessRender(step);
    step++;

    PIA_SAMPLE_LOG("Inet_Login completed.\n");

    return true;
}


void FinalizePiaModules()
{
    PIA_SAMPLE_LOG("FinalizePiaModules\n");

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_RESET,
        false,
        false,
        0);

    // Finish cleaning up the Session instance before cleaning up the Facade instance.
    Session_CleanupSession();

    /* Process to log out from the server. */
    Inet_Logout();

    /* Processes Internet finalization (disconnection). */
    Inet_FinalizeNetwork();

    // Finalizes PiaClone.
    FinalizePiaClone();

    // Finalizes PiaSession.
    FinalizePiaSession();

    // Finalizes PiaTransport.
    FinalizePiaTransport();

    /* Finalizes PiaInet. */
    Inet_FinalizePiaInet();

    /* NEX-related finalization. */
    Nex_FinalizeNex();
}
