﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SampleConfig.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include "SampleLib/SampleLib.h"

#include <pia/platform.h>

#include <pia/inet.h>
#include <pia/inet/inet_NexFacade.h>
#include <pia/inet/inet_NexNetworkFactory.h>

#include <pia/transport.h>
#include <pia/session.h>
#include <pia/clone.h>
#include <pia/reckoning.h>

#define FRAME_PER_SECOND 60

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

const char PROGRAM_NAME[] = "clone/TeamMatchmake";

/*!
    Enumerates possible states of the sample program.
 */
enum SampleSceneNo
{
    SCENE_PREPARE_INET,                     //!< This is the state where preparations are made for Internet communication. It has no relation to P2P communication.
    SCENE_SELECT_TEAM,                      //!< This is the state for selecting the team the local station belongs to.
    SCENE_MATCH_MAKE_TEAM,                  //!< This is the state for making a team session and for joining the team session.
    SCENE_COMMUNICATE_TEAM,                 //!< This is the state where peers communicate in a team session. It is also possible to matchmake with another team and migrate to the game session.
    SCENE_CHANGING_SESSION_PREPROCESSING,   //!< This is the state of preprocessing prior to migrating between a team session and a game session.
    SCENE_CHANGING_SESSION,                 //!< This is the state of waiting for migration between a team session and a game session.
    SCENE_CHANGING_SESSION_POSTPROCESSING,  //!< This is the state of postprocessing after migrating between a team session and a game session.
    SCENE_COMMUNICATE_GAME,                 //!< This is the state where peers communicate in a game session.
    SCENE_NETWORK_ERROR,                    //!< This is the state showing network error codes. Assume that the error or EULA applet normally will start in this state.
    SCENE_RESET                             //!< Defines the state where all information managed by the demo has been reset to the default values.
};

/*!
    Enumerates communication errors handled by the sample program.
 */
enum SampleErrorType
{
    ERROR_UNKNOWN = 0,              //!< Specifies that no error occurred.
    INITIALIZE_ERROR_WIRELESS_OFF,  //!< Specifies that initialization failed because wireless mode is off.
    INITIALIZE_ERROR_OTHER_REASON,  //!< Specifies that initialization failed for some other reason.
    MATCHMAKE_ERROR,                //!< Could not establish the session.
    SESSION_LOST_ERROR,             //!< Disconnected from the session during communication.
    CONNECTION_LOST_ERROR,          //!< The network connection was disconnected.
    APPLICATION_ERROR               //!< Application Errors
};

void SetCommunicationErrorType(SampleErrorType error);  //Sets the communication error type.
void SetNetworkErrorResult(nn::Result result);          //This function sets the <tt>Result</tt> value to be converted into a network error code.
void SetResetOfNetworkIsNecessary(bool isNecessary);    //This function sets whether it will be necessary to reset the network (for example, logging in again) when there is a communication error.

void InitializePiaTransport();
void FinalizePiaTransport();
void InitializePiaSession();
void FinalizePiaSession();
void InitializePiaClone();
void FinalizePiaClone();
bool InitializePiaModules();
void FinalizePiaModules();

/* ---< State by process. >--- */
SampleSceneNo SceneSelectTeam();
SampleSceneNo SceneMatchmake();
SampleSceneNo SceneCommunicationTeam();
SampleSceneNo SceneChangingSessionPreprocessing();
SampleSceneNo SceneChangingSession();
SampleSceneNo SceneChangingSessionPostprocessing();
SampleSceneNo SceneCommunicationGame();
SampleSceneNo SceneNetworkError();

/* ---< The  enumerated type for distinguishing team affiliations >--- */
enum SampleTeamType
{
    TEAM_ALPHA = 0,
    TEAM_BRAVO,
    TEAM_CHARLIE,
    TEAM_DELTA,
    TEAM_TYPE_NUM
};
SampleTeamType GetCurrentTeamType();

/* ---< Definitions of structures used in the sample. >--- */
struct RectPosition  //!< Stores rectangle position coordinates.
{
    u16 x;
    u16 y;
};

//The string for displaying the role screen within a session.
const char* SESSION_ROLE(size_t index);

// In the scene where communication occurs with another party in the session, this is the setting for the interval [frame] to send the local host information.
// When set to 0, it sends in every game frame.
const u32 SAMPLE_INTERVAL_SENDING_STATION_STATUS = 3;

const u8 SAMPLE_TEAM_MEMBER_MAX = 4;    //!< The maximum number of participants on a single team.
const u8 SAMPLE_TEAM_NUM_IN_GAME = 3;   //!< The number of teams to merge in battle matchmaking.
const u8 SAMPLE_GAME_PLAYER_MAX = SAMPLE_TEAM_MEMBER_MAX * SAMPLE_TEAM_NUM_IN_GAME;  //!< The maximum number of participants after teams have been merged.

const bool ENABLE_HOST_MIGRATION = true;  //!< Enables or disables the host migration feature.

/* Thread Priority Settings */
#if NN_PIA_CTR
const s32 SEND_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;          //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#elif NN_PIA_WIN
#error  "Not implemented."
#elif NN_PIA_CAFE
const s32 SEND_THREAD_PRIORITY = OS_PRIORITY_APP_DEFAULT - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;  //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#else
#error  "Illegal platform."
#endif

/* NEX-related settings. */
// Condition group used when building or searching for matchmaking sessions on a NEX game server.
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
// The game mode for team formation.
const u32 MATCHMAKE_GAME_MODE_TEAM_BASE = 45100;
const u32 MATCHMAKE_GAME_MODE_TEAM = MATCHMAKE_GAME_MODE_TEAM_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;
// The game mode for battling (for game sessions).
const u32 MATCHMAKE_GAME_MODE_GAME_BASE = 45200;
const u32 MATCHMAKE_GAME_MODE_GAME = MATCHMAKE_GAME_MODE_GAME_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;
// Index of the attribute to use for the team type.
const u32 MATCHMAKE_ATTR_INDEX_TEAM_TYPE = 0;  
const u32 MATCHMAKE_ATTR_INDEX_USERPASSWORD = 1;

/* Memory-related settings */
const u32 MEMORY_BLOCK_SIZE_PIA = 2 * 1024 * 1024;      //!< Defines the amount of memory allocated to Pia.
const u32 MEMORY_BLOCK_SIZE_NEX = 1024 * 1024;          //!< Specifies the memory size provided to NEX.
const u32 MEMORY_BLOCK_SIZE_NEX_EMERGENCY = 50 * 1024;  //!< Specifies the memory size provided to NEX for emergencies.

/* Drawing-related settings */
const size_t SCREEN_WIDTH = 400;   //!< Defines the horizontal resolution of the screen.
const size_t SCREEN_HEIGHT = 240;  //!< Defines the vertical resolution of the upper screen.

const int FONT_SIZE = 8;  //!< Defines the size of the font to draw.

const size_t SAMPLE_USER_NAME_LENGTH = 11;  //!< The maximum number of characters for <tt>UserName</tt> drawing, including the terminating character.

const char* GetPiaVersionString();  //!< Function for getting the string for drawing the Pia library version information.

// Function for setting the <tt>Result</tt> used to indicate Internet-related errors.
void SetNetworkErrorResult(nn::Result result);

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel();

// Initializes the state-synchronization object.
void InitObjects();

// Game states:
enum GAME_STATE
{
    GAME_STATE_NONE = 0,
    GAME_STATE_TEAM,
    GAME_STATE_PREPARING_GAME,
    GAME_STATE_GAME,
    GAME_STATE_PREPARING_TEAM
};
extern GAME_STATE g_GameState;

// Joint Session Types
enum JOINT_SESSION_TYPE
{
    JOINT_SESSION_TYPE_NONE = 0,
    JOINT_SESSION_TYPE_RANDOM,
    JOINT_SESSION_TYPE_CREATE,
    JOINT_SESSION_TYPE_CREATE_PRIVATE,
    JOINT_SESSION_TYPE_JOIN,
    JOINT_SESSION_TYPE_JOIN_PRIVATE,
    JOINT_SESSION_TYPE_DESTROY,
    JOINT_SESSION_TYPE_LEAVE
};
extern JOINT_SESSION_TYPE g_JointSessionType;

// The session state.
bool IsInSession();
bool IsInJointSession();
bool IsSessionMigratingToSession();
bool IsSessionMigratingToJointSession();
bool IsSessionMigrating();
extern bool g_IsOpened;

// Callback that determines whether to approve or reject the join-in request.
bool SampleJoinApprovalCallback(const nn::pia::transport::Station::IdentificationToken *pToken);

// Token data.
struct SampleTokenData
{
#if 0
    u32 sessionId;
#endif
    u8 teamType;
};

// User password.
extern const wchar_t* USER_PASSWORD[];
const size_t USER_PASSWORD_LENGTH = 5;
extern u8 g_UserPasswordIndexForJoint;

void ResetEventChecker();
void ResetErrorOccurredFlag();

#if PIA_TEST_CODE
bool IsEnableHostMigration();
void IsEnableHostMigration(bool b);

void SetMaxPlayerNum(u8 num);
u8 GetMaxPlayerNum();

enum TestParamIndex
{
    INDEX_LATENCY_EMULATION = 0,
    INDEX_PACKET_LOSS_EMULATION,
    INDEX_ENABLE_HOST_MIGRATION,
    INDEX_JOINT_SEQUENCE,
    INDEX_CLOSE_SESSION_SEQUENCE,
    INDEX_OPEN_SESSION_SEQUENCE,
    INDEX_LOCAL_STATION_INDEX,
    INDEX_APPLICATION_DATA_HEAD,
    INDEX_APPLICATION_DATA_SIZE,
    INDEX_JOIN_JOINT_SESSION_ID
};

const u8 DELAY_DISPATCH_NUM = 100;
#endif

// Clone-related.
enum CLONE_ID_OFFSET
{
    CLONE_ID_OFFSET_STATION = 0x0,
    CLONE_ID_OFFSET_COMMON_OBJECT = 0x10000,
    CLONE_ID_OFFSET_TEAM_CONSTITUTION_MANAGER = 0x20000
};
const u32 COMMON_OBJECT_NUM = SAMPLE_TEAM_NUM_IN_GAME * (1 + 1) + (1 + 1); //!< Each team Atom, Seq + common Atom in game, Seq.

const int SAMPLE_PROGRAMPARAM_GETJOINSEQUENCE = 101;
const int SAMPLE_PROGRAMPARAM_OPENCLOSESESSION_COMPLETE = 104;

enum OperationType
{
    OPERATION_TYPE_NONE = 0,
    OPERATION_TYPE_END_OPERATION,
    OPERATION_TYPE_OPEN_SESSION,
    OPERATION_TYPE_CLOSE_SESSION,
    OPERATION_TYPE_OPEN_JOINT_SESSION,
    OPERATION_TYPE_CLOSE_JOINT_SESSION
};

void SetLocalOperationType(u32 type);
u32 GetOperationType(nn::pia::StationId stationId);
void UpdateStatusManager();
