﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneChangingSession.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>
#include <pia/session.h>
#include <pia/util.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Clone_Setup.h"

#include "SampleConfig.h"
#include "EventChecker.h"
#include "TeamConstitutionManager.h"


namespace
{
// Enumerated type for indicating asynchronous processes waiting when teams are being merged and demerged.
enum SampleWaitingAsyncProcessing
{
    SAMPLE_PASSIVE_PROCESSING = 0,
    SAMPLE_JOIN_RANDOM_JOINT_SESSION,
    SAMPLE_CREATE_JOINT_SESSION,
    SAMPLE_JOIN_JOINT_SESSION,
    SAMPLE_BROWSE_JOINT_SESSION,
    SAMPLE_DESTROY_JOINT_SESSION,
    SAMPLE_LEAVE_JOINT_SESSION,
    SAMPLE_OPEN_PARTICIPATION,
    SAMPLE_CLOSE_PARTICIPATION,
    SAMPLE_LEAVE_SESSION,
    SAMPLE_OPEN_SESSION,
    SAMPLE_CLOSE_SESSION,
    SAMPLE_OPEN_JOINT_SESSION,
    SAMPLE_CLOSE_JOINT_SESSION
};
}

// Processing before starting session migration.
SampleSceneNo SceneChangingSessionPreprocessing()
{
    nn::Result result;

    u32 frameNo = 0;
    bool isExit = false;
    bool errorOccurred = false;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_CHANGING_SESSION_PREPROCESSING,
        true,
        false,
        0,
        0
        );

    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
    nn::pia::StationId jointHostStationId = nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId();

    switch (g_GameState)
    {
    case GAME_STATE_TEAM:
    {
        // Before starting migration from team session to game session.
        if (localStationId != hostStationId)
        {
            // Given the specifications for this demo, execution on a non-host should never reach here for any status other than STATUS_MIGRATING_SESSION.
            PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
            // This state is reached when passive session migration is started.
        }

        // Register callback for determining whether to approve a join.
        nn::pia::session::Session::GetInstance()->RegisterJoiningApprovalCallback(SampleJoinApprovalCallback);
        isExit = true;
    }
    break;
    case GAME_STATE_GAME:
    {
        // Before starting migration from game session to team session.
        if ((localStationId != jointHostStationId && g_JointSessionType == JOINT_SESSION_TYPE_DESTROY)
        ||  (localStationId != hostStationId && g_JointSessionType == JOINT_SESSION_TYPE_LEAVE))
        {
            // Given the specifications for this demo, execution on a non-host should never reach here for any status other than STATUS_MIGRATING_SESSION.
            PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
            // This state is reached when passive session migration is started.
        }

        isExit = true;
    }
    break;
    default:
    {
        PIA_SAMPLE_ASSERT(false);
    }
    break;
    }

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Execute dispatch.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Check for internal NEX errors.
        if (Nex_CheckInternalError())
        {
            SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
            errorOccurred = true;
            isExit = true;
        }
#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_JOINT_SEQUENCE, nn::pia::session::Session::GetInstance()->GetJointSequence());

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
            PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());

            if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
            }
            else
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
            }
        }
#endif

        //Screen rendering.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        {
            //Check the role of the local station.
            int roleIndex = 0;  //Unknown.
            localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
            if (localStationId != nn::pia::STATION_ID_INVALID)
            {
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    roleIndex = 1;  // 1: Host.
                }
                else
                {
                    roleIndex = 2;  // 2: Client.
                }
            }

            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Changing Session Preprocessing");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum:%2d  Role=%s", nn::pia::session::Session::GetInstance()->GetValidStationIdNum(), SESSION_ROLE(roleIndex));
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationId:0x%016llX", localStationId.GetValue());
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Processing is in progress...");
            line += FONT_SIZE;

            line = SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    if (errorOccurred)
    {
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    // Scene transition branches based on post-process <tt>SessionStatus</tt>.
    switch (nn::pia::session::Session::GetInstance()->GetStatus())
    {
    case nn::pia::session::Session::STATUS_NOT_CONNECTED:
    {
        // Return to team selection.
        Session_CleanupSession();
        return SCENE_SELECT_TEAM;
    }
    case nn::pia::session::Session::STATUS_CONNECTED_SESSION:
    case nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION:
    case nn::pia::session::Session::STATUS_MIGRATING_SESSION:
    {
        return SCENE_CHANGING_SESSION;
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_SESSION:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_SESSION\n");
        SetCommunicationErrorType(SESSION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultSessionConnectionIsLost());
        SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
        if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            // Leaves the network in a state where it does not need to be reset, because the process can return to matchmaking when you clean up and restart the Session object.
            SetResetOfNetworkIsNecessary(false);
        }
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_NETWORK:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_NETWORK\n");
        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    default:
    {
        // This should never happen, given the specifications for this demo.
        PIA_SAMPLE_ASSERTMSG(false, "Unexpected MatchmakeStatus:%d", nn::pia::session::Session::GetInstance()->GetStatus());

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    }
}

// Scene in which asynchronous process is waiting during session migration
SampleSceneNo SceneChangingSession()
{
    nn::Result result = nn::ResultSuccess();

    u32 frameNo = 0;
    bool isExit = false;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_CHANGING_SESSION,
        true,
        false,
        0,
        0
        );

    nn::pia::util::HandlingType handlingType = nn::pia::util::HANDLING_VOID;
    SampleWaitingAsyncProcessing waitingProcess = SAMPLE_PASSIVE_PROCESSING;

    // Starts asynchronous process for migration.
    switch (g_GameState)
    {
    case GAME_STATE_TEAM:
    {  // Migrates from team session to game session.
        if (nn::pia::session::Session::GetInstance()->IsHost())
        {
            // Host.
#if PIA_TEST_CODE
            u16 maxParticipantNum = GetMaxPlayerNum();
#else
            u16 maxParticipantNum = SAMPLE_GAME_PLAYER_MAX;
#endif

            // Settings for the matchmaking session to create (for merging).
            nn::pia::inet::NexCreateSessionSetting createSessionSettingForGame;
            {
                // Sets the game mode to "game."
                createSessionSettingForGame.SetGameMode(MATCHMAKE_GAME_MODE_GAME);
                // All of the merged members will participate, so set the maximum number of participants to "the maximum number after merging."
                createSessionSettingForGame.SetMaxParticipants(maxParticipantNum);
                createSessionSettingForGame.SetMinParticipants(1);
                createSessionSettingForGame.SetAttribute(MATCHMAKE_ATTR_INDEX_TEAM_TYPE, static_cast<u32>(GetCurrentTeamType()));
#if PIA_TEST_CODE
                // Configures the test system to open the game to participation when the session is created.
                createSessionSettingForGame.SetOpenParticipation(true);
#else
                // Configures the settings to close the game to participation when session is created.
                createSessionSettingForGame.SetOpenParticipation(false);
#endif
            }
            // Random-match search condition settings.
            const size_t cSearchCriteriaNum = 1;
            nn::pia::inet::NexSessionSearchCriteria nexSearchCriteriaArray[cSearchCriteriaNum];
            nexSearchCriteriaArray[0].SetGameMode(MATCHMAKE_GAME_MODE_GAME);
            nexSearchCriteriaArray[0].SetMaxParticipants(maxParticipantNum);
            // Searches for team types that are different from own team type.
            const size_t attributeArraySize = TEAM_TYPE_NUM - 1;
            u32 attributeArray[attributeArraySize];
            size_t idx = 0;
            for (u32 i = 0; i < TEAM_TYPE_NUM; ++i)
            {
                if (static_cast<SampleTeamType>(i) != GetCurrentTeamType())
                {
                    attributeArray[idx] = i;
                    ++idx;
                }
            }
            nexSearchCriteriaArray[0].SetAttribute(MATCHMAKE_ATTR_INDEX_TEAM_TYPE, attributeArray, attributeArraySize);

            if (g_JointSessionType == JOINT_SESSION_TYPE_RANDOM)
            {
                result = nn::pia::session::Session::GetInstance()->JoinRandomJointSessionAsync(
                    &createSessionSettingForGame, nexSearchCriteriaArray, cSearchCriteriaNum);
                PIA_SAMPLE_LOG("Call JoinRandomJointSessionAsync.\n");
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Call JoinRandomJointSessionAsync is failure.\n");
                    PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                    isExit = true;
                }
                else
                {
                    g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOINT_RANDOM);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOINT_RANDOM);

                    waitingProcess = SAMPLE_JOIN_RANDOM_JOINT_SESSION;
                }
            }
#if PIA_TEST_CODE
            else if (g_JointSessionType == JOINT_SESSION_TYPE_CREATE
            ||       g_JointSessionType == JOINT_SESSION_TYPE_CREATE_PRIVATE)
            {
                createSessionSettingForGame.SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, (g_JointSessionType == JOINT_SESSION_TYPE_CREATE_PRIVATE) ? 1 : 0);
                if (g_JointSessionType == JOINT_SESSION_TYPE_CREATE_PRIVATE)
                {
                    createSessionSettingForGame.SetSessionUserPassword(USER_PASSWORD[g_UserPasswordIndexForJoint], USER_PASSWORD_LENGTH);
                }
                result = nn::pia::session::Session::GetInstance()->CreateJointSessionAsync(&createSessionSettingForGame);
                PIA_SAMPLE_LOG("Call CreateJointSessionAsync.\n");
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Call CreateJointSessionAsync is failure.\n");
                    PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                    isExit = true;
                }
                else
                {
                    g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_CREATE_JOINT);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_CREATE_JOINT);

                    waitingProcess = SAMPLE_CREATE_JOINT_SESSION;
                }
            }
            else if (g_JointSessionType == JOINT_SESSION_TYPE_JOIN
            ||       g_JointSessionType == JOINT_SESSION_TYPE_JOIN_PRIVATE)
            {
                nexSearchCriteriaArray[0].SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, (g_JointSessionType == JOINT_SESSION_TYPE_JOIN_PRIVATE) ? 1 : 0);
                nexSearchCriteriaArray[0].SetExcludeUserPasswordSet(!(g_JointSessionType == JOINT_SESSION_TYPE_JOIN_PRIVATE));
                nexSearchCriteriaArray[0].SetResultRange(0, 4);
#if PIA_TEST_CODE
                nexSearchCriteriaArray[0].SetOpenedOnly(false);
#endif

                result = nn::pia::session::Session::GetInstance()->BrowseSessionAsync(&nexSearchCriteriaArray[0]);
                PIA_SAMPLE_LOG("Call BrowseSessionAsync.\n");
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Call BrowseSessionAsync is failure.\n");
                    PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                    isExit = true;
                }
                else
                {
                    waitingProcess = SAMPLE_BROWSE_JOINT_SESSION;
                }
            }
#endif
        }
        else
        {
            // Given the specifications for this demo, execution on a non-host should never reach here for any status other than STATUS_MIGRATING_SESSION.
            PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
            // This state is reached when passive session migration is started.
            // Just wait to advance; asynchronous processing has begun within the library.
            waitingProcess = SAMPLE_PASSIVE_PROCESSING;
        }
        PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_PREPARING_GAME);
        g_GameState = GAME_STATE_PREPARING_GAME;
    }
    break;
    case GAME_STATE_GAME:
    {  // Migrates from game session to team session.
        if (g_JointSessionType == JOINT_SESSION_TYPE_DESTROY)
        {
            // Joint session host.
            if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
            {
                // Leave from game session and return to team session.
                PIA_SAMPLE_LOG("Call DestroyJointSessionAsync.\n");
                result = nn::pia::session::Session::GetInstance()->DestroyJointSessionAsync();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Call DestroyJointSessionAsync is failure.\n");
                    PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                    isExit = true;
                }
                else
                {
                    // Set anticipated event.
                    g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT);
                    g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_DESTROY_JOINT);
                    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
                    nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
                    for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
                    {
                        if (g_TeamConstitutionManager.GetTeamIndex(*it) != g_TeamConstitutionManager.GetTeamIndex(pSession->GetLocalStationId()))
                        {
                            g_SessionMigrationEventChecker.SetExpectedLeaveEventStationId(*it);
                        }
                    }

                    waitingProcess = SAMPLE_DESTROY_JOINT_SESSION;
                }
            }
            else
            {
                // No flag other than <tt>DestroyJointSession</tt> should be set.
                PIA_SAMPLE_ASSERTMSG((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT) != 0,
                    "DestroyJointSession is not processing!?");
                // Given the specifications for this demo, companion stations should not come here for any status other than STATUS_MIGRATING_SESSION.
                PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
            }
        }
#if PIA_TEST_CODE
        else if (g_JointSessionType == JOINT_SESSION_TYPE_LEAVE)
        {
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                // Check whether <tt>DestroyJointSession</tt> started before the joint session.
                if (g_SessionMigrationEventChecker.IsCheckEnabled() == false)
                {
                    PIA_SAMPLE_LOG("Call LeaveJointSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->LeaveJointSessionAsync();
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("Call LeaveJointSessionAsync is failure.\n");
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        isExit = true;
                    }
                    else
                    {
                        // Set anticipated event.
                        g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
                        g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_LEAVE_JOINT);
                        g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_LEAVE_JOINT);
                        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
                        nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
                        for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
                        {
                            if (g_TeamConstitutionManager.GetTeamIndex(*it) != g_TeamConstitutionManager.GetTeamIndex(pSession->GetLocalStationId()))
                            {
                                g_SessionMigrationEventChecker.SetExpectedLeaveEventStationId(*it);
                            }
                        }

                        waitingProcess = SAMPLE_LEAVE_JOINT_SESSION;
                    }
                }
                else
                {
                    // No flag other than <tt>DestroyJointSession</tt> should be set.
                    PIA_SAMPLE_ASSERTMSG((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT) != 0,
                        "DestroyJointSession is not processing!?");
                    // Given the specifications for this demo, companion stations should not come here for any status other than STATUS_MIGRATING_SESSION.
                    PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
                    PIA_SAMPLE_LOG("Received DestroyJointSession Invitation before calling LeaveJointSessionAsync.\n");
                }
            }
            else
            {
                PIA_SAMPLE_ASSERTMSG((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_LEAVE_JOINT) != 0,
                    "LeaveJointSession is not processing!?");
                // Given the specifications for this demo, companion stations should not come here for any status other than STATUS_MIGRATING_SESSION.
                PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION);
            }
        }
#endif
        PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_PREPARING_TEAM);
        g_GameState = GAME_STATE_PREPARING_TEAM;
    }
    break;
    default:
    {
        PIA_SAMPLE_ASSERT(false);
    }
    break;
    }

#if PIA_TEST_CODE
    bool isStopJointSessionProcess = false;
#endif
    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Execute dispatch.
#if PIA_TEST_CODE
        bool isSkipDispatch = isStopJointSessionProcess && (nn::pia::session::Mesh::GetInstance()->GetMeshState() == nn::pia::session::Mesh::STATE_DISCONNECTED);
        if (!isSkipDispatch)
        {
            nn::pia::common::Scheduler::GetInstance()->Dispatch();
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }
#else
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        nn::nex::Scheduler::GetInstance()->DispatchAll();
#endif

        // Check for internal NEX errors.
        if (Nex_CheckInternalError())
        {
            result = nn::pia::ResultNetworkConnectionIsLost();
            isExit = true;
        }

#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_JOINT_SEQUENCE, nn::pia::session::Session::GetInstance()->GetJointSequence());

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
            PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());

            if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
            }
            else
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
            }
        }

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            // Clients that are passively included in joint sessions can call the LeaveSession function, but this is not recommended.
            if (waitingProcess == SAMPLE_PASSIVE_PROCESSING)
            {
                PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());
                nn::pia::session::Session::GetInstance()->LeaveSessionAsync();
                waitingProcess = SAMPLE_LEAVE_SESSION;
            }
        }
        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            // Makes the process crash, for testing purposes.
            // Your application should wait for the asynchronous process to end and then respond appropriately to the failure.
            result = nn::pia::ResultNetworkConnectionIsLost();
            isExit = true;
        }
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT)
        &&  nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
        {
            PIA_SAMPLE_LOG("isStopJointSessionProcess is set.\n");
            isStopJointSessionProcess = true;
        }
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP)
        &&  nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            PIA_SAMPLE_LOG("isStopJointSessionProcess is unset.\n");
            isStopJointSessionProcess = false;
        }
#endif


        // Check for the completion of the asynchronous process,
        switch (waitingProcess)
        {
        case SAMPLE_JOIN_RANDOM_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedJoinRandomJointSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetJoinRandomJointSessionResult();
                if (result.IsSuccess())
                {
#if PIA_TEST_CODE
                    g_IsOpened = true;
#else
                    g_IsOpened = (nn::pia::session::Session::GetInstance()->IsJointSessionHost()) ? false : true;
#endif
                }
            }
        }
        break;
#if PIA_TEST_CODE
        case SAMPLE_CREATE_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedCreateJointSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetCreateJointSessionResult();
                if (result.IsSuccess())
                {
#if PIA_TEST_CODE
                    g_IsOpened = true;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_JOIN_JOINT_SESSION_ID, nn::pia::session::Session::GetInstance()->GetJointSessionId());
#else
                    g_IsOpened = false;
#endif
                }
            }
        }
        break;
        case SAMPLE_BROWSE_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedBrowseSession())
            {
                result = nn::pia::session::Session::GetInstance()->GetBrowseSessionResult();
                nn::pia::session::ISessionInfoList* pSessionInfoList = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("BrowseSession failed.\n");
                    result = nn::pia::ResultSessionIsNotFound();
                    isExit = true;
                }
                else if (pSessionInfoList->GetSize() == 0)
                {
                    PIA_SAMPLE_LOG("Session is not found.\n");
                    result = nn::pia::ResultSessionIsNotFound();
                    isExit = true;
                }
                else
                {
#if PIA_TEST_CODE
                    nn::pia::inet::NexSessionInfo* info = static_cast<nn::pia::inet::NexSessionInfo*>(*(pSessionInfoList->Begin()));
                    size_t bufSize = 0;
                    const int appDataBufSize = 16;
                    u8 applicationData[appDataBufSize];

                    bufSize = info->GetApplicationDataSize();
                    info->GetApplicationDataBuf(&applicationData, appDataBufSize);

                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_APPLICATION_DATA_HEAD, applicationData[0]);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_APPLICATION_DATA_SIZE, bufSize);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_JOIN_JOINT_SESSION_ID, info->GetSessionId());

                    PIA_SAMPLE_LOG("JointSessionId: %d\n", info->GetSessionId());
                    for (int i=0;i<appDataBufSize;++i)
                    {
                        PIA_SAMPLE_LOG("App Data buf[%d]: %d\n", i, applicationData[i]);
                    }
                    PIA_SAMPLE_LOG("App Data Size: %d\n", bufSize);
#endif

                    nn::pia::inet::NexJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(*(pSessionInfoList->Begin()));
                    if (g_JointSessionType == JOINT_SESSION_TYPE_JOIN_PRIVATE)
                    {
                        joinSessionSetting.SetSessionUserPassword(USER_PASSWORD[g_UserPasswordIndexForJoint], USER_PASSWORD_LENGTH);
                    }

                    PIA_SAMPLE_LOG("Call JoinJointSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->JoinJointSessionAsync(&joinSessionSetting);
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("Call JoinJointSessionAsync is failure.\n");
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        isExit = true;
                    }
                    else
                    {
                        g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
                        g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOIN_JOINT);
                        g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOIN_JOINT);
                        waitingProcess = SAMPLE_JOIN_JOINT_SESSION;
                    }
                }
            }
        }
        break;
        case SAMPLE_JOIN_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedJoinJointSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetJoinJointSessionResult();
                if (result.IsSuccess())
                {
                    g_IsOpened = true;
            }
        }
        }
        break;
        case SAMPLE_LEAVE_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedLeaveJointSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetLeaveJointSessionResult();
                if (result.IsSuccess())
                {
                    g_IsOpened = false;
            }
        }
        }
        break;
#endif
        case SAMPLE_DESTROY_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedDestroyJointSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetDestroyJointSessionResult();
                if (result.IsSuccess())
                {
                    g_IsOpened = false;
            }
        }
        }
        break;
        case SAMPLE_PASSIVE_PROCESSING:
        {
            // Client that did not start the asynchronous process itself (the process was started passively on instruction from the host).
            // Check the progress of the process using the "session migrating" flag.
            if (!IsSessionMigrating())
            {
                result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
                if (result.IsFailure())
                {
                    if (result == nn::pia::ResultNetworkConnectionIsLost())
                    {
                        PIA_SAMPLE_LOG("CheckConnectionError failed: ResultNetworkConnectionIsLost\n");
                    }
                    else if (result == nn::pia::ResultSessionConnectionIsLost())
                    {
                        PIA_SAMPLE_LOG("CheckConnectionError failed: ResultSessionConnectionIsLost\n");
                    }
                }
                switch (g_JointSessionType)
                {
                case JOINT_SESSION_TYPE_CREATE:
                case JOINT_SESSION_TYPE_CREATE_PRIVATE:
                {
#if PIA_TEST_CODE
                    g_IsOpened = true;
#else
                    g_IsOpened = false;
#endif
                }
                break;
                case JOINT_SESSION_TYPE_JOIN:
                case JOINT_SESSION_TYPE_JOIN_PRIVATE:
                {
                    g_IsOpened = true;
                }
                break;
                case JOINT_SESSION_TYPE_RANDOM:
                {
#if PIA_TEST_CODE
                    g_IsOpened = true;
#else
                    g_IsOpened = (nn::pia::session::Session::GetInstance()->IsJointSessionHost()) ? false : true;
#endif
                }
                break;

                case JOINT_SESSION_TYPE_LEAVE:
                case JOINT_SESSION_TYPE_DESTROY:
                {
                    g_IsOpened = false;
                }
                break;
                }
                isExit = true;
            }
        }
        break;
#if PIA_TEST_CODE
        case SAMPLE_LEAVE_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedLeaveSession())
            {
                isExit = true;
                result = nn::pia::session::Session::GetInstance()->GetLeaveSessionResult();
            }
        }
        break;
#endif
        }

        if (result.IsFailure())
        {
            if (result == nn::pia::ResultSessionIsNotFound())
            {
                // Directly sets the type of handling for the given sample.
                handlingType = nn::pia::util::HANDLING_LEAVE_AND_CLEANUP;
            }
            else
            {
                handlingType = nn::pia::util::GetHandlingType(result);
            }
            if (result == nn::pia::ResultNetworkIsFull())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            }
            else
            {
                SetCommunicationErrorType(MATCHMAKE_ERROR);
            }
            SetNetworkErrorResult(result);
        }

        //Screen rendering.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        {
            //Check the role of the local station.
            int roleIndex = 0;  //Unknown.
            nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
            if (localStationId != nn::pia::STATION_ID_INVALID)
            {
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    roleIndex = 1;  // 1: Host.
                }
                else
                {
                    roleIndex = 2;  // 2: Client.
                }
            }

            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Changing Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum:%2d  Role=%s", nn::pia::session::Session::GetInstance()->GetValidStationIdNum(), SESSION_ROLE(roleIndex));
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationId:0x%016llX", localStationId.GetValue());
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Processing is in progress...");
            line += FONT_SIZE;

            line = SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    if (handlingType != nn::pia::util::HANDLING_VOID)
    {
        if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP
        ||  handlingType == nn::pia::util::HANDLING_LEAVE)
        {
            SampleSessionAsyncStatus status = Session_LeaveSession();
            if (status == SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                SetResetOfNetworkIsNecessary(false);
            }
        }
        else if (handlingType == nn::pia::util::HANDLING_LOGOUT
        ||       handlingType == nn::pia::util::HANDLING_RECONNECT
        ||       handlingType == nn::pia::util::HANDLING_REINITIALIZE)
        {
            SetResetOfNetworkIsNecessary(true);
        }
        else if (handlingType == nn::pia::util::HANDLING_PROGRAMMING_ERROR)
        {
            PIA_SAMPLE_ASSERT(false);
        }
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    // Scene transition branches based on post-process <tt>SessionStatus</tt>.
    switch (nn::pia::session::Session::GetInstance()->GetStatus())
    {
    case nn::pia::session::Session::STATUS_NOT_CONNECTED:
    {
        // Return to team selection.
        Session_CleanupSession();
        return SCENE_SELECT_TEAM;
    }
    case nn::pia::session::Session::STATUS_CONNECTED_SESSION:
    case nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION:
    {
        switch (g_GameState)
        {
        case GAME_STATE_PREPARING_GAME:
            {
                if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
                {
                    // Did session migration continuously occur?
                    PIA_SAMPLE_LOG("Session migrated continuously? GAME_STATE_PREPARING_GAME, STATUS_CONNECTED_SESSION\n");
                    PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_TEAM);
                    g_GameState = GAME_STATE_TEAM;
                    return SCENE_CHANGING_SESSION_POSTPROCESSING;
                }
                PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION);
                PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_GAME);
                g_GameState = GAME_STATE_GAME;
                return SCENE_CHANGING_SESSION_POSTPROCESSING;
            }
        case GAME_STATE_PREPARING_TEAM:
            {
                if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION)
                {
                    // Did session migration continuously occur?
                    PIA_SAMPLE_LOG("Session migrated continuously? GAME_STATE_PREPARING_TEAM, STATUS_CONNECTED_JOINT_SESSION\n");
                    PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_GAME);
                    g_GameState = GAME_STATE_GAME;
                    return SCENE_CHANGING_SESSION_POSTPROCESSING;
                }
                PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION);
                PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_TEAM);
                g_GameState = GAME_STATE_TEAM;
                return SCENE_CHANGING_SESSION_POSTPROCESSING;
            }
        default:
            {
                PIA_SAMPLE_ASSERT(false);
                Session_CleanupSession();
                return SCENE_NETWORK_ERROR;
            }
        }
    }
    case nn::pia::session::Session::STATUS_MIGRATING_SESSION:
    {
        PIA_SAMPLE_LOG("STATUS_MIGRATING_SESSION\n");
        // Immediately after joining a joint session, the disbanding of that joint session began.
        PIA_SAMPLE_ASSERT(waitingProcess == SAMPLE_JOIN_RANDOM_JOINT_SESSION
                       || waitingProcess == SAMPLE_CREATE_JOINT_SESSION
                       || waitingProcess == SAMPLE_JOIN_JOINT_SESSION);
        // Skip the post-processing to transition to a communication scenario in the game.
        PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_GAME);
        PIA_SAMPLE_ASSERT(g_GameState == GAME_STATE_PREPARING_GAME);
        g_GameState = GAME_STATE_GAME;
        return SCENE_COMMUNICATE_GAME;
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_SESSION:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_SESSION\n");
        SetCommunicationErrorType(SESSION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultSessionConnectionIsLost());
        SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
        if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            // Leaves the network in a state where it does not need to be reset, because the process can return to matchmaking when you clean up and restart the Session object.
            SetResetOfNetworkIsNecessary(false);
        }
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_NETWORK:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_NETWORK\n");
        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    default:
    {
        // This should never happen, given the specifications for this demo.
        PIA_SAMPLE_ASSERTMSG(false, "Unexpected MatchmakeStatus:%d", nn::pia::session::Session::GetInstance()->GetStatus());

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    }
}

// Processing after finishing session migration.
SampleSceneNo SceneChangingSessionPostprocessing()
{
    nn::Result result = nn::ResultSuccess();

    u32 frameNo = 0;
    bool isExit = false;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_CHANGING_SESSION_POSTPROCESSING,
        true,
        false,
        0,
        0
        );

    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();

    nn::pia::util::HandlingType handlingType = nn::pia::util::HANDLING_VOID;
    SampleWaitingAsyncProcessing waitingProcess = SAMPLE_PASSIVE_PROCESSING;

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Updates the clock for CloneProtocol.
        Clone_GetCloneProtocol()->UpdateClock();

        // Execute dispatch.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        if (waitingProcess == SAMPLE_PASSIVE_PROCESSING)
        {
            result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
            if (result.IsFailure())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                isExit = true;
            }
            // Check for internal NEX errors.
            else if (Nex_CheckInternalError())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                result = nn::pia::ResultNetworkConnectionIsLost();
                isExit = true;
            }
            // Waits for <tt>CloneProtocol</tt> to become active.
            else if (Clone_GetCloneProtocol()->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE)
            {
                switch (g_GameState)
                {
                case GAME_STATE_TEAM:
                {  // Migration to team session is complete.

                    // Unregister callback for determining whether to approve a join.
                    nn::pia::session::Session::GetInstance()->UnregisterJoiningApprovalCallback();

#if PIA_TEST_CODE
#else
                    if (nn::pia::session::Session::GetInstance()->IsHost())
                    {
                        PIA_SAMPLE_LOG("Call OpenSessionAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("Call OpenSessionAsync failed.\n");
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                            isExit = true;
                        }
                        else
                        {
                            waitingProcess = SAMPLE_OPEN_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_SESSION));
                        }
                    }
                    else
#endif
                    {
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                        isExit = true;
                    }
                }
                break;
                case GAME_STATE_GAME:
                {  // Migration to game session is complete.

                    // Unregister callback for determining whether to approve a join.
                    nn::pia::session::Session::GetInstance()->UnregisterJoiningApprovalCallback();

#if PIA_TEST_CODE
#else
                    if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
                    {
                        // Checks the number of teams participating in the joint session.
                        nn::pia::StationId jointSessionHostStationId = nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId();
                        u32 sessionIdOfJointSessionHost = nn::pia::session::Session::GetInstance()->GetSessionId(jointSessionHostStationId);
                        bool isInSingleTeam = true;
                        for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
                                                                          it != nn::pia::session::Session::GetInstance()->End(); ++it)
                        {
                            if (sessionIdOfJointSessionHost != nn::pia::session::Session::GetInstance()->GetSessionId(*it))
                            {
                                isInSingleTeam = false;
                                break;
                            }
                        }
                        // If there is one team, it means the joint session was created in the closed state, so open it to participation.
                        if (isInSingleTeam)
                        {
                            PIA_SAMPLE_LOG("Call OpenJointSessionAsync.\n");
                            result = nn::pia::session::Session::GetInstance()->OpenJointSessionAsync();
                            if (result.IsFailure())
                            {
                                PIA_SAMPLE_LOG("Call OpenJointSessionAsync failed.\n");
                                PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                                isExit = true;
                            }
                            else
                            {
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_JOINT_SESSION));
                                waitingProcess = SAMPLE_OPEN_JOINT_SESSION;
                            }
                        }
                        else
                        {
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                            isExit = true;
                        }
                    }
                    else
#endif
                    {
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                        isExit = true;
                    }
                }
                break;
                default:
                {
                    PIA_SAMPLE_ASSERT(false);
                }
                break;
                }
            }
            // The session-migration process began while waiting for <tt>CloneProtocol</tt> to become active.
            else if (IsSessionMigrating())
            {
                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                isExit = true;
            }
        }

#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_JOINT_SEQUENCE, nn::pia::session::Session::GetInstance()->GetJointSequence());

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
            PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());
            if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
            }
            else
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
            }
        }
#endif

        // Check for the completion of the asynchronous process,
        switch (waitingProcess)
        {
        case SAMPLE_OPEN_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenSession())
            {
                result = nn::pia::session::Session::GetInstance()->GetOpenSessionResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("OpenSession succeeded.\n");
                    g_IsOpened = true;
                }
                else
                {
                    PIA_SAMPLE_LOG("OpenSession failed.\n");
                }
                isExit = true;
                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
            }
        }
        break;
        case SAMPLE_OPEN_JOINT_SESSION:
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenJointSession())
            {
                result = nn::pia::session::Session::GetInstance()->GetOpenJointSessionResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("OpenJointSession succeeded.\n");
                    g_IsOpened = true;
                }
                else
                {
                    PIA_SAMPLE_LOG("OpenJointSession failed.\n");
                }
                isExit = true;
                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
            }
        }
        break;
        }
        if (result.IsFailure())
        {
            if (result == nn::pia::ResultNetworkConnectionIsLost())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            }
            else if (result == nn::pia::ResultSessionConnectionIsLost())
            {
                SetCommunicationErrorType(SESSION_LOST_ERROR);
            }
            else if (result == nn::pia::ResultSessionIsNotFound())
            {
                SetCommunicationErrorType(SESSION_LOST_ERROR);
            }
            else if (result == nn::pia::ResultNexInternalError() || result == nn::pia::ResultGameServerProcessAborted())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            }
            else
            {
                PIA_TRACE_RESULT(result);
                PIA_SAMPLE_ASSERT(false);
            }
            SetNetworkErrorResult(result);
            handlingType = nn::pia::util::GetHandlingType(result);
        }

        UpdateStatusManager();

        //Screen rendering.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        {
            //Check the role of the local station.
            int roleIndex = 0;  //Unknown.
            localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
            if (localStationId != nn::pia::STATION_ID_INVALID)
            {
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    roleIndex = 1;  // 1: Host.
                }
                else
                {
                    roleIndex = 2;  // 2: Client.
                }
            }

            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Changing Session Postprocessing");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum:%2d  Role=%s", nn::pia::session::Session::GetInstance()->GetValidStationIdNum(), SESSION_ROLE(roleIndex));
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationId:%lld", localStationId.GetValue());
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Processing is in progress...");
            line += FONT_SIZE;

            line = SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    if (handlingType != nn::pia::util::HANDLING_VOID)
    {
        if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP
        ||  handlingType == nn::pia::util::HANDLING_LEAVE)
        {
            SampleSessionAsyncStatus status = Session_LeaveSession();
            if (status == SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                SetResetOfNetworkIsNecessary(false);
            }
        }
        else if (handlingType == nn::pia::util::HANDLING_LOGOUT
        ||       handlingType == nn::pia::util::HANDLING_RECONNECT
        ||       handlingType == nn::pia::util::HANDLING_REINITIALIZE)
        {
            SetResetOfNetworkIsNecessary(true);
        }
        else if (handlingType == nn::pia::util::HANDLING_PROGRAMMING_ERROR)
        {
            PIA_SAMPLE_ASSERT(false);
        }
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    // Scene transition branches based on post-process <tt>SessionStatus</tt>.
    switch (nn::pia::session::Session::GetInstance()->GetStatus())
    {
    case nn::pia::session::Session::STATUS_NOT_CONNECTED:
    {
        // Return to team selection.
        Session_CleanupSession();
        return SCENE_SELECT_TEAM;
    }
    case nn::pia::session::Session::STATUS_CONNECTED_SESSION:
    case nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION:
    case nn::pia::session::Session::STATUS_MIGRATING_SESSION:
    {
        switch (g_GameState)
        {
        case GAME_STATE_GAME:
            {
                return SCENE_COMMUNICATE_GAME;
            }
        case GAME_STATE_TEAM:
            {
                return SCENE_COMMUNICATE_TEAM;
            }
        default:
            {
                PIA_SAMPLE_ASSERT(false);
                Session_CleanupSession();
                return SCENE_NETWORK_ERROR;
            }
        }
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_SESSION:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_SESSION\n");
        SetCommunicationErrorType(SESSION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultSessionConnectionIsLost());
        SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
        if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            // Leaves the network in a state where it does not need to be reset because the process can return to matchmaking when you clean up and restart the Session object.
            SetResetOfNetworkIsNecessary(false);
        }
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    case nn::pia::session::Session::STATUS_DISCONNECTED_NETWORK:
    {
        PIA_SAMPLE_LOG("STATUS_DISCONNECTED_NETWORK\n");
        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    default:
    {
        // This should never happen, given the specifications for this demo.
        PIA_SAMPLE_ASSERTMSG(false, "Unexpected MatchmakeStatus:%d", nn::pia::session::Session::GetInstance()->GetStatus());

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }
    }
}
