﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneCommunication.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/inet.h>
#include <pia/transport.h>
#include <pia/session.h>
#include <pia/util.h>

#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Clone_Setup.h"
#include "SampleLib/SampleLib.h"
#include "SampleObjectStatus.h"

#include "EventChecker.h"
#include "TeamConstitutionManager.h"
#include "ObjectManager.h"
#include "StatusManager.h"

#include "Shared_Nex_Setup.h"
#include <nex.h>
/*!

    @brief  This CPP file contains processing that is independent of the network type (<tt>session</tt> + <tt>transport</tt>).

*/


SampleTeamType GetCurrentTeamType();
const nn::pia::sampleframework::Color GetColorOfTeamType(SampleTeamType teamType);

//
// Global Variables
//

// For managing the station status.
StatusManager g_StatusManager;

// For checking events.
EventChecker g_SessionMigrationEventChecker;
EventChecker g_HostMigrationEventChecker;


namespace
{

enum SessionState
{
    SESSION_STATE_NONE = 0,
    SESSION_STATE_SESSION,
    SESSION_STATE_JOINT_SESSION,
    SESSION_STATE_MIGRATING_TO_SESSION,
    SESSION_STATE_MIGRATING_TO_JOINT_SESSION
};
SessionState s_SessionState = SESSION_STATE_NONE;
bool s_IsNeedModifyAttribute = false;

bool s_HostChangedEventOccurred = false;

bool s_IsLeaveEventForLocalOccurred = false;

bool s_IsErrorOccurred = false;

bool s_IsLeaveProcessing = false;

const size_t SAMPLE_MAX_SEND_THREAD_BUFFER_NUM = 16;
const size_t SAMPLE_MAX_RECEIVE_THREAD_BUFFER_NUM = 32;

void BindObjectToStatus()
{
    // StationIndex must re-bind Object to StationStatus at varied times.
    for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
    {
        SampleObjectStatus *pStatus = g_StatusManager.GetStationStatus(i);
        pStatus->pObject = g_ObjectManager.GetReceiveObject(pStatus->stationId);
    }
}

void AssignCommonObjectsStatus()
{
    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        g_StatusManager.AssignCommonObjectStatus(OBJECT_STATUS_TYPE_TEAM_ATOMIC, i, g_ObjectManager.GetTeamAtomicObject(i));
        g_StatusManager.AssignCommonObjectStatus(OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL, i, g_ObjectManager.GetTeamSequentialObject(i));
    }
    g_StatusManager.AssignCommonObjectStatus(OBJECT_STATUS_TYPE_COMMON_ATOMIC, SAMPLE_TEAM_NUM_IN_GAME, g_ObjectManager.GetCommonAtomicObject());
    g_StatusManager.AssignCommonObjectStatus(OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL, SAMPLE_TEAM_NUM_IN_GAME, g_ObjectManager.GetCommonSequentialObject());
}

const size_t cColorOfRectangleNum = 8;
const nn::pia::sampleframework::Color COLOR_OF_RECTANGLE[cColorOfRectangleNum] =
    {
        nn::pia::sampleframework::COLOR_SILVER,
        nn::pia::sampleframework::COLOR_RED,
        nn::pia::sampleframework::COLOR_LIME,
        nn::pia::sampleframework::COLOR_OLIVE,
        nn::pia::sampleframework::COLOR_PURPLE,
        nn::pia::sampleframework::COLOR_TEAL,
        nn::pia::sampleframework::COLOR_MAROON,
        nn::pia::sampleframework::COLOR_GREEN
    };

// The field settings when drawing a rectangle.
const s32 cTeamFieldWidth = 340 / SAMPLE_TEAM_NUM_IN_GAME;
const s32 cTeamFieldHeight = 70;
const s32 cTeamFieldMargin = 5;
const s32 cLeftTopX = 20;
const s32 cLeftTopY = 48;

void DrawFrame(s32 x, s32 y, s32 width, s32 height, const nn::pia::sampleframework::Color& col)
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(x, y, 2, height, col);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(x, y, width, 2, col);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(x + width - 2, y, 2, height, col);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(x, y + height - 2, width, 2, col);
}

void DrawTeamFrame(int teamIndex, const nn::pia::sampleframework::Color& teamColor)
{
    s32 posLeftTopX = cLeftTopX + (cTeamFieldWidth + cTeamFieldMargin) * teamIndex;
    s32 posLeftTopY = cLeftTopY;

    s32 width = cTeamFieldWidth;
    s32 height = cTeamFieldHeight;

    // Adjusts only the thickness of the border so the border extends outside the assumed region.
    DrawFrame(posLeftTopX - 2, posLeftTopY - 2, width + 4, height + 4, teamColor);
}

const s32 cRectSize = 8;

RectPosition MoveRectPosition(SampleObjectStatus* pStatus, s32 diffX, s32 diffY)
{
    s32 xBuf;
    s32 yBuf;

    if (pStatus->type == OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL 
    ||  pStatus->type == OBJECT_STATUS_TYPE_COMMON_ATOMIC)
    {
        const s32 cWidth = cTeamFieldWidth * SAMPLE_TEAM_NUM_IN_GAME;
        xBuf = (static_cast<s32>(pStatus->pos.x) + diffX + cWidth) % cWidth;
        yBuf = (static_cast<s32>(pStatus->pos.y));
    }
    else
    {
        // Adjusts the maximum value so edge of the field does not stick out.
        const s32 cWidth = cTeamFieldWidth - cRectSize;
        const s32 cHeight = cTeamFieldHeight - cRectSize;

        xBuf = (static_cast<s32>(pStatus->pos.x) + diffX + cWidth) % cWidth;
        yBuf = (static_cast<s32>(pStatus->pos.y) + diffY + cHeight) % cHeight;
    }
    RectPosition ret = { static_cast<u16>(xBuf), static_cast<u16>(yBuf) };
    return ret;
}

void StartCloneProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(Clone_GetCloneProtocol()));
    PIA_SAMPLE_ASSERT(Clone_GetCloneProtocol()->GetState() == nn::pia::clone::CloneProtocol::STATE_INACTIVE);
    nn::Result r = Clone_GetCloneProtocol()->Start();
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
}

void DrawStationStatus(const SampleObjectStatus& target)
{
    int teamIndex = g_TeamConstitutionManager.GetTeamIndex(target.stationId);
    // Stations that have not decided on a team are not drawn.
    if (target.type == OBJECT_STATUS_TYPE_STATION && teamIndex == -1)
    {
        return;
    }
    if ((target.type == OBJECT_STATUS_TYPE_TEAM_ATOMIC 
    ||   target.type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL)
    &&   target.teamIndex == -1)
    {
        return;
    }

    // Adjusts the drawing position based on the team.
    s32 offsetX, offsetY;
    if (target.type == OBJECT_STATUS_TYPE_STATION)
    {
        offsetX = cLeftTopX + (cTeamFieldWidth + cTeamFieldMargin) * teamIndex;
        offsetY = cLeftTopY;
    }
    else if (target.type == OBJECT_STATUS_TYPE_TEAM_ATOMIC 
    ||       target.type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL)
    {
        offsetX = cLeftTopX + (cTeamFieldWidth + cTeamFieldMargin) * target.teamIndex;
        offsetY = cLeftTopY;
    }
    else
    {
        offsetX = cLeftTopX;
        offsetY = cLeftTopY;
    }

    u16 targetId;
    if (target.type == OBJECT_STATUS_TYPE_STATION)
    {
        // Gets and uses the index in the team.
        targetId = g_TeamConstitutionManager.GetMemberIndex(target.stationId);
    }
    else
    {
        // A non-station object.
        targetId = SAMPLE_TEAM_MEMBER_MAX;
    }

    // Draws rectangle using the position information.
    s32 xBuf = offsetX + target.pos.x;
    s32 yBuf = offsetY + target.pos.y;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        xBuf, yBuf, cRectSize, cRectSize, COLOR_OF_RECTANGLE[targetId % cColorOfRectangleNum]);
    if (&target == g_StatusManager.GetOperationTargetStatus())
    {
        const nn::pia::sampleframework::Color cOperationTargetColor(1.0f, 0.3f, 0.0f, 1.0f);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(cOperationTargetColor);
    }
    switch(target.type)
    {
    case OBJECT_STATUS_TYPE_STATION:
        {
            
#if PIA_TEST_CODE
            nn::pia::transport::Station* pStationBuf = nn::pia::transport::StationManager::GetInstance()->GetStation(target.stationId);
            int rtt = 0;
            if (PIA_IS_VALID_POINTER(pStationBuf))
            {
                // Gets the RTT from the station instance corresponding to <tt>StationId</tt>.
                rtt = pStationBuf->GetRtt();
                if(rtt == nn::pia::transport::Station::INVALID_RTT)
                {
                    rtt = 0;
                }
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                xBuf + cRectSize, yBuf, "<-%X(%d), RTT:0x%03X", targetId, target.operationType, rtt);
#else
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                xBuf + cRectSize, yBuf, "<-%X(%d)", targetId, target.operationType);
#endif
        }
        break;
    case OBJECT_STATUS_TYPE_TEAM_ATOMIC:
        {
            if (target.pObject->GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_LOCKED)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-A!");
            }
            else if (target.pObject->GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_TRY_LOCK)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-A?");
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-A");
            }
        }
        break;
    case OBJECT_STATUS_TYPE_COMMON_ATOMIC:
        {
            if (target.pObject->GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_LOCKED)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-CA!");
            }
            else if (target.pObject->GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_TRY_LOCK)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-CA?");
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                    xBuf + cRectSize, yBuf, "<-CA");
            }
        }
        break;
    case OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL:
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                xBuf + cRectSize, yBuf, "<-Q");
        }
        break;
    case OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL:
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                xBuf + cRectSize, yBuf, "<-CQ");
        }
        break;
    }
    if (&target == g_StatusManager.GetOperationTargetStatus())
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    }
}

enum SampleWaitingAsyncProcessing
{
    SAMPLE_LEAVE_SESSION,
    SAMPLE_MODIFY_ATTRIBUTE,
    SAMPLE_OPEN_PARTICIPATION,
    SAMPLE_CLOSE_PARTICIPATION,
    SAMPLE_OPEN_SESSION,
    SAMPLE_CLOSE_SESSION,
    SAMPLE_OPEN_JOINT_SESSION,
    SAMPLE_CLOSE_JOINT_SESSION,
    SAMPLE_UPDATE_JOINT_SESSION_SETTING,
    SAMPLE_UPDATE_JOINT_SESSION_APPLICATION_DATA
#if PIA_TEST_CODE
    ,SAMPLE_OPEN_JOINT_SESSION_PARTICIPATION,
    SAMPLE_CLOSE_JOINT_SESSION_PARTICIPATION
#endif
};
}  // end of anonymous namespace


//The string for displaying the role screen within a session.
const char* SESSION_ROLE(size_t index)
{
    switch (index)
    {
    case 1:
        return "Host";
    case 2:
        return "Client.";
    }
    return "Unknown.";
}

//The string for displaying the role screen within a joint session.
const char* JOINT_SESSION_ROLE(size_t index)
{
    switch (index)
    {
    case 1:
        return "JointSessionHost";
    case 2:
        return "JointSessionClient";
    }
    return "Unknown.";
}

void ResetErrorOccurredFlag()
{
    s_IsErrorOccurred = false;
}

void UpdateSendData(s32 diffX, s32 diffY)
{
    // Sets data to the object.
    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(localStationId);
    PIA_SAMPLE_ASSERT(pStationStatus != NULL);
    SampleObjectStatus* pOperationTargetStatus = g_StatusManager.GetOperationTargetStatus();
    PIA_SAMPLE_ASSERT(pOperationTargetStatus != NULL);

    if (g_ObjectManager.GetSendObject()->m_ElementIncrementNumber.IsReadyToSetValue())
    {
        g_ObjectManager.GetSendObject()->m_ElementIncrementNumber.SetValue(pStationStatus->latestIncrementNumber + 1);
    }
    pStationStatus->stationNumber = nn::pia::session::Session::GetInstance()->GetStationNum();

    if (g_ObjectManager.GetSendObject()->m_ElementIsStartJointSession.IsReadyToSetValue())
    {
        g_ObjectManager.GetSendObject()->m_ElementIsStartJointSession.SetValue(pStationStatus->isStartJointSession);
    }
    if (g_ObjectManager.GetSendObject()->m_ElementOperationType.IsReadyToSetValue())
    {
        g_ObjectManager.GetSendObject()->m_ElementOperationType.SetValue(pStationStatus->operationType);
    }
    if (g_StatusManager.GetOperationTargetStatus()->type == OBJECT_STATUS_TYPE_STATION && g_StatusManager.GetOperationTargetStatus()->stationId == localStationId)
    {
        if (g_ObjectManager.GetSendObject()->m_ElementRectPosition.IsReadyToSetValue())
        {
            g_ObjectManager.GetSendObject()->m_ElementRectPosition.SetValue(MoveRectPosition(pStationStatus, diffX, diffY));
        }
        if (g_ObjectManager.GetSendObject()->m_ElementStationNumber.IsReadyToSetValue())
        {
            g_ObjectManager.GetSendObject()->m_ElementStationNumber.SetValue(pStationStatus->stationNumber);
        }
    }
    else
    {
        if (g_ObjectManager.GetSendObject()->m_ElementRectPosition.IsReadyToSetValue())
        {
            g_ObjectManager.GetSendObject()->m_ElementRectPosition.SetValue(MoveRectPosition(pStationStatus, 1, 0));
        }
    }

    if (pOperationTargetStatus->type == OBJECT_STATUS_TYPE_TEAM_ATOMIC ||
        pOperationTargetStatus->type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL ||
        pOperationTargetStatus->type == OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL ||
        pOperationTargetStatus->type == OBJECT_STATUS_TYPE_COMMON_ATOMIC
        )
    {
        SampleObjectStatus* pCommonObjectStatus = g_StatusManager.GetCommonObjectStatusByType(pOperationTargetStatus->type, GetCurrentTeamType());
        PIA_SAMPLE_ASSERT(pCommonObjectStatus != NULL);
        Object* pObj = pCommonObjectStatus->pObject;
        PIA_SAMPLE_ASSERT(pObj != NULL);
        if (pObj->m_ElementIncrementNumber.IsReadyToSetValue())
        {
            pObj->m_ElementIncrementNumber.SetValue(pCommonObjectStatus->latestIncrementNumber + 1);
        }
        if (pObj->m_ElementRectPosition.IsReadyToSetValue())
        {
            pObj->m_ElementRectPosition.SetValue(MoveRectPosition(pCommonObjectStatus, diffX, diffY));
        }
        if (pObj->m_ElementStationNumber.IsReadyToSetValue())
        {
            pObj->m_ElementStationNumber.SetValue(pCommonObjectStatus->stationNumber);
        }

        if (pObj->m_ElementIsStartJointSession.IsReadyToSetValue())
        {
            pObj->m_ElementIsStartJointSession.SetValue(pCommonObjectStatus->isStartJointSession);
        }
        if (pObj->m_ElementOperationType.IsReadyToSetValue())
        {
            pObj->m_ElementOperationType.SetValue(pCommonObjectStatus->operationType);
        }
    }
}

void SetLocalOperationType(u32 type)
{
    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(localStationId);
    if (pStationStatus != NULL)
    {
        PIA_SAMPLE_LOG("SetLocalOperationType: %d -> %d\n", pStationStatus->operationType, type);
        pStationStatus->operationType = type;
        UpdateSendData(0,0);
    }}

u32 GetOperationType(nn::pia::StationId targetStationId)
{
    SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(targetStationId);
    if (pStationStatus != NULL)
    {
        return pStationStatus->operationType;
    }
    return OPERATION_TYPE_NONE;
}

void UpdateStatusManager()
{
    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(localStationId);
    if (pStationStatus != NULL)
    {
        UpdateSendData(0,0);
    }
    g_StatusManager.Update();
}

bool CheckSessionStatus()
{
    bool retval = true;
#if PIA_TEST_CODE
#else
    u32 stationNum = nn::pia::session::Session::GetInstance()->GetStationNum();
    u32 participantNum = nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum();

    // Whether there are any stations attempting to join or leave the session.
    if (stationNum != participantNum)
    {
        retval = false;
    }
    // Whether the stations that have joined the session are all synchronized (whether the connection process is complete).
    for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
        it != nn::pia::session::Session::GetInstance()->End();
        ++it)
    {
        SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(*it);
        if (pStationStatus->stationId != nn::pia::STATION_ID_INVALID)
        {
            if (pStationStatus->stationNumber != stationNum)
            {
                retval = false;
            }
        }
    }
#endif
    return retval;
}

void SampleSessionEventCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    PIA_SAMPLE_LOG("%s is call EventType: %d, StationId: %lld\n", __func__, eventType, stationId.GetValue());

    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG(" EventType: EVENT_JOIN\n");
        nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
        if (s_SessionState == SESSION_STATE_NONE || IsInSession() || IsInJointSession())
        {
            // The  session-joining event.

            if (stationId == localStationId)
            {
                // The local station's session-joining event.
                PIA_SAMPLE_ASSERT(s_SessionState == SESSION_STATE_NONE);
                if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
                {
                    s_SessionState = SESSION_STATE_SESSION;
                }
                else if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION)
                {
                    s_SessionState = SESSION_STATE_JOINT_SESSION;
                }
                else
                {
                    PIA_SAMPLE_ASSERTMSG(false, "Session status is invalid. [%d]", nn::pia::session::Session::GetInstance()->GetStatus());
                }
                s_IsNeedModifyAttribute = false;
                s_IsLeaveEventForLocalOccurred = false;

                // Resets the team roster.
                g_TeamConstitutionManager.Reset();

                // Assigns the status of the common object.
                AssignCommonObjectsStatus();

                // Starts communication for synchronizing states.
                StartCloneProtocol();
            }
        }
        else if (IsSessionMigrating())
        {
            // In the process of session migration.
        }

        // Assigns a new StationStatus because there was a new join-in event.
        g_StatusManager.AssignStationStatus(stationId, g_ObjectManager.GetReceiveObject(stationId));

        // Updates the team roster.
        if (!g_TeamConstitutionManager.JoinEvent(stationId))
        {
            // A different session has joined for the same team.
            // Handled as an error.
            s_IsErrorOccurred = true;
        }

        g_SessionMigrationEventChecker.SetJoinEventStationId(stationId);
        // Check if stationId can be converted to a valid station index.
        nn::pia::StationIndex stationIndex;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        PIA_SAMPLE_ASSERT(stationIndex != nn::pia::STATION_INDEX_INVALID);
        NN_PIA_DUMMY_PARAM_IN_RELEASE(result);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG(" EventType: EVENT_LEAVE\n");
        PIA_SAMPLE_LOG("host: %lld(0x%016llx), jointSessionHost: %lld(0x%016llx)\n", 
            nn::pia::session::Session::GetInstance()->GetHostStationId().GetValue(),
            nn::pia::session::Session::GetInstance()->GetHostStationId().GetValue(),
            nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId().GetValue(),
            nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId().GetValue());
        // Cleans up StationStatus that has received a leave event.
        SampleObjectStatus* pStatus = g_StatusManager.GetStationStatus(stationId);
        PIA_SAMPLE_ASSERT(pStatus != NULL);  // Has a leave event come from an unconnected peer?
        pStatus->Clear();

        // Updates the team roster.
        g_TeamConstitutionManager.LeaveEvent(stationId);

        nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
        if (stationId == localStationId)
        {
            // The local station's session-leaving event.
            s_SessionState = SESSION_STATE_NONE;
            g_StatusManager.ClearCommonObjectStatus();
            s_IsLeaveEventForLocalOccurred = true;
            s_IsNeedModifyAttribute = false;
        }

        if ((stationId == nn::pia::session::Session::GetInstance()->GetHostStationId() ||
             stationId == nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId()) &&
            !s_IsLeaveProcessing)
        {
            // The host or the joint session host left.
            if (!g_HostMigrationEventChecker.IsCheckEnabled())
            {
                // Start checking for host migration event.
                g_HostMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_HOST_MIGRATION);
            }
            PIA_SAMPLE_ASSERT(g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION);
            if (stationId == nn::pia::session::Session::GetInstance()->GetHostStationId())
            {
                g_HostMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_SESSION_HOST_CHANGED);
            }
            if (stationId == nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId())
            {
                g_HostMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED);
            }
        }

        g_SessionMigrationEventChecker.SetLeaveEventStationId(stationId);
         //Check if stationId can be converted to a valid station index.
        nn::pia::StationIndex stationIndex;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->CheckTransportStatus();
        if(result.IsSuccess())
        {
            result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
            if(result.IsSuccess())
            {
                PIA_SAMPLE_ASSERT(stationIndex != nn::pia::STATION_INDEX_INVALID);
            }
        }
    }
    break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
    {
        // The host has changed.
        PIA_SAMPLE_LOG(" EventType: EVENT_SESSION_HOST_CHANGED\n");

        if (!IsSessionMigrating())
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }
        else
        {
            // Update the team roster after the session migration process is done.
            s_HostChangedEventOccurred = true;
        }

        PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0);

        g_HostMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_SESSION_HOST_CHANGED);
        // Check host data.
        PIA_SAMPLE_ASSERT(stationId != nn::pia::STATION_ID_INVALID);
        PIA_SAMPLE_ASSERT(stationId == nn::pia::session::Session::GetInstance()->GetHostStationId());
        if (stationId == nn::pia::session::Session::GetInstance()->GetLocalStationId())
        {
            PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->IsHost());
        }
    }
    break;
    case nn::pia::session::Session::EVENT_JOINT_SESSION_HOST_CHANGED:
    {
        // The joint session host has changed.
        PIA_SAMPLE_LOG(" EventType: EVENT_JOINT_SESSION_HOST_CHANGED\n");

        // Updates the team roster.
        if (!g_TeamConstitutionManager.SessionHostChangedEvent())
        {
            s_IsErrorOccurred = true;
        }

        if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
        {
            // Change the matchmaking session attribute to the joint session host's team.
            s_IsNeedModifyAttribute = true;
        }

        PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED) != 0);

        g_HostMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED);
        // Check host data.
        PIA_SAMPLE_ASSERT(stationId != nn::pia::STATION_ID_INVALID);
        PIA_SAMPLE_ASSERT(stationId == nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId());
        if (stationId == nn::pia::session::Session::GetInstance()->GetLocalStationId())
        {
            PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->IsJointSessionHost());
        }
    }
    break;
    // The following events are related to session migration.
    case nn::pia::session::Session::EVENT_START_SESSION_JOINT_RANDOM:
    {
        // Joint session start event.
        PIA_SAMPLE_LOG(" EventType: EVENT_START_SESSION_JOINT_RANDOM\n");
        PIA_SAMPLE_ASSERT(IsInSession());
        s_SessionState = SESSION_STATE_MIGRATING_TO_JOINT_SESSION;
        g_JointSessionType = JOINT_SESSION_TYPE_RANDOM;
        s_HostChangedEventOccurred = false;

        // Set anticipated event.
        if (nn::pia::session::Session::GetInstance()->IsHost() == false)
        {
            g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOINT_RANDOM);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOINT_RANDOM);
        }
        // Set the generated event.
        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOINT_RANDOM);
    }
    break;
    case nn::pia::session::Session::EVENT_END_SESSION_JOINT_RANDOM:
    {
        // Joint session end event.
        PIA_SAMPLE_LOG(" EventType: EVENT_END_SESSION_JOINT_RANDOM\n");
        PIA_SAMPLE_ASSERT(IsSessionMigratingToJointSession());
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_JOINT_RANDOM) != 0);
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_END_SESSION_JOINT_RANDOM) != 0);
        s_SessionState = SESSION_STATE_JOINT_SESSION;

        // StationIndex has changed, so re-bind Object to StationStatus.
        BindObjectToStatus();

        if (s_HostChangedEventOccurred)
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }

        // Starts communication for synchronizing states.
        StartCloneProtocol();

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOINT_RANDOM);

        // Check event.
        g_SessionMigrationEventChecker.Trace();
        PIA_SAMPLE_ASSERT(g_SessionMigrationEventChecker.GetEventFlag() == g_SessionMigrationEventChecker.GetExpectedEventFlag());
        g_SessionMigrationEventChecker.CheckStationIdList();
        g_SessionMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
    }
    break;
    case nn::pia::session::Session::EVENT_START_SESSION_CREATE_JOINT:
    {
        // Joint session start event.
        PIA_SAMPLE_LOG(" EventType: EVENT_START_SESSION_CREATE_JOINT\n");
        PIA_SAMPLE_ASSERT(IsInSession());
        s_SessionState = SESSION_STATE_MIGRATING_TO_JOINT_SESSION;
        g_JointSessionType = JOINT_SESSION_TYPE_CREATE;
        s_HostChangedEventOccurred = false;

        // Set anticipated event.
        if (nn::pia::session::Session::GetInstance()->IsHost() == false)
        {
            g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_CREATE_JOINT);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_CREATE_JOINT);
        }
        // Set the generated event.
        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_START_SESSION_CREATE_JOINT);
    }
    break;
    case nn::pia::session::Session::EVENT_END_SESSION_CREATE_JOINT:
    {
        // Joint session end event.
        PIA_SAMPLE_LOG(" EventType: EVENT_END_SESSION_CREATE_JOINT\n");
        PIA_SAMPLE_ASSERT(IsSessionMigratingToJointSession());
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_CREATE_JOINT) != 0);
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_END_SESSION_CREATE_JOINT) != 0);
        s_SessionState = SESSION_STATE_JOINT_SESSION;

        // StationIndex has changed, so re-bind Object to StationStatus.
        BindObjectToStatus();

        if (s_HostChangedEventOccurred)
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }

        // Starts communication for synchronizing states.
        StartCloneProtocol();

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_END_SESSION_CREATE_JOINT);

        // Check event.
        g_SessionMigrationEventChecker.Trace();
        PIA_SAMPLE_ASSERT(g_SessionMigrationEventChecker.GetEventFlag() == g_SessionMigrationEventChecker.GetExpectedEventFlag());
        g_SessionMigrationEventChecker.CheckStationIdList();
        g_SessionMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
    }
    break;
    case nn::pia::session::Session::EVENT_START_SESSION_JOIN_JOINT:
    {
        // Joint session start event.
        PIA_SAMPLE_LOG(" EventType: EVENT_START_SESSION_JOIN_JOINT\n");
        PIA_SAMPLE_ASSERT(IsInSession());
        s_SessionState = SESSION_STATE_MIGRATING_TO_JOINT_SESSION;
        g_JointSessionType = JOINT_SESSION_TYPE_JOIN;
        s_HostChangedEventOccurred = false;

        // Set anticipated event.
        if (nn::pia::session::Session::GetInstance()->IsHost() == false)
        {
            g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOIN_JOINT);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOIN_JOINT);
        }
        // Set the generated event.
        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_START_SESSION_JOIN_JOINT);
    }
    break;
    case nn::pia::session::Session::EVENT_END_SESSION_JOIN_JOINT:
    {
        // Joint session end event.
        PIA_SAMPLE_LOG(" EventType: EVENT_END_SESSION_JOIN_JOINT\n");
        PIA_SAMPLE_ASSERT(IsSessionMigratingToJointSession());
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_JOIN_JOINT) != 0);
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_END_SESSION_JOIN_JOINT) != 0);
        s_SessionState = SESSION_STATE_JOINT_SESSION;

        // StationIndex has changed, so re-bind Object to StationStatus.
        BindObjectToStatus();

        if (s_HostChangedEventOccurred)
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }

        // Starts communication for synchronizing states.
        StartCloneProtocol();

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_END_SESSION_JOIN_JOINT);

        // Check event.
        g_SessionMigrationEventChecker.Trace();
        PIA_SAMPLE_ASSERT(g_SessionMigrationEventChecker.GetEventFlag() == g_SessionMigrationEventChecker.GetExpectedEventFlag());
        g_SessionMigrationEventChecker.CheckStationIdList();
        g_SessionMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
    }
    break;
    case nn::pia::session::Session::EVENT_START_SESSION_DESTROY_JOINT:
    {
        // Joint session dissolution start event.
        PIA_SAMPLE_LOG(" EventType: EVENT_START_SESSION_DESTROY_JOINT\n");
        PIA_SAMPLE_ASSERT(IsInJointSession());
        s_SessionState = SESSION_STATE_MIGRATING_TO_SESSION;
        g_JointSessionType = JOINT_SESSION_TYPE_DESTROY;
        s_HostChangedEventOccurred = false;
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost() && nn::pia::session::Session::GetInstance()->IsHost())
        {
            // Session hosts that are not the joint session host are removed from the team structure of the other session upon disbanding the joint session, so set a flag.
            // 
            s_HostChangedEventOccurred = true;
        }

        // Set anticipated event.
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost())
        {
            g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_DESTROY_JOINT);

            // At this point, the stations for another team should have left (although there will be cases where team information is not yet received)
            nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
            nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
            int localTeamIndex = g_TeamConstitutionManager.GetTeamIndex(pSession->GetLocalStationId());
            NN_PIA_DUMMY_PARAM_IN_RELEASE(localTeamIndex);
            for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
            {
                int teamIndex = g_TeamConstitutionManager.GetTeamIndex(*it);
                NN_PIA_DUMMY_PARAM_IN_RELEASE(teamIndex);
                PIA_SAMPLE_LOG(" StationId:%lld is in Team%d. Myteam is %d.\n", it->GetValue(), teamIndex, localTeamIndex);
                PIA_SAMPLE_ASSERT(localTeamIndex == -1 || teamIndex == -1 || teamIndex == localTeamIndex);
            }

            // End the host migration checking (that for EVENT_JOINT_SESSION_HOST_CHANGED) that might have occurred due to leaving the disbanded joint session.
            if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
            {
                g_HostMigrationEventChecker.UnsetExpectedEventFlag(EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED);
                if (g_HostMigrationEventChecker.GetExpectedEventFlag() == 0)
                {
                    g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
                }
            }
        }

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT);
    }
    break;
    case nn::pia::session::Session::EVENT_END_SESSION_DESTROY_JOINT:
    {
        // Joint session dissolution end event.
        PIA_SAMPLE_LOG(" EventType: EVENT_END_SESSION_DESTROY_JOINT\n");
        PIA_SAMPLE_ASSERT(IsSessionMigratingToSession());
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_DESTROY_JOINT) != 0);
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_END_SESSION_DESTROY_JOINT) != 0);
        s_SessionState = SESSION_STATE_SESSION;
        s_IsNeedModifyAttribute = false;

        // StationIndex has changed, so re-bind Object to StationStatus.
        BindObjectToStatus();

        if (s_HostChangedEventOccurred)
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }

        // Starts communication for synchronizing states.
        StartCloneProtocol();

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_END_SESSION_DESTROY_JOINT);

        // Check event.
        g_SessionMigrationEventChecker.Trace();
        PIA_SAMPLE_ASSERT(g_SessionMigrationEventChecker.GetEventFlag() == g_SessionMigrationEventChecker.GetExpectedEventFlag());
        g_SessionMigrationEventChecker.CheckStationIdList();
        g_SessionMigrationEventChecker.CheckExpectedLeaveStationIdList();
        g_SessionMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
    }
    break;
    case nn::pia::session::Session::EVENT_START_SESSION_LEAVE_JOINT:
    {
        // Joint session dissolution start event.
        PIA_SAMPLE_LOG(" EventType: EVENT_START_SESSION_LEAVE_JOINT\n");
        PIA_SAMPLE_ASSERT(IsInJointSession());
        s_SessionState = SESSION_STATE_MIGRATING_TO_SESSION;
        g_JointSessionType = JOINT_SESSION_TYPE_LEAVE;
        s_HostChangedEventOccurred = false;
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost() && nn::pia::session::Session::GetInstance()->IsHost())
        {
            // Session hosts that are not the joint session host are removed from the team structure of the other session upon disbanding the joint session, so set a flag.
            // 
            s_HostChangedEventOccurred = true;
        }

        // Set anticipated event.
        if (!nn::pia::session::Session::GetInstance()->IsHost())
        {
            g_SessionMigrationEventChecker.CheckBegin(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_START_SESSION_LEAVE_JOINT);
            g_SessionMigrationEventChecker.SetExpectedEventFlag(EventChecker::EVENT_BIT_END_SESSION_LEAVE_JOINT);

            // At this point, the stations for another team should have left (although there will be cases where team information is not yet received).
            nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
            nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
            int localTeamIndex = g_TeamConstitutionManager.GetTeamIndex(pSession->GetLocalStationId());
            NN_PIA_DUMMY_PARAM_IN_RELEASE(localTeamIndex);
            for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
            {
                int teamIndex = g_TeamConstitutionManager.GetTeamIndex(*it);
                NN_PIA_DUMMY_PARAM_IN_RELEASE(teamIndex);
                PIA_SAMPLE_LOG(" StationId: %lld is in Team %d. Myteam is %d.\n", it->GetValue(), teamIndex, localTeamIndex);
                PIA_SAMPLE_ASSERT(localTeamIndex == -1 || teamIndex == -1 || teamIndex == localTeamIndex);
            }
        }

        // End the check for EVENT_JOINT_SESSION_HOST_CHANGED, which is a check for host migration that probably was triggered by a station leaving the joint session.
        if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
        {
            g_HostMigrationEventChecker.UnsetExpectedEventFlag(EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED);
            if (g_HostMigrationEventChecker.GetExpectedEventFlag() == 0)
            {
                g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
            }
        }

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_START_SESSION_LEAVE_JOINT);
    }
    break;
    case nn::pia::session::Session::EVENT_END_SESSION_LEAVE_JOINT:
    {
        // Joint session dissolution end event.
        PIA_SAMPLE_LOG(" EventType: EVENT_END_SESSION_LEAVE_JOINT\n");
        PIA_SAMPLE_ASSERT(IsSessionMigratingToSession());
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_START_SESSION_LEAVE_JOINT) != 0);
        PIA_SAMPLE_ASSERT((g_SessionMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_END_SESSION_LEAVE_JOINT) != 0);
        s_SessionState = SESSION_STATE_SESSION;
        s_IsNeedModifyAttribute = false;

        // StationIndex has changed, so re-bind Object to StationStatus.
        BindObjectToStatus();

        if (s_HostChangedEventOccurred)
        {
            // Updates the team roster.
            if (!g_TeamConstitutionManager.SessionHostChangedEvent())
            {
                s_IsErrorOccurred = true;
            }
        }

        // Starts communication for synchronizing states.
        StartCloneProtocol();

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_END_SESSION_LEAVE_JOINT);

        // Check event.
        g_SessionMigrationEventChecker.Trace();
        PIA_SAMPLE_ASSERT(g_SessionMigrationEventChecker.GetEventFlag() == g_SessionMigrationEventChecker.GetExpectedEventFlag());
        g_SessionMigrationEventChecker.CheckStationIdList();
        g_SessionMigrationEventChecker.CheckExpectedLeaveStationIdList();
        g_SessionMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_SESSION_MIGRATION);
    }
    break;
    case nn::pia::session::Session::EVENT_SESSION_MIGRATION_FAILED:
    {
        // Session migration failed event.
        PIA_SAMPLE_LOG(" EventType: EVENT_SESSION_MIGRATION_failed.\n");
        PIA_SAMPLE_ASSERT(s_SessionState == SESSION_STATE_NONE && s_IsLeaveEventForLocalOccurred);
        NN_PIA_DUMMY_PARAM_IN_RELEASE(s_IsLeaveEventForLocalOccurred);

        g_SessionMigrationEventChecker.SetEventFlag(EventChecker::EVENT_BIT_SESSION_MIGRATION_FAILED);
    }
    break;
    }
}

bool SampleJoinApprovalCallback(const nn::pia::transport::Station::IdentificationToken *pToken)
{
    NN_PIA_DUMMY_PARAM(pToken);
    return true;

#if 0
    // Callback that determines whether to approve or reject the join-in request. (Called only by the session host.)
    // In this sample, this is used to prevent the same team from joining 2 or more different sessions while participating in a joint session.
    // 

    if (pToken == NULL)
    {
        PIA_SAMPLE_ASSERT(false);
        return false;
    }

    SampleTokenData tokenData;
    memcpy(&tokenData, pToken->data, sizeof(tokenData));
    
    u32 curTeamSessionId = g_TeamConstitutionManager.GetTeamSessionId(tokenData.teamType);
    if (curTeamSessionId == 0 || curTeamSessionId == tokenData.sessionId)
    {
        return true;
    }
    else
    {
        return false;
    }
#endif
}

bool IsInSession()
{
    return s_SessionState == SESSION_STATE_SESSION;
}

bool IsInJointSession()
{
    return s_SessionState == SESSION_STATE_JOINT_SESSION;
}

bool IsSessionMigratingToSession()
{
    return s_SessionState == SESSION_STATE_MIGRATING_TO_SESSION;
}

bool IsSessionMigratingToJointSession()
{
    return s_SessionState == SESSION_STATE_MIGRATING_TO_JOINT_SESSION;
}

bool IsSessionMigrating()
{
    return IsSessionMigratingToSession() || IsSessionMigratingToJointSession();
}

void InitializePiaTransport()
{
    nn::Result result;

    // * The Factory instance is not required except for initialization, so it is allocated with an automatic variable.
    nn::pia::inet::NexNetworkFactory nexFactory;

    // These parameters are borrowed from the transport demo.
    // For more information about the parameters, see the Programming Manual.
    nn::pia::transport::Transport::Setting transportSetting;

    transportSetting.pFactory = &nexFactory;
    transportSetting.maxStationNum = SAMPLE_GAME_PLAYER_MAX;
    transportSetting.maxSendThreadBufferNum = SAMPLE_MAX_SEND_THREAD_BUFFER_NUM;
    transportSetting.maxReceiveThreadBufferNum = SAMPLE_MAX_RECEIVE_THREAD_BUFFER_NUM;

#if PIA_TEST_CODE
    nn::pia::transport::Transport::DebugSetting debugSetting;
    debugSetting.sendThreadLatencyEmulationPacketNum = 128;
    debugSetting.receiveThreadLatencyEmulationPacketNum = 128;
    debugSetting.bPacketLossEmulation = true;
    result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
#endif

    result = Transport_InitializePiaTransport(transportSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //Priority settings for the send/receive thread. (For information about priorities, see SampleConfig.h.)
    //The priority of the send/receive thread does not necessarily need to be changed, but in this sample, the priority has been changed to show the change method.
    //Priority has been changed.
    nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(pThreadStreamManager);  //If the transport initialization has finished, the instance has probably been created.
    pThreadStreamManager->SetSendThreadPriority(SEND_THREAD_PRIORITY);
    pThreadStreamManager->SetReceiveThreadPriority(RECEIVE_THREAD_PRIORITY);

#if PIA_TEST_CODE
    pThreadStreamManager->SetSendThreadLatencyEmulation(100, 100);
    pThreadStreamManager->SetReceiveThreadLatencyEmulation(0, 0);
    PIA_SAMPLE_LOG("Enable latency emulation.\n");

    pThreadStreamManager->SetSendThreadPacketLossRatio(10);
    pThreadStreamManager->SetReceiveThreadPacketLossRatio(10);
    PIA_SAMPLE_LOG("Enable pakcet loss emulation.\n");
#endif
}

void FinalizePiaTransport()
{
    Transport_FinalizePiaTransport(0, 0, NULL);
}

void InitializePiaSession()
{
    nn::Result result;

    // * The Factory instance is not required except for initialization, so it is allocated with an automatic variable.
    nn::pia::inet::NexNetworkFactory nexFactory;

    nn::pia::session::Session::Setting sessionSetting;
    sessionSetting.pNetworkFactory = &nexFactory;
    sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
    sessionSetting.browsedSessionInfoListNum = 4;
#if PIA_TEST_CODE
    result = Session_InitializePiaSession(sessionSetting, false, false, true);
#else
    result = Session_InitializePiaSession(sessionSetting);
#endif
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Initializes the status before the event callback is registered.
    g_StatusManager.ClearStationStatus();
    g_StatusManager.ClearCommonObjectStatus();

    // Registers the event callback for Session.
    nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionEventCallback);
}

void FinalizePiaSession()
{
    if (nn::pia::session::Session::GetInstance() != NULL)
    {
        // Unregisters the event callback for Session.
        nn::pia::session::Session::GetInstance()->UnregisterSessionEventCallback();
    }
    Session_FinalizePiaSession();
}

void InitializePiaClone()
{
    Clone_InitializePiaClone();
}

void FinalizePiaClone()
{
    Clone_FinalizePiaClone();
}

/*!
    @brief  Handles the state in which communication is performed using a <tt>PiaTransport</tt> instance.
    
    When a host shuts down a network, the host enters the 'select network' state, but the clients enter the error state.
    
    @return  Returns the number of the next state (scene) to execute.
 */

SampleSceneNo SceneCommunicationTeam()
{
    nn::Result result = nn::ResultSuccess();

    u32 frameNo = 0;
    bool isExit = false;
    bool goToGame = false;

#if PIA_TEST_CODE
    bool isDelayDispatch = false;
    bool isStopDispatch = false;
    u8 delayCount = 0;
#endif

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_COMMUNICATE_TEAM,
        true,
        nn::pia::session::Session::GetInstance()->IsHost(),
        nn::pia::session::Session::GetInstance()->GetStationNum(),
        0
        );

    nn::pia::transport::StationConnectionInfo localInfo;
    nn::pia::transport::StationConnectionInfo hostInfo;

    if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
    {
        result = nn::pia::session::Session::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        result = nn::pia::session::Session::GetInstance()->GetHostStationConnectionInfo(&hostInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
    }

    // Gets the team color based on the local station's team type.
    nn::pia::sampleframework::Color myTeamColor = GetColorOfTeamType(GetCurrentTeamType());

    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();

    SampleWaitingAsyncProcessing waitingProcess;
    bool asyncProcessingFlag = false;
    nn::pia::util::HandlingType handlingType = nn::pia::util::HANDLING_VOID;
    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        if (!asyncProcessingFlag && s_IsErrorOccurred)
        {
            SetCommunicationErrorType(APPLICATION_ERROR);
            handlingType = nn::pia::util::HANDLING_LEAVE_AND_CLEANUP;
            isExit = true;
        }

        if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
        {
            // Checking for host migration event.

            // Should not be waiting for joint session host migration.
            PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED) == 0);
            nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
            if (nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(hostStationId))
            {
                // Host station ID is valid.
                if (hostStationId == nn::pia::session::Session::GetInstance()->GetLocalStationId())
                {
                    // Local station is new host.
                    if (nn::pia::session::Session::GetInstance()->IsHost())
                    {
                        // The EVENT_SESSION_HOST_CHANGED event should already have occurred.
                        PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0);
                        g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
                    }
                }
                else
                {
                    // Local station is new client.
                    // The EVENT_SESSION_HOST_CHANGED event should already have occurred.
                    PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0);
                    g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
                }
            }
        }

#if PIA_TEST_CODE
#else
        bool isOperationReady = false;
        OperationType localOperationType = static_cast<OperationType>(GetOperationType(nn::pia::session::Session::GetInstance()->GetLocalStationId()));
        if (localOperationType == OPERATION_TYPE_END_OPERATION)
        {
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                // After OpenSession/CloseSession has ended, the session host checks that clients have completed.
                bool isAllCompleted = true;
                for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
                    it != nn::pia::session::Session::GetInstance()->End(); ++ it)
                {
                    OperationType type = static_cast<OperationType>(GetOperationType(*it));
                    if (type != OPERATION_TYPE_END_OPERATION)
                    {
                        isAllCompleted = false;
                    }
                }
                if (isAllCompleted)
                {
                    PIA_SAMPLE_LOG("Operation all completed.\n");
                    SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                }
            }
            else if (GetOperationType(nn::pia::session::Session::GetInstance()->GetHostStationId()) == OPERATION_TYPE_NONE)
            {
                // After the session host has synchronized the completion state, the client states are also transitioned.
                PIA_SAMPLE_LOG("Operation all completed.\n");
                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
            }
        }
        else if (localOperationType == OPERATION_TYPE_NONE)
        {
            bool isAllReady = true;
            for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
                it != nn::pia::session::Session::GetInstance()->End(); ++ it)
            {
                OperationType type = static_cast<OperationType>(GetOperationType(*it));
                if (type != OPERATION_TYPE_NONE)
                {
                    isAllReady = false;
                }
            }
            if (isAllReady)
            {
                isOperationReady = true;
            }
        }
#endif

        if (!asyncProcessingFlag && !s_IsErrorOccurred)
        {
#if PIA_TEST_CODE
#else
            if (!nn::pia::session::Session::GetInstance()->IsHost())
            {
                if (localOperationType == OPERATION_TYPE_NONE)
                {
                    OperationType hostOperationType = static_cast<OperationType>(GetOperationType(nn::pia::session::Session::GetInstance()->GetHostStationId()));
                    switch (hostOperationType)
                    {
                        case OPERATION_TYPE_OPEN_SESSION:
                        {
                            PIA_SAMPLE_LOG("Call OpenSessionAsync because Host is processsing OpenSessionAsync.\n");
                            result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                            if (result.IsSuccess())
                            {
                                asyncProcessingFlag = true;
                                waitingProcess = SAMPLE_OPEN_SESSION;
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_SESSION));
                            }
                            else if (result == nn::pia::ResultSessionInconsistentState())
                            {
                                PIA_SAMPLE_LOG("OpenSession failed. ResultSessionInconsistentState\n");
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Call OpenSessionAsync failed. Result:0x%08X\n", result.GetValue());
                                PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                        }
                        break;
                        case OPERATION_TYPE_CLOSE_SESSION:
                        {
                            PIA_SAMPLE_LOG("Call CloseSessionAsync because Host is processsing CloseSessionAsync.\n");
                            result = nn::pia::session::Session::GetInstance()->CloseSessionAsync();
                            if (result.IsSuccess())
                            {
                                asyncProcessingFlag = true;
                                waitingProcess = SAMPLE_CLOSE_SESSION;
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_SESSION));
                            }
                            else if (result == nn::pia::ResultSessionInconsistentState())
                            {
                                PIA_SAMPLE_LOG("CloseSession failed. ResultSessionInconsistentState\n");
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Call CloseSessionAsync failed. Result:0x%08X\n", result.GetValue());
                                PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                        }
                        break;
                    }
                }
            }
#endif
            if (!asyncProcessingFlag)
            {
                // X: JointSession (Host)
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
                {
                    if (nn::pia::session::Session::GetInstance()->IsHost() && CheckSessionStatus())
                    {
                        goToGame = true;
                        isExit = true;
#if PIA_TEST_CODE
                        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_JOIN;
                        }
                        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_JOIN_PRIVATE;
                            g_UserPasswordIndexForJoint = TEAM_TYPE_NUM;
                            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_START))
                            {
                                g_UserPasswordIndexForJoint = static_cast<u8>(GetCurrentTeamType());
                            }
                        }
                        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_CREATE;
                        }
                        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_CREATE_PRIVATE;
                            g_UserPasswordIndexForJoint = TEAM_TYPE_NUM;
                        }
                        else
#endif
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_RANDOM;
                        }
                        // Reports start of <tt>Joint</tt>.
                        nn::pia::StationId mylocalStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
                        SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(mylocalStationId);
                        PIA_SAMPLE_ASSERT(pStationStatus != NULL);

                        pStationStatus->isStartJointSession = true;
                        pStationStatus->isUpdataStartJointSession = true;
                        UpdateSendData(0, 0);
                    }
                }
                // B: LeaveSession
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
                {
                    PIA_SAMPLE_LOG("Call LeaveSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->LeaveSessionAsync();
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("Call LeaveSession failed.\n");
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        asyncProcessingFlag = true;
                        waitingProcess = SAMPLE_LEAVE_SESSION;
                        s_IsLeaveProcessing = true;

                        // An EVENT_SESSION_HOST_CHANGED event will not be notified upon starting to leave a session during host migration. Host migration checking ends.
                        if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
                        {
                            g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
                        }
                    }
                }
                // No handling yet for the necessary results for cleanup, logging out, etc.
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
                {
#if PIA_TEST_CODE
                    // Disguise the open/close states so <tt>OpenSessionAsync</tt> is called when the state is <tt>Open</tt>, and <tt>CloseSessionAsync</tt> is called when the state is <tt>Close</tt>.
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        g_IsOpened = true;
                    }

                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                    {
                        g_IsOpened = false;
                    }

                    if (g_IsOpened)
                    {
                        result = nn::pia::session::Session::GetInstance()->CloseSessionAsync();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("Call CloseSessionAsync\n");
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_CLOSE_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call CloseSessionAsync failed. Result:0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                    else
                    {
                        result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("Call OpenSessionAsync\n");
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_OPEN_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("OpenSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call OpenSessionAsync failed. Result:0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
#else
                    if (isOperationReady && nn::pia::session::Session::GetInstance()->IsHost())
                    {
                        if (g_IsOpened)
                        {
                            result = nn::pia::session::Session::GetInstance()->CloseSessionAsync();
                            if (result.IsSuccess())
                            {
                                PIA_SAMPLE_LOG("Call CloseSessionAsync\n");
                                asyncProcessingFlag = true;
                                waitingProcess = SAMPLE_CLOSE_SESSION;
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_SESSION));
                            }
                            else if (result == nn::pia::ResultSessionInconsistentState())
                            {
                                PIA_SAMPLE_LOG("CloseSession failed. ResultSessionInconsistentState\n");
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Call CloseSessionAsync failed. Result:0x%08X\n", result.GetValue());
                                PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                        }
                        else
                        {
                            result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                            if (result.IsSuccess())
                            {
                                PIA_SAMPLE_LOG("Call OpenSessionAsync\n");
                                asyncProcessingFlag = true;
                                waitingProcess = SAMPLE_OPEN_SESSION;
                                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_SESSION));
                            }
                            else if (result == nn::pia::ResultSessionInconsistentState())
                            {
                                PIA_SAMPLE_LOG("OpenSession failed. ResultSessionInconsistentState\n");
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Call OpenSessionAsync failed. Result:0x%08X\n", result.GetValue());
                                PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                                s_IsErrorOccurred = true;
                                isExit = true;
                            }
                        }
                    }
#endif
                }
            }
        }

#if PIA_TEST_CODE
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
            PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());
            if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
            }
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
            }
            else
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
            }
        }
        if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT) && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
        {
            isDelayDispatch = !isDelayDispatch;
            delayCount = 0;
            isStopDispatch = false;
        }
        if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP) && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            PIA_SAMPLE_LOG("Dispatch stoped.");
            isStopDispatch = !isStopDispatch;
            isDelayDispatch = false;
            delayCount = 0;
        }
#endif

        if (asyncProcessingFlag)
        {
#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_CLOSE_SESSION_SEQUENCE, nn::pia::session::Session::GetInstance()->GetCloseSessionSequence());
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_OPEN_SESSION_SEQUENCE, nn::pia::session::Session::GetInstance()->GetOpenSessionSequence());
#endif

            // Checks the progress of the asynchronous process.
            switch (waitingProcess)
            {
                case SAMPLE_LEAVE_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedLeaveSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetLeaveSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("LeaveSession succeeded.\n");
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("LeaveSession failed. Result: 0x%08X\n", result.GetValue());
                        }
                        asyncProcessingFlag = false;
                        s_IsLeaveProcessing = false;
                    }
                }
                break;
                case SAMPLE_OPEN_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedOpenSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetOpenSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("OpenSession succeeded.\n");
                            g_IsOpened = true;
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("OpenSession failed. ResultSessionInconsistentState\n");
                        }
                        else if (result == nn::pia::ResultSessionConnectionIsLost())
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. ResultSessionConnectionIsLost\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("OpenSession failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
                    }
                }
                break;
                case SAMPLE_CLOSE_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedCloseSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetCloseSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("CloseSession succeeded.\n");
                            g_IsOpened = false;
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. ResultSessionInconsistentState\n");
                        }
                        else if (result == nn::pia::ResultSessionConnectionIsLost())
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. ResultSessionConnectionIsLost\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
                    }
                }
                break;
            }
            if (result.IsFailure())
            {
                s_IsErrorOccurred = true;
                isExit = true;
            }
        }

        if (!s_IsErrorOccurred)
        {
            // R: ChangeOperationTarget
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                g_StatusManager.ChangeOperationTarget(GetCurrentTeamType());
            }
            // A: AtomicLockOperation
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                g_StatusManager.AtomicLockOperation();
            }
        
            s32 diffX = 0, diffY = 0;
            // Checks for input to move rectangle.
            {
                f32 analogX, analogY;
                nn::pia::sampleframework::SampleFramework::GetInstance()->GetAnalog(&analogX, &analogY);
                if (analogX != 0.0f || analogY != 0.0f)
                {
                    const f32 cAnalogMoveBase = 3.0f;
                    diffX = static_cast<s32>(cAnalogMoveBase * analogX);
                    diffY = static_cast<s32>(cAnalogMoveBase * analogY * (-1.0f));  // Up/down on analog stick is reversed from up/down when drawing. (Negative and positive are reversed.)
                }
                else
                {
                    const s32 cMoveBase = 2;
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
                    {
                        diffX -= cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
                    {
                        diffX += cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        diffY -= cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                    {
                        diffY += cMoveBase;
                    }

                    // Always move, even if there is no input for that, so communication is easier to grasp.
                    if (diffX == 0 && diffY == 0)
                    {
                        diffX = 1;
                    }
                }
            }

            // The sending process before the Dispatch process.
            if (IsInSession()) // Do not send while the session is migrating.
            {
                // Checks the sending interval.
                if ((frameNo % (SAMPLE_INTERVAL_SENDING_STATION_STATUS + 1)) == 0)
                {
                    // Updates the data to send.
                    UpdateSendData(diffX, diffY);
                }
            }

            // Updates the clock for CloneProtocol.
            Clone_GetCloneProtocol()->UpdateClock();

#if PIA_TEST_CODE
            // Reduce the frequency of dispatch for testing, so that it is easier to pinpoint a certain timing.
            if(isDelayDispatch)
            {
                ++delayCount;
                delayCount = delayCount % DELAY_DISPATCH_NUM;
            }

            if( (!isDelayDispatch || (isDelayDispatch && delayCount == DELAY_DISPATCH_NUM-1)) && !isStopDispatch)
            {
#endif
            // Execute Dispatch.
            nn::pia::common::Scheduler::GetInstance()->Dispatch();
            nn::nex::Scheduler::GetInstance()->DispatchAll();
#if PIA_TEST_CODE
            }
#endif

            // Checks whether the team roster has changed.
            g_TeamConstitutionManager.Update();

            // The receiving process after the Dispatch process.
            // Updates the status based on the received data.
            g_StatusManager.Update();

            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE_TEAM,
                true,
                nn::pia::session::Session::GetInstance()->IsHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum(),
                0
                );
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SAMPLE_PROGRAMPARAM_OPENCLOSESESSION_COMPLETE, g_IsOpened ? 1 : 0);

            // The Session status is not checked while waiting for an asynchronous process.
            if (!asyncProcessingFlag)
            {
                // Check the 'session migrating ' flag.
                if (IsSessionMigrating())
                {
                    // Generated when merging into a game session.
                    PIA_SAMPLE_LOG("ChangingSessionFlag became true.\n");
                    goToGame = true;
                    isExit = true;
                }
                // The check for a disconnection is done by CheckConnection.

                // Session status check.
                result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
                if (result.IsFailure())
                {
                    if (result == nn::pia::ResultNetworkConnectionIsLost())
                    {
                        PIA_SAMPLE_LOG("CheckConnection: Session::CheckConnection: Disconnected NETWORK\n");
                    }
                    else if (result == nn::pia::ResultSessionConnectionIsLost())
                    {
                        PIA_SAMPLE_LOG("CheckConnection: Session::CheckConnection: Disconnected SESSION\n");
                    }
                    else
                    {
                        // This value is unlikely.
                        PIA_SAMPLE_ASSERT(false);
                        PIA_SAMPLE_LOG("Session::CheckConnection: Unexpected result: 0x%08x\n", result.GetValue());
                    }
                    s_IsErrorOccurred = true;
                    isExit = true;
                }
            }
        }

        //Screen rendering.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            //String.
            {
                //Check the role of the local station.
                int roleIndex = 0;  //Unknown;
                localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
                if (localStationId != nn::pia::STATION_ID_INVALID)
                {
                    if (nn::pia::session::Session::GetInstance()->IsHost())
                    {
                        roleIndex = 1;  // 1:Host
                    }
                    else
                    {
                        roleIndex = 2;  // 2:Client
                    }
                }

                int line = 0;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Communication (Team)");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum: %2d, Role: %-6s", 
                    nn::pia::session::Session::GetInstance()->GetStationNum(), 
                    SESSION_ROLE(roleIndex));
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationId: %lld, TeamMemberIndex: %2d", 
                    localStationId.GetValue(), 
                    g_TeamConstitutionManager.GetMemberIndex(localStationId));
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  SessionId: %d", nn::pia::session::Session::GetInstance()->GetSessionId());
                line += FONT_SIZE;

                line = SCREEN_HEIGHT - (FONT_SIZE * 11);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: JointSession (Host)");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: LeaveSession");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: ChangeOperationTarget");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: TryLock (Atomic)");
                line += FONT_SIZE;
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  L: %s%s", 
                        g_IsOpened ? "CloseSessionAsync" : "OpenSessionAsync", 
                        (asyncProcessingFlag && (waitingProcess == SAMPLE_CLOSE_SESSION || waitingProcess == SAMPLE_OPEN_SESSION)) ? " (wait other)" : "");
                }
                line += FONT_SIZE;

#if PIA_TEST_CODE
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[OpenSession  Seq] %d", 
                    nn::pia::session::Session::GetInstance()->GetOpenSessionSequence());
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[CloseSession Seq] %d",
                    nn::pia::session::Session::GetInstance()->GetCloseSessionSequence());
                line += FONT_SIZE;
#else
                line += FONT_SIZE * 2;
#endif
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
                line += FONT_SIZE;
            }

            // Draws the frame for distinguishing the team.
            DrawTeamFrame(GetCurrentTeamType(), myTeamColor);

            // Draws StationStatus.
            {
                for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
                {
                    // Draws all StationStatus objects with a valid registered StationId.
                    if (g_StatusManager.GetStationStatus(i)->stationId != nn::pia::STATION_ID_INVALID)
                    {
                        DrawStationStatus(*g_StatusManager.GetStationStatus(i));
                    }
                }
            }

            // Draws common objects.
            for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
            {
                // Draws only local station's team objects.
                if (g_StatusManager.GetCommonObjectStatus(i)->type == OBJECT_STATUS_TYPE_TEAM_ATOMIC ||
                    g_StatusManager.GetCommonObjectStatus(i)->type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL)
                {
                    if (g_StatusManager.GetCommonObjectStatus(i)->teamIndex == GetCurrentTeamType())
                    {
                        DrawStationStatus(*g_StatusManager.GetCommonObjectStatus(i));
                    }
                }
            }
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    if (s_IsErrorOccurred)
    {
        PIA_SAMPLE_ASSERT(asyncProcessingFlag == false);

        if (handlingType == nn::pia::util::HANDLING_VOID)
        {
            handlingType = nn::pia::util::GetHandlingType(result);
            SetNetworkErrorResult(result);
            
            if (result == nn::pia::ResultSessionConnectionIsLost())
            {
                SetCommunicationErrorType(SESSION_LOST_ERROR);
            }
            else if (result == nn::pia::ResultNetworkConnectionIsLost())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            }
            else
            {
                SetCommunicationErrorType(MATCHMAKE_ERROR);
            }
        }

        if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP)
        {
            // An EVENT_SESSION_HOST_CHANGED event will not be notified upon starting to leave a session during host migration. Host migration checking ends.
            if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
            {
                PIA_SAMPLE_LOG("cancel HostMigrationEventChecker (Team)");
                g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
            }
            SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
            if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                // Leaves the network in a state where it does not need to be reset, because the process can return to matchmaking when you clean up and restart the Session object.
                SetResetOfNetworkIsNecessary(false);
            }
        }

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    if (goToGame)
    {
        return SCENE_CHANGING_SESSION_PREPROCESSING;
    }

    // Return to team selection.
    Session_CleanupSession();
    return SCENE_SELECT_TEAM;
}


SampleSceneNo SceneCommunicationGame()
{
    nn::Result result = nn::ResultSuccess();

    u32 frameNo = 0;
    bool isExit = false;
    bool backToTeam = false;

#if PIA_TEST_CODE
    bool isDelayDispatch = false;
    bool isStopDispatch = false;
    u8 delayCount = 0;
#endif

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_COMMUNICATE_GAME,
        true,
        nn::pia::session::Session::GetInstance()->IsJointSessionHost(),
        nn::pia::session::Session::GetInstance()->GetStationNum(),
        0
        );

    nn::pia::transport::StationConnectionInfo localInfo;
    nn::pia::transport::StationConnectionInfo jointSessionHostInfo;

    if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_JOINT_SESSION)
    {
        result = nn::pia::session::Session::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        result = nn::pia::session::Session::GetInstance()->GetJointSessionHostStationConnectionInfo(&jointSessionHostInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
    }

    SampleWaitingAsyncProcessing waitingProcess;
    bool asyncProcessingFlag = false;

    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    nn::pia::util::HandlingType handlingType = nn::pia::util::HANDLING_VOID;

    // 
    nn::pia::StationId mylocalStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    SampleObjectStatus* pStationStatus = g_StatusManager.GetStationStatus(mylocalStationId);
    PIA_SAMPLE_ASSERT(pStationStatus != NULL);

    pStationStatus->isStartJointSession = false;
    pStationStatus->isUpdataStartJointSession = true;
    g_StatusManager.Update();

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
        {
            // Checking for host migration event.
            bool isHostCheckEnd = false;
            bool isJointSessionHostCheckEnd = false;

            if ((g_HostMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0)
            {
                // Check the host migration of the host.
                nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
                if (nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(hostStationId))
                {
                    // Host station ID is valid.
                    if (hostStationId == nn::pia::session::Session::GetInstance()->GetLocalStationId())
                    {
                        // Local station is new host.
                        if (nn::pia::session::Session::GetInstance()->IsHost())
                        {
                            // The EVENT_SESSION_HOST_CHANGED event should already have occurred.
                            PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0);
                            isHostCheckEnd = true;
                        }
                    }
                    else
                    {
                        // Local station is new client.
                        // The EVENT_SESSION_HOST_CHANGED event should already have occurred.
                        PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_SESSION_HOST_CHANGED) != 0);
                        isHostCheckEnd = true;
                    }
                }
            }
            else
            {
                isHostCheckEnd = true;
            }

            if ((g_HostMigrationEventChecker.GetExpectedEventFlag() & EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED) != 0)
            {
                // Check the host migration of the joint session host.
                nn::pia::StationId jointSessionHostStationId = nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId();
                if (nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(jointSessionHostStationId))
                {
                    // Joint session host station ID is valid.
                    if (jointSessionHostStationId == nn::pia::session::Session::GetInstance()->GetLocalStationId())
                    {
                        // Local station is new joint session host.
                        if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
                        {
                            // The EVENT_JOINT_SESSION_HOST_CHANGED event should already have occurred.
                            PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED) != 0);
                            isJointSessionHostCheckEnd = true;
                        }
                    }
                    else
                    {
                        // Local station is not the new joint session host.
                        // The EVENT_JOINT_SESSION_HOST_CHANGED event should already have occurred.
                        PIA_SAMPLE_ASSERT((g_HostMigrationEventChecker.GetEventFlag() & EventChecker::EVENT_BIT_JOINT_SESSION_HOST_CHANGED) != 0);
                        isJointSessionHostCheckEnd = true;
                    }
                }
            }
            else
            {
                isJointSessionHostCheckEnd = true;
            }

            if (isHostCheckEnd && isJointSessionHostCheckEnd)
            {
                g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
            }
        }

#if PIA_TEST_CODE
#else
        bool isOperationReady = false;
        OperationType localOperationType = static_cast<OperationType>(GetOperationType(nn::pia::session::Session::GetInstance()->GetLocalStationId()));
        if (localOperationType == OPERATION_TYPE_END_OPERATION)
        {
            if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
            {
                // After OpenJointSession/CloseJointSession has ended, the joint session host checks that clients have completed.
                bool isAllCompleted = true;
                for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
                    it != nn::pia::session::Session::GetInstance()->End(); ++ it)
                {
                    OperationType type = static_cast<OperationType>(GetOperationType(*it));
                    if (type != OPERATION_TYPE_END_OPERATION)
                    {
                        isAllCompleted = false;
                    }
                }
                if (isAllCompleted)
                {
                    PIA_SAMPLE_LOG("Operation completed.\n");
                    SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
                }
            }
            else if (GetOperationType(nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId()) == OPERATION_TYPE_NONE)
            {
                // After the joint session host has synchronized the completion state, the client states are also transitioned.
                PIA_SAMPLE_LOG("Operation completed.\n");
                SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_NONE));
            }
        }
        else if (localOperationType == OPERATION_TYPE_NONE)
        {
            bool isAllReady = true;
            for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin();
                it != nn::pia::session::Session::GetInstance()->End(); ++ it)
            {
                OperationType type = static_cast<OperationType>(GetOperationType(*it));
                if (type != OPERATION_TYPE_NONE)
                {
                    isAllReady = false;
                }
            }
            if (isAllReady)
            {
                isOperationReady = true;
            }
        }
#endif

        if (!asyncProcessingFlag && s_IsErrorOccurred)
        {
            SetCommunicationErrorType(APPLICATION_ERROR);
            handlingType = nn::pia::util::HANDLING_LEAVE_AND_CLEANUP;
            isExit = true;
        }
        else if (!asyncProcessingFlag && !s_IsErrorOccurred)
        {
#if PIA_TEST_CODE
#else
            if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost())
            {
                if (localOperationType == OPERATION_TYPE_NONE)
                {
                    // After the joint session host has called OpenJointSessionAsync/CloseJointSessionAsync, the local station also calls the function.
                    OperationType jointSessionHostOperationType = static_cast<OperationType>(GetOperationType(nn::pia::session::Session::GetInstance()->GetJointSessionHostStationId()));
                    switch (jointSessionHostOperationType)
                    {
                    case OPERATION_TYPE_OPEN_JOINT_SESSION:
                        PIA_SAMPLE_LOG("Call OpenJointSessionAsync because Host is processsing OpenJointSessionAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->OpenJointSessionAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_OPEN_JOINT_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_JOINT_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("OpenJointSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call OpenJointSessionAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        break;
                    case OPERATION_TYPE_CLOSE_JOINT_SESSION:
                        PIA_SAMPLE_LOG("Call CloseJointSessionAsync because Host is processsing CloseJointSessionAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->CloseJointSessionAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_CLOSE_JOINT_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_JOINT_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("CloseJointSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call CloseJointSessionAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        break;
                    default:
                        break;
                    }
                }
            }
            else if (CheckSessionStatus() && isOperationReady)
            {
                // In this sample, the session is closed when it is detected that more than the specified number of teams are participating.
                // However, when multiple teams join simultaneously, the session may not close in time, so check the number of teams after the join-in process has completed.
                // 
                if (g_IsOpened && g_TeamConstitutionManager.GetTeamNum() >= SAMPLE_TEAM_NUM_IN_GAME)
                {
                    PIA_SAMPLE_LOG("Call CloseJointSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->CloseJointSessionAsync();
                    if (result.IsSuccess())
                    {
                        asyncProcessingFlag = true;
                        waitingProcess = SAMPLE_CLOSE_JOINT_SESSION;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_JOINT_SESSION));
                    }
                    else if (result == nn::pia::ResultSessionInconsistentState())
                    {
                        PIA_SAMPLE_LOG("CloseJointSession failed. ResultSessionInconsistentState\n");
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("Call CloseJointSessionAsync failed. Result: 0x%08X\n", result.GetPrintableBits());
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                }
                // Opens the session because there are fewer than the specified number of teams.
                else if (!g_IsOpened && g_TeamConstitutionManager.GetTeamNum() < SAMPLE_TEAM_NUM_IN_GAME)
                {
                    PIA_SAMPLE_LOG("Call OpenJointSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->OpenJointSessionAsync();
                    if (result.IsSuccess())
                    {
                        asyncProcessingFlag = true;
                        waitingProcess = SAMPLE_OPEN_JOINT_SESSION;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_JOINT_SESSION));
                    }
                    else if (result == nn::pia::ResultSessionInconsistentState())
                    {
                        PIA_SAMPLE_LOG("OpenJointSession failed. ResultSessionInconsistentState\n");
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("Call OpenJointSessionAsync failed. Result: 0x%08X\n", result.GetPrintableBits());
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                }
            }
#endif
            if (!asyncProcessingFlag)
            {
                // X: DestroyJointSession (JointSessionHost)
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
                {
#if PIA_TEST_CODE
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        if (nn::pia::session::Session::GetInstance()->IsHost() && CheckSessionStatus())
                        {
                            g_JointSessionType = JOINT_SESSION_TYPE_LEAVE;
                            backToTeam = true;
                            isExit = true;
                        }
                    }
                    else
#endif
                    if (nn::pia::session::Session::GetInstance()->IsJointSessionHost() && CheckSessionStatus())
                    {
                        g_JointSessionType = JOINT_SESSION_TYPE_DESTROY;
                        backToTeam = true;
                        isExit = true;
                    }

                    if (backToTeam)
                    {
                        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
                        nn::pia::session::Session::StationIdIterator itEnd = pSession->End();
                        for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != itEnd; ++it)
                        {
                            if (g_TeamConstitutionManager.GetTeamIndex(*it) != g_TeamConstitutionManager.GetTeamIndex(pSession->GetLocalStationId()))
                            {
                                g_SessionMigrationEventChecker.SetExpectedLeaveEventStationId(*it);
                            }
                        }
                    }
                }
                // B: LeaveSession
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
                {
                    PIA_SAMPLE_LOG("Call LeaveSessionAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->LeaveSessionAsync();
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("Call LeaveSession failed.\n");
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        asyncProcessingFlag = true;
                        waitingProcess = SAMPLE_LEAVE_SESSION;
                        s_IsLeaveProcessing = true;

                        // An EVENT_SESSION_HOST_CHANGED event will not be notified upon starting to leave a session during host migration. Host migration checking ends.
                        if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
                        {
                            g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
                        }
                    }
                }
                // Changes the matchmaking session attribute 'team type' after host migration.
                else if (s_IsNeedModifyAttribute)
                {
                    PIA_SAMPLE_LOG("Call ModifyJointSessionAttributeAsync.\n");
                    result = nn::pia::session::Session::GetInstance()->ModifyJointSessionAttributeAsync(MATCHMAKE_ATTR_INDEX_TEAM_TYPE, static_cast<u32>(GetCurrentTeamType()));
                    if (result.IsSuccess())
                    {
                        asyncProcessingFlag = true;
                        waitingProcess = SAMPLE_MODIFY_ATTRIBUTE;
                        s_IsNeedModifyAttribute = false;
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("Call ModifyJointSessionAttributeAsync failed. Result: 0x%08X\n", result.GetValue());
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        s_IsErrorOccurred = true;
                        isExit = true;
                    }
                }
#if PIA_TEST_CODE
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
                {
                    // Disguise the open/close states so OpenSessionAsync is called when the state is Open, and CloseSessionAsync is called when the state is Close.
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        g_IsOpened = true;
                    }

                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                    {
                        g_IsOpened = false;
                    }

                    if (g_IsOpened)
                    {
                        PIA_SAMPLE_LOG("Call CloseJointSessionAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->CloseJointSessionAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_CLOSE_JOINT_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_CLOSE_JOINT_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("CloseJointSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call CloseJointSessionAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("Call OepnJointSessionAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->OpenJointSessionAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_OPEN_JOINT_SESSION;
                            SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_OPEN_JOINT_SESSION));
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("OpenJointSession failed. ResultSessionInconsistentState\n");
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call OpenJointSessionAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                }
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
                {
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        PIA_SAMPLE_LOG("Call OpenJointSessionParticipationAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->OpenJointSessionParticipationAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_OPEN_JOINT_SESSION_PARTICIPATION;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call OpenJointSessionParticipationAsync failed. Result: 0x%08X\n", result.GetValue());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                    else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                    {
                        PIA_SAMPLE_LOG("Call CloseJointSessionParticipationAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->CloseJointSessionParticipationAsync();
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_CLOSE_JOINT_SESSION_PARTICIPATION;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call CloseJointSessionParticipationAsync failed. Result: 0x%08X\n", result.GetValue());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                }
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
                {
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
                    {
                        // When only some of the parameters are updated, there is no need to set the unnecessary ones.
                        nn::pia::inet::NexUpdateSessionSetting updateSetting;
                        const u32 MATCHMAKE_ATTR_INDEX = 6;
                        u32 attr[nn::pia::inet::ATTRIBUTE_MAX_SIZE];
                        for (int i=0; i < nn::pia::inet::ATTRIBUTE_MAX_SIZE;i++)
                        {
                            if(i != MATCHMAKE_ATTR_INDEX)
                            {
                                attr[i] = 111 * i;
                            }
                        }
                        updateSetting.SetAttributes(attr);
                        const wchar_t description[] = L"newdescription";
                        updateSetting.SetDescription(description);

                        const wchar_t pass[] = L"newpassword";
                        updateSetting.SetSessionUserPassword(pass, 12);
                        updateSetting.SetOpenParticipation(false);
                        updateSetting.SetSelectionPriority(3);
                        updateSetting.SetSessionType(nn::pia::inet::SESSION_TYPE_ANYBODY);
                        updateSetting.SetMaxParticipants(8);
                        updateSetting.SetMinParticipants(2);

                        const u32 applicationDataSize = 16;
                        u8 applicationData[applicationDataSize];

                        for(int i=0;i<applicationDataSize;i++)
                        {
                            applicationData[i] = 100 + i;
                        }
                        updateSetting.SetApplicationData(applicationData, sizeof(applicationData));

                        PIA_SAMPLE_LOG("Call UpdateJointSessionSettingAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->UpdateJointSessionSettingAsync(&updateSetting);
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_UPDATE_JOINT_SESSION_SETTING;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call UpdateJointSessionSettingAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                    else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
                    {
                        const u32 applicationDataSize = 16;
                        u8 applicationData[applicationDataSize];

                        for(int i=0;i<applicationDataSize;i++)
                        {
                            applicationData[i] = 100 + i;
                        }

                        PIA_SAMPLE_LOG("Call UpdateJointSessionApplicationDataAsync.\n");
                        result = nn::pia::session::Session::GetInstance()->UpdateJointSessionApplicationDataAsync(applicationData, sizeof(applicationData));
                        if (result.IsSuccess())
                        {
                            asyncProcessingFlag = true;
                            waitingProcess = SAMPLE_UPDATE_JOINT_SESSION_APPLICATION_DATA;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Call UpdateJointSessionApplicationDataAsync failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                            s_IsErrorOccurred = true;
                            isExit = true;
                        }
                    }
                }
#endif
            }
        }

#if PIA_TEST_CODE
        // The dispatch interval sometimes is adjusted even during the asynchronous process, so operate at some location removed from the checking of the asynchronous process.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT) 
        &&  nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
        {
            isDelayDispatch = !isDelayDispatch;
            delayCount = 0;
            isStopDispatch = false;
        }
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP) 
        &&  nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            PIA_SAMPLE_LOG("Dispatch stoped.");
            isStopDispatch = !isStopDispatch;
            isDelayDispatch = false;
            delayCount = 0;
        }
#endif

        if (asyncProcessingFlag)
        {
            // Checks the progress of the asynchronous process.
            switch (waitingProcess)
            {
                case SAMPLE_LEAVE_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedLeaveSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetLeaveSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("LeaveSession succeeded.\n");
                            isExit = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("LeaveSession failed. Result: 0x%08X\n", result.GetValue());
                        }
                        asyncProcessingFlag = false;
                        s_IsLeaveProcessing = false;
                    }
                }
                break;
                case SAMPLE_MODIFY_ATTRIBUTE:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedModifyJointSessionAttribute())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetModifyJointSessionAttributeResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("ModifyJointSessionAttribute succeeded.\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("ModifyJointSessionAttribute failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                    }
                }
                break;
                case SAMPLE_OPEN_JOINT_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedOpenJointSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetOpenJointSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("OpenJointSession succeeded.\n");
                            g_IsOpened = true;
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("OpenJointSession failed. ResultSessionInconsistentState\n");
                        }
                        else if (result == nn::pia::ResultSessionConnectionIsLost())
                        {
                            PIA_SAMPLE_LOG("OpenJointSession failed. ResultSessionConnectionIsLost\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("OpenJointSession failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
                    }
                }
                break;
                case SAMPLE_CLOSE_JOINT_SESSION:
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompletedCloseJointSession())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetCloseJointSessionResult();
                        if (result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("CloseJointSession succeeded.\n");
                            g_IsOpened = false;
                        }
                        else if (result == nn::pia::ResultSessionInconsistentState())
                        {
                            PIA_SAMPLE_LOG("CloseJointSession failed. ResultSessionInconsistentState\n");
                        }
                        else if (result == nn::pia::ResultSessionConnectionIsLost())
                        {
                            PIA_SAMPLE_LOG("CloseSession failed. ResultSessionConnectionIsLost\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("CloseJointSession failed. Result: 0x%08X\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                        SetLocalOperationType(static_cast<u32>(OPERATION_TYPE_END_OPERATION));
                    }
                }
                break;
#if PIA_TEST_CODE
                case SAMPLE_OPEN_JOINT_SESSION_PARTICIPATION:
                {
                    if(nn::pia::session::Session::GetInstance()->IsCompletedOpenJointSessionParticipation())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetOpenJointSessionParticipationResult();
                        if(result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("OpenJointSessionParticipation succeeded.\n");
                            g_IsOpened = true;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("OpenJointSessionParticipation failed. result = %x\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                    }
                }
                break;
                case SAMPLE_CLOSE_JOINT_SESSION_PARTICIPATION:
                {
                    if(nn::pia::session::Session::GetInstance()->IsCompletedCloseJointSessionParticipation())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetCloseJointSessionParticipationResult();
                        if(result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("CloseJointSessionParticipation succeeded.\n");
                            g_IsOpened = false;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("CloseJointSessionParticipation failed. result = %x\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                    }
                }
                break;
                case SAMPLE_UPDATE_JOINT_SESSION_SETTING:
                {
                    if(nn::pia::session::Session::GetInstance()->IsCompletedUpdateJointSessionSetting())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetUpdateJointSessionSettingResult();
                        if(result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("UpdateJointSessionSetting succeeded.\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("UpdateJointSessionSetting failed. result = %x\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                    }
                }
                break;
                case SAMPLE_UPDATE_JOINT_SESSION_APPLICATION_DATA:
                {
                    if(nn::pia::session::Session::GetInstance()->IsCompletedUpdateJointSessionApplicationData())
                    {
                        result = nn::pia::session::Session::GetInstance()->GetUpdateJointSessionApplicationDataResult();
                        if(result.IsSuccess())
                        {
                            PIA_SAMPLE_LOG("UpdateJointSessionApplicationData succeeded.\n");
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("UpdateJointSessionApplicationData failed. result = %x\n", result.GetValue());
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
                        }
                        asyncProcessingFlag = false;
                    }
                }
                break;
#endif
            }
            if (result.IsFailure())
            {
                s_IsErrorOccurred = true;
                isExit = true;
            }
        }
        else
        {
            // Check the 'session migrating' flag.
            if (IsSessionMigrating())
            {
                // Generated when splitting from a game session to a team session.
                PIA_SAMPLE_LOG("ChangingSessionFlag became true.\n");
                backToTeam = true;
                isExit = true;
            }

            // The check for a disconnection is done by CheckConnection.
            result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
            SetNetworkErrorResult(result);
            if (result.IsFailure())
            {
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("CheckConnection: Session::CheckConnection: Disconnected NETWORK\n");
                }
                else if (result == nn::pia::ResultSessionConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("CheckConnection: Session::CheckConnection: Disconnected SESSION\n");
                }
                else
                {
                    // This value is unlikely.
                    PIA_SAMPLE_LOG("Session::CheckConnection: Unexpected result: 0x%08x\n", result.GetValue());
                    PIA_SAMPLE_ASSERT(false);
                }
                s_IsErrorOccurred = true;
                isExit = true;
            }
        }

        if (!s_IsErrorOccurred)
        {
            // R: ChangeOperationTarget
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                g_StatusManager.ChangeOperationTarget(GetCurrentTeamType());
            }
            // A: AtomicLockOperation
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                g_StatusManager.AtomicLockOperation();
            }
        
#if PIA_TEST_CODE
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
                PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());
                if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
                {
                    pMgr->SetSendThreadPacketLossRatio(100);
                    pMgr->SetReceiveThreadPacketLossRatio(10);
                    PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
                }
                else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
                {
                    pMgr->SetSendThreadPacketLossRatio(10);
                    pMgr->SetReceiveThreadPacketLossRatio(100);
                    PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
                }
                else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    pMgr->SetSendThreadPacketLossRatio(10);
                    pMgr->SetReceiveThreadPacketLossRatio(10);
                    PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
                }
                else 
                {
                    pMgr->SetSendThreadPacketLossRatio(100);
                    pMgr->SetReceiveThreadPacketLossRatio(100);
                    PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
                }
            }
#endif

            s32 diffX = 0, diffY = 0;
            // Checks for input to move rectangle.
            {
                f32 analogX, analogY;
                nn::pia::sampleframework::SampleFramework::GetInstance()->GetAnalog(&analogX, &analogY);
                if (analogX != 0.0f || analogY != 0.0f)
                {
                    const f32 cAnalogMoveBase = 3.0f;
                    diffX = static_cast<s32>(cAnalogMoveBase * analogX);
                    diffY = static_cast<s32>(cAnalogMoveBase * analogY * (-1.0f));  // Up/down on analog stick is reversed from up/down when drawing. (Negative and positive are reversed.)
                }
                else
                {
                    const s32 cMoveBase = 2;
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
                    {
                        diffX -= cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
                    {
                        diffX += cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                    {
                        diffY -= cMoveBase;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                    {
                        diffY += cMoveBase;
                    }

                    // Always move, even if there is no input for that, so communication is easier to grasp.
                    if (diffX == 0 && diffY == 0)
                    {
                        diffX = 1;
                    }
                }
            }

            // The sending process before the Dispatch process.
            if (IsInJointSession()) // Do not send while the session is migrating.
            {
                // Checks the sending interval.
                if ((frameNo % (SAMPLE_INTERVAL_SENDING_STATION_STATUS + 1)) == 0)
                {
                    // Updates the data to send.
                    UpdateSendData(diffX, diffY);
                }
            }

            // Updates the clock for CloneProtocol.
            Clone_GetCloneProtocol()->UpdateClock();

#if PIA_TEST_CODE
            // Reduce the frequency of dispatch for testing, so that it is easier to pinpoint a certain timing.
            if (isDelayDispatch)
            {
                ++delayCount;
                delayCount = delayCount % DELAY_DISPATCH_NUM;
            }

            if ((!isDelayDispatch || (isDelayDispatch && delayCount == DELAY_DISPATCH_NUM-1)) && !isStopDispatch)
            {
#endif

            // Execute Dispatch.
            nn::pia::common::Scheduler::GetInstance()->Dispatch();
            nn::nex::Scheduler::GetInstance()->DispatchAll();

#if PIA_TEST_CODE
            }
#endif

            // Checks whether the team roster has changed.
            g_TeamConstitutionManager.Update();

            // The receiving process after the Dispatch process.
            // Updates the status based on the received data.
            g_StatusManager.Update();

            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE_GAME,
                true,
                nn::pia::session::Session::GetInstance()->IsJointSessionHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum(),
                0
                );

            // Screen rendering.
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

                // String.
                {
                    // Check the role of the local station.
                    int roleIndex = 0;  //Unknown;
                    int jointRoleIndex = 0;  //Unknown;
                    localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
                    if (localStationId != nn::pia::STATION_ID_INVALID)
                    {
                        if (nn::pia::session::Session::GetInstance()->IsHost())
                        {
                            roleIndex = 1;  // 1:Host
                        }
                        else
                        {
                            roleIndex = 2;  // 2:Client
                        }

                        if (nn::pia::session::Session::GetInstance()->IsJointSessionHost())
                        {
                            jointRoleIndex = 1;  // 1:JointSessionHost
                        }
                        else
                        {
                            jointRoleIndex = 2;  // 2:JointSessionClient
                        }
                    }

                    int line = 0;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Communication (Game), TeamNum: %d", g_TeamConstitutionManager.GetTeamNum());
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum: %2d, Role:%-6s %-18s", 
                        nn::pia::session::Session::GetInstance()->GetStationNum(), 
                        SESSION_ROLE(roleIndex), JOINT_SESSION_ROLE(jointRoleIndex));
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationId: %lld, TeamMemberIndex: %2d", 
                        localStationId.GetValue(), 
                        g_TeamConstitutionManager.GetMemberIndex(localStationId));
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  SessionId: %d, JointSessionId: %d", 
                        nn::pia::session::Session::GetInstance()->GetSessionId(), 
                        nn::pia::session::Session::GetInstance()->GetJointSessionId());
                    line += FONT_SIZE;

                    line = SCREEN_HEIGHT - (FONT_SIZE * 10);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: DestroyJointSession (JointSessionHost)");
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: LeaveSession");
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: ChangeOperationTarget");
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: TryLock (Atomic)");
                    line += FONT_SIZE;

#if PIA_TEST_CODE
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[OpenJointSession  Seq] %d", 
                        nn::pia::session::Session::GetInstance()->GetOpenJointSessionSequence());
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[CloseJointSession Seq] %d", 
                        nn::pia::session::Session::GetInstance()->GetCloseJointSessionSequence());
                    line += FONT_SIZE;
#else
                    line += FONT_SIZE * 2;
#endif
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
                    line += FONT_SIZE;
                }

                // Draws StationStatus.
                u8 joinedTeamFlag = 0x0;
                for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
                {
                    SampleObjectStatus* pStatus = g_StatusManager.GetStationStatus(i);
                    if (pStatus->stationId != nn::pia::STATION_ID_INVALID)
                    {
                        DrawStationStatus(*pStatus);
                        int teamIndex = g_TeamConstitutionManager.GetTeamIndex(pStatus->stationId);
                        if (teamIndex >= 0)
                        {
                            joinedTeamFlag |= 0x1 << teamIndex;
                        }
                    }
                }

                // Draws the frame for distinguishing the team.
                for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
                {
                    if (joinedTeamFlag & (0x1 << i))
                    {
                        DrawTeamFrame(i, GetColorOfTeamType(static_cast<SampleTeamType>(i)));
                    }
                }

                // Draws common objects.
                for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
                {
                    SampleObjectStatus* pStatus = g_StatusManager.GetCommonObjectStatus(i);
                    if (pStatus->type == OBJECT_STATUS_TYPE_TEAM_ATOMIC
                    ||  pStatus->type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL)
                    {
                        if (pStatus->teamIndex < 0 || (joinedTeamFlag & (0x1 << pStatus->teamIndex)) == 0)
                        {
                            continue;
                        }
                    }
                    DrawStationStatus(*pStatus);
                }
            }
        }
#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_CLOSE_SESSION_SEQUENCE, nn::pia::session::Session::GetInstance()->GetCloseJointSessionSequence());
        // PIA_SAMPLE_LOG("CloseSession sq: %d.\n", nn::pia::session::Session::GetInstance()->GetCloseSessionSequence());
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_OPEN_SESSION_SEQUENCE, nn::pia::session::Session::GetInstance()->GetCloseJointSessionSequence());
        // PIA_SAMPLE_LOG("OpenSession sq: %d.\n", nn::pia::session::Session::GetInstance()->GetOpenSessionSequence());
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_LOCAL_STATION_INDEX, nn::pia::session::Session::GetInstance()->GetLocalStationIndex());
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SAMPLE_PROGRAMPARAM_OPENCLOSESESSION_COMPLETE, g_IsOpened ? 1 : 0);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    if (s_IsErrorOccurred)
    {
        PIA_SAMPLE_ASSERT(asyncProcessingFlag == false);

        if (handlingType == nn::pia::util::HANDLING_VOID)
        {
            handlingType = nn::pia::util::GetHandlingType(result);
            SetNetworkErrorResult(result);
            
            if (result == nn::pia::ResultSessionConnectionIsLost())
            {
                SetCommunicationErrorType(SESSION_LOST_ERROR);
            }
            else if (result == nn::pia::ResultNetworkConnectionIsLost())
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            }
            else
            {
                SetCommunicationErrorType(MATCHMAKE_ERROR);
            }
        }

        if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP)
        {
            // An EVENT_SESSION_HOST_CHANGED event will not be notified upon starting to leave a session during host migration. Host migration checking ends.
            if (g_HostMigrationEventChecker.GetCheckTarget() == EventChecker::CHECK_TARGET_HOST_MIGRATION)
            {
                PIA_SAMPLE_LOG("cancel HostMigrationEventChecker (Game)");
                g_HostMigrationEventChecker.CheckEnd(EventChecker::CHECK_TARGET_HOST_MIGRATION);
            }

            SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
            if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                // Leaves the network in a state where it does not need to be reset, because the process can return to matchmaking when you clean up and restart the Session object.
                SetResetOfNetworkIsNecessary(false);
            }
        }

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    if (backToTeam)
    {
        return SCENE_CHANGING_SESSION_PREPROCESSING;
    }

    // Return to team selection.
    Session_CleanupSession();
    return SCENE_SELECT_TEAM;
}
