﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneError.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/util.h>

#include "SampleLib/SampleLib.h"

#include "Shared_Inet_Setup.h"

#include "SampleConfig.h"

namespace
{
SampleErrorType s_ErrorType = ERROR_UNKNOWN;  // Variable that holds the errors that occurred most recently.

const char* ERROR_INFO_STRING[] =
    {
        "Unknown Reason",
        "Initialize Error (wireless OFF)",
        "Initialize Error (other reason)",
        "Matchmake Failed",
        "Session Lost",
        "Connection Lost",
        "Application Error"
    };

nn::Result s_NetworkErrorResult = nn::ResultSuccess();

bool s_ResetOfNetworkIsNecessary = true;
}

/*!
    @brief  This function sets the error reason before jumping to the error state.
 */
void SetCommunicationErrorType(SampleErrorType error)
{
    s_ErrorType = error;
}

// Pass the result of errors that occur when connecting to the Internet.
// The Result passed here is then passed to the network error code conversion API and output to the screen.
// 
void SetNetworkErrorResult(nn::Result result)
{
    s_NetworkErrorResult = result;
}

nn::Result GetNetworkErrorResult(void)
{
    return s_NetworkErrorResult;
}

//This function sets whether it will be necessary to reset the network (for example, logging in again) before jumping to the error state.
void SetResetOfNetworkIsNecessary(bool isNecessary)
{
    s_ResetOfNetworkIsNecessary = isNecessary;
}

bool GetResetOfNetworkIsNecessary()
{
    return s_ResetOfNetworkIsNecessary;
}


/*!
    @brief  This scene displays the network error code for errors occurring in the Internet connection environment.
    
    In a normal program, pass the network error codes to the Error/EULA Viewer.
 */
SampleSceneNo SceneNetworkError()
{
    // Convert the error <tt>Result</tt> to a network error code.
    // Normally, you would pass this error code to the error/EULA applet.
    // However, the error/EULA applet cannot be used unless development hardware is set to <tt>prod</tt> mode.
    // For simplicity, this sample demo does not use the error/EULA applet.
    // Not used.
    nn::Result result = GetNetworkErrorResult();
    PIA_SAMPLE_LOG_RESULT(result);
    bool isPiaError = nn::pia::util::IsPiaResult(result);
    u32 networkErrorCode = (isPiaError) ? nn::pia::util::NetworkErrorCodeConverter::Convert(result) : Inet_GetNetworkErrorCode();

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_NETWORK_ERROR, false);

    int frameNo = 0;
    bool isExit = false;

    // Rendering
    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Exit the loop if the A Button is pressed.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            isExit = true;
        }

        int line;

        // Renders the screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A network error occured.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
        line += FONT_SIZE;
        line += FONT_SIZE;

        if (s_ErrorType != APPLICATION_ERROR)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Result: 0x%08X", GetNetworkErrorResult().GetPrintableBits());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Level:       %4d", GetNetworkErrorResult().GetLevel());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Summary:     %4d", GetNetworkErrorResult().GetSummary());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Module:      %4d", GetNetworkErrorResult().GetModule());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Description: %4d", GetNetworkErrorResult().GetDescription());
            line += FONT_SIZE;
            line += FONT_SIZE;

            if (networkErrorCode != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Network error code: %07d", networkErrorCode);
                line += FONT_SIZE;
            }
            if (Inet_GetLoginNetworkErrorCode() != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Login error code: %07d", Inet_GetLoginNetworkErrorCode());
                line += FONT_SIZE;
            }
        }
        line = SCREEN_HEIGHT - (FONT_SIZE * 4);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        nn::nex::Scheduler::GetInstance()->DispatchAll();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        frameNo++;
    }

    //Clears the error information, just in case.
    s_ErrorType = ERROR_UNKNOWN;
    SetNetworkErrorResult(nn::ResultSuccess());

    SampleSceneNo nextSceneNo;
    if (GetResetOfNetworkIsNecessary())
    {
        nextSceneNo = SCENE_RESET;
    }
    else
    {
        nextSceneNo = SCENE_SELECT_TEAM;
    }

    // Initializes the flag indicating whether the network needs to be reset.
    SetResetOfNetworkIsNecessary(true);

    return nextSceneNo;
}
