﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>
#include <pia/session.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"

#include "SampleConfig.h"
#include "EventChecker.h"
#include "TeamConstitutionManager.h"

namespace
{
SampleTeamType s_CurrentTeamType = TEAM_ALPHA;

const nn::pia::sampleframework::Color cTeamColorAlpha(1.0f, 1.0f, 0.0f, 1.0f);
const nn::pia::sampleframework::Color cTeamColorBravo(1.0f, 0.0f, 1.0f, 1.0f);
const nn::pia::sampleframework::Color cTeamColorCharlie(0.0f, 0.0f, 1.0f, 1.0f);
const nn::pia::sampleframework::Color cTeamColorDelta(1.0f, 0.0f, 0.0f, 1.0f);
const nn::pia::sampleframework::Color cTeamColor[] = {
    cTeamColorAlpha, cTeamColorBravo, cTeamColorCharlie, cTeamColorDelta
};
const char* cTeamName[] = {
    "TEAM ALPHA", "TEAM BRAVO", "TEAM CHARLIE", "TEAM DELTA"
};

const nn::pia::sampleframework::Color cSelectMarkColor(0.8f, 0.8f, 0.8f, 1.0f);
const s32 cSelectMarkSize = 20;

#if PIA_TEST_CODE
u32 s_MaxSilenceTime = nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT;
#endif

void DrawTeamExplanatoryNotes(
    s32 posLeftTopX,
    s32 posLeftTopY,
    const char* teamNameStr,
    const nn::pia::sampleframework::Color& teamColor)
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        posLeftTopX + 5, posLeftTopY + 5, cSelectMarkSize - 10, cSelectMarkSize - 10, teamColor);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
        posLeftTopX + cSelectMarkSize + 5, posLeftTopY + 6, teamNameStr);
}

void DrawSelectMark(s32 posLeftTopX, s32 posLeftTopY)
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        posLeftTopX, posLeftTopY, 2, cSelectMarkSize, cSelectMarkColor);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        posLeftTopX, posLeftTopY, cSelectMarkSize, 2, cSelectMarkColor);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        posLeftTopX + cSelectMarkSize - 2, posLeftTopY, 2, cSelectMarkSize, cSelectMarkColor);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
        posLeftTopX, posLeftTopY + cSelectMarkSize - 2, cSelectMarkSize, 2, cSelectMarkColor);
}

}

// Function for using information related to team type in communicating scenes.
SampleTeamType GetCurrentTeamType()
{
    return s_CurrentTeamType;
}

const nn::pia::sampleframework::Color GetColorOfTeamType(SampleTeamType teamType)
{
    if (teamType < TEAM_TYPE_NUM)
    {
        return cTeamColor[teamType];
    }
    return nn::pia::sampleframework::COLOR_GRAY;
}


SampleSceneNo SceneSelectTeam()
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_SELECT_TEAM,
        false,
        false,
        0);

    u32 frameNo = 0;
    bool isExit = false;
    bool errorOccurred = false;
    bool logout = false;

    s_CurrentTeamType = TEAM_ALPHA;
    SampleTeamType selectTeamBuf = s_CurrentTeamType;

#if PIA_TEST_CODE
    IsEnableHostMigration(true);
    SetMaxPlayerNum(SAMPLE_GAME_PLAYER_MAX);
    s_MaxSilenceTime = nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT;
#endif

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Up/Down Buttons: Change team type.
        const u32 cTeamTypeNum = static_cast<u32>(SAMPLE_TEAM_NUM_IN_GAME);
        u32 selectMarkPosBuf = static_cast<u32>(selectTeamBuf);
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            selectMarkPosBuf = (selectMarkPosBuf + cTeamTypeNum - 1) % cTeamTypeNum;
        }
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            selectMarkPosBuf = (selectMarkPosBuf + 1) % cTeamTypeNum;
        }
        selectTeamBuf = static_cast<SampleTeamType>(selectMarkPosBuf);

        // A Button: Select team.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            isExit = true;
        }
        // START: Logout.
        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            logout = true;
            isExit = true;
        }
#if PIA_TEST_CODE
        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            IsEnableHostMigration(!IsEnableHostMigration());
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_ENABLE_HOST_MIGRATION, IsEnableHostMigration() ? 1 : 0);
            PIA_SAMPLE_LOG("Set Host Migration %s\n", IsEnableHostMigration() ? "true" : "false");
        }
        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            SetMaxPlayerNum(SAMPLE_GAME_PLAYER_MAX / SAMPLE_TEAM_NUM_IN_GAME + 1);
            PIA_SAMPLE_LOG("Set Max Player Num: %d\n", GetMaxPlayerNum());
        }
        else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            s_MaxSilenceTime = 60 * 1000;
            PIA_SAMPLE_LOG("Set Max Silence Time: %d\n", s_MaxSilenceTime);
        }
#endif

        // Execute dispatch.
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Checks the AC network connection status, game server connection status, and NEX internal errors.
        if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
        {
            errorOccurred = true;
            isExit = true;
        }

        //Screen rendering.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Select Team.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Please select your team type.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        s32 xBuf, yBuf;
        xBuf = 30;
        yBuf = line;

        for (u32 i = 0; i < cTeamTypeNum; i++)
        {
            DrawTeamExplanatoryNotes(xBuf, yBuf, cTeamName[i], cTeamColor[i]);
            if (selectTeamBuf == static_cast<SampleTeamType>(i))
            {
                DrawSelectMark(xBuf, yBuf);
            }
            yBuf += cSelectMarkSize;
        }

        line = SCREEN_HEIGHT - (FONT_SIZE * 6);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Change team type.");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Decide team type.");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  START: Logout.");
        line += FONT_SIZE;
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        ++frameNo;
    }

    s_CurrentTeamType = selectTeamBuf;

    if (errorOccurred)
    {
        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
        SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        return SCENE_NETWORK_ERROR;
    }
    if (logout)
    {
        return SCENE_RESET;
    }

    return SCENE_MATCH_MAKE_TEAM;
}


SampleSceneNo SceneMatchmake()
{
    PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_PREPARING_TEAM);
    g_GameState = GAME_STATE_PREPARING_TEAM;
    g_SessionMigrationEventChecker.Reset();
    g_HostMigrationEventChecker.Reset();

    ResetErrorOccurredFlag();
    g_IsOpened = false;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_MATCH_MAKE_TEAM,
        false,
        false,
        0);

    PIA_SAMPLE_LOG("ScenMatchmake start.\n");

    // Unregister callback for determining whether to approve a join.
    nn::pia::session::Session::GetInstance()->UnregisterJoiningApprovalCallback();

    // Put team information in token.
    SampleTokenData tokenData;
#if 0
    tokenData.sessionId = nn::pia::session::Session::GetInstance()->GetSessionId();
#endif
    tokenData.teamType = static_cast<u8>(GetCurrentTeamType());
    nn::pia::transport::Station::IdentificationToken token;
    memcpy(token.data, &tokenData, sizeof(tokenData));

    nn::pia::session::Session::StartupSetting startupSetting;
    startupSetting.pToken = &token;
#if PIA_TEST_CODE
    startupSetting.bUsingHostMigration = IsEnableHostMigration();
    startupSetting.maxSilenceTime = s_MaxSilenceTime;
#else
    startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;
#endif
    nn::Result r = Session_StartupSession(startupSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);

    // Settings for the matchmaking session to create for a team.
    nn::pia::inet::NexCreateSessionSetting createSessionSetting;
    createSessionSetting.SetGameMode(MATCHMAKE_GAME_MODE_TEAM);  // Set the game mode to "team."
    createSessionSetting.SetMaxParticipants(SAMPLE_TEAM_MEMBER_MAX);
    createSessionSetting.SetMinParticipants(1);
    createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX_TEAM_TYPE, static_cast<u32>(s_CurrentTeamType));
    createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, 0);
    createSessionSetting.SetOpenParticipation(true);

    // Auto-match search condition settings.
    const size_t cSearchCriteriaNum = 1;
    nn::pia::inet::NexSessionSearchCriteria nexSearchCriteriaArray[cSearchCriteriaNum];
    nexSearchCriteriaArray[0].SetGameMode(MATCHMAKE_GAME_MODE_TEAM);
    nexSearchCriteriaArray[0].SetMaxParticipants(SAMPLE_TEAM_MEMBER_MAX);
    // Search for <tt>TeamType</tt> same as own.
    nexSearchCriteriaArray[0].SetAttribute(MATCHMAKE_ATTR_INDEX_TEAM_TYPE, static_cast<u32>(s_CurrentTeamType));
    nexSearchCriteriaArray[0].SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, 0);

#if PIA_TEST_CODE
    bool isDelayDispatch = false;
    bool isStopDispatch = false;
    u8 delayCount = 0;
#endif

    u32 frameNo = 0;
    bool isExit = false;
    bool errorOccurred = false;
    bool backToSelectTeam = false;
    bool logout = false;

    bool asyncProcessingFlag = false;

    while (!isExit)
    {
        if (asyncProcessingFlag)
        {
            if (nn::pia::session::Session::GetInstance()->IsCompletedJoinRandomSession())
            {
                isExit = true;
                if (nn::pia::session::Session::GetInstance()->GetJoinRandomSessionResult().IsFailure())
                {
                    errorOccurred = true;
                    SetNetworkErrorResult(nn::pia::session::Session::GetInstance()->GetJoinRandomSessionResult());
                    PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetJoinRandomSessionResult() != nn::pia::ResultMeshIsFull());
                }
            }
#if PIA_TEST_CODE
            else if (nn::pia::session::Session::GetInstance()->IsCompletedCreateSession())
            {
                isExit = true;
                if (nn::pia::session::Session::GetInstance()->GetCreateSessionResult().IsFailure())
                {
                    errorOccurred = true;
                    SetNetworkErrorResult(nn::pia::session::Session::GetInstance()->GetCreateSessionResult());
                    PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetCreateSessionResult() != nn::pia::ResultMeshIsFull());
                }
            }
            else if (nn::pia::session::Session::GetInstance()->IsCompletedBrowseSession())
            {
                if (nn::pia::session::Session::GetInstance()->GetBrowseSessionResult().IsFailure())
                {
                    errorOccurred = true;
                    isExit = true;
                    SetNetworkErrorResult(nn::pia::session::Session::GetInstance()->GetBrowseSessionResult());
                }
                else if (nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize() > 0)
                {
                    nn::pia::inet::NexJoinSessionSetting joinSessionSetting;

                    joinSessionSetting.SetSessionInfoPtr(*(nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin()));
                    joinSessionSetting.SetSessionUserPassword(USER_PASSWORD[s_CurrentTeamType], USER_PASSWORD_LENGTH);

                    nn::Result result = nn::pia::session::Session::GetInstance()->JoinSessionAsync(&joinSessionSetting);
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("call JoinSession is failure.\n");
                        if (result != nn::pia::ResultNetworkConnectionIsLost())
                        {
                            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                        }
                        errorOccurred = true;
                        isExit = true;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("Not found.\n");
                    asyncProcessingFlag = false;
                }
            }
            else if (nn::pia::session::Session::GetInstance()->IsCompletedJoinSession())
            {
                isExit = true;
                if (nn::pia::session::Session::GetInstance()->GetJoinSessionResult().IsFailure())
                {
                    errorOccurred = true;
                    SetNetworkErrorResult(nn::pia::session::Session::GetInstance()->GetJoinSessionResult());
                    PIA_SAMPLE_ASSERT(nn::pia::session::Session::GetInstance()->GetJoinSessionResult() != nn::pia::ResultMeshIsFull());
                }
            }

            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SAMPLE_PROGRAMPARAM_GETJOINSEQUENCE, nn::pia::session::Session::GetInstance()->GetJoinAutoMatchmakeSequence());
#endif
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        if (!asyncProcessingFlag)
        {
            // A Button: Start AutoMatchmake.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->JoinRandomSessionAsync(
                    &createSessionSetting, nexSearchCriteriaArray, cSearchCriteriaNum);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("call JoinRandomSessionAsync is failure.\n");
                    if (result != nn::pia::ResultNetworkConnectionIsLost())
                    {
                        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                    }
                    errorOccurred = true;
                    isExit = true;
                }
                else
                {
                    PIA_SAMPLE_LOG("call JoinRandomSessionAsync is success.\n");
                    asyncProcessingFlag = true;
                }
            }
#if PIA_TEST_CODE
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
                {
                    createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, 1);
                    createSessionSetting.SetSessionUserPassword(USER_PASSWORD[s_CurrentTeamType], USER_PASSWORD_LENGTH);

                    nn::Result result = nn::pia::session::Session::GetInstance()->CreateSessionAsync(&createSessionSetting);
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("call CreateSession is failure.\n");
                        if (result != nn::pia::ResultNetworkConnectionIsLost())
                        {
                            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                        }
                        errorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        asyncProcessingFlag = true;
                    }
                }
                else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
                {
                    nexSearchCriteriaArray[0].SetAttribute(MATCHMAKE_ATTR_INDEX_USERPASSWORD, 1);
                    nexSearchCriteriaArray[0].SetExcludeUserPasswordSet(false);
                    nexSearchCriteriaArray[0].SetResultRange(0, 4);
                    nn::Result result = nn::pia::session::Session::GetInstance()->BrowseSessionAsync(nexSearchCriteriaArray);
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("call BrowseSession is failure.\n");
                        if (result != nn::pia::ResultNetworkConnectionIsLost())
                        {
                            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                        }
                        errorOccurred = true;
                        isExit = true;
                    }
                    else
                    {
                        asyncProcessingFlag = true;
                    }
                }
            }
#endif
            // B Button: Back to SelectTeam.
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                backToSelectTeam = true;
                isExit = true;
            }
            // START: Logout.
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
            {
                logout = true;
                isExit = true;
            }
        }


#if PIA_TEST_CODE
        if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT) && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
        {
            isDelayDispatch = !isDelayDispatch;
            PIA_SAMPLE_LOG("matchmake Dispatch delay %s\n", isDelayDispatch ? "enable":"disable");
            delayCount = 0;
            isStopDispatch = false;
        }
        if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP) && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            isStopDispatch = !isStopDispatch;
            PIA_SAMPLE_LOG("matchmake Dispatch %s.\n", isStopDispatch ? "stop":"resume");
            isDelayDispatch = false;
            delayCount = 0;
        }
#endif


#if PIA_TEST_CODE
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
            PIA_SAMPLE_LOG("[JointSequence] %d\n", nn::pia::session::Session::GetInstance()->GetJointSequence());
            if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(10);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 100%, Receive 10%).\n");
            }
            else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
            {
                pMgr->SetSendThreadPacketLossRatio(10);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send 10%, Receive 100%).\n");
            }
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                pMgr->SetSendThreadPacketLossRatio(0);
                pMgr->SetReceiveThreadPacketLossRatio(0);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 10%).\n");
            }
            else
            {
                pMgr->SetSendThreadPacketLossRatio(100);
                pMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("Enable packet loss emulation (Send/Receive 100%).\n");
            }
        }
#endif

#if PIA_TEST_CODE
        // Reduce the frequency of dispatch for testing, so that it is easier to pinpoint a certain timing.
        if(isDelayDispatch)
        {
            ++delayCount;
            delayCount = delayCount % DELAY_DISPATCH_NUM;
        }

        if( (!isDelayDispatch || (isDelayDispatch && delayCount == DELAY_DISPATCH_NUM-1)) && !isStopDispatch)
        {
#endif
        // Execute dispatch.
        nn::nex::Scheduler::GetInstance()->DispatchAll();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
#if PIA_TEST_CODE
        }
#endif

        // Check for internal NEX errors.
        if (Nex_CheckInternalError())
        {
            errorOccurred = true;
            isExit = true;
        }

        //Screen rendering.
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Matchmake.");
        line += FONT_SIZE;
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  CurrentTeamType:");
        line += FONT_SIZE;
        const s32 xBuf = 30;
        switch (s_CurrentTeamType)
        {
        case TEAM_ALPHA:
            DrawTeamExplanatoryNotes(xBuf, line, "TEAM ALPHA", cTeamColorAlpha);
            break;
        case TEAM_BRAVO:
            DrawTeamExplanatoryNotes(xBuf, line, "TEAM BRAVO", cTeamColorBravo);
            break;
        case TEAM_CHARLIE:
            DrawTeamExplanatoryNotes(xBuf, line, "TEAM CHARLIE", cTeamColorCharlie);
            break;
        case TEAM_DELTA:
            DrawTeamExplanatoryNotes(xBuf, line, "TEAM DELTA", cTeamColorDelta);
            break;
        }
        line += cSelectMarkSize + FONT_SIZE;

        if (!asyncProcessingFlag)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Please select operation.");
            line += FONT_SIZE;
        }
        else
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Processing is in progress...");
            line += FONT_SIZE;
        }

        line = SCREEN_HEIGHT - (FONT_SIZE * 8);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += FONT_SIZE;
        if (!asyncProcessingFlag)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Start JoinRandomSession.");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Back to SelectTeam.");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  START: Logout.");
            line += FONT_SIZE;
        }
        else
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Can not accept operation now.");
            line += FONT_SIZE;
            line += FONT_SIZE * 2;
        }
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        ++frameNo;
    }

    if (errorOccurred)
    {
        if (asyncProcessingFlag
#if PIA_TEST_CODE
        && !nn::pia::session::Session::GetInstance()->IsCompletedCreateSession()
        && !nn::pia::session::Session::GetInstance()->IsCompletedBrowseSession()
        && !nn::pia::session::Session::GetInstance()->IsCompletedJoinSession()
#endif
        && !nn::pia::session::Session::GetInstance()->IsCompletedJoinRandomSession())
        {
            // If interrupted because of abnormal communication conditions.
            SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        }

        SetCommunicationErrorType(MATCHMAKE_ERROR);

        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    if (backToSelectTeam)
    {
        Session_CleanupSession();
        return SCENE_SELECT_TEAM;
    }
    if (logout)
    {
        return SCENE_RESET;
    }

    nn::pia::session::Session::Status sessionStatus = nn::pia::session::Session::GetInstance()->GetStatus();
    // NicDown is invoked at the same time as a successful session join, and there are cases where a socket error is detected in the sending and receiving thread.
    if (sessionStatus == nn::pia::session::Session::STATUS_DISCONNECTED_NETWORK)
    {
        return SCENE_RESET;
    }
    // If session migration is started immediately after a session join process starts,
    // there may be cases where session migration will not start.
    // When the session reconstruction that occurs as part of session migration happens at the same time as a
    // successful session join, the session state may end up disconnected even though the join process succeeded.
    if (sessionStatus == nn::pia::session::Session::STATUS_DISCONNECTED_SESSION)
    {
        SampleSessionAsyncStatus leaveSessionAsyncStatus = Session_LeaveSession();
        if (leaveSessionAsyncStatus == SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            // Leaves the network in a state where it does not need to be reset, because the process can return to matchmaking when you clean up and restart the Session object.
            SetResetOfNetworkIsNecessary(false);
        }
        SetNetworkErrorResult(nn::pia::ResultSessionConnectionIsLost());
        SetCommunicationErrorType(MATCHMAKE_ERROR);
        Session_CleanupSession();
        return SCENE_NETWORK_ERROR;
    }

    PIA_SAMPLE_LOG("GameState %d -> %d\n", g_GameState, GAME_STATE_TEAM);
    g_GameState = GAME_STATE_TEAM;
    g_IsOpened = true;
    return SCENE_COMMUNICATE_TEAM;
}
