﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     StatusManager.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include "StatusManager.h"
#include "Shared_Clone_Setup.h"

namespace
{
}


StatusManager::StatusManager()
{
    m_pOperationTargetStatus = &m_StationStatus[0];
}

void StatusManager::ClearStationStatus()
{
    for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
    {
        m_StationStatus[i].Clear();
    }
}

void StatusManager::ClearCommonObjectStatus()
{
    for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
    {
        m_CommonObjectStatus[i].Clear();
    }
}

void StatusManager::Update()
{
    // Updates the status based on the received data.
    for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
    {
        m_StationStatus[i].Update();
    }

    for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
    {
        m_CommonObjectStatus[i].Update();
    }
}

void StatusManager::AssignStationStatus(nn::pia::StationId stationId, ReceiveObject* pReceiveObject)
{
    // Searches for unused <tt>StationStatus</tt>.
    int idx = -1;
    for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
    {
        if (m_StationStatus[i].stationId == nn::pia::STATION_ID_INVALID)
        {
            idx = i;
            break;
        }
    }
    PIA_SAMPLE_ASSERT(idx != -1);
    m_StationStatus[idx].Clear();
    m_StationStatus[idx].type = OBJECT_STATUS_TYPE_STATION;
    m_StationStatus[idx].stationId = stationId;
    m_StationStatus[idx].pObject = pReceiveObject;
}

SampleObjectStatus* StatusManager::GetStationStatus(nn::pia::StationId stationId)
{
    // Returns a <tt>StationStatus</tt> that has the specified <tt>StationID</tt>.
    for (int i = 0; i < SAMPLE_GAME_PLAYER_MAX; ++i)
    {
        if (m_StationStatus[i].stationId == stationId)
        {
            return &m_StationStatus[i];
        }
    }
    return NULL;
}

SampleObjectStatus* StatusManager::GetStationStatus(int idx)
{
    PIA_SAMPLE_ASSERT(idx < SAMPLE_GAME_PLAYER_MAX);
    return &m_StationStatus[idx];
}

void StatusManager::AssignCommonObjectStatus(SampleObjectStatusType type, int teamIndex, Object* pObject)
{
    // Searches for unused <tt>CommonObjectStatus</tt>.
    int idx = -1;
    for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
    {
        if (m_CommonObjectStatus[i].type == OBJECT_STATUS_TYPE_NONE)
        {
            idx = i;
            break;
        }
    }
    PIA_SAMPLE_ASSERT(idx != -1);
    m_CommonObjectStatus[idx].Clear();
    m_CommonObjectStatus[idx].teamIndex = teamIndex;
    m_CommonObjectStatus[idx].type = type;
    m_CommonObjectStatus[idx].pObject = pObject;
    m_CommonObjectStatus[idx].pos.y = 24 + type * 12;
}

SampleObjectStatus* StatusManager::GetCommonObjectStatus(int idx)
{
    PIA_SAMPLE_ASSERT(idx < COMMON_OBJECT_NUM);
    return &m_CommonObjectStatus[idx];
}

SampleObjectStatus* StatusManager::GetCommonObjectStatusByType(SampleObjectStatusType type, int teamIndex)
{
    for (int i = 0; i < COMMON_OBJECT_NUM; ++i)
    {
        if (m_CommonObjectStatus[i].type == type)
        {
            if (type == OBJECT_STATUS_TYPE_TEAM_ATOMIC || type == OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL)
            {
                PIA_SAMPLE_ASSERT(teamIndex != -1);
                if (teamIndex != m_CommonObjectStatus[i].teamIndex)
                {
                    continue;
                }
            }
            return &m_CommonObjectStatus[i];
        }
    }
    return NULL;
}

void StatusManager::ChangeOperationTarget(int localTeamIndex)
{
    PIA_SAMPLE_ASSERT(m_pOperationTargetStatus != NULL);
    switch(m_pOperationTargetStatus->type)
    {
    case OBJECT_STATUS_TYPE_STATION:
        {
            m_pOperationTargetStatus = GetCommonObjectStatusByType(OBJECT_STATUS_TYPE_TEAM_ATOMIC, localTeamIndex);
        }
        break;
    case OBJECT_STATUS_TYPE_TEAM_ATOMIC:
        {
            m_pOperationTargetStatus = GetCommonObjectStatusByType(OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL, localTeamIndex);
        }
        break;
    case OBJECT_STATUS_TYPE_TEAM_SEQUENTIAL:
        {
            if (g_GameState == GAME_STATE_GAME)
            {
                m_pOperationTargetStatus = GetCommonObjectStatusByType(OBJECT_STATUS_TYPE_COMMON_ATOMIC);
            }
            else
            {
                m_pOperationTargetStatus = GetStationStatus(nn::pia::session::Session::GetInstance()->GetLocalStationId());
            }
        }
        break;
    case OBJECT_STATUS_TYPE_COMMON_ATOMIC:
        {
            m_pOperationTargetStatus = GetCommonObjectStatusByType(OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL);
        }
        break;
    case OBJECT_STATUS_TYPE_COMMON_SEQUENTIAL:
    default:
        {
            m_pOperationTargetStatus = GetStationStatus(nn::pia::session::Session::GetInstance()->GetLocalStationId());
        }
    }
    PIA_SAMPLE_ASSERT(m_pOperationTargetStatus != NULL);
    return;
}

void StatusManager::AtomicLockOperation()
{
    if (m_pOperationTargetStatus->type != OBJECT_STATUS_TYPE_TEAM_ATOMIC &&
        m_pOperationTargetStatus->type != OBJECT_STATUS_TYPE_COMMON_ATOMIC)
    {
        return;
    }

    nn::Result result;
    nn::pia::clone::AtomicSharingClone::LockStatus lockStatus = m_pOperationTargetStatus->pObject->GetLockStatus();
    switch(lockStatus)
    {
    case nn::pia::clone::AtomicSharingClone::LOCK_STATUS_LOCKED:
    case nn::pia::clone::AtomicSharingClone::LOCK_STATUS_TRY_LOCK:
        {
            result = m_pOperationTargetStatus->pObject->Unlock();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Unlock failed.\n");
            }
        }
        break;
    case nn::pia::clone::AtomicSharingClone::LOCK_STATUS_UNLOCKED:
        {
            result = m_pOperationTargetStatus->pObject->TryLock();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("TryLock failed.\n");
            }
        }
        break;
    }
}
