﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     TeamConstitutionManager.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include "TeamConstitutionManager.h"
#include "Shared_Clone_Setup.h"

TeamConstitutionManager g_TeamConstitutionManager;

namespace
{
u8 GetTeamType(nn::pia::StationId stationId)
{
    nn::pia::transport::Station* pStation = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
    if (pStation == NULL)
    {
        return SAMPLE_TEAM_NUM_IN_GAME;
    }
    nn::pia::transport::Station::IdentificationToken token;
    nn::Result result = pStation->GetIdentificationToken(&token);
    if (result.IsFailure())
    {
        return SAMPLE_TEAM_NUM_IN_GAME;
    }
    SampleTokenData tokenData;
    memcpy(&tokenData, token.data, sizeof(tokenData));
    return tokenData.teamType;
}
}

// Reverses the <tt>TeamConstitution</tt>bytes so it can be serialized.
template <>
void nn::pia::clone::ReverseSerializePolicy<TeamConstitutionManager::TeamConstitution>::Reverse(TeamConstitutionManager::TeamConstitution* pTo, const TeamConstitutionManager::TeamConstitution* cpFrom)
{
    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        ReverseValue(&(pTo->sessionIdTable[i]), &(cpFrom->sessionIdTable[i]));
        for (int j = 0; j < SAMPLE_TEAM_MEMBER_MAX; ++j)
        {
            // Stops if the alignment is wrong, so do not reference the values pTo->stationIdTable[i][j] and cpFrom->stationIdTable[i][j], or directly assign values.
            // 
            ReverseValue(&(pTo->stationIdTable[i][j]), &(cpFrom->stationIdTable[i][j]));
        }
    }
}

TeamConstitutionManager::TeamConstitutionManager()
{
    m_IsNeedSend = false;
    m_LatestInfoClock = nn::pia::clone::INVALID_CLOCK;
    m_Clone.RegisterElement(&m_ElementR, 0x01);
}


void TeamConstitutionManager::Setup()
{
    nn::Result result = Clone_GetCloneProtocol()->RegisterAtomicSharingClone(&m_Clone, CLONE_ID_OFFSET_TEAM_CONSTITUTION_MANAGER + 0x0);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    NN_PIA_DUMMY_PARAM_IN_RELEASE(result);
}

void TeamConstitutionManager::Reset()
{
    m_TeamConstitution.Clear();
    m_IsNeedSend = false;
    m_LatestInfoClock = nn::pia::clone::INVALID_CLOCK;
}

void TeamConstitutionManager::Update()
{
    if (Clone_GetCloneProtocol()->GetState() != nn::pia::clone::CloneProtocol::STATE_ACTIVE)
    {
        return;
    }

    nn::Result result;
    if (nn::pia::session::Session::GetInstance()->IsJointSessionHost() || (IsInSession() && nn::pia::session::Session::GetInstance()->IsHost()))
    {
        // Sends if the information on the composition of the team has been updated.
        if (m_IsNeedSend)
        {
            if (m_Clone.GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_LOCKED)
            {
                // Sends if all members are synchronized.
                nn::pia::transport::StationManager::Iterator itEnd = nn::pia::transport::StationManager::GetInstance()->End();
                for (nn::pia::transport::StationManager::Iterator it = nn::pia::transport::StationManager::GetInstance()->Begin(); it != itEnd; ++it)
                {
                    if (!Clone_GetCloneProtocol()->IsActiveStation(it->GetStationId()))
                    {
                        return;
                    }
                }
                if (!m_Clone.IsConnectedWithAll())
                {
                    return;
                }

                m_TeamConstitution.Trace();
                PIA_SAMPLE_ASSERT(m_ElementR.IsReadyToSetValue());
                result = m_ElementR.SetValue(m_TeamConstitution);
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Send Team Constitution succeeded.\n");
                    m_IsNeedSend = false;
                }
                else
                {
                    PIA_SAMPLE_LOG("Send Team Constitution failed.\n");
                    PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidTiming());
                }
            }
            else if (m_Clone.GetLockStatus() == nn::pia::clone::AtomicSharingClone::LOCK_STATUS_UNLOCKED)
            {
                result = m_Clone.TryLock();
                PIA_SAMPLE_ASSERT(result.IsSuccess());
            }
        }
    }
    else
    {
        // Updates the team information if new information on the composition of the team has been received.
        if (m_ElementR.IsValidValue())
        {
            nn::pia::clone::ClockValue clock = m_ElementR.GetClock();
            if (clock != m_LatestInfoClock)
            {
                m_LatestInfoClock = clock;
                m_TeamConstitution = m_ElementR.GetValue();
                PIA_SAMPLE_LOG("Receive Team Constitution\n");
                m_TeamConstitution.Trace();
                return;
            }
        }
    }
    return;
}

bool TeamConstitutionManager::JoinEvent(nn::pia::StationId stationId)
{
    PIA_SAMPLE_LOG("JoinEvent called GameState:%d\n", g_GameState);
    if (IsInSession() || IsSessionMigratingToSession() || IsSessionMigratingToJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsHost())
        {
            return true;
        }
    }
    else if (IsInJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost())
        {
            return true;
        }
    }
    else
    {
        PIA_SAMPLE_ASSERT(false);
    }

    bool isSuccess = AssignTeamMember(stationId);
    if (!isSuccess)
    {
        // A different session has joined for the same team.
        return false;
    }

    m_IsNeedSend = true;
    return true;
}

void TeamConstitutionManager::LeaveEvent(nn::pia::StationId stationId)
{
    PIA_SAMPLE_LOG("LeaveEvent called GameState:%d\n", g_GameState);
    if (IsInSession() || IsSessionMigratingToSession() || IsSessionMigratingToJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsHost())
        {
            return;
        }
    }
    else if (IsInJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost())
        {
            return;
        }
    }
    else
    {
        PIA_SAMPLE_ASSERT(false);
    }

    bool isSuccess = CancelTeamMember(stationId);
    PIA_SAMPLE_ASSERT(isSuccess);
    NN_PIA_DUMMY_PARAM_IN_RELEASE(isSuccess);

    m_IsNeedSend = true;
}

bool TeamConstitutionManager::SessionHostChangedEvent()
{
    PIA_SAMPLE_LOG("SessionHostChangedEvent called GameState:%d\n", g_GameState);
    if (IsInSession() || IsSessionMigratingToSession() || IsSessionMigratingToJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsHost())
        {
            return true;
        }
    }
    else if (IsInJointSession())
    {
        if (!nn::pia::session::Session::GetInstance()->IsJointSessionHost())
        {
            return true;
        }
    }
    else
    {
        PIA_SAMPLE_ASSERT(false);
    }

    // Skip while the session is migrating.
    if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_MIGRATING_SESSION)
    {
        return true;
    }

    // The new hosts updates the team roster information and sends it to the clients.
    bool isSuccess = true;

    // The local system deletes any stations that have left the team but are still included in the team roster information it has.
    // Also deleted when there is a team error, as it is possible to re-join on a different team.
    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        for (int j = 0; j < SAMPLE_TEAM_MEMBER_MAX; ++j)
        {
            nn::pia::StationId stationId = GetStationId(i, j);
            if (stationId != nn::pia::STATION_ID_INVALID)
            {
                if (nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(stationId) == false)
                {
                    isSuccess &= CancelTeamMember(stationId);
                }
                else if (GetTeamType(stationId) != i)
                {
                    PIA_SAMPLE_LOG("Different team type: %lld (%d != %d)\n", stationId.GetValue(), GetTeamType(stationId), i);
                    isSuccess &= CancelTeamMember(stationId);
                }
            }
        }
    }
    PIA_SAMPLE_ASSERT(isSuccess);

    // The local system adds any stations that are not included in the team roster information it has.
    nn::pia::session::Session::StationIdIterator itEnd = nn::pia::session::Session::GetInstance()->End();
    for (nn::pia::session::Session::StationIdIterator it = nn::pia::session::Session::GetInstance()->Begin(); it != itEnd; ++it)
    {
        if (nn::pia::transport::StationManager::GetInstance()->GetStation(*it) == NULL)
        {
            // Already left.
            continue;
        }
        if (g_TeamConstitutionManager.GetTeamIndex(*it) == -1)
        {
            // Add because not included in the team roster.
            isSuccess &= AssignTeamMember(*it);
        }
    }
    if (!isSuccess)
    {
        // A different session has joined for the same team.
        return false;
    }
    m_IsNeedSend = true;
    return true;
}

bool TeamConstitutionManager::AssignTeamMember(nn::pia::StationId stationId)
{
    u8 teamType = GetTeamType(stationId);
    PIA_SAMPLE_ASSERT(teamType < SAMPLE_TEAM_NUM_IN_GAME);

    u32 sessionId = nn::pia::session::Session::GetInstance()->GetSessionId(stationId);
    if (m_TeamConstitution.sessionIdTable[teamType] != nn::pia::session::INVALID_SESSION_ID && m_TeamConstitution.sessionIdTable[teamType] != sessionId)
    {
        // A different session has joined for the same team.
        PIA_SAMPLE_LOG("different sessionId:0x%08x (teamType:%d sessionId:0x%08x)\n", sessionId, teamType, m_TeamConstitution.sessionIdTable[teamType]);
        return false;
    }
    m_TeamConstitution.sessionIdTable[teamType] = sessionId;

    for (int i = 0; i < SAMPLE_TEAM_MEMBER_MAX; ++i)
    {
        if (m_TeamConstitution.stationIdTable[teamType][i] == nn::pia::STATION_ID_INVALID.GetValue())
        {
            m_TeamConstitution.stationIdTable[teamType][i] = stationId.GetValue();
            return true;
        }
    }
    PIA_SAMPLE_ASSERT(false);
    return false;
}

bool TeamConstitutionManager::CancelTeamMember(nn::pia::StationId stationId)
{
    if (stationId == nn::pia::STATION_ID_INVALID)
    {
        return false;
    }

    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        for (int j = 0; j < SAMPLE_TEAM_MEMBER_MAX; ++j)
        {
            if (m_TeamConstitution.stationIdTable[i][j] == stationId.GetValue())
            {
                m_TeamConstitution.stationIdTable[i][j] = nn::pia::STATION_ID_INVALID.GetValue();

                // Check for other members on this team.
                bool isEmpty = true;
                for (int k = 0; k < SAMPLE_TEAM_MEMBER_MAX; ++k)
                {
                    if (m_TeamConstitution.stationIdTable[i][k] != nn::pia::STATION_ID_INVALID.GetValue())
                    {
                        isEmpty = false;
                        break;
                    }
                }
                if (isEmpty)
                {
                    // No one was there, so update sessionIdTable.
                    m_TeamConstitution.sessionIdTable[i] = nn::pia::session::INVALID_SESSION_ID;
                }
                return true;
            }
        }
    }

    // Returns success because, upon host migration, an exit event for any station that is no longer on the team may occur after the new host updates the team roster.
    return true;
}

nn::pia::StationId TeamConstitutionManager::GetStationId(int teamIndex, int memberIndex) const
{
    if (teamIndex < 0 || teamIndex >= SAMPLE_TEAM_NUM_IN_GAME)
    {
        return nn::pia::STATION_ID_INVALID;
    }

    if (memberIndex < 0 || memberIndex >= SAMPLE_TEAM_MEMBER_MAX)
    {
        return nn::pia::STATION_ID_INVALID;
    }

    return nn::pia::StationId(m_TeamConstitution.stationIdTable[teamIndex][memberIndex]);
}

int TeamConstitutionManager::GetTeamIndex(nn::pia::StationId stationId) const
{
    if (stationId == nn::pia::STATION_ID_INVALID)
    {
        return -1;
    }

    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        for (int j = 0; j < SAMPLE_TEAM_MEMBER_MAX; ++j)
        {
            if (m_TeamConstitution.stationIdTable[i][j] == stationId.GetValue())
            {
                return i;
            }
        }
    }
    return -1;
}

int TeamConstitutionManager::GetMemberIndex(nn::pia::StationId stationId) const
{
    if (stationId == nn::pia::STATION_ID_INVALID)
    {
        return -1;
    }

    for (int i = 0; i < SAMPLE_TEAM_NUM_IN_GAME; ++i)
    {
        for (int j = 0; j < SAMPLE_TEAM_MEMBER_MAX; ++j)
        {
            if (m_TeamConstitution.stationIdTable[i][j] == stationId.GetValue())
            {
                return j;
            }
        }
    }
    return -1;
}

u32 TeamConstitutionManager::GetTeamSessionId(int teamIndex) const
{
    if (teamIndex < 0 || teamIndex >= SAMPLE_TEAM_NUM_IN_GAME)
    {
        return nn::pia::session::INVALID_SESSION_ID;
    }
    return m_TeamConstitution.sessionIdTable[teamIndex];
}

u32 TeamConstitutionManager::GetTeamMemberNum(int teamIndex) const
{
    if (teamIndex < 0 || teamIndex >= SAMPLE_TEAM_NUM_IN_GAME)
    {
        return 0;
    }
    int num = 0;
    for (int i = 0; i < SAMPLE_TEAM_MEMBER_MAX; ++i)
    {
        if (GetStationId(teamIndex, i) != nn::pia::STATION_ID_INVALID)
        {
            ++num;
        }
    }
    return num;
}

u32 TeamConstitutionManager::GetTeamNum() const
{
    u32 teamNum = 0;
    for (int i=0;i < SAMPLE_TEAM_NUM_IN_GAME;i++)
    {
        if (m_TeamConstitution.sessionIdTable[i] != nn::pia::session::INVALID_SESSION_ID)
        {
            for (int j=0;j < SAMPLE_TEAM_MEMBER_MAX;j++)
            {
                if (m_TeamConstitution.stationIdTable[i][j] != nn::pia::STATION_ID_INVALID.GetValue())
                {
                    ++teamNum;
                    break;
                }
            }
        }
    }
    return teamNum;
}
