﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/version.h>
#include <pia/common.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "Shared_Common_Setup.h"

#include "SampleConfig.h"
#include "TeamConstitutionManager.h"
#include "ObjectManager.h"

GAME_STATE g_GameState = GAME_STATE_NONE;
bool g_IsOpened = false;
JOINT_SESSION_TYPE g_JointSessionType = JOINT_SESSION_TYPE_NONE;
const wchar_t* USER_PASSWORD[] = {
    L"alpha",
    L"bravo",
    L"charlie",
    L"delta",
    L"joint"
};
u8 g_UserPasswordIndexForJoint = 0;

namespace
{
/* ---< Prepare memory to allocate to Pia. >--- */
// The memory region must be 4-byte aligned.
// In this sample, sufficiently large values are used.
PIA_ATTRIBUTE_ALIGN(4) bit8 s_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA];
/* ---< End of memory preparations. >--- */

// When set to <tt>true</tt>, suppresses the impact on processing speed when several trace flags of the Pia library have been enabled.
// However, the order relationship with logs of other libraries (NEX, and so on) may be switched.
const bool SWITCH_USE_CACHED_PRINT = true;

// Variables for saving the Pia library version information character string.
char s_SamplePiaVersionInfoString[SCREEN_WIDTH / FONT_SIZE] = {};

#if PIA_TEST_CODE
bool s_IsEnableHostMigration = true;
u8 s_MaxPlayerNum = SAMPLE_GAME_PLAYER_MAX;
#endif

}  // end of anonymous namespace

// Function that gets the Pia library version information string.
const char* GetPiaVersionString()
{
    return s_SamplePiaVersionInfoString;
}

#if PIA_TEST_CODE
bool IsEnableHostMigration()
{
    return s_IsEnableHostMigration;
}

void IsEnableHostMigration(bool b)
{
    s_IsEnableHostMigration = b;
}

void SetMaxPlayerNum(u8 num)
{
    s_MaxPlayerNum = num;
}

u8 GetMaxPlayerNum()
{
    return s_MaxPlayerNum;
}
#endif

#if NN_PIA_CTR
u32 GetInternetLinkLevel();
#endif

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel()
{
#if NN_PIA_CTR
    // Get the signal strength according to the communication method, and return a value from 0 to 3.
    return GetInternetLinkLevel();
#elif NN_PIA_WIN
    // Always 0.
    return 0;
#elif NN_PIA_CAFE
    // There is no way to get the signal strength, so it is always 0.
    return 0;
#else
#error  "Illegal platform."
#endif
}

PIA_SAMPLE_MAIN_FUNC()
{
    PIA_SAMPLE_LOG("Sample program Start.\n");

    // Creates the Pia library version information string.
    ::std::snprintf(s_SamplePiaVersionInfoString, SCREEN_WIDTH / FONT_SIZE, "Pia %d.%d.%d r%d", PIA_VERSION_MAJOR, PIA_VERSION_MINOR, PIA_VERSION_MICRO, PIA_VERSION_REVISION);

    nn::Result result;

    // Initialize the render framework for the Pia sample program.
    nn::pia::sampleframework::SampleFramework::Setting setting;
    setting.bUpperScreenOnly = true;
    nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAME_PER_SECOND);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.0f, 0.0f, 0.2f, 1.0f));
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_PREPARE_INET, false);
#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(INDEX_ENABLE_HOST_MIGRATION, IsEnableHostMigration() ? 1 : 0);
#endif

    while (1)
    {
        SampleSceneNo nextSceneNo = SCENE_PREPARE_INET;

        /* ---< Setup of the features used in the PiaCommon module >--- */
        u64 traceFlags = 0;
        if (SWITCH_USE_CACHED_PRINT)
        {
            traceFlags |=
                nn::pia::common::TRACE_FLAG_COMMON |
                nn::pia::common::TRACE_FLAG_LOCAL |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_SESSION |
                nn::pia::common::TRACE_FLAG_INET |
                nn::pia::common::TRACE_FLAG_CLONE |
                nn::pia::common::TRACE_FLAG_SESSION_JOINT |
#if PIA_TEST_CODE
                nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
                nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
                nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
                nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
                nn::pia::common::TRACE_FLAG_INET_INIT |
                nn::pia::common::TRACE_FLAG_INET_NAT |
                nn::pia::common::TRACE_FLAG_INET_NAT_DETAIL |
                nn::pia::common::TRACE_FLAG_INET_NAT_PROBE |
                nn::pia::common::TRACE_FLAG_INET_SOCKET |
#endif
#if 0
                nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_SEND |
                nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_RECEIVE |
#endif
                nn::pia::common::TRACE_FLAG_ALWAYS;
        }
        else
        {
            traceFlags |= nn::pia::common::TRACE_FLAG_ALWAYS;
        }
        Common_InitializePiaCommon(s_PiaMemoryBlock, MEMORY_BLOCK_SIZE_PIA, traceFlags, SWITCH_USE_CACHED_PRINT);

        AppTrace::GetInstance()->SetFlag(SAMPLE_TRACE_ALWAYS);

        // Specify error handling for the Pia library.
        ErrorHandling::Initialize();
        /* ---<Setup of the features used in the end of the PiaCommon module. >--- */

        bool isExit = false;

        /* Game loop */
        while (isExit == false)
        {
            switch (nextSceneNo)
            {
            case SCENE_PREPARE_INET:
            {
                if (InitializePiaModules())
                {
                    g_ObjectManager.RegisterAll();
                    g_TeamConstitutionManager.Setup();
                    nextSceneNo = SCENE_SELECT_TEAM;
                }
                else
                {
                    nextSceneNo = SCENE_NETWORK_ERROR;
                }
            }
            break;
            case SCENE_SELECT_TEAM:
            {
                nextSceneNo = SceneSelectTeam();
            }
            break;
            case SCENE_MATCH_MAKE_TEAM:
            {
                nextSceneNo = SceneMatchmake();
            }
            break;
            case SCENE_COMMUNICATE_TEAM:
            {
                nextSceneNo = SceneCommunicationTeam();
            }
            break;
            case SCENE_CHANGING_SESSION_PREPROCESSING:
            {
                nextSceneNo = SceneChangingSessionPreprocessing();
            }
            break;
            case SCENE_CHANGING_SESSION:
            {
                nextSceneNo = SceneChangingSession();
            }
            break;
            case SCENE_CHANGING_SESSION_POSTPROCESSING:
            {
                nextSceneNo = SceneChangingSessionPostprocessing();
            }
            break;
            case SCENE_COMMUNICATE_GAME:
            {
                nextSceneNo = SceneCommunicationGame();
            }
            break;
            case SCENE_NETWORK_ERROR:
            {
                nextSceneNo = SceneNetworkError();
            }
            break;
            case SCENE_RESET:
            {
                FinalizePiaModules();
                isExit = true;
            }
            break;
            }
        }

        /* ---< Finalize PiaCommon. >--- */
        ErrorHandling::Finalize();
        Common_FinalizePiaCommon();
        /* ---< End of PiaCommon-related finalization. >--- */
    }
    //Control does not arrive here.
    //nn::pia::sampleframework::SampleFramework::DestroyInstance();
    //PIA_SAMPLE_LOG("Sample program end.\n");

}  // end of nnMain

// When DDL files are not used with NEX, a definition of the empty InitDOClasses function is required.
namespace nn
{
namespace nex
{
void InitDOClasses()
{
}
}
}
