﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Object.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "Object.h"
#include <pia/session.h>
#include "Shared_Clone_Setup.h"

namespace
{

const f32 VEL_MAX = 6.f;
const f32 DEF_ACC = 0.5f;


const f32 DECREASE_VEL = 0.15f;

const f32 SQUARE_SIZE = 10.f;
const f32 DISP_WIDTH = 400.f - SQUARE_SIZE;
const f32 DISP_HEIGHT = 240.f - SQUARE_SIZE;


EventStatusArg   cEventStatusArg;


s32 cEventStatusIndex;

enum EEventStatusType{
    EST_UNRELIABLE,
    EST_LATENCY,
    EST_CHASE_TYPE,
    EST_CHASE_RATIO,
    EST_VEL_SCALE,
    EST_STOP_THRESHOLD,
    EST_RECKONING_THRESHOLD,

#if DELAY_TEST_ON
    EST_DELAY_CLOCK,
#endif
    EST_MAX
};

/// Method for Interpolating Position
enum EEventChaseType{
    ECT_NONE,  //< Do not interpolate.
    ECT_CHASE, //< Chase using ratio.
    ECT_MAX
};


}  // end of anonymous namespace

// Clamp. orgValue is a value between minValue and maxValue.
template <typename T>
static T Clamp(T orgValue, T minValue, T maxValue)
{
    if (orgValue < minValue)
    {
        return minValue;
    }
    if (maxValue < orgValue)
    {
        return maxValue;
    }
    return orgValue;
}

static f32 getMaxValue(){

    return VEL_MAX * cEventStatusArg.mVelScale;
}

static f32 getDefAcc(){

    return DEF_ACC * cEventStatusArg.mVelScale;
}

static f32 getDecreaseVel(){

    return ( 1.0f + DECREASE_VEL * cEventStatusArg.mVelScale * cEventStatusArg.mVelScale );
}

/// Copied from sead.
static bool chaseUtl ( f32* current, f32 target, const f32 step )
{

    // The code is shorter if you implement something similar to vector to calculate diff, but then one run through the function involves a lot of computations and comparisons, so we implement it this way.
    const f32 before = *current;
    f32 after;
    if (before < target) {
        // Approaches the positive direction.
        after = before + step;
        if (target <= after ||  // If target has been reached.
            after < before)     // If the after calculation result has overflowed.
        {
            *current = target;
            return true;
        } else {
            *current = after;
            return false;
        }
    } else if (before > target) {
        // Approaches the negative direction.
        after = before - step;
        if (after <= target ||  // If target has been reached.
            before < after)     // If the after calculation result has underflowed.
        {
            *current = target;
            return true;
        } else {
            *current = after;
            return false;
        }
    } else {
        // current and target are the same from the start.
        return true;
    }
}

static bool vecChaseDecrease(nn::pia::reckoning::Vector3f* pVec,
               const nn::pia::reckoning::Vector3f& rTarget,
               f32 ratio,
               f32 min,
               f32 max/* = std::numeric_limits<f32>::max()*/)
{

    PIA_ASSERT( 0.0f <= ratio );
    PIA_ASSERT( 1.0f >= ratio );
    PIA_ASSERT(min >= 0);
    PIA_ASSERT(max >= 0);
    PIA_ASSERT(min <= max);

    const f32   stepX = fabs( pVec->x - rTarget.x ) * ratio; //sqrt( pVec->SquaredDistance( rTarget )  ) * ratio;

    const f32   stepY = fabs( pVec->y - rTarget.y ) * ratio; 

    const f32   stepZ = fabs( pVec->z - rTarget.z ) * ratio; 

    bool blX = chaseUtl( &(pVec->x) , rTarget.x , Clamp( stepX , min, max ) );
    bool blY = chaseUtl( &(pVec->y) , rTarget.y , Clamp( stepY , min, max ) );
    bool blZ = chaseUtl( &(pVec->z) , rTarget.z , Clamp( stepZ , min,  max ) );
    return blX && blY && blZ ;
}


// Adjust the initial value of <tt>Strategy</tt> so movements are smooth.
// You can see how jerky movements are when set to <tt>10.f</tt>.
//nn::pia::reckoning::Simple3dReckoningStrategy Object::s_ReckoningStrategy(1.f);

nn::pia::reckoning::Simple3dReckoningStrategy Object::s_ReckoningStrategy(1.f);



////////////////////////////////////////////////////////////////////


Operation Object::s_Operation;

// Reverse the Status bytes so it can be serialized.
template <>
void nn::pia::clone::ReverseSerializePolicy<Status>::Reverse(Status* pTo, const Status* cpFrom)
{
    ReverseValue(&pTo->m_X, &cpFrom->m_X);
    ReverseValue(&pTo->m_Y, &cpFrom->m_Y);
}

void Operation::Update(bool isHold)
{
    nn::pia::reckoning::Vector3f::Mul(&m_Acc, getDefAcc() );
    nn::pia::reckoning::Vector3f::Add(&m_Vel, m_Acc);
    nn::pia::reckoning::Vector3f::Add(&m_Pos, m_Vel);

    if (!isHold)
    {
        m_Acc.MakeZero();
        m_Vel /= getDecreaseVel();
    }
}

void Operation::ClampValues()
{
    m_Vel.x = Clamp(m_Vel.x, -getMaxValue() , getMaxValue() );
    m_Vel.y = Clamp(m_Vel.y, -getMaxValue() , getMaxValue() );
    m_Pos.x = Clamp(m_Pos.x, 0.f, DISP_WIDTH);
    m_Pos.y = Clamp(m_Pos.y, 0.f, DISP_HEIGHT);
}

bool Operation::IsStop()
{

    f32 distance = m_Pos.SquaredDistance(m_OldPos);
    f32 thresholdStop = m_ThresholdStop * m_ThresholdStop;

    return  ( distance <= thresholdStop );
}

void Operation::SetThresholdStop(f32 th)
{
    if (th < 0.f)
    {
        m_ThresholdStop = 0.f;
    }
    else
    {
        m_ThresholdStop = th;
    }
}

void Object::RegisterElementCore(nn::pia::clone::CloneBase* pClone)
{
    m_ElementR.SetReckoningStrategy(&s_ReckoningStrategy);
    pClone->RegisterElement(&m_ElementR, 0x01);
    pClone->RegisterElement(&m_ElementU, 0x02);


    cEventStatusArg.mIsUnreliable = false;
    cEventStatusArg.mLatency = 400;
    cEventStatusArg.mChaseType = ECT_CHASE;
    cEventStatusArg.mChaseRatio = 0.06f;
    cEventStatusArg.mVelScale = 1.0f;
    cEventStatusArg.mStopThreshold = 0.1f;
    cEventStatusArg.mReckoningThreshold = 1.0f;

#if DELAY_TEST_ON
    cEventStatusArg.mDelayClock = 0.f;
#endif
    cEventStatusIndex = 0;


}

bool Object::CleanupCore(nn::pia::clone::CloneBase* pClone)
{
    if (pClone->IsRegisteredWithProtocol())
    {
        Unregister();
    }
    else
    {
        nn::Result r;
        // Only call UnregisterElement if already unregistered from Clone.
        if (m_ElementR.IsRegisteredWithCloneBase())
        {
            PIA_SAMPLE_PRINT("Clone Element Id:%x - Unregister\n", m_ElementR.GetId());
            r = pClone->UnregisterElement(&m_ElementR);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
        }
        if (m_ElementU.IsRegisteredWithCloneBase())
        {
            PIA_SAMPLE_PRINT("Clone Element Id:%x - Unregister\n", m_ElementU.GetId());
            r = pClone->UnregisterElement(&m_ElementU);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
        }
    }
    bool ret = !m_ElementR.IsRegisteredWithCloneBase() &&
               !m_ElementU.IsRegisteredWithCloneBase() ;
    if (ret)
    {
        PIA_SAMPLE_PRINT("Clone Type:%s - Cleanup Process is Succeeded.\n", GetCloneTypeText());
    }
    return ret;
}

void Object::ResetReckoningStrategy()
{
    nn::Result r = m_ElementR.ResetReckoningStrategy();
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    NN_PIA_DUMMY_PARAM(r);
}

bool Object::IsLocalStation()
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    if (pSession && m_StationIndex == pSession->GetLocalStationIndex())
    {
        return true;
    }
    return false;
}

void Object::UpdateElement()
{
    if (!cEventStatusArg.mIsUnreliable  && m_ElementR.IsReadyToSetValue())
    {

        bool isStop = s_Operation.IsStop();

        m_ElementR.SetValue(s_Operation.m_Pos, &isStop);

        s_Operation.m_OldPos = s_Operation.m_Pos;
    }
    if ( cEventStatusArg.mIsUnreliable && m_ElementU.IsReadyToSetValue())
    {
        Status v;
        v.m_X = s_Operation.m_Pos.x;
        v.m_Y = s_Operation.m_Pos.y;
        m_ElementU.SetValue(v);
    }



}

const char* Object::GetCurrentMode()
{
    return cEventStatusArg.mIsUnreliable ? "Unr" : "Rck";
}

bool Object::isActive(){

    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    return  (pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE);
}


SendObject::SendObject()
{
    RegisterElementCore(&m_Clone);
}

void SendObject::Register()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();
    pCloneProtocol->RegisterSendClone(&m_Clone, m_CloneId);
}

void SendObject::Unregister()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();
    pCloneProtocol->UnregisterSendClone(m_CloneId);
}


void SendObject::Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();


    // Control local station's square.
    bool isHold = false;

    f32 accX, accY;
    pSampleFramework->GetAnalog(&accX, &accY);
    const f32 min = 0.1f;
    if (accX >= min || accX <= -min)
    {
        s_Operation.m_Acc.x = accX;
        isHold = true;
    }
    if (accY >= min || accY <= -min)
    {
        s_Operation.m_Acc.y = -accY;
        isHold = true;
    }
    s_Operation.Update(isHold);
    s_Operation.ClampValues();
}

void SendObject::Update()
{
    UpdateElement();
}

bool SendObject::Cleanup()
{
    return CleanupCore(&m_Clone);
}

int SendObject::Print(int y, const nn::pia::sampleframework::Color& cColor)
{
    PrintValue(y, cColor);
    return y;
}

void SendObject::PrintValue(int y, const nn::pia::sampleframework::Color& cColor)
{
    // Use local coordinates to display square controlled by local station.
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    pSampleFramework->SetFontColor(cColor);
    nn::pia::reckoning::Vector3f* pos = &s_Operation.m_Pos;

    pSampleFramework->DrawText( 40, y, "%6.2f, %6.2f", pos->x, pos->y);
    pSampleFramework->DrawRect(pos->x, pos->y, static_cast<s32>(SQUARE_SIZE), static_cast<s32>(SQUARE_SIZE), cColor);
}

ReceiveObject::ReceiveObject()
{
    RegisterElementCore(&m_Clone);
    mDrawPos = s_Operation.m_Pos;
}

void ReceiveObject::Register()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();
    pCloneProtocol->RegisterReceiveClone(&m_Clone, m_StationIndex, m_CloneId);


}

void ReceiveObject::Unregister()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();
    pCloneProtocol->UnregisterReceiveClone(m_StationIndex, m_CloneId);
}

void ReceiveObject::Operate()
{
}

void ReceiveObject::Update()
{
    UpdateElement();
}

bool ReceiveObject::Cleanup()
{
    return CleanupCore(&m_Clone);
}

int ReceiveObject::Print(int y, const nn::pia::sampleframework::Color& cColor)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    pSampleFramework->SetFontColor(cColor);
    if (m_Clone.IsRegisteredWithProtocol())
    {
        pSampleFramework->DrawText( 0, y, "R %1d ", m_StationIndex );

        // Show RTT.
        nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
        nn::pia::StationId stationId;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationId(&stationId, m_Clone.GetSendStationIndex());
        if (result.IsSuccess())
        {
            nn::pia::transport::Station* pStation = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
            if (pStation && pSession->GetLocalStationIndex() != m_Clone.GetSendStationIndex())
            {
                s32 rtt = pStation->GetRtt( 3 );
                pSampleFramework->DrawText(170, y, "RTT: %d", rtt);
            }
        }
    }
    PrintValue(y, cColor);
    y += 15;
    return y;
}

void ReceiveObject::PrintValue(int y, const nn::pia::sampleframework::Color& cColor)
{
    if (IsLocalStation())
    {
        return;
    }
    // Show coordinates of squares received from remote stations.
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::reckoning::Vector3f v;
    bool isValid = false;

#if DELAY_TEST_ON
    s_ReckoningStrategy.setGetTiming( cEventStatusArg.mDelayClock );
#endif

    if (!cEventStatusArg.mIsUnreliable && m_ElementR.IsValidValue())
    {

        v = m_ElementR.GetValue();
        if( cEventStatusArg.mChaseType == ECT_NONE ){
            mDrawPos = v;
        }else if( cEventStatusArg.mChaseType == ECT_CHASE ){
            vecChaseDecrease( &mDrawPos , v , cEventStatusArg.mChaseRatio , 0.1f , 10.0f );
        }

        isValid = true;
    }
    if (m_ElementU.IsValidValue())
    {
        Status s = m_ElementU.GetValue();

        if (cEventStatusArg.mIsUnreliable)
        {
            v.x = s.m_X;
            v.y = s.m_Y;
            mDrawPos.x  = v.x;
            mDrawPos.y  = v.y;
            isValid = true;
        }
    }
    if (isValid)
    {
        pSampleFramework->DrawText(40, y, "%6.2f, %6.2f", v.x, v.y);
        pSampleFramework->DrawRect(mDrawPos.x, mDrawPos.y, static_cast<s32>(SQUARE_SIZE), static_cast<s32>(SQUARE_SIZE), cColor);
    }
#if DELAY_TEST_ON
    s_ReckoningStrategy.offGetTiming();
#endif

}


SequentialObject::SequentialObject()
{

    m_Clone.RegisterElement(&m_ElementREX, 0x4 );
}

void SequentialObject::Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    NN_PIA_DUMMY_PARAM(pSampleFramework);


    if (m_ElementREX.IsReadyToSetValue())
    {

    /// Altered items are skewed once L is pressed.
    if( pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP )){

        cEventStatusIndex = ( cEventStatusIndex - 1 + EST_MAX ) % EST_MAX;
    }else if( pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN )){

        cEventStatusIndex = ( cEventStatusIndex + 1 ) % EST_MAX;
    }



    // Switch EventStatusArg.
    s32 count = 0;
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT ))
    {
        count ++;
    }else if( pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT ) ){
        count --;

    }

    if( count != 0 ){


            switch( cEventStatusIndex ){

                case EST_UNRELIABLE:
                    {
                        cEventStatusArg.mIsUnreliable = !cEventStatusArg.mIsUnreliable;
                    }

                break;
                case EST_LATENCY:
                    {
                        cEventStatusArg.mLatency += ( 50 * count);

                        cEventStatusArg.mLatency = Clamp( static_cast<s32>(cEventStatusArg.mLatency ), 0 , 600 );

                    }
                break;
                case EST_CHASE_TYPE:
                        cEventStatusArg.mChaseType +=( count + ECT_MAX);
                        cEventStatusArg.mChaseType %= ECT_MAX;

                break;


                case EST_CHASE_RATIO:
                        cEventStatusArg.mChaseRatio += 0.01f * count;

                        cEventStatusArg.mChaseRatio = Clamp( static_cast<f32>(cEventStatusArg.mChaseRatio ), 0.02f , 0.4f );

                break;

                case EST_VEL_SCALE:
                        cEventStatusArg.mVelScale += 0.1f * count;

                        cEventStatusArg.mVelScale = Clamp( static_cast<f32>(cEventStatusArg.mVelScale ), 0.7f , 1.5f );
                break;

                case EST_STOP_THRESHOLD:
                        cEventStatusArg.mStopThreshold += 0.05f * count;

                        cEventStatusArg.mStopThreshold = Clamp( static_cast<f32>(cEventStatusArg.mStopThreshold ), 0.0f , 1.0f );

                break;

                case EST_RECKONING_THRESHOLD:
                        cEventStatusArg.mReckoningThreshold += 0.1f * count;
                        cEventStatusArg.mReckoningThreshold = Clamp( static_cast<f32>(cEventStatusArg.mReckoningThreshold ), 0.0f , 3.0f );

                break;
#if DELAY_TEST_ON
                case EST_DELAY_CLOCK:
                        cEventStatusArg.mDelayClock += 40.f * count;

                        cEventStatusArg.mDelayClock = Clamp( static_cast<f32>(cEventStatusArg.mDelayClock ), 0.0f , 280.0f );

                break;
#endif
            }

            /// Settings for each separate unit. The settings must also be configured on the peer CAT-DEVs.
            nn::pia::transport::ThreadStreamManager::GetInstance()->SetSendThreadLatencyEmulation( cEventStatusArg.mLatency / 2 , cEventStatusArg.mLatency / 2 );

            nn::pia::transport::ThreadStreamManager::GetInstance()->SetReceiveThreadLatencyEmulation( 0, 0 );

            nn::Result r = m_ElementREX.SetValue( cEventStatusArg );


            if (r.IsSuccess())
            {
//                PIA_SAMPLE_PRINT("Send SequentialObject \n");
            }

    }

    }
}

void SequentialObject::Update()
{
//    UpdateElement();

    if (m_ElementREX.IsValidValue())
    {
        cEventStatusArg = m_ElementREX.GetValue();
        /// Settings for each separate unit. The settings must also be configured on the peers.
        nn::pia::transport::ThreadStreamManager::GetInstance()->SetSendThreadLatencyEmulation( cEventStatusArg.mLatency / 2 , cEventStatusArg.mLatency / 2 );

        nn::pia::transport::ThreadStreamManager::GetInstance()->SetReceiveThreadLatencyEmulation( 0, 0 );


        Object::GetOperation()->SetThresholdStop( cEventStatusArg.mStopThreshold );
        Object::GetReckoningStrategy()->SetThreshold( cEventStatusArg.mReckoningThreshold );
    }

}

bool SequentialObject::Cleanup()
{

    nn::Result r;
    if (m_ElementREX.IsRegisteredWithCloneBase())
    {
        PIA_SAMPLE_PRINT("Clone Element Id:%x - Unregister\n", m_ElementREX.GetId());
        r = m_Clone.UnregisterElement(&m_ElementREX);
        PIA_SAMPLE_ASSERT(r.IsSuccess());
    }

    bool ret =  !m_ElementREX.IsRegisteredWithCloneBase();
    if (ret)
    {
        PIA_SAMPLE_PRINT("Clone Type:%s - Cleanup Process is Succeeded.\n", GetCloneTypeText());
    }
//    return CleanupCore(&m_Clone);
     return ret;
}

void SequentialObject::RegisterClone()
{
    nn::Result result = Clone_GetCloneProtocol()->RegisterSequentialSharingClone(&m_Clone, m_CloneId);
    if (result.IsSuccess())
    {
        PIA_SAMPLE_PRINT("RegisterSequentialSharingClone Success (cloneId:%d)\n", m_CloneId);
    }
    else
    {
        PIA_SAMPLE_PRINT("RegisterSequentialSharingClone Failure (cloneId:%d)\n", m_CloneId);
        PIA_SAMPLE_ASSERT(false);
    }
}

void SequentialObject::UnregisterClone()
{
    nn::Result result = Clone_GetCloneProtocol()->UnregisterSequentialSharingClone(&m_Clone);
    if (result.IsSuccess())
    {
        PIA_SAMPLE_PRINT("UnregisterSequentialSharingClone Success (cloneId:%d)\n", m_CloneId);
    }
    else
    {
        PIA_SAMPLE_PRINT("UnregisterSequentialSharingClone Failure (cloneId:%d)\n", m_CloneId);
        PIA_SAMPLE_ASSERT(false);
    }
}

int SequentialObject::Print(int y, const nn::pia::sampleframework::Color& cColor)
{

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::sampleframework::Color cColorLY = cColor;

    if( !isActive() ){
        cColorLY = nn::pia::sampleframework::Color( 0.4f , 0.4f, 0.4f , 1.0f );
    }

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor( cColorLY );

    const s32 beginX = 250;
    s32 posY = 155;

    pSampleFramework->DrawText(beginX -10 , posY + 10 * cEventStatusIndex  , "* ");

    if( cEventStatusArg.mIsUnreliable ){
        pSampleFramework->DrawText(beginX, posY , "Unreliable ");
    }else{
        pSampleFramework->DrawText(beginX, posY , "Reckoning ");
    }

    posY += 10;
    pSampleFramework->DrawText(beginX, posY , "Latency       %d ", cEventStatusArg.mLatency );

    posY += 10;
    if( cEventStatusArg.mChaseType == ECT_NONE ){
    pSampleFramework->DrawText(beginX, posY , "ChaseType     None");

    }else{
    pSampleFramework->DrawText(beginX, posY , "ChaseType    Chase");

    }
    
    posY += 10;
    pSampleFramework->DrawText(beginX, posY , "ChaseRatio    %0.2f ", cEventStatusArg.mChaseRatio );

    posY += 10;
    pSampleFramework->DrawText(beginX, posY , "VelScale      %1.1f ", cEventStatusArg.mVelScale );

    posY += 10;
    pSampleFramework->DrawText(beginX, posY , "StopThreshold %1.1f ", cEventStatusArg.mStopThreshold );

    posY += 10;
    pSampleFramework->DrawText(beginX, posY , "Threshold     %1.1f ", cEventStatusArg.mReckoningThreshold );

    posY += 10;
#if DELAY_TEST_ON
    pSampleFramework->DrawText(beginX, posY , "DelayClock    %3.0f ", cEventStatusArg.mDelayClock );
#endif

    return y;
}
