﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Object.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/reckoning.h>
#include <pia/clone.h>
#include <pia/transport.h>
#include "SampleLib/SampleLib.h"


#define DELAY_TEST_ON  0

struct EventStatusArg{
    
    bool mIsUnreliable;
    s32 mLatency;

    s32 mChaseType;//< Smooths out the sent data.

    f32 mChaseRatio;
    f32 mVelScale;
    f32 mStopThreshold;
    f32 mReckoningThreshold;
#if DELAY_TEST_ON
    f32 mDelayClock;
#endif
};


class Operation
{
public:
    nn::pia::reckoning::Vector3f m_Pos, m_OldPos, m_Vel, m_Acc;
    f32 m_ThresholdStop;

    Operation()
    {
        m_Pos.MakeZero();
        m_OldPos.MakeZero();
        m_Vel.MakeZero();
        m_Acc.MakeZero();
        m_ThresholdStop = 0.1f;
    }
    void Update(bool isHold);
    void ClampValues();
    bool IsStop();
    f32 GetThresholdStop()
    {
        return m_ThresholdStop;
    }
    void SetThresholdStop(f32 th);
};

struct Status
{
    f32 m_X;
    f32 m_Y;
};

class Object
{
public:
    virtual void Register()
    {
    }
    virtual void Unregister()
    {
    }
    virtual void RegisterElement()
    {
    }
    void RegisterElementCore(nn::pia::clone::CloneBase* pClone);
    virtual void Operate()
    {
    }
    virtual void Update()
    {
    }
    virtual bool Cleanup()
    {
        return true;
    }
    bool CleanupCore(nn::pia::clone::CloneBase* pClone);
    void ResetReckoningStrategy();
    bool IsLocalStation();
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor)
    {
        NN_PIA_DUMMY_PARAM(cColor);
        return y;
    }
    virtual void PrintValue(int, const nn::pia::sampleframework::Color& cColor)
    {
        NN_PIA_DUMMY_PARAM(cColor);
    }

    void SetStationIndex(nn::pia::StationIndex index)
    {
        m_StationIndex = index;
    }
    nn::pia::StationIndex GetStationIndex()
    {
        return m_StationIndex;
    }
    void SetCloneId(nn::pia::clone::CloneBase::Id id)
    {
        m_CloneId = id;
    }
    nn::pia::clone::CloneBase::Id GetCloneId()
    {
        return m_CloneId;
    }



    static nn::pia::reckoning::Simple3dReckoningStrategy* GetReckoningStrategy()
    {
        return &s_ReckoningStrategy;
    }

    static Operation* GetOperation()
    {
        return &s_Operation;
    }


protected:
    virtual const char* GetCloneTypeText() const = 0;
    void UpdateElement();
    const char* GetCurrentMode();

    bool isActive();

protected:
    nn::pia::clone::CloneBase::Id m_CloneId;
    nn::pia::StationIndex m_StationIndex;

    nn::pia::reckoning::ReckoningCloneElement<nn::pia::reckoning::Simple3dReckoningStrategy> m_ElementR;
    static nn::pia::reckoning::Simple3dReckoningStrategy s_ReckoningStrategy;

    // Element for comparison.
    nn::pia::clone::UnreliableCloneElement<Status, nn::pia::clone::NetworkByteOrderSerializePolicy<Status> > m_ElementU;
    // Element for notification.

    nn::pia::clone::ReliableCloneElement<EventStatusArg> m_ElementREX;

    static Operation s_Operation;
};

class SendObject : public Object
{
public:
    SendObject();
    virtual void Register();
    virtual void Unregister();
    virtual void RegisterElement()
    {
        RegisterElementCore(&m_Clone);
    }
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    virtual void PrintValue(int y, const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::SendClone m_Clone;

protected:
    virtual const char* GetCloneTypeText() const
    {
        return "S";
    }
};

class ReceiveObject : public Object
{
public:
    ReceiveObject();
    void InitCloneTypeText()
    {
        ::std::snprintf(m_CloneTypeText, sizeof(m_CloneTypeText), "R%1d", m_StationIndex);
    }
    virtual void Register();
    virtual void Unregister();
    virtual void RegisterElement()
    {
        RegisterElementCore(&m_Clone);
    }
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual int Print(int y, const nn::pia::sampleframework::Color& cColor);
    virtual void PrintValue(int y, const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::ReceiveClone m_Clone;

protected:
    virtual const char* GetCloneTypeText() const
    {
        return m_CloneTypeText;
    }
    char m_CloneTypeText[4];

    nn::pia::reckoning::Vector3f mDrawPos;
};


// A game object that uses SequentialSharingClone.
// To share the game state.
class SequentialObject : public Object
{
public:
    SequentialObject();
    virtual void Operate();
    virtual void Update();
    virtual bool Cleanup();
    virtual void RegisterClone();
    virtual void UnregisterClone();
    virtual bool IsCloneRegistered() const
    {
        return m_Clone.IsRegisteredWithProtocol();
    }
    virtual int Print(int y , const nn::pia::sampleframework::Color& cColor);
    nn::pia::clone::SequentialSharingClone m_Clone;

protected:
    virtual nn::pia::clone::CloneBase* GetClonePtr()
    {
        return &m_Clone;
    }
    virtual const nn::pia::clone::CloneBase& GetClone() const
    {
        return m_Clone;
    }
    virtual const char* GetCloneTypeText() const
    {
        return "Q";
    }


};
