﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Common_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"
#include "Shared_Common_Setup.h"

#include <pia/common/common_Trace.h>
#include <pia/common/common_Scheduler.h>
#include <pia/common/common_Time.h>

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

namespace
{
// Buffer size that caches the log output.
#ifdef PIA_PROFILING_TEST_BUILD
const u32 CACHED_PRINT_BUFFER_SIZE = 64 * 1024;
#else
#if NN_PIA_CTR
const u32 CACHED_PRINT_BUFFER_SIZE = 64 * 1024;
#else
const u32 CACHED_PRINT_BUFFER_SIZE = 512 * 1024;
#endif
#endif

// Handler for hooking exceptions.
#if NN_PIA_CTR
void setExceptionHandlers()
{
    // Not implemented for CTR.
}
#elif NN_PIA_WIN
void setExceptionHandlers()
{
    // Not implemented for Windows...
}
#elif NN_PIA_CAFE
OSExceptionCallback s_PrevDsiExceptionCallback = NULL;
OSExceptionCallback s_PrevIsiExceptionCallback = NULL;
OSExceptionCallback s_PrevProgramExceptionCallback = NULL;


void execCachedPrintFlush()
{
    // Execute the flushing process if an instance of CachedPrint exists.
    nn::pia::common::CachedPrint* p = nn::pia::common::CachedPrint::GetInstance();
    if (PIA_IS_VALID_POINTER(p))
    {
        p->Flush();
    }
}

// Guaranteed to return FALSE if exceptions are not processed.
#define CALL_PREV_CALLBACK(cb)         \
    if (PIA_IS_VALID_POINTER(cb))      \
    {                                  \
        return cb(interruptedContext); \
    }                                  \
    return FALSE

BOOL dsiExceptionHandler(OSContext* interruptedContext)
{
    execCachedPrintFlush();
    CALL_PREV_CALLBACK(s_PrevDsiExceptionCallback);
}

BOOL isiExceptionHandler(OSContext* interruptedContext)
{
    execCachedPrintFlush();
    CALL_PREV_CALLBACK(s_PrevIsiExceptionCallback);
}

BOOL programExceptionHandler(OSContext* interruptedContext)
{
    execCachedPrintFlush();
    CALL_PREV_CALLBACK(s_PrevProgramExceptionCallback);
}

void setExceptionHandlers()
{
    // Handler for hooking exceptions (Cafe).
    s_PrevDsiExceptionCallback = OSSetExceptionCallback(OS_EXCEPTION_DSI, dsiExceptionHandler);
    s_PrevIsiExceptionCallback = OSSetExceptionCallback(OS_EXCEPTION_ISI, isiExceptionHandler);
    s_PrevProgramExceptionCallback = OSSetExceptionCallback(OS_EXCEPTION_PROGRAM, programExceptionHandler);
}
#endif  // end of NN_PIA_CAFE


}  // end of anonymous namespace


nn::Result Common_InitializePiaCommon(
    u8* pPiaHeapBuffer,
    size_t szPiaHeapBufferSize,
    u64 traceFlags,
    bool enableCachedPrint,
    bool enableExceptionHook)
{
    PIA_SAMPLE_LOG("Common_InitializePiaCommon:traceFlags:0x%016llx, enableCachedPrint:%d, enableExceptionHook:%d\n", traceFlags, enableCachedPrint, enableExceptionHook);

    if (enableExceptionHook)
    {
        setExceptionHandlers();
    }

    nn::Result result;
    result = nn::pia::common::Initialize(pPiaHeapBuffer, szPiaHeapBufferSize);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::common::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    if (enableCachedPrint)
    {
        result = nn::pia::common::CachedPrint::CreateInstance(CACHED_PRINT_BUFFER_SIZE);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    result = nn::pia::common::Trace::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);


#if PIA_TEST_CODE
    u64 traceFlagsForPiaTest =
        traceFlags |
        nn::pia::common::TRACE_FLAG_COMMON |
        nn::pia::common::TRACE_FLAG_LOCAL |
        nn::pia::common::TRACE_FLAG_TRANSPORT |
//        nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_SEND |
//        nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_RECEIVE |
        nn::pia::common::TRACE_FLAG_SESSION |
        nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
        nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
        nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
        nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
        nn::pia::common::TRACE_FLAG_SESSION_JOINT |
        nn::pia::common::TRACE_FLAG_INET |
        nn::pia::common::TRACE_FLAG_INET_NAT |
        nn::pia::common::TRACE_FLAG_INET_NAT_DETAIL |
        nn::pia::common::TRACE_FLAG_INET_NAT_PROBE |
        nn::pia::common::TRACE_FLAG_INET_SOCKET |
        nn::pia::common::TRACE_FLAG_ALWAYS;
#if 1
    nn::pia::common::Trace::GetInstance()->SetFlag(traceFlagsForPiaTest);
#else
    // n1769 Trace is disabled here on a trial basis to examine how log/trace output increases the CPU load on the PC.
    nn::pia::common::Trace::GetInstance()->SetFlag(0);
#endif
#else
    nn::pia::common::Trace::GetInstance()->SetFlag(traceFlags);
#endif
    nn::pia::common::Trace::GetInstance()->SetPrintTime(true);

    result = nn::pia::common::Scheduler::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::common::WatermarkManager::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    nn::pia::common::WatermarkManager::GetInstance()->EnableAllWatermark();

    result = nn::pia::common::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Common_InitializePiaCommon:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());
    return result;
}

void Common_FinalizePiaCommon()
{
    PIA_SAMPLE_LOG("Common_FinalizePiaCommon\n");

    // With Pia, Finalize and DestroyInstance always succeed.
    nn::pia::common::WatermarkManager::DestroyInstance();
    nn::pia::common::Scheduler::DestroyInstance();
    nn::pia::common::Trace::DestroyInstance();
    if (nn::pia::common::CachedPrint::GetInstance() != NULL)
    {
        nn::pia::common::CachedPrint::GetInstance()->Flush();
        nn::pia::common::CachedPrint::DestroyInstance();
    }
    nn::pia::common::Finalize();
}
