﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Inet_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#if NN_PIA_CTR
#include <nn/version.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/ac.h>
#if (NN_VERSION_MAJOR == 2 && NN_VERSION_MINOR == 4 && NN_VERSION_MICRO >= 1) || NN_VERSION_MAJOR >= 3
#include <nn/socket_Initialize.h>
#else
#include <nn/socket.h>
#endif
#elif NN_PIA_CAFE
#include <cafe/net/so.h>
#include <nn/ac.h>
#include <nn/act.h>
#include <nn/fp.h>
#endif

#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"

namespace
{
// The network factory to specify for Transport or Session.
#if ENABLE_SERVERLESS
nn::pia::inet::LanNetworkFactory s_NetworkFactory;
#elif ENABLE_LAN_NETWORK
nn::pia::inet::NexLanNetworkFactory s_NetworkFactory;
#else
nn::pia::inet::NexNetworkFactory s_NetworkFactory;
#endif

// Network error code.
u32 s_NetworkErrorCode = 0;

// Specify the login network error code.
s32 s_LoginNetworkErrorCode = 0;

#if NN_PIA_CTR

/*
    Number of threads to handle the socket API.
    When using nn::nex::Core::ThreadModeUnsafeTransportBuffer,
    you need a total of five threads: two for NEX RV and three for Pia.

    When using nn::nex::Core::ThreadModeUnsafeUser,
    you need a total of four threads: one for NEX RV and three for Pia.
*/
const s32 SOCKET_SESSION_MAX = 5;
// The size to use when allocating the send or receive buffer for one socket.
const s32 SOCKET_BUFFER_SIZE = 64 * 1024;
// Specify the size of the working memory for the socket library.
const size_t SOCKET_BUFFER_TOTAL_SIZE = SOCKET_BUFFER_SIZE * SOCKET_SESSION_MAX;
// The working memory that is passed to the socket library. (Requires a 4096-byte aligned buffer.)
PIA_ATTRIBUTE_ALIGN(4096) u8 s_SocketBuffer[SOCKET_BUFFER_TOTAL_SIZE];

// The event that fires when a disconnection notification is received.
nn::os::Event s_ACDisconnectedEvent;

#elif NN_PIA_CAFE

// The connection status flag.
bool s_ACConnected = false;
bool s_UpdateACStatusRun = false;
// The thread and thread stack for periodically checking the connection status of the AC.
PIA_ATTRIBUTE_ALIGN(8) Thread s_UpdateACStatusThread;
PIA_ATTRIBUTE_ALIGN(4) u8 s_UpdateACStatusThreadStack[4 * 1024];

int UpdateACStatus(int argc, void* argv)
{
    NN_PIA_DUMMY_PARAM(argc);
    NN_PIA_DUMMY_PARAM(argv);

    while (s_UpdateACStatusRun)
    {
        bool IsConnected;
        nn::Result result = nn::ac::IsApplicationConnected(&IsConnected);
        if (result.IsSuccess())
        {
            // Save results so that they can be accessed from the main thread.
            s_ACConnected = IsConnected;
        }
        else
        {
            // Set the status to "disconnected" if the <tt>nn::ac::Initialize</tt> function has not been called.
            s_ACConnected = false;
        }

        // Loops approximately every 100 msec.
        OSSleepMilliseconds(100);
    }
    return 0;
}

#if ENABLE_SERVERLESS
#else
OSEvent s_FriendPresenceLoginEvent;
//-----------------------------------------------------------------------------
// Callback at completion of the asynchronous process for logging in to the friend server.
//-----------------------------------------------------------------------------
void FriendPresenceLoginCompletionCallback(nn::Result result, void* pContext)
{
    nn::Result* pResult = static_cast<nn::Result*>(pContext);
    *pResult = result;

    OSSignalEvent(&s_FriendPresenceLoginEvent);
}
#endif
#endif

void PrintIfConfig()
{
    nn::pia::common::InetAddress host, netmask, dns1, dns2, gateway;

#if NN_PIA_CTR
    s32 ret;
    nn::socket::InAddr addr1, addr2;

    NN_UNUSED_VAR(ret);

    ret = nn::socket::GetPrimaryAddress(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2));
    PIA_SAMPLE_ASSERT(ret == 0);
    host.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr1.addr));
    netmask.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr2.addr));

    ret = nn::socket::GetResolver(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2));
    PIA_SAMPLE_ASSERT(ret == 0);
    dns1.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr1.addr));
    dns2.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr2.addr));

    ret = nn::socket::GetDefaultGateway(reinterpret_cast<u8*>(&addr1));
    PIA_SAMPLE_ASSERT(ret == 0);
    gateway.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr1.addr));

#elif NN_PIA_CAFE
    nn::Result result;
    nn::ac::IpAddress addr;

    result = nn::ac::GetAssignedAddress(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    host.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr));

    result = nn::ac::GetAssignedSubnet(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    netmask.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr));

    result = nn::ac::GetAssignedPreferedDns(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    dns1.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr));

    result = nn::ac::GetAssignedAlternativeDns(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    dns2.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr));

    result = nn::ac::GetAssignedGateway(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    gateway.SetAddress(nn::pia::common::ByteOrder::NetworkToHost32(addr));
#endif
    nn::pia::common::String string;

    host.GetAddressString(&string);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: host   : %s\n", string.GetCharArray());

    netmask.GetAddressString(&string);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: netmask: %s\n", string.GetCharArray());

    dns1.GetAddressString(&string);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: dns1   : %s\n", string.GetCharArray());

    dns2.GetAddressString(&string);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: dns2   : %s\n", string.GetCharArray());

    gateway.GetAddressString(&string);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: gateway: %s\n", string.GetCharArray());

#if NN_PIA_CTR
#ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
    PIA_SAMPLE_LOG("\n");
    nn::socket::DumpRoutingTable();
#endif
#endif
}
};

u32 Inet_GetLocalAddress()
{
#if NN_PIA_CTR
    s32 ret;
    nn::socket::InAddr addr1, addr2;

    NN_UNUSED_VAR(ret);

    ret = nn::socket::GetPrimaryAddress(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2));
    PIA_SAMPLE_ASSERT(ret == 0);
    return addr1.addr;
#elif NN_PIA_CAFE
    nn::Result result;
    nn::ac::IpAddress addr;

    result = nn::ac::GetAssignedAddress(&addr);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    return addr;
#elif NN_PIA_WIN
    //Get the hostname.
    char hostname[256];
    if (gethostname(hostname, sizeof(hostname)) != 0)
    {
        return 0;
    }

    //Get IP address from the hostname.
    HOSTENT* hostent = gethostbyname(hostname);
    if (hostent == NULL)
    {
        return 0;
    }
    IN_ADDR inaddr;
    memcpy(&inaddr, hostent->h_addr_list[0], sizeof(struct in_addr));
    return ntohl(inaddr.S_un.S_addr);
#else
    return 0;
#endif
}


nn::Result Inet_InitializeNetwork()
{
    s_NetworkErrorCode = 0;

#if NN_PIA_CTR
    nn::Result result;
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(2000));

    // You must call fs::Initialize, but you do not need to call fs::Finalize.
    nn::fs::Initialize();

    PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Initialize\n");
    result = nn::ac::Initialize();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Initialize failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    nn::ac::Config config;
    nn::ac::CreateDefaultConfig(&config);

#ifndef RELEASE_VERSION
    nn::ac::DebugSetNetworkArea(&config, nn::ac::NETWORK_AREA_AUTO);
#endif

    PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Connect:\n");
    result = nn::ac::Connect(config);
    if (result.IsFailure())
    {
        u32 errorCode;
        nn::ac::GetLastErrorCode(&errorCode);
        PIA_SAMPLE_LOG("nn::ac::Connect() failed, errcode=%u\n", errorCode);
        s_NetworkErrorCode = errorCode;

        PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Connect failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    // Register the event for receiving disconnection notifications.
    s_ACDisconnectedEvent.Initialize(true);
    nn::ac::CTR::RegisterDisconnectEvent(&s_ACDisconnectedEvent);

    size_t requiredMemorySize = nn::socket::GetRequiredMemorySize(SOCKET_BUFFER_SIZE, SOCKET_SESSION_MAX);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: socket::Initialize requiredMemorySize: %d, SOCKET_BUFFER: %d, SOCKET_SESSION_MAX: %d\n", 
        requiredMemorySize, SOCKET_BUFFER_SIZE, SOCKET_SESSION_MAX);
    result = nn::socket::Initialize(reinterpret_cast<uptr>(s_SocketBuffer), requiredMemorySize, SOCKET_BUFFER_SIZE, SOCKET_SESSION_MAX);
    PIA_SAMPLE_ASSERT(requiredMemorySize < sizeof(s_SocketBuffer));

    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork: socket::Initialize failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }
#elif NN_PIA_CAFE
    nn::Result result;

    // Initialize the Account library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: act::Initialize\n");
    result = nn::act::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());

    // Initialize the Friend Presence library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: fp::Initialize\n");
    result = nn::fp::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());

    // Initialize the AC library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Initialize\n");
    result = nn::ac::Initialize();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Initialize: failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    // Connect to the network using the default profile in the System Settings.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Connect:\n");
    result = nn::ac::Connect();
    if (result.IsFailure())
    {
        nn::ac::ErrorCode errorCode;
        nn::ac::GetLastErrorCode(&errorCode);
        PIA_SAMPLE_LOG("nn::ac::Connect() failed, errcode=%u\n", errorCode);
        s_NetworkErrorCode = errorCode;

        PIA_SAMPLE_LOG("Inet_InitializeNetwork: ac::Connect failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    // Set the flag status to "connected."
    s_ACConnected = false;
    // Start the thread for checking the AC status.
    s_UpdateACStatusRun = true;
    s_UpdateACStatusThread.Start(UpdateACStatus, NULL, reinterpret_cast<void*>(s_UpdateACStatusThreadStack), sizeof(s_UpdateACStatusThreadStack), 0, OS_PRIORITY_APP_DEFAULT);

    // Initialize the socket library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork: SOInit\n");
    if (SOInit() != 0)
    {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork: SOInit failed, err=%d\n", SOLastError());
        return SampleFailedNetworkInterfaceUp();
    }

#elif NN_PIA_WIN
    // With Windows, do nothing because the operating system starts the network interface automatically.

    // Initialize Winsock.
    WORD wVersionRequested = MAKEWORD(2, 2);  // Request the version of WinSock to use.
    WSADATA wsaData;                          // Winsock data structure.
    int nErrorStatus;

    nErrorStatus = WSAStartup(wVersionRequested, &wsaData);
    if (nErrorStatus != 0)
    {
        PIA_BASIC_REPORT("Winsock initialization failed.\n");
        return SampleFailedNetworkInterfaceUp();
    }
    else
    {
        // Check whether it is the same as the requested version.
        if (LOBYTE(wsaData.wVersion) == 2 && HIBYTE(wsaData.wVersion) == 2)
        {
            // Display version information.
            PIA_BASIC_REPORT("version %d.%d is required\n",
                             LOBYTE(wsaData.wVersion),
                             HIBYTE(wsaData.wVersion));
            PIA_BASIC_REPORT("latest version is %d.%d\n",
                             LOBYTE(wsaData.wHighVersion),
                             HIBYTE(wsaData.wHighVersion));
        }
    }
#endif

    // Print the enabled network interface information.
    PrintIfConfig();

    return nn::ResultSuccess();
}

void Inet_FinalizeNetwork()
{
#if NN_PIA_CTR
    nn::Result result;

    // Finalize the socket library.
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: socket::Finalize\n");
    result = nn::socket::Finalize();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork: socket::Finalize failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
    }

    // Finalize the event for receiving disconnection notifications.
    s_ACDisconnectedEvent.Finalize();

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Close\n");
    result = nn::ac::Close();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Close failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        //n1589: If it returns here, the next AC initialization will fail.
    }

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Finalize\n");
    result = nn::ac::Finalize();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Finalize failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
    }

// You must call fs::Initialize, but you do not need to call fs::Finalize.
#elif NN_PIA_CAFE
    nn::Result result;

    // If the <tt>ac::Connect</tt> function has succeeded, finalize the socket library, and join the thread for checking the AC status.
    if (s_UpdateACStatusRun)
    {
        // Finalize the socket library.
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork: SOFinish\n");
        int ret = SOFinish();
        if (ret != 0)
        {
            PIA_SAMPLE_LOG("Inet_FinalizeNetwork: SOFinish failed, err=%d\n", ret);
        }

        // Terminate and join the thread for checking AC status.
        s_UpdateACStatusRun = false;
        s_UpdateACStatusThread.Join();
        s_UpdateACStatusThread.Finalize();
    }

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Close\n");
    result = nn::ac::Close();
    if (result.IsFailure())
    {
        nn::ac::ErrorCode errorCode;
        nn::ac::GetLastErrorCode(&errorCode);
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Close failed, errcode=%u\n", errorCode);
        PIA_SAMPLE_LOG_RESULT(result);
    }

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: ac::Finalize\n");
    nn::ac::Finalize();
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: fp::Finalize\n");
    nn::fp::Finalize();
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork: act::Finalize\n");
    nn::act::Finalize();

#elif NN_PIA_WIN
    // Stop Winsock.
    WSACleanup();
// With Windows, do nothing because the network interface is controlled by the operating system.
#endif
}

bool Inet_IsConnected()
{
#if NN_PIA_CTR
    if (s_ACDisconnectedEvent.Wait(0))
    {
        return false;
    }
    else
    {
        return true;
    }
#elif NN_PIA_CAFE
    return s_ACConnected;
#elif NN_PIA_WIN
    return true;
#endif
}


//-----------------------------------------------------------------------------
// Log in to the game server.
// Log in to the friend server at the same time.
//-----------------------------------------------------------------------------
nn::Result Inet_Login(u32 timeout)
{
    nn::Result result;

#if ENABLE_SERVERLESS
    result = nn::pia::inet::LanFacade::GetInstance()->Bind(Inet_GetLocalAddress());
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("LanFacade::Bind failed.\n");
        return SampleFailedNetworkInterfaceUp();
    }
#else
    PIA_SAMPLE_LOG("Inet_Login\n");

    NN_PIA_COMMON_SET_PROGRAM_INFORMATION();

    nn::nex::NgsFacade* pNgsFacade = Nex_GetNgsFacadeInstance();
    s_LoginNetworkErrorCode = 0;

#if ENABLE_LAN_NETWORK
    PIA_SAMPLE_LOG("Inet_Login: using LAN network\n");

    //The appropriate account name.
    const int NAMELEN = 20;
    qChar buffer[NAMELEN + 1] = { '\0' };
    if (buffer[0] == 0)
    {
        for (int i = 0; i < NAMELEN; i++)
        {
            int rndNum = nn::nex::Platform::GetRandomNumber(52);
            bool isBig = (rndNum < 26);
            buffer[i] = ((isBig ? 'A' : 'a') + static_cast<nn::nex::qChar>(rndNum % 26));
        }
        buffer[NAMELEN] = '\0';
    }

    nn::nex::StringStream account;
    account << buffer;

    //Information about the server to connect to.
    //n2788: Address of the LAN matchmaking server to use for development with the Pia library.
    qChar* pHost = _T("10.149.23.200");
    qUnsignedInt16 port = 9110;
    u32 gameId = MATCHMAKE_GAME_ID;
    qChar* pAccessKey = _T("Test");
    qChar8* pPassword = "password";

    //Create an account.
    {
        nn::nex::ProtocolCallContext pccCreate;
        nn::nex::qResult ret = pNgsFacade->DebugCreateAccount(&pccCreate, pHost, port, account.CStr(), pPassword, pAccessKey, 30000);
        if (!ret)
        {
            PIA_SAMPLE_LOG("NgsFacade::DebugCreateAccount() call failed.\n");
            return SampleFailedNexLogin();
        }

        pccCreate.Wait();
        if (pccCreate.GetState() != nn::nex::CallContext::CallSuccess)
        {
            PIA_SAMPLE_LOG("NgsFacade::DebugCreateAccount() failed.\n");
            //return NULL;
        }
    }

    nn::nex::ProtocolCallContext oContext;
    if (!pNgsFacade->DebugLoginDirect(&oContext, pHost, port, account.CStr(), pPassword, pAccessKey, nn::nex::TimeInterval(timeout)))
    {
        // Failed to start the login process.
        PIA_SAMPLE_LOG("Inet_Login: Login failed.\n");
        return SampleFailedNexLogin();
    }
#else
#if NN_PIA_CTR
    nn::os::Event event(true);

    // Initialize the friend library.
    result = nn::friends::Initialize();
    if (result.IsFailure())
    {
        // This can fail due to the internal state of the daemon.
        PIA_SAMPLE_LOG("Inet_Login: friends::Initialize failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    // Log in to the friend server.
    result = nn::friends::Login(&event);
    if (result.IsFailure())
    {
        // This can fail if you are not connected to an AP.
        PIA_SAMPLE_LOG("Inet_Login: friends::Login failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    // Wait for asynchronous login processing to end.
    // Put code here for readability.
    event.Wait();

    // Get the results of the asynchronous login process.
    result = nn::friends::GetLastResponseResult();
    if (result.IsFailure())
    {
        // This can fail due to the internal state of the daemon.
        PIA_SAMPLE_LOG("Inet_Login: friends::GetLastResponseResult failed, Error Code: %d\n", nn::friends::ResultToErrorCode(result));
        s_NetworkErrorCode = nn::friends::ResultToErrorCode(result);
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    PIA_SAMPLE_LOG("Inet_Login: friends::Login succeeded, MyPrincipalId: %d\n", nn::friends::GetMyPrincipalId());

    // Log in to the game server as soon as you can.
    if (!nn::friends::HasLoggedIn())
    {
        PIA_SAMPLE_LOG("Inet_Login: HasLoggedIn failed.\n");
        return SampleFailedFriendsLogin();
    }
#elif NN_PIA_CAFE
    // Logging in to the friend server is not necessary, but log in if you are using Friend Presence.
    nn::Result asyncResult;

    // Initializes events in auto mode.
    OSInitEvent(&s_FriendPresenceLoginEvent, FALSE, OS_EVENT_AUTO);
    PIA_SAMPLE_LOG("Inet_Login: fp::LoginAsync called.\n");
    result = nn::fp::LoginAsync(FriendPresenceLoginCompletionCallback, &asyncResult);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_Login: fp::LoginAsync start failed, Error Code: %d\n", nn::fp::ResultToErrorCode(result));
        s_NetworkErrorCode = nn::fp::ResultToErrorCode(result);
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }
    // Waits for the event signaling the completion of the <tt>nn::fp::LoginAsync</tt> function.
    OSWaitEvent(&s_FriendPresenceLoginEvent);
    if (asyncResult.IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_Login: fp::LoginAsync async failed, Error Code: %d\n", nn::fp::ResultToErrorCode(asyncResult));
        s_NetworkErrorCode = nn::fp::ResultToErrorCode(asyncResult);
        PIA_SAMPLE_LOG_RESULT(asyncResult);
        return asyncResult;
    }

    PIA_SAMPLE_LOG("Inet_Login: fp::LoginAsync succeeded, MyPrincipalId: %d\n", nn::fp::GetMyPrincipalId());
#endif

    PIA_SAMPLE_LOG("Logging in to Game Server...\n");

    u32 gameId = MATCHMAKE_GAME_ID;
    const qChar* pAccessKey = MATCHMAKE_ACCESS_KEY;

    // Log in to the game server.
    nn::nex::ProtocolCallContext oContext;
    if (!pNgsFacade->Login(&oContext, gameId, pAccessKey, nn::nex::TimeInterval(timeout)))
    {
        // Failed to start the login process.
        PIA_SAMPLE_LOG("Inet_Login: Login failed.\n");
        return SampleFailedNexLogin();
    }
#endif  // ENABLE_LAN_NETWORK

    // Wait for asynchronous login processing to end.
    // nn::nex::Scheduler::Dispatch is called in ProtocolCallContext::Wait, and the asynchronous processing progresses.
    // Put code here for readability.
    oContext.Wait();
    if (!oContext.GetOutcome())
    {
        u32 networkErrorCode;
        networkErrorCode = nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(oContext.GetOutcome());
        PIA_SAMPLE_LOG("Inet_Login: Login failed. NetworkErrorCode: %d, LoginErrorCode: %d\n",
                       networkErrorCode,
                       pNgsFacade->GetLastLoginErrorCode());
        s_NetworkErrorCode = networkErrorCode;
        s_LoginNetworkErrorCode = pNgsFacade->GetLastLoginErrorCode();
        return SampleFailedNexLogin();
    }

    if (!pNgsFacade->IsConnected())
    {
        PIA_SAMPLE_LOG("Inet_Login: NgsFacade::IsConnected failed.\n");
        return SampleFailedNexLogin();
    }
    // If logging in succeeds, it does not become NULL.
    PIA_SAMPLE_ASSERT(pNgsFacade->GetCredentials() != NULL);

    if (PIA_IS_VALID_POINTER(nn::pia::inet::NexFacade::GetInstance()))
    {
        // Sets the NEX game server information in <tt>NexFacade</tt>.
        nn::pia::inet::NexFacade::LoginInfo loginInfo;
        loginInfo.gameServerId = gameId;
        loginInfo.pNgsFacade = pNgsFacade;

        result = nn::pia::inet::NexFacade::GetInstance()->Bind(&loginInfo);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    PIA_SAMPLE_LOG("Inet_Login: Login succeeded\n");
#endif

#if !NN_PIA_WIN && !ENABLE_SERVERLESS
    // For the community ID
    nn::nex::MatchmakeExtensionClient* pMatchmakeClient = Nex_GetMatchmakeExtensionClientInstance();
    nn::nex::Credentials* pCredentials = pNgsFacade->GetCredentials();
    NN_PIA_DUMMY_PARAM(pMatchmakeClient);
    NN_PIA_DUMMY_PARAM(pCredentials);
    PIA_SAMPLE_ASSERT(pCredentials != NULL);
    PIA_SAMPLE_ASSERT( pMatchmakeClient->Bind(pCredentials) );
#endif

    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------
// Log out from the game server.
//-----------------------------------------------------------------------------
nn::Result Inet_Logout()
{
#if ENABLE_SERVERLESS
    if (nn::pia::inet::LanFacade::GetInstance())
    {
        nn::pia::inet::LanFacade::GetInstance()->Unbind();
    }
#else

#if !NN_PIA_WIN
    // For the community ID
    nn::nex::MatchmakeExtensionClient* pMatchmakeClient = Nex_GetMatchmakeExtensionClientInstance();
    if(pMatchmakeClient)
    {
        PIA_SAMPLE_ASSERT(pMatchmakeClient->Unbind());
    }
#endif

    // Clears the NEX game server information from <tt>NexFacade</tt>.
    if (nn::pia::inet::NexFacade::GetInstance())
    {
        nn::pia::inet::NexFacade::GetInstance()->Unbind();
    }

    // Log out from the game server.
    nn::nex::NgsFacade* pNgsFacade = Nex_GetNgsFacadeInstance();
    if (pNgsFacade)
    {
        nn::nex::ProtocolCallContext ccLogout;
        PIA_SAMPLE_LOG("Inet_Logout: Logout\n");
        pNgsFacade->Logout(&ccLogout);
        ccLogout.Wait();
    }
#if ENABLE_LAN_NETWORK
#else
#if NN_PIA_CTR
    PIA_SAMPLE_LOG("Inet_Logout: friends::Logout\n");
    // Log out from the friend server.
    if (nn::friends::IsInitialized())  // Because it apparently triggers an assert in the SDK if you call the Logout function while n1769 is not initialized...
    {
        PIA_SAMPLE_LOG("Inet_Logout: friends::Logout called.\n");
        nn::friends::Logout();
    }

    // Terminate the friend library.
    if (nn::friends::IsInitialized())  // Because it apparently triggers an assert in the SDK if you call the Finalize function while n1769 is not initialized...
    {
        nn::friends::Finalize();
    }
#elif NN_PIA_CAFE
    // Log out from the friend server.
    PIA_SAMPLE_LOG("Inet_Logout: fp::Logout\n");
    if (nn::fp::IsInitialized())
    {
        PIA_SAMPLE_LOG("Inet_Logout: fp::Logout called.\n");
        nn::fp::Logout();
    }
#endif
#endif

    PIA_SAMPLE_LOG("Inet_Logout: Logout succeeded\n");
#endif
    return nn::ResultSuccess();
}

u32 Inet_GetNetworkErrorCode()
{
    return s_NetworkErrorCode;
}

s32 Inet_GetLoginNetworkErrorCode()
{
    return s_LoginNetworkErrorCode;
}

#if NN_PIA_CTR
nn::Result Inet_UpdateGameMode(nn::friends::GameMode gameMode, const wchar_t* pDescription)
{
    nn::Result result = nn::friends::UpdateGameMode(gameMode, pDescription);
    if (result.IsFailure())
    {
        s_NetworkErrorCode = nn::friends::ResultToErrorCode(result);
    }
    return result;
}
#elif NN_PIA_CAFE
nn::Result Inet_UpdateGameMode(nn::fp::GameMode gameMode, const wchar_t* pDescription)
{
    nn::Result result = nn::fp::UpdateGameMode(&gameMode, pDescription);
    if (result.IsFailure())
    {
        s_NetworkErrorCode = nn::fp::ResultToErrorCode(result);
    }
    return result;
}
#endif

nn::Result Inet_InitializePiaInet(size_t mtuSize)
{
    PIA_SAMPLE_LOG("Inet_InitializePiaInet\n");
    nn::Result result;

    nn::pia::inet::Setting setting;
    setting.mtuSize = mtuSize;

    result = nn::pia::inet::Initialize(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::inet::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

#if ENABLE_SERVERLESS
    result = nn::pia::inet::LanFacade::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#else
    result = nn::pia::inet::NexFacade::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#endif

    result = nn::pia::inet::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

#if ENABLE_LAN_NETWORK
    nn::pia::inet::NexFacade::EnableLanNetwork(true);
#endif

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Inet_InitializePiaInet: Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());
    return result;
}

void Inet_FinalizePiaInet()
{
    PIA_SAMPLE_LOG("Inet_FinalizePiaInet\n");

// With Pia, Finalize and DestroyInstance always succeed.
#if ENABLE_SERVERLESS
    nn::pia::inet::LanFacade::DestroyInstance();
#else
    nn::pia::inet::NexFacade::DestroyInstance();
#endif
    nn::pia::inet::Finalize();

    PIA_SAMPLE_LOG("Inet_FinalizePiaInet finished\n");
}

nn::pia::transport::NetworkFactory* Inet_GetNetworkFactory()
{
    return &s_NetworkFactory;
}
