﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/util.h>
#include "SampleLib/SampleLib.h"
#include "Object.h"

#include <cstdio>

#if NN_PIA_CTR
#define USE_PIA_LOCAL 1
#else
#define USE_PIA_LOCAL 0
#endif


#if USE_PIA_LOCAL
#include <pia/local.h>
#include "Shared_Local_Setup.h"
#endif

#include "Shared_Basic_Setup.h"
#include "Shared_Clone_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Reckoning_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Transport_Setup.h"

#define PROGRAM_TITLE "SimpleReckoning"

#if NN_PIA_WIN
#define snprintf(...) _snprintf(__VA_ARGS__)
#endif

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif


namespace
{
const bool ENABLE_HOST_MIGRATION = false;  //!<  Enables or disables the host migration feature. It is disabled permanently in this demo.
const int MAX_STATIONS = 8;                // Defines the maximum number of stations.
const int cObjectNum = 1 + MAX_STATIONS;
const nn::pia::clone::CloneBase::Id cCloneId = 0;

static SendObject s_SendObject;
static ReceiveObject s_ReceiveObject[MAX_STATIONS];
static Object* s_apObject[cObjectNum];
static SequentialObject s_SeqObject;

static int s_Cursor = cObjectNum;
static int s_HoldCounterUp = 0;
static int s_HoldCounterDown = 0;
static int s_HoldCounterRight = 0;
static int s_HoldCounterLeft = 0;
static int s_PrintSendPacketNum = 0;
static int s_PrintReceivePacketNum = 0;
static int s_LastSendPacketNum = 0;
static int s_LastReceivePacketNum = 0;

// Indicates the mode that is executing.
// With inet, starts with automatch, and then determines whether to be the host or a client.
enum GameMode
{
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_MASTER,
    GAME_MODE_LOCAL_CLIENT,
#endif
    GAME_MODE_INET_AUTOMATCH,
    GAME_MODE_INET_MASTER,
    GAME_MODE_INET_CLIENT,
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_ERROR,
#endif
    GAME_MODE_INET_ERROR,
    GAME_MODE_EXIT
};

const size_t MEMSIZE = 10 * 1024 * 1024;             // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.

// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const size_t SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const size_t SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

#define SAMPLE_NEX_THREAD_MODE nn::nex::Core::ThreadModeUnsafeTransportBuffer

const u32 MATCHMAKE_GAME_MODE = 55000;
#if USE_PIA_LOCAL
const u32 SAMPLE_TITLE_ID = 0xFF340;
const u8 SAMPLE_SUB_ID = 0x00;
//  Using <tt>sizeof</tt>, so use [ ] even to define <tt>char</tt>.
const char SAMPLE_PASSPHRASE[] = "aaabbccc0123456789";
const size_t SAMPLE_PASSPHRASE_SIZE = sizeof(SAMPLE_PASSPHRASE);
const char SAMPLE_SIGNATURE_KEY_DATA[] = "simplereckoning_key";
const size_t SAMPLE_SIGNATURE_KEY_DATA_SIZE = sizeof(SAMPLE_SIGNATURE_KEY_DATA);
#endif

}  // end of anonymous namespace


static nn::pia::StationIndex ConvertStationIdToDataIndex(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
    if(result.IsSuccess())
    {
        PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
        return stationIndex;
    }
    else
    {
        // When there is no longer a host and the session is forced to end, <ttt>Draw</tt> can be called in a flash
        // before the state ends, so do not throw an assertion on the <tt>result.IsSuccess</tt> function.
        return nn::pia::STATION_INDEX_INVALID;
    }
}

// Determines the mode to execute.
static GameMode DecideGameMode()
{
#if USE_PIA_LOCAL
    PIA_SAMPLE_PRINT("Please select game mode.\n [A]LocalHost\n [B]LocalClient\n [X]InetAutomatch\n [START]Quit\n");
#else
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
#endif
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
#if USE_PIA_LOCAL
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalHost\n");
            gameMode = GAME_MODE_LOCAL_MASTER;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalClient\n");
            gameMode = GAME_MODE_LOCAL_CLIENT;
            isDecided = true;
        }
#endif

        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        pSampleFramework->DrawText(10, 20, "Select game mode.");
#if USE_PIA_LOCAL
        pSampleFramework->DrawText(10, 30, "A: LocalHost");
        pSampleFramework->DrawText(10, 40, "B: LocalClient");
        pSampleFramework->DrawText(10, 50, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 60, "START: Quit this application");
#else
        pSampleFramework->DrawText(10, 30, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 40, "START: Quit this application");
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return gameMode;
        }
    }
}

static void RegisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex index = ConvertStationIdToDataIndex(stationId);
    PIA_SAMPLE_ASSERT(index < MAX_STATIONS);

    ReceiveObject& r = s_ReceiveObject[index];
    r.Register();
}

static void UnregisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex index = ConvertStationIdToDataIndex(stationId);
    PIA_SAMPLE_ASSERT(index < MAX_STATIONS);

    ReceiveObject& r = s_ReceiveObject[index];
    r.Unregister();
}

// Declarations of callback functions.
static void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG("Sample Event JOIN detect. StationId(%016llx)\n", stationId.GetValue());
        RegisterStationId(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect. StationId(%016llx)\n", stationId.GetValue());
        UnregisterStationId(stationId);
    }
    break;
    }
}

// Initialize the objects.
static void InitObjects()
{
    int i = 0;
    {
        s_SendObject.SetCloneId(cCloneId);
        s_apObject[i] = &s_SendObject;
        ++i;
    }
    for (int j = 0; j < MAX_STATIONS; ++j)
    {
        s_ReceiveObject[j].SetCloneId(cCloneId);
        s_ReceiveObject[j].SetStationIndex(static_cast<nn::pia::StationIndex>(j));
        s_ReceiveObject[j].InitCloneTypeText();
        s_apObject[i] = &s_ReceiveObject[j];
        ++i;
    }

    {
        s_SeqObject.SetCloneId( cCloneId );
    }
    PIA_SAMPLE_ASSERT(i == cObjectNum);
}

static void RegisterSendObject()
{
    nn::pia::StationIndex sidx;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&sidx, nn::pia::session::Session::GetInstance()->GetLocalStationId());
    if (result.IsFailure())
    {
        sidx = nn::pia::STATION_INDEX_INVALID;
    }

    s_SendObject.SetStationIndex(sidx);
    s_SendObject.Register();

    if( !s_SeqObject.IsCloneRegistered() ){
        s_SeqObject.RegisterClone();
    }
}

// Operation by key input.
static bool Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    // Exit the GameLoop function if START is pressed.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        return true;
    }

    s_SendObject.Operate();
    s_SeqObject.Operate();
    {
//        f32 threshold = Object::GetReckoningStrategy()->GetThreshold();
//        f32 thStop = Object::GetOperation()->GetThresholdStop();

        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_UP))
        {
            ++s_HoldCounterUp;
        }
        else if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
        {
            ++s_HoldCounterDown;
        }
        else
        {
            s_HoldCounterUp = 0;
            s_HoldCounterDown = 0;
        }

        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
        {
            ++s_HoldCounterRight;
        }
        else if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
        {
            ++s_HoldCounterLeft;
        }
        else
        {
            s_HoldCounterRight = 0;
            s_HoldCounterLeft = 0;
        }

    }
    {
        nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            switch (pCloneProtocol->GetState())
            {
            case nn::pia::clone::CloneProtocol::STATE_INACTIVE:
            {
                nn::Result r = pCloneProtocol->Start();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            }
            break;

            case nn::pia::clone::CloneProtocol::STATE_START:
            case nn::pia::clone::CloneProtocol::STATE_ACTIVE:
            {
                nn::Result r = pCloneProtocol->Stop();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            }
            break;
            }
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            if (pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE &&
                pCloneProtocol->IsRegulatingClock() == false)
            {
                nn::Result r = pCloneProtocol->RegulateClock();
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            }
        }
    }

    return false;
}

static nn::pia::sampleframework::Color GetColor(int index)
{
    nn::pia::sampleframework::Color result(1.f, 1.f, 1.f, 1.f);
    switch (index)
    {
    case 0:
    {
        result = nn::pia::sampleframework::Color(1.f, 0.0f, 0.0f, 1.f);
    }
    break;
    case 1:
    {
        result = nn::pia::sampleframework::Color(0.4f, 0.4f, 0.9f, 1.f);
    }
    break;
    case 2:
    {
        result = nn::pia::sampleframework::Color(0.f, 0.5f, 0.f, 1.f);
    }
    break;
    case 3:
    {
        result = nn::pia::sampleframework::Color(0.5f, 0.5f, 0.f, 1.f);
    }
    break;
    case 4:
    {
        result = nn::pia::sampleframework::Color(0.f, 1.f, 1.f, 1.f);
    }
    break;
    case 5:
    {
        result = nn::pia::sampleframework::Color(1.f, 0.f, 1.f, 1.f);
    }
    break;
    case 6:
    {
        result = nn::pia::sampleframework::Color(1.f, 1.f, 0.f, 1.f);
    }
    break;
    default:
    {
        result = nn::pia::sampleframework::Color(0.5f, 0.5f, 0.5f, 1.f);
    }
    break;
    }
    return result;
}

// Draw on the screen.
static void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    const nn::pia::sampleframework::Color cCursorColor(1.0f, 0.3f, 0.3f, 1.0f);
    const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);
    const nn::pia::sampleframework::Color cColorR(1.0f, 0.4f, 0.4f, 1.0f);
    const nn::pia::sampleframework::Color cColorG(0.4f, 1.0f, 0.4f, 1.0f);
    const nn::pia::sampleframework::Color cColorB(0.4f, 0.4f, 1.0f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);
    {
        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "LocalHost";
            }
            else
            {
                cGameModeText = "LocalClient";
            }
        }
        break;
#endif
        case GAME_MODE_INET_AUTOMATCH:
        {
            cGameModeText = "InetAutomatch";
        }
        break;
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "InetHost";
            }
            else
            {
                cGameModeText = "InetClient";
            }
        }
        break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%d", pSession->GetStationNum());
        pSampleFramework->DrawText(300, 20, "Index:%d", ConvertStationIdToDataIndex(pSession->GetLocalStationId()));
    }

    {
        int idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    {
        pSampleFramework->SetFontColor(s_Cursor == cObjectNum ? cCursorColor : cDefaultColor);

        const char* cpStateText = "";
        switch (pCloneProtocol->GetState())
        {
        case nn::pia::clone::CloneProtocol::STATE_INACTIVE:
        {
            cpStateText = "Inactive";
        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_START:
        {
            cpStateText = "Start";
        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_ACTIVE:
        {
            cpStateText = "Active";

        }
        break;
        case nn::pia::clone::CloneProtocol::STATE_STOP:
        {
            cpStateText = "Stop";
        }
        break;
        }
        pSampleFramework->DrawText(10, 225, "State:%s", cpStateText);

        pSampleFramework->SetFontColor(cDefaultColor);
    }

    {
        nn::pia::clone::ClockValue clock = pCloneProtocol->GetClock();
        if (clock == nn::pia::clone::INVALID_CLOCK)
        {
            pSampleFramework->DrawText(10, 210, "Clock: --------");
        }
        else
        {
            pSampleFramework->DrawText(10, 210, "Clock: %8.2f%c", static_cast<f32>(clock) / 1000.0f, pCloneProtocol->IsRegulatingClock() ? 'R' : ' ');
        }
    }
/*
    {
        pSampleFramework->DrawText(10, 205, "Stop Threshold:%6.2f", Object::GetOperation()->GetThresholdStop());
        pSampleFramework->DrawText(10, 215, "Threshold:%6.2f", Object::GetReckoningStrategy()->GetThreshold());
    }
*/
    {
        if (pSampleFramework->GetFrameCount() % 30 == 0)
        {
            {
                int packetNum = nn::pia::transport::ThreadStreamManager::GetInstance()->GetSendThreadSendPacketNum();
                s_PrintSendPacketNum = packetNum - s_LastSendPacketNum;
                s_LastSendPacketNum = packetNum;
            }
            {
                int packetNum = nn::pia::transport::ThreadStreamManager::GetInstance()->GetReceiveThreadReceivePacketNum();
                s_PrintReceivePacketNum = packetNum - s_LastReceivePacketNum;
                s_LastReceivePacketNum = packetNum;
            }
        }

        pSampleFramework->DrawText(150, 225 - 8, "Send : %3d", s_PrintSendPacketNum);
        pSampleFramework->DrawText(150, 225, "Recv : %3d", s_PrintReceivePacketNum);
    }

    {
        int y = 30;
        for (int i = 0; i < MAX_STATIONS; ++i)
        {

            Object* o = &s_ReceiveObject[i];
            nn::pia::sampleframework::Color color = GetColor(i);
            nn::pia::StationIndex idx = o->GetStationIndex();
            // Print the Send object before the line feed.
            if (s_SendObject.GetStationIndex() == idx)
            {
                y = s_SendObject.Print(y, color );
            }
            y = o->Print(y, color );  // Line feed here.
        }
    }
    pSampleFramework->SetFontColor(cDefaultColor);

    s_SeqObject.Print( 0 , cDefaultColor );
}

static void Update()
{
    nn::pia::clone::CloneProtocol* pCloneProtocol = Clone_GetCloneProtocol();

    if (pCloneProtocol->GetState() != nn::pia::clone::CloneProtocol::STATE_INACTIVE)
    {
        pCloneProtocol->UpdateClock();
    }

    if (pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_START ||
        pCloneProtocol->GetState() == nn::pia::clone::CloneProtocol::STATE_ACTIVE)
    {
        if (pCloneProtocol->GetError() != nn::pia::clone::CloneProtocol::ERROR_TYPE_NOTHING)
        {
            // Detected an error, so stop.
            const char* cpErrorText = "";
            switch (pCloneProtocol->GetError())
            {
            case nn::pia::clone::CloneProtocol::ERROR_TYPE_CLOCK_OVERFLOW:
            {
                cpErrorText = "ClockOverflow";
            }
            break;
            case nn::pia::clone::CloneProtocol::ERROR_TYPE_DROP_EVENT:
            {
                cpErrorText = "DropEvent";
            }
            break;
            }
            PIA_SAMPLE_PRINT("Error: %s\n", cpErrorText);
            pCloneProtocol->Stop();
        }
    }

    for (int i = 0; i < cObjectNum; ++i)
    {
        s_apObject[i]->Update();
    }

    s_SeqObject.Update();
}

// Main loop during communication.
static void MainLoop(GameMode gameMode, bool* pIsLeaveNeeded)
{
    RegisterSendObject();

    {

    // The Send Latency Settings
    nn::pia::transport::ThreadStreamManager::GetInstance()->SetSendThreadLatencyEmulation( 200, 200 );

        nn::pia::transport::ThreadStreamManager::GetInstance()->SetReceiveThreadLatencyEmulation( 0, 0 );

    }

    nn::Result result;
    while (true)
    {
        bool isExit = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        isExit |= Operate();

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (gameMode == GAME_MODE_INET_MASTER || gameMode == GAME_MODE_INET_CLIENT)
        {
            if (Inet_IsConnected() == false)
            {
                // Ends because the AC has disconnected.
                isExit = true;
                *pIsLeaveNeeded = false;
            }
        }

        result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
        if (result.IsSuccess() &&
            nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
        {
            Update();
        }
        else
        {
            // Ends because the session was lost.
            isExit = true;
            nn::pia::util::HandlingType handlingType = nn::pia::util::GetHandlingType(result);
            if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP)
            {
                *pIsLeaveNeeded = true;
            }
            else if(handlingType == nn::pia::util::HANDLING_RECONNECT)
            {
                *pIsLeaveNeeded = false;
            }
        }



        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        Draw(gameMode);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            return;
        }
    }
}

static void ResetReckoningStrategy()
{
    bool isCleanupAll = true;

    do
    {
        for (int i = 0; i < cObjectNum; ++i)
        {
            // Unregister each Element of each Clone.
            isCleanupAll &= s_apObject[i]->Cleanup();
        }
    } while (!isCleanupAll);

    if (isCleanupAll)
    {
        PIA_SAMPLE_PRINT("Called ResetReckoningStrategy()\n");
        for (int i = 0; i < cObjectNum; ++i)
        {
            s_apObject[i]->ResetReckoningStrategy();
        }
        // Re-register.
        for (int i = 0; i < cObjectNum; ++i)
        {
            s_apObject[i]->RegisterElement();
        }
    }
}

// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;
    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(60);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.1f, 0.1f, 0.1f, 1.0f));
    }

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    InitObjects();



    while (true)
    {
        GameMode gameMode = DecideGameMode();
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            nn::pia::common::Log::GetGlobalObject()->SetMask(nn::pia::common::Log::LEVEL_ERROR);

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            {
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID, ENABLE_HOST_MIGRATION);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
            {
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }



            {
                // Insert debugSetting.
//            nn::Result result;

            nn::pia::transport::Transport::DebugSetting debugSetting;
            debugSetting.sendThreadLatencyEmulationPacketNum = 128;
            debugSetting.receiveThreadLatencyEmulationPacketNum = 128;
            debugSetting.bPacketLossEmulation = false;
            result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            }
            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            nn::pia::transport::Transport::Setting setting;
            setting.pFactory = pFactory;
            setting.maxStationNum = MAX_STATIONS;
            setting.maxSendThreadBufferNum = 16;
            setting.maxReceiveThreadBufferNum = 24;


            result = Transport_InitializePiaTransport(setting);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            // Set up clone.
            Clone_InitializePiaClone();

            // Set up reckoning.
            Reckoning_InitializePiaReckoning();

            // Sets up the session.
            result = Session_InitializePiaSession(
                pFactory,
                nn::pia::session::NetworkTopology_FullMesh);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionCallback);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }

        // Connect to the network connections and create a session.
        {
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the matchmaking session to create.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, SAMPLE_PASSPHRASE_SIZE);
                        createNetworkSetting.m_PassphraseLength = SAMPLE_PASSPHRASE_SIZE;
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, SAMPLE_SIGNATURE_KEY_DATA_SIZE);
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
                    searchCriteriaList[0].SetResultRange(0, 8);
                }

                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connect to the Internet
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }
                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
                if (isHost)
                {
                    // Became the host as a result of the matching.
                    gameMode = GAME_MODE_INET_MASTER;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;

            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        bool isLeaveNeeded = false;

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
        {
            MainLoop(gameMode, &isLeaveNeeded);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }
        }
        break;
        }


        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            {
                if (isLeaveNeeded)
                {
                    // Leaves the session.
                    // In this sample, the return value is ignored because finalization will take place later.
                    Session_LeaveSession();
                }
            }
            break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
            }
            break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_Logout();
                Inet_FinalizeNetwork();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Reckoning_FinalizePiaReckoning();
            Clone_FinalizePiaClone();
            ResetReckoningStrategy();
            Transport_FinalizePiaTransport();
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
                Local_FinalizePiaLocal();
            }
            break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
            Common_FinalizePiaCommon();
        }
    }

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
