﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SampleConfig.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>

#include <pia/inet.h>

#if NN_PIA_CTR
#include <pia/local.h>
#include <nn/friends.h>
#elif NN_PIA_CAFE
#include <nn/fp.h>
#else
// This sample program will run only on the CTR or Cafe platforms.
#error  "Invalid platform."
#endif

#include <pia/transport.h>
#include <pia/session.h>
#include <SampleLib/Assert.h>

#include "Shared_Inet_Setup.h"

#define FRAME_PER_SECOND 60

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

#define ENABLE_OPEN_CLOSE_SESSION 0

const char PROGRAM_NAME[] = "session/advanced";

/*!
    @brief  Enumerates possible states of the sample program.
 */
enum SampleSceneNo
{
    SCENE_START,                  //!<  The initial state. It has no relation to communication.
    SCENE_PREPARE_INET,           //!<  This is the state where preparations are made for Internet communication. It has no relation to communication.
    SCENE_PREPARE_LOCAL,          //!<  This is the state where preparations are made for local communication. It has no relation to communication.
    SCENE_SELECT_MATCHMAKE_TYPE,  //!<  This is the state that specifies how the session is constructed.
    SCENE_RANDOM_MATCHMAKE,       //!<  This is the state that uses the auto-matchmaking feature to join a session.
    SCENE_BROWSE_MATCHMAKE,       //!<  This is the state that creates a session to join, or searches for a created session and joins it, for each specified network type.
    SCENE_FRIENDS_MATCHMAKE,      //!<  This is the state that uses the friend presence feature to join a session.
    SCENE_COMMUNICATE,            //!<  In this state, communication with peers occurs within the session.
    SCENE_ERROR,                  //!<  This is the state that is transitioned to when problems occur. This demo simulates the type of error messages found in an actual application window.
    SCENE_NETWORK_ERROR,          //!<  This is the state showing network error codes. Assume that the error viewer or EULA applet will normally start in this state.
    SCENE_RESET,                  //!<  Defines the state where all information managed by the demo has been reset to the default values.
    SCENE_DUMMY_SESSION,          //!<  This is the state for building, destroying, and updating dummy sessions for debugging.
    SCENE_EXIT                    //!<  Exits the demo.
};

/*!
    @brief  An enumerated type indicating the network type.
 */
enum NetworkType
{
    TYPE_LOCAL,    //!<  Indicates local (UDS) communication.
    TYPE_INTERNET  //!<  Indicates Internet communication.
};

/*!
    @brief  The enumerated type indicating how to construct the session.
 */
enum MatchmakeType
{
    MATCHMAKE_RANDOM,  //!<  Uses the random matchmaking feature.
    MATCHMAKE_BROWSE,  //!<  Creates a session to join, or searches for a created session and joins it.
    MATCHMAKE_FRIENDS  //!<  Uses the friend presence feature to join a session that is owned by a friend.
};

/*!
    @brief  Enumerates communication errors handled by the sample program.
 */
enum SampleErrorType
{
    UNKNOWN = 0,                     //!<  Unknown error.
    INITIALIZE_ERROR_NETWORK_IF_UP,  //!<  The network interface failed to launch. (For CTR systems, this includes when wireless mode is turned off.)
    INITIALIZE_ERROR_OTHER_REASON,   //!<  Specifies that initialization failed for some other reason.
    LOGIN_ERROR,                     //!<  Specifies that the session could not be established.
    CONNECTION_LOST_ERROR,           //!<  Specifies that the network connection was disconnected.
    SESSION_LOST_ERROR,              //!<  Specifies that the session was disconnected during communication.
    SESSION_STARTUP_ERROR,
    SESSION_CREATE_ERROR,
    SESSION_BROWSE_ERROR,
    SESSION_JOIN_ERROR,
    SESSION_LEAVE_ERROR,
    SESSION_DESTROY_ERROR
};

enum MatchmakeCommand
{
    MC_NONE = 0,
    MC_CREATE,
    MC_BROWSE,
    MC_BROWSE_OWNER,
    MC_BROWSE_PARTICIPANT,
    MC_JOIN,
    MC_REJOIN,
    MC_BACK,
    MC_JOIN_PERSISTENT_GATHERING
};


void SetCommunicationErrorType(SampleErrorType error);  //Sets the communication error type.
void SetNetworkErrorResult(nn::Result result);          //This function sets the <tt>Result</tt> value to be converted into a network error code.
SampleSceneNo GetNextSceneByErrorResult(nn::Result result, SampleErrorType errorType);
SampleSceneNo SetGameMode(u32 mode, u32 type, bool isJoinable, const wchar_t* pDescription);
void ResetGameMode();
void UpdateGameMode();

u32 GetIncrementalGameModeValueByBitRateCheckMode();

void Dispatch();

void SetUnreliableProtocolHandle(u32 handle);
u32 GetUnreliableProtocolHandle();
void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId);

wchar_t* GetSystemPassword();
size_t GetSystemPasswordLength();
u32 GetSessionId();
void ResetRejoinSettings();

/* ---< State by process. >--- */
SampleSceneNo SceneStart();
SampleSceneNo ScenePrepareInternet();
SampleSceneNo ScenePrepareLocal();
SampleSceneNo SceneDummySession();
SampleSceneNo SceneSelectMatchmakeType();
SampleSceneNo SceneRandomMatchmake();
SampleSceneNo SceneBrowseMatchmake();
SampleSceneNo SceneFriendsMatchmake();
SampleSceneNo SceneCommunication();
SampleSceneNo SceneError();
SampleSceneNo SceneNetworkError();
SampleSceneNo SceneReset();

// Search for Communities
u32 FindPersistentGatheringId(u32 targetPid, nn::nex::MatchmakeExtensionClient* pClient);

// In the scene where communication occurs with another party in the session, this is the setting for the interval [frame] to send the local host information.
// When set to 0, it sends in every game frame.
const u32 SAMPLE_INTERVAL_SENDING_NODE_STATUS = 0;

/* Session-related. */
const u8 SAMPLE_MAX_ENTRY = 8;  //!<  Maximum number of connections. In an actual application, set according to the application's specifications.
const u8 SESSION_LIST_MAX = 5;  //!<  The maximum number of sessions to display.
const u8 FRIENDS_LIST_MAX = 5;  //!<  The maximum number of friends to display.

const wchar_t SAMPLE_DESCRIPTION_RANDOM[] = L"RandomMatchmake";
const wchar_t SAMPLE_DESCRIPTION_BROWSE[] = L"BrowseMatchmake";
const wchar_t SAMPLE_DESCRIPTION_FRIENDS[] = L"FriendsMatchmake";

const bool ENABLE_SAMPLE_HOST_MIGRATION = true;

/* Thread priority settings. */
#if NN_PIA_CTR
const s32 SEND_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;          //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#elif NN_PIA_CAFE
const s32 SEND_THREAD_PRIORITY = OS_PRIORITY_APP_DEFAULT - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;  //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#endif

/* NEX-Related Settings */
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 35300;  // Game mode.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;
const u32 MATCHMAKE_ATTR_INDEX = 0;     //!<  Index of the attribute to use.

/* Memory-Related Settings */
const u32 MEMORY_BLOCK_SIZE_PIA = 2 * 1024 * 1024;      //!<  Defines the amount of memory allocated to Pia.
const u32 MEMORY_BLOCK_SIZE_NEX = 1024 * 1024;          //!<  Defines the amount of memory allocated to NEX.
const u32 MEMORY_BLOCK_SIZE_NEX_EMERGENCY = 50 * 1024;  //!<  Defines the amount of memory allocated to NEX for emergencies.

#if NN_PIA_CTR
/* Local-related. */
const bit32 SAMPLE_TITLE_ID = 0xFF311;      //!<  Defines the unique ID of this program.
const u8 SAMPLE_SUB_ID = 0x00;              //!<  Defines the sub-ID.
const char SAMPLE_PASSPHRASE[] = __FILE__;  //!<  Defines the passphrase used to generate the encryption key for UDS communication. In an actual application, make sure that the passphrase is complex and sufficiently long.
#endif

/* Friend-related */
const u32 FRIENDS_MATCHMAKE_GAME_MODE = 0;
const u64 FRIENDS_MATCHMAKE_GAME_MODE_MASK = 0x0000000000000001;
#if NN_PIA_CTR
typedef nn::friends::FriendKey FriendId;
typedef nn::friends::FriendPresence FriendPresence;
typedef nn::friends::GameMode FriendGameMode;

const u32 FRIEND_JOIN_AVAILABILITY_JOINABLE = nn::friends::JOIN_AVAILABILITY_JOINABLE;
const u32 FRIEND_JOIN_AVAILABILITY_NOT_JOINABLE = nn::friends::JOIN_AVAILABILITY_NOT_JOINABLE;
#elif NN_PIA_CAFE
typedef nn::fp::PrincipalId FriendId;
typedef nn::fp::FriendPresence FriendPresence;
typedef nn::fp::GameMode FriendGameMode;

const u32 FRIEND_JOIN_AVAILABILITY_JOINABLE = nn::fp::JOIN_AVAILABILITY_JOINABLE;
const u32 FRIEND_JOIN_AVAILABILITY_NOT_JOINABLE = nn::fp::JOIN_AVAILABILITY_NOT_JOINABLE;
#endif
const wchar_t SAMPLE_USER_PASSWORD[] = L"password";
const size_t SAMPLE_USER_PASSWORD_LENGTH = 8;
extern bool g_IsFromFriendList;
extern FriendId g_FriendIdForJoinIn;

// Event handler for handling notifications dispatched to friends who are logged in to the same game server.
class SampleNexNofiticationEventHandler : public nn::nex::NotificationEventHandler
{
public:
    SampleNexNofiticationEventHandler()
    {
    }
    virtual ~SampleNexNofiticationEventHandler()
    {
    }

    virtual void ProcessNotificationEvent(const nn::nex::NotificationEvent& oEvent)
    {
        NN_PIA_DUMMY_PARAM(oEvent);
        PIA_SAMPLE_LOG("Received notification event. Type:%u, Param1:%d, Param2:%d\n",
            oEvent.GetType(), oEvent.GetParam1(), oEvent.GetParam2());
    }
};

extern SampleNexNofiticationEventHandler g_NexNotificationHandler;


/* Drawing-Related Settings */
const size_t UPPER_SCREEN_WIDTH = 400;   //!<  Defines the horizontal resolution of the upper screen.
const size_t UPPER_SCREEN_HEIGHT = 240;  //!<  Defines the vertical resolution of the upper screen.

const size_t LOWER_SCREEN_WIDTH = 320;   //!<  Defines the horizontal resolution of the lower screen.
const size_t LOWER_SCREEN_HEIGHT = 240;  //!<  Defines the vertical resolution of the lower screen.

const int FONT_SIZE = 8;  //!<  Defines the size of the font to draw.

const size_t SAMPLE_USER_NAME_LENGTH = 11;  //!<  The maximum number of characters for <tt>UserName</tt> drawing, including the terminating character.

const char* GetPiaVersionString();  //!<  Function for getting the string for drawing the Pia library version information.

extern u32 g_FrameNo;
extern NetworkType g_NetworkType;
extern bool g_IsEnableHostMigration;
extern bool g_IsEnableBitRateCheck;

extern u32 g_JoinedPersistentGathering;

void piaTestCallback();

const int SAMPLE_PROGRAMPARAM_GETJOINSEQUENCE = 101;
const int SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT = 102;
const int SAMPLE_PROGRAMPARAM_ISCOMPLETED_CLOSEPARTICIPATION = 103;
const int SAMPLE_PROGRAMPARAM_HOST_MIGRATION_STATE = 110;
