﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneBrowseMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/inet.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/session.h>

#include "SampleConfig.h"
#include "SampleLib/SampleLib.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"

#include "Shared_Nex_Setup.h"

namespace
{
// Array for storing the search results history.
nn::pia::inet::NexSessionInfo s_NexSessionInfoBufferArray[SESSION_LIST_MAX];
u32 s_NexSessionIdBufferArray[SESSION_LIST_MAX];
#if NN_PIA_CTR
nn::pia::local::UdsSessionInfo s_UdsSessionInfoBufferArray[SESSION_LIST_MAX];
#endif
// Number of history entries retained.
u8 s_SessionInfoHistoryNum = 0;
// Whether the SessionId is used as a history (for Internet communication only).
bool s_IsSessionIdHistoryUsed = false;
// Whether the SessionInfo is used as a history (for Internet communication and local communication).
bool s_IsSessionInfoHistoryUsed = false;


const u32 SAMPLE_APPLICATION_DATA_BUFFER_SIZE = 4;
u8 s_applicationDataBuffer[SAMPLE_APPLICATION_DATA_BUFFER_SIZE];
#if NN_PIA_CTR
const char SIGNATURE_KEY_DATA[] = "signature";
struct SampleBeaconApplicationData
{
    char programName[64];
    char playerName[nn::pia::transport::Station::PLAYER_NAME_MAX_DATA_SIZE];
};

// A shared process when getting UserName.
void CheckAndGetUserName(char userNameDst[], const nn::cfg::UserName& userNameSrc, size_t dstMaxSize)
{
    if (userNameSrc.isNgUserName)
    {
        // Profanity
        ::std::strncpy(userNameDst, "NG_WORD", dstMaxSize);
    }
    else
    {
        // Convert UserName from wchar_t to char for display on the screen.
        // The PiaSampleFramework does not support the display of wide characters.
        int ret = ::std::wcstombs(userNameDst, userNameSrc.userName, dstMaxSize);
        if (ret == -1)
        {
            // Conversion failed.
            ::std::strncpy(userNameDst, "NOT ASCII", dstMaxSize);
        }
    }

    // For safety, explicitly insert a terminating character.
    userNameDst[dstMaxSize - 1] = '\0';
}
#endif
#if PIA_TEST_CODE
void PiaTestCallbackForJoinSession()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    PIA_ASSERT(PIA_IS_VALID_POINTER(pSampleFramework));

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

    pSampleFramework->SetProgramParam( SAMPLE_PROGRAMPARAM_GETJOINSEQUENCE, pSession->GetJoinSequence() );
}
#endif
}


SampleSceneNo SceneBrowseMatchmakeLoop(
    nn::pia::session::CreateSessionSetting* pCreateSessionSetting,
    nn::pia::session::JoinSessionSetting* pJoinSessionSetting,
    nn::pia::session::SessionSearchCriteria* pSearchCriteria)
{
    SampleSceneNo nextSceneNo = SCENE_BROWSE_MATCHMAKE;
    nn::Result result;

    while (nextSceneNo == SCENE_BROWSE_MATCHMAKE)
    {
        u32 targetSession = 0;
        MatchmakeCommand command = MC_NONE;

        // Rejoins the session.
        if (GetSessionId() != nn::pia::session::INVALID_SESSION_ID)
        {
            // Clears the information for rejoining.
            ResetRejoinSettings();

            // Rejoining with a system password is only done using Internet communication.
            if (g_NetworkType == TYPE_INTERNET)
            {
                command = MC_JOIN;
            }
        }

        while (command == MC_NONE)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

            // Up/Down Buttons: Select a session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                targetSession = (SESSION_LIST_MAX + targetSession - 1) % SESSION_LIST_MAX;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                targetSession = (targetSession + 1) % SESSION_LIST_MAX;
            }
            // X Button: Create a new session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                command = MC_CREATE;
            }
            // Y Button: Start a session search.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                command = MC_BROWSE;
            }
#if PIA_TEST_CODE
            // Left/Right Buttons: Packet loss emulation.
            nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pThreadStreamManager));
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                pThreadStreamManager->SetSendThreadPacketLossRatio(100);
                pThreadStreamManager->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_LOG("set PacketLossRatio to 100.\n");
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                pThreadStreamManager->SetSendThreadPacketLossRatio(0);
                pThreadStreamManager->SetReceiveThreadPacketLossRatio(0);
                PIA_SAMPLE_LOG("set PacketLossRatio to 0.\n");
            }
#else
            if (g_NetworkType == TYPE_INTERNET)
            {
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
                {
                    command = MC_JOIN_PERSISTENT_GATHERING;
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R))
                {
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
                    {
                        command = MC_BROWSE_OWNER;
                    }
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
                    {
                        command = MC_BROWSE_PARTICIPANT;
                    }
                }
            }
#endif
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                // Whether to use the search results history.
                s_IsSessionInfoHistoryUsed = (s_IsSessionInfoHistoryUsed) ? false : true;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT) && g_NetworkType == TYPE_INTERNET)
            {
                s_IsSessionIdHistoryUsed = (s_IsSessionIdHistoryUsed) ? false : true;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
            {
                // Clears the search results history.
                for (int i=0; i<SESSION_LIST_MAX; i++)
                {
                    s_NexSessionInfoBufferArray[i].Clear();
                    s_NexSessionIdBufferArray[i] = nn::pia::session::INVALID_SESSION_ID;
#if NN_PIA_CTR
                    s_UdsSessionInfoBufferArray[i].Clear();
#endif
                }
                s_SessionInfoHistoryNum = 0;
            }
            // B Button: Reset.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                command = MC_BACK;
            }
            // A Button: Joins the selected session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                // For Internet communication only, it is possible to retain the session ID as search results history and specify it when joining a session.
                if (s_IsSessionIdHistoryUsed && g_NetworkType == TYPE_INTERNET)
                {
                    for (int i = 0; i < s_SessionInfoHistoryNum; i++)
                    {
                        if (i == targetSession && s_NexSessionIdBufferArray[i] != nn::pia::session::INVALID_SESSION_ID)
                        {
                            static_cast<nn::pia::inet::NexJoinSessionSetting*>(pJoinSessionSetting)->SetSessionId(s_NexSessionIdBufferArray[i]);
                            command = MC_JOIN;
                            break;
                        }
                    }
                }
                else
                {
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = NULL;
                    // When joining using the search results history.
                    if (s_IsSessionInfoHistoryUsed)
                    {
                        for (int i = 0; i < s_SessionInfoHistoryNum; i++)
                        {
                            if (i == targetSession)
                            {
#if NN_PIA_CTR
                                if (g_NetworkType == TYPE_LOCAL)
                                {
                                    pTargetSessionInfo = &s_UdsSessionInfoBufferArray[i];
                                }
                                else
#endif
                                {
                                    pTargetSessionInfo = &s_NexSessionInfoBufferArray[i];
                                }
                                break;
                            }
                        }
                    }
                    // When joining using the search results of the previous search.
                    else
                    {
                        nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                        size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                        for (int i = 0; i < foundSessionNum; i++)
                        {
                            if (i == targetSession)
                            {
                                pTargetSessionInfo = *it;
                                break;
                            }
                            ++it;
                        }
                    }
                    if (pTargetSessionInfo)
                    {
                        pJoinSessionSetting->SetSessionInfoPtr(pTargetSessionInfo);
                        command = MC_JOIN;
                    }
                }
            }

#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
                1,
                nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize());
#endif

            // Lower screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %d", PROGRAM_NAME, g_FrameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: BrowseMatchmake.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::session::ISessionInfoList* sessionInfoList = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList();
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Session List ] Found:%01d ", sessionInfoList->GetSize());
            line += FONT_SIZE;
            nn::pia::session::ISessionInfoList::Iterator it = sessionInfoList->Begin();

            for (int i = 0; i < SESSION_LIST_MAX; i++)
            {
                if (s_IsSessionIdHistoryUsed && g_NetworkType == TYPE_INTERNET)
                {
                    if (i == targetSession)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                    }
                    else
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                    }
                    if (i < s_SessionInfoHistoryNum)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " GID=%d", s_NexSessionIdBufferArray[i]);
                    }
                    else
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  -----");
                    }
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                }
                else
                {
                    nn::pia::session::ISessionInfo* pInfo = NULL;
                    if (s_IsSessionInfoHistoryUsed)
                    {
                        if (i < s_SessionInfoHistoryNum)
                        {
#if NN_PIA_CTR
                            if (g_NetworkType == TYPE_LOCAL)
                            {
                                pInfo = &s_UdsSessionInfoBufferArray[i];
                            }
                            else
#endif
                            {
                                pInfo = &s_NexSessionInfoBufferArray[i];
                            }
                        }
                    }
                    else
                    {
                        if (it != sessionInfoList->End())
                        {
                            pInfo = (*it);
                        }
                    }
                    if (pInfo != NULL)
                    {
                        if (i == targetSession)
                        {
                            if (g_NetworkType == TYPE_INTERNET)
                            {
                                nn::pia::inet::NexSessionInfo* pTargetNexSessionInfo = static_cast<nn::pia::inet::NexSessionInfo*>(pInfo);
                                u8 applicationData[SAMPLE_APPLICATION_DATA_BUFFER_SIZE];
                                size_t appDataSize = pTargetNexSessionInfo->GetApplicationDataSize();
                                size_t size = (appDataSize < sizeof(applicationData)) ? appDataSize : sizeof(applicationData);
                                result = pTargetNexSessionInfo->GetApplicationDataBuf(applicationData, size);
                                if (result.IsFailure())
                                {
                                    memset(applicationData, 0, sizeof(applicationData));
                                }
                                u32 attribute;
                                result = pTargetNexSessionInfo->GetAttribute(&attribute, MATCHMAKE_ATTR_INDEX);
                                if (result.IsFailure())
                                {
                                    attribute = 0;
                                }

                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "%sMode=%d GID=%d [%d/%d-%d][%s]",
#if (NEX_CURRENT_VERSION_NUMBER >= NEX_VERSION_NUMBER(3, 8, 0))
                                    pTargetNexSessionInfo->IsRestrictedByUserPassword() ? "*" : " ",
#else
                                    " ",
#endif
                                    pTargetNexSessionInfo->GetGameMode(),
                                    pTargetNexSessionInfo->GetSessionId(),
                                    pTargetNexSessionInfo->GetCurrentParticipants(),
                                    pTargetNexSessionInfo->GetMinParticipants(),
                                    pTargetNexSessionInfo->GetMaxParticipants(),
                                    pTargetNexSessionInfo->IsOpened() ? "Opened" : "Closed");
                                line += FONT_SIZE;

                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 16),
                                    "Type=%d Host=%d Target=%d",
                                    pTargetNexSessionInfo->GetSessionType(),
                                    pTargetNexSessionInfo->GetHostPrincipalId(),
                                    pTargetNexSessionInfo->GetTargetPrincipalId());
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 15),
                                    "Date:%04d/%02d/%02d-%02d:%02d:%02d Prior=%d",
                                    pTargetNexSessionInfo->GetStartedTime().GetYear(),
                                    pTargetNexSessionInfo->GetStartedTime().GetMonth(),
                                    pTargetNexSessionInfo->GetStartedTime().GetDay(),
                                    pTargetNexSessionInfo->GetStartedTime().GetHour(),
                                    pTargetNexSessionInfo->GetStartedTime().GetMinute(),
                                    pTargetNexSessionInfo->GetStartedTime().GetSecond(),
                                    pTargetNexSessionInfo->GetSelectionPriority());
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 14),
                                    "ApplicationData = %d %d %d %d", applicationData[0], applicationData[1], applicationData[2], applicationData[3]);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 13),
                                    "Attribute[%d] = %d", MATCHMAKE_ATTR_INDEX, attribute);
                            }
#if NN_PIA_CTR
                            else if (g_NetworkType == TYPE_LOCAL)
                            {
                                nn::pia::local::LocalSessionInfo* pTargetLocalSessionInfo = static_cast<nn::pia::local::LocalSessionInfo*>(pInfo);

                                // Gets the session's signal strength.
                                u8 radioStrength = 0;
                                result = pTargetLocalSessionInfo->GetRadioStrength(&radioStrength);
                                if (result.IsFailure())
                                {
                                    radioStrength = 0;
                                }

                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Mode=%d GID=%d [%d/%d][%s][%d]",
                                    pTargetLocalSessionInfo->GetGameMode(),
                                    pTargetLocalSessionInfo->GetSessionId(),
                                    pTargetLocalSessionInfo->GetCurrentParticipants(),
                                    pTargetLocalSessionInfo->GetMaxParticipants(),
                                    pTargetLocalSessionInfo->IsOpened() ? "Opened" : "Closed",
                                    radioStrength);
                                line += FONT_SIZE;

                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

                                // Gets the information for stations currently participating in sessions.
                                nn::pia::local::UdsStationInfo udsStationInfoList[SAMPLE_MAX_ENTRY];
                                nn::pia::local::LocalSessionStationInfo sessionStationInfo;
                                sessionStationInfo.SetStationInfoListBuf(udsStationInfoList, SAMPLE_MAX_ENTRY);
                                result = pTargetLocalSessionInfo->GetSessionStationInfo(&sessionStationInfo);
                                if (result.IsSuccess())
                                {
                                    char userName[nn::cfg::CFG_USER_NAME_LENGTH];
                                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 21), "[ UserName List ]");
                                    for (int j = 0; j < SAMPLE_MAX_ENTRY; j++)
                                    {
                                        if (udsStationInfoList[j].m_Role != nn::pia::local::LocalStationInfo::STATION_INFO_ROLE_NONE)
                                        {
                                            bool isHost = udsStationInfoList[j].m_Role == nn::pia::local::LocalStationInfo::STATION_INFO_ROLE_HOST;
                                            CheckAndGetUserName(userName, udsStationInfoList[j].m_UserName, nn::cfg::CFG_USER_NAME_LENGTH);
                                            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * (20 - j)),
                                                " %-6s:%s\n", (isHost ? "Host" : "Client"), userName);
                                            NN_PIA_DUMMY_PARAM_IN_RELEASE(isHost);
                                        }
                                    }
                                }

                                // Gets the BSSID for the session to join.
                                bit8 bssidBuf[nn::uds::MAC_ADDRESS_SIZE];
                                result = pTargetLocalSessionInfo->GetBssid(bssidBuf);
                                if (result.IsFailure())
                                {
                                    for (int j = 0; j < nn::uds::MAC_ADDRESS_SIZE; j++)
                                    {
                                        bssidBuf[j] = 0;
                                    }
                                }

                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 11),
                                    " Bssid=%02x:%02x:%02x:%02x:%02x:%02x", bssidBuf[0], bssidBuf[1], bssidBuf[2], bssidBuf[3], bssidBuf[4], bssidBuf[5]);

                                // Gets the beacon data for the session.
                                SampleBeaconApplicationData beaconApplicationData;
                                size_t appDataSize = 0;
                                result = pTargetLocalSessionInfo->GetApplicationDataSize(&appDataSize);
                                if (result.IsFailure())
                                {
                                    appDataSize = 0;
                                }
                                PIA_SAMPLE_ASSERT(appDataSize == sizeof(beaconApplicationData));
                                result = pTargetLocalSessionInfo->GetApplicationDataBuf(&beaconApplicationData, sizeof(beaconApplicationData));
                                if (result.IsFailure())
                                {
                                    beaconApplicationData.playerName[0] = '\0';
                                    beaconApplicationData.programName[0] = '\0';
                                }

                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 10),
                                    " Title=%s", beaconApplicationData.programName);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, LOWER_SCREEN_HEIGHT - (FONT_SIZE * 9),
                                    " PlayerName(Host)=%s", beaconApplicationData.playerName);
                            }
#endif
                        }
                        else
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Mode=%d GID=%d [%d/%d][%s]",
                                pInfo->GetGameMode(),
                                pInfo->GetSessionId(),
                                pInfo->GetCurrentParticipants(),
                                pInfo->GetMaxParticipants(),
                                pInfo->IsOpened() ? "Opened" : "Closed");
                            line += FONT_SIZE;
                        }
                        ++it;
                    }
                    else
                    {
                        if (i == targetSession)
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                        }
                        else
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                        }
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  -----");
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                        line += FONT_SIZE;
                    }
                }
            }
            if (g_NetworkType == TYPE_INTERNET)
            {
                line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 12);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "Persistent Gathering Id: %d", g_JoinedPersistentGathering);
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Select Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  LEFT : Use SessionInfoHistory toggle");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  RIGHT: Use SessionIdHistory toggle");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Create New Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Browse Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R+X: Browse Session By Owner");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R+Y: Browse Session By Participant");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Join this Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Back");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  L: Create/Join PersistentGathering");
                line += FONT_SIZE;
            }
            else
            {
                line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 8);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Select Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  LEFT : Use SessionInfoHistory toggle");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Create New Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Browse Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Join this Session");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Back");
                line += FONT_SIZE;
            }

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            ++g_FrameNo;
            Dispatch();

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

            // Checks the connection status.
            result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
            if (result.IsFailure())
            {
                SampleErrorType errorType = (result == nn::pia::ResultSessionConnectionIsLost()) ? SESSION_LOST_ERROR : CONNECTION_LOST_ERROR;
                return GetNextSceneByErrorResult(result, errorType);
            }

        }  // while (command == MC_NONE)

        if (command == MC_CREATE)
        {
            SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_CreateSession(pCreateSessionSetting, &result);
            if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                PIA_SAMPLE_LOG_RESULT(result);
                nextSceneNo = GetNextSceneByErrorResult(result, SESSION_CREATE_ERROR);
            }
            else
            {
                nextSceneNo = SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode(),
                                          nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY,
                                          false,
                                          SAMPLE_DESCRIPTION_BROWSE);
            }
        }
        else if (command == MC_BROWSE)
        {
            nn::pia::session::SessionSearchCriteria* pTargetSearchCriteria = pSearchCriteria;
            SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_BrowseSession(pTargetSearchCriteria, &result);
            if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                PIA_SAMPLE_LOG_RESULT(result);
                nextSceneNo = GetNextSceneByErrorResult(result, SESSION_BROWSE_ERROR);
            }
            else
            {
                nn::pia::session::ISessionInfoList* sessionInfoList = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList();
                for (nn::pia::session::ISessionInfoList::Iterator it = sessionInfoList->Begin(); it != sessionInfoList->End(); ++it)
                {
                    if (s_SessionInfoHistoryNum < SESSION_LIST_MAX)
                    {
#if NN_PIA_CTR
                        if (g_NetworkType == TYPE_LOCAL)
                        {
                            nn::pia::local::UdsSessionInfo* pInfo = static_cast<nn::pia::local::UdsSessionInfo*>(*it);
                            s_UdsSessionInfoBufferArray[s_SessionInfoHistoryNum].Copy(*pInfo);
                        }
                        else
#endif
                        {
                            nn::pia::inet::NexSessionInfo* pInfo = static_cast<nn::pia::inet::NexSessionInfo*>(*it);
                            s_NexSessionInfoBufferArray[s_SessionInfoHistoryNum].Copy(*pInfo);
                            s_NexSessionIdBufferArray[s_SessionInfoHistoryNum] = pInfo->GetSessionId();
                        }
                        ++s_SessionInfoHistoryNum;
                    }
                }
            }
        }
        if (command == MC_BROWSE_OWNER)
        {
#if NN_PIA_CTR
            const u32 TargetPrincipalId = 1800164835;
#elif NN_PIA_CAFE
            const u32 TargetPrincipalId = 1799852706;
#else
            const u32 TargetPrincipalId = 0;
#endif
            // Specifies the owner of a matchmaking session and performs a search. Change the specified principle ID to an appropriate value.
            nn::pia::inet::NexSessionSearchCriteriaOwner searchCriteria;
            searchCriteria.SetOwnerPrincipalId(TargetPrincipalId);
            searchCriteria.SetResultRange(0, SESSION_LIST_MAX);

            SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_BrowseSession(&searchCriteria, &result);
            if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                PIA_SAMPLE_LOG_RESULT(result);
                nextSceneNo = GetNextSceneByErrorResult(result, SESSION_BROWSE_ERROR);
            }
            else
            {
                nn::pia::session::ISessionInfoList* sessionInfoList = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList();
                for (nn::pia::session::ISessionInfoList::Iterator it = sessionInfoList->Begin(); it != sessionInfoList->End(); ++it)
                {
                    if (s_SessionInfoHistoryNum < SESSION_LIST_MAX)
                    {
                        nn::pia::inet::NexSessionInfo* pInfo = static_cast<nn::pia::inet::NexSessionInfo*>(*it);
                        s_NexSessionInfoBufferArray[s_SessionInfoHistoryNum].Copy(*pInfo);
                        s_NexSessionIdBufferArray[s_SessionInfoHistoryNum] = pInfo->GetSessionId();
                        ++s_SessionInfoHistoryNum;
                    }
                }
            }
        }
        if (command == MC_BROWSE_PARTICIPANT)
        {
#if NN_PIA_CTR
            const u32 TargetPrincipalId[3] = { 1800164835, 1800164834, 1800164809 };
#elif NN_PIA_CAFE
            const u32 TargetPrincipalId[3] = { 1799852706, 1799916875, 1799916873 };
#else
            const u32 TargetPrincipalId[3] = { 0, 0, 0 };
#endif
            // Specifies a user and performs a search. Change the specified principle ID to an appropriate value.
            nn::pia::inet::NexSessionSearchCriteriaParticipant searchCriteria;
            searchCriteria.SetParticipantPrincipalId(TargetPrincipalId, 3);

            SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_BrowseSession(&searchCriteria, &result);
            if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                PIA_SAMPLE_LOG_RESULT(result);
                nextSceneNo = GetNextSceneByErrorResult(result, SESSION_BROWSE_ERROR);
            }
            else
            {
                nn::pia::session::ISessionInfoList* sessionInfoList = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList();
                for (nn::pia::session::ISessionInfoList::Iterator it = sessionInfoList->Begin(); it != sessionInfoList->End(); ++it)
                {
                    if (s_SessionInfoHistoryNum < SESSION_LIST_MAX)
                    {
                        nn::pia::inet::NexSessionInfo* pInfo = static_cast<nn::pia::inet::NexSessionInfo*>(*it);
                        s_NexSessionInfoBufferArray[s_SessionInfoHistoryNum].Copy(*pInfo);
                        s_NexSessionIdBufferArray[s_SessionInfoHistoryNum] = pInfo->GetSessionId();
                        ++s_SessionInfoHistoryNum;
                    }
                }
            }
        }
        else if (command == MC_JOIN)
        {
#if PIA_TEST_CODE
            RegisterPiaTestCallback(PiaTestCallbackForJoinSession);
#endif
            SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinSession(pJoinSessionSetting, &result);
            if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                PIA_SAMPLE_LOG_RESULT(result);
                if (sampleSessionAsyncStatus == SAMPLE_SESSION_ASYNC_CANCELED)
                {
                    nextSceneNo = SCENE_SELECT_MATCHMAKE_TYPE;
                    Session_CleanupSession();
                }
                else
                {
                    nextSceneNo = GetNextSceneByErrorResult(result, SESSION_JOIN_ERROR);
                }
            }
            else
            {
                nextSceneNo = SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode(),
                                          nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY,
                                          false,
                                          SAMPLE_DESCRIPTION_BROWSE);
            }
#if PIA_TEST_CODE
            RegisterPiaTestCallback(NULL);
#endif
        }
        else if (command == MC_BACK)
        {
#if NN_PIA_CTR
            if (g_NetworkType == TYPE_LOCAL)
            {
                nextSceneNo = SCENE_RESET;
            }
            else
#endif
            {
                nextSceneNo = SCENE_SELECT_MATCHMAKE_TYPE;
                Session_CleanupSession();
            }
        }
#if ENABLE_SERVERLESS
#else
        else if (g_NetworkType != TYPE_LOCAL)
        {
            if (command == MC_JOIN_PERSISTENT_GATHERING)
            {
                // The target community is the one in which the specified PID is participating.
                // Change the PID value to the PID of the device you are using.
                const size_t targetPidListSize = 4;
                u32 targetPidList[targetPidListSize] = {1111111110, 1111111111, 1111111112, 1111111113};

                nn::nex::MatchmakeExtensionClient* pClient = Nex_GetMatchmakeExtensionClientInstance();
                // Searching

                // First find the community the local station is participating in.
                u32 localPid = pClient->GetCredentials()->GetPrincipalID();
                nn::nex::GatheringID joinedPersistentGatheringId = FindPersistentGatheringId(localPid, pClient);

                // If participating in the community.
                if (joinedPersistentGatheringId != 0)
                {
                    g_JoinedPersistentGathering = joinedPersistentGatheringId;

                }
                // If not participating in the community.
                else
                {
                    // Search for communities that other PIDs are participating in.
                    bool foundPersistentGathering = false;
                    for (int i = 0; i < targetPidListSize; i++)
                    {
                        if(targetPidList[i] != localPid)
                        {
                            joinedPersistentGatheringId = FindPersistentGatheringId(targetPidList[i], pClient);
                        }

                        if (joinedPersistentGatheringId != 0)
                        {
                            foundPersistentGathering = true;
                            break;
                        }
                    }

                    nn::nex::ProtocolCallContext oContext;
                    // Join if there is a hit.
                    if (foundPersistentGathering)
                    {
                        if(pClient->JoinPersistentGathering(&oContext, joinedPersistentGatheringId))
                        {
                            oContext.Wait();
                            if (oContext.GetOutcome())
                            {
                                g_JoinedPersistentGathering = joinedPersistentGatheringId;
                                PIA_SAMPLE_LOG("Join persistent gathering success. (targetPId: %d)\n", joinedPersistentGatheringId);
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Join persistent callcontext return false. (targetPid: %d)\n", joinedPersistentGatheringId);
                            }
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Join persistent call failure. (targetPid: %d)\n", joinedPersistentGatheringId);
                        }
                    }
                    // Create if not.
                    else
                    {
                        // Community parameter settings
                        nn::nex::PersistentGathering persistentGathering;
                        persistentGathering.SetPersistentGatheringType(nn::nex::PersistentGatheringType::Open);
                        persistentGathering.SetMaxParticipants(0);
                        persistentGathering.SetParticipationStartDate(nn::nex::DateTime::Never);
                        persistentGathering.SetParticipationEndDate(nn::nex::DateTime::Never);

                        if (pClient->CreatePersistentGathering(&oContext, persistentGathering, &joinedPersistentGatheringId))
                        {
                            oContext.Wait();

                            if(oContext.GetOutcome())
                            {
                                g_JoinedPersistentGathering = joinedPersistentGatheringId;
                                PIA_SAMPLE_LOG("Create persistent gathering success. (created Gid: %d)\n", joinedPersistentGatheringId);
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("Create persistent callcontext return false. (targetPid: %d)\n", joinedPersistentGatheringId);
                            }
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("Create persistent call failure. (targetPid: %d)\n", joinedPersistentGatheringId);
                        }
                    }
                }
                nn::pia::inet::NexCreateSessionSetting* pNexCreateSessionSetting = reinterpret_cast<nn::pia::inet::NexCreateSessionSetting*>(pCreateSessionSetting);
                pNexCreateSessionSetting->SetCommunityId(g_JoinedPersistentGathering);
                nn::pia::inet::NexSessionSearchCriteria* pNexSearchCriteria = reinterpret_cast<nn::pia::inet::NexSessionSearchCriteria*>(pSearchCriteria);
                pNexSearchCriteria->SetCommunityId(g_JoinedPersistentGathering);
                pNexSearchCriteria++;
                pNexSearchCriteria->SetCommunityId(g_JoinedPersistentGathering);
            }
        }
#endif
    }  // while (nextSceneNo == SCENE_BROWSE_MATCHMAKE)

    return nextSceneNo;
}

u32 FindPersistentGatheringId(u32 targetPid, nn::nex::MatchmakeExtensionClient* pClient)
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::qList<nn::nex::PersistentGathering> myPersistentGatheringList;

    nn::nex::GatheringID joinedPersistentGatheringId = 0;

    if(pClient->FindPersistentGatheringByParticipant(&oContext, targetPid, nn::nex::ResultRange(), &myPersistentGatheringList))
    {
        oContext.Wait();
        if(oContext.GetOutcome())
        {
            if(!myPersistentGatheringList.empty())
            {
                nn::nex::qList<nn::nex::PersistentGathering>::const_iterator it = myPersistentGatheringList.begin();

                joinedPersistentGatheringId = myPersistentGatheringList.front().GetID();
                PIA_SAMPLE_LOG("Found joined persistent gathering: %d\n", joinedPersistentGatheringId);

                while(it != myPersistentGatheringList.end())
                {
                    PIA_SAMPLE_LOG("-------------------------------------------\n");
                    PIA_SAMPLE_LOG("GatheringID : %u\n", it->GetID());
                    PIA_SAMPLE_LOG("  PersistentGatheringCode : %llu\n", nn::nex::MatchmakeExtensionClient::GatheringIdToPersistentGatheringCode(it->GetID()));
                    PIA_SAMPLE_LOG("  PersistentGatheringType : %u\n", it->GetPersistentGatheringType());
                    PIA_SAMPLE_LOG("  Owner : %u\n", it->GetOwnerPID());
                    PIA_SAMPLE_LOG("  MaxParticipants : %u\n", it->GetMaxParticipants());
                    PIA_SAMPLE_LOG("  Attribute_0 : %u\n", it->GetAttribute(0));
                    PIA_SAMPLE_LOG("  Attribute_1 : %u\n", it->GetAttribute(1));
                    PIA_SAMPLE_LOG("  Attribute_2 : %u\n", it->GetAttribute(2));
                    PIA_SAMPLE_LOG("  Attribute_3 : %u\n", it->GetAttribute(3));
                    PIA_SAMPLE_LOG("  Attribute_4 : %u\n", it->GetAttribute(4));
                    PIA_SAMPLE_LOG("  Attribute_5 : %u\n", it->GetAttribute(5));
                    PIA_SAMPLE_LOG("  Participation Count : %u\n", it->GetParticipationCount());
                    PIA_SAMPLE_LOG("  MatchmakeSession Count : %u\n", it->GetMatchmakeSessionCount());
                    PIA_SAMPLE_LOG("-------------------------------------------\n");

                    ++it;
                }

                return joinedPersistentGatheringId;
            }
            else
            {
                PIA_SAMPLE_LOG("Persistent gathering search result is empty. (targetPid: %d)\n", targetPid);
                return 0;
            }
        }
        else
        {
            PIA_SAMPLE_LOG("FindPersistentGatheringByParticipant callContext return false. (targetPid: %d)\n", targetPid);
            return 0;
        }
    }
    else
    {
        PIA_SAMPLE_LOG("FindPersistentGatheringByParticipant call failure. (targetPid: %d)\n", targetPid);
        return 0;
    }
}

SampleSceneNo SceneBrowseMatchmake()
{
    nn::Result result;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_BROWSE_MATCHMAKE,
        false,
        false,
        0);

#if NN_PIA_CTR
    if (g_NetworkType == TYPE_LOCAL)
    {
        // Settings for the session being created.
        nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
        {
            createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
            createNetworkSetting.m_MaxEntry = SAMPLE_MAX_ENTRY;
            createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
            memcpy(createNetworkSetting.m_Passphrase, const_cast<char*>(SAMPLE_PASSPHRASE), sizeof(SAMPLE_PASSPHRASE));
            createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);

            SampleBeaconApplicationData beaconApplicationData;
            ::std::memset(&beaconApplicationData, 0x00, sizeof(beaconApplicationData));
            ::std::memcpy(beaconApplicationData.programName, PROGRAM_NAME, sizeof(PROGRAM_NAME));

            nn::pia::transport::Station::PlayerName playerName;
            Session_GetLocalPlayerName(&playerName);

            int ret = ::std::wcstombs(beaconApplicationData.playerName,
                                      reinterpret_cast<wchar_t*>(playerName.data),
                                      nn::pia::transport::Station::PLAYER_NAME_MAX_DATA_SIZE);
            if (ret == -1)
            {
                // Conversion failed.
                ::std::strncpy(beaconApplicationData.playerName, "NOT ASCII", nn::pia::transport::Station::PLAYER_NAME_MAX_DATA_SIZE);
            }

            memcpy(createNetworkSetting.m_ApplicationData, &beaconApplicationData, sizeof(beaconApplicationData));
            createNetworkSetting.m_ApplicationDataSize = sizeof(beaconApplicationData);
        }
        nn::pia::local::UdsCreateSessionSetting createSessionSetting;
        createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
        result = createSessionSetting.SetSignatureKey(SIGNATURE_KEY_DATA, sizeof(SIGNATURE_KEY_DATA));
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        createSessionSetting.SetMaxParticipants(SAMPLE_MAX_ENTRY);

        // Settings (besides SessionId) for the session to join.
        nn::pia::local::UdsJoinSessionSetting joinSessionSetting;

        result = joinSessionSetting.SetPassphrase(const_cast<char*>(SAMPLE_PASSPHRASE), sizeof(SAMPLE_PASSPHRASE));
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        result = joinSessionSetting.SetSignatureKey(SIGNATURE_KEY_DATA, sizeof(SIGNATURE_KEY_DATA));
        PIA_SAMPLE_ASSERT(result.IsSuccess());

        // Search condition settings
        nn::pia::local::LocalSessionSearchCriteria searchCriteria;

        searchCriteria.SetMaxParticipants(SAMPLE_MAX_ENTRY);
        searchCriteria.SetSubId(SAMPLE_SUB_ID);
        searchCriteria.SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
        searchCriteria.SetVacantOnly(true);
        searchCriteria.SetOpenedOnly(true);
        searchCriteria.SetResultRange(0, SESSION_LIST_MAX);

        return SceneBrowseMatchmakeLoop(&createSessionSetting, &joinSessionSetting, &searchCriteria);
    }
    else
#endif
    {
        // Settings for the matchmaking session to create.
#if ENABLE_SERVERLESS
        nn::pia::inet::LanCreateSessionSetting createSessionSetting;
        createSessionSetting.SetSessionType(nn::pia::inet::LAN_SESSION_TYPE_ANYBODY);
#else
        nn::pia::inet::NexCreateSessionSetting createSessionSetting;
        createSessionSetting.SetSessionType(nn::pia::inet::SESSION_TYPE_ANYBODY);
#endif
        createSessionSetting.SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode());
        createSessionSetting.SetMaxParticipants(SAMPLE_MAX_ENTRY);
        createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX, g_IsEnableHostMigration);
        createSessionSetting.SetOpenParticipation(true);
        createSessionSetting.SetDescription(SAMPLE_DESCRIPTION_BROWSE);
#if ENABLE_SERVERLESS
#else
        // Set the community ID.
        createSessionSetting.SetCommunityId(g_JoinedPersistentGathering);
#endif
        for (int i = 0; i < SAMPLE_APPLICATION_DATA_BUFFER_SIZE; i++)
        {
            s_applicationDataBuffer[i] = i;
        }
        createSessionSetting.SetApplicationData(s_applicationDataBuffer, sizeof(s_applicationDataBuffer));
        // Check the application data that was set.
        {
            u8 checkBuffer[SAMPLE_APPLICATION_DATA_BUFFER_SIZE];
            size_t dataSize;
            createSessionSetting.GetApplicationData(checkBuffer, &dataSize, sizeof(checkBuffer));
            PIA_SAMPLE_ASSERT(dataSize == SAMPLE_APPLICATION_DATA_BUFFER_SIZE);
            PIA_SAMPLE_ASSERT(!std::memcmp(checkBuffer, s_applicationDataBuffer, SAMPLE_APPLICATION_DATA_BUFFER_SIZE));
        }

        // Settings when joining the session.
#if ENABLE_SERVERLESS
        nn::pia::inet::LanJoinSessionSetting joinSessionSetting;
#else
        nn::pia::inet::NexJoinSessionSetting joinSessionSetting;
        if (GetSessionId() != nn::pia::session::INVALID_SESSION_ID)
        {
            joinSessionSetting.SetSessionId(GetSessionId());
            joinSessionSetting.SetSessionSystemPassword(GetSystemPassword(), GetSystemPasswordLength());
        }
#endif

        // Search condition settings
#if ENABLE_SERVERLESS
        nn::pia::inet::LanSessionSearchCriteria searchCriteriaList[2];
#else
        nn::pia::inet::NexSessionSearchCriteria searchCriteriaList[2];
#endif
        // Specifies a single attribute.
        searchCriteriaList[0].SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode());
        searchCriteriaList[0].SetMaxParticipants(SAMPLE_MAX_ENTRY);
        searchCriteriaList[0].SetAttribute(MATCHMAKE_ATTR_INDEX, g_IsEnableHostMigration);
        searchCriteriaList[0].SetResultRange(0, SESSION_LIST_MAX);
        searchCriteriaList[0].SetOpenedOnly(false);
#if ENABLE_SERVERLESS
#else
        searchCriteriaList[0].SetCommunityId(g_JoinedPersistentGathering);
#endif
        // Specifies an attribute range.
        // Example of code for specifying multiple attributes.
        searchCriteriaList[1].SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode());
        searchCriteriaList[1].SetMaxParticipantsRange(SAMPLE_MAX_ENTRY, SAMPLE_MAX_ENTRY);
        searchCriteriaList[1].SetAttributeRange(MATCHMAKE_ATTR_INDEX, 2, 72);
        searchCriteriaList[1].SetResultRange(0, SESSION_LIST_MAX);
#if ENABLE_SERVERLESS
#else
        searchCriteriaList[1].SetCommunityId(g_JoinedPersistentGathering);
#endif
        return SceneBrowseMatchmakeLoop(&createSessionSetting, &joinSessionSetting, searchCriteriaList);
    }
}
