﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneCommunication.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleConfig.h"

#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/inet.h>
#include <pia/transport.h>
#include <pia/session.h>

#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"
#include "SampleLib/SampleLib.h"

namespace
{

struct RectPosition  //!<  Stores rectangle position coordinates.
{
    u16 x;
    u16 y;
};

// Station status information.
struct SampleStationStatus
{
    nn::pia::StationId stationId;
    u8 colorIndex;
    RectPosition pos;
    u32 latestIncrementNumber;
    s32 rtt;
    nn::pia::transport::Station::IdentificationInfo idInfo;

    SampleStationStatus()
    {
        stationId = nn::pia::STATION_ID_INVALID;
        colorIndex = 0;
        pos.x = 0;
        pos.y = 0;
        latestIncrementNumber = 0;
        rtt = 0;
    }

    static size_t GetSerializedSize()
    {
        return sizeof(u64) + sizeof(u8) + sizeof(u16) + sizeof(u16) + sizeof(u32);
    }

    void Serialize(void* pBuf)
    {
        PIA_SAMPLE_ASSERT(pBuf != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pBuf) % 4 == 0);

        u8* p = static_cast<u8*>(pBuf);

        *reinterpret_cast<u64*>(p) = nn::pia::common::ByteOrder::HostToNetwork64(stationId.GetValue());
        p += 8;
        *p = colorIndex;
        p += 1;
        *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.x);
        p += 2;
        *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.y);
        p += 2;
        *reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(latestIncrementNumber);
    }

    void Deserialize(const void* pData)
    {
        PIA_SAMPLE_ASSERT(pData != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pData) % 4 == 0);

        const u8* p = static_cast<const u8*>(pData);

        stationId = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(p));
        p += 8;
        colorIndex = *p;
        p += 1;
        pos.x = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        pos.y = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        latestIncrementNumber = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
    }

    void UpdateInfo()
    {
        // Update RTT.
        nn::pia::transport::Station* pStationBuf = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
        if (PIA_IS_VALID_POINTER(pStationBuf))
        {
            // Gets the RTT from the station instance corresponding to StationId.
            rtt = pStationBuf->GetRtt();
        }
        else
        {
            // Set to an invalid value if you cannot get the station instance.
            rtt = nn::pia::transport::Station::INVALID_RTT;
        }
    }
};

// Array of station statuses.
SampleStationStatus s_StationStatus[SAMPLE_MAX_ENTRY];

// Unreliable protocol handle.
u32 s_UnreliableProtocolHandle = 0;

// Session ID set with a session system password, and the password.
u32 s_SessionId = nn::pia::session::INVALID_SESSION_ID;
wchar_t s_SystemPassword[nn::pia::inet::SESSION_SYSTEM_PASSWORD_LENGTH];

// Rectangle color to associate with station index.
const nn::pia::sampleframework::Color COLOR_OF_RECTANGLE[] =
{
    nn::pia::sampleframework::COLOR_BLACK,
    nn::pia::sampleframework::COLOR_FUCHSIA,
    nn::pia::sampleframework::COLOR_GRAY,
    nn::pia::sampleframework::COLOR_WHITE,
    nn::pia::sampleframework::COLOR_MAROON,
    nn::pia::sampleframework::COLOR_RED,
    nn::pia::sampleframework::COLOR_OLIVE,
    nn::pia::sampleframework::COLOR_YELLOW,
    nn::pia::sampleframework::COLOR_GREEN,
    nn::pia::sampleframework::COLOR_LIME,
    nn::pia::sampleframework::COLOR_TEAL,
    nn::pia::sampleframework::COLOR_NAVY,
    nn::pia::sampleframework::COLOR_PURPLE
};

// The string for displaying the role screen within a session.
const char* SESSION_ROLE[] = {
    "Unknown",
    "Host",
    "Client"
};

// Indicates the open status of the session.
enum SessionOpenStatus
{
    eStatusUnknown,
    eStatusClosed,
    eStatusOpened
};

// The strings for displaying the open status of the session.
const char* SESSION_OPEN_STATUS[] = {
    "Unknown",
    "Closed",
    "Opened"
};

}  // end of anonymous namespace


// Sets the UnreliableProtocolHandle that was obtained when PiaTransport was initialized.
void SetUnreliableProtocolHandle(u32 handle)
{
    s_UnreliableProtocolHandle = handle;
}

u32 GetUnreliableProtocolHandle()
{
    return s_UnreliableProtocolHandle;
}

u16 GetStationStatusArrayIndex(const nn::pia::StationId& stationId)
{
    // First make sure there are no matches.
    for (size_t i = 0; i < SAMPLE_MAX_ENTRY; i++)
    {
        if (s_StationStatus[i].stationId == stationId)
        {
            return static_cast<u16>(i);
        }
    }
    // If no match is found, check for an empty element.
    for (size_t i = 0; i < SAMPLE_MAX_ENTRY; i++)
    {
        if (s_StationStatus[i].stationId == nn::pia::STATION_ID_INVALID)
        {
            return static_cast<u16>(i);
        }
    }
    // Control should never reach here
    return 0xffff;
}

void ActivateStationStatus(const nn::pia::StationId& stationId)
{
    u16 idx = GetStationStatusArrayIndex(stationId);
    PIA_SAMPLE_LOG("ActivateStationStatus: idx=%d, stationId:%lld\n", idx, stationId.GetValue());
    PIA_SAMPLE_ASSERT(idx < SAMPLE_MAX_ENTRY);

    if (idx < SAMPLE_MAX_ENTRY)
    {
        // Notifications for connection events for the same station should not occur multiple times.
        PIA_SAMPLE_ASSERT(s_StationStatus[idx].stationId == nn::pia::STATION_ID_INVALID);

        nn::pia::StationIndex stationIndex;
        u8 colorIndex = 0;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        if (result.IsSuccess())
        {
            colorIndex = static_cast<u8>(stationIndex) + 1;
        }

        s_StationStatus[idx].stationId = stationId;
        s_StationStatus[idx].colorIndex = colorIndex;
        s_StationStatus[idx].pos.x = 0;
        s_StationStatus[idx].pos.y = FONT_SIZE * colorIndex;
        s_StationStatus[idx].latestIncrementNumber = 0;
        s_StationStatus[idx].rtt = nn::pia::transport::Station::INVALID_RTT;

        nn::pia::transport::StationManager* pManager = nn::pia::transport::StationManager::GetInstance();
        PIA_SAMPLE_ASSERT(pManager != NULL);

        if (pManager)
        {
            nn::pia::transport::Station* pStation = pManager->GetStation(stationId);
            PIA_SAMPLE_ASSERT(pStation != NULL);

            if (pStation)
            {
                if (pStation->GetState() == nn::pia::transport::Station::STATE_PARTICIPATING)
                {
                    result = pStation->GetIdentificationToken(&s_StationStatus[idx].idInfo.token);
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                    result = pStation->GetPlayerName(&s_StationStatus[idx].idInfo.name);
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }
                else
                {
                    PIA_SAMPLE_ASSERT(false);
                }
            }
        }
    }
}

void DeactivateStationStatus(const nn::pia::StationId& stationId)
{
    u16 idx = GetStationStatusArrayIndex(stationId);
    PIA_SAMPLE_LOG("DeactivateStationStatus: idx=%d, stationId:%lld\n", idx, stationId.GetValue());
    PIA_SAMPLE_ASSERT(idx < SAMPLE_MAX_ENTRY);

    if (idx < SAMPLE_MAX_ENTRY)
    {
        // There should not be multiple notifications of a disconnection event for the same station.
        PIA_SAMPLE_ASSERT(s_StationStatus[idx].stationId != nn::pia::STATION_ID_INVALID);

        s_StationStatus[idx].stationId = nn::pia::STATION_ID_INVALID;
        s_StationStatus[idx].colorIndex = 0;
        s_StationStatus[idx].pos.x = 0;
        s_StationStatus[idx].pos.y = 0;
        s_StationStatus[idx].latestIncrementNumber = 0;
        s_StationStatus[idx].rtt = nn::pia::transport::Station::INVALID_RTT;

        ::std::memset(&s_StationStatus[idx].idInfo, 0, sizeof(nn::pia::transport::Station::IdentificationInfo));
    }
}

void UpdateStationStatus()
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != pSession->End(); ++it)
    {
        u16 statusArrayIndex = GetStationStatusArrayIndex(*it);

        nn::pia::StationIndex stationIndex;
        u8 colorIndex = 0;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, *it);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        if (result.IsSuccess())
        {
            colorIndex = static_cast<u8>(stationIndex) + 1;
        }

        // Update rectangle color index.
        s_StationStatus[statusArrayIndex].colorIndex = colorIndex;
        s_StationStatus[statusArrayIndex].pos.y = FONT_SIZE * colorIndex;

        // Checks information linked to the station.
        nn::pia::transport::StationManager* pManager = nn::pia::transport::StationManager::GetInstance();
        PIA_SAMPLE_ASSERT(pManager != NULL);
        if (pManager)
        {
            nn::pia::transport::Station* pStation = pManager->GetStation(*it);

            if (pStation)
            {
                if (pStation->GetState() == nn::pia::transport::Station::STATE_PARTICIPATING)
                {
                    // Token.
                    nn::pia::transport::Station::IdentificationToken token;
                    result = pStation->GetIdentificationToken(&token);
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                    if (::std::memcmp(&token, &s_StationStatus[statusArrayIndex].idInfo.token, sizeof(token)))
                    {
                        PIA_SAMPLE_ASSERT(false);
                    }

                    // Player name.
                    nn::pia::transport::Station::PlayerName playerName;
                    result = pStation->GetPlayerName(&playerName);
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                    if (::std::memcmp(&playerName, &s_StationStatus[statusArrayIndex].idInfo.name, sizeof(playerName)))
                    {
                        PIA_SAMPLE_ASSERT(false);
                    }
                }
                else
                {
                    PIA_SAMPLE_ASSERT(false);
                }
            }
            else
            {
                PIA_SAMPLE_LOG("Station may be left during session migration. StationId:%lld\n", it->GetValue());
            }
        }
    }
}

bool CheckStationStatus()
{
    bool bResult = true;
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    for (nn::pia::session::Session::StationIdIterator it = pSession->Begin(); it != pSession->End(); ++it)
    {
        u16 idx = GetStationStatusArrayIndex(*it);
        PIA_SAMPLE_LOG("CheckStationStatus: StationIndex=%d, StationId:%lld(%lld)\n", idx, s_StationStatus[idx].stationId.GetValue(), it->GetValue());
        if (s_StationStatus[idx].stationId != *it)
        {
            bResult = false;
        }
    }
    return bResult;
}

bool CheckGetLocalAndHostStationId()
{
    bool bResult = true;
    nn::pia::StationId localStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    if (localStationId == nn::pia::STATION_ID_INVALID)
    {
        bResult = false;
    }
    else
    {
        nn::pia::transport::StationConnectionInfo stationConnectionInfo;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&stationConnectionInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        NN_PIA_DUMMY_PARAM(result);
        stationConnectionInfo.Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
    }
    nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
    if (hostStationId == nn::pia::STATION_ID_INVALID)
    {
        bResult = false;
    }
    else
    {
        nn::pia::transport::StationConnectionInfo stationConnectionInfo;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&stationConnectionInfo);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        NN_PIA_DUMMY_PARAM(result);
        stationConnectionInfo.Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
    }
    PIA_SAMPLE_LOG("CheckGetLocalAndHostStationId: Local=%lld, Host:%lld\n", localStationId.GetValue(), hostStationId.GetValue());
    return bResult;
}

u32 GetSessionId()
{
    return s_SessionId;
}

wchar_t* GetSystemPassword()
{
    return s_SystemPassword;
}

size_t GetSystemPasswordLength()
{
    return sizeof(s_SystemPassword) / sizeof(wchar_t);
}

void PrintSystemPassword()
{
    PIA_SAMPLE_LOG("-----------------------------------\n");
    PIA_SAMPLE_LOG("s_SessionId : %d\n", s_SessionId);
    for (u32 i = 0; i < nn::pia::inet::SESSION_SYSTEM_PASSWORD_LENGTH; i++)
    {
        PIA_SAMPLE_LOG("s_SystemPassword[%d] : %u\n", i, s_SystemPassword[i]);
    }
    PIA_SAMPLE_LOG("-----------------------------------\n");
}

void ResetRejoinSettings()
{
    PIA_SAMPLE_LOG("Reset settings for rejoin.\n");
    s_SessionId = nn::pia::session::INVALID_SESSION_ID;
    for (u32 i = 0; i < nn::pia::inet::SESSION_SYSTEM_PASSWORD_LENGTH; i++)
    {
        s_SystemPassword[i] = '\0';
    }
}

s32 GetSessionTime()
{
    s32 time = -1;
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    if (pSession)
    {
        time = static_cast<s32>(pSession->GetTime());
    }
    return time;
}

// Declarations of callback functions.
void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_LOG("[%d] Sample Event Callback called. State:%d SessionId:%d PID:%d StationNum:%d(%d) ClockPulse:%d\n",
        GetSessionTime(),
        pSession->GetStatus(),
        pSession->GetSessionId(),
        pSession->GetPrincipalId(),
        pSession->GetStationNum(),
        pSession->GetValidStationIdNum(),
        pSession->GetSyncClockPulseInterval());
    PIA_SAMPLE_LOG("    LocalStationId:%lld HostStationId:%lld\n",
        pSession->GetLocalStationId().GetValue(),
        pSession->GetHostStationId().GetValue())

    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_ASSERT(CheckGetLocalAndHostStationId());
        PIA_SAMPLE_LOG("Sample Event JOIN detect. StationId(%lld)\n", stationId.GetValue());
        ActivateStationStatus(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect. StationId(%lld)\n", stationId.GetValue());
        DeactivateStationStatus(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
    {
        PIA_SAMPLE_LOG("Sample Event Host Change detect. StationId(%lld)\n", stationId.GetValue());
        if(nn::pia::session::Session::GetInstance()->GetLocalStationId() == stationId)
        {
            if(!nn::pia::session::Session::GetInstance()->IsHost())
            {
                PIA_SAMPLE_LOG("HostChangeEvent that localStaionId is host detect but IsHost is false.\n");
            }
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0, (u32) stationId.GetValue());
    }
    break;
    case nn::pia::session::Session::EVENT_HOST_MIGRATION_FAILED:
    {
        // Disconnections due to host migration failures are detected by the CheckConnectionError function, so here we just output the log.
        PIA_SAMPLE_LOG("Sample Event Host Migration Failed detect. StationId(%lld)\n", stationId.GetValue());
    }
    break;
    case nn::pia::session::Session::EVENT_SET_SESSION_SYSTEM_PASSWORD:
    {
        PIA_SAMPLE_LOG("Sample Event System Password Set detect. StationId(%lld)\n", stationId.GetValue());
        s_SessionId = nn::pia::session::Session::GetInstance()->GetSessionId();
        pSession->GetSessionSystemPassword(GetSystemPassword());
        PrintSystemPassword();
    }
    break;
    case nn::pia::session::Session::EVENT_CLEAR_SESSION_SYSTEM_PASSWORD:
    {
        PIA_SAMPLE_LOG("Sample Event System Password Clear detect. StationId(%lld)\n", stationId.GetValue());
        ResetRejoinSettings();
    }
    break;
    }
}

bool SampleJoinApprovalCallback(const nn::pia::transport::Station::IdentificationToken *pToken)
{
    // Callback that determines whether to approve or reject the join-in request. (Called only by the session host.)
    bool isJoinApproval = false;

    // This sample does not check the content of the token data and gives approval as long as it is set.
    if (pToken != NULL)
    {
        // Output log of the content of the identification token of the station where the passed request originated.
        PIA_SAMPLE_LOG("Sample Join Approval Callback RequesterToken Data:%s\n", pToken->data);
        isJoinApproval = true;
    }
    PIA_SAMPLE_LOG("Sample Join Approval Callback %s\n", (isJoinApproval) ? "Approved" : "Denied");

    return isJoinApproval;
}

// Gets Session::DisconnectReason and outputs it to the log.
void PrintSessionDisconnectReason()
{
    const size_t cReasonStrSize = 32;
    char reasonStr[cReasonStrSize] = {};
    nn::pia::session::Session::DisconnectReason reason = nn::pia::session::Session::GetInstance()->GetSessionDisconnectReason();
    switch (reason)
    {
    case nn::pia::session::Session::NOT_YET_COMMUNICATED:
        ::std::strncpy(reasonStr, "NOT_YET_COMMUNICATED", cReasonStrSize);
        break;
    case nn::pia::session::Session::OPERATION_OF_OWN:
        ::std::strncpy(reasonStr, "OPERATION_OF_OWN", cReasonStrSize);
        break;
    case nn::pia::session::Session::OPERATION_OF_OTHER:
        ::std::strncpy(reasonStr, "OPERATION_OF_OTHER", cReasonStrSize);
        break;
    case nn::pia::session::Session::KICKOUT_BY_USER:
        ::std::strncpy(reasonStr, "KICKOUT_BY_USER", cReasonStrSize);
        break;
    case nn::pia::session::Session::KICKOUT_BY_SYSTEM:
        ::std::strncpy(reasonStr, "KICKOUT_BY_SYSTEM", cReasonStrSize);
        break;
    case nn::pia::session::Session::INCONSISTENT_INFO:
        ::std::strncpy(reasonStr, "INCONSISTENT_INFO", cReasonStrSize);
        break;
    case nn::pia::session::Session::MIGRATION_FAIL:
        ::std::strncpy(reasonStr, "MIGRATION_FAIL", cReasonStrSize);
        break;
    case nn::pia::session::Session::EXTERNAL_FACTOR:
        ::std::strncpy(reasonStr, "EXTERNAL_FACTOR", cReasonStrSize);
        break;
    case nn::pia::session::Session::MIGRATION_FATAL_ERROR:
        ::std::strncpy(reasonStr, "MIGRATION_FATAL_ERROR", cReasonStrSize);
        break;
    case nn::pia::session::Session::UNKNOWN_REASON:
    default:
        ::std::strncpy(reasonStr, "UNKNOWN_REASON", cReasonStrSize);
        break;
    }
    PIA_SAMPLE_LOG(" DisconnectReason:%d (%s)\n", reason, reasonStr);
}

/*!
@brief  Handles the state in which communication is performed using a <tt>PiaTransport</tt> instance.

When a host shuts down a network, it enters the select network state, but clients enter the error state.

@return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo SceneCommunication()
{
    bool isExit = false;

    // The session opens right after it is created, so set the host status to eStatusOpened.
    SessionOpenStatus sessionStatus = (nn::pia::session::Session::GetInstance()->IsHost()) ? eStatusOpened : eStatusUnknown;

    nn::pia::transport::StationConnectionInfo localInfo;
    nn::pia::transport::StationConnectionInfo masterInfo;

    nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
    nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&masterInfo);

    SampleSceneNo nextSceneNo = SCENE_SELECT_MATCHMAKE_TYPE;

    bool isAsyncProcessInProgress = false;
    bool isSessionLost = false;
    bool isRequiredLeaveSession = false;

    // This flag is used to check the bitrate detection feature's measurement completion until it is obtained once. This is due to the host's bitrate detection feature's measurement completion timing not being the same as the session participation success.
    bool isCompletedMeasuringBitRate = false;
    if(!nn::pia::session::Session::GetInstance()->IsHost())
    {
        // Clients will have completed obtaining this before the session participation is completed.
        isCompletedMeasuringBitRate = true;
    }

    // Set the callback for determining whether to approve a join.
    nn::pia::session::Session::GetInstance()->RegisterJoiningApprovalCallback(SampleJoinApprovalCallback);

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
    SCENE_COMMUNICATE,
    true,
    nn::pia::session::Session::GetInstance()->IsHost(),
    nn::pia::session::Session::GetInstance()->GetStationNum());

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        if (isAsyncProcessInProgress)
        {
            // Checks the PiaSession asynchronous processing.
#if !ENABLE_OPEN_CLOSE_SESSION
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenParticipation())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetOpenParticipationResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Open Participate Complete success\n");
                    sessionStatus = eStatusOpened;
                }
                else
                {
                    PIA_SAMPLE_LOG("Open Participate Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedCloseParticipation())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetCloseParticipationResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Close Participate Complete success\n");
                    sessionStatus = eStatusClosed;
                }
                else
                {
                    PIA_SAMPLE_LOG("Close Participate Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
#else
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenSession())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetOpenSessionResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Open Session Complete success\n");
                    sessionStatus = eStatusOpened;
                }
                else
                {
                    PIA_SAMPLE_LOG("Open Session Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedCloseSession())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetCloseSessionResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Close Session Complete success\n");
                    sessionStatus = eStatusClosed;
                }
                else
                {
                    PIA_SAMPLE_LOG("Close Session Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
#endif
            if (nn::pia::session::Session::GetInstance()->IsCompletedGenerateSystemPassword())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetGenerateSystemPasswordResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Generate System Password Complete success\n");
                }
                else
                {
                    PIA_SAMPLE_LOG("Generate System Password Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedClearSystemPassword())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetClearSystemPasswordResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Clear System Password Complete success\n");
                }
                else
                {
                    PIA_SAMPLE_LOG("Clear System Password Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedModifyAttribute())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetModifyAttributeResult();
                if(result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("Get Modify Attribute Complete success\n");
                }
                else
                {
                    PIA_SAMPLE_LOG("Get Modify Attribute Complete failed\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }
                isAsyncProcessInProgress = false;
            }
        }
        else
        {
            // Checks the session status.
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
            if (result.IsFailure())
            {
                if (result == nn::pia::ResultSessionConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("CheckConnection: Session Disconnected\n");
                    PrintSessionDisconnectReason();
                    isRequiredLeaveSession = true;
                    isAsyncProcessInProgress = true;
                    isSessionLost = true;
                    isExit = true;
                }
                else
                {
                    PIA_SAMPLE_LOG("CheckConnection: Connection is lost\n");
                    nextSceneNo = GetNextSceneByErrorResult(result, CONNECTION_LOST_ERROR);
                    isExit = true;
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_COMMUNICATE, false);
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE,
                true,
                nn::pia::session::Session::GetInstance()->IsHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum());

#if PIA_TEST_CODE
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
                    SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
                    nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());

                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
                    SAMPLE_PROGRAMPARAM_HOST_MIGRATION_STATE,
                    reinterpret_cast<nn::pia::common::Job*>(nn::pia::session::Mesh::GetInstance()->GetProcessHostMigrationJobPtr())->GetState());
#endif
            }
        }

        // Draw to the lower screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            // String.
            {
                // Check the role of the local station.
                int roleIndex = 0;  // Unknown.
                if (nn::pia::session::Session::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID)
                {
                    roleIndex = (nn::pia::session::Session::GetInstance()->IsHost() ? 1 : 2);  // 1: Host, 2: Client.
                }

                int line = 0;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, g_FrameNo);
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Communication");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Time=%08d SessionId=%08X",
                    GetSessionTime(), nn::pia::session::Session::GetInstance()->GetSessionId());
                line += FONT_SIZE;
                if( nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum() != 0)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum=%d Matchmake=%d Role=%s",
                        nn::pia::session::Session::GetInstance()->GetStationNum(), nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum(), SESSION_ROLE[roleIndex]);
                }
                else
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum=%d Role=%s",
                        nn::pia::session::Session::GetInstance()->GetStationNum(), SESSION_ROLE[roleIndex]);
                }
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Participation=%s", SESSION_OPEN_STATUS[sessionStatus]);
                line += FONT_SIZE;

                for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
                {
                    const u8 DRAW_BASE = line;

                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(10, DRAW_BASE + (10 * i), 8, 8, COLOR_OF_RECTANGLE[s_StationStatus[i].colorIndex]);
                    if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                    {
                        s_StationStatus[i].UpdateInfo();
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "Id=%lld, %05X, RTT:%4d", s_StationStatus[i].stationId.GetValue(), s_StationStatus[i].latestIncrementNumber, s_StationStatus[i].rtt);
                    }
                    else
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "No Connection");
                    }
                }

#if PIA_TEST_CODE
                line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 16);
#else
                line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 15);
#endif
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Network Information ]");
                line += FONT_SIZE;

                // Screen display of communication-related data (for debugging).
                nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
                nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&masterInfo);

                nn::pia::common::String mainStr;
                nn::pia::common::String subStr;
                localInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                localInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Local  Main:%s", mainStr.GetCharArray());
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "          Sub:%s", subStr.GetCharArray());
                line += FONT_SIZE;
                masterInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                masterInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Host Main:%s", mainStr.GetCharArray());
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "          Sub:%s", subStr.GetCharArray());
                line += FONT_SIZE;

                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    B: LeaveSession");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    Y: ModifyAttribute");
                line += FONT_SIZE;
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    L: Open/Close Participation");
                    line += FONT_SIZE;
                    if (g_NetworkType == TYPE_INTERNET)
                    {
#if ENABLE_SERVERLESS
#else
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "START: Generate/Clear SystemPassword");
                        line += FONT_SIZE;
#endif
                    }
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
            }
        }
        // Draw to the upper screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
            {
                if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                {
                    // Gets the player name
                    char name[SAMPLE_USER_NAME_LENGTH];
                    for (int j = 0; j < SAMPLE_USER_NAME_LENGTH - 1; j++)
                    {
                        if (j < s_StationStatus[i].idInfo.name.length)
                        {
                            // Enable only the ASCII range to display.
                            if (s_StationStatus[i].idInfo.name.data[j] >= 0x20 && s_StationStatus[i].idInfo.name.data[j] < 0x7f)
                            {
                                name[j] =  static_cast<char>(s_StationStatus[i].idInfo.name.data[j]);
                            }
                            else
                            {
                                name[j] = '?';
                            }
                        }
                    }

                    if (SAMPLE_USER_NAME_LENGTH - 1 <= s_StationStatus[i].idInfo.name.length)
                    {
                        name[SAMPLE_USER_NAME_LENGTH - 1] = '\0';
                    }
                    else
                    {
                        name[s_StationStatus[i].idInfo.name.length] = '\0';
                    }

                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
                        s_StationStatus[i].pos.x, s_StationStatus[i].pos.y, 8, 8, COLOR_OF_RECTANGLE[s_StationStatus[i].colorIndex]);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                        s_StationStatus[i].pos.x + 10, s_StationStatus[i].pos.y, "<-ID:%lld (%s)", s_StationStatus[i].stationId.GetValue(), name);
                }
            }

            int line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        }

        // Execute dispatch.
        Dispatch();
        if (!isAsyncProcessInProgress && !isExit)
        {
            // Checks the session status.
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
            if (result.IsFailure())
            {
                if (result == nn::pia::ResultSessionConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("CheckConnection: Session Disconnected\n");
                    PrintSessionDisconnectReason();
                    isRequiredLeaveSession = true;
                    isAsyncProcessInProgress = true;
                    isSessionLost = true;
                    isExit = true;
                }
                else
                {
                    PIA_SAMPLE_LOG("CheckConnection: Connection is lost\n");
                    nextSceneNo = GetNextSceneByErrorResult(result, CONNECTION_LOST_ERROR);
                    isExit = true;
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_COMMUNICATE, false);
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE,
                true,
                nn::pia::session::Session::GetInstance()->IsHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum());
            }
        }

        // Get measurement result from the bitrate detection feature
        if(nn::pia::session::Session::GetInstance()->IsHost() && !isCompletedMeasuringBitRate)
        {
            s32 bitRate = 0;
            if(nn::pia::session::Session::GetInstance()->IsCompleteBitRateCheck())
            {
                if(nn::pia::session::Session::GetInstance()->GetUplinkBitRate(&bitRate).IsSuccess())
                {
                    isCompletedMeasuringBitRate = true;
                    PIA_SAMPLE_LOG("bitRate: %d\n", bitRate);
                }
            }
        }

#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
            SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
            nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
            SAMPLE_PROGRAMPARAM_ISCOMPLETED_CLOSEPARTICIPATION,
            nn::pia::session::Session::GetInstance()->IsCompletedCloseParticipation() ? 1 : 0 );

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
            SAMPLE_PROGRAMPARAM_HOST_MIGRATION_STATE,
            reinterpret_cast<nn::pia::common::Job*>(nn::pia::session::Mesh::GetInstance()->GetProcessHostMigrationJobPtr())->GetState());
#endif

        nn::pia::StationId myStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();

        // Receive.
        // Executed after Dispatch (first time) to facilitate receiving data with the Scheduler::Dispatch function.
        if (isExit == false)
        {
            // Because the size of the data the other party sends generally cannot be anticipated,
            // allocate a receive buffer the size of MAX_DATA_SIZE of UnreliableProtocol.
            PIA_ATTRIBUTE_ALIGN(4) static bit8 receiveBuffer[nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE];

            nn::Result result;
            // Because the number of received data is unknown, the receive process repeats until the buffer is empty.
            do
            {
                nn::pia::StationId sourceStationId;
                size_t receivedDataSize = 0;

                result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Receive(&sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));

                if (result.IsSuccess())
                {
                    u16 id = GetStationStatusArrayIndex(sourceStationId);
                    if (id < SAMPLE_MAX_ENTRY)
                    {
                        // You may need to check whether the received data is correct.
                        s_StationStatus[id].Deserialize(&receiveBuffer);
                    }
                }
                else if (result == nn::pia::ResultNotInCommunication())
                {
                    // Failed to receive data due to a disconnection.
                    // This failure can occur if an asynchronous session delete or leave process completes due to an immediately preceding Dispatch process.
                    // If the status is normal, the session status check shows DISCONNECTED when the next game frame starts.
                    PIA_SAMPLE_LOG("UnreliableProtocol::Receive failed. (IsNotInCommunication)\n");
                    break;
                }
                else if (result != nn::pia::ResultNoData())
                {
                    // Other reception failed.
                    // In this demo, this is not an issue.
                }
            } while (result != nn::pia::ResultNoData());
        }

        int move_x = 0;  // Amount of movement of the rectangle on the local station (x).
        int move_y = 0;  // Amount of movement of the rectangle on the local station (y).

        // Leave the session.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            if (!isAsyncProcessInProgress)
            {
                // If leaving the session as a result of a user operation, first clear the rejoin information and then leave.
                ResetRejoinSettings();
                ResetGameMode();
                isRequiredLeaveSession = true;
                isExit = true;
                isAsyncProcessInProgress = true;
            }
        }
        // Change attribute value of the session.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            if (!isAsyncProcessInProgress)
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->ModifyAttributeAsync(0, 255);
                isAsyncProcessInProgress = true;
                if (result.IsFailure())
                {
                    isAsyncProcessInProgress = false;
                    PIA_SAMPLE_LOG_RESULT(result);
                    PIA_SAMPLE_LOG("ModifyAttribute failed\n");
                }
            }
        }
#if PIA_TEST_CODE
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
            PIA_SAMPLE_ASSERT(pThreadStreamManager);  // If the transport initialization has finished, the instance has probably been created.

            s32 ratio = (pThreadStreamManager->GetSendThreadPacketLossRatio() == 100) ? 0 : 100;
            pThreadStreamManager->SetSendThreadPacketLossRatio(ratio);
            pThreadStreamManager->SetReceiveThreadPacketLossRatio(ratio);
            PIA_SAMPLE_LOG("Packet dropped emulation ratio (send) = %d\n", ratio);
            PIA_SAMPLE_LOG("Packet dropped emulation ratio (recv) = %d\n", ratio);
        }
#endif
        // Handle user input.
        // Opening and closing a session.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
        {
            if (!isAsyncProcessInProgress)
            {
                // eStatusUnknown (when you have joined as a client, the session will be in an open state
                if (sessionStatus == eStatusOpened || sessionStatus == eStatusUnknown)
                {
#if ENABLE_OPEN_CLOSE_SESSION
                        nn::Result result = nn::pia::session::Session::GetInstance()->CloseSessionAsync();
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_LOG("CloseSession failed\n");
                        }
#else
                        nn::Result result = nn::pia::session::Session::GetInstance()->CloseParticipationAsync();
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_LOG("CloseParticipation failed\n");
                        }
#endif
                    }
                else
                {
#if ENABLE_OPEN_CLOSE_SESSION
                        nn::Result result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_LOG("OpenSession failed\n");
                        }
#else
                        nn::Result result = nn::pia::session::Session::GetInstance()->OpenParticipationAsync();
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG("OpenParticipation failed\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                        }
#endif
                    }
                }
            }
        if (g_NetworkType == TYPE_INTERNET)
        {
#if ENABLE_SERVERLESS
// The following are not supported if matchmaking without a server.
#else
            // Sets and clears a system password.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
            {
                if (!isAsyncProcessInProgress)
                {
                    if (s_SessionId == nn::pia::session::INVALID_SESSION_ID)
                    {
                        nn::Result result = nn::pia::session::Session::GetInstance()->GenerateSystemPasswordAsync(GetSystemPassword());
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_LOG("GenerateSystemPassword failed\n");
                        }
                    }
                    else
                    {
                        nn::Result result = nn::pia::session::Session::GetInstance()->ClearSystemPasswordAsync();
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            isAsyncProcessInProgress = false;
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_LOG("ClearSystemPassword failed\n");
                        }
                    }
                }
            }
 #endif
        }
        // Move the rectangle.
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
            {
                move_x = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                move_x = 1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
            {
                move_y = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
            {
                move_y = 1;
            }
        }

        // Locked send process.
        // Executed before Dispatch (second time) for the convenience of having send processing performed by the Scheduler::Dispatch() function.
        if (isExit == false)
        {
            u16 id = GetStationStatusArrayIndex(myStationId);
            if (id < SAMPLE_MAX_ENTRY)
            {
                // Update the data of the local station.
                if (move_x != 0)
                {
                    s_StationStatus[id].pos.x = (s_StationStatus[id].pos.x + UPPER_SCREEN_WIDTH + move_x) % UPPER_SCREEN_WIDTH;
                }
                else
                {
                    // Move the rectangle even when there is no input to indicate that stations are communicating.
                    s_StationStatus[id].pos.x = (s_StationStatus[id].pos.x + UPPER_SCREEN_WIDTH + 1) % UPPER_SCREEN_WIDTH;
                }
                if (move_y != 0)
                {
                    s_StationStatus[id].pos.y = (s_StationStatus[id].pos.y + UPPER_SCREEN_HEIGHT + move_y) % UPPER_SCREEN_HEIGHT;
                }
                s_StationStatus[id].latestIncrementNumber = g_FrameNo;

                if ((g_FrameNo % (SAMPLE_INTERVAL_SENDING_NODE_STATUS + 1)) == 0)
                {
                    u16 stationNum = nn::pia::session::Session::GetInstance()->GetStationNum();
                    if (stationNum > 1)
                    {
                        // Use SendToAll to send as a batch process.
                        SampleStationStatus sendData;
                        s_StationStatus[id].Serialize(&sendData);
                        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->SendToAll(reinterpret_cast<bit8*>(&sendData), SampleStationStatus::GetSerializedSize()).IsFailure())
                        {
                            // Send failure.
                            // In this demo, this is not an issue.
                            // PIA_SAMPLE_LOG("Send failed.\n");
                        }
                    }
                }
            }
        }

        // Execute Dispatch (second time: for send processing).
        Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        g_FrameNo++;
    }

    if (isRequiredLeaveSession)
    {
        SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_LeaveSession();
        if (sampleSessionAsyncStatus == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
        {
            nextSceneNo = GetNextSceneByErrorResult(nn::pia::ResultNetworkConnectionIsLost(), SESSION_LEAVE_ERROR);
        }
        else if (isSessionLost)
        {
            nextSceneNo = GetNextSceneByErrorResult(nn::pia::ResultSessionConnectionIsLost(), SESSION_LOST_ERROR);
        }
        isAsyncProcessInProgress = false;
    }

    // Unregister the callback for determining whether to approve a join.
    nn::pia::session::Session::GetInstance()->UnregisterJoiningApprovalCallback();

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0, 0);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
    SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
    nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());

#if PIA_TEST_CODE
    {
    nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(pThreadStreamManager);  // If the transport initialization has finished, the instance has probably been created.

    s32 ratio = 0;
    pThreadStreamManager->SetSendThreadPacketLossRatio(ratio);
    pThreadStreamManager->SetReceiveThreadPacketLossRatio(ratio);
    PIA_SAMPLE_LOG("Packet dropped emulation ratio (send) = %d\n", ratio);
    PIA_SAMPLE_LOG("Packet dropped emulation ratio (recv) = %d\n", ratio);
    }
#endif

    Session_CleanupSession();
    return nextSceneNo;
}
