﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "SampleLib/SampleLib.h"

#include "SampleConfig.h"

#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include <pia/inet/inet_NexDebugSession.h>

namespace
{
    const int DummySessionMaxSize = 20;
    u32 s_DummySessionIdArray[DummySessionMaxSize];
    u32 s_DummySessionIdSize = 0;

    const int DummySessionCommand_Create = 1;
    const int DummySessionCommand_Destroy = 2;
    const int DummySessionCommand_Update = 3;
    const int DummySessionCommand_Back = 4;

// NEX-related.
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;
};

SampleSceneNo SceneDummySession()
{
    SampleSceneNo nextSceneNo = SCENE_DUMMY_SESSION;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_DUMMY_SESSION, false, false, 0);

    nn::Result result;

    bool isLoggedIn = false;

    // Initializes NEX.
    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    PIA_SAMPLE_LOG("Nex_InitializeNex succeeded.\n");

    // Initializes PiaInet.
    nn::pia::inet::Setting setting;
    setting.mtuSize = nn::pia::inet::DEFAULT_MTU_SIZE;
    result = nn::pia::inet::Initialize(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    result = nn::pia::inet::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    result = nn::pia::inet::NexFacade::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    result = nn::pia::inet::NexDebugSession::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    result = nn::pia::inet::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    PIA_SAMPLE_LOG("PiaInet Initialize succeeded.\n");

    // Initializes the network connection.
    result = Inet_InitializeNetwork();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Net_InitializeNetwork failed.\n");
        nextSceneNo = SCENE_START;
    }
    // Log in to the game server.
    else if (Inet_Login().IsFailure())
    {
        PIA_SAMPLE_LOG("Inet_Login failed.\n");
        nextSceneNo = SCENE_START;
    }
    else
    {
        PIA_SAMPLE_LOG("Inet_Login succeeded.\n");
        isLoggedIn = true;
        // Starts <tt>NexDebugSession</tt> if login to the game server succeeds.
        result = nn::pia::inet::NexDebugSession::GetInstance()->Startup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        PIA_SAMPLE_LOG("NexDebugSession started.\n");
    }

    while (nextSceneNo == SCENE_DUMMY_SESSION)
    {
        int command = 0;
        while (command == 0)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

            // A Button: Creates a new dummy session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                command = DummySessionCommand_Create;
            }
            // X Button: Destroys a dummy session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                command = DummySessionCommand_Destroy;
            }
            // Y Button: Updates dummy session settings.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                command = DummySessionCommand_Update;
            }
            // B Button: Back
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                command = DummySessionCommand_Back;
                nextSceneNo = SCENE_START;
            }

            // Lower screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %d", PROGRAM_NAME, g_FrameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: DummySession.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 5);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Create DummySession");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Destroy DummySession");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Update DummySession");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Back");
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            ++g_FrameNo;

            Dispatch();

            // Checks the connection status.
            if (!Inet_IsConnected())
            {
                PIA_SAMPLE_LOG("Inet_IsConnected disconnected.\n");
                nextSceneNo = SCENE_START;
                break;
            }
        }
        if (command == DummySessionCommand_Create)
        {
            // Settings for the matchmaking session to create.
            nn::pia::inet::NexCreateSessionSetting createSessionSetting;

            createSessionSetting.SetSessionType(nn::pia::inet::SESSION_TYPE_ANYBODY);
            createSessionSetting.SetGameMode(MATCHMAKE_GAME_MODE + MATCHMAKE_BROWSE + GetIncrementalGameModeValueByBitRateCheckMode());
            createSessionSetting.SetMaxParticipants(SAMPLE_MAX_ENTRY);
            createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX, g_IsEnableHostMigration);
            createSessionSetting.SetOpenParticipation(true);
            createSessionSetting.SetDescription(SAMPLE_DESCRIPTION_BROWSE);

            result = nn::pia::inet::NexDebugSession::GetInstance()->CreateDummySessionAsync(createSessionSetting);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            // Waits for the asynchronous process <tt>CreateDummySessionAsync</tt> to complete.
            int loopCounter = 0;
            while (true)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "creating dummy session... %d", loopCounter++);
                nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                ++g_FrameNo;

                Dispatch();
                if (nn::pia::inet::NexDebugSession::GetInstance()->IsCompletedCreateDummySession())
                {
                    result = nn::pia::inet::NexDebugSession::GetInstance()->GetCreateDummySessionResult();
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("CreateDummySessionAsync failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        nextSceneNo = SCENE_START;
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("CreateDummySessionAsync success. SessionId=%d\n",
                            nn::pia::inet::NexDebugSession::GetInstance()->GetCreatedDummySessionId());
                        s_DummySessionIdArray[s_DummySessionIdSize] = nn::pia::inet::NexDebugSession::GetInstance()->GetCreatedDummySessionId();
                        ++s_DummySessionIdSize;
                    }
                    break;
                }
            }
        }
        if (command == DummySessionCommand_Destroy)
        {
            if (s_DummySessionIdSize > 0)
            {
                result = nn::pia::inet::NexDebugSession::GetInstance()->DestroyDummySessionAsync(s_DummySessionIdArray[s_DummySessionIdSize - 1]);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                // Waits for the asynchronous process <tt>DestroyDummySessionAsync</tt> to complete.
                int loopCounter = 0;
                while (true)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "destroying dummy session... %d", loopCounter++);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                    ++g_FrameNo;

                    Dispatch();
                    if (nn::pia::inet::NexDebugSession::GetInstance()->IsCompletedDestroyDummySession())
                    {
                        result = nn::pia::inet::NexDebugSession::GetInstance()->GetDestroyDummySessionResult();
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("DestroyDummySessionAsync failed.\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                            nextSceneNo = SCENE_START;
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("DestroyDummySessionAsync succeeded. SessionId=%d\n",
                                s_DummySessionIdArray[s_DummySessionIdSize - 1]);
                            --s_DummySessionIdSize;
                            s_DummySessionIdArray[s_DummySessionIdSize] = 0;
                        }
                        break;
                    }
                }
            }
        }
        if (command == DummySessionCommand_Update)
        {
            if (s_DummySessionIdSize > 0)
            {
                const u32 UpdateSessionSettingSize = 7;
                nn::pia::inet::NexUpdateSessionSetting updateSessionSetting[UpdateSessionSettingSize];
                updateSessionSetting[0].SetDescription(L"UpdateSession0");
                updateSessionSetting[1].SetOpenParticipation(false);
                updateSessionSetting[2].SetMinParticipants(2);
                updateSessionSetting[3].SetMaxParticipants(6);
                u8 appData[4] = { 0xDE, 0xAD, 0xBE, 0xEF };
                updateSessionSetting[4].SetApplicationData(appData, sizeof(appData));
                u32 attributes[6] = { 0xf, 0xe, 0xe, 0xd, 0xb, 0xa };
                updateSessionSetting[5].SetAttributes(attributes);
                updateSessionSetting[6].SetMinParticipants(1);
                updateSessionSetting[6].SetMaxParticipants(8);
                updateSessionSetting[6].SetOpenParticipation(true);
                updateSessionSetting[6].SetDescription(L"UpdateSession6");
                u32 attributes2[6] = { 0xc, 0xa, 0xf, 0xe, 0xb, 0xa };
                updateSessionSetting[6].SetAttributes(attributes2);
                appData[0] = 0xCA;
                appData[1] = 0xFE;
                updateSessionSetting[6].SetApplicationData(appData, sizeof(appData));

                for (u32 i = 0; i < s_DummySessionIdSize && i < UpdateSessionSettingSize && nextSceneNo == SCENE_DUMMY_SESSION; i++)
                {
                    u32 sessionId = s_DummySessionIdArray[i];
                    result = nn::pia::inet::NexDebugSession::GetInstance()->UpdateDummySessionAsync(sessionId, updateSessionSetting[i]);
                    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                    // Waits for the asynchronous process <tt>UpdateDummySessionAsync</tt> to complete.
                    int loopCounter = 0;
                    while (true)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "updating dummy session... %d", loopCounter++);
                        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                        ++g_FrameNo;

                        Dispatch();
                        if (nn::pia::inet::NexDebugSession::GetInstance()->IsCompletedUpdateDummySession())
                        {
                            result = nn::pia::inet::NexDebugSession::GetInstance()->GetUpdateDummySessionResult();
                            if (result.IsFailure())
                            {
                                PIA_SAMPLE_LOG("UpdateDummySessionAsync failed. sessionId=%d\n", sessionId);
                                PIA_SAMPLE_LOG_RESULT(result);
                                nextSceneNo = SCENE_START;
                            }
                            else
                            {
                                PIA_SAMPLE_LOG("UpdateDummySessionAsync succeeded. SessionId=%d\n", sessionId);
                            }
                            break;
                        }
                    }
                }
            }
        }
    }

    if (isLoggedIn)
    {
        // Cleans up <tt>NexDebugSession</tt> before logging out from the game server.
        nn::pia::inet::NexDebugSession::GetInstance()->Cleanup();
        // Log out from the game server.
        Inet_Logout();
    }
    // Network connection termination processing.
    Inet_FinalizeNetwork();
    // Destroys <tt>NexDebugSession</tt> singleton.
    nn::pia::inet::NexDebugSession::DestroyInstance();
    // Destroys <tt>NexFacade</tt> singleton.
    nn::pia::inet::NexFacade::DestroyInstance();
    // Finalizes PiaInet.
    nn::pia::inet::Finalize();
    // Finalize NEX.
    Nex_FinalizeNex();

    return nextSceneNo;
}
