﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneError.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/util.h>

#include "SampleLib/SampleLib.h"

#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"
#include "SampleConfig.h"

namespace
{
SampleErrorType s_ErrorType = UNKNOWN;  //Variable that holds the errors that occurred most recently.

const char* ERROR_INFO_STRING[] =
    {
        "Unknown Reason",
        "Initialize Error (Network IF up)",
        "Initialize Error (other reason)",
        "Login failed",
        "Connection was lost",
        "Session was lost",
        "StartupSession failed",
        "CreateSession failed",
        "BrowseSession failed",
        "JoinSession failed",
        "LeaveSession failed",
        "DestroySession failed"
    };

nn::Result s_NetworkErrorResult = nn::ResultSuccess();
}


/*!
    @brief  This function sets the error reason before jumping to the error state.
 */
void SetCommunicationErrorType(SampleErrorType error)
{
    s_ErrorType = error;
}

/*!
    @brief  This state displays a description for a recoverable error (such as a communication error).
    
    It only displays error information collected when the error occurred.
 */
SampleSceneNo SceneError()
{
    nn::Result result;
    bool isExit = false;
    SampleSceneNo nextScene = (s_ErrorType == SESSION_LOST_ERROR) ? SCENE_SELECT_MATCHMAKE_TYPE : SCENE_RESET;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_ERROR, false);
    ResetGameMode();

    // Main loop.
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        //Handles user input. Pad information is collected by the BeginFrame function.
        {
            // A:    Restart.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                isExit = true;
            }
        }

        // Renders to the screens.
        int line;

        // Lower screen (standard).
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, g_FrameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Communication Error Occurred.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
        line += FONT_SIZE;

        line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 2);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        // Execute Dispatch.
        Dispatch();
        g_FrameNo++;
    }
    //Clears the error information, just in case.
    s_ErrorType = UNKNOWN;
    SetNetworkErrorResult(nn::ResultSuccess());
    return nextScene;
}


// Pass the result of errors that occur when connecting to the Internet.
// The Result passed here is then passed to the network error code conversion API and output to the screen.
// 
void SetNetworkErrorResult(nn::Result result)
{
    s_NetworkErrorResult = result;
}


nn::Result GetNetworkErrorResult(void)
{
    return s_NetworkErrorResult;
}


/*!
    @brief  This scene displays the network error code for errors occurring in the Internet connection environment.
    Processing transitions to this state when an error is generated that requires logging in again.
    In a normal program, pass the network error codes to the Error/EULA Viewer.
 */
SampleSceneNo SceneNetworkError()
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_NETWORK_ERROR, false);
    ResetGameMode();

    // Convert the error <tt>Result</tt> to a network error code.
    // Normally, you would pass this error code to the error/EULA applet.
    // However, the error/EULA applet cannot be used unless development hardware is set to <tt>prod</tt> mode.
    // For simplicity, this sample demo does not use the error/EULA applet.
    // Not used.
    nn::Result result = GetNetworkErrorResult();
    PIA_SAMPLE_LOG_RESULT(result);
    bool isPiaError = nn::pia::util::IsPiaResult(result);
    u32 networkErrorCode = (isPiaError) ? nn::pia::util::NetworkErrorCodeConverter::Convert(result) : Inet_GetNetworkErrorCode();

    bool isExit = false;

    // Rendering
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Exit the loop if the A Button is pressed.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            isExit = true;
        }

        int line;

        // Render the lower screen
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, g_FrameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Network Error Occurred.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Result: 0x%08X", result.GetPrintableBits());
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Level:       %4d", result.GetLevel());
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Summary:     %4d", result.GetSummary());
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Module:      %4d", result.GetModule());
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Description: %4d", result.GetDescription());
        line += FONT_SIZE;
        line += FONT_SIZE;

        if (networkErrorCode != 0)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Network error code: %07d", networkErrorCode);
            line += FONT_SIZE;
        }
        if (!isPiaError && Inet_GetLoginNetworkErrorCode() != 0)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Login error code: %07d", Inet_GetLoginNetworkErrorCode());
            line += FONT_SIZE;
        }

        line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 2);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        // Execute Dispatch.
        Dispatch();

        g_FrameNo++;
    }

    //Clears the error information, just in case.
    s_ErrorType = UNKNOWN;
    SetNetworkErrorResult(nn::ResultSuccess());

    nn::pia::util::HandlingType type = nn::pia::util::GetHandlingType(result);

    if (type == nn::pia::util::HANDLING_LOGOUT
    ||  type == nn::pia::util::HANDLING_RECONNECT
    ||  type == nn::pia::util::HANDLING_REINITIALIZE)
    {
        // Transitions to SCENE_RESET in order to clean up, log out, reconnect, and reinitialize.
        return SCENE_RESET;
    }
    else if (type == nn::pia::util::HANDLING_CLEANUP)
    {
        // Transitions to SCENE_SELECT_MATCHMAKE_TYPE to clean up because logging in again is not necessary.
        Session_CleanupSession();
        return SCENE_SELECT_MATCHMAKE_TYPE;
    }
    else
    {
        // Any other Pia errors that require handling should have been handled by this point.
        if (isPiaError)
        {
            // Unexpected errors.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            return SCENE_RESET;
        }
        else
        {
            // Non-Pia error.
            return SCENE_RESET;
        }
    }
}
