﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: ScenePrepareInternet.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"

#include "SampleConfig.h"


/*!

    @brief  CPP file with processing specific to Internet communication.

*/

namespace
{
// NEX-related.
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

const char* stepStrings[] = {
    "Initialize NEX",
    "Initialize Pia modules",
    "Initialize Network",
    "Login server",
    "Retrieve Notification Event"
};  // Four steps.

enum InitializeStep
{
    STEP_INITIALIZE_NEX = 0,
    STEP_INITIALIZE_PIA_MODULES,
    STEP_INITIALIZE_NETWORK,
    STEP_LOGIN_SERVER,
    STEP_RETRIEVE_NOTIFICATION_DATA,
    STEP_MAX
};

}  // end of anonymous namespace

SampleSceneNo ScenePrepareInternet()
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_PREPARE_INET,
        false,
        false,
        0,
        0);

    nn::Result result;
    u32 step = STEP_INITIALIZE_NEX;

    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    ++step;

    nn::nex::qList<nn::nex::NotificationEvent> listNotificationEvent;

    while (step < STEP_MAX)
    {
        switch (step)
        {
        case STEP_INITIALIZE_PIA_MODULES:
            // Initializes PiaInet.
            result = Inet_InitializePiaInet();
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            // Initializes PiaTransport.
            {
                u32 unreliableProtocolHandle;
                nn::pia::transport::Transport::Setting transportSetting;
                transportSetting.pFactory = Inet_GetNetworkFactory();
                transportSetting.maxStationNum = SAMPLE_MAX_ENTRY;
                transportSetting.maxSendThreadBufferNum = 16;
                transportSetting.maxReceiveThreadBufferNum = 32;

                result = Transport_InitializePiaTransport(
                    transportSetting,
                    0,
                    &unreliableProtocolHandle,
                    0,
                    NULL);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                // Sets the UnreliableProtocolHandle to use for communication.
                SetUnreliableProtocolHandle(unreliableProtocolHandle);
            }
            // Initializes PiaSession.
            {
                nn::pia::session::Session::Setting sessionSetting;
                sessionSetting.pNetworkFactory = Inet_GetNetworkFactory();
                sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;

                if(g_IsEnableBitRateCheck)
                {
                    sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Enable;
                    PIA_SAMPLE_LOG("BitRate check mode: %s\n", g_IsEnableBitRateCheck ? "Enable" : "Disable");
                }
                else
                {
                    sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Disable;
                    PIA_SAMPLE_LOG("BitRate check mode: %s\n", g_IsEnableBitRateCheck ? "Enable" : "Disable");
                }

                sessionSetting.browsedSessionInfoListNum = SESSION_LIST_MAX;
                result = Session_InitializePiaSession(sessionSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                // Sets the event callback.
                result = nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionCallback);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
            ++step;
            break;

        case STEP_INITIALIZE_NETWORK:
            result = Inet_InitializeNetwork();
            if (result.IsFailure())
            {
                return GetNextSceneByErrorResult(result, INITIALIZE_ERROR_NETWORK_IF_UP);
            }
            ++step;
            break;

        case STEP_LOGIN_SERVER:
            // Register an event handler to receive event notifications from the game server, and then log in.
            result = Inet_Login();
            if (result.IsFailure())
            {
                // Login fails.
                return GetNextSceneByErrorResult(result, LOGIN_ERROR);
            }
#if ENABLE_SERVERLESS
#else
            // Register an event handler to receive event notifications from the game server.
            Nex_GetNgsFacadeInstance()->RegisterNotificationEventHandler(&g_NexNotificationHandler);

            // Gets the event that has already been notified to the game server.
            listNotificationEvent.clear();
            result = nn::pia::inet::NexFacade::GetInstance()->RetrieveNotificationDataAsync(0xf, &listNotificationEvent);
            if (result.IsFailure())
            {
                return GetNextSceneByErrorResult(result, LOGIN_ERROR);
            }
#endif
            ++step;
            break;

        case STEP_RETRIEVE_NOTIFICATION_DATA:
#if ENABLE_SERVERLESS
            ++step;
#else
            if (nn::pia::inet::NexFacade::GetInstance()->IsCompletedRetrieveNotificationData())
            {
                result = nn::pia::inet::NexFacade::GetInstance()->GetRetrieveNotificationDataResult();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Retrieve notification event process is failed. Result:%d\n", result.GetValue());
                }
                else
                {
                    for (nn::nex::qList<nn::nex::NotificationEvent>::const_iterator it = listNotificationEvent.begin(); it != listNotificationEvent.end(); ++it)
                    {
                        PIA_SAMPLE_LOG("Retrieved notification event. Type:%u, Param1:%d, Param2:%d\n", it->GetType(), it->GetParam1(), it->GetParam2());
                    }
                }
                ++step;
            }
#endif
            break;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Lower screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo]", PROGRAM_NAME);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Inet Initializing process.");
        line += FONT_SIZE;

        line += FONT_SIZE;

        // Progress display.
        // Steps that have been processed are displayed in white; steps that have not been processed are displayed in red.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   START");
            line += FONT_SIZE;
            for (u32 i = 0; i < STEP_MAX; i++)
            {
                if (i == step)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "     |");
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                }
                else
                {
                    line += FONT_SIZE;
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", stepStrings[i]);
                line += FONT_SIZE;
            }
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   END");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        Dispatch();
        ++g_FrameNo;
    }
    return SCENE_SELECT_MATCHMAKE_TYPE;
}
