﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


/*
Advanced Session Sample Demo

Overview
  This is a more advanced demo of the features of the nn::pia::session::Session class.
  It performs random matchmaking, browse matchmaking, and
  friend join-in matchmaking using Internet communication.

  Friend join-in matchmaking supports joining the session of a friend who is running the same game,
  as well as joining a friend's session from the HOME Menu.

  The following features are supported while in a communicating state.
   Host Migration
   Opening and closing a session.
   Rejoining using a matchmaking system password.

  In addition to these Internet communication matchmaking techniques, the 3DS supports browse matchmaking over local communication.

  The Wii U can be switched to LAN matchmaking by setting ENABLE_SERVERLESS.

Note
  Even if presence is updated, it can take roughly 10 seconds from the previous update before friends are notified.
  Because it is possible for friend presence-related change notifications to be delayed due to network conditions,
  note that even if the friend presence library determines that it is possible to join, that is not always
  the case.

  With the Nintendo 3DS, the friend presence join-in feature may not work correctly
  if Home Menu icon cache remains in the system. Be sure to
  reset the icon cache when using the join-in feature.

  This sample uses the friend presence join-in feature to join sessions where a friend is a participant.
  This requires creating a master ROM with a properly configured join-in game ID
  and join-in game mode mask, and starting it from the HOME Menu.
  For more information about the procedure of creating a master ROM,
  see the <em>Master ROM Creation and Submission Guidelines</em> for each platform.
*/


#include <pia/version.h>

#include "SampleLib/SampleLib.h"

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"

#if NN_PIA_CTR
#include "Shared_Local_Setup.h"
#endif

#include "SampleConfig.h"

#if NN_PIA_CAFE
#include <cafe/sysapp.h>
#endif

namespace
{
// The memory region to allocate to Pia. It must be 4-byte aligned.
// In this sample, sufficiently large values are used.
PIA_ATTRIBUTE_ALIGN(4) bit8 s_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA];

// Variables for saving the Pia library version information character string.
char s_SamplePiaVersionInfoString[UPPER_SCREEN_WIDTH / FONT_SIZE] = {};

const wchar_t SAMPLE_DESCRIPTION[] = L"NotInCommunication";
const wchar_t* s_pDescription;
FriendGameMode s_FriendGameMode;
}  // end of anonymous namespace

// Declarations of global variables.
u32 g_FrameNo = 0;
NetworkType g_NetworkType = TYPE_INTERNET;
bool g_IsEnableHostMigration = false;
bool g_IsFromFriendList = false;
FriendId g_FriendIdForJoinIn;
bool g_IsEnableBitRateCheck = false;
u32 g_JoinedPersistentGathering = 0;
SampleNexNofiticationEventHandler g_NexNotificationHandler;

// Function that gets the Pia library version information string.
const char* GetPiaVersionString()
{
    return s_SamplePiaVersionInfoString;
}

u32 GetIncrementalGameModeValueByBitRateCheckMode()
{
    if(g_IsEnableBitRateCheck)
    {
        return 1000;
    }
    else
    {
        return 0;
    }
}

void Dispatch()
{
    if (g_NetworkType == TYPE_INTERNET && (g_FrameNo % 2) == 0)
    {
        if (nn::nex::Scheduler::GetInstance())
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }
    }
    nn::pia::common::Scheduler::GetInstance()->Dispatch();
}

PIA_SAMPLE_MAIN_FUNC()
{
    PIA_SAMPLE_LOG("Sample program Start.\n");

    // Determine whether launched from the friend list
    g_IsFromFriendList = false;
#if NN_PIA_CTR
    if (nn::friends::IsFromFriendList(&g_FriendIdForJoinIn))
    {
        // Get the presence information and check at some later point whether joinable.
        g_IsFromFriendList = true;
    }
#elif NN_PIA_CAFE
    SysArgSlot slots[2] = { { SYS_ARG_ID_JOIN_PID }, { SYS_ARG_ID_TERMINATOR } };
    SysCallerData callerData;

    if (SYSGetArguments(slots, &callerData) == SYS_ERROR_NONE && slots[0]._base.type == SYS_ARG_TYPE_U32)
    {
        g_FriendIdForJoinIn = slots[0]._u32.value;
        // Get the presence information and check at some later point whether joinable.
        g_IsFromFriendList = true;
    }
#endif

    // Creates the Pia library version information string.
    ::std::snprintf(s_SamplePiaVersionInfoString,
                    UPPER_SCREEN_WIDTH / FONT_SIZE,
                    "Pia %d.%d.%d r%d",
                    PIA_VERSION_MAJOR,
                    PIA_VERSION_MINOR,
                    PIA_VERSION_MICRO,
                    PIA_VERSION_REVISION);

    // Clears the information for rejoining.
    ResetRejoinSettings();

    ResetGameMode();

    nn::Result result;

#if PIA_TEST_CODE
    // Debugging feature (packet loss emulation)
    // Make sure that this does not get called in the retail version ROM.
    {
        nn::pia::transport::Transport::DebugSetting transportDebugSetting;
        transportDebugSetting.bPacketLossEmulation = true;
        result = nn::pia::transport::Transport::SetDebugSetting(transportDebugSetting);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
    }
#endif

    // Initialize the render framework for the Pia sample program.
    nn::pia::sampleframework::SampleFramework::CreateInstance();
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAME_PER_SECOND);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.0f, 0.0f, 0.2f, 1.0f));
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_START, false);

    Basic_SetProgramTitle("Session Advanced");

    while (true)
    {
        SampleSceneNo nextSceneNo = SCENE_START;

        // Setup of the features used in the PiaCommon module.
        result = Common_InitializePiaCommon(
            s_PiaMemoryBlock,
            MEMORY_BLOCK_SIZE_PIA,
            nn::pia::common::TRACE_FLAG_COMMON |
            nn::pia::common::TRACE_FLAG_LOCAL |
            nn::pia::common::TRACE_FLAG_TRANSPORT |
            nn::pia::common::TRACE_FLAG_SESSION |
            nn::pia::common::TRACE_FLAG_SESSION_JOINT |
            nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
            nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
            nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
            nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
            nn::pia::common::TRACE_FLAG_INET |
            nn::pia::common::TRACE_FLAG_ALWAYS,
            true,
            true);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        // Because the processes for each feature are clearly separated in this sample, the processes in the main thread are divided into several scenes.
        bool isExit = false;

/* Game loop. */
        while (isExit == false)
        {
            switch (nextSceneNo)
            {
            case SCENE_START:
            {
                nextSceneNo = SceneStart();
            }
            break;
            case SCENE_PREPARE_INET:
            {
                nextSceneNo = ScenePrepareInternet();
            }
            break;
            case SCENE_PREPARE_LOCAL:
            {
                nextSceneNo = ScenePrepareLocal();
            }
            break;
            case SCENE_DUMMY_SESSION:
            {
                nextSceneNo = SceneDummySession();
            }
            break;
            case SCENE_SELECT_MATCHMAKE_TYPE:
            {
                nextSceneNo = SceneSelectMatchmakeType();
            }
            break;
            case SCENE_RANDOM_MATCHMAKE:
            {
                nextSceneNo = SceneRandomMatchmake();
            }
            break;
            case SCENE_BROWSE_MATCHMAKE:
            {
                nextSceneNo = SceneBrowseMatchmake();
            }
            break;
            case SCENE_FRIENDS_MATCHMAKE:
            {
                nextSceneNo = SceneFriendsMatchmake();
            }
            break;
            case SCENE_COMMUNICATE:
            {
                nextSceneNo = SceneCommunication();
            }
            break;
            case SCENE_ERROR:
            {
                nextSceneNo = SceneError();
                Session_CleanupSession();
            }
            break;
            case SCENE_NETWORK_ERROR:
            {
                nextSceneNo = SceneNetworkError();
            }
            break;
            case SCENE_RESET:
            {
                nextSceneNo = SceneReset();
            }
            break;
            case SCENE_EXIT:
            {
                isExit = true;
            }
            break;
            }
        }

        // Finalizes PiaCommon.
        Common_FinalizePiaCommon();
    }
    // Control does not arrive here.
    // nn::pia::sampleframework::SampleFramework::DestroyInstance();
    // PIA_SAMPLE_LOG("End the sample program.\n");
}

SampleSceneNo SceneStart()
{
    nn::Result result;
    SampleSceneNo nextSceneNo = SCENE_START;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_START,
        false,
        false,
        0);

    g_FrameNo = 0;

    if (g_IsFromFriendList)
    {
        g_NetworkType = TYPE_INTERNET;
        nextSceneNo = SCENE_PREPARE_INET;
    }
    else
    {
        while (nextSceneNo == SCENE_START)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

#if NN_PIA_CTR
            // A Button:
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                g_NetworkType = TYPE_LOCAL;
                nextSceneNo = SCENE_PREPARE_LOCAL;
            }
#endif
            // X Button:
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                g_NetworkType = TYPE_INTERNET;
                nextSceneNo = SCENE_PREPARE_INET;
            }
            // B Button.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                g_NetworkType = TYPE_INTERNET;
                nextSceneNo = SCENE_DUMMY_SESSION;
            }
            // L Button:
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                g_IsEnableBitRateCheck = !g_IsEnableBitRateCheck;
                PIA_SAMPLE_LOG("Change BitRate check mode: %s\n", g_IsEnableBitRateCheck ? "Enable" : "Disable");
            }

            // Lower screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %d", PROGRAM_NAME, g_FrameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Select NetworkType.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 6);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Start Internet session");
            line += FONT_SIZE;
#if NN_PIA_CTR
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Start Local session");
            line += FONT_SIZE;
#endif
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Start Dummy session");
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            ++g_FrameNo;
        }
    }

    return nextSceneNo;
}

SampleSceneNo SceneReset()
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_RESET, false);

    // The process comes here before creating a Session instance if <tt>ac::Connect</tt> has failed, so do a check.
    if (nn::pia::session::Session::GetInstance())
    {
        Session_CleanupSession();
        nn::pia::session::Session::GetInstance()->UnregisterSessionEventCallback();
    }
    if (g_NetworkType == TYPE_INTERNET)
    {
        Nex_GetNgsFacadeInstance()->UnregisterNotificationEventHandler(&g_NexNotificationHandler);
        Inet_Logout();
        Inet_FinalizeNetwork();
    }

    Session_FinalizePiaSession();
    Transport_FinalizePiaTransport(GetUnreliableProtocolHandle(), 0);
    if (g_NetworkType == TYPE_INTERNET)
    {
        Inet_FinalizePiaInet();
        Nex_FinalizeNex();
    }
#if NN_PIA_CTR
    else
    {
        Local_FinalizePiaLocal();
    }
#endif

    return SCENE_START;
}

SampleSceneNo GetNextSceneByErrorResult(nn::Result result, SampleErrorType errorType)
{
    SampleSceneNo nextScene;

    // The socket failed to initialize, the network connection was dropped, etc.
    if (result == SampleFailedNetworkInterfaceUp() ||
        result == nn::pia::ResultNetworkConnectionIsLost() ||
        result == nn::pia::ResultSessionConnectionIsLost())
    {
        SetCommunicationErrorType(errorType);
        nextScene = SCENE_ERROR;
    }
    else if (result == nn::pia::ResultInvalidArgument() ||
             result == nn::pia::ResultInvalidState() ||
             result == nn::pia::ResultIncompatibleFormat())
    {
        // Programming error.
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        SetCommunicationErrorType(UNKNOWN);
        nextScene = SCENE_ERROR;
    }
    else
    {
        SetCommunicationErrorType(errorType);
        SetNetworkErrorResult(result);
        nextScene = SCENE_NETWORK_ERROR;
    }
    return nextScene;
}

void ResetGameMode()
{
    nn::Result result;

    s_FriendGameMode.joinGameMode = 0;
    s_FriendGameMode.matchmakeSystemType = nn::nex::MATCHMAKE_SYSTEM_TYPE_INVALID;
    s_FriendGameMode.joinGroupId = 0;
    s_FriendGameMode.ownerPrincipalId = 0;
    s_FriendGameMode.joinAvailabilityFlag = FRIEND_JOIN_AVAILABILITY_NOT_JOINABLE;
    s_pDescription = SAMPLE_DESCRIPTION;
#if NN_PIA_CTR
    if (nn::friends::IsInitialized())
#elif NN_PIA_CAFE
    if (nn::fp::IsInitialized())
#endif
    {
        result = Inet_UpdateGameMode(s_FriendGameMode, s_pDescription);
        if (result.IsFailure())
        {
            PIA_SAMPLE_LOG("Inet_UpdateGameMode is failed.\n");
        }
    }
}

void UpdateGameMode()
{
    nn::Result result;

    if (s_FriendGameMode.joinAvailabilityFlag == FRIEND_JOIN_AVAILABILITY_JOINABLE)
    {
        s_FriendGameMode.joinGroupId = nn::pia::session::Session::GetInstance()->GetSessionId();
        nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
        nn::pia::transport::Station* pHostStation = nn::pia::transport::StationManager::GetInstance()->GetStation(hostStationId);
        pHostStation->GetPrincipalId(&s_FriendGameMode.ownerPrincipalId);
    }
#if NN_PIA_CTR
    if (nn::friends::IsInitialized())
#elif NN_PIA_CAFE
    if (nn::fp::IsInitialized())
#endif
    {
        result = Inet_UpdateGameMode(s_FriendGameMode, s_pDescription);
        if (result.IsFailure())
        {
            PIA_SAMPLE_LOG("Inet_UpdateGameMode is failed.\n");
        }
    }
}

SampleSceneNo SetGameMode(u32 mode, u32 type, bool isJoinable, const wchar_t* pDescription)
{
    nn::Result result;

    s_FriendGameMode.joinGameMode = mode;
    s_FriendGameMode.matchmakeSystemType = type;
    s_pDescription = pDescription;
    if (isJoinable)
    {
        s_FriendGameMode.joinGroupId = nn::pia::session::Session::GetInstance()->GetSessionId();
        nn::pia::StationId hostStationId = nn::pia::session::Session::GetInstance()->GetHostStationId();
        nn::pia::transport::Station* pHostStation = nn::pia::transport::StationManager::GetInstance()->GetStation(hostStationId);
        pHostStation->GetPrincipalId(&s_FriendGameMode.ownerPrincipalId);
        s_FriendGameMode.joinAvailabilityFlag = FRIEND_JOIN_AVAILABILITY_JOINABLE;
    }
    else
    {
        s_FriendGameMode.joinGroupId = 0;
        s_FriendGameMode.ownerPrincipalId = 0;
        s_FriendGameMode.joinAvailabilityFlag = FRIEND_JOIN_AVAILABILITY_NOT_JOINABLE;
    }
#if NN_PIA_CTR
    if (nn::friends::IsInitialized())
#elif NN_PIA_CAFE
    if (nn::fp::IsInitialized())
#endif
    {
        result = Inet_UpdateGameMode(s_FriendGameMode, s_pDescription);
        if (result.IsFailure())
        {
            SetNetworkErrorResult(result);
            return SCENE_NETWORK_ERROR;
        }
    }
    return SCENE_COMMUNICATE;
}


// When DDL files are not used with NEX, a definition of the empty InitDOClasses function is required.
namespace nn
{
namespace nex
{
void InitDOClasses()
{
}
}
}
