/*---------------------------------------------------------------------------*
  Project: Pia
  File: SampleConfig.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#if NN_PIA_CTR
#define USE_PIA_LOCAL 1
#elif NN_PIA_CAFE
#define USE_PIA_LOCAL 0
#elif NN_PIA_WIN
#define USE_PIA_LOCAL 0
#endif

// Checks whether the macro is appropriate for the platform.
#if NN_PIA_CAFE
#if USE_PIA_LOCAL == 1
#error  "PiaLocal is not available in Cafe."
#endif
#endif

#if USE_PIA_LOCAL
#include <pia/local.h>
#endif // End of USE_PIA_LOCAL.
#include <pia/inet.h>

#include <pia/transport.h>
#include <pia/session.h>

#include "Shared_Nex_Setup.h"

#define USE_NEX_AUTOMATCH 0

#define FRAME_PER_SECOND 60

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

const char PROGRAM_NAME[] = "Mesh";
const size_t SAMPLE_USER_NAME_LENGTH = 11;  //!< The maximum number of characters for <tt>UserName</tt> drawing, including the terminating character.

/*!
    @brief  Enumerates possible states of the sample program.
*/
enum SampleSceneNo
{
    SCENE_SELECT_NETWORK_TYPE,  //!< Network type selection scene. This state is simply waiting for user input and is not related to communication.
    SCENE_MATCH_MAKE,           //!< This state creates sessions for joining and handles the joining process for each of the specified network types.
    SCENE_PREPARE_SESSION,      //!< This state initializes <tt>transport</tt> and <tt>session</tt>, runs the <tt>Session::CreateMesh</tt> and <tt>Session::JoinMesh</tt> functions, and waits for asynchronous processing to finish.
    SCENE_COMMUNICATE,          //!< In this state, communication with peers occurs within the session.
    SCENE_ERROR,                //!< This is the state that is transitioned to when problems occur. This demo simulates the type of error messages found in an actual application window.
    SCENE_NETWORK_ERROR,        //!< This is the state showing network error codes. Assume that the error viewer or EULA applet will normally start in this state.
    SCENE_RESET                 //!< Defines the state where all information managed by the demo has been reset to the default values.
#if PIA_TEST_CODE
    ,
    SCENE_WAIT
#endif
};

/*!
    @brief  An enumerated type indicating the network type.
*/
enum NetworkType
{
    TYPE_UNKNOWN,
    TYPE_LOCAL,    //!< Indicates local (UDS) communication.
    TYPE_INTERNET  //!< Indicates Internet communication.
};

/*!
    @brief  Enumerates communication errors handled by the sample program.
*/
enum SampleErrorType
{
    NO_ERROR = 0,                   //!< Specifies that no error occurred.
    INITIALIZE_ERROR_WIRELESS_OFF,  //!< Specifies that initialization failed because wireless mode is off.
    INITIALIZE_ERROR_OTHER_REASON,  //!< Specifies that initialization failed for some other reason.
    MATCH_MAKE_ERROR,               //!< Specifies that the session could not be established.
    SESSION_LOST_ERROR,             //!< Specifies that the session was disconnected during communication.
    CONNECTION_LOST_ERROR           //!< Specifies that the network connection was disconnected.
};

void SetCommunicationErrorType(SampleErrorType error);  //Sets the communication error type.
void SetNetworkErrorResult(nn::Result result);          //This function sets the <tt>Result</tt> value to be converted into a network error code.

bool InitializePiaLocal();
bool InitializePiaInet();
void InitializeTransportAndSession(NetworkType nwType);

void FinalizePiaLocal();
void FinalizePiaInet();
void FinalizePiaTransport();
void FinalizePiaSession();
void CleanupPiaSession();

u32 GetIncrementalGameModeValueByBitRateCheckMode();

// For use by <tt>Inet</tt>.
nn::Result NexFacadeStartup();
bool OpenParticipation();
bool GetHostStationConnectionInfo(nn::pia::transport::StationConnectionInfo* pInfo);

/* ---< State by process. >--- */
SampleSceneNo sceneSelectNetworkType(NetworkType* pNwType);
SampleSceneNo SceneLocalMatchmake(bool* pIsSessionHost);
SampleSceneNo SceneInetMatchmake(bool* pIsSessionHost);
SampleSceneNo ScenePrepareSession(NetworkType nwType, bool isSessionHost);
SampleSceneNo SceneCommunication(NetworkType nwType);
SampleSceneNo SceneError();
SampleSceneNo SceneNetworkError();

/*!
    @brief  Method used for the ID token within the sample programs.
*/
enum SampleTokenType
{
    TOKEN_TYPE_ALPHA,
    TOKEN_TYPE_BETA,
    TOKEN_TYPE_NUM  // Number of token use methods.
};

void SetLocalIdentificationToken(SampleTokenType tType);

/* ---< Definitions of structures used in the sample. >--- */

struct RectPosition  //!< Stores rectangle position coordinates.
{
    u16 x;
    u16 y;
};

struct SampleStationStatus
{
    nn::pia::StationId stationId;
    RectPosition pos;
    u32 latestIncrementNumber;
    u32 principalId;
    u8 name[SAMPLE_USER_NAME_LENGTH];
    u8 useRoundRobin;
    s32 rtt;
    nn::pia::StationId routeInfoId;

    void Serialize(void* pBuf)
    {
        PIA_SAMPLE_ASSERT(pBuf != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pBuf) % 4 == 0);

        u8* p = static_cast<u8*>(pBuf);

*reinterpret_cast<u64*>(p) = nn::pia::common::ByteOrder::HostToNetwork64(stationId.GetValue());
        p += 8;
*reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.x);
        p += 2;
*reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.y);
        p += 2;
*reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(latestIncrementNumber);
        p += 4;
*reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(principalId);
        p += 4;
        for (int i = 0; i < SAMPLE_USER_NAME_LENGTH - 1; i++)
        {
*p = name[i];
            ++p;
        }
*p = useRoundRobin;
    }

    void Deserialize(const void* pData)
    {
        PIA_SAMPLE_ASSERT(pData != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pData) % 4 == 0);

        const u8* p = static_cast<const u8*>(pData);

        stationId = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(p));
        p += 8;
        pos.x = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        pos.y = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        latestIncrementNumber = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
        p += 4;
        principalId = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
        p += 4;
        for (int i = 0; i < SAMPLE_USER_NAME_LENGTH - 1; i++)
        {
            name[i] = *p;
            ++p;
        }
        name[SAMPLE_USER_NAME_LENGTH - 1] = 0;  // Terminating character.
        useRoundRobin = *p;
    }

    void UpdateInfo(bool enableRelay)
    {
        // Update RTT.
        nn::pia::transport::Station* pStationBuf = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
        if (PIA_IS_VALID_POINTER(pStationBuf))
        {
            // Gets the RTT from the station instance corresponding to <tt>StationId</tt>.
            rtt = pStationBuf->GetRtt();
        }
        else
        {
            // Set to an invalid value if you cannot get the station instance.
            rtt = nn::pia::transport::Station::INVALID_RTT;
        }

        // Update the connection route information.
        if (enableRelay)
        {
            // If the relay feature is enabled, get the connection route information using the function for session debugging.
            // For the relay route information, StationIndex.
            nn::pia::StationIndex stationIndex, routeInfoIndex;
            nn::Result r = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);

            r = nn::pia::session::Mesh::GetInstance()->GetConnectionRoute(stationIndex, &routeInfoIndex);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);

            r = nn::pia::transport::Transport::GetInstance()->ConvertToStationId(&routeInfoId, routeInfoIndex);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
        else
        {
            // If the relay feature is disabled, the direct route is the only route.
            routeInfoId = stationId;
        }
    }
};

/* LocalNetwork-related settings. */
const bit32 SAMPLE_TITLE_ID = 0xFF310;       //!< Defines the unique ID of this program.
const u8 SAMPLE_SUB_ID = 0x00;               //!< Defines the sub-ID.
const u8 SAMPLE_CHANNEL = 11;                //!< Defines the channel used by the sample program. This channel is explicitly set to 11 because the first run of debugging hardware cannot communicate successfully on channel 1.
const bool ENABLE_AROUND_NETWORK_SEARCH = true;  //!< Defines whether to enable Nearby Network Search.

const char SAMPLE_PASSPHRASE[] = __FILE__;  //!< Defines the passphrase used to generate the encryption key for UDS communication. In an actual application, make sure that the passphrase is complex and sufficiently long.

const u8 SAMPLE_MAX_ENTRY = 5;    //!< Maximum number of connections. In an actual application, set according to the application's specifications.
const u8 GATHERING_LIST_MAX = 5;  //!< Maximum number of found networks to show.

#if USE_PIA_LOCAL
//!< Options when sending. If you use <tt>FORCE_DIRECT_BC</tt>, communication is low latency among UDS clients, but problems occur with hidden terminals.
const bit8 UDS_SEND_OPTION = nn::uds::FORCE_DIRECT_BC;
//!< Options when receiving. Basically, when using PiaSession, specify NO_WAIT.
const bit8 UDS_RECEIVE_OPTION = nn::uds::NO_WAIT;
//!< Application data to set to the beacon.
struct SampleBeaconApplicationData
{
    u32 value;
    char programName[64];  //Specifies information exchanged as a text string. A program name is entered in this demo.
};
#endif  // End of USE_PIA_LOCAL.

/* Thread priority settings. */
#if NN_PIA_CTR
const s32 SEND_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
#elif NN_PIA_CAFE
const s32 SEND_THREAD_PRIORITY = OS_PRIORITY_APP_DEFAULT - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
#elif NN_PIA_WIN
const s32 SEND_THREAD_PRIORITY = THREAD_PRIORITY_NORMAL - 1;  //The transport send thread.
#else
#error  "Invalid platform.
#endif
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1; //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.

/* NEX-Related Settings */
// Condition group used when building or searching for matchmaking sessions on a NEX game server.
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 35000;  // Game mode.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + (MATCHMAKE_GAME_MODE_OFFSET != 0) ? 100 + MATCHMAKE_GAME_MODE_OFFSET : MATCHMAKE_GAME_MODE_OFFSET;
const u32 MATCHMAKE_ATTR_INDEX = 0;      //!< Index of the attribute to use.
const u32 MATCHMAKE_ATTR_VAL_MAX = 100;  //!< Maximum value to set for the attribute to use. (In this demo, it is a random number greater than 0 and less than the maximum value.)

/* Memory-Related Settings */
const u32 MEMORY_BLOCK_SIZE_PIA = 2 * 1024 * 1024;      //!< Defines the amount of memory allocated to Pia.
const u32 MEMORY_BLOCK_SIZE_NEX = 1024 * 1024;          //!< Defines the amount of memory allocated to NEX.
const u32 MEMORY_BLOCK_SIZE_NEX_EMERGENCY = 50 * 1024;  //!< Defines the amount of memory allocated to NEX for emergencies.

/* Drawing-Related Settings */
const size_t UPPER_SCREEN_WIDTH = 400;   //!< Defines the horizontal resolution of the upper screen.
const size_t UPPER_SCREEN_HEIGHT = 240;  //!< Defines the vertical resolution of the upper screen.

const size_t LOWER_SCREEN_WIDTH = 320;   //!< Defines the horizontal resolution of the lower screen.
const size_t LOWER_SCREEN_HEIGHT = 240;  //!< Defines the vertical resolution of the lower screen.

const int FONT_SIZE = 8;  //!< Defines the size of the font to draw.

const char* GetPiaVersionString();  //!< Function for getting the string for drawing the Pia library version information.

// Function for setting the <tt>Result</tt> used to indicate Internet-related errors.
void SetNetworkErrorResult(nn::Result result);

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel(NetworkType nwType);

#if PIA_TEST_CODE
void SetUsingInvalidSignature(bool isUsing);

enum ParamKey
{
    PARAM_KEY_ENABLE_HOST_MIGRATION = 0x00,
    PARAM_KEY_CURSOR_POS,
    PARAM_KEY_FOUND_SESSION_NUM,
    PARAM_KEY_GATHERING_LIST_MAX,
    PARAM_KEY_APPLICATION_DATA,
    PARAM_KEY_ALLOW_PARTICIPATING_STATE,
    PARAM_KEY_STATION_ID,
    PARAM_KEY_NETWORK_TYPE,
    PARAM_KEY_GETJOINMESHSEQUENCE = 0x30,
    PARAM_KEY_HOST_MIGRATION_STATE = 0x40,
    PARAM_KEY_IS_AROUND_NETWORK_SEARCH_ACTIVATED = 0x70,
    PARAM_KEY_FOUND_AROUND_NETWORK_NUM = 0x71,
    PARAM_KEY_FOUND_AROUND_NETWORK_1_APPLICATION_DATA = 0x80,
    PARAM_KEY_FOUND_AROUND_NETWORK_1_PARTICIPANTS = 0x81,
    PARAM_KEY_FOUND_AROUND_NETWORK_2_APPLICATION_DATA = 0x90,
    PARAM_KEY_FOUND_AROUND_NETWORK_2_PARTICIPANTS = 0x91
};
#endif
