﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneCommunication.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/inet.h>
#include <pia/transport.h>
#include <pia/session.h>

#if NN_PIA_CTR
#include "Shared_Local_Setup.h"
#endif
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Nex_Matchmake.h"
#include "Shared_Session_Mesh.h"
#include "SampleLib/SampleLib.h"
#include "SampleConfig.h"

#include <nex.h>

/*!

    @brief  This CPP file contains processing that is independent of the network type (<tt>session</tt> + <tt>transport</tt>).

*/

extern u8 s_GameModeOption;  // Option value for changing <tt>MATCHMAKE_GAME_MODE/SAMPLE_SUB_ID</tt>.
extern bool s_EnableHostMigration;  //!< Enables or disables the host migration feature. Fixed to enabled when the <tt>PiaSession</tt> host migration feature is used.
extern bool s_IsEnableBitRateCheck;

bool s_UsingInvalidSignature;

u32 GetJoinedGatheringID();  // The get function for displaying the <tt>GatheringID</tt> on the screen. (Implemented in <tt>SceneInetMatchmake.cpp</tt>.)
#if USE_PIA_LOCAL
// Functions related to the display of <tt>UserName</tt>. (Implemented in <tt>SceneLocalMatchmake.cpp</tt>.)
void CheckAndGetUserName(char userNameDst[], const nn::cfg::UserName& userNameSrc, size_t dstMaxSize);  // Converts <tt>UserName</tt> into a <tt>char</tt> array.
void SaveCurrentHostUserName(nn::pia::common::StationAddress hostStationAddress);                       // Gets UserName from the host's StationAddress and saves it.
void GetHostUserName(char hostUserName[]);                                                              // Gets the UserName of the UDS host in order to display it on the screen.
void UpdateBeaconData();
static nn::pia::local::UdsAroundNetworkInfo s_AroundNetworkInfoList[GATHERING_LIST_MAX];                // Information buffer for networks discovered by nearby network search.
#endif

// Variables, constants, and functions for dynamically changing the network topology setting of the session.
size_t s_TargetNetworkTopology = 0;
const size_t NETWORK_TOPOLOGY_TYPE_NUM = 2;
const nn::pia::session::NetworkTopology NETWORK_TOPOLOGY_TYPE_TARGETS[NETWORK_TOPOLOGY_TYPE_NUM] = {
    nn::pia::session::NetworkTopology_FullMesh,
    nn::pia::session::NetworkTopology_RelayMesh
};
const char* NETWORK_TOPOLOGY_TYPE_STRINGS[NETWORK_TOPOLOGY_TYPE_NUM] = {
    "FullMesh",
    "RelayMesh"
};

size_t GetNetworkTopologyTypeNum()
{
    return NETWORK_TOPOLOGY_TYPE_NUM;
}

nn::pia::session::NetworkTopology GetNetworkTopologyType(size_t index)
{
    return NETWORK_TOPOLOGY_TYPE_TARGETS[index];
}

const char* GetNetworkTopologyTypeString(size_t index)
{
    return NETWORK_TOPOLOGY_TYPE_STRINGS[index];
}

bool IsEnableRelay(size_t index)
{
    return index > 0;
}

namespace
{
SampleStationStatus s_StationStatus[SAMPLE_MAX_ENTRY];

u32 s_UnreliableProtocolHandle = 0;
u32 s_RoundRobinUnreliableProtocolHandle = 0;
bool s_EnableRoundRobin = false;
bool s_InitializeTransport = false;
bool s_InitializeSession = false;
#if PIA_TEST_CODE
bool s_IsKickNextJoinStation = false;
u32 s_KickoutTargetStation = 0;
#endif

const nn::pia::sampleframework::Color COLOR_OF_RECTANGLE[12] =
    {
        nn::pia::sampleframework::COLOR_FUCHSIA,
        nn::pia::sampleframework::COLOR_GRAY,
        nn::pia::sampleframework::COLOR_WHITE,
        nn::pia::sampleframework::COLOR_MAROON,
        nn::pia::sampleframework::COLOR_RED,
        nn::pia::sampleframework::COLOR_OLIVE,
        nn::pia::sampleframework::COLOR_YELLOW,
        nn::pia::sampleframework::COLOR_GREEN,
        nn::pia::sampleframework::COLOR_LIME,
        nn::pia::sampleframework::COLOR_TEAL,
        nn::pia::sampleframework::COLOR_NAVY,
        nn::pia::sampleframework::COLOR_PURPLE
    };

void sessionCallback(nn::pia::session::Mesh::EventType eventType, nn::pia::StationId stationId)
{
    const size_t strSize = 128;
    static char eventString[strSize] = {};

    nn::pia::StationIndex stationIndex;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
    NN_PIA_DUMMY_PARAM(result);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    PIA_SAMPLE_LOG("[sessionCallback] statId:%llu statIdx:%u\n", stationId.GetValue(), stationIndex);

    switch (eventType)
    {
    case nn::pia::session::Mesh::EVENT_JOIN:
    {
        nn::pia::StationId localStationId = nn::pia::session::Mesh::GetInstance()->GetLocalStationId();
        PIA_SAMPLE_ASSERT(localStationId != nn::pia::STATION_ID_INVALID);
        NN_PIA_DUMMY_PARAM_IN_RELEASE(localStationId);
        nn::pia::StationId hostStationId = nn::pia::session::Mesh::GetInstance()->GetHostStationId();
        PIA_SAMPLE_ASSERT(hostStationId != nn::pia::STATION_ID_INVALID);
        NN_PIA_DUMMY_PARAM_IN_RELEASE(hostStationId);

        PIA_SAMPLE_ASSERT(nn::pia::session::Mesh::GetInstance()->CheckStationIdIsValid(stationId));
        PIA_SAMPLE_ASSERT(s_StationStatus[stationIndex].stationId == nn::pia::STATION_ID_INVALID);  //Notifications for connection events for the same station should not occur multiple times.

        s_StationStatus[stationIndex].stationId = stationId;  //Only shows whether it is valid.
        s_StationStatus[stationIndex].pos.x = 0;
        s_StationStatus[stationIndex].pos.y = 0;
        s_StationStatus[stationIndex].latestIncrementNumber = 0;
        s_StationStatus[stationIndex].useRoundRobin = 0;
        s_StationStatus[stationIndex].rtt = nn::pia::transport::Station::INVALID_RTT;

        nn::pia::transport::Station::PlayerName name;
        nn::pia::transport::Station::IdentificationToken token;

        result = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId)->GetPrincipalId(&s_StationStatus[stationIndex].principalId);
        PIA_SAMPLE_ASSERT(result.IsSuccess());

        result = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId)->GetPlayerName(&name);
        PIA_SAMPLE_ASSERT(result.IsSuccess());

        for (int i = 0; i < SAMPLE_USER_NAME_LENGTH - 1; i++)
        {
            if (i < name.length)
            {
                // Enable only the ASCII range to display.
                s_StationStatus[stationIndex].name[i] = (name.data[i] >= 0x20 && name.data[i] < 0x7f) ? static_cast<u8>(name.data[i]) : '?';
            }
        }

        result = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId)->GetIdentificationToken(&token);
        PIA_SAMPLE_ASSERT(result.IsSuccess());

        ::std::snprintf(eventString, strSize, "%s is call(%d[%s], %hu) - Name:%s PID:%08x Token:%s", __func__, eventType, "JOIN", stationIndex, s_StationStatus[stationIndex].name, s_StationStatus[stationIndex].principalId, token.data);

#if PIA_TEST_CODE
        if (s_IsKickNextJoinStation)
        {
            s_IsKickNextJoinStation = false;
            nn::pia::session::Mesh::GetInstance()->KickoutStation(stationId);
        }
#endif
    }
    break;
    case nn::pia::session::Mesh::EVENT_LEAVE:
    {
        PIA_SAMPLE_ASSERT(nn::pia::session::Mesh::GetInstance()->CheckStationIdIsValid(stationId));
        PIA_SAMPLE_ASSERT(s_StationStatus[stationIndex].stationId != nn::pia::STATION_ID_INVALID);  //Disconnection events also should not occur multiple times for the same station.

        s_StationStatus[stationIndex].stationId = nn::pia::STATION_ID_INVALID;  //Only shows whether it is valid.
        s_StationStatus[stationIndex].pos.x = 0;
        s_StationStatus[stationIndex].pos.y = 0;
        s_StationStatus[stationIndex].latestIncrementNumber = 0;
        s_StationStatus[stationIndex].principalId = 0;
        s_StationStatus[stationIndex].useRoundRobin = 0;
        s_StationStatus[stationIndex].rtt = nn::pia::transport::Station::INVALID_RTT;
        ::std::memset(s_StationStatus[stationIndex].name, 0, sizeof(s_StationStatus[stationIndex].name));

        ::std::snprintf(eventString, strSize, "%s is call(%d[%s], %hu, stationId::0x%llx)", __func__, eventType, "LEAVE", stationIndex, s_StationStatus[stationIndex].stationId.GetValue());
    }
    break;
    case nn::pia::session::Mesh::EVENT_HOST_CHANGED:
    {
// Transfer of session host privileges due to host migration.
#if USE_PIA_LOCAL
        // Get and save the <tt>UserName</tt> of the UDS host.
        nn::pia::transport::Station* pStation = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
        if (pStation != NULL)
        {
            SaveCurrentHostUserName(pStation->GetStationAddress());
        }
#else
// Does nothing.
#endif
        ::std::snprintf(eventString, strSize, "%s is call(%d[%s], %hu)", __func__, eventType, "HOST_CHANGED", stationIndex);
    }
    break;
    case nn::pia::session::Mesh::EVENT_MIGRATION_FAIL:
    {
        // The host migration process failed.
        // Does nothing.
        ::std::snprintf(eventString, strSize, "%s is call(%d[%s], %hu)", __func__, eventType, "MIGRATION_FAIL", stationIndex);
    }
    break;
    default:
    {
        ::std::snprintf(eventString, strSize, "%s is call(%d[%s], %hu)", __func__, eventType, "unknown", stationIndex);
    }
    break;
    }

    PIA_SAMPLE_LOG("%s\n", eventString);
}

bit8 s_LocalTokenData[nn::pia::transport::Station::IDENTIFICATION_TOKEN_MAX_DATA_SIZE] = {};

bool joinApprovalCallback(const nn::pia::transport::Station::IdentificationToken* pRequesterToken)
{
    // Callback that determines whether to approve or reject the join-in request. (Called only by the session host.)

    // Output log of the content of the identification token of the station where the passed request originated.
    PIA_SAMPLE_LOG("[joinApprovalCallback] RequesterToken Data:%s\n", pRequesterToken->data);

    bool joinApproval = false;

    // This sample makes the determination by comparing the first 20 characters with one's own identification token data. If they are the same, it approves the request.
    if (::std::memcmp(pRequesterToken->data, s_LocalTokenData, 20) == 0)
    {
        joinApproval = true;
    }

    PIA_SAMPLE_LOG("[joinApprovalCallback] %s\n", (joinApproval) ? "Approved" : "Denied");

    return joinApproval;
}

//The string for displaying the role screen within a session.
const char* SESSION_ROLE[] = {
    "Unknown",
    "Host",
    "Client"
};

#if USE_PIA_LOCAL
const char* GetAllowParticipatingStateString()
{
    switch(nn::pia::local::LocalNetwork::GetInstance()->GetAllowParticipatingState())
    {
    case nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_NONE:
        return "None";
    case nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_ALLOW:
        return "Allow";
    case nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_DISALLOW:
        return "Disallow";
    case nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_ALLOW_ONLY_RECONNECTION:
        return "Reconnection";
    }
    return "---";
}
#endif
}  // end of anonymous namespace

void FinalizePiaTransport();    //Handles cleanup of the PiaTransport class (forward declaration).
void CleanupPiaSession();       //Handles cleanup of the Session class (forward declaration).
void FinalizePiaSession();      //Handles cleanup of the PiaSession class (forward declaration).
void ResetMatchMakingClient();  //The cleanup function called in order to return to server matchmaking from communicating. (Implemented in <tt>SceneInetMatchmake.cpp</tt>.)


//When an API fails that has never failed before for a reason other than a problem in implementation, stops with an assert.
SampleSceneNo ScenePrepareSession(NetworkType nwType, bool isSessionHost)
{
    nn::Result result;

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_PREPARE_SESSION,
        false,
        false,
        0,
        GetLinkLevel(nwType));
#endif

    {
        nn::pia::transport::StationConnectionInfo stationConnectionInfo;
        // Sets its own identification token.
        nn::pia::transport::Station::IdentificationToken token;
        ::std::memcpy(token.data, s_LocalTokenData, nn::pia::transport::Station::IDENTIFICATION_TOKEN_MAX_DATA_SIZE);

        //Registers the event callback.
        //Before callback registration, initialize the structure that the callback accesses.
        for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
        {
            s_StationStatus[i].stationId = nn::pia::STATION_ID_INVALID;  //Only shows whether it is valid.
            s_StationStatus[i].pos.x = 0;
            s_StationStatus[i].pos.y = 0;
            s_StationStatus[i].latestIncrementNumber = 0;
            s_StationStatus[i].principalId = 0;
            s_StationStatus[i].useRoundRobin = 0;
            s_StationStatus[i].rtt = nn::pia::transport::Station::INVALID_RTT;
            s_StationStatus[i].routeInfoId = nn::pia::STATION_ID_INVALID;
            ::std::memset(s_StationStatus[i].name, 0, sizeof(s_StationStatus[i].name));
        }
        nn::pia::session::Mesh::GetInstance()->RegisterMeshEventCallback(sessionCallback);

        //Registers the callback for determining whether to approve or reject a join request.
        nn::pia::session::Mesh::GetInstance()->RegisterJoiningMeshApprovalCallback(joinApprovalCallback);

        //Set the pulse interval for the synchronization clock.
        //The default value is 2000 ms. This generally does not need to be changed, but we set it to 1000 ms here.
        nn::pia::session::Mesh::GetInstance()->SetSyncClockPulseInterval(1000);
        PIA_SAMPLE_ASSERT(nn::pia::session::Mesh::GetInstance()->GetSyncClockPulseInterval() == 1000);

        switch (nwType)  //Because only the <tt>Facade</tt> startup must be executed with this timing, the code is dependent on the network type.
        {
        case TYPE_LOCAL:
        {
#if USE_PIA_LOCAL
            // Sets up <tt>LocalFacade</tt>.
            result = nn::pia::local::LocalFacade::GetInstance()->Startup();
            if (result.IsFailure())
            {
                // The status of being disconnected from the UDS network.
                // It might also fail if the local communication host migration process is in progress.
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                return SCENE_ERROR;
            }

            // Start the Session instance after setting up the LocalFacade instance.
            {
                // Configure the packet with the signature feature.
                char signatureKeyData[] = "session_advanced key";
                nn::pia::common::SignatureSetting sampleSignatureSetting(
                    nn::pia::common::SignatureSetting::MODE_HMAC_MD5,  // Signature mode.
                    static_cast<const void*>(signatureKeyData),        // Signature key.
                    sizeof(signatureKeyData)                           // Signature key length.
                    );
#if PIA_TEST_CODE
                char dummySignatureKeyData[] = "dummyKey";
                if (s_UsingInvalidSignature)
                {
                    sampleSignatureSetting.Set(
                        nn::pia::common::SignatureSetting::MODE_HMAC_MD5, 
                        static_cast<const void*>(dummySignatureKeyData), 
                        sizeof(dummySignatureKeyData)
                        );
                }
#endif

                // Local communication uses the default setting.
                nn::pia::session::Mesh::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = s_EnableHostMigration;
                startupSetting.signatureSetting = sampleSignatureSetting;
                startupSetting.pCryptoSetting = NULL;
                startupSetting.pToken = &token;
                result = Session_StartupMesh(startupSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }

            if (isSessionHost)
            {
                result = nn::pia::session::Mesh::GetInstance()->CreateMeshAsync();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                // Waits for asynchronous processing in the same way as <tt>JoinMesh</tt>.
                PIA_SAMPLE_LOG("Start CreateMesh async.\n");
            }
            else
            {
                // Gets the <tt>StationConnectionInfo</tt> for the mesh host.
                // For local communications, the mesh host is always the host of the local network (the UDS host).
                result = nn::pia::local::LocalFacade::GetInstance()->GetHostStationConnectionInfo(&stationConnectionInfo);
                stationConnectionInfo.Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                //Because this process takes time, display a message such as "Please wait."
                if (nn::pia::session::Mesh::GetInstance()->JoinMeshAsync(stationConnectionInfo).IsFailure())
                {
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    return SCENE_ERROR;
                }
                PIA_SAMPLE_LOG("Start JoinMesh async.\n");
            }
#endif  // end of USE_PIA_LOCAL
        }
        break;
        case TYPE_INTERNET:
        {
            // NexFacade setup.
            result = NexFacadeStartup();
            if (result.IsFailure())
            {
                if (result == nn::pia::ResultDnsFailed())
                {
                    // DNS resolution failed.
                    PIA_SAMPLE_LOG("SceneInetMatchmake:NexFacadeStartup failed. (DnsFailed)\n");
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                }
                else if (result == nn::pia::ResultNatCheckFailed())
                {
                    // The NAT check failed. (It failed to create a socket, or NEX returned an empty address.)
                    PIA_SAMPLE_LOG("SceneInetMatchmake:NexFacadeStartup failed. (NatCheckFailed)\n");
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                }
                else if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    // Connection is in an abnormal state.
                    PIA_SAMPLE_LOG("SceneInetMatchmake:NexFacadeStartup failed. (ConnectionIsLost)\n");
                    SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                }
                else if (result == nn::pia::ResultCancelled())
                {
                    // The startup of NexFacade is cancelled.
                    PIA_SAMPLE_LOG("SceneInetMatchmake:NexFacadeStartup failed. (Cancelled)\n");
                    return SCENE_RESET;
                }
                else
                {
                    //Unexpected failure.
                    PIA_SAMPLE_LOG("SceneInetMatchmake:NexFacadeStartup failed. (Unknown)\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    PIA_SAMPLE_ASSERT(false);

                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                }

                SetNetworkErrorResult(result);
                return SCENE_NETWORK_ERROR;
            }

            // Start the <tt>Session</tt> instance after setting up the <tt>NexFacade</tt> instance.
            {
                // Configure the packet with the signature feature.
                // Convert the shared session key obtained from the server during matchmaking so that it can be used as the signature key.
                nn::pia::common::SignatureSetting* pSignatureSetting = Nex_GetMatchmakeSessionSignatureSetting();
                PIA_SAMPLE_ASSERT(pSignatureSetting);
                nn::pia::common::CryptoSetting cryptoSetting;
                cryptoSetting.m_Mode = nn::pia::common::CryptoSetting::MODE_NOTHING;

                nn::pia::session::Mesh::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = s_EnableHostMigration;
#if PIA_TEST_CODE
                char dummySignatureKeyData[32];
                ::memset(dummySignatureKeyData, 0, sizeof(dummySignatureKeyData));

                if (s_UsingInvalidSignature)
                {
                    startupSetting.signatureSetting.Set(
                        nn::pia::common::SignatureSetting::MODE_HMAC_MD5, 
                        static_cast<const void*>(dummySignatureKeyData), 
                        sizeof(dummySignatureKeyData)
                        );
                }
                else
                {
                    startupSetting.signatureSetting = *pSignatureSetting;
                }
#else
                startupSetting.signatureSetting = *pSignatureSetting;
#endif
                startupSetting.pCryptoSetting = &cryptoSetting;
                if(s_IsEnableBitRateCheck)
                {
                    startupSetting.uplinkBitRateLowerLimit = 200 * 1000;
                    startupSetting.bitRateCheckPacketSize = 1024;
                }
                startupSetting.pToken = &token;
                result = Session_StartupMesh(startupSetting);

                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }

            if (isSessionHost)
            {
                result = nn::pia::session::Mesh::GetInstance()->CreateMeshAsync();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                // Waits for asynchronous processing in the same way as <tt>JoinMesh</tt>.
                PIA_SAMPLE_LOG("Start CreateMesh async.\n");
            }
            else
            {
                //Gets the <tt>StationConnectionInfo</tt> for the session host.
                //Because this process takes time, display a message such as "Please wait."
                if (!GetHostStationConnectionInfo(&stationConnectionInfo))
                {
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    SetNetworkErrorResult(SampleFailedGetHostStationConnectionInfo());
                    return SCENE_NETWORK_ERROR;
                }
                result = nn::pia::session::Mesh::GetInstance()->JoinMeshAsync(stationConnectionInfo);
                if (result.IsFailure())
                {
                    if (result == nn::pia::ResultStationConnectionFailed())
                    {
                        // Failed to connect.
                        PIA_SAMPLE_LOG("!!!Join process Failed (StationConnectionFailed)!!!\n");
                    }

                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    SetNetworkErrorResult(result);
                    return SCENE_NETWORK_ERROR;
                }
            }
        }
        break;
        }
        // Cycles a render loop because JoinMesh takes time and nn::pia::common::Scheduler::Dispatch is required.
        int frameNo = 0;
        u32 linkLevel = GetLinkLevel(nwType);
        while ((isSessionHost && nn::pia::session::Mesh::GetInstance()->IsCompletedCreateMesh() == false) || (!isSessionHost && nn::pia::session::Mesh::GetInstance()->IsCompletedJoinMesh() == false))
        {
            bool errorOccurred = false;

            if (nwType == TYPE_INTERNET)
            {
                // Checks the AC network connection status, game server connection status, and NEX internal errors.
                if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
                {
                    errorOccurred = true;
                    result = nn::pia::ResultNetworkConnectionIsLost();
                }
                // To maintain communication with the NEX game server, you must call nn::nex::Scheduler::Dispatch.
                // In this demo, it is called in every frame. To reduce processing, you can call nn::nex::Scheduler::Dispatch less often.
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
#if USE_PIA_LOCAL
            else
            {
                // Checks the local network connection status.
                if (!Local_IsConnected())
                {
                    errorOccurred = true;
                    result = nn::pia::ResultNetworkConnectionIsLost();
                }
            }
#endif

            nn::pia::common::Scheduler::GetInstance()->Dispatch();

            if (!errorOccurred)
            {
                //Status check for the network connection.
                result = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
            }
            if (result.IsFailure())
            {
                errorOccurred = true;
                //If the network connection is abnormal:
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    //Returns when the network connection is interrupted because the wireless switch is turned off or communication fails.
                    PIA_SAMPLE_LOG("Connection Lost when waiting Create/Join process.\n");
                }
                else
                {
                    //Unexpected state.
                    PIA_SAMPLE_LOG("Transport::CheckConnectionError is unknown status when waiting Create/Join process.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    PIA_SAMPLE_ASSERT(false);
                }
            }
            if (errorOccurred)
            {
                // Call Cleanup because an error occurred during Session processing.
                nn::pia::session::Mesh::GetInstance()->Cleanup();

                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                if (nwType == TYPE_INTERNET)
                {
                    SetNetworkErrorResult(result);
                    return SCENE_NETWORK_ERROR;
                }
                else
                {
                    return SCENE_ERROR;
                }
            }
            
#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam( PARAM_KEY_GETJOINMESHSEQUENCE, nn::pia::session::Mesh::GetInstance()->GetJoinMeshSequence() );
#endif

            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
            int line = 0;

            //Lower screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Prepare Session (%d)", nn::pia::session::Mesh::GetInstance()->GetJoinMeshSequence());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(LOWER_SCREEN_WIDTH - (FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology));
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Please Wait...(loop=%d)", frameNo);
            line += FONT_SIZE;
            line += FONT_SIZE;

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 2);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Cancel JoinMesh");
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            //Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            frameNo++;

            // Update signal strength information periodically.
            if (frameNo % FRAME_PER_SECOND == 0)
            {
                linkLevel = GetLinkLevel(nwType);
#if PIA_TEST_CODE
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                    SCENE_PREPARE_SESSION,
                    false,
                    false,
                    0,
                    linkLevel);
#endif
            }

            //Handle user input.
            {
                // B: Cancel JoinMesh
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
                {
                    result = nn::pia::session::Mesh::GetInstance()->CancelJoinMeshAsync();
                    if (result.IsSuccess())
                    {
                        PIA_SAMPLE_LOG("CancelJoinMeshAsync is success.");
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("CancelJoinMeshAsync is failure.");
                        PIA_SAMPLE_LOG_RESULT(result);
                    }
                }
            }
        }

        //Confirm the result for asynchronous processing.
        {
            if (isSessionHost)
            {
                result = nn::pia::session::Mesh::GetInstance()->GetCreateMeshResult();
            }
            else
            {
                result = nn::pia::session::Mesh::GetInstance()->GetJoinMeshResult();
                if(result.IsSuccess())
                {
                    s32 bitRate = 0;
                    if(nn::pia::session::Mesh::GetInstance()->GetUplinkBitRate(&bitRate).IsSuccess())
                    {
                        PIA_SAMPLE_LOG("bitRate : %d\n", bitRate);
                    }
                }
            }
            if (!result.IsSuccess())
            {
                PIA_SAMPLE_LOG("Async process is failure.\n");
                if (result == nn::pia::ResultInvalidState())
                {
                    // It was not a state where connection was possible.
                    PIA_SAMPLE_LOG("Join process Failed (InvalidState)\n");
                }
                else if (result == nn::pia::ResultCancelled())
                {
                    // Canceled by the user.
                    PIA_SAMPLE_LOG("Join process Failed (Cancelled)\n");
                }
                else if (result == nn::pia::ResultStationConnectionFailed())
                {
                    // Failed to connect.
                    PIA_SAMPLE_LOG("Join process Failed (StationConnectionFailed)\n");
                }
                else if (result == nn::pia::ResultStationConnectionNatTraversalFailedUnknown())
                {
                    // The NAT-traversal process failed. The NAT type is unknown.
                    PIA_SAMPLE_LOG("Join process Failed (StationConnectionNatTraversalFailedUnknown)\n");
                }
                else if (result == nn::pia::ResultNatTraversalFailedBothEim())
                {
                    // NAT traversal failed. The NAT type was EIM on both the local station and the peer with which the connection failed.
                    PIA_SAMPLE_LOG("Join process Failed (NatTraversalFailedBothEim)\n");
                }
                else if (result == nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm())
                {
                    // NAT traversal failed. The NAT type was EIM on the local station and EDM on the peer with which the connection failed.
                    PIA_SAMPLE_LOG("Join process Failed (NatTraversalFailedLocalEimRemoteEdm)\n");
                }
                else if (result == nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim())
                {
                    // NAT traversal failed. The NAT type was EDM on the local station and EIM on the peer with which the connection failed.
                    PIA_SAMPLE_LOG("Join process Failed (NatTraversalFailedLocalEdmRemoteEim)\n");
                }
                else if (result == nn::pia::ResultNatTraversalFailedBothEdm())
                {
                    // NAT traversal failed. The NAT type was EDM on both the local station and the peer with which the connection failed.
                    PIA_SAMPLE_LOG("Join process Failed (NatTraversalFailedBothEdm)\n");
                }
                else if (result == nn::pia::ResultNatTraversalRequestTimeout())
                {
                    // Nat transversal has failed and the process has timed out.
                    PIA_SAMPLE_LOG("Join process Failed (ResultNatTraversalFailedRequestTimeout)\n");
                }
                else if (result == nn::pia::ResultJoinRequestDenied())
                {
                    // The session host denied the connection.
                    PIA_SAMPLE_LOG("Join process Failed (JoinRequestDenied)\n");
                }
                else if (result == nn::pia::ResultMeshIsFull())
                {
                    // The session that you wanted to participate in was full, so you could not participate.
                    PIA_SAMPLE_LOG("Join process Failed (MeshIsFull)\n");
                }
                else if (result == nn::pia::ResultInvalidSystemMessage())
                {
                    // Because an invalid system message was received, the process was ended.
                    PIA_SAMPLE_LOG("Join process Failed (InvalidSystemMessage)\n");
                }
                else if (result == nn::pia::ResultIncompatibleFormat())
                {
                    // The communication format of the peer is incompatible. Could not connect.
                    PIA_SAMPLE_LOG("Join process Failed (IncompatibleFormat)\n");
                }
                else if (result == nn::pia::ResultFailedStartupProtocol())
                {
                    // The system protocol start process failed.
                    PIA_SAMPLE_LOG("Create process Failed (FailedStartupProtocol)\n");
                }
                else if (result == nn::pia::ResultRelayFailedNoCandidate())
                {
                    // The relay connection failed because there were no candidates for a relay path.
                    PIA_SAMPLE_LOG("Join process Failed (RelayFailedNoCandidate)\n");
                }
                else if (result == nn::pia::ResultRelayFailedRttLimit())
                {
                    // The relay connection failed because the relay path exceeded the RTT limit.
                    PIA_SAMPLE_LOG("Join process Failed (RelayFailedRttLimit)\n");
                }
                else if (result == nn::pia::ResultRelayFailedRelayNumLimit())
                {
                    // The relay connection failed because the relay path exceeded the relay request limit.
                    PIA_SAMPLE_LOG("Join process Failed (RelayFailedRelayNumLimit)\n");
                }
                else if (result == nn::pia::ResultRelayFailedUnknown())
                {
                    // The relay connection failed (unknown reason).
                    PIA_SAMPLE_LOG("Join process Failed (RelayFailedUnknown)\n");
                }
                else
                {
                    // Any other errors.
                    PIA_SAMPLE_LOG("Join process Failed (Unknown)\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                }

                if (result == nn::pia::ResultCancelled())
                {
                    // On cancellation, return to the matchmaking screen.
                    CleanupPiaSession();

                    switch (nwType)
                    {
                    case TYPE_LOCAL:
                    {
#if USE_PIA_LOCAL
                        if (nn::pia::local::LocalFacade::GetInstance() != NULL)
                        {
                            nn::pia::local::LocalFacade::GetInstance()->Cleanup();
                        }
                        // Shut down the UDS network.
                        bool isSuccess = Local_DisconnectNetwork();
                        if (!isSuccess)
                        {
                            SetCommunicationErrorType(MATCH_MAKE_ERROR);
                            return SCENE_ERROR;
                        }
                        Local_CleanupPiaLocal();
#endif
                    }
                    break;
                    case TYPE_INTERNET:
                    {
                        // Finish cleanup of NexFacade before finalizing transport.
                        if (nn::pia::inet::NexFacade::GetInstance() != NULL)
                        {
                            nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                            nn::pia::inet::NexFacade::GetInstance()->Cleanup();
                        }
                        // End the matchmaking session.
                        ResetMatchMakingClient();
                    }
                    break;
                    }
                    return SCENE_MATCH_MAKE;
                }
                else
                {
                    // Treated as a matchmaking error.
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    if (nwType == TYPE_INTERNET)
                    {
                        SetNetworkErrorResult(result);
                        return SCENE_NETWORK_ERROR;
                    }
                    else
                    {
                        return SCENE_ERROR;
                    }
                }
            }
        }

        if (isSessionHost && nwType == TYPE_INTERNET)
        {
            // Use OpenParticipation because the Session (P2P communication) preparations are finished.
            //Because this process takes time, display a message such as "Please wait."
            if (!OpenParticipation())
            {
                PIA_SAMPLE_LOG("ScenePrepareSession:OpenParticipation failed.\n");
                SetNetworkErrorResult(SampleFailedOpenParticipation());
                SetCommunicationErrorType(MATCH_MAKE_ERROR);
                return SCENE_NETWORK_ERROR;
            }
        }
    }
    return SCENE_COMMUNICATE;  //Execution succeeded. Start communication.
}
/*!
    @brief  Handles the state in which communication is performed using a <tt>PiaTransport</tt> instance.
    
    When a host shuts down a network, it enters the select network state, but clients enter the error state.
    
    @return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo SceneCommunication(NetworkType nwType)
{
    nn::Result result;

    bool isExit = false;          //Represents the condition for exiting the loop so that <tt>EndFrame</tt> does not need to be considered.
    bool isExitManually = false;  //Specifies the flag for detecting cases other than manual disconnections from the network. Disconnections not intended by the local station are treated as errors.
    u32 linkLevel = GetLinkLevel(nwType);
    int frameNo = 0;

    bool isBackToMatchmake = true;  //Represents a flag indicating whether to return to server matchmaking.

    nn::pia::StationId myStationId = nn::pia::session::Mesh::GetInstance()->GetLocalStationId();

    nn::pia::StationIndex myStationIndex;
    result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&myStationIndex, myStationId);
    PIA_SAMPLE_ASSERT(result.IsSuccess());

    for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
        if (s_StationStatus[i].stationId == myStationId)  //Update the data of the local station.
        {
            s_StationStatus[i].pos.x = 0;
            s_StationStatus[i].pos.y = (10 * i);
        }

    // For log output and screen display (For debugging.) of communication-related data.
    nn::pia::transport::StationConnectionInfo localInfo;
    nn::pia::transport::StationConnectionInfo hostInfo;

    nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
    nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&hostInfo);

#if USE_PIA_LOCAL
    u8 channel = 0;
    u32 foundAroundNetworkNum = 0;
#endif

    // Because the display data changes significantly, the output changes depending on the network type.
    if (nwType == TYPE_INTERNET)
    {
        PIA_SAMPLE_LOG("SceneCommunication: print Connection Information.(inet) - START\n");
        PIA_SAMPLE_LOG("  NatProperty Mapping:%d, Filtering:%d, PortIncrement:%d\n",
                          nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyMapping(),
                          nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyFiltering(),
                          nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyPortIncrement());

        nn::pia::common::String mainStr;
        nn::pia::common::String subStr;
        localInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
        localInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
        PIA_SAMPLE_LOG("  LocalAddress: Main[%s] Sub[%s]\n", mainStr.GetCharArray(), subStr.GetCharArray());
        hostInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
        hostInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
        PIA_SAMPLE_LOG("  HostAddress: Main[%s] Sub[%s]\n", mainStr.GetCharArray(), subStr.GetCharArray());
        PIA_SAMPLE_LOG("SceneCommunication: print Connection Information.(inet) - END\n");
    }
#if USE_PIA_LOCAL
    else
    {
        PIA_SAMPLE_LOG("SceneCommunication: print Connection Information.(local) - START\n");

        nn::pia::local::UdsStationInfo infoBuf;
        char localNameBuf[SAMPLE_USER_NAME_LENGTH] = {};
        char hostNameBuf[SAMPLE_USER_NAME_LENGTH] = {};
        // Gets the UDS UserName of localStation.
        if (nn::pia::local::LocalNetwork::GetInstance()->GetStationInfo(&infoBuf, localInfo.GetMainLocation().GetStationAddress()).IsSuccess())
        {
            CheckAndGetUserName(localNameBuf, infoBuf.m_UserName, SAMPLE_USER_NAME_LENGTH);
        }
        else
        {
            // GetStationInfo failed.
            ::std::strncpy(localNameBuf, "GET FAILED", SAMPLE_USER_NAME_LENGTH);
        }
        // Gets the UDS UserName of hostStation.
        if (nn::pia::local::LocalNetwork::GetInstance()->GetStationInfo(&infoBuf, hostInfo.GetMainLocation().GetStationAddress()).IsSuccess())
        {
            CheckAndGetUserName(hostNameBuf, infoBuf.m_UserName, SAMPLE_USER_NAME_LENGTH);
        }
        else
        {
            // GetStationInfo failed.
            ::std::strncpy(hostNameBuf, "GET FAILED", SAMPLE_USER_NAME_LENGTH);
        }
        PIA_SAMPLE_LOG("  UserName  Local=%s, Host=%s\n", localNameBuf, hostNameBuf);
        PIA_SAMPLE_LOG("SceneCommunication: print Connection Information.(local) - END\n");

        // Saves the <tt>UserName</tt> of the UDS host.
        SaveCurrentHostUserName(hostInfo.GetMainLocation().GetStationAddress());

        // Gets the wireless channel.
        channel = nn::pia::local::LocalNetwork::GetInstance()->GetNetworkChannel();
    }
#endif
#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_COMMUNICATE,
        nn::pia::session::Mesh::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID,
        nn::pia::session::Mesh::GetInstance()->IsHost(),
        nn::pia::session::Mesh::GetInstance()->GetStationNum(),
        linkLevel);
#endif

    // Main loop.
    //Because data is sent and received in separate threads, only update the local station data, and then render.
    while (isExit == false)
    {
        bool errorOccurred = false;

        //Disconnection detected.
        if (nwType == TYPE_INTERNET)
        {
            // Checks the AC network connection status, game server connection status, and NEX internal errors.
            if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
            {
                errorOccurred = true;
                result = nn::pia::ResultNetworkConnectionIsLost();
            }
        }
#if USE_PIA_LOCAL
        else
        {
            // Checks the local network connection status.
            if (!Local_IsConnected())
            {
                errorOccurred = true;
                result = nn::pia::ResultNetworkConnectionIsLost();
            }
        }
#endif
        if (!errorOccurred)
        {
            //Status check for the network connection.
            result = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
        }
        if (result.IsFailure())
        {
            errorOccurred = true;
            //If the network connection is abnormal:
            if (result == nn::pia::ResultNetworkConnectionIsLost())
            {
                // Returns when the network connection is interrupted because the wireless switch is turned off or communication fails.
                PIA_SAMPLE_LOG("Connection Lost when scene Communication.\n");
            }
            else
            {
                //Unexpected state.
                PIA_SAMPLE_LOG("Transport::CheckConnectionError is unknown status when scene Communication.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                PIA_SAMPLE_ASSERT(false);
            }
        }
        if (errorOccurred)
        {
            // Call Cleanup because an error occurred during Session processing.
            nn::pia::session::Mesh::GetInstance()->Cleanup();

            SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            if (nwType == TYPE_INTERNET)
            {
                SetNetworkErrorResult(result);
                return SCENE_NETWORK_ERROR;
            }
            else
            {
                return SCENE_ERROR;
            }
        }
        //Session status check.
        //In this sample, operation assumes that the session always ends in a disconnect state at the end even for a normal finalization.
        if (nn::pia::session::Mesh::GetInstance()->CheckConnectionError() == nn::pia::ResultMeshConnectionIsLost())
        {
            PIA_SAMPLE_LOG("SessionState became DISCONNECTED.\n");
            break;  //In this case, exiting is forced.
        }

        // Periodically update signal strength information.
        if (frameNo % FRAME_PER_SECOND == 0)
        {
            linkLevel = GetLinkLevel(nwType);
        }
#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
            SCENE_COMMUNICATE,
            nn::pia::session::Mesh::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID,
            nn::pia::session::Mesh::GetInstance()->IsHost(),
            nn::pia::session::Mesh::GetInstance()->GetStationNum(),
            linkLevel);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam( 
            PARAM_KEY_STATION_ID, 
            static_cast<u32>(nn::pia::session::Mesh::GetInstance()->GetLocalStationId().GetValue()));
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
            PARAM_KEY_HOST_MIGRATION_STATE, 
            reinterpret_cast<nn::pia::common::Job*>(nn::pia::session::Mesh::GetInstance()->GetProcessHostMigrationJobPtr())->GetState());

#if USE_PIA_LOCAL
        if (nwType == TYPE_LOCAL)
        {
            if (nn::pia::local::LocalNetwork::GetInstance()->IsConnected())
            {
                PIA_SAMPLE_ASSERT(nn::pia::local::LocalNetwork::GetInstance()->GetAllowParticipatingState() != nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_NONE);
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_ALLOW_PARTICIPATING_STATE, nn::pia::local::LocalNetwork::GetInstance()->GetAllowParticipatingState());
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_IS_AROUND_NETWORK_SEARCH_ACTIVATED, nn::pia::local::LocalNetwork::GetInstance()->IsAroundNetworkSearchActivated() ? 1 : 0);
        }
#endif
#endif

        if (nwType == TYPE_INTERNET)
        {
            // To maintain communication with the NEX game server, you must call nn::nex::Scheduler::Dispatch.
            // In this demo, it is called in every frame. To reduce processing, you can call nn::nex::Scheduler::Dispatch less often.
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        //Just in case, checks whether there are duplicates of <tt>StationAddress</tt>.
        nn::pia::transport::StationManager* pStationManager = nn::pia::transport::StationManager::GetInstance();
        if (pStationManager)
        {
            // Check whether there is a station with the same <tt>StationAddress</tt> among the valid stations belonging to the <tt>StationManager</tt>.
            int i, j;
            i = 0;
            for (nn::pia::transport::StationManager::Iterator it = pStationManager->Begin(); it != pStationManager->End(); ++it)
            {
                nn::pia::transport::StationManager::Iterator itBuf = it;
                j = i + 1;
                for (++itBuf; itBuf != pStationManager->End(); ++itBuf)
                {
                    if (it->GetStationAddress() == itBuf->GetStationAddress())
                    {
                        PIA_SAMPLE_LOG("[TEST] found same StationAddress.\n");
                        PIA_SAMPLE_LOG("Station Idx:%d [0x%08x]:\n", i, &(*it));
                        it->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
                        PIA_SAMPLE_LOG("Station Idx:%d [0x%08x]:\n", j, &(*itBuf));
                        itBuf->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);

                        nn::pia::session::Mesh::GetInstance()->TraceStationConnectionInfoList(nn::pia::common::TRACE_FLAG_ALWAYS);

                        PIA_SAMPLE_ASSERT(false);
                    }
                    ++j;
                }
                ++i;
            }
        }

        //Draw to the lower screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            //String.
            {
                //Check the role of the local station.
                int roleIndex = 0 /*Unknown.*/;
                if (nn::pia::session::Mesh::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID)
                {
                    roleIndex = (nn::pia::session::Mesh::GetInstance()->IsHost() ? 1 /*Host*/ : 2 /*Client.*/);
                }

                int line = 0;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Communication");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   (StationId=%d, Role=%s)", myStationIndex, SESSION_ROLE[roleIndex]);  //line +=FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(LOWER_SCREEN_WIDTH - (FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
                line += FONT_SIZE;

#if USE_PIA_LOCAL
                // List networks found by AroundNetworkSearch.
                if (nwType == TYPE_LOCAL && ENABLE_AROUND_NETWORK_SEARCH)
                {
                    // AroundNetworkSearch process. (Searches for communicating networks.)
                    // Because getting a complete list takes time, updating it more than a few times per second makes no sense. Here, the list is updated once every 10 frames.
                    // Only clients can list nearby networks due to CTR hardware limitations. A process for sharing the list via communications can be used
                    //  The list cannot be created when there is only a single host, even though code is implemented in such a way that even the host can list surrounding networks.
                    if ((frameNo % 10) == 0)
                    {
                        result = nn::pia::local::LocalNetwork::GetInstance()->GetAroundNetworkInfoList(s_AroundNetworkInfoList, &foundAroundNetworkNum, GATHERING_LIST_MAX);
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("GetAroundNetworkInfoList failed.\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                            PIA_SAMPLE_ASSERT(false);
                        }
                    }

                    line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 16);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[Around Network] Found %d (%s [R+v])", foundAroundNetworkNum, nn::pia::local::LocalNetwork::GetInstance()->IsAroundNetworkSearchActivated() ? "ON":"OFF");
                    line += FONT_SIZE;

#if PIA_TEST_CODE
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_FOUND_AROUND_NETWORK_NUM, foundAroundNetworkNum);
#endif
                    for (int i = 0; i < GATHERING_LIST_MAX; ++i)
                    {
                        if (i < foundAroundNetworkNum)
                        {
                            SampleBeaconApplicationData beaconApplicationData;
                            size_t beaconDataSize = 0;
                            result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationData(&beaconApplicationData, &beaconDataSize, sizeof(beaconApplicationData), &s_AroundNetworkInfoList[i].m_NetworkDescription);
                            if (result == nn::pia::ResultNoData())
                            {
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  ----");
                            }
                            else
                            {
                                PIA_SAMPLE_ASSERT(result.IsSuccess());
                                char userName[SAMPLE_USER_NAME_LENGTH];
                                CheckAndGetUserName(userName, s_AroundNetworkInfoList[i].m_NodeInformation[0].m_UserName, SAMPLE_USER_NAME_LENGTH);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  CH%02d:0x%08x[%02d/%02d]:%s",
                                    s_AroundNetworkInfoList[i].m_NetworkDescription.GetChannel(),
                                    beaconApplicationData.value,
                                    s_AroundNetworkInfoList[i].m_NetworkDescription.GetCurrentParticipants(),
                                    s_AroundNetworkInfoList[i].m_NetworkDescription.GetMaxParticipants(), 
                                    userName
                                    );
                            }
#if PIA_TEST_CODE
                            if (i < 2)
                            {
                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_FOUND_AROUND_NETWORK_1_APPLICATION_DATA + 0x10 * i, beaconApplicationData.value);
                                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_FOUND_AROUND_NETWORK_1_PARTICIPANTS + 0x10 * i, s_AroundNetworkInfoList[i].m_NetworkDescription.GetCurrentParticipants());
                            }
#endif
                        }
                        else
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  ----");
                        }
                        line +=FONT_SIZE;
                    }
                }
#endif

                if (nwType == TYPE_LOCAL)
                {
                    line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 10);
                }
                else
                {
                    line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 12);
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Network Information ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology));
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
                line += FONT_SIZE;

                //Screen display of communication-related data (For debugging.).
                nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
                nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&hostInfo);
                if (nwType == TYPE_INTERNET)
                {
                    line -= FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(FONT_SIZE * 23, line, "GID:%d", GetJoinedGatheringID());
                    line += FONT_SIZE;

                    u8 mapping = nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyMapping();
                    u8 filtering = nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyFiltering();
                    u32 portIncrement = nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyPortIncrement();
                    u16 interfacePort = nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyInterfacePort();
                    u16 publicPort = nn::pia::inet::NexFacade::GetInstance()->GetNatPropertyPublicPort();

                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(
                        0,
                        line,
                        "  Nat:%d-%d-%d  Port:%d-%d",
                        mapping,
                        filtering,
                        portIncrement,
                        interfacePort,
                        publicPort);

                    line += FONT_SIZE;

                    nn::pia::common::String mainStr;
                    nn::pia::common::String subStr;
                    localInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                    localInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Local Main:%s", mainStr.GetCharArray());
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "         Sub:%s", subStr.GetCharArray());
                    line += FONT_SIZE;
                    hostInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                    hostInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Host  Main:%s", mainStr.GetCharArray());
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "         Sub:%s", subStr.GetCharArray());
                    line += FONT_SIZE;
                }
#if USE_PIA_LOCAL
                else
                {
                    static char hostUserName[SAMPLE_USER_NAME_LENGTH];
                    GetHostUserName(hostUserName);
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Host UserName: %s", hostUserName);
                    line += FONT_SIZE;

                    u32 beaconValue = 0;
                    if (nn::pia::local::LocalNetwork::GetInstance()->IsConnected() && !nn::pia::local::LocalNetwork::GetInstance()->IsDuringHostMigration())
                    {
                        SampleBeaconApplicationData beaconApplicationData;
                        size_t beaconDataSize = 0;
                        result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationDataFromBeacon(&beaconApplicationData, &beaconDataSize, sizeof(beaconApplicationData));
                        if (result.IsSuccess())
                        {
                            beaconValue = beaconApplicationData.value;
                        }
                        else
                        {
                            PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidConnection() || result == nn::pia::ResultNotInCommunication());
                        }
                    }
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  CH: %02d   Beacon: 0x%08x [R+^]", channel, beaconValue);
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Participation %s [R(+X)+>]", GetAllowParticipatingStateString());
                    line += FONT_SIZE;
                }
#endif

                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: LeaveMesh/DestroyMesh");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R+B: BackToMenu");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Toggle RoundRobin(Current:%s)", (s_EnableRoundRobin ? "ON" : "OFF"));
                line += FONT_SIZE;
            }

            for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
            {
                const u8 DRAW_BASE = 30;

                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(10, DRAW_BASE + (10 * i), 8, 8, COLOR_OF_RECTANGLE[i]);
                if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                {
                    nn::pia::StationIndex index;
                    result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&index, s_StationStatus[i].stationId);
                    if (result.IsSuccess())
                    {
                        s_StationStatus[i].UpdateInfo(IsEnableRelay(s_TargetNetworkTopology));
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "Id=%02X, %05X, RTT:%4d", index, s_StationStatus[i].latestIncrementNumber, s_StationStatus[i].rtt);
                    }
                }
                else
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "No Connection");
                }
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        }
        //Draw to the upper screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
            {
                if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                {
                    nn::pia::StationIndex index;
                    result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&index, s_StationStatus[i].stationId);
                    if (result.IsSuccess())
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(s_StationStatus[i].pos.x, s_StationStatus[i].pos.y, 8, 8, COLOR_OF_RECTANGLE[index]);
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor((s_StationStatus[i].useRoundRobin) ? nn::pia::sampleframework::COLOR_YELLOW : nn::pia::sampleframework::COLOR_WHITE);
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(s_StationStatus[i].pos.x + 10, s_StationStatus[i].pos.y, "<-ID:%02X(%s)", index, s_StationStatus[i].name);
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                    }
                }
            }

            int line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }

        // Execute Dispatch (first time: for receive processing).
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        //Receive.
        // Executed after Dispatch (first time) to facilitate receiving data with the Scheduler::Dispatch function.
        {
            // Because the size of the data the other party sends generally cannot be anticipated, allocate a receive buffer the size of MAX_DATA_SIZE of UnreliableProtocol.
            PIA_ATTRIBUTE_ALIGN(4) static bit8 receiveBuffer[nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE];

            //There may be stations that use both UnreliableProtocol and RoundRobinUnreliableProtocol to send data, so both UnreliableProtocol and RoundRobinUnreliableProtocol are used, in that order, to receive data.
            // 

            //Because the number of received data is unknown, the receive process repeats until the buffer is empty.

            //UnreliableProtocol
            do
            {
                nn::pia::StationId sourceStationId;
                size_t receivedDataSize = 0;

                result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Receive(&sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));

                if (result.IsSuccess())
                {
                    if (sourceStationId != nn::pia::STATION_ID_INVALID)
                    {
                        nn::pia::StationIndex sourceStationIndex;
                        result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&sourceStationIndex, sourceStationId);
                        if (result.IsSuccess() && s_StationStatus[sourceStationIndex].stationId != nn::pia::STATION_ID_INVALID)
                        {
                            //You may need to check whether the received data is correct.
                            s_StationStatus[sourceStationIndex].Deserialize(&receiveBuffer);
                        }
                    }
                }
                else if (result == nn::pia::ResultNotInCommunication())
                {
                    //Failed to receive data due to a disconnection.
                    //This failure can occur if an asynchronous session delete/leave process completes due to an immediately preceding Dispatch process.
                    //If the status is normal, the session status check shows DISCONNECTED when the next game frame starts.
                    PIA_SAMPLE_LOG("UnreliableProtocol::Receive failed. (IsNotInCommunication)\n");
                    break;
                }
                else if (result != nn::pia::ResultNoData())
                {
                    //Other reception failed.
                    //In this demo, this is not an issue.
                }
            } while (result != nn::pia::ResultNoData());
            //RoundRobinUnreliableProtocol
            do
            {
                nn::pia::StationId sourceStationId;

                size_t receivedDataSize = 0;

                result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle)->Receive(&sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));

                if (result.IsSuccess())
                {
                    if (sourceStationId != nn::pia::STATION_ID_INVALID)
                    {
                        nn::pia::StationIndex sourceStationIndex;
                        result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&sourceStationIndex, sourceStationId);
                        if (result.IsSuccess() && s_StationStatus[sourceStationIndex].stationId != nn::pia::STATION_ID_INVALID)
                        {
                            //You may need to check whether the received data is correct.
                            s_StationStatus[sourceStationIndex].Deserialize(&receiveBuffer);
                        }
                    }
                }
                else if (result == nn::pia::ResultNotInCommunication())
                {
                    //Failed to receive data due to a disconnection.
                    //This failure can occur if an asynchronous session delete/leave process completes due to an immediately preceding Dispatch process.
                    //If the status is normal, the session status check shows DISCONNECTED when the next game frame starts.
                    PIA_SAMPLE_LOG("RoundRobinUnreliableProtocol::Receive failed. (IsNotInCommunication)\n");
                    break;
                }
                else if (result != nn::pia::ResultNoData())
                {
                    //Other reception failed.
                    //In this demo, this is not an issue.
                }
            } while (result != nn::pia::ResultNoData());
        }

        int move_x = 0;  //Amount of movement of the rectangle on the local station (x).
        int move_y = 0;  //Amount of movement of the rectangle on the local station (y).

        //Handle user input.
        {
            // B: The process of leaving the session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                if (nn::pia::session::Mesh::GetInstance()->IsHost())
                {
                    if (s_EnableHostMigration)
                    {
                        // If the host and host migration is enabled, use <tt>LeaveMeshWithHostMigrationAsync</tt>.
                        if (nn::pia::session::Mesh::GetInstance()->LeaveMeshWithHostMigrationAsync().IsSuccess())
                        {
                            PIA_SAMPLE_LOG("Start LeaveMeshWithHostMigration.\n");
                            isExitManually = true;
                            isExit = true;
                            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R))
                            {
                                // Quit without going back to matchmaking.
                                isBackToMatchmake = false;
                            }
                        }
                    }
                    else
                    {
                        // If the host and host migration is disabled, then <tt> DestroyMeshAsync</tt>.
                        // Only fails if not the host, so only check only success for the returned value.
                        if (nn::pia::session::Mesh::GetInstance()->DestroyMeshAsync().IsSuccess())
                        {
                            PIA_SAMPLE_LOG("Start DestroyMesh.\n");
                            isExitManually = true;
                            isExit = true;
                            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R))
                            {
                                // Quit without going back to matchmaking.
                                isBackToMatchmake = false;
                            }
                        }
                    }
                }
                else
                {
                    // If a client, then <tt>LeaveMeshAsync</tt>.
                    // Leave fails only when it is not Client, so the result value checks only successes.
                    // When already disconnected, it breaks at the next loop.
                    if (nn::pia::session::Mesh::GetInstance()->LeaveMeshAsync().IsSuccess())
                    {
                        PIA_SAMPLE_LOG("Start LeaveMesh.\n");
                        isExitManually = true;
                        isExit = true;
                        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R))
                        {
                            // Quit without going back to matchmaking.
                            isBackToMatchmake = false;
                        }
                    }
                }
            }

            // A: Enable or disable <tt>RoundRobin</tt>.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                s_EnableRoundRobin = !s_EnableRoundRobin;
            }

#if PIA_TEST_CODE
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_L))
            {
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
                {
                    s_KickoutTargetStation = (s_KickoutTargetStation == SAMPLE_MAX_ENTRY - 1) ? 0 : s_KickoutTargetStation + 1;
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
                {
                    s_KickoutTargetStation = (s_KickoutTargetStation == 0) ? SAMPLE_MAX_ENTRY - 1 : s_KickoutTargetStation - 1;
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_START))
                {
                    PIA_SAMPLE_LOG("KickoutStation: Target=%d, %lld.\n", s_KickoutTargetStation, s_StationStatus[s_KickoutTargetStation].stationId.GetValue());
                    if (s_StationStatus[s_KickoutTargetStation].stationId != nn::pia::STATION_ID_INVALID 
                    &&  s_StationStatus[s_KickoutTargetStation].stationId != myStationId)
                    {
                        nn::pia::session::Mesh::GetInstance()->KickoutStation(s_StationStatus[s_KickoutTargetStation].stationId);
                    }
                    s_KickoutTargetStation = 0;
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
                {
                    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                    pMgr->SetSendThreadPacketLossRatio(100);
                    pMgr->SetReceiveThreadPacketLossRatio(0);
                    PIA_SAMPLE_LOG("Send Pakcet loss ratio = 100.\n");
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                    pMgr->SetSendThreadPacketLossRatio(0);
                    pMgr->SetReceiveThreadPacketLossRatio(100);
                    PIA_SAMPLE_LOG("Recv Pakcet loss ratio = 100.\n");
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
                {
                    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                    pMgr->SetSendThreadPacketLossRatio(100);
                    pMgr->SetReceiveThreadPacketLossRatio(100);
                    PIA_SAMPLE_LOG("Pakcet loss ratio = 100.\n");
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
                {
                    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                    pMgr->SetSendThreadPacketLossRatio(0);
                    pMgr->SetReceiveThreadPacketLossRatio(0);
                    PIA_SAMPLE_LOG("Pakcet loss ratio = 0.\n");
                }
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP) 
                &&  nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
                {
                    s_IsKickNextJoinStation = true;
                }
            }
#endif

#if USE_PIA_LOCAL
            if (nwType == TYPE_LOCAL)
            {
                // R + Up: Updates the beacon data.
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R) &&
                    nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
                {
                    UpdateBeaconData();
                }

                // R+→: Toggle between join permitted and denied (R+X+→: Only re-connection permitted)
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R) &&
                    nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_X))
                    {
                        result = nn::pia::local::LocalNetwork::GetInstance()->DisallowParticipating(false);
                    }
                    else if (nn::pia::local::LocalNetwork::GetInstance()->GetAllowParticipatingState() == nn::pia::local::LocalNetwork::ALLOW_PARTICIPATING_STATE_ALLOW)
                    {
                        result = nn::pia::local::LocalNetwork::GetInstance()->DisallowParticipating();
                    }
                    else
                    {
                        result = nn::pia::local::LocalNetwork::GetInstance()->AllowParticipating();
                    }
                    if (result.IsFailure())
                    {
                        // Handle during a check of connection status.
                        PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidConnection());
                    }
                }

                if (ENABLE_AROUND_NETWORK_SEARCH)
                {
                    if (nn::pia::local::LocalNetwork::GetInstance()->IsHost())
                    {
                        // R+↓: Toggle nearby network search on and off.
                        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_R) &&
                            nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
                        {
                            if (!nn::pia::local::LocalNetwork::GetInstance()->IsAroundNetworkSearchActivated())
                            {
                                result = nn::pia::local::LocalNetwork::GetInstance()->StartAroundNetworkSearch();
                                PIA_ASSERT(result.IsSuccess());
                            }
                            else
                            {
                                result = nn::pia::local::LocalNetwork::GetInstance()->StopAroundNetworkSearch();
                                PIA_ASSERT(result.IsSuccess());
                            }
                        }
                    }
                }
            }
#endif
        }
        //Move the rectangle.
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
            {
                move_x = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                move_x = 1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
            {
                move_y = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
            {
                move_y = 1;
            }
        }

        //Locked send process.
        // Executed before Dispatch (second time) for the convenience of having send processing performed by the Scheduler::Dispatch function.
        {
            //Update the data of the local station.
            if (move_x != 0)
            {
                s_StationStatus[myStationIndex].pos.x = (s_StationStatus[myStationIndex].pos.x + UPPER_SCREEN_WIDTH + move_x) % UPPER_SCREEN_WIDTH;
            }
            else
            {
                //Move the rectangle even when there is no input to indicate that stations are communicating.
                s_StationStatus[myStationIndex].pos.x = (s_StationStatus[myStationIndex].pos.x + UPPER_SCREEN_WIDTH + 1) % UPPER_SCREEN_WIDTH;
            }
            if (move_y != 0)
            {
                s_StationStatus[myStationIndex].pos.y = (s_StationStatus[myStationIndex].pos.y + UPPER_SCREEN_HEIGHT + move_y) % UPPER_SCREEN_HEIGHT;
            }
            s_StationStatus[myStationIndex].latestIncrementNumber = frameNo;

#if 0
            // Use SendToAll to send as a batch process.
            SampleStationStatus sendData;
            s_StationStatus[myStationIndex].useRoundRobin = (s_EnableRoundRobin) ? 1 : 0;
            s_StationStatus[myStationIndex].Serialize(&sendData);
            if (s_EnableRoundRobin)
            {
                if (nn::pia::transport::Transport::GetInstance()->
                     GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle)->
                     SendToAll(reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                {
                    //Send failure.
                    //In this demo, this is not an issue.
                    //PIA_SAMPLE_LOG("Send Fail\n");
                }
            }
            else
            {
                if (nn::pia::transport::Transport::GetInstance()->
                     GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->
                     SendToAll(reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                {
                    //Send failure.
                    //In this demo, this is not an issue.
                    //PIA_SAMPLE_LOG("Send Fail\n");
                }
            }
#else
            // Send only to valid parties using Send.
            for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
            {
                // Send to valid <tt>StationIds</tt> other than the local station.
                if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID && s_StationStatus[i].stationId != myStationId)
                {
                    SampleStationStatus sendData;
                    s_StationStatus[myStationIndex].useRoundRobin = (s_EnableRoundRobin) ? 1 : 0;
                    s_StationStatus[myStationIndex].Serialize(&sendData);
                    if (s_EnableRoundRobin)
                    {
                        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle)->Send(s_StationStatus[i].stationId, reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                        {
                            //Send failure.
                            //In this demo, this is not an issue.
                            //PIA_SAMPLE_LOG("Send failed.\n");
                        }
                    }
                    else
                    {
                        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Send(s_StationStatus[i].stationId, reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                        {
                            //Send failure.
                            //In this demo, this is not an issue.
                            //PIA_SAMPLE_LOG("Send failed.\n");
                        }
                    }
                }
            }
#endif
        }

        // Execute Dispatch (second time: for send processing).
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        frameNo++;
    }

#if PIA_TEST_CODE
    {
        nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

        pMgr->SetSendThreadPacketLossRatio(0);
        pMgr->SetReceiveThreadPacketLossRatio(0);
        PIA_SAMPLE_LOG("Pakcet loss ratio = 0.\n");
    }
#endif

    if (isExitManually)
    {
        frameNo = 0;
        while (true)
        {
            bool errorOccurred = false;

            //Disconnection detected.
            if (nwType == TYPE_INTERNET)
            {
                // Checks the AC network connection status, game server connection status, and NEX internal errors.
                if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
                {
                    errorOccurred = true;
                    result = nn::pia::ResultNetworkConnectionIsLost();
                }
            }
#if USE_PIA_LOCAL
            else
            {
                // Checks the local network connection status.
                if (!Local_IsConnected())
                {
                    errorOccurred = true;
                    result = nn::pia::ResultNetworkConnectionIsLost();
                }
            }
#endif
            if (!errorOccurred)
            {
                //Status check for the network connection.
                result = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
            }
            if (result.IsFailure())
            {
                errorOccurred = true;
                //If the network connection is abnormal:
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    // Returns when the network connection is interrupted because the wireless switch is turned off or communication fails.
                    PIA_SAMPLE_LOG("Connection Lost when scene Destroy/Leave Session.\n");
                }
                else
                {
                    //Unexpected state.
                    PIA_SAMPLE_LOG("Transport::CheckConnectionError is unknown status when scene Destroy/Leave Session.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    PIA_SAMPLE_ASSERT(false);
                }
            }
            if (errorOccurred)
            {
                // Call Cleanup because an error occurred during Session processing.
                nn::pia::session::Mesh::GetInstance()->Cleanup();

                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                if (nwType == TYPE_INTERNET)
                {
                    SetNetworkErrorResult(result);
                    return SCENE_NETWORK_ERROR;
                }
                else
                {
                    return SCENE_ERROR;
                }
            }

            // Check if asynchronous process is finished before disconnection.
            if (nn::pia::session::Mesh::GetInstance()->IsCompletedDestroyMesh() || nn::pia::session::Mesh::GetInstance()->IsCompletedLeaveMesh() || nn::pia::session::Mesh::GetInstance()->IsCompletedLeaveMeshWithHostMigration())
            {
                PIA_SAMPLE_LOG("Destroy/Leave/LeaveWithHostMigration completed.\n");
                break;
            }

            // Periodically update signal strength information.
            if (frameNo % FRAME_PER_SECOND == 0)
            {
                linkLevel = GetLinkLevel(nwType);
            }
#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE,
                nn::pia::session::Mesh::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID,
                nn::pia::session::Mesh::GetInstance()->IsHost(),
                nn::pia::session::Mesh::GetInstance()->GetStationNum(),
                linkLevel);
#endif

            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

            if (nwType == TYPE_INTERNET)
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
            nn::pia::common::Scheduler::GetInstance()->Dispatch();

            //Lower screen.
            int line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Destroy/Leave Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(LOWER_SCREEN_WIDTH - (FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Please Wait...(loop=%d)", frameNo);
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

            //Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            frameNo++;
        }
    }

    // The session processing is finished, so Cleanup runs.
    nn::pia::session::Mesh::GetInstance()->Cleanup();

    if (isExitManually)
    {
        // Disconnected by user operation.

        if (isBackToMatchmake)
        {
            CleanupPiaSession();

            switch (nwType)
            {
            case TYPE_LOCAL:
            {
#if USE_PIA_LOCAL
                if (nn::pia::local::LocalFacade::GetInstance() != NULL)
                {
                    nn::pia::local::LocalFacade::GetInstance()->Cleanup();
                }
                // Shut down the UDS network.
                bool isSuccess = Local_DisconnectNetwork(s_EnableHostMigration);
                if (!isSuccess)
                {
                    SetCommunicationErrorType(SESSION_LOST_ERROR);
                    return SCENE_ERROR;
                }
                Local_CleanupPiaLocal();
#endif
            }
            break;
            case TYPE_INTERNET:
            {
                // Finish cleanup of NexFacade before finalizing transport.
                if (nn::pia::inet::NexFacade::GetInstance() != NULL)
                {
                    nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                    nn::pia::inet::NexFacade::GetInstance()->Cleanup();
                }
                // End the matchmaking session.
                ResetMatchMakingClient();
            }
            break;
            }
            return SCENE_MATCH_MAKE;
        }

        return SCENE_RESET;  //Jump directly to the reset process because the station disconnected due to user input.
    }
    //else
    // Handle disconnection as a communication error if disconnected for a reason other than user input.
    SetCommunicationErrorType(SESSION_LOST_ERROR);

    if (nwType == TYPE_INTERNET)
    {
        SetNetworkErrorResult(SampleMeshStateDisconnected());
        return SCENE_NETWORK_ERROR;
    }
    else
    {
        return SCENE_ERROR;
    }
}

void InitializeTransportAndSession(NetworkType nwType)
{
// * The Factory instance is not required except for initialization, so it is allocated with an automatic variable.
#if USE_PIA_LOCAL
    nn::pia::local::UdsNetworkFactory udsFactory;
#endif
    nn::pia::inet::NexNetworkFactory nexFactory;
    nn::Result result;

    // Initialize nn::pia::transport.
    if (s_InitializeTransport == false)
    {
#if PIA_TEST_CODE
        nn::pia::transport::Transport::DebugSetting transportDebugSetting;
        transportDebugSetting.bPacketLossEmulation = true;
        result = nn::pia::transport::Transport::SetDebugSetting(transportDebugSetting);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
#endif

        nn::pia::transport::Transport::Setting transportSetting;
        if (nwType == TYPE_INTERNET)
        {
            transportSetting.pFactory = &nexFactory;
        }
#if USE_PIA_LOCAL
        else
        {
            transportSetting.pFactory = &udsFactory;
        }
#endif
        transportSetting.maxStationNum = SAMPLE_MAX_ENTRY;
        transportSetting.maxSendThreadBufferNum = 16;
        transportSetting.maxReceiveThreadBufferNum = 32;

        result = nn::pia::transport::Initialize();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::transport::BeginSetup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::transport::Transport::CreateInstance(transportSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        //The port number is hard-coded to 0.
        s_UnreliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::UnreliableProtocol>(0);
        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Initialize();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        s_RoundRobinUnreliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(0);
        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle)->Initialize();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::transport::EndSetup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        //Priority settings for the send/receive thread. (For information about priorities, see SampleConfig.h.)
        //The priority of the send/receive thread does not necessarily need to be changed, but in this sample, the priority has been changed to show the change method.
        //Priority has been changed.
        nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
        PIA_SAMPLE_ASSERT(pThreadStreamManager);  //If the transport initialization has finished, the instance has probably been created.
        pThreadStreamManager->SetSendThreadPriority(SEND_THREAD_PRIORITY);
        pThreadStreamManager->SetReceiveThreadPriority(RECEIVE_THREAD_PRIORITY);

        s_InitializeTransport = true;
    }

    if (s_InitializeSession == false)
    {
        nn::pia::session::Mesh::Setting sessionSetting;

        if (nwType == TYPE_INTERNET)
        {
            sessionSetting.pNetworkFactory = &nexFactory;

            // The feature to measure bandwidth
            if(s_IsEnableBitRateCheck)
            {
                sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Enable;
            }
            else
            {
                sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Disable;
            }
        }

#if USE_PIA_LOCAL
        else
        {
            sessionSetting.pNetworkFactory = &udsFactory;
            // The feature to measure bandwidth
            sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Disable;
        }
#endif
        // Dynamically changes the network topology.
        sessionSetting.networkTopology = GetNetworkTopologyType(s_TargetNetworkTopology);
        


        result = nn::pia::session::Initialize();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::session::BeginSetup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::session::Mesh::CreateInstance(sessionSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::session::EndSetup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        s_InitializeSession = true;
    }
}

void FinalizePiaTransport()
{
    if (nn::pia::transport::Transport::GetInstance() != NULL)
    {
        //Finalize nn::pia::transport::UnreliableProtocol.
        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle) != NULL)
        {
            nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Finalize();
            nn::pia::transport::Transport::GetInstance()->DestroyProtocol(s_UnreliableProtocolHandle);
            s_UnreliableProtocolHandle = 0;
        }

        //Finalize <tt>nn::pia::transport::RoundRobinUnreliableProtocol</tt>.
        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle) != NULL)
        {
            nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_RoundRobinUnreliableProtocolHandle)->Finalize();
            nn::pia::transport::Transport::GetInstance()->DestroyProtocol(s_RoundRobinUnreliableProtocolHandle);
            s_RoundRobinUnreliableProtocolHandle = 0;
        }
    }

    //Finalize nn::pia::Transport.
    nn::pia::transport::Transport::DestroyInstance();  //The function to destroy the <tt>transport</tt> instance can be run without any problems, even if the instance has not been created.
    nn::pia::transport::Finalize();

    s_InitializeTransport = false;
}

void CleanupPiaSession()
{
    if (nn::pia::session::Mesh::GetInstance() != NULL)
    {
        nn::pia::session::Mesh::GetInstance()->UnregisterJoiningMeshApprovalCallback();
        nn::pia::session::Mesh::GetInstance()->UnregisterMeshEventCallback();
        nn::pia::session::Mesh::GetInstance()->Cleanup();
    }
}

void FinalizePiaSession()
{
    //Finalize nn::pia::session.
    if (nn::pia::session::Mesh::GetInstance() != NULL)
    {
        nn::pia::session::Mesh::DestroyInstance();
    }
    nn::pia::session::Finalize();

    s_InitializeSession = false;
}


// String for token type screen display.
const char* TOKEN_TYPE_STRING[TOKEN_TYPE_NUM] = {
    "TYPE_ALPHA",
    "TYPE_BETA"
};

void SetLocalIdentificationToken(SampleTokenType tType)
{
    int randomSeed;

#if NN_PIA_CTR
    randomSeed = static_cast<int>(nn::os::Tick::GetSystemCurrent());
#elif NN_PIA_CAFE
    randomSeed = static_cast<int>(OSGetTick());
#endif
    ::std::srand(randomSeed);

    ::std::memset(s_LocalTokenData, 0x00, sizeof(s_LocalTokenData));

    // Sets its own identification token data.
    // This sample handles the data as a string, but if the data has a size within <tt>IDENTIFICATION_TOKEN_MAX_DATA_SIZE [byte]</tt>, the application can use it freely.
    switch (tType)
    {
    case TOKEN_TYPE_ALPHA:
    {
        ::std::snprintf(reinterpret_cast<char*>(s_LocalTokenData), sizeof(s_LocalTokenData), "advanced [Type A] Id:0x%08X", ::std::rand());
    }
    break;
    case TOKEN_TYPE_BETA:
    {
        ::std::snprintf(reinterpret_cast<char*>(s_LocalTokenData), sizeof(s_LocalTokenData), "ADVANCED-<TYPE_B> Id:0x%08X", ::std::rand());
    }
    break;
    default:
    {
        PIA_SAMPLE_LOG("Unknown TokenType.\n");
    }
    break;
    }

    PIA_SAMPLE_LOG("SetLocalIdentificationToken:\"%s\"\n", s_LocalTokenData);
}
