﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneError.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/util.h>

#include "SampleLib/SampleLib.h"

#include "Shared_Inet_Setup.h"

#include "SampleConfig.h"

namespace
{
SampleErrorType s_ErrorType = NO_ERROR;  //Variable that holds the errors that occurred most recently.

const char* ERROR_INFO_STRING[] =
    {
        "Unknown Reason",
        "Initialize Error (wireless OFF)",
        "Initialize Error (other reason)",
        "Matchmake Failed",
        "Session Lost",
        "Connection Lost"
    };

nn::Result s_NetworkErrorResult = nn::ResultSuccess();
}

/*!
    @brief  This function sets the error reason before jumping to the error state.
 */
void SetCommunicationErrorType(SampleErrorType error)
{
    s_ErrorType = error;
}

/*!
    @brief  This state displays a description for a recoverable error (such as a communication error).
    
    It only displays error information collected when the error occurred.
 */
SampleSceneNo SceneError()
{
    nn::Result result;
    bool isExit = false;

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_ERROR, false);
#endif

    int frameNo = 0;
    // Main loop.
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        //Handles user input. Pad information is collected by the BeginFrame function.
        {
            // A:    Restart.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                isExit = true;
            }
        }

        // Renders to the screens.
        int line;
        {
            // Lower screen (standard).
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Communication Error Occurred.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
            line += FONT_SIZE;

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 2);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        {
            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        frameNo++;
    }

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_WAIT, false);
#endif

    //Clears the error information, just in case.
    s_ErrorType = NO_ERROR;
    SetNetworkErrorResult(nn::ResultSuccess());
    return SCENE_RESET;
}


// Pass the result of errors that occur when connecting to the Internet.
// Pass the result to the function for conversion to a network error code, and then display it.
// 
void SetNetworkErrorResult(nn::Result result)
{
    s_NetworkErrorResult = result;
}


nn::Result GetNetworkErrorResult(void)
{
    return s_NetworkErrorResult;
}


/*!
    @brief  This scene displays the network error code for errors occurring in the Internet connection environment.
    
    In a normal program, pass the network error codes to the Error/EULA Viewer.
 */
SampleSceneNo SceneNetworkError()
{
    // Convert the error <tt>Result</tt> to a network error code.
    // Normally, you would pass this error code to the error/EULA applet.
    // However, the error/EULA applet cannot be used unless development hardware is set to <tt>prod</tt> mode.
    // For simplicity, this sample demo does not use the error/EULA applet.
    // Not used.
    nn::Result result = GetNetworkErrorResult();
    bool isPiaError = nn::pia::util::IsPiaResult(result);
    if (isPiaError && nn::pia::util::GetViewerType(result) == nn::pia::util::VIEWER_SHOULD_USE)
    {
        // Usually you would call the error/EULA applet and pass the network error code here or something similar, but in this sample we only print the Result as a string.
        // 
        PIA_SAMPLE_LOG("Network error Result: %s\n", nn::pia::util::GetResultString(result));
    }
    u32 networkErrorCode = (isPiaError) ? nn::pia::util::NetworkErrorCodeConverter::Convert(result) : Inet_GetNetworkErrorCode();

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_NETWORK_ERROR, false);
#endif

    int frameNo = 0;
    // Rendering
    while (1)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Exit the loop if the A Button is pressed.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            break;
        }

        int line;
        {
            // Render the lower screen
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Network Error Occurred.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Result: 0x%08X", GetNetworkErrorResult().GetPrintableBits());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Level:       %4d", GetNetworkErrorResult().GetLevel());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Summary:     %4d", GetNetworkErrorResult().GetSummary());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Module:      %4d", GetNetworkErrorResult().GetModule());
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Description: %4d", GetNetworkErrorResult().GetDescription());
            line += FONT_SIZE;
            line += FONT_SIZE;

            if (networkErrorCode != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Network error code: %d", networkErrorCode);
                line += FONT_SIZE;
            }
            if (Inet_GetLoginNetworkErrorCode() != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Login error code: %d", Inet_GetLoginNetworkErrorCode());
                line += FONT_SIZE;
            }

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 2);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        {
            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
            line += FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        frameNo++;
    }

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_WAIT, false);
#endif

    //Clears the error information, just in case.
    s_ErrorType = NO_ERROR;
    SetNetworkErrorResult(nn::ResultSuccess());

    return SCENE_RESET;
}
