﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneInetMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>

#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Nex_Matchmake.h"

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include <nn/ac.h>

#if NN_PIA_CTR
#include <nn/friends.h>
#elif NN_PIA_CAFE
#include <nn/fp.h>
#else
#error  "Invalid platform."
#endif

#if ENABLE_SERVERLESS
#error  "ENABLE_SERVERLESS is available only in using Session."
#endif

#include "SampleConfig.h"

/*!

    @brief  CPP file with processing specific to Internet communication.

*/

extern u8 s_GameModeOption;  // The option value that changes MATCHMAKE_GAME_MODE.
extern bool s_EnableHostMigration;  //!<  Enables or disables the host migration feature. Fixed to enabled when the PiaSessionhost migration feature is used.

extern bool s_UsingInvalidSignature;

// Dynamically change the network topology setting of the Session.
extern size_t s_TargetNetworkTopology;
const char* GetNetworkTopologyTypeString(size_t index);

extern bool s_IsEnableBitRateCheck;

u32 GetIncrementalGameModeValueByBitRateCheckMode()
{
    if(s_IsEnableBitRateCheck)
    {
        return 300;
    }
    else
    {
        return 0;
    }
}

namespace
{
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

// Related to NEX matchmaking.
nn::nex::GatheringID s_JoinedGatheringId = INVALID_GATHERINGID;  // Gathering ID to use for matchmaking
nn::nex::PrincipalID s_OwnerPID = INVALID_PRINCIPALID;           // Specify the principal ID of the host.

nn::pia::common::Event s_SearchStarter;    // Event for waiting for the search start of searchThread.
bool s_SearchProcessingFlag = false;       // The flag for detecting that searchThread is running a search process.
bool s_SearchThreadTerminateFlag = false;  // The flag for searchThread termination.
Thread s_SearchThread;
#if NN_PIA_CAFE
const size_t STACK_SIZE = 8192;
u32 s_SearchThreadStack[STACK_SIZE / sizeof(u32)];
#endif

// Session search thread.
void SearchThreadFunc(nn::nex::qList<nn::nex::GatheringHolder>* plistGatheringHolder)
{
    PIA_SAMPLE_LOG("SearchThreadFunc:start.\n");

    // When a thread starts, the search process runs once, but it is handled as specification.
    // Set a flag to indicate that the main thread shows that a search is running at startup.
    s_SearchProcessingFlag = true;

    // Search process.
    while (!s_SearchThreadTerminateFlag)
    {
        // Blocked until an internal search completes.
        nn::Result result = Nex_SearchMatchmakeSession(
            MATCHMAKE_GAME_MODE + s_GameModeOption + GetIncrementalGameModeValueByBitRateCheckMode(),
            SAMPLE_MAX_ENTRY,
            GATHERING_LIST_MAX,
            plistGatheringHolder);
        if (result.IsFailure())
        {
            // Search failed.
            PIA_SAMPLE_LOG("SearchThreadFunc:Nex_SearchMatchmakeSession failed.\n");
            s_SearchProcessingFlag = false;
            s_SearchStarter.Wait();
            continue;
        }

        // Clears the flag during processing and waits for notice of the start of the next search.
        s_SearchProcessingFlag = false;
        s_SearchStarter.Wait();
    }

    PIA_SAMPLE_LOG("SearchThreadFunc:end.\n");
}

#if NN_PIA_CTR
void SearchThreadFuncWrapper(void* pParam)
{
    SearchThreadFunc(static_cast<nn::nex::qList<nn::nex::GatheringHolder>*>(pParam));
}
#elif NN_PIA_WIN
#error  "Not implemented."
#elif NN_PIA_CAFE
int SearchThreadFuncWrapper(int, void* pParam)
{
    SearchThreadFunc(static_cast<nn::nex::qList<nn::nex::GatheringHolder>*>(pParam));
    return 0;
}
#else
#error  "Invalid platform."
#endif


// Displays the progress of <tt>InitializePiaInet</tt> on the screen.
void InitializingProcessRender(u32 processingStep)
{
    // Gets the signal strength information.
    u32 linkLevel = GetLinkLevel(TYPE_INTERNET);
#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_WAIT,
        false,
        false,
        0,
        linkLevel);
#endif

    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

    // Lower screen.
    nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    int line = 0;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo]", PROGRAM_NAME);
    line += FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Inet Initializing process.");
    line += FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(LOWER_SCREEN_WIDTH - (FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
    line += FONT_SIZE;

    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology));
    line += FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
    line += FONT_SIZE;
    line += FONT_SIZE;

    // Progress display.
    // Steps that have been processed are displayed in white; steps that have not been processed are displayed in red.
    {
        const char* stepStrings[] = {
            "Initialize NEX",
            "Initialize inet module",
            "Initialize Network",
            "Login server"
        };  // Four steps.
        const u32 STEP_MAX = 4;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   START");
        line += FONT_SIZE;
        for (u32 step = 0; step < STEP_MAX; step++)
        {
            if (step == processingStep)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "     |");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
            }
            else
            {
                line += FONT_SIZE;
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", stepStrings[step]);
            line += FONT_SIZE;
        }
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   END");
        line += FONT_SIZE;
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

    nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

    // Upper screen.
    nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

    line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
    line += FONT_SIZE;

    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
}

}  // end of anonymous namespace

// Get function to display GatheringID on the screen in the "communicating" state.
u32 GetJoinedGatheringID()
{
    return s_JoinedGatheringId;
}

#if NN_PIA_CTR
// Function for getting the signal strength of the communication with the access point, and returning a converted value in the range from 0 through 3.
u32 GetInternetLinkLevel()
{
    nn::ac::CTR::LinkLevel acLinkLevel = nn::ac::CTR::GetLinkLevel();

    switch (acLinkLevel)
    {
    case nn::ac::CTR::LINK_LEVEL_3:
    {
        return 3;
    }
    case nn::ac::CTR::LINK_LEVEL_2:
    {
        return 2;
    }
    case nn::ac::CTR::LINK_LEVEL_1:
    {
        return 1;
    }
    case nn::ac::CTR::LINK_LEVEL_0:
    default:
    {
        return 0;
    }
    }
}
#endif

/* ---< Setup of the features used in the PiaInet module. >--- */
// Performs Internet connection processing, logging on to the game server, and so on.
// Because nn::ac::Connect blocks for long periods, we do not recommend running it on the main thread.
// This sample program runs it here to emphasize process readability.
bool InitializePiaInet()
{
    nn::Result result;

    // Because the blocking process continues for a long time, the progress is displayed on the screen.
    u32 step = 0;
    InitializingProcessRender(step);
    step++;

/* Start processing NEX-related initialization. */

    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

/* Finished processing NEX-related initialization. */

    InitializingProcessRender(step);
    step++;

/* Start of the Pia modules initialization process. */

    result = Inet_InitializePiaInet();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    InitializeTransportAndSession(TYPE_INTERNET);

/* End of the Pia modules initialization process. */

    InitializingProcessRender(step);
    step++;

/* Preparing for Internet communication. */

    result = Inet_InitializeNetwork();
    if (result.IsFailure())
    {
#if NN_PIA_CTR
        if (result == nn::ac::ResultWifiOff())
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_WIRELESS_OFF);
        }
        else
#elif NN_PIA_WIN
#error  "Not implemented."
#elif NN_PIA_CAFE
// n1769: Cafe AC does not have ResultWifiOff, so the if statement of ↑ is skipped.
#else
#error  "Invalid platform."
#endif
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        }
        SetNetworkErrorResult(result);

        return false;
    }

/* Finished preparing Internet connection processing. */

    InitializingProcessRender(step);
    step++;

/* Start of the process of logging in to the server. */

    result = Inet_Login();
    if (result.IsFailure())
    {
        // Log in fails.
        SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        SetNetworkErrorResult(result);
        return false;
    }

    result = Nex_InitializeMatchmake();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

/* End of the process of logging in to the server. */

    InitializingProcessRender(step);
    step++;

    return true;
}

void FinalizePiaInet()
{
    // Finish cleaning up the Session instance before cleaning up the Facade instance.
    CleanupPiaSession();

    // Finish Cleanup of Facade before finalizing transport.
    if (nn::pia::inet::NexFacade::GetInstance() != NULL)
    {
        nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
        nn::pia::inet::NexFacade::GetInstance()->Cleanup();
    }

     // Finishes the matchmaking session.
    Nex_LeaveMatchmakeSession(
        s_JoinedGatheringId,
        s_OwnerPID,
        true);
    s_JoinedGatheringId = INVALID_GATHERINGID;

    Nex_FinalizeMatchmake();

/* Process to log out from the server. */
    Inet_Logout();

/* Processes Internet finalization (disconnection). */
    Inet_FinalizeNetwork();

    // Finalizes PiaSession.
    FinalizePiaSession();

    // Finalizes PiaTransport.
    FinalizePiaTransport();

/* Finalizes PiaInet. */
    Inet_FinalizePiaInet();

/* NEX-related finalization. */
    Nex_FinalizeNex();
}

/*!
    @brief  Builds an Internet communication session using a <tt>PiaInet</tt> instance.

    In this state, you can query the game server and search for sessions, and either join a found session or build a new one.
    When this state ends (even if it succeeds), a connection has not yet been established. The connected state begins after building or joining a <tt>PiaSession</tt> session.

    @return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo SceneInetMatchmake(bool* pIsSessionHost)
{
    nn::Result result;

    // Check the settings of <tt>MATCHMAKE_GAME_MODE</tt> and <tt>GameModeOption</tt>.
    PIA_SAMPLE_LOG("SceneInetMatchmake: Matchmake settings -> MATCHMAKE_GAME_MODE:%04d GameModeOption:%02d IncrementalByBitRateCheckMode:%02d\n", MATCHMAKE_GAME_MODE, s_GameModeOption, GetIncrementalGameModeValueByBitRateCheckMode());

    // Save the location for the session search results.
    nn::nex::qList<nn::nex::GatheringHolder> listGatheringHolder;

    // Start the session search thread.
    s_SearchStarter.Initialize(false);
    s_SearchProcessingFlag = false;
    s_SearchThreadTerminateFlag = false;
#if NN_PIA_CTR
    s_SearchThread.Start(SearchThreadFuncWrapper, &listGatheringHolder, 8192);
#elif NN_PIA_WIN
#error  "Not implemented."
#elif NN_PIA_CAFE
    s_SearchThread.Start(SearchThreadFuncWrapper, &listGatheringHolder, s_SearchThreadStack, STACK_SIZE, 0, 15);
#else
#error  "Invalid platform."
#endif

    s_JoinedGatheringId = INVALID_GATHERINGID;

    size_t targetSession = 0;
    size_t foundSession = 0;

    bool isExit = false;
    bool isLogout = false;
    bool isCreateNewNetwork = false;
    bool errorOccurred = false;

    u32 linkLevel = GetLinkLevel(TYPE_INTERNET);
#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_MATCH_MAKE,
        false,
        false,
        0,
        linkLevel);
    s_UsingInvalidSignature = false;
#endif

    int frameNo = 0;
    // Main loop.
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Checks the AC network connection status, game server connection status, and NEX internal errors.
        if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
        {
            errorOccurred = true;
            isExit = true;
        }
        // Handles user input. Pad information is collected by the BeginFrame function.
        {
            // Up/Down Buttons: Select a session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                targetSession = (GATHERING_LIST_MAX + targetSession - 1) % GATHERING_LIST_MAX;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                targetSession = (targetSession + 1) % GATHERING_LIST_MAX;
            }

            // A Button: Joins the selected session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                if (s_SearchProcessingFlag)
                {
                    // Do not perform anything while searching.
                }
                else
                {
                    isExit = true;
                    isCreateNewNetwork = false;
                }
            }
            // B Button: Log out.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                if (s_SearchProcessingFlag)
                {
                    // Do not perform anything while searching.
                }
                else
                {
                    isExit = true;
                    isLogout = true;
                }
            }
            // X Button: Create a new session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                if (s_SearchProcessingFlag)
                {
                    // Do not perform anything while searching.
                }
                else
                {
                    isExit = true;
                    isCreateNewNetwork = true;
                }
            }
            // Y Button: Start a session search.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                // Start a session search with the session search thread.
                s_SearchStarter.Signal();
                s_SearchProcessingFlag = true;
            }
            // R Button: Enable or disable host migration.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                s_EnableHostMigration ^= true;
#if PIA_TEST_CODE
                PIA_SAMPLE_LOG("Host Migration %s\n", s_EnableHostMigration ? "Enabled" : "Disabled");
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_ENABLE_HOST_MIGRATION, s_EnableHostMigration ? 1 : 0);
#endif
            }
#if PIA_TEST_CODE
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                s_UsingInvalidSignature = true;
            }
#endif
        }

        // Call nn::nex::Scheduler::DispatchAll to maintain the connection with the NEX server.
        // Do not perform anything here because the search process causes server communication during a search.
        if (!s_SearchProcessingFlag)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Draw process.
        // Lower screen (standard).
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Search, Join or Create Session.");
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(LOWER_SCREEN_WIDTH - (FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology));
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
        line += FONT_SIZE;
        line += FONT_SIZE;

        if (isExit)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Please Wait...");
            line += FONT_SIZE;
        }
        else
        {
            // Check if a session search is in progress.
            if (s_SearchProcessingFlag)
            {
                // A session search is in progress because of a different thread.
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Session List ] Search in progress...");
                line += FONT_SIZE;
            }
            else
            {
                // List up sessions.
                foundSession = listGatheringHolder.size();

                nn::nex::qList<nn::nex::GatheringHolder>::iterator it = listGatheringHolder.begin();
                nn::nex::MatchmakeSession* pMatchmakeSession;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Session List ] Found:%01d ", foundSession);
                line += FONT_SIZE;

                for (int i = 0; i < GATHERING_LIST_MAX; i++)
                {
                    if (i == targetSession)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                    }

                    if (i < foundSession)
                    {
                        PIA_SAMPLE_ASSERT(it != listGatheringHolder.end());
                        pMatchmakeSession = reinterpret_cast<nn::nex::MatchmakeSession*>(it->Get());
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GID=%08X, HostPID=%08X [%02d/%02d]", pMatchmakeSession->GetID(), pMatchmakeSession->GetHostPID(), pMatchmakeSession->GetParticipantCount(), pMatchmakeSession->GetMaxParticipants());
                        line += FONT_SIZE;
                        ++it;
                    }
                    else
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  -----");
                        line += FONT_SIZE;
                    }
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                }
            }

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 7);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Select Matchmake Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Join this Matchmake Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Logout and BackToMenu");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Create New Matchmake Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Search Matchmake Session");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: Toggle Host Migration: %s", s_EnableHostMigration ? "ENABLE" : "DISABLE");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;

        // Update signal strength information periodically.
        if (frameNo % FRAME_PER_SECOND == 0)
        {
            linkLevel = GetLinkLevel(TYPE_INTERNET);
#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_MATCH_MAKE,
                false,
                false,
                0,
                linkLevel);
#endif
        }
    }

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_WAIT, false);
#endif

    // After exiting the loop, join the session search thread.
    s_SearchThreadTerminateFlag = true;
    s_SearchStarter.Signal();
    s_SearchThread.Join();
    s_SearchThread.Finalize();
    s_SearchStarter.Finalize();

    // Disconnection detected.
    if (errorOccurred)
    {
        // With this demo, all of these errors are lumped together and treated as matchmaking errors.
        SetCommunicationErrorType(MATCH_MAKE_ERROR);
        SetNetworkErrorResult(nn::pia::ResultNetworkConnectionIsLost());
        return SCENE_NETWORK_ERROR;
    }

    // Finalize and log off.
    if (isLogout)
    {
        return SCENE_RESET;
    }

#if !USE_NEX_AUTOMATCH
    *pIsSessionHost = isCreateNewNetwork;

    if (isCreateNewNetwork)
    {
        // Create a new session.
        result = Nex_CreateMatchmakeSession(
            &s_JoinedGatheringId,
            MATCHMAKE_GAME_MODE + s_GameModeOption + GetIncrementalGameModeValueByBitRateCheckMode(),
            SAMPLE_MAX_ENTRY,
            "PiaSession simple");
        if (result.IsFailure())
        {
            // When the program has been properly implemented, the only causes are that wireless is turned off, the system is entering Sleep Mode, or server communication has timed out.
            // With this demo, all of these errors are lumped together and treated as matchmaking errors.
            SetCommunicationErrorType(MATCH_MAKE_ERROR);
            SetNetworkErrorResult(SampleFailedCreateMatchmakeSession());
            return SCENE_NETWORK_ERROR;
        }
    }
    else
    {
        // Join an existing session.
        if (listGatheringHolder.size() > targetSession)
        {
            // Get the GatheringID of the session specified by targetSession.
            nn::nex::qList<nn::nex::GatheringHolder>::iterator it = listGatheringHolder.begin();
            ::std::advance(it, targetSession);
            PIA_SAMPLE_ASSERT(it != listGatheringHolder.end());
            s_JoinedGatheringId = it->Get()->GetID();
            if (s_JoinedGatheringId != INVALID_GATHERINGID)
            {
                // Processes joining the matchmaking session.
                result = Nex_JoinMatchmakeSession(s_JoinedGatheringId);
                if (result.IsFailure())
                {
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    SetNetworkErrorResult(SampleFailedJoinMatchmakeSession());
                    return SCENE_NETWORK_ERROR;
                }
            }
            else
            {
                // GatheringID is an invalid value.
                return SCENE_MATCH_MAKE;  // Transitions to the same scene again.
            }
        }
        else
        {
            // The specification of the participation location is odd. (An index without any session information is specified.)
            return SCENE_MATCH_MAKE;  // Transitions to the same scene again.
        }
    }
#else
    nn::nex::PrincipalID ownerPID;

    result = Nex_AutoMatchmake(
        &s_JoinedGatheringId,
        &ownerPID,
        MATCHMAKE_GAME_MODE + s_GameModeOption + GetIncrementalGameModeValueByBitRateCheckMode(),
        SAMPLE_MAX_ENTRY,
        "PiaSession simple"
        );
    if (result.IsFailure())
    {
        SetCommunicationErrorType(MATCH_MAKE_ERROR);
        SetNetworkErrorResult(SampleFailedJoinMatchmakeSession());
        return SCENE_NETWORK_ERROR;
    }
    *pIsSessionHost = (ownerPID == Nex_GetNgsFacadeInstance()->GetCredentials()->GetPrincipalID());
    NN_PIA_DUMMY_PARAM_IN_RELEASE(isCreateNewNetwork);
#endif

    return SCENE_PREPARE_SESSION;
}


nn::Result NexFacadeStartup()
{
    nn::Result result = nn::pia::inet::NexFacade::GetInstance()->Startup();
    if (result.IsSuccess())
    {
        result = nn::pia::inet::NexFacade::GetInstance()->StartNatSessionAsync();
        if (result.IsSuccess())
        {
            // Waits for NexFacade::StartNatSession to end.
            while (!nn::pia::inet::NexFacade::GetInstance()->IsCompletedStartNatSession() && result.IsSuccess())
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();

                nn::pia::common::Scheduler::GetInstance()->Dispatch();

                // Checks the AC network connection status, game server connection status, and NEX internal errors.
                if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
                {
                    result = nn::pia::ResultNetworkConnectionIsLost();
                }
                else
                {
                    result = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
                }

                nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

                // Cancel asynchronous process
                if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
                {
                    nn::pia::inet::NexFacade::GetInstance()->CancelStartNatSessionAsync();
                }

                nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            }

            if (result.IsFailure())
            {
                // Connection is in an abnormal state.
                PIA_SAMPLE_LOG("Transport::CheckConnectionError became abnormal when waiting StartNatSession process.\n");

                // Implement StopNatSession because there could be an interruption in the middle of StartNatSession asynchronous processing.
                nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                nn::pia::inet::NexFacade::GetInstance()->Cleanup();
                return result;
            }

            result = nn::pia::inet::NexFacade::GetInstance()->GetStartNatSessionResult();

            if (result.IsSuccess())
            {
                // Transmit the matchmaking session GatheringID to NexFacade.
                nn::pia::inet::NexFacade::GetInstance()->UpdateGatheringID(s_JoinedGatheringId);
            }
            else
            {
                // The asynchronous process of StartNatSession failed.
                PIA_SAMPLE_LOG("NexFacade::StartNatSession process failed.\n");

                nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                nn::pia::inet::NexFacade::GetInstance()->Cleanup();
            }
        }
        else
        {
            // StartNatSession failed.
            PIA_SAMPLE_LOG("NexFacade::StartNatSession failed.\n");

            nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
            nn::pia::inet::NexFacade::GetInstance()->Cleanup();
        }
    }
    else
    {
        // Startup failed.
        PIA_SAMPLE_LOG("NexFacade::Startup failed.\n");
        nn::pia::inet::NexFacade::GetInstance()->Cleanup();
    }
    return result;
}


bool OpenParticipation()
{
    if (s_JoinedGatheringId == INVALID_GATHERINGID)
    {
        // The GatheringId that should have been saved is odd.
        return false;
    }

    nn::Result result = Nex_OpenParticipation(s_JoinedGatheringId);
    if (result.IsFailure())
    {
        return false;
    }
    return true;
}


bool GetHostStationConnectionInfo(nn::pia::transport::StationConnectionInfo* pInfo)
{
    // Gets the StationURL of the session host.
    nn::nex::qList<nn::nex::StationURL> lstURLs;

    nn::Result result = Nex_GetHostStationURLs(s_JoinedGatheringId, &lstURLs);
    if (result.IsSuccess())
    {
        // Successfully obtained StationURL.
        // Converts the StationURL to StationConnectionInfo.
        if (nn::pia::inet::NexFacade::ConvertNexStationUrlToStationConnectionInfo(lstURLs, pInfo).IsSuccess())
        {
            // Conversion succeeded.
            return true;
        }
    }
    return false;
}


void ResetMatchMakingClient()
{
    // To go to the matchmaking scene again, put the matchmaking session in order.
    // (Do not log off from the server or disconnect Internet communications.)

/* Starts finalization of MatchMakingClient. */
    // Finishes the matchmaking session.
    Nex_LeaveMatchmakeSession(
        s_JoinedGatheringId,
        s_OwnerPID,
        true);
    s_JoinedGatheringId = INVALID_GATHERINGID;
    Nex_FinalizeMatchmake();
/* Completed finalization of MatchMakingClient. */

/* Starts the initialization process for MatchMakingClient. */
    Nex_InitializeMatchmake();
/* Completed initialization of MatchMakingClient. */
}
