﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneLocalMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/local.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "SampleConfig.h"

#include "Shared_Local_Setup.h"

#include <string.h>

/*!
    @brief  CPP file with processing specific to local communication.


*/

extern u8 s_GameModeOption;  // The option value that changes <tt>SAMPLE_SUB_ID</tt>.
extern bool s_EnableHostMigration;  //!< Enables or disables the host migration feature. Fixed to enabled when the <tt>PiaSession</tt> host migration feature is used.

extern bool s_UsingInvalidSignature;

// Dynamically change the network topology setting of the <tt>Session</tt>.
extern size_t s_TargetNetworkTopology;
const char* GetNetworkTopologyTypeString(size_t index);

namespace
{
// For storing the <tt>UserName</tt> of the current UDS host.
nn::os::CriticalSection s_HostUserNameLock;
char s_CurrentHostUserName[SAMPLE_USER_NAME_LENGTH];
}  // end of anonymous namespace

// A shared process when getting <tt>UserName</tt>.
void CheckAndGetUserName(char userNameDst[], const nn::cfg::UserName& userNameSrc, size_t dstMaxSize)
{
    if (userNameSrc.isNgUserName)
    {
        // Profanity
        ::std::strncpy(userNameDst, "NG_WORD", dstMaxSize);
    }
    else
    {
        // Convert <tt>UserName</tt> from <tt>wchar_t</tt> to <tt>char</tt> for display on the screen.
        // (The <tt>PiaSampleFramework</tt> does not support the display of wide characters.)
        int ret = ::std::wcstombs(userNameDst, userNameSrc.userName, dstMaxSize);
        if (ret == -1)
        {
            // Conversion failed.
            ::std::strncpy(userNameDst, "NOT ASCII", dstMaxSize);
        }
    }

    // For safety, explicitly insert a terminating character.
    userNameDst[dstMaxSize - 1] = '\0';
}

// API for storing the <tt>UserName</tt> of current UDS host.
void SaveCurrentHostUserName(nn::pia::common::StationAddress hostStationAddress)
{
    if (nn::pia::local::LocalNetwork::GetInstance() == NULL)
    {
        return;
    }

    static nn::pia::local::UdsStationInfo s_StationInfoBuf;

    s_HostUserNameLock.Enter();

    if (nn::pia::local::LocalNetwork::GetInstance()->GetStationInfo(&s_StationInfoBuf, hostStationAddress).IsSuccess())
    {
        CheckAndGetUserName(s_CurrentHostUserName, s_StationInfoBuf.m_UserName, SAMPLE_USER_NAME_LENGTH);
    }
    else
    {
        // GetStationInfo failed.
        ::std::strncpy(s_CurrentHostUserName, "GET FAILED", SAMPLE_USER_NAME_LENGTH);
    }

    s_HostUserNameLock.Leave();
}

// The get function for displaying the <tt>UserName</tt> of the UDS host on the screen during communicating states.
// Copies the <tt>UserName</tt> of the current UDS host to the array passed as an argument.
void GetHostUserName(char hostUserName[])
{
    s_HostUserNameLock.Enter();

    ::std::strncpy(hostUserName, s_CurrentHostUserName, SAMPLE_USER_NAME_LENGTH);

    s_HostUserNameLock.Leave();
}

// Function for getting the signal strength for the currently connected network and returning a converted value in the range from 0 through 3.
u32 GetLocalLinkLevel()
{
    nn::pia::local::LocalNetwork* pLocalNetwork = nn::pia::local::LocalNetwork::GetInstance();

    if (pLocalNetwork == NULL)
    {
        return 0;
    }

    return pLocalNetwork->GetLinkLevel();
}

void UpdateBeaconData()
{
    if (!nn::pia::local::LocalNetwork::GetInstance()->IsConnected() || nn::pia::local::LocalNetwork::GetInstance()->IsDuringHostMigration())
    {
        return;
    }

    // The host sets beacon data.
    if (nn::pia::local::LocalNetwork::GetInstance()->IsHost())
    {
        SampleBeaconApplicationData beaconApplicationData;
        beaconApplicationData.value = static_cast<u32>(nn::os::Tick::GetSystemCurrent());
        ;
        ::std::memset(beaconApplicationData.programName, 0x00, sizeof(beaconApplicationData.programName));
        ::std::memcpy(beaconApplicationData.programName, PROGRAM_NAME, sizeof(PROGRAM_NAME));

        nn::Result result = nn::pia::local::LocalNetwork::GetInstance()->SetApplicationDataToBeacon(&beaconApplicationData, sizeof(beaconApplicationData));
        if (result.IsFailure())
        {
            // Detected when checking connection if function fails due to connection issues.
            PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidConnection());
            return;
        }
        PIA_SAMPLE_LOG("Set BeaconData (Update):0x%08X, %s\n", beaconApplicationData.value, beaconApplicationData.programName);
    }

    // Gets the beacon data.
    {
        SampleBeaconApplicationData beaconApplicationData;
        size_t beaconDataSize = 0;
        nn::Result result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationDataFromBeacon(&beaconApplicationData, &beaconDataSize, sizeof(beaconApplicationData));
        if (result.IsFailure())
        {
            // Detected when checking connection if function fails due to connection issues.
            PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidConnection() || result == nn::pia::ResultNotInCommunication());
            return;
        }

        size_t beaconDataSize2 = 0;
        result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationDataSizeFromBeacon(&beaconDataSize2);
        if (result.IsFailure())
        {
            // Detected when checking connection if function fails due to connection issues.
            PIA_SAMPLE_ASSERT(result == nn::pia::ResultInvalidConnection() || result == nn::pia::ResultNotInCommunication());
            return;
        }

        PIA_SAMPLE_ASSERT(beaconDataSize == beaconDataSize2 && beaconDataSize == sizeof(beaconApplicationData));

        PIA_SAMPLE_LOG("Get BeaconData (Update):0x%08X, %s (size:%d)\n", beaconApplicationData.value, beaconApplicationData.programName, beaconDataSize);
#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_APPLICATION_DATA, beaconApplicationData.value);
#endif
    }
}

// Sets up the features used in the PiaLocal module.
bool InitializePiaLocal()
{
    nn::Result result;

    s_HostUserNameLock.Initialize();

    result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID + s_GameModeOption, s_EnableHostMigration, ENABLE_AROUND_NETWORK_SEARCH);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    if (ENABLE_AROUND_NETWORK_SEARCH)
    {
        nn::pia::local::LocalAroundNetworkSearchSetting setting;
        setting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
        setting.m_SubId = SAMPLE_SUB_ID;
        setting.m_ScanInterval = 200;
        setting.m_ScanTime = 20;
        setting.m_NetworkInfoLifeTime = 5000;
        nn::pia::local::LocalNetwork::GetInstance()->SetAroundNetworkSearchSetting(setting);
    }

    InitializeTransportAndSession(TYPE_LOCAL);

    return result.IsSuccess();
}

void FinalizePiaLocal()
{
    // Finish cleaning up the Session instance before cleaning up the Facade instance.
    CleanupPiaSession();

    // Finish Cleanup of Facade before finalizing transport.
    if (nn::pia::local::LocalFacade::GetInstance() != NULL)
    {
        nn::pia::local::LocalFacade::GetInstance()->Cleanup();
    }

    // Finalizes PiaSession.
    FinalizePiaSession();

    // Finalizes PiaTransport.
    FinalizePiaTransport();

    // PiaLocal-related finalization processes.
    Local_CleanupPiaLocal();
    Local_FinalizePiaLocal();

    s_HostUserNameLock.Finalize();
}

/*!
    @brief  Builds a local-communication session using a <tt>LocalNetwork</tt> instance.
    
    In this state, the program continually searches for UDS sessions in the background, and either joins a found session or builds a new one.
    When this state ends successfully, the "joined local communication session" state begins.
    
    @return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo SceneLocalMatchmake(bool* pIsSessionHost)
{
    nn::Result result;

    // Confirms the settings of <tt>SAMPLE_SUB_ID</tt> and <tt>GameModeOption</tt>.
    PIA_SAMPLE_LOG("SceneLocalMatchmake: Matchmake settings -> SAMPLE_SUB_ID:%02x GameModeOption:%02d\n", SAMPLE_SUB_ID, s_GameModeOption);

    // PiaLocal startup.
    result = Local_StartupPiaLocal();
    if (result.IsFailure())
    {
        if (result == nn::pia::ResultInvalidConnection())
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_WIRELESS_OFF);
        }
        else
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        }
        return SCENE_ERROR;
    }

    size_t targetSession = 0;
    size_t foundSession = 0;

    //Static declarations that don't accumulate in the thread stack.
    static nn::pia::local::UdsNetworkDescription s_NetworkDescriptionList[GATHERING_LIST_MAX];
    static u8 s_RadioStrengthList[GATHERING_LIST_MAX];
    static nn::pia::local::UdsStationInfo s_StationInfoListBuf[SAMPLE_MAX_ENTRY];
    static char s_HostUserNameList[GATHERING_LIST_MAX][SAMPLE_USER_NAME_LENGTH];

    bool isExit = false;
    bool isBackToMenu = false;
    bool isCreateNewNetwork = false;
    bool isErrorOccurred = false;
    bool isStartScan = true;
    bool isScanning = false;

    u32 communicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);

#if PIA_TEST_CODE
    bool isNeedProgramParamUpdate = true;
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_MATCH_MAKE, false);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_GATHERING_LIST_MAX, GATHERING_LIST_MAX);
    s_UsingInvalidSignature = false;
#endif

    u32 frameNo = 0;
    // Main loop.
    while (!isExit || isScanning)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        //Handles user input. Pad information is collected by the BeginFrame function.
        {
            // Up/Down: Select a session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                targetSession = (GATHERING_LIST_MAX + targetSession - 1) % GATHERING_LIST_MAX;
#if PIA_TEST_CODE
                isNeedProgramParamUpdate = true;
#endif
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                targetSession = (targetSession + 1) % GATHERING_LIST_MAX;
#if PIA_TEST_CODE
                isNeedProgramParamUpdate = true;
#endif
            }

            // A:    Joins the selected session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                isExit = true;
                isCreateNewNetwork = false;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                isExit = true;
                isBackToMenu = true;
            }
            // X:    Create a new session.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                isExit = true;
                isCreateNewNetwork = true;
            }
            // Y:    Search for sessions.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                isStartScan= true;
#if PIA_TEST_CODE
                foundSession = 0;
                for (int i = 0; i < GATHERING_LIST_MAX; ++i)
                {
                    s_NetworkDescriptionList[i] = nn::pia::local::UdsNetworkDescription();
                }
                isNeedProgramParamUpdate = true;
#endif
            }
#if PIA_TEST_CODE
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                s_UsingInvalidSignature = true;
            }
#endif
        }

        // Scanning-related.
        if (isScanning)
        {
            // Check for the completion of the asynchronous process,
            if (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedScanNetwork())
            {
                isScanning = false;
#if PIA_TEST_CODE
                isNeedProgramParamUpdate = true;
#endif
                result = nn::pia::local::LocalNetwork::GetInstance()->GetScanNetworkResult();
                if (result.IsSuccess())
                {
                    if (!isExit)
                    {
                        // Gets scan results.
                        result = nn::pia::local::LocalNetwork::GetInstance()->GetNetworkDescriptionList(s_NetworkDescriptionList, &foundSession, GATHERING_LIST_MAX);
                        if (result.IsFailure())
                        {
                            foundSession = 0;
                        }
                        else
                        {
                            // Gets the information for the stations participating in the discovered networks, and stores the <tt>UserName</tt> of the host.
                            for (u32 i = 0; i < foundSession; ++i)
                            {
                                result = nn::pia::local::LocalNetwork::GetInstance()->GetStationInfoList(s_StationInfoListBuf, SAMPLE_MAX_ENTRY, i);
                                if (result.IsSuccess())
                                {
                                    ::std::strncpy(s_HostUserNameList[i], "NO NAME", SAMPLE_USER_NAME_LENGTH);

                                    // Searchers the host's <tt>StationInfo</tt> and saves the <tt>UserName</tt>.
                                    for (u32 j = 0; j < SAMPLE_MAX_ENTRY; ++j)
                                    {
                                        if (s_StationInfoListBuf[j].m_Role == nn::pia::local::LocalStationInfo::STATION_INFO_ROLE_HOST)
                                        {
                                            CheckAndGetUserName(s_HostUserNameList[i], s_StationInfoListBuf[j].m_UserName, SAMPLE_USER_NAME_LENGTH);
                                            break;
                                        }
                                    }
                                }
                                else
                                {
                                    ::std::strncpy(s_HostUserNameList[i], "GET FAILED", SAMPLE_USER_NAME_LENGTH);
                                }

                                result = nn::pia::local::LocalNetwork::GetInstance()->GetRadioStrength(&s_RadioStrengthList[i], i);
                                PIA_SAMPLE_ASSERT(result.IsSuccess());
                            }
                        }
                    }
                }
                else
                {
                    // Scanning failed.
                    PIA_SAMPLE_LOG("ScanNetwork Failure\n");
                    foundSession = 0;

                    if (result == nn::pia::ResultInvalidConnection())
                    {
                        // LocalNetwork must be reinitialized.
                        isExit = true;
                        isErrorOccurred = true;
                    }
                }
            }
        }
        else if (!isExit && isStartScan)
        {
            isStartScan = false;
            // Start scan
            result = nn::pia::local::LocalNetwork::GetInstance()->ScanNetworkAsync(communicationId, SAMPLE_SUB_ID + s_GameModeOption);
            if (result.IsSuccess())
            {
                isScanning = true;
                PIA_SAMPLE_LOG("ScanNetworkAsync succeeded.\n");
            }
            else
            {
                PIA_SAMPLE_LOG("ScanNetworkAsync Failure\n");
            }
        }

        // Renders to the screens.
        // Lower screen (standard).
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Search, Join or Create Network.");
        line += FONT_SIZE;
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology));
        line += FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
        line += FONT_SIZE;
        line += FONT_SIZE;

        if (isExit)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Please Wait...");
            line += FONT_SIZE;
        }
        else
        {
            // List up sessions.
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Network List ] Found:%01d ", foundSession);
            line += FONT_SIZE;

            for (u32 i = 0; i < GATHERING_LIST_MAX; i++)
            {
                if (i == targetSession)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                }

                if (i < foundSession)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  CH%02d:Radio%d Host:%-10s [%02d/%02d]", s_NetworkDescriptionList[i].GetChannel(), s_RadioStrengthList[i], s_HostUserNameList[i], s_NetworkDescriptionList[i].GetCurrentParticipants(), s_NetworkDescriptionList[i].GetMaxParticipants());
                    line += FONT_SIZE;
                }
                else
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  -----");
                    line += FONT_SIZE;
                }
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            }

            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 6);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Select Network");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Join this Network");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: BackToMenu");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Create New Network");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Search Network");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

#if PIA_TEST_CODE
        if (isNeedProgramParamUpdate)
        {
            isNeedProgramParamUpdate = false;
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_CURSOR_POS, targetSession);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_FOUND_SESSION_NUM, foundSession);
            if (targetSession < foundSession)
            {
                SampleBeaconApplicationData beaconApplicationData;
                size_t beaconDataSize = 0;
                result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationData(&beaconApplicationData, &beaconDataSize, sizeof(beaconApplicationData), &(s_NetworkDescriptionList[targetSession]));
                PIA_SAMPLE_ASSERT(result.IsSuccess());
                if (beaconDataSize != sizeof(SampleBeaconApplicationData))
                {
                    PIA_SAMPLE_LOG("%d:%d\n", targetSession, foundSession);
                    PIA_SAMPLE_LOG("size:%d (expected:%d)\n", beaconDataSize, sizeof(SampleBeaconApplicationData));
                    PIA_SAMPLE_LOG("data:%s\n", beaconApplicationData.programName);
                }
                PIA_SAMPLE_ASSERT(beaconDataSize == sizeof(SampleBeaconApplicationData));
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_APPLICATION_DATA, beaconApplicationData.value);
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_APPLICATION_DATA, 0);
            }
        }
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_WAIT, false);
#endif

    if (isErrorOccurred)
    {
        SetCommunicationErrorType(MATCH_MAKE_ERROR);
        return SCENE_ERROR;
    }

    //Finalize and return to the menu.
    if (isBackToMenu)
    {
        return SCENE_RESET;
    }

    if (isCreateNewNetwork)
    {
        // Create a new session.
        nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
        {
            createNetworkSetting.m_SubId = SAMPLE_SUB_ID + s_GameModeOption;
            createNetworkSetting.m_MaxEntry = SAMPLE_MAX_ENTRY;
            createNetworkSetting.m_LocalCommunicationId = communicationId;
            ::std::memcpy(createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
            createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
            createNetworkSetting.m_Channel = SAMPLE_CHANNEL;

            SampleBeaconApplicationData beaconApplicationData;
            beaconApplicationData.value = static_cast<u32>(nn::os::Tick::GetSystemCurrent());
            ::std::memset(beaconApplicationData.programName, 0x00, sizeof(beaconApplicationData.programName));
            ::std::memcpy(beaconApplicationData.programName, PROGRAM_NAME, sizeof(PROGRAM_NAME));
            memcpy(createNetworkSetting.m_ApplicationData, &beaconApplicationData, sizeof(beaconApplicationData));
            createNetworkSetting.m_ApplicationDataSize = sizeof(beaconApplicationData);
            PIA_SAMPLE_LOG("Set BeaconData:0x%08X, %s\n", beaconApplicationData.value, beaconApplicationData.programName);
#if PIA_TEST_CODE
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_APPLICATION_DATA, beaconApplicationData.value);
#endif
        }

        bool isSuccess = Local_CreateNetwork(&createNetworkSetting);
        if (isSuccess)
        {
            // Successfully created network.
            PIA_SAMPLE_LOG("CreateNetwork succeeded.\n");
            *pIsSessionHost = true;
        }
        else
        {
            // Failed to create network.
            PIA_SAMPLE_LOG("CreateNetwork Failure\n");
            // When the program has been properly implemented, failure occurs only when the wireless is off, or the system is entering Sleep Mode.
            // With this demo, all of these errors are lumped together and treated as matchmaking errors.
            SetCommunicationErrorType(MATCH_MAKE_ERROR);
            return SCENE_ERROR;
        }
    }
    else
    {
        // Join an existing session.
        nn::pia::local::UdsConnectNetworkSetting connectNetworkSetting;
        {
            ::std::memcpy(connectNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
            connectNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
            connectNetworkSetting.m_pLocalNetworkDescription = &(s_NetworkDescriptionList[targetSession]);

            // Output the BSSID of the station attempting to join the session to the log.
            bit8 bssidBuf[nn::uds::MAC_ADDRESS_SIZE];
            s_NetworkDescriptionList[targetSession].GetBssid(bssidBuf);
            PIA_SAMPLE_LOG("BSSID:");
            for (int i = 0; i < nn::uds::MAC_ADDRESS_SIZE; ++i)
            {
                PIA_SAMPLE_LOG(" %02x", bssidBuf[i]);
            }
            PIA_SAMPLE_LOG("\n");
        }
        bool isSuccess = Local_ConnectNetwork(&connectNetworkSetting);
        if (isSuccess)
        {
            // Successfully joined network.
            PIA_SAMPLE_LOG("ConnectNetwork succeeded.\n");
            *pIsSessionHost = false;

            // Gets the beacon data.
            {
                SampleBeaconApplicationData beaconApplicationData;
                size_t beaconDataSize = 0;
                result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationData(&beaconApplicationData, &beaconDataSize, sizeof(beaconApplicationData), connectNetworkSetting.m_pLocalNetworkDescription);
                PIA_SAMPLE_ASSERT(result.IsSuccess());

                size_t beaconDataSize2 = 0;
                result = nn::pia::local::LocalNetwork::GetInstance()->GetApplicationDataSize(&beaconDataSize2, connectNetworkSetting.m_pLocalNetworkDescription);
                PIA_SAMPLE_ASSERT(result.IsSuccess());

                PIA_SAMPLE_ASSERT(beaconDataSize == beaconDataSize2 && beaconDataSize == sizeof(beaconApplicationData));

                PIA_SAMPLE_LOG("Get BeaconData:0x%08X, %s (size:%d)\n", beaconApplicationData.value, beaconApplicationData.programName, beaconDataSize);
#if PIA_TEST_CODE
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_APPLICATION_DATA, beaconApplicationData.value);
#endif
            }
        }
        else
        {
            // Failed to join network.
            PIA_SAMPLE_LOG("ConnectNetwork Failure\n");

            // With this demo, all of these errors are lumped together and treated as matchmaking errors.
            SetCommunicationErrorType(MATCH_MAKE_ERROR);
            return SCENE_ERROR;
        }
    }

    return SCENE_PREPARE_SESSION;
}
