﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneSelectNetworkType.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "SampleConfig.h"

#if USE_PIA_LOCAL
#include <pia/local.h>
#endif


namespace
{
}  // end of anonymous namespace

u8 s_GameModeOption = 0;  // Defines the option value added to <tt>MATCHMAKE_GAME_MODE</tt> and <tt>SAMPLE_SUB_ID</tt> to add variation.
bool s_EnableHostMigration = true;  //!< Enables or disables the host migration feature. Fixed to enabled when the <tt>PiaSession</tt> host migration feature is used.
bool s_IsEnableBitRateCheck = false;

// Dynamically change the network topology setting of the <tt>Session</tt>.
extern size_t s_TargetNetworkTopology;
size_t GetNetworkTopologyTypeNum();
const char* GetNetworkTopologyTypeString(size_t index);
bool IsEnableRelay(size_t index);

// This macro changes the game mode option to avoid matching sessions with different network topology settings.
#define CURRENT_GAME_MODE_OPTION(_option) \
    static_cast<u8>(_option + (s_TargetNetworkTopology* GAME_MODE_OPTION_MAX))

extern const char* TOKEN_TYPE_STRING[];

/*!
    @brief  Handles the state used to select a local or internet connection.
           
    @return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo sceneSelectNetworkType(NetworkType* pNwType)
{
    nn::Result result;

    s_GameModeOption = 0;
    u8 optionCursorPosition = 0;          // Specifies the position of the cursor used to change the game mode option.
    const u8 GAME_MODE_POSITION_MAX = 4;  // Specifies the number of digits representing the cursor position.
    const u8 GAME_MODE_OPTION_MAX = 20;   // Specifies the maximum value of the game mode option.

    s_TargetNetworkTopology = 0;

    size_t targetTokenType = 0;

    bool isExit = false;

    s_IsEnableBitRateCheck = false;

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_SELECT_NETWORK_TYPE, false);
#endif

    int frameNo = 0;
    // Main loop.
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        //Handles user input. Pad information is collected by the BeginFrame function.
        {
            // Right/Left: Moves the cursor used to change the game mode.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                optionCursorPosition = (GAME_MODE_POSITION_MAX + optionCursorPosition - 1) % GAME_MODE_POSITION_MAX;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                optionCursorPosition = (optionCursorPosition + 1) % GAME_MODE_POSITION_MAX;
            }
            u8 optionDiff = 1;
            for (int i = optionCursorPosition; i < GAME_MODE_POSITION_MAX - 1; i++)
            {
                optionDiff *= 10;
            }
            // Up/Down: Select the method for using ID tokens, or change the game mode value.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                if (optionCursorPosition == 0)
                {
                    targetTokenType = (TOKEN_TYPE_NUM + targetTokenType - 1) % TOKEN_TYPE_NUM;
                }
                else if (optionCursorPosition == 1)
                {
                    s_TargetNetworkTopology = (GetNetworkTopologyTypeNum() + s_TargetNetworkTopology - 1) % GetNetworkTopologyTypeNum();
                }
                else
                {
                    s_GameModeOption = (s_GameModeOption + optionDiff) % GAME_MODE_OPTION_MAX;
                }
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                if (optionCursorPosition == 0)
                {
                    targetTokenType = (targetTokenType + 1) % TOKEN_TYPE_NUM;
                }
                else if (optionCursorPosition == 1)
                {
                    s_TargetNetworkTopology = (s_TargetNetworkTopology + 1) % GetNetworkTopologyTypeNum();
                }
                else
                {
                    s_GameModeOption = (GAME_MODE_OPTION_MAX + s_GameModeOption - optionDiff) % GAME_MODE_OPTION_MAX;
                }
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                s_IsEnableBitRateCheck = !s_IsEnableBitRateCheck;
                PIA_SAMPLE_LOG("Change BitRate check mode: %s\n", s_IsEnableBitRateCheck ? "Enable" : "Disable");
            }

#if USE_PIA_LOCAL
            // A:    Selects Local.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                // The settings used for relay communications are not set for local communication.
                if (!IsEnableRelay(s_TargetNetworkTopology))
                {
                    isExit = true;
                    *pNwType = TYPE_LOCAL;

                    s_IsEnableBitRateCheck = false;
#if PIA_TEST_CODE
                    if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
                    {
                        s_EnableHostMigration = false;
                    }
                    else
                    {
                        s_EnableHostMigration = true;
                    }
                    PIA_SAMPLE_LOG("Host Migration %s\n", s_EnableHostMigration ? "Enabled" : "Disabled");
#endif
                }
            }
#endif
            // B:    Select Internet (NEX).
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                isExit = true;
                *pNwType = TYPE_INTERNET;

                // Host migration is on by default.
                s_EnableHostMigration = true;
#if PIA_TEST_CODE
                PIA_SAMPLE_LOG("Host Migration %s\n", s_EnableHostMigration ? "Enabled" : "Disabled");
#endif
            }
        }

#if PIA_TEST_CODE
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_ENABLE_HOST_MIGRATION, s_EnableHostMigration ? 1 : 0);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(PARAM_KEY_NETWORK_TYPE, *pNwType);
#endif

        // Renders to the screens.
        // Lower screen (standard).
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
        line += FONT_SIZE;

        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Select network-type.");
            line += FONT_SIZE;
            line += FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Please Select NetworkType:");
            line += FONT_SIZE;
#if USE_PIA_LOCAL
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   A: UDS Local communication");
            line += FONT_SIZE;
#endif
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   B: Internet (using NEX matchmake)");
            line += FONT_SIZE;

            line += FONT_SIZE * 2;
            if (optionCursorPosition == GAME_MODE_POSITION_MAX - 1 || optionCursorPosition == GAME_MODE_POSITION_MAX - 2)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText((17 + optionCursorPosition) * FONT_SIZE, line, "v");
                line += FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
            }
            else
            {
                line += FONT_SIZE;
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption);
            line += FONT_SIZE;

            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Using TokenType:");
            line += FONT_SIZE;
            for (int i = 0; i < TOKEN_TYPE_NUM; i++)
            {
                if (i == targetTokenType)
                {
                    if (optionCursorPosition == 0)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                    }
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " > %s", TOKEN_TYPE_STRING[i]);
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                }
                else
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", TOKEN_TYPE_STRING[i]);
                    line += FONT_SIZE;
                }
            }

            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology Setting:");
            line += FONT_SIZE;
            for (int i = 0; i < GetNetworkTopologyTypeNum(); i++)
            {
                if (i == s_TargetNetworkTopology)
                {
                    if (optionCursorPosition == 1)
                    {
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
                    }
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " > %s", GetNetworkTopologyTypeString(i));
                    line += FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
                }
                else
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", GetNetworkTopologyTypeString(i));
                    line += FONT_SIZE;
                }
            }


            line = LOWER_SCREEN_HEIGHT - (FONT_SIZE * 5);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP/DOWN: Select using TokenType");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "           Select NetworkTopology");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "           Change GameMode option");
            line += FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  LEFT/RIGHT: Move cursor");
            line += FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = UPPER_SCREEN_HEIGHT - FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", GetPiaVersionString());
        line += FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        frameNo++;
    }

    // Set the local station identification token to match the token type.
    SetLocalIdentificationToken(static_cast<SampleTokenType>(targetTokenType));
    // Revises the game mode value so matching is not done with a different network topology.
    s_GameModeOption = CURRENT_GAME_MODE_OPTION(s_GameModeOption);

#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_WAIT, false);
#endif

    return SCENE_MATCH_MAKE;
}
