﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Session_Mesh.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"
#include "Shared_Basic_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Session_Mesh.h"

#if NN_PIA_CAFE
#include <cafe.h>
#include <nn/act.h>
#include <nn/fp.h>
#elif NN_PIA_CTR
#include <nn/friends.h>
#endif

#include <pia/transport.h>

// Macro for exiting while loop when a network disconnect is detected.
#define BREAK_IF_CONNECTION_IS_LOST()                                                                            \
    \
{                                                                                                         \
        nn::Result connectionStatusResult = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();\
        if (connectionStatusResult.IsSuccess())                                                                  \
        {                                                                                                        \
        }                                                                                                        \
        else if (connectionStatusResult == nn::pia::ResultNetworkConnectionIsLost())                             \
        {                                                                                                        \
            PIA_SAMPLE_LOG("ResultNetworkConnectionIsLost is detected.  Break.\n");                              \
            break;                                                                                               \
        }                                                                                                        \
        else                                                                                                     \
        {                                                                                                        \
            PIA_SAMPLE_LOG_RESULT(connectionStatusResult);                                                            \
            PIA_SAMPLE_ASSERTMSG(false, "Transport::CheckConnectionError() returned unexpected Result.");        \
        }                                                                                                        \
    \
}



nn::Result Session_InitializePiaSession_ForMesh(nn::pia::transport::NetworkFactory* pFactory,
                                                nn::pia::session::NetworkTopology topology,
                                                nn::pia::session::BitRateCheckMode bitRateCheckMode)
{
    nn::pia::session::Mesh::Setting setting;
    setting.pNetworkFactory = pFactory;
    setting.networkTopology = topology;
    setting.bitRateCheckMode = bitRateCheckMode;

    return Session_InitializePiaSession_ForMesh(setting);
}

nn::Result Session_InitializePiaSession_ForMesh(const nn::pia::session::Mesh::Setting& setting)
{
    return Session_InitializePiaSession_ForMesh(setting, false);
}

nn::Result Session_InitializePiaSession_ForMesh(
    const nn::pia::session::Mesh::Setting& setting,
    bool isEnableRelayEmulation)
{
    PIA_SAMPLE_LOG("Session_InitializePiaSession_ForMesh\n");
    nn::Result result;

    result = nn::pia::session::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    {
        // For debug feature settings.
        nn::pia::session::Mesh::DebugSetting debugSetting;
        // Settings for connection route emulation for debugging.
        debugSetting.isEnableRelayEmulation = isEnableRelayEmulation;

        // Make sure that this does not get called in the retail version ROM.
        result = nn::pia::session::Mesh::SetDebugSetting(debugSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    result = nn::pia::session::Mesh::CreateInstance(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Session_InitializePiaSession_ForMesh:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());
    return result;
}

void Session_FinalizePiaSession_ForMesh()
{
    PIA_SAMPLE_LOG("Session_FinalizePiaSession_ForMesh\n");

    // With Pia, Finalize and DestroyInstance always succeed.
    if (nn::pia::session::Mesh::GetInstance() != NULL)
    {
        nn::pia::session::Mesh::DestroyInstance();
    }
    nn::pia::session::Finalize();
}


void Session_GetLocalPlayerName_ForMesh(nn::pia::transport::Station::PlayerName* pPlayerName)
{

    nn::Result result;

#if NN_PIA_CAFE
    //Uses the name of the Mii character.
    char16 miiName[ACT_MII_NAME_SIZE];
    std::memset(miiName, 0, sizeof(miiName));

    result = nn::act::GetMiiName(miiName);
    if (result.IsSuccess())
    {
        for (int i = 0; i < ACT_MII_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (miiName[i] != 0)
            {
                pPlayerName->data[i] = miiName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    nn::fp::MyPresence myPresence;
    result = nn::fp::GetMyPresence(&myPresence);
    if (result.IsSuccess())
    {
        pPlayerName->language = myPresence.consoleInfo.language;
    }
    else
    {
        pPlayerName->language = nn::fp::LANGUAGE_INVALID;
    }
#elif NN_PIA_CTR
    //Uses the name of the Mii character.
    char16 screenName[nn::friends::SCREEN_NAME_SIZE];
    result = nn::friends::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    result = nn::friends::GetMyScreenName(screenName);
    if (result.IsSuccess() && screenName[0] != 0)  // Also exclude cases where a Mii character is not set.
    {
        for (int i = 0; i < nn::friends::SCREEN_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (screenName[i] != 0)
            {
                pPlayerName->data[i] = screenName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    pPlayerName->language = 0;
#else
    std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
    pPlayerName->length = 9;
    pPlayerName->language = 0;
#endif
}

nn::Result Session_StartupMesh(nn::pia::session::Mesh::StartupSetting startupSetting)
{
    PIA_SAMPLE_LOG("signature GetKeySize : %d\n", startupSetting.signatureSetting.GetKeySize());

    nn::Result result;
    nn::pia::transport::Station::PlayerName playerName;

    Session_GetLocalPlayerName_ForMesh(&playerName);
    startupSetting.pPlayerName = &playerName;    

    result = nn::pia::session::Mesh::GetInstance()->Startup(startupSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    return result;
}


void Session_CleanupMesh()
{
    PIA_SAMPLE_LOG("Session_CleanupMesh\n");
    nn::pia::session::Mesh::GetInstance()->Cleanup();
#if NN_PIA_CTR
    nn::friends::Finalize();
#endif
}

bool Session_CreateMesh()
{
    PIA_SAMPLE_LOG("Session_CreateMesh\n");

    nn::Result result;

    // Creation of mesh.
    result = nn::pia::session::Mesh::GetInstance()->CreateMeshAsync();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Wait for <tt>CreateMeshAsync</tt>  to complete.
    int loopCounter = 0;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "creating session... %d", loopCounter++);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        BREAK_IF_CONNECTION_IS_LOST();

        if (nn::pia::session::Mesh::GetInstance()->IsCompletedCreateMesh())
        {
            result = nn::pia::session::Mesh::GetInstance()->GetCreateMeshResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_CreateMesh: CreateMeshAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                PIA_SAMPLE_ASSERT(result == nn::pia::ResultFailedStartupProtocol());
                return false;
            }

            PIA_SAMPLE_LOG("Session_CreateMesh: CreateMeshAsync succeeded.\n");
            return true;
        }
    }

    PIA_SAMPLE_LOG("Session_CreateMesh: Connection is lost.\n");
    return false;
}

void Session_DestroyMesh()
{
    PIA_SAMPLE_LOG("Session_DestroyMesh\n");

    if (nn::pia::session::Mesh::GetInstance()->CheckConnectionError() == nn::pia::ResultMeshConnectionIsLost())
    {
        PIA_SAMPLE_LOG("Session_DestroyMesh: ResultMeshConnectionIsLost.  DestroyMeshAsync() is not called.\n");
        return;
    }

    // Destruction of mesh.
    nn::Result result;
    result = nn::pia::session::Mesh::GetInstance()->DestroyMeshAsync();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Wait for <tt>DestroyMesh</tt> to complete.
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "destroying session...");

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        BREAK_IF_CONNECTION_IS_LOST();

        if (nn::pia::session::Mesh::GetInstance()->IsCompletedDestroyMesh())
        {
            // The destruction of the mesh has completed.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(nn::pia::session::Mesh::GetInstance()->GetDestroyMeshResult());
            PIA_SAMPLE_LOG("Session_DestroyMesh: DestroyMeshAsync succeeded.\n");
            return;
        }
    }
}


void Session_LeaveMeshWithHostMigration()
{
    PIA_SAMPLE_LOG("Session_LeaveMeshWithHostMigration\n");

    if (nn::pia::session::Mesh::GetInstance()->CheckConnectionError() == nn::pia::ResultMeshConnectionIsLost())
    {
        PIA_SAMPLE_LOG("Session_LeaveMeshWithHostMigration: ResultMeshConnectionIsLost.  LeaveMeshWithHostMigrationAsync() is not called.\n");
        return;
    }

    // Issue instruction to begin host migration and leave the mesh.
    nn::Result result;
    result = nn::pia::session::Mesh::GetInstance()->LeaveMeshWithHostMigrationAsync();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Waits for <tt>LeaveMeshWithHostMigration</tt>  to complete.
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "leaving session(with HM)...");

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        BREAK_IF_CONNECTION_IS_LOST();

        if (nn::pia::session::Mesh::GetInstance()->IsCompletedLeaveMeshWithHostMigration())
        {
            // Leaving the mesh has completed.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(nn::pia::session::Mesh::GetInstance()->GetLeaveMeshWithHostMigrationResult());
            PIA_SAMPLE_LOG("Session_LeaveMeshWithHostMigration: LeaveMeshWithHostMigrationAsync succeeded.\n");
            return;
        }
    }
}


bool Session_JoinMesh(const nn::pia::transport::StationConnectionInfo& stationConnectionInfo)
{
    PIA_SAMPLE_LOG("Session_JoinMesh\n");

    nn::Result result;

    // Joins a mesh.
    result = nn::pia::session::Mesh::GetInstance()->JoinMeshAsync(stationConnectionInfo);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_JoinMesh: JoinMeshAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultStationConnectionFailed());
        return false;
    }

    // Waits for the <tt>JoinMeshAsync(/tt> function to complete.
    int loopCounter = 0;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // The job's dispatch process (execution of job inserted by JoinMesh).
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "joinning session... %d", loopCounter++);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        BREAK_IF_CONNECTION_IS_LOST();

        if (nn::pia::session::Mesh::GetInstance()->IsCompletedJoinMesh())
        {
            result = nn::pia::session::Mesh::GetInstance()->GetJoinMeshResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_JoinMesh: JoinMeshAsync() failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                PIA_SAMPLE_ASSERT(result == nn::pia::ResultStationConnectionFailed() ||
                                  result == nn::pia::ResultStationConnectionNatTraversalFailedUnknown() ||
                                  result == nn::pia::ResultNatTraversalFailedBothEim() ||
                                  result == nn::pia::ResultNatTraversalFailedBothEdm() ||
                                  result == nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm() ||
                                  result == nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim() ||
                                  result == nn::pia::ResultJoinRequestDenied() ||
                                  result == nn::pia::ResultMeshIsFull() ||
                                  result == nn::pia::ResultRelayFailedNoCandidate() ||
                                  result == nn::pia::ResultRelayFailedRttLimit() ||
                                  result == nn::pia::ResultRelayFailedRelayNumLimit() ||
                                  result == nn::pia::ResultRelayFailedUnknown() ||
                                  result == nn::pia::ResultNatTraversalRequestTimeout() ||
                                  result == nn::pia::ResultInvalidSystemMessage());
                return false;
            }

            PIA_SAMPLE_LOG("Session_JoinMesh: JoinMeshAsync() succeeded.\n");
            return true;
        }
    }

    PIA_SAMPLE_LOG("Session_JoinMesh: Connection is lost.\n");
    return false;
}

void Session_LeaveMesh()
{
    PIA_SAMPLE_LOG("Session_LeaveMesh\n");

    if (nn::pia::session::Mesh::GetInstance()->CheckConnectionError() == nn::pia::ResultMeshConnectionIsLost())
    {
        PIA_SAMPLE_LOG("Session_LeaveMesh: ResultMeshConnectionIsLost.  LeaveMeshAsync() is not called.\n");
        return;
    }

    // Leaves the mesh.
    nn::Result result = nn::pia::session::Mesh::GetInstance()->LeaveMeshAsync();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Waits for the <tt>LeaveMesh</tt> function to complete.
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "leaving session...");

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        BREAK_IF_CONNECTION_IS_LOST();

        // Determines whether the <tt>LeaveMesh</tt> function succeeded.
        if (nn::pia::session::Mesh::GetInstance()->IsCompletedLeaveMesh())
        {
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(nn::pia::session::Mesh::GetInstance()->GetLeaveMeshResult());
            PIA_SAMPLE_LOG("Session_LeaveMesh: LeaveMeshAsync succeeded.\n");
            return;
        }
    }
}
