﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/version.h>
#include <pia/common.h>


#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"

#include "SampleConfig.h"

#if USE_PIA_LOCAL
#include <pia/local.h>
#endif

namespace
{
/* ---< Prepare memory to allocate to Pia. >--- */
// The memory region must be 4-byte aligned.
// In this sample, sufficiently large values are used.
PIA_ATTRIBUTE_ALIGN(4) bit8 s_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA];
/* ---< End of memory preparations. >--- */

// Variables for saving the Pia library version information character string.
char s_SamplePiaVersionInfoString[UPPER_SCREEN_WIDTH / FONT_SIZE] = {};


void InitializePiaCommon(void)
{
    /* ---< Setup of the features used in the PiaCommon module >--- */
    nn::Result result = Common_InitializePiaCommon(
        s_PiaMemoryBlock,
        MEMORY_BLOCK_SIZE_PIA,
        nn::pia::common::TRACE_FLAG_COMMON |
            nn::pia::common::TRACE_FLAG_LOCAL |
            nn::pia::common::TRACE_FLAG_TRANSPORT |
            nn::pia::common::TRACE_FLAG_SESSION |
            nn::pia::common::TRACE_FLAG_INET |
//n1589:test:off
#if PIA_TEST_CODE
        nn::pia::common::TRACE_FLAG_INET_NAT |
        nn::pia::common::TRACE_FLAG_INET_NAT_DETAIL |
        nn::pia::common::TRACE_FLAG_INET_NAT_PROBE |
        nn::pia::common::TRACE_FLAG_INET_SOCKET |

        nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_SEND |
        nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_RECEIVE |

        nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
        nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
        nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
        nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
#endif
            nn::pia::common::TRACE_FLAG_ALWAYS,
        true,
        true);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    /* ---<Setup of the features used in the end of the PiaCommon module. >--- */
}


void FinalizePiaCommon(void)
{
    /* ---< Finalize PiaCommon. >--- */
    Common_FinalizePiaCommon();
    /* ---< End of PiaCommon-related finalization. >--- */
}


}  // end of anonymous namespace

// Function that gets the Pia library version information string.
const char* GetPiaVersionString()
{
    return s_SamplePiaVersionInfoString;
}

#if NN_PIA_CTR
// Functions for getting the signal strength according to the communication method.
u32 GetLocalLinkLevel();
u32 GetInternetLinkLevel();
#endif

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel(NetworkType nwType)
{
#if NN_PIA_CTR
    // Get the signal strength according to the communication method, and return a value from 0 to 3.
    switch (nwType)
    {
    case TYPE_LOCAL:
    {
        return GetLocalLinkLevel();
    }
    case TYPE_INTERNET:
    {
        return GetInternetLinkLevel();
    }
    default:
    {
        return 0;
    }  // Invalid argument.
    }
#elif NN_PIA_WIN
    // Always 0.
    return 0;
#elif NN_PIA_CAFE
    // There is no way to get the signal strength, so it is always 0.
    return 0;
#else
#error  "Illegal platform."
#endif
}

PIA_SAMPLE_MAIN_FUNC()
{
    PIA_SAMPLE_LOG("Sample program Start.\n");

    // Creates the Pia library version information string.
    ::std::snprintf(s_SamplePiaVersionInfoString, UPPER_SCREEN_WIDTH / FONT_SIZE, "Pia %d.%d.%d r%d", PIA_VERSION_MAJOR, PIA_VERSION_MINOR, PIA_VERSION_MICRO, PIA_VERSION_REVISION);

    Basic_SetProgramTitle(PROGRAM_NAME);

    nn::Result result;

    // Initialize the render framework for the Pia sample program.
    nn::pia::sampleframework::SampleFramework::CreateInstance();
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAME_PER_SECOND);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.0f, 0.0f, 0.2f, 1.0f));
#if PIA_TEST_CODE
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_SELECT_NETWORK_TYPE, false);
#endif

    bool piaCommonInitialized = false;

    while (true)
    {
        SampleSceneNo nextSceneNo = SCENE_ERROR;

        if (!piaCommonInitialized)
        {
            InitializePiaCommon();
            piaCommonInitialized = true;
        }

        // Because the processes for each feature are clearly separated in this sample, the processes in the main thread are divided into several scenes.
        bool isExit = false;

        NetworkType nwType = TYPE_UNKNOWN;  //!< Available for switching exit processing.
        nextSceneNo = SCENE_SELECT_NETWORK_TYPE;
        bool isSessionHost = false;

        while (isExit == false)
        {
            switch (nextSceneNo)
            {
            case SCENE_SELECT_NETWORK_TYPE:
            {
                nextSceneNo = sceneSelectNetworkType(&nwType);

                switch (nwType)
                {
#if USE_PIA_LOCAL
                case TYPE_LOCAL:
                {
                    if (InitializePiaLocal() == false)  //Initialization error. (The cause should already be set in a function.)
                    {
                        nextSceneNo = SCENE_ERROR;
                    }
                }
                break;
#endif  // end of USE_PIA_LOCAL
                case TYPE_INTERNET:
                {
                    if (InitializePiaInet() == false)  //Initialization error. (The cause should already be set in a function.)
                    {
                        nextSceneNo = SCENE_NETWORK_ERROR;
                    }
                }
                break;
                }
            }
            break;
            case SCENE_MATCH_MAKE:
            {
                switch (nwType)
                {
#if USE_PIA_LOCAL
                case TYPE_LOCAL:
                {
                    nextSceneNo = SceneLocalMatchmake(&isSessionHost);
                }
                break;
#endif  // end of USE_PIA_LOCAL
                case TYPE_INTERNET:
                {
                    nextSceneNo = SceneInetMatchmake(&isSessionHost);
                }
                break;
                }
            }
            break;
            case SCENE_PREPARE_SESSION:
            {
                nextSceneNo = ScenePrepareSession(nwType, isSessionHost);
            }
            break;
            case SCENE_COMMUNICATE:
            {
                nextSceneNo = SceneCommunication(nwType);
            }
            break;
            case SCENE_ERROR:
            {
                nextSceneNo = SceneError();
            }
            break;
            case SCENE_NETWORK_ERROR:
            {
                nextSceneNo = SceneNetworkError();
            }
            break;
            case SCENE_RESET:
            {
                switch (nwType)
                {
#if USE_PIA_LOCAL
                case TYPE_LOCAL:
                {
                    FinalizePiaLocal();
                }
                break;
#endif  // end of USE_PIA_LOCAL
                case TYPE_INTERNET:
                {
                    FinalizePiaInet();
                }
                break;
                }
                isExit = true;
            }
            break;
            }
        }

        if (piaCommonInitialized)
        {
            FinalizePiaCommon();
            piaCommonInitialized = false;
        }
    }
    //Control does not arrive here.
    //nn::pia::sampleframework::SampleFramework::DestroyInstance();
    //PIA_SAMPLE_LOG("Sample program end.\n");
}  // end of nnMain

// When DDL files are not used with NEX, a definition of the empty InitDOClasses function is required.
namespace nn
{
namespace nex
{
void InitDOClasses()
{
}
}
}
