﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     PrepareInternet.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>
#include <pia/transport/transport_StationConnectionInfo.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "Shared_Basic_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"

#include "SampleConfig.h"


#if NN_PIA_CTR
#include <nn/os.h>
#include <nn/ac.h>
#elif NN_PIA_CAFE
#include <nn/ac.h>
#endif


/*!

    @brief  CPP file with processing specific to Internet communication.

*/

//extern u8 s_GameModeOption; // Option values for changing MATCHMAKE_GAME_MODE.

// Dynamically change the network topology setting of the <tt>Session</tt>.
//extern size_t s_TargetNetworkTopology;
//const char* GetNetworkTopologyTypeString(size_t index);

namespace
{
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

// Displays the progress of <tt>InitializePiaInet</tt> on the screen.
void InitializingProcessRender(u32 processingStep)
{
    // Gets the signal strength information.
    u32 linkLevel = GetLinkLevel();
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_PREPARE_INET,
        false,
        false,
        0,
        linkLevel);

    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

    // Lower screen.
    nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    int line = 0;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo]", PROGRAM_NAME);
    line += BASIC_FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Inet Initializing process.");
    line += BASIC_FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(BASIC_LOWER_SCREEN_WIDTH - (BASIC_FONT_SIZE * 9), line, "LinkLv:%d", linkLevel);
    line += BASIC_FONT_SIZE;

    line += BASIC_FONT_SIZE;

    // Progress display.
    // Steps that have been processed are displayed in white; steps that have not been processed are displayed in red.
    {
        const char* stepStrings[] = {
            "Initialize Network",
            "Initialize NEX",
            "Login server",
            "Initialize inet module"
        };  // 4 step.
        const u32 STEP_MAX = 4;

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   START");
        line += BASIC_FONT_SIZE;
        for (u32 step = 0; step < STEP_MAX; step++)
        {
            if (step == processingStep)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "     |");
                line += BASIC_FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_RED);
            }
            else
            {
                line += BASIC_FONT_SIZE;
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   %s", stepStrings[step]);
            line += BASIC_FONT_SIZE;
        }
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   END");
        line += BASIC_FONT_SIZE;
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

    nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

    // Upper screen.
    nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

    line = BASIC_UPPER_SCREEN_HEIGHT - BASIC_FONT_SIZE;
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", Basic_GetPiaVersionString());
    line += BASIC_FONT_SIZE;

    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
}

}  // end of anonymous namespace

#if NN_PIA_CTR
// Function for getting the signal strength of the communication with the access point, and returning a converted value in the range from 0 through 3.
u32 GetInternetLinkLevel()
{
    nn::ac::CTR::LinkLevel acLinkLevel = nn::ac::CTR::GetLinkLevel();

    switch (acLinkLevel)
    {
    case nn::ac::CTR::LINK_LEVEL_3:
        return 3;
    case nn::ac::CTR::LINK_LEVEL_2:
        return 2;
    case nn::ac::CTR::LINK_LEVEL_1:
        return 1;
    case nn::ac::CTR::LINK_LEVEL_0:
    default:
        return 0;
    }
}
#endif

/* ---< Setup of the features used in the <tt>PiaInet</tt> module. >--- */
// Performs Internet connection processing, logging on to the game server, and so on.
// Because <tt>nn::ac::Connect</tt> blocks for long periods, we do not recommend running it on the main thread.
// This sample program runs it here to emphasize process readability.
bool InitializePiaInet()
{
    nn::Result result;

    // Because the blocking process continues for a long time, the progress is displayed on the screen.
    u32 step = 0;
    InitializingProcessRender(step);
    step++;

    /* Start processing NEX-related initialization. */

    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    /* Finished processing NEX-related initialization. */

    InitializingProcessRender(step);
    step++;

    PIA_SAMPLE_LOG("NEX initialize completed.\n");

    /* Started processing PiaInet initialization. */
    result = Inet_InitializePiaInet();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    
    PIA_SAMPLE_LOG("Inet_InitializePiaInet completed.\n");
    
    InitializePiaSession();
    PIA_SAMPLE_LOG("InitializePiaSession completed.\n");

    /* Finished processing <tt>PiaInet</tt> initialization. */

    InitializingProcessRender(step);
    step++;

    /* Preparing for Internet communication. */
    result = Inet_InitializeNetwork();
    if (result.IsFailure())
    {
#if NN_PIA_CTR
        if (result == nn::ac::ResultWifiOff())
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_WIRELESS_OFF);
        }
        else
#endif
        {
            SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        }
        PIA_SAMPLE_LOG("Inet_InitializeNetwork failed.\n");

        SetNetworkErrorResult(result);

        return false;
    }
    PIA_SAMPLE_LOG("Inet_InitializeNetwork completed.\n");

    /* Finished preparing Internet connection processing. */

    InitializingProcessRender(step);
    step++;

    /* Start of the process of logging in to the server. */

    result = Inet_Login();
    if (result.IsFailure())
    {
        // Log in fails.
        SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        SetNetworkErrorResult(result);
        PIA_SAMPLE_LOG("Inet_Login failed.\n");
        return false;
    }

    PIA_SAMPLE_LOG("Inet_Login completed.\n");

    /* End of the process of logging in to the server. */

    InitializingProcessRender(step);
    step++;

    return true;
}


void FinalizePiaInet()
{
    /* Process to log out from the server. */
    Inet_Logout();

    /* Processes Internet finalization (disconnection). */
    Inet_FinalizeNetwork();

    FinalizePiaSession();

    /* Finalizes PiaInet. */
    Inet_FinalizePiaInet();

    /* NEX-related finalization. */
    Nex_FinalizeNex();
}
