﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SampleConfig.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/platform.h>

#include <pia/inet.h>
#include <pia/inet/inet_NexFacade.h>
#include <pia/inet/inet_NexNetworkFactory.h>

#include <pia/transport.h>
#include <pia/session.h>

#define FRAME_PER_SECOND 60

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

// Set to 1 to use the matchmaking analysis feature.
#define ENABLE_MATCHMAKE_ANALYSE 0
#define ENABLE_BIT_RATE_CHECK 0

const char PROGRAM_NAME[] = "session/RandomMatchmake";

/*!
    Enumerates possible states of the sample program.
 */
enum SampleSceneNo
{
    SCENE_PREPARE_INET,     //!< This is the state where preparations are made for Internet communication. It has no relation to communication.
    SCENE_MATCH_MAKE,       //!< This scene creates sessions for joining and handles the joining process for each of the specified network types.
    SCENE_COMMUNICATE,      //!< In this state, communication with peers occurs within the session.
    SCENE_ERROR,            //!< This is the state that is transitioned to when problems occur. This demo simulates the type of error message window found in an actual application.
    SCENE_NETWORK_ERROR,    //!< This is the state showing network error codes. Assume that the error or EULA applet normally will start in this state.
    SCENE_RESET             //!< Defines the state where all information managed by the demo has been reset to the default values.
};

/*!
    @brief  An enumerated type indicating the network type.
 */
enum NetworkType
{
    TYPE_LOCAL,    //!< Indicates local (UDS) communication.
    TYPE_INTERNET  //!< Indicates Internet communication.
};

/*!
    Enumerates communication errors handled by the sample program.
 */
enum SampleErrorType
{
    ERROR_NONE = 0,                 //!< Specifies that no error occurred.
    INITIALIZE_ERROR_WIRELESS_OFF,  //!< Specifies that initialization failed because wireless mode is off.
    INITIALIZE_ERROR_OTHER_REASON,  //!< Specifies that initialization failed for some other reason.
    MATCH_MAKE_ERROR,               //!< Could not establish the session.
    SESSION_LOST_ERROR,             //!< Disconnected from the session during communication.
    CONNECTION_LOST_ERROR           //!< The network connection was disconnected.
};

void SetCommunicationErrorType(SampleErrorType error);  //Sets the communication error type.
void SetNetworkErrorResult(nn::Result result);          //This function sets the <tt>Result</tt> value to be converted into a network error code.

bool InitializePiaInet();
void FinalizePiaInet();

void InitializePiaSession();
void CleanupPiaSession();
void FinalizePiaSession();

/* ---< State by process. >--- */
SampleSceneNo SceneMatchmake();
SampleSceneNo SceneCommunication();
SampleSceneNo SceneMeshConnectionError();
SampleSceneNo SceneNetworkError();

/* ---< Definitions of structures used in the sample. >--- */

struct RectPosition  //!< Stores rectangle position coordinates.
{
    u16 x;
    u16 y;
};

struct SampleStationStatus
{
    nn::pia::StationId stationId;
    RectPosition pos;
    u32 latestIncrementNumber;
    s32 rtt;
    nn::pia::StationIndex routeInfoIndex;

    void Serialize(void* pBuf)
    {
        PIA_SAMPLE_ASSERT(pBuf != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pBuf) % 4 == 0);

        u8* p = static_cast<u8*>(pBuf);

        *reinterpret_cast<u64*>(p) = nn::pia::common::ByteOrder::HostToNetwork64(stationId.GetValue());
        p += 8;
        *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.x);
        p += 2;
        *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.y);
        p += 2;
        *reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(latestIncrementNumber);
    }

    void Deserialize(const void* pData)
    {
        PIA_SAMPLE_ASSERT(pData != NULL);
        PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pData) % 4 == 0);

        const u8* p = static_cast<const u8*>(pData);

        stationId = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(p));
        p += 8;
        pos.x = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        pos.y = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
        p += 2;
        latestIncrementNumber = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
    }

    void UpdateInfo(bool enableRelay)
    {
        // Update RTT.
        nn::pia::transport::Station* pStationBuf = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
        if (PIA_IS_VALID_POINTER(pStationBuf))
        {
            // Gets the RTT from the station instance corresponding to <tt>StationId</tt>.
            rtt = pStationBuf->GetRtt();
        }
        else
        {
            // Set to an invalid value if you cannot get the station instance.
            rtt = nn::pia::transport::Station::INVALID_RTT;
        }

        // Gets the <tt>stationIndex</tt> using the function for debugging <tt>Transport</tt> to get the relay connection route information.
        nn::pia::StationIndex stationIndex;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
        // Sets an invalid value when getting the value fails.
        if (result.IsFailure())
        {
            routeInfoIndex = nn::pia::STATION_INDEX_INVALID;
        }

        // Update the connection route information.
        if (enableRelay)
        {
            // If the relay feature is enabled, get the connection route information using the function for session debugging.
            nn::Result r = nn::pia::session::Session::GetInstance()->GetConnectionRoute(stationIndex, &routeInfoIndex);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
        else
        {
            // If the relay feature is disabled, the direct route is the only route.
            routeInfoIndex = stationIndex;
        }
    }
};


// In the scene where communication occurs with another party in the session, this is the setting for the interval [frame] to send the local host information.
// When set to 0, it sends in every game frame.
const u32 SAMPLE_INTERVAL_SENDING_NODE_STATUS = 0;

const u8 SAMPLE_MAX_ENTRY = 5;    //!< Maximum number of connections. In an actual application, set according to the application's specifications.
const u8 GATHERING_LIST_MAX = 5;  //!< Maximum number of found networks to show.

const u32 SAMPLE_INITIAL_RATING_VALUE = 1500;
const u8 SAMPLE_INITIAL_SELECTION_PRIORITY = 0;

const u32 SAMPLE_MIN_SCORE_SETTING_INDEX = 2;

const bool ENABLE_HOST_MIGRATION = false;  //!< Enables or disables the host migration feature. Fixed to enabled when the <tt>PiaSession</tt> host migration feature is used.

/* Thread Priority Settings */
#if NN_PIA_CTR
const s32 SEND_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;          //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#elif NN_PIA_CAFE
const s32 SEND_THREAD_PRIORITY = OS_PRIORITY_APP_DEFAULT - 1;  //Priority of the transport send thread. A priority higher than the main processing is desirable.
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY - 1;  //Priority of the transport send thread. Sets the priority high so that opportunities are not lost.
#endif

/* NEX-related settings. */
// Condition group used when building or searching for matchmaking sessions on a NEX game server.
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 35500;  // Game mode.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;
const u32 MATCHMAKE_ATTR_INDEX = 0;     //!< Index of the attribute to use.
const u32 MATCHMAKE_ATTR_VAL = 72;      //!< Sample value set as the attribute to use.

/* Memory-related settings */
const u32 MEMORY_BLOCK_SIZE_PIA = 2 * 1024 * 1024;      //!< Defines the amount of memory allocated to Pia.
const u32 MEMORY_BLOCK_SIZE_NEX = 1024 * 1024;          //!< Specifies the memory size provided to NEX.
const u32 MEMORY_BLOCK_SIZE_NEX_EMERGENCY = 50 * 1024;  //!< Specifies the memory size provided to NEX for emergencies.

/* Drawing-related settings */
const size_t SAMPLE_USER_NAME_LENGTH = 11;  //!< The maximum number of characters for <tt>UserName</tt> drawing, including the terminating character.

// Function for setting the <tt>Result</tt> used to indicate Internet-related errors.
void SetNetworkErrorResult(nn::Result result);

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel();

const int SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT = 102;
const int SAMPLE_PROGRAMPARAM_MATCHMAKE_SELECT_METHOD = 105;
const int SAMPLE_PROGRAMPARAM_SCORE_BASED_PARAM = 106;
