﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneCommunication.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_CriticalSection.h>
#include <pia/common/common_ScopedLock.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/inet.h>
#include <pia/transport.h>
#include <pia/session.h>

#if NN_PIA_CAFE
#include <nn/act.h>
#endif

#include "Shared_Basic_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"
#include "SampleLib/SampleLib.h"
#include "SampleConfig.h"

#include <nex.h>
/*!

    @brief  This CPP file contains processing that is independent of the network type (<tt>session</tt> + <tt>transport</tt>).

*/

extern u8 s_GameModeOption;  // Option value for changing <tt>MATCHMAKE_GAME_MODE/SAMPLE_SUB_ID</tt>.

namespace
{
SampleStationStatus s_StationStatus[SAMPLE_MAX_ENTRY];

u32 s_UnreliableProtocolHandle = 0;

const nn::pia::sampleframework::Color COLOR_OF_RECTANGLE[12] =
    {
        nn::pia::sampleframework::COLOR_FUCHSIA,
        nn::pia::sampleframework::COLOR_GRAY,
        nn::pia::sampleframework::COLOR_WHITE,
        nn::pia::sampleframework::COLOR_MAROON,
        nn::pia::sampleframework::COLOR_RED,
        nn::pia::sampleframework::COLOR_OLIVE,
        nn::pia::sampleframework::COLOR_YELLOW,
        nn::pia::sampleframework::COLOR_GREEN,
        nn::pia::sampleframework::COLOR_LIME,
        nn::pia::sampleframework::COLOR_TEAL,
        nn::pia::sampleframework::COLOR_NAVY,
        nn::pia::sampleframework::COLOR_PURPLE
    };

//The string for displaying the role screen within a session.
const char* SESSION_ROLE[] = {
    "Unknown",
    "Host",
    "Client"
};

#if ENABLE_MATCHMAKE_ANALYSE
bit8 s_TokenLock[nn::pia::transport::Station::IDENTIFICATION_TOKEN_MAX_DATA_SIZE] = "Lock";
bit8 s_TokenUnlock[nn::pia::transport::Station::IDENTIFICATION_TOKEN_MAX_DATA_SIZE] = "Unlock";
bool s_IsApproved = true;
#endif
}  // end of anonymous namespace

void FinalizePiaSession();  //Handles cleanup of the PiaSession class (forward declaration).

void ActivateStationStatus(const nn::pia::StationId stationId)
{
    nn::pia::StationIndex index;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&index, stationId);
    NN_PIA_DUMMY_PARAM(result);
    PIA_SAMPLE_ASSERT(result.IsSuccess());

    s_StationStatus[index].stationId = stationId;
    s_StationStatus[index].pos.x = 0;
    s_StationStatus[index].pos.y = 0;
    s_StationStatus[index].latestIncrementNumber = 0;
    s_StationStatus[index].rtt = nn::pia::transport::Station::INVALID_RTT;
    s_StationStatus[index].routeInfoIndex = nn::pia::STATION_INDEX_INVALID;
}

void DeactivateStationStatus(const nn::pia::StationId stationId)
{
    nn::pia::StationIndex index;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&index, stationId);
    NN_PIA_DUMMY_PARAM(result);
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    PIA_SAMPLE_ASSERT(index < SAMPLE_MAX_ENTRY);  // Has a leave event come from an unconnected peer?

    s_StationStatus[index].stationId = nn::pia::STATION_ID_INVALID;  //Only shows whether it is valid.
    s_StationStatus[index].pos.x = 0;
    s_StationStatus[index].pos.y = 0;
    s_StationStatus[index].latestIncrementNumber = 0;
    s_StationStatus[index].rtt = nn::pia::transport::Station::INVALID_RTT;
    s_StationStatus[index].routeInfoIndex = nn::pia::STATION_INDEX_INVALID;
}

void sampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, const nn::pia::StationId stationId)
{
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG("Sample Event JOIN detect.\n");
        ActivateStationStatus(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect.\n");
        DeactivateStationStatus(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
    {
        PIA_SAMPLE_LOG("Sample Event Host Change detect.\n");
    }
    break;
    case nn::pia::session::Session::EVENT_HOST_MIGRATION_FAILED:
    {
        PIA_SAMPLE_LOG("Sample Event Host Migration Failed detect.\n");
        // This error is handled by the <tt>Session::CheckConnectionError</tt> function, so just output it to the log.
    }
    break;
    }
}

#if ENABLE_MATCHMAKE_ANALYSE
bool joinApprovalCallback(const nn::pia::transport::Station::IdentificationToken* pRequesterToken)
{
    bit8 header = 'L';
    if( *(pRequesterToken->data) == header)
    {
        return false;
    }

    return true;
}
#endif

void InitializePiaSession()
{
    PIA_SAMPLE_LOG("InitializePiaTransport\n");

    nn::Result result;

#if PIA_TEST_CODE
    // To specify the debugging feature (packet loss emulation)
    // Make sure that this does not get called in the retail version ROM.
    nn::pia::transport::Transport::DebugSetting transportDebugSetting;
    transportDebugSetting.bPacketLossEmulation = true;
    nn::pia::transport::Transport::SetDebugSetting(transportDebugSetting);
#endif

    // These parameters are borrowed from the transport demo.
    // For more information about the parameters, see the Programming Manual.
    nn::pia::transport::Transport::Setting transportSetting;

    transportSetting.pFactory = Inet_GetNetworkFactory();
    transportSetting.maxStationNum = SAMPLE_MAX_ENTRY;
    transportSetting.maxSendThreadBufferNum = 16;
    transportSetting.maxReceiveThreadBufferNum = 32;

    result = nn::pia::transport::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::Transport::CreateInstance(transportSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //The port number is hard-coded to 0.
    s_UnreliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::UnreliableProtocol>(0);
    result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

#if NN_PIA_CTR || NN_PIA_CAFE
    //Priority settings for the send/receive thread. (For information about priorities, see SampleConfig.h.)
    //The priority of the send/receive thread does not necessarily need to be changed, but in this sample, the priority has been changed to show the change method.
    //Priority has been changed.
    nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(pThreadStreamManager);  //If the transport initialization has finished, the instance has probably been created.
    pThreadStreamManager->SetSendThreadPriority(SEND_THREAD_PRIORITY);
    pThreadStreamManager->SetReceiveThreadPriority(RECEIVE_THREAD_PRIORITY);
#endif

    PIA_SAMPLE_LOG("InitializePiaSession\n");

#if ENABLE_MATCHMAKE_ANALYSE
    {
        // To specify the debugging feature (the matchmaking analysis feature)
        // Make sure that this does not get called in the retail version ROM.
        nn::pia::session::Session::DebugSetting debugSetting;
        debugSetting.isEnableMatchmakeAnalysis = true;
        result = nn::pia::session::Session::SetDebugSetting(debugSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }
#endif

    nn::pia::session::Session::Setting sessionSetting;

    sessionSetting.pNetworkFactory = Inet_GetNetworkFactory();

    // Sets the network topology to full mesh.
    sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;

#if ENABLE_BIT_RATE_CHECK
    // Measure the bandwidth when the session is established.
    sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Enable;
#else
    sessionSetting.bitRateCheckMode = nn::pia::session::BitRateCheck_Disable;
#endif

    result = nn::pia::session::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::Session::CreateInstance(sessionSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //Registers the event callback.
    //Before callback registration, initialize the structure that the callback accesses.
    for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
    {
        s_StationStatus[i].stationId = nn::pia::STATION_ID_INVALID;  //Only shows whether it is valid.
        s_StationStatus[i].pos.x = 0;
        s_StationStatus[i].pos.y = 0;
        s_StationStatus[i].latestIncrementNumber = 0;
        s_StationStatus[i].rtt = nn::pia::transport::Station::INVALID_RTT;
        s_StationStatus[i].routeInfoIndex = nn::pia::STATION_INDEX_INVALID;
    }

    // Registers the session event callback.
    nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(sampleSessionCallback);

#if ENABLE_MATCHMAKE_ANALYSE
    nn::pia::session::Session::GetInstance()->RegisterJoiningApprovalCallback(joinApprovalCallback);
#endif
}

/*!
    @brief  Handles the state in which communication is performed using a <tt>PiaTransport</tt> instance.
    
    When a host shuts down a network, it enters the select network state, but clients enter the error state.
    
    @return  Returns the number of the next state (scene) to execute.
 */
SampleSceneNo SceneCommunication()
{
    u32 frameNo = 0;
    bool isExit = false;
    bool isOpenParticipation = true;  // In this sample, the default value is <tt>true</tt> because created sessions are open to participation.
    bool isWaitOtherStation = false;  // Stores whether OpenSession and CloseSession are waiting for other stations.

    nn::pia::transport::StationConnectionInfo localInfo;
    nn::pia::transport::StationConnectionInfo masterInfo;

    nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
    nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&masterInfo);

    SampleSceneNo nextSceneNo = SCENE_MATCH_MAKE;

    bool isAsyncProcessInProgress = false;

    u8 priority = SAMPLE_INITIAL_SELECTION_PRIORITY;
    u32 matchmakeAttribute = MATCHMAKE_ATTR_VAL;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_COMMUNICATE,
        true,
    nn::pia::session::Session::GetInstance()->IsHost(),
    nn::pia::session::Session::GetInstance()->GetStationNum());

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Checks the PiaSession asynchronous processing.
        if(isAsyncProcessInProgress)
        {
#if ENABLE_MATCHMAKE_ANALYSE
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenParticipation())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetOpenParticipationResult();
                if (result.IsSuccess() && !isOpenParticipation)
                {
                    PIA_SAMPLE_LOG("Open participation complete and succeeded.\n");
                    // Add the flag approving/denying participation and displays on screen.
                    isOpenParticipation = true;
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedCloseParticipation())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetCloseParticipationResult();
                if (result.IsSuccess() && isOpenParticipation)
                {
                    PIA_SAMPLE_LOG("Close participation complete and succeeded.\n");
                    // Add the flag approving/denying participation and displays on screen.
                    isOpenParticipation = false;
                }
                isAsyncProcessInProgress = false;
            }
#else
            if (nn::pia::session::Session::GetInstance()->IsCompletedOpenSession())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetOpenSessionResult();
                if (result.IsSuccess() && !isOpenParticipation)
                {
                    PIA_SAMPLE_LOG("Open session complete and succeeded.\n");
                    // Add the flag approving/denying participation and displays on screen.
                    isOpenParticipation = true;
                }
                isWaitOtherStation = false;
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedCloseSession())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetCloseSessionResult();
                if (result.IsSuccess() && isOpenParticipation)
                {
                    PIA_SAMPLE_LOG("Close session complete and succeeded.\n");
                    // Add the flag approving/denying participation and displays on screen.
                    isOpenParticipation = false;
                }
                isWaitOtherStation = false;
                isAsyncProcessInProgress = false;
            }
#endif
            if (nn::pia::session::Session::GetInstance()->IsCompletedModifyAttribute())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetModifyAttributeResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("ModifyAttribute Complete and succeeded.\n");
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedUpdateSessionSetting())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetUpdateSessionSettingResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("UpdateSessionSetting Complete and succeeded.\n");
                }
                isAsyncProcessInProgress = false;
            }
            if (nn::pia::session::Session::GetInstance()->IsCompletedUpdateApplicationData())
            {
                nn::Result result = nn::pia::session::Session::GetInstance()->GetUpdateApplicationDataResult();
                if (result.IsSuccess())
                {
                    PIA_SAMPLE_LOG("UpdateApplicationData Complete and succeeded.\n");
                }
                isAsyncProcessInProgress = false;
            }            
        }
        else
        {
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();

            if(result.IsFailure())
            {
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("Session::CheckConnection: Disconnected NETWORK\n");
                    nextSceneNo = SCENE_NETWORK_ERROR;
                    SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                    SetNetworkErrorResult(result);
                }
                else if (result == nn::pia::ResultSessionConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("Session::CheckConnection: Disconnected SESSION\n");

                    SampleSessionAsyncStatus status = Session_LeaveSession();
                    if (status == SAMPLE_SESSION_ASYNC_SUCCESS || status == SAMPLE_SESSION_ASYNC_DISCONNECTED_SESSION)
                    {
                        nextSceneNo = SCENE_ERROR;
                        SetCommunicationErrorType(SESSION_LOST_ERROR);
                        SetNetworkErrorResult(result);
                    }
                    else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
                    {
                        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                        nextSceneNo = SCENE_NETWORK_ERROR;
                    }
                }
                isExit = true;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_COMMUNICATE, false);
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE,
                true,
                nn::pia::session::Session::GetInstance()->IsHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum());

                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
                SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
                nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());
            }

        }

        //Draw to the lower screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            //String.
            {
                int line = 0;

                {
                    for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
                    {
                        const u8 DRAW_BASE = line;

                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(10, DRAW_BASE + (10 * i), 8, 8, COLOR_OF_RECTANGLE[i]);
                        if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "Idx=%02X, %05X, RTT:%4d", i, s_StationStatus[i].latestIncrementNumber, s_StationStatus[i].rtt);
                        }
                        else
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(20, DRAW_BASE + (10 * i), "No Connection");
                        }
                    }
                }

                line = BASIC_LOWER_SCREEN_HEIGHT - (BASIC_FONT_SIZE * 12);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Network Information ]");
                line += BASIC_FONT_SIZE;
                //nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  NetworkTopology: %s", GetNetworkTopologyTypeString(s_TargetNetworkTopology)); line +=BASIC_FONT_SIZE;
                //nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  GameMode option: %02d", s_GameModeOption); line +=BASIC_FONT_SIZE;

                //Screen display of communication-related data (For debugging.).
                nn::pia::transport::Transport::GetInstance()->GetLocalStationConnectionInfo(&localInfo);
                nn::pia::transport::Transport::GetInstance()->GetHostStationConnectionInfo(&masterInfo);

                nn::pia::common::String mainStr;
                nn::pia::common::String subStr;
                localInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                localInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Local  Main:%s", mainStr.GetCharArray());
                line += BASIC_FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "          Sub:%s", subStr.GetCharArray());
                line += BASIC_FONT_SIZE;
                masterInfo.GetMainLocation().GetStationAddress().GetInetAddress().GetAddressString(&mainStr);
                masterInfo.GetSubLocation().GetStationAddress().GetInetAddress().GetAddressString(&subStr);
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Host Main:%s", mainStr.GetCharArray());
                line += BASIC_FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "          Sub:%s", subStr.GetCharArray());
                line += BASIC_FONT_SIZE;

                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
                line += BASIC_FONT_SIZE;

                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: LeaveSession");
                line += BASIC_FONT_SIZE;
#if !ENABLE_MATCHMAKE_ANALYSE
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: %s %s", isOpenParticipation ? "CloseSession":"OpenSession", isWaitOtherStation ? "(Wait Other Station)" : "");
                line += BASIC_FONT_SIZE;
#endif
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Y: Change Matchmake Selection Priority");
                    line += BASIC_FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  X: Modify Attribute[%d] Value", MATCHMAKE_ATTR_INDEX);
                    line += BASIC_FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  RIGHT+L: UpdateSessionSetting");
                    line += BASIC_FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  LEFT +L: UpdateApplicationData");
                    line += BASIC_FONT_SIZE;
#if ENABLE_MATCHMAKE_ANALYSE
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: %s", isOpenParticipation ? "CloseParticipation":"OpenParticipation");
                    line += BASIC_FONT_SIZE;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  START: ApprovalCallback(%s)", s_IsApproved ? "Set Deny" : "Set approval");
                    line += BASIC_FONT_SIZE;
#endif
                }
                //nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R+B/R+X: BackToMatchmake(OnlyInternet)"); line +=BASIC_FONT_SIZE;


                nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
            }
        }
        //Draw to the upper screen.
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            {
                for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
                {
                    if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
                    {
                        s_StationStatus[i].UpdateInfo(false);

                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(s_StationStatus[i].pos.x, s_StationStatus[i].pos.y, 8, 8, COLOR_OF_RECTANGLE[i]);
                        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(s_StationStatus[i].pos.x + 10, s_StationStatus[i].pos.y, "<-ID:%02X", i);
                    }
                }
            }

            int line = BASIC_UPPER_SCREEN_HEIGHT - BASIC_FONT_SIZE * 8;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", Basic_GetPiaVersionString());
            line += BASIC_FONT_SIZE;

            //Check the role of the local station.
            int roleIndex = 0;  //Unknown;

            if (nn::pia::session::Session::GetInstance()->GetLocalStationId() != nn::pia::STATION_ID_INVALID)
            {
                roleIndex = (nn::pia::session::Session::GetInstance()->IsHost() ? 1 : 2);  // 1:Host, 2:Client
            }

            // Matchmaking information.
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Communication");
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  StationNum %d   Role=%s", nn::pia::session::Session::GetInstance()->GetStationNum(), SESSION_ROLE[roleIndex]);
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  SessionID %d", nn::pia::session::Session::GetInstance()->GetSessionId());
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Participation=%s", isOpenParticipation ? "open" : "close");
            line += BASIC_FONT_SIZE;
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Matchmake Selection Priority=%d", priority);
                line += BASIC_FONT_SIZE;
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Attribute[%d] Value:%d", MATCHMAKE_ATTR_INDEX, matchmakeAttribute);
                line += BASIC_FONT_SIZE;
            }
            nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        }

        // Execute Dispatch.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        if (nn::nex::Scheduler::GetInstance())
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        //Receive.
        // Executed after Dispatch (first time) to facilitate receiving data with the Scheduler::Dispatch function.
        {
            // Because the size of the data the other party sends generally cannot be anticipated, allocate a receive buffer the size of MAX_DATA_SIZE of UnreliableProtocol.
            // 
            PIA_ATTRIBUTE_ALIGN(4) static bit8 receiveBuffer[nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE];

            nn::Result result;
            //Because the number of received data is unknown, the receive process repeats until the buffer is empty.
            do
            {
                nn::pia::StationId sourceStationId;
                size_t receivedDataSize = 0;

                result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Receive(&sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));

                if (result.IsSuccess())
                {
                    if (sourceStationId != nn::pia::STATION_ID_INVALID)
                    {
                        //You may need to check whether the received data is correct.
                        nn::pia::StationIndex sourceStatinIndex;
                        result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&sourceStatinIndex, sourceStationId);
                        if (result.IsSuccess() && s_StationStatus[sourceStatinIndex].stationId != nn::pia::STATION_ID_INVALID)
                        {
                            s_StationStatus[sourceStatinIndex].Deserialize(&receiveBuffer);
                        }
                    }
                }
                else if (result == nn::pia::ResultNotInCommunication())
                {
                    //Failed to receive data due to a disconnection.
                    //This failure can occur if an asynchronous session delete/leave process completes due to an immediately preceding Dispatch process.
                    //If the status is normal, the session status check shows DISCONNECTED when the next game frame starts.
                    PIA_SAMPLE_LOG("UnreliableProtocol::Receive failed. (IsNotInCommunication)\n");
                    break;
                }
                else if (result != nn::pia::ResultNoData())
                {
                    //Other reception failed.
                    //In this demo, this is not an issue.
                }
            } while (result != nn::pia::ResultNoData());
        }
        int move_x = 0;  //Amount of movement of the rectangle on the local station (x).
        int move_y = 0;  //Amount of movement of the rectangle on the local station (y).

        //Move the rectangle.
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
            {
                move_x = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                move_x = 1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_UP))
            {
                move_y = -1;
            }
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
            {
                move_y = 1;
            }
        }

        //Locked send process.
        // Executed before Dispatch (second time) for the convenience of having send processing performed by the Scheduler::Dispatch function.
        {
            nn::pia::StationId myStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
            nn::pia::StationIndex myStationIndex;
            nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&myStationIndex, myStationId);

            // If the <tt>ConvertToStationIndex</tt> function fails, the local station has been disconnected from the session already, so do not bother sending.
            if (result.IsSuccess())
            {
                //Update the data of the local station.
                if (move_x != 0)
                {
                    s_StationStatus[myStationIndex].pos.x = (s_StationStatus[myStationIndex].pos.x + BASIC_UPPER_SCREEN_WIDTH + move_x) % BASIC_UPPER_SCREEN_WIDTH;
                }
                else
                {
                    //Move the rectangle even when there is no input to indicate that stations are communicating.
                    s_StationStatus[myStationIndex].pos.x = (s_StationStatus[myStationIndex].pos.x + BASIC_UPPER_SCREEN_WIDTH + 1) % BASIC_UPPER_SCREEN_WIDTH;
                }
                if (move_y != 0)
                {
                    s_StationStatus[myStationIndex].pos.y = (s_StationStatus[myStationIndex].pos.y + BASIC_UPPER_SCREEN_HEIGHT + move_y) % BASIC_UPPER_SCREEN_HEIGHT;
                }
                s_StationStatus[myStationIndex].latestIncrementNumber = frameNo;

                if ((frameNo % (SAMPLE_INTERVAL_SENDING_NODE_STATUS + 1)) == 0)
                {
#if 1
                    u16 stationNum = nn::pia::session::Session::GetInstance()->GetStationNum();
                    if (stationNum > 1)
                    {
                        // Use SendToAll to send as a batch process.
                        SampleStationStatus sendData;
                        s_StationStatus[myStationIndex].Serialize(&sendData);
                        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->SendToAll(reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                        {
                            //Send failure.
                            //In this demo, this is not an issue.
                            //PIA_SAMPLE_LOG("Send Fail\n");
                        }
                    }
#else
                    // Send only to valid parties using Send.
                    for (int i = 0; i < SAMPLE_MAX_ENTRY; i++)
                    {
                        // Send to valid <tt>StationIds</tt> other than the local station.
                        if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID && s_StationStatus[i].stationId != myStationId)
                        {
                            SampleStationStatus sendData;
                            s_StationStatus[myStationIndex].Serialize(&sendData);
                            if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Send(s_StationStatus[i].stationId, reinterpret_cast<bit8*>(&sendData), sizeof(SampleStationStatus)).IsFailure())
                            {
                                //Send failure.
                                //In this demo, this is not an issue.
                                //PIA_SAMPLE_LOG("Send failed. Dest StationIndex:%d\n", index);
                            }
                        }
                    }
#endif
                }
            }
        }

        // Execute Dispatch (second time: for send processing).
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        // Session connection status check.
        {
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();

            if (result.IsFailure())
            {
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("Session::CheckConnection: Disconnected NETWORK\n");
                    nextSceneNo = SCENE_NETWORK_ERROR;
                    SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                    SetNetworkErrorResult(result);
                }
                else if (result == nn::pia::ResultSessionConnectionIsLost())
                {
                    PIA_SAMPLE_LOG("Session::CheckConnection: Disconnected SESSION\n");

                    SampleSessionAsyncStatus status = Session_LeaveSession();
                    if (status == SAMPLE_SESSION_ASYNC_SUCCESS || status == SAMPLE_SESSION_ASYNC_DISCONNECTED_SESSION)
                    {
                        nextSceneNo = SCENE_ERROR;
                        SetCommunicationErrorType(SESSION_LOST_ERROR);
                        SetNetworkErrorResult(result);
                    }
                    else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
                    {
                        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                        nextSceneNo = SCENE_NETWORK_ERROR;
                    }
                }
                isExit = true;
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_COMMUNICATE, false);
            }
            else
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
                SCENE_COMMUNICATE,
                true,
                nn::pia::session::Session::GetInstance()->IsHost(),
                nn::pia::session::Session::GetInstance()->GetStationNum());

                nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
                SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
                nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());
            }
        }


        //Handle user input.
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                if (!isAsyncProcessInProgress)
                {
                    SampleSessionAsyncStatus status = Session_LeaveSession();
                    if (status == SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        isExit = true;
                        isAsyncProcessInProgress = false;
                    }
                    else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_SESSION)
                    {
                        isExit = true;
                    }
                    else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
                    {
                        SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                        isExit = true;
                        nextSceneNo = SCENE_NETWORK_ERROR;
                    }
                }
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                u8 temporaryPriority = 1 - priority;
                if (nn::pia::session::Session::GetInstance()->IsUpdatableSelectionPriority(temporaryPriority))
                {
                    priority = temporaryPriority;
                    nn::pia::session::Session::GetInstance()->UpdateSelectionPriority(priority);
                }
            }

#if ENABLE_MATCHMAKE_ANALYSE
           if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
            {
                nn::pia::transport::Station::IdentificationToken token;
                bit8* pTokenData;
                if(s_IsApproved)
                {
                    s_IsApproved = false;
                    pTokenData = s_TokenLock;
                }
                else
                {
                    s_IsApproved = true;
                    pTokenData = s_TokenUnlock;
                }
                ::std::memcpy(token.data, pTokenData, nn::pia::transport::Station::IDENTIFICATION_TOKEN_MAX_DATA_SIZE);

                nn::pia::session::Session::GetInstance()->SetMatchmakeAnalysisToken(&token);
            }

           if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    if (!isAsyncProcessInProgress)
                    {
                        nn::Result result;
                        if(isOpenParticipation)
                        {
                            result = nn::pia::session::Session::GetInstance()->CloseParticipationAsync();
                        }
                        else
                        {
                            result = nn::pia::session::Session::GetInstance()->OpenParticipationAsync();
                        }
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("%s failed.\n", isOpenParticipation ? "CloseParticipation":"OpenParticipation");
                            PIA_SAMPLE_LOG_RESULT(result);
                            isAsyncProcessInProgress = false;
                        }
                    }
                }
            }
#else
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
            {
                if (!isAsyncProcessInProgress)
                {
                    nn::Result result;
                    if(isOpenParticipation)
                    {
                        result = nn::pia::session::Session::GetInstance()->CloseSessionAsync();
                    }
                    else
                    {
                        result = nn::pia::session::Session::GetInstance()->OpenSessionAsync();
                    }
                    isAsyncProcessInProgress = true;
                    isWaitOtherStation = true;
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("%s failed.\n", isOpenParticipation ? "CloseSession":"OpenSession");
                        PIA_SAMPLE_LOG_RESULT(result);
                        isAsyncProcessInProgress = false;
                        isWaitOtherStation = false;
                    }
                }
            }
#endif
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
            {
                if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    if (nn::pia::session::Session::GetInstance()->IsHost() && !isAsyncProcessInProgress)
                    {
                        // When only some of the parameters are updated, there is no need to set the unnecessary ones.
                        nn::pia::inet::NexUpdateSessionSetting updateSetting;
                        matchmakeAttribute = 100 - matchmakeAttribute;
                        u32 attr[nn::pia::inet::ATTRIBUTE_MAX_SIZE];
                        attr[MATCHMAKE_ATTR_INDEX] = matchmakeAttribute;
                        for (int i=0; i < nn::pia::inet::ATTRIBUTE_MAX_SIZE;i++)
                        {
                            if(i != MATCHMAKE_ATTR_INDEX)
                            {
                                attr[i] = 111 * i;
                            }
                        }
                        updateSetting.SetAttributes(attr);
                        const wchar_t description[] = L"newdescription";
                        updateSetting.SetDescription(description);

                        const wchar_t pass[] = L"newpassword";
                        updateSetting.SetSessionUserPassword(pass, 12);
                        updateSetting.SetOpenParticipation(false);
                        updateSetting.SetSelectionPriority(3);
                        updateSetting.SetSessionType(nn::pia::inet::SESSION_TYPE_ANYBODY);
                        updateSetting.SetMaxParticipants(8);
                        updateSetting.SetMinParticipants(2);

                        const u32 applicationDataSize = 16;
                        u8 applicationData[applicationDataSize];

                        for(int i=0;i<applicationDataSize;i++)
                        {
                            applicationData[i] = 100 + i;
                        }
                        updateSetting.SetApplicationData(applicationData, sizeof(applicationData));

                        const u32 applicationDataVerificationBufSize = 32;
                        u8 applicationDataVerificationBuf[applicationDataVerificationBufSize];
                        size_t dataSize = 0;
                        updateSetting.GetApplicationData(applicationDataVerificationBuf, &dataSize, applicationDataVerificationBufSize);

                        // Update parameters for scored matchmaking
#if NN_PIA_CTR
                        nn::cfg::CTR::Initialize();
                        nn::cfg::CTR::SimpleAddressId addrId;
                        nn::cfg::CTR::GetSimpleAddressId(&addrId);
                        u32 cfgCountryCode = static_cast<u32>(addrId.GetCountryCode());
                        nn::cfg::CTR::Finalize();
#elif NN_PIA_CAFE
                        nn::act::Initialize();
                        u32 addrId = nn::act::GetSimpleAddressId();
                        u32 cfgCountryCode = 0;
                        if (addrId != ACT_INVALID_SIMPLE_ADDRESS_ID)
                        {
                            cfgCountryCode = static_cast<u8>((addrId & 0xff000000) >> 24);
                        }
                        nn::act::Finalize();
#else
                        u32 cfgCountryCode = 0;
#endif
                        updateSetting.SetViolationRate(1);
                        updateSetting.SetDisconnectionRate(1);
                        updateSetting.SetRatingValue(2015);
                        updateSetting.SetNintendoCountryCode(cfgCountryCode);
                        updateSetting.SetUpdateGeoIp(true);
                        
                        NN_PIA_DUMMY_PARAM(applicationDataVerificationBuf);
                        NN_PIA_DUMMY_PARAM(dataSize);
                        PIA_SAMPLE_LOG("--- Confirm ApplicationData ----\n");
                        for (int i=0;i < dataSize; i++)
                        {
                            PIA_SAMPLE_LOG("%d,", applicationDataVerificationBuf[i]);
                        }
                        PIA_SAMPLE_LOG("\n--- Confirm ApplicationData ----\n");
                        
                        nn::Result result = nn::pia::session::Session::GetInstance()->UpdateSessionSettingAsync(&updateSetting);
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("UpdateSessionSettingAsync failed.\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                            isAsyncProcessInProgress = false;
                        }
                    }
                }
                else if(nn::pia::sampleframework::SampleFramework::GetInstance()->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
                {
                    if (nn::pia::session::Session::GetInstance()->IsHost() && !isAsyncProcessInProgress)
                    {
                        const u32 applicationDataSize = 16;
                        u8 applicationData[applicationDataSize];

                        for(int i=0;i<applicationDataSize;i++)
                        {
                            applicationData[i] = 100 + i;
                        }

                        nn::Result result = nn::pia::session::Session::GetInstance()->UpdateApplicationDataAsync(applicationData, sizeof(applicationData));
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("UpdateApplicationDataAsync failed.\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                            isAsyncProcessInProgress = false;
                        }
                    }
                }
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    if (!isAsyncProcessInProgress)
                    {
                        matchmakeAttribute = 100 - matchmakeAttribute;
                        nn::Result result = nn::pia::session::Session::GetInstance()->ModifyAttributeAsync(MATCHMAKE_ATTR_INDEX, matchmakeAttribute);
                        isAsyncProcessInProgress = true;
                        if (result.IsFailure())
                        {
                            PIA_SAMPLE_LOG("ModifyAttributeAsync failed.\n");
                            PIA_SAMPLE_LOG_RESULT(result);
                            isAsyncProcessInProgress = false;
                        }

                    }
                }
            }

#if PIA_TEST_CODE
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_SAMPLE_ASSERT(pThreadStreamManager);  //If the transport initialization has finished, the instance has probably been created.

                s32 ratio = 100;
                pThreadStreamManager->SetSendThreadPacketLossRatio(ratio);
                pThreadStreamManager->SetReceiveThreadPacketLossRatio(ratio);
                PIA_SAMPLE_LOG("Packet dropped emulation ratio (send) = %d\n", ratio);
                PIA_SAMPLE_LOG("Packet dropped emulation ratio (recv) = %d\n", ratio);
            }
            else if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_SAMPLE_ASSERT(pThreadStreamManager);  //If the transport initialization has finished, the instance has probably been created.

                s32 ratio = 0;
                pThreadStreamManager->SetSendThreadPacketLossRatio(ratio);
                pThreadStreamManager->SetReceiveThreadPacketLossRatio(ratio);
                PIA_SAMPLE_LOG("Packet dropped emulation ratio (send) = %d\n", ratio);
                PIA_SAMPLE_LOG("Packet dropped emulation ratio (recv) = %d\n", ratio);
            }
#endif
        }

        frameNo++;
    }

    // Clears the rendering to the upper screen.
    nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

    // Set the number of participants.
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(
    SAMPLE_PROGRAMPARAM_MATCHMAKESESSION_PARTICAPANT,
    nn::pia::session::Session::GetInstance()->GetMatchmakeSessionParticipantNum());

    // clean up
    CleanupPiaSession();

    return nextSceneNo;
}

void CleanupPiaSession()
{
    Session_CleanupSession();
}

//Prevent problems from occurring even if finalized recklessly.
void FinalizePiaSession()
{
    // Finalize nn::pia::session.
    PIA_SAMPLE_LOG("FinalizePiaSession\n");
    Session_FinalizePiaSession();

    PIA_SAMPLE_LOG("FinalizePiaTransport\n");
    if (nn::pia::transport::Transport::GetInstance() != NULL)
    {
        //Finalize nn::pia::transport::UnreliableProtocol.
        if (nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle) != NULL)
        {
            nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Finalize();
            nn::pia::transport::Transport::GetInstance()->DestroyProtocol(s_UnreliableProtocolHandle);
            s_UnreliableProtocolHandle = 0;
        }
    }

    //Finalize nn::pia::Transport.
    nn::pia::transport::Transport::DestroyInstance();  //The function to destroy the <tt>transport</tt> instance can be run without any problems, even if the instance has not been created.
    nn::pia::transport::Finalize();
}
