﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SceneError.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/util.h>

#include "SampleLib/SampleLib.h"

#include "Shared_Inet_Setup.h"
#include "Shared_Basic_Setup.h"
#include "SampleConfig.h"

namespace
{
SampleErrorType s_ErrorType = ERROR_NONE;  //Variable that holds the errors that occurred most recently.

const char* ERROR_INFO_STRING[] =
    {
        "Unknown Reason",
        "Initialize Error (wireless OFF)",
        "Initialize Error (other reason)",
        "Matchmake Failed",
        "Session Lost",
        "Connection Lost"
    };

nn::Result s_NetworkErrorResult = nn::ResultSuccess();
}

/*!
    @brief  This function sets the error reason before jumping to the error state.
 */
void SetCommunicationErrorType(SampleErrorType error)
{
    s_ErrorType = error;
}

// Pass the result of errors that occur when connecting to the Internet.
// Pass the result to the function for conversion to a network error code, and then display it.
// 
void SetNetworkErrorResult(nn::Result result)
{
    s_NetworkErrorResult = result;
}


nn::Result GetNetworkErrorResult(void)
{
    return s_NetworkErrorResult;
}

/*!
    @brief  This state displays a description for a recoverable error (such as a communication error).
    
    It only displays error information collected when the error occurred.
 */
SampleSceneNo SceneMeshConnectionError()
{
    nn::Result result;
    bool isExit = false;

    s32 networkErrorCode = nn::pia::util::NetworkErrorCodeConverter::Convert(GetNetworkErrorResult());

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_ERROR, false);

    int frameNo = 0;
    // Main loop.
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        //Handles user input. Pad information is collected by the BeginFrame function.
        {
            // A:    Restart.
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                isExit = true;
            }
        }

        // Renders to the screens.
        int line;
        {
            // Lower screen (standard).
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Communication Error Occurred.");
            line += BASIC_FONT_SIZE;
            line += BASIC_FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
            line += BASIC_FONT_SIZE;
            line += BASIC_FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Network error code: %07d", networkErrorCode);
            line += BASIC_FONT_SIZE;
            line = BASIC_LOWER_SCREEN_HEIGHT - (BASIC_FONT_SIZE * 2);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
            line += BASIC_FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        {
            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = BASIC_UPPER_SCREEN_HEIGHT - BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", Basic_GetPiaVersionString());
            line += BASIC_FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        // Execute Dispatch.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        if (nn::nex::Scheduler::GetInstance())
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        frameNo++;
    }
    //Clears the error information, just in case.
    s_ErrorType = ERROR_NONE;
    SetNetworkErrorResult(nn::ResultSuccess());
    return SCENE_MATCH_MAKE;
}


/*!
    @brief  This scene displays the network error code for errors occurring in the Internet connection environment.
    Processing transitions to this state when an error is generated that requires logging in again.
    In a normal program, pass the network error codes to the Error/EULA Viewer.
 */
SampleSceneNo SceneNetworkError()
{
    // Convert the error <tt>Result</tt> to a network error code.
    // Normally, you would pass this error code to the error/EULA applet.
    // However, the error/EULA applet cannot be used unless development hardware is set to <tt>prod</tt> mode.
    // For simplicity, this sample demo does not use the error/EULA applet.
    // Not used.
    nn::Result result = GetNetworkErrorResult();
    PIA_SAMPLE_LOG_RESULT(result);
    bool isPiaError = nn::pia::util::IsPiaResult(result);
    u32 networkErrorCode = (isPiaError) ? nn::pia::util::NetworkErrorCodeConverter::Convert(result) : Inet_GetNetworkErrorCode();

    bool isExit = false;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_NETWORK_ERROR, false);

    int frameNo = 0;
    // Rendering
    while (isExit == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Exit the loop if the A Button is pressed.
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            isExit = true;
        }

        int line;
        {
            // Render the lower screen
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
            nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);

            line = 0;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %08X", PROGRAM_NAME, frameNo);
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, " Scene: A Network Error Occurred.");
            line += BASIC_FONT_SIZE;
            line += BASIC_FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  %s", ERROR_INFO_STRING[s_ErrorType]);
            line += BASIC_FONT_SIZE;
            line += BASIC_FONT_SIZE;

            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Result: 0x%08X", GetNetworkErrorResult().GetPrintableBits());
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Level:       %4d", GetNetworkErrorResult().GetLevel());
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Summary:     %4d", GetNetworkErrorResult().GetSummary());
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Module:      %4d", GetNetworkErrorResult().GetModule());
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "   Description: %4d", GetNetworkErrorResult().GetDescription());
            line += BASIC_FONT_SIZE;
            line += BASIC_FONT_SIZE;

            if (networkErrorCode != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Network error code: %07d", networkErrorCode);
                line += BASIC_FONT_SIZE;
            }
            if (Inet_GetLoginNetworkErrorCode() != 0)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Login error code: %07d", Inet_GetLoginNetworkErrorCode());
                line += BASIC_FONT_SIZE;
            }

            line = BASIC_LOWER_SCREEN_HEIGHT - (BASIC_FONT_SIZE * 2);
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Exit and Restart.");
            line += BASIC_FONT_SIZE;
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();
        {
            // Upper screen.
            nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

            line = BASIC_UPPER_SCREEN_HEIGHT - BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", Basic_GetPiaVersionString());
            line += BASIC_FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        frameNo++;
    }

    //Clears the error information, just in case.
    s_ErrorType = ERROR_NONE;
    SetNetworkErrorResult(nn::ResultSuccess());

    return SCENE_RESET;
}
