/*---------------------------------------------------------------------------*
  Project: Pia
  File: SceneMatchmake.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/common/common_Event.h>
#include <pia/inet.h>
#include <pia/session.h>
#include <pia/util.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"

#include "SampleConfig.h"
#include "Shared_Basic_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"

#if NN_PIA_CAFE
#include <nn/act.h>
#endif

namespace
{
enum SampleSessionSelectMethod
{
    SAMPLE_SESSION_SELECT_METHOD_ROMDOM = 0,
    SAMPLE_SESSION_SELECT_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY,
    SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED,
    SAMPLE_SESSION_SELECT_METHOD_MAX
};

const char* SELECT_METHOD[] =
{
    "Random",
    "BroadenRange+Priority",
    "ScoreBased"
};

const u32 SAMPLE_RATING_MIN = 800;
const u32 SAMPLE_RATING_MAX = 3000;

enum SampleScoreSettintIndex
{
    SAMPLE_SCORE_INDEX_RATING = SAMPLE_MIN_SCORE_SETTING_INDEX,
    SAMPLE_SCORE_INDEX_DISCONNECTION_RATE,
    SAMPLE_SCORE_INDEX_VIOLATION_RATE,
    SAMPLE_SCORE_INDEX_GEO_IP,
    SAMPLE_SCORE_INDEX_COUNTRY_CODE,
    SAMPLE_SCORE_INDEX_SESSION_DURAION,
    SAMPLE_SCORE_INDEX_SELECTION_PRIORITY,
    SAMPLE_MAX_SCORE_SETTING_INDEX = SAMPLE_SCORE_INDEX_SELECTION_PRIORITY
};

const char* SCORE_SETTING[] =
{
    "Rating",
    "Disconnecion Rate",
    "Violation Rate",
    "Geo Ip",
    "Country Code",
    "Session Duration",
    "Selection Priority"
};

const u32 SAMPLE_GEO_IP_MAX = 9;

const char* GEO_IP_LOCATION[] =
{
    "Tokyo",
    "Osaka",
    "New York",
    "San Flancisco",
    "Vancouver",
    "London",
    "Paris",
    "Berlin",
    "Sydney",
};

#if !ENABLE_SERVERLESS
const wchar_t* GEO_IP_LIST[] =
{
    L"202.232.224.105", // Tokyo
    L"133.1.138.144",   // Osaka
    L"128.59.105.24",   // New York
    L"130.212.10.204",  // San Francisco
    L"137.82.130.49",   // Vancouver
    L"128.86.130.195",  // London
    L"193.55.96.121",   // Paris
    L"160.45.170.10",   // Berlin
    L"129.78.5.11",     // Sydney
};
#endif
}


SampleSceneNo SceneMatchmake()
{
    nn::Result result;

    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(
        SCENE_MATCH_MAKE,
        false,
        false,
        0);

    // Session startup.
    nn::pia::session::Session::StartupSetting startupSetting;
    startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;
    startupSetting.cryptoMode = nn::pia::common::CryptoSetting::MODE_NOTHING;   // Do not use encrypted communications.
    startupSetting.maxSilenceTime = 8000;                       // Sets the maximum time of no communications to 8000 ms.
    startupSetting.keepAliveSendingInterval = 1500;             // Sets the interval for sending keep-alive signals to 1500 ms.
    startupSetting.pToken = NULL;                               // Do not set an identification token.

#if ENABLE_BIT_RATE_CHECK
    startupSetting.uplinkBitRateLowerLimit = 200 * 1000;
    startupSetting.bitRateCheckPacketSize = 1024;
#endif

    result = Session_StartupSession(startupSetting);
    if (result.IsFailure())
    {
        // Report <tt>result</tt> for error scene.
        SetCommunicationErrorType(INITIALIZE_ERROR_OTHER_REASON);
        SetNetworkErrorResult(result);
        return SCENE_NETWORK_ERROR;
    }

    SampleSessionSelectMethod selectionMethodType = SAMPLE_SESSION_SELECT_METHOD_ROMDOM;
    u32 rating = SAMPLE_INITIAL_RATING_VALUE;

    u32 scoreSettingIndex = SAMPLE_MIN_SCORE_SETTING_INDEX;
    u32 disconnectionRate = 0;
    u32 violationRate = 0;
    u32 geoIpIndex = 0;
    u32 countryCode = 1;    // 1: Japan
    u8 selectionPriority = 0;

    u32 scoreBasicParam = 0;

    // Settings for the matchmaking session to create.
#if ENABLE_SERVERLESS
    nn::pia::inet::LanCreateSessionSetting createSessionSetting;
#else
    nn::pia::inet::NexCreateSessionSetting createSessionSetting;
#endif
    createSessionSetting.SetGameMode(MATCHMAKE_GAME_MODE);
    createSessionSetting.SetMaxParticipants(SAMPLE_MAX_ENTRY);
    createSessionSetting.SetAttribute(MATCHMAKE_ATTR_INDEX, MATCHMAKE_ATTR_VAL);
    createSessionSetting.SetOpenParticipation(true);  // Allows participation as soon as session is created.
    createSessionSetting.SetSelectionPriority(SAMPLE_INITIAL_SELECTION_PRIORITY);

    // Search condition settings
    const size_t searchSettingNum = nn::pia::inet::SEARCH_CRITERIA_MAX_NUM;
#if ENABLE_SERVERLESS
    nn::pia::inet::LanSessionSearchCriteria sessionSearchCriteriaList[searchSettingNum];
#else
    nn::pia::inet::NexSessionSearchCriteria sessionSearchCriteriaList[searchSettingNum];
#endif

    // Specifies a single attribute.
    sessionSearchCriteriaList[0].SetGameMode(MATCHMAKE_GAME_MODE);
    sessionSearchCriteriaList[0].SetMaxParticipants(SAMPLE_MAX_ENTRY);
    sessionSearchCriteriaList[0].SetAttribute(MATCHMAKE_ATTR_INDEX, MATCHMAKE_ATTR_VAL);

    // Specifies multiple attributes.
    // The code example is for specifying multiple attributes, but the search is made on the above-described specification for a single attribute, so this is not used for actual searching.
    sessionSearchCriteriaList[1].SetGameMode(MATCHMAKE_GAME_MODE);
    sessionSearchCriteriaList[1].SetMaxParticipantsRange(SAMPLE_MAX_ENTRY, SAMPLE_MAX_ENTRY);
    sessionSearchCriteriaList[1].SetMinParticipantsRange(1, 1);
    const size_t arraySize = 5;
    u32 attributeArray[arraySize] = { 2, 4, 6, 11, 28 };
    sessionSearchCriteriaList[1].SetAttribute(MATCHMAKE_ATTR_INDEX, attributeArray, arraySize);

    u32 frameNo = 0;
    bool isExit = false;
    SampleSceneNo nextScene = SCENE_MATCH_MAKE;

    while (!isExit)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Lower screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(false);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        int line = 0;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[%s Demo] %d", PROGRAM_NAME, frameNo);
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Scene: Matchmake.");
        line += BASIC_FONT_SIZE;
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Selection Method:", SELECT_METHOD[selectionMethodType]);
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, line, "  %s.", SELECT_METHOD[selectionMethodType]);
        line += BASIC_FONT_SIZE;
        if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  Rating: %u", rating);
            line += BASIC_FONT_SIZE;
        }
        else if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED)
        {
            line += BASIC_FONT_SIZE;
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    Score Setting: %d %s", scoreSettingIndex, SCORE_SETTING[scoreSettingIndex - SAMPLE_MIN_SCORE_SETTING_INDEX]);
            line += BASIC_FONT_SIZE;
            switch (scoreSettingIndex)
            {
                case SAMPLE_SCORE_INDEX_RATING:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    Rating: %u", rating);
                    line += BASIC_FONT_SIZE;
                    break;
                case SAMPLE_SCORE_INDEX_DISCONNECTION_RATE:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    DisconnectionRate: %u", disconnectionRate);
                    line += BASIC_FONT_SIZE;
                    break;
                case SAMPLE_SCORE_INDEX_VIOLATION_RATE:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    ViolationRate: %u", violationRate);
                    line += BASIC_FONT_SIZE;
                    break;
                case SAMPLE_SCORE_INDEX_SESSION_DURAION:
                    // Nothing to display.
                    break;
                case SAMPLE_SCORE_INDEX_SELECTION_PRIORITY:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    SelectionPriority: %u", selectionPriority);
                    line += BASIC_FONT_SIZE;
                    break;
                case SAMPLE_SCORE_INDEX_GEO_IP:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    Geo IP: %u. %s", geoIpIndex, GEO_IP_LOCATION[geoIpIndex]);
                    line += BASIC_FONT_SIZE;
                    break;
                case SAMPLE_SCORE_INDEX_COUNTRY_CODE:
                    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "    Country Code: %u", countryCode);
                    line += BASIC_FONT_SIZE;
                    break;
                default:
                    break;
            }
        }

        line = BASIC_LOWER_SCREEN_HEIGHT - (BASIC_FONT_SIZE * 7);
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Manual ]");
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  A: Start RandomMatchmake");
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  B: Logout");
        line += BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  R: Change Selection Method");
        line += BASIC_FONT_SIZE;
        if(selectionMethodType != SAMPLE_SESSION_SELECT_METHOD_ROMDOM)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  UP, DOWN: Parameter Adjustment");
            line += BASIC_FONT_SIZE;
        }
        if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "  RIGHT, LEFT: Change Score Setting");
            line += BASIC_FONT_SIZE;
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->SwapBuffers();

        ++frameNo;

        // Upper screen.
        nn::pia::sampleframework::SampleFramework::GetInstance()->ChangeRenderTarget(true);

        line = BASIC_UPPER_SCREEN_HEIGHT - BASIC_FONT_SIZE;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(0, line, "[ Version ] %s", Basic_GetPiaVersionString());
        line += BASIC_FONT_SIZE;

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        // Performs <tt>dispatch</tt>.
        if (nn::nex::Scheduler::GetInstance())
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

#if PIA_TEST_CODE
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_L))
        {
            // Intentionally make it so that the search conditions do not yield any results.
            attributeArray[4] = 50;
            sessionSearchCriteriaList[1].SetAttribute(MATCHMAKE_ATTR_INDEX, attributeArray, arraySize);
        }
#endif
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            if(selectionMethodType != SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED)
            {
                createSessionSetting.SetAttribute(nn::pia::inet::SESSION_BROADEN_RANGE_ATTRIBUTE_INDEX, rating);

                sessionSearchCriteriaList[0].SetAttribute(nn::pia::inet::SESSION_BROADEN_RANGE_ATTRIBUTE_INDEX, rating);
                sessionSearchCriteriaList[1].SetAttribute(nn::pia::inet::SESSION_BROADEN_RANGE_ATTRIBUTE_INDEX, rating);
            }
#if !ENABLE_SERVERLESS
            else if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED)
            {
                PIA_SAMPLE_LOG("score Setting Index: %d\n", scoreSettingIndex);

                createSessionSetting.SetRatingValue(rating);
                createSessionSetting.SetDisconnectionRate(disconnectionRate);
                createSessionSetting.SetViolationRate(violationRate);
                createSessionSetting.SetSelectionPriority(selectionPriority);

                sessionSearchCriteriaList[0].SetScoreSettingIndex(scoreSettingIndex);
                sessionSearchCriteriaList[0].SetRatingValue(rating);
                sessionSearchCriteriaList[0].SetDisconnectionRate(disconnectionRate);
                sessionSearchCriteriaList[0].SetViolationRate(violationRate);

                sessionSearchCriteriaList[1].SetScoreSettingIndex(scoreSettingIndex);
                sessionSearchCriteriaList[1].SetRatingValue(rating);
                sessionSearchCriteriaList[1].SetDisconnectionRate(disconnectionRate);
                sessionSearchCriteriaList[1].SetViolationRate(violationRate);

                if(scoreSettingIndex == SAMPLE_SCORE_INDEX_GEO_IP)
                {
                    createSessionSetting.UseGeoIp(true);
                    createSessionSetting.SetOverrideIpAddress(GEO_IP_LIST[geoIpIndex], wcslen(GEO_IP_LIST[geoIpIndex]));
                    createSessionSetting.SetNintendoCountryCode(countryCode);

                    sessionSearchCriteriaList[0].SetUseGeoIp(true);
                    sessionSearchCriteriaList[0].SetOverrideIpAddress(GEO_IP_LIST[geoIpIndex], wcslen(GEO_IP_LIST[geoIpIndex]));
                    sessionSearchCriteriaList[0].SetNintendoCountryCode(countryCode);

                    sessionSearchCriteriaList[1].SetUseGeoIp(true);
                    sessionSearchCriteriaList[1].SetOverrideIpAddress(GEO_IP_LIST[geoIpIndex], wcslen(GEO_IP_LIST[geoIpIndex]));
                    sessionSearchCriteriaList[1].SetNintendoCountryCode(countryCode);
                }
                else if(scoreSettingIndex == SAMPLE_SCORE_INDEX_COUNTRY_CODE)
                {
                    createSessionSetting.UseGeoIp(false);
                    createSessionSetting.SetNintendoCountryCode(countryCode);

                    sessionSearchCriteriaList[0].SetUseGeoIp(false);
                    sessionSearchCriteriaList[0].SetNintendoCountryCode(countryCode);

                    sessionSearchCriteriaList[1].SetUseGeoIp(false);
                    sessionSearchCriteriaList[1].SetNintendoCountryCode(countryCode);
                }
            }
#endif

            SampleSessionAsyncStatus status = Session_JoinRandomSession(
                &createSessionSetting,
                sessionSearchCriteriaList,
                searchSettingNum,
                &result);
            if (status == SAMPLE_SESSION_ASYNC_SUCCESS)
            {
                nextScene = SCENE_COMMUNICATE;
                isExit = true;
                continue;
            }
            else if (status == SAMPLE_SESSION_ASYNC_FAILED)
            {
                SetNetworkErrorResult(result);
                nn::pia::util::HandlingType handlingType = nn::pia::util::GetHandlingType(result);

                switch(handlingType)
                {
                case nn::pia::util::HANDLING_LOGOUT:
                case nn::pia::util::HANDLING_RECONNECT:
                case nn::pia::util::HANDLING_REINITIALIZE:
                    SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                    nextScene = SCENE_NETWORK_ERROR;
                    isExit = true;
                    break;
                case nn::pia::util::HANDLING_LEAVE:
                case nn::pia::util::HANDLING_LEAVE_AND_CLEANUP:
                case nn::pia::util::HANDLING_RETRY:
                    // The result of the JoinRandomSessionAsync asynchronous process is never any of the statuses above.
                case nn::pia::util::HANDLING_CLEANUP:
                    SetCommunicationErrorType(MATCH_MAKE_ERROR);
                    nextScene = SCENE_ERROR;
                    isExit = true;
                    break;
                case nn::pia::util::HANDLING_IGNORABLE:
                    // The JoinRandomSessionAsync asynchronous process never returns a Result handled in this manner.
                    PIA_SAMPLE_LOG("ignorable result\n");
                    break;
                case nn::pia::util::HANDLING_PROGRAMMING_ERROR:
                case nn::pia::util::HANDLING_VOID:
                default:
                    PIA_TRACE_RESULT(result);
                    PIA_SAMPLE_LOG("programming error\n");
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                    break;
                }
            }
            else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_SESSION)
            {
                SetCommunicationErrorType(MATCH_MAKE_ERROR);
                nextScene = SCENE_ERROR;
                isExit = true;
            }
            else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
            {
                SetCommunicationErrorType(CONNECTION_LOST_ERROR);
                nextScene = SCENE_NETWORK_ERROR;
                isExit = true;
            }
            else if (status == SAMPLE_SESSION_ASYNC_CANCELED)
            {
                CleanupPiaSession();
                nextScene = SCENE_MATCH_MAKE;
                isExit = true;
            }
        }

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            CleanupPiaSession();
            nextScene = SCENE_RESET;
            isExit = true;
        }

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            selectionMethodType = static_cast<SampleSessionSelectMethod>( (selectionMethodType + 1) % SAMPLE_SESSION_SELECT_METHOD_MAX);

            switch (selectionMethodType)
            {
            case SAMPLE_SESSION_SELECT_METHOD_ROMDOM:
                rating = SAMPLE_INITIAL_RATING_VALUE;
                sessionSearchCriteriaList[0].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_RANDOM);
                sessionSearchCriteriaList[1].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_RANDOM);

                scoreBasicParam = rating;
                break;
            case SAMPLE_SESSION_SELECT_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY:
                rating = SAMPLE_INITIAL_RATING_VALUE;
                sessionSearchCriteriaList[0].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY);
                sessionSearchCriteriaList[1].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY);

                scoreBasicParam = rating;
               break;
            case SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED:
                scoreSettingIndex = SAMPLE_SCORE_INDEX_RATING;
                rating = SAMPLE_INITIAL_RATING_VALUE;
                disconnectionRate = 0;
                violationRate = 0;
                geoIpIndex = 0;
                selectionPriority = 0;
                sessionSearchCriteriaList[0].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_SCORE_BASED);
                sessionSearchCriteriaList[1].SetRandomSessionSelectionMethod(nn::pia::inet::SELECTION_METHOD_SCORE_BASED);

                scoreBasicParam = rating;
                break;
            default:
                break;
            }
        }

        if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                rating = rating + 100;
                if (rating > SAMPLE_RATING_MAX)
                {
                    rating = SAMPLE_RATING_MAX;
                }
                scoreBasicParam = rating;
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                rating = rating - 100;
                if (rating < SAMPLE_RATING_MIN)
                {
                    rating = SAMPLE_RATING_MIN;
                    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0, rating);
                }
                scoreBasicParam = rating;
            }
        }
        else if(selectionMethodType == SAMPLE_SESSION_SELECT_METHOD_SCORE_BASED)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                ++scoreSettingIndex;
                if(scoreSettingIndex > SAMPLE_MAX_SCORE_SETTING_INDEX)
                {
                    scoreSettingIndex = SAMPLE_MIN_SCORE_SETTING_INDEX;
                }
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                --scoreSettingIndex;
                if(scoreSettingIndex < SAMPLE_MIN_SCORE_SETTING_INDEX)
                {
                    scoreSettingIndex = SAMPLE_MAX_SCORE_SETTING_INDEX;
                }
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                switch (scoreSettingIndex)
                {
                case SAMPLE_SCORE_INDEX_RATING:
                    rating = rating + 100;
                    if (rating > SAMPLE_RATING_MAX)
                    {
                        rating = SAMPLE_RATING_MAX;
                    }
                    break;
                case SAMPLE_SCORE_INDEX_DISCONNECTION_RATE:
                    // Because rating occur in four stages, at 0 to 1%, 1 to 3%, 3 to 5%, and 5 to 100%, with increases at 2% levels, when it reaches 8%, it is set to go around.
                    disconnectionRate = (disconnectionRate + 2) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_VIOLATION_RATE:
                    // Because rating occur in four stages, at 0 to 1%, 1 to 3%, 3 to 5%, and 5 to 100%, with increases at 2% levels, when it reaches 8%, it is set to go around.
                    violationRate = (violationRate + 2) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_SELECTION_PRIORITY:
                    selectionPriority = (selectionPriority + 2) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_GEO_IP:
                    geoIpIndex = (geoIpIndex + 1) % SAMPLE_GEO_IP_MAX;
                    break;
                case SAMPLE_SCORE_INDEX_COUNTRY_CODE:
                    countryCode = 50 - countryCode; // 49(US) <-> 1(JP)
                    break;
                default:
                    break;
                }
            }

            if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                switch (scoreSettingIndex)
                {
                case SAMPLE_SCORE_INDEX_RATING:
                    rating = rating - 100;
                    if (rating < SAMPLE_RATING_MIN)
                    {
                        rating = SAMPLE_RATING_MIN;
                        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0, rating);
                    }
                    break;
                case SAMPLE_SCORE_INDEX_DISCONNECTION_RATE:
                    disconnectionRate = (disconnectionRate + 6) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_VIOLATION_RATE:
                    violationRate = (violationRate + 6) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_GEO_IP:
                    geoIpIndex = (geoIpIndex + SAMPLE_GEO_IP_MAX - 1) % SAMPLE_GEO_IP_MAX;
                    break;
                case SAMPLE_SCORE_INDEX_SELECTION_PRIORITY:
                    selectionPriority = (selectionPriority + 6) % 8;
                    break;
                case SAMPLE_SCORE_INDEX_COUNTRY_CODE:
                    countryCode = 50 - countryCode; // 49(US) <-> 1(JP)
                    break;
                default:
                    break;
                }
            }

            switch (scoreSettingIndex)
            {
                case SAMPLE_SCORE_INDEX_RATING:
                    scoreBasicParam = rating;
                    break;
                case SAMPLE_SCORE_INDEX_DISCONNECTION_RATE:
                    scoreBasicParam = disconnectionRate;
                    break;
                case SAMPLE_SCORE_INDEX_VIOLATION_RATE:
                    scoreBasicParam = violationRate;
                    break;
                case SAMPLE_SCORE_INDEX_GEO_IP:
                    scoreBasicParam = geoIpIndex;
                    break;
                case SAMPLE_SCORE_INDEX_SELECTION_PRIORITY:
                    scoreBasicParam = selectionPriority;
                    break;
                case SAMPLE_SCORE_INDEX_COUNTRY_CODE:
                    scoreBasicParam = countryCode;
                    break;
                default:
                    break;
            }
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SAMPLE_PROGRAMPARAM_MATCHMAKE_SELECT_METHOD, selectionMethodType + (scoreSettingIndex * 100) );
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(SAMPLE_PROGRAMPARAM_SCORE_BASED_PARAM, scoreBasicParam);

        // Checks the connection status.
        // AC
        if (!Inet_IsConnected())
        {
            PIA_SAMPLE_LOG("Inet_IsConnected Disconnected \n");
            nn::pia::common::CachedPrint::GetInstance()->Flush();
            nextScene = SCENE_NETWORK_ERROR;
            SetCommunicationErrorType(CONNECTION_LOST_ERROR);
            isExit = true;
        }
    }

    return nextScene;
}
