﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/version.h>
#include <pia/common.h>

#include "SampleLib/Defines.h"
#include "SampleLib/ErrorHandling.h"
#include "SampleLib/SampleLib.h"
#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"

#include "SampleConfig.h"


namespace
{
// The memory region must be 4-byte aligned.
// In this sample, sufficiently large values are used.
PIA_ATTRIBUTE_ALIGN(4) bit8 s_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA];

// When set to <tt>true</tt>, suppresses the impact on processing speed when several trace flags of the Pia library have been enabled.
// However, the order relationship with logs of other libraries (NEX, and so on) may be switched.
const bool SWITCH_USE_CACHED_PRINT = true;

}  // end of anonymous namespace

#if NN_PIA_CTR
// Functions for getting the signal strength according to the communication method.
u32 GetInternetLinkLevel();
#endif

// Function that brings together the processes for getting the signal strength.
u32 GetLinkLevel()
{
#if NN_PIA_CTR
    // Get the signal strength according to the communication method, and return a value from 0 to 3.
    return GetInternetLinkLevel();
#elif NN_PIA_WIN
    // Always 0.
    return 0;
#elif NN_PIA_CAFE
    // There is no way to get the signal strength, so it is always 0.
    return 0;
#else
#error  "Illegal platform."
#endif
}

PIA_SAMPLE_MAIN_FUNC()
{
    PIA_SAMPLE_LOG("Sample program Start.\n");
    Basic_SetProgramTitle(PROGRAM_NAME);

    nn::Result result;

    // Initialize the render framework for the Pia sample program.
    nn::pia::sampleframework::SampleFramework::CreateInstance();
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAME_PER_SECOND);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(nn::pia::sampleframework::Color(0.0f, 0.0f, 0.2f, 1.0f));
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramStatus(SCENE_PREPARE_INET, false);

    while (1)
    {
        SampleSceneNo nextSceneNo = SCENE_ERROR;

        // Setup of the features used in the PiaCommon module.
        result = Common_InitializePiaCommon(
            s_PiaMemoryBlock,
            MEMORY_BLOCK_SIZE_PIA,
            nn::pia::common::TRACE_FLAG_COMMON |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
#if PIA_TEST_CODE
            nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_SEND |
            nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_RECEIVE |
#endif
                nn::pia::common::TRACE_FLAG_SESSION |
#if PIA_TEST_CODE
            nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
            nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
            nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
            nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
#endif
                nn::pia::common::TRACE_FLAG_INET |
#if PIA_TEST_CODE
            nn::pia::common::TRACE_FLAG_INET_NAT |
            nn::pia::common::TRACE_FLAG_INET_NAT_DETAIL |
            nn::pia::common::TRACE_FLAG_INET_NAT_PROBE |
            nn::pia::common::TRACE_FLAG_INET_SOCKET |
#endif
                nn::pia::common::TRACE_FLAG_ALWAYS,
            SWITCH_USE_CACHED_PRINT,
            true);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        // Specify error handling for the Pia library.
        ErrorHandling::Initialize();

        // Because the processes for each feature are clearly separated in this sample, the processes in the main thread are divided into several scenes.
        bool isExit = false;

        nextSceneNo = SCENE_PREPARE_INET;
        //bool isSessionHost = false;

        /* Game loop */
        while (isExit == false)
        {
            switch (nextSceneNo)
            {
            case SCENE_PREPARE_INET:
            {
                if (InitializePiaInet())
                {
                    nextSceneNo = SCENE_MATCH_MAKE;
                }
                else
                {
                    nextSceneNo = SCENE_NETWORK_ERROR;
                }
            }
            break;
            case SCENE_MATCH_MAKE:
            {
                nextSceneNo = SceneMatchmake();
            }
            break;
            case SCENE_COMMUNICATE:
            {
                nextSceneNo = SceneCommunication();
            }
            break;
            case SCENE_ERROR:
            {
                nextSceneNo = SceneMeshConnectionError();
                CleanupPiaSession();
            }
            break;
            case SCENE_NETWORK_ERROR:
            {
                nextSceneNo = SceneNetworkError();
            }
            break;
            case SCENE_RESET:
            {
                FinalizePiaInet();
                isExit = true;
            }
            break;
            }
        }

        /* PiaCommon-related finalization. */
        ErrorHandling::Finalize();

        Common_FinalizePiaCommon();
        /* End of PiaCommon-related finalization. */
    }
    //Control does not arrive here.
    //nn::pia::sampleframework::SampleFramework::DestroyInstance();
    //PIA_SAMPLE_LOG("Sample program end.\n");

}  // end of nnMain

// When DDL files are not used with NEX, a definition of the empty InitDOClasses function is required.
namespace nn
{
namespace nex
{
void InitDOClasses()
{
}
}
}
