﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/util.h>

#include "SampleLib/SampleLib.h"

#include "SampleConfig.h"
#include "ControllerData.h"

#include <cstdio>

#if USE_PIA_LOCAL
#include "Shared_Local_Setup.h"
#endif

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Sync_Setup.h"

namespace
{
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.

static ControllerReader* s_pControllerReader;
static ControllerAccessor s_ControllerAccessor[MAX_STATIONS];

static bit32 USING_DATA_ID_BITMAP_ALL = (0x1 << DATA_ID_NUM) - 1;
static bit32 s_MyUsingDataIdBitmap;
static s32 s_Delay;
static u32 s_SendPeriod;
static int s_CursorConfigId;

// For changing the delay.
static u32 s_ChangeDelayInput;
static u32 s_ChangeDelay;

// For changing the send interval.
static u32 s_ChangeSendPeriod;

nn::pia::StationId s_StationIdList[MAX_STATIONS];

enum ConfigId
{
#if PIA_TEST_CODE
    CONFIG_DELAY = DATA_ID_NUM - 1,
#else
    CONFIG_DELAY = DATA_ID_NUM,
#endif
    CONFIG_SENDPERIOD,
    CONFIG_NUM
};

// Indicates the mode that is executing.
// <tt>inet</tt> starts with AUTOMATCH and then, after matching, decides whether to be the host or a client.
enum GameMode
{
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_HOST,
    GAME_MODE_LOCAL_CLIENT,
#endif
    GAME_MODE_INET_AUTOMATCH,
    GAME_MODE_INET_HOST,
    GAME_MODE_INET_CLIENT,

#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_ERROR,
#endif
    GAME_MODE_INET_ERROR,
    GAME_MODE_EXIT
};

nn::pia::StationIndex ConvertStationIdToStationIndex(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
    if (result.IsFailure())
    {
        stationIndex = nn::pia::STATION_INDEX_INVALID;
    }
    return stationIndex;
}

void ClearStationIdList()
{
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        s_StationIdList[i] = nn::pia::STATION_ID_INVALID;
    }
}

void RegisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(s_StationIdList[stationIndex] == nn::pia::STATION_ID_INVALID);
    s_StationIdList[stationIndex] = stationId;
}

void UnregisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(s_StationIdList[stationIndex] != nn::pia::STATION_ID_INVALID);
    s_StationIdList[stationIndex] = nn::pia::STATION_ID_INVALID;
}

// Declarations of callback functions.
void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG("Sample Event JOIN detect. StationId(0x%016llx)\n", stationId.GetValue());
        RegisterStationId(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect. StationId(0x%016llx)\n", stationId.GetValue());
        UnregisterStationId(stationId);
    }
    break;
    }
}

// Determines the mode to execute.
GameMode DecideGameMode()
{
#if USE_PIA_LOCAL
    PIA_SAMPLE_PRINT("Please select game mode.\n [A]LocalHost\n [B]LocalClient\n [X]InetAutomatch\n [START]Quit\n");
#else
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
#endif
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
#if USE_PIA_LOCAL
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalHost\n");
            gameMode = GAME_MODE_LOCAL_HOST;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalClient\n");
            gameMode = GAME_MODE_LOCAL_CLIENT;
            isDecided = true;
        }
#endif

        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        pSampleFramework->DrawText(10, 20, "Select game mode.");
#if USE_PIA_LOCAL
        pSampleFramework->DrawText(10, 30, "A: LocalHost");
        pSampleFramework->DrawText(10, 40, "B: LocalClient");
        pSampleFramework->DrawText(10, 50, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 60, "START: Quit this application");
#else
        pSampleFramework->DrawText(10, 30, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 40, "START: Quit this application");
#endif

#if PIA_TEST_CODE
        pSampleFramework->SetProgramStatus(isDecided ? 0 : 1, false);
        pSampleFramework->SetProgramParam(0x01, isDecided ? 1 : 0);
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return gameMode;
        }
    }
}


#if PIA_TEST_CODE
//Up Button: 1, Right Button: 2, Down Button: 4, Left Button: 8
static bit32 s_KeyBit[4] = {
    nn::pia::sampleframework::BUTTON_UP,
    nn::pia::sampleframework::BUTTON_RIGHT,
    nn::pia::sampleframework::BUTTON_DOWN,
    nn::pia::sampleframework::BUTTON_LEFT
};

void CheckTestInput()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L) &&
        pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
    {
        u32 changeDelay = 0;
        for (u32 i = 0; i < 4; i++)
        {
            if (pSampleFramework->IsTrigger(s_KeyBit[i]))
            {
                changeDelay += (0x01 << i);
            }
        }

        s_ChangeDelayInput = changeDelay;
        s_ChangeDelay = s_ChangeDelayInput;
    }
}
#endif


// Operation by key input.
bool Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    nn::Result r;

    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        // Exit the GameLoop function if START is pressed.
        PIA_SAMPLE_PRINT("Exit GameLoop\n");
        return true;
    }

    switch (pSyncProtocol->GetState())
    {
    case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
    {
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            // The A Button starts synchronous communication.
            r = pSyncProtocol->Start(s_MyUsingDataIdBitmap, s_Delay, s_SendPeriod);
            PIA_SAMPLE_PRINT("Start %s\n", r.IsSuccess() ? "Success" : "Failure");
        }

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            --s_CursorConfigId;
            if (s_CursorConfigId < 0)
            {
                s_CursorConfigId = CONFIG_NUM - 1;
            }
        }
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            ++s_CursorConfigId;
            if (s_CursorConfigId >= CONFIG_NUM)
            {
                s_CursorConfigId = 0;
            }
        }

        if (s_CursorConfigId < CONFIG_DELAY)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT) || pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                s_MyUsingDataIdBitmap ^= (0x1 << s_CursorConfigId);
            }
        }
        else if (s_CursorConfigId == CONFIG_DELAY)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                if (s_Delay > 1)
                {
                    --s_Delay;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                if (s_Delay < DELAY_MAX)
                {
                    ++s_Delay;
                }
            }
        }
        else if (s_CursorConfigId == CONFIG_SENDPERIOD)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                if (s_SendPeriod > 1)
                {
                    --s_SendPeriod;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                ++s_SendPeriod;
            }
        }
    }
    break;

    case nn::pia::sync::SyncProtocol::STATE_WAITING:
    case nn::pia::sync::SyncProtocol::STATE_STARTING:
    case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
    case nn::pia::sync::SyncProtocol::STATE_ENDED_ALONE:
    {
        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R) && pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            // End synchronization if the R and X Buttons are pressed.
            r = pSyncProtocol->End();
            PIA_SAMPLE_PRINT("End %s\n", r.IsSuccess() ? "Success" : "Failure");
        }

        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R) && pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            // End synchronization on the local station only if the R and Y Buttons are pressed.
            r = pSyncProtocol->EndAlone();
        }

        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                // Increment the delay by 1 if the R and Right Buttons are pressed.
                if (s_ChangeDelayInput < pSyncProtocol->GetMaxDelay())
                {
                    ++s_ChangeDelayInput;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                // Decrement the delay by 1 if the R and Left Buttons are pressed.
                if (s_ChangeDelayInput > 0)
                {
                    --s_ChangeDelayInput;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                // Ask to change the delay if the R and A Buttons are pressed.
                s_ChangeDelay = s_ChangeDelayInput;
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                // Increment the send interval by 1 if the R and Up Buttons are pressed.
                ++s_ChangeSendPeriod;
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                // Decrement the send interval by 1 if the R and Down Buttons are pressed.
                if (s_ChangeSendPeriod > 1)
                {
                    --s_ChangeSendPeriod;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                // Change the send interval if the R and B Buttons are pressed.
                pSyncProtocol->SetSendPeriod(s_ChangeSendPeriod);
            }
        }

#if PIA_TEST_CODE
        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L))
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                nn::Result result;
                nn::pia::transport::ThreadStreamManager* threadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_ASSERT(PIA_IS_VALID_POINTER(threadStreamMgr));
                result = threadStreamMgr->SetSendThreadPacketLossRatio(0);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                result = threadStreamMgr->SetReceiveThreadPacketLossRatio(0);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                nn::Result result;
                nn::pia::transport::ThreadStreamManager* threadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_ASSERT(PIA_IS_VALID_POINTER(threadStreamMgr));
                result = threadStreamMgr->SetSendThreadPacketLossRatio(100);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                result = threadStreamMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
        }
#endif

#if PIA_TEST_CODE
        CheckTestInput();
#endif
    }
    break;
#if PIA_TEST_CODE
    case nn::pia::sync::SyncProtocol::STATE_ENDING:
    {
        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L))
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                nn::Result result;
                nn::pia::transport::ThreadStreamManager* threadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_ASSERT(PIA_IS_VALID_POINTER(threadStreamMgr));
                result = threadStreamMgr->SetSendThreadPacketLossRatio(0);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                result = threadStreamMgr->SetReceiveThreadPacketLossRatio(0);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                nn::Result result;
                nn::pia::transport::ThreadStreamManager* threadStreamMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_ASSERT(PIA_IS_VALID_POINTER(threadStreamMgr));
                result = threadStreamMgr->SetSendThreadPacketLossRatio(100);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                result = threadStreamMgr->SetReceiveThreadPacketLossRatio(100);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
        }
    }
    break;
#endif
    }

    return false;
}


static u32 sDispatchCount = 0;
static s32 sDispatchTime = 0;

// Draw on the screen.
void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    static const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    static const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);

    {
        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_HOST:
        case GAME_MODE_LOCAL_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "LocalHost";
            }
            else
            {
                cGameModeText = "LocalClient";
            }
        }
        break;
#endif
        case GAME_MODE_INET_AUTOMATCH:
        {
            cGameModeText = "InetAutomatch";
        }
        break;
        case GAME_MODE_INET_HOST:
        case GAME_MODE_INET_CLIENT:
        {
            if (pSession->IsHost())
            {
                cGameModeText = "InetHost";
            }
            else
            {
                cGameModeText = "InetClient";
            }
        }
        break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%d", pSession->GetStationNum());
        pSampleFramework->DrawText(290, 20, "Index:%d", ConvertStationIdToStationIndex(pSession->GetLocalStationId()));
    }

    {
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    {
        // Displays the state of SyncProtocol.
        const char* cpState = "";
        switch (pSyncProtocol->GetState())
        {
        case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
            cpState = "NOT_SYNCHRONIZED";
            break;
        case nn::pia::sync::SyncProtocol::STATE_WAITING:
            cpState = "WAITING";
            break;
        case nn::pia::sync::SyncProtocol::STATE_STARTING:
            cpState = "STARTING";
            break;
        case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
            cpState = "SYNCHRONIZED";
            break;
        case nn::pia::sync::SyncProtocol::STATE_ENDING:
            cpState = "ENDING";
            break;
        case nn::pia::sync::SyncProtocol::STATE_ENDED_ALONE:
            cpState = "ENDED_ALONE";
            break;
        }
        pSampleFramework->DrawText(10, 30, "State:%s", cpState);

        if (pSyncProtocol->GetLastEndReason() != nn::pia::sync::SyncProtocol::END_REASON_NOTHING)
        {
            const char* cpEndReason = "";
            switch (pSyncProtocol->GetLastEndReason())
            {
            case nn::pia::sync::SyncProtocol::END_REASON_END:
                cpEndReason = "END";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_OTHER_STATION:
                cpEndReason = "OTHER_STATION";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_CHANGE_CONNECTION:
                cpEndReason = "CHANGE_CONNECTION";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_TIMEOUT:
                cpEndReason = "TIMEOUT";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_END_ALONE:
                cpEndReason = "END_ALONE";
                break;
            }
            pSampleFramework->DrawText(200, 30, "EndReason:%s", cpEndReason);
        }
    }

    {
        // Displays the possible operations for the given state.
        switch (pSyncProtocol->GetState())
        {
        case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
        {
            pSampleFramework->DrawText(10, 40, "A:Start START:Exit");
        }
        break;
        case nn::pia::sync::SyncProtocol::STATE_WAITING:
        case nn::pia::sync::SyncProtocol::STATE_STARTING:
        case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
        {
            pSampleFramework->DrawText(10, 40, "R+X:End R+Y:EndAlone");
            pSampleFramework->DrawText(10, 50, "START:Exit");
        }
        break;
        case nn::pia::sync::SyncProtocol::STATE_ENDING:
        {
            pSampleFramework->DrawText(10, 40, "START:Exit");
        }
        break;
        }
    }

    if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED)
    {
        pSampleFramework->DrawText(10, 60, "Key");
        pSampleFramework->DrawText(10, 70, "Touch");
        pSampleFramework->DrawText(10, 80, "Stick");
        pSampleFramework->DrawText(10, 90, "Delay");
        pSampleFramework->DrawText(10, 100, "SendPeriod");

        for (s32 i = 0; i < CONFIG_NUM; ++i)
        {
            s32 y = 60 + i * 10;
            if (i < CONFIG_DELAY)
            {
                pSampleFramework->DrawText(100, y, "%c %s", (i == s_CursorConfigId) ? '>' : ' ', ((0x1 << i) & s_MyUsingDataIdBitmap) ? "ON" : "OFF");
            }
            else if (i == CONFIG_DELAY)
            {
                pSampleFramework->DrawText(100, y, "%c %d", (i == s_CursorConfigId) ? '>' : ' ', s_Delay);
            }
            else if (i == CONFIG_SENDPERIOD)
            {
                pSampleFramework->DrawText(100, y, "%c %d", (i == s_CursorConfigId) ? '>' : ' ', s_SendPeriod);
            }
        }
    }
    else
    {

        {
            // Shows the column titles.
            pSampleFramework->DrawText(5, 70, "UP");
            pSampleFramework->DrawText(5, 80, "DOWN");
            pSampleFramework->DrawText(5, 90, "LEFT");
            pSampleFramework->DrawText(5, 100, "RIGHT");
            pSampleFramework->DrawText(5, 110, "A");
            pSampleFramework->DrawText(5, 120, "B");
            pSampleFramework->DrawText(5, 130, "X");
            pSampleFramework->DrawText(5, 140, "Y");
            pSampleFramework->DrawText(5, 155, "TouchX");
            pSampleFramework->DrawText(5, 165, "TouchY");
            pSampleFramework->DrawText(5, 180, "StickX");
            pSampleFramework->DrawText(5, 190, "StickY");
            pSampleFramework->DrawText(5, 200, "EndAln");
        }

        {
            // Displays the synchronized controller information of each station.
            for (s32 i = 0; i < MAX_STATIONS; ++i)
            {
                nn::pia::StationId stationId = s_StationIdList[i];

                s32 x = 55 + i * 43;

                pSampleFramework->SetFontColor(pSyncProtocol->CheckEntry(stationId) ? cDefaultColor : cInvalidColor);
                pSampleFramework->DrawText(x, 60, " [%dP]", i + 1);

                if (pSyncProtocol->CheckEntry(stationId) == false)
                {
                    continue;
                }

                bit32 usingDataIdBitmap = pSyncProtocol->GetUsingDataIdBitmap(stationId);

                if (usingDataIdBitmap & (0x1 << DATA_ID_KEY))
                {
                    for (s32 j = 0; j < KEY_NUM; ++j)
                    {
                        s32 y = 70 + j * 10;
                        Key key = static_cast<Key>(j);
                        pSampleFramework->DrawText(x, y, " %c%c%c", s_ControllerAccessor[i].IsTrigger(key) ? 'T' : '-', s_ControllerAccessor[i].IsHold(key) ? 'H' : '-', s_ControllerAccessor[i].IsRelease(key) ? 'R' : '-');
                    }
                }

                if (usingDataIdBitmap & (0x1 << DATA_ID_TOUCH))
                {
                    if (s_ControllerAccessor[i].IsTouch())
                    {
                        pSampleFramework->DrawText(x, 155, "%4d", s_ControllerAccessor[i].GetTouchX());
                        pSampleFramework->DrawText(x, 165, "%4d", s_ControllerAccessor[i].GetTouchY());
                    }
                    else
                    {
                        pSampleFramework->DrawText(x, 155, " ---");
                        pSampleFramework->DrawText(x, 165, " ---");
                    }
                }

                if (usingDataIdBitmap & (0x1 << DATA_ID_STICK))
                {
                    pSampleFramework->DrawText(x, 180, "%5.2f", s_ControllerAccessor[i].GetStickX());
                    pSampleFramework->DrawText(x, 190, "%5.2f", s_ControllerAccessor[i].GetStickY());
                }
            }

            pSampleFramework->SetFontColor(cDefaultColor);
        }

        {
            // Display the end frame.
            for (s32 i = 0; i < MAX_STATIONS; ++i)
            {
                nn::pia::StationId stationId = s_StationIdList[i];
                if (stationId == nn::pia::STATION_ID_INVALID)
                {
                    continue;
                }

                s32 x = 55 + i * 43;
                u32 endFrameNo = 0;
                if (pSyncProtocol->GetState() != nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED &&
                    pSyncProtocol->GetState() != nn::pia::sync::SyncProtocol::STATE_WAITING)
                {
                    nn::Result result = pSyncProtocol->GetEndFrameNo(stationId, &endFrameNo);
                    if (result.IsSuccess())
                    {
                        pSampleFramework->DrawText(x, 200, "%4d", endFrameNo);
                    }
                    else
                    {
                        if (pSyncProtocol->CheckEntry(stationId) == false)
                        {
                            continue;
                        }
                        pSampleFramework->DrawText(x, 200, " ---");
                    }
                }
            }
        }
        {
            // Display the delay.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(180, 30, "Delay:%d", pSyncProtocol->GetDelay());
            }

            // Display the requested delay.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                if (ENABLE_CHANGE_DELAY)
                {
                    pSampleFramework->DrawText(180, 40, "R+[<>A]:ChangeDelay:%d", s_ChangeDelayInput);
                }
                else
                {
                    pSampleFramework->DrawText(180, 40, "R+[<>A]:ChangeDelay:-");
                }
            }

            // Display the send interval.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(290, 30, "SendPeriod:%d", pSyncProtocol->GetSendPeriod());
            }

            // Display the send interval to set.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(180, 50, "R+[^vB]:ChangeSendPeriod:%d", s_ChangeSendPeriod);
            }

            // Display the frame number.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(260, 10, "Frame: %d", pSyncProtocol->GetFrameNo());
            }

#if 0
            // Displays the number of leading frames.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(180, 40, "LeadingFrame:");
                for (u32 i = 0; i < pSession->GetStationNum(); i++)
                {
                    pSampleFramework->DrawText(280 + i*20, 40, " %d", pSyncProtocol->EstimateLeadingFrameNum(s_StationIdList[i], FRAME_PER_SECONDS));
                }
            }
#endif

            // To check on process slowdowns, display the number of dispatch attempts and time required.
            pSampleFramework->DrawText(10, 220, "Dispatch: %3d (%4dms)", sDispatchCount, sDispatchTime);
        }
    }
}

#if PIA_TEST_CODE
void CheckSyncData()
{
    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    // Baseline station for comparison.
    u32 baseIndex = 0;
    for (u32 i = 0; i < SYNC_CHECK_BUFFER_NUM; ++i)
    {
        if (pSyncProtocol->CheckEntry(s_StationIdList[i]))
        {
            baseIndex = i;
            break;
        }
    }

    // Check the data that station j sent about station i.
    for (u32 i = 0; i < SYNC_CHECK_BUFFER_NUM; ++i)
    {
        if (!pSyncProtocol->CheckEntry(s_StationIdList[i]))
        {
            continue;
        }

        // Check based on the data sent from the station <tt>baseIndex</tt>.
        u32 holdFlag = s_ControllerAccessor[baseIndex].GetSyncCheckHoldFlag(i);
        u32 triggerFlag = s_ControllerAccessor[baseIndex].GetSyncCheckTriggerFlag(i);
        u32 releaseFlag = s_ControllerAccessor[baseIndex].GetSyncCheckReleaseFlag(i);
        u16 touchX = s_ControllerAccessor[baseIndex].GetSyncCheckTouchX(i);
        u16 touchY = s_ControllerAccessor[baseIndex].GetSyncCheckTouchY(i);
        u8 isTouch = s_ControllerAccessor[baseIndex].GetSyncCheckIsTouch(i);
        f32 stickX = s_ControllerAccessor[baseIndex].GetSyncCheckStickX(i);
        f32 stickY = s_ControllerAccessor[baseIndex].GetSyncCheckStickY(i);

        for (u32 j = baseIndex + 1; j < MAX_STATIONS; ++j)
        {
            if (!pSyncProtocol->CheckEntry(s_StationIdList[j]))
            {
                continue;
            }

            if (holdFlag != s_ControllerAccessor[j].GetSyncCheckHoldFlag(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (triggerFlag != s_ControllerAccessor[j].GetSyncCheckTriggerFlag(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (releaseFlag != s_ControllerAccessor[j].GetSyncCheckReleaseFlag(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (touchX != s_ControllerAccessor[j].GetSyncCheckTouchX(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (touchY != s_ControllerAccessor[j].GetSyncCheckTouchY(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (isTouch != s_ControllerAccessor[j].GetSyncCheckIsTouch(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (stickX != s_ControllerAccessor[j].GetSyncCheckStickX(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
            if (stickY != s_ControllerAccessor[j].GetSyncCheckStickY(i))
            {
                PIA_SAMPLE_ASSERT(false);
            }
        }
    }
}

void UpdatePiaTestData(GameMode gameMode, bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    bool isInet = false;
    switch (gameMode)
    {
    case GAME_MODE_INET_HOST:
    case GAME_MODE_INET_CLIENT:
    case GAME_MODE_INET_AUTOMATCH:
    {
        isInet = true;
    }
    break;
    }

    nn::pia::session::Session::Status sessionStatus = pSession->GetStatus();
    bool isConnected = pSession->CheckConnectionError().IsSuccess() &&
                       sessionStatus == nn::pia::session::Session::STATUS_CONNECTED_SESSION;
    pSampleFramework->SetProgramStatus(
        isExit ? 0 : (isInet ? 4 : 3),
        isConnected,
        pSession->GetLocalStationId() == pSession->GetHostStationId(),
        pSession->GetStationNum());


    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    // Session.
    pSampleFramework->SetProgramParam(0x00, ConvertStationIdToStationIndex(pSession->GetLocalStationId()));

    // Overall SyncProtocol.
    pSampleFramework->SetProgramParam(0x10, pSyncProtocol->GetState());
    pSampleFramework->SetProgramParam(0x11, pSyncProtocol->GetLastEndReason());
    pSampleFramework->SetProgramParam(0x12, pSyncProtocol->GetFrameNo());
    pSampleFramework->SetProgramParam(0x13, pSyncProtocol->GetDelay());
    pSampleFramework->SetProgramParam(0x14, pSyncProtocol->GetSendPeriod());

    // Local station settings.
    pSampleFramework->SetProgramParam(0x20, s_CursorConfigId);
    pSampleFramework->SetProgramParam(0x21, s_MyUsingDataIdBitmap);
    pSampleFramework->SetProgramParam(0x22, s_Delay);
    pSampleFramework->SetProgramParam(0x23, s_SendPeriod);
    pSampleFramework->SetProgramParam(0x24, s_ChangeDelayInput);
    pSampleFramework->SetProgramParam(0x25, s_ChangeSendPeriod);

    // Each individual station.
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        nn::pia::StationId stationId = s_StationIdList[i];
        u32 idOfs = (i + 8) << 4;

        pSampleFramework->SetProgramParam(idOfs + 0x0, pSyncProtocol->CheckEntry(stationId) ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x1, pSyncProtocol->GetUsingDataIdBitmap(stationId));

        pSampleFramework->SetProgramParam(idOfs + 0x2, s_ControllerAccessor[i].GetHoldFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x3, s_ControllerAccessor[i].GetTriggerFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x4, s_ControllerAccessor[i].GetReleaseFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x5, s_ControllerAccessor[i].IsTouch() ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x6, s_ControllerAccessor[i].GetTouchX());
        pSampleFramework->SetProgramParam(idOfs + 0x7, s_ControllerAccessor[i].GetTouchY());
        pSampleFramework->SetProgramParam(idOfs + 0x8, static_cast<s32>(s_ControllerAccessor[i].GetStickX() * 1000.0f));
        pSampleFramework->SetProgramParam(idOfs + 0x9, static_cast<s32>(s_ControllerAccessor[i].GetStickY() * 1000.0f));
    }
}
#endif

// Main loop during communication.
void MainLoop(GameMode gameMode, bool* pIsLeaveNeeded)
{
    nn::Result r;

    s_pControllerReader->Clear();
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        s_ControllerAccessor[i].Clear();
    }

    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    while (true)
    {
        bool isExit = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        sDispatchCount = 1;
        nn::pia::common::Time start;
        start.SetNow();

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount() % NEX_DISPATCH_PERIOD == 0)
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (gameMode == GAME_MODE_INET_HOST || gameMode == GAME_MODE_INET_CLIENT)
        {
            if (Inet_IsConnected() == false)
            {
                // Ends because the AC has disconnected.
                isExit = true;
                *pIsLeaveNeeded = false;
            }
        }

        nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
        if (result.IsFailure())
        {
            // Ends because the communication has been disconnected.
            isExit = true;
            nn::pia::util::HandlingType handlingType = nn::pia::util::GetHandlingType(result);
            PIA_SAMPLE_ASSERT(handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP || handlingType == nn::pia::util::HANDLING_RECONNECT);
            if (handlingType == nn::pia::util::HANDLING_LEAVE_AND_CLEANUP)
            {
                *pIsLeaveNeeded = true;
            }
            else
            {
                *pIsLeaveNeeded = false;
            }
        }

        if (pSyncProtocol->NeedSetData())
        {
            // Get the data to send and sets it to SyncProtocol.
            s_pControllerReader->Read();

#if PIA_TEST_CODE
            // Set the data for detecting desync.
            for (u32 i = 0; i < SYNC_CHECK_BUFFER_NUM; ++i)
            {
                nn::pia::StationId stationId = s_StationIdList[i];
                if (pSyncProtocol->CanGetData() && pSyncProtocol->CheckEntry(stationId))
                {
                    s_pControllerReader->SetSyncCheckKeyData(
                        i,
                        s_ControllerAccessor[i].GetHoldFlag(),
                        s_ControllerAccessor[i].GetTriggerFlag(),
                        s_ControllerAccessor[i].GetReleaseFlag());
                    s_pControllerReader->SetSyncCheckTouchData(
                        i,
                        s_ControllerAccessor[i].GetTouchX(),
                        s_ControllerAccessor[i].GetTouchY(),
                        s_ControllerAccessor[i].IsTouch());
                    s_pControllerReader->SetSyncCheckStickData(
                        i,
                        s_ControllerAccessor[i].GetStickX(),
                        s_ControllerAccessor[i].GetStickY());
                }
            }

#if 0
            // Force detection of desync.
            if (pSyncProtocol->GetFrameNo() == 100 &&
                nn::pia::session::Session::GetInstance()->IsHost())
            {
                s_pControllerReader->SetSyncCheckKeyData(
                    0,
                    s_ControllerAccessor[0].GetHoldFlag() + 1,
                    s_ControllerAccessor[0].GetTriggerFlag(),
                    s_ControllerAccessor[0].GetReleaseFlag()
                );
            }
#endif

#endif

            for (u32 i = 0; i < DATA_ID_NUM; ++i)
            {
                DataId dataId = static_cast<DataId>(i);
                if (pSyncProtocol->NeedSetData(dataId))
                {
                    r = pSyncProtocol->SetData(dataId, s_pControllerReader->GetData(dataId));
                    PIA_SAMPLE_ASSERT(r.IsSuccess());
                }
            }

            if (ENABLE_CHANGE_DELAY)
            {
                // Request delay change.
                r = pSyncProtocol->RequestToChangeDelay(s_ChangeDelay);
                PIA_SAMPLE_ASSERT(r.IsSuccess());
                s_ChangeDelay = 0;
            }
        }

        if (!isExit)
        {
            // Accept operation.
            isExit = Operate();
            if (isExit)
            {
                *pIsLeaveNeeded = true;
            }
        }

        if (!isExit)
        {
            // Advance one frame.
            r = pSyncProtocol->Step();
            if (r.IsFailure())
            {
                if (r == nn::pia::ResultDataIsNotArrivedYet())
                {
                    // Because the data has yet to arrive, repeatedly call Dispatch until it arrives.

                    while (true)
                    {
                        Draw(gameMode);
#if PIA_TEST_CODE
                        UpdatePiaTestData(gameMode, isExit);
#endif

                        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

                        ++sDispatchCount;
                        if (nn::nex::Scheduler::GetInstance() != NULL)
                        {
                            nn::nex::Scheduler::GetInstance()->DispatchAll();
                        }

                        nn::pia::common::Scheduler::GetInstance()->Dispatch();

                        if (gameMode == GAME_MODE_INET_HOST || gameMode == GAME_MODE_INET_CLIENT)
                        {
                            if (Inet_IsConnected() == false)
                            {
                                // Ends because the AC has disconnected.
                                break;
                            }
                        }

                        if (nn::pia::session::Session::GetInstance()->CheckConnectionError().IsFailure())
                        {
                            // Ends because the communication has been disconnected.
                            break;
                        }

                        // The frame has not advanced since the data was set, so it is not necessary to set it here.
                        PIA_SAMPLE_ASSERT(pSyncProtocol->NeedSetData() == false);

                        if (isExit == false)
                        {
                            // Accept operation.
                            isExit = Operate();
                            if (isExit)
                            {
                                *pIsLeaveNeeded = true;
                            }
                        }

                        r = pSyncProtocol->Step();
                        if (r != nn::pia::ResultDataIsNotArrivedYet())
                        {
                            break;
                        }
                    }
                }
                else if (r == nn::pia::ResultDataIsNotSet())
                {
                    // The data is set previously, so this result does not occur.
                    PIA_SAMPLE_ASSERT(false);
                }
                else
                {
                    // Other errors are ignored here.
                }
            }
        }

        // If you are using the feature to change the delay during synchronization, you may not be able
        // to get data from some frames. You must check whether you can get data in the current frame.
        bool isNoDataFrame = false;
        if (pSyncProtocol->CanGetData())
        {
            r = pSyncProtocol->GetIsNoDataFrame(&isNoDataFrame);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
        }

        // Get the received data. The local station also uses the data obtained here.
        for (u32 i = 0; i < MAX_STATIONS; ++i)
        {
            nn::pia::StationId stationId = s_StationIdList[i];
            s_ControllerAccessor[i].Clear();

            if (isNoDataFrame)
            {
                continue;
            }

            if (pSyncProtocol->CanGetData() && pSyncProtocol->CheckEntry(stationId))
            {
                bit32 usingDataIdBitmap = pSyncProtocol->GetUsingDataIdBitmap(stationId);
                for (u32 j = 0; j < DATA_ID_NUM; ++j)
                {
                    DataId dataId = static_cast<DataId>(j);
                    if ((0x1 << dataId) & usingDataIdBitmap)
                    {
                        r = pSyncProtocol->GetData(stationId, dataId, s_ControllerAccessor[i].GetDataPtr(dataId));
                        PIA_SAMPLE_ASSERT(r.IsSuccess());

                        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSyncProtocol->GetDataPtr(stationId, dataId)));
                        if (dataId == DATA_ID_KEY)
                        {
                            PIA_SAMPLE_ASSERT(*static_cast<const u32*>(pSyncProtocol->GetDataPtr(stationId, dataId)) == *static_cast<u32*>(s_ControllerAccessor[i].GetDataPtr(dataId)));
                        }
                    }
                }
            }
        }

#if PIA_TEST_CODE
        // Run a desync check using the data for detecting desync.
        if (pSyncProtocol->CanGetData())
        {
            CheckSyncData();
        }
#endif

        nn::pia::common::Time end;
        end.SetNow();
        sDispatchTime = (end - start).GetMSec();

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        Draw(gameMode);

#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, isExit);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0x01, isExit ? 1 : 0);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            return;
        }
    }
}


}  // end of anonymous namespace


// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;

    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        setting.bRawTouchpad = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAME_PER_SECONDS);

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    // Controller-related initialization is conducted by the <tt>SampleFramework::CreateInstance</tt> function,
    // so create a <tt>ControllerReader</tt> instance after that has been done.
    ControllerReader controllerReader;
    s_pControllerReader = &controllerReader;

    while (true)
    {
        ClearStationIdList();

        GameMode gameMode = DecideGameMode();
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
#if PIA_TEST_CODE
            u64 traceFlags =
                nn::pia::common::TRACE_FLAG_SYNC |
                nn::pia::common::TRACE_FLAG_LOCAL |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_SESSION |
                nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
                nn::pia::common::TRACE_FLAG_SESSION_LEAVE |
                nn::pia::common::TRACE_FLAG_SESSION_UPDATE;
#else
            u64 traceFlags = 0;
#endif
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlags, true);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            {
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID, static_cast<bool>(ENABLE_HOST_MIGRATION));
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
            {
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            {
                nn::pia::transport::Transport::Setting setting;
                setting.pFactory = pFactory;
                setting.maxStationNum = MAX_STATIONS;
                setting.maxSendThreadBufferNum = 16;
                setting.maxReceiveThreadBufferNum = 32;

#if PIA_TEST_CODE
                nn::pia::transport::Transport::DebugSetting debugSetting;
                debugSetting.sendThreadLatencyEmulationPacketNum = 128;
                debugSetting.receiveThreadLatencyEmulationPacketNum = 512;
                debugSetting.bPacketLossEmulation = true;
                result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#endif

                result = Transport_InitializePiaTransport(setting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }

            // Sets up sync.
            {
                nn::pia::sync::SyncProtocol::Setting setting;
                setting.m_MaxDelay = DELAY_MAX;
                setting.m_TimeoutFrame = 240;
                setting.m_DataUnitSize[DATA_ID_KEY] = sizeof(KeyData);
                setting.m_DataUnitSize[DATA_ID_TOUCH] = sizeof(TouchData);
                setting.m_DataUnitSize[DATA_ID_STICK] = sizeof(StickData);
#if PIA_TEST_CODE
                setting.m_DataUnitSize[DATA_ID_SYNC_CHECK] = sizeof(SyncCheckData);
#endif
                // Enables or disables use of the change delay feature while communication is synchronized.
                // If this is used, SyncProtocol::RequestToChangeDelay must be called every frame.
                // For notes on using this feature, see Advanced Features of PiaSync in the Pia Programming Manual: PiaSync Guide.
                setting.m_IsUsingChangeDelay = ENABLE_CHANGE_DELAY;

#if PIA_TEST_CODE
                setting.m_DataCompressionLevel = nn::pia::sync::SyncProtocol::COMPRESSION_LEVEL_NONE;
#else
                setting.m_DataCompressionLevel = nn::pia::sync::SyncProtocol::COMPRESSION_LEVEL_MIDDLE;
#endif

                Sync_InitializePiaSync(setting);
            }

            // Sets up the session.
            result = Session_InitializePiaSession(
                pFactory,
                nn::pia::session::NetworkTopology_FullMesh);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionCallback);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }


        // Connect to the network connections and create a session.
        {
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the session being created.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                        createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
                    searchCriteriaList[0].SetResultRange(0, 8);
                }
                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_BrowseSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connect to the Internet
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus asyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
                if (isHost)
                {
                    // Became the host as a result of matching.
                    gameMode = GAME_MODE_INET_HOST;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        s_MyUsingDataIdBitmap = USING_DATA_ID_BITMAP_ALL;
        s_Delay = 8;
        s_SendPeriod = FRAME_PER_SECONDS / 30;
        s_CursorConfigId = 0;
        s_ChangeDelay = 0;
        s_ChangeDelayInput = 0;
        s_ChangeSendPeriod = 1;

        // Calculate the maximum data size for each data ID to fit the synchronization data in one packet.
        {
            size_t size = Sync_GetSyncProtocol()->CalcMaxSyncDataSizeSentInOnePacket(s_Delay, DATA_ID_NUM);
            PIA_SAMPLE_PRINT("MaxSyncDataSizeSentInOnePacket = %d\n", size);
        }

        bool isLeaveNeeded = false;

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_HOST:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_HOST:
        case GAME_MODE_INET_CLIENT:
        {
            MainLoop(gameMode, &isLeaveNeeded);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }
        }
        break;
        }

        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            {
                if (isLeaveNeeded)
                {
                    // Leave the session.
                    // In this sample, the return value is ignored because finalization will take place later.
                    Session_LeaveSession();
                }
            }
            break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
            }
            break;
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_Logout();
                Inet_FinalizeNetwork();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, true);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetProgramParam(0x01, 0);
#endif

        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Sync_FinalizePiaSync();
            Transport_FinalizePiaTransport();
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
                Local_FinalizePiaLocal();
            }
            break;
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
            Common_FinalizePiaCommon();
        }
    }

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
