/*---------------------------------------------------------------------------*
  Project: Pia
  File: Shared_Local_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "Shared_Local_Setup.h"
#include "Shared_Basic_Setup.h"

#include <pia/local.h>

namespace
{
// Receiving buffer for UDS.
// The buffer must be 4096-byte aligned, and its size must be a multiple of 4096.
const u32 UDS_RECEIVE_BUFFER_SIZE = 4096 * 20;

// Buffer for UDS scanning.
// Approximately 1 KB per network is needed.
const u32 UDS_SCAN_BUFFER_SIZE = 5 * 1024;
nn::pia::local::UdsNetworkFactory s_NetworkFactory;

const char TEST_PASSPHRASE[] = "aaabbccc0123456789!";

u32 s_TitleId = 0;
u8 s_SubId = 0;
}

u32 Local_GetTitleId()
{
    return s_TitleId;
}

u8 Local_GetSubId()
{
    return s_SubId;
}

nn::Result Local_InitializePiaLocal(u32 titleId, u8 subId, const nn::pia::local::UdsNetworkSetting& networkSetting)
{
    PIA_SAMPLE_LOG("Local_InitializePiaLocal\n");

    s_TitleId = titleId;
    s_SubId = subId;

    nn::Result result;

    result = nn::pia::local::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::local::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::local::LocalNetwork::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::local::LocalNetwork::GetInstance()->Initialize(networkSetting);
    if (result.IsFailure())
    {
        // But even if it fails, it still calls nn::pia::local::EndSetup.
        nn::Result resultBuf = nn::pia::local::EndSetup();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(resultBuf);

        return result;
    }
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::local::LocalFacade::CreateInstance();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::local::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Local_InitializePiaLocal:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());

    return result;
}

nn::Result Local_InitializePiaLocal(u32 titleId, u8 subId, bool enableHostMigration, bool enableAroundNetworkSearch)
{
    // Structure for LocalNetwork settings.
    nn::pia::local::UdsNetworkSetting networkSetting;
    networkSetting.m_IsEnableHostMigration = enableHostMigration;  // Enables or disables the host migration feature.
    networkSetting.m_IsEnableAroundNetworkSearch = enableAroundNetworkSearch;  // Enables or disables the nearby network search feature.
    networkSetting.m_pUserName = NULL;                             // User name
    networkSetting.m_ReceiveBufferSize = UDS_RECEIVE_BUFFER_SIZE;  // Receive buffer size.
    networkSetting.m_pReceiveBuffer = NULL;                        // Receive buffer. If <tt>NULL</tt> is specified, a buffer is allocated within the library.
    networkSetting.m_ReceiveOption = nn::uds::NO_WAIT;             // Options when receiving. Basically, when using PiaSession, specify NO_WAIT.
    networkSetting.m_ScanBufferSize = UDS_SCAN_BUFFER_SIZE;        // Scan buffer size.
    networkSetting.m_SendOption = nn::uds::NO_WAIT;                // Options when sending.
    networkSetting.m_ApplicationVersion = 1;                       // Application version.

    return Local_InitializePiaLocal(titleId, subId, networkSetting);
}

void Local_FinalizePiaLocal()
{
    PIA_SAMPLE_LOG("Local_FinalizePiaLocal\n");

    // With Pia, Finalize and DestroyInstance always succeed.
    nn::pia::local::LocalFacade::DestroyInstance();
    if (PIA_IS_VALID_POINTER(nn::pia::local::LocalNetwork::GetInstance()))
    {
        nn::pia::local::LocalNetwork::GetInstance()->Finalize();
    }
    nn::pia::local::LocalNetwork::DestroyInstance();
    nn::pia::local::Finalize();
}

bool Local_CreateNetwork(size_t maxStations)
{
    u32 communicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(Local_GetTitleId());
    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
    createNetworkSetting.m_SubId = Local_GetSubId();
    createNetworkSetting.m_LocalCommunicationId = communicationId;
    createNetworkSetting.m_MaxEntry = maxStations;
    createNetworkSetting.m_Channel = 0;
    createNetworkSetting.m_ApplicationDataSize = 0;
    memcpy(createNetworkSetting.m_Passphrase, const_cast<char*>(TEST_PASSPHRASE), sizeof(TEST_PASSPHRASE));
    createNetworkSetting.m_PassphraseLength = sizeof(TEST_PASSPHRASE);
    return Local_CreateNetwork(&createNetworkSetting);
}

bool Local_CreateNetwork(nn::pia::local::LocalCreateNetworkSetting* pCreateNetworkSetting)
{
    nn::Result r;

    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
    nn::pia::common::Scheduler::GetInstance()->Dispatch();
    r = nn::pia::local::LocalNetwork::GetInstance()->CreateNetworkAsync(pCreateNetworkSetting);
    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("LocalNetwork::CreateNetworkAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        return false;
    }

    // Wait for CreateNetworkAsync to complete.
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "creating network...");

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedCreateNetwork())
        {
            r = nn::pia::local::LocalNetwork::GetInstance()->GetCreateNetworkResult();
            if (r.IsFailure())
            {
                PIA_SAMPLE_LOG("LocalNetwork::CreateNetworkAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(r);
                return false;
            }
            else
            {
                PIA_SAMPLE_LOG("LocalNetwork::CreateNetworkAsync succeeded.\n");
                break;
            }
        }
    }

    return true;
}

bool Local_SearchAndConnectToNetwork(nn::pia::transport::StationConnectionInfo* pStationConnectionInfo)
{
    nn::Result r;
    nn::pia::local::UdsNetworkDescription description;
    size_t gatheringNum = 0;
    u32 communicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(Local_GetTitleId());

    u32 searchRetryCnt = 0;
    const u32 searchRetryCntMax = 10;

    // Network search.
    bool isAsyncStarted = false;
    while (true)
    {
        bool isExit = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            // The search is ended if START is pressed.
            isExit = true;
            if (isAsyncStarted)
            {
                r = nn::pia::local::LocalNetwork::GetInstance()->CancelScanNetworkAsync();
                PIA_SAMPLE_LOG("Local_SearchAndConnectToNetwork canceled.\n");
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
        }

        if (isExit == false)
        {
            if (isAsyncStarted == false)
            {
                r = nn::pia::local::LocalNetwork::GetInstance()->ScanNetworkAsync(communicationId, Local_GetSubId());
                if (r.IsSuccess())
                {
                    isAsyncStarted = true;
                }
            }
            else if (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedScanNetwork())
            {
                r = nn::pia::local::LocalNetwork::GetInstance()->GetScanNetworkResult();
                if (r.IsSuccess())
                {
                    r = nn::pia::local::LocalNetwork::GetInstance()->GetNetworkDescriptionList(&description, &gatheringNum, 1);
                    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
                    if (gatheringNum > 0)
                    {
                        // Found a network.
                        isExit = true;
                    }
                    else
                    {
                        ++searchRetryCnt;
                        if (searchRetryCnt > searchRetryCntMax)
                        {
                            PIA_SAMPLE_LOG("Local_SearchAndConnectToNetwork timeout\n");
                            isExit = true;
                        }
                    }
                }
                else
                {
                    gatheringNum = 0;

                    if (r == nn::pia::ResultInvalidConnection())
                    {
                        // LocalNetwork must be reinitialized.
                        isExit = true;
                    }
                }

                isAsyncStarted = false;
            }
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "searching Local network...");
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 30, "START: cancel");
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            break;
        }
    }

    // Wait for asynchronous processes to complete.
    while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedScanNetwork() == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "waiting ScanNetwork completed...");
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }

    if (gatheringNum <= 0)
    {
        // Failure if leaving even though a network could not be found.
        return false;
    }

    // Join the network.
    nn::pia::local::UdsConnectNetworkSetting connectNetworkSetting;
    memcpy(connectNetworkSetting.m_Passphrase, const_cast<char*>(TEST_PASSPHRASE), sizeof(TEST_PASSPHRASE));
    connectNetworkSetting.m_PassphraseLength = sizeof(TEST_PASSPHRASE);
    connectNetworkSetting.m_pLocalNetworkDescription = &description;
    bool isSuccess = Local_ConnectNetwork(&connectNetworkSetting);
    if (isSuccess == false)
    {
        return false;
    }

    // Start of LocalFacade. (The local network must have been built when this API is called.)
    if (nn::pia::local::LocalNetwork::GetInstance()->IsConnected() == false)
    {
        PIA_SAMPLE_LOG("LocalNetwork::IsConnected() == false\n");
        return false;
    }
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::local::LocalFacade::GetInstance()));
    r = nn::pia::local::LocalFacade::GetInstance()->Startup();
    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("LocalFacade::Startup failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        PIA_SAMPLE_ASSERT(false);
        return false;
    }

    //Gets the <tt>StationConnectionInfo</tt> for the session host.
    r = nn::pia::local::LocalFacade::GetInstance()->GetHostStationConnectionInfo(pStationConnectionInfo);
    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("LocalFacade::GetHostStationConnectionInfo failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        PIA_SAMPLE_ASSERT(false);
        return false;
    }

    return true;
}

bool Local_ScanNetwork(u32 localCommunicationId, u8 subId)
{
    nn::Result r;
    r = nn::pia::local::LocalNetwork::GetInstance()->ScanNetworkAsync(localCommunicationId, subId);
    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("ScanNetworkAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        return false;
    }

    while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedScanNetwork() == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            // Cancels the network search.
            r = nn::pia::local::LocalNetwork::GetInstance()->CancelScanNetworkAsync();
            if (r.IsSuccess())
            {
                PIA_SAMPLE_LOG("CancelScanNetworkAsync succeeded.\n");
            }
            else
            {
                PIA_SAMPLE_LOG("CancelScanNetworkAsync Failure\n");
            }
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "searching...");
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }

    r = nn::pia::local::LocalNetwork::GetInstance()->GetScanNetworkResult();
    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("ScanNetworkAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        return false;
    }

    return true;
}

bool Local_ConnectNetwork(nn::pia::local::LocalConnectNetworkSetting* pConnectNetworkSetting)
{
    nn::Result r;

    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
    nn::pia::common::Scheduler::GetInstance()->Dispatch();
    r = nn::pia::local::LocalNetwork::GetInstance()->ConnectNetworkAsync(pConnectNetworkSetting);
    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("ConnectNetworkAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        return false;
    }

    while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedConnectNetwork() == false)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        if (nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            // Cancel the attempt to join the network.
            r = nn::pia::local::LocalNetwork::GetInstance()->CancelConnectNetworkAsync();
            if (r.IsSuccess())
            {
                PIA_SAMPLE_LOG("CancelConnectNetworkAsync succeeded.\n");
            }
            else
            {
                PIA_SAMPLE_LOG("CancelConnectNetworkAsync Failure\n");
            }
        }
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "connecting...");
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }

    r = nn::pia::local::LocalNetwork::GetInstance()->GetConnectNetworkResult();
    if (r.IsFailure())
    {
        PIA_SAMPLE_LOG("ConnectNetworkAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(r);
        return false;
    }

    return true;
}

bool Local_DisconnectNetwork(bool isMigration)
{
    if (!PIA_IS_VALID_POINTER(nn::pia::local::LocalNetwork::GetInstance()))
    {
        return false;
    }

    if (!nn::pia::local::LocalNetwork::GetInstance()->IsConnected())
    {
        return false;
    }

    if (nn::pia::local::LocalNetwork::GetInstance()->IsHost())  //For the host.
    {
        PIA_SAMPLE_LOG("IsHost\n");
        if (isMigration)
        {
            // Performs manual host migration.
            nn::Result result = nn::pia::local::LocalNetwork::GetInstance()->DisconnectNetworkAsync();
            if (result.IsSuccess())
            {
                while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedDisconnectNetwork() == false)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                    nn::pia::common::Scheduler::GetInstance()->Dispatch();
                    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                }
                result = nn::pia::local::LocalNetwork::GetInstance()->GetDisconnectNetworkResult();
                return result.IsSuccess();
            }
            else
            {
                return false;
            }
        }
        else
        {
            // Destroys network.
            nn::Result result = nn::pia::local::LocalNetwork::GetInstance()->DestroyNetworkAsync();
            if (result.IsSuccess())
            {
                while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedDestroyNetwork() == false)
                {
                    nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                    nn::pia::common::Scheduler::GetInstance()->Dispatch();
                    nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                }
                result = nn::pia::local::LocalNetwork::GetInstance()->GetDestroyNetworkResult();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("DestroyNetwork failure\n");
                }
                return result.IsSuccess();
            }
            else
            {
                return false;
            }
        }
    }
    else if (nn::pia::local::LocalNetwork::GetInstance()->IsClient())  //For clients.
    {
        PIA_SAMPLE_LOG("IsClient\n");
        nn::Result result = nn::pia::local::LocalNetwork::GetInstance()->DisconnectNetworkAsync();
        if (result.IsSuccess())
        {
            while (nn::pia::local::LocalNetwork::GetInstance()->IsCompletedDisconnectNetwork() == false)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                nn::pia::common::Scheduler::GetInstance()->Dispatch();
                nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
            }
            result = nn::pia::local::LocalNetwork::GetInstance()->GetDisconnectNetworkResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("DisconnectNetwork failure\n");
            }
            return result.IsSuccess();
        }
        else
        {
            return false;
        }
    }

    return true;
}

nn::Result Local_StartupPiaLocal()
{
    if (!PIA_IS_VALID_POINTER(nn::pia::local::LocalNetwork::GetInstance()))
    {
        PIA_RETURN_RESULT(nn::pia::ResultInvalidState);
    }

    return nn::pia::local::LocalNetwork::GetInstance()->Startup();
}

void Local_CleanupPiaLocal()
{
    if (PIA_IS_VALID_POINTER(nn::pia::local::LocalNetwork::GetInstance()))
    {
        nn::pia::local::LocalNetwork::GetInstance()->Cleanup();
    }
}

void Local_CleanupNetwork()
{
    // Clean up LocalFacade.
    if (PIA_IS_VALID_POINTER(nn::pia::local::LocalFacade::GetInstance()))
    {
        nn::pia::local::LocalFacade::GetInstance()->Cleanup();
    }
}

nn::pia::transport::NetworkFactory* Local_GetNetworkFactory()
{
    return &s_NetworkFactory;
}

bool Local_IsConnected()
{
    if (PIA_IS_VALID_POINTER(nn::pia::local::LocalNetwork::GetInstance()))
    {
        // Returns <tt>true</tt> during host migration.
        return nn::pia::local::LocalNetwork::GetInstance()->IsConnected() || nn::pia::local::LocalNetwork::GetInstance()->IsDuringHostMigration();
    }
    else
    {
        return false;
    }
}
