﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Demo.h"


#if NN_PIA_CTR
#define USE_PIA_LOCAL 1
#else
#define USE_PIA_LOCAL 0
#endif

#if USE_PIA_LOCAL
#include <pia/local.h>
#include "Shared_Local_Setup.h"
#endif

#include <pia/util.h>

#include <cstdio>

#if NN_PIA_WIN
#define snprintf(...) _snprintf(__VA_ARGS__)
#endif

#ifndef ENABLE_PROFILING
#define ENABLE_PROFILING 0
#endif

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

#define PROGRAM_TITLE "ReliableAndUnreliable"


namespace
{
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 30000;  // Use different values here for each Pia sample program.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;

const size_t MAX_STATIONS = 4;  // Defines the maximum number of stations.

#if USE_PIA_LOCAL
const u32 SAMPLE_TITLE_ID = 0xFF300;  // For the transport sample.
const u8 SAMPLE_SUB_ID = 0x00;
const char SAMPLE_PASSPHRASE[] = "aaabbccc0123456789";
const char SAMPLE_SIGNATURE_KEY_DATA[] = "ReliableAndUnreliable_key";
#endif

// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const u32 SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const u32 SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

#define SAMPLE_NEX_THREAD_MODE nn::nex::Core::ThreadModeUnsafeTransportBuffer

const size_t LARGE_BUFFER_SIZE = 1 * 1024 * 1024;  // The maximum size when sending and receiving the large data with ReliableLarge.
static bit32 s_LargeBuffer[LARGE_BUFFER_SIZE / sizeof(bit32)];
const size_t LARGE_DATA_OFFSET_MAX = 1 * 1024 * 1024;
static bit32 s_BasicLargeData[(LARGE_DATA_OFFSET_MAX + LARGE_BUFFER_SIZE) / sizeof(bit32)];

const size_t MEMSIZE = 10 * 1024 * 1024;             // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.


const u16 UNRELIABLE_PROTOCOL_PORT = 0;      // Defines the port number used by UnreliableProtocol.
const u16 RELIABLE_PROTOCOL_PORT = 0;        // Defines the port number used by ReliableProtocol.
const u16 RELIABLE_PROTOCOL_LARGE_PORT = 1;  // Defines the port number used by <tt>ReliableProtocol</tt> for sending a large amount of data.

u32 s_hUnreliable = 0;
u32 s_hReliable = 0;
u32 s_hReliableLarge = 0;

#if PIA_TEST_CODE
struct OptionalParameter s_OptionalParam;
const s32 OPTIONAL_PARAM_OFFSET = 200;
#endif

nn::pia::StationId conv2StationId(nn::pia::StationIndex idx)
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    if (!PIA_IS_VALID_POINTER(pTransport))
    {
        PIA_SAMPLE_ASSERTMSG(false, "conv2StationId(): Transport instance is not ready?");
        return nn::pia::STATION_ID_INVALID;
    }

    nn::pia::StationId id;
    nn::Result ret = pTransport->ConvertToStationId(&id, idx);
    if (ret.IsFailure())
    {
        if (ret==nn::pia::ResultNotFound())
        {
            return nn::pia::STATION_ID_INVALID;
        }
        else
        {
            PIA_SAMPLE_PRINT("Result = 0x%08x\n", ret.GetValue());
            PIA_SAMPLE_ASSERTMSG(false, "conv2StationId(): ConvertToStationId() failed.");
            return nn::pia::STATION_ID_INVALID;
        }
    }

    return id;
}


nn::pia::StationIndex conv2StationIndex(nn::pia::StationId id)
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    if (!PIA_IS_VALID_POINTER(pTransport))
    {
        PIA_SAMPLE_ASSERTMSG(false, "conv2StationIndex: Transport instance is not ready?");
        return nn::pia::STATION_INDEX_INVALID;
    }

    nn::pia::StationIndex idx;
    nn::Result ret = pTransport->ConvertToStationIndex(&idx, id);
    if (ret.IsFailure())
    {
        if (ret==nn::pia::ResultNotFound())
        {
            return nn::pia::STATION_INDEX_INVALID;
        }
        else
        {
            PIA_SAMPLE_PRINT("Result = 0x%08x\n", ret.GetValue());
            PIA_SAMPLE_ASSERTMSG(false, "conv2StationIndex: ConvertToStationIndex() failed.");
            return nn::pia::STATION_INDEX_INVALID;
        }
    }

    return idx;
}


// Create and initialize protocols.
nn::Result CreateAndInitializeProtocols()
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Create and initialize an UnreliableProtocol instance. (The arrival of data is not guaranteed.)
    {
        PIA_SAMPLE_ASSERT(s_hUnreliable == 0);
        s_hUnreliable = pTransport->CreateProtocol<nn::pia::transport::UnreliableProtocol>(UNRELIABLE_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hUnreliable != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        pTransport->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable)->Initialize();
    }

    // Create and initialize a ReliableProtocol instance. (Arrival of data is guaranteed.)
    {
        PIA_SAMPLE_ASSERT(s_hReliable == 0);
        s_hReliable = pTransport->CreateProtocol<nn::pia::transport::ReliableProtocol>(RELIABLE_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hReliable != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        const u32 SEND_BUF_NUM = 64;
        const u32 RECV_BUF_NUM = 64;
        pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable)->Initialize(SEND_BUF_NUM, RECV_BUF_NUM);
    }

    // Create and initialize a <tt>ReliableProtocolLarge</tt> instance. (Arrival of data is guaranteed.)
    {
        PIA_SAMPLE_ASSERT(s_hReliableLarge == 0);
        s_hReliableLarge = pTransport->CreateProtocol<nn::pia::transport::ReliableProtocol>(RELIABLE_PROTOCOL_LARGE_PORT);
        PIA_SAMPLE_ASSERT(s_hReliableLarge != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        const u32 SEND_BUF_NUM = 1000;
        const u32 RECV_BUF_NUM = 1000;
        pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliableLarge)->Initialize(SEND_BUF_NUM, RECV_BUF_NUM);
    }

    return nn::ResultSuccess();
}


// Clean up protocols.
void FinalizeAndDestroyProtocols()
{
    // Get the Transport singleton.
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Finalize <tt>ReliableProtocolLarge</tt>.
    nn::pia::transport::ReliableProtocol* pReliableLarge = pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliableLarge);
    if (PIA_IS_VALID_POINTER(pReliableLarge))
    {
        pReliableLarge->Finalize();
        pTransport->DestroyProtocol(s_hReliableLarge);
        s_hReliableLarge = 0;
    }

    // Finalize ReliableProtocol.
    nn::pia::transport::ReliableProtocol* pReliable = pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable);
    if (PIA_IS_VALID_POINTER(pReliable))
    {
        pReliable->Finalize();
        pTransport->DestroyProtocol(s_hReliable);
        s_hReliable = 0;
    }

    // Finalize UnreliableProtocol
    nn::pia::transport::UnreliableProtocol* pUnreliable = pTransport->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable);
    if (PIA_IS_VALID_POINTER(pUnreliable))
    {
        pUnreliable->Finalize();
        pTransport->DestroyProtocol(s_hUnreliable);
        s_hUnreliable = 0;
    }
}


nn::pia::transport::UnreliableProtocol* GetUnreliableProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable);
}


nn::pia::transport::ReliableProtocol* GetReliableProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable);
}


nn::pia::transport::ReliableProtocol* GetReliableProtocolLarge()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliableLarge);
}


// Creates source data for large data.
void InitBasicLargeData()
{
    std::srand(0x12345678);
    u32 size = sizeof(s_BasicLargeData) / sizeof(bit32);
    for (u32 i = 0; i < size; ++i)
    {
        s_BasicLargeData[i] = static_cast<bit32>(std::rand());
    }
}


// Creates the large data. The content has no particular meaning.
u32 CreateLargeData(u32 key)
{
    // The first 4 bytes are the key itself. Based on that, the size and remaining data are determined by random numbers.

    std::srand(key);

    u32 size;

#if PIA_TEST_CODE
    size = 128 * 1024;
    PIA_SAMPLE_ASSERT(size < LARGE_BUFFER_SIZE);
#else
#if ENABLE_PROFILING
    if (true)
#else
    if (std::rand() & 0x1)
#endif
    {
        size = LARGE_BUFFER_SIZE;
    }
    else
    {
        size = (std::rand() * sizeof(u32)) % LARGE_BUFFER_SIZE;
        if (size <= 0)
        {
            size = LARGE_BUFFER_SIZE;
        }
    }
#endif

    u32 offset = (std::rand() * sizeof(u32)) % LARGE_DATA_OFFSET_MAX;
    const void* cpData = reinterpret_cast<void*>(reinterpret_cast<u32>(s_BasicLargeData) + offset);
    std::memcpy(s_LargeBuffer, cpData, size);
    s_LargeBuffer[0] = key;

    return size;
}

// Checks the content of the large data.
bool CheckLargeData(u32 dataSize)
{
    if (dataSize <= 0)
    {
        return false;
    }
    if (dataSize % 4 != 0)
    {
        return false;
    }

    // The first 4 bytes are taken as the key, and then it checks the size and remaining data calculated from random numbers.

    u32 key = s_LargeBuffer[0];
    std::srand(key);

    u32 size;
#if PIA_TEST_CODE
    size = 128 * 1024;
    PIA_SAMPLE_ASSERT(size < LARGE_BUFFER_SIZE);
#else
#if ENABLE_PROFILING
    if (true)
#else
    if (std::rand() & 0x1)
#endif
    {
        size = LARGE_BUFFER_SIZE;
    }
    else
    {
        size = (std::rand() * sizeof(u32)) % LARGE_BUFFER_SIZE;
        if (size <= 0)
        {
            size = LARGE_BUFFER_SIZE;
        }
    }
#endif

    if (size != dataSize)
    {
        return false;
    }

    u32 offset = (std::rand() * sizeof(u32)) % LARGE_DATA_OFFSET_MAX;
    const void* cpData = reinterpret_cast<void*>(reinterpret_cast<u32>(s_BasicLargeData) + offset + sizeof(bit32));
    return (std::memcmp(&(s_LargeBuffer[1]), cpData, size - sizeof(bit32)) == 0);
}


// Enumerated type to distinguish <tt>Protocol</tt>.
enum ProtocolType
{
    PROTOCOL_UNRELIABLE,
    PROTOCOL_RELIABLE,
    PROTOCOL_RELIABLE_LARGE,
    PROTOCOL_NUM
};

static nn::pia::common::Time s_CurrentTime;

// Data managing the send/receive state.
class StationData
{
public:
    nn::pia::StationId m_StationId;
    struct ProtocolData
    {
        bool m_IsSendRepeat;
        bool m_IsSendOnce;
        u32 m_SendNum;
        u32 m_ReceiveNum;
        nn::pia::common::Time m_LastReceivedTime;
    };
    ProtocolData m_ProtocolData[PROTOCOL_NUM];

    void Clear()
    {
        for (int i = 0; i < PROTOCOL_NUM; ++i)
        {
            m_ProtocolData[i].m_IsSendRepeat = false;
            m_ProtocolData[i].m_IsSendOnce = false;
            m_ProtocolData[i].m_SendNum = 0;
            m_ProtocolData[i].m_ReceiveNum = 0;
            m_ProtocolData[i].m_LastReceivedTime = s_CurrentTime;
        }
    }
};

static const u32 STATION_DATA_NUM = MAX_STATIONS + 1;
static const int STATION_ALL_IDX = MAX_STATIONS;
static StationData s_StationData[STATION_DATA_NUM];
static int s_TargetStationIdx;
static int s_TargetProtocolType;
static bool s_IsLogEnable;

#if ENABLE_PROFILING
static bool s_IsCountReceive = false;
static u32 s_ReceivedSize = 0;
static u32 s_ReceivedTime = 0;
#endif // end of ENABLE_PROFILING


// Initializes the data for send/receive states.
void InitializeStationData()
{
    for (int i = 0; i < STATION_DATA_NUM; ++i)
    {
        s_StationData[i].m_StationId = nn::pia::STATION_ID_INVALID;
        s_StationData[i].Clear();
    }
    s_StationData[STATION_ALL_IDX].m_StationId = nn::pia::STATION_ID_ALL;
    s_StationData[STATION_ALL_IDX].Clear();

    s_TargetProtocolType = 0;
    s_TargetStationIdx = STATION_ALL_IDX;
    s_IsLogEnable = true;
}


// Process during connection and disconnection.
void SessionEventCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    nn::pia::transport::TransportAnalyzer* pAnalyzer = nn::pia::transport::TransportAnalyzer::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pAnalyzer));

    nn::Result r;
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
        s_StationData[conv2StationIndex(stationId)].Clear();
        s_StationData[conv2StationIndex(stationId)].m_StationId = stationId;
        r = pAnalyzer->SetMarker("JoinEvent");
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        break;

    case nn::pia::session::Session::EVENT_LEAVE:
        if (stationId == s_StationData[s_TargetStationIdx].m_StationId)
        {
            // The send destination where the cursor is located is disconnected. Move the cursor to the All location.
            s_TargetStationIdx = STATION_ALL_IDX;
        }
        r = pAnalyzer->SetMarker("LeaveEvent");
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
        // Session host privileges to the session have been migrated.
        if (stationId == s_StationData[s_TargetStationIdx].m_StationId)
        {
            // The send destination where the cursor is located is disconnected. Move the cursor to the All location.
            s_TargetStationIdx = STATION_ALL_IDX;
        }
        r = pAnalyzer->SetMarker("HostChangedEvent");
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        break;
    case nn::pia::session::Session::EVENT_HOST_MIGRATION_FAILED:
        // Host migration fails.
        r = pAnalyzer->SetMarker("HostMigrationFailedEvent");
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        break;
    default:
        PIA_SAMPLE_ASSERTMSG(false, "Unsupported event(%d).", eventType);
        break;
    }
}


// Operation by key input.
bool Operate()
{
    bool isExit = false;

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    // Switches the send destination of the operation target with the Up/Down Buttons.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP) || pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
    {
        int mlt = pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP) ? -1 : 1;
        nn::pia::StationId localStationId = pSession->GetLocalStationId();
        for (int i = 1; i < STATION_DATA_NUM; i++)
        {
            int idx = (s_TargetStationIdx + i * mlt + STATION_DATA_NUM) % STATION_DATA_NUM;
            nn::pia::StationId stationId = s_StationData[idx].m_StationId;
            if ((stationId != localStationId && pSession->CheckStationIdIsValid(stationId)) || stationId == nn::pia::STATION_ID_ALL)
            {
                s_TargetStationIdx = idx;
                break;
            }
        }
    }

    // Switches the Protocol of the operation target with the Left/Right Buttons.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
    {
        --s_TargetProtocolType;
        if (s_TargetProtocolType < 0)
        {
            s_TargetProtocolType = PROTOCOL_NUM - 1;
        }
    }
    else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
    {
        ++s_TargetProtocolType;
        if (s_TargetProtocolType >= PROTOCOL_NUM)
        {
            s_TargetProtocolType = 0;
        }
    }

    // Sets the one-time send flag with the A Button.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
    {
        s_StationData[s_TargetStationIdx].m_ProtocolData[s_TargetProtocolType].m_IsSendOnce = true;
    }

    // Enables or disables the continuous send flag with the X Button.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
    {
        s_StationData[s_TargetStationIdx].m_ProtocolData[s_TargetProtocolType].m_IsSendRepeat ^= true;
    }

    // Enables or disables the log output during sending and receiving with the Y Button.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
    {
        s_IsLogEnable ^= true;
    }

    // Exit the GameLoop function if START is pressed.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        isExit = true;
    }

#if ENABLE_PROFILING
    // Toggle throughput metering with the L Button.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L))
    {
        if (s_IsCountReceive)
        {
            s_IsCountReceive = false;
        }
        else
        {
            s_IsCountReceive = true;
            s_ReceivedSize = 0;
            s_ReceivedTime = 0;
        }
    }
#endif

    return isExit;
}


// <tt>UnreliableProtocol</tt> send processing.
void SendUnreliable(nn::pia::transport::UnreliableProtocol* pUnreliable)
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    for (int i = 0; i < STATION_DATA_NUM; ++i)
    {
        nn::pia::StationId stationId = s_StationData[i].m_StationId;
        if (stationId == pSession->GetLocalStationId())
        {
            continue;
        }

        StationData::ProtocolData* pProtocolData = &s_StationData[i].m_ProtocolData[PROTOCOL_UNRELIABLE];
        if (pProtocolData->m_IsSendOnce || pProtocolData->m_IsSendRepeat)
        {
            nn::Result r;
            if (stationId == nn::pia::STATION_ID_ALL)
            {
                // Send to everyone who is connected.
                // Sends data of the maximum size that can be sent with <tt>UnreliableProtocol</tt>.
                static const u32 BUFFER_SIZE = nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE;
                char buffer[BUFFER_SIZE];
                memset(buffer, 0xff, BUFFER_SIZE);
                int size = snprintf(buffer, BUFFER_SIZE, "Unreliable Broadcast %llu => All (No.%u)", pSession->GetLocalStationId().GetValue(), pProtocolData->m_SendNum);
                size_t limit = pUnreliable->GetDataSizeLimit();
                PIA_SAMPLE_ASSERT(size < static_cast<int>(limit));
                NN_PIA_DUMMY_PARAM_IN_RELEASE(size);

                // The actual send process.
                r = pUnreliable->SendToAll(reinterpret_cast<bit8*>(buffer), limit);
            }
            else
            {
                // Sends to a specified unit.
                // Sends data of a size of the created string.
                if (pSession->CheckStationIdIsValid(stationId) == false)
                {
                    continue;
                }
                static const u32 BUFFER_SIZE = 128;
                char buffer[BUFFER_SIZE];
                int size = snprintf(buffer, BUFFER_SIZE, "Unreliable Unicast %llu => %llu (No.%u)", pSession->GetLocalStationId().GetValue(), stationId.GetValue(), pProtocolData->m_SendNum);
                PIA_SAMPLE_ASSERT(size < BUFFER_SIZE);

                // The actual send process.
                r = pUnreliable->Send(stationId, reinterpret_cast<bit8*>(buffer), size + 1);
            }

            if (r.IsSuccess())
            {
                // Send succeeded, so it counts the number of sends.
                ++(pProtocolData->m_SendNum);
                pProtocolData->m_IsSendOnce = false;
            }
            else
            {
                // Other errors require the program fixes.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
        }
    }
}


// <tt>ReliableProtocol</tt> send processing.
void SendReliable(nn::pia::transport::ReliableProtocol* pReliable)
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    for (int i = 0; i < STATION_DATA_NUM; ++i)
    {
        nn::pia::StationId stationId = s_StationData[i].m_StationId;
        if (stationId == pSession->GetLocalStationId())
        {
            continue;
        }

        // For transmissions to individuals, send after confirming that communicating with the send destination is occurring.
        if (stationId != nn::pia::STATION_ID_ALL && pReliable->IsInCommunication(stationId) == false)
        {
            continue;
        }

        StationData::ProtocolData* pProtocolData = &s_StationData[i].m_ProtocolData[PROTOCOL_RELIABLE];
        if (pProtocolData->m_IsSendOnce || pProtocolData->m_IsSendRepeat)
        {
            // Sends with the same interface for both individual and send to all transmissions.
            static const u32 BUFFER_SIZE = 128;
            char buffer[BUFFER_SIZE];
            int size = snprintf(buffer, BUFFER_SIZE, "Reliable %llu => %llu (No.%u)", pSession->GetLocalStationId().GetValue(), stationId.GetValue(), pProtocolData->m_SendNum);
            PIA_SAMPLE_ASSERT(size < BUFFER_SIZE);

            // The actual send process.
            nn::Result r = pReliable->Send(stationId, buffer, size + 1);

            if (r.IsSuccess())
            {
                // Send succeeded, so it counts the number of sends.
                ++(pProtocolData->m_SendNum);
                pProtocolData->m_IsSendOnce = false;
            }
            else if (r == nn::pia::ResultBufferIsFull())
            {
                // The send buffer is full.
                // The <tt>Once</tt> flag is also left enabled, to send again next time.
                if (s_IsLogEnable)
                {
                    PIA_SAMPLE_LOG("ReliableProtocol BufferIsFull to:%llu No:%u\n", stationId.GetValue(), pProtocolData->m_SendNum);
                }
            }
            else if (r == nn::pia::ResultNotInCommunication())
            {
                // The program does not run this process if you check in advance with <tt>IsInCommunication</tt>.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
            else
            {
                // Other errors require the program fixes.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
        }
    }
}


// Send processing for large data with <tt>ReliableProtocol</tt>. The same as the <tt>SendReliable</tt> function, except for the method to create the send data.
void SendReliableLarge(nn::pia::transport::ReliableProtocol* pReliableLarge)
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    for (int i = 0; i < STATION_DATA_NUM; ++i)
    {
        nn::pia::StationId stationId = s_StationData[i].m_StationId;
        if (stationId == pSession->GetLocalStationId())
        {
            continue;
        }

        // For transmissions to individuals, send after confirming that communicating with the send destination is occurring.
        if (stationId != nn::pia::STATION_ID_ALL && pReliableLarge->IsInCommunication(stationId) == false)
        {
            continue;
        }

        StationData::ProtocolData* pProtocolData = &s_StationData[i].m_ProtocolData[PROTOCOL_RELIABLE_LARGE];
        if (pProtocolData->m_IsSendOnce || pProtocolData->m_IsSendRepeat)
        {
            nn::pia::StationIndex targetIdx = conv2StationIndex(stationId);
            PIA_SAMPLE_ASSERT(targetIdx != nn::pia::STATION_INDEX_INVALID);
            nn::pia::StationIndex localIdx  = conv2StationIndex(pSession->GetLocalStationId());
            PIA_SAMPLE_ASSERT(localIdx != nn::pia::STATION_INDEX_INVALID);

            // Sends with the same interface for both individual and send to all transmissions.
            u32 key = (static_cast<u32>(targetIdx) << 24) | (static_cast<u32>(localIdx) << 16) | (pProtocolData->m_SendNum & 0xffff);
            u32 size = CreateLargeData(key);
            PIA_SAMPLE_ASSERT(CheckLargeData(size));

            // The actual send process.
            nn::Result r = pReliableLarge->Send(stationId, s_LargeBuffer, size);

            if (r.IsSuccess())
            {
                // Send succeeded, so it counts the number of sends.
                ++(pProtocolData->m_SendNum);
                pProtocolData->m_IsSendOnce = false;
            }
            else if (r == nn::pia::ResultBufferIsFull())
            {
                // The send buffer is full.
                // The <tt>Once</tt> flag is also left enabled, to send again next time.
                // Because it takes time until there is room in the send buffer, do not print.
            }
            else if (r == nn::pia::ResultNotInCommunication())
            {
                // The program does not run this process if you check in advance with <tt>IsInCommunication</tt>.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
            else
            {
                // Other errors require the program fixes.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
            }
        }
    }
}


// <tt>UnreliableProtocol</tt> receive processing.
void ReceiveUnreliable(nn::pia::transport::UnreliableProtocol* pUnreliable)
{
    while (true)
    {
        static const u32 BUFFER_SIZE = nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE;
        char buffer[BUFFER_SIZE];
        nn::pia::StationId stationId;
        size_t receivedSize;

        // The actual receive process.
        nn::Result r = pUnreliable->Receive(&stationId, reinterpret_cast<bit8*>(buffer), &receivedSize, BUFFER_SIZE);

        if (r.IsSuccess())
        {
            // Because the receive data was obtained, it counts the number of receptions.
            StationData::ProtocolData* pProtocolData = &s_StationData[conv2StationIndex(stationId)].m_ProtocolData[PROTOCOL_UNRELIABLE];
            if (s_IsLogEnable)
            {
                nn::pia::common::TimeSpan span = s_CurrentTime - pProtocolData->m_LastReceivedTime;
                NN_PIA_DUMMY_PARAM_IN_RELEASE(span);
                PIA_SAMPLE_LOG("Unreliable Received from %llu. Data(%d bytes; span %d ms): [%s]\n", stationId.GetValue(), receivedSize, span.GetMSec(), buffer);
            }
            ++(pProtocolData->m_ReceiveNum);
            pProtocolData->m_LastReceivedTime = s_CurrentTime;
        }
        else if (r == nn::pia::ResultNoData())
        {
            // Because there is no more receive data, the receive process ends.
            break;
        }
        else
        {
            // Other errors require the program fixes.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
    }
}


// <tt>ReliableProtocol</tt> receive processing.
// Returns <tt>true</tt> if communications are in a state where continuing is not possible.
bool ReceiveReliable(nn::pia::transport::ReliableProtocol* pReliable)
{
    while (true)
    {
        static const u32 BUFFER_SIZE = 128;
        char buffer[BUFFER_SIZE];
        nn::pia::StationId stationId;
        size_t receivedSize;

        // The actual receive process.
        nn::Result r = pReliable->Receive(&stationId, reinterpret_cast<bit8*>(buffer), &receivedSize, BUFFER_SIZE);

        if (r.IsSuccess())
        {
            // Because the receive data was obtained, it counts the number of receptions.
            StationData::ProtocolData* pProtocolData = &s_StationData[conv2StationIndex(stationId)].m_ProtocolData[PROTOCOL_RELIABLE];
            if (s_IsLogEnable)
            {
                nn::pia::common::TimeSpan span = s_CurrentTime - pProtocolData->m_LastReceivedTime;
                NN_PIA_DUMMY_PARAM_IN_RELEASE(span);
                PIA_SAMPLE_LOG("Reliable Received from %llu. Data(%d bytes; span %d ms): [%s]\n", stationId.GetValue(), receivedSize, span.GetMSec(), buffer);
            }
            ++(pProtocolData->m_ReceiveNum);
            pProtocolData->m_LastReceivedTime = s_CurrentTime;
        }
        else if (r == nn::pia::ResultNoData())
        {
            // Because there is no more receive data, the receive process ends.
            break;
        }
        else if (r == nn::pia::ResultBrokenData())
        {
            // Data of an unexpected size arrived.
            // When in this state, receiving with this <tt>ReliableProtocol</tt> cannot continue, so communications are terminated.
            return true;
        }
        else if (r == nn::pia::ResultNotInCommunication())
        {
            // Does not go here if participating in a session.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
        else
        {
            // Other errors require the program fixes.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
    }

    return false;
}


// Receive processing for large data with <tt>ReliableProtocol</tt>.
// Returns <tt>true</tt> if communications are in a state where continuing is not possible.
// The same as the <tt>ReceiveReliable function</tt>, except the handling of the receive data is different.
bool ReceiveReliableLarge(nn::pia::transport::ReliableProtocol* pReliableLarge)
{
    while (true)
    {
        nn::pia::StationId stationId;
        size_t receivedSize;

        // The actual receive process.
        nn::Result r = pReliableLarge->Receive(&stationId, reinterpret_cast<bit8*>(s_LargeBuffer), &receivedSize, LARGE_BUFFER_SIZE);
        if (r.IsSuccess())
        {
            // Because the receive data was obtained, it counts the number of receptions.
            bool isValidData = CheckLargeData(receivedSize);
            StationData::ProtocolData* pProtocolData = &s_StationData[conv2StationIndex(stationId)].m_ProtocolData[PROTOCOL_RELIABLE_LARGE];
            nn::pia::common::TimeSpan span = s_CurrentTime - pProtocolData->m_LastReceivedTime;
            NN_PIA_DUMMY_PARAM_IN_RELEASE(span);
            if (s_IsLogEnable)
            {
                PIA_SAMPLE_LOG("ReliableLarge Received from %llu. Data(%d bytes; span %dms): %s\n", stationId.GetValue(), receivedSize, span.GetMSec(), isValidData ? "OK" : "NG");
            }
            NN_PIA_DUMMY_PARAM_IN_RELEASE(isValidData);

            // Data that does not pass the check should not exist as long as communication is done between samples and an assert is used.
            PIA_SAMPLE_ASSERTMSG(isValidData, "received data is broken");

            ++(pProtocolData->m_ReceiveNum);
            pProtocolData->m_LastReceivedTime = s_CurrentTime;

#if ENABLE_PROFILING
            if (s_IsCountReceive)
            {
                s_ReceivedSize += receivedSize;
                s_ReceivedTime += span.GetMSec();
            }
#endif
        }
        else if (r == nn::pia::ResultNoData())
        {
            // Because there is no more receive data, the receive process ends.
            break;
        }
        else if (r == nn::pia::ResultBrokenData())
        {
            // Data of an unexpected size arrived.
            // When in this state, receiving with this <tt>ReliableProtocol</tt> cannot continue, so communications are terminated.
            return true;
        }
        else if (r == nn::pia::ResultNotInCommunication())
        {
            // Does not go here if participating in a session.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
        else
        {
            // Other errors require the program fixes.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(r);
        }
    }

    return false;
}


// Draw on the screen.
void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    static const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    static const nn::pia::sampleframework::Color cCursorColor(1.0f, 0.3f, 0.3f, 1.0f);
    static const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);
    {
        pSampleFramework->DrawText(10, 10, "PiaTransport ReliableAndUnreliable");
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                cGameModeText = "LocalHost";
            }
            else
            {
                cGameModeText = "LocalClient";
            }
            break;
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                cGameModeText = "InetHost";
            }
            else
            {
                cGameModeText = "InetClient";
            }
            break;
        case GAME_MODE_INET_AUTOMATCH:
            PIA_SAMPLE_ASSERTMSG(false, "InetAutomatch?");
            cGameModeText = "InetAutomatch";
            break;
        default:
            PIA_SAMPLE_ASSERTMSG(false, "Unknown game mode? (%d)", gameMode);
            cGameModeText = "???";
            break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%d", pSession->GetStationNum());
        pSampleFramework->DrawText(320, 20, "Log:%s", s_IsLogEnable ? "ON" : "OFF");
        pSampleFramework->DrawText(10, 30, "A: send once  X: toggle to send");
        pSampleFramework->DrawText(10, 40, "Y: toggle to log  START:quit");
    }

    {
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    {
        pSampleFramework->DrawText(10, 60, "Idx");
        pSampleFramework->DrawText(50, 60, "Unreliable");
        pSampleFramework->DrawText(150, 60, "Reliable");
        pSampleFramework->DrawText(250, 60, "ReliableLarge");
        pSampleFramework->DrawText(360, 60, "RTT");
    }

    nn::pia::StationId localStationId = pSession->GetLocalStationId();
    nn::pia::StationId masterStationId = pSession->GetHostStationId();

    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        int y1 = 70 + i * 22;
        int y2 = 80 + i * 22;
        nn::pia::StationId stationId = s_StationData[i].m_StationId;

        if (stationId == masterStationId)
        {
            pSampleFramework->DrawText(2, y1, "*");
        }
        pSampleFramework->DrawText(10, y1, "%02x", conv2StationIndex(stationId));

        bool isValid = pSession->CheckStationIdIsValid(stationId);
        if (isValid == false)
        {
            pSampleFramework->SetFontColor(cInvalidColor);
            pSampleFramework->DrawText(30, y1, "disconnect");
            pSampleFramework->SetFontColor(cDefaultColor);
            continue;
        }

        if (i == conv2StationIndex(localStationId))
        {
            pSampleFramework->SetFontColor(cInvalidColor);
            pSampleFramework->DrawText(30, y1, "me");
            pSampleFramework->SetFontColor(cDefaultColor);
            continue;
        }

        for (int j = 0; j < PROTOCOL_NUM; ++j)
        {
            int x = 70 + j * 100;
            if (i == s_TargetStationIdx && j == s_TargetProtocolType)
            {
                pSampleFramework->SetFontColor(cCursorColor);
                pSampleFramework->DrawText(x - 10, y1, ">");
            }
            pSampleFramework->DrawText(x, y1, "%6d%c", s_StationData[i].m_ProtocolData[j].m_SendNum, s_StationData[i].m_ProtocolData[j].m_IsSendRepeat ? '@' : ' ');
            pSampleFramework->DrawText(x, y2, "%6d", s_StationData[i].m_ProtocolData[j].m_ReceiveNum);
            if (i == s_TargetStationIdx && j == s_TargetProtocolType)
            {
                pSampleFramework->SetFontColor(cDefaultColor);
            }
        }

        // Round trip time (RTT) display.
        {
            int x = 360;
            nn::pia::transport::StationManager* pMgr = nn::pia::transport::StationManager::GetInstance();
            if (PIA_IS_VALID_POINTER(pMgr))
            {
                nn::pia::transport::Station* pStation = pMgr->GetStation(conv2StationId(static_cast<nn::pia::StationIndex>(i)));
                if (PIA_IS_VALID_POINTER(pStation))
                {
                    pSampleFramework->DrawText(x, y1, "%dms", pStation->GetRtt());
                    pSampleFramework->DrawText(x, y1 + 10, "[%d]", pStation->GetRttSamplingNum());
                }
            }
        }
    }

    {
        int i = STATION_ALL_IDX;
        int y1 = 70 + i * 22;
        pSampleFramework->DrawText(10, y1, "All");

        for (int j = 0; j < PROTOCOL_NUM; ++j)
        {
            int x = 70 + j * 100;
            if (MAX_STATIONS == s_TargetStationIdx && j == s_TargetProtocolType)
            {
                pSampleFramework->SetFontColor(cCursorColor);
                pSampleFramework->DrawText(x - 10, y1, ">");
            }
            pSampleFramework->DrawText(x, y1, "%6d%c", s_StationData[i].m_ProtocolData[j].m_SendNum, s_StationData[i].m_ProtocolData[j].m_IsSendRepeat ? '@' : ' ');
            if (MAX_STATIONS == s_TargetStationIdx && j == s_TargetProtocolType)
            {
                pSampleFramework->SetFontColor(cDefaultColor);
            }
        }
    }

    // Renders the number of packets that were successfully sent and received on the screen.
    {
        // Render position.
        int x = 10;
        int y = 200;

        nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

        pSampleFramework->DrawText(x, y, "Send packets: %5d", pMgr->GetSendThreadSendPacketNum());
        pSampleFramework->DrawText(x, y + 12, "Recv packets: %5d", pMgr->GetReceiveThreadReceivePacketNum());
    }

    // Draws the amount of data that was successfully sent and received on the screen.
    {
        // Render position.
        int x = 180;
        int y = 200;

        nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

        pSampleFramework->DrawText(x, y, "Send data(bytes): %8lld", pMgr->GetSendThreadSendDataSize());
        pSampleFramework->DrawText(x, y + 12, "Recv data(bytes): %8lld", pMgr->GetReceiveThreadReceiveDataSize());
    }
}


// Clears the send/receive counter.
void ClearSendRecvPacketNum(void)
{
    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

    pMgr->ClearSendThreadSendPacketNum();
    pMgr->ClearReceiveThreadReceivePacketNum();
}


// Clears the send/receive data counter.
void ClearSendRecvDataSize(void)
{
    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

    pMgr->ClearSendThreadSendDataSize();
    pMgr->ClearReceiveThreadReceiveDataSize();
}


#if PIA_TEST_CODE
void UpdatePiaTestData(GameMode gameMode, bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_ASSERT(PIA_IS_VALID_POINTER(pSession));

    bool isInet = false;
    switch (gameMode)
    {
    case GAME_MODE_INET_MASTER:
    case GAME_MODE_INET_CLIENT:
    case GAME_MODE_INET_AUTOMATCH:
        isInet = true;
        break;
    }

    pSampleFramework->SetProgramStatus(
        isExit ? 0 : (isInet ? 4 : 3),
        pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION,
        pSession->GetLocalStationId() == pSession->GetHostStationId(),
        pSession->GetStationNum());

    // Session.
    pSampleFramework->SetProgramParam(0x00, conv2StationIndex(pSession->GetLocalStationId()));

    // Local station settings.
    pSampleFramework->SetProgramParam(0x08, s_TargetStationIdx);
    pSampleFramework->SetProgramParam(0x09, s_TargetProtocolType);

#if ENABLE_PROFILING
    // Measurement
    pSampleFramework->SetProgramParam(0x0c, s_IsCountReceive ? 1 : 0);
    pSampleFramework->SetProgramParam(0x0d, s_ReceivedSize);
    pSampleFramework->SetProgramParam(0x0e, s_ReceivedTime);
#endif

    // Each individual station.
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        nn::pia::StationId sid = conv2StationId(static_cast<nn::pia::StationIndex>(i));
        size_t sIdx = i;
        u32 idOfs = (i + 1) << 4;

        bool isValid = pSession->CheckStationIdIsValid(sid);
        pSampleFramework->SetProgramParam(idOfs + 0x0, isValid ? 1 : 0);

        pSampleFramework->SetProgramParam(idOfs + 0x4, s_StationData[sIdx].m_ProtocolData[PROTOCOL_UNRELIABLE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x5, s_StationData[sIdx].m_ProtocolData[PROTOCOL_UNRELIABLE].m_ReceiveNum);
        pSampleFramework->SetProgramParam(idOfs + 0x6, s_StationData[sIdx].m_ProtocolData[PROTOCOL_UNRELIABLE].m_SendNum);

        pSampleFramework->SetProgramParam(idOfs + 0x8, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x9, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE].m_ReceiveNum);
        pSampleFramework->SetProgramParam(idOfs + 0xa, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE].m_SendNum);

        pSampleFramework->SetProgramParam(idOfs + 0xc, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE_LARGE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0xd, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE_LARGE].m_ReceiveNum);
        pSampleFramework->SetProgramParam(idOfs + 0xe, s_StationData[sIdx].m_ProtocolData[PROTOCOL_RELIABLE_LARGE].m_SendNum);
    }

    // To all stations.
    {
        u32 idOfs = (STATION_ALL_IDX + 1) << 4;

        pSampleFramework->SetProgramParam(idOfs + 0x4, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_UNRELIABLE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x6, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_UNRELIABLE].m_SendNum);

        pSampleFramework->SetProgramParam(idOfs + 0x8, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_RELIABLE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0xa, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_RELIABLE].m_SendNum);

        pSampleFramework->SetProgramParam(idOfs + 0xc, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_RELIABLE_LARGE].m_IsSendRepeat ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0xe, s_StationData[STATION_ALL_IDX].m_ProtocolData[PROTOCOL_RELIABLE_LARGE].m_SendNum);
    }
}
#endif


// Main loop during communication.
void MainLoop(GameMode gameMode)
{
    nn::Result r;

    while (true)
    {
        bool isExit = false;
        bool bCallLeaveSession = false;

        s_CurrentTime.SetNow();

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        isExit |= Operate();

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::Result connectionStatus = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
        if (connectionStatus.IsFailure())
        {
            // Ends because the communication has been disconnected.
            isExit = true;
        }

        // Check the session connection status.
        {
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();

            // Error handling.
            if (result.IsFailure())
            {
                enum nn::pia::util::HandlingType handling = nn::pia::util::GetHandlingType(result);
                switch(handling)
                {
                case nn::pia::util::HANDLING_LEAVE_AND_CLEANUP:
                    PIA_SAMPLE_LOG("Exit MainLoop() with LeaveSession().\n");
                    bCallLeaveSession = true;
                    isExit = true;
                    break;
                case nn::pia::util::HANDLING_RECONNECT:
                    PIA_SAMPLE_LOG("Exit MainLoop().\n");
                    isExit = true;
                    break;
                default:
                    // The GetHandlingType function is capable of returning values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT,
                    // but when the Result objects that can be returned by the Session::CheckConnectionError function are considered,
                    // values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT are never returned.
                    // An assert statement is provided for that reason.
                    PIA_SAMPLE_ASSERTMSG(false, "Unexpected result.\n");
                    break;
                }
            }
        }

        if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
        {
            // Send/receive processing takes place if during communication.

            SendUnreliable(GetUnreliableProtocol());
            SendReliable(GetReliableProtocol());
            SendReliableLarge(GetReliableProtocolLarge());

            ReceiveUnreliable(GetUnreliableProtocol());
            isExit |= ReceiveReliable(GetReliableProtocol());
            isExit |= ReceiveReliableLarge(GetReliableProtocolLarge());
        }
        else
        {
            // Ends because the session was lost.
            isExit = true;
        }

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        Draw(gameMode);

#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, isExit);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        if (isExit)
        {
            if (bCallLeaveSession)
            {
                Session_LeaveSession();
            }

            ClearSendRecvPacketNum();
            ClearSendRecvDataSize();

            return;
        }
    }
}


}  // end of anonymous namespace


// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;
    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(30);
    }

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    InitBasicLargeData();

    s_CurrentTime.SetNow();

    while (true)
    {
        nn::pia::common::MemoryLeakChecker leakChecker;

#if PIA_TEST_CODE
        s_OptionalParam.Initialize(OPTIONAL_PARAM_OFFSET);
        GameMode gameMode = DecideGameMode("PiaTransport ReliableAndUnreliable", &s_OptionalParam);
#else
        GameMode gameMode = DecideGameMode("PiaTransport ReliableAndUnreliable");
#endif
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
#if 0
            u64 traceFlag = 0;
#else
            u64 traceFlag =
                nn::pia::common::TRACE_FLAG_COMMON |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_LOCAL |
                nn::pia::common::TRACE_FLAG_SESSION;
#endif
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlag, true);
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            result = leakChecker.Set();
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            {
                PIA_SAMPLE_LOG("subId: %x\n", SAMPLE_SUB_ID);
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID, true);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Local_InitializePiaLocal() failed.\n");
                    Local_FinalizePiaLocal();

                    result = leakChecker.Check();
                    PIA_SAMPLE_ASSERT(result.IsSuccess());

                    Common_FinalizePiaCommon();
                    continue;
                }
                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            nn::pia::transport::Transport::Setting transportSetting;
            transportSetting.pFactory = pFactory;
            transportSetting.maxStationNum = MAX_STATIONS;
            transportSetting.maxSendThreadBufferNum = 16;
            transportSetting.maxReceiveThreadBufferNum = 32;
            transportSetting.analysisInterval = 5;  // in seconds.

#if PIA_TEST_CODE
            // Do not call the <tt>SetDebugSetting</tt> function if you are not using
            // the latency emulation feature or packet loss emulation.
            {
                nn::pia::transport::Transport::DebugSetting debugSetting;
                debugSetting.sendThreadLatencyEmulationPacketNum = 0;
                debugSetting.receiveThreadLatencyEmulationPacketNum = 0;
                debugSetting.bPacketLossEmulation = s_OptionalParam.bPacketLossEmulation;
                result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
                PIA_SAMPLE_ASSERT(result.IsSuccess());
            }
#endif

            result = Transport_InitializePiaTransport(
                transportSetting,
                0,
                NULL,
                0,
                NULL,
                64,
                64,
                CreateAndInitializeProtocols,
                32);
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            {
                // Explicitly specify the priority of <tt>PiaTransport</tt> send/receive threads and sleep times.
                nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                pMgr->SetSendThreadPriority(8);
                pMgr->SetSendThreadSleepTimeSpan(5);
                pMgr->SetReceiveThreadPriority(8);
                pMgr->SetReceiveThreadSleepTimeSpan(5);

                // Confirm set values.
                PIA_SAMPLE_LOG("Send thread priority:      %d\n", pMgr->GetSendThreadPriority());
                PIA_SAMPLE_LOG("Send thread sleep span:    %d\n", pMgr->GetSendThreadSleepTimeSpan());
                PIA_SAMPLE_LOG("Receive thread priority:   %d\n", pMgr->GetReceiveThreadPriority());
                PIA_SAMPLE_LOG("Receive thread sleep span: %d\n", pMgr->GetReceiveThreadSleepTimeSpan());
            }

            {
                // Provides an example of explicitly specifying/getting the initial interval for sending RTT pulses.
                // If the default settings are suitable for your purposes, there is no need to call the SetInitialRttPulseInterval function.
                nn::pia::transport::Station::SetInitialRttPulseInterval(200);  // The value is in milliseconds.
                PIA_SAMPLE_LOG("Initial RTT pulse interval: %d\n", nn::pia::transport::Station::GetInitialRttPulseInterval());
            }

            {
                // To pass the transport analysis text to a function that is proprietary to your application,
                // specify a pointer to the callback function with a non-null value for <tt>RegisterPrintCallback()</tt>.
                nn::pia::transport::TransportAnalyzer* pAnalyzer = nn::pia::transport::TransportAnalyzer::GetInstance();
                PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pAnalyzer));
                pAnalyzer->RegisterPrintCallback(NULL);
            }

#if PIA_TEST_CODE
            if(s_OptionalParam.bPacketLossEmulation)
            {
                nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
                PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));

                pMgr->SetSendThreadPacketLossRatio(s_OptionalParam.sendPacketLossPercentage);
                pMgr->SetReceiveThreadPacketLossRatio(s_OptionalParam.recvPacketLossPercentage);
            }

            // Initializes the session.
            nn::pia::session::Session::Setting sessionSetting;
            sessionSetting.pNetworkFactory = pFactory;
            sessionSetting.browsedSessionInfoListNum = 16;

            if(s_OptionalParam.bRelayEmulation)
            {
                // If you are performing relay emulation, you need to call the settings function before creating the session instance.
                nn::pia::session::Session::DebugSetting debugSetting;
                debugSetting.isEnableRelayEmulation = true;
                nn::Result ret = nn::pia::session::Session::SetDebugSetting(debugSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(ret);
                sessionSetting.networkTopology = nn::pia::session::NetworkTopology_RelayMesh;
            }
            else
            {
                sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
            }
            result = Session_InitializePiaSession(sessionSetting, s_OptionalParam.bRelayEmulation);
#else
            // Sets up the session.
            result = Session_InitializePiaSession(pFactory, nn::pia::session::NetworkTopology_FullMesh);
#endif

        }

        // Connect to the network connections and create a session.
        {
            nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SessionEventCallback);
            InitializeStationData();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = true;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the matchmaking session to create.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                        createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = true;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
#if PIA_TEST_CODE
                    searchCriteriaList[0].SetResultRange(0, 16);
#else
                    searchCriteriaList[0].SetResultRange(0, 8);
#endif
                }

                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_BrowseSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connects to the Internet.
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = true;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }
                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

#if PIA_TEST_CODE
                if(s_OptionalParam.bRelayEmulation)
                {
                    SetRelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3);
                }
#endif

                bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
                if (isHost)
                {
                    // Became the host as a result of the matching.
                    gameMode = GAME_MODE_INET_MASTER;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;

            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
            MainLoop(gameMode);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }

            break;
        }


        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
                // Leave the session.
                Session_LeaveSession();
                break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
                break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
                Inet_Logout();
                Inet_FinalizeNetwork();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }


        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Transport_FinalizePiaTransport(0, 0, FinalizeAndDestroyProtocols);
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
                Local_FinalizePiaLocal();
                break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            result = leakChecker.Check();
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            Common_FinalizePiaCommon();
        }
    }

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
