﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Demo.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "Shared_Demo.h"

#include "SampleLib/SampleLib.h"

#include <pia/session.h>


#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif


#if PIA_TEST_CODE
const s32 IDX_RELAY_EMU          = 0;
const s32 IDX_PACKET_LOSS_EMU    = 1;
const s32 IDX_SEND_PACKET_LOSS   = 2;
const s32 IDX_RECV_PACKET_LOSS   = 3;
const s32 IDX_RELAY_PRESET       = 4;
const s32 IDX_LATENCY_EMU        = 5;
const s32 IDX_LATENCY_PRESET     = 6;
const s32 IDX_DEMANDING_BIT_RATE = 7;
const s32 IDX_END                = 8;

struct ParameterSetting
{
    s32 value;          // This value is the object of operations.
    s32 minimum;        // The smallest value of value.
    s32 maximum;        // The greatest value of value.
    s32 increment;      // Increment amount
    const char* pName;  // Name of the item.
};

ParameterSetting s_ParameterSetting[IDX_END] =
{
    {0       , 0       ,   1       , 1       , "Relay emulation"            },
    {0       , 0       ,   1       , 1       , "Packet loss emulation"      },
    {0       , 0       , 100       , 1       , "Send packet loss percentage"},
    {0       , 0       , 100       , 1       , "Recv packet loss percentage"},
    {0       , 0       ,  10       , 1       , "Relay preset"               },
    {0       , 0       ,   1       , 1       , "Latency emulation"          },
    {0       , 0       ,  10       , 1       , "Latency preset"             },
    {200*1000, 100*1000,  1000*1000, 100*1000, "Demanding bit rate"         },
};

s32 s_CursorIdx = 0;


void initializePiaTest()
{
    s_CursorIdx = 0;
}


void operatePiaTest(OptionalParameter* pOpt)
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pOpt));

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSampleFramework));

    // Move the cursor up/down with the +Control Pad (this could loop)
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
    {
        s_CursorIdx = (s_CursorIdx + IDX_END - 1) % IDX_END;
    }
    else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
    {
        s_CursorIdx = (s_CursorIdx + 1) % IDX_END;
    }

    // Increase/decrease the value by pressing the left/right button on the +Control Pad.
    struct ParameterSetting& param = s_ParameterSetting[s_CursorIdx];
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
    {
        if(param.value < param.maximum)
        {
            param.value += param.increment;
        }
    }
    else if(pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
    {
        if(param.value > param.minimum)
        {
            param.value -= param.increment;
        }
    }
}


void updatePiaTestData(s32 offset)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    pSampleFramework->SetProgramParam(offset+0, s_CursorIdx);
    pSampleFramework->SetProgramParam(offset+1, s_ParameterSetting[s_CursorIdx].value);
}


void drawPiaTest()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSampleFramework));

    // Starting point and height of the XY coordinate.
    s32 x = 10;
    s32 y = 110;
    const s32 H = 10;

    // Cursor rendering (animated).
    static int calls = 0;
    int shift = (calls / 10) % 2;
    pSampleFramework->DrawText(x - 7 + shift, y + H * s_CursorIdx, ">");
    ++calls;

    // Rendering of each item.
    for (int i=0; i<IDX_END; ++i)
    {
        const struct ParameterSetting& param = s_ParameterSetting[i];
        pSampleFramework->DrawText(x, y, "%s: %d", param.pName, param.value);
        y += H;
    }
}
#endif  // end of PIA_TEST_CODE


// Determines the mode to execute.
GameMode DecideGameMode(const char* pTitle, OptionalParameter* pOpt)
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTitle));

#if PIA_TEST_CODE
    initializePiaTest();
#else
    NN_PIA_DUMMY_PARAM(pOpt);
#endif

#if ENABLE_PIA_LOCAL
    PIA_SAMPLE_PRINT("Please select game mode.\n [A]LocalHost\n [B]LocalClient\n [X]InetAutomatch\n [START]Quit\n");
#else
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
#endif
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
#if ENABLE_PIA_LOCAL
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalHost\n");
            gameMode = GAME_MODE_LOCAL_MASTER;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalClient\n");
            gameMode = GAME_MODE_LOCAL_CLIENT;
            isDecided = true;
        }
#endif
#if PIA_TEST_CODE
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            PIA_SAMPLE_PRINT("GameMode: InetHost\n");
            gameMode = GAME_MODE_INET_MASTER;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L))
        {
            PIA_SAMPLE_PRINT("GameMode: InetClient\n");
            gameMode = GAME_MODE_INET_CLIENT;
            isDecided = true;
        }
#endif

#if PIA_TEST_CODE
        operatePiaTest(pOpt);
        updatePiaTestData(pOpt->offset);
#endif

        pSampleFramework->DrawText(10, 10, pTitle);
        pSampleFramework->DrawText(10, 20, "Select game mode.");
#if ENABLE_PIA_LOCAL
        pSampleFramework->DrawText(10, 30, "A: LocalHost");
        pSampleFramework->DrawText(10, 40, "B: LocalClient");
        pSampleFramework->DrawText(10, 50, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 60, "START: Quit this application");
#else
        pSampleFramework->DrawText(10, 30, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 40, "START: Quit this application");
#endif

#if PIA_TEST_CODE
        pSampleFramework->DrawText(10, 70, "R: InetHost");
        pSampleFramework->DrawText(10, 80, "L: InetClient");
#endif

#if PIA_TEST_CODE
        drawPiaTest();

        pSampleFramework->SetProgramStatus(isDecided ? 0 : 1, false);
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
#if PIA_TEST_CODE
            PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pOpt));
            pOpt->bRelayEmulation          = s_ParameterSetting[IDX_RELAY_EMU].value;
            pOpt->bPacketLossEmulation     = s_ParameterSetting[IDX_PACKET_LOSS_EMU].value;
            pOpt->sendPacketLossPercentage = s_ParameterSetting[IDX_SEND_PACKET_LOSS].value;
            pOpt->recvPacketLossPercentage = s_ParameterSetting[IDX_RECV_PACKET_LOSS].value;
            pOpt->relayPreset              = s_ParameterSetting[IDX_RELAY_PRESET].value;
            pOpt->bLatencyEmulation        = s_ParameterSetting[IDX_LATENCY_EMU].value;
            pOpt->latencyPreset            = s_ParameterSetting[IDX_LATENCY_PRESET].value;
            pOpt->demandingBitRate         = s_ParameterSetting[IDX_DEMANDING_BIT_RATE].value;
#endif

            return gameMode;
        }
    }
}


void SetRelayEmulation(nn::pia::StationIndex c1, nn::pia::StationIndex c2)
{
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

    // Initialize relay emulation before configuring in case the previous settings remain.
    pSession->ResetRelayEmulation();

    // Relay connection.
    PIA_BASIC_REPORT("Set RelayEmulation: %d - %d\n", c1, c2);
    nn::Result r = pSession->RelayEmulation(c1, c2, nn::pia::session::Mesh::EMULATION_TYPE_RELAY);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
}
