﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Demo.h"

#define ENABLE_HOST_MIGRATION 1

#if NN_PIA_CTR
#define USE_PIA_LOCAL 1
#else
#define USE_PIA_LOCAL 0
#endif

#if USE_PIA_LOCAL
#include <pia/local.h>
#include "Shared_Local_Setup.h"
#endif

#include <pia/util.h>

#include <cstdio>


#ifndef ENABLE_PROFILING
#define ENABLE_PROFILING 0
#endif

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

#define PROGRAM_TITLE "ReliableBroadcast"


namespace
{
const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 30100;  // Use different values here for each Pia sample program.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;

const size_t MAX_STATIONS = 4;  // Defines the maximum number of stations.

const size_t ORIGINAL_DATA_SIZE = 10 * 1024 * 1024;               // The size of the source data being sent or received.
static bit32 s_OriginalData[ORIGINAL_DATA_SIZE / sizeof(u32)];    // The source data that is sent or received.
const size_t RECEIVE_BUFFER_SIZE = 8 * 1024 * 1024;               // The size of the receive buffer.
static bit32 s_ReceiveBuffer[RECEIVE_BUFFER_SIZE / sizeof(u32)];  // Receive buffer.

#if USE_PIA_LOCAL
const u32 SAMPLE_TITLE_ID = 0xFF301;  // For the transport sample.
const u8 SAMPLE_SUB_ID = 0x00;
const char SAMPLE_PASSPHRASE[] = "aaabbccc0123456789";
const char SAMPLE_SIGNATURE_KEY_DATA[] = "ReliableBroadcast_key";
#endif

// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const u32 SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const u32 SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

#define SAMPLE_NEX_THREAD_MODE nn::nex::Core::ThreadModeUnsafeTransportBuffer

const size_t MEMSIZE = 1 * 1024 * 1024;              // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.


const u16 RELIABLE_BROADCAST_PROTOCOL_PORT = 0;  // Defines the port number used by ReliableBroadcastProtocol.

u32 s_hReliableBroadcast = 0;

#if PIA_TEST_CODE
struct OptionalParameter s_OptionalParam;
const s32 OPTIONAL_PARAM_OFFSET = 200;
#endif


nn::pia::StationId conv2StationId(nn::pia::StationIndex idx)
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    if (!PIA_IS_VALID_POINTER(pTransport))
    {
        PIA_SAMPLE_ASSERTMSG(false, "conv2StationId(): Transport instance is not ready?");
        return nn::pia::STATION_ID_INVALID;
    }

    nn::pia::StationId id;
    nn::Result ret = pTransport->ConvertToStationId(&id, idx);
    if (ret.IsFailure())
    {
        if (ret==nn::pia::ResultNotFound())
        {
            return nn::pia::STATION_ID_INVALID;
        }
        else
        {
            PIA_SAMPLE_PRINT("Result = 0x%08x\n", ret.GetValue());
            PIA_SAMPLE_ASSERTMSG(false, "conv2StationId(): ConvertToStationId() failed.");
            return nn::pia::STATION_ID_INVALID;
        }
    }

    return id;
}


nn::pia::StationIndex conv2StationIndex(nn::pia::StationId id)
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    if (!PIA_IS_VALID_POINTER(pTransport))
    {
        PIA_SAMPLE_ASSERTMSG(false, "conv2StationIndex: Transport instance is not ready?");
        return nn::pia::STATION_INDEX_INVALID;
    }

    nn::pia::StationIndex idx;
    nn::Result ret = pTransport->ConvertToStationIndex(&idx, id);
    if (ret.IsFailure())
    {
        if (ret==nn::pia::ResultNotFound())
        {
            return nn::pia::STATION_INDEX_INVALID;
        }
        else
        {
            PIA_SAMPLE_PRINT("Result = 0x%08x\n", ret.GetValue());
            PIA_SAMPLE_ASSERTMSG(false, "conv2StationIndex: ConvertToStationIndex() failed.");
            return nn::pia::STATION_INDEX_INVALID;
        }
    }

    return idx;
}


// Create and initialize the protocol.
nn::Result CreateAndInitializeProtocol()
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Create and initialize ReliableBroadcastProtocol.
    {
        PIA_SAMPLE_ASSERT(s_hReliableBroadcast == 0);
        s_hReliableBroadcast = pTransport->CreateProtocol<nn::pia::transport::ReliableBroadcastProtocol>(RELIABLE_BROADCAST_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hReliableBroadcast != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        pTransport->GetProtocol<nn::pia::transport::ReliableBroadcastProtocol>(s_hReliableBroadcast)->Initialize();
    }

    return nn::ResultSuccess();
}


// Clean up the protocol.
void FinalizeAndDestroyProtocol()
{
    // Get the Transport singleton.
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Finalization of <tt>ReliableBroadcastProtocol</tt>.
    nn::pia::transport::ReliableBroadcastProtocol* pReliableBroadcastProtocol = pTransport->GetProtocol<nn::pia::transport::ReliableBroadcastProtocol>(s_hReliableBroadcast);
    if (PIA_IS_VALID_POINTER(pReliableBroadcastProtocol))
    {
        pReliableBroadcastProtocol->Finalize();
        pTransport->DestroyProtocol(s_hReliableBroadcast);
        s_hReliableBroadcast = 0;
    }
}


nn::pia::transport::ReliableBroadcastProtocol* GetReliableBroadcastProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableBroadcastProtocol>(s_hReliableBroadcast);
}


// Creates the source data to send.
void InitOriginalData()
{
    std::srand(0x12345678);
    u32 size = ORIGINAL_DATA_SIZE / sizeof(bit32);
    for (u32 i = 0; i < size; ++i)
    {
        s_OriginalData[i] = static_cast<bit32>(std::rand());
    }
}


// Checks the content of data.
bool CheckReceivedData(u32 offset, u32 size)
{
    if (offset + size > ORIGINAL_DATA_SIZE)
    {
        // Invalid data is data that attempts to reference until the source data is out of range.
        return false;
    }

    const void* cpData = reinterpret_cast<void*>(reinterpret_cast<u32>(s_OriginalData) + offset);
    return (std::memcmp(s_ReceiveBuffer, cpData, size) == 0);
}


static nn::pia::StationId s_ReceiveStationId;                                      // Specifies the sender of receive data.
static size_t s_ReceiveBufferSize;                                                 // The size of the specified receive buffer.
static size_t s_SendDataSize;                                                      // The size of the data to send.
static nn::pia::transport::ReliableBroadcastProtocol::Configuration s_SendConfig;  // The setting when sending.
static bool s_IsCompletionChecked;                                                 // Have you checked for send/receive completion?
static const size_t MESSAGE_SIZE = 32;                                             // The size of the message for display.
static char s_Message[MESSAGE_SIZE];                                               // The message for display.

#if ENABLE_PROFILING
static u32 s_TimeElapsed = 0;  // Time required for previous process (ms).
#endif

// The target of the operation.
enum OperationTarget
{
    OPERATION_RECEIVE_STATION_ID = 0,
    OPERATION_RECEIVE_BUFFER_SIZE,
    OPERATION_START_RECEIVE,
    OPERATION_REJECT_REQUEST,
    OPERATION_SEND_SIZE,
    OPERATION_CONFIG_SEND_TO_NEWCOMER,
    OPERATION_CONFIG_SYNCHRONIZE_COMPLETION,
    OPERATION_CONFIG_SEND_ENDLESS,
    OPERATION_CONFIG_FAIL_IF_REJECTED,
    OPERATION_CONFIG_MEDIATE_REQUEST,
    OPERATION_START_SEND,
    OPERATION_THROUGHPUT_LIMIT,
    OPERATION_CANCEL,
    OPERATION_FINISH_SEND_ENDLESS,
    OPERATION_RESET,
    OPERATION_TARGET_NUM
};
OperationTarget s_OperationTarget;

enum OperationGroup
{
    OPERATION_GROUP_RECEIVE = 0,
    OPERATION_GROUP_SEND,
    OPERATION_GROUP_PROGRESS,
    OPERATION_GROUP_SETTING,
    OPERATION_GROUP_NUM
};

OperationGroup GetOperationGroup(OperationTarget target)
{
    switch (target)
    {
    case OPERATION_RECEIVE_STATION_ID:
    case OPERATION_RECEIVE_BUFFER_SIZE:
    case OPERATION_START_RECEIVE:
    case OPERATION_REJECT_REQUEST:
        return OPERATION_GROUP_RECEIVE;

    case OPERATION_SEND_SIZE:
    case OPERATION_CONFIG_SEND_TO_NEWCOMER:
    case OPERATION_CONFIG_SYNCHRONIZE_COMPLETION:
    case OPERATION_CONFIG_SEND_ENDLESS:
    case OPERATION_CONFIG_FAIL_IF_REJECTED:
    case OPERATION_CONFIG_MEDIATE_REQUEST:
    case OPERATION_START_SEND:
        return OPERATION_GROUP_SEND;

    case OPERATION_THROUGHPUT_LIMIT:
    case OPERATION_CANCEL:
    case OPERATION_FINISH_SEND_ENDLESS:
    case OPERATION_RESET:
        return OPERATION_GROUP_PROGRESS;

    default:
        PIA_SAMPLE_ASSERT(false);
        return OPERATION_GROUP_PROGRESS;
    }
}

static u32 s_LineFeed[OPERATION_GROUP_NUM];
static const u32 cLineFeedSize = 9;
static nn::pia::common::Time s_StartTime;

// Initializes the application settings.
void InitializeGameData()
{
    s_ReceiveStationId = nn::pia::STATION_ID_INVALID;
    s_ReceiveBufferSize = 8 * 1024 * 1024;
    s_SendDataSize = 1 * 1024 * 1024;
    s_SendConfig = nn::pia::transport::ReliableBroadcastProtocol::Configuration();
    s_IsCompletionChecked = false;
    memset(s_Message, 0, sizeof(s_Message));
    s_OperationTarget = OPERATION_RECEIVE_STATION_ID;
    for (u32 i = 0; i < OPERATION_GROUP_NUM; ++i)
    {
        s_LineFeed[i] = 0;
    }
    s_LineFeed[GetOperationGroup(s_OperationTarget)] = cLineFeedSize;

    std::srand(static_cast<u32>(nn::pia::common::Time::GetTime().GetTick()));
}


// Process during connection and disconnection.
void SessionEventCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId)
{
    nn::Result r;
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
        break;
    case nn::pia::session::Session::EVENT_LEAVE:
        break;
    case nn::pia::session::Session::EVENT_SESSION_HOST_CHANGED:
        break;
    case nn::pia::session::Session::EVENT_HOST_MIGRATION_FAILED:
        // Host migration fails.
        break;
    default:
        PIA_SAMPLE_ASSERTMSG(false, "Unsupported event(%d).", eventType);
        break;
    }
}


// Sets the message for display.
void SetMessage(const char* cpFormat, ...)
{
    va_list va;
    va_start(va, cpFormat);
    vsnprintf(s_Message, MESSAGE_SIZE - 1, cpFormat, va);
    va_end(va);
    s_Message[MESSAGE_SIZE - 1] = '\0';
    PIA_SAMPLE_LOG("%s\n", s_Message);
}


// Starts sending data.
void StartSend()
{
    // Determines an offset value using a random number and considers send data to be data in a position displaced by the amount of that offset from the source data.
    // Offset data is transferred to the receive data side as user data.

    u32 offset = static_cast<u32>(std::rand() & 0xFFFFFFFC) % (ORIGINAL_DATA_SIZE - s_SendDataSize);
    const void* cpData = reinterpret_cast<void*>(reinterpret_cast<u32>(s_OriginalData) + offset);
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    nn::Result r = pProtocol->StartSend(cpData, s_SendDataSize, s_SendConfig, offset);
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    SetMessage("start send data");
    s_StartTime.SetNow();
#if ENABLE_PROFILING
    s_TimeElapsed = 0;
#endif
    s_OperationTarget = OPERATION_CANCEL;
    s_IsCompletionChecked = false;
}


// Starts receiving data.
void StartReceive()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();

    if (s_ReceiveStationId == nn::pia::STATION_ID_INVALID)
    {
        // Starts receiving the highest priority receive request data.
        const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpRequest = pProtocol->GetRequest();
        if (cpRequest != NULL)
        {
            nn::pia::StationId srcId = cpRequest->GetSourceStationId();
            PIA_SAMPLE_ASSERT(srcId != nn::pia::STATION_ID_INVALID);
            nn::Result r = pProtocol->StartReceive(s_ReceiveBuffer, s_ReceiveBufferSize, srcId);
            NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
            SetMessage("start receive data");
            s_StartTime.SetNow();
            s_OperationTarget = OPERATION_CANCEL;
            s_IsCompletionChecked = false;
        }
    }
    else if (nn::pia::session::Session::GetInstance()->CheckStationIdIsValid(s_ReceiveStationId))
    {
        // Starts receiving from the specified station.
        const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpRequest = pProtocol->GetRequest(s_ReceiveStationId);
        if (cpRequest != NULL)
        {
            nn::Result r = pProtocol->StartReceive(s_ReceiveBuffer, s_ReceiveBufferSize, s_ReceiveStationId);
            NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
            SetMessage("start receive data");
            s_StartTime.SetNow();
#if ENABLE_PROFILING
            s_TimeElapsed = 0;
#endif
            s_OperationTarget = OPERATION_CANCEL;
            s_IsCompletionChecked = false;
        }
        else
        {
            nn::Result r = pProtocol->StartReceive(s_ReceiveBuffer, s_ReceiveBufferSize, s_ReceiveStationId);
            NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
            PIA_SAMPLE_ASSERT(r.IsSuccess());
            SetMessage("start receive data");
            s_StartTime.SetNow();
#if ENABLE_PROFILING
            s_TimeElapsed = 0;
#endif
            s_OperationTarget = OPERATION_CANCEL;
            s_IsCompletionChecked = false;
        }
    }
}


// Refuses the receive request.
void RejectRequest()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    nn::Result r = pProtocol->RejectRequest();
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
    SetMessage("reject request");
    s_OperationTarget = OPERATION_RECEIVE_STATION_ID;
    s_IsCompletionChecked = false;
}


// Stops the current send/receive operation.
void Cancel()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    switch (pProtocol->GetState())
    {
    case nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING:
    {
        // Stops sending.
        nn::Result r = pProtocol->Cancel();
        NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        SetMessage("cancel send data");
    }
    break;

    case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVING:
    {
        // Stops receiving.
        nn::Result r = pProtocol->Cancel();
        NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
        PIA_SAMPLE_ASSERT(r.IsSuccess());
        SetMessage("cancel receive data");
    }
    break;
    }
}


// Stops endless sending.
void FinishSendEndless()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    PIA_SAMPLE_ASSERT(pProtocol->GetState() == nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING);
    const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpTransferSetting = pProtocol->GetCurrentTransferSetting();
    NN_PIA_DUMMY_PARAM_IN_RELEASE(cpTransferSetting);
    PIA_SAMPLE_ASSERT(cpTransferSetting != NULL);
    PIA_SAMPLE_ASSERT(cpTransferSetting->m_Configuration.IsSendEndless());
    nn::Result r = pProtocol->FinishSendEndless();
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    SetMessage("finish SendEndless");
}


// Resets the send/receive complete state.
void Reset()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    PIA_SAMPLE_ASSERT(pProtocol->IsRunning() == false);
    nn::Result r = pProtocol->Reset();
    NN_PIA_DUMMY_PARAM_IN_RELEASE(r);
    PIA_SAMPLE_ASSERT(r.IsSuccess());
    SetMessage("reset");
    s_IsCompletionChecked = false;
}


// Determines if the specified operation target is in a state that allows operation now.
bool IsValidOperationTarget(OperationTarget operationTarget)
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();

    switch (operationTarget)
    {
    case OPERATION_RECEIVE_STATION_ID:
    case OPERATION_RECEIVE_BUFFER_SIZE:
        return (pProtocol->IsRunning() == false);

    case OPERATION_START_RECEIVE:
        if (pProtocol->IsRunning())
        {
            return false;
        }
        if (s_ReceiveStationId == nn::pia::STATION_ID_INVALID)
        {
            return (pProtocol->GetRequest() != NULL);
        }
        else
        {
            return true;
        }

    case OPERATION_REJECT_REQUEST:
        if (pProtocol->IsRunning())
        {
            return false;
        }
        if (s_ReceiveStationId == nn::pia::STATION_ID_INVALID)
        {
            return (pProtocol->GetRequest() != NULL);
        }
        else
        {
            return (pProtocol->GetRequest(s_ReceiveStationId) != NULL);
        }

    case OPERATION_SEND_SIZE:
    case OPERATION_CONFIG_SEND_TO_NEWCOMER:
    case OPERATION_CONFIG_SYNCHRONIZE_COMPLETION:
    case OPERATION_CONFIG_SEND_ENDLESS:
    case OPERATION_CONFIG_FAIL_IF_REJECTED:
    case OPERATION_CONFIG_MEDIATE_REQUEST:
    case OPERATION_START_SEND:
        return (pProtocol->IsRunning() == false);

    case OPERATION_THROUGHPUT_LIMIT:
        return true;

    case OPERATION_CANCEL:
        return pProtocol->IsRunning();

    case OPERATION_FINISH_SEND_ENDLESS:
        return ((pProtocol->GetState() == nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING) &&
                (pProtocol->GetCurrentTransferSetting()->m_Configuration.IsSendEndless()));

    case OPERATION_RESET:
        return (pProtocol->IsRunning() == false);

    default:
        return false;
    }
}


// Operation by key input.
bool Operate()
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();

    // Exit the GameLoop function if START is pressed.
    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        return true;
    }

    // The count for key repeat.
    static s32 keyCountL = -1;
    static s32 keyCountR = -1;
    {
        if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
        {
            ++keyCountL;
            keyCountR = -1;
        }
        else if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
        {
            keyCountL = -1;
            ++keyCountR;
        }
        else
        {
            keyCountL = -1;
            keyCountR = -1;
        }
    }

    // Switches the operation target using Up/Down Buttons.
    {
        int offset = 0;
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            offset = -1;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            offset = 1;
        }
        else if (IsValidOperationTarget(s_OperationTarget) == false)
        {
            offset = 1;
        }

        if (offset != 0)
        {
            int target = s_OperationTarget;
            while (true)
            {
                target += offset;
                if (target < 0)
                {
                    target = OPERATION_TARGET_NUM - 1;
                }
                else if (target >= OPERATION_TARGET_NUM)
                {
                    target = 0;
                }

                if (IsValidOperationTarget(static_cast<OperationTarget>(target)))
                {
                    s_OperationTarget = static_cast<OperationTarget>(target);
                    return false;
                }
            }
        }
    }

    switch (s_OperationTarget)
    {
    case OPERATION_RECEIVE_STATION_ID:
        // Switches the receive target using Left/Right Buttons.
        {
            int offset = 0;
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                offset = 1;
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                offset = -1;
            }

            // If <tt>s_ReceiveStationId</tt> is <tt>nn::pia::STATION_ID_INVALID</tt>, indicates the highest priority receive request.
            // Skips itself and switches the target station.

            if (offset != 0)
            {
                int stIdx = static_cast<int>(conv2StationIndex(s_ReceiveStationId));
                while (true)
                {
                    if (stIdx == nn::pia::STATION_INDEX_INVALID)
                    {
                        if (offset > 0)
                        {
                            stIdx = 0;
                        }
                        else
                        {
                            stIdx = MAX_STATIONS - 1;
                        }
                    }
                    else
                    {
                        stIdx += offset;
                        if (stIdx < 0 || stIdx >= MAX_STATIONS)
                        {
                            s_ReceiveStationId = nn::pia::STATION_ID_INVALID;
                            break;
                        }
                    }

                    if (stIdx != conv2StationIndex(pSession->GetLocalStationId()))
                    {
                        s_ReceiveStationId = conv2StationId(static_cast<nn::pia::StationIndex>(stIdx));
                        break;
                    }
                }
            }
        }
        break;

    case OPERATION_RECEIVE_BUFFER_SIZE:
    case OPERATION_SEND_SIZE:
        // Switches the send/receive size using the Left/Right Buttons.
        // When changing the number of bytes for an item, the number normally changes in increments of 256 KB. However, if you hold down the L Button, you can change the number in increments of 1 KB, and if you hold down the R Button, you can change it in increments of 4 bytes.
        {
            size_t* pSize = (s_OperationTarget == OPERATION_RECEIVE_BUFFER_SIZE) ? &s_ReceiveBufferSize : &s_SendDataSize;

            u32 unit = 256 * 1024;
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L))
            {
                unit = 1024;
            }
            else if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
            {
                unit = 4;
            }

            if (keyCountL == 0 || keyCountL > 15)
            {
                if (*pSize > unit)
                {
                    if (*pSize % unit > 0)
                    {
                        *pSize -= (*pSize % unit);
                    }
                    else
                    {
                        *pSize -= unit;
                    }
                }
            }
            else if (keyCountR == 0 || keyCountR > 15)
            {
                *pSize = *pSize - (*pSize % unit) + unit;
                if (*pSize > RECEIVE_BUFFER_SIZE)
                {
                    *pSize = RECEIVE_BUFFER_SIZE;
                }
            }
        }
        break;

    case OPERATION_THROUGHPUT_LIMIT:
        // Switches the throughput upper limit value using the Left/Right Buttons.
        // When changing the number of bytes for an item, the number normally changes in increments of 256 KB. However, if you hold down the R Button, you can change the number in increments of 4 bytes.
        {
            u32 unit = 256;
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
            {
                unit = 4;
            }

            u32 limit = pProtocol->GetThroughputLimit();
            if (keyCountL == 0 || keyCountL > 15)
            {
                if (limit > 0)
                {
                    if (limit % unit > 0)
                    {
                        limit -= (limit % unit);
                    }
                    else
                    {
                        limit -= unit;
                    }
                }
            }
            else if (keyCountR == 0 || keyCountR > 15)
            {
                limit = limit - (limit % unit) + unit;
                static const u32 THROUGHPUT_LIMIT_MAX = 10 * 1024;
                if (limit > THROUGHPUT_LIMIT_MAX)
                {
                    limit = THROUGHPUT_LIMIT_MAX;
                }
            }
            pProtocol->SetThroughputLimit(limit);
        }
        break;

    case OPERATION_CONFIG_SEND_TO_NEWCOMER:
    case OPERATION_CONFIG_SYNCHRONIZE_COMPLETION:
    case OPERATION_CONFIG_SEND_ENDLESS:
    case OPERATION_CONFIG_FAIL_IF_REJECTED:
    case OPERATION_CONFIG_MEDIATE_REQUEST:
        // Switches between yes and no using the Left/Right Buttons.
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT) || pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                switch (s_OperationTarget)
                {
                case OPERATION_CONFIG_SEND_TO_NEWCOMER:
                    s_SendConfig.SetSendToNewcomer(!s_SendConfig.IsSendToNewcomer());
                    break;
                case OPERATION_CONFIG_SYNCHRONIZE_COMPLETION:
                    s_SendConfig.SetSynchronizeCompletion(!s_SendConfig.IsSynchronizeCompletion());
                    break;
                case OPERATION_CONFIG_SEND_ENDLESS:
                    s_SendConfig.SetSendEndless(!s_SendConfig.IsSendEndless());
                    break;
                case OPERATION_CONFIG_FAIL_IF_REJECTED:
                    s_SendConfig.SetFailIfRejected(!s_SendConfig.IsFailIfRejected());
                    break;
                case OPERATION_CONFIG_MEDIATE_REQUEST:
                    s_SendConfig.SetMediateRequest(!s_SendConfig.IsMediateRequest());
                    break;
                }
            }
        }
        break;

    case OPERATION_START_RECEIVE:
    case OPERATION_REJECT_REQUEST:
    case OPERATION_START_SEND:
    case OPERATION_CANCEL:
    case OPERATION_FINISH_SEND_ENDLESS:
    case OPERATION_RESET:
        // Executes a process using the A Button.
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                switch (s_OperationTarget)
                {
                case OPERATION_START_RECEIVE:
                    StartReceive();
                    break;
                case OPERATION_REJECT_REQUEST:
                    RejectRequest();
                    break;
                case OPERATION_START_SEND:
                    StartSend();
                    break;
                case OPERATION_CANCEL:
                    Cancel();
                    break;
                case OPERATION_FINISH_SEND_ENDLESS:
                    FinishSendEndless();
                    break;
                case OPERATION_RESET:
                    Reset();
                    break;
                }
            }
        }
        break;
    }

    return false;
}


// Have you checked for send/receive completion?
void CheckCompletion()
{
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();
    switch (pProtocol->GetState())
    {
    case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVE_SUCCESS:
        if (s_IsCompletionChecked == false)
        {
            const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpSetting = pProtocol->GetCurrentTransferSetting();
            PIA_SAMPLE_ASSERT(cpSetting != NULL);

            // Checks to determine whether the content of the received data is correct.
            bool isValidData = CheckReceivedData(cpSetting->m_UserData, cpSetting->m_DataSize);
            NN_PIA_DUMMY_PARAM_IN_RELEASE(isValidData);
            nn::pia::StationId srcId = cpSetting->GetSourceStationId();
            NN_PIA_DUMMY_PARAM_IN_RELEASE(srcId);
            PIA_SAMPLE_ASSERT(srcId != nn::pia::STATION_ID_INVALID);
            PIA_SAMPLE_ASSERTMSG(isValidData, "received a wrong data (source:0x%16llx offset:%d, size:%d)", srcId.GetValue(), cpSetting->m_UserData, cpSetting->m_DataSize);

            nn::pia::common::Time finishTime;
            finishTime.SetNow();

#if ENABLE_PROFILING
            s_TimeElapsed = (finishTime - s_StartTime).GetMSec();
#endif
            SetMessage("receive success (%.1fs)", (finishTime - s_StartTime).GetSecF32());
            s_IsCompletionChecked = true;
        }
        break;

    case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVE_FAILURE:
        if (s_IsCompletionChecked == false)
        {
            SetMessage("receive failure");
            s_IsCompletionChecked = true;
        }
        break;

    case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_SUCCESS:
        if (s_IsCompletionChecked == false)
        {
            nn::pia::common::Time finishTime;
            finishTime.SetNow();

#if ENABLE_PROFILING
            s_TimeElapsed = (finishTime - s_StartTime).GetMSec();
#endif
            SetMessage("send success (%.1fs)", (finishTime - s_StartTime).GetSecF32());
            s_IsCompletionChecked = true;
        }
        break;

    case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_FAILURE:
        if (s_IsCompletionChecked == false)
        {
            SetMessage("send failure");
            s_IsCompletionChecked = true;
        }
        break;
    }
}


#define PRINT(x, ...)                                    \
    {                                                    \
        pSampleFramework->DrawText((x), y, __VA_ARGS__); \
    }
#define NEXT_LINE()         \
    {                       \
        y += cLineFeedSize; \
    }
#define NEXT_LINE_SUB(group)    \
    {                           \
        y += s_LineFeed[group]; \
    }

#define PRINT_COLOR(color, x, ...)                       \
    \
{                                                 \
        pSampleFramework->SetFontColor(color);           \
        pSampleFramework->DrawText((x), y, __VA_ARGS__); \
        pSampleFramework->SetFontColor(cDefaultColor);   \
    \
}

#define PRINT_CURSOR(target, x, ...)                      \
    \
{                                                  \
        if (s_OperationTarget == (target))                \
        {                                                 \
            PRINT_COLOR(cCursorColor, (x), __VA_ARGS__);  \
        }                                                 \
        else if (IsValidOperationTarget(target) == false) \
        {                                                 \
            PRINT_COLOR(cInvalidColor, (x), __VA_ARGS__); \
        }                                                 \
        else                                              \
        {                                                 \
            PRINT((x), __VA_ARGS__);                      \
        }                                                 \
    \
}


// Draw on the screen.
void Draw(GameMode gameMode)
{
    // Updates the width of the new line character.
    {
        OperationGroup group = GetOperationGroup(s_OperationTarget);
        for (u32 i = 0; i < OPERATION_GROUP_NUM; ++i)
        {
            if (i == group)
            {
                if (s_LineFeed[i] < cLineFeedSize)
                {
                    ++s_LineFeed[i];
                }
            }
            else
            {
                if (s_LineFeed[i] > 0)
                {
                    --s_LineFeed[i];
                }
            }
        }
    }

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();

    static const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    static const nn::pia::sampleframework::Color cCursorColor(1.0f, 0.3f, 0.3f, 1.0f);
    static const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);
    static const nn::pia::sampleframework::Color cBarBackColor(0.2f, 0.2f, 0.2f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);

    s32 y = 4;

    // Common displays.
    {
        PRINT(4, "PiaTransport ReliableBroadcast");
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        PRINT(380, "%c", cAnimLetter[idx]);
        NEXT_LINE();

        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                cGameModeText = "LocalHost";
            }
            else
            {
                cGameModeText = "LocalClient";
            }
            break;
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
            if (nn::pia::session::Session::GetInstance()->IsHost())
            {
                cGameModeText = "InetHost";
            }
            else
            {
                cGameModeText = "InetClient";
            }
            break;
        case GAME_MODE_INET_AUTOMATCH:
            PIA_SAMPLE_ASSERTMSG(false, "InetAutomatch?");
            cGameModeText = "InetAutomatch";
            break;
        default:
            PIA_SAMPLE_ASSERTMSG(false, "Unknown game mode? (%d)", gameMode);
            cGameModeText = "???";
            break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        PRINT(4, "GameMode:%s", cGameModeText);
        NEXT_LINE();

        PRINT(4, "StationNum:%d", pSession->GetStationNum());
        PRINT(180, "Local:0x%02x  Host:0x%02x",
              conv2StationIndex(pSession->GetLocalStationId()) & 0xff,
              conv2StationIndex(pSession->GetHostStationId()) & 0xff);
        NEXT_LINE();
    }

    y += 3;

    // State.
    {
        const char* cpStateName = "";
        switch (pProtocol->GetState())
        {
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_INVALID:
            cpStateName = "Invalid";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_WAIT:
            cpStateName = "Wait";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING:
            cpStateName = "Sending";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_SUCCESS:
            cpStateName = "Send Success";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_FAILURE:
            cpStateName = "Send Failure";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVING:
            cpStateName = "Receiving";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVE_SUCCESS:
            cpStateName = "Receive Success";
            break;
        case nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVE_FAILURE:
            cpStateName = "Receive Failure";
            break;
        }
        PRINT(4, "State: %s", cpStateName);
        NEXT_LINE();
    }

    y += 3;

    // Receive setting.
    if (s_LineFeed[OPERATION_GROUP_RECEIVE] > 0)
    {
        PRINT(4, "-Receive");
        NEXT_LINE();

        PRINT(20, "request:");
        if (s_ReceiveStationId == nn::pia::STATION_ID_INVALID)
        {
            PRINT_CURSOR(OPERATION_RECEIVE_STATION_ID, 220, "top priority");
        }
        else
        {
            PRINT_CURSOR(OPERATION_RECEIVE_STATION_ID, 220, "station 0x%02x", conv2StationIndex(s_ReceiveStationId) & 0xff);
            if (pSession->CheckStationIdIsValid(s_ReceiveStationId) == false)
            {
                PRINT(300, "disconnected");
            }
        }
        NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

        const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpRequest = (s_ReceiveStationId == nn::pia::STATION_ID_INVALID) ? pProtocol->GetRequest() : pProtocol->GetRequest(s_ReceiveStationId);

        if (cpRequest != NULL)
        {
            PRINT(20, "source stationId:");
            nn::pia::StationId srcId = cpRequest->GetSourceStationId();
            PRINT(220, "0x%02x", conv2StationIndex(srcId) & 0xff);
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "data size:");
            PRINT(220, "%7dB (%4.2fMB)", cpRequest->m_DataSize, static_cast<f32>(cpRequest->m_DataSize) / (1024.0f * 1024.0f));
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "offset:");
            PRINT(220, "%7dB", cpRequest->m_UserData);
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "send to newcomer:");
            PRINT(220, cpRequest->m_Configuration.IsSendToNewcomer() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "synchronize completion:");
            PRINT(220, cpRequest->m_Configuration.IsSynchronizeCompletion() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "send endless:");
            PRINT(220, cpRequest->m_Configuration.IsSendEndless() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "fail if rejected:");
            PRINT(220, cpRequest->m_Configuration.IsFailIfRejected() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "mediate request:");
            PRINT(220, cpRequest->m_Configuration.IsMediateRequest() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);
        }
        else
        {
            PRINT(20, "source stationId:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "data size:");
            PRINT(220, " ----- B ( -- MB)");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "offset:");
            PRINT(220, " ----- B");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "send to newcomer:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "synchronize completion:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "send endless:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "fail if rejected:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

            PRINT(20, "mediate request:");
            PRINT(220, "-");
            NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);
        }

        PRINT(20, "receive buffer size:");
        PRINT_CURSOR(OPERATION_RECEIVE_BUFFER_SIZE, 220, "%7dB (%4.2fMB)", s_ReceiveBufferSize, static_cast<f32>(s_ReceiveBufferSize) / (1024.0f * 1024.0f));
        NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

        PRINT_CURSOR(OPERATION_START_RECEIVE, 20, "> Start receive");
        NEXT_LINE_SUB(OPERATION_GROUP_RECEIVE);

        PRINT_CURSOR(OPERATION_REJECT_REQUEST, 20, "> Reject request");
    }
    else
    {
        PRINT_COLOR(cInvalidColor, 4, "+Receive");
        NEXT_LINE();

        PRINT_COLOR(cInvalidColor, 20, "...");
    }
    NEXT_LINE();

    // Send setting.
    if (s_LineFeed[OPERATION_GROUP_SEND] > 0)
    {
        PRINT(4, "-Send");
        NEXT_LINE();

        PRINT(20, "data size:");
        PRINT_CURSOR(OPERATION_SEND_SIZE, 220, "%7dB (%4.2fMB)", s_SendDataSize, static_cast<f32>(s_SendDataSize) / (1024.0f * 1024.0f));
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT(20, "send to newcomer:");
        PRINT_CURSOR(OPERATION_CONFIG_SEND_TO_NEWCOMER, 220, s_SendConfig.IsSendToNewcomer() ? "yes" : "no");
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT(20, "synchronize completion:");
        PRINT_CURSOR(OPERATION_CONFIG_SYNCHRONIZE_COMPLETION, 220, s_SendConfig.IsSynchronizeCompletion() ? "yes" : "no");
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT(20, "send endless:");
        PRINT_CURSOR(OPERATION_CONFIG_SEND_ENDLESS, 220, s_SendConfig.IsSendEndless() ? "yes" : "no");
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT(20, "fail if rejected:");
        PRINT_CURSOR(OPERATION_CONFIG_FAIL_IF_REJECTED, 220, s_SendConfig.IsFailIfRejected() ? "yes" : "no");
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT(20, "mediate request:");
        PRINT_CURSOR(OPERATION_CONFIG_MEDIATE_REQUEST, 220, s_SendConfig.IsMediateRequest() ? "yes" : "no");
        NEXT_LINE_SUB(OPERATION_GROUP_SEND);

        PRINT_CURSOR(OPERATION_START_SEND, 20, "> Start send");
    }
    else
    {
        PRINT_COLOR(cInvalidColor, 4, "+Send");
        NEXT_LINE();

        PRINT_COLOR(cInvalidColor, 20, "...");
    }
    NEXT_LINE();

    // Progress display.
    if (s_LineFeed[OPERATION_GROUP_PROGRESS] > 0)
    {
        PRINT(4, "-Progress");
        NEXT_LINE();

        const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpTransferSetting = pProtocol->GetCurrentTransferSetting();
        if (cpTransferSetting != NULL)
        {
            PRINT(20, "source stationId:");
            nn::pia::StationId srcId = cpTransferSetting->GetSourceStationId();
            PRINT(220, "0x%02x", conv2StationIndex(srcId) & 0xff);
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            {
                u32 total = cpTransferSetting->m_DataSize;
                u32 current = pProtocol->GetProgress();
                f32 rate = static_cast<f32>(current) / static_cast<f32>(total);
                u32 barLength = static_cast<u32>(rate * 300);
                pSampleFramework->DrawRect(20, y, barLength, 8, cDefaultColor);
                pSampleFramework->DrawRect(20 + barLength, y, 300 - barLength, 8, cBarBackColor);
                NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

                PRINT(20, "%7dB / %7dB (%5.1f%%)", current, total, rate * 100.0f);
                NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);
            }

            switch (pProtocol->GetState())
            {
            case nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING:
            case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_SUCCESS:
            case nn::pia::transport::ReliableBroadcastProtocol::STATE_SEND_FAILURE:
                for (u32 i = 0; i < MAX_STATIONS; ++i)
                {
                    nn::pia::StationId stationId = conv2StationId(static_cast<nn::pia::StationIndex>(i));
                    if (stationId == pSession->GetLocalStationId())
                    {
                        continue;
                    }

                    const nn::pia::sampleframework::Color& color = pProtocol->IsSendTarget(stationId) ? cDefaultColor : cInvalidColor;

                    PRINT_COLOR(color, 30, "%02x:", conv2StationIndex(stationId) & 0xff);
                    if (pSession->CheckStationIdIsValid(stationId))
                    {
                        u32 current = pProtocol->GetSendProgress(stationId);
                        f32 rate = static_cast<f32>(current) / static_cast<f32>(cpTransferSetting->m_DataSize);
                        u32 barLength = static_cast<u32>(rate * 200);
                        pSampleFramework->DrawRect(50, y, barLength, 8, color);
                        pSampleFramework->DrawRect(50 + barLength, y, 200 - barLength, 8, cBarBackColor);

                        PRINT_COLOR(color, 260, "%7dB (%5.1f%%)", current, rate * 100.0f);
                    }
                    else
                    {
                        PRINT_COLOR(color, 50, "disconnected");
                    }
                    NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);
                }
                break;
            }

            PRINT(20, "offset:");
            PRINT(220, "%7dB", cpTransferSetting->m_UserData);
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            PRINT(20, "send to newcomer:");
            PRINT(220, cpTransferSetting->m_Configuration.IsSendToNewcomer() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            PRINT(20, "synchronize completion:");
            PRINT(220, cpTransferSetting->m_Configuration.IsSynchronizeCompletion() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            PRINT(20, "send endless:");
            PRINT(220, cpTransferSetting->m_Configuration.IsSendEndless() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            PRINT(20, "fail if rejected:");
            PRINT(220, cpTransferSetting->m_Configuration.IsFailIfRejected() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

            PRINT(20, "mediate request:");
            PRINT(220, cpTransferSetting->m_Configuration.IsMediateRequest() ? "yes" : "no");
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);
        }
        else if (pProtocol->GetState() == nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVING ||
                 pProtocol->GetState() == nn::pia::transport::ReliableBroadcastProtocol::STATE_RECEIVE_FAILURE)
        {

            PRINT(20, "source stationId:");
            PRINT(220, "0x%02x", conv2StationIndex(s_ReceiveStationId) & 0xff);
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);
        }

        PRINT(20, "throughput limit:");
        PRINT_CURSOR(OPERATION_THROUGHPUT_LIMIT, 220, "%6dB", pProtocol->GetThroughputLimit());
        NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);

        if (pProtocol->IsRunning())
        {
            PRINT_CURSOR(OPERATION_CANCEL, 20, "> Cancel");
        }
        else
        {
            PRINT_CURSOR(OPERATION_RESET, 20, "> Reset");
        }

        if (pProtocol->GetState() == nn::pia::transport::ReliableBroadcastProtocol::STATE_SENDING)
        {
            NEXT_LINE_SUB(OPERATION_GROUP_PROGRESS);
            PRINT_CURSOR(OPERATION_FINISH_SEND_ENDLESS, 20, "> Finish send endless");
        }
    }
    else
    {
        PRINT_COLOR(cInvalidColor, 4, "+Progress");
        NEXT_LINE();

        PRINT_COLOR(cInvalidColor, 20, "...");
    }

    // Message.
    {
        y = 226;
        PRINT(4, "Message: %s", s_Message);
    }
}


#if PIA_TEST_CODE
void UpdatePiaTestData(GameMode gameMode, bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::transport::ReliableBroadcastProtocol* pProtocol = GetReliableBroadcastProtocol();

    bool isInet = false;
    switch (gameMode)
    {
    case GAME_MODE_INET_MASTER:
    case GAME_MODE_INET_CLIENT:
    case GAME_MODE_INET_AUTOMATCH:
        isInet = true;
        break;
    }

    pSampleFramework->SetProgramStatus(
        isExit ? 0 : (isInet ? 4 : 3),
        pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION,
        pSession->GetLocalStationId() == pSession->GetHostStationId(),
        pSession->GetStationNum());

    // Session.
    pSampleFramework->SetProgramParam(0x00, conv2StationIndex(pSession->GetLocalStationId()));

    // Operations.
    pSampleFramework->SetProgramParam(0x08, s_OperationTarget);

    // Status.
    pSampleFramework->SetProgramParam(0x10, pProtocol->GetState());
    pSampleFramework->SetProgramParam(0x11, pProtocol->IsRunning() ? 1 : 0);

    // Are operations possible?
    for (int i = 0; i < OPERATION_TARGET_NUM; ++i)
    {
        pSampleFramework->SetProgramParam(0x20 + i, IsValidOperationTarget(static_cast<OperationTarget>(i)) ? 1 : 0);
    }

    // Receive setting.
    pSampleFramework->SetProgramParam(0x30, conv2StationIndex(s_ReceiveStationId));
    pSampleFramework->SetProgramParam(0x31, s_ReceiveBufferSize);

    // Send setting.
    pSampleFramework->SetProgramParam(0x40, s_SendDataSize);
    pSampleFramework->SetProgramParam(0x41, s_SendConfig.IsSendToNewcomer());
    pSampleFramework->SetProgramParam(0x42, s_SendConfig.IsSynchronizeCompletion());
    pSampleFramework->SetProgramParam(0x43, s_SendConfig.IsSendEndless());
    pSampleFramework->SetProgramParam(0x44, s_SendConfig.IsFailIfRejected());
    pSampleFramework->SetProgramParam(0x45, s_SendConfig.IsMediateRequest());

    // Progress.
    pSampleFramework->SetProgramParam(0x50, pProtocol->GetThroughputLimit());
    pSampleFramework->SetProgramParam(0x51, pProtocol->GetProgress());
#if ENABLE_PROFILING
    pSampleFramework->SetProgramParam(0x52, s_TimeElapsed);
#endif

    // State of each peer station.
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        nn::pia::StationId sid = conv2StationId(static_cast<nn::pia::StationIndex>(i));
        u32 idOfs = (i + 6) << 4;

        if (pSession->CheckStationIdIsValid(sid) && sid != pSession->GetLocalStationId())
        {
            pSampleFramework->SetProgramParam(idOfs + 0x0, 1);

            pSampleFramework->SetProgramParam(idOfs + 0x4, pProtocol->GetSendProgress(sid));

            const nn::pia::transport::ReliableBroadcastProtocol::TransferSetting* cpRequest = pProtocol->GetRequest(sid);
            if (cpRequest != NULL)
            {
                pSampleFramework->SetProgramParam(idOfs + 0x8, 1);
                pSampleFramework->SetProgramParam(idOfs + 0x9, cpRequest->m_DataSize);
                pSampleFramework->SetProgramParam(idOfs + 0xa, cpRequest->m_Configuration.IsSendToNewcomer());
                pSampleFramework->SetProgramParam(idOfs + 0xb, cpRequest->m_Configuration.IsSynchronizeCompletion());
                pSampleFramework->SetProgramParam(idOfs + 0xc, cpRequest->m_Configuration.IsSendEndless());
                pSampleFramework->SetProgramParam(idOfs + 0xd, cpRequest->m_Configuration.IsFailIfRejected());
                pSampleFramework->SetProgramParam(idOfs + 0xe, cpRequest->m_Configuration.IsMediateRequest());
                pSampleFramework->SetProgramParam(idOfs + 0xf, cpRequest->m_UserData);
            }
            else
            {
                for (u32 j = 0x8; j <= 0xf; ++j)
                {
                    pSampleFramework->SetProgramParam(idOfs + j, 0);
                }
            }
        }
        else
        {
            for (u32 j = 0; j < 16; ++j)
            {
                pSampleFramework->SetProgramParam(idOfs + j, 0);
            }
        }
    }
}
#endif

// Main loop during communication.
void MainLoop(GameMode gameMode)
{
    nn::Result r;
    while (true)
    {
        bool isExit = false;
        bool bCallLeaveSession = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::Result connectionStatus = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
        if (connectionStatus.IsFailure())
        {
            // Ends because the communication has been disconnected.
            isExit = true;
        }

        // Check Transport connection status
        nn::Result result = nn::pia::transport::Transport::GetInstance()->CheckConnectionError();
        if (result.IsSuccess())
        {
            // If Transport has no problems, check the Session connection status.
            result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
        }
        else
        {
            PIA_SAMPLE_LOG("Transport::CheckConnectionError() failed.\n");
        }

        // Error handling.
        if (result.IsFailure())
        {
            enum nn::pia::util::HandlingType handling = nn::pia::util::GetHandlingType(result);
            switch(handling)
            {
            case nn::pia::util::HANDLING_LEAVE_AND_CLEANUP:
                PIA_SAMPLE_LOG("Exit MainLoop() with LeaveSession().\n");
                bCallLeaveSession = true;
                isExit = true;
                break;
            case nn::pia::util::HANDLING_RECONNECT:
                PIA_SAMPLE_LOG("Exit MainLoop().\n");
                isExit = true;
                break;
            default:
                // The <tt>GetHandlingType()</tt> function is capable of returning values other than <tt>HANDLING_LEAVE_AND_CLEANUP</tt> and
                // <tt>HANDLING_RECONNECT</tt>, but when the <tt>Result</tt> objects that can be returned by the
                // <tt>Transport::CheckConnectionError</tt> and <tt>Session::CheckConnectionError</tt> functions are considered,
                // values other than <tt>HANDLING_LEAVE_AND_CLEANUP</tt> and <tt>HANDLING_RECONNECT</tt> are never returned.
                // An assert statement is provided for that reason.
                PIA_SAMPLE_ASSERTMSG(false, "Unexpected result.\n");
                break;
            }
        }

        if (nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION)
        {
            // Send/receive processing takes place if during communication.
            isExit |= Operate();

            CheckCompletion();
        }
        else
        {
            // Ends because the session was lost.
            isExit = true;
        }

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        Draw(gameMode);

#if PIA_TEST_CODE
        UpdatePiaTestData(gameMode, isExit);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            if (bCallLeaveSession)
            {
                Session_LeaveSession();
            }

            return;
        }
    }
}

}  // end of anonymous namespace


// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;
    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
        nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(60);
    }

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    InitOriginalData();

    while (true)
    {
#if PIA_TEST_CODE
        s_OptionalParam.Initialize(OPTIONAL_PARAM_OFFSET);
        GameMode gameMode = DecideGameMode("PiaTransport ReliableBroadcast", &s_OptionalParam);
#else
        GameMode gameMode = DecideGameMode("PiaTransport ReliableBroadcast");
#endif
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
#if 0
            u64 traceFlag = 0;
#else
            u64 traceFlag =
                nn::pia::common::TRACE_FLAG_COMMON |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_LOCAL |
                nn::pia::common::TRACE_FLAG_SESSION;
#endif
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlag, true);
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            {
#if ENABLE_HOST_MIGRATION
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID, true);
#else
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID);
#endif
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Local_InitializePiaLocal() failed.\n");
                    Local_FinalizePiaLocal();
                    Common_FinalizePiaCommon();
                    continue;
                }
                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            nn::pia::transport::Transport::Setting transportSetting;
            transportSetting.pFactory = pFactory;
            transportSetting.maxStationNum = MAX_STATIONS;
            transportSetting.maxSendThreadBufferNum = 16;
            transportSetting.maxReceiveThreadBufferNum = 32;

            result = Transport_InitializePiaTransport(
                transportSetting,
                0,
                NULL,
                0,
                NULL,
                64,
                64,
                CreateAndInitializeProtocol,
                32);
            PIA_SAMPLE_ASSERT(result.IsSuccess());

            // Sets up the session.
            nn::pia::session::Session::Setting sessionSetting;
            sessionSetting.pNetworkFactory = pFactory;
            sessionSetting.browsedSessionInfoListNum = 16;

#if PIA_TEST_CODE
            if(s_OptionalParam.bRelayEmulation)
            {
                // If you are performing relay emulation, you need to call the settings function before creating the session instance.
                nn::pia::session::Session::DebugSetting debugSetting;
                debugSetting.isEnableRelayEmulation = true;
                nn::Result ret = nn::pia::session::Session::SetDebugSetting(debugSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(ret);
                sessionSetting.networkTopology = nn::pia::session::NetworkTopology_RelayMesh;
            }
            else
            {
                sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
            }
            result = Session_InitializePiaSession(sessionSetting, s_OptionalParam.bRelayEmulation);

#else
            sessionSetting.networkTopology = nn::pia::session::NetworkTopology_FullMesh;
            result = Session_InitializePiaSession(sessionSetting);
#endif

            PIA_SAMPLE_ASSERT(result.IsSuccess());
        }


        // Connect to the network connections and create a session.
        {
            nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SessionEventCallback);
            InitializeGameData();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the matchmaking session to create.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                        createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
                    searchCriteriaList[0].SetResultRange(0, 16);
                }

                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_BrowseSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connects to the Internet.
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }
                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus sampleSessionAsyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (sampleSessionAsyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

#if PIA_TEST_CODE
                if(s_OptionalParam.bRelayEmulation)
                {
                    SetRelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3);
                }
#endif

                bool isHost = nn::pia::session::Session::GetInstance()->IsHost();
                if (isHost)
                {
                    // Became the host as a result of the matching.
                    gameMode = GAME_MODE_INET_MASTER;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;

            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_MASTER:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_MASTER:
        case GAME_MODE_INET_CLIENT:
            MainLoop(gameMode);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }

            break;
        }


        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
                // Leave the session.
                Session_LeaveSession();
                break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
                break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
                Inet_Logout();
                Inet_FinalizeNetwork();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }


        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Transport_FinalizePiaTransport(0, 0, FinalizeAndDestroyProtocol);
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_MASTER:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
                Local_FinalizePiaLocal();
                break;
#endif
            case GAME_MODE_INET_MASTER:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
            Common_FinalizePiaCommon();
        }
    }

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
