﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     reckoning_Simple3dReckoningStrategy.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/reckoning/reckoning_Simple3dReckoningStrategy.h>

#include <pia/common/common_Trace.h>

namespace nn
{
namespace pia
{
namespace reckoning
{

/*!
@brief  Sets the default threshold value.
*/
Simple3dReckoningStrategy::Simple3dReckoningStrategy(f32 threshold)
  : m_Threshold(threshold)
{
}

void Simple3dReckoningStrategy::SetThreshold(f32 threshold)
{
    // Restrict the range so the threshold value does not take a value smaller than 0.
    if (threshold < 0.f)
    {
        m_Threshold = 0.f;
    }
    else
    {
        m_Threshold = threshold;
    }
}


static void calcParameter(float* b, const float& v, const float& t) { *b = v / t; }



static void calcEstimation(float* out, const float& a, const float& b, const float& t) {
    *out = (a ) * t + b;
}


/*
void Simple3dReckoningStrategy::calcEstimationDelay( Value* pValue ,  const Accessor& accessor , float clock ) const {


    s32 beginIndex = 0;
    while( beginIndex < accessor.GetNum() ){
        if( clock > ( float)( accessor.GetClock( beginIndex ) ) ){

            break;
        } else {

            beginIndex ++;
        }
    }

    s32 index0,index1;

    float tt;

    Value baseValue;
    if( beginIndex == 0 ){
        index0 = 0 ;
        index1 = 1;

        tt = clock - (float)( accessor.GetClock( index0 ) );

        baseValue = accessor.GetSample( index0 ).m_Value;
    }else if( beginIndex == accessor.GetNum() ){
        index0 = beginIndex -2;
        index1 = beginIndex - 1;

        tt = clock - (float)( accessor.GetClock( index1 ) );
        baseValue = accessor.GetSample( index1 ).m_Value;

    } else {

        index0 = beginIndex - 1;
        index1 = beginIndex;

        tt = clock - (float)( accessor.GetClock( index1 ) );
        baseValue = accessor.GetSample( index1 ).m_Value;

    }


    Value value1 = accessor.GetSample( index1 ).m_Value;
    nn::pia::reckoning::Vector3f::Mul( &value1 , -1.0f );

    Value diff = accessor.GetSample( index0 ).m_Value;
    nn::pia::reckoning::Vector3f::Add( &diff , value1 );


    nn::pia::reckoning::Vector3f::Mul( &diff , tt / (float)( accessor.GetClock( index0 ) - accessor.GetClock( index1 ) ) );


    nn::pia::reckoning::Vector3f::Add( &baseValue , diff );
    
    *pValue = baseValue;

}
*/



bool Simple3dReckoningStrategy::Estimate(Value* pValue, const Accessor& accessor,
                                    nn::pia::clone::ClockValue clock, const Work* cpWork) const {

//    PIA_SAMPLE_PRINT(" Estimate \n");
    // If there are no sample values, do not use the estimate value.
    if (accessor.GetNum() <= 0) {
        return false;
    }

    // If the current time is earlier than the time corresponding to the earliest sample value, use the sample value as the estimate value.
    // 
    if (clock < accessor.GetClock(accessor.GetNum() - 1)) {
        *pValue = accessor.GetSample(accessor.GetNum() - 1).m_Value;
        return true;
    }

    float tClock = (float)(int)(clock - cpWork->m_BaseClock);

    // If you pass the stop flag when calling the `SetValue` function, the most recent sample value will be used as the estimate value.
    // 


    if (accessor.GetSample(0).m_IsStop) {
        *pValue = accessor.GetSample(0).m_Value;
        return true;
    }


    // Use the pre-calculated values to perform the estimation.

    calcEstimation(&pValue->x, cpWork->m_Param.x, cpWork->m_ParamBase.x,  tClock );
    calcEstimation(&pValue->y, cpWork->m_Param.y, cpWork->m_ParamBase.y,  tClock );
    calcEstimation(&pValue->z, cpWork->m_Param.z, cpWork->m_ParamBase.z,  tClock );
/*
    if( accessor.GetNum() >= 3 ){


        float calcClock;

        if( m_IsGetTiming == true ){
            calcClock = (float)( clock ) - m_DelayClock;
        } else {

            calcClock = (float)( clock);
       }

        calcEstimationDelay( pValue , accessor , calcClock );

    } else {

    calcEstimation(&pValue->x, cpWork->m_Param.x, cpWork->m_ParamBase.x,  tClock );
    calcEstimation(&pValue->y, cpWork->m_Param.y, cpWork->m_ParamBase.y,  tClock );
    calcEstimation(&pValue->z, cpWork->m_Param.z, cpWork->m_ParamBase.z,  tClock );

    }
*/
    return true;
}



bool Simple3dReckoningStrategy::CheckSample(bool* pIsReliable, const Value& value,
                                       const SetValueArg* cpSetValueArg, const Accessor& accessor,
                                       nn::pia::clone::ClockValue clock, const Work* cpWork) const {

    // If there are no sample values, specify the reliable communication protocol and send the value set with the `SetValue` function.
    if (accessor.GetNum() <= 0) {
        *pIsReliable = true;
        return true;
    }

    // If the Stop flag was not set for the previous sample value but is set for this sample value, use reliable communication to send the value set by SetValue().
    // 
    if (accessor.GetSample(0).m_IsStop != true && *cpSetValueArg == true) {
        *pIsReliable = true;
        return true;
    }

    // If the estimate value cannot be calculated because there aren't enough samples, send the value set with the `SetValue` function.
    Value estimatedValue;
    bool r = Estimate(&estimatedValue, accessor, clock, cpWork);
    if (r == false) {
        return true;
    }

    // If the difference between the estimated value and the value set with the `SetValue` function is larger than the specified threshold value, send the value set with the `SetValue` function.

    f32 dis = value.SquaredDistance(estimatedValue);
    bool isCheck =  dis >= m_Threshold * m_Threshold;

    return isCheck;
}


void Simple3dReckoningStrategy::MakeSample(Sample* pSample, const Value& value,
                                      const SetValueArg* cpSetValueArg, const Accessor&,
                                      nn::pia::clone::ClockValue, const Work*) const {

    // Use the value set with the `SetValue` function as a sample value.
    (*pSample).m_Value = value;
    //
    (*pSample).m_IsStop = *cpSetValueArg;
}

/*!
@brief  Determines the coefficients of a second-order evaluation function and saves them to <tt>Work</tt>.
*/
void Simple3dReckoningStrategy::OnUpdateSample(const Accessor& accessor, int, Work* pWork) const {

    // When there are no sample values:
    if (accessor.GetNum() <= 0) {
        pWork->m_Param.MakeZero();
        pWork->m_ParamBase.MakeZero();
        pWork->m_BaseClock = 0;
        return;
    }

    Value v0 = accessor.GetSample(0).m_Value;

//  PIA_SAMPLE_PRINT("OnUpdate Sample %f \n", v0.x );
    nn::pia::clone::ClockValue c0 = accessor.GetClock(0);

    pWork->m_Param.MakeZero();
    pWork->m_ParamBase = v0;
    pWork->m_BaseClock = c0;


    // Determines the coefficient for linear interpolation.
    if (accessor.GetNum() >= 2) {
        nn::pia::clone::ClockValue c1 = accessor.GetClock(1);

        PIA_ASSERT(c0 != c1);
        {
            float t = (float)(int)(c0 - c1);
            Value paramBase = pWork->m_ParamBase;
            nn::pia::reckoning::Vector3f value1 = accessor.GetSample(1).m_Value;
            nn::pia::reckoning::Vector3f::Mul(&value1 , -1.f);
            nn::pia::reckoning::Vector3f::Add(&paramBase, value1 );

            calcParameter(&pWork->m_Param.x, paramBase.x, t);
            calcParameter(&pWork->m_Param.y, paramBase.y, t);
            calcParameter(&pWork->m_Param.z, paramBase.z, t);

            return;
        }
    }

}

// Print information useful for debugging.
void Simple3dReckoningStrategy::Trace(u64 flag) const
{
#if NN_PIA_ENABLE_TRACE
    PIA_TRACE(flag, "Simple3dReckoningStrategy::Trace() called.");
#else   // else of NN_PIA_ENABLE_TRACE
    NN_PIA_DUMMY_PARAM(flag);
#endif  // end of NN_PIA_ENABLE_TRACE
}
}
}
}  // end of namespace nn::pia::reckoning
