﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: ProfilingTest_Local.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "ProfilingTest_Local.h"
#include "ProfilingTest_Main.h"
#include "ProfilingTest_P2P.h"

#include "Shared_Nex_Setup.h"
#include "Shared_Local_Setup.h"

#include "Shared_Session_Setup.h"

#if ENABLE_NEX_NETZ
#include <pia/local/local_NexVirtualNetwork.h>
#else
#include <pia/transport/transport_Transport.h>
#include <pia/session/session_Session.h>
#endif

//Channel used when building UDS networks.
//Specify 0, 1, 6, or 11.
const u8 UDS_NETWORK_CHANNEL = 11;
const u32 TEST_TITLE_ID = SAMPLE_TITLE_ID;  // For session.
const u8 TEST_SUB_ID = SAMPLE_SUB_ID;
const char TEST_PASSPHRASE[] = "aaabbccc0123456789!";

static u32 communicationId;

nn::Result InitializeLocal()
{
    nn::Result result;

#if ENABLE_NEX_NETZ
    result = Nex_InitializeNex(
        SAMPLE_NEX_MEMORY_SIZE,
        SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
        SAMPLE_NEX_THREAD_MODE);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("InitializeLocal:Nex_InitializeNex: failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        // Clean up using <tt>FinalizeLocal</tt>.
        return SampleNexInvalidState();
    }

#if ENABLE_MEMORY_PROFILING && USE_MULTIUNITHEAP //n1769
    Nex_EnableMultiUnitHeapPeakRecording();
#endif

    // NAT traversal processing timeout settings.
    nn::nex::RootTransport::SetNATTraversalTimeout(MATCHMAKE_PARTICIPANTS_MAX * 5000);

    // Disable a LAN session.
    nn::nex::LANSessionDiscovery::Disable();

#if NN_PIA_NEX_3
//  NEX3 is not defined- - -
#else
    nn::nex::ProductFacade::UseUserDO(false);
#endif

#if ENABLE_TRACE_NEX
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);

    nn::nex::TraceLog::GetInstance()->SetFlag(
        TRACE_ALWAYS
#if ENABLE_TRACE_NEX_DETAIL
        |
        TRACE_NAT | TRACE_UDS
#if 0
        | TRACE_INIT
        | TRACE_TERMINATE
        | TRACE_FAULT

        | D_TRACE_NAT_TRAVERSAL
        | D_TRACE_TRANSPORT_EXPORT

        | TRACE_TIME
        | TRACE_CONNECTION
        | TRACE_TEARDOWN
        | TRACE_UPDATE
        | TRACE_TRANSPORT
        | TRACE_PHBDR
        | TRACE_OPERATION
        | TRACE_DUPSPACE
        | TRACE_LOADBALANCING
        | TRACE_SESSIONDISCOVERY
        | TRACE_SCHEDULER

        | D_TRACE_TEST
        | D_TRACE_DO_MIGRATION
        | D_TRACE_THREAD
        | D_TRACE_CONSTRUCT
        | D_TRACE_TRANSPORT
        | D_TRACE_TRANSPORT_END_POINT
        | D_TRACE_DO_MESSAGE_BUNDLING
        | D_TRACE_TRANSPORT_SEND_MULTI
        | D_TRACE_TRANSPORT_BUNDLING
        | D_TRACE_TRANSPORT_RELIABLE
        | D_TRACE_TRANSPORT_ROUTING
        | D_TRACE_TRANSPORT_PACKET
        | D_TRACE_TRANSPORT_HOST_MIGRATION
        | D_TRACE_UDS
#endif
#endif
        );
#endif  // ENABLE_TRACE_NEX

#endif  // ENABLE_NEX_NETZ

    result = Local_InitializePiaLocal(TEST_TITLE_ID, TEST_SUB_ID);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("InitializeLocal: Local_InitializePiaLocal failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        // Clean up using <tt>FinalizeLocal</tt>.
        return SampleNexInvalidState();
    }

    PIA_SAMPLE_LOG("InitializeLocal: succeeded\n");
    return nn::ResultSuccess();
}

void FinalizeLocal()
{
    Local_FinalizePiaLocal();
#if ENABLE_NEX_NETZ
    Nex_FinalizeNex();
#endif
}

//-----------------------------------------------------------------------------
// Main loop.
//-----------------------------------------------------------------------------
void MainLoopLocal()
{
    nn::Result result;

    for (g_MatchMakeLoopCnt = 0; g_MatchMakeLoopCnt < MATCHMAKE_TRY_NUM; ++g_MatchMakeLoopCnt)
    {
        communicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(TEST_TITLE_ID);

#if ENABLE_AUTO_SESSION_REPEAT
        s32 sleep = (static_cast<u32>(nn::os::Tick::GetSystemCurrent()) % 100) * 5 + 200;
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(sleep));
        bool isHost = static_cast<u32>(nn::os::Tick::GetSystemCurrent()) % MATCHMAKE_PARTICIPANTS_MAX == 0;

        g_State = STATE_SETUP_AUTOMATCH;
        result = ProcessAutoMatchmakeLocal(isHost);

#else  // ENABLE_AUTO_SESSION_REPEAT
        PIA_SAMPLE_PRINT("Press A : Start Local Host\n");
        PIA_SAMPLE_PRINT("Press Y : Start Local Client\n");
        PIA_SAMPLE_PRINT("Press B : Logout\n");
        bit32 button = nn::pia::sampleframework::SampleFramework::GetInstance()->WaitForButton(
            nn::pia::sampleframework::BUTTON_A | nn::pia::sampleframework::BUTTON_B | nn::pia::sampleframework::BUTTON_Y);

        if (button == nn::pia::sampleframework::BUTTON_B)
        {
            break;
        }

        g_State = STATE_SETUP_AUTOMATCH;
        result = ProcessAutoMatchmakeLocal(button == nn::pia::sampleframework::BUTTON_A);

#endif  // ENABLE_AUTO_SESSION_REPEAT

        // Ends here if matchmaking succeeds.
        if (result.IsSuccess())
        {
            PIA_SAMPLE_LOG("MainLoopLocal: ProcessAutoMatchmakeLocal succeeded\n");
        }
        else
        {
            PIA_SAMPLE_LOG("MainLoopLocal: ProcessAutoMatchmakeLocal failed.\n");
            ShowErrorCodeLoop(result);
            break;
        }
    }

    g_State = STATE_CLEANUP;
}

#if ENABLE_NEX_NETZ
void UdsNetworkTerminator(nn::pia::local::NexVirtualNetwork* pNexVirtualNetwork)
{
    if (pNexVirtualNetwork)
    {
        nn::nex::VirtualSocketDriver::UnregisterVirtualNetwork();
        // Before destroying the <tt>NexVirtualNetwork</tt> instance, unregister <tt>VirturalSocketDriver</tt> settings.
        qDelete pNexVirtualNetwork;
    }
}
#endif

nn::Result ProcessAutoMatchmakeLocal(bool isHost)
{
    nn::Result result = nn::ResultSuccess();
    bool bSuccess = false;
    PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: %d\n", isHost);

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_ASSERT(pSession != NULL);

    nn::pia::session::Session::StartupSetting startupSetting;
    startupSetting.bUsingHostMigration = static_cast<bool>(ENABLE_HOST_MIGRATION);

    result = Session_StartupSession(startupSetting);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: Session::GetInstance()->Startup failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    if (isHost)
    {
        g_State = STATE_CREATING_SESSION;

        nn::pia::local::UdsCreateNetworkSetting setting;
        {
            setting.m_LocalCommunicationId = communicationId;
            setting.m_SubId = TEST_SUB_ID;
            setting.m_MaxEntry = MATCHMAKE_PARTICIPANTS_MAX;
            setting.m_Channel = UDS_NETWORK_CHANNEL;
            memcpy(setting.m_Passphrase, TEST_PASSPHRASE, sizeof(TEST_PASSPHRASE));
            setting.m_PassphraseLength = sizeof(TEST_PASSPHRASE);
        }
        nn::pia::local::UdsCreateSessionSetting createSessionSetting;
        createSessionSetting.SetLocalCreateNetworkSetting(setting);
        createSessionSetting.SetMaxParticipants(MATCHMAKE_PARTICIPANTS_MAX);
        createSessionSetting.SetMinParticipants(1);
#if ENABLE_SIGNATURE
        char signatureKeyData[] = "ProfilingTest key";
        createSessionSetting.SetSignatureKey(signatureKeyData, sizeof(signatureKeyData));
#endif

        SampleSessionAsyncStatus asyncStatus;
        asyncStatus = Session_CreateSession(&createSessionSetting, &result);
        if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: CreateP2PSession failed.\n");
            PIA_SAMPLE_LOG_RESULT(result);
            return result;
        }

        bSuccess = true;
    }
    else
    {
        g_State = STATE_JOINING_SESSION;

        // Search for sessions.
        nn::pia::session::ISessionInfo* pTargetGatheringHolder = NULL;
        const u32 scanRetryCnt = 10;
        u32 scanCnt = 0;
        while (true)
        {
            nn::pia::local::LocalSessionSearchCriteria criteria;
            criteria.SetMaxParticipants(MATCHMAKE_PARTICIPANTS_MAX);
            criteria.SetSubId(TEST_SUB_ID);
            criteria.SetLocalCommunicationId(communicationId);
            criteria.SetResultRange(0, 1);
            result = pSession->BrowseSessionAsync(&criteria);
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: BrowseSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                return result;
            }

            while (true)
            {
                if (pSession->IsCompletedBrowseSession())
                {
                    result = pSession->GetBrowseSessionResult();
                    if (result.IsFailure())
                    {
                        PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: BrowseSessionAsync failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        return result;
                    }
                    else
                    {
                        break;
                    }
                }
                result = CalcAsyncProcess();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: Connection is lost\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    return result;
                }
            }

            // Gets the first session.
            nn::pia::session::ISessionInfoList* pGatheringHolderList = pSession->GetBrowsedSessionInfoList();
            if (pGatheringHolderList->GetSize() > 0)
            {
                nn::pia::session::ISessionInfoList::Iterator it = pSession->GetBrowsedSessionInfoList()->Begin();
                pTargetGatheringHolder = *it;
                bSuccess = true;
                break;
            }

            ++scanCnt;
            if (scanCnt > scanRetryCnt)
            {
                PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: TimeOut\n");
                result = nn::pia::ResultCancelled();
                PIA_SAMPLE_LOG_RESULT(result);
                return result;
            }
        }

        // Joins a session.
        PIA_SAMPLE_ASSERT(bSuccess);
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetGatheringHolder));
        nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
        joinSessionSetting.SetSessionInfoPtr(pTargetGatheringHolder);
        joinSessionSetting.SetPassphrase(const_cast<char*>(TEST_PASSPHRASE), sizeof(TEST_PASSPHRASE));
#if ENABLE_SIGNATURE
        char signatureKeyData[] = "ProfilingTest key";
        joinSessionSetting.SetSignatureKey(signatureKeyData, sizeof(signatureKeyData));
#endif
        bool asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
        if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
        {
            PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
            PIA_SAMPLE_LOG_RESULT(result);
            return result;
        }
    }

    // Wait for enough players to join.
    if (bSuccess)
    {
        result = WaitForParticipantLoop();
        if (result.IsFailure())
        {
            PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: WaitForParticipantLoop failed.\n");
            PIA_SAMPLE_LOG_RESULT(result);
            bSuccess = false;
        }
    }

    // Start the main sending/receiving processing.
    if (bSuccess)
    {
        result = GameLoop();
        if (result.IsFailure())
        {
            PIA_SAMPLE_LOG("ProcessAutoMatchmakeLocal: GameLoop failed.\n");
            PIA_SAMPLE_LOG_RESULT(result);
            bSuccess = false;
        }
    }

    result = pSession->LeaveSessionAsync();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("LeaveSessionAsync: failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    while (true)
    {
        if (pSession->IsCompletedLeaveSession())
        {
            result = pSession->GetLeaveSessionResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("ProcessMatchmakeInternet: LeaveSession failed");
                PIA_SAMPLE_LOG_RESULT(result);
            }
            break;
        }

        result = CalcAsyncProcess();
        if (result.IsFailure())
        {
            PIA_SAMPLE_LOG("ProcessMatchmakeInternet: Connection is lost\n");
            PIA_SAMPLE_LOG_RESULT(result);
        }
    }

    Session_CleanupSession();

    return result;
}
