﻿/*---------------------------------------------------------------------------*
Project:  Pia
File:     ProfilingTest_MemoryMonitor.cpp

Copyright Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.
*---------------------------------------------------------------------------*/


#include "ProfilingTest_MemoryMonitor.h"
#include "ProfilingTest_Common.h"

#include "ProfilingTest_MultiUnitHeapManager.h"
#include "SampleLib/SampleLib.h"

#include <pia/common/common_Time.h>

#if USE_MULTIUNITHEAP
template <class T>
class MultiUnitHeapManager;
typedef MultiUnitHeapManager<nn::nex::ThreadUnsafeMultiUnitHeap> TAllocator;
#else
class HeapManager;
typedef HeapManager TAllocator;
#endif

namespace
{

}  // End of anonymous namespace


///< Instantiates the object.
MemoryMonitor::MemoryMonitor(void)
  : m_memRec(m_workBuf, WORKBUF_SIZE),
    m_currMemUsage(0),
    m_prevMemUsage(0),
    m_AllocationMonitoring(false)
{
}


///< Destroys the object.
MemoryMonitor::~MemoryMonitor(void)
{
}

void MemoryMonitor::StartAllocationMonitor(void)
{
    m_AllocationMonitoring = true;
#if USE_MEMORY_MONITOR
    // We set 1 here because we are eager to trace all memory allocation (Is this too much?).
    nn::nex::MemoryManager::SetAllocationMonitoringLowerLimit(1);
#endif
}

void MemoryMonitor::StopAllocationMonitor(void)
{
    m_AllocationMonitoring = false;
#if USE_MEMORY_MONITOR
    // The feature is actually turned off
    nn::nex::MemoryManager::SetAllocationMonitoringLowerLimit(0xffffffff);
#endif
}


void MemoryMonitor::Update(void)
{
#if USE_MULTIUNITHEAP
    m_prevMemUsage = m_currMemUsage;
    m_currMemUsage = TAllocator::GetUsage();
    unsigned int t = static_cast<unsigned int>(static_cast<s32>(nn::pia::common::Time::GetTime() - m_standardTime));
    m_memRec.Record(m_currMemUsage, t);

// Automatic mode. As soon as the memory usage exceeds 1MB, switches to the hunting mode to determine how the memory is allocated.
#if 0
    const size_t BORDER_LINE = 1024 * 1024;  // 1MB
    if ((m_prevMemUsage <= BORDER_LINE) && (m_currMemUsage > BORDER_LINE))
    {
        PIA_LOG(nn::pia::common::Log::LEVEL_ALWAYS, ("Memory BORDER_LINE is exceeded!  MemoryManager::SetAllocationMonitoringLowerLimit() will be called."));
#if USE_MEMORY_MONITOR
        nn::nex::MemoryManager::SetAllocationMonitoringLowerLimit(1);  // Set 1 because very eager to trace all memory allocation from here (too much?).
#endif
    }
#endif

#else
// Do nothing.
#endif
}

void MemoryMonitor::PrintAllocationResult(void) const
{
#if USE_MEMORY_MONITOR
    if (!m_AllocationMonitoring)
    {
        nn::nex::MemoryManager::PrintAllocationMonitoringResult();
    }
#endif
}

void MemoryMonitor::PrintMemoryUsage(void) const
{
#if USE_MULTIUNITHEAP
    if (m_prevMemUsage > m_currMemUsage)
    {
        PIA_LOG(nn::pia::common::Log::LEVEL_ALWAYS, ("MultiUnitHeap memory usage: %u bytes (-%u bytes)"), m_currMemUsage, m_prevMemUsage - m_currMemUsage);
    }
    else if (m_prevMemUsage < m_currMemUsage)
    {
        PIA_LOG(nn::pia::common::Log::LEVEL_ALWAYS, ("MultiUnitHeap memory usage: %u bytes (+%u bytes)"), m_currMemUsage, m_currMemUsage - m_prevMemUsage);
    }
    else
    {
        // Do nothing.
    }
#else
//    PIA_SAMPLE_TRACE_STREAM(SAMPLE_TRACE_ALWAYS, ("Caution.  MultiUnitHeap is not ready."));
#endif
}


void MemoryMonitor::Render(void) const
{
    const s32 X = 160;
    const s32 Y = 128;
    const s32 WIDTH = 64;
    const s32 HEIGHT = 64;

    // Render the base
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(X, Y, WIDTH + 1, HEIGHT, nn::pia::sampleframework::COLOR_BLACK);

    // Renders the memory consumption amount with a vertical bar graph.
    const int LIMIT = (SPAN < m_memRec.GetCurrentEntries()) ? SPAN : m_memRec.GetCurrentEntries();
    for (int i = 0; i < LIMIT; ++i)
    {
        int amount = 0;
        bool ret = m_memRec.GetAmount(&amount, i);
        if (!ret)
        {
            PIA_SAMPLE_TRACE_STREAM(
                SAMPLE_TRACE_ALWAYS,
                ("ERROR: MemoryMonitor::Render: m_memRec.GetAmount failed."));
        }
        SYSTEMCHECK(ret);

        const size_t H = amount / 1024 / 32;
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(
            X + 2 * i,
            Y + HEIGHT - H,
            1,
            H,
            nn::pia::sampleframework::COLOR_AQUA);
    }

    // Also display in numbers
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(X, Y + HEIGHT, "%u KB", m_currMemUsage / 1024);
}
