﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     ProfilingTest_P2P.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "ProfilingTest_P2P.h"
#include "Shared_Nex_Matchmake.h"
#include <pia/common/common_ByteOrder.h>
#include <pia/common/common_CallContext.h>
#include <pia/transport.h>
#include <pia/inet/inet_definitions.h>

#if ENABLE_NEX_NETZ
#else
#include <pia/inet/inet_NexFacade.h>
#endif

namespace
{
u32 s_SendCount = 0;
u32 s_ReceiveCount = 0;
u32 s_JoinTime = 0;

u32 s_SendCountReliable = 0;
u32 s_SendFailedCountReliable = 0;
u32 s_ReceiveCountReliable = 0;

#if ENABLE_NEX_NETZ
const u8 TYPE_UNRELIABLE = 0x00;
const u8 TYPE_RELIABLE = 0x01;

u16 GetConnectedStationNum()
{
    u16 connectedStationNum = 0;
    for (nn::nex::Station::SelectionIterator station; !station.EndReached(); ++station)
    {
        if (!station->IsLocal())
        {
            if (station->IsConnected())
            {
                connectedStationNum++;
            }
        }
    }
    return connectedStationNum;
}
#if USE_DIRECTSTREAM
std::map<nn::nex::ConnectionID, nn::nex::CallContext*> s_pSendReliableCallContext;
#endif
std::map<nn::nex::ConnectionID, u32> s_DropCount;
std::map<nn::nex::ConnectionID, u32> s_LatestReceiveSendCount;
#else
std::map<u64, u32> s_DropCount;
std::map<u64, u32> s_LatestReceiveSendCount;
#endif
};

//-----------------------------------------------------------------------------
// Create a P2P Session.
//-----------------------------------------------------------------------------
nn::Result CreateP2PSession()
{
    nn::Result result;

#if ENABLE_NEX_NETZ
#if NN_PIA_NEX_3
#if ENABLE_LOCAL
// A shared session key is required for Internet communication, but it does not need to be set for local communication.
#else
    // Enable the shared session key obtained from the server during matchmaking to be used as the signature key.
    nn::nex::qVector<nn::nex::qByte>* pSessionSignatureKey = Nex_GetMatchmakeSessionSignatureKey();
    PIA_SAMPLE_ASSERT(pSessionSignatureKey);
    nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(*pSessionSignatureKey);
#endif
#endif
    if (!nn::nex::Session::CreateSession(NULL))
    {
        PIA_SAMPLE_LOG("CreateP2PSession: CreateSession failed.\n");
        return SampleNexInvalidState();
    }
    PIA_SAMPLE_LOG("CreateP2PSession: succeeded\n");
    return nn::ResultSuccess();
#else
    return nn::pia::ResultNotImplemented();
#endif
}

#if ENABLE_NEX_NETZ

// n1769 Be sure to terminate NetZ instance.
void TerminateNetZ(nn::nex::NetZ* pNetZ)
{
    nn::nex::CallContext callContext;
    pNetZ->Terminate(&callContext);

#if 0
    callContext.Wait();
#else
    while (1)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

#if NN_PIA_CAFE
#if USE_CLEAR_DATACACHE
        ClearDataCache();
#endif
#endif

        NexDispatch();
        PiaDispatch();

        if (callContext.GetState() != nn::nex::CallContext::CallInProgress)
        {
            break;
        }

        DrawScreen();
#if PIA_TEST_CODE
        UpdatePiaTestData();
#endif
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }
#endif
}

nn::Result DestroyP2PSession(nn::nex::NetZ* pNetZ)
{
    nn::Result result;
    nn::pia::common::Time startTime = nn::pia::common::Time::GetTime();

    // n1769
    TerminateNetZ(pNetZ);

    PIA_SAMPLE_PRINT("DestroyP2PSession: succeeded: destroy session time:%d [ms]\n", static_cast<s32>(nn::pia::common::Time::GetTime() - startTime));
    return nn::ResultSuccess();
}

#else

nn::Result DestroyP2PSession()
{
#if 0
    nn::Result result;
    nn::pia::common::Time startTime = nn::pia::common::Time::GetTime();

    PIA_SAMPLE_PRINT("DestroyP2PSession: succeeded: destroy session time:%d [ms]\n", (s32)(nn::pia::common::Time::GetTime() - startTime));
    return nn::ResultSuccess();
#else
    return nn::pia::ResultNotImplemented();
#endif
}

#endif

//-----------------------------------------------------------------------------
// Join a P2P Session
//-----------------------------------------------------------------------------
#if ENABLE_NEX_NETZ

nn::Result JoinP2PSession(const nn::nex::qList<nn::nex::StationURL>& lstURL)
{
    nn::Result result;
    nn::pia::common::Time startTime = nn::pia::common::Time::GetTime();
#if NN_PIA_NEX_3
#if ENABLE_LOCAL
// A shared session key is required for Internet communication, but it does not need to be set for local communication.
#else
    // Enable the shared session key obtained from the server during matchmaking to be used as the signature key.
    nn::nex::qVector<nn::nex::qByte>* pSessionSignatureKey = Nex_GetMatchmakeSessionSignatureKey();
    PIA_SAMPLE_ASSERT(pSessionSignatureKey);
    nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(*pSessionSignatureKey);
#endif
#endif
    nn::nex::CallContext callContext;
    if (!nn::nex::Session::JoinSession(&callContext, lstURL))
    {
        PIA_SAMPLE_LOG("JoinP2PSession: JoinMesh failed.\n");
        return SampleNexInvalidState();
    }

    result = WaitCallContext(callContext);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("JoinP2PSession: WaitCallContext failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    if (callContext.GetState() != nn::nex::CallContext::CallSuccess)
    {
        PIA_SAMPLE_LOG("JoinP2PSession: JoinMesh failed:%d\n", callContext.GetState());
        return SampleNexInvalidState();
    }

    // Waits until everyone can connect
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        NexDispatch();
        PiaDispatch();

        u16 stationNum = GetStationNum();
        u16 connectedStationNum = GetConnectedStationNum();
        if (stationNum == connectedStationNum + 1)
        {
            PIA_SAMPLE_LOG("JoinP2PSession: connectedStationNum:%d\n", connectedStationNum);
            break;
        }

        DrawScreen();
#if PIA_TEST_CODE
        UpdatePiaTestData();
#endif
        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }

    s_JoinTime = static_cast<s32>(nn::pia::common::Time::GetTime() - startTime);
    PIA_SAMPLE_PRINT("JoinP2PSession: succeeded: join time:%d [ms]\n", s_JoinTime);

    return nn::ResultSuccess();
}

#else

nn::Result JoinP2PSession(const nn::nex::qList<nn::nex::StationURL>& lstURL)
{
    nn::Result result;
    nn::pia::transport::StationConnectionInfo connectionInfo;
    result = nn::pia::inet::NexFacade::ConvertNexStationUrlToStationConnectionInfo(lstURL, &connectionInfo);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("JoinP2PSession: ConvertNexStationUrlToStationConnectionInfo failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    return JoinP2PSession(connectionInfo);
}

nn::Result JoinP2PSession(const nn::pia::transport::StationConnectionInfo& connectionInfo)
{
#if 0
    nn::Result result;
    nn::pia::common::Time startTime = nn::pia::common::Time::GetTime();

    s_JoinTime = (s32)(nn::pia::common::Time::GetTime() - startTime);
    PIA_SAMPLE_PRINT("JoinP2PSession: succeeded: join time:%d [ms]\n", s_JoinTime);

    return nn::ResultSuccess();
#else
    NN_PIA_DUMMY_PARAM(connectionInfo);
    return nn::pia::ResultNotImplemented();
#endif
}

#endif

#if ENABLE_NEX_NETZ
#else
nn::Result LeaveP2PSession()
{
#if 0
    nn::Result result;
    nn::pia::common::Time startTime = nn::pia::common::Time::GetTime();

    PIA_SAMPLE_PRINT("LeaveP2PSession: succeeded: leave time:%d [ms]\n", (s32)(nn::pia::common::Time::GetTime() - startTime));
    return nn::ResultSuccess();
#else
    return nn::pia::ResultNotImplemented();
#endif
}
#endif

//-----------------------------------------------------------------------------
// Send.
//-----------------------------------------------------------------------------
#if ENABLE_NEX_NETZ
nn::Result SendP2PDataNetZ()
{
    u8 testSendBuffer[SEND_DATA_SIZE];

    PIA_SAMPLE_TRACE(SAMPLE_TRACE_SEND, "SendP2PDataNetZ:%d", s_SendCount);

    // Store our own connection ID as a test
    nn::nex::ConnectionID myConnectionID = nn::nex::Station::GetLocalInstance()->GetConnectionID();
    myConnectionID = nn::pia::common::ByteOrder::HostToNetwork32(myConnectionID);
    std::memcpy(&testSendBuffer[0], &myConnectionID, sizeof(nn::nex::ConnectionID));
    //n1769 Enter data to distinguish between Reliable/Unreliable.
    testSendBuffer[sizeof(nn::nex::ConnectionID)] = TYPE_UNRELIABLE;
    std::memcpy(&testSendBuffer[sizeof(nn::nex::ConnectionID) + sizeof(u8)], &s_SendCount, sizeof(u32));

#if USE_DIRECTSTREAM
    bool bResult = nn::nex::DirectStream::GetInstance()->SendAll(nn::nex::DirectStream::Unreliable, testSendBuffer, SEND_DATA_SIZE);
#else
    bool bResult = nn::nex::DirectTransport::GetInstance()->SendUnreliableAll(testSendBuffer, SEND_DATA_SIZE);
#endif
    if (!bResult)
    {
        PIA_SAMPLE_LOG("SendP2PDataNetZ: SendUnreliableAll failed.\n");
        return SampleNexInvalidState();
    }

    s_SendCount++;
    return nn::ResultSuccess();
}


nn::Result SendP2PDataReliableNetZ()
{
    u8 testSendBuffer[SEND_DATA_SIZE_RELIABLE];

    PIA_SAMPLE_TRACE(SAMPLE_TRACE_SEND, "SendP2PDataReliableNetZ:%d", s_SendCountReliable);

    // Store our own connection ID as a test
    nn::nex::ConnectionID myConnectionID = nn::nex::Station::GetLocalInstance()->GetConnectionID();
    myConnectionID = nn::pia::common::ByteOrder::HostToNetwork32(myConnectionID);
    std::memcpy(&testSendBuffer[0], &myConnectionID, sizeof(nn::nex::ConnectionID));
    //n1769 Enter data to distinguish between Reliable/Unreliable.
    testSendBuffer[sizeof(nn::nex::ConnectionID)] = TYPE_RELIABLE;

#if USE_DIRECTSTREAM
    nn::nex::Station::SelectionIterator itStations;
    bool bResult = true;

    for (itStations.GotoStart(); !itStations.EndReached(); ++itStations)
    {
        if (!itStations->IsLocal() && itStations->IsConnected())
        {
            // Send to stations other than the local host
            nn::nex::ConnectionID dstID = itStations->GetConnectionID();

#if USE_CONTEXT
            if (s_pSendReliableCallContext.find(dstID) == s_pSendReliableCallContext.end())
            {
                // If there is no context for the destination station, create it.
                s_pSendReliableCallContext[dstID] = qNew nn::nex::CallContext();
            }
            nn::nex::CallContext* pContext = s_pSendReliableCallContext[dstID];
            nn::nex::CallContext::State state = pContext->GetState();

            if (state == nn::nex::CallContext::CallInProgress)
            {
                //PIA_SAMPLE_LOG("Send Reliable dstID:%d in progress\n", dstID);
                continue;
            }
            else
            {
                if (state == nn::nex::CallContext::CallFailure)
                {
                    PIA_SAMPLE_LOG("Send Reliable error. dstID:%d Result:%08x\n", dstID, pContext->GetOutcome().GetReturnCode());
                    bResult = false;
                    s_SendFailedCountReliable++;
                }
                else if (state == nn::nex::CallContext::CallCancelled)
                {
                    PIA_SAMPLE_LOG("Send Reliable cancelled. dstID:%d\n", dstID);
                }
                else if (state == nn::nex::CallContext::CallSuccess)
                {
                    if (pContext->GetOutcome())
                    {
                        s_SendCountReliable++;
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("Send Reliable error. dstID:%d Result:%08x\n", dstID, pContext->GetOutcome().GetReturnCode());
                        bResult = false;
                        s_SendFailedCountReliable++;
                    }
                }
                pContext->Reset();

                // If not asynchronously processing, send next packet
                qBool result = nn::nex::DirectStream::GetInstance()->Send(pContext, dstID, nn::nex::DirectStream::Reliable, testSendBuffer, SEND_DATA_SIZE_RELIABLE);
                if (!result)
                {
                    PIA_SAMPLE_LOG("Send Reliable error. dstID:%d\n", dstID);
                }
            }
#else
            qBool result = nn::nex::DirectStream::GetInstance()->Send(dstID, nn::nex::DirectStream::Reliable, testSendBuffer, SEND_DATA_SIZE_RELIABLE);
            if (!result)
            {
                PIA_SAMPLE_LOG("Send Reliable error. dstID:%d\n", dstID);
                s_SendFailedCountReliable++;
            }
            else
            {
                s_SendCountReliable++;
            }
#endif
        }
    }
    if (!bResult)
    {
        return SampleNexInvalidState();
    }
    return nn::ResultSuccess();
#else
    bool bResult = nn::nex::DirectTransport::GetInstance()->SendReliableAll(testSendBuffer, SEND_DATA_SIZE_RELIABLE);

    if (!bResult)
    {
        s_SendFailedCountReliable++;
        return SampleNexInvalidState();
    }
    s_SendCountReliable++;
    return nn::ResultSuccess();
#endif
}

#if USE_DIRECTSTREAM
void CancelSendP2PDataReliableNetZ()
{
    // Uncompleted contexts are canceled
    std::map<nn::nex::ConnectionID, nn::nex::CallContext*>::iterator it;
    for (it = s_pSendReliableCallContext.begin(); it != s_pSendReliableCallContext.end(); ++it)
    {
        nn::nex::CallContext* pContext = it->second;
        if (pContext->GetState() == nn::nex::CallContext::CallInProgress)
        {
            pContext->Cancel();
            PIA_SAMPLE_LOG("Cancel Context. dstID:%d\n", it->first);
        }
        qDelete pContext;
    }
}
#endif

#else

nn::Result SendP2PData(u32 unreliableProtocolHandle)
{
    u8 testSendBuffer[SEND_DATA_SIZE];
    nn::Result result;

    if (GetStationNum() <= 1)
    {
        return nn::ResultSuccess();
    }

    PIA_SAMPLE_TRACE(SAMPLE_TRACE_SEND, "SendP2PData:SendCount:%d", s_SendCount);

    nn::pia::transport::UnreliableProtocol* pUnreliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(unreliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pUnreliableProtocol != NULL);

    nn::pia::transport::Station* pLocalStation = nn::pia::transport::StationManager::GetInstance()->GetLocalStation();
    PIA_SAMPLE_ASSERT(pLocalStation != NULL);

    nn::pia::StationId localStationId = pLocalStation->GetStationId();
    // Writes serialized data to the send buffer.

    // Sender ID.
    *reinterpret_cast<u64*>(&testSendBuffer[0]) = nn::pia::common::ByteOrder::HostToNetwork64(localStationId.GetValue());
    // Number of sends.
    *reinterpret_cast<u32*>(&testSendBuffer[8]) = nn::pia::common::ByteOrder::HostToNetwork32(s_SendCount);

    result = pUnreliableProtocol->SendToAll(testSendBuffer, SEND_DATA_SIZE);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("SendP2PData: pUnreliableProtocol->SendToAll failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    s_SendCount++;
    return nn::ResultSuccess();
}

nn::Result SendP2PDataR2(u32 r2UnreliableProtocolHandle, u32 stationNum)
{
    u8 testSendBuffer[SEND_DATA_SIZE];
    nn::Result result;

    if (GetStationNum() <= 1)
    {
        return nn::ResultSuccess();
    }

    PIA_SAMPLE_TRACE(SAMPLE_TRACE_SEND, "SendP2PData:SendCount:%d", s_SendCount);

    nn::pia::transport::RoundRobinUnreliableProtocol* pUnreliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(r2UnreliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pUnreliableProtocol != NULL);

    nn::pia::transport::Station* pLocalStation = nn::pia::transport::StationManager::GetInstance()->GetLocalStation();
    PIA_SAMPLE_ASSERT(pLocalStation != NULL);

    nn::pia::StationId localStationId = pLocalStation->GetStationId();
    // Writes serialized data to the send buffer.

    // Sender ID.
    *reinterpret_cast<u64*>(&testSendBuffer[0]) = nn::pia::common::ByteOrder::HostToNetwork64(localStationId.GetValue());
    // Number of sends.
    *reinterpret_cast<u32*>(&testSendBuffer[8]) = nn::pia::common::ByteOrder::HostToNetwork32(s_SendCount);

    result = pUnreliableProtocol->SendToRoundRobin(testSendBuffer, SEND_DATA_SIZE, stationNum);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("SendP2PData: pUnreliableProtocol->SendToRoundRobin failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        return result;
    }

    s_SendCount++;
    return nn::ResultSuccess();
}


//n1769
nn::Result SendP2PDataReliable(u32 reliableProtocolHandle)
{
    u8 testSendBuffer[SEND_DATA_SIZE_RELIABLE];
    nn::Result result;

    if (GetStationNum() <= 1)
    {
        return nn::ResultSuccess();
    }

    PIA_SAMPLE_TRACE(SAMPLE_TRACE_SEND, "SendP2PDataReliable:SendCount:%d", s_SendCountReliable);

    nn::pia::transport::ReliableProtocol* pReliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(reliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pReliableProtocol != NULL);

    nn::pia::transport::Station* pLocalStation = nn::pia::transport::StationManager::GetInstance()->GetLocalStation();
    PIA_SAMPLE_ASSERT(pLocalStation != NULL);

    nn::pia::StationId localStationId = pLocalStation->GetStationId();

    // Sender ID.
    *reinterpret_cast<u64*>(&testSendBuffer[0]) = nn::pia::common::ByteOrder::HostToNetwork64(localStationId.GetValue());

    result = pReliableProtocol->Send(nn::pia::STATION_ID_ALL, testSendBuffer, SEND_DATA_SIZE_RELIABLE);
    if (result.IsSuccess())
    {
        s_SendCountReliable++;
    }
    else
    {
        s_SendFailedCountReliable++;
        if (result == nn::pia::ResultBufferIsFull())
        {
            // The send buffer is full. Anticipated in the <tt>Send</tt> with the next frame.
        }
        else
        {
            PIA_SAMPLE_LOG("SendP2PDataReliable: pReliableProtocol->Send() failed.\n");
            PIA_SAMPLE_LOG_RESULT(result);
        }
    }

    return nn::ResultSuccess();
}


nn::Result SendP2PDataVoice()
{
    PIA_ASSERTMSG(false, "Not implemented.");
    return nn::ResultSuccess();
}
#endif

//-----------------------------------------------------------------------------
// Receive.
//-----------------------------------------------------------------------------
#if ENABLE_NEX_NETZ

void ParseReceiveP2PData(nn::nex::StationID sourceId, u8* buffer, u32 size)
{
    NN_PIA_DUMMY_PARAM(sourceId);

    nn::nex::StationID value = *(reinterpret_cast<nn::nex::StationID*>(buffer));
    value = nn::pia::common::ByteOrder::HostToNetwork32(value);

    switch (buffer[sizeof(nn::nex::ConnectionID)])
    {
    case TYPE_UNRELIABLE:
        if (size != SEND_DATA_SIZE)
        {
            PIA_SAMPLE_LOG("ParseReceiveP2PData: invalid size:%d\n", size);
        }
        else
        {
            u32 sendCount;
            std::memcpy(&sendCount, &buffer[sizeof(nn::nex::ConnectionID) + sizeof(u8)], sizeof(u32));
            PIA_SAMPLE_TRACE(
                SAMPLE_TRACE_RECEIVE,
                "ParseReceiveP2PData:connectionID:%x,size:%d,srcId:%d,seq:%d,ReceiveCount:%d",
                sourceId,
                size,
                value,
                sendCount,
                s_ReceiveCount);
            s_ReceiveCount++;

            if (sendCount - s_LatestReceiveSendCount[value] == 0)
            {
                if (s_LatestReceiveSendCount[value] != 0)
                {
                    PIA_SAMPLE_LOG("ParseReceiveP2PData: receive same data [%08x][%08x]\n", value, sendCount);
                }
            }
            else if (sendCount - s_LatestReceiveSendCount[value] > 1)
            {
                //PIA_SAMPLE_LOG("ParseReceiveP2PData: drop data? [%08x][%08x:%08x]\n", value, sendCount, s_LatestReceiveSendCount[value]);
                s_DropCount[value] += sendCount - s_LatestReceiveSendCount[value] - 1;
            }
            s_LatestReceiveSendCount[value] = sendCount;
        }
        break;
    case TYPE_RELIABLE:
        if (size != SEND_DATA_SIZE_RELIABLE)
        {
            PIA_SAMPLE_LOG("ParseReceiveP2PData: invalid size:%d\n", size);
        }
        else
        {
            PIA_SAMPLE_TRACE(
                SAMPLE_TRACE_RECEIVE,
                "P2PReceiveCallback:connectionID:%x,size:%d,srcId:%d,ReceiveCountReliable:%d",
                sourceId,
                size,
                value,
                s_ReceiveCountReliable);
            s_ReceiveCountReliable++;
        }
        break;
    default:
        PIA_SAMPLE_ASSERTMSG(false, "Unknown data type.");
        break;
    }
}

#if USE_DIRECTSTREAM
void ReceiveP2PData()
{
    // Vector for receiving DirectStream.
    nn::nex::qVector<nn::nex::DirectStream::Data> recvData;

    // Clear <tt>Vector</tt>.
    recvData.clear();

    // Receive.
    nn::nex::DirectStream::GetInstance()->GetReceivedData(recvData);

    nn::nex::qVector<nn::nex::DirectStream::Data>::iterator it = recvData.begin();
    while (it != recvData.end())
    {
        ParseReceiveP2PData((*it).id, (*it).buffer.Get(), (*it).buffer.GetSize());
        ++it;
    }
}

#else
//-----------------------------------------------------------------------------
// Receive callback
//-----------------------------------------------------------------------------
void P2PReceiveCallback(nn::nex::StationID sourceId, u8* buffer, u32 size)
{
    ParseReceiveP2PData(sourceId, buffer, size);
}
#endif  // USE_DIRECTSTREAM

#else

nn::Result ReceiveP2PDataImpl(nn::pia::transport::UnreliableProtocol* pUnreliableProtocol)
{

    const size_t BUFFER_SIZE = 1500;
    nn::pia::StationId sourceId;
    char buffer[BUFFER_SIZE];

    while (true)
    {
        size_t receivedSize;
        nn::Result result = pUnreliableProtocol->Receive(&sourceId, reinterpret_cast<bit8*>(buffer), &receivedSize, sizeof(buffer));
        if (result.IsFailure())
        {
            if (result == nn::pia::ResultNoData())
            {
                // Do nothing.
                break;
            }
            else
            {
                PIA_SAMPLE_LOG("ReceiveP2PDataImpl: pUnreliableProtocol->Receive failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                return result;
            }
        }

        s_ReceiveCount++;

        // Receive data operation
        // Restore the StationId from the receive data
        nn::pia::StationId sourceIdData = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(&buffer[0]));
        u64 sourceIdDataKey = sourceIdData.GetValue();
        PIA_SAMPLE_TRACE(SAMPLE_TRACE_RECEIVE, "ReceiveP2PDataImpl:receivedSize:%d,sourceId:0x%llx,sourceIdData:0x%llx\n", receivedSize, sourceId.GetValue(), sourceIdDataKey);
        // Restores the `sendCount` from the receive data.
        u32 sendCount = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u32*>(&buffer[8]));
        if ((sendCount - s_LatestReceiveSendCount[sourceIdDataKey]) == 0)
        {
            if (s_LatestReceiveSendCount[sourceIdDataKey] != 0)
            {
                PIA_SAMPLE_LOG("ReceiveP2PDataImpl: receive same data Id:%llx Seq:%08x\n", sourceIdDataKey, sendCount);
            }
        }
        else if ((sendCount - s_LatestReceiveSendCount[sourceIdDataKey]) > 1)
        {
            //            PIA_SAMPLE_LOG("ReceiveP2PData: drop data? Id:%d CurrentSeq:%08x PrevSeq:%08x Drop:%d\n",
            //                sourceIdData+1, sendCount, s_LatestReceiveSendCount[sourceIdDataKey], sendCount - s_LatestReceiveSendCount[sourceIdDataKey] - 1);
            s_DropCount[sourceIdDataKey] += sendCount - s_LatestReceiveSendCount[sourceIdDataKey] - 1;
        }
        s_LatestReceiveSendCount[sourceIdDataKey] = sendCount;
    }

    return nn::ResultSuccess();
}

nn::Result ReceiveP2PDataR2(u32 r2UnreliableProtocolHandle)
{
    nn::pia::transport::RoundRobinUnreliableProtocol* pUnreliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(r2UnreliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pUnreliableProtocol != NULL);

    return ReceiveP2PDataImpl(static_cast<nn::pia::transport::UnreliableProtocol*>(pUnreliableProtocol));
}

nn::Result ReceiveP2PData(u32 unreliableProtocolHandle)
{
    nn::pia::transport::UnreliableProtocol* pUnreliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(unreliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pUnreliableProtocol != NULL);

    return ReceiveP2PDataImpl(pUnreliableProtocol);
}

nn::Result ReceiveP2PDataReliable(u32 reliableProtocolHandle)
{
    nn::pia::transport::ReliableProtocol* pReliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(reliableProtocolHandle);
    PIA_SAMPLE_ASSERT(pReliableProtocol != NULL);

    const size_t BUFFER_SIZE = 1500;
    nn::pia::StationId sourceId;
    char buffer[BUFFER_SIZE];

    while (true)
    {
        size_t receivedSize;
        nn::Result result = pReliableProtocol->Receive(&sourceId, buffer, &receivedSize, sizeof(buffer));
        if (result.IsFailure())
        {
            if (result == nn::pia::ResultNoData())
            {
                // Do nothing.
                break;
            }
            else
            {
                PIA_SAMPLE_LOG("ReceiveP2PDataReliable: pReliableProtocol->Receive failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                return result;
            }
        }

        s_ReceiveCountReliable++;

        // Receive data operation
        // Restore the StationId from the receive data
        nn::pia::StationId sourceIdData = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(&buffer[0]));
        PIA_SAMPLE_TRACE(SAMPLE_TRACE_RECEIVE, "ReceiveP2PDataReliable:receivedSize:%d,sourceId:0x%llx,sourceIdData:0x%llx\n", receivedSize, sourceId.GetValue(), sourceIdData.GetValue());
    }

    return nn::ResultSuccess();
}

#endif

#if ENABLE_NEX_NETZ
#else
#if ENABLE_LATENCY_PROFILING
nn::Result ProfileLatency(bool isReliable, u32 protocolHandle, LatencyStatistic* pStatistic, u32 statisticNum)
{
    struct Data
    {
        u64 m_TickTime;
        nn::pia::StationId m_StationId;
    };

    nn::pia::transport::UnreliableProtocol* pUnreliable = NULL;
    nn::pia::transport::ReliableProtocol* pReliable = NULL;
    if (isReliable)
    {
        pReliable = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(protocolHandle);
        PIA_SAMPLE_ASSERT(pReliable != NULL);
    }
    else
    {
        pUnreliable = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(protocolHandle);
        PIA_SAMPLE_ASSERT(pUnreliable != NULL);
    }

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    u64 currentTick = nn::pia::common::Time::GetTime().GetTick();
    nn::pia::StationId myStationId = pSession->GetLocalStationId();

    nn::Result r;

    while (true)
    {
        Data data;
        nn::pia::StationId srcStationId;
        size_t dataSize;

        if (isReliable)
        {
            r = pReliable->Receive(&srcStationId, &data, &dataSize, sizeof(data));
            if (r.IsSuccess())
            {
            }
            else if (r == nn::pia::ResultNoData())
            {
                break;
            }
            else
            {
                PIA_SAMPLE_LOG("ProfileLatency:pReliable->Receive failed.\n");
                PIA_SAMPLE_LOG_RESULT(r);
                return r;
            }
        }
        else
        {
            r = pUnreliable->Receive(&srcStationId, reinterpret_cast<bit8*>(&data), &dataSize, sizeof(data));
            if (r.IsSuccess())
            {
            }
            else if (r == nn::pia::ResultNoData())
            {
                break;
            }
            else
            {
                PIA_SAMPLE_LOG("ProfileLatency:pUnreliable->Receive failed.\n");
                PIA_SAMPLE_LOG_RESULT(r);
                return r;
            }
        }

        if (dataSize != sizeof(data))
        {
            // Ignores because the data was unexpected
            continue;
        }

        if (data.m_StationId == myStationId)
        {
            // Measures because it is data from the local host
            nn::pia::StationIndex srcStationIndex;
            nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&srcStationIndex, srcStationId);
            if (result.IsSuccess() && srcStationIndex < static_cast<nn::pia::StationIndex>(statisticNum))
            {
                nn::pia::common::TimeSpan duration;
                duration.SetTick(currentTick - data.m_TickTime);
                pStatistic[srcStationIndex].Record(duration);
            }
        }
        else if (data.m_StationId == srcStationId)
        {
            // Sends back immediately because it is data generated by the other host
            if (isReliable)
            {
                if (pReliable->IsInCommunication(srcStationId))
                {
                    r = pReliable->Send(srcStationId, &data, sizeof(data));
                    if (r.IsSuccess())
                    {
                    }
                    else if (r == nn::pia::ResultBufferIsFull())
                    {
                        // This is discarded because it cannot be sent back
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("ProfileLatency:pReliable->Send failed.\n");
                        PIA_SAMPLE_LOG_RESULT(r);
                        return r;
                    }
                }
            }
            else
            {
                r = pUnreliable->Send(srcStationId, reinterpret_cast<bit8*>(&data), sizeof(data));
                if (r.IsSuccess())
                {
                }
                else
                {
                    PIA_SAMPLE_LOG("ProfileLatency:pUnreliable->Send failed.\n");
                    PIA_SAMPLE_LOG_RESULT(r);
                    return r;
                }
            }
        }
        else
        {
            // Discarded because it is not the local host's nor the other host's
        }
    }

    // Send local host first data sometimes.
    {
        static u32 sendCount = 0;
        const u32 c_SendSpan = 60;
        PIA_SAMPLE_ASSERT(statisticNum <= c_SendSpan);

        nn::pia::StationIndex myStationIndex;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&myStationIndex, myStationId);
        if (result.IsSuccess())
        {
            nn::pia::StationIndex dstStationIndex = static_cast<nn::pia::StationIndex>(sendCount % c_SendSpan);
            if ((dstStationIndex < static_cast<nn::pia::StationIndex>(statisticNum)) &&
                (dstStationIndex != myStationIndex) &&
                (pSession->CheckStationIdIsValid(dstStationIndex)))
            {
                Data data;
                data.m_TickTime = currentTick;
                data.m_StationId = myStationId;

                if (isReliable)
                {
                    if (pReliable->IsInCommunication(dstStationIndex))
                    {
                        r = pReliable->Send(dstStationIndex, &data, sizeof(data));
                        if (r.IsSuccess())
                        {
                        }
                        else if (r == nn::pia::ResultBufferIsFull())
                        {
                            // Quit because it cannot be sent.
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("ProfileLatency:pReliable->Send failed.\n");
                            PIA_SAMPLE_LOG_RESULT(r);
                            return r;
                        }
                    }
                }
                else
                {
                    r = pUnreliable->Send(dstStationIndex, reinterpret_cast<bit8*>(&data), sizeof(data));
                    if (r.IsSuccess())
                    {
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("ProfileLatency:pUnreliable->Send failed.\n");
                        PIA_SAMPLE_LOG_RESULT(r);
                        return r;
                    }
                }
            }

            ++sendCount;
        }
    }

    return nn::ResultSuccess();
}
#endif
#endif

//-----------------------------------------------------------------------------
// Display connection information
//-----------------------------------------------------------------------------
void TraceP2PSession()
{
#if ENABLE_CPU_PROFILING
#else

#if ENABLE_NEX_NETZ
    PIA_SAMPLE_LOG("-----------------------------\n");
    nn::nex::Station::SelectionIterator itStations;
    itStations.GotoStart();
    while (!itStations.EndReached())
    {

        PIA_SAMPLE_LOG(
            "Station %8x: URL = %ls, RTT = %d\n",
            itStations->GetHandle().GetValue(),
            itStations->GetStationURL(0),
            itStations->GetRTTAverage());

        ++itStations;
    }
    PIA_SAMPLE_LOG("\n");
#endif

#endif
}

u16 GetStationNum()
{
#if ENABLE_NEX_NETZ
    return nn::nex::Station::GetNbLocalStations();
#else
    return nn::pia::session::Session::GetInstance()->GetStationNum();
#endif
}

bool IsSessionHost()
{
#if ENABLE_NEX_NETZ
    return nn::nex::Session::GetRole() == SAMPLE_SESSION_ROLE_HOST;
#else
    return nn::pia::session::Session::GetInstance()->IsHost();
#endif
}

u32 GetJoinTime()
{
    return s_JoinTime;
}

u32 GetP2PSendCount()
{
    return s_SendCount;
}

u32 GetP2PReceiveCount()
{
    return s_ReceiveCount;
}

#if ENABLE_NEX_NETZ
u32 GetP2PDropCount(nn::nex::ConnectionID id)
#else
u32 GetP2PDropCount(nn::pia::StationId id)
#endif
{
#if ENABLE_NEX_NETZ
    nn::nex::ConnectionID idKey = id;
#else
    u64 idKey = id.GetValue();
#endif
    if (s_DropCount.find(idKey) == s_DropCount.end())
    {
        return 0;
    }
    return s_DropCount[idKey];
}

u32 GetP2PDropCount()
{
    u32 sum = 0;

#if ENABLE_NEX_NETZ
    for (std::map<nn::nex::ConnectionID, u32>::iterator it = s_DropCount.begin(); it != s_DropCount.end(); ++it)
#else
    for (std::map<u64, u32>::iterator it = s_DropCount.begin(); it != s_DropCount.end(); ++it)
#endif
    {
        sum += it->second;
    }
    return sum;
}

u32 GetP2PSendCountReliable()
{
    return s_SendCountReliable;
}

u32 GetP2PSendFailedCountReliable()
{
    return s_SendFailedCountReliable;
}

u32 GetP2PReceiveCountReliable()
{
    return s_ReceiveCountReliable;
}

void ClearCounters()
{
    s_SendCount = 0;
    s_ReceiveCount = 0;

    s_DropCount.clear();
    s_LatestReceiveSendCount.clear();

    s_SendCountReliable = 0;
    s_SendFailedCountReliable = 0;
    s_ReceiveCountReliable = 0;
}
