﻿/*---------------------------------------------------------------------------*
Project:  Pia
File:     ProfilingTest_StatisticalRecord.cpp

Copyright Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.
*---------------------------------------------------------------------------*/


#include "ProfilingTest_Common.h"
#include "ProfilingTest_StatisticalRecord.h"

namespace
{

}  // end of anonymous namespace


///< Instantiates the object.
StatisticalRecord::StatisticalRecord(void* pMem, size_t sz)
  : m_pEntries(NULL),
    m_maxEntries(0),
    m_currEntries(0),
    m_currIndex(0)
{
    if (pMem == NULL)
    {
        PIA_SAMPLE_LOG("NULL pointer is detected\n");
        return;
    }

    if (sz == 0)
    {
        PIA_SAMPLE_LOG("Memory size is zero.\n");
        return;
    }

    m_maxEntries = sz / sizeof(struct Entry);
    m_pEntries = static_cast<struct Entry*>(pMem);
}


///< Destroys the object.
StatisticalRecord::~StatisticalRecord(void)
{
}


void StatisticalRecord::Reset(void)
{
    m_currEntries = 0;
    m_currIndex = 0;
}


void StatisticalRecord::Record(TAmount amount, TTime t)
{
    m_pEntries[m_currIndex].amount = amount;
    m_pEntries[m_currIndex].t = t;

    // Updates the number of recorded entries.
    if (m_currEntries < GetMaxEntries())
    {
        ++m_currEntries;
    }

    // Increments the index. It also checks whether the index loops back to the beginning.
    ++m_currIndex;
    if (m_currIndex >= GetMaxEntries())
    {
        m_currIndex = 0;
    }
}


bool StatisticalRecord::GetAmount(TAmount* pAmount, size_t n) const
{
    if (pAmount == NULL)
    {
        PIA_SAMPLE_LOG("pAmount is NULL.");
        return false;
    }
    else if (n >= GetCurrentEntries())
    {
        PIA_SAMPLE_LOG("n(=%u) is out of boundary.", n);
        return false;
    }
    else
    {
        size_t idx = (GetMaxEntries() + m_currIndex - 1 - n) % GetMaxEntries();
        *pAmount = m_pEntries[idx].amount;
        return true;
    }
}


bool StatisticalRecord::IsMonotoneIncreasing(unsigned int timeSpan) const
{
    if (timeSpan == 0)
    {
        PIA_SAMPLE_LOG("Can not determine increase or decrease because timeSpan = 0.");
        return false;
    }

    if (timeSpan >= GetCurrentEntries())
    {
        PIA_SAMPLE_LOG("timeSpan(=%u) is too large.", timeSpan);
        return false;
    }

    bool ret = false;
    // Support for never used warning.
    NN_PIA_DUMMY_PARAM(ret);
    for (int i = 0; i < timeSpan; ++i)
    {
        TAmount a1 = 0;
        ret = GetAmount(&a1, i);
        SYSTEMCHECK(ret);

        TAmount a0 = 0;
        ret = GetAmount(&a0, i + 1);
        SYSTEMCHECK(ret);

        // There can be no areas without a monotonous increase.
        if (a0 >= a1)
        {
            return false;
        }
    }

    // Fortunately, a determination of monotonous increase is made if you get this far.
    return true;
}


bool StatisticalRecord::Test(void)
{
    PIA_SAMPLE_LOG("StatisticalRecord::Test() begin.");

    const size_t MAX_ENTRYS = 128;   // Maximum number of entries expected.
    const int AMOUNT_OFFSET = 1000;  // The value to add to time t.
    const int BUFSIZE = 1024;
    static int buf[BUFSIZE / sizeof(int)];

    StatisticalRecord sr(buf, sizeof(buf));

    // Check the feature for getting the maximum number of entries.
    if (sr.GetMaxEntries() != MAX_ENTRYS)
    {
        PIA_SAMPLE_LOG("Failure.  Entries: %u???", sr.GetMaxEntries());
        return false;
    }

    // Confirm that no values can be extracted.
    {
        TAmount amount = 0;
        bool ret = sr.GetAmount(&amount, 0);
        if (ret)
        {
            PIA_SAMPLE_LOG("GetAmount() Success...  Unexpected behavior!");
            return false;
        }
    }

    // Check whether you can get the recorded values.
    sr.Record(AMOUNT_OFFSET + 0, 0);
    TAmount amount = 0;
    bool ret = sr.GetAmount(&amount, 0);
    if (!ret)
    {
        PIA_SAMPLE_LOG("Failure.  GetAmount() failed.");
        return false;
    }

    // Check whether the values obtained are correct.
    if (amount != AMOUNT_OFFSET)
    {
        PIA_SAMPLE_LOG("Failure.  GetAmount: %u???", amount);
        return false;
    }

    // Check the number of entries.
    if (sr.GetCurrentEntries() != 1)
    {
        PIA_SAMPLE_LOG("Failure.  GetCurrentEntries(): %u???", sr.GetCurrentEntries());
        return false;
    }

    // Write one cycle through the ring buffer.
    for (size_t i = 1; i <= MAX_ENTRYS; ++i)
    {
        sr.Record(i + AMOUNT_OFFSET, i);
    }

    // Check the number of elements.
    if (sr.GetCurrentEntries() != MAX_ENTRYS)
    {
        PIA_SAMPLE_LOG("Failure.  GetCurrentEntries(): %u", sr.GetCurrentEntries());
        return false;
    }

    // Check the values as you go back through them.
    for (size_t i = 0; i < MAX_ENTRYS; ++i)
    {
        ret = sr.GetAmount(&amount, i);
        if (!ret)
        {
            PIA_SAMPLE_LOG("Failure.  GetAmount() failed.");
            return false;
        }
        if (amount != static_cast<TAmount>(AMOUNT_OFFSET + MAX_ENTRYS - i))
        {
            PIA_SAMPLE_LOG("Failure. Unexpected value.  GetAmount(): %d", amount);
            return false;
        }
    }

    ret = sr.IsMonotoneIncreasing(3);
    if (!ret)
    {
        PIA_SAMPLE_LOG("Failure. Should be increased.");
        return false;
    }

    // This time add flat values.
    for (size_t i = 0; i < MAX_ENTRYS / 2; ++i)
    {
        sr.Record(10000, 1000 + i);
    }
    ret = sr.IsMonotoneIncreasing(MAX_ENTRYS / 4);
    if (ret)
    {
        PIA_SAMPLE_LOG("Failure. Should be flat.");
        return false;
    }

    // Confirm that values can no longer be obtained if they are reset.
    sr.Reset();
    ret = sr.GetAmount(&amount, 0);
    if (ret)
    {
        PIA_SAMPLE_LOG("GetAmount() Success...  Unexpected behavior!");
        return false;
    }

    PIA_SAMPLE_LOG("StatisticalRecord::Test() end.  <<<<<<Congratulations.  All green!>>>>>>");
    return true;
}
