﻿/*---------------------------------------------------------------------------*
Project:  Pia
File:     ProfilingTest_StatisticalRecord.h

Copyright Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.
*---------------------------------------------------------------------------*/


#ifndef STATISTICALRECORD_H_
#define STATISTICALRECORD_H_

#include "Shared_Nex_Setup.h"

typedef int TAmount;         // "Volume" type.
typedef unsigned int TTime;  // Type indicating "time."


/*!
@brief  This class gets recorded stats.

Specifically, this class serves to record the amount of free memory to this class for every frame to examine how memory usage increases.
Instances of this class consume a fixed amount of memory, and old data is regularly discarded.
 


*/
class StatisticalRecord
{
public:
    /*!
    @brief  Instantiates the object. Takes the memory to record and its size as arguments.

           The working memory provided here must be held until the destructor is called.

    @param[in] pMem  Starting address of the memory block. The buffer must be 4-byte aligned.
    @param[in] sz  Size of the memory block.
     */
    StatisticalRecord(void* pMem, size_t sz);


    /*!
@brief  Destructor.
     */
    ~StatisticalRecord(void);


    /*!
    @brief  Discards all of the recorded data and reverts to a blank state.
     */
    void Reset(void);


    /*!
    @brief  Returns the maximum number of entries that can be recorded in an instance.

           The number of entries that can be recorded varies depending on the amount of memory provided to the constructor.
           Naturally, the number of entries that can be recorded increases if more memory is provided.

    @return  Returns the maximum number of entries that can be recorded.
     */
    size_t GetMaxEntries(void) const
    {
        return m_maxEntries;
    }


    /*!
    @brief  Records the amount to time <tt><var>t</var></tt>.

    @param[in] amount  Provides the amount you want to record. Negative numbers can also be provided.
    @param[in] t  Provide the time at the point this function is called.
     */
    void Record(TAmount amount, TTime t);

    /*!
    @brief  Gets the nth recorded amount (going back from the most recent).

            Returns the most recent value if you specify zero as the argument.

    @param[out] pAmount  A pointer to the variable that stores the amount. NULL cannot be used.
    @param[in] n

    @return  Returns true if pAmount is a non-NULL pointer and n is within range.
     */
    bool GetAmount(TAmount* pAmount, size_t n) const;


    /*!
    @brief  Returns the number of entries currently recorded.

    @return  The number of entries recorded by the <tt>@ref Record</tt> function.
     */
    size_t GetCurrentEntries(void) const
    {
        return m_currEntries;
    }


    /*!
    @brief  Determines whether the recently recorded amounts are increasing monotonously.

           Goes back a number of records equal to timeSpan from the most recent record, and then checks the trend from that point to the most recent record.
           Returns true if the amount increases monotonously.

    @param[in] timeSpan  Provides the span of time to cover.

    @return  True if a monotonous increase, false otherwise.
     
*/
    bool IsMonotoneIncreasing(unsigned int timeSpan) const;


    /*!
    @brief  Executes a single unit test of the class.

    @return  True if the test passes. Returns false if the test fails.
     */
    static bool Test(void);

private:
    struct Entry
    {
        TAmount amount;  // Some kind of "amount."
        TTime t;         // Time at the time of recording.
    };

    ///< The default constructor, copy constructor, and the assignment operator are private.
    StatisticalRecord(void);
    StatisticalRecord(const StatisticalRecord&);
    StatisticalRecord& operator=(const StatisticalRecord&);

    struct Entry* m_pEntries;  // Array to record the "volume" and "time" as a pair.
    size_t m_maxEntries;       // Maximum number of entries. Depends on the amount of given memory.
    size_t m_currEntries;      // The number of entries currently recorded. Of course, this does not exceed m_maxEntries.
    size_t m_currIndex;        // When Record function is called next, data is written in this index location.
};

#endif  // end of STATISTICALRECORD_H_
