﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Session_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"
#include "Shared_Basic_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Session_Setup.h"

#if NN_PIA_CAFE
#include <cafe.h>
#include <nn/act.h>
#include <nn/fp.h>
#elif NN_PIA_CTR
#include <nn/friends.h>
#endif

#include <pia/inet.h>
#if NN_PIA_CTR
#include <pia/local.h>
#endif
#include <pia/transport.h>

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif


#if PIA_TEST_CODE
PiaTestCallback s_PiaTestCallback = NULL;

PiaTestCallback RegisterPiaTestCallback(PiaTestCallback cb)
{
    PiaTestCallback prev = s_PiaTestCallback;
    s_PiaTestCallback = cb;
    return prev;
}
#else
PiaTestCallback RegisterPiaTestCallback(PiaTestCallback){ return NULL; }
#endif

namespace
{
    bool g_EnableBitRateCheck = false;
}

nn::Result Session_InitializePiaSession(
    nn::pia::transport::NetworkFactory* pFactory,
    nn::pia::session::NetworkTopology topology)
{
    nn::pia::session::Session::Setting setting;
    setting.pNetworkFactory = pFactory;
    setting.networkTopology = topology;      // Configure the network topology.
    setting.browsedSessionInfoListNum = 8;  // The number of browsable sessions

    return Session_InitializePiaSession(setting);
}

nn::Result Session_InitializePiaSession(
    const nn::pia::session::Session::Setting& setting)
{
    return Session_InitializePiaSession(setting, false);
}

nn::Result Session_InitializePiaSession(
    const nn::pia::session::Session::Setting& setting,
    bool isEnableRelayEmulation,
    bool isEnableMatchmakeAnalysis,
    bool isDisableCheckMaxSilenceTime)
{
    NN_PIA_DUMMY_PARAM(isEnableRelayEmulation);
    NN_PIA_DUMMY_PARAM(isEnableMatchmakeAnalysis);
    NN_PIA_DUMMY_PARAM(isDisableCheckMaxSilenceTime);

    PIA_SAMPLE_LOG("Session_InitializePiaSession\n");
    nn::Result result;

    result = nn::pia::session::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

#if PIA_TEST_CODE
    {
        // For debug feature settings.
        nn::pia::session::Session::DebugSetting debugSetting;
        // Settings for connection route emulation for debugging.
        debugSetting.isEnableRelayEmulation = isEnableRelayEmulation;
        debugSetting.isEnableMatchmakeAnalysis = isEnableMatchmakeAnalysis;
        debugSetting.isDisableCheckMaxSilenceTime = isDisableCheckMaxSilenceTime;

        // Make sure that this does not get called in the retail version ROM.
        result = nn::pia::session::Session::SetDebugSetting(debugSetting);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }
#endif
    
    g_EnableBitRateCheck = (setting.bitRateCheckMode == nn::pia::session::BitRateCheck_Enable);

    result = nn::pia::session::Session::CreateInstance(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Session_InitializePiaSession:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());

    return result;
}

void Session_FinalizePiaSession()
{
    PIA_SAMPLE_LOG("Session_FinalizePiaSession\n");

    // With Pia, Finalize and DestroyInstance always succeed.
    if (nn::pia::session::Session::GetInstance() != NULL)
    {
        nn::pia::session::Session::GetInstance()->Cleanup();
        nn::pia::session::Session::DestroyInstance();
    }
    else
    {
        PIA_SAMPLE_LOG("Session_FinalizeSession: Session instance is not created yet.\n");
    }
    nn::pia::session::Finalize();
}


void Session_GetLocalPlayerName(nn::pia::transport::Station::PlayerName* pPlayerName)
{
    nn::Result result;

#if NN_PIA_CAFE
    //Uses the name of the Mii character.
    char16 miiName[ACT_MII_NAME_SIZE];
    std::memset(miiName, 0, sizeof(miiName));

    result = nn::act::GetMiiName(miiName);
    if (result.IsSuccess())
    {
        for (int i = 0; i < ACT_MII_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (miiName[i] != 0)
            {
                pPlayerName->data[i] = miiName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    nn::fp::MyPresence myPresence;
    result = nn::fp::GetMyPresence(&myPresence);
    if (result.IsSuccess())
    {
        pPlayerName->language = myPresence.consoleInfo.language;
    }
    else
    {
        pPlayerName->language = nn::fp::LANGUAGE_INVALID;
    }
#elif NN_PIA_CTR
    //Uses the name of the Mii character.
    char16 screenName[nn::friends::SCREEN_NAME_SIZE];
    result = nn::friends::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    result = nn::friends::GetMyScreenName(screenName);
    nn::friends::Finalize();
    if (result.IsSuccess() && screenName[0] != 0)  // Also exclude cases where a Mii character is not set.
    {
        for (int i = 0; i < nn::friends::SCREEN_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (screenName[i] != 0)
            {
                pPlayerName->data[i] = screenName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    pPlayerName->language = 0;
#else
    std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
    pPlayerName->length = 9;
    pPlayerName->language = 0;
#endif
}

nn::Result Session_StartupSession(nn::pia::session::Session::StartupSetting startupSetting)
{
    nn::pia::transport::Station::PlayerName playerName;
    Session_GetLocalPlayerName(&playerName);

    startupSetting.pPlayerName = &playerName;
    PIA_SAMPLE_LOG("Session_StartupSession\n");

    nn::Result result = nn::pia::session::Session::GetInstance()->Startup(startupSetting);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_StartupSession: Startup failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
    }
    return result;
}


void Session_CleanupSession()
{
    if (nn::pia::session::Session::GetInstance())
    {
        PIA_SAMPLE_LOG("Session_CleanupSession\n");
        nn::pia::session::Session::GetInstance()->Cleanup();
    }
    else
    {
        PIA_SAMPLE_LOG("Session_CleanupSession: Session instance is not created yet.\n");
    }
}

SampleSessionAsyncStatus Session_JoinRandomSession(
    nn::pia::session::CreateSessionSetting* pCreateSessionSetting,
    nn::pia::session::SessionSearchCriteria searchCriteriaArray[],
    size_t searchCriteriaNum,
    nn::Result* pAsyncResult,
    bool bRender)
{
    PIA_SAMPLE_LOG("Session_JoinRandomSession\n");

    PIA_SAMPLE_ASSERT(pAsyncResult != NULL);
    // Returns <tt>ResultSuccess</tt> if the function has escaped while the asynchronous process is still uncompleted.
    nn::Result result = nn::ResultSuccess();
    SampleSessionAsyncStatus asyncStatus = SAMPLE_SESSION_ASYNC_SUCCESS;

    // Starts random matchmaking.
    result = nn::pia::session::Session::GetInstance()->JoinRandomSessionAsync(pCreateSessionSetting, searchCriteriaArray, searchCriteriaNum);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_JoinRandomSession: call JoinRandomSessionAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
        *pAsyncResult = result;
        return SAMPLE_SESSION_ASYNC_FAILED;
    }

    // Waits for the asynchronous process <tt>JoinRandomSessionAsync</tt> to complete.
    int loopCounter = 0;
    bool isCanceled = false;
    while (true)
    {
        if (bRender)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
        }

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (bRender)
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "random matchmaking session... %d (%d)", loopCounter++, nn::pia::session::Session::GetInstance()->GetJoinAutoMatchmakeSequence() );
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 30, "  B: Cancel");

            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        }
        else
        {
#if NN_PIA_CTR
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(16));
#elif NN_PIA_CAFE
            OSSleepMilliseconds(16);
#elif NN_PIA_WIN
            Sleep(16);
#endif
        }
        
#if PIA_TEST_CODE
        if (PIA_IS_VALID_POINTER(s_PiaTestCallback))
        {
            s_PiaTestCallback();
        }
#endif

        if (nn::pia::session::Session::GetInstance()->IsCompletedJoinRandomSession())
        {
            result = nn::pia::session::Session::GetInstance()->GetJoinRandomSessionResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_JoinRandomSession: JoinRandomSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                *pAsyncResult = result;
                return isCanceled ? SAMPLE_SESSION_ASYNC_CANCELED : SAMPLE_SESSION_ASYNC_FAILED;
            }
            else
            {
                PIA_SAMPLE_LOG("Session_JoinRandomSession: JoinRandomSessionAsync succeeded.\n");
                s32 bitRate = 0;
                if (g_EnableBitRateCheck)
                {
                    if (nn::pia::session::Session::GetInstance()->IsCompleteBitRateCheck())
                    {
                        if (nn::pia::session::Session::GetInstance()->GetUplinkBitRate(&bitRate).IsSuccess())
                        {
                            PIA_SAMPLE_LOG("bitRate: %d\n", bitRate);
                        }
                        else
                        {
                            PIA_SAMPLE_LOG("bitRate measuring failed.\n");
                        }
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("bitRate measuring is not complete.\n");
                    }
                }
                *pAsyncResult = result;
                return asyncStatus;
            }
        }
        else
        {
            // Cancellation process
            if (!isCanceled && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                result = nn::pia::session::Session::GetInstance()->CancelJoinRandomSessionAsync();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: CancelJoinRandomSessionAsync failed.\n");
                    PIA_SAMPLE_ASSERT(false);
                }
                else
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: CancelJoinRandomSessionAsync success. Begin Canceling...\n");
                    isCanceled = true;
                }
            }
        }
    }
}

SampleSessionAsyncStatus Session_CreateSession(
    nn::pia::session::CreateSessionSetting* pCreateSessionSetting,
    nn::Result* pAsyncResult)
{
    PIA_SAMPLE_LOG("Session_CreateSession\n");

    PIA_SAMPLE_ASSERT(pAsyncResult != NULL);
    // Returns <tt>ResultSuccess</tt> if the function has escaped while the asynchronous process is still uncompleted.
    nn::Result result = nn::ResultSuccess();
    SampleSessionAsyncStatus asyncStatus = SAMPLE_SESSION_ASYNC_SUCCESS;

    // Creates a session.
    result = nn::pia::session::Session::GetInstance()->CreateSessionAsync(pCreateSessionSetting);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_CreateSession: call CreateSessionAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
        *pAsyncResult = result;
        return SAMPLE_SESSION_ASYNC_FAILED;
    }

    // Waits for the asynchronous process <tt>CreateSessionAsync</tt> to complete.
    int loopCounter = 0;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "creating session... %d", loopCounter++);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

#if PIA_TEST_CODE
        if (PIA_IS_VALID_POINTER(s_PiaTestCallback))
        {
            s_PiaTestCallback();
        }
#endif

        if (nn::pia::session::Session::GetInstance()->IsCompletedCreateSession())
        {
            result = nn::pia::session::Session::GetInstance()->GetCreateSessionResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_CreateSession: CreateSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                *pAsyncResult = result;
                return SAMPLE_SESSION_ASYNC_FAILED;
            }
            else
            {
                PIA_SAMPLE_LOG("Session_CreateSession: CreateSessionAsync succeeded.\n");
                *pAsyncResult = result;
                return asyncStatus;
            }
        }
    }
}

SampleSessionAsyncStatus Session_BrowseSession(
    nn::pia::session::SessionSearchCriteria* pSearchCriteria,
    nn::Result* pAsyncResult)
{
    PIA_SAMPLE_LOG("Session_BrowseSession\n");

    PIA_SAMPLE_ASSERT(pAsyncResult != NULL);
    // Returns <tt>ResultSuccess</tt> if the function has escaped while the asynchronous process is still uncompleted.
    nn::Result result = nn::ResultSuccess();
    SampleSessionAsyncStatus asyncStatus = SAMPLE_SESSION_ASYNC_SUCCESS;

    // Searches for a session
    result = nn::pia::session::Session::GetInstance()->BrowseSessionAsync(pSearchCriteria);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_BrowseSession: call BrowseSessionAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
        *pAsyncResult = result;
        return SAMPLE_SESSION_ASYNC_FAILED;
    }

    // Waits for the asynchronous process <tt>BrowseSessionAsync</tt> to complete.
    int loopCounter = 0;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "browsing session... %d", loopCounter++);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

#if PIA_TEST_CODE
        if (PIA_IS_VALID_POINTER(s_PiaTestCallback))
        {
            s_PiaTestCallback();
        }
#endif

        if (nn::pia::session::Session::GetInstance()->IsCompletedBrowseSession())
        {
            result = nn::pia::session::Session::GetInstance()->GetBrowseSessionResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_BrowseSession: BrowseSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                *pAsyncResult = result;
                return SAMPLE_SESSION_ASYNC_FAILED;
            }
            else
            {
                PIA_SAMPLE_LOG("Session_BrowseSession: BrowseSessionAsync succeeded.\n");
                *pAsyncResult = result;
                return asyncStatus;
            }
        }
    }
}

SampleSessionAsyncStatus Session_JoinSession(
    nn::pia::session::JoinSessionSetting* pJoinSessionSetting,
    nn::Result* pAsyncResult)
{
    PIA_SAMPLE_LOG("Session_JoinSession\n");

    PIA_SAMPLE_ASSERT(pAsyncResult != NULL);
    // Returns <tt>ResultSuccess</tt> if the function has escaped while the asynchronous process is still uncompleted.
    nn::Result result = nn::ResultSuccess();
    SampleSessionAsyncStatus asyncStatus = SAMPLE_SESSION_ASYNC_SUCCESS;

    // Joins a session.
    result = nn::pia::session::Session::GetInstance()->JoinSessionAsync(pJoinSessionSetting);
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_JoinSession: call JoinSessionAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        PIA_SAMPLE_ASSERT(result == nn::pia::ResultNetworkConnectionIsLost());
        *pAsyncResult = result;
        return SAMPLE_SESSION_ASYNC_FAILED;
    }

    // Waits for the asynchronous process <tt>JoinSessionAsync</tt> to complete.
    int loopCounter = 0;
    bool isCanceled = false;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "joining session... %d (%d)", loopCounter++, nn::pia::session::Session::GetInstance()->GetJoinSequence() );
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 30, "  B: Cancel");

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
        
#if PIA_TEST_CODE
        if (PIA_IS_VALID_POINTER(s_PiaTestCallback))
        {
            s_PiaTestCallback();
        }
#endif

        if (nn::pia::session::Session::GetInstance()->IsCompletedJoinSession())
        {
            result = nn::pia::session::Session::GetInstance()->GetJoinSessionResult();
            if (result.IsFailure())
            {
                PIA_SAMPLE_LOG("Session_JoinSession: JoinSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);
                *pAsyncResult = result;
                return isCanceled ? SAMPLE_SESSION_ASYNC_CANCELED : SAMPLE_SESSION_ASYNC_FAILED;
            }
            else
            {
                PIA_SAMPLE_LOG("Session_JoinSession: JoinSessionAsync succeeded.\n");

                if (g_EnableBitRateCheck)
                {
                    s32 bitRate = 0;
                    if (nn::pia::session::Session::GetInstance()->GetUplinkBitRate(&bitRate).IsSuccess())
                    {
                        PIA_SAMPLE_LOG("bitRate: %d\n", bitRate);
                    }
                }                
                *pAsyncResult = result;
                return asyncStatus;
            }
        }
        else
        {
            // Cancellation process
            if (!isCanceled && nn::pia::sampleframework::SampleFramework::GetInstance()->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                result = nn::pia::session::Session::GetInstance()->CancelJoinSessionAsync();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: CancelJoinSessionAsync failed.\n");
                    PIA_SAMPLE_ASSERT(false);
                }
                else
                {
                    PIA_SAMPLE_LOG("Session_JoinSession: CancelJoinSessionAsync success. Begin Canceling...\n");
                    isCanceled = true;
                }
            }
        }
    }
}

SampleSessionAsyncStatus Session_LeaveSession()
{
    PIA_SAMPLE_LOG("Session_LeaveSession\n");

    // Destroys a session.
    nn::Result result;
    result = nn::pia::session::Session::GetInstance()->LeaveSessionAsync();
    if (result.IsFailure())
    {
        PIA_SAMPLE_LOG("Session_LeaveSession: call LeaveSessionAsync failed.\n");
        PIA_SAMPLE_LOG_RESULT(result);
        if (result == nn::pia::ResultNetworkConnectionIsLost())
        {
            // The asynchronous process will fail if the network is disconnected.
            return SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK;
        }
        PIA_SAMPLE_ASSERT(false);
    }

    SampleSessionAsyncStatus asyncStatus = SAMPLE_SESSION_ASYNC_SUCCESS;

    // Waits for the asynchronous process <tt>LeaveSessionAsync</tt> to complete.
    int loopCounter = 0;
    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

        // Dispatching is also required when nn::nex::Scheduler is created.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }

        // Dispatch process.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, Basic_GetProgramTitle());
        nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 20, "leaving session... %d", loopCounter++);

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

#if PIA_TEST_CODE
        if (PIA_IS_VALID_POINTER(s_PiaTestCallback))
        {
            s_PiaTestCallback();
        }
#endif

        if (nn::pia::session::Session::GetInstance()->IsCompletedLeaveSession())
        {
            // The destruction of the session is complete.
            result = nn::pia::session::Session::GetInstance()->GetLeaveSessionResult();
            if (result.IsFailure())
            {
                // Also only output log when process fails.
                PIA_SAMPLE_LOG("Session_LeaveSession: LeaveSessionAsync failed.\n");
                PIA_SAMPLE_LOG_RESULT(result);

                // When the next error occurs, will need to reconnect to the network.
                if (result == nn::pia::ResultNetworkConnectionIsLost())
                {
                    return SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK;
                }
            }
            else
            {
                PIA_SAMPLE_LOG("Session_LeaveSession: LeaveSessionAsync succeeded.\n");
            }
            return asyncStatus;
        }
    }
}
