﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Transport_Setup.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"
#include "Shared_Transport_Setup.h"

namespace
{
const size_t MAX_STATIONS = 12;
}

// Prototype declaration for common processes.
nn::Result transport_InitializePiaTransportImpl(
    const nn::pia::transport::Transport::Setting& setting,
    u16 unreliableProtocolPort,
    u32* pUnreliableProtocolHandle,
    u16 reliableProtocolPort,
    u32* pReliableProtocolHandle,
    size_t reliableSendBufNum,
    size_t reliableRecvBufNum,
    InitializeAdditionalProtocolFunction pAdditionalFunction,
    size_t unreliableRecvBufNum);


nn::Result Transport_InitializePiaTransport(const nn::pia::transport::Transport::Setting& setting)
{
    return transport_InitializePiaTransportImpl(
        setting,
        0,
        NULL,
        0,
        NULL,
        0,
        0,
        NULL,
        0);
}


nn::Result Transport_InitializePiaTransport(
    const nn::pia::transport::Transport::Setting& setting,
    u16 unreliableProtocolPort,
    u32* pUnreliableProtocolHandle,  //n1769: When <tt>NULL</tt> is passed, the <tt>UnreliableProtocol</tt> initialization process is not finished.
    u16 reliableProtocolPort,
    u32* pReliableProtocolHandle,  //n1769: When <tt>NULL</tt> is passed, the <tt>ReliableProtocol</tt> initialization process is not finished.
    size_t reliableSendBufNum,
    size_t reliableRecvBufNum,
    InitializeAdditionalProtocolFunction pAdditionalFunction,
    size_t unreliableRecvBufNum)
{
    return transport_InitializePiaTransportImpl(
        setting,
        unreliableProtocolPort,
        pUnreliableProtocolHandle,
        reliableProtocolPort,
        pReliableProtocolHandle,
        reliableSendBufNum,
        reliableRecvBufNum,
        pAdditionalFunction,
        unreliableRecvBufNum);
}


nn::Result Transport_InitializePiaTransport(
    nn::pia::transport::NetworkFactory* pFactory,
    u16 unreliableProtocolPort,
    u32* pUnreliableProtocolHandle,
    u16 reliableProtocolPort,
    u32* pReliableProtocolHandle,
    size_t reliableSendBufNum,
    size_t reliableRecvBufNum,
    InitializeAdditionalProtocolFunction pAdditionalFunction,
    size_t unreliableRecvBufNum)
{

    nn::pia::transport::Transport::Setting setting;
    setting.pFactory = pFactory;
    setting.maxStationNum = MAX_STATIONS;
    setting.maxSendThreadBufferNum = 16;
    setting.maxReceiveThreadBufferNum = 24;

    return transport_InitializePiaTransportImpl(
        setting,
        unreliableProtocolPort,
        pUnreliableProtocolHandle,
        reliableProtocolPort,
        pReliableProtocolHandle,
        reliableSendBufNum,
        reliableRecvBufNum,
        pAdditionalFunction,
        unreliableRecvBufNum);
}


nn::Result transport_InitializePiaTransportImpl(
    const nn::pia::transport::Transport::Setting& setting,
    u16 unreliableProtocolPort,
    u32* pUnreliableProtocolHandle,
    u16 reliableProtocolPort,
    u32* pReliableProtocolHandle,
    size_t reliableSendBufNum,
    size_t reliableRecvBufNum,
    InitializeAdditionalProtocolFunction pAdditionalFunction,
    size_t unreliableRecvBufNum)
{
    PIA_SAMPLE_LOG("Transport_InitializePiaTransport\n");
    nn::Result result;

    result = nn::pia::transport::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::Transport::CreateInstance(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    if (pUnreliableProtocolHandle != NULL)
    {
        *pUnreliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::UnreliableProtocol>(unreliableProtocolPort);
        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(*pUnreliableProtocolHandle)->Initialize(unreliableRecvBufNum);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    //n1769: <tt>ReliableProtocol</tt> initialization processing.
    if (pReliableProtocolHandle != NULL)
    {
        const size_t SEND_BUF_NUM = reliableSendBufNum;
        const size_t RECV_BUF_NUM = reliableRecvBufNum;
        *pReliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::ReliableProtocol>(reliableProtocolPort);
        PIA_SAMPLE_ASSERT(*pReliableProtocolHandle != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(*pReliableProtocolHandle)->Initialize(SEND_BUF_NUM, RECV_BUF_NUM);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    //m2004: Allows use of any protocol.
    if (pAdditionalFunction != NULL)
    {
        result = pAdditionalFunction();
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }

    result = nn::pia::transport::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Print the amount of consumed memory.
    PIA_SAMPLE_PRINT("Transport_InitializePiaTransport:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());
    return result;
}

void Transport_FinalizePiaTransport()
{
    Transport_FinalizePiaTransport(
        0,
        0,
        NULL);
}

void Transport_FinalizePiaTransport(
    u32 unreliableProtocolHandle,
    u32 reliableProtocolHandle,
    FinalizeAdditionalProtocolFunction pAdditionalFunction)
{
    PIA_SAMPLE_LOG("Transport_FinalizePiaTransport\n");

    if (nn::pia::transport::Transport::GetInstance() != NULL)
    {
        //n1769: Finalize ReliableProtocol.
        if (reliableProtocolHandle != 0)
        {
            nn::pia::transport::ReliableProtocol* pReliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(reliableProtocolHandle);
            if (pReliableProtocol != NULL)
            {
                pReliableProtocol->Finalize();
                nn::pia::transport::Transport::GetInstance()->DestroyProtocol(reliableProtocolHandle);
            }
        }

        if (unreliableProtocolHandle != 0)
        {
            nn::pia::transport::UnreliableProtocol* pUnreliableProtocol = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(unreliableProtocolHandle);
            if (pUnreliableProtocol != NULL)
            {
                pUnreliableProtocol->Finalize();
                nn::pia::transport::Transport::GetInstance()->DestroyProtocol(unreliableProtocolHandle);
            }
        }
    }

    //n2004: Allows use of any protocol.
    if (pAdditionalFunction != NULL)
    {
        pAdditionalFunction();
    }

    // With Pia, Finalize and DestroyInstance always succeed.
    if (nn::pia::transport::Transport::GetInstance() != NULL)
    {
        nn::pia::transport::Transport::DestroyInstance();
    }
    nn::pia::transport::Finalize();

    PIA_SAMPLE_LOG("Transport_FinalizePiaTransport:finished\n");
}
