﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     SampleStationStatus.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include <pia/common.h>
#include <pia/inet.h>
#include <pia/transport.h>
#include <pia/session.h>

#include "SampleStationStatus.h"

#include "SampleLib/SampleLib.h"


void SampleStationStatus::Startup(nn::pia::StationId id, nn::pia::StationIndex index)
{
    stationId = id;
    stationIndex = index;
    pos.x = 0;
    pos.y = 0;

    routeInfoId = nn::pia::STATION_ID_INVALID;

    rtt = nn::pia::transport::Station::INVALID_RTT;

    this->ResetPacketLossCheck();

    reliableSendCount = 0;
    reliableReceiveCount = 0;
}


void SampleStationStatus::Cleanup()
{
    if (stationId != nn::pia::STATION_ID_INVALID)
    {
        // The valid <tt>StationStatus</tt> outputs information before cleanup.
        PIA_SAMPLE_LOG("SampleStationStatus::Cleanup: StationId:0x%016llX StationIndex:%X Pos:(%d, %d) RouteInfo:0x%016llX RTT:%d\n",
                       stationId.GetValue(),
                       stationIndex,
                       pos.x,
                       pos.y,
                       routeInfoId.GetValue(),
                       rtt);
        PIA_SAMPLE_LOG("  UnreliableProtocol: First:0x%05X Latest:0x%05X TotalCount:0x%05X LossCount:0x%05X LossRate:%5.2f%%\n",
                       unreliable.firstNum,
                       unreliable.latestNum,
                       (unreliable.latestNum - unreliable.firstNum),
                       unreliable.lossCount,
                       unreliable.CalcPacketLoss() * 100.0f);
        PIA_SAMPLE_LOG("  RoundRobinUP      : First:0x%05X Latest:0x%05X TotalCount:0x%05X LossCount:0x%05X LossRate:%5.2f%%\n",
                       roundRobin.firstNum,
                       roundRobin.latestNum,
                       (roundRobin.latestNum - roundRobin.firstNum),
                       roundRobin.lossCount,
                       roundRobin.CalcPacketLoss() * 100.0f);
        PIA_SAMPLE_LOG("  ReliableProtocol  : SendCount:%2d  ReceiveCount:%2d\n",
                       reliableSendCount,
                       reliableReceiveCount);
    }

    stationId = nn::pia::STATION_ID_INVALID;
    stationIndex = nn::pia::STATION_INDEX_INVALID;

    pos.x = 0;
    pos.y = 0;

    routeInfoId = nn::pia::STATION_ID_INVALID;

    rtt = nn::pia::transport::Station::INVALID_RTT;

    this->ResetPacketLossCheck();

    reliableSendCount = 0;
    reliableReceiveCount = 0;
}


void SampleStationStatus::Serialize(void* pBuf, bool isRoundRobin) const
{
    PIA_SAMPLE_ASSERT(pBuf != NULL);
    PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pBuf) % 4 == 0);

    u8* p = static_cast<u8*>(pBuf);

    *reinterpret_cast<u64*>(p) = nn::pia::common::ByteOrder::HostToNetwork64(stationId.GetValue());
    p += 8;
    *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.x);
    p += 2;
    *reinterpret_cast<u16*>(p) = nn::pia::common::ByteOrder::HostToNetwork16(pos.y);
    p += 2;
    if (isRoundRobin)
    {
        *reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(roundRobin.latestNum);
    }
    else
    {
        *reinterpret_cast<u32*>(p) = nn::pia::common::ByteOrder::HostToNetwork32(unreliable.latestNum);
    }
    p += 4;
    PIA_SAMPLE_ASSERT(p == static_cast<u8*>(pBuf) + SERIALIZED_SIZE);
}


void SampleStationStatus::Deserialize(const void* pData, bool isRoundRobin)
{
    PIA_SAMPLE_ASSERT(pData != NULL);
    PIA_SAMPLE_ASSERT(reinterpret_cast<u32>(pData) % 4 == 0);

    const u8* p = static_cast<const u8*>(pData);

    stationId = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(p));
    p += 8;
    pos.x = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
    p += 2;
    pos.y = nn::pia::common::ByteOrder::NetworkToHost16(*reinterpret_cast<const u16*>(p));
    p += 2;

    if (isRoundRobin)
    {
        s32 calcBuf = roundRobin.latestNum;
        roundRobin.latestNum = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
        if (roundRobin.firstNum == 0)
        {
            roundRobin.firstNum = roundRobin.latestNum;
        }
        else
        {
            if (static_cast<s32>(roundRobin.latestNum) - calcBuf - 1 > 0)
            {
                // There is packet loss if the send count has increased by more than 1.
                // (But with the UnreliableProtocol, because there is a possibility for packets to arrive out of order, there actually might not have been any packet loss.)
                roundRobin.lossCount += static_cast<s32>(roundRobin.latestNum) - calcBuf - 1;
            }
        }
    }
    else
    {
        s32 calcBuf = unreliable.latestNum;
        unreliable.latestNum = nn::pia::common::ByteOrder::NetworkToHost32(*reinterpret_cast<const u32*>(p));
        if (unreliable.firstNum == 0)
        {
            unreliable.firstNum = unreliable.latestNum;
        }
        else
        {
            if (static_cast<s32>(unreliable.latestNum) - calcBuf - 1 > 0)
            {
                // There is packet loss if the send count has increased by more than 1.
                // (But with the UnreliableProtocol, because there is a possibility for packets to arrive out of order, there actually might not have been any packet loss.)
                unreliable.lossCount += static_cast<s32>(unreliable.latestNum) - calcBuf - 1;
            }
        }
    }
}


void SampleStationStatus::UpdateInfo(bool enableRelay)
{
    // Update RTT.
    nn::pia::transport::Station* pStationBuf = nn::pia::transport::StationManager::GetInstance()->GetStation(stationId);
    if (PIA_IS_VALID_POINTER(pStationBuf))
    {
        // Gets the RTT from the station instance corresponding to <tt>StationId</tt>.
        rtt = pStationBuf->GetRtt();
    }
    else
    {
        // Set to an invalid value if you cannot get the station instance.
        rtt = nn::pia::transport::Station::INVALID_RTT;
    }

    // Update the connection route information.
    if (enableRelay)
    {
        nn::pia::StationIndex stationIndex;
        nn::pia::StationIndex routeIndex;
        nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
        if (result.IsSuccess())
        {
            // If the relay feature is enabled, get the connection route information using the function for session debugging.
            result = nn::pia::session::Mesh::GetInstance()->GetConnectionRoute(stationIndex, &routeIndex);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = nn::pia::transport::Transport::GetInstance()->ConvertToStationId(&routeInfoId, routeIndex);
            if (result == nn::pia::ResultNotFound())
            {
                // Even after the connection is established, a peer for which <tt>StationIndex</tt> cannot be converted to <tt>StationId</tt> might become a relay base.
                // But this is only temporary and will be resolved (the value will be able to be converted to <tt>StationId</tt>) if the process proceeds.
                // To indicate that this is the situation, set <tt>routeInfoId</tt> to STATION_ID_INVALID.
                routeInfoId = nn::pia::STATION_ID_INVALID;
            }
            else
            {
                // There is not other reason for the process to fail.
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
        }
    }
    else
    {
        // If the relay feature is disabled, the direct route is the only route.
        routeInfoId = stationId;
    }
}


void SampleStationStatus::ResetPacketLossCheck()
{
    unreliable.firstNum = 0;
    unreliable.latestNum = 0;
    unreliable.lossCount = 0;

    roundRobin.firstNum = 0;
    roundRobin.latestNum = 0;
    roundRobin.lossCount = 0;
}
