﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#include "SampleStationStatus.h"
#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Nex_Matchmake.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Session_Mesh.h"

#if NN_PIA_CTR
#include <nn/version.h>
#elif NN_PIA_CAFE
#include <sdk_ver.h>
#endif

#if NN_PIA_NEX_3
#include <OnlineCore/src/version/nex_version.h>
#else
#include <version/nex_version.h>
#endif

// This is for development use only. Do not use.
// Toggling this macro is enabled only in non-releases for CTR or only in Debug for Cafe.
#define ENABLE_DEBUG_CLIENT 0

#if ENABLE_DEBUG_CLIENT
#if NN_PIA_CTR
#pragma diag_suppress 1300
#endif
#include <RendezVous/Services/Jugem/Debug/src/Client/DebugClient.h>
#if NN_PIA_CTR
#pragma diag_warning 1300
#endif
#endif

#include <pia/util.h>

#include <cstdio>

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

#define PROGRAM_TITLE "TopologyTest"

namespace
{
// The group of operational settings for the demo.

// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const u32 SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const u32 SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

#define SAMPLE_NEX_THREAD_MODE nn::nex::Core::ThreadModeUnsafeTransportBuffer

const u32 NEX_DISPATCH_PERIOD = 3;

const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 35700;  // Game mode.
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;

#if ENABLE_DEBUG_CLIENT
nn::nex::DebugClient* s_pDebugClient = NULL;
#endif


bool IsConnected()
{
    // Status check for network connection to AC.
    if (!Inet_IsConnected())
    {
        PIA_SAMPLE_LOG("IsConnected: Unreach for internet.\n");
        return false;
    }

    // Status check for NEX system error.
    if (Nex_CheckInternalError())
    {
        PIA_SAMPLE_LOG("IsConnected: NEX internal error occurs.\n");
        return false;
    }

    return true;
}

#if NN_PIA_CAFE
const size_t MAX_STATIONS = 16;                      // Defines the maximum number of stations.
const size_t MEMSIZE = 8 * 1024 * 1024;              // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
#elif NN_PIA_CTR
const size_t MAX_STATIONS = 12;                      // Defines the maximum number of stations.
const size_t MEMSIZE = 8 * 1024 * 1024;              // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
#endif
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.

const size_t MAX_SEND_THREAD_BUFFER_NUM = 24;
const size_t MAX_RECEIVE_THREAD_BUFFER_NUM = 32;
const size_t FRAMES_PER_SECOND = 30;

const u16 UNRELIABLE_PROTOCOL_PORT = 0;     // Defines the port number used by UnreliableProtocol.
const u16 R2_UNRELIABLE_PROTOCOL_PORT = 0;  // Defines the port number used by RoundRobinUnreliableProtocol.
const u16 RELIABLE_PROTOCOL_PORT = 0;       // Defines the port number used by ReliableProtocol.

static u32 s_hUnreliable;    // Specifies the handle corresponding to UnreliableProtocol.
static u32 s_hR2Unreliable;  // Specifies the handle corresponding to RoundRobinUnreliableProtocol.
static u32 s_hReliable;      // Specifies the handle corresponding to ReliableProtocol.

// Create and initialize protocols.
nn::Result CreateAndInitializeProtocols()
{
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Create and initialize an UnreliableProtocol instance. (The arrival of data is not guaranteed.)
    {
        PIA_SAMPLE_ASSERT(s_hUnreliable == 0);
        s_hUnreliable = pTransport->CreateProtocol<nn::pia::transport::UnreliableProtocol>(UNRELIABLE_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hUnreliable != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        const u32 UP_RECV_BUF_NUM = 32;
        pTransport->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable)->Initialize(UP_RECV_BUF_NUM);
    }

    // Create and initialize RoundRobinUnreliableProtocol (communication without guaranteed arrival of data).
    {
        s_hR2Unreliable = pTransport->CreateProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(R2_UNRELIABLE_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hR2Unreliable != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        const u32 R2UP_RECV_BUF_NUM = 32;
        pTransport->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_hR2Unreliable)->Initialize(R2UP_RECV_BUF_NUM);
    }

    // Create and initialize a ReliableProtocol instance. (Arrival of data is guaranteed.)
    {
        PIA_SAMPLE_ASSERT(s_hReliable == 0);
        s_hReliable = pTransport->CreateProtocol<nn::pia::transport::ReliableProtocol>(RELIABLE_PROTOCOL_PORT);
        PIA_SAMPLE_ASSERT(s_hReliable != 0);  // The CreateProtocol function returns 0 if the same protocol ID already exists. Confirm that CreateProtocol has not returned 0.
        const u32 RP_SEND_BUF_NUM = 64;
        const u32 RP_RECV_BUF_NUM = 64;
        pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable)->Initialize(RP_SEND_BUF_NUM, RP_RECV_BUF_NUM);
    }

    return nn::ResultSuccess();
}

// Clean up protocols.
void FinalizeAndDestroyProtocols()
{
    // Get the Transport singleton.
    nn::pia::transport::Transport* pTransport = nn::pia::transport::Transport::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTransport));

    // Finalize ReliableProtocol.
    nn::pia::transport::ReliableProtocol* pReliable = pTransport->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable);
    if (PIA_IS_VALID_POINTER(pReliable))
    {
        pReliable->Finalize();
        pTransport->DestroyProtocol(s_hReliable);
        s_hReliable = 0;
    }

    // Finalization of RoundRobinUnreliableProtocol.
    nn::pia::transport::RoundRobinUnreliableProtocol* pR2Unreliable = pTransport->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_hR2Unreliable);
    if (PIA_IS_VALID_POINTER(pR2Unreliable))
    {
        pR2Unreliable->Finalize();
        pTransport->DestroyProtocol(s_hR2Unreliable);
        s_hR2Unreliable = 0;
    }

    // Finalize UnreliableProtocol
    nn::pia::transport::UnreliableProtocol* pUnreliable = pTransport->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable);
    if (PIA_IS_VALID_POINTER(pUnreliable))
    {
        pUnreliable->Finalize();
        pTransport->DestroyProtocol(s_hUnreliable);
        s_hUnreliable = 0;
    }
}

nn::pia::transport::UnreliableProtocol* GetUnreliableProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_hUnreliable);
}

nn::pia::transport::RoundRobinUnreliableProtocol* GetRoundRobinUnreliableProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::RoundRobinUnreliableProtocol>(s_hR2Unreliable);
}

nn::pia::transport::ReliableProtocol* GetReliableProtocol()
{
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::transport::Transport::GetInstance()));
    return nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_hReliable);
}

const size_t UNRELIABLE_SEND_DATA_SIZE = 100;  // Packet size when sending data using Unreliable communications.
const size_t RELIABLE_SEND_DATA_SIZE = 128;    // Packet size when sending data using Reliable communications.

static u32 s_SendingStationStatusPeriod = 1;  // Interval for sending StationStatus (can dynamically change).
static bool s_UseSendToAll = false;           // This sets whether to use SendToAll when sending via UnreliableProtocol/ReliableProtocol (or STATION_ID_ALL when sending via ReliableProtocol).
static bool s_EnableRoundRobin = false;       // This sets whether to use RoundRobin for sending. (Can dynamically change.)
static u32 s_RoundRobinSendStationNum = 2;    // The number of stations to send to at the same time when using RoundRobin. (Can dynamically change.)

static bool s_SendReliableOnce = false;  // Flag for sending all at once with Reliable communication.

static size_t s_StationStatusIndexOffset = 0;   // The starting index when drawing the StationStatus screen. (Can dynamically change.)
static bool s_StationStatusDrawVebose = false;  // To switch to the detailed view when drawing the StationStatus screen. (Can dynamically change.)

struct SamplePreset  // This structure is for the default settings that match the network topology.
{
    nn::pia::session::NetworkTopology topology;  // Network topology settings.
    const char* pTopologyString;                 // String to output for the name of the topology.

    bool bHostMigration;               // Host migration settings.
    const char* pHostMigrationString;  // String to output for the name of the host migration settings.

    u32 gameModeOption;  // Optional value so no matching takes place with other presets.

    u32 sendingPeriod;  // Interval for sending StationStatus.

    bool useSendToAll;             // This sets whether to use SendToAll when sending (or STATION_ID_ALL in the case of ReliableProtocol).
    bool enableRoundRobin;         // This sets whether to use RoundRobin for sending.
    u32 roundRobinSendStationNum;  // The number of stations to send to at the same time when using RoundRobin.

    bool useMesh;

    nn::pia::sampleframework::Color bgColor;  // Used to change background color so changes to presets are readily discernible.

    // Indicates whether the relay feature is enabled in this network topology.
    bool IsEnableRelay() const
    {
        return topology == nn::pia::session::NetworkTopology_RelayMesh;
    }

    // Indicates whether host migration is enabled.
    bool IsEnableHostMigration() const
    {
        return bHostMigration;
    }

    bool IsUseMesh() const
    {
        return useMesh;
    }
};

static size_t s_TargetPresetIndex = 0;  // The index of the preset to use.

const size_t MAX_PRESET_NUM = 12;
const SamplePreset PRESET_ARRAY[MAX_PRESET_NUM] =
    {
        {  // Preset 0:
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh",
            false, "None",
            0, 1, true, false, 2, false,
            nn::pia::sampleframework::Color(0.0f, 0.0f, 0.3f, 1.0f) },
        {  // Preset 1:
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh",
            false, "None",
            2, 1, false, false, 2, false,
            nn::pia::sampleframework::Color(0.0f, 0.3f, 0.0f, 1.0f) },
        {  // Preset 2:
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh",
            true, "Enable",
            3, 1, false, false, 2, false,
            nn::pia::sampleframework::Color(0.3f, 0.0f, 0.0f, 1.0f) },
        {  // Preset 3:
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh",
            true, "Enable",
            4, 1, true, false, 2, false,
            nn::pia::sampleframework::Color(0.3f, 0.0f, 0.3f, 1.0f) },
        {  // Preset 4:
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh",
            true, "Enable",
            5, 1, false, false, 2, false,
            nn::pia::sampleframework::Color(0.0f, 0.3f, 0.3f, 1.0f) },
        {  // Preset 5:
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh",
            true, "Enable",
            6, 1, false, false, 2, false,
            nn::pia::sampleframework::Color(0.3f, 0.3f, 0.0f, 1.0f) },

        {  // Preset 6: FullMesh NoHM Mesh
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh(M)",
            false, "None",
            7, 1, true, false, 2, true,
            nn::pia::sampleframework::Color(0.0f, 0.0f, 0.6f, 1.0f) },
        {  // Preset 7: RelayMesh NoHM Mesh
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh(M)",
            false, "None",
            8, 1, false, false, 2, true,
            nn::pia::sampleframework::Color(0.0f, 0.6f, 0.0f, 1.0f) },
        {  // Preset 8: RelayMesh HM Mesh
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh(M)",
            true, "Enable",
            9, 1, false, false, 2, true,
            nn::pia::sampleframework::Color(0.6f, 0.0f, 0.0f, 1.0f) },
        {  // Preset 9: FullMesh HM Mesh
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh(M)",
            true, "Enable",
            10, 1, true, false, 2, true,
            nn::pia::sampleframework::Color(0.6f, 0.0f, 0.6f, 1.0f) },
        {  // Preset10: RelayMesh HM Mesh
            nn::pia::session::NetworkTopology_RelayMesh, "RelayMesh(M)",
            true, "Enable",
            11, 1, true, false, 2, true,
            nn::pia::sampleframework::Color(0.0f, 0.6f, 0.6f, 1.0f) },
        {  // Preset11: FullMesh HM Mesh
            nn::pia::session::NetworkTopology_FullMesh, "FullMesh(M)",
            true, "Enable",
            12, 1, false, false, 2, true,
            nn::pia::sampleframework::Color(0.6f, 0.6f, 0.0f, 1.0f) }
    };

// When the relay feature is enabled, this sets the maximum number of transfer requests that can be assigned to one station.
const u16 RELAY_SETTING_MAX_NUM_OF_RELAY_ROUTES = 2;
// When the relay feature is enabled, this sets the maximum total RTT for the communication route.
const u16 RELAY_SETTING_MAX_RTT = nn::pia::session::NN_PIA_SESSION_MAX_RTT_DEFAULT;

void SetRelaySettings(u16 maxNumOfRelayRoutes, u16 maxRtt)
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        // Changes settings related to the relay communication feature.
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::session::Mesh::GetInstance()));

        nn::Result result;
        result = nn::pia::session::Mesh::GetInstance()->SetMaxNumOfRelayRoutesAssignedToStation(maxNumOfRelayRoutes);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::session::Mesh::GetInstance()->SetMaxRtt(maxRtt);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }
    else
    {
        // Changes settings related to the relay communication feature.
        PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(nn::pia::session::Session::GetInstance()));

        nn::Result result;
        result = nn::pia::session::Session::GetInstance()->SetMaxNumOfRelayRoutesAssignedToStation(maxNumOfRelayRoutes);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

        result = nn::pia::session::Session::GetInstance()->SetMaxRtt(maxRtt);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    }
}

static bool s_EnableLatencyEmulation = false;  // Indicates whether to use the latency emulation feature.

struct LatencyEmulationSettings  // For managing the settings for latency emulation.
{
    u32 sendMin;
    u32 sendMax;
    u32 receiveMin;
    u32 receiveMax;
};
const LatencyEmulationSettings c_DefaultLatencyEmulationSettings = {
    100, 150, 100, 150
};
static LatencyEmulationSettings s_LatencyEmulationCurrent = c_DefaultLatencyEmulationSettings;
static LatencyEmulationSettings s_LatencyEmulationBuffer = c_DefaultLatencyEmulationSettings;

void SetLatencyEmulation(u32 sendMin, u32 sendMax, u32 recvMin, u32 recvMax)
{
    // Specifies the latency emulation time of the PiaTransport send/receive threads.
    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pMgr));
    pMgr->SetSendThreadLatencyEmulation(sendMin, sendMax);
    pMgr->SetReceiveThreadLatencyEmulation(recvMin, recvMax);
}

static u32 s_KickoutStationTargetIndex = 0;  // Index value of target StationStatus when KickoutStation is called.


// Related to the presets for the relay emulation feature settings.
static const u32 RELAY_EMULATION_PRESET_MAX = 17;
static u32 s_RelayEmulationPresetNum = 0;

u32 GetRelayEmulationPresetNum()
{
    return s_RelayEmulationPresetNum;
}

void SetRelayEmulationPresetNum(u32 num)
{
    if (num <= RELAY_EMULATION_PRESET_MAX)
    {
        s_RelayEmulationPresetNum = num;
    }
}

void RelayEmulationSetting()
{
    // Configures the relay emulation feature based on the preset number.

    nn::pia::session::Mesh* pSession = nn::pia::session::Mesh::GetInstance();
    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pSession));

    nn::Result result;
    const nn::pia::session::Mesh::EmulationType cTypeRelay = nn::pia::session::Mesh::EMULATION_TYPE_RELAY;

    // Initialize relay emulation before configuring in case the previous settings remain.
    pSession->ResetRelayEmulation();

    switch (s_RelayEmulationPresetNum)
    {
    case 1:
        // 2-3 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-3\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 2:
        // 2-4 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 3:
        // 2-4 and 3-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 3-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 4:
        // 3-5 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 3-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 5:
        // 2-5 and 3-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-5, 3-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 6:
        // 2-6 and 3-6 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-6, 3-6\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_6, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_6, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 7:
        // 3-4 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 3-4\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 8:
        // 2-3 and 2-4 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-3, 2-4\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 9:
        // 2-4, 3-5, and 4-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 3-5, 4-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_4, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 10:
        // 2-4 and 2-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 2-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 11:
#if NN_PIA_CAFE
        // 2-16 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-16\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_16, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#elif NN_PIA_CTR
        // 2-12 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-12\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_12, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#endif
        break;
    case 12:
        // 2-4 and 3-4 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 3-4\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 13:
        // 2-3 and 3-4 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-3, 3-4\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 14:
        // 2-4 and 4-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 4-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_4, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 15:
        // 2-5 is a relay connection.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 16:
        // 2-4 and 4-6 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-4, 4-6\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_4, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_4, nn::pia::STATION_INDEX_6, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 17:
        // 2-3 and 3-5 are relay connections.
        PIA_SAMPLE_PRINT("RelayEmulation: 2-3, 3-5\n");
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_2, nn::pia::STATION_INDEX_3, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        result = pSession->RelayEmulation(nn::pia::STATION_INDEX_3, nn::pia::STATION_INDEX_5, cTypeRelay);
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        break;
    case 0:
    default:
        // No relay emulation.
        PIA_SAMPLE_PRINT("RelayEmulation is nothing.\n");
        break;
    }
}


// Related to the presets for the session settings.
static const u32 SESSION_SETTING_PRESET_MAX = 2;
u32 s_SessionSettingPresetNum = 0;
u32 s_SessionMaxSilenceTime = 10000;
u32 s_SessionKeepAliveSendingInterval = 1000;

u32 GetSessionSettingPresetNum()
{
    return s_SessionSettingPresetNum;
}

void SetSessionSettingPresetNum(u32 num)
{
    if (num < SESSION_SETTING_PRESET_MAX)
    {
        s_SessionSettingPresetNum = num;
        switch (s_SessionSettingPresetNum)
        {
        case 1:
            s_SessionMaxSilenceTime = 60000;
            s_SessionKeepAliveSendingInterval = 1000;
            break;
        default:
            s_SessionMaxSilenceTime = 10000;
            s_SessionKeepAliveSendingInterval = 1000;
            break;
        }
    }
}

// Gets Session::DisconnectReason and outputs it to the log.
void CheckSessionDisconnectReason()
{
    const size_t cReasonStrSize = 32;
    char reasonStr[cReasonStrSize] = {};
    nn::pia::session::Mesh::DisconnectReason reason = nn::pia::session::Mesh::GetInstance()->GetDisconnectReason();
    switch (reason)
    {
    case nn::pia::session::Mesh::NOT_YET_COMMUNICATED:
        ::std::strncpy(reasonStr, "NOT_YET_COMMUNICATED", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::OPERATION_OF_OWN:
        ::std::strncpy(reasonStr, "OPERATION_OF_OWN", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::OPERATION_OF_OTHER:
        ::std::strncpy(reasonStr, "OPERATION_OF_OTHER", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::KICKOUT_BY_USER:
        ::std::strncpy(reasonStr, "KICKOUT_BY_USER", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::KICKOUT_BY_SYSTEM:
        ::std::strncpy(reasonStr, "KICKOUT_BY_SYSTEM", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::INCONSISTENT_INFO:
        ::std::strncpy(reasonStr, "INCONSISTENT_INFO", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::MIGRATION_FAIL:
        ::std::strncpy(reasonStr, "MIGRATION_FAIL", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::EXTERNAL_FACTOR:
        ::std::strncpy(reasonStr, "EXTERNAL_FACTOR", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::MIGRATION_FATAL_ERROR:
        ::std::strncpy(reasonStr, "MIGRATION_FATAL_ERROR", cReasonStrSize);
        break;
    case nn::pia::session::Mesh::UNKNOWN_REASON:
    default:
        ::std::strncpy(reasonStr, "UNKNOWN_REASON", cReasonStrSize);
        break;
    }
    PIA_SAMPLE_LOG("PiaSession DisconnectReason:%d (%s)\n", reason, reasonStr);
}


// Indicates the mode that is executing.
// With inet, starts with automatch, and then determines whether to be the host or a client.
enum GameMode
{
    GAME_MODE_INET_AUTOMATCH,
    GAME_MODE_INET_HOST,
    GAME_MODE_INET_CLIENT,
    GAME_MODE_INET_ERROR,
    GAME_MODE_EXIT
};

void InitializeSettings()
{
#if PIA_TEST_CODE
    s_LatencyEmulationCurrent = c_DefaultLatencyEmulationSettings;
    s_LatencyEmulationBuffer = c_DefaultLatencyEmulationSettings;
#endif
}

// Determines the mode to execute.
GameMode DecideGameMode()
{
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;

    InitializeSettings();

    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            s_EnableLatencyEmulation = !s_EnableLatencyEmulation;
        }
        else
        {
            // Up/Down on the +Control Pad: Changes the preset.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                s_TargetPresetIndex = (s_TargetPresetIndex + 1) % MAX_PRESET_NUM;

                pSampleFramework->SetClearColor(PRESET_ARRAY[s_TargetPresetIndex].bgColor);
            }
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                s_TargetPresetIndex = (s_TargetPresetIndex + MAX_PRESET_NUM - 1) % MAX_PRESET_NUM;

                pSampleFramework->SetClearColor(PRESET_ARRAY[s_TargetPresetIndex].bgColor);
            }
            // Left/Right on the +Control Pad: Changes settings of relay emulation feature.
            if (!pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
            {
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
                {
                    SetRelayEmulationPresetNum(GetRelayEmulationPresetNum() - 1);
                }
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    SetRelayEmulationPresetNum(GetRelayEmulationPresetNum() + 1);
                }
            }
            // R Button + left/right on +Control Pad: Changes setting of session feature.
            else
            {
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
                {
                    SetSessionSettingPresetNum(GetSessionSettingPresetNum() - 1);
                }
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    SetSessionSettingPresetNum(GetSessionSettingPresetNum() + 1);
                }
            }
        }

        // Draw to the upper screen.
        pSampleFramework->ChangeRenderTarget(true);

        const s32 cLineHeight = 10;
        s32 line = 10;

        pSampleFramework->DrawText(10, line, PROGRAM_TITLE);
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "Select game mode.");
        line += cLineHeight;

        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[X] Login");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[Y] Toggle LatencyEmulation");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[UP/DOWN] Change preset");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[LEFT/RIGHT] Change RelayEmulation");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[START] Quit this application");
        line += cLineHeight;

        {
            u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
            const char cAnimLetter[] = { '-', '\\', '|', '/' };
            pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
        }

        pSampleFramework->SwapBuffers();

        // Draw to the lower screen.
        pSampleFramework->ChangeRenderTarget(false);

        line = 10;
        pSampleFramework->DrawText(10, line, "Preset index: %d", s_TargetPresetIndex);
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "NetworkTopology  : %s", PRESET_ARRAY[s_TargetPresetIndex].pTopologyString);
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "HostMigration    : %s", PRESET_ARRAY[s_TargetPresetIndex].pHostMigrationString);
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "Sending Period   : %d", PRESET_ARRAY[s_TargetPresetIndex].sendingPeriod);
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "Use SendToAll    : %s", PRESET_ARRAY[s_TargetPresetIndex].useSendToAll ? "true" : "false");
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "Enable RoundRobin: %s", PRESET_ARRAY[s_TargetPresetIndex].enableRoundRobin ? "true" : "false");
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "RoundRobin Send Station Num : %d", PRESET_ARRAY[s_TargetPresetIndex].roundRobinSendStationNum);
        line += cLineHeight;
        pSampleFramework->DrawText(20, line, "Use Mesh         : %s", PRESET_ARRAY[s_TargetPresetIndex].useMesh ? "true" : "false");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "Enable LatencyEmulation: %s", s_EnableLatencyEmulation ? "true" : "false");
        line += cLineHeight;

        line += cLineHeight;
        if (PRESET_ARRAY[s_TargetPresetIndex].topology != nn::pia::session::NetworkTopology_RelayMesh || GetRelayEmulationPresetNum() == 0)
        {
            SetRelayEmulationPresetNum(0);
            pSampleFramework->DrawText(10, line, "RelayEmulation Setting: OFF");
            line += cLineHeight;
        }
        else
        {
            pSampleFramework->DrawText(10, line, "RelayEmulation Setting: %d", GetRelayEmulationPresetNum());
            line += cLineHeight;
        }
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "Session Setting: %d", GetSessionSettingPresetNum());
        line += cLineHeight;

#if PIA_TEST_CODE
        {
            pSampleFramework->SetProgramStatus(isDecided ? 0 : 1, false);
            pSampleFramework->SetProgramParam(0x00, nn::pia::STATION_INDEX_INVALID);
            pSampleFramework->SetProgramParam(0x01, nn::pia::STATION_INDEX_INVALID);

            pSampleFramework->SetProgramParam(0x10, s_TargetPresetIndex);
            pSampleFramework->SetProgramParam(0x11, s_EnableLatencyEmulation ? 1 : 0);
            pSampleFramework->SetProgramParam(0x12, GetRelayEmulationPresetNum());
            pSampleFramework->SetProgramParam(0x13, GetSessionSettingPresetNum());
        }
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return gameMode;
        }
    }
}


// Game loop while logged in.
bool LoginLoop()
{
    PIA_SAMPLE_PRINT("[X]StartAutomatch\n [START]Logout\n");
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    bool isStart = false;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("Start Automatch\n");
            isStart = true;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Logout\n");
            isStart = false;
            isDecided = true;
        }

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount() % NEX_DISPATCH_PERIOD == 0)
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
        }
        if (!IsConnected())
        {
            PIA_SAMPLE_PRINT("Connection is lost\n");
            isStart = false;
            isDecided = true;
        }

        // Draw to the upper screen.
        pSampleFramework->ChangeRenderTarget(true);

        const s32 cLineHeight = 10;
        s32 line = 10;

        pSampleFramework->DrawText(10, line, PROGRAM_TITLE);
        line += cLineHeight;
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[X] Start Automatch");
        line += cLineHeight;
        pSampleFramework->DrawText(10, line, "[START] Logout");
        line += cLineHeight;

        {
            u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
            const char cAnimLetter[] = { '-', '\\', '|', '/' };
            pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
        }

        pSampleFramework->SwapBuffers();

        pSampleFramework->ChangeRenderTarget(false);

#if PIA_TEST_CODE
        {
            pSampleFramework->SetProgramStatus(isDecided ? 0 : 2, false);
            pSampleFramework->SetProgramParam(0x00, nn::pia::STATION_INDEX_INVALID);
            pSampleFramework->SetProgramParam(0x01, nn::pia::STATION_INDEX_INVALID);
        }
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return isStart;
        }
    }
}


// Displays the settings.
void PrintSetting()
{
    PIA_SAMPLE_PRINT("********** Print Setting **********\n");

    PIA_SAMPLE_PRINT("Application : %s\n", PROGRAM_TITLE);

#if NN_PIA_DEBUG
    PIA_SAMPLE_PRINT("Build : Debug\n");
#elif NN_PIA_DEVELOP
    PIA_SAMPLE_PRINT("Build : Development\n");
#elif NN_PIA_RELEASE
    PIA_SAMPLE_PRINT("Build : Release\n");
#endif

#if NN_PIA_CTR
    PIA_SAMPLE_PRINT("SDK Version : %s\n", NN_CURRENT_SDK_VERSION);
    PIA_SAMPLE_PRINT("Firmware Version : %s\n", NN_CURRENT_FIRMWARE_VERSION);
#elif NN_PIA_CAFE
    PIA_SAMPLE_PRINT("SDK Version : %s\n", CAFE_OS_SDK_VERSION_STRING);
#endif
    PIA_SAMPLE_PRINT("Nex Version : %s\n", NEX_VERSION_STRING);

#if NN_PIA_CTR
    nn::fnd::DateTime dateTime = nn::fnd::DateTime::GetNow();
    PIA_SAMPLE_PRINT("Date : %d/%02d/%02d %02d:%02d:%02d\n",
                     dateTime.GetYear(),
                     dateTime.GetMonth(),
                     dateTime.GetDay(),
                     dateTime.GetHour(),
                     dateTime.GetMinute(),
                     dateTime.GetSecond());
#elif NN_PIA_CAFE
    OSCalendarTime calenderTime;
    OSTicksToCalendarTime(OSGetTime(), &calenderTime);
    PIA_SAMPLE_PRINT("Date : %d/%02d/%02d %02d:%02d:%02d\n",
                     calenderTime.year,
                     calenderTime.mon + 1,
                     calenderTime.mday,
                     calenderTime.hour,
                     calenderTime.min,
                     calenderTime.sec);
#endif

#if NN_PIA_CTR
    nn::ac::ApType apType;
    nn::ac::Initialize();
    nn::Result r = nn::ac::GetConnectingApType(&apType);
    const char* apTypeText = "Unknown";
    if (r.IsSuccess())
    {
        switch (apType)
        {
        case nn::ac::CTR::AP_TYPE_NONE:
            apTypeText = "None";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_1:
            apTypeText = "UserSetting1";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_2:
            apTypeText = "UserSetting2";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_3:
            apTypeText = "UserSetting3";
            break;
        case nn::ac::CTR::AP_TYPE_NINTENDO_WIFI_USB_CONNECTOR:
            apTypeText = "USBConnector";
            break;
        case nn::ac::CTR::AP_TYPE_NINTENDO_ZONE:
            apTypeText = "NintendoZone";
            break;
        case nn::ac::CTR::AP_TYPE_WIFI_STATION:
            apTypeText = "WiFiStation";
            break;
        case nn::ac::CTR::AP_TYPE_HOTSPOT:
            apTypeText = "HotSpot";
            break;
        };
    }
    PIA_SAMPLE_PRINT("AP type : %s\n", apTypeText);
#endif

#if NN_PIA_CTR
    nn::cfg::Initialize();
    PIA_SAMPLE_PRINT("TransferableId : %016llx\n", nn::cfg::CTR::GetTransferableId(0xFF000));
#endif

    PIA_SAMPLE_PRINT("Frame per second : %d\n", FRAMES_PER_SECOND);
    PIA_SAMPLE_PRINT("Dispatch timeout : %d ms\n", 0);

    PIA_SAMPLE_PRINT("Matchmake participants max : %d\n", MAX_STATIONS);

    PIA_SAMPLE_PRINT("NEX thread mode : %d\n", nn::nex::Core::GetThreadMode());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadSleepSpan : %d ms\n", nn::nex::RootTransport::GetTransportBufferThreadSleepSpan());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadSendPriority : %d\n", nn::nex::RootTransport::GetTransportBufferThreadSendPriority());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadRecvPriority : %d\n", nn::nex::RootTransport::GetTransportBufferThreadRecvPriority());
    PIA_SAMPLE_PRINT("NEX dispatch period in p2p session: %d\n", NEX_DISPATCH_PERIOD);

    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    if (pMgr)
    {
        PIA_SAMPLE_PRINT("Send thread priority : %d\n", pMgr->GetSendThreadPriority());
        PIA_SAMPLE_PRINT("Send thread sleep span : %d ms\n", pMgr->GetSendThreadSleepTimeSpan());
        PIA_SAMPLE_PRINT("Receive thread priority : %d\n", pMgr->GetReceiveThreadPriority());
        PIA_SAMPLE_PRINT("Receive thread sleep span : %d ms\n", pMgr->GetReceiveThreadSleepTimeSpan());
    }

    PIA_SAMPLE_PRINT("Preset index : %d\n", s_TargetPresetIndex);
    PIA_SAMPLE_PRINT(" NetworkTopology  : %s\n", PRESET_ARRAY[s_TargetPresetIndex].pTopologyString);
    PIA_SAMPLE_PRINT(" HostMigration    : %s\n", PRESET_ARRAY[s_TargetPresetIndex].pHostMigrationString);
    PIA_SAMPLE_PRINT(" Sending Period   : %d\n", PRESET_ARRAY[s_TargetPresetIndex].sendingPeriod);
    PIA_SAMPLE_PRINT(" Use SendToAll    : %s\n", PRESET_ARRAY[s_TargetPresetIndex].useSendToAll ? "true" : "false");
    PIA_SAMPLE_PRINT(" Enable RoundRobin: %s\n", PRESET_ARRAY[s_TargetPresetIndex].enableRoundRobin ? "true" : "false");
    PIA_SAMPLE_PRINT(" Use Mesh         : %s\n", PRESET_ARRAY[s_TargetPresetIndex].useMesh ? "true" : "false");
    PIA_SAMPLE_PRINT(" RoundRobin Send Station Num : %d\n", PRESET_ARRAY[s_TargetPresetIndex].roundRobinSendStationNum);
    if (PRESET_ARRAY[s_TargetPresetIndex].IsEnableRelay())
    {
        PIA_SAMPLE_PRINT(" RelaySetting - MaxNumOfRelayRoutes:%d, MaxRtt:%d\n", RELAY_SETTING_MAX_NUM_OF_RELAY_ROUTES, RELAY_SETTING_MAX_RTT);
    }

    PIA_SAMPLE_PRINT("Enable LatencyEmulation: %s\n", s_EnableLatencyEmulation ? "true" : "false");
    if (s_EnableLatencyEmulation)
    {
        PIA_SAMPLE_PRINT(" SendMin:%d, SendMax:%d, RecvMin:%d, RecvMax:%d\n",
                         s_LatencyEmulationCurrent.sendMin,
                         s_LatencyEmulationCurrent.sendMax,
                         s_LatencyEmulationCurrent.receiveMin,
                         s_LatencyEmulationCurrent.receiveMax);
    }

    PIA_SAMPLE_PRINT("SendDataSize - Unreliable:%d, Reliable:%d\n", UNRELIABLE_SEND_DATA_SIZE, RELIABLE_SEND_DATA_SIZE);

    PIA_SAMPLE_PRINT("***********************************\n");
}


#if NN_PIA_CAFE
// Clears the data cache.
void ClearDataCache()
{
    const u32 DC_SIZE_MAIN = 2 * 1024 * 1024;
    const u32 DC_SIZE_SUB = 512 * 1024;
    u32 range = (OSGetCoreId() == 1) ? DC_SIZE_MAIN : DC_SIZE_SUB;
    // Memory for clearing cache.
    PIA_ATTRIBUTE_ALIGN(32) static u8 s_DataCacheClearBuffer[DC_SIZE_MAIN];
    DCZeroRange(static_cast<void*>(s_DataCacheClearBuffer), range);
}
#endif


SampleStationStatus s_StationStatus[MAX_STATIONS];

u16 GetStationStatusArrayIndex(const nn::pia::StationId& stationId)
{
    // Converts StationId into StationIndex, and uses it as an index value.
    nn::pia::StationIndex stationIndexBuf;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndexBuf, stationId);
    if (result.IsFailure())
    {
        // Must not be called when conversion fails.
        PIA_SAMPLE_LOG("ConvertToStationIndex is failure. StationId:0x%016llX", stationId.GetValue());
        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        return static_cast<u16>(nn::pia::STATION_INDEX_INVALID);
    }
    return static_cast<u16>(stationIndexBuf);
}

nn::pia::StationIndex GetStationIndexByStationStatus(const nn::pia::StationId& targetStationId)
{
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        if (s_StationStatus[i].stationId == targetStationId)
        {
            return s_StationStatus[i].stationIndex;
        }
    }
    return nn::pia::STATION_INDEX_INVALID;
}

void sessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    PIA_SAMPLE_LOG("%s is call(%d, %lld)\n", PIA_CODE_POSITION_FUNC, eventType, stationId.GetValue());
    u16 id;

    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        id = GetStationStatusArrayIndex(stationId);
        s_StationStatus[id].Startup(stationId, static_cast<nn::pia::StationIndex>(id));
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        id = GetStationStatusArrayIndex(stationId);
        PIA_SAMPLE_ASSERT(id < MAX_STATIONS);

        s_StationStatus[id].Cleanup();
    }
    break;
    }
}

void meshCallback(nn::pia::session::Mesh::EventType eventType, nn::pia::StationId stationId)
{
    PIA_SAMPLE_LOG("%s is call(%d, %lld)\n", PIA_CODE_POSITION_FUNC, eventType, stationId.GetValue());
    u16 id;

    switch (eventType)
    {
    case nn::pia::session::Mesh::EVENT_JOIN:
    {
        id = GetStationStatusArrayIndex(stationId);
        s_StationStatus[id].Startup(stationId, static_cast<nn::pia::StationIndex>(id));
    }
    break;
    case nn::pia::session::Mesh::EVENT_LEAVE:
    {
        id = GetStationStatusArrayIndex(stationId);
        PIA_SAMPLE_ASSERT(id < MAX_STATIONS);

        s_StationStatus[id].Cleanup();
    }
    break;
    }
}


const size_t UPPER_SCREEN_WIDTH = 400;   //!<  Defines the horizontal resolution of the upper screen.
const size_t UPPER_SCREEN_HEIGHT = 240;  //!<  Defines the vertical resolution of the upper screen.

const u16 MAX_RECT_POS_X = UPPER_SCREEN_WIDTH / 2;   //!<  Maximum value of the rectangle's x-coordinate.
const u16 MAX_RECT_POS_Y = UPPER_SCREEN_HEIGHT / 2;  //!<  Maximum value of the rectangle's y-coordinate.
const u16 RECT_SIZE = 8;

const size_t COLOR_OF_RECTANGLE_NUM = 16;
const nn::pia::sampleframework::Color COLOR_OF_RECTANGLE[COLOR_OF_RECTANGLE_NUM] =
    {
        nn::pia::sampleframework::Color(1.0f, 0.5f, 1.0f, 1.0f),
        nn::pia::sampleframework::Color(1.0f, 1.0f, 0.5f, 1.0f),
        nn::pia::sampleframework::Color(0.5f, 1.0f, 1.0f, 1.0f),
        nn::pia::sampleframework::Color(1.0f, 0.3f, 0.5f, 1.0f),
        nn::pia::sampleframework::Color(0.5f, 1.0f, 0.3f, 1.0f),
        nn::pia::sampleframework::Color(0.3f, 0.5f, 1.0f, 1.0f),
        nn::pia::sampleframework::COLOR_MAROON,
        nn::pia::sampleframework::COLOR_RED,
        nn::pia::sampleframework::COLOR_OLIVE,
        nn::pia::sampleframework::COLOR_YELLOW,
        nn::pia::sampleframework::COLOR_GREEN,
        nn::pia::sampleframework::COLOR_LIME,
        nn::pia::sampleframework::COLOR_TEAL,
        nn::pia::sampleframework::COLOR_NAVY,
        nn::pia::sampleframework::COLOR_PURPLE,
        nn::pia::sampleframework::COLOR_FUCHSIA
    };

const nn::pia::sampleframework::Color DEFAULT_FONT_COLOR(1.0f, 1.0f, 1.0f, 1.0f);
const nn::pia::sampleframework::Color INVALID_COLOR(0.4f, 0.4f, 0.4f, 1.0f);

// Draw a rectangle to the screen.
void DrawStationRect(SampleStationStatus& stationStatus, s32 offsetX, s32 offsetY)
{
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(DEFAULT_FONT_COLOR);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawRect(offsetX + stationStatus.pos.x, offsetY + stationStatus.pos.y, RECT_SIZE, RECT_SIZE, COLOR_OF_RECTANGLE[stationStatus.stationIndex]);
    nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(offsetX + stationStatus.pos.x + 10, offsetY + stationStatus.pos.y, "<-Idx:%02X", stationStatus.stationIndex);
}

// Draws the StationStatus screen. (You can switch between simple and detailed versions.)
s32 DrawStationStatus(SampleStationStatus& stationStatus, bool verbose, s32 offsetX, s32 offsetY)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    const s32 cLineHeight = 9;
    s32 line = offsetY;

    pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);
    if (stationStatus.stationId != nn::pia::STATION_ID_INVALID)
    {
        pSampleFramework->DrawRect(offsetX, offsetY, RECT_SIZE, RECT_SIZE, COLOR_OF_RECTANGLE[stationStatus.stationIndex]);
        nn::pia::StationId localStationId = nn::pia::session::Mesh::GetInstance()->GetLocalStationId();
        nn::pia::StationIndex localStationIndex = GetStationIndexByStationStatus(localStationId);
        if (stationStatus.stationId == localStationId)
        {
            // The local host.
            pSampleFramework->DrawText(offsetX + 10, line, "%02X LocalStation", stationStatus.stationIndex);
        }
        else if (stationStatus.stationId == stationStatus.routeInfoId)
        {
            // Peer with direct connection routes.
            pSampleFramework->DrawText(offsetX + 10, line, "%02X <===== %02X, RTT:%3d", stationStatus.stationIndex, localStationIndex, stationStatus.rtt);
        }
        else if (stationStatus.routeInfoId == nn::pia::STATION_ID_INVALID)
        {
            // Although this is a relay connection, the StationId of the relay station has not been obtained yet.
            pSampleFramework->DrawText(offsetX + 10, line, "%02X <=\?\?<= %02X, RTT:%3d", stationStatus.stationIndex, localStationIndex, stationStatus.rtt);
        }
        else
        {
            // Relay connection.
            nn::pia::StationIndex routeInfoStationIndex = GetStationIndexByStationStatus(stationStatus.routeInfoId);
            pSampleFramework->DrawText(offsetX + 10, line, "%02X <=%02X<= %02X, RTT:%3d", stationStatus.stationIndex, routeInfoStationIndex, localStationIndex, stationStatus.rtt);
        }
        line += cLineHeight * 1;

        if (verbose)
        {
            // Detailed version.
            pSampleFramework->DrawText(offsetX + 18, line, "UP:%05X, %5.2f%%", stationStatus.unreliable.latestNum, stationStatus.unreliable.CalcPacketLoss() * 100.0f);
            line += cLineHeight * 1;
            pSampleFramework->DrawText(offsetX + 18, line, "R2:%05X, %5.2f%%", stationStatus.roundRobin.latestNum, stationStatus.roundRobin.CalcPacketLoss() * 100.0f);
            line += cLineHeight * 1;
            pSampleFramework->DrawText(offsetX + 18, line, "RP:%5X, %5X", stationStatus.reliableSendCount, stationStatus.reliableReceiveCount);
            line += cLineHeight * 1;
        }
    }
    else
    {
        pSampleFramework->DrawRect(offsetX, offsetY, RECT_SIZE, RECT_SIZE, INVALID_COLOR);
        pSampleFramework->DrawText(offsetX + 10, line, "No Connection");
        line += cLineHeight * 1;
        if (verbose)
        {
            // So it is understood there is an invalid value for the detailed version.
            pSampleFramework->SetFontColor(INVALID_COLOR);

            pSampleFramework->DrawText(offsetX + 18, line, "UP:-----, --.--%%");
            line += cLineHeight * 1;
            pSampleFramework->DrawText(offsetX + 18, line, "R2:-----, --.--%%");
            line += cLineHeight * 1;
            pSampleFramework->DrawText(offsetX + 18, line, "RP:    -,     -");
            line += cLineHeight * 1;
        }
    }

    return line;  // Return the Y coordinate of the next line and use for the next drawing process.
}

nn::pia::StationId GetLocalStationId()
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->GetLocalStationId();
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->GetLocalStationId();
    }
}

nn::pia::StationId GetHostStationId()
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->GetHostStationId();
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->GetHostStationId();
    }
}

u16 GetStationNum()
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->GetStationNum();
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->GetStationNum();
    }
}

bool IsHost()
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->IsHost();
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->IsHost();
    }
}

bool IsJoined()
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->CheckConnectionError().IsSuccess();
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION;
    }
}

nn::Result KickoutStation(nn::pia::StationId stationId)
{
    if (PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
    {
        return nn::pia::session::Mesh::GetInstance()->KickoutStation(stationId);
    }
    else
    {
        return nn::pia::session::Session::GetInstance()->KickoutStation(stationId);
    }
}

enum SampleProtocolType
{
    UNRELIABLE = 0,
    ROUNDROBIN_UNRELIABLE,
    RELIABLE,
    SAMPLE_PROTOCOL_TYPE_MAX
};

// The process for receiving data.
void ReceiveProcess()
{
    // Allocate a receive buffer of size MAX_DATA_SIZE of UnreliableProtocol.
    PIA_ATTRIBUTE_ALIGN(4) static bit8 receiveBuffer[nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE];
    nn::pia::StationId sourceStationId;
    size_t receivedDataSize;

    // Check the receive buffers for all protocols that might be used for sending and receiving in this demo.
    const int USE_PROTOCOL_NUM = 3;
    SampleProtocolType useProtocolList[USE_PROTOCOL_NUM] = {
        UNRELIABLE,
        ROUNDROBIN_UNRELIABLE,
        RELIABLE
    };
    size_t protocolTypeIndex = 0;

    // Because we do not know how much data will be received, repeat the process of receiving until the buffers for all protocols are empty.
    while (true)
    {
        nn::Result recvResult;
        switch (useProtocolList[protocolTypeIndex])
        {
        case UNRELIABLE:
        {
            // UnreliableProtocol
            recvResult = GetUnreliableProtocol()->Receive(
                &sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));
        }
        break;
        case ROUNDROBIN_UNRELIABLE:
        {
            // RoundRobinUnreliableProtocol
            recvResult = GetRoundRobinUnreliableProtocol()->Receive(
                &sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));
        }
        break;
        case RELIABLE:
        {
            // ReliableProtocol
            recvResult = GetReliableProtocol()->Receive(
                &sourceStationId, receiveBuffer, &receivedDataSize, sizeof(receiveBuffer));
        }
        break;
        default:
        {
            PIA_SAMPLE_ASSERT(false);
        }
        }

        if (recvResult.IsSuccess())
        {
            u16 index = GetStationStatusArrayIndex(sourceStationId);
            PIA_SAMPLE_ASSERT(index < MAX_STATIONS);

            // You may need to check whether the received data is correct.

            if (useProtocolList[protocolTypeIndex] == UNRELIABLE)
            {
                // With Unreliable protocols, the position of the rectangle is sent and received.
                s_StationStatus[index].Deserialize(&receiveBuffer, false);
            }
            else if (useProtocolList[protocolTypeIndex] == ROUNDROBIN_UNRELIABLE)
            {
                // With Unreliable protocols, the position of the rectangle is sent and received.
                s_StationStatus[index].Deserialize(&receiveBuffer, true);
            }
            else if (useProtocolList[protocolTypeIndex] == RELIABLE)
            {
#if NN_PIA_UNDER_DEVELOP
                nn::pia::StationId stationIdInMessage;
                nn::pia::StationIndex stationIndexInMessage;

                stationIdInMessage = nn::pia::common::ByteOrder::NetworkToHost64(*reinterpret_cast<const u64*>(&receiveBuffer[0]));
                stationIndexInMessage = static_cast<nn::pia::StationIndex>(receiveBuffer[8]);

                // Checks the message content and the sender's StationId.
                PIA_SAMPLE_ASSERT(stationIdInMessage == sourceStationId);
                PIA_SAMPLE_ASSERT(stationIndexInMessage == index);
#endif // NN_PIA_UNDER_DEVELOP
                // Count the number of times ReliableProtocol has received data.
                ++(s_StationStatus[index].reliableReceiveCount);
            }
        }
        else if (recvResult == nn::pia::ResultNoData())
        {
            // The receive buffer of the target protocol is
            // empty, so check the next protocol.
            ++protocolTypeIndex;
        }
        else if (recvResult == nn::pia::ResultNotInCommunication())
        {
            // Failed to receive data due to a disconnection.
            // You reach this point in cases where you have been kicked out.
            break;
        }
#if 0
            else if ( useProtocolList[protocolTypeIndex] == RELIABLE
                && recvResult == nn::pia::ResultBrokenData() )
            {
                // End communication if an unexpected amount of data is received with ReliableProtocol.
                PIA_SAMPLE_LOG("ReliableProtocol::Receive failed. (BrokenData)\n");
                break;
            }
#endif
        else
        {
            // Other reception failed.
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(recvResult);
        }

        if (protocolTypeIndex >= USE_PROTOCOL_NUM)
        {
            // Reception completed for all protocols.
            break;
        }
    }

    // End reception.
}

// The process for sending data.
void SendProcess()
{
    nn::Result sendResult;

    u32 frameCount = nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount();
    if (frameCount % s_SendingStationStatusPeriod == 0)
    {
        // Send the local station's StationStatus via Unreliable communication.
        nn::pia::StationId localStationId = GetLocalStationId();
        PIA_ATTRIBUTE_ALIGN(4) static bit8 sendBuffer[UNRELIABLE_SEND_DATA_SIZE] = {};
        PIA_COMPILE_ASSERT(UNRELIABLE_SEND_DATA_SIZE >= SampleStationStatus::SERIALIZED_SIZE);

        u32 localStatusIndex;
        bool sendIsNeeded = false;

        // Search for local station's StationStatus and create data to send.
        // Sending is not necessary if there are no other valid stations, so check this too.
        for (int i = 0; i < MAX_STATIONS; ++i)
        {
            if (s_StationStatus[i].stationId == localStationId)
            {
                // Create data for sending.
                s_StationStatus[i].Serialize(sendBuffer, s_EnableRoundRobin);
                localStatusIndex = i;
            }
            else if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
            {
                sendIsNeeded = true;
            }
        }

        if (!sendIsNeeded)
        {
            // Sending is not necessary because there are no other valid stations.
            // Essentially, the session is created with the host as the only participant.
        }
        else if (s_EnableRoundRobin)
        {
            // RoundRobin sending.
            sendResult = GetRoundRobinUnreliableProtocol()->SendToRoundRobin(sendBuffer, sizeof(sendBuffer), s_RoundRobinSendStationNum);
            if (sendResult.IsSuccess())
            {
                // Sending was successful, so increment the send counter.
                ++(s_StationStatus[localStatusIndex].roundRobin.latestNum);
            }
            // In this demo, there is little worry about the failure of Unreliable sending of data.
        }
        else
        {
            bool sendIsSuccess = false;
            if (s_UseSendToAll)
            {
                // Use SendToAll to send as a batch process.
                sendResult = GetUnreliableProtocol()->SendToAll(sendBuffer, sizeof(sendBuffer));
                sendIsSuccess = sendResult.IsSuccess();
                // In this demo, there is little worry about the failure of Unreliable sending of data.
            }
            else
            {
                // Send only to valid parties using Send.
                for (int i = 0; i < MAX_STATIONS; ++i)
                {
                    // Peers other than the local station with a valid StationId.
                    if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID && s_StationStatus[i].stationId != localStationId)
                    {
                        sendResult = GetUnreliableProtocol()->Send(s_StationStatus[i].stationId, sendBuffer, sizeof(sendBuffer));
                        // Treat as successful if data was sent successfully to at least one peer.
                        sendIsSuccess |= sendResult.IsSuccess();
                        // In this demo, there is little worry about the failure of Unreliable sending of data.
                    }
                }
            }

            if (sendIsSuccess)
            {
                // Sending was successful, so increment the send counter.
                ++(s_StationStatus[localStatusIndex].unreliable.latestNum);
            }
        }
    }

    // Send by Reliable protocol.
    if (s_SendReliableOnce)
    {
        nn::pia::StationId localStationId = GetLocalStationId();
        nn::pia::StationIndex localStationIndex = GetStationIndexByStationStatus(localStationId);
        PIA_ATTRIBUTE_ALIGN(4) static bit8 sendBuffer[RELIABLE_SEND_DATA_SIZE] = {};
        // Create data for Reliable sending.
        // Store the StationId of the local station in the first 8 bytes.
        // Store the StationIndex of the local station in the next byte.
        *reinterpret_cast<u64*>(&sendBuffer[0]) = nn::pia::common::ByteOrder::HostToNetwork64(localStationId.GetValue());
        sendBuffer[8] = static_cast<bit8>(localStationIndex);

        if (s_UseSendToAll)
        {
            // Send all at once using STATION_ID_ALL.
            sendResult = GetReliableProtocol()->Send(nn::pia::STATION_ID_ALL, sendBuffer, sizeof(sendBuffer));
            if (sendResult.IsSuccess())
            {
                // Increment the Reliable send counter of the local station's StationStatus.
                for (int i = 0; i < MAX_STATIONS; ++i)
                {
                    if (s_StationStatus[i].stationId == localStationId)
                    {
                        ++(s_StationStatus[i].reliableSendCount);
                        break;
                    }
                }
            }
            else
            {
                // Sending failed.
                PIA_SAMPLE_LOG("ReliableProtocol::Send(STATION_ID_ALL) is failure. Result:0x%08X\n", sendResult.GetPrintableBits());
            }
        }
        else
        {
            // Send to every valid peer.
            for (int i = 0; i < MAX_STATIONS; ++i)
            {
                // Peers other than the local station with a valid StationId.
                if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID && s_StationStatus[i].stationId != localStationId)
                {
                    sendResult = GetReliableProtocol()->Send(s_StationStatus[i].stationId, sendBuffer, sizeof(sendBuffer));
                    if (sendResult.IsSuccess())
                    {
                        // Sending was successful, so increment the Reliable send counter.
                        ++(s_StationStatus[i].reliableSendCount);
                    }
                    else
                    {
                        // Sending failed.
                        PIA_SAMPLE_LOG("ReliableProtocol::Send(Id:%lld) is failure. Result:0x%08X\n", s_StationStatus[i].stationId.GetValue(), sendResult.GetPrintableBits());
                    }
                }
            }
        }

        // Once sent, clear the send flag (regardless of whether the send was successful).
        s_SendReliableOnce = false;
    }
}

static size_t s_ChangeSettingTargetIndex = 0, s_CursorIndex = 0;
enum ChangeSettingTarget
{
    UNRELIABLE_SENDING_PERIOD = 0,  // Frequency of Unreliable sending (two-digit value).
    ROUND_ROBIN_SEND_STATION_NUM,   // Number of stations for RoundRobin sending (two-digit value).
    LATENCY_EMULATION_SETTING,      // Latency emulation settings (three-digit value x 4).
    KICKOUT_STATION,                // KickoutStation target setting and function call (two-digit value).
    CHANGE_SETTING_TARGET_NUM
};
const size_t MAX_CURSOR_INDEX_ARRAY[CHANGE_SETTING_TARGET_NUM] = {
    2, 2, 12, 2
};

// Operation by key input.
bool Operate(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::Result r;

    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        // Exit the GameLoop function if START is pressed.
        PIA_SAMPLE_PRINT("Exit GameLoop\n");
        return true;
    }

    switch (gameMode)
    {
    case GAME_MODE_INET_HOST:
    case GAME_MODE_INET_CLIENT:
    {
        s32 move_x = 0, move_y = 0;
        bool resetPacketLossCheck = false;
        if (!pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L | nn::pia::sampleframework::BUTTON_R))
        {
            // +Control Pad: Moves the rectangle.
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_LEFT))
            {
                move_x = -1;
            }
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                move_x = 1;
            }
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_UP))
            {
                move_y = -1;
            }
            if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_DOWN))
            {
                move_y = 1;
            }

            // A Button: Sets the Reliable transfer flag.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                s_SendReliableOnce = true;
            }

            // B Button: Switches between simple and detailed view of StationStatus.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
            {
                s_StationStatusDrawVebose = !s_StationStatusDrawVebose;
            }

            // X Button: Toggles use of RoundRobin sending.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                s_EnableRoundRobin = !s_EnableRoundRobin;
            }

            // Y Button: Resets the counter for checking for packet loss.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                resetPacketLossCheck = true;
            }
        }
        else if (pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L) && !pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
        {
            // L + +Control Pad up/down: Scrolls the StationStatus display.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
            {
                s_StationStatusIndexOffset = (s_StationStatusIndexOffset > 0) ? s_StationStatusIndexOffset - 1 : 0;
            }
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
            {
                s_StationStatusIndexOffset = (s_StationStatusIndexOffset < MAX_STATIONS - 1) ? s_StationStatusIndexOffset + 1 : s_StationStatusIndexOffset;
            }
        }
        else if (!pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_L) && pSampleFramework->IsHold(nn::pia::sampleframework::BUTTON_R))
        {
            // R + X/Y Buttons: Changes the target for changing settings.
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
            {
                s_ChangeSettingTargetIndex = (s_ChangeSettingTargetIndex + 1) % CHANGE_SETTING_TARGET_NUM;
                s_CursorIndex = 0;
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
            {
                s_ChangeSettingTargetIndex = (s_ChangeSettingTargetIndex + CHANGE_SETTING_TARGET_NUM - 1) % CHANGE_SETTING_TARGET_NUM;
                s_CursorIndex = 0;
            }
            // R + A Button: Applies changes to latency emulation settings / Calls KickoutStation.
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
            {
                if (s_EnableLatencyEmulation && static_cast<ChangeSettingTarget>(s_ChangeSettingTargetIndex) == LATENCY_EMULATION_SETTING)
                {
                    // Apply settings after the Buffer value has been copied to Current.
                    s_LatencyEmulationCurrent = s_LatencyEmulationBuffer;
                    SetLatencyEmulation(
                        s_LatencyEmulationCurrent.sendMin, s_LatencyEmulationCurrent.sendMax, s_LatencyEmulationCurrent.receiveMin, s_LatencyEmulationCurrent.receiveMax);
                }
                else if (static_cast<ChangeSettingTarget>(s_ChangeSettingTargetIndex) == KICKOUT_STATION)
                {
                    // Call to KickoutStation.
                    // Normally there should be branches, like "do not call KickoutStation if the StationId is for a non-participating station,"
                    // but to simplify the code (and to demonstrate that the call can fail), KickoutStation is simply called in this demo.
                    nn::pia::StationId stationId = s_StationStatus[s_KickoutStationTargetIndex].stationId;
                    r = KickoutStation(stationId);
                    if (r.IsSuccess())
                    {
                        PIA_SAMPLE_LOG("KickoutStation(TargetId:%lld) is success.\n", stationId.GetValue());
                    }
                    else
                    {
                        PIA_SAMPLE_LOG("KickoutStation(TargetId:%lld) is failure. Result:0x%08X\n", stationId.GetValue(), r.GetPrintableBits());
                    }
                }
            }
            else
            {
                int horizontal = 0, vertical = 0;
                // R + +Control Pad: Dynamically changes the values for various settings.
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
                {
                    horizontal = -1;
                }
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
                {
                    horizontal = 1;
                }
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
                {
                    vertical = -1;
                }
                if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
                {
                    vertical = 1;
                }

                if (horizontal != 0)
                {
                    s_CursorIndex = (s_CursorIndex + MAX_CURSOR_INDEX_ARRAY[s_ChangeSettingTargetIndex] + horizontal) % MAX_CURSOR_INDEX_ARRAY[s_ChangeSettingTargetIndex];
                }
                if (vertical != 0)
                {
                    switch (static_cast<ChangeSettingTarget>(s_ChangeSettingTargetIndex))
                    {
                    case UNRELIABLE_SENDING_PERIOD:  // Frequency of Unreliable sending.
                    {
                        const u32 MAX_SEND_PERIOD = FRAMES_PER_SECOND, MIN_SEND_PERIOD = 1;
                        u32 calcBuf = 1;
                        for (int i = 0; i < s_CursorIndex; ++i)
                        {
                            calcBuf *= 10;
                        }
                        if (vertical > 0)
                        {
                            s_SendingStationStatusPeriod = (s_SendingStationStatusPeriod - MIN_SEND_PERIOD < calcBuf) ? MIN_SEND_PERIOD : s_SendingStationStatusPeriod - calcBuf;
                        }
                        else
                        {
                            s_SendingStationStatusPeriod = (MAX_SEND_PERIOD - s_SendingStationStatusPeriod < calcBuf) ? MAX_SEND_PERIOD : s_SendingStationStatusPeriod + calcBuf;
                        }
                    }
                    break;
                    case ROUND_ROBIN_SEND_STATION_NUM:  // Number of stations for RoundRobin sending.
                    {
                        const u32 MAX_R2_SEND_STATION_NUM = MAX_STATIONS, MIN_R2_SEND_STATION_NUM = 1;
                        u32 calcBuf = 1;
                        for (int i = 0; i < s_CursorIndex; ++i)
                        {
                            calcBuf *= 10;
                        }
                        if (vertical > 0)
                        {
                            s_RoundRobinSendStationNum = (s_RoundRobinSendStationNum - MIN_R2_SEND_STATION_NUM < calcBuf) ? MIN_R2_SEND_STATION_NUM : s_RoundRobinSendStationNum - calcBuf;
                        }
                        else
                        {
                            s_RoundRobinSendStationNum = (MAX_R2_SEND_STATION_NUM - s_RoundRobinSendStationNum < calcBuf) ? MAX_R2_SEND_STATION_NUM : s_RoundRobinSendStationNum + calcBuf;
                        }
                    }
                    break;
                    case LATENCY_EMULATION_SETTING:  // Latency emulation settings.
                    {
                        if (s_EnableLatencyEmulation)
                        {
                            const u32 MAX_LIMIT = 500, MIN_LIMIT = 0;
                            u32 digit = 2 - s_CursorIndex % 3;
                            u32 calcBuf = 1;
                            for (int i = 0; i < digit; ++i)
                            {
                                calcBuf *= 10;
                            }
                            if (s_CursorIndex < 3)
                            {
                                if (vertical > 0)
                                {
                                    s_LatencyEmulationBuffer.sendMin = (s_LatencyEmulationBuffer.sendMin - MIN_LIMIT < calcBuf) ? MIN_LIMIT : s_LatencyEmulationBuffer.sendMin - calcBuf;
                                }
                                else
                                {
                                    s_LatencyEmulationBuffer.sendMin = (s_LatencyEmulationBuffer.sendMax - s_LatencyEmulationBuffer.sendMin < calcBuf) ? s_LatencyEmulationBuffer.sendMax : s_LatencyEmulationBuffer.sendMin + calcBuf;
                                }
                            }
                            else if (s_CursorIndex < 6)
                            {
                                if (vertical > 0)
                                {
                                    s_LatencyEmulationBuffer.sendMax = (s_LatencyEmulationBuffer.sendMax - s_LatencyEmulationBuffer.sendMin < calcBuf) ? s_LatencyEmulationBuffer.sendMin : s_LatencyEmulationBuffer.sendMax - calcBuf;
                                }
                                else
                                {
                                    s_LatencyEmulationBuffer.sendMax = (MAX_LIMIT - s_LatencyEmulationBuffer.sendMax < calcBuf) ? MAX_LIMIT : s_LatencyEmulationBuffer.sendMax + calcBuf;
                                }
                            }
                            else if (s_CursorIndex < 9)
                            {
                                if (vertical > 0)
                                {
                                    s_LatencyEmulationBuffer.receiveMin = (s_LatencyEmulationBuffer.receiveMin - MIN_LIMIT < calcBuf) ? MIN_LIMIT : s_LatencyEmulationBuffer.receiveMin - calcBuf;
                                }
                                else
                                {
                                    s_LatencyEmulationBuffer.receiveMin = (s_LatencyEmulationBuffer.receiveMax - s_LatencyEmulationBuffer.receiveMin < calcBuf) ? s_LatencyEmulationBuffer.receiveMax : s_LatencyEmulationBuffer.receiveMin + calcBuf;
                                }
                            }
                            else  // if( s_CursorIndex < 12 )
                            {
                                if (vertical > 0)
                                {
                                    s_LatencyEmulationBuffer.receiveMax = (s_LatencyEmulationBuffer.receiveMax - s_LatencyEmulationBuffer.receiveMin < calcBuf) ? s_LatencyEmulationBuffer.receiveMin : s_LatencyEmulationBuffer.receiveMax - calcBuf;
                                }
                                else
                                {
                                    s_LatencyEmulationBuffer.receiveMax = (MAX_LIMIT - s_LatencyEmulationBuffer.receiveMax < calcBuf) ? MAX_LIMIT : s_LatencyEmulationBuffer.receiveMax + calcBuf;
                                }
                            }
                        }
                    }
                    break;
                    case KICKOUT_STATION:
                    {
                        u32 calcBuf = 1;
                        for (int i = 0; i < s_CursorIndex; ++i)
                        {
                            calcBuf *= 0x10;
                        }
                        if (vertical > 0)
                        {
                            s_KickoutStationTargetIndex = static_cast<u32>((s_KickoutStationTargetIndex < calcBuf) ? 0 : s_KickoutStationTargetIndex - calcBuf);
                        }
                        else
                        {
                            s_KickoutStationTargetIndex = static_cast<u32>((MAX_STATIONS - s_KickoutStationTargetIndex <= calcBuf) ? MAX_STATIONS - 1 : s_KickoutStationTargetIndex + calcBuf);
                        }
                    }
                    break;
                    }
                }
            }
        }

        if (move_x == 0 && move_y == 0)
        {
            // Move the rectangle so it is easy to discern that communication is ongoing, even when there is no input that moves the rectangle.
            move_x = 1;
        }

        nn::pia::StationId localStationId = GetLocalStationId();
        for (int i = 0; i < MAX_STATIONS; ++i)
        {
            if (s_StationStatus[i].stationId == localStationId)
            {
                // Update the data of the local station.
                s_StationStatus[i].pos.x = (s_StationStatus[i].pos.x + MAX_RECT_POS_X + move_x) % MAX_RECT_POS_X;
                s_StationStatus[i].pos.y = (s_StationStatus[i].pos.y + MAX_RECT_POS_Y + move_y) % MAX_RECT_POS_Y;
            }
            else if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
            {
                if (resetPacketLossCheck)
                {
                    s_StationStatus[i].ResetPacketLossCheck();
                }
            }
        }
    }
    break;
    }

    return false;
}


// This is the process for drawing parts where settings are dynamically changing.
void DrawChangeSetting(s32 offsetX, s32 offsetY)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    const nn::pia::sampleframework::Color cCommandColor(1.0f, 0.8f, 0.0f, 1.0f);

    s32 left = offsetX, top = offsetY;

    pSampleFramework->SetFontColor(cCommandColor);
    pSampleFramework->DrawText(left + 5, top, "[R+Y]");
    pSampleFramework->DrawText(left, top + 10, "<<");
    pSampleFramework->DrawText(left + 250, top, "[R+X]");
    pSampleFramework->DrawText(left + 280, top + 10, ">>");
    top += 10;

    pSampleFramework->DrawText(left, top + 15, "[R+</>]MoveCursor [R+^/v]Add/Subtract");

    pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);

    switch (static_cast<ChangeSettingTarget>(s_ChangeSettingTargetIndex))
    {
    case UNRELIABLE_SENDING_PERIOD:  // Frequency of Unreliable sending.
    {
        left += 30;
        pSampleFramework->DrawText(left, top, "<Unreliable sending period>");
        left -= 10;
        top += 40;
        pSampleFramework->DrawText(left, top, "SendingPeriod: %02d", s_SendingStationStatusPeriod);

        pSampleFramework->SetFontColor(cCommandColor);
        pSampleFramework->DrawText(left + 120 + (8 - s_CursorIndex * 8), top - 8, "v");
    }
    break;
    case ROUND_ROBIN_SEND_STATION_NUM:  // Number of stations for RoundRobin sending.
    {
        left += 30;
        pSampleFramework->DrawText(left, top, "<RoundRobin send Station num>");
        left -= 10;
        top += 40;
        pSampleFramework->DrawText(left, top, "RoundRobinSendStationNum: %02d", s_RoundRobinSendStationNum);

        pSampleFramework->SetFontColor(cCommandColor);
        pSampleFramework->DrawText(left + 208 + (8 - s_CursorIndex * 8), top - 8, "v");
    }
    break;
    case LATENCY_EMULATION_SETTING:  // Latency emulation settings.
    {
        left += 30;
        pSampleFramework->DrawText(left, top, "<LatencyEmulation settings>");

        if (!s_EnableLatencyEmulation)
        {
            pSampleFramework->SetFontColor(INVALID_COLOR);
        }

        size_t digit = 2 - s_CursorIndex % 3;
        size_t target = (s_CursorIndex < 3) ? 0 : (s_CursorIndex < 6) ? 1 : (s_CursorIndex < 9) ? 2 : 3;
        s32 cursorLeft, cursorTop;
        left -= 10;
        top += 40;
        pSampleFramework->DrawText(left, top, "SendMin   :%03d", s_LatencyEmulationBuffer.sendMin);
        if (target == 0)
        {
            cursorLeft = left + 88 + (16 - digit * 8);
            cursorTop = top - 8;
        }
        left += 140;
        pSampleFramework->DrawText(left, top, "SendMax   :%03d", s_LatencyEmulationBuffer.sendMax);
        if (target == 1)
        {
            cursorLeft = left + 88 + (16 - digit * 8);
            cursorTop = top - 8;
        }
        left -= 140;
        top += 20;
        pSampleFramework->DrawText(left, top, "ReceiveMin:%03d", s_LatencyEmulationBuffer.receiveMin);
        if (target == 2)
        {
            cursorLeft = left + 88 + (16 - digit * 8);
            cursorTop = top - 8;
        }
        left += 140;
        pSampleFramework->DrawText(left, top, "ReceiveMax:%03d", s_LatencyEmulationBuffer.receiveMax);
        if (target == 3)
        {
            cursorLeft = left + 88 + (16 - digit * 8);
            cursorTop = top - 8;
        }

        if (s_EnableLatencyEmulation)
        {
            pSampleFramework->SetFontColor(cCommandColor);
            pSampleFramework->DrawText(cursorLeft, cursorTop, "v");

            pSampleFramework->DrawText(left - 140, top + 15, "[R+A] Apply this Setting");
        }
    }
    break;
    case KICKOUT_STATION:  // KickoutStation target setting and function call.
    {
        left += 30;
        pSampleFramework->DrawText(left, top, "<Kickout of Station>");

        left -= 10;
        top += 40;
        pSampleFramework->DrawText(left, top, "Kickout Target Index: 0x%02X", s_KickoutStationTargetIndex);

        pSampleFramework->SetFontColor(cCommandColor);
        pSampleFramework->DrawText(left + 192 + (8 - s_CursorIndex * 8), top - 8, "v");

        pSampleFramework->DrawText(left, top + 15, "[R+A] Execute KickoutStation");
    }
    break;
    }
}


// Draw on the screen.
void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::StationId localStationId = GetLocalStationId();
    nn::pia::StationId hostStationId = GetHostStationId();
    u16 stationNum = GetStationNum();
    bool isHost = IsHost();

    pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);

    // Draw to the upper screen.
    pSampleFramework->ChangeRenderTarget(true);
    {
        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
        case GAME_MODE_INET_AUTOMATCH:
            cGameModeText = "InetAutomatch";
            break;
        case GAME_MODE_INET_HOST:
            cGameModeText = "InetHost";
            break;
        case GAME_MODE_INET_CLIENT:
            cGameModeText = "InetClient";
            break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%2d Idx:%02X %s",
            stationNum, GetStationIndexByStationStatus(localStationId), isHost ? "Host" : "Client");
        pSampleFramework->DrawText(180, 29, "Local:%lld Host:%lld", localStationId.GetValue(), hostStationId.GetValue());
    }

    {
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    // Display Control Method
    {
        const s32 cLineHeight = 9;
        s32 left = 200;
        s32 line = 40;
        pSampleFramework->DrawText(left, line, "[START]End communication");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[X]Toggle RoundRobin");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[Y]Reset PacketLossCheck");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[A]Reliable Send once");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[B]Toggle Status verbose");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[L+^/v]Scroll Status");
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "[R+*]Change settings");
        line += cLineHeight;
    }

    // Draw rectangle, and draw StationStatus.
    s32 drawStatusOffsetY = 40;
    const s32 RECT_FIELD_LEFT = UPPER_SCREEN_WIDTH / 2;
    const s32 RECT_FIELD_TOP = UPPER_SCREEN_HEIGHT / 2;

    bool drawStatusFlag = false;
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        if (i == s_StationStatusIndexOffset)
        {
            drawStatusFlag = true;
        }

        if (s_StationStatus[i].stationId != nn::pia::STATION_ID_INVALID)
        {
            // Update information for valid stations before drawing. (Determine the network topology setting based on whether the relay feature is enabled or disabled.)
            s_StationStatus[i].UpdateInfo(PRESET_ARRAY[s_TargetPresetIndex].IsEnableRelay());

            DrawStationRect(s_StationStatus[i], RECT_FIELD_LEFT, RECT_FIELD_TOP);
        }

        if (drawStatusFlag)
        {
            s32 yBuf = DrawStationStatus(s_StationStatus[i], s_StationStatusDrawVebose, 10, drawStatusOffsetY);
            if (yBuf + (yBuf - drawStatusOffsetY) > UPPER_SCREEN_HEIGHT)
            {
                // The next status would be drawn off the screen, so do not draw.
                drawStatusFlag = false;
            }
            drawStatusOffsetY = yBuf;
        }
    }

    // The border of the region for the rectangle.
    pSampleFramework->DrawLine(RECT_FIELD_LEFT, RECT_FIELD_TOP, RECT_FIELD_LEFT, RECT_FIELD_TOP + MAX_RECT_POS_Y, INVALID_COLOR);
    pSampleFramework->DrawLine(RECT_FIELD_LEFT, RECT_FIELD_TOP, RECT_FIELD_LEFT + MAX_RECT_POS_X, RECT_FIELD_TOP, INVALID_COLOR);
    pSampleFramework->DrawLine(RECT_FIELD_LEFT + MAX_RECT_POS_X, RECT_FIELD_TOP, RECT_FIELD_LEFT + MAX_RECT_POS_X, RECT_FIELD_TOP + MAX_RECT_POS_Y, INVALID_COLOR);
    pSampleFramework->DrawLine(RECT_FIELD_LEFT, RECT_FIELD_TOP + MAX_RECT_POS_Y - 1, RECT_FIELD_LEFT + MAX_RECT_POS_X, RECT_FIELD_TOP + MAX_RECT_POS_Y - 1, INVALID_COLOR);

    pSampleFramework->SwapBuffers();

    // Draw to the lower screen.
    pSampleFramework->ChangeRenderTarget(false);

    // Display of settings.
    {
        const s32 cLineHeight = 9;
        s32 left = 10;
        s32 line = 10;

        pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);

        pSampleFramework->DrawText(left, line, "MaxStations:%2d", MAX_STATIONS);
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "SendDataSize");
        pSampleFramework->DrawText(left + 105, line, "- Unreliable:%4d", UNRELIABLE_SEND_DATA_SIZE);
        line += cLineHeight;
        pSampleFramework->DrawText(left + 105, line, "- Reliable  :%4d", RELIABLE_SEND_DATA_SIZE);
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "NetworkTopology:%s", PRESET_ARRAY[s_TargetPresetIndex].pTopologyString);
        line += cLineHeight;
        if (PRESET_ARRAY[s_TargetPresetIndex].IsEnableRelay())
        {
            pSampleFramework->DrawText(left + 15, line, "MaxNumOfRelayRoutes:%2d,  MaxRtt:%3d", RELAY_SETTING_MAX_NUM_OF_RELAY_ROUTES, RELAY_SETTING_MAX_RTT);
            line += cLineHeight;
        }
        else
        {
            pSampleFramework->SetFontColor(INVALID_COLOR);
            pSampleFramework->DrawText(left + 15, line, "MaxNumOfRelayRoutes:--,  MaxRtt:---");
            line += cLineHeight;
            pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);
        }
        pSampleFramework->DrawText(left, line, "HostMigrationMode:%s", PRESET_ARRAY[s_TargetPresetIndex].pHostMigrationString);
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "LatencyEmulation: %s", s_EnableLatencyEmulation ? "ON" : "OFF");
        line += cLineHeight;
        if (s_EnableLatencyEmulation)
        {
            pSampleFramework->DrawText(left + 15, line, "SendMin:%3d, SendMax:%3d", s_LatencyEmulationCurrent.sendMin, s_LatencyEmulationCurrent.sendMax);
            line += cLineHeight;
            pSampleFramework->DrawText(left + 15, line, "RecvMin:%3d, RecvMax:%3d", s_LatencyEmulationCurrent.receiveMin, s_LatencyEmulationCurrent.receiveMax);
            line += cLineHeight;
        }
        else
        {
            pSampleFramework->SetFontColor(INVALID_COLOR);
            pSampleFramework->DrawText(left + 15, line, "SendMin:---, SendMax:---");
            line += cLineHeight;
            pSampleFramework->DrawText(left + 15, line, "RecvMin:---, RecvMax:---");
            line += cLineHeight;
            pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);
        }
        pSampleFramework->DrawText(left, line, "UnreliableSendingPeriod:%d", s_SendingStationStatusPeriod);
        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "RoundRobin:%3s SendStationNum:%2d", s_EnableRoundRobin ? "ON" : "OFF", s_RoundRobinSendStationNum);
        line += cLineHeight;

        line += cLineHeight;
        pSampleFramework->DrawText(left, line, "Change Setting:");
        line += cLineHeight;
        DrawChangeSetting(left, line + 10);
    }

    pSampleFramework->SetFontColor(DEFAULT_FONT_COLOR);
}


#if PIA_TEST_CODE
void UpdatePiaTestData(bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::StationId localStationId = GetLocalStationId();
    nn::pia::StationId hostStationId = GetHostStationId();
    u16 stationNum = GetStationNum();
    bool isHost = IsHost();
    bool isJoined = IsJoined();

    pSampleFramework->SetProgramStatus(isExit ? 0 : 3, isJoined, isHost, stationNum);

    pSampleFramework->SetProgramParam(0x00, GetStationIndexByStationStatus(localStationId));
    pSampleFramework->SetProgramParam(0x01, GetStationIndexByStationStatus(hostStationId));

    pSampleFramework->SetProgramParam(0x20, s_ChangeSettingTargetIndex);
    pSampleFramework->SetProgramParam(0x21, s_CursorIndex);

    pSampleFramework->SetProgramParam(0x30, s_SendingStationStatusPeriod);
    pSampleFramework->SetProgramParam(0x31, s_RoundRobinSendStationNum);
    pSampleFramework->SetProgramParam(0x32, s_LatencyEmulationBuffer.sendMin);
    pSampleFramework->SetProgramParam(0x33, s_LatencyEmulationBuffer.sendMax);
    pSampleFramework->SetProgramParam(0x34, s_LatencyEmulationBuffer.receiveMin);
    pSampleFramework->SetProgramParam(0x35, s_LatencyEmulationBuffer.receiveMax);
    pSampleFramework->SetProgramParam(0x36, s_KickoutStationTargetIndex);
    pSampleFramework->SetProgramParam(0x37, nn::pia::session::Mesh::GetInstance()->GetProcessUpdateMeshJobStepSequence());

    pSampleFramework->SetProgramParam(0x40, s_LatencyEmulationCurrent.sendMin);
    pSampleFramework->SetProgramParam(0x41, s_LatencyEmulationCurrent.sendMax);
    pSampleFramework->SetProgramParam(0x42, s_LatencyEmulationCurrent.receiveMin);
    pSampleFramework->SetProgramParam(0x43, s_LatencyEmulationCurrent.receiveMax);

    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        u8 idOfs = i * 0x4 + 0x80;
        pSampleFramework->SetProgramParam(idOfs + 0x0, s_StationStatus[i].stationIndex);
        pSampleFramework->SetProgramParam(idOfs + 0x1, GetStationIndexByStationStatus(s_StationStatus[i].routeInfoId));
        pSampleFramework->SetProgramParam(idOfs + 0x2, s_StationStatus[i].rtt);
    }
}
#endif

// Main loop during communication.
GameMode MainLoop(GameMode gameMode)
{
    nn::Result r;

    nn::pia::StationId localStationId = GetLocalStationId();
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        if (s_StationStatus[i].stationId == localStationId)  // Update the data of the local station.
        {
            s_StationStatus[i].pos.x = 0;
            s_StationStatus[i].pos.y = (RECT_SIZE * i);

            // Update the send and receive count values now, so that the count value becomes 1 the first time data is sent.
            s_StationStatus[i].unreliable.latestNum = 1;
            s_StationStatus[i].roundRobin.latestNum = 1;
        }
    }
    s_StationStatusIndexOffset = 0;
    s_StationStatusDrawVebose = false;

    // Apply the selected presets.
    s_SendingStationStatusPeriod = PRESET_ARRAY[s_TargetPresetIndex].sendingPeriod;
    s_UseSendToAll = PRESET_ARRAY[s_TargetPresetIndex].useSendToAll;
    s_EnableRoundRobin = PRESET_ARRAY[s_TargetPresetIndex].enableRoundRobin;
    s_RoundRobinSendStationNum = PRESET_ARRAY[s_TargetPresetIndex].roundRobinSendStationNum;

    bool isError = false;
    while (true)
    {
        bool isExit = false;
        bool bCallLeaveSession = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

#if NN_PIA_CAFE
        // Puts into a state where no cache hit occurs for the operating core data cache.
        ClearDataCache();
#endif

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount() % NEX_DISPATCH_PERIOD == 0)
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
        {
            nn::pia::session::Session::Status status = nn::pia::session::Session::GetInstance()->GetStatus();
            if (status == nn::pia::session::Session::STATUS_DISCONNECTED_NETWORK)
            {
                // Ends because the communication has been disconnected.
                isExit = true;
                isError = true;
            }
        }
        else
        {
            // Checks the AC network connection status, game server connection status, and NEX internal errors.
            if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
            {
                // Ends because the communication has been disconnected.
                isExit = true;
                isError = true;
            }
        }

        switch (gameMode)
        {
        case GAME_MODE_INET_HOST:
        case GAME_MODE_INET_CLIENT:
            if (!IsConnected())
            {
                // Ends because the communication has been disconnected.
                isExit = true;
                isError = true;
            }
            break;
        }

        {
            if (IsJoined())
            {
                // Receive.
                ReceiveProcess();

                // If currently communicating, operations are received.
                isExit |= Operate(gameMode);

                // Send.
                SendProcess();
            }
            else
            {
                // Ends because the session was lost.
                isExit = true;

                // Log reason why session was disconnected.
                CheckSessionDisconnectReason();
            }

            Draw(gameMode);
        }

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        // Check the session connection status.
        if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
        {
            nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();

            // Error handling.
            if (result.IsFailure())
            {
                enum nn::pia::util::HandlingType handling = nn::pia::util::GetHandlingType(result);
                switch(handling)
                {
                case nn::pia::util::HANDLING_LEAVE_AND_CLEANUP:
                    PIA_SAMPLE_LOG("Exit MainLoop() with LeaveSession().\n");
                    bCallLeaveSession = true;
                    isExit = true;
                    break;
                case nn::pia::util::HANDLING_RECONNECT:
                    PIA_SAMPLE_LOG("Exit MainLoop().\n");
                    isExit = true;
                    break;
                default:
                    // The GetHandlingType function is capable of returning values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT,
                    // but when the Result objects that can be returned by the Session::CheckConnectionError function are considered,
                    // values other than HANDLING_LEAVE_AND_CLEANUP and HANDLING_RECONNECT are never returned.
                    // An assert statement is provided for that reason.
                    PIA_SAMPLE_ASSERTMSG(false, "Unexpected result.\n");
                    break;
                }
            }
        }
        else
        {
            if (nn::pia::session::Mesh::GetInstance()->CheckConnectionError().IsFailure())
            {
                isExit = true;
                isError = true;
            }
        }


#if PIA_TEST_CODE
        UpdatePiaTestData(isExit);
#endif

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            if (bCallLeaveSession)
            {
                Session_LeaveSession();
            }
            break;
        }
    }

    if (isError)
    {
        return GAME_MODE_INET_ERROR;
    }
    else
    {
        return gameMode;
    }
}
}  // end of anonymous namespace


// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;

#if NN_PIA_CTR
    nn::os::InitializeAlarmSystem();
#endif

    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = false;
        setting.bRawTouchpad = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetClearColor(PRESET_ARRAY[s_TargetPresetIndex].bgColor);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(FRAMES_PER_SECOND);

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    while (true)
    {
        GameMode gameMode = DecideGameMode();
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        bool logined = false;

        // Initialization process for each Pia module.
        // Perform in this order.
        {
            u64 traceFlag =
                nn::pia::common::TRACE_FLAG_COMMON |
                nn::pia::common::TRACE_FLAG_TRANSPORT |
                nn::pia::common::TRACE_FLAG_TRANSPORT_RELAY |
                //                nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_SEND |
                //                nn::pia::common::TRACE_FLAG_TRANSPORT_PACKET_RECEIVE |
                //                nn::pia::common::TRACE_FLAG_TRANSPORT_PROTOCOLMESSAGE_SEND |
                //                nn::pia::common::TRACE_FLAG_TRANSPORT_PROTOCOLMESSAGE_RECEIVE |
                nn::pia::common::TRACE_FLAG_SESSION |
                nn::pia::common::TRACE_FLAG_SESSION_CONNECT |
                nn::pia::common::TRACE_FLAG_SESSION_UPDATE |
                nn::pia::common::TRACE_FLAG_SESSION_MIGRATION |
                nn::pia::common::TRACE_FLAG_INET |
                //                nn::pia::common::TRACE_FLAG_INET_NAT |
                //                nn::pia::common::TRACE_FLAG_INET_NAT_PROBE |
                //                nn::pia::common::TRACE_FLAG_INET_NAT_DETAIL |
                //                nn::pia::common::TRACE_FLAG_INET_SOCKET_SEND |
                nn::pia::common::TRACE_FLAG_ALWAYS;
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlag, true);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
            case GAME_MODE_INET_AUTOMATCH:
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                pFactory = Inet_GetNetworkFactory();

#if ENABLE_DEBUG_CLIENT
                s_pDebugClient = qNew nn::nex::DebugClient();
                PIA_SAMPLE_ASSERT(s_pDebugClient != NULL);
#endif
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }

            // Use Factory to set up Transport.
            // For Internet communication, use nn::pia::inet::NexNetworkFactory.
            nn::pia::transport::Transport::Setting transportSetting;
            transportSetting.pFactory = pFactory;
            transportSetting.maxStationNum = MAX_STATIONS;
            transportSetting.maxSendThreadBufferNum = MAX_SEND_THREAD_BUFFER_NUM;
            transportSetting.maxReceiveThreadBufferNum = MAX_RECEIVE_THREAD_BUFFER_NUM;
            if (s_EnableLatencyEmulation)
            {
                // Specifies the number of buffers for the latency emulation feature.
                // If you are not using latency emulation, either set these values
                // to zero or avoid calling the Transport::SetDebugSetting function.
                nn::pia::transport::Transport::DebugSetting debugSetting;
                {
                    debugSetting.sendThreadLatencyEmulationPacketNum = 128;
                    debugSetting.receiveThreadLatencyEmulationPacketNum = 128;
                    debugSetting.bPacketLossEmulation = false;
                }
                result = nn::pia::transport::Transport::SetDebugSetting(debugSetting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
            result = Transport_InitializePiaTransport(
                transportSetting,
                0,
                NULL,
                0,
                NULL,
                0,
                0,
                CreateAndInitializeProtocols,
                0);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

            PrintSetting();

            // Sets up the session.
            if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
            {
                nn::pia::session::Session::Setting sessionSetting;
                nn::pia::session::Session::DebugSetting sessionDebugSetting;

                sessionSetting.pNetworkFactory = pFactory;
                sessionSetting.networkTopology = PRESET_ARRAY[s_TargetPresetIndex].topology;

                // When using RelayMesh topology, configure so that the relay emulation feature is enabled for debugging.
                if (sessionSetting.networkTopology == nn::pia::session::NetworkTopology_RelayMesh)
                {
                    sessionDebugSetting.isEnableRelayEmulation = true;
                    nn::pia::session::Session::SetDebugSetting(sessionDebugSetting);
                }
                result = Session_InitializePiaSession(sessionSetting, sessionDebugSetting.isEnableRelayEmulation);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }
            else
            {
                nn::pia::session::Mesh::Setting sessionSetting;
                nn::pia::session::Mesh::DebugSetting sessionDebugSetting;

                sessionSetting.pNetworkFactory = pFactory;
                sessionSetting.networkTopology = PRESET_ARRAY[s_TargetPresetIndex].topology;

                // When using RelayMesh topology, configure so that the relay emulation feature is enabled for debugging.
                if (sessionSetting.networkTopology == nn::pia::session::NetworkTopology_RelayMesh)
                {
                    sessionDebugSetting.isEnableRelayEmulation = true;
                    nn::pia::session::Mesh::SetDebugSetting(sessionDebugSetting);
                }
                result = Session_InitializePiaSession_ForMesh(sessionSetting, sessionDebugSetting.isEnableRelayEmulation);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }

            if (s_EnableLatencyEmulation)
            {
                // Apply the latency emulation settings.
                // Overwrite the value of Current to Buffer.
                SetLatencyEmulation(
                    s_LatencyEmulationCurrent.sendMin, s_LatencyEmulationCurrent.sendMax, s_LatencyEmulationCurrent.receiveMin, s_LatencyEmulationCurrent.receiveMax);
                s_LatencyEmulationBuffer = s_LatencyEmulationCurrent;
            }

            if (PRESET_ARRAY[s_TargetPresetIndex].IsEnableRelay())
            {
                // If the relay feature is enabled for the network topology, apply the settings related to this feature.
                SetRelaySettings(RELAY_SETTING_MAX_NUM_OF_RELAY_ROUTES, RELAY_SETTING_MAX_RTT);
            }
        }

        // Connect to the Internet and log in.
        {
            switch (gameMode)
            {
            case GAME_MODE_INET_AUTOMATCH:
                logined = false;
                if (Inet_InitializeNetwork().IsSuccess())
                {
                    if (Inet_Login().IsSuccess())
                    {
                        logined = true;
#if ENABLE_DEBUG_CLIENT
                        nn::nex::Credentials* pCredentials = Nex_GetNgsFacadeInstance()->GetCredentials();
                        PIA_SAMPLE_ASSERT(pCredentials != NULL);
                        if(!s_pDebugClient->Bind(pCredentials))
                        {
                            PIA_SAMPLE_ASSERTMSG(false, "DebugClient Bind failure.");
                        }

                        // Make it possible to rejoin the same session through random matching.
                        nn::nex::ProtocolCallContext protocolCallContext;
                        s_pDebugClient->SetExcludeJoinedMatchmakeSession(&protocolCallContext, true);
                        PIA_SAMPLE_LOG("SetExcludeJoinedMatchmake is Complete\n");
#endif
                    }
                }
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }


        while (logined)
        {
            bool isStart = LoginLoop();
            if (isStart == false)
            {
                break;
            }

            // Process for establishing a session.
            {
                switch (gameMode)
                {
                case GAME_MODE_INET_AUTOMATCH:
                {
                    SampleSessionAsyncStatus status;
                    // Register the Session event callback function.
                    // Before callback registration, initialize the StationStatus that the callback accesses.
                    for (int i = 0; i < MAX_STATIONS; ++i)
                    {
                        s_StationStatus[i].Startup(nn::pia::STATION_ID_INVALID, nn::pia::STATION_INDEX_INVALID);
                    }
                    if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
                    {
                        nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(sessionCallback);

                        // Start Session.
                        nn::pia::session::Session::StartupSetting startupSetting;
                        startupSetting.bUsingHostMigration = PRESET_ARRAY[s_TargetPresetIndex].bHostMigration;
                        startupSetting.cryptoMode = nn::pia::common::CryptoSetting::MODE_NOTHING;
                        startupSetting.maxSilenceTime = s_SessionMaxSilenceTime;
                        startupSetting.keepAliveSendingInterval = s_SessionKeepAliveSendingInterval;

                        result = Session_StartupSession(startupSetting);
                        if(result == nn::pia::ResultNetworkConnectionIsLost())
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                            break;
                        }
                        else
                        {
                            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
                        }

                        // Configure the relay emulation feature after startup.
                        RelayEmulationSetting();

                        nn::pia::inet::NexCreateSessionSetting createSessionSetting;
                        createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                        createSessionSetting.SetGameMode(MATCHMAKE_GAME_MODE + PRESET_ARRAY[s_TargetPresetIndex].gameModeOption);
                        createSessionSetting.SetOpenParticipation(true);

                        nn::pia::inet::NexSessionSearchCriteria sessionSearchCriteria;
                        sessionSearchCriteria.SetGameMode(MATCHMAKE_GAME_MODE + PRESET_ARRAY[s_TargetPresetIndex].gameModeOption);
                        sessionSearchCriteria.SetMaxParticipants(MAX_STATIONS);
                        sessionSearchCriteria.SetOpenedOnly(true);
                        sessionSearchCriteria.SetVacantOnly(true);

                         status = Session_JoinRandomSession(
                            &createSessionSetting,
                            &sessionSearchCriteria,
                            1,
                            &result);
                        if (status == SAMPLE_SESSION_ASYNC_SUCCESS)
                        {
                            // Asynchronous process succeeded.
                            if (IsHost())
                            {
                                // Became the host as a result of matching.
                                gameMode = GAME_MODE_INET_HOST;
                            }
                            else
                            {
                                // Became a client as a result of matching.
                                gameMode = GAME_MODE_INET_CLIENT;
                            }
                        }
                        else if (status == SAMPLE_SESSION_ASYNC_FAILED)
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                        }
                        else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_SESSION)
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                        }
                        else if (status == SAMPLE_SESSION_ASYNC_DISCONNECTED_NETWORK)
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                        }
                    }
                    else
                    {
                        nn::nex::PrincipalID ownerPID;
                        u32 joinedGatheringId;

                        // Initialize matchmaking session.
                        result = Nex_InitializeMatchmake();
                        PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                        // Matchmaking process.
                        result = Nex_AutoMatchmake(
                            &joinedGatheringId,
                            &ownerPID,
                            MATCHMAKE_GAME_MODE + PRESET_ARRAY[s_TargetPresetIndex].gameModeOption,
                            MAX_STATIONS);
                        if (result.IsFailure())
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                            break;
                        }
                        Nex_SetOwnerPID(ownerPID);

                        // Startup process for the NexFacade class.
                        nn::pia::inet::NexFacade::GetInstance()->Startup();
                        result = nn::pia::inet::NexFacade::GetInstance()->StartNatSessionAsync();
                        if (result.IsFailure())
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                            break;
                        }
                        while (true)
                        {
                            nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();
                            nn::nex::Scheduler::GetInstance()->DispatchAll();
                            nn::pia::common::Scheduler::GetInstance()->Dispatch();

                            // Checks the AC network connection status, game server connection status, and NEX internal errors.
                            if (!Inet_IsConnected() || !Nex_IsNgsConnected() || Nex_CheckInternalError())
                            {
                                result = nn::pia::ResultNetworkConnectionIsLost();
                                break;
                            }

                            if (nn::pia::inet::NexFacade::GetInstance()->IsCompletedStartNatSession())
                            {
                                result = nn::pia::inet::NexFacade::GetInstance()->GetStartNatSessionResult();
                                break;
                            }
                            nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                        }
                        if (result.IsFailure())
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                            break;
                        }
                        nn::pia::inet::NexFacade::GetInstance()->UpdateGatheringID(joinedGatheringId);

                        nn::pia::session::Mesh::GetInstance()->RegisterMeshEventCallback(meshCallback);

                        // Startup process for the Mesh class.
                        nn::pia::session::Mesh::StartupSetting startupSetting;
                        startupSetting.maxSilenceTime = s_SessionMaxSilenceTime;
                        startupSetting.keepAliveSendingInterval = s_SessionKeepAliveSendingInterval;
                        startupSetting.bUsingHostMigration = PRESET_ARRAY[s_TargetPresetIndex].bHostMigration;
                        startupSetting.signatureSetting = *(Nex_GetMatchmakeSessionSignatureSetting());

                        result = Session_StartupMesh(startupSetting);
                        //result = Session_StartupMesh(
                        //    s_SessionMaxSilenceTime,
                        //    s_SessionKeepAliveSendingInterval,
                        //    PRESET_ARRAY[s_TargetPresetIndex].bHostMigration,
                        //    Nex_GetMatchmakeSessionSignatureSetting());
                        if (result.IsFailure())
                        {
                            gameMode = GAME_MODE_INET_ERROR;
                            break;
                        }

                        RelayEmulationSetting();

                        if (ownerPID == Nex_GetNgsFacadeInstance()->GetCredentials()->GetPrincipalID())
                        {
                            // Build mesh.
                            if (!Session_CreateMesh())
                            {
                                gameMode = GAME_MODE_INET_ERROR;
                                break;
                            }

                            result = Nex_OpenParticipation(joinedGatheringId);
                            if (result.IsFailure())
                            {
                                gameMode = GAME_MODE_INET_ERROR;
                            }
                            else
                            {
                                gameMode = GAME_MODE_INET_HOST;
                            }
                        }
                        else
                        {
                            // Gets the StationURL of the session host.
                            nn::nex::qList<nn::nex::StationURL> lstURLs;
                            nn::pia::transport::StationConnectionInfo stationConnectionInfo;

                            result = Nex_GetHostStationURLs(joinedGatheringId, &lstURLs);
                            if (result.IsFailure())
                            {
                                gameMode = GAME_MODE_INET_ERROR;
                                break;
                            }
                            // Converts the <tt>StationURL</tt> to <tt>StationConnectionInfo</tt>.
                            result = nn::pia::inet::NexFacade::ConvertNexStationUrlToStationConnectionInfo(lstURLs, &stationConnectionInfo);
                            if (result.IsFailure())
                            {
                                gameMode = GAME_MODE_INET_ERROR;
                                break;
                            }
                            // Join mesh.
                            if (!Session_JoinMesh(stationConnectionInfo))
                            {
                                gameMode = GAME_MODE_INET_ERROR;
                            }
                            else
                            {
                                gameMode = GAME_MODE_INET_CLIENT;
                            }
                        }
                    }
                }
                break;

                default:
                    PIA_SAMPLE_ASSERT(false);
                    break;
                }
            }

            // The main process.
            switch (gameMode)
            {
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
                gameMode = MainLoop(gameMode);

                // Print the measurement results from the watermark feature.
                if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
                {
                    nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
                }

                break;
            }

            // Session termination process.
            {
                // Finalize the session.
                if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
                {
                    switch (gameMode)
                    {
                    case GAME_MODE_INET_HOST:
                    case GAME_MODE_INET_CLIENT:
                        // Leave the session.
                        Session_LeaveSession();
                        break;
                    case GAME_MODE_INET_ERROR:
                        break;
                    default:
                        PIA_SAMPLE_ASSERT(false);
                        break;
                    }
                    // Clean up the Session class.
                    Session_CleanupSession();
                    nn::pia::session::Session::GetInstance()->UnregisterSessionEventCallback();
                }
                else
                {
                    u32 joinedGatheringId = nn::pia::inet::NexFacade::GetInstance()->GetGatheringID();
                    switch (gameMode)
                    {
                    case GAME_MODE_INET_HOST:
                    case GAME_MODE_INET_CLIENT:
                        if (nn::pia::session::Mesh::GetInstance()->IsHost())
                        {
                            if (PRESET_ARRAY[s_TargetPresetIndex].IsEnableHostMigration())
                            {
                                Session_LeaveMeshWithHostMigration();
                            }
                            else
                            {
                                Session_DestroyMesh();
                            }
                        }
                        else
                        {
                            Session_LeaveMesh();
                        }
                        break;
                    case GAME_MODE_INET_ERROR:
                        break;
                    default:
                        PIA_SAMPLE_ASSERT(false);
                        break;
                    }
                    // Cleanup process for the Mesh class.
                    Session_CleanupMesh();
                    nn::pia::session::Mesh::GetInstance()->UnregisterMeshEventCallback();
                    // Cleanup process for the NexFacade class.
                    nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                    nn::pia::inet::NexFacade::GetInstance()->Cleanup();
                    // Leave the matchmaking session.
                    Nex_LeaveMatchmakeSession(
                        joinedGatheringId,
                        Nex_GetOwnerPID(),
                        PRESET_ARRAY[s_TargetPresetIndex].IsEnableHostMigration());

                    Nex_FinalizeMatchmake();
                }

                gameMode = GAME_MODE_INET_AUTOMATCH;
            }
        }


        // Logs out and disconnects from the Internet.
        {
            switch (gameMode)
            {
            case GAME_MODE_INET_AUTOMATCH:
#if ENABLE_DEBUG_CLIENT
                if (s_pDebugClient)
                {
                    s_pDebugClient->Unbind();
                }
#endif
                Inet_Logout();
                Inet_FinalizeNetwork();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
        }


        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            if (!PRESET_ARRAY[s_TargetPresetIndex].IsUseMesh())
            {
                Session_FinalizePiaSession();
            }
            else
            {
                Session_FinalizePiaSession_ForMesh();
            }
            Transport_FinalizePiaTransport(0, 0, FinalizeAndDestroyProtocols);
            switch (gameMode)
            {
            case GAME_MODE_INET_AUTOMATCH:
#if ENABLE_DEBUG_CLIENT
                if (s_pDebugClient)
                {
                    qDelete s_pDebugClient;
                    s_pDebugClient = NULL;
                }
#endif
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
                break;
            default:
                PIA_SAMPLE_ASSERT(false);
                break;
            }
            Common_FinalizePiaCommon();
        }
    }

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
