﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: main.cpp

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#include "SampleLib/SampleLib.h"

#include "controller.h"

#if NN_PIA_NEX_3
#include <OnlineCore/src/version/nex_version.h>
#else
#include <version/nex_version.h>
#endif

#include <cstdio>

#if NN_PIA_CTR
#define USE_PIA_LOCAL 1
#else
#define USE_PIA_LOCAL 0
#endif

#if USE_PIA_LOCAL
#include <pia/local.h>
#include "Shared_Local_Setup.h"
#endif

#include "Shared_Basic_Setup.h"
#include "Shared_Common_Setup.h"
#include "Shared_Inet_Setup.h"
#include "Shared_Nex_Setup.h"
#include "Shared_Transport_Setup.h"
#include "Shared_Session_Setup.h"
#include "Shared_Sync_Setup.h"

#if NN_PIA_CTR
#include <nn/version.h>
#elif NN_PIA_CAFE
#include <sdk_ver.h>
#endif

#ifdef PIA_TEST_BUILD
// Code for automatic verification. Ignore this code as a sample.
#define PIA_TEST_CODE 1
#else
#define PIA_TEST_CODE 0
#endif

#define PROGRAM_TITLE "SyncProtocolProfiling"

namespace
{
const u32 CPU_PROFILING_PERIOD = 10;

const size_t MAX_STATIONS = 8;  // Defines the maximum number of stations.

static u32 s_NexDispatchPeriod = 3;
static u32 s_FramePerSecond = 30;

const size_t MEMSIZE = 2 * 1024 * 1024;              // Memory size, in bytes, provided with the nn::pia::common::Initialize function.
PIA_ATTRIBUTE_ALIGN(4) static u8 s_Memory[MEMSIZE];  // Memory provided with the nn::pia::common::Initialize function.

// When not using NetZ, a small size is sufficient.
// Adjust it according to the application specifications.
const u32 SAMPLE_NEX_MEMORY_SIZE = 1 * 1024 * 1024;
const u32 SAMPLE_NEX_EMERGENCY_MEMORY_SIZE = 50 * 1024;

const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

const u32 MATCHMAKE_GAME_MODE_OFFSET = 0;
const u32 MATCHMAKE_GAME_MODE_BASE = 40100;
const u32 MATCHMAKE_GAME_MODE = MATCHMAKE_GAME_MODE_BASE + 100000 * NN_PIA_PLATFORM + 1000000 * PIA_TEST_CODE + MATCHMAKE_GAME_MODE_OFFSET;
#if USE_PIA_LOCAL
const u32 SAMPLE_TITLE_ID = 0xFF321;
const u8 SAMPLE_SUB_ID = 0x00;
const char SAMPLE_PASSPHRASE[] = "aaabbccc0123456789!";
const char SAMPLE_SIGNATURE_KEY_DATA[] = "syncprotocol key";
#endif

const bool ENABLE_HOST_MIGRATION = false;  //!<  Enables or disables the host migration feature. It is disabled permanently in this demo.

static ControllerReader* s_pControllerReader;
static ControllerAccessor s_ControllerAccessor[MAX_STATIONS];

static bit32 s_MyUsingDataIdBitmap;
static u32 s_Delay;
static u32 s_SendPeriod;
static int s_CursorConfigId;

const bit32 USING_DATA_ID_BITMAP_ALL = (0x1 << DATA_ID_NUM) - 1;
const u32 DELAY_MAX = 20;

static s32 s_StartProfileFrame = -1;
static u32 s_LastProfiledFrame = 0;

nn::pia::StationId s_StationIdList[MAX_STATIONS];

enum ConfigId
{
    CONFIG_DELAY = DATA_ID_NUM,
    CONFIG_SENDPERIOD,
    CONFIG_NUM
};

// Indicates the mode that is executing.
// <tt>inet</tt> starts with AUTOMATCH and then, after matching, decides whether to be the host or a client.
enum GameMode
{
#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_HOST,
    GAME_MODE_LOCAL_CLIENT,
#endif
    GAME_MODE_INET_AUTOMATCH,
    GAME_MODE_INET_HOST,
    GAME_MODE_INET_CLIENT,

#if USE_PIA_LOCAL
    GAME_MODE_LOCAL_ERROR,
#endif
    GAME_MODE_INET_ERROR,
    GAME_MODE_EXIT
};

#if NN_PIA_RELEASE && (NN_PIA_CTR || NN_PIA_CAFE)
void PrintCallback(const char* pStr, s32 len)
{
#if NN_PIA_CTR
    nn::dbg::detail::PutString(pStr, len);
#elif NN_PIA_CAFE
    OSReport(pStr);
#endif
}
#endif

nn::pia::StationIndex ConvertStationIdToStationIndex(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex;
    nn::Result result = nn::pia::transport::Transport::GetInstance()->ConvertToStationIndex(&stationIndex, stationId);
    if (result.IsFailure())
    {
        stationIndex = nn::pia::STATION_INDEX_INVALID;
    }
    return stationIndex;
}

void ClearStationIdList()
{
    for (int i = 0; i < MAX_STATIONS; ++i)
    {
        s_StationIdList[i] = nn::pia::STATION_ID_INVALID;
    }
}

void RegisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(s_StationIdList[stationIndex] == nn::pia::STATION_ID_INVALID);
    s_StationIdList[stationIndex] = stationId;
}

void UnregisterStationId(nn::pia::StationId stationId)
{
    nn::pia::StationIndex stationIndex = ConvertStationIdToStationIndex(stationId);
    PIA_SAMPLE_ASSERT(stationIndex < MAX_STATIONS);
    PIA_SAMPLE_ASSERT(s_StationIdList[stationIndex] != nn::pia::STATION_ID_INVALID);
    s_StationIdList[stationIndex] = nn::pia::STATION_ID_INVALID;
}

// Declarations of callback functions.
void SampleSessionCallback(nn::pia::session::Session::SessionEventType eventType, nn::pia::StationId stationId)
{
    switch (eventType)
    {
    case nn::pia::session::Session::EVENT_JOIN:
    {
        PIA_SAMPLE_LOG("Sample Event JOIN detect. StationId(0x%016llx)\n", stationId.GetValue());
        RegisterStationId(stationId);
    }
    break;
    case nn::pia::session::Session::EVENT_LEAVE:
    {
        PIA_SAMPLE_LOG("Sample Event LEAVE detect. StationId(0x%016llx)\n", stationId.GetValue());
        UnregisterStationId(stationId);
    }
    break;
    }
}

// Determines the mode to execute.
GameMode DecideGameMode()
{
#if USE_PIA_LOCAL
    PIA_SAMPLE_PRINT("Please select game mode.\n [A]LocalHost\n [B]LocalClient\n [X]InetAutomatch\n [START]Quit\n");
#else
    PIA_SAMPLE_PRINT("Please select game mode.\n [X]InetAutomatch\n [START]Quit\n");
#endif
    PIA_SAMPLE_PRINT(" Waiting input\n");

    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    GameMode gameMode;
    bool isDecided = false;
    while (true)
    {
        pSampleFramework->BeginFrame();

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_X))
        {
            PIA_SAMPLE_PRINT("GameMode: InetAutomatch\n");
            gameMode = GAME_MODE_INET_AUTOMATCH;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
        {
            PIA_SAMPLE_PRINT("Exit\n");
            gameMode = GAME_MODE_EXIT;
            isDecided = true;
        }
#if USE_PIA_LOCAL
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalHost\n");
            gameMode = GAME_MODE_LOCAL_HOST;
            isDecided = true;
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_B))
        {
            PIA_SAMPLE_PRINT("GameMode: LocalClient\n");
            gameMode = GAME_MODE_LOCAL_CLIENT;
            isDecided = true;
        }
#endif
#if PIA_TEST_CODE
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_Y))
        {
            if (s_FramePerSecond == 30)
            {
                s_FramePerSecond = 60;
                s_NexDispatchPeriod = 6;
            }
            else
            {
                s_FramePerSecond = 30;
                s_NexDispatchPeriod = 3;
            }
            pSampleFramework->SetFramePerSecond(s_FramePerSecond);
            PIA_SAMPLE_PRINT("FreamePerSecond: %d\n", s_FramePerSecond);
        }
#endif

        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        pSampleFramework->DrawText(10, 20, "Select game mode.");
#if USE_PIA_LOCAL
        pSampleFramework->DrawText(10, 30, "A: LocalHost");
        pSampleFramework->DrawText(10, 40, "B: LocalClient");
        pSampleFramework->DrawText(10, 50, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 60, "START: Quit this application");
#else
        pSampleFramework->DrawText(10, 30, "X: InetAutomatch");
        pSampleFramework->DrawText(10, 40, "START: Quit this application");
#endif

#if PIA_TEST_CODE
        pSampleFramework->DrawText(10, 80, "%d fps", s_FramePerSecond);

        pSampleFramework->SetProgramStatus(isDecided ? 0 : 1, false);
        pSampleFramework->SetProgramParam(0xff, s_FramePerSecond);
#endif

        pSampleFramework->EndFrame();

        if (isDecided)
        {
            return gameMode;
        }
    }
}


// Displays the profiling settings (static).
void PrintStaticSetting(const nn::pia::sync::SyncProtocol::Setting& setting)
{
    PIA_SAMPLE_PRINT("******** Profiling Setting ********\n");

    PIA_SAMPLE_PRINT("Application : " PROGRAM_TITLE "\n");

#if NN_PIA_DEBUG
    PIA_SAMPLE_PRINT("Build : Debug\n");
#elif NN_PIA_DEVELOP
    PIA_SAMPLE_PRINT("Build : Development\n");
#elif NN_PIA_RELEASE
    PIA_SAMPLE_PRINT("Build : Release\n");
#endif

#if NN_PIA_CTR
    PIA_SAMPLE_PRINT("SDK Version : %s\n", NN_CURRENT_SDK_VERSION);
    PIA_SAMPLE_PRINT("Firmware Version : %s\n", NN_CURRENT_FIRMWARE_VERSION);
#elif NN_PIA_CAFE
    PIA_SAMPLE_PRINT("SDK Version : %s\n", CAFE_OS_SDK_VERSION_STRING);
#endif
    PIA_SAMPLE_PRINT("Nex Version : %s\n", NEX_VERSION_STRING);

    PIA_SAMPLE_PRINT("Library : Pia\n");

#if NN_PIA_CTR
    nn::fnd::DateTime dateTime = nn::fnd::DateTime::GetNow();
    PIA_SAMPLE_PRINT("Date : %d/%02d/%02d %02d:%02d:%02d\n",
                     dateTime.GetYear(),
                     dateTime.GetMonth(),
                     dateTime.GetDay(),
                     dateTime.GetHour(),
                     dateTime.GetMinute(),
                     dateTime.GetSecond());
#elif NN_PIA_CAFE
    OSCalendarTime calenderTime;
    OSTicksToCalendarTime(OSGetTime(), &calenderTime);
    PIA_SAMPLE_PRINT("Date : %d/%02d/%02d %02d:%02d:%02d\n",
                     calenderTime.year,
                     calenderTime.mon + 1,
                     calenderTime.mday,
                     calenderTime.hour,
                     calenderTime.min,
                     calenderTime.sec);
#endif

#if NN_PIA_CTR
    nn::ac::ApType apType;
    nn::ac::Initialize();
    nn::Result r = nn::ac::GetConnectingApType(&apType);
    const char* apTypeText = "Unknown";
    if (r.IsSuccess())
    {
        switch (apType)
        {
        case nn::ac::CTR::AP_TYPE_NONE:
            apTypeText = "None";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_1:
            apTypeText = "UserSetting1";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_2:
            apTypeText = "UserSetting2";
            break;
        case nn::ac::CTR::AP_TYPE_USER_SETTING_3:
            apTypeText = "UserSetting3";
            break;
        case nn::ac::CTR::AP_TYPE_NINTENDO_WIFI_USB_CONNECTOR:
            apTypeText = "USBConnector";
            break;
        case nn::ac::CTR::AP_TYPE_NINTENDO_ZONE:
            apTypeText = "NintendoZone";
            break;
        case nn::ac::CTR::AP_TYPE_WIFI_STATION:
            apTypeText = "WiFiStation";
            break;
        case nn::ac::CTR::AP_TYPE_HOTSPOT:
            apTypeText = "HotSpot";
            break;
        };
    }
    PIA_SAMPLE_PRINT("AP type : %s\n", apTypeText);
#endif

#if NN_PIA_CTR
    nn::cfg::Initialize();
    PIA_SAMPLE_PRINT("TransferableId : %016llx\n", nn::cfg::CTR::GetTransferableId(0xFF000));
#endif

    PIA_SAMPLE_PRINT("Frame per second : %d\n", s_FramePerSecond);
    PIA_SAMPLE_PRINT("Dispatch timeout : %d ms\n", 0);

    PIA_SAMPLE_PRINT("Matchmake participants max : %d\n", MAX_STATIONS);

    PIA_SAMPLE_PRINT("NEX thread mode : %d\n", nn::nex::Core::GetThreadMode());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadSleepSpan : %d ms\n", nn::nex::RootTransport::GetTransportBufferThreadSleepSpan());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadSendPriority : %d\n", nn::nex::RootTransport::GetTransportBufferThreadSendPriority());
    PIA_SAMPLE_PRINT("NEX TransportBufferThreadRecvPriority : %d\n", nn::nex::RootTransport::GetTransportBufferThreadRecvPriority());
    PIA_SAMPLE_PRINT("NEX dispatch period in p2p session: %d\n", s_NexDispatchPeriod);

    nn::pia::transport::ThreadStreamManager* pMgr = nn::pia::transport::ThreadStreamManager::GetInstance();
    if (pMgr)
    {
        PIA_SAMPLE_PRINT("Send thread priority : %d\n", pMgr->GetSendThreadPriority());
        PIA_SAMPLE_PRINT("Send thread sleep span : %d ms\n", pMgr->GetSendThreadSleepTimeSpan());
        PIA_SAMPLE_PRINT("Receive thread priority : %d\n", pMgr->GetReceiveThreadPriority());
        PIA_SAMPLE_PRINT("Receive thread sleep span : %d ms\n", pMgr->GetReceiveThreadSleepTimeSpan());
    }

    PIA_SAMPLE_PRINT("SyncProtocol::Setting::m_DataUnitSize :");
    for (u32 i = 0; i < nn::pia::sync::SyncProtocol::DATA_ID_NUM; ++i)
    {
        PIA_SAMPLE_PRINT(" %d", setting.m_DataUnitSize[i]);
    }
    PIA_SAMPLE_PRINT("\n");
    PIA_SAMPLE_PRINT("SyncProtocol::Setting::m_MaxDelay : %d\n", setting.m_MaxDelay);
    PIA_SAMPLE_PRINT("SyncProtocol::Setting::m_TimeoutFrame : %d\n", setting.m_TimeoutFrame);

    PIA_SAMPLE_PRINT("***********************************\n");
}


// Displays the profiling settings (dynamic).
void PrintDynamicSetting(GameMode gameMode)
{

    PIA_SAMPLE_PRINT("******** Profiling Setting ********\n");

    switch (gameMode)
    {
#if USE_PIA_LOCAL
    case GAME_MODE_LOCAL_HOST:
    case GAME_MODE_LOCAL_CLIENT:
        PIA_SAMPLE_PRINT("Game mode : local\n");
        PIA_SAMPLE_PRINT("channel %d\n", nn::pia::local::LocalNetwork::GetInstance()->GetNetworkChannel());
        break;
#endif
    case GAME_MODE_INET_HOST:
    case GAME_MODE_INET_CLIENT:
        PIA_SAMPLE_PRINT("Game mode : inet\n");
        break;
    }

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    PIA_SAMPLE_PRINT("Game participants num : %d\n", pSession->GetStationNum());
    PIA_SAMPLE_PRINT("Station : %s\n", pSession->GetHostStationId() == pSession->GetLocalStationId() ? "Host" : "Client");
    PIA_SAMPLE_PRINT("Index : %d\n", ConvertStationIdToStationIndex(pSession->GetLocalStationId()));

    PIA_SAMPLE_PRINT("FPS : %d\n", s_FramePerSecond);

    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();
    PIA_SAMPLE_PRINT("UsingDataIdBitmap :");
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        PIA_SAMPLE_PRINT(" %02x", pSyncProtocol->GetUsingDataIdBitmap(static_cast<nn::pia::StationId>(i)));
    }
    PIA_SAMPLE_PRINT("\n");

    PIA_SAMPLE_PRINT("Delay : %d\n", pSyncProtocol->GetDelay());
    PIA_SAMPLE_PRINT("SendPeriod : %d\n", pSyncProtocol->GetSendPeriod());

    PIA_SAMPLE_PRINT("***********************************\n");
}


#if NN_PIA_CAFE
// Clears the data cache.
void ClearDataCache()
{
    const u32 DC_SIZE_MAIN = 2 * 1024 * 1024;
    const u32 DC_SIZE_SUB = 512 * 1024;
    u32 range = (OSGetCoreId() == 1) ? DC_SIZE_MAIN : DC_SIZE_SUB;
    // Memory for clearing cache.
    PIA_ATTRIBUTE_ALIGN(32) static u8 s_DataCacheClearBuffer[DC_SIZE_MAIN];
    DCZeroRange(static_cast<void*>(s_DataCacheClearBuffer), range);
}
#endif


// Operation by key input.
bool Operate(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    nn::Result r;

    if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_START))
    {
        // Exit the GameLoop function if START is pressed.
        PIA_SAMPLE_PRINT("Exit GameLoop\n");
        return true;
    }

    switch (pSyncProtocol->GetState())
    {
    case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_A))
        {
            // The A Button starts synchronous communication.
            r = pSyncProtocol->Start(s_MyUsingDataIdBitmap, s_Delay, s_SendPeriod);
            PIA_SAMPLE_PRINT("Start %s\n", r.IsSuccess() ? "Success" : "Failure");
        }

        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_UP))
        {
            --s_CursorConfigId;
            if (s_CursorConfigId < 0)
            {
                s_CursorConfigId = CONFIG_NUM - 1;
            }
        }
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_DOWN))
        {
            ++s_CursorConfigId;
            if (s_CursorConfigId >= CONFIG_NUM)
            {
                s_CursorConfigId = 0;
            }
        }

        if (s_CursorConfigId < DATA_ID_NUM)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT) || pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                s_MyUsingDataIdBitmap ^= (0x1 << s_CursorConfigId);
            }
        }
        else if (s_CursorConfigId == CONFIG_DELAY)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                if (s_Delay > 1)
                {
                    --s_Delay;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                if (s_Delay < DELAY_MAX)
                {
                    ++s_Delay;
                }
            }
        }
        else if (s_CursorConfigId == CONFIG_SENDPERIOD)
        {
            if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_LEFT))
            {
                if (s_SendPeriod > 1)
                {
                    --s_SendPeriod;
                }
            }
            else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_RIGHT))
            {
                ++s_SendPeriod;
            }
        }

        break;

    case nn::pia::sync::SyncProtocol::STATE_WAITING:
    case nn::pia::sync::SyncProtocol::STATE_STARTING:
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            // The R Button ends synchronous communication.
            r = pSyncProtocol->End();
            PIA_SAMPLE_PRINT("End %s\n", r.IsSuccess() ? "Success" : "Failure");
        }
        break;

    case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
        if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_R))
        {
            // The R Button ends synchronous communication.
            r = pSyncProtocol->End();
            PIA_SAMPLE_PRINT("End %s\n", r.IsSuccess() ? "Success" : "Failure");
        }
        else if (pSampleFramework->IsTrigger(nn::pia::sampleframework::BUTTON_L))
        {
            // The L Button starts CPU profiling.
            PrintDynamicSetting(gameMode);
            if (nn::pia::common::CachedPrint::GetInstance() != NULL)
            {
                nn::pia::common::CachedPrint::GetInstance()->Flush();
            }
            s_StartProfileFrame = nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount();
            s_LastProfiledFrame = 0;
            Profiler::GetInstance()->StartProfilingTarget(CPU_PROFILING_PERIOD);
        }
        break;
    }

    return false;
}


static u32 sDispatchCount = 0;
static s32 sDispatchTime = 0;

// Draw on the screen.
void Draw(GameMode gameMode)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();
    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();
    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    static const nn::pia::sampleframework::Color cDefaultColor(1.0f, 1.0f, 1.0f, 1.0f);
    static const nn::pia::sampleframework::Color cInvalidColor(0.4f, 0.4f, 0.4f, 1.0f);

    pSampleFramework->SetFontColor(cDefaultColor);

    {
        pSampleFramework->DrawText(10, 10, PROGRAM_TITLE);
        const char* cGameModeText = NULL;
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_HOST:
            cGameModeText = "LocalHost";
            break;
        case GAME_MODE_LOCAL_CLIENT:
            cGameModeText = "LocalClient";
            break;
#endif
        case GAME_MODE_INET_AUTOMATCH:
            cGameModeText = "InetAutomatch";
            break;
        case GAME_MODE_INET_HOST:
            cGameModeText = "InetHost";
            break;
        case GAME_MODE_INET_CLIENT:
            cGameModeText = "InetClient";
            break;
        }
        PIA_SAMPLE_ASSERT(cGameModeText != NULL);
        pSampleFramework->DrawText(10, 20, "GameMode:%s", cGameModeText);
        pSampleFramework->DrawText(180, 20, "StationNum:%d", pSession->GetStationNum());
        pSampleFramework->DrawText(290, 20, "Index:%d", ConvertStationIdToStationIndex(pSession->GetLocalStationId()));
    }

    {
        u32 idx = (pSampleFramework->GetFrameCount() / 4) % 4;
        const char cAnimLetter[] = { '-', '\\', '|', '/' };
        pSampleFramework->DrawText(380, 10, "%c", cAnimLetter[idx]);
    }

    {
        // Displays the state of SyncProtocol.
        const char* cpState = "";
        switch (pSyncProtocol->GetState())
        {
        case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
            cpState = "NOT_SYNCHRONIZED";
            break;
        case nn::pia::sync::SyncProtocol::STATE_WAITING:
            cpState = "WAITING";
            break;
        case nn::pia::sync::SyncProtocol::STATE_STARTING:
            cpState = "STARTING";
            break;
        case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
            cpState = "SYNCHRONIZED";
            break;
        case nn::pia::sync::SyncProtocol::STATE_ENDING:
            cpState = "ENDING";
            break;
        }
        pSampleFramework->DrawText(10, 30, "State:%s", cpState);

        if (pSyncProtocol->GetLastEndReason() != nn::pia::sync::SyncProtocol::END_REASON_NOTHING)
        {
            const char* cpEndReason = "";
            switch (pSyncProtocol->GetLastEndReason())
            {
            case nn::pia::sync::SyncProtocol::END_REASON_END:
                cpEndReason = "END";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_OTHER_STATION:
                cpEndReason = "OTHER_STATION";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_CHANGE_CONNECTION:
                cpEndReason = "CHANGE_CONNECTION";
                break;
            case nn::pia::sync::SyncProtocol::END_REASON_TIMEOUT:
                cpEndReason = "TIMEOUT";
                break;
            }
            pSampleFramework->DrawText(200, 30, "EndReason:%s", cpEndReason);
        }
    }

    {
        // Displays the possible operations for the given state.
        switch (pSyncProtocol->GetState())
        {
        case nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED:
            pSampleFramework->DrawText(10, 40, "A:Start START:Exit");
            break;

        case nn::pia::sync::SyncProtocol::STATE_WAITING:
        case nn::pia::sync::SyncProtocol::STATE_STARTING:
        case nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED:
            pSampleFramework->DrawText(10, 40, "R:End START:Exit");
            break;

        case nn::pia::sync::SyncProtocol::STATE_ENDING:
            pSampleFramework->DrawText(10, 40, "START:Exit");
            break;
        }
    }

    if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_NOT_SYNCHRONIZED)
    {
        pSampleFramework->DrawText(10, 60, "Key");
        pSampleFramework->DrawText(10, 70, "Touch");
        pSampleFramework->DrawText(10, 80, "Stick");
        pSampleFramework->DrawText(10, 90, "Delay");
        pSampleFramework->DrawText(10, 100, "SendPeriod");

        for (s32 i = 0; i < CONFIG_NUM; ++i)
        {
            s32 y = 60 + i * 10;
            if (i < CONFIG_DELAY)
            {
                pSampleFramework->DrawText(100, y, "%c %s", (i == s_CursorConfigId) ? '>' : ' ', ((0x1 << i) & s_MyUsingDataIdBitmap) ? "ON" : "OFF");
            }
            else if (i == CONFIG_DELAY)
            {
                pSampleFramework->DrawText(100, y, "%c %d", (i == s_CursorConfigId) ? '>' : ' ', s_Delay);
            }
            else if (i == CONFIG_SENDPERIOD)
            {
                pSampleFramework->DrawText(100, y, "%c %d", (i == s_CursorConfigId) ? '>' : ' ', s_SendPeriod);
            }
        }
    }
    else
    {

        {
            // Shows the column titles.
            pSampleFramework->DrawText(5, 70, "UP");
            pSampleFramework->DrawText(5, 80, "DOWN");
            pSampleFramework->DrawText(5, 90, "LEFT");
            pSampleFramework->DrawText(5, 100, "RIGHT");
            pSampleFramework->DrawText(5, 110, "A");
            pSampleFramework->DrawText(5, 120, "B");
            pSampleFramework->DrawText(5, 130, "X");
            pSampleFramework->DrawText(5, 140, "Y");
            pSampleFramework->DrawText(5, 155, "TouchX");
            pSampleFramework->DrawText(5, 165, "TouchY");
            pSampleFramework->DrawText(5, 180, "StickX");
            pSampleFramework->DrawText(5, 190, "StickY");
        }

        {
            // Displays the synchronized controller information of each station.
            for (s32 i = 0; i < MAX_STATIONS; ++i)
            {
                nn::pia::StationId stationId = s_StationIdList[i];

                s32 x = 55 + i * 43;

                pSampleFramework->SetFontColor(pSyncProtocol->CheckEntry(stationId) ? cDefaultColor : cInvalidColor);
                pSampleFramework->DrawText(x, 60, " [%dP]", i + 1);

                if (pSyncProtocol->CheckEntry(stationId) == false)
                {
                    continue;
                }

                bit32 usingDataIdBitmap = pSyncProtocol->GetUsingDataIdBitmap(stationId);

                if (usingDataIdBitmap & (0x1 << DATA_ID_KEY))
                {
                    for (s32 j = 0; j < KEY_NUM; ++j)
                    {
                        s32 y = 70 + j * 10;
                        Key key = static_cast<Key>(j);
                        pSampleFramework->DrawText(x, y, " %c%c%c", s_ControllerAccessor[i].IsTrigger(key) ? 'T' : '-', s_ControllerAccessor[i].IsHold(key) ? 'H' : '-', s_ControllerAccessor[i].IsRelease(key) ? 'R' : '-');
                    }
                }

                if (usingDataIdBitmap & (0x1 << DATA_ID_TOUCH))
                {
                    if (s_ControllerAccessor[i].IsTouch())
                    {
                        pSampleFramework->DrawText(x, 155, "%4d", s_ControllerAccessor[i].GetTouchX());
                        pSampleFramework->DrawText(x, 165, "%4d", s_ControllerAccessor[i].GetTouchY());
                    }
                    else
                    {
                        pSampleFramework->DrawText(x, 155, " ---");
                        pSampleFramework->DrawText(x, 165, " ---");
                    }
                }

                if (usingDataIdBitmap & (0x1 << DATA_ID_STICK))
                {
                    pSampleFramework->DrawText(x, 180, "%5.2f", s_ControllerAccessor[i].GetStickX());
                    pSampleFramework->DrawText(x, 190, "%5.2f", s_ControllerAccessor[i].GetStickY());
                }
            }

            pSampleFramework->SetFontColor(cDefaultColor);
        }

        {
            // Display the delay.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(180, 30, "Delay:%d", pSyncProtocol->GetDelay());
            }

            // Display the send interval.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(290, 30, "SendPeriod:%d", pSyncProtocol->GetSendPeriod());
            }

            // Display the frame number.
            if (pSyncProtocol->GetState() == nn::pia::sync::SyncProtocol::STATE_SYNCHRONIZED)
            {
                pSampleFramework->DrawText(260, 10, "Frame: %d", pSyncProtocol->GetFrameNo());
            }

            // To check on process slowdowns, display the number of dispatch attempts and time required.
            pSampleFramework->DrawText(10, 220, "Dispatch: %3d (%4dms)", sDispatchCount, sDispatchTime);
        }
    }
}


// This is a GameLoop for making comparisons when the communication is not synchronous.
void DummyLoop()
{
    s_pControllerReader->Clear();
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        s_ControllerAccessor[i].Clear();
    }

    bool isProfiling = false;

    while (true)
    {
        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

#if NN_PIA_CAFE
        // Puts into a state where no cache hit occurs for the operating core data cache.
        ClearDataCache();
#endif

        sDispatchCount = 1;
        nn::pia::common::Time start;
        start.SetNow();

        s_pControllerReader->Read();

        nn::pia::common::Time end;
        end.SetNow();
        sDispatchTime = (end - start).GetMSec();

        if (isProfiling)
        {
            if (Profiler::GetInstance()->IsCounting() == false)
            {
                nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
                break;
            }
        }
        else
        {
            PIA_SAMPLE_PRINT("start profiling dummy loop\n");
            isProfiling = true;
            Profiler::GetInstance()->StartProfilingBase(CPU_PROFILING_PERIOD);
        }

        if (Profiler::GetInstance()->IsCounting())
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, "profiling ...");
        }
        else
        {
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();
    }
}


#if PIA_TEST_CODE
void UpdatePiaTestData(GameMode gameMode, bool isExit)
{
    nn::pia::sampleframework::SampleFramework* pSampleFramework = nn::pia::sampleframework::SampleFramework::GetInstance();

    nn::pia::session::Session* pSession = nn::pia::session::Session::GetInstance();

    bool isInet = false;
    switch (gameMode)
    {
    case GAME_MODE_INET_HOST:
    case GAME_MODE_INET_CLIENT:
    case GAME_MODE_INET_AUTOMATCH:
        isInet = true;
        break;
    }

    pSampleFramework->SetProgramStatus(
        isExit ? 0 : (isInet ? 4 : 3),
        pSession->GetStatus() == nn::pia::session::Session::STATUS_CONNECTED_SESSION,
        pSession->GetLocalStationId() == pSession->GetHostStationId(),
        pSession->GetStationNum());


    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    // Session.
    pSampleFramework->SetProgramParam(0x00, pSession->GetLocalStationIndex());

    // Overall SyncProtocol.
    pSampleFramework->SetProgramParam(0x04, pSyncProtocol->GetState());
    pSampleFramework->SetProgramParam(0x05, pSyncProtocol->GetLastEndReason());
    pSampleFramework->SetProgramParam(0x06, pSyncProtocol->GetFrameNo());
    pSampleFramework->SetProgramParam(0x07, pSyncProtocol->GetDelay());

    // Local station settings.
    pSampleFramework->SetProgramParam(0x08, s_MyUsingDataIdBitmap);
    pSampleFramework->SetProgramParam(0x09, s_CursorConfigId);
    pSampleFramework->SetProgramParam(0x0b, s_Delay);

    // Profile.
    pSampleFramework->SetProgramParam(0x0c, s_LastProfiledFrame);
    pSampleFramework->SetProgramParam(0x0d, (s_StartProfileFrame >= 0) ? 1 : 0);

    // Each individual station.
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        nn::pia::StationId stationId = s_StationIdList[i];
        u32 idOfs = (i + 1) << 4;

        pSampleFramework->SetProgramParam(idOfs + 0x0, pSyncProtocol->CheckEntry(stationId) ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x1, pSyncProtocol->GetUsingDataIdBitmap(stationId));

        pSampleFramework->SetProgramParam(idOfs + 0x4, s_ControllerAccessor[i].GetHoldFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x5, s_ControllerAccessor[i].GetTriggerFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x6, s_ControllerAccessor[i].GetReleaseFlag());
        pSampleFramework->SetProgramParam(idOfs + 0x7, s_ControllerAccessor[i].IsTouch() ? 1 : 0);
        pSampleFramework->SetProgramParam(idOfs + 0x8, s_ControllerAccessor[i].GetTouchX());
        pSampleFramework->SetProgramParam(idOfs + 0x9, s_ControllerAccessor[i].GetTouchY());
        pSampleFramework->SetProgramParam(idOfs + 0xa, static_cast<s32>(s_ControllerAccessor[i].GetStickX() * 1000.0f));
        pSampleFramework->SetProgramParam(idOfs + 0xb, static_cast<s32>(s_ControllerAccessor[i].GetStickY() * 1000.0f));
    }
}
#endif

// Main loop during communication.
void MainLoop(GameMode gameMode, bool* pIsLeaveNeeded)
{
    nn::Result r;

    s_pControllerReader->Clear();
    for (u32 i = 0; i < MAX_STATIONS; ++i)
    {
        s_ControllerAccessor[i].Clear();
    }
    s_LastProfiledFrame = 0;

    nn::pia::sync::SyncProtocol* pSyncProtocol = Sync_GetSyncProtocol();

    while (true)
    {
        bool isExit = false;

        nn::pia::sampleframework::SampleFramework::GetInstance()->BeginFrame();

#if NN_PIA_CAFE
        // Puts into a state where no cache hit occurs for the operating core data cache.
        ClearDataCache();
#endif

        sDispatchCount = 1;
        nn::pia::common::Time start;
        start.SetNow();

        // For GAME_MODE_INET_XXX, nn::nex::Scheduler is created, and that dispatch must be performed.
        if (nn::nex::Scheduler::GetInstance() != NULL)
        {
            if (nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount() % s_NexDispatchPeriod == 0)
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();
            }
        }

        // Call the nn::pia::common::Scheduler::Dispatch function at regular intervals.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (pSyncProtocol->NeedSetData())
        {
            // Get the data to send and sets it to SyncProtocol.
            s_pControllerReader->Read();
            for (u32 i = 0; i < DATA_ID_NUM; ++i)
            {
                DataId dataId = static_cast<DataId>(i);
                if (pSyncProtocol->NeedSetData(dataId))
                {
                    r = pSyncProtocol->SetData(dataId, s_pControllerReader->GetData(dataId));
                    PIA_SAMPLE_ASSERT(r.IsSuccess());
                }
            }
        }

        // Advance one frame.
        r = pSyncProtocol->Step();
        if (r.IsFailure())
        {
            if (r == nn::pia::ResultDataIsNotArrivedYet())
            {
                // Because the data has yet to arrive, repeatedly call Dispatch until it arrives.

                while (true)
                {
                    nn::Result connectionStatus;

                    if (gameMode == GAME_MODE_INET_HOST || gameMode == GAME_MODE_INET_CLIENT)
                    {
                        if (Inet_IsConnected() == false)
                        {
                            // Ends because the AC has disconnected.
                            break;
                        }
                    }

                    if (nn::pia::session::Session::GetInstance()->CheckConnectionError().IsFailure())
                    {
                        // Ends because the communication has been disconnected.
                        break;
                    }

                    nn::pia::sampleframework::SampleFramework::GetInstance()->SleepCurrentThreadMilliseconds(1000 / s_FramePerSecond);
                    ++sDispatchCount;
                    if (nn::nex::Scheduler::GetInstance() != NULL)
                    {
                        nn::nex::Scheduler::GetInstance()->DispatchAll();
                    }

                    nn::pia::common::Scheduler::GetInstance()->Dispatch();

                    // The frame has not advanced since the data was set, so it is not necessary to set it here.
                    PIA_SAMPLE_ASSERT(pSyncProtocol->NeedSetData() == false);

                    r = pSyncProtocol->Step();
                    if (r != nn::pia::ResultDataIsNotArrivedYet())
                    {
                        break;
                    }
                }
            }
            else if (r == nn::pia::ResultDataIsNotSet())
            {
                // The data is set previously, so this result does not occur.
                PIA_SAMPLE_ASSERT(false);
            }
            else
            {
                // Other errors are ignored here.
            }
        }

        // Get the received data. The local station also uses the data obtained here.
        for (u32 i = 0; i < MAX_STATIONS; ++i)
        {
            nn::pia::StationId stationId = s_StationIdList[i];
            s_ControllerAccessor[i].Clear();

            if (pSyncProtocol->CanGetData() && pSyncProtocol->CheckEntry(stationId))
            {
                bit32 usingDataIdBitmap = pSyncProtocol->GetUsingDataIdBitmap(stationId);
                for (u32 j = 0; j < DATA_ID_NUM; ++j)
                {
                    DataId dataId = static_cast<DataId>(j);
                    if ((0x1 << dataId) & usingDataIdBitmap)
                    {
                        r = pSyncProtocol->GetData(stationId, dataId, s_ControllerAccessor[i].GetDataPtr(dataId));
                        PIA_SAMPLE_ASSERT(r.IsSuccess());
                        if (dataId == DATA_ID_KEY)
                        {
                            PIA_SAMPLE_ASSERT(*static_cast<const u32*>(pSyncProtocol->GetDataPtr(stationId, dataId)) == *static_cast<u32*>(s_ControllerAccessor[i].GetDataPtr(dataId)));
                        }
                    }
                }
            }
        }

        nn::pia::common::Time end;
        end.SetNow();
        sDispatchTime = (end - start).GetMSec();

        // Calls nn::pia::common::Scheduler::Dispatch one more time to allow what was sent to the protocol to be processed immediately.
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        if (gameMode == GAME_MODE_INET_HOST || gameMode == GAME_MODE_INET_CLIENT)
        {
            if (Inet_IsConnected() == false)
            {
                // Ends because the AC has disconnected.
                isExit = true;
                *pIsLeaveNeeded = false;
            }
        }

        nn::Result result = nn::pia::session::Session::GetInstance()->CheckConnectionError();
        if (result.IsFailure())
        {
            // Ends because the communication has been disconnected.
            isExit = true;
            if (result == nn::pia::ResultSessionConnectionIsLost())
            {
                *pIsLeaveNeeded = true;
            }
            else if (result == nn::pia::ResultNetworkConnectionIsLost())
            {
                *pIsLeaveNeeded = false;
            }
        }

        if (Profiler::GetInstance()->IsCounting())
        {
            nn::pia::sampleframework::SampleFramework::GetInstance()->DrawText(10, 10, "profiling ...");
        }
        else
        {
            if (s_StartProfileFrame >= 0)
            {
                s_LastProfiledFrame = nn::pia::sampleframework::SampleFramework::GetInstance()->GetFrameCount() - s_StartProfileFrame;
                PIA_SAMPLE_PRINT("profile frame : %d\n", s_LastProfiledFrame);
                s_StartProfileFrame = -1;
            }

            if (isExit == false)
            {
                // If currently communicating, operations are received.
                isExit = Operate(gameMode);
                if (isExit)
                {
                    *pIsLeaveNeeded = true;
                }
            }

            Draw(gameMode);
#if PIA_TEST_CODE
            UpdatePiaTestData(gameMode, isExit);
#endif
        }

        nn::pia::sampleframework::SampleFramework::GetInstance()->EndFrame();

        if (isExit)
        {
            return;
        }
    }
}


}  // end of anonymous namespace


// Entry point.
PIA_SAMPLE_MAIN_FUNC()
{
    nn::Result result;

#if NN_PIA_CTR
    nn::os::InitializeAlarmSystem();
#endif

    {
        nn::pia::sampleframework::SampleFramework::Setting setting;
        setting.bUpperScreenOnly = true;
        setting.bRawTouchpad = true;
        nn::pia::sampleframework::SampleFramework::CreateInstance(setting);
    }
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFontColor(nn::pia::sampleframework::COLOR_WHITE);
    nn::pia::sampleframework::SampleFramework::GetInstance()->SetFramePerSecond(s_FramePerSecond);

    PIA_SAMPLE_PRINT(PROGRAM_TITLE "\n");
    Basic_SetProgramTitle(PROGRAM_TITLE);

    // Controller-related initialization is conducted by the <tt>SampleFramework::CreateInstance</tt> function,
    // so create a <tt>ControllerReader</tt> instance after that has been done.
    ControllerReader controllerReader;
    s_pControllerReader = &controllerReader;

    Profiler::CreateInstance();

    PIA_SAMPLE_PRINT("start profiling idle\n");
    Profiler::GetInstance()->StartProfilingIdle(CPU_PROFILING_PERIOD);

    DummyLoop();

    while (true)
    {
        ClearStationIdList();

        GameMode gameMode = DecideGameMode();
        if (gameMode == GAME_MODE_EXIT)
        {
            break;
        }

        // Initialization process for each Pia module.
        // Perform in this order.
        {
            u64 traceFlags = 0;
            result = Common_InitializePiaCommon(s_Memory, MEMSIZE, traceFlags, true);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
#if NN_PIA_RELEASE && (NN_PIA_CTR || NN_PIA_CAFE)
            nn::pia::common::RegisterInternalPrintCallback(PrintCallback);
#endif

            nn::pia::transport::NetworkFactory* pFactory = NULL;
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            {
                result = Local_InitializePiaLocal(SAMPLE_TITLE_ID, SAMPLE_SUB_ID);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Local_GetNetworkFactory();
            }
            break;
#endif
            case GAME_MODE_INET_AUTOMATCH:
            {
                // Initialize NEX before initializing PiaInet.
                result = Nex_InitializeNex(
                    SAMPLE_NEX_MEMORY_SIZE,
                    SAMPLE_NEX_EMERGENCY_MEMORY_SIZE,
                    SAMPLE_NEX_THREAD_MODE);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                result = Inet_InitializePiaInet();
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

                pFactory = Inet_GetNetworkFactory();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Use Factory to set up Transport.
            // Uses nn::pia::local::UdsNetworkFactory for local communications, and uses nn::pia::inet::NexNetworkFactory for Internet communications.
            {
                nn::pia::transport::Transport::Setting setting;
                setting.pFactory = pFactory;
                setting.maxStationNum = MAX_STATIONS;
                setting.maxSendThreadBufferNum = 16;
                setting.maxReceiveThreadBufferNum = 32;

                result = Transport_InitializePiaTransport(setting);
                PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            }

            // Sets up sync.
            {
                nn::pia::sync::SyncProtocol::Setting setting;
                setting.m_MaxDelay = DELAY_MAX;
                setting.m_TimeoutFrame = 240;
                setting.m_DataUnitSize[DATA_ID_KEY] = sizeof(KeyData);
                setting.m_DataUnitSize[DATA_ID_TOUCH] = sizeof(TouchData);
                setting.m_DataUnitSize[DATA_ID_STICK] = sizeof(StickData);
                Sync_InitializePiaSync(setting);
                PrintStaticSetting(setting);
            }

            // Sets up the session.
            result = Session_InitializePiaSession(
                pFactory,
                nn::pia::session::NetworkTopology_FullMesh);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
            result = nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback(SampleSessionCallback);
            PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
        }


        // Connect to the network connections and create a session.
        {
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Create a session.
                // Settings for the session being created.
                nn::pia::local::UdsCreateSessionSetting createSessionSetting;
                {
                    nn::pia::local::UdsCreateNetworkSetting createNetworkSetting;
                    {
                        createNetworkSetting.m_SubId = SAMPLE_SUB_ID;
                        createNetworkSetting.m_LocalCommunicationId = nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID);
                        createNetworkSetting.m_MaxEntry = MAX_STATIONS;
                        memcpy(&createNetworkSetting.m_Passphrase, SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                        createNetworkSetting.m_PassphraseLength = sizeof(SAMPLE_PASSPHRASE);
                    }
                    createSessionSetting.SetLocalCreateNetworkSetting(createNetworkSetting);
                    createSessionSetting.SetMaxParticipants(MAX_STATIONS);
                    result = createSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    PIA_SAMPLE_ASSERT(result.IsSuccess());
                }

                SampleSessionAsyncStatus asyncStatus = Session_CreateSession(&createSessionSetting, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_CreateSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;

            case GAME_MODE_LOCAL_CLIENT:
            {
                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Searches for sessions.
                // Search condition settings
                const size_t searchSettingNum = 1;
                nn::pia::local::LocalSessionSearchCriteria searchCriteriaList[searchSettingNum];
                {
                    searchCriteriaList[0].SetMaxParticipants(MAX_STATIONS);
                    searchCriteriaList[0].SetSubId(SAMPLE_SUB_ID);
                    searchCriteriaList[0].SetLocalCommunicationId(nn::pia::local::LocalNetwork::GetInstance()->CreateLocalCommunicationId(SAMPLE_TITLE_ID));
                    searchCriteriaList[0].SetResultRange(0, 8);
                }
                SampleSessionAsyncStatus asyncStatus = Session_BrowseSession(&searchCriteriaList[0], &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_BrowseSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }

                // Joins a discovered session.
                size_t foundSessionNum = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->GetSize();
                if (foundSessionNum > 0)
                {
                    PIA_SAMPLE_LOG("session found\n");
                    nn::pia::session::ISessionInfoList::Iterator it = nn::pia::session::Session::GetInstance()->GetBrowsedSessionInfoList()->Begin();
                    nn::pia::session::ISessionInfo* pTargetSessionInfo = *it;
                    PIA_SAMPLE_ASSERT(PIA_IS_VALID_POINTER(pTargetSessionInfo));

                    nn::pia::local::UdsJoinSessionSetting joinSessionSetting;
                    joinSessionSetting.SetSessionInfoPtr(pTargetSessionInfo);
                    joinSessionSetting.SetPassphrase(SAMPLE_PASSPHRASE, sizeof(SAMPLE_PASSPHRASE));
                    joinSessionSetting.SetSignatureKey(SAMPLE_SIGNATURE_KEY_DATA, sizeof(SAMPLE_SIGNATURE_KEY_DATA));
                    asyncStatus = Session_JoinSession(&joinSessionSetting, &result);
                    if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                    {
                        PIA_SAMPLE_LOG("Session_JoinSession: failed.\n");
                        PIA_SAMPLE_LOG_RESULT(result);
                        gameMode = GAME_MODE_LOCAL_ERROR;
                        break;
                    }
                }
                else
                {
                    PIA_SAMPLE_LOG("no session found\n");
                    gameMode = GAME_MODE_LOCAL_ERROR;
                    break;
                }
            }
            break;
#endif  // USE_PIA_LOCAL

            case GAME_MODE_INET_AUTOMATCH:
            {
                // Connect to the Internet
                result = Inet_InitializeNetwork();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_InitializeNetwork: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Log in to the game server.
                result = Inet_Login();
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Inet_Login: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Start Session.
                nn::pia::session::Session::StartupSetting startupSetting;
                startupSetting.bUsingHostMigration = ENABLE_HOST_MIGRATION;

                result = Session_StartupSession(startupSetting);
                if (result.IsFailure())
                {
                    PIA_SAMPLE_LOG("Session_StartupSession failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                // Match automatically.
                nn::pia::inet::NexCreateSessionSetting setting;
                {
                    setting.SetGameMode(MATCHMAKE_GAME_MODE);
                    setting.SetMaxParticipants(MAX_STATIONS);
                    setting.SetOpenParticipation(true);
                }

                const size_t searchSettingNum = 1;
                nn::pia::inet::NexSessionSearchCriteria criteria[searchSettingNum];
                {
                    criteria[0].SetGameMode(MATCHMAKE_GAME_MODE);
                    criteria[0].SetMaxParticipants(MAX_STATIONS);
                }
                SampleSessionAsyncStatus asyncStatus = Session_JoinRandomSession(&setting, criteria, searchSettingNum, &result);
                if (asyncStatus != SAMPLE_SESSION_ASYNC_SUCCESS)
                {
                    PIA_SAMPLE_LOG("Session_JoinRandomSession: failed.\n");
                    PIA_SAMPLE_LOG_RESULT(result);
                    gameMode = GAME_MODE_INET_ERROR;
                    break;
                }

                if (nn::pia::session::Session::GetInstance()->IsHost())
                {
                    // Became the host as a result of matching.
                    gameMode = GAME_MODE_INET_HOST;
                }
                else
                {
                    // Became a client as a result of matching.
                    gameMode = GAME_MODE_INET_CLIENT;
                }
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        s_MyUsingDataIdBitmap = USING_DATA_ID_BITMAP_ALL;
        s_Delay = 5;
        s_SendPeriod = s_FramePerSecond / 30;
        s_CursorConfigId = 0;

        bool isLeaveNeeded = false;

        // The main process.
        switch (gameMode)
        {
#if USE_PIA_LOCAL
        case GAME_MODE_LOCAL_HOST:
        case GAME_MODE_LOCAL_CLIENT:
#endif
        case GAME_MODE_INET_HOST:
        case GAME_MODE_INET_CLIENT:
        {
            MainLoop(gameMode, &isLeaveNeeded);

            // Print the measurement results from the watermark feature.
            if (PIA_IS_VALID_POINTER(nn::pia::common::WatermarkManager::GetInstance()))
            {
                nn::pia::common::WatermarkManager::GetInstance()->Trace(nn::pia::common::TRACE_FLAG_ALWAYS);
            }
        }
        break;
        }

        // End the session and disconnect from the network.
        {
            // Finalize the session.
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            {
                if (isLeaveNeeded)
                {
                    // Leave the session.
                    // In this sample, the return value is ignored because finalization will take place later.
                    Session_LeaveSession();
                }
            }
            break;
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_ERROR:
#endif
            case GAME_MODE_INET_ERROR:
                break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }

            // Clean up the Session class.
            Session_CleanupSession();

            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
            }
            break;
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_Logout();
                Inet_FinalizeNetwork();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
        }

        // Finalize each Pia module.
        // Finalize in reverse order of initialization.
        {
            Session_FinalizePiaSession();
            Sync_FinalizePiaSync();
            Transport_FinalizePiaTransport();
            switch (gameMode)
            {
#if USE_PIA_LOCAL
            case GAME_MODE_LOCAL_HOST:
            case GAME_MODE_LOCAL_CLIENT:
            case GAME_MODE_LOCAL_ERROR:
            {
                Local_FinalizePiaLocal();
            }
            break;
#endif
            case GAME_MODE_INET_HOST:
            case GAME_MODE_INET_CLIENT:
            case GAME_MODE_INET_ERROR:
            {
                Inet_FinalizePiaInet();
                Nex_FinalizeNex();
            }
            break;
            default:
            {
                PIA_SAMPLE_ASSERT(false);
            }
            break;
            }
            Common_FinalizePiaCommon();
        }
    }

    Profiler::DestroyInstance();

    nn::pia::sampleframework::SampleFramework::DestroyInstance();

    PIA_SAMPLE_LOG("end.\n");
}
