﻿/*---------------------------------------------------------------------------*
  Project: Pia
  File: session_Session.h

  Copyright Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#pragma once

#include <pia/session/session_definitions.h>

#include <pia/common/common_CryptoSetting.h>

#include <pia/transport.h>

#include <pia/session/session_Mesh.h>


// Forward declarations (common).
namespace nn
{
namespace pia
{
namespace common
{
class CallContext;
typedef CallContext AsyncContext;
}
}
}

namespace nn
{
namespace pia
{
namespace transport
{
class NetworkFactory;
class StationConnectionInfo;
}
}
}

namespace nn
{
namespace pia
{
namespace session
{

// Forward declarations.
class CreateSessionSetting;
class JoinSessionSetting;
class SessionStatusCheckJob;
class CreateSessionJob;
class AutoMatchmakeJob;
class BrowseMatchmakeJob;
class JoinSessionJob;
class DestroySessionJob;
class LeaveSessionJob;
class OpenParticipationJob;
//class OpenSessionJob;
class CloseParticipationJob;
//class CloseSessionJob;
class ConfigParticipationJobBase;
class GenerateMatchmakeSystemPasswordJob;
class ClearMatchmakeSystemPasswordJob;
class JointSessionJob;
class ModifyAttributeJob;
class UpdateSessionSettingJob;
class UpdateApplicationDataJob;

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
class RefereeStartGameJob;
class RefereeEndGameJob;
class RefereeEndGameWithoutReportJob;
class RefereeGetOrCreateStatsJob;
class IMatchmakeReferee;
class IEndGamePersonalReport;
#endif

class StationIdStatusTable;
class SessionProtocol;
class IMatchmakeSession;
class ISessionInfoList;
class MeshLayerController;
class MatchNetworkFactory;
class SessionSearchCriteria;
class ISessionInfo;
class ISessionInfoList;
class MeshEventListenerForSession;
class UpdateSessionSetting;

/*!
  @class  nn::pia::session::Session
  @brief  This class contains processes for matchmaking and for creating and joining sessions used for P2P communication.

  @details  This class implements the singleton pattern. Only a single instance can be created.

            Some of the internal processes of the <tt>Session</tt> class are registered as jobs in <tt>@ref nn::pia::common::Scheduler</tt>
            so you must make periodic calls to <tt>@ref nn::pia::common::Scheduler::Dispatch</tt>

  @date 2015-06-30  Removed the session merge feature.
  @date 2015-05-20  Added the <tt>Session::GetUplinkBitRate</tt> function.
  @date 2015-05-20  Added the <tt>bitRateCheckMode</tt> member to the <tt>Setting</tt> structure to enable configuration of the bit rate detection feature.
  @date 2015-05-20  Added the <tt>uplinkBitRateLowerLimit</tt> and <tt>bitRateCheckPacketSize</tt> members to the <tt>StartupSetting</tt> structure to enable configuration of the bit rate detection feature.
  @date 2015-01-20  Added the <tt>GetMatchmakeSessionParticipantNum()</tt>, <tt>CancelJoinSessionAsync()</tt>, and <tt>CancelJoinRandomSessionAsync()</tt> functions.
  @date 2015-01-20  Added a description of the specification changes to <tt>CloseParticipationAsync</tt>.
  @date 2015-01-20  Noted that <tt>ResultGameServerProcessAborted</tt> is sometimes returned in the results of asynchronous processes related to the sessions feature.
  @date 2015-01-07  Added the <tt>GetJoinSequence()</tt> and <tt>GetJoinMeshSequence()</tt> functions.
  @date 2014-11-05  Deleted <tt>STATUS_CONNECTED_NETWORK</tt> from the return values for the <tt>GetStatus()</tt> function.
  @date 2014-11-05  Changed the <tt>CheckConnection()</tt> function to <tt>CheckConnectionError()</tt>.
  @date 2014-11-05  Added <tt>EVENT_JOINT_SESSION_HOST_CHANGED</tt>.
  @date 2014-11-05  Added a feature for changing the search conditions for joint sessions.
  @date 2014-11-05  Added the <tt>GetJointSessionStationId()</tt> function.
  @date 2014-10-07  Added functions and constants related to joint session features that had been removed.
  @date 2014-09-18  Changed the specifications so that <tt>ResultAlreadyExists</tt> is returned if you call <tt>CreateInstance()</tt> when an instance already exists.
  @date 2014-08-11  Removed functions and constants related to the joint session feature, and added functions and constants related to the session merge feature.
  @date 2014-02-27  Initial version.
*/
class Session : public nn::pia::common::RootObject
{
    friend class SessionStatusCheckJob;
    friend class CreateSessionJob;
    friend class AutoMatchmakeJob;
    friend class JoinSessionJob;
    friend class JointSessionJob;
    friend class DestroySessionJob;
    friend class LeaveSessionJob;
    friend class OpenParticipationJob;
    friend class OpenSessionJob;
    friend class CloseParticipationJob;
    friend class CloseSessionJob;
    friend class SessionProtocol;
    friend class StationIdStatusTable;
    friend class MeshEventListenerForSession;
    friend class MeshLayerController;
    friend class UpdateApplicationDataJob;
    friend class ModifyAttributeJob;
    friend class GenerateMatchmakeSystemPasswordJob;
    friend class ClearMatchmakeSystemPasswordJob;

public:
/*!
        @brief  Stores parameters passed to the <tt>@ref nn::pia::session::Session::CreateInstance()</tt> function.
        @see  CreateInstance
*/
    struct Setting
    {
        Setting(void)
          : pNetworkFactory(NULL),
            networkTopology(NetworkTopology_FullMesh),
            browsedSessionInfoListNum(0)
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
            ,bitRateCheckMode(BitRateCheck_Disable)
#endif
        {
        }

        transport::NetworkFactory* pNetworkFactory;  //!<  Pointer to a <tt>NetworkFactory</tt> specified for each network.
        session::NetworkTopology networkTopology;    //!<  Specifies the network topology setting for the session.
        u8 browsedSessionInfoListNum;                //!<  Size of the list that stores the session search results.
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        session::BitRateCheckMode bitRateCheckMode;  //!<  Settings for the bitrate detection feature.
#endif
    };

/*!
    @brief  (For debugging.) This structure stores the debug feature settings that are passed to <tt>@ref nn::pia::session::Session::SetDebugSetting</tt>.
*/
    struct DebugSetting
    {
        DebugSetting(bool argIsEnableRelayEmulation = false, bool argIsEnableMatchmakeAnalysis = false, bool argIsDisableCheckMaxSilenceTime = false)
          : isEnableRelayEmulation(argIsEnableRelayEmulation),
            isEnableMatchmakeAnalysis(argIsEnableMatchmakeAnalysis),
            isDisableCheckMaxSilenceTime(argIsDisableCheckMaxSilenceTime)
        {
        }
        bool isEnableRelayEmulation;  //!<  Settings for the connection route emulation feature for debugging.
        bool isEnableMatchmakeAnalysis; //!<  Settings for the matchmaking analysis feature.
        bool isDisableCheckMaxSilenceTime; //!<  Does not check the range set for the maximum time without communication.
    };

/*!
    @brief  This structure stores settings passed to <tt>@ref nn::pia::session::Session::Startup</tt>.
*/
    struct StartupSetting
    {
        StartupSetting()
            : bUsingHostMigration(true),
            pPlayerName(NULL),
            cryptoMode(common::CryptoSetting::MODE_NOTHING),
            maxSilenceTime(nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT),
            keepAliveSendingInterval(nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT),
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
            uplinkBitRateLowerLimit(nn::pia::session::NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT),
            bitRateCheckPacketSize(0),
            bitRateMeasuringSpan(1000),
            isSkipBitRateCheck(false),
#endif
            pToken(NULL)
        {
        }
        bool bUsingHostMigration;                                           //!<  Sets the use of the host migration feature. The default is <tt>true</tt> (enabled).
        const nn::pia::transport::Station::PlayerName* pPlayerName;         //!<  Specifies the name and language code associated with the local station. Other members of the mesh access this information.
        nn::pia::common::CryptoSetting::Mode cryptoMode;                    //!<  Specifies the encryption setting mode. Defaults to <tt>common::CryptoSetting::MODE_NOTHING</tt>.
        u32 maxSilenceTime;                                                 //!<  Specifies the maximum allowable time without communication (in milliseconds). The default value is <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT</tt>. The maximum value that can be set is <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_MAX</tt>, the minimum value is  <tt>session::NN_PIA_SESSION_MAX_SILENCE_TIME_MIN</tt>.
        u32 keepAliveSendingInterval;                                       //!<  Specifies the send interval for keep-alive (in milliseconds). The default value is <tt>session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</tt>.
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        s32 uplinkBitRateLowerLimit;    //!<  Specifies the minimum rate required by the application when using the bitrate detection feature. The value is in bits per second. The default value is <tt>session::NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT</tt>.
        size_t bitRateCheckPacketSize;  //!<  Specifies the IP packet size used for testing when using the bitrate detection feature. The value is in bytes. The expected value is the average size of IP packet sent and received by the application. It must be a multiple of 4. The default is <tt>0</tt>.
        s32 bitRateMeasuringSpan;       //!<  Sets the time to spend measuring when the bit rate detection feature is used. The value is in milliseconds. The default is <tt>1000 msec</tt>. Changing this value from the default is not recommended.
        bool isSkipBitRateCheck;        //!<  Set to skip detection processes when using bitrate detection. Specify <tt>true</tt> to skip. The default is <tt>false</tt>.
#endif
        const nn::pia::transport::Station::IdentificationToken* pToken;     //!<  Specifies a pointer to the identification token indicating the local device. If <tt>NULL</tt> is specified, an identification token with all data set to <tt>NULL</tt> is used. The default is <tt>NULL</tt>.
    };

/*
        @cond  PRIVATE
        @brief  Enumerates <tt>Session</tt> states.
*/
    enum ModuleState
    {
        INITIALIZED = 0,  //!<  Initialization complete
        STARTUP_COMPLETE  //!<  Ready for matchmaking
    };
    //! @endcond

/*!
        @cond  PRIVATE
        @brief  Enumerates matchmaking states.
*/
    enum MatchmakeStatus
    {
        NOT_IN_SESSION = 0,         //!<  Not connected.
        JOINED_MATCHMAKE,           //!<  Joined a matchmaking session.
        ESTABLISHED_SESSION,        //!<  Joined a session.
        CHANGING_SESSION,           //!<  Being transferred to a session by the joint session feature. (Includes a period with no communication.)
        ESTABLISHED_JOINT_SESSION   //!<  Already joined the joint session.
    };
    //! @endcond

/*!
        @cond  PRIVATE
        @brief  Enumerates the connection status.
*/
    enum ConnectionStatus
    {
        NOT_CONNECTED,         //!<  Not communicating.
        CONNECTED,             //!<  Communicating.
        DISCONNECTED_SESSION,  //!<  Disconnected from the session. After leaving the session with <tt>@ref LeaveSessionAsync</tt> be sure to call <tt>@ref Cleanup</tt>.
        DISCONNECTED_NETWORK   //!<  Disconnected from the network. Log out and then log in again.
    };
    //! @endcond

/*!
        @cond  PRIVATE
        @brief  Enumerates the join-in processes.
*/
    enum ProcessType {
        PROCESS_TYPE_NONE = 0,                  //!<  If no processes are taking place. (Value for initialization.)
        PROCESS_TYPE_MERGE_SESSION,             //!<  Process that takes all of the participants in the current session to join a different session. Specifies the session to join.
        PROCESS_TYPE_MERGE_RANDOM_SESSION,      //!<  Process that takes all of the participants in the current session to join a different session. The session to join is determined by random matchmaking.
        PROCESS_TYPE_DEMERGE_SESSION,           //!<  Process that splits a session and makes the new session independent.
        PROCESS_TYPE_EMIGRATE_SESSION,          //!<  Process that migrates stations to another session. Specifies the destination session.
        PROCESS_TYPE_EMIGRATE_RANDOM_SESSION,   //!<  Process that migrates stations to another session. The destination session is determined by random matchmaking.
        PROCESS_TYPE_CREATE_JOINT_SESSION,      //!<  Process that creates a joint session, and brings all of the participants in the current session into it.
        PROCESS_TYPE_JOIN_JOINT_SESSION,        //!<  Process that specifies a joint session, and brings all of the participants in the current session into it.
        PROCESS_TYPE_JOIN_RANDOM_JOINT_SESSION, //!<  Process that specifies a joint session through random matchmaking and brings all of the participants in the current session into it.
        PROCESS_TYPE_DESTROY_JOINT_SESSION,     //!<  Process that disbands a joint session.
        PROCESS_TYPE_LEAVE_JOINT_SESSION,       //!<  Process that leaves a joint session.
        PROCESS_TYPE_MAX
    };
/* @endcond */

/*!
        @cond  PRIVATE
        @brief  Enumerated type for distinguishing asynchronous process types for leaving a mesh.
*/
    enum LeaveMeshProcessType {
        LEAVE_PROCESS_TYPE_UNKNOWN = 0,
        LEAVE_PROCESS_TYPE_LEAVE,
        LEAVE_PROCESS_TYPE_LEAVE_WITH_HM,
        LEAVE_PROCESS_TYPE_DESTROY
    };
/* @endcond */

/*!
    @brief  Enumerates reasons why a <tt>Session</tt> object cannot communicate.

    @details  This information can be retrieved using the <tt>@ref GetSessionDisconnectReason</tt> function.
*/
    enum DisconnectReason
    {
        UNKNOWN_REASON = 0,    //!<  Specifies that the reason for the disconnection is unknown. One way to get this value is by calling the <tt>Session::Cleanup</tt> function during communication (without destroying or leaving the session).
        NOT_YET_COMMUNICATED,  //!<  Specifies that no communication has been performed. You get this value if you never build or join a session after calling the <tt>Session::Startup</tt> function.
        OPERATION_OF_OWN,      //!<  Specifies that the disconnection was caused by the local station. You get this value when the local station left the session by calling <tt>Session::LeaveSessionAsync</tt> or similar function.
        OPERATION_OF_OTHER,    //!<  Specifies that the disconnection was caused by another station in the session. One way to get this value is when the local station is a client, and the host destroys the session.
        KICKOUT_BY_USER,       //!<  Specifies that the local station was kicked out from the session because of a call to the <tt>Session::KickoutStation</tt> function.
        KICKOUT_BY_SYSTEM,     //!<  Specifies that the disconnection from the session was caused by the library. One way to get this value is when relaying is enabled, and the library disconnects you from the session because it restricts the number of relay requests.
        INCONSISTENT_INFO,     //!<  Specifies that the disconnection was caused by a desync. (The session information was no longer consistent between participants.) One way to get this value is when the station counts do not match.
        MIGRATION_FAIL,        //!<  Specifies that the disconnection was caused by a failed host migration.
        EXTERNAL_FACTOR,       //!<  Specifies that the disconnection was caused by something other than the session. One way to get this value is when the network connection is lost.
        MIGRATION_FATAL_ERROR  //!<  Specifies that a fatal error occurred during host migration. This error can occur when the library detects that a session has broken up.
    };

/*!
        @brief  Enumerates session states.

        @details  This information can be retrieved using the <tt>Session::GetStatus</tt> function.
*/
    enum Status
    {
        STATUS_NOT_CONNECTED,           //!<  Not communicating.
        STATUS_CONNECTED_SESSION,       //!<  Joined a session.
        STATUS_MIGRATING_SESSION,       //!<  Being transferred to a session by the joint session feature. (Includes a period with no communication.)
        STATUS_CONNECTED_JOINT_SESSION, //!<  Joined a joint session.
        STATUS_DISCONNECTED_SESSION,    //!<  Disconnected from the session. Call <tt>@ref Cleanup</tt> after calling <tt>@ref LeaveSessionAsync</tt>.
        STATUS_DISCONNECTED_NETWORK     //!<  Disconnected from the network. Log out and then log in again.
    };

/*!
        @brief  This enumerated type is passed as a parameter to the callback that sends notifications about changes in the session state.

        @details  It represents the possible changes to the session state.
*/
    enum SessionEventType
    {
        EVENT_JOIN = 0,                         //!<  Indicates that a station has joined.
        EVENT_LEAVE,                            //!<  A station has left.
        EVENT_SESSION_HOST_CHANGED,             //!<  Indicates that the host of the session has changed. (Only occurs when host migration is enabled.)
        EVENT_HOST_MIGRATION_FAILED,            //!<  Indicates that host migration failed. (Only occurs when host migration is enabled.)
        EVENT_JOINT_SESSION_HOST_CHANGED,       //!<  Indicates that the host of the joint session has changed. (Only occurs when using the joint session feature with host migration enabled.)
        EVENT_START_SESSION_CREATE_JOINT = 11,  //!<  Indicates that the process of creating a joint session has started. (Only occurs when using the joint session feature.)
        EVENT_END_SESSION_CREATE_JOINT,         //!<  Indicates that the process of creating a joint session has finished. (Only occurs when using the joint session feature.)
        EVENT_START_SESSION_JOIN_JOINT,         //!<  Indicates that the process of joining a joint session has started. (Only occurs when using the joint session feature.)
        EVENT_END_SESSION_JOIN_JOINT,           //!<  Indicates that the process of joining a joint session has finished. (Only occurs when using the joint session feature.)
        EVENT_START_SESSION_JOINT_RANDOM,       //!<  Indicates that a joint session has started through random matchmaking. (Only occurs when using the joint session feature.)
        EVENT_END_SESSION_JOINT_RANDOM,         //!<  Indicates that a joint session started through random matchmaking has been completed. (Only occurs when using the joint session feature.)
        EVENT_START_SESSION_DESTROY_JOINT,      //!<  Indicates that the process of disbanding a joint session has started. (Only occurs when using the joint session feature.)
        EVENT_END_SESSION_DESTROY_JOINT,        //!<  Indicates that the process of disbanding a joint session has finished. (Only occurs when using the joint session feature.)
        EVENT_START_SESSION_LEAVE_JOINT,        //!<  Indicates that the process of leaving a joint session has started. (Only occurs when using the joint session feature.)
        EVENT_END_SESSION_LEAVE_JOINT,          //!<  Indicates that the process of leaving a joint session has finished. (Only occurs when using the joint session feature.)
        EVENT_SESSION_MIGRATION_FAILED,         //!<  Indicates that the session migration process for the joint session feature failed. (Only occurs when using the joint session feature.)
        EVENT_SET_SESSION_SYSTEM_PASSWORD,      //!<  A system password was set on the session.
        EVENT_CLEAR_SESSION_SYSTEM_PASSWORD     //!<  The system password of the session was cleared.
    };

/*!
    @brief  This callback function type is called when the session state changes.

    @details  <b>Parameters</b>
           <ul>
           <li><tt><var>eventType</var></tt>: Enumerated type indicating what the state change was. For more information, see the <tt>@ref SessionEventType</tt> enumerated type.</li>
           <li><tt><var>id</var></tt>: The <tt>StationId</tt> of the station that changed its state.</li>
           </ul>

*/
    typedef void (*SessionEventCallback)(SessionEventType eventType, StationId id);

/*!
    @brief  Declares the type of the callback function for the session host to invoke when determining whether to approve join requests from clients.

    @details  This function defines the callback that is invoked when the local station is the session host and it receives a join request from a client.
             The return value is used to decide whether to approve or deny the join request.

    @param[in] pRequesterToken  Specifies a pointer to the ID token of the station that sent the join request.

    @return  Returns <tt>true</tt> to approve the join request, or <tt>false</tt> otherwise.
*/
    typedef bool (*JoiningApprovalCallback)(const nn::pia::transport::Station::IdentificationToken* pRequesterToken);


    typedef common::ObjList<StationId> TStationIdList;

/*!
     @brief  An iterator for accessing valid <tt>StationId</tt> objects in a session (non-<tt>const</tt> version).
*/
    class StationIdIterator
    {
    public:
        StationIdIterator(TStationIdList::Iterator it) :
            m_Iterator(it)
        {
        }
        StationId& operator*(void)
        {
            return *m_Iterator;
        }
        StationId* operator->(void)
        {
            return &(*m_Iterator);
        }
        bool operator==(const StationIdIterator& rhs) const
        {
            return m_Iterator == rhs.m_Iterator;
        }
        bool operator!=(const StationIdIterator& rhs) const
        {
            return !(*this == rhs);
        }
        StationIdIterator& operator++(void)
        {
            ++m_Iterator;
            return *this;
        }
    private:
        TStationIdList::Iterator m_Iterator;
    };

/*!
     @brief  An iterator for accessing valid <tt>StationId</tt> objects in a session (<tt>const</tt> version).
*/
    class StationIdConstIterator
    {
    public:
        StationIdConstIterator(TStationIdList::ConstIterator it)
          : m_ConstIterator(it)
        {
        }
        const StationId& operator*(void)
        {
            return *m_ConstIterator;
        }
        const StationId* operator->(void)
        {
            return &(*m_ConstIterator);
        }
        bool operator==(const StationIdConstIterator& rhs) const
        {
            return m_ConstIterator == rhs.m_ConstIterator;
        }
        bool operator!=(const StationIdConstIterator& rhs) const
        {
            return !(*this == rhs);
        }
        StationIdConstIterator& operator++(void)
        {
            ++m_ConstIterator;
            return *this;
        }

    private:
        TStationIdList::ConstIterator m_ConstIterator;
    };


/*!
     @name  Singleton Instance Control
     @{
*/

/*!
    @brief  Creates the <tt>Session</tt> class instance (singleton pattern).

    @details  If the relay emulation feature for debugging has been set by the <tt>@ref SetDebugSetting</tt> function,
              and the network topology is not set to <tt>NetworkTopology_RelayMesh</tt>, the call to this function fails.

    @attention  Do not call <tt>@ref SetDebugSetting</tt> in retail ROMs.

    @param[in] setting  Parameters to set when creating the instance.

    @return  Returns a <tt>Result</tt> value that indicates success if the instance is created successfully.

    @retval nn::pia::ResultNotInitialized  Indicates that the <tt>session</tt> module is not initialized. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was created at the wrong time. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultAlreadyExists  Indicates that an instance has already been created. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates a problem in the parameter values or the combination thereof. Programming error. Fix your program so that this error is not returned.
*/
    static nn::Result CreateInstance(Session::Setting setting);


/*!
    @brief  Frees the <tt>Session</tt> class instance (singleton pattern).
*/
    static void DestroyInstance();

/*!
    @brief  Gets the <tt>Session</tt> class instance (singleton pattern).

    @details  Returns a null pointer if the instance has not yet been
    created and initialized using the <tt>@ref CreateInstance</tt> function.

    @return  Returns a pointer to the <tt>Session</tt> instance.
*/
    static Session* GetInstance()
    {
        return s_pInstance;
    }


/*!
    @brief  (For debugging.) Configures the <tt>Session</tt> debugging feature.

    @details  This function can be called while the <tt>Session</tt> instance does not exist. Call this function before creating the <tt>Session</tt> instance with <tt>@ref CreateInstance</tt>.

    If the <tt>isEnableRelayEmulation</tt> member of the <tt>DebugSetting</tt> structure is set to <tt>true</tt>, the <tt>Session</tt> will be configured to use the relay emulation feature for debugging.
    When the session is configured to use the relay emulation feature, the <tt>@ref CreateInstance</tt> function will not succeed unless the network topology is set to <tt>NetworkTopology_RelayMesh</tt>.

    If the <tt>isEnableMatchmakeAnalysis</tt> member of the <tt>DebugSetting</tt> structure is set to <tt>true</tt>, the <tt>Session</tt> will be configured to use the matchmaking analysis feature.

    If this function is not called, all related debugging features are disabled.

    @attention  Do not call this function in retail ROMs.

    @return  Returns the result of the function. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. You must make sure that the implementation of this function in your application does not return any errors.
            If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Indicates that the function cannot be executed in this state. This result is returned when the <tt>Session</tt> instance has already been created. Programming error. Fix your program so that this error is not returned.
*/
    static nn::Result SetDebugSetting(const DebugSetting& debugSetting);

/*!
     end of name Singleton Instance Control
     @}
*/


/*!
     @name  Finish Initialization
     @{
*/

/*!
        @brief  Executes the startup procedure.

        @details  Call this function after the <tt>session::EndSetup()</tt> function has been called and the <tt>Session</tt> initialization process has ended.

        The use of host migration is set by the <tt>bUsingHostMigration</tt> member of the <tt>StartupSetting</tt> structure. Set to <tt>true</tt> to enable. The default is <tt>true</tt>.
        Specify the name and language code associated with the local station in the <tt>pPlayerName</tt> member of <tt>StartupSetting</tt>. The name must be no longer than 16 characters (not including the terminating null character).
    @if CAFE_DOC
        Set the language code to the language used by the application. Use one of the values defined in <tt>nn::fp::Language</tt>.
        The name and language code you specify are shared with all members of the session. When the local station joins a session, all session participants are added to Player History.
        When another client joins the session, the new participant's name and language code are looked up and automatically added to Player History using the
        <tt>nn::fp::AddRecentPlayRecord()</tt> function from within the library.
        For more information about Player History, see the <i>Pia Programming Manual</i> for the Friend Presence library
        and the <i>Pia API Reference</i> for the <tt>nn::fp::AddRecentPlayRecord()</tt> function.
    @endif
    @if CTR_DOC
            Specify a language code if your application determines the languages of other regions. You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage()</tt> function if you want to use the language setting of the system.
    @endif

        You can set the communication encryption mode in the <tt>cryptoMode</tt> member of the <tt>StartupSetting</tt> structure. The default value is <tt>common::CryptoSetting::MODE_NOTHING</tt>.
    @if CTR_DOC
        You do not need to configure encryption for local communication because the data is encrypted in the UDS communication layer of the SDK.
    @endif

        Specify the maximum allowable time without communication in the <tt>maxSilenceTime</tt> member of <tt>StartupSetting</tt>. If the period without communication with the station exceeds the specified period, that station is treated as disconnected.
        Specify the keep-alive send interval in the <tt>keepAliveSendingInterval</tt> member of the <tt>StartupSetting</tt> structure. Keep alive is a feature that keeps connections alive during periods of no communication by having the system send data to stations the application is not sending data to.

        The default values for the <tt>maxSilenceTime</tt> and <tt>keepAliveSendingInterval</tt> members of <tt>StartupSetting</tt> are as follows. (These values are used if you do not modify the <tt>maxSilenceTime</tt> and <tt>keepAliveSendingInterval</tt> members of <tt>StartupSetting</tt>.)
        <ul>
        <li> <tt>maxSilenceTime</tt>: <tt>nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT</tt></li>
        <li> <tt>keepAliveSendingInterval</tt>: <tt>nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</tt></li>
        </ul>
        If you customize these values, make sure to test the actual results as you adjust them.

        In the <tt>pToken</tt> member of the <tt>StartupSetting</tt> structure, you can set an identification token that indicates the local host. The default is <tt>NULL</tt>.
        This identification token is shared with peers during the connection process, and it is associated with and maintained in <tt>@ref nn::pia::transport::Station</tt>.
        You are not required to specify an identification token if you do not need to use one.

        The <tt>@ref Startup</tt> function is the only way to specify and change an identification token that indicates the local host.
        There is no supported method for changing an identification token while a session is in use.

        If the bitrate detection feature is enabled, you must set the <tt>uplinkBitRateLowerLimit</tt> and <tt>bitRateCheckPacketSize</tt> members of the <tt>@ref StartupSetting</tt> structure.
        Both members default to an invalid value. Set the minimum bitrate (in bits per second) required by the application in <tt>uplinkBitRateLowerLimit</tt> and the IP packet size (in bytes) to use for measurements in <tt>bitRateCheckPacketSize</tt>.

        @attention  Do not transition to sleep mode during a call to this function during local communication.

        @param[in] startupSetting  Specifies the settings to use for startup.

        @return  Returns a <tt>Result</tt> value that indicates success if the process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This result is only returned during local communication. Programming error. Fix your program so that this error is not returned.

        @see  Cleanup
*/
    nn::Result Startup(StartupSetting& startupSetting);

    /// @cond PRIVATE
/*!
        @brief  Executes the startup procedure.

        @details  Call this function after the <tt>session::EndSetup()</tt> function has been called and the <tt>Session</tt> initialization process has ended.

        Host migration is configured with the first argument. Set to <tt>true</tt> to enable.
        Specify the name and language code associated with the local station in the second parameter. The name must be no longer than 16 characters (not including the terminating null character).
    @if CAFE_DOC
        Set the language code to the language used by the application. Use one of the values defined in <tt>nn::fp::Language</tt>.
        The name and language code you specify are shared with all members of the session. When the local station joins a session, all session participants are added to Player History.
        When another client joins the session, the new participant's name and language code are looked up and automatically added to Player History using the
        <tt>nn::fp::AddRecentPlayRecord()</tt> function from within the library.
        For more information about Player History, see the <i>Pia Programming Manual</i> for the Friend Presence library
        and the <i>Pia API Reference</i> for the <tt>nn::fp::AddRecentPlayRecord()</tt> function.
    @endif
    @if CTR_DOC
            Specify a language code if your application determines the languages of other regions. You can use the value returned by the <tt>nn::cfg::CTR::GetLanguage()</tt> function if you want to use the language setting of the system.
    @endif

        You can set the communication encryption mode in the third parameter.
    @if CTR_DOC
        You do not need to configure encryption for local communication because the data is encrypted in the UDS communication layer of the SDK.
    @endif

        Specify the maximum allowable time without communication in the fourth parameter. If the period without communication with the station exceeds the specified period, that station is treated as disconnected.
        Specify the keep-alive send interval in the fifth parameter. Keep alive is a feature that keeps connections alive during periods of no communication by having the system send data to stations the application is not sending data to.

        The default values for parameters four and five are listed below. The following defaults are used if you call an overload without configuration arguments.
        <ul>
        <li>Fourth parameter: <tt><var>nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT</var></tt></li>
        <li>Fifth parameter: <tt><var>nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT</var></tt></li>
        </ul>
        If you customize these values, make sure to test the actual results as you adjust them.

        In the fifth parameter, you can set an identification token that indicates the local host.
        This identification token is shared with peers during the connection process, and it is associated with and maintained in <tt>@ref nn::pia::transport::Station</tt>.
        You are not required to specify an identification token if you do not need to use one.

        The <tt>@ref Startup</tt> function is the only way to specify and change an identification token that indicates the local host.
        There is no supported method for changing an identification token while a session is in use.

        @attention  Do not transition to sleep mode during a call to this function during local communication.

        @param[in] bUsingHostMigration  Specifies whether host migration will be used.
        @param[in] pPlayerName  Specifies the name and language code associated with the local station. Other members of the session access this information.
        @param[in] maxSilenceTime  Specifies the maximum allowable time without communication (in milliseconds).
        @param[in] keepAliveSendingInterval  Specifies the send interval for keep-alive (in milliseconds).
        @param[in] cryptoMode  Specifies the encryption mode.
        @param[in] pToken  Specifies a pointer to the identification token indicating the local device. If <tt>NULL</tt> is specified, an identification token with all data set to <tt>NULL</tt> is used.

        @return  Returns a <tt>Result</tt> value that indicates success if the process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This result is only returned during local communication. Programming error. Fix your program so that this error is not returned.

        @see  Cleanup
*/
/*
    nn::Result Startup(
        bool bUsingHostMigration,
        const nn::pia::transport::Station::PlayerName* pPlayerName,
        nn::pia::common::CryptoSetting::Mode cryptoMode = common::CryptoSetting::MODE_NOTHING,
        u32 maxSilenceTime = nn::pia::session::NN_PIA_SESSION_MAX_SILENCE_TIME_DEFAULT,
        u32 keepAliveSendingInterval = nn::pia::session::NN_PIA_SESSION_KEEP_ALIVE_INTERVAL_DEFAULT,
#if NN_PIA_EXPERIMENT_BAND_WIDTH_CHECK
        s32 bandwidthLimit = nn::pia::session::NN_PIA_SESSION_INVALID_UPLINK_BIT_RATE_LOWER_LIMIT,
#endif
        const nn::pia::transport::Station::IdentificationToken* pToken = NULL);
*/
    /// @endcond

/*!
        @brief  Executes the cleanup procedure.

        @details  Call this function after the session ends or an error occurs to restore the <tt>Session</tt> to a usable state.
                    If <tt>@ref CheckConnectionError</tt> has returned <tt>@ref ResultSessionConnectionIsLost</tt>
                    call <tt>@ref LeaveSessionAsync</tt> before calling this function.

        Call <tt>Startup</tt> again if you want to use your <tt>Session</tt> instance again after you have called the <tt>@ref Cleanup</tt> function.

        @see  Startup
*/
    void Cleanup();

/*!
     end of name Finish Initialization
     @}
*/

/*!
     @name  Session Operations
     @{
*/

/*!
    @brief  Creates or joins a session using random matchmaking.
    @details
           This function cannot be used with local communications.
           Starts an asynchronous session creation or joining process after finding a random match. The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedJoinRandomSession</tt> to check whether the process has completed and <tt>@ref GetJoinRandomSessionResult</tt> to check the result.

           Multiple search criteria can be specified. The search starts from the first criteria and works its way down the list until it finds a session to join.
           Up to <tt>@ref inet::SEARCH_CRITERIA_MAX_NUM</tt> search criteria can be specified. If you try to set more criteria than the maximum, <tt>nn::pia::ResultInvalidArgument</tt> is returned.
           Sessions that are closed or full are automatically excluded from the search.
           In addition, the last three sessions that the local station joined as a client are remembered by the server and excluded from the search. This record expires after one hour.

    @param[in] pCreateSessionSetting  Specifies the settings to use when creating a session.
    @param[in] pSearchCriteria  Specifies the list of search criteria to use when searching for sessions.
    @param[in] criteriaNum  Specifies the number of criteria to use when searching for sessions.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedJoinRandomSession, <tt>GetJoinRandomSessionResult</tt>
*/
    nn::Result JoinRandomSessionAsync(CreateSessionSetting* pCreateSessionSetting, SessionSearchCriteria* pSearchCriteria, size_t criteriaNum);

/*!
    @brief  Checks the completion of <tt>JoinRandomSessionAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
              This function checks the completion of asynchronous processes started with <tt>@ref JoinRandomSessionAsync</tt>.
    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  JoinRandomSessionAsync, <tt>GetJoinRandomSessionResult</tt>
*/
    bool IsCompletedJoinRandomSession() const;

/*!
    @brief  Gets the result of <tt>JoinRandomSessionAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
              This function gets the results after an asynchronous process started by <tt>@ref JoinRandomSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultNatCheckFailed  The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultDnsFailed  DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultSessionIsNotFound  The target session was lost or closed during the asynchronous process. Perform cleanup.
    @retval nn::pia::ResultCancelled  The asynchronous process was canceled. Perform cleanup.
    @retval nn::pia::ResultStationConnectionFailed  The connection process failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultStationConnectionNatTraversalFailedUnknown  The NAT traversal connection process failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEim  The NAT traversal connection process failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm  The NAT traversal connection process failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim  The NAT traversal connection process failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEdm  The NAT traversal connection process failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultNatTraversalRequestTimeout  The NAT traversal connection process failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
    @retval nn::pia::ResultJoinRequestDenied  The session host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultMeshIsFull  Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedNoCandidate  The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRttLimit  The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRelayNumLimit  The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedUnknown  The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultInvalidSystemMessage  The process was canceled because an invalid response was received from the session host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultIncompatibleFormat  Indicates that the peer is using an incompatible communication format. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultSessionWrongState  The state of the joined session was irregular. Perform cleanup.

    @see  JoinRandomSessionAsync, IsCompletedJoinRandomSession
*/
    nn::Result GetJoinRandomSessionResult() const;

/*!
    @brief  Cancels <tt>JoinRandomSessionAsync</tt> asynchronous processes.

    @details  This function cancels asynchronous processes started by <tt>JoinRandomSessionAsync</tt>.
                The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
                The library quickly acts to cancel the asynchronous process after this function is called, but check
                using the <tt>@ref IsCompletedJoinRandomSession()</tt> function that the cancellation has completed.

    @return  Returns the result of canceling <tt>JoinRandomSessionAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when no <tt>JoinRandomSessionAsync</tt> asynchronous processing is being performed. Programming error. Fix your program so that this error is not returned.

    @see  JoinRandomSessionAsync, IsCompletedJoinRandomSession, GetJoinRandomSessionResult
*/
    nn::Result CancelJoinRandomSessionAsync();

/*!
    @brief  Searches for sessions.
    @details

           This function starts an asynchronous process that searches for sessions.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedBrowseSession</tt> to check whether the process has completed and <tt>@ref GetBrowseSessionResult</tt> to check the result.
           If the search process succeeds, the list of browsable sessions is updated and can be retrieved by calling <tt>@ref GetBrowsedSessionInfoList</tt>.

    @param[in] pSearchCriteria  Specifies the search criteria to use when searching for sessions.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedBrowseSession, GetBrowseSessionResult, GetBrowsedSessionInfoList
*/
    nn::Result BrowseSessionAsync(SessionSearchCriteria* pSearchCriteria);

/*!
    @brief  Checks the completion of <tt>BrowseSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref BrowseSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  BrowseSessionAsync, GetBrowseSessionResult, GetBrowsedSessionInfoList
*/
    bool IsCompletedBrowseSession() const;

/*!
    @brief  Gets the result of <tt>BrowseSessionAsync</tt> asynchronous processing.
    @details  This function gets the results after an asynchronous process started by <tt>@ref BrowseSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultCancelled  The asynchronous process that was running was canceled. Perform cleanup.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @if CTR_DOC
    @retval nn::pia::ResultErrorOccurred  An internal error occurred. This result is only returned during local communication. Perform cleanup.
    @endif

    @see  BrowseSessionAsync, IsCompletedBrowseSession, GetBrowsedSessionInfoList
*/
    nn::Result GetBrowseSessionResult() const;

/*!
    @brief  Gets the list of sessions that were found by the <tt>BrowseSessionAsync</tt> asynchronous process.
    @details  This function gets the session list after an asynchronous process started by <tt>@ref BrowseSessionAsync</tt> ends and the result from <tt>@ref GetBrowseSessionResult</tt> indicates success.

    @return  Returns a pointer to the list of sessions that were found.

    @see  BrowseSessionAsync, IsCompletedBrowseSession, GetBrowseSessionResult
*/
    session::ISessionInfoList* GetBrowsedSessionInfoList();

/*!
    @brief  Creates a session.
    @details

           This function starts an asynchronous process that creates a session.
           The local station becomes the host of the created session.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedCreateSession</tt> to check whether the process has completed and <tt>@ref GetCreateSessionResult</tt> to check the result.

    @param[in] pCreateSessionSetting  Specifies the settings to use when creating the session.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedCreateSession, GetCreateSessionResult
*/
    nn::Result CreateSessionAsync(CreateSessionSetting* pCreateSessionSetting);

/*!
    @brief  Checks the completion of <tt>CreateSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref CreateSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CreateSessionAsync, GetCreateSessionResult
*/
    bool IsCompletedCreateSession() const;

/*!
        @brief  Gets the result of <tt>CreateSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref CreateSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultInvalidState  Indicates an internal state that prevents the asynchronous process from advancing. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultDnsFailed  DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
        @retval nn::pia::ResultNatCheckFailed  The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
        @retval nn::pia::ResultCancelled  The asynchronous process was canceled. Perform cleanup.
        @retval nn::pia::ResultTimeOut  Timed out. This result is only returned during local communication. Perform cleanup.
        @retval nn::pia::ResultErrorOccurred  An internal error occurred. This result is only returned during local communication. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultSessionWrongState  The state of the joined session was irregular. Perform cleanup.

        @see  CreateSessionAsync, IsCompletedCreateSession
*/
    nn::Result GetCreateSessionResult() const;

    // Session joining.
/*!
    @brief  Joins a session.
    @details

           This function starts an asynchronous process that joins a session.

           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedJoinSession</tt> to check whether the process has completed and <tt>@ref GetJoinSessionResult</tt> to check the result.

    @param pJoinSessionSetting  Specifies the settings of the session to join.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedJoinSession, GetJoinSessionResult
*/
    nn::Result JoinSessionAsync(const JoinSessionSetting* pJoinSessionSetting);

/*!
    @brief  Checks the completion of <tt>JoinSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref JoinSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  JoinSessionAsync, GetJoinSessionResult
*/
    bool IsCompletedJoinSession() const;

/*!
        @brief  Gets the result of <tt>JoinSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref JoinSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultInvalidState  Returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost or closed during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultMatchmakeSessionIsFull  The matchmaking session is full. Perform cleanup.
        @retval nn::pia::ResultDeniedByParticipants  The local station is blacklisted by a session participant. Perform cleanup.
        @retval nn::pia::ResultParticipantInBlackList  A participant on the local station's blacklist is in the session. Perform cleanup.
        @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultNatCheckFailed  The NAT check failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
        @retval nn::pia::ResultDnsFailed  DNS resolution failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
        @retval nn::pia::ResultCancelled  The asynchronous process was canceled. Perform cleanup.
        @retval nn::pia::ResultInvalidState  Returned when the station is not in a connection-ready state. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultStationConnectionFailed  The connection process failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultStationConnectionNatTraversalFailedUnknown  The NAT traversal connection process failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultNatTraversalFailedBothEim  The NAT traversal connection process failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm  The NAT traversal connection process failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim  The NAT traversal connection process failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultNatTraversalFailedBothEdm  The NAT traversal connection process failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultNatTraversalRequestTimeout  The NAT traversal connection process failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors. Perform cleanup.
        @retval nn::pia::ResultJoinRequestDenied  The session host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultMeshIsFull  Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultRelayFailedNoCandidate  The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultRelayFailedRttLimit  The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultRelayFailedRelayNumLimit  The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultRelayFailedUnknown  The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultInvalidSystemMessage  The process was canceled because an invalid response was received from the session host. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultSessionUserPasswordUnmatch  The user password specified when joining does not match the user password set on the session. Perform cleanup.
        @retval nn::pia::ResultSessionSystemPasswordUnmatch  The system password specified when joining does not match the system password set on the session. Perform cleanup.
        @retval nn::pia::ResultSessionIsClosed  Attempted to join a session that is closed. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
        @retval nn::pia::ResultHostIsNotFriend  Attempted to join a session with a host that is not a friend. Perform cleanup.
        @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
        @retval nn::pia::ResultIncompatibleFormat  Indicates that the peer is using an incompatible communication format. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultSessionWrongState  The state of the joined session was irregular. Perform cleanup.
        @retval nn::pia::ResultTimeOut  Timed out.
                                                             @if CTR_DOC
                                                             It could also be returned when the passphrase set with <tt>@ref nn::pia::local::UdsJoinSessionSetting</tt> is incorrect.
                                                             This result is only returned during local communication.
                                                             @endif Perform cleanup.
        @retval nn::pia::ResultDifferentVersion  Attempted to join a session with a different local network version or application version. This result is only returned during local communication. Perform cleanup.
        @retval nn::pia::ResultErrorOccurred  An internal error occurred. This result is only returned during local communication. Perform cleanup.

        @see  JoinSessionAsync, IsCompletedJoinSession
*/
    nn::Result GetJoinSessionResult() const;

/*!
    @brief  Cancels <tt>JoinSessionAsync</tt> asynchronous processes.

    @details  This function cancels asynchronous processes started by <tt>JoinSessionAsync</tt>.
                The cancel process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
                The library quickly acts to cancel the asynchronous process after this function is called, but check
                using the <tt>@ref IsCompletedJoinSession()</tt> function that the cancellation has completed.

    @return  Returns the result of canceling <tt>JoinSessionAsync</tt> asynchronous processing. On success, returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval ResultInvalidState  Returned when no <tt>JoinSessionAsync</tt> asynchronous processing is being performed. Programming error. Fix your program so that this error is not returned.

    @see  JoinSessionAsync, IsCompletedJoinSession, GetJoinSessionResult
*/
    nn::Result CancelJoinSessionAsync();

/*!
    @brief  Leaves a session.
    @details

           This function starts an asynchronous process that leaves a session.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedLeaveSession</tt> to check whether the process has completed and <tt>@ref GetLeaveSessionResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedLeaveSession, GetLeaveSessionResult
*/
    nn::Result LeaveSessionAsync();

/*!
    @brief  Checks the completion of <tt>LeaveSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref LeaveSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  LeaveSessionAsync, GetLeaveSessionResult
*/
    bool IsCompletedLeaveSession() const;

/*!
        @brief  Gets the result of <tt>LeaveSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref LeaveSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running and <tt>Startup</tt> has been called. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultErrorOccurred  An internal error occurred. This result is only returned during local communication. Perform cleanup.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  LeaveSessionAsync, IsCompletedLeaveSession
*/
    nn::Result GetLeaveSessionResult() const;

    // Starting to accept session participants.
/*!
    @brief  Starts accepting participants in the session.
    @details

           This function starts an asynchronous process that accepts participants.
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedOpenParticipation</tt> to check whether the process has completed and <tt>@ref GetOpenParticipationResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedOpenParticipation, GetOpenParticipationResult, IsHost
*/
    nn::Result OpenParticipationAsync();

/*!
    @brief  Checks the completion of <tt>OpenParticipationAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref OpenParticipationAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  OpenParticipationAsync, GetOpenParticipationResult
*/
    bool IsCompletedOpenParticipation() const;

/*!
        @brief  Gets the result of <tt>OpenParticipationAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref OpenParticipationAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  OpenParticipationAsync, IsCompletedOpenParticipation
*/
    nn::Result GetOpenParticipationResult() const;

    // Starting to accept session participants.
/*!
    @brief  Starts accepting participants in the session.
    @details

           This function starts an asynchronous process that accepts participants.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedOpenSession</tt> to check whether the process has completed and <tt>@ref GetOpenSessionResult</tt> to check the result.
           All stations participating in the session must call this function.
           The asynchronous process ends after it can be verified that all stations that have finished joining the session have called this function.
           After asynchronous processes complete, the matchmaking session is in an open state.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>CloseSessionAsync</tt>. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.

    @see  IsCompletedOpenSession, GetOpenSessionResult, IsHost
*/
    nn::Result OpenSessionAsync();

/*!
    @brief  Checks for completion of the <tt>OpenSessionAsync</tt> asynchronous process.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref OpenSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  OpenSessionAsync, GetOpenSessionResult
*/
    bool IsCompletedOpenSession() const;

/*!
        @brief  Gets the result of the <tt>OpenSessionAsync</tt> asynchronous process.
        @details  This function gets the results after an asynchronous process started by <tt>@ref OpenSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. This result also can be returned when the host has created and joined a joint session and passively begun an asynchronous process. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>CloseSessionAsync</tt>. <tt>OpenSessionAsync</tt> was not called by all stations participating in the session within 60 seconds after the function was called. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.
        @retval nn::pia::ResultSessionInconsistentState  Indicates that the process took longer than expected to complete. If the station is a client, make it check whether the other clients have called <tt>OpenSessionAsync</tt>. Perform cleanup after calling <tt>LeaveSessionAsync</tt>.

        @see  OpenSessionAsync, IsCompletedOpenSession
*/
    nn::Result GetOpenSessionResult() const;

    // Closing a session to new participants.
/*!
    @brief  Closes a session to new participants.
    @details

           This function starts an asynchronous process that closes a session to new participants.
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedCloseParticipation</tt> to check whether the process has completed and <tt>@ref GetCloseParticipationResult</tt> to check the result.

           It is possible to imagine a case where some station attempts to join the session immediately before this function is called and ends up joining after the asynchronous process has completed.
           For this reason, the asynchronous process ends after waiting for the session to close and for the number of participants in the matchmaking session to match the number of participants in the mesh.
           The asynchronous process also ends if the numbers do not match after a certain amount of time (15 seconds). In this case, <tt>@ref GetCloseParticipationResult</tt>
           returns <tt>ResultNegligibleFault</tt>. There could be attempts to join after the asynchronous process has ended.
           Handle these attempts as appropriate in the application.


    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedCloseParticipation, GetCloseParticipationResult, IsHost
*/
    nn::Result CloseParticipationAsync();

/*!
    @brief  Checks the completion of <tt>CloseParticipationAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref CloseParticipationAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CloseParticipationAsync, GetCloseParticipationResult
*/
    bool IsCompletedCloseParticipation() const;

/*!
        @brief  Gets the result of <tt>CloseParticipationAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref OpenParticipationAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultNegligibleFault  A certain amount of time passed without the matchmaking sessions and number of participants matching. Handle appropriately in the application.

        @see  CloseParticipationAsync, IsCompletedCloseParticipation
*/
    nn::Result GetCloseParticipationResult() const;

    // Closing a session to new participants.
/*!
    @brief  Closes a session to new participants.
    @details

           This function starts an asynchronous process that closes a session to new participants.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedCloseSession</tt> to check whether the process has completed and <tt>@ref GetCloseSessionResult</tt> to check the result.
           All stations participating in the session must call this function.

           After it is confirmed that the numbers of matchmaking session and mesh participants match and that the stations that have completed connection have called this function, asynchronous processes finish.
           After it is complete, the matchmaking session enters a closed state. Advance processing under the assumption that after asynchronous processing is complete, the number of new participants will not increase.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>OpenSessionAsync</tt>. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.

    @see  IsCompletedCloseSession, GetCloseSessionResult, IsHost
*/
    nn::Result CloseSessionAsync();

/*!
    @brief  Checks the completion of <tt>CloseSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref CloseSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CloseSessionAsync, GetCloseSessionResult
*/
    bool IsCompletedCloseSession() const;

/*!
        @brief  Gets the result of <tt>CloseSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref CloseSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. This result also can be returned when the host has created and joined a joint session and passively begun an asynchronous process. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>OpenSessionAsync</tt>. <tt>CloseSessionAsync</tt> was not called by all stations participating in the session within 60 seconds after the function was called. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.
        @retval nn::pia::ResultSessionInconsistentState  Indicates that the process took longer than expected to complete. If the station is a client, have it check whether the other clients have called <tt>CloseSessionAsync</tt>. Perform cleanup after calling <tt>LeaveSessionAsync</tt>.

        @see  CloseSessionAsync, IsCompletedCloseSession
*/
    nn::Result GetCloseSessionResult() const;

    // Generating a system password for a session
/*!
    @brief  Generates a system password on the server and sets the password on the session.
    @details
           This function cannot be used with local communications.
           This function asynchronously generates a system password on the server and sets the password on the session.
           The configured password is stored in the <tt><var>pSystemPassword</var></tt> parameter after the asynchronous process is complete.
           The client is notified by the <tt>Session::EVENT_SET_SESSION_SYSTEM_PASSWORD</tt> event and can get the configured password using <tt>@ref GetSessionSystemPassword</tt>.
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedGenerateSystemPassword</tt> to check whether the process has completed and <tt>@ref GetGenerateSystemPasswordResult</tt> to check the result.

    @param[out] pSystemPassword  Stores the system password that was set.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedGenerateSystemPassword, GetGenerateSystemPasswordResult, GetSessionSystemPassword, IsHost
*/
    nn::Result GenerateSystemPasswordAsync(wchar_t* pSystemPassword);
/*!
    @brief  Checks the completion of <tt>GenerateSystemPasswordAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref GenerateSystemPasswordAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  GenerateSystemPasswordAsync, GetGenerateSystemPasswordResult
*/
    bool IsCompletedGenerateSystemPassword() const;
/*!
        @brief  Gets the result of <tt>GenerateSystemPasswordAsync</tt> asynchronous processing.
        @details
                 This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref GenerateSystemPasswordAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  GenerateSystemPasswordAsync, IsCompletedGenerateSystemPassword
*/
    nn::Result GetGenerateSystemPasswordResult() const;

    //! @cond PRIVATE
    void SetSessionSystemPassword(const wchar_t* pSystemPassword, size_t length);
    void ResetSessionSystemPassword();
    //! @endcond
/*!
    @brief  Gets the system password that was set on the session with <tt>GenerateSystemPasswordAsync</tt>.
    @details

              This function cannot be used with local communications.
              Gets the system password that was set on the session with <tt>GenerateSystemPasswordAsync</tt>.
              You can get the system password set for the session after <tt>Session::EVENT_SET_SESSION_SYSTEM_PASSWORD</tt>
              has been reported to the registered event callback.

    @param[out] pSystemPassword  Stores the system password that was set.

    @see  GenerateSystemPasswordAsync
*/
    void GetSessionSystemPassword(wchar_t* pSystemPassword);

/*!
    @brief  This function clears the system password that was set on the session.
    @details  This function cannot be used with local communications.

           This function asynchronously clears the system password that was set on the session.
           Clients are notified by the <tt>Session::EVENT_CLEAR_SESSION_SYSTEM_PASSWORD</tt> event.
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedClearSystemPassword</tt> to check whether the process has completed and <tt>@ref GetClearSystemPasswordResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.

    @see  IsCompletedClearSystemPassword, GetClearSystemPasswordResult, IsHost
*/
    nn::Result ClearSystemPasswordAsync();
/*!
    @brief  Checks the completion of <tt>ClearSystemPasswordAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref ClearSystemPasswordAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  ClearSystemPasswordAsync, GetClearSystemPasswordResult
*/
    bool IsCompletedClearSystemPassword() const;
/*!
        @brief  Gets the result of <tt>ClearSystemPasswordAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref ClearSystemPasswordAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  ClearSystemPasswordAsync, IsCompletedClearSystemPassword
*/
    nn::Result GetClearSystemPasswordResult() const;

/*!
    @brief  Updates the specified session search attribute.
    @details  This function cannot be used with local communications.

           This function asynchronously updates the specified session search attribute.
           Only the session host can execute this function.
           The <tt><var>attributeIndex</var></tt> parameter can be set to a value between <tt>0</tt> and <tt>@ref nn::pia::inet::ATTRIBUTE_MAX_SIZE</tt> - <tt>1</tt>.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedModifyAttribute</tt> to check whether the process has completed and <tt>@ref GetModifyAttributeResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  The value specified in <tt><var>attributeIndex</var></tt> is outside the allowed range. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedModifyAttribute, GetModifyAttributeResult, IsHost
*/
    nn::Result ModifyAttributeAsync(u32 attributeIndex, u32 newValue);
/*!
    @brief  Checks the completion of <tt>ModifyAttributeAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref ModifyAttributeAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  ModifyAttributeAsync, GetModifyAttributeResult
*/
    bool IsCompletedModifyAttribute() const;
/*!
        @brief  Gets the result of <tt>ModifyAttributeAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref ModifyAttributeAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  ModifyAttributeAsync, IsCompletedModifyAttribute
*/
    nn::Result GetModifyAttributeResult() const;

 /*!
    @brief  Updates the specified session settings.
    @details  This function cannot be used with local communications.
           This function asynchronously updates the specified session settings.
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedUpdateSessionSetting</tt> to check whether the process has completed and <tt>@ref GetUpdateSessionSettingResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @param[in] pUpdateSessionSetting  Pointer to the <tt>UpdateSessionSetting</tt> derived class for gathering the settings to change.

    @see  UpdateSessionSetting, IsCompletedUpdateSessionSetting, GetUpdateSessionSettingResult, IsHost
*/
    nn::Result UpdateSessionSettingAsync(UpdateSessionSetting* pUpdateSessionSetting);
/*!
    @brief  Checks the completion of <tt>UpdateSessionSettingAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref UpdateSessionSettingAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  UpdateSessionSettingAsync, GetUpdateSessionSettingResult
*/
    bool IsCompletedUpdateSessionSetting() const;
/*!
        @brief  Gets the result of <tt>UpdateSessionSettingAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref UpdateSessionSettingAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  UpdateSessionSettingAsync, IsCompletedUpdateSessionSetting
*/
    nn::Result GetUpdateSessionSettingResult() const;

/*!
    @brief  Updates application-defined data.
    @details

           Asynchronously updates application-defined data (the data set in beacons during local communication).
           Only the session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedUpdateApplicationData</tt> to check whether the process has completed and <tt>@ref GetUpdateApplicationDataResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Either the specified <tt><var>pApplicationData</var></tt> is <tt>Null</tt> or the <tt><var>dataSize</var></tt> has exceeded the value that can be set. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @param[in] pApplicationData  The application-defined data to update.
    @param[in] dataSize  The size of the specified application-defined data.

    @see  IsCompletedUpdateApplicationData, GetUpdateApplicationDataResult, IsHost
*/
    nn::Result UpdateApplicationDataAsync(const void* pApplicationData, size_t dataSize);
/*!
    @brief  Checks the completion of <tt>UpdateApplicationDataAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref UpdateApplicationDataAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  UpdateApplicationDataAsync, GetUpdateApplicationDataResult
*/
    bool IsCompletedUpdateApplicationData() const;
/*!
        @brief  Gets the result of <tt>UpdateApplicationDataAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref UpdateApplicationDataAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultGameServerMaintenance  Indicates that the server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.

        @see  UpdateApplicationDataAsync, IsCompletedUpdateApplicationData
*/
    nn::Result GetUpdateApplicationDataResult() const;

/*!
        @brief  Updates the matchmaking priority when a session with the same requirements is found.
        @details  This function cannot be used with local communications.
                 If the value is being updated multiple times during the session with something other than <tt>0</tt> or
                 <tt>nn::pia::inet::SESSION_MAX_SELECTION_PRIORITY</tt>, you must wait 30 seconds between calls.
                 Use <tt>nn::pia::session::Session::IsUpdatableSelectionPriority</tt> to check whether an update can be made.
                 When searches are performed using <tt>nn::pia::inet::SELECTION_METHOD_BROADEN_RANGE_WITH_SELECTION_PRIORITY</tt>, matchmaking opportunities are prioritized according to the greatest matchmaking priority value.
                 When searches are performed using <tt>nn::pia::inet::SELECTION_METHOD_SCORE_BASED</tt>, the set score calculation is used.
                 You can check the settings by checking the progress values in the NMAS matchmaking session list.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidArgument  The specified argument exceeds the maximum allowed value. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultInvalidState  Cannot execute because either the LAN matching feature is being used or the station is not the host. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultTemporaryUnavailable  Called within 30 seconds of the previous call with values other than <tt>0</tt> or <tt>SESSION_MAX_SELECTION_PRIORITY</tt>. Use <tt>@ref nn::pia::session::Session::IsUpdatableSelectionPriority</tt> to confirm whether an update is possible before calling.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

        @param[in] priority  The progress in the game.

        @see  IsUpdatableSelectionPriority
*/
    nn::Result UpdateSelectionPriority(u8 priority) const;
/*!
        @brief  Returns whether <tt>UpdateSelectionPriority</tt> can be used for the value being updated.

        @param[in] priority  The progress in the game to be updated.
*/
    bool IsUpdatableSelectionPriority(u8 priority) const;

    /// @cond PRIVATE
    MeshLayerController* GetMeshLayerControllerPtr()
    {
        return m_pMeshLayerController;
    }
    /// @endcond

/*!
     end of name Session Operations
     @}
*/


/*!
     @name  Analyzing Session Information
     @{
*/

/*!
        @brief  Gets whether the local station is the session host.

        @details  This function returns whether the local station is the host of a standard session.

        @return  Returns <tt>true</tt> if the local station is the host, or <tt>false</tt> otherwise.
*/
    bool IsHost() const;

/*!
        @cond  PRIVATE
        @brief  Gets whether the local station is the mesh host.

        @details  This function returns whether the local station is the host of the mesh.

        @return  Returns <tt>true</tt> if the local station is the host, or <tt>false</tt> otherwise.
*/
    bool IsMeshHost() const;
    /// @endcond

/*!
        @cond  PRIVATE
        @brief  Gets whether the local station is the session owner.

        @return  Returns <tt>true</tt> if the local station is the session owner, or <tt>false</tt> otherwise.
*/
    bool IsOwner() const;
    /// @endcond

    /// @cond PRIVATE
/*!
        @brief  Gets the <tt>StationIndex</tt> of the local station.

        @details  This function returns <tt>STATION_INDEX_INVALID</tt> if the local station's <tt>StationIndex</tt> has not been established.

        @return  Returns the <tt>StationIndex</tt> representing the local station.
*/
    nn::pia::StationIndex GetLocalStationIndex() const;

/*!
        @brief  Gets the <tt>StationIndex</tt> of the host.

        @details  This function returns <tt>STATION_INDEX_INVALID</tt> if the local station is not a participant in a session.
                 When participating in a session or a joint session, this function returns a <tt>StationIndex</tt> that specifies the session host.

    @return  Returns the <tt>StationIndex</tt> that represents the host of the session where the local station is a participant.
*/
    nn::pia::StationIndex GetHostStationIndex() const;
    /// @endcond

/*!
    @brief  Checks whether an arbitrary <tt>StationId</tt> within the session is currently valid.

    @param[in] id  Specifies the <tt>StationId</tt> being checked.

    @return  Returns <tt>true</tt> if the specified <tt>StationId</tt> is valid, or <tt>false</tt> otherwise.
*/
    bool CheckStationIdIsValid(nn::pia::StationId id) const;

    /// @cond PRIVATE
    bool CheckStationIndexIsValid(nn::pia::StationIndex index) const;
    /// @endcond


/*!
    @brief  Gets the <tt>StationId</tt> that represents the local station in the session.

    @details  This function returns <tt>STATION_ID_INVALID</tt> if the local station is not a participant in a session.

    @return  Returns the <tt>StationId</tt> that specifies the local device.
*/
    StationId GetLocalStationId() const
    {
        return m_LocalStationId;
    }

/*!
    @brief  Gets the <tt>StationId</tt> that represents the host of the session where the local station is a participant.

    @details  This function returns <tt>STATION_ID_INVALID</tt> if the local station is not a participant in a session.
             When participating in a session or a joint session, this function returns a <tt>StationId</tt> that specifies the session host.

    @return  Returns the <tt>StationId</tt> that represents the host of the session where the local station is a participant.
*/
    StationId GetHostStationId() const
    {
        return m_HostStationId;
    }

/*!
    @brief  Gets the <tt>StationConnectionInfo</tt> that specifies the local device.

    @details  When participating in a session or a joint session, this function returns a <tt>StationConnectionInfo</tt> that represents the station itself.

    @param[out] pInfoOut  Specifies a pointer to the object that holds the address obtained for the session host.
    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
    @retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates not participating in a session. Programming error. Fix your program so that this error is not returned.
    @retval ResultTemporaryUnavailable  Session transition in progress. Handle appropriately in the application.

    @return  Returns the <tt>StationConnectionInfo</tt> that specifies the local device.
*/
    nn::Result GetLocalStationConnectionInfo(transport::StationConnectionInfo* pInfoOut);

/*!
    @brief  Gets the <tt>StationConnectionInfo</tt> that represents the host of the session where the local station is a participant.

    @details  When participating in a session or a joint session, this function returns a <tt>StationConnectionInfo</tt> that specifies the session host.

    @param[out] pInfoOut  Specifies a pointer to the object that holds the address obtained for the session host.
    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
    @retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates not participating in a session. Programming error. Fix your program so that this error is not returned.
    @retval ResultTemporaryUnavailable  Session transition in progress. Handle appropriately in the application.

    @return  Returns the <tt>StationConnectionInfo</tt> that represents the host of the session where the local station is a participant.
*/
    nn::Result GetHostStationConnectionInfo(transport::StationConnectionInfo* pInfoOut);

/*!
    @brief  Gets the number of valid stations connected to the current session.

    @return  Returns the number of connected stations.
*/
    u16 GetStationNum() const;

/*!
    @brief  Gets the number of stations currently participating in the matchmaking session.

    @details  When <tt>@ref GetStatus</tt> returns <tt>STATUS_CONNECTED_SESSION</tt>, this function returns the number of stations participating in the matchmaking session.
                When <tt>@ref GetStatus</tt> returns <tt>STATUS_MIGRATING_SESSION</tt>, this function returns an undefined value, so do not call it in such cases.
                When <tt>@ref GetStatus</tt> returns <tt>STATUS_CONNECTED_JOINT_SESSION</tt>, this function returns the number of stations participating in the joint matchmaking session.

    @return  Returns the number of stations participating in the matchmaking session.
*/
    u16 GetMatchmakeSessionParticipantNum() const;

/*!
    @brief  Gets the number of valid <tt>StationId</tt> instances in the session.

    @details  This function returns <tt>0</tt> if the local station is not a participant in a session.

    @return  Returns the number of valid <tt>StationId</tt> instances in the session.
*/
    u16 GetValidStationIdNum() const;


/*!
    @brief  Checks whether communication is possible with the session.

    @details  The <tt>Result</tt> object returned by this function changes if there are any irregularities in the session or the connection.
                Periodically call this function to check the status.

    @details  If communication is possible, the detailed status can be retrieved using <tt>@ref GetStatus</tt>.

    @details  If communication with the mesh becomes impossible after joining the session, the detailed reason can be checked by calling <tt>@ref GetSessionDisconnectReason</tt>.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if communication is possible. Otherwise, one of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
*/
    nn::Result CheckConnectionError() const;


/*!
    @brief  Gets a <tt>@ref Status</tt> object that specifies the status of the session.

    @details  The <tt>@ref Status</tt> object returned by this function changes if there are any irregularities in the session or the connection.
                Periodically call this function to check the status.

    @return  Returns the current <tt>@ref Status</tt>.
*/
    Status GetStatus() const;

/*!
    @brief  Determines whether the bitrate value may be found using the bitrate detection feature.

    @return  Returns <tt>true</tt> if the bitrate value can be obtained. <tt>true</tt> is also returned when detection fails on a timeout, or cancel.
*/
    bool IsCompleteBitRateCheck();

/*!
    @brief  Gets the uplink bitrate of the connection measured by the bitrate detection feature.

    @details  Note that the value returned by this function is simply a measurement of the amount of data received over a set
             amount of time and does not necessarily represent the actual bandwidth of the connection.
             Even if the connection is capable of a high bitrate, if an application has
             low requirements, the amount of data sent and received during the measurement
             process is also small, which results in a lower bits-per-second value.

    @param[out] pBps  Stores the measured bitrate. The value is in bits per second. A negative value is set if the bitrate could not be measured.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the call succeeds.

    @retval ResultNotInitialized  Indicates that the instance has not been initialized. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidArgument  Indicates that an argument is invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Indicates that the function was called at the wrong time. It is possible that the function was called while the bitrate detection feature was disabled. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetUplinkBitRate(s32* pBps);

    /// @cond PRIVATE
/*!
    @brief  Gets a <tt>MatchmakeStatus</tt> object that specifies the matchmaking status of the session.

    @return  Returns the current <tt>MatchmakeStatus</tt>.
*/
    MatchmakeStatus GetMatchmakeStatus() const
    {
        return m_MatchmakeStatus;
    }
    /// @endcond

    /// @cond PRIVATE
/*!
    @brief  Gets a <tt>ConnectionStatus</tt> object that specifies the status of the connection.

    @return  Returns the current <tt>ConnectionStatus</tt>.
*/
    ConnectionStatus GetConnectionStatus() const
    {
        return m_ConnectionStatus;
    }
    /// @endcond

    /// @cond PRIVATE
    // private
    StationIdStatusTable* GetStationIdStatusTablePtr()
    {
        return m_pStationIdStatusTable;
    }

    // private
    SessionProtocol* GetSessionProtocolPtr()
    {
        return m_pSessionProtocol;
    }
    /// @endcond

/*!
    @brief  Gets the session ID.

    @return  Returns the session ID of the currently joined session. Returns <tt>INVALID_SESSION_ID</tt> if not a participant in a session.
*/
    u32 GetSessionId() const
    {
        return GetCurrentSessionId();
    }

/*!
    @brief  Gets the session ID for the session in which the station with the specified station ID is participating.
    @details  This function cannot be used with local communications.
                Gets the session ID for the session in which the station with the specified station ID is participating.

    @param[in] id  The station ID.

    @return  Returns the session ID for the session in which the station with the specified station ID is participating. Returns <tt>INVALID_SESSION_ID</tt> if that station is not a participant in a session.
*/
    u32 GetSessionId(StationId id) const;

    /// @cond PRIVATE
    // <tt>SessionId</tt> Getting and Setting
    void SetSessionId(u32 sessionId)
    {
        SetCurrentSessionId(sessionId);
    }

    void SetPrincipalId(u32 principalId)
    {
        m_PrincipalId = principalId;
    }
    u32 GetPrincipalId() const
    {
        return m_PrincipalId;
    }
    /// @endcond

/*!
    @brief  Gets an iterator for the valid <tt>StationId</tt> objects in the session (non-<tt>const</tt> version).

    @return  Returns an iterator that points to the start of a list of <tt>StationId</tt> objects that are valid at the time the function is called.
    @see  End
*/
    StationIdIterator Begin()
    {
        return StationIdIterator(m_ValidStationIdList.Begin());
    }

/*!
    @brief  Gets an iterator for the valid <tt>StationId</tt> objects in the session (non-<tt>const</tt> version).

    @return  Returns an iterator that points to the position after the last element in a list of <tt>StationId</tt> objects that are valid at the time the function is called.
    @see  Begin
*/
    StationIdIterator End()
    {
        return StationIdIterator(m_ValidStationIdList.End());
    }

/*!
    @brief  Gets an iterator for the valid <tt>StationId</tt> objects in the session (non-<tt>const</tt> version).

    @return  Returns an iterator that points to the start of a list of <tt>StationId</tt> objects that are valid at the time the function is called.
    @see  End
*/
    StationIdConstIterator Begin() const
    {
        return StationIdConstIterator(m_ValidStationIdList.ConstBegin());
    }

/*!
    @brief  Gets an iterator for the valid <tt>StationId</tt> objects in the session (non-<tt>const</tt> version).

    @return  Returns an iterator that points to the position after the last element in a list of <tt>StationId</tt> objects that are valid at the time the function is called.
    @see  Begin
*/
    StationIdConstIterator End() const
    {
        return StationIdConstIterator(m_ValidStationIdList.ConstEnd());
    }

/*! @cond PRIVATE */
    ///  Whether StationIdTable is being used.
    bool IsUsingStationIdTable() const;
/*! @endcond */

/*!
     end of name Analyzing Session Information
     @}
*/

/*!
     @name  Joint Session Feature
     @{
*/

/*!
    @brief  Joins a joint session by using random matchmaking.
    @details  This function cannot be used with local communications.
            This function can only be used by the host.
            Starts the asynchronous process that handles joining a joint session through random matchmaking.
            The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedJoinRandomJointSession</tt> to check whether the process has completed and <tt>@ref GetJoinRandomJointSessionResult</tt> to check the result.

            When the host starts an asynchronous process using this function, the process that handles joining a joint session through random matchmaking starts automatically on the client.
            To track the progress of the automatically started processes, either handle the state change event or call <tt>@ref GetStatus</tt> and check the <tt>@ref Status</tt> object that is returned.

            Some API functions, such as those related to sending and receiving, are temporarily unavailable during the asynchronous process.

    @param[in] pCreateSessionSetting  Specifies a pointer to the <tt>CreateSessionSetting</tt> object that specifies the settings for the new session to create.
    @param[in] pSearchCriterias  Specifies a pointer to the beginning of the array of criteria to use when searching for a session.
    @param[in] criteriaNum  Specifies the number of elements in the criteria array to use when searching for a session.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running and that the local station is the host. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedJoinRandomJointSession, GetJoinRandomJointSessionResult
*/
    nn::Result JoinRandomJointSessionAsync(
        CreateSessionSetting* pCreateSessionSetting,
        SessionSearchCriteria* pSearchCriterias,
        size_t criteriaNum);

/*!
    @brief  Checks the completion of <tt>JoinRandomJointSessionAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
                This function checks the completion of asynchronous processes started with <tt>@ref JoinRandomJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  JoinRandomJointSessionAsync, GetJoinRandomJointSessionResult
*/
    bool IsCompletedJoinRandomJointSession() const;

/*!
        @brief  Gets the result of <tt>JoinRandomJointSessionAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                    This function gets the results after an asynchronous process started by <tt>@ref JoinRandomJointSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the asynchronous process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the local station is not in a joint session-ready state. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionIsNotFound  The target session was lost or closed during the asynchronous process. Perform cleanup.
    @retval nn::pia::ResultMatchmakeSessionIsFull  The target matchmaking session is full. Perform cleanup.
    @retval nn::pia::ResultDeniedByParticipants  The local station is blacklisted by a session participant. Perform cleanup.
    @retval nn::pia::ResultParticipantInBlackList  A participant on the local station's blacklist is in the session. Perform cleanup.
    @retval nn::pia::ResultSessionUserPasswordUnmatch  The user password specified when joining does not match the user password set on the session. Perform cleanup.
    @retval nn::pia::ResultSessionSystemPasswordUnmatch  The system password specified when joining does not match the system password set on the session. Perform cleanup.
    @retval nn::pia::ResultSessionIsClosed  Attempted to join a session that is closed. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultHostIsNotFriend  Attempted to join a session with a host that is not a friend. Perform cleanup.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
    @retval nn::pia::ResultCompanionStationIsLeft  Lost the connection with the companion station. Perform cleanup.
    @retval nn::pia::ResultCompanionStationDenied  The companion station was not in a joint session-ready state and denied the connection. Perform cleanup.
    @retval nn::pia::ResultSessionMigrationFailed  Failed to migrate the session. Perform cleanup.
    @retval nn::pia::ResultStationConnectionFailed  The connection process failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultStationConnectionNatTraversalFailedUnknown  The NAT traversal connection process failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEim  The NAT traversal connection process failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm  The NAT traversal connection process failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim  The NAT traversal connection process failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEdm  The NAT traversal connection process failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalRequestTimeout  The NAT traversal connection process failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultJoinRequestDenied  The mesh host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultMeshIsFull  Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedNoCandidate  The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRttLimit  The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRelayNumLimit  The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedUnknown  The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultInvalidSystemMessage  The process was canceled because an invalid response was received from the mesh host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultIncompatibleFormat  Indicates that the peer is using an incompatible communication format. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultSessionWrongState  The state of the joined session was irregular. Perform cleanup.

    @see  JoinRandomJointSessionAsync, IsCompletedJoinRandomJointSession
*/
    nn::Result GetJoinRandomJointSessionResult() const;

    //! @cond PRIVATE
/*!
    @brief  Creates a joint session.
    @details  This function cannot be used with local communications.
            This function can only be used by the host.
            It starts an asynchronous process that creates a joint session.
            The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedCreateJointSession</tt> to check whether the process has completed and <tt>@ref GetCreateJointSessionResult</tt> to check the result.

            When the host starts the asynchronous process using this function, the interal process for creating the joint session starts automatically on the client.
            To track the progress of the automatically started processes, either handle the state change event or call <tt>@ref GetStatus</tt> and check the <tt>@ref Status</tt> object that is returned.

            Some API functions, such as those related to sending and receiving, are temporarily unavailable during the asynchronous process.

    @param[in] pCreateSessionSetting  Specifies a pointer to the <tt>CreateSessionSetting</tt> object that specifies the settings for the new session to create.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running and that the local station is the host. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedCreateJointSession, GetCreateJointSessionResult
*/
    nn::Result CreateJointSessionAsync(CreateSessionSetting* pCreateSessionSetting);

/*!
    @brief  Checks for completion of the <tt>CreateJointSessionAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
                This function checks the completion of asynchronous processes started with <tt>@ref CreateJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CreateJointSessionAsync, GetCreateJointSessionResult
*/
    bool IsCompletedCreateJointSession() const;

/*!
        @brief  Gets the result of the <tt>CreateJointSessionAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                    This function gets the results after an asynchronous process started by <tt>@ref CreateJointSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the asynchronous process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the local station is not in a joint session-ready state. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
    @retval nn::pia::ResultCompanionStationIsLeft  Lost the connection with the companion station. Perform cleanup.
    @retval nn::pia::ResultCompanionStationDenied  The companion station was not in a joint session-ready state and denied the connection. Perform cleanup.
    @retval nn::pia::ResultSessionMigrationFailed  Failed to migrate the session. Perform cleanup.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultSessionWrongState  The created session is in an irregular state. Perform cleanup.

    @see  CreateJointSessionAsync, IsCompletedCreateJointSession
*/
    nn::Result GetCreateJointSessionResult() const;

/*!
    @brief  Joins the specified joint session.
    @details  This function cannot be used with local communications.
            This function can only be used by the host.
            Starts the asynchronous process to join the specified joint session.
            The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedJoinJointSession</tt> to check whether the process has completed and <tt>@ref GetJoinJointSessionResult</tt> to check the result.

            When the host starts an asynchronous process using this function, the process that handles joining a joint session through random matchmaking starts automatically on the client.
            To track the progress of the automatically started processes, either handle the state change event or call <tt>@ref GetStatus</tt> and check the <tt>@ref Status</tt> object that is returned.

            Some API functions, such as those related to sending and receiving, are temporarily unavailable during the asynchronous process.

    @param pJoinSessionSetting  Specifies the settings of the session to join.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running and that the local station is the host. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedJoinJointSession, GetJoinJointSessionResult
*/
    nn::Result JoinJointSessionAsync(JoinSessionSetting* pJoinSessionSetting);

/*!
    @brief  Checks for completion of the <tt>JoinJointSessionAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
                This function checks the completion of asynchronous processes started with <tt>@ref JoinJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  JoinJointSessionAsync, GetJoinJointSessionResult
*/
    bool IsCompletedJoinJointSession() const;

/*!
        @brief  Gets the result of the <tt>JoinJointSessionAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                    This function gets the results after an asynchronous process started by <tt>@ref JoinJointSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the asynchronous process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the local station is not in a joint session-ready state. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionIsNotFound  The target session was lost or closed during the asynchronous process. Perform cleanup.
    @retval nn::pia::ResultMatchmakeSessionIsFull  The target matchmaking session is full. Perform cleanup.
    @retval nn::pia::ResultDeniedByParticipants  The local station is blacklisted by a session participant. Perform cleanup.
    @retval nn::pia::ResultParticipantInBlackList  A participant on the local station's blacklist is in the session. Perform cleanup.
    @retval nn::pia::ResultSessionUserPasswordUnmatch  The user password specified when joining does not match the user password set on the session. Perform cleanup.
    @retval nn::pia::ResultSessionSystemPasswordUnmatch  The system password specified when joining does not match the system password set on the session. Perform cleanup.
    @retval nn::pia::ResultSessionIsClosed  Attempted to join a session that is closed. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultHostIsNotFriend  Attempted to join a session with a host that is not a friend. Perform cleanup.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
    @retval nn::pia::ResultCompanionStationIsLeft  Lost the connection with the companion station. Perform cleanup.
    @retval nn::pia::ResultCompanionStationDenied  The companion station was not in a joint session-ready state and denied the connection. Perform cleanup.
    @retval nn::pia::ResultSessionMigrationFailed  Failed to migrate the session. Perform cleanup.
    @retval nn::pia::ResultStationConnectionFailed  The connection process failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultStationConnectionNatTraversalFailedUnknown  The NAT traversal connection process failed. The NAT type is unknown. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEim  The NAT traversal connection process failed. The NAT type was EIM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEimRemoteEdm  The NAT traversal connection process failed. The NAT type was EIM for the local station and EDM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedLocalEdmRemoteEim  The NAT traversal connection process failed. The NAT type was EDM for the local station and EIM for the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalFailedBothEdm  The NAT traversal connection process failed. The NAT type was EDM for both the local station and the peer to which the connection failed. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultNatTraversalRequestTimeout  The NAT traversal connection process failed. The NAT traversal request timed out. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Perform cleanup.
    @retval nn::pia::ResultJoinRequestDenied  The mesh host denied the connection request. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultMeshIsFull  Failed to join the mesh because it was full. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedNoCandidate  The relay connection failed because there were no relay route candidates. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRttLimit  The relay connection failed because the relay route RTT limit was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedRelayNumLimit  The relay connection failed because the limit on the number of relay route relay requests was exceeded. Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultRelayFailedUnknown  The relay connection failed. (Details unknown.) Perform cleanup. Handle appropriately in the application. When an error occurs in Internet communication, If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Do not use the error/EULA applet (CTR) for local communication errors.
    @retval nn::pia::ResultInvalidSystemMessage  The process was canceled because an invalid response was received from the mesh host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultIncompatibleFormat  Indicates that the peer is using an incompatible communication format. Returned when you try to communicate with a ROM that uses a different version of Pia that does not have communication compatibility. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultSessionWrongState  The state of the joined session was irregular. Perform cleanup.

    @see  JoinJointSessionAsync, IsCompletedJoinJointSession
*/
    nn::Result GetJoinJointSessionResult() const;

/*!
    @brief  Leaves a joint session.
    @details  This function cannot be used with local communications.
            This function is only available to the joint session host.
            Starts the asynchronous process of leaving a joint session.

            The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedLeaveJointSession</tt> to check whether the process has completed and <tt>@ref GetLeaveJointSessionResult</tt> to check the result.

            If the host has started the process of leaving a joint session with this function, the internal process for leaving the joint session also begins automatically for the companion stations.
            To track the progress of the automatically started processes, either handle the state change event or call <tt>@ref GetStatus</tt> and check the <tt>@ref Status</tt> object that is returned.

            Some API functions, such as those related to sending and receiving, are temporarily unavailable during the asynchronous process.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedLeaveJointSession, GetLeaveJointSessionResult
*/
    nn::Result LeaveJointSessionAsync();

/*!
    @brief  Checks for completion of the <tt>LeaveJointSessionAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
                This function checks the completion of asynchronous processes started with <tt>@ref LeaveJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  LeaveJointSessionAsync, GetLeaveJointSessionResult
*/
    bool IsCompletedLeaveJointSession() const;

/*!
        @brief  Gets the result of the <tt>LeaveJointSessionAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                    This function gets the results after an asynchronous process started by <tt>@ref LeaveJointSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the asynchronous process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Returns when the process cannot be made. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
    @retval nn::pia::ResultCompanionStationIsLeft  Lost the connection with the companion station. Perform cleanup.
    @retval nn::pia::ResultCompanionStationDenied  The companion station was not in a joint disband-ready state and denied the connection. Perform cleanup.
    @retval nn::pia::ResultSessionMigrationFailed  Failed to migrate the session. Perform cleanup.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

    @see  LeaveJointSessionAsync, IsCompletedLeaveJointSession
*/
    nn::Result GetLeaveJointSessionResult() const;
    //! @endcond

/*!
    @brief  Disbands a joint session.
    @details  This function cannot be used with local communications.
            This function is only available to the joint session host.
            It starts an asynchronous process that disbands a joint session.

            The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
            Use <tt>@ref IsCompletedDestroyJointSession</tt> to check whether the process has completed and <tt>@ref GetDestroyJointSessionResult</tt> to check the result.

            If the host starts a process to disband a joint session with this function, the companion-side processes for disbanding start automatically.
            To track the progress of the automatically started processes, either handle the state change event or call <tt>@ref GetStatus</tt> and check the <tt>@ref Status</tt> object that is returned.

            Some API functions, such as those related to sending and receiving, are temporarily unavailable during the asynchronous process.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.

    @see  IsCompletedDestroyJointSession, GetDestroyJointSessionResult
*/
    nn::Result DestroyJointSessionAsync();

/*!
    @brief  Checks the completion of <tt>DestroyJointSessionAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
                This function checks the completion of asynchronous processes started with <tt>@ref DestroyJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  DestroyJointSessionAsync, GetDestroyJointSessionResult
*/
    bool IsCompletedDestroyJointSession() const;

/*!
        @brief  Gets the result of <tt>DestroyJointSessionAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                    This function gets the results after an asynchronous process started by <tt>@ref DestroyJointSessionAsync</tt> ends.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the asynchronous process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Returns when the process cannot be made. This result is also returned when the asynchronous process has not completed or has not started. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that unexpected arguments were passed to the function when starting the asynchronous process. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultGameServerMaintenance  Indicates that the game server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.
    @retval nn::pia::ResultCompanionStationIsOffline  A companion client was not logged in to the game server. Perform cleanup.
    @retval nn::pia::ResultCompanionStationIsLeft  Lost the connection with the companion station. Perform cleanup.
    @retval nn::pia::ResultCompanionStationDenied  The companion station was not in a joint disband-ready state and denied the connection. Perform cleanup.
    @retval nn::pia::ResultSessionMigrationFailed  Failed to migrate the session. Perform cleanup.
    @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
    @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

    @see  DestroyJointSessionAsync, IsCompletedDestroyJointSession
*/
    nn::Result GetDestroyJointSessionResult() const;

    // Starts accepting participants in the joint session.
/*!
    @brief  Starts accepting participants in the joint session.
    @details  This function cannot be used with local communications.
              This function is only available to the joint session host.
              This function starts an asynchronous process that accepts joint session participants.
              The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
              Use <tt>@ref IsCompletedOpenJointSessionParticipation</tt> to check whether the process has completed and <tt>@ref GetOpenJointSessionParticipationResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedOpenJointSessionParticipation, GetOpenJointSessionParticipationResult, IsJointSessionHost
*/
    nn::Result OpenJointSessionParticipationAsync();

/*!
    @brief  Checks for completion of the <tt>OpenJointSessionParticipationAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref OpenJointSessionParticipationAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  OpenJointSessionParticipationAsync, GetOpenJointSessionParticipationResult
*/
    bool IsCompletedOpenJointSessionParticipation() const;

/*!
        @brief  Gets the result of the <tt>OpenJointSessionParticipationAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref OpenJointSessionParticipationAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  OpenJointSessionParticipationAsync, IsCompletedOpenJointSessionParticipation
*/
    nn::Result GetOpenJointSessionParticipationResult() const;

/*!
    @brief  Starts accepting participants in the joint session.
    @details

           This function starts an asynchronous process that accepts joint session participants.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedOpenJointSession</tt> to check whether the process has completed and <tt>@ref GetOpenJointSessionResult</tt> to check the result.
           All stations participating in the joint session must call this function.
           The asynchronous process ends after it can be verified that all stations that have finished joining the session have called this function.
           After asynchronous processes complete, the matchmaking session is in an open state.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>CloseJointSessionAsync</tt>. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.

    @see  IsCompletedOpenParticipation, GetOpenJointSessionResult, IsHost
*/
    nn::Result OpenJointSessionAsync();

/*!
    @brief  Checks the completion of <tt>OpenJointSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref OpenJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  OpenJointSessionAsync, GetOpenJointSessionResult
*/
    bool IsCompletedOpenJointSession() const;

/*!
        @brief  Gets the result of <tt>OpenJointSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref OpenJointSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. This result also can be returned when the host has disbanded a joint session and passively begun an asynchronous process. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>CloseJointSessionAsync</tt>. <tt>OpenJointSessionAsync</tt> was not called by all stations participating in the session within 60 seconds after the function was called. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.
        @retval nn::pia::ResultSessionInconsistentState  Indicates that the process took longer than expected to complete. If the station is a client, make it check whether the other clients have called <tt>OpenJointSessionAsync</tt>. Perform cleanup after calling <tt>LeaveSessionAsync</tt>.

        @see  OpenJointSessionAsync, IsCompletedOpenJointSession
*/
    nn::Result GetOpenJointSessionResult() const;

    // Closing a joint session to new participants
/*!
    @brief  Closes a joint session to new participants.
    @details  This function cannot be used with local communications.
              This function is only available to the joint session host.
              This function starts an asynchronous process that closes a joint session to new participants.
              The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
              Use <tt>@ref IsCompletedCloseJointSessionParticipation</tt> to check whether the process has completed and <tt>@ref GetCloseJointSessionParticipationResult</tt> to check the result.

              It is possible to imagine a case where a station attempts to join the joint session immediately before this function is called and ends up joining after the asynchronous process has completed.
              For this reason, the asynchronous process ends after waiting for the session to close and the number of participants in the matchmaking session to match the number of participants in the mesh.
              The asynchronous process also ends if the numbers do not match after a certain amount of time (15 seconds). In this case, <tt>@ref GetCloseJointSessionParticipationResult</tt>
              returns <tt>ResultNegligibleFault</tt>. There could be attempts to join after the asynchronous process has ended.
              Handle these attempts as appropriate in the application.


    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedCloseJointSessionParticipation, GetCloseJointSessionParticipationResult, IsJointSessionHost
*/
    nn::Result CloseJointSessionParticipationAsync();

/*!
    @brief  Checks for completion of the <tt>CloseJointSessionParticipationAsync</tt> asynchronous process.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref CloseJointSessionParticipationAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CloseJointSessionParticipationAsync, GetCloseJointSessionParticipationResult
*/
    bool IsCompletedCloseJointSessionParticipation() const;

/*!
        @brief  Gets the result of the <tt>CloseJointSessionParticipationAsync</tt> asynchronous process.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref CloseJointSessionParticipationAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultNegligibleFault  A certain amount of time passed without the matchmaking sessions and number of participants matching. Handle appropriately in the application.

        @see  CloseJointSessionParticipationAsync, IsCompletedCloseJointSessionParticipation
*/
    nn::Result GetCloseJointSessionParticipationResult() const;

/*!
    @brief  Closes a joint session to new participants.
    @details

           This function starts an asynchronous process that closes a joint session to new participants.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedCloseJointSession</tt> to check whether the process has completed and <tt>@ref GetCloseJointSessionResult</tt> to check the result.
           All stations participating in the joint session must call this function.

           After it is confirmed that the numbers of matchmaking session and mesh participants match and that the stations that have completed connection have called this function, asynchronous processes finish.
           After it is complete, the matchmaking session enters a closed state. Advance processing under the assumption that after asynchronous processing is complete, the number of new participants will not increase.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that communication is not already taking place. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
    @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>OpenJointSessionAsync</tt>. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.

    @see  IsCompletedCloseJointSession, GetCloseJointSessionResult, IsHost
*/
    nn::Result CloseJointSessionAsync();

/*!
    @brief  Checks the completion of <tt>CloseJointSessionAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref CloseJointSessionAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  CloseJointSessionAsync, GetCloseJointSessionResult
*/
    bool IsCompletedCloseJointSession() const;

/*!
        @brief  Gets the result of <tt>CloseJointSessionAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref CloseJointSessionAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. This result also can be returned when the host has disbanded a joint session and passively begun an asynchronous process. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultSessionInvalidState  Indicates that the host has called <tt>OpenJointSessionAsync</tt>. <tt>CloseJointSessionAsync</tt> was not called by all stations participating in the session within 60 seconds after the function was called. Programming error. Fix your program so that this error is not returned. For error handling in the release version, perform cleanup after calling <tt>LeaveSessionAsync</tt>.
        @retval nn::pia::ResultSessionInconsistentState  Indicates that the process took longer than expected to complete. If the station is a client, have it check whether the other clients have called <tt>CloseJointSessionAsync</tt>. Perform cleanup after calling <tt>LeaveSessionAsync</tt>.

        @see  CloseJointSessionAsync, IsCompletedCloseJointSession
*/
    nn::Result GetCloseJointSessionResult() const;

/*!
    @brief  Updates the joint session search conditions.
    @details  This function cannot be used with local communications.

           Updates the joint session search conditions.
           Only the joint session host can execute this function.
           The <tt><var>attributeIndex</var></tt> parameter can be set to a value between <tt>0</tt> and <tt>@ref nn::pia::inet::ATTRIBUTE_MAX_SIZE</tt> - <tt>1</tt>.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedModifyJointSessionAttribute</tt> to check whether the process has completed and <tt>@ref GetModifyJointSessionAttributeResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  The value specified in <tt><var>attributeIndex</var></tt> is outside the allowed range. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @see  IsCompletedModifyJointSessionAttribute, GetModifyJointSessionAttributeResult, IsJointSessionHost
*/
    nn::Result ModifyJointSessionAttributeAsync(u32 attributeIndex, u32 newValue);
/*!
    @brief  Checks the completion of <tt>ModifyJointSessionAttributeAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref ModifyJointSessionAttributeAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  ModifyJointSessionAttributeAsync, GetModifyJointSessionAttributeResult
*/
    bool IsCompletedModifyJointSessionAttribute() const;
/*!
        @brief  Gets the result of <tt>ModifyJointSessionAttributeAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref ModifyJointSessionAttributeAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  ModifyJointSessionAttributeAsync, IsCompletedModifyJointSessionAttribute
*/
    nn::Result GetModifyJointSessionAttributeResult() const;

/*!
    @brief  Updates the joint session settings.
    @details  This function cannot be used with local communications.
           Updates the joint session settings asynchronously.
           Only the joint session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedUpdateJointSessionSetting</tt> to check whether the process has completed and <tt>@ref GetUpdateJointSessionSettingResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. This function is not supported in local communication. Programming error. Fix your program so that this error is not returned.
                                       For Internet communication, make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the joint session host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @param[in] pUpdateSessionSetting  Pointer to the <tt>UpdateSessionSetting</tt> derived class for gathering the settings to change.

    @see  UpdateSessionSetting, IsCompletedUpdateJointSessionSetting, GetUpdateJointSessionSettingResult, IsJointSessionHost
*/
    nn::Result UpdateJointSessionSettingAsync(UpdateSessionSetting* pUpdateSessionSetting);

/*!
    @brief  Checks the completion of <tt>UpdateJointSessionSettingAsync</tt> asynchronous processing.
    @details  This function cannot be used with local communications.
             This function checks the completion of asynchronous processes started with <tt>@ref UpdateJointSessionSettingAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  UpdateJointSessionSettingAsync, GetUpdateJointSessionSettingResult
*/
    bool IsCompletedUpdateJointSessionSetting() const;
/*!
        @brief  Gets the result of <tt>UpdateJointSessionSettingAsync</tt> asynchronous processing.
        @details  This function cannot be used with local communications.
                 This function gets the results after an asynchronous process started by <tt>@ref UpdateJointSessionSettingAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.

        @see  UpdateJointSessionSettingAsync, IsCompletedUpdateJointSessionSetting
*/
    nn::Result GetUpdateJointSessionSettingResult() const;

/*!
    @brief  Updates joint session application-defined data.
    @details

           Asynchronously updates joint session application-defined data.
           Only the joint session host can execute this function.
           The started process advances by periodically calling the <tt>@ref nn::pia::common::Scheduler::Dispatch</tt> function.
           Use <tt>@ref IsCompletedUpdateJointSessionApplicationData</tt> to check whether the process has completed and <tt>@ref GetUpdateJointSessionApplicationDataResult</tt> to check the result.

    @return  Returns a <tt>Result</tt> value that indicates success if the asynchronous process starts successfully. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

    @retval nn::pia::ResultInvalidArgument  Either the specified <tt><var>pApplicationBuffer</var></tt> is <tt>Null</tt> or the specified <tt><var>dataSize</var></tt> has exceeded the value that can be set. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that no other asynchronous processes are running, that <tt>Startup</tt> has been called, and that the local station is the joint session host. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.

    @param[in] pApplicationBuffer  The application-defined data to update.
    @param[in] dataSize  The size of the specified application-defined data.

    @see  IsCompletedUpdateJointSessionApplicationData, GetUpdateJointSessionApplicationDataResult, IsJointSessionHost
*/
    nn::Result UpdateJointSessionApplicationDataAsync(const void* pApplicationBuffer, size_t dataSize);

/*!
    @brief  Checks the completion of <tt>UpdateJointSessionApplicationDataAsync</tt> asynchronous processing.
    @details  This function checks the completion of asynchronous processes started with <tt>@ref UpdateJointSessionApplicationDataAsync</tt>.

    @return  Returns <tt>true</tt> if the asynchronous process has completed, or <tt>false</tt> otherwise. Also returns <tt>false</tt> when there has been no asynchronous processing.

    @see  UpdateJointSessionApplicationDataAsync, GetUpdateJointSessionApplicationDataResult
*/
    bool IsCompletedUpdateJointSessionApplicationData() const;

/*!
        @brief  Gets the result of <tt>UpdateJointSessionApplicationDataAsync</tt> asynchronous processing.
        @details  This function gets the results after an asynchronous process started by <tt>@ref UpdateJointSessionApplicationDataAsync</tt> ends.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt>. If this function call fails, one or more of the following <tt>Result</tt> values is returned.

        @retval nn::pia::ResultInvalidState  Indicates that the instance was not in a valid state when the function was called. Make sure that that no other asynchronous processes are running. Programming error. Fix your program so that this error is not returned.
        @retval nn::pia::ResultNetworkConnectionIsLost  The network is not available. The wireless switch may be turned off, there may be a problem with the access point, or the connection to the game server may have been lost. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Reconnect to AC.
        @retval nn::pia::ResultSessionConnectionIsLost  Disconnected from the mesh. Call <tt>LeaveSessionAsync</tt> and then call <tt>Cleanup</tt>. Perform cleanup.
        @retval nn::pia::ResultSessionIsNotFound  The target session was lost during the asynchronous process. Perform cleanup.
        @retval nn::pia::ResultNexInternalError  This is an internal NEX error. The <tt>nn::pia::util::NetworkErrorCodeConverter::Convert</tt> function gets a NEX network error code. Check the NEX section of the network error code list. After logging out, reinitialize Pia and NEX. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Handle appropriately in the application.
        @retval nn::pia::ResultGameServerProcessAborted  A game server-related process failed. Execute up to the logout process from the NEX server.
        @retval nn::pia::ResultGameServerMaintenance  Indicates that the server is under maintenance. If this error occurs during Internet communication, convert to a network error code and display in the error viewer applet/EULA applet. However, during local communication, do not display it as an error with error code in the error viewer applet/EULA applet. Execute up to the logout process from the NEX server.

        @see  UpdateJointSessionApplicationDataAsync, IsCompletedUpdateJointSessionApplicationData
*/
    nn::Result GetUpdateJointSessionApplicationDataResult() const;

/*!
    @brief  Gets the ID of the joint session.

    @if CTR_DOC
    @details  This function always returns <tt>INVALID_SESSION_ID</tt> during local communication.
    @endif

    @return  Returns the session ID of the currently joined joint session. Returns <tt>INVALID_SESSION_ID</tt> if not a participant in a joint session.
*/
    u32 GetJointSessionId() const;

/*!
        @brief  Gets whether the local station is the joint session host.

        @details  This function returns whether the local station is the host of the joint session.

        @return  Returns <tt>true</tt> if the local station is the host, or <tt>false</tt> otherwise.
*/
    bool IsJointSessionHost() const;

/*!
        @cond  PRIVATE
        @brief  Gets the <tt>StationIndex</tt> of the joint session host.

        @details  This function returns <tt>STATION_INDEX_INVALID</tt> if the local station is not a participant in a joint session.
                 When participating in a joint session, this function returns a <tt>StationIndex</tt> that specifies the joint session host.

    @return  Returns the <tt>StationIndex</tt> that represents the host of the joint session where the local station is a participant.
*/
    StationIndex GetJointSessionHostStationIndex() const;
    //! @endcond

/*!
    @brief  Gets the <tt>StationId</tt> that represents the joint session host.

    @details  This function returns <tt>STATION_ID_INVALID</tt> if the local station is not a participant in a joint session.
             When participating in a joint session, this function returns a <tt>StationId</tt> that specifies the joint session host.

    @return  Returns the <tt>StationId</tt> that represents the joint session host.
*/
    StationId GetJointSessionHostStationId() const;

/*!
    @brief  Gets the <tt>StationConnectionInfo</tt> that represents the host of the joint session where the local station is a participant.

    @details  When participating in a joint session, this function returns a <tt>StationConnectionInfo</tt> that specifies the joint session host.

    @param[out] pInfoOut  A pointer to the object that holds the address representing the joint session host.
    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if execution succeeds. You must make sure that the implementation of this function in your application does not return any errors.
    @retval ResultInvalidArgument  Indicates that one or more arguments are invalid. Programming error. Fix your program so that this error is not returned.
    @retval ResultInvalidState  Not participating in a joint session. Programming error. Fix your program so that this error is not returned.

    @return  Returns the <tt>StationConnectionInfo</tt> that represents the host of the joint session where the local station is a participant.
*/
    nn::Result GetJointSessionHostStationConnectionInfo(transport::StationConnectionInfo* pInfoOut);

/*!
        @cond  PRIVATE
*/
    int GetJointSequence();
    //! @endcond

/*!
        @cond  PRIVATE
*/
    int GetJoinSequence();
    int GetJoinAutoMatchmakeSequence();
    int GetJoinMeshSequence();
    u8 GetCloseSessionSequence();
    u8 GetCloseJointSessionSequence();
    u8 GetOpenSessionSequence();
    u8 GetOpenJointSessionSequence();
    void CleanupJobsWhenJointSession();
    //! @endcond

/*!
     end of name Joint Session Feature
     @}
*/


/*!
     @name  Registering and Unregistering Callback Functions
     @{
*/

/*!
        @brief  Registers the callback function that is called when the session state changes.

        @details  This function registers the callback that is called when the session state changes (when a station joins or exits).
        Do not include any processes that block for long periods of time in the registering function.

        @param[in] callback  Specifies a pointer to the callback function being registered.

        @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the process succeeds. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
        @retval nn::pia::ResultInvalidState  Returned when an event notification callback is already registered. Programming error. Fix your program so that this error is not returned.

        @see  UnregisterSessionEventCallback
*/
    nn::Result RegisterSessionEventCallback(SessionEventCallback callback);

/*!
        @brief  Unregisters the callback function registered for changes to the session state.

        @see  RegisterSessionEventCallback
*/
    void UnregisterSessionEventCallback();


    /// @cond PRIVATE
    static void NoticeEventCallback(session::Session::SessionEventType eventType, StationIndex index);

    void CheckStatusAtChangeMeshState();
    /// @endcond

/*!
        @brief  Registers the callback function for determining whether to approve or deny a join request when received by the session host.

        @details  This function defines the callback that is invoked when the local station is the session host and it receives a join request from a client.
                 The callback function is called by <tt>@ref nn::pia::common::Scheduler::Dispatch</tt>.
                 Do not include any processes that block for long periods of time in the registering function.

        @param[in] callback  Specifies a pointer to the callback function being registered.

        @see  UnregisterJoiningApprovalCallback
*/
    void RegisterJoiningApprovalCallback(JoiningApprovalCallback callback);

/*!
        @brief  Unregisters the callback function that is called to determine whether to approve or deny a join request.

        @see  RegisterJoiningApprovalCallback
*/
    void UnregisterJoiningApprovalCallback();

    //! @cond PRIVATE
    static bool SessionJoiningApprovalCallback(const nn::pia::transport::Station::IdentificationToken* pRequesterToken);
    MeshEventListener* GetMeshEventListenerPtr();
    transport::Transport::DisconnectionEventListener GetDisconnectionEventListener();
    static u8 CheckHostCandidateRankingCallback(StationIndex targetStationIndex, bool isHost);
    static u32 GetSessionIdCallback();
    //! @endcond

/*!
     end of name Registering and Unregistering Callback Functions
     @}
*/

    /// @cond PRIVATE
    // private
    void JoinNotice(StationId targetStationId);
    /// @endcond

/*!
     @name  Synchronous Clock
     @{
*/

/*!
    @brief  Specifies the synchronous clock (the time, in milliseconds, since the session was created).

    @details  The synchronous clock is synchronized among all the stations in the session.
                Clients joined in the session periodically send a clock-synchronization pulse to the session host.
                The host then responds to these pulses, ensuring synchronization between the host and clients.
                By default, synchronization pulses are sent every two seconds,
                but the application can also change this interval. For more information, see <tt>SetSyncClockPulseInterval()</tt> and
                <tt>GetSyncClockPulseInterval()</tt>.

    @return  Returns the elapsed time, in milliseconds, since the session was created.
                This function returns a negative value if the station is not
                synchronized with the session host, such as when it has just joined a session
                and not yet received a response from the host.

    @see  SetSyncClockPulseInterval, GetSyncClockPulseInterval
*/
    s64 GetTime(void) const;


/*!
    @brief  Specifies the send interval of the synchronization pulse for the synchronous clock.

    @param[in] interval  Specifies the interval at which the synchronization pulse is sent. The value is in milliseconds.
    @see  GetTime, GetSyncClockPulseInterval
*/
    void SetSyncClockPulseInterval(s32 interval);


/*!
    @brief  Gets the send interval of the synchronization pulse for the synchronous clock.

    @return  Returns the interval at which synchronization pulses are sent. The value is in milliseconds.
    @see  GetTime, SetSyncClockPulseInterval
*/
    s32 GetSyncClockPulseInterval(void) const;

/*!
     end of name Synchronous Clock
     @}
*/

/*!
     @name  Relay Feature
     @{
*/

/*!
    @brief  Sets the maximum number of forwarding requests that can be assigned to a single station when relaying is enabled.

    @details  The library uses this value as a metric when searching for and selecting relay paths.
                If you do not set a value, the library uses [<tt><var>MaximumSessionParticipants</var></tt>] × [<tt><var>MaximumSessionParticipants</var></tt>] as the default.

                We recommend that you set this to the same value for all participants in a session. (Operation is not guaranteed if different values are set on different clients.)

                This function is only available when the network topology allows relaying, and <tt>@ref Startup</tt> has not yet been called.

    @param[in] num  Specifies the maximum value.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval nn::pia::ResultInvalidState  Indicates that relay connections are disabled or the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetMaxNumOfRelayRoutesAssignedToStation(u16 num);


/*!
    @brief  Sets the maximum total RTT for a relay route when relaying is enabled.

    @details  The library uses this value as a metric when searching for and selecting relay paths.
                If you do not set a value, the library uses the default value of <tt>nn::pia::session::NN_PIA_SESSION_MAX_RTT_DEFAULT</tt>.

                We recommend that you set this to the same value for all participants in a session. (Operation is not guaranteed if different values are set on different clients.)

                This function is only available when the network topology allows relaying, and <tt>@ref Startup</tt> has not yet been called.

    @param[in] rtt  Specifies the maximum value. The value is in milliseconds.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval nn::pia::ResultInvalidState  Indicates that relay connections are disabled or the <tt>Startup</tt> function has already been called. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result SetMaxRtt(u16 rtt);

/*!
    @brief  Gets the connection path between the local station and an arbitrary station index when relaying is enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.
                The connection path between the local station and <tt><var>targetIndex</var></tt> is stored in <tt><var>pRelayIndex</var></tt>, according to the following rules.
                (The value of <tt><var>pRelayIndex</var></tt> does not change if the function could not retrieve the connection path.)
                <ul>
                <li>Direct connection: <tt><var>targetIndex</var></tt></li>
                <li>Relay connection: <tt>StationIndex</tt> of the relay requester.</li>
                <li>Local station: <tt>STATION_INDEX_INVALID</tt></li>
                <li>Not connected: <tt>STATION_INDEX_INVALID</tt></li>
                </ul>

    @details  Even if you can get a valid value for the requester's <tt>StationIndex</tt>, it might not be possible to convert the value to a <tt>StationId</tt>.
                When this occurs, <tt>@ref transport::Transport::ConvertToStationId</tt> returns <tt>ResultNotFound</tt>.
                This state is only temporary and you can convert to a <tt>StationId</tt> after the internal process advances normally.

    @param[in] targetIndex  Specifies the <tt>StationIndex</tt> with the connection path to retrieve.
    @param[out] pRelayIndex  Specifies the <tt>StationIndex</tt> indicating the connection path.

    @return  Returns a <tt>Result</tt> value indicating success if the value can be retrieved successfully.
    @retval nn::pia::ResultInvalidState  Indicates that relay connections are disabled. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  The specified <tt>StationIndex</tt> is invalid, or the pointer specified in the arguments is invalid. Programming error. Fix your program so that this error is not returned.
*/
    nn::Result GetConnectionRoute(StationIndex targetIndex, StationIndex* pRelayIndex);

/*!
    @brief  Specifies connection path emulation settings between any two <tt>StationIndex</tt> objects when relaying is enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.

    @details  When you call this function, make sure that you have enabled the connection path emulation feature by using the <tt>@ref Session::SetDebugSetting</tt> function.

    @details  If <tt>EMULATION_TYPE_RELAY</tt> is set, the connection method for the target <tt>StationIndex</tt> is always a relay connection
                (even when, for example, a direct connection is possible).

    @details  Call this function when the terminal is not connected to a session.
                If you call this function when the terminal has already joined a session, there is no guarantee that changing the connection path emulation settings will work.

    @param[in] index1  Specifies the first <tt>StationIndex</tt>.
    @param[in] index2  Specifies the first <tt>StationIndex</tt>.
    @param[in] type  Specifies the connection method emulation setting.

    @return  Returns a <tt>Result</tt> value indicating success if the setting succeeds.
    @retval nn::pia::ResultInvalidState  Indicates that relay connections are disabled. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that the <tt>StationIndex</tt> is invalid, or the specified <tt>EmulationType</tt> is invalid. Programming error. Fix your program so that this error is not returned.

    @see  SetDebugSetting, ResetRelayEmulation
*/
    nn::Result RelayEmulation(StationIndex index1, StationIndex index2, Mesh::EmulationType type);

/*!
    @brief  Initializes connection path emulation settings when relay connections are enabled. (For debugging.)

    @details  This function is for debugging. Do not include it in retail products.

    @details  All connections are initialized to use the connection path emulation configured by the <tt>@ref RelayEmulation</tt> function.
                (This is equivalent to setting <tt>EMULATION_TYPE_NONE</tt> for all combinations of <tt>StationIndex</tt>.)

    @details  Call this function when the terminal is not connected to a session.
                If you call this function when the terminal has already joined a session, there is no guarantee that changing the connection path emulation settings will work.

    @see  SetDebugSetting, RelayEmulation
*/
    void ResetRelayEmulation();

/*!
     end of name Relay Feature
     @}
*/

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    nn::Result GetRefereeStatsAsync(u32 initialRatingValue, u32 category);
    bool IsCompletedGetRefereeStats();
    nn::Result GetGetRefereeStatsResult();

    nn::Result StartGameAsync(u32 category);
    bool IsCompletedStartGame();
    nn::Result GetStartGameResult();

    nn::Result EndGameAsync(session::IEndGamePersonalReport* pReportArray, size_t reportArraySize);
    bool IsCompletedEndGame();
    nn::Result GetEndGameResult();

    nn::Result EndGameWithoutReportAsync();
    bool IsCompletedEndWithoutReportGame();
    nn::Result GetEndGameWithoutReportResult();

    // Returns <tt>true</tt> from when <tt>StartGameAsync</tt> is called to when <tt>EndGameAsync</tt> or <tt>EndGameWithoutReportAsync</tt> is called.
    bool IsInGame();
#endif

/*!
     @name  Other
     @{
*/

/*!
        @brief  Gets whether host migration is enabled.

        @details  The return value is determined based on the argument value specified in <tt>nn::pia::session::Session::Startup</tt>.
        @return  Returns <tt>true</tt> if <tt>@ref nn::pia::session::Session::Startup</tt> was called with host migration enabled.

*/
    bool IsUsingHostMigration() const
    {
        return m_IsUsingHostMigration;
    };

    /// @cond PRIVATE
/*!
    @brief  Kicks a station from the session. (Host only.)

    @details  This function can only be used by the host.

    @details  A kick notice is conveyed to the station assigned the <tt>StationIndex</tt> specified
                in the argument, and that station leaves the mesh upon receiving the notice.

    @details  Because of latency in transmitting the disconnect notification and other factors, there may be a delay between successfully calling this function, and the station leaving the mesh.

    @param[in] index  Specifies the <tt>StationIndex</tt> assigned to the station being kicked.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the station is successfully kicked from the session. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Indicates that a node other than the host called the function. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that the <tt>StationIndex</tt> specified in the arguments belongs to the mesh host, or that no station in the mesh has that ID. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInProgress  Indicates that the process to kick the station having the <tt>StationIndex</tt> specified in the arguments has already started. Handle appropriately in the application.
    @retval nn::pia::ResultBufferIsFull  Indicates that the buffer for kicking stations or the buffer for sending kick notifications is full. It may succeed if you try again after some time has passed. Handle appropriately in the application.
*/
    nn::Result KickoutStation(nn::pia::StationIndex index);
    /// @endcond

/*!
    @brief  Kicks a station from the mesh. (Host only.)

    @details  This function can only be used by the host.

    @details  A kick notice is conveyed to the station assigned the <tt>StationId</tt> specified in
                the argument, and that station leaves the mesh upon receiving the notice.

    @details  Because of latency in transmitting the disconnect notification and other factors, there may be a delay between successfully calling this function, and the station leaving the mesh.

    @param[in] id  Specifies the station ID assigned to the station being kicked.

    @return  Returns a <tt>Result</tt> value for which the <tt>IsSuccess()</tt> function returns <tt>true</tt> if the station is successfully kicked from the mesh. If this function call fails, one or more of the following <tt>Result</tt> values is returned.
    @retval nn::pia::ResultInvalidState  Indicates that a node other than the host called the function. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInvalidArgument  Indicates that the <tt>StationId</tt> specified in the arguments belongs to the mesh host, or that no station in the mesh has that ID. Programming error. Fix your program so that this error is not returned.
    @retval nn::pia::ResultInProgress  Indicates that the process to kick the station having the <tt>StationId</tt> specified in the arguments has already started. Handle appropriately in the application.
    @retval nn::pia::ResultBufferIsFull  Indicates that the buffer for kicking stations or the buffer for sending kick notifications is full. It may succeed if you try again after some time has passed. Handle appropriately in the application.
*/
    nn::Result KickoutStation(nn::pia::StationId id);

/*!
    @brief  Gets the reason why the mesh was disconnected.

    @details  This function always returns <tt>UNKNOWN_REASON</tt> if the mesh is in a state that allows for communication.

    @return  Returns a <tt>DisconnectReason</tt> value indicating the reason why the mesh disconnected.

    @see  CheckConnectionError
*/
    DisconnectReason GetSessionDisconnectReason() const;

/*!
    @brief  Prints information that is useful for debugging.

    @param[in] flag  Specifies the bitwise OR of trace flags. For more information, see the <tt>@ref TraceFlag</tt> API reference.
*/
    virtual void Trace(u64 flag) const;

/*!
    @brief  Sets the <tt>transport::Station::IdentificationToken</tt> used to determine whether inconsistency is detected in the session participation approval callback in the matchmaking analysis feature.

    @param[in] pToken  Token used to determine whether inconsistency is detected in the session participation approval callback.
*/
    void SetMatchmakeAnalysisToken(nn::pia::transport::Station::IdentificationToken* pToken);

/*!
     end of name Others
     @}
*/


    /// @cond PRIVATE
    bool SaveWaitingStationIndex();
    u32 UpdateWaitingStationIndexNum();
    void SetWaitingOtherClientsFlag(bool flag);
    /// @endcond

    /// @cond PRIVATE
    IMatchmakeSession* GetCurrentMatchmakeSession()
    {
        return m_pMatchmakeSessionPtrArray[currentIndex()];
    }
    IMatchmakeSession* GetBufferMatchmakeSession()
    {
        return m_pMatchmakeSessionPtrArray[bufferIndex()];
    }
    const IMatchmakeSession* GetCurrentMatchmakeSession() const
    {
        return m_pMatchmakeSessionPtrArray[currentIndex()];
    }
    const IMatchmakeSession* GetBufferMatchmakeSession() const
    {
        return m_pMatchmakeSessionPtrArray[bufferIndex()];
    }
    ISessionInfoList* GetSessionInfoListPtr()
    {
        return m_pSessionInfoList;
    }
#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    IMatchmakeReferee* GetMatchmakeReferee()
    {
        return m_pMatchmakeRefereePtr;
    }
#endif

    u32 GetCurrentSessionId() const
    {
        return m_SessionId[currentIndex()];
    }
    u32 GetBufferSessionId() const
    {
        return m_SessionId[bufferIndex()];
    }
    u8 GetCurrentMaxParticipantNum() const
    {
        return m_MaxParticipantNum[currentIndex()];
    }
    u8 GetBufferMaxParticipantNum() const
    {
        return m_MaxParticipantNum[bufferIndex()];
    }

    bool IsLeaveIgnorable()
    {
        return m_LeaveIgnorableFlag;
    }
    bool IsPostponeJoinEvent()
    {
        return m_PostponeJoinEventFlag;
    }

    void IsHostMigrateOccurred(bool b);
    bool IsHostMigrateOccurred();
    nn::pia::transport::Station::IdentificationToken* GetMatchmakeAnalysisToken();
    void NotifyOwnershipChange(u32 sessionId, u32 ownerId);
    void NotifyUpdateHost(u32 sessionId, u32 hostId);
    void NotifyLeave(u32 sessionId, u32 principalId);
    void NotifyUnregister(u32 sessionId);
    void SetUsingLastGidOption(bool b);
    /// @endcond

    /// @cond PRIVATE
    bool IsTargetMatchmakeSessionOpen()const;
    bool IsOpenSession(u32 sessionId = 0)const;
    void SetOpenSession(u32 sessionId, bool flg);
    u32 GetConfigParticipationTargetSessionId()const;
    IMatchmakeSession* GetConfigParticipationTargetMatchmakeSession();
    const IMatchmakeSession* GetConfigParticipationTargetMatchmakeSession()const;
    ModuleState GetModuleState() const
    {
        return m_ModuleState;
    }
    /// @endcond

protected:
    bool IsProcessingJoin() const;
    bool IsWaitingOtherClients() const;
    bool IsPostponeNotice() const;

private:
    //! @cond PRIVATE
/*
    @brief  The constructor. This constructor is private because the singleton pattern is used.
*/
    Session(Session::Setting setting);


/*!
    @brief  The destructor. This destructor is private because the singleton pattern is used.
*/
    ~Session();

/*!
    @brief  This copy constructor is private because the singleton pattern is used.
*/
    Session(const Session&);


/*!
    @brief  This assignment operator is private because the singleton pattern is used.
*/
    Session& operator=(const Session&);
    //! @endcond

    bool UpdateStationIdToBecomeHost();
    bool UpdateStationIdToBecomeClient();

    bool AddNewStationId(StationId stationId, StationIndex stationIndex, u32 uniqueId, u32 sessionId);
    bool RemoveStationId(StationId stationId, StationIndex stationIndex, u32 uniqueId, u32 sessionId);

    bool SendMatchStationInfoToOtherClient();

    void CleanupStatus(bool isPostponeCleanupLocalStationId = false);

    void ChangeStatusToDisconnectedSession();
    void ChangeStatusToDisconnectedNetwork();

    void CancelJoinProcessByFatalError();

    // A pointer to a singleton instance.
    static Session* s_pInstance;

    static DebugSetting s_DebugSetting;

    bool m_IsUsingHostMigration;

    //
    bool m_IsNecessaryServer;

    MeshLayerController* m_pMeshLayerController;

    SessionStatusCheckJob* m_pSessionStatusCheckJob;
    CreateSessionJob* m_pCreateSessionJob;
    AutoMatchmakeJob* m_pAutoMatchmakeJob;
    BrowseMatchmakeJob* m_pBrowseMatchmakeJob;
    JoinSessionJob* m_pJoinSessionJob;
    LeaveSessionJob* m_pLeaveSessionJob;
    DestroySessionJob* m_pDestroySessionJob;
    OpenParticipationJob* m_pOpenParticipationJob;
    //OpenSessionJob* m_pOpenSessionJob;
    CloseParticipationJob* m_pCloseParticipationJob;
    //CloseSessionJob* m_pCloseSessionJob;
    ConfigParticipationJobBase* m_pConfigParticipationJob;

    GenerateMatchmakeSystemPasswordJob* m_pGenerateMatchmakeSystemPasswordJob;
    ClearMatchmakeSystemPasswordJob* m_pClearMatchmakeSystemPasswordJob;

    u32 m_SessionPasswordSize;
    wchar_t* m_pSessionPassword;

    JointSessionJob* m_pJointSessionJob;

    ModifyAttributeJob* m_pModifyAttributeJob;
    UpdateSessionSettingJob* m_pUpdateSessionSettingJob;
    UpdateApplicationDataJob* m_pUpdateApplicationDataJob;

#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    RefereeStartGameJob* m_pRefereeStartGameJob;
    RefereeEndGameJob* m_pRefereeEndGameJob;
    RefereeEndGameWithoutReportJob* m_pRefereeEndGameWithoutReportJob;
    RefereeGetOrCreateStatsJob* m_pRefereeGetOrCreateStatsJob;
#endif

    SessionSearchCriteria* m_pPrimarySessionSearchCriteria;
    SessionSearchCriteria* m_pSecondarySessionSearchCriteria;

    SessionEventCallback m_SessionEventCallback;
    JoiningApprovalCallback m_JoiningApprovalCallback;

    StationIdStatusTable* m_pStationIdStatusTable;

    u32 m_SessionProtocolHandle;
    SessionProtocol* m_pSessionProtocol;

    ModuleState m_ModuleState;
    MatchmakeStatus m_MatchmakeStatus;

    ConnectionStatus m_ConnectionStatus;

    StationId m_LocalStationId;
    StationId m_HostStationId;
    StationId m_JointSessionHostStationId;

    u32 m_PrincipalId;

    enum AsyncProcessingName
    {
        ASYNC_NOTHING = 0,
        ASYNC_AUTO_MATCHMAKE,
        ASYNC_BROWSE_SESSION,
        ASYNC_CREATE_SESSION,
        ASYNC_JOIN_SESSION,
        ASYNC_LEAVE_SESSION,
        ASYNC_OPEN_PARTICIPATION,
        ASYNC_OPEN_SESSION,
        ASYNC_CLOSE_PARTICIPATION,
        ASYNC_CLOSE_SESSION,
        ASYNC_GENERATE_MATCHMAKE_SYSTEM_PASSWORD,
        ASYNC_CLEAR_MATCHMAKE_SYSTEM_PASSWORD,
        ASYNC_AUTO_MATCHMAKE_GAME,
        ASYNC_BROWSE_GAME_SESSION,
        ASYNC_CREATE_GAME_SESSION,
        ASYNC_JOIN_GAME_SESSION,
        ASYNC_LEAVE_GAME_SESSION,
        ASYNC_OPEN_GAME_SESSION_PARTICIPATION,
        ASYNC_CLOSE_GAME_SESSION_PARTICIPATION,
        ASYNC_MERGE_SESSION,
        ASYNC_MERGE_RANDOM_SESSION,
        ASYNC_DEMERGE_SESSION,
        ASYNC_CREATE_JOINT_SESSION,
        ASYNC_JOIN_JOINT_SESSION,
        ASYNC_JOIN_RANDOM_JOINT_SESSION,
        ASYNC_DESTROY_JOINT_SESSION,
        ASYNC_LEAVE_JOINT_SESSION,
        ASYNC_OPEN_JOINT_SESSION_PARTICIPATION,
        ASYNC_CLOSE_JOINT_SESSION_PARTICIPATION,
        ASYNC_OPEN_JOINT_SESSION,
        ASYNC_CLOSE_JOINT_SESSION,
        ASYNC_UPDATE_SESSION_SETTING,
        ASYNC_UPDATE_JOINT_SESSION_SETTING,
        ASYNC_UPDATE_APPLICATION_DATA,
        ASYNC_UPDATE_JOINT_SESSION_APPLICATION_DATA,
#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
        ASYNC_START_GAME,
        ASYNC_END_GAME,
        ASYNC_END_GAME_WITHOUT_REPORT,
        ASYNC_GET_OR_CREATE_REFEREE_STATS,
#endif
        ASYNC_MODIFY_MATCHMAKE_SESSION_ATTRIBUTE,
        ASYNC_MODIFY_JOINT_SESSION_ATTRIBUTE
    };

    void ResetInternalAsyncState();
    bool IsCompletedAsyncProcess(AsyncProcessingName processingName) const;
    nn::Result GetAsyncProcessResult(AsyncProcessingName processingName) const;
    nn::Result CancelAsyncProcess(AsyncProcessingName processingName);

    nn::Result OpenParticipationCore(u32 sessionId, IMatchmakeSession* pMatchmakeSession, AsyncProcessingName asyncProcessName);
    nn::Result OpenSessionCore(u32 sessionId, IMatchmakeSession* pMatchmakeSession, AsyncProcessingName asyncProcessName);
    nn::Result CloseParticipationCore(u32 sessionId, IMatchmakeSession* pMatchmakeSession, AsyncProcessingName asyncProcessName);
    nn::Result CloseSessionCore(u32 sessionId, IMatchmakeSession* pMatchmakeSession, AsyncProcessingName asyncProcessName);

    nn::Result ModifyAttributeAsyncCore(u32 attributeIndex, u32 newValue, u32 targetSessionId, nn::pia::session::IMatchmakeSession* ptargetMatchmakeSession);
    nn::Result UpdateSessionSettingAsyncCore(UpdateSessionSetting* pUpdateSessionSetting, u32 targetSessionId);
    nn::Result UpdateApplicationDataAsyncCore(const void* pApplicationBuffer, size_t dataSize, u32 targetSessionId);

    common::CallContext m_InternalContext;
    AsyncProcessingName m_CurrentAsyncProcessingName;

    MeshEventListenerForSession* m_pMeshEventListenerForSession;

    void CallSessionEventCallback(SessionEventType eventType, StationId stationId);

    bit32 m_WaitingStationIndexBitmap;
    bool m_WaitingOtherClientsStationIdFlag;
    bool LeaveNoticeToWaitingStationIndex(StationIndex leaveStationIndex);
    void ClearWaitingStationIndex();

    bool m_LeaveIgnorableFlag;
    bool m_PostponeJoinEventFlag;
    void SetLeaveIgnorableFlag(bool isLeaveIgnorable)
    {
        m_LeaveIgnorableFlag = isLeaveIgnorable;
    }
    void SetPostponeJoinEvent(bool isPostponeJoinEvent)
    {
        m_PostponeJoinEventFlag = isPostponeJoinEvent;
    }

    u8 currentIndex() const
    {
        return m_CurrentMatchmakeSessionIndex;
    }
    u8 bufferIndex() const
    {
        return (m_CurrentMatchmakeSessionIndex == 0 ? 1 : 0);
    }
    void SetCurrentSessionId(u32 sessionId)
    {
        m_SessionId[currentIndex()] = sessionId;
    }
    void SetBufferSessionId(u32 sessionId)
    {
        m_SessionId[bufferIndex()] = sessionId;
    }
    void SetCurrentMaxParticipantNum(u8 num)
    {
        m_MaxParticipantNum[currentIndex()] = num;
    }
    void SetBufferMaxParticipantNum(u8 num)
    {
        m_MaxParticipantNum[bufferIndex()] = num;
    }
    void SwapMatchmakeSessionBuffer()
    {
        m_CurrentMatchmakeSessionIndex = bufferIndex();
    }

    u8 m_MaxParticipantNum[2];
    u8 m_CurrentMatchmakeSessionIndex;
    IMatchmakeSession* m_pMatchmakeSessionPtrArray[2];
    ISessionInfoList* m_pSessionInfoList;
    u32 m_SessionId[2];
#if NN_PIA_ENABLE_MATCHMAKE_REFEREE
    IMatchmakeReferee* m_pMatchmakeRefereePtr;
#endif
    void ChangeStationIdConvertEnableFlag(bool flag);

    TStationIdList m_ValidStationIdList;
    void* m_pValidStationIdListBuffer;
    void validateStationId(const StationId& targetStationId);
    void invalidateStationId(const StationId& targetStationId);

    u32 m_SessionIdOfJointSessionHost;

    // Flag for the matchmaking analysis feature.
    // Flag that detects events indicating that host migration has occurred.
    bool m_IsHostMigrateOccurred;
    nn::pia::transport::Station::IdentificationToken m_MatchmakeAnalysisToken;

    static const int CLOSE_SESSION_LIST_NUM = 4;
    u32 m_CloseSessionId[CLOSE_SESSION_LIST_NUM];

    bool m_IsUsingLastGidOption;

};
}
}
}  // end of namespace nn::pia::session
